#!/bin/python2
# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================
import sys
import os.path

bench_dir = os.path.dirname(os.path.abspath(__file__))
top_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))

sys.path.append(os.path.abspath(top_dir))
sys.path.append(os.path.join(top_dir, '../resources/antlr/antlr-python-runtime-3.1.1/'))
sys.path.insert(0, os.path.join(top_dir, '../resources'))

import subprocess
import time
import re

from runscripts.basic import BaseScript
from utils import log, misc, procmon

import adder.gen
import sensorfilter.gen

class BenchScript(BaseScript):
    """This is a script for automated benchmarking."""

    __actions = []
    #__benchmarks = ['adder', 'sensorFilter', 'sensorFilter-ND', 'sensorFilter-Hybrid']
    __benchmarks = ['mc-bdd', 'mc-bmc', 'ftg', 'perf-fte', 'perf', 'perf-nd', 'perf-hybrid']
    __perf_tools = ['mrmc', 'imca', 'sim-fast', 'sim-slow']

    __ranges = {'mc-bdd':2,
                'mc-bmc':2,
                'ftg':2,
                'perf-fte':1,
                'perf':2,
                'perf-nd':2,
                'perf-hybrid':2}

    def gen_bench(self, bench, num):
        if bench == 'mc-bdd':
            self.gen_adder(num)
        elif bench == 'mc-bmc':
            self.gen_adder(num)
        elif bench == 'ftg':
            self.gen_adder(num)
        elif bench == 'perf-fte':
            self.gen_sf_ft(num)
        elif bench == 'perf':
            self.gen_sf(num)
        elif bench == 'perf-nd':
            self.gen_sf_nd(num)
        elif bench == 'perf-hybrid':
            self.gen_sf_hybrid(num)
        else:
            assert False, "Invalid benchmark to generate"

    def gen_adder(self, num):
        (slim_fname, props_fname, err_fname, fi_fname) = adder.gen.main(num, True)
        self.models = [slim_fname, err_fname]
        self.models = map(self.gen_file, self.models)
        self.props = self.gen_file(props_fname)
        self.fixml = self.gen_file(fi_fname)

    def gen_sf(self, num):
        (slim_fname, props_fname, fi_fname) = sensorfilter.gen.main(num, 100)
        self.models = [ self.gen_file(slim_fname) ]
        self.models += map(self.bench_file, ["sensorfilter/sensorfilterCore.slim", "sensorfilter/sensorfilterErr.slim"])
        self.props = self.gen_file(props_fname)
        self.fixml = self.gen_file(fi_fname)

    def gen_sf_nd(self, num):
        (slim_fname, props_fname, fi_fname) = sensorfilter.gen.main(num, 100)
        self.models = [ self.gen_file(slim_fname) ]
        self.models += map(self.bench_file, ["sensorfilter/sensorfilterCore-ND.slim", "sensorfilter/sensorfilterErr-ND.slim"])
        self.models = map(self.bench_file, self.models)
        self.props = self.gen_file(props_fname)
        self.fixml = self.gen_file(fi_fname)

    def gen_sf_hybrid(self, num):
        (slim_fname, props_fname, fi_fname) = sensorfilter.gen.main(num, 100)
        self.models = [ self.gen_file(slim_fname) ]
        self.models += map(self.bench_file, ["sensorfilter/sensorfilterCore-Hybrid.slim", "sensorfilter/sensorfilterErr-Hybrid.slim"])
        self.models = map(self.bench_file, self.models)
        self.props = self.gen_file(props_fname)
        self.fixml = self.gen_file(fi_fname)

    def gen_sf_ft(self, num):
        if num > 14:
            log.system.warning("Too high value for FTE, using 14")
            num = 14
        self.models = ["sensorfilter/faulttrees/sensorfilter-ft-%d.xml"%num, "sensorfilter/faulttrees/full-sensorfilter-ft.rates_map"]
        self.models = map(self.bench_file, self.models)

    def bench_file(self, rel_path):
        """Returns absolute path to a file part of benchmark"""
        return os.path.join(bench_dir, rel_path)

    def gen_file(self, rel_path):
        """Returns absolute path to a file generated by benchmark"""
        # CWD is already output dir, so abspath suffices
        return os.path.abspath(rel_path)

    def exec_bench(self, bench):
        if bench == 'mc-bdd':
            return self.mc_bdd()
        elif bench == 'mc-bmc':
            return self.mc_bmc()
        elif bench == 'ftg':
            return self.ftg()
        elif bench == 'perf-fte':
            return self.fte(self.perf_tool)
        elif bench == 'perf' or\
             bench == 'perf-nd' or\
             bench == 'perf-hybrid':
            if self.perf_tool.startswith("sim"):
                return self.sf_sim(self.perf_tool)
            else:
                return self.sf_imc(self.perf_tool)
        else:
            assert False, "Invalid benchmark to execute"

    def mc_bdd(self):
        script = os.path.join(top_dir, "runscripts", "model_checking__srs_015.py")
        args = [script]
        args += self.models
        args.append("-p%s" % self.props)
        args.append("-i%s" % self.fixml)
        return self._execute(args)

    def mc_bmc(self, bound=20):
        script = os.path.join(top_dir, "runscripts", "model_checking__srs_015.py")
        args = [script]
        args += self.models
        args.append("-p%s" % self.props)
        args.append("-i%s" % self.fixml)
        args.append("--bmc")
        args.append("--bmc-length=%d" % bound)
        return self._execute(args)

    def ftg(self):
        script = os.path.join(top_dir, "runscripts", "fault_tree_generation__srs_017.py")
        args = [script]
        args += self.models
        args.append("-p%s" % self.props)
        args.append("-i%s" % self.fixml)
        args.append("--property-num=0")
        args.append("--dynamic=true")
        return self._execute(args)

    def fte(self, tool):
        script = os.path.join(top_dir, "runscripts", "dynamic_fault_tree_evaluation__srs_025.py")
        args = [script]
        args += self.models  # FT and rates map
        args.append("--tool=%s" % tool)
        args.append("10")         # mission time
        args.append("/dev/null")  # event file
        args.append("/dev/null")  # rates file
        return self._execute(args)

    def sf_imc(self, tool):
        script = os.path.join(top_dir, "runscripts", "performability_evaluation__srs_022.py")
        args = [script]
        args += self.models
        args.append("-p%s" % self.props)
        args.append("-i%s" % self.fixml)
        args.append("--tool=%s" % tool)
        args.append("--error-bound=0.001")
        return self._execute(args)

    def sf_sim(self, tool):
        if tool == 'sim-slow':
            confidence = 0.98
            error_bound = 0.01
        else:
            confidence = 0.95
            error_bound = 0.05
        script = os.path.join(top_dir, "runscripts", "mc_simulation.py")
        args = [script]
        args += self.models
        args.append("-p%s" % self.props)
        args.append("-i%s" % self.fixml)
        args.append("--confidence=%f" % confidence)
        args.append("--error-bound=%f" % error_bound)
        return self._execute(args)

    def bench(self, parameter):
        if parameter < self.__ranges[self.benchmark]:
            log.user.info("Skipping benchmark of size %d: too small", parameter)
            return
        if self.benchmark == 'perf-fte' and parameter > 14:
            log.user.info("Skipping FTE of size %d: too large", parameter)
            return

        cwd = os.getcwd()

        # Name of benchmark executed
        self.benchmark_id = self.benchmark + '-' + str(parameter)
        if (self.benchmark.startswith('perf')):
            self.benchmark_id += '-' + self.perf_tool

        # Store benchmark results in separate dir
        self.output_dir = misc.get_out_fname(self.benchmark_id)
        if not os.path.isdir(self.output_dir):
            os.makedirs(self.output_dir)
        # Everything created in out dir
        os.chdir(self.output_dir)

        log.user.info("generating benchmark %s size %d" % (self.benchmark,parameter))
        self.gen_bench(self.benchmark, parameter)

        log.user.info("executing benchmark %s size %d" % (self.benchmark,parameter))
        start_time = time.time()
        (ret, cpu_time, mem) = self.exec_bench(self.benchmark)
        end_time = time.time()

        time_diff = end_time - start_time
        log.user.info("executing took " + str(time_diff) + " seconds")

        with open(self.output_file, 'a+') as file:
            file.write("%s : %s %s %s\n" % (self.benchmark_id,str(time_diff),str(cpu_time), str(mem)))

        os.chdir(cwd)

    def __init__(self, derived_actions=[]):
        super(BenchScript, self).__init__(
            derived_actions=self.__actions+derived_actions)

        self.benchmark = self.__benchmarks[0]
        self.parameter = 0
        self.range = False
        self.perf_tool = self.__perf_tools[0]

        self.output_file = None
        self.output_dir = None
        self.benchmark_id = ""

        # Filenames for runscript inputs
        self.models = []
        self.props = None
        self.fixml = None

    def get_usage(self):
        return "usage: %prog [options]"

    def register_options(self, opt_parser):
        super(BenchScript, self).register_options(opt_parser)

        opt_parser.add_option("-b", "--benchmark",
                              dest="BENCHMARK",
                              action="store", default=self.__benchmarks[0],
                              help="Benchmark to execute")
        opt_parser.add_option("-p", "--parameter",
                              dest="PARAMETER",
                              action="store", default=0,
                              help="Parameter value for the benchmark (complexity).")
        opt_parser.add_option("--range",
                              dest="RANGE",
                              action="store_true",
                              help="When given, will benchmark from the lowest value up to parameter.")
        opt_parser.add_option("-t", "--tool",
                              dest="PERFTOOL",
                              action="store", default=self.__perf_tools[0],
                              help="Tool to use for performability")

    def check_options(self, opt_parser, options):
        super(BenchScript, self).check_options(opt_parser, options)

        if opt_parser.values.BENCHMARK:
            if not opt_parser.values.BENCHMARK in self.__benchmarks:
                opt_parser.error("Benchmark must be one of " + ', '.join(self.__benchmarks))

        if opt_parser.values.PERFTOOL:
            if not opt_parser.values.PERFTOOL in self.__perf_tools:
                opt_parser.error("Perf tool must be one of " + ', '.join(self.__perf_tools))

        if opt_parser.values.PERFTOOL.startswith('sim') and opt_parser.values.BENCHMARK == 'perf-fte':
            opt_parser.error("FTE not possible with the simulator")
        if opt_parser.values.PERFTOOL == 'mrmc' and opt_parser.values.BENCHMARK in ['perf-nd', 'perf-hybrid']:
            opt_parser.error("MRMC cannot handle non-deterministic/hybrid models")

        if not options.BASE_FILE_NAME:
            options.BASE_FILE_NAME = opt_parser.values.BENCHMARK

    def store_options(self, opt_parser, options):
        super(BenchScript, self).store_options(opt_parser, options)

        if opt_parser.values.BENCHMARK:
            self.benchmark = opt_parser.values.BENCHMARK
        if opt_parser.values.PARAMETER:
            self.parameter = int(opt_parser.values.PARAMETER)
        if opt_parser.values.RANGE:
            self.range = True
        if opt_parser.values.PERFTOOL:
            self.perf_tool = opt_parser.values.PERFTOOL

    def action(self):
        super(BenchScript, self).action()

        # File to dump results
        self.output_file = os.path.abspath(misc.get_out_fname("timing.txt"))
        if self.benchmark.startswith('perf'):
            self.output_file = os.path.abspath(misc.get_out_fname("timing-%s.txt" % self.perf_tool))

        if self.range:
            for i in range(1, self.parameter+1):
                self.bench(i)
        else:
            self.bench(self.parameter)

        log.user.info("Benchmarking has finished, the results can be found in: %s" % self.output_file)

    def _execute(self, args):
        """Execute args in the shell and return return code. stdout and err will be written
        to output_dir"""
        f_stdout = open('stdout', 'w')
        f_stderr = open('stderr', 'w')
        try:
            args = [sys.executable] + args
            log.user.info("Executing: " + ' '.join(args))
            proc = subprocess.Popen(args, stdout=f_stdout, stderr=f_stderr, stdin=None, cwd=os.getcwd())
            maxmem = 0
            while proc.poll() is None:
                pids = get_pid_tree(proc.pid)
                mem = 0
                cpu_time = 0
                for pid in pids:
                    if pid == proc.pid: continue
                    memtime = procmon.get_pid_time_mem(pid)
                    if not memtime: continue
                    (pid_time, pid_mem) = memtime
                    cpu_time += pid_time
                    mem += pid_mem
                if mem > maxmem:
                    maxmem = mem
                time.sleep(1)
            log.user.info("Executing took " + str(cpu_time) + " seconds, consumed max " + str(maxmem) + "KB memory")
            return (proc.returncode, cpu_time, mem)
        except Exception as e:
            log.system.error("Exception: " + str(e))
        finally:
            f_stdout.close()
            f_stderr.close()

def get_pid_tree(pid):
    """Use pstree utility to retrieve all PIDs that are child of given PID
    PID must be an active process"""
    proc = subprocess.Popen(['/usr/bin/pstree', '-p', str(pid)], stdout=subprocess.PIPE, stderr=subprocess.PIPE, stdin=None)
    (stdout, stderr) = proc.communicate()

    pat = re.compile(r'\((\d+)\)')
    pids = re.findall(pat, stdout)
    pids = map(int, pids)
    return pids

if __name__ == "__main__":
    main = BenchScript()
    main.run()
