# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================

import gtk
import gobject
import sys
import os
import traceback
import cgi

debug_info = None

def setup_path():
    """Sets up the python include paths to include needed directories"""
    import os.path 
    from src.utils.globals import TOPDIR

    # src.local_root is used by gui components 
    # .. is used for accessing all the compass tools chain
    sys.path = [os.path.join(TOPDIR, "src/local_root"),
                os.path.join(TOPDIR, ".."),
                os.path.join(TOPDIR, '../../resources/antlr/antlr-python-runtime-3.1.1/')
                ] + sys.path
    return

def setup_application():
    """Sets up options, logging, and any other application-level
    setting"""
    import utils.log
    utils.log.init()
    return


def check_requirements():
    """Checks versions and other requirements"""
    import gtkmvc; gtkmvc.require("1.2.0")
    return

def log_internal_errors(etype, evalue, etb):
    global debug_info
    from utils import log
    from utils.log import log_icd
    from src.utils import debug
    log.system.error("Exception occurred: " + str(evalue))
    @log_icd
    def display_error(etype, evalue, etb):
        messageDlg = gtk.MessageDialog(parent=None,
                                flags=gtk.DIALOG_MODAL,
                                type=gtk.MESSAGE_ERROR,
                                buttons=gtk.BUTTONS_CLOSE,
                                message_format=None)

        def _handle_clicked(self, *args):
            self.destroy()

        messageDlg.connect('response', _handle_clicked)

        messageDlg.set_title("Internal Error")
        error_msg = "Exception occurred: " + str(evalue)
        error_msg += "\nPlease check the output console for more details"
        messageDlg.set_markup(cgi.escape(error_msg))

        msgContent = messageDlg.get_content_area()
        expander = gtk.Expander()
        expander.set_label("Exception details")

        detailLabel = gtk.Label()
        detailLabel.set_text("".join(traceback.format_exception(etype, evalue, etb, 10)))

        expander.add(detailLabel)
        msgContent.pack_end(expander, False, False, 0)

        messageDlg.show_all()

    gobject.idle_add(display_error, etype, evalue, etb)

    if debug_info is not None:
        debug.dump_debug_info(debug_info)

def install_exception_hook():
    """Add an global exception handler for logging the exception"""

    def new_hook(etype, evalue, etb):
        log_internal_errors(etype, evalue, etb)
        sys.excepthook = new_hook # set the hook again

    sys.excepthook = new_hook
    pass

def register_debug_info(debug):
    global debug_info
    debug_info = debug

def main():
    from src.models.application import ApplModel as ApplModel
    from src.ctrls.application import ApplCtrl as ApplCtrl
    from src.views.application import ApplView as ApplView

    m = ApplModel()
    v = ApplView()
    c = ApplCtrl(m,v)

    def load_files_in_project(m, v, c):
        # Store the current CWD, files are loaded form here
        start_dir = os.getcwd()
        new_cwd = start_dir

        # slim files
        if options.SLIM_FILE_NAMES:
            try: m.slim.add_slim_models(options.SLIM_FILE_NAMES)
            except IOError, e:
                err_dia = gtk.MessageDialog(parent=v.get_top_widget(),
                                            buttons=gtk.BUTTONS_CLOSE,
                                            flags=gtk.DIALOG_MODAL,
                                            type=gtk.MESSAGE_ERROR)
                err_dia.set_markup("<big><b>File not found</b></big>")
                err_dia.format_secondary_markup(str(e))
                err_dia.run()
                err_dia.destroy()
                pass
            # Store the possibly new CWD
            new_cwd = os.getcwd()

        #Restore CWD to initial
        os.chdir(start_dir)

        # injections:
        if options.FAULT_INJECTION_FILE_NAME:
            m.slim.load_fault_injections(options.FAULT_INJECTION_FILE_NAME)
            pass

        # mission
        if options.MISSION_FILE_NAME:
            m.mission.get_mission_doc(options.MISSION_FILE_NAME)
            c.on_menu_tfpg_item_enabled()
            pass

        # tfpg
        if options.TFPG_FILE_NAME:
            tfpg_doc = m.tfpg.get_tfpg_doc(options.TFPG_FILE_NAME)
            c.tfpg_activate = True
            c.on_menu_tfpg_item_enabled()
            pass

        # tfpg assoc
        if options.TFPG_ASSOC_FILE_NAME:
            c.tfpg.tfpg_slim_assoc_ctrl.get_assoc_doc_from_shell(tfpg_doc, options.TFPG_ASSOC_FILE_NAME)
            pass

        # loading of properties is not supported in property model...
        if options.PROPERTIES_FILE_NAME:
            c.properties.load_properties_file(options.PROPERTIES_FILE_NAME)

        # Finally set CWD to path of last sSLIM file
        os.chdir(new_cwd)

        return False

    import options
    gobject.idle_add(load_files_in_project, m, v, c)
    install_exception_hook()
    gtk.main()
    return

if __name__ == "__main__":
    setup_path()

    from src.utils.optparser import GUIOptions
    options = GUIOptions()
    options.parse()

    setup_application()
    check_requirements()

    # backup the original exception hook
    orig_hook = sys.excepthook

    main()

    #restore the original exception hook
    sys.excepthook = orig_hook
    pass
