# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================


import thread
import gtk
import gobject
import re
import xml.sax.saxutils
from gtkmvc import Controller

import property_management.src.constants as prop_const
from utils.exceptions import NusmvControlledException

from src.ctrls.activity import ActivityCtrl, StartStopActivityCtrl
from src.ctrls.trace import TraceCtrl, FilterBarCtrl
from src.views.trace import TraceView, FilterBarView
from src.ctrls.trace import ExtensibleTraceCtrl, GuidedSimTransitionCtrl
from src.views.trace import ExtensibleTraceView, GuidedSimTransitionView
from src.models.trace import TraceViewModel
from src.models.correctness import ModelSimulationModel
from src.models.filtered_properties import FilteredPropertiesModel

import slim_compiler.src.smv.api as smvapi
from slim_compiler.src.smv import support
from slim_compiler.src.dataStructure import nominalbehaviour


class CorrectnessCtrl (ActivityCtrl):
    """Controller for the Correctness activity. Two sub-controllers
    control model checking and model simulation panels."""
    def __init__(self, model, view):
        ActivityCtrl.__init__(self,
                              CorrectnessSubCtrl(self, model, view.subview),
                              model, view)

        # this is used to control the type of selection of properties
        self.mode = None

        self._mc_col_title = "MC"
        return

    def register_view(self, view):
        ActivityCtrl.register_view(self, view)

        self.set_mode("sim")

        tv = view['tv_properties']
        tv.set_model(self.model.filtered_props.get_properties_list())

        # selection
        def __rend_select(col, rend, model, iter):
            col_num = FilteredPropertiesModel.SELECT_MULTI
            # in earlier version multi was not allowed in mc:
            #col_num = 1 if self.mode == "mc" else 2
            rend.set_property("active", model[iter][col_num])
            return

        rend1 = gtk.CellRendererToggle()
        rend1.set_property("activatable", True)
        rend1.connect("toggled", self.on_prop_toggled)
        col = gtk.TreeViewColumn('Name', rend1)
        col.set_cell_data_func(rend1, __rend_select)

        # name
        rend2 = gtk.CellRendererText()
        col.pack_start(rend2)
        col.set_cell_data_func(rend2, lambda c,r,m,i: r.set_property('text', m[i][FilteredPropertiesModel.PROPERTY].name))
        tv.append_column(col)

        # status
        def __prop_status(col, rend, model , iter):
            result = model[iter][FilteredPropertiesModel.MC_RESULT]
            if result is None: rend.set_property('stock-id', None)
            elif result.truth: rend.set_property('stock-id', gtk.STOCK_YES)
            elif result.truth is None:
                if result.sat_length is None:
                    rend.set_property('stock-id', gtk.STOCK_DIALOG_ERROR)
                else:
                    rend.set_property('stock-id', gtk.STOCK_DIALOG_QUESTION)
            else: rend.set_property('stock-id', gtk.STOCK_NO)
            return
        rend = gtk.CellRendererPixbuf()
        col = gtk.TreeViewColumn(self._mc_col_title, rend)
        col.set_cell_data_func(rend, __prop_status)
        tv.append_column(col)

        # formula
        rend = gtk.CellRendererText()
        col = gtk.TreeViewColumn('Formula', rend)
        col.set_cell_data_func(rend, lambda c,r,m,i: r.set_property('text', m[i][FilteredPropertiesModel.PROPERTY].description))
        tv.append_column(col)

        # no selection for this tv:
        tv.get_selection().set_mode(gtk.SELECTION_NONE)

        # to process double click that shows the corresponding mc result in mc mode
        tv.connect('row-activated', self.on_row_activated, self)
        return

    def set_mode(self, mode):
        """Set current modality: model checking or simulation"""
        if mode == self.mode: return
        assert mode in ("sim", "dc", "mc", "zeno", "td")
        self.mode = mode
        self.view['tv_properties'].queue_draw()

        if mode is "mc":
            nominal = not self.model.model_check.extended_model

            if self.model.model_check.use_bdd: 
                if self.model.slim_model.timed_domain:
                    logics = (prop_const.Logic.CTL,
                              prop_const.Logic.LTL,
                              prop_const.Logic.MITL,)
     
                else:
                    logics = (prop_const.Logic.CTL,
                              prop_const.Logic.LTL,)
                      
            else: logics =(prop_const.Logic.LTL,prop_const.Logic.MITL,)  
            

        elif mode is "dc":
            nominal  = not self.model.model_dc.extended_model
            logics = self.model.filtered_props.get_filtering() # don't care

        elif mode is "sim":
            nominal  = not self.model.model_sim.extended_model
            logics = None # propositional only

        elif mode is "zeno":
            nominal  = not self.model.model_zeno.extended_model
            logics = None

        elif mode is "td":
            nominal  = not self.model.model_td.extended_model
            logics = None

        else: assert False, "unexpected mode"

        self.model.filtered_props.set_filtering(logics, nominal)

        # tv_properties is not enabled for Deadlock Checking, Zeno Analysis and Time Divergence
        self.view['tv_properties'].set_sensitive(mode != "dc" and mode!= "zeno" and mode!= "td")

        return

    # ---------------------------------------------------------------
    #        Signal handlers
    # ---------------------------------------------------------------
    def on_prop_toggled(self, rend, path):
        child_path = self.model.filtered_props.get_properties_list().convert_path_to_child_path(path)
        self.model.props_model.toggle_property(child_path, multi=True)
        # In earlier version, it was not possible to select multiple properties in MC:
        #self.model.props_model.toggle_property(child_path, multi=(self.mode == "sim"))
        return

    def on_row_activated(self, tv, path, column, who):
        if who is self and self.mode == "mc" and column.get_title() == self._mc_col_title:
            mc_result = self.model.filtered_props.get_properties_list()[path][FilteredPropertiesModel.MC_RESULT]
            self.model.model_check.result = mc_result
            pass
        return

    pass # end of class
# ----------------------------------------------------------------------


class CorrectnessSubCtrl (Controller):
    """Sub-controller for the correctness activity sub-view"""
    def __init__(self, parent, model, view):
        Controller.__init__(self, model, view)

        # chief controller
        self.parent = parent

        # sub-controllers (view is shared)
        self.model_check = ModelCheckingCtrl(self, model.model_check, view.mc)
        self.model_sim = ModelSimulationCtrl(self, model.model_sim, view.sim)
        self.dc = DeadlockCheckingCtrl(self, model.model_dc, view.dc)
        self.zeno_analysis = ZenoAnalysisCtrl(self, model.model_zeno, view.zeno)
        self.td = TimeDivergenceCtrl(self, model.model_td, view.td)

        return

    def register_view(self, view):

        return

    # ---------------------------------------------------------------
    #        Signal handlers
    # ---------------------------------------------------------------
    def on_nb_correctness_switch_page(self, nb, page, num):
        assert num >= 0 and num < 5
        self.parent.set_mode({ 0 : "sim", 1 : "dc", 2: "mc", 3: "zeno", 4: "td"}[num])
        return

    # ---------------------------------------------------------------
    #        Notifications from observable properties
    # ---------------------------------------------------------------


    pass # end of class
# ----------------------------------------------------------------------

class ModelCheckingCtrl (StartStopActivityCtrl):

    def __init__(self, parent, model, view):
        StartStopActivityCtrl.__init__(self, model, view,
                                       model.run_model_checking,
                                       model.stop_nusmv)

        # this sub-ctrl handles the result view
        self.results_ctrl = ResultsCtrl(model, view.mc_results)

        self.parent = parent

        # we observe the number of failures into the slim model to
        # enable the corresponding button.
        self.model.slim_model.register_observer(self)

        self.prev_trace_ctrl = None # this is used to dispose controllers

        # we also observe the number of selected properties in the
        # properties model, to enable model checking button
        # accordingly
        self.model.filtered_props.register_observer(self)

        self.tvm = TraceViewModel() #the model of all (filterbar, tracemodel, extensibletracemodel..)

        self.fbar_view = FilterBarView(windowed=False)
        self.fbar = self.fbar_view['TBFilter']
        self.fbar_ctrl = FilterBarCtrl(self.tvm,
                                       self.fbar_view)

        return

    def register_view(self, view):
        StartStopActivityCtrl.register_view(self, view)

        # sets sensitivity of sat stuff depending on the selected tecnology
        # [MP] self.view['table_sat'].set_sensitive(not self.model.use_bdd)
        self.view.enable_sat_options(not self.model.use_bdd)
        return

    def register_adapters(self):
        StartStopActivityCtrl.register_adapters(self)

        self.adapt("use_bdd")
        self.adapt("sat_bound")
        self.adapt("klive_bound")
        self.adapt("use_sbmc")
        self.adapt("sbmc_complete")
        self.adapt("extended_model", "cb_mc_extended_model")
        return

    # ---------------------------------------------------------------
    #        Signal and other callbacks handlers
    # ---------------------------------------------------------------

    # ---------------------------------------------------------------
    #        Notifications from observable properties
    # ---------------------------------------------------------------

    def property_multi_selected_num_value_change(self, model, old, new):
    
        
        rows = [row for row in self.model.filtered_props
                if row[FilteredPropertiesModel.SELECT_MULTI]]

        ordered_rows = sorted(rows,
                              key=lambda r: r[
                                  FilteredPropertiesModel.PROPERTY].
                              property.get_logic() ==
                              prop_const.Logic.MITL)
        # gets the list of selected properties (currently only one
        # property can be selected)
        properties = [row[FilteredPropertiesModel.PROPERTY].property
                      for row in ordered_rows]
                      
        logics = []          
        for property in properties:
            logics.append(property.get_logic())
          
        # NO Properties selected 
        if not logics:
            self.view["sb_klive_bound"].set_sensitive(True)       
            self.view["cb_sbmc_complete"].set_sensitive(True)    
            self.view["cb_use_sbmc"].set_sensitive(True)    
           
        # Timed + Non Timed Properties selected
        elif prop_const.Logic.MITL in logics:
            if prop_const.Logic.LTL in logics or prop_const.Logic.CTL in logics:
                self.view["sb_klive_bound"].set_sensitive(True)       
                self.view["cb_sbmc_complete"].set_sensitive(True)    
                self.view["cb_use_sbmc"].set_sensitive(True)
            else:
                # Timed Properties    
                self.view["cb_sbmc_complete"].set_active(False)
                self.view["cb_sbmc_complete"].set_sensitive(False)
                self.view["cb_use_sbmc"].set_active(False)
                self.view["cb_use_sbmc"].set_sensitive(False)
                self.view["sb_klive_bound"].set_sensitive(True)  
            
        else:
            # Non Timed Properties    
            self.view["sb_klive_bound"].set_sensitive(False)
            self.view["cb_sbmc_complete"].set_sensitive(True)    
            self.view["cb_use_sbmc"].set_sensitive(True)
 
        self.view.enable_model_checking(new>0)
        return

    def property_failures_value_change(self, model, old, new):
        """Sets the sensibility of button for enabling checking of
        extended model if there are failures in the slim model"""
        self.view.enable_model_extension(new > 0)

    # when bdd is used, sat stuff is grayed out, and property
    # filtering is set accordingly (no CTL for sat solving)
    def property_use_bdd_value_change(self, model, old, new):

        # LTL is always enabled, CTL is enabled only when using BDDs
        logics = (prop_const.Logic.CTL,
                  prop_const.Logic.LTL,
                  prop_const.Logic.MITL,) if new else (prop_const.Logic.LTL,prop_const.Logic.MITL,)

        # whether to consider extended model or not
        nominal = not self.model.extended_model

        self.model.filtered_props.set_filtering(logics, nominal)

        # [MP] self.view['table_sat'].set_sensitive(not new)
        self.view.enable_sat_options(not new)
        return

    def property_extended_model_value_change(self, model, old, new):
        # LTL is always enabled, CTL is enabled only when using BDDs
        logics = (prop_const.Logic.CTL,
                  prop_const.Logic.LTL,
                  prop_const.Logic.MITL,) if new else (prop_const.Logic.LTL,prop_const.Logic.MITL,)

        # whether to consider extended model or not
        nominal = not self.model.extended_model

        self.model.filtered_props.set_filtering(logics, nominal)
        return

    def property_can_use_bdd_value_change(self, model, old, new):
        if not new:
            # checks if bdd is currenlty selected, and switch to
            # SAT if needed (this can happens when the user loads a
            # slim model which has infinite domain)
            if self.view['rb_use_bdd'].get_active():
                self.view['rb_use_bdd'].set_active(False)
                self.view['rb_use_sat'].set_active(True)
                pass
            pass

        self.view['rb_use_bdd'].set_sensitive(new)
        return

    def property_result_value_change(self, model, old, new):
        """this is used to process results of model checking"""
        if isinstance(new, NusmvControlledException):
            self.view.show_exception(new.get_title(), new.get_description())
            return

        if self.prev_trace_ctrl != None:
            self.prev_trace_ctrl.relieve_model(self.prev_trace_ctrl.model)
            self.prev_trace_ctrl = None
            pass

        if new is None:
            self.view.mc_results.show_no_results()
            return

        # If truth is none, either bounds have been reached OR an error occurred.
        # Errors are however not clearly passed by nusmv processes, so for now check
        # if sat_length is none, and if so assume some error occurred and refer the
        # user to the log
        if new.truth is None and new.error is None and new.sat_length is None:
            self.view.show_exception("Unknown SMV error", "Check the logs for details")
            return

        unk_msg = ("true up to bound %d" % new.sat_length) \
            if new.sat_length is not None else "true"

        if new.error is None:
            truth, stock_id = { None : (unk_msg , gtk.STOCK_DIALOG_QUESTION),
                                True : ("true", gtk.STOCK_YES),
                                False : ("false", gtk.STOCK_NO),
                                } [new.truth]
        else: return

        desc = """The <tt>%s</tt> property:<span background="#feffbd"><tt>%s</tt></span> has been found <b>%s</b>.""" % (new.type_name, xml.sax.saxutils.escape(new.formula_text), truth)

        trace_view = None

        if new.truth is False:
            desc += """ A counter-example is shown below."""

            trace_view = TraceView(windowed=False, filters=False)
            self.tvm.add_traces(new.trace)
            self.tvm.set_active_trace(new.trace)

            self.prev_trace_ctrl = TraceCtrl(self.tvm, trace_view)
            assert self.fbar_ctrl
            assert self.fbar_view
            pass

        self.view.show_mc_result(truth, desc, trace_view, self.fbar_view, stock_id)
        return

    pass #end of class
# ----------------------------------------------------------------------


class ModelSimulationCtrl (StartStopActivityCtrl):
    """
    This class is a controller for the model simulation pane (ModelSimulationView) that
    handles some controllers (such as jump, restart and run simulation), a trace,
    in guided or random simulation mode, depending on the simulation type of the model,
    and a filter bar, which appears only when a trace is shown.
    """
    # error and info messages
    init_message = 'Click on the <span foreground="dark green" weight="bold">Pick</span> button'
    existing_message = 'Click first on the <span foreground="dark green" weight="bold">Pick</span> button'
    existing_pick_message = 'Now click on the <span foreground="dark green" weight="bold">column</span> that you would use as starting point'
    no_trace_message = 'There is not a trace from which you can choose a starting point!\nSelect another options, or generate first a new trace'
    too_strong_costraints = '''<span foreground="red" weight="bold">It was not possible to simulate.</span>
Constraints may be <b>too strong</b>,
or the simulator has reached a <b>deadlock state</b> in the model.'''
    extend_trace_message = 'Now you can extend the trace clicking on the button <b>run model simulation</b>'
    begin_message = 'You can either simulate from scratch, or start a guided simulation'
    simulate_message = 'Now you can simulate from the trace shown below'
    simulation_stopped_by_user = 'The simulation <b>was stopped!</b> Cannot add new steps'

    PICK_DELAY = 30000 # 30 seconds

    def __init__(self, parent, model, view):
        StartStopActivityCtrl.__init__(self, model, view,
                                       model.run_model_simulation,
                                       model.stop_simulation)
        self.parent = parent
        # we observe the number of failures into the slim model to
        # enable the corresponding button.
        self.model.slim_model.register_observer(self)

        self.prev_trace_ctrl = None # this is used to dispose controllers
        self.startStep = -1 #invalid step
        # this sub-ctrl handles the result view
        self.results_ctrl = ResultsCtrl(model, view.sim_results)

        self.tvm = TraceViewModel()
        # set filter toolbar
        self.fbar_view = FilterBarView(windowed=False)
        self.fbar = self.fbar_view['TBFilter']
        self.fbar_ctrl = FilterBarCtrl(self.tvm,
                                       self.fbar_view)

        self.wait_for_picking = False
        self._timeout = None
        self.prev_trace_ctrl_extensible = None
        #view['HBFromInitExisting'].set_sensitive(True)

        self.dtc = None # deadlock trace controller
        return

    def register_view(self, view):
        StartStopActivityCtrl.register_view(self, view)
        return

    def register_adapters(self):
        StartStopActivityCtrl.register_adapters(self)
        self.adapt("sim_len", "sb_sim_len")
        self.adapt("extended_model", "cb_sim_extended_model")
        return


    # ---------------------------------------------------------------
    #        Signal handlers
    # ---------------------------------------------------------------
    def property_extended_model_value_change(self, model, old, new):
        self.model.filtered_props.set_filtering(None, not new)
        return

    def on_cb_sim_type_changed(self, cb):
        if cb.get_active() == 0: # random sim
            self.model.set_type(self.model.RANDOM)
            self.model.pick_unset() #unset picking
            self.model.can_simulate = True
            self.view.select_random()
            pass

        elif cb.get_active() == 1: # transition-based sim
            self.model.set_type(self.model.TRANS_GUIDED)
            self.view.select_guided()
            if self.prev_trace_ctrl_extensible is not None:
                self.prev_trace_ctrl_extensible.view.select_trans_based_guided()
                pass

            self.model.calculate_guided_structure()
            self.model.can_simulate = self.model.trans_guided.get_can_simulate()
            pass

        elif cb.get_active() == 2: # value-based sim
            self.model.can_simulate = True
            self.model.set_type(self.model.VALUE_GUIDED)
            self.view.select_guided()
            if self.prev_trace_ctrl_extensible is not None:
                self.prev_trace_ctrl_extensible.view.select_value_based_guided()
                pass
            pass

        else: assert False, "No possible choice: %d" % cb.get_active()
        return

    def on_button_sim_restart_clicked(self, btn):
        self.model.pick_initial_state()
        return

    def goto_state(self, btn):
        # pick from existing
        self.prev_trace_ctrl.clear_step_selected()
        self.view.set_info_label(self.existing_pick_message)
        self.model.pick_unset()
        self.pick_wait() #wait for picking
        return

    def __choose_message(self, result):
        # helper for other functions
        if result:
            return self.existing_message
        else:
            return self.no_trace_message

    #FUNCTION FOR PICKING
    #helper fun for picking
    def _pick_expired(self):
        self.wait_for_picking = False
        self._timeout = None
        self.view.clear_info_label()
        return
    #end

    def pick_wait(self):
        """
        wait 30 seconds, if none call pick_set() in this time,
        it isn't possible to call pick_set(), until pick_wait is recalled.
        """
        if self._timeout:
            return #do not add more than one timeout
        self.wait_for_picking = True
        #use timeout_add
        self._timeout = gobject.timeout_add( ModelSimulationCtrl.PICK_DELAY , self._pick_expired )
        return

    #end helper function for picking

    def step_picked(self, col, step):
        #CALLBACK for the col_clicked signal
        #if expired
        if not self.wait_for_picking:
            return

        #if not expired
        if self.prev_trace_ctrl:
            self.startStep = step+1
            #jumping to the last col will get the same result as the previous one
            if self.startStep == self.model.result.steps():
                return
             #here i ask the confirm
            md = gtk.MessageDialog(None,
                                   gtk.DIALOG_DESTROY_WITH_PARENT, gtk.MESSAGE_QUESTION,
                                   gtk.BUTTONS_YES_NO
                                   )
            md.set_markup("<big><b>Do you want to jump?</b></big>\nJumping to a previous state will slice the current simulation at the chosen state (at step %d). Following steps will be no longer available." % self.startStep )
            res = md.run()
            if res == gtk.RESPONSE_YES:
                self.model.pick_set( self.startStep )
                self.view.set_info_label(self.simulate_message)
                pass
            md.destroy()
            pass

        self._pick_expired() # now the user have to reclick on the jump button in order to jump
        return

    # ---------------------------------------------------------------
    #        Notifications from observable properties
    # ---------------------------------------------------------------

    def property_busy_value_change(self, model, old, new):
        StartStopActivityCtrl.property_busy_value_change(self, model, old, new)

        #hide or show all the buttons
        self.view.enable_options(new <= 0)

        if self.model.can_simulate is False:
            self.view.get_start_button().set_sensitive(False)
            pass

        if model.RANDOM != self.model.sim_type:
            self.view.select_guided()
            pass
        return

    def property_sim_type_value_change(self, model, old, new):
        if new in (model.TRANS_GUIDED, model.VALUE_GUIDED):
            self.view.sim_results.show_results_page(1)
        else: # random
            assert new == model.RANDOM
            self.view.sim_results.show_results_page(0)
            pass

        return

    def property_can_simulate_value_change(self, model, old, new):
        if self.model.busy == 0:
            self.view.get_start_button().set_sensitive(new)
            pass
        return

    def property_failures_value_change(self, model, old, new):
        """Sets the sensibility of button for enabling checking of
        extended model if there are failures in the slim model"""
        self.view.enable_model_extension(new > 0)

    def property_deadlock_found_signal_emit(self, model, trace_model):
        dtm = TraceViewModel()
        dtm.add_traces(trace_model)
        dtm.set_active_trace(trace_model)
        dtv = TraceView(windowed=False, filters=True)
        self.dtc = TraceCtrl(dtm, dtv)

        self.view.show_deadlock(dtv)
        self._pick_expired()
        return

    def property_invalid_trace_signal_emit(self, model, arg):
        """
        This signal is emitted when the read_trace function fails.
        """
        self.view.sim_results.show_page(3)
        self._pick_expired()
        return

    def property_new_trace_signal_emit(self, model, arg):
        """
        This signal is emitted when the user try to extend a simulation
        after the fault injections are introduced or deleted from the model.
        It shows a little warning.
        """
        self.view.sim_results.show_no_results(
            desc="Warning! <b>The trace will be recalculated from scratch</b>, because the model is changed"
            )
        self._pick_expired()
        return

    def property_sim_stopped_signal_emit(self, model, arg):
        page_showed = self.view.sim_results.get_active_page()
        if page_showed == 0: # the "no results to show" page
            self.view.sim_results.show_no_results(self.simulation_stopped_by_user)
        elif page_showed == 1: # if there is the "results" page actually showed
            self.view.sim_results.change_results_description(self.simulation_stopped_by_user)
        # if there is another page showed do nothing
        return

    def property_sim_failed_signal_emit(self, model, arg):
        """
        This is received when the simulation fails, either in
        the picking stage, or in the simulation stage
        """
        if self.model is not model: return

        # if simulation is guided, then the previous trace (if any)
        # is kept, and the user is notified with a message.
        #
        # Otherwise, the trace is reset and a panel is show telling
        # the user that constraints are too strong if there are
        # selected constraints, or that a deadlock has been found
        # otherwise.
        self._pick_expired()

        error_message = str(arg) if arg else self.too_strong_costraints

        if model.get_type() in (model.TRANS_GUIDED, model.VALUE_GUIDED):
            self.view.sim_results.change_results_description( error_message )
        elif model.get_type() == model.RANDOM:
            self.model.result = None # obscure the previous trace
            self.view.sim_results.show_default_error_results( desc=error_message )
        else:
            raise Exception( "Unhandled type of simulation "+str(model.get_type()) )

        return

    def property_slim_changed_signal_emit(self, model, arg):
        """
        This is received when the slim model changes, but the result property in the model is already None.
        In this case it is necessary to reset the view showing the "no result to show" page.
        (Note that if the result is not None, the event is triggered by the property_result_value_change()
        method)
        """
        self.view['button_sim_restart'].set_sensitive(False)
        self.view['button_sim_jump'].set_sensitive(False)
        self.view.sim_results.show_no_results()
        self._pick_expired()
        return

    def property_result_value_change(self, model, old, new):
        """this is used to process results of model simulation"""

        gstc = None
        gstv = None
        pos_random = None
        pos_guided = None
        if isinstance(new, NusmvControlledException):
            self.view.show_exception(new.get_title(), new.get_description())
            return

        if self.prev_trace_ctrl != None:
            # saves the current scrolling positions before relieving them
            pos_random = self.prev_trace_ctrl.view.get_scroll_pos()
            pos_guided = self.prev_trace_ctrl_extensible.view.get_scroll_pos()

            # saves the transition-based guided sim controller for being reused later
            gstc = self.prev_trace_ctrl_extensible.get_trans_sim_guided_ctrl()
            gstv = self.prev_trace_ctrl_extensible.view.detach_trans_sim_guided_view()

            self.prev_trace_ctrl.relieve_model(self.prev_trace_ctrl.model)
            self.prev_trace_ctrl_extensible.relieve_model(self.prev_trace_ctrl_extensible.model)

            self.prev_trace_ctrl = None
            self.prev_trace_ctrl_extensible = None
            pass

        if new is None:
            # In this case I have to do the same stuff that does this signal callback
            # The parameter is actually not used, so they are not important
            self.property_slim_changed_signal_emit(None, None)
            return

        trace_view = None

        if new.trace is not None:

            #desc = """A simulation <span foreground="dark green">exists</span> and it is shown by the following <b>trace</b>"""
            desc = ""
            #set model
            self.tvm.add_traces(new.trace)
            self.tvm.set_active_trace(new.trace)

            # creates the guided trace view, and scrolls it to the
            # previous position if needed:
            if gstv is None: gstv = GuidedSimTransitionView()
            trace_view_guided = ExtensibleTraceView(gstv, windowed=False, filters=False)
            if self.model.get_type() == ModelSimulationModel.TRANS_GUIDED:
                trace_view_guided.select_trans_based_guided()
            elif self.model.get_type() == ModelSimulationModel.VALUE_GUIDED:
                trace_view_guided.select_value_based_guided()
                pass

            if pos_guided is not None: trace_view_guided.set_scroll_pos(pos_guided)

            # here the guided simulation controller is created and stored
            if gstc is None: gstc = GuidedSimTransitionCtrl(self.model.trans_guided,
                                                            trace_view_guided.trans_view,
                                                            self)

            self.prev_trace_ctrl_extensible = ExtensibleTraceCtrl(self.tvm, trace_view_guided, gstc)

            # creates the random trace view, and scrolls it to the
            # previous position if needed:
            trace_view = TraceView(windowed=False, filters=False) #change
            if pos_random is not None: trace_view.set_scroll_pos(pos_random)

            # here the corresponding random simulation controller is created and sorted:
            self.prev_trace_ctrl = TraceCtrl(self.tvm, trace_view)

            #set the function to be called when someone click a column
            self.prev_trace_ctrl.set_on_col_click_fun( self.step_picked )
            self.prev_trace_ctrl_extensible.set_on_col_click_fun( self.step_picked )
        else:
            desc = """A simulation <span foreground="red">does not exist</span>.
It might be due to unconsistent constraints within the model, or within the given properties."""
            pass

        self.view.show_sim_result(desc,
                                  trace_view,
                                  trace_view_guided,
                                  self.fbar_view)

        return

    pass # end of class
# ----------------------------------------------------------------------


class DeadlockCheckingCtrl (StartStopActivityCtrl):

    def __init__(self, parent, model, view):
        StartStopActivityCtrl.__init__(self, model, view,
                                       model.run_deadlock_checking,
                                       model.stop_nusmv)

        # we observe the slim model to reset previous results
        self.parent = parent

        # we observe the number of failures into the slim model to
        # enable the corresponding button.
        self.model.slim_model.register_observer(self)

        return

    def register_view(self, view):
        StartStopActivityCtrl.register_view(self, view)
        return

    # ---------------------------------------------------------------
    #        Signal and other callbacks handlers
    # ---------------------------------------------------------------

    # ---------------------------------------------------------------
    #        Notifications from observable properties
    # ---------------------------------------------------------------
    def property_extended_model_value_change(self, model, old, new):
        # [MP] deadlock sub-model does not have a filtered_properties attribute
        pass

    def property_available_value_change(self, model, old, new):
        if not new: self.view.show_not_avail()
        else: self.view.show_ready()
        return

    def property_failures_value_change(self, model, old, new):
        """Sets the sensibility of button for enabling checking of
        extended model if there are failures in the slim model"""
        self.view.enable_model_extension(new > 0)

    def property_status_value_change(self, model, old, new):
        if model is self.model:
            if new == "running": self.view.show_running()
            elif new == "error": self.view.show_error()
            elif new == "yes": self.view.show_yes()
            elif new == "no": self.view.show_no()
            else: self.view.show_ready()
            pass

        return

    def register_adapters(self):
        StartStopActivityCtrl.register_adapters(self)
        self.adapt("extended_model", "cb_dc_extended_model")
        return

    pass #end of class
# ----------------------------------------------------------------------


# LB: the new class for ZENO Analysis Ctrl

class ZenoAnalysisCtrl (StartStopActivityCtrl):

    zeno_counter = 0

    # Observable properties
    __observables__ = ("zeno_counter", )


    def __init__(self, parent, model, view):
        StartStopActivityCtrl.__init__(self, model, view,
                                       self.run_zeno_analysis,
                                       model.stop_zeno_analysis)

        # we observe the slim model to reset previous results
        self.parent = parent

        # The result ctrl: it can contains the trace for zeno and non-zeno results
        self.results_ctrl = None

        # It is the liststore of modes and results. It is composed of:
        # enabled (bool), slim_mode (string), smv_mode (string), 
        #message/result (string) and colour (string).
        self.modes_results_liststore = gtk.ListStore(bool, str, str, str, str, str)

        # we observe the number of failures into the slim model to
        # enable the corresponding button.
        self.model.slim_model.register_observer(self)

        return


    def register_view(self, view):
        StartStopActivityCtrl.register_view(self, view)

        view['win_zeno_detached'].connect("delete-event",
                                        self.on_window_zeno_detached_delete_event)

        self.trace_is_detached = False

        return


    def on_window_zeno_detached_delete_event(self, win, event):
        """When the delete button is clicked, the hbox results is re-attached."""
        self.view.attach()
        self.trace_is_detached = False
        return True


    def on_button_zeno_detach_clicked(self, button):
        """When the detach button is clicked, the hbox results is detached."""
        self.view.detach()
        self.trace_is_detached = True
        return


    def initialize_zeno_analysis_tab(self, modes_results_ls):
        """Initializes the liststore of modes and results in the relate Treeview.
            @param   modes_results_ls:  The list of enum modes from the slim_model loaded.
            @type    modes_results_ls:  list of string

            @return   self.modes_results_liststore:  The liststore of modes and results.
            @rtype    self.modes_results_liststore:  gtk.ListStore
        """

        for mode_res in modes_results_ls:
            self.modes_results_liststore.append([False, mode_res[0], mode_res[1], mode_res[2], "", "#C0C0C0"])

        # The treeview of enabled, enum modes and results is created
        # and the columns will be filled

        tv = self.view['tv_zeno_modes_results']
        tv.set_model(model=self.modes_results_liststore)


        # Column Enabled
        renderer_toggle_enabled = gtk.CellRendererToggle()
        renderer_toggle_enabled.connect("toggled", self.on_cell_toggled)
        column_toggle_enabled = gtk.TreeViewColumn("Enabled", renderer_toggle_enabled, active=0)

        self.toggle_button = gtk.CheckButton("Enabled")
        self.toggle_button.show()
        column_toggle_enabled.set_widget(self.toggle_button)
        column_toggle_enabled.set_clickable(True)
        column_toggle_enabled.connect("clicked", self.on_header_enabled_clicked)
        column_toggle_enabled.set_resizable(True)

        tv.append_column(column_toggle_enabled)

        # Column Components
        renderer_text_components = gtk.CellRendererText()
        column_text_components = gtk.TreeViewColumn("Components", renderer_text_components, text=2, foreground=5)
        column_text_components.set_resizable(True)
        tv.append_column(column_text_components)

        # Column Modes
        renderer_text_modes = gtk.CellRendererText()
        column_text_modes = gtk.TreeViewColumn("Modes", renderer_text_modes, text=1, foreground=5)
        column_text_modes.set_resizable(True)
        
        tv.append_column(column_text_modes)

        # Column Results
        renderer_text_results = gtk.CellRendererText()
        column_text_results = gtk.TreeViewColumn("Results", renderer_text_results, text=4, foreground=5)
        column_text_results.set_resizable(True)
        
        tv.append_column(column_text_results)


        # To control selection of results
        sel = self.view['tv_zeno_modes_results'].get_selection()
        sel.set_mode(gtk.SELECTION_MULTIPLE)
        sel.connect('changed', self.on_results_selection_changed)

        return self.modes_results_liststore



    def get_modes_results_liststore(self):
        """Gets the liststore of modes and results.
            @return   self.modes_results_liststore:  The liststore of modes and results.
            @rtype    self.modes_results_liststore:  gtk.ListStore
        """
        return self.modes_results_liststore


    def run_zeno_analysis(self):
        """Runs Zeno Analysis."""

        # When the zeno analysis runs once again, the view and the Treeview are cleared
        if self.zeno_counter >= 1:
            self.view.clear()

        # Performs the Zeno Analysis
        self.model.analyze_zeno(self.modes_results_liststore)


        # Populates the Treeview with the results of Zeno Analysis:
        # for each enabled mode there is a result/message created
        # at the end of Zeno Analysis

        modes = self.get_modes_from_walker()

        for message in self.view.get_messages_list():

            for mode in modes:
            
                if message[0] == mode[0]:

                    for mode_ls in self.modes_results_liststore:
                        if mode_ls[0] is True and mode[0] == mode_ls[1]:
                            mode_ls[4] = message[1]


        self.zeno_counter += 1

        return


    def property_fdir_comp_counter_value_change(self, model, old, new):

        self.clear()        
        # Initialize the treeview of modes
        if old < new:
            self.modes_results_liststore = self.initialize_zeno_analysis_tab(self.get_modes_from_walker())

        pass



    def property_root_counter_value_change(self, model, old, new):

        self.clear()        
        # Initialize the treeview of modes
        if old < new:
            self.modes_results_liststore = self.initialize_zeno_analysis_tab(self.get_modes_from_walker())

        pass


    def get_modes_from_walker(self):
        """Gets the modes list from walker.
            @return   modes:  The modes list of Zeno Analysis.
	    @rtype    modes:  list of strings
        """
        modes = []
        list_impl = [] # List of fdir_componenst implementations

        for fdirs in self.model.slim_model.fdir_components: 
            if fdirs[2] is False:
                list_impl.append(str(fdirs[3].name))

        for name, impl, zeno_state_slim, zeno_state_smv in \
            self.model.get_zeno_states(self.model.slim_model.get_extended_root_impl()):

            if str(impl.name) not in list_impl:
                m1 = re.match("mode = mode:(\w+)", zeno_state_slim)
                assert m1
                modes.append([m1.group(1), name, zeno_state_smv])
       
        return sorted(modes, key=lambda lst: str.lower(str(lst[1])))


    # ---------------------------------------------------------------
    #        Signal and other callbacks handlers
    # ---------------------------------------------------------------

    def on_header_enabled_clicked(self, widget):
        """Enables/Disables all the rows of the Modes/Results Treeview.
            @param   widget:  The Enabled Treeview column.
            @type    widget:  gtk.TreeViewColumn
        """
        is_active = not self.toggle_button.get_active()

        idx = 0
        while idx < len(self.modes_results_liststore):

            self.modes_results_liststore[idx][0] = is_active

            # Sets the colour of enabled modes: BLACK
            if is_active:
                self.modes_results_liststore[idx][5] = "#000000"

            # Sets the colour of disabled modes: GREY
            else:
                self.modes_results_liststore[idx][5] = "#C0C0C0"
            idx += 1

        self.toggle_button.set_active(is_active)

        return


    def on_cell_toggled(self, widget, path):
        """Enables/Disables the row selected.
            @param   widget:  The cell of the toggle checkbox.
            @type    widget:  gtk.CellRendererToggle
            @param   path:  The path of the selected row of the treeview.
            @type    path:  int

        """
        # Sets the colour of enabled modes: BLACK
        if not self.modes_results_liststore[path][0]:
            self.modes_results_liststore[path][5] = "#000000"

        # Sets the colour of disabled modes: GREY
        else:
            self.modes_results_liststore[path][5] = "#C0C0C0"

        self.get_modes_results_liststore()[path][0] = \
            not self.get_modes_results_liststore()[path][0]

        return


    def on_results_selection_changed(self, selection):
        """Changes the results selected.
            @param   selection:  The selection of results.
            @type    selection:  gtk.TreeSelection
        """

        # Gets the tree_model and the path of the results selected
        tree_model, paths = selection.get_selected_rows()

        # If a mode is not selected, the previous result/message is cleared
        for mode_ls in self.modes_results_liststore:
            if mode_ls[0] is False:
                self.view.display_zeno_results(self.model.last_mode, None, None, "no trace", None)

        for zeno_mode in self.model.zeno_results_list:

            # Selects a row and gets the value of this row.
            value = ""
            for path in paths:
                tree_iter = tree_model.get_iter(path)
                value = tree_model.get_value(tree_iter,1)
                break

            # Checks if the modes are in row.

            if str(zeno_mode[1]) == value:

                # If the result is None, no traces are shown and the button detach is hidden.
                if zeno_mode[0].result is None:
                    self.view.display_zeno_results(self.model.last_mode, None, None, "no trace", None)
                    return

                # Else the traces are created.

                # Instantiate the various handlers

                tvm = TraceViewModel()
                trace = zeno_mode[0].result.trace
                tvm.add_traces(trace)
                tvm.set_active_trace(trace)
                trace_view = TraceView(windowed=False, filters=True)
                self.results_ctrl = TraceCtrl(tvm, trace_view)

                # synchronize scrollbars of trace windows

                swtop = trace_view['sw_top']

                adj_v = gtk.Adjustment(
                    value=0.00,
                    lower=0.00,
                    upper=100.00,
                    step_incr=1.00,
                    page_incr=10.00,
                    page_size=10.00)

                adj_h = gtk.Adjustment(
                  value=0.00,
                  lower=0.00,
                  upper=100.00,
                  step_incr=1.00,
                  page_incr=10.00,
                  page_size=10.00)

                swtop.set_hadjustment( adj_h )
                swtop.set_vadjustment( adj_v )

                # The trace and the button detach are shown.
                self.view.display_zeno_results(self.model.last_mode, None, None, "change trace", trace_view)

        return


    # ---------------------------------------------------------------
    #        Notifications from observable properties
    # ---------------------------------------------------------------

    def clear(self):
        """Clears the view and lists of modes, results"""
        self.view.clear()
        self.view.clear_treeview()
        self.modes_results_liststore.clear()
        self.model.zeno_results_list = []
        return


    def property_extended_model_value_change(self, model, old, new):
        # [MP] deadlock sub-model does not have a filtered_properties attribute
        pass


    def property_infinite_domain_value_change(self, model, old, new):
        """When the domain of the model is infinite, MathSAT and BMC
        must be used and the user will be given the option of the BCM bound."""

        if new: self.view['button_run_zeno_analysis'].set_sensitive(True)
        else: self.view['button_run_zeno_analysis'].set_sensitive(False)

        return


    def property_failures_value_change(self, model, old, new):
        """Sets the sensibility of button for enabling checking of
        extended model if there are failures in the slim model"""
        self.view.enable_model_extension(new > 0)
        return


    def register_adapters(self):
        StartStopActivityCtrl.register_adapters(self)
        self.adapt("extended_model", "cb_zeno_extended_model")
        self.adapt("sat_bound", "sb_zeno_sat_bound")
        self.adapt("k_zeno", "sb_zeno_k_zeno")
        return


    def property_slim_files_num_value_change(self, model, old, new):
        """This notification is called when the number of loaded
        slim files (into the observed slim model) is changed. It is
        used to enable/disable some features based on the presence
        of at least one slim model."""

        self.clear()

        if model.infinite_domain:

            if old < new:
                # Initialize the treeview of modes
                self.modes_results_liststore = self.initialize_zeno_analysis_tab(self.get_modes_from_walker())

        return


    def property_zeno_result_value_change(self, model, old, new):
        """When a new diagnosability analysis result is created, this method
            will be advised to display it.
        """

        # relieve old results

        if self.results_ctrl != None:
            self.results_ctrl.relieve_model(self.results_ctrl.model)
            self.results_ctrl = None
            pass

        if new is None: return

        assert old != new, "Toolchain is broken (DIAG)."

        if new.success in ("error", "unreachable", "stop", "selection"):
            self.view.display_zeno_results(self.model.last_mode, None, None, new.success, None)
            self.zeno_counter += 1
            return

        elif new.success == "unknown_not_found":
            self.view.display_zeno_results(self.model.last_mode, self.model.last_steps, None, new.success, None)
            self.zeno_counter += 1
            return

        elif new.success == "unknown_not_proved":
            self.view.display_zeno_results(self.model.last_mode, None, self.model.last_time, new.success, None)
            self.zeno_counter += 1
            return

        else:

            if self.model.zeno_result.result is None:
                self.zeno_counter += 1
                return

            # Instantiate the various handlers
            tvm = TraceViewModel()
            trace = self.model.zeno_result.result.trace
            tvm.add_traces(trace)
            tvm.set_active_trace(trace)
            trace_view = TraceView(windowed=False, filters=True)
            self.results_ctrl = TraceCtrl(tvm, trace_view)

            # synchronize scrollbars of trace windows

            swtop = trace_view['sw_top']

            adj_v = gtk.Adjustment(
                value=0.00,
                lower=0.00,
                upper=100.00,
                step_incr=1.00,
                page_incr=10.00,
                page_size=10.00)

            adj_h = gtk.Adjustment(
              value=0.00,
              lower=0.00,
              upper=100.00,
              step_incr=1.00,
              page_incr=10.00,
              page_size=10.00)

            swtop.set_hadjustment( adj_h )
            swtop.set_vadjustment( adj_v )

            # The trace and the button detach are shown.
            self.view.display_zeno_results(self.model.last_mode, None, None, new.success, trace_view)

        return


    pass #end of class
# ----------------------------------------------------------------------


# LB: the new class for TimeDivergence Ctrl

class TimeDivergenceCtrl (StartStopActivityCtrl):

    td_counter = 0

    # Observable properties
    __observables__ = ("td_counter", )


    def __init__(self, parent, model, view):
        StartStopActivityCtrl.__init__(self, model, view,
                                       self.run_time_divergence,
                                       model.stop_time_divergence)

        # we observe the slim model to reset previous results
        self.parent = parent

        # The result ctrl: it can contains the trace for unbounded results
        self.results_ctrl = None

        # It is the liststore of modes and results. It is composed of:
        # enabled (bool), clock (string), bound (str), time_scale unit (str),
        # message/result (string) and colour (string).
        self.clocks_results_liststore = gtk.ListStore(bool, str, str, str, str, str)

        self.balloon = gtk.Window()
        self.balloon.set_decorated(False)
        self.balloon.set_accept_focus(False)
        self.balloon.set_property("skip-taskbar-hint", True)
        self.txtview_in_slim_assoc = gtk.TextView()
        self.txtview_in_slim_assoc.set_property("wrap-mode",gtk.WRAP_WORD)
        self.txtbuf = gtk.TextBuffer()
        self.txtview_in_slim_assoc.set_buffer(self.txtbuf)
        self.balloon.add(self.txtview_in_slim_assoc)

        # we observe the number of failures into the slim model to
        # enable the corresponding button.
        self.model.slim_model.register_observer(self)
        return


    def register_view(self, view):

        StartStopActivityCtrl.register_view(self, view)

        view['win_td_detached'].connect("delete-event",
                                        self.on_window_td_detached_delete_event)

        self.trace_is_detached = False

        return


    def on_window_td_detached_delete_event(self, win, event):
        """When the delete button is clicked, the hbox results is re-attached."""
        self.view.attach()
        self.trace_is_detached = False
        return True


    def on_button_td_detach_clicked(self, button):
        """When the detach button is clicked, the hbox results is detached."""
        self.view.detach()
        self.trace_is_detached = True
        return


    def initialize_td_analysis_tab(self, clocks_results_ls):
        """Initializes the liststore of clocks, bounds and results in the relate Treeview.
            @param   clocks_results_ls:  The list of clocks from the slim_model loaded.
            @type    clocks_results_ls:  list of string

            @return   self.clocks_results_liststore:  The liststore of clocks and results.
            @rtype    self.clocks_results_liststore:  gtk.ListStore
        """

        for clock_res in clocks_results_ls:
    
            dsc = support.get_subcomponent_from_path(self.model.slim_model.get_extended_root_impl(), clock_res)
            assert isinstance(dsc, nominalbehaviour.DataSubcomponent)
            assert isinstance(dsc.data_type, nominalbehaviour.Clock)
            
            
            time_unit = dsc.data_type.time_unit

            if time_unit is None:   
                self.clocks_results_liststore.append([False, clock_res, "0.0", "", "", "#C0C0C0"])
                
            else:
                # Converts the time_unit from int to string
                import slim_compiler.src.dataStructure.expressions as expressions
                time_type = expressions.TimeUnitType.time_unit_type_to_str(time_unit)
                self.clocks_results_liststore.append([False, clock_res, "0.0", time_type, "", "#C0C0C0"])


        # The treeview of enabled, enum modes and results is created
        # and the columns will be filled

        tv = self.view['tv_td_clocks_results']
        tv.set_model(model=self.clocks_results_liststore)


        # Column Enabled
        renderer_toggle_enabled = gtk.CellRendererToggle()
        renderer_toggle_enabled.connect("toggled", self.on_cell_toggled)
        column_toggle_enabled = gtk.TreeViewColumn("Enabled", renderer_toggle_enabled, active=0)

        self.toggle_button = gtk.CheckButton("Enabled")
        self.toggle_button.show()
        column_toggle_enabled.set_widget(self.toggle_button)
        column_toggle_enabled.set_clickable(True)
        column_toggle_enabled.connect("clicked", self.on_header_enabled_clicked)
        column_toggle_enabled.set_resizable(True)
        
        tv.append_column(column_toggle_enabled)

        # Column Clocks
        renderer_text_clocks = gtk.CellRendererText()
        column_text_clocks = gtk.TreeViewColumn("Clocks", renderer_text_clocks, text=1, foreground=5)
        column_text_clocks.set_resizable(True)
        
        tv.append_column(column_text_clocks)


        # Column Bounds
        renderer_editable_bounds = gtk.CellRendererText()
        renderer_editable_bounds.set_property("editable", True)

        column_editable_bounds = gtk.TreeViewColumn("Bounds", renderer_editable_bounds, text=2, foreground=5)
        column_editable_bounds.set_resizable(True)
        
        tv.append_column(column_editable_bounds)

        renderer_editable_bounds.connect("edited",
                                         self.bounds_edited,
                                         self.clocks_results_liststore,
                                         'tv_td_clocks_results')

        # Column Time Scales Unit

        time_scale_units_liststore = gtk.ListStore(str)
        time_scale_units_liststore.append(["msec"])
        time_scale_units_liststore.append(["sec"])
        time_scale_units_liststore.append(["min"])
        time_scale_units_liststore.append(["hour"])
        time_scale_units_liststore.append(["day"])

        renderer_time_scale_unit = gtk.CellRendererCombo()
        renderer_time_scale_unit.set_property("editable", True)
        renderer_time_scale_unit.set_property("has-entry", False)
        renderer_time_scale_unit.set_property("model", time_scale_units_liststore)
        renderer_time_scale_unit.set_property("text-column", 0)
        renderer_time_scale_unit.connect("edited", self.on_combo_time_scale_changed, self.clocks_results_liststore)

        column_time_scale_unit = gtk.TreeViewColumn("Time Scale Unit", renderer_time_scale_unit, text=3, foreground=5)
        column_time_scale_unit.set_resizable(True)
        
        tv.append_column(column_time_scale_unit)


        # Column Results
        renderer_text_results = gtk.CellRendererText()
        column_text_results = gtk.TreeViewColumn("Results", renderer_text_results, text=4, foreground=5)
        column_text_results.set_resizable(True)
        
        tv.append_column(column_text_results)


        # To control selection of results
        sel = self.view['tv_td_clocks_results'].get_selection()
        sel.set_mode(gtk.SELECTION_MULTIPLE)
        sel.connect('changed', self.on_results_selection_changed)

        return self.clocks_results_liststore



    def get_clocks_results_liststore(self):
        """Gets the liststore of clocks and results.
            @return   self.clocks_results_liststore:  The liststore of clocks and results.
            @rtype    self.clocks_results_liststore:  gtk.ListStore
        """
        return self.clocks_results_liststore



    def run_time_divergence(self):
        """Runs Time Divergence."""

        # When the time divergence runs once again, the view and the Treeview are cleared
        if self.td_counter >= 1:
            self.view.clear()


        # Performs the Zeno Analysis
        self.model.analyze_time_divergence(self.clocks_results_liststore)


        # Populates the Treeview with the results of Time Divergence:
        # for each enabled clock there is a result/message created
        # at the end of Time Divergence

        clocks = self.get_clocks_from_walker()

        for message in self.view.get_messages_list():

            for clock in clocks:

                if message[0] == clock:

                    for clock_ls in self.clocks_results_liststore:
                        if clock_ls[0] is True and clock == clock_ls[1]:
                            clock_ls[4] = message[1]


        self.td_counter += 1

        return



    # ---------------------------------------------------------------
    #        Signal and other callbacks handlers
    # ---------------------------------------------------------------

    def on_header_enabled_clicked(self, widget):
        """Enables/Disables all the rows of the Modes/Results Treeview.
            @param   widget:  The Enabled Treeview column.
            @type    widget:  gtk.TreeViewColumn
        """
        is_active = not self.toggle_button.get_active()

        idx = 0
        while idx < len(self.clocks_results_liststore):
            self.clocks_results_liststore[idx][0] = is_active

            # Sets the colour of enabled modes: BLACK
            if is_active:
        
                # Bug #589: Commented the useless collection of internal clocks
#                clk = self.clocks_results_liststore[idx][1].split(".")
#                if clk[-1] == nominalbehaviour.TIME_DELAY_CLK_NAME:
#                    self.clocks_results_liststore[idx][5] = "#0000ff"
#                else:
#                    self.clocks_results_liststore[idx][5] = "#000000"
                
                self.clocks_results_liststore[idx][5] = "#000000"
                
            # Sets the colour of disabled modes: GREY
            else:
                self.clocks_results_liststore[idx][5] = "#C0C0C0"
            idx += 1

        self.toggle_button.set_active(is_active)

        return



    def on_cell_toggled(self, widget, path):
        """Enables/Disables the row selected.
            @param   widget:  The cell of the toggle checkbox.
            @type    widget:  gtk.CellRendererToggle
            @param   path:  The path of the selected row of the treeview.
            @type    path:  int

        """
        # Sets the colour of enabled modes: BLACK
        if not self.clocks_results_liststore[path][0]:
        
            # Bug #589: Commented the useless collection of internal clocks
#            clk = self.clocks_results_liststore[path][1].split(".")
#            if clk[-1] == nominalbehaviour.TIME_DELAY_CLK_NAME:
#                self.clocks_results_liststore[path][5] = "#0000ff"
#            else:
#                self.clocks_results_liststore[path][5] = "#000000"

            self.clocks_results_liststore[path][5] = "#000000"
            
        # Sets the colour of disabled modes: GREY
        else:
            self.clocks_results_liststore[path][5] = "#C0C0C0"

        self.get_clocks_results_liststore()[path][0] = \
            not self.get_clocks_results_liststore()[path][0]

        return



    def on_combo_time_scale_changed(self, widget, path, text, liststore):
        """Changes the new time_scale units of the clock selected in the Treeview.
            @param    widget: The widget of slim_expression.
            @type     widget: gtk.CellRenderedText
            @param    path: The path of the time-scale unit.
            @type     path: int
            @param    text: The text of the time-scale unit.
            @type     text: string
            @param    liststore: The liststore of the time-scale unit.
            @type     liststore: gtk.ListStore
        """
        liststore[path][3] = text
        return



    def on_results_selection_changed(self, selection):
        """Changes the results selected.
            @param   selection:  The selection of results.
            @type    selection:  gtk.TreeSelection
        """

        # Gets the tree_model and the path of the results selected
        tree_model, paths = selection.get_selected_rows()

        # If a mode is not selected, the previous result/message is cleared
        for clock_ls in self.clocks_results_liststore:

            if clock_ls[0] is False:
                self.view.display_td_results(self.model.last_clock, None, "no trace", None)

        for td_clock in self.model.td_results_list:

            # Selects a row and gets the value of this row.
            value = ""
            for path in paths:
                tree_iter = tree_model.get_iter(path)
                value = tree_model.get_value(tree_iter,1)
                break

            # Checks if the modes are in row.

            if str(td_clock[1]) == value:

                # If the result is None, no traces are shown and the button detach is hidden.
                if td_clock[0].result is None:
                    self.view.display_td_results(self.model.last_clock, None, "no trace", None)
                    return

                # Else the traces are created.

                # Instantiate the various handlers

                tvm = TraceViewModel()
                trace = td_clock[0].result.trace
                tvm.add_traces(trace)
                tvm.set_active_trace(trace)
                trace_view = TraceView(windowed=False, filters=True)
                self.results_ctrl = TraceCtrl(tvm, trace_view)

                # synchronize scrollbars of trace windows

                swtop = trace_view['sw_top']

                adj_v = gtk.Adjustment(
                    value=0.00,
                    lower=0.00,
                    upper=100.00,
                    step_incr=1.00,
                    page_incr=10.00,
                    page_size=10.00)

                adj_h = gtk.Adjustment(
                  value=0.00,
                  lower=0.00,
                  upper=100.00,
                  step_incr=1.00,
                  page_incr=10.00,
                  page_size=10.00)

                swtop.set_hadjustment( adj_h )
                swtop.set_vadjustment( adj_v )

                # The trace and the button detach are shown.
                self.view.display_td_results(self.model.last_clock, None, "change trace", trace_view)

        return


    def bounds_edited(self, widget, path, text, liststore, tv_name):
        """Function called when the user inserts or updates a bound
            associated to a clock.

            @param    widget: The widget of bound.
            @type     widget: gtk.CellRenderedText
            @param    path: The path of the bound.
            @type     path: int
            @param    text: The text of the bound.
            @type     text: string
            @param    liststore: The liststore of the bound.
            @type     liststore: gtk.ListStore
            @param    tv_name: The name of the window (monitored or not monitored)
            @type     tv_name: string
        """

        # If there is not entry: ERROR
        if len(text) == 0:
            self.show_bound_error("The value length must be at least 1", tv_name, path)

        import re
        
        # The entry could be:
        # - an integer number with at most 8 digits
        # - a real number with at most 8 digits and at most 2 decimal digits
        regex = re.compile('^(?<![\d.])(\d{1,8}|\d{0,8}\.\d{1,2})?(?![\d.])$')

        if regex.match(text):

            # If the entry is < 0: ERROR
            if float(text) < 0:
                self.show_bound_error("The value must be >= 0", tv_name, path)

            # If the entry is > -1: OK
            else:
                liststore[path][2] = text
                self.balloon.hide_all()

        # If the entry is a string
        else:
            error_message = "The value must be:\n - an integer number with at most 8 digits\n" \
                            " - a real number with at most 8 digits and at most 2 decimal digits."
            self.show_bound_error(error_message, tv_name, path)

        return



    def show_bound_error(self, text, tv_name, path):
        """Function which shows the error derived from the insertion of a wrong slim expression.
            @param    text: The text of the wrong slim_expression.
            @type     text: string
            @param    tv_name: The name of the window (monitored or not monitored)
            @type     tv_name: string
            @param    path: The path of the wrong slim_expression.
            @type     path: int
            """

        window_x, window_y = self.view[tv_name].get_parent_window().get_origin()
        bottom_left_x = self.view[tv_name].allocation.x + window_x
        bottom_left_y = self.view[tv_name].allocation.y + window_y + 50 + (int(path)*22)

        self.balloon.set_position(gtk.WIN_POS_MOUSE)
        self.txtview_in_slim_assoc.modify_base(gtk.STATE_NORMAL, gtk.gdk.color_parse("indian red"))
        self.txtbuf.set_text(text)
        self.balloon.resize(400, 30)
        self.balloon.show_all()
        self.balloon.move(bottom_left_x, bottom_left_y)

        return



    def property_fdir_comp_counter_value_change(self, model, old, new):

        self.clear()        
        # Initialize the treeview of modes
        if old < new:
            self.clocks_results_liststore = self.initialize_td_analysis_tab(self.get_clocks_from_walker())

        pass
    

    def property_root_counter_value_change(self, model, old, new):

        self.clear()        
        # Initialize the treeview of modes
        if old < new:
            self.clocks_results_liststore = self.initialize_td_analysis_tab(self.get_clocks_from_walker())

        pass



    def get_clocks_from_walker(self):
        """Gets the clocks list from walker.
            @return   clocks:  The clocks list of Time Divergence
	        @rtype    clocks:  list of strings
        """

        list_impl = [] # List of fdir_componenst implementations

        for fdirs in self.model.slim_model.fdir_components: 
            if fdirs[2] is False:
                list_impl.append(str(fdirs[3].name))


        # Gets the clocks from the slim_model
        from slim_compiler.src.smv.walkers import ClockCollectorWalker

        walker = ClockCollectorWalker()
        walker.walk(self.model.slim_model.get_extended_root_impl())

        clocks_old = [clock[0] for clock in walker.get_found_clocks()]
        
        clocks = []
        for clock in walker.get_found_clocks():

            if str(clock[3].name) not in list_impl:
                clocks.append(clock[0])

        # Bug #589: Commented the useless collection of internal clocks
#        internal_clks_ordered = []
#        clks_ordered = []

#        for clk in clocks:
#            clock = clk.split(".")
#            if clock[-1] == nominalbehaviour.TIME_DELAY_CLK_NAME:
#                internal_clks_ordered.append(clk)
#            else: clks_ordered.append(clk)

#        clocks = clks_ordered + internal_clks_ordered

        return clocks


    # ---------------------------------------------------------------
    #        Notifications from observable properties
    # ---------------------------------------------------------------


    def clear(self):
        """Clears the view and lists of modes, results"""
        self.view.clear()
        self.view.clear_treeview()
        self.clocks_results_liststore.clear()
        self.model.td_results_list = []
        return


    def property_extended_model_value_change(self, model, old, new):
        # [MP] deadlock sub-model does not have a filtered_properties attribute
        pass


    def property_infinite_domain_value_change(self, model, old, new):
        """When the domain of the model is infinite, MathSAT and BMC
        must be used and the user will be given the option of the BCM bound."""

        if new: self.view['button_run_td'].set_sensitive(True)
        else: self.view['button_run_td'].set_sensitive(False)

        return


    def property_failures_value_change(self, model, old, new):
        """Sets the sensibility of button for enabling checking of
        extended model if there are failures in the slim model"""
        self.view.enable_model_extension(new > 0)
        return


    def register_adapters(self):
        StartStopActivityCtrl.register_adapters(self)
        self.adapt("extended_model", "cb_td_extended_model")
        self.adapt("sat_bound", "sb_td_sat_bound")
        return


    def property_slim_files_num_value_change(self, model, old, new):
        """This notification is called when the number of loaded
        slim files (into the observed slim model) is changed. It is
        used to enable/disable some features based on the presence
        of at least one slim model."""

        self.clear()

        if model.infinite_domain:

            if old < new:
                # Initialize the treeview of modes
                self.clocks_results_liststore = self.initialize_td_analysis_tab(self.get_clocks_from_walker())

        return


    def property_td_result_value_change(self, model, old, new):
        """When a new time divergence result is created, this method
            will be advised to display it.
        """

        # relieve old results

        if self.results_ctrl != None:
            self.results_ctrl.relieve_model(self.results_ctrl.model)
            self.results_ctrl = None
            pass

        if new is None: return

        assert old != new, "Toolchain is broken (DIAG)."

        if new.success in ("error", "unknown", "stop", "selection"):
            self.view.display_td_results(self.model.last_clock, None, new.success, None)
            self.td_counter += 1
            return

        elif new.success == "bounded":
            self.view.display_td_results(self.model.last_clock, self.model.last_time, new.success, None)
            self.td_counter += 1
            return

        else:

            if self.model.td_result.result is None:
                self.td_counter += 1
                return


            # Instantiate the various handlers
            tvm = TraceViewModel()
            trace = self.model.td_result.result.trace
            tvm.add_traces(trace)
            tvm.set_active_trace(trace)
            trace_view = TraceView(windowed=False, filters=True)
            self.results_ctrl = TraceCtrl(tvm, trace_view)

            # synchronize scrollbars of trace windows

            swtop = trace_view['sw_top']

            adj_v = gtk.Adjustment(
                value=0.00,
                lower=0.00,
                upper=100.00,
                step_incr=1.00,
                page_incr=10.00,
                page_size=10.00)

            adj_h = gtk.Adjustment(
              value=0.00,
              lower=0.00,
              upper=100.00,
              step_incr=1.00,
              page_incr=10.00,
              page_size=10.00)

            swtop.set_hadjustment( adj_h )
            swtop.set_vadjustment( adj_v )

            # The trace and the button detach are shown.
            self.view.display_td_results(self.model.last_clock, None, new.success, trace_view)

        return


    pass #end of class
# ----------------------------------------------------------------------


class ResultsCtrl (Controller):
    """This class handles the panels (class ResultsView) showing
    results of model checking, simulation, etc. That panel has some
    features that are handled by this controller, like the
    capability of being embedded or raised to its own window."""

    def register_view(self, view):
        view['window_detached'].connect("delete-event",
                                        self.on_windows_detached_delete_event)

        view['window_filters_detached'].connect("delete-event",
                                                self.on_windows_filters_detached_delete_event)

        self.trace_is_detached = False
        self.filters_is_detached = False

        return

    # ---------------------------------------------------------------
    #        Signal handlers
    # ---------------------------------------------------------------
    def on_windows_detached_delete_event(self, win, event):
        self.view.attach()
        self.trace_is_detached = False
        return True

    def on_button_detach_clicked(self, button):
        if self.filters_is_detached: #reattach it
            self.on_windows_filters_detached_delete_event(self.view['window_filters_detached'], None)

        self.view.detach()
        self.trace_is_detached = True
        return

    def on_button_attach_clicked(self, button):
        self.view.attach()
        self.trace_is_detached = False
        return

    def on_button_sim_restart_clicked(self, btn):
        assert isinstance( self.model, ModelSimulationModel ), "The model of the result ctrl must be a modelsimulationmodel!"
        self.model.prev_trace_fname = None # forget the previous value
        self.model.result = None
        self.model.pick_unset() # if there is a pick set, forget it
        self.view.show_no_results() # show the "no results" tab. Necessary in the case in which result is not changed

        # forget about the previous deadlock trace, if any
        if self.dtc:
            self.dtc.relieve_model(self.dtc.model)
            self.dtc = None
            pass

        return

    #filters
    def on_windows_filters_detached_delete_event(self, win, event):
        w = self.view['window_filters_detached']
        child = w.get_child()
        assert child, "No child in this window!"

        w.remove(child)
        self.view['vbox_res'].pack_end(child,expand=False,padding=6)
        w.hide()

        self.filters_is_detached = False
        return True


    pass # end of class
# ----------------------------------------------------------------------
