# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================


from gtkmvc import Controller
from gtkmvc.adapters import Adapter
import gtk
import slim_compiler.src.model_extension.fault_injection as fault_injection
from src.views.balloon import Balloon
from src.views.fault_injection import FaultInjectionView
from delayedcall import DelayedCall


class FaultInjectionCtrl (Controller):
    """This is the controller for the dialog used to prompt the
    user for the fault injection when requested. The
    model used by this controller are the subclasses of FaultInjection"""


    def __init__(self, model, view):
        Controller.__init__(self, model, view)
        model.fill_error_implementations()
        
        self.effect_valid = False
        
        return


    def register_view(self, view):
        # error implementions entry
        box = view['box_err_impl']
        box.set_model(self.model.error_implementations)
        box.set_text_column(0)
        entry = view['box_err_impl'].child
        compl = gtk.EntryCompletion()
        compl.set_model(self.model.error_implementations)
        compl.set_text_column(0)
        entry.set_completion(compl)
        entry.connect('activate', self.on_change_entry_err_impl)
        box.connect('changed', self.on_change_entry_err_impl)
        # error states entry
        box = view['combo_box_err_state']
        box.set_model(self.model.error_states)
        cell = gtk.CellRendererText()
        box.pack_start(cell, True)
        box.add_attribute(cell, 'text', 0)        
        box.connect('changed', self.on_change_box_err_states)
        # instances entry
        box = view['box_instance']
        box.set_model(self.model.nominal_instances)
        box.set_text_column(0)
        entry = view['box_instance'].child
        compl = gtk.EntryCompletion()
        compl.set_model(self.model.nominal_instances)
        compl.set_text_column(0)
        entry.set_completion(compl)
        entry.connect('changed', self.on_change_entry_instance) 
        # data combobox
        box = view['combo_box_data']
        box.set_model(self.model.datas)
        cell = gtk.CellRendererText()
        box.pack_start(cell, True)
        box.add_attribute(cell, 'text', 0)        
        box.connect('changed', self.on_change_box_data)
        # effect entry
        entry = view['entry_effect']
        entry.connect('changed', self.on_change_entry_effect)
        entry.connect('activate', self.on_activate_entry_effect) 
        
        # this window is used to display errors in a tooltip-like fashion
        self.balloon = Balloon(self.view[FaultInjectionView.top])
        self.delay = DelayedCall(1000, self.balloon.change_color_and_balloon)
        return
        


    def check_and_set_visibility_of_widgets(self):
        # check valid error impl
        entry = self.view['box_err_impl'].child
        m = self.model.listStore_as_dict(self.model.error_implementations)
        err_impl_valid = entry.get_text() in m
        # check valid error state
        box = self.view['combo_box_err_state']
        err_state_valid = box.get_active_text() != None
        # check valid instance
        entry = self.view['box_instance'].child
        m = self.model.listStore_as_dict(self.model.nominal_instances)
        instance_valid = entry.get_text() in m
        # check valid data 
        box = self.view['combo_box_data']
        assignment = box.get_active_text()
        data_valid = box.get_active_text() != None
        # set sensitivity of the input elements
        self.view['combo_box_err_state'].set_sensitive(err_impl_valid)
        self.view['box_instance'].set_sensitive(err_impl_valid and err_state_valid)
        self.view['combo_box_data'].set_sensitive(err_impl_valid and err_state_valid and instance_valid)
        
        if err_impl_valid and err_state_valid and instance_valid and data_valid:
            self.view['entry_effect'].set_sensitive(True)
        else:
            self.view['entry_effect'].set_sensitive(False)
            self.balloon.hide_all()
        # set sensitivity of inject button
        effect_valid = False
        errors = []
        if err_impl_valid and err_state_valid and instance_valid and data_valid:
            assignment = self.view['combo_box_data'].get_active_text() + " := " + self.view['entry_effect'].get_text()
            errors = fault_injection.get_errors_in_fault_effect(
                            assignment,
                            self.model.nominal_instance.implementation)            
            effect_valid = errors == []
        btn_inject = self.view['button_fi_inject']    
        btn_inject.set_sensitive(effect_valid)   
        self.effect_valid = effect_valid     
        # set the color of the effect entry
        if effect_valid:
            self.delay.immediately_change_color_and_balloon(
                                               self.view['entry_effect'], None)
        else:
            if self.view['entry_effect'].get_text() != "":
                self.delay.change_color_and_balloon(self.view['entry_effect'],
                                            [str(error) for error in errors])
            else:
                #if there is no input, make entry white and hide the balloon
                self.delay.reset_delay()
                self.balloon.hide_all()
                self.view['entry_effect'].modify_base(gtk.STATE_NORMAL, gtk.gdk.color_parse("white"))
        return (err_impl_valid, err_state_valid, instance_valid, data_valid, effect_valid)


    # signals 
    def on_change_entry_err_impl(self, editable):
        err_impl_valid, _, _, _, _ = self.check_and_set_visibility_of_widgets()
        if err_impl_valid: 
            m = self.model.listStore_as_dict(self.model.error_implementations)
            self.model.error_implementation = m[editable.get_active_text()]
            self.model.fill_error_states(self.model.error_implementation)
            self.model.fill_nominal_instances(self.model.error_implementation)
            # could be that the currently chosen error implementation is invalid with the 
            # currently entered nominal instance, rerun checking and setting of visibilities
            self.check_and_set_visibility_of_widgets()
        return


    def on_change_box_err_states(self, combobox):
        _, err_state_valid, instance_valid, _, _ = self.check_and_set_visibility_of_widgets()
        if err_state_valid: 
            self.model.error_state = combobox.get_active_text()
            # could be that the currently chosen error state is invalid with the
            # currently entered data port, rerun checking and setting of visibilities
            if instance_valid: 
                self.model.fill_data(self.model.nominal_instance, self.model.error_state)
                self.check_and_set_visibility_of_widgets()            
        return


    def on_change_entry_instance(self, editable):
        _, _, instance_valid, _, _ = self.check_and_set_visibility_of_widgets()
        if instance_valid:
            m = self.model.listStore_as_dict(self.model.nominal_instances)
            self.model.nominal_instance = m[editable.get_text()]
            self.model.fill_data(self.model.nominal_instance, self.model.error_state)
        return
    
    
    def on_change_box_data(self, combobox): 
        _, _, _, data_valid, _ = self.check_and_set_visibility_of_widgets()        
        if data_valid: self.model.data = combobox.get_active_text()
        return        


    def on_change_entry_effect(self, editable):
        _, _, _, _, effect_valid = self.check_and_set_visibility_of_widgets()            
        if effect_valid: self.model.effect = editable.get_text()
        return

    def on_activate_entry_effect(self, editable):
        if self.effect_valid:
            # Simulate clicking of fault injection
            self.on_button_fi_inject_clicked(None)
        return
        
        
    def on_button_fi_inject_clicked(self, button):
        self.view['fault_injection'].response(gtk.RESPONSE_APPLY)
        self.view['fault_injection'].destroy()
        return                
        
    
    def on_button_fi_discard_clicked(self, button):
        self.view['fault_injection'].response(gtk.RESPONSE_CANCEL)
        self.view['fault_injection'].destroy()            
        return
        
                
    pass
