# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================


import gtk
import gobject
from gtkmvc import Controller

import property_management.src.constants as prop_const
from utils.exceptions import NusmvControlledException
import property_management.src.api as prop_api

from src.models.filtered_properties import FilteredPropertiesModel
from src.ctrls.activity import ActivityCtrl, StartStopActivityCtrl
from src.ctrls.correctness import ModelCheckingCtrl, ResultsCtrl
from src.ctrls.performability import *
from utils import fsap_process
from src.ctrls.trace import TraceCtrl, FilterBarCtrl
from src.views.trace import TraceView, FilterBarView
from src.models.trace import TraceViewModel
from xml.sax.saxutils import escape

class FdirCtrl (ActivityCtrl):
    """Controller for the Correctness activity. Two sub-controllers
    control model checking and model simulation panels."""
    def __init__(self, model, view):
        ActivityCtrl.__init__(self, 
                              FdirSubCtrl(self, model, view.subview), 
                              model, view)

        # Sub-controllesrs
        self.fda = FdirFdaCtrl(model.fda, view.fda)
        self.diag = FdirDiagCtrl(model.diag, view.diag)
        self.fia = FdirFiaCtrl(model.fia, view.fia)      
        self.fra = FdirFraCtrl(model.fra, view.fra)
        self.fca = FdirFcaCtrl(model.fca, view.fca)

        # this is used to control the type of selection of properties
        self.mode = None
        return

    def register_view(self, view):
        ActivityCtrl.register_view(self, view)
        
        self.set_mode("fda")

        tv = view['tv_properties']
        tv.set_model(self.model.filtered_props.get_properties_list())

        # selection (does not matter for pfda and pfra since it is disabled)
        def __rend_select(col, rend, model, iter):
            col_num = FilteredPropertiesModel.SELECT_SINGLE \
                if self.mode in ("fda", "fia", "fra", "diag", "fca") \
                else FilteredPropertiesModel.SELECT_MULTI
            rend.set_property("active", model[iter][col_num])
            return

        rend1 = gtk.CellRendererToggle()
        rend1.set_property("activatable", True)
        rend1.connect("toggled", self.on_prop_toggled)
        col = gtk.TreeViewColumn('Name', rend1)
        col.set_cell_data_func(rend1, __rend_select)
        # and name
        rend2 = gtk.CellRendererText()
        col.pack_start(rend2)
        col.set_cell_data_func(rend2, lambda c,r,m,i: r.set_property('text', 
                                               m[i][FilteredPropertiesModel.PROPERTY].name))        
        tv.append_column(col)

        rend = gtk.CellRendererText()
        col = gtk.TreeViewColumn('Formula', rend)
        col.set_cell_data_func(rend, lambda c,r,m,i: r.set_property('text', 
                                               m[i][FilteredPropertiesModel.PROPERTY].description))
        tv.append_column(col)

        # no selection for this tv:
        tv.get_selection().set_mode(gtk.SELECTION_NONE)
        return

    def set_mode(self, mode):
        """Set current modality"""
        if mode == self.mode: return
        assert mode in ("fda", "fia", "fra", "diag", "fca")
        self.mode = mode
        self.view['tv_properties'].queue_draw()

        self.view['tv_properties'].set_sensitive(True)
        if mode in ("fda", "diag"): self.model.filtered_props.set_filtering(None) # only propositionals
        elif mode in ("fia", "fca"): self.view['tv_properties'].set_sensitive(False)
        elif mode in ("fra",): self.model.filtered_props.set_filtering((prop_const.Logic.CTL, prop_const.Logic.LTL,))
        else: self.model.filtered_props.restore_logic_filtering() # FIXME for pfda

    def drag_data_get_data(self, treeview, context, selection, target_id,
                           etime):
        treeselection = treeview.get_selection()
        model, iter = treeselection.get_selected()
        data = model.get_value(0, iter)
        selection.set(selection.target, 8, data)

        return

    # ---------------------------------------------------------------
    #        Signal handlers
    # ---------------------------------------------------------------    
    def on_prop_toggled(self, rend, path):
        child_path = self.model.filtered_props.get_properties_list().convert_path_to_child_path(path)
        self.model.props_model.toggle_property(child_path, multi=(self.mode not in ("fda", "fra", "diag", "fca")))
        return

    pass # end of class
# ----------------------------------------------------------------------


class FdirSubCtrl (Controller):
    """Sub-controller for the fdir activity sub-view"""
    def __init__(self, parent, model, view):
        Controller.__init__(self, model, view)

        # chief controller
        self.parent = parent

        # sub-controllers (all share the same view)
        self.fda = FdirFdaCtrl(model.fda, view.fda)
        self.diag = FdirDiagCtrl(model.diag, view.diag)
        self.fia = FdirFiaCtrl(model.fia, view.fia)      
        self.fra = FdirFraCtrl(model.fra, view.fra)
        self.fca = FdirFcaCtrl(model.fca, view.fca)
        return

    # ---------------------------------------------------------------
    #        Signal handlers
    # ---------------------------------------------------------------    
    def on_nb_fdir_switch_page(self, nb, page, num):
        """Switch page depending on the fdir tab:
           0 -> Fault Detection Analysis
           1 -> Fault Isolation Analysis
           2 -> Fault Recovery Analysis
           3 -> Diagnosability Analysis
           4 -> Fault Coverage Analysis
        """
 
        assert num >= 0 and num < 5
        self.parent.set_mode("fda fia fra diag fca".split()[num])
        return

    # ---------------------------------------------------------------
    #        Notifications from observable properties
    # ---------------------------------------------------------------    

    pass # end of class
# ----------------------------------------------------------------------


class FdirFdaCtrl (StartStopActivityCtrl):
    """This is a sub-controller for handling FDA activities"""
        
    def __init__(self, model, view):
        StartStopActivityCtrl.__init__(self, model, view, 
                                       self.run_fda, model.stop_fsap)

        # this controller is interested in observing if the slim
        # model has infinite domain
        self.observe_model(model.slim_model)

        # we also observe the number of selected properties in the
        # properties model, to enable fault detection button
        # accordingly
        self.model.filtered_props.register_observer(self)
        return

    def register_view(self, view): 
        StartStopActivityCtrl.register_view(self, view)

        tv = view['tv_fda_list']
        tv.set_model(self.model.fda_list)

        rend = gtk.CellRendererText()
        col = gtk.TreeViewColumn('Alarm', rend, text=0)
        tv.append_column(col)  
        
        # We set the Detection Delay as 0.0 (by default)
        self.view['entry_detect_delay'].set_text("0.0")   
        
        # The event related the Detection Delay                           
        self.view['entry_detect_delay'].connect('changed', self.on_entry_detect_delay_changed)
        self.view['entry_detect_delay'].set_sensitive(False) # Finite delay as default

        self.view['combobox_detection_delay'].set_active(0)
        
        ## The time unit combobox must be selectable only in case the slim model uses time scaling
        self.view['combobox_detection_delay'].set_sensitive(self.model.slim_model.compile_info.uses_time_scaling)

        return

    def register_adapters(self):
        self.adapt("sat_bound", "fda_sat_bound")
        return

    def run_fda(self):
        """Makes the model generate fda and then shows the
        resulting generated list of alarms"""

        # If the model is discrete, we do not have a delay
        if not self.model.slim_model.infinite_domain:
            delay = None
            self.model.detect_delay = None
        elif self.view['radiobutton_fda_finite_delay'].get_active():
            delay = -1.0
            self.model.detect_delay = -1.0
        else:
            if self.view['combobox_detection_delay'].get_sensitive():
                delay = self.model.get_delay(self.view['entry_detect_delay'].get_text(),
                                             self.view['combobox_detection_delay'].get_active_text())
            else:
                delay = self.model.get_delay(self.view['entry_detect_delay'].get_text(), None)
            self.model.detect_delay = delay
        fname = self.model.generate_fda(detection_delay=delay)
        if fname is None: 
            self.view['label_fda_info'].set_markup(
                "There are <span foreground='red'>no alarms</span> in model. Fault Detection Analysis is <b>meaningless</b>")
            return
        
        res = self.model.load_fda(fname)
        if not res:
            self.view['label_fda_info'].set_markup("The list of alarms with the given property is <b>empty</b>")

        else: 
            self.view['label_fda_info'].set_markup("The given property produced a list alarms")

            row = [ row for row in self.model.filtered_props \
                        if row[FilteredPropertiesModel.SELECT_SINGLE] ][0]
            self.model.store_fda(row)
        return

    def property_single_selected_value_change(self, model, old, new):
        self.view.enable_fda(new)
        if not new: return

        rows = filter( 
            lambda row: row [ FilteredPropertiesModel.SELECT_SINGLE ],
            self.model.filtered_props )

        assert 1 == len(rows), "Exactly one row should be selected."
                      
        # restore previous results selecting prop
        fda = rows[0] [FilteredPropertiesModel.FDA_LIST]
        self.model.restore_fda(fda)
        
        return

    # observable properties
    def property_infinite_domain_value_change(self, model, old, new):
        self.view.enable_sat(not new)
        return
        
        
        
    # LB: this is the event triggered by the Detection Delay
    def on_entry_detect_delay_changed(self, entry):
        """Changes the new Detection Delay.
            @param   entry:  The new value of the Detection Delay.
            @type    entry:  gtk.Entry
        """
        
        # If the number is real or integer
        is_real = False
        
        # If there is not entry: ERROR
        if len(entry.get_text()) == 0:
            self.view.change_color_and_balloon(entry, ["The value length must be at least 1"])
            # At the moment the RUN CHECK Button is always enabled
            
        import re
        
        # If the model is Hybrid
        if self.model.slim_model.infinite_domain:
            regex = re.compile('^[0-9.]+$') # The entry must be a real number 
            is_real = True
        else:
            regex = re.compile('^[0-9]+$') # The entry must be a integer number 
                
                
        if regex.match(entry.get_text()):

            # If the entry is a real number
            if is_real: 
        
                # If the entry is < 0: ERROR
                if float(entry.get_text()) < 0:   
                    self.view.change_color_and_balloon(entry, ["The value must be >= 0.0"])
                    # At the moment the RUN CHECK Button is always enabled
                    
                
                else: # If the entry is > -1: OK
                    
                    # Set the max length of entry as 2 decimal (eg. 2.34 or 6.78)
                    float_entry = "%.2f" % float(entry.get_text())
                    entry.set_max_length(len(str(float_entry)))
                        
                    self.view.change_color_and_balloon(entry, None)
                    # At the moment the RUN CHECK Button is always enabled
                        
                    self.model.detect_delay = float(entry.get_text())
                    
            else: # If the entry is an integer number
             
                # If the entry is a < 0: ERROR
                if int(entry.get_text()) < 0:   
                    self.view.change_color_and_balloon(entry, ["The value must be >= 0.0"])
                    # At the moment the RUN CHECK Button is always enabled
                    
                
                else: # If the entry is > -1: OK
                    
                    self.view.change_color_and_balloon(entry, None)
                    # At the moment the RUN CHECK Button is always enabled
                        
                    self.model.detect_delay = int(entry.get_text())
                  
            
        else: # If the entry is a string 
                
            self.view.change_color_and_balloon(entry, ["The value must be >= 0.0"])
            # At the moment the RUN CHECK Button is always enabled

        # Set the max lenght of entry as 10 characters
        entry.set_max_length(10)
        
        return 


    def on_radiobutton_fda_finite_delay_toggled(self, widget):
        """Function called when the user selects FINITE delay"""
        self.view['entry_detect_delay'].set_editable(not widget.get_active())
        self.view['entry_detect_delay'].set_sensitive(not widget.get_active())
        self.view['combobox_detection_delay'].set_sensitive(self.model.slim_model.compile_info.uses_time_scaling and not widget.get_active())
        return


    pass # end of class
# ----------------------------------------------------------------------


class FdirDiagCtrl (StartStopActivityCtrl):
    """This is a sub-controller for handling DIAG activities"""

    def __init__(self, model, view):
        StartStopActivityCtrl.__init__(self, model, view, 
                                       self.run_diag, model.stop_fsap)
        self.results_ctrl_A = None
        self.results_ctrl_B = None
        # for observing infinite_domain (for showing/hiding
        # the SAT bound selection box) and for observing if
        # there are any changes in the slim model (for "resetting"
        # diagnosability analysis).
        self.observe_model(model.slim_model)

        # we observe the property panel for enabling/disabling diag.analysis
        self.model.filtered_props.register_observer(self)
        return



    def register_view(self, view): 
        StartStopActivityCtrl.register_view(self, view)

        # bind restrictions view and store
        tv_res = view['tv_restrictions']
        tv_res.set_model(self.model.restrictions_store)
        tv_res.set_reorderable(True)
        tv_res.get_selection().set_mode(gtk.SELECTION_SINGLE)
        tv_res.append_column(self.model.restriction_col)

        # whenever there's a change in the property manager (beside addition
        # of new properties), the status of diagnosability analysis is reset.
        slim_model_properties = self.model.filtered_props.props_model.get_properties_tree()
        slim_model_properties.connect("row_deleted", lambda tm, pa: self.reset_diag_analysis())
        
        # We set the Detection Delay as 0 (by default) 
        self.view['entry_diag_delay'].set_text("0.0")   
        
        # The event related the Diagnosability Delay                          
        self.view['entry_diag_delay'].connect('changed', self.on_entry_diag_delay_changed)

        # Set default active time unit
        self.view['combobox_diag_time_scale'].set_active(0)
        ## The time unit combobox must be selectable only in case the slim model uses time scaling
        self.view['combobox_diag_time_scale'].set_sensitive(self.model.slim_model.compile_info.uses_time_scaling)

        # At first, finite delay is active, hence the delay text field must not be selctable
        self.view['entry_diag_delay'].set_editable(False)
        self.view['entry_diag_delay'].set_sensitive(False)
              
        return



    def register_adapters(self):
        self.adapt("sat_bound")
        return



    def run_diag(self):
        """Makes the model run the diagnosability test on
        the selected property and restrictions, and displays
        the results."""

        if not self.model.check_observations():
            self.view['label_diag_info'].set_markup(
                "There are <span foreground='red'>no observables</span> in the model. Diagnosability Analysis is <b>meaningless</b>")
            return

        if self.view.trace_window:
            self.view.on_trace_window_delete(None, None)

        # If the model is discrete we let the delay to its default value of -1.0;
        # if the model is continuous and FINITE is enabled, the delay is -1.0
        if not self.model.slim_model.infinite_domain or self.view['radiobutton_diag_finite_delay'].get_active():
            self.model.diag_delay = -1.0
        else:
            if self.view['combobox_diag_time_scale'].get_sensitive():
                self.model.set_delay(self.view['entry_diag_delay'].get_text(),
                                     self.view['combobox_diag_time_scale'].get_active_text())
            else:
                self.model.set_delay(self.view['entry_diag_delay'].get_text(), None)
        self.model.go_diag()
        return 


    def reset_diag_analysis(self):
        """Hides trace results and wipes path restrictions and context."""
        self.model.restrictions_store.clear()
        self.on_button_diagctx_del_clicked(None)
        return



    def get_selected_property(self):
        properties = [ row[FilteredPropertiesModel.PROPERTY].property \
                           for row in self.model.filtered_props \
                           if row[FilteredPropertiesModel.SELECT_SINGLE] ]

        if len(properties) == 0: return None # nothing is selected

        assert len(properties) == 1, "Toolset broken. Diagnosability Analysis can \
                                      accept only single property selections."

        return properties[0]



    # LB: this is the event triggered by the Detection Delay
    def on_entry_diag_delay_changed(self, entry):
        """Changes the new Diagnosability Delay.
            @param   entry:  The new value of the Diagnosability Delay.
            @type    entry:  gtk.Entry
        """
        
        # If the number is real or integer
        is_real = False
        
        # If there is not entry: ERROR
        if len(entry.get_text()) == 0:
            self.view.change_color_and_balloon(entry, ["The value length must be at least 1"])
            # At the moment the RUN CHECK Button is always enabled
            
        import re
        
        # If the model is Hybrid
        if self.model.slim_model.infinite_domain:
            regex = re.compile('^[0-9.]+$') # The entry must be a real number 
            is_real = True
        else:
            regex = re.compile('^[0-9]+$') # The entry must be a integer number 
                
                
        if regex.match(entry.get_text()):

            # If the entry is a real number
            if is_real: 
        
                # If the entry is < 0: ERROR
                if float(entry.get_text()) < 0.0:   
                    self.view.change_color_and_balloon(entry, ["The value must be >= 0.0"])
                    # At the moment the RUN CHECK Button is always enabled
                    
                
                else: # If the entry is >= 0.0: OK
                    
                    # Set the max length of entry as 2 decimal (eg. 2.34 or 6.78)
                    float_entry = "%.2f" % float(entry.get_text())
                    entry.set_max_length(len(str(float_entry)))
                        
                    self.view.change_color_and_balloon(entry, None)
                    # At the moment the RUN CHECK Button is always enabled
                        
                    self.model.detect_delay = float(entry.get_text())
                    
            else: # If the entry is an integer number
             
                # If the entry is a < 0: ERROR
                if float(entry.get_text()) < 0:   
                    self.view.change_color_and_balloon(entry, ["The value must be >= 0"])
                    # At the moment the RUN CHECK Button is always enabled                    
                
                else: # If the entry is >= 0: OK
                    
                    self.view.change_color_and_balloon(entry, None)
                    # At the moment the RUN CHECK Button is always enabled
                        
                    self.model.detect_delay = float(entry.get_text())                  
            
        else: # If the entry is a string 
                
            self.view.change_color_and_balloon(entry, ["The value must be >= 0.0"])

        
        # Set the max lenght of entry as 10 characters
        entry.set_max_length(10)
          
        return 
                
        
    # ---------------------------------------------------------------
    #        Signal handlers
    # ---------------------------------------------------------------

    def on_res_add_clicked(self, widget):
        """The selected item in the properties view is added to the
        restrictions list."""
        prop = self.get_selected_property()
        if prop is None: return
        self.model.restrictions_store.append((prop,))
        return

    def on_res_del_clicked(self, widget):
        """The selected item in the restrictions view is removed."""
        selection = self.view['tv_restrictions'].get_selection()
        (model, iter) = selection.get_selected()

        if model is None or iter is None: return

        path = model.get_path(iter)
        model.remove(iter)
        selection.select_path(path)
        if not selection.path_is_selected(path):
            row = path[0]-1
            if row >= 0:
                selection.select_path((row,))
        return

    def on_button_diagctx_add_clicked(self, widget):
        prop = self.get_selected_property()
        if prop is None: return
        prop_desc = prop._get_description()
        self.model.diag_ctx = prop
        self.view['entry_diagctx_text'].set_text(prop_desc)
        return

    def on_button_diagctx_del_clicked(self, widget):
        self.model.diag_ctx = None
        self.view['entry_diagctx_text'].set_text("")
        return

    def on_radiobutton_diag_finite_delay_toggled(self, widget):
        """Function called when the user selects FINITE delay"""
        self.view['entry_diag_delay'].set_editable(not widget.get_active())
        self.view['entry_diag_delay'].set_sensitive(not widget.get_active())
        self.view['combobox_diag_time_scale'].set_sensitive(self.model.slim_model.compile_info.uses_time_scaling and not widget.get_active())
        return
        

    # ---------------------------------------------------------------
    #        Notifications from observable properties
    # ---------------------------------------------------------------

    def property_single_selected_value_change(self, model, old, new):
        self.view.enable_diag(new)
        return

    def property_infinite_domain_value_change(self, model, old, new):
        """When the domain of the model is infinite, MathSAT and BMC
        must be used and the user will be given the option of the BCM bound."""

        # if new:
        #     self.view['vbox_satbound'].show()
        # else:
        #     self.view['vbox_satbound'].hide()
        self.view['vbox_satbound'].set_sensitive(new)
        return
           
    def property_changed_value_change(self, model, old, new):
        """Whenever something in the Slim model changes, Diagnosability
        Analysis will be reset as the present results might be incorrect."""
        self.reset_diag_analysis()
        return

    def property_diag_result_value_change(self, model, old, new):
        """When a new diagnosability analysis result is created, this method
        will be advised to display it."""

        # If the result is not already shown (notice that it is not clear why this functions is called twice
        # each time with the same parameters)
        if not self.model.diag_result_shown:

            # relieve old results
            if self.results_ctrl_A != None:
                assert self.results_ctrl_B is not None, "Toolchain is broken (DIAG)."
                self.results_ctrl_A.relieve_model(self.results_ctrl_A.model)
                self.results_ctrl_A = None
                self.results_ctrl_B.relieve_model(self.results_ctrl_B.model)
                self.results_ctrl_B = None
                pass
        
            if new is None:
                return

            assert old != new, "Toolchain is broken (DIAG)."

            # if now traces have to be shown

            if new.success in ("Error", "Unknown", "True"):
                self.view.display_diag_results(self.model.last_property_name, new.success, None, None)
                return

            # instantiate the various handlers

            tvm_A = TraceViewModel()
            tvm_B = TraceViewModel()
            
            trace_A = self.model.diag_result.result_A.trace
            trace_B = self.model.diag_result.result_B.trace
            
            tvm_A.add_traces(trace_A)
            tvm_B.add_traces(trace_B)

            tvm_A.set_active_trace(trace_A)
            tvm_B.set_active_trace(trace_B)

            trace_view_A = TraceView(windowed=False, filters=False)
            trace_view_B = TraceView(windowed=False, filters=False)

            self.results_ctrl_A = TraceCtrl(tvm_A, trace_view_A)
            self.results_ctrl_B = TraceCtrl(tvm_B, trace_view_B)

            # synchronize scrollbars of trace windows
            swtop_A = trace_view_A['sw_top']
            swtop_B = trace_view_B['sw_top']

            adj_v = gtk.Adjustment(
                value=0.00,
                lower=0.00,
                upper=100.00,
                step_incr=1.00,
                page_incr=10.00,
                page_size=10.00)
            
            adj_h = gtk.Adjustment(
                value=0.00,
                lower=0.00,
                upper=100.00,
                step_incr=1.00,
                page_incr=10.00,
                page_size=10.00)

            swtop_A.set_hadjustment( adj_h )
            swtop_B.set_hadjustment( adj_h )

            swtop_A.set_vadjustment( adj_v )
            swtop_B.set_vadjustment( adj_v )

            # tell the view to display the traces

            self.view.display_diag_results(self.model.last_property_name, new.success, trace_view_A, trace_view_B)
            
            # Notify the model that we have shown the result
            self.model.diag_result_shown = True
            #print self.model.diag_result_shown
            return

    pass # end of class
# ----------------------------------------------------------------------


class FdirFiaCtrl (StartStopActivityCtrl):
    """This is a sub-controller for handling FIA activities"""
        
    def __init__(self, model, view):
        StartStopActivityCtrl.__init__(self, model, view, 
                                       self.run_fia, model.stop_fsap)

        # this controller is interested in observing if the slim
        # model has infinite domain
        self.observe_model(model.slim_model)

        return

    def register_view(self, view): 
        StartStopActivityCtrl.register_view(self, view)

        tv = view['tv_fia_list']
        tv.set_model(self.model.ftrees_list)
        #double click == click on show_ftree_button
        tv.connect('row-activated', lambda tv, path, tvc: self.on_button_show_ftree_clicked(None) )

        rend = gtk.CellRendererText()
        col = gtk.TreeViewColumn('Num', rend, text=0)
        tv.append_column(col)

        rend = gtk.CellRendererText()
        col = gtk.TreeViewColumn('Alarm', rend, text=1)
        tv.append_column(col)

        rend = gtk.CellRendererText()
        col = gtk.TreeViewColumn('Event File Name', rend, text=2)
        tv.append_column(col)        

        rend = gtk.CellRendererText()
        col = gtk.TreeViewColumn('Gate File Name', rend, text=3)
        tv.append_column(col)        

        sel = tv.get_selection() #here
        sel.connect("changed", self.on_tv_fia_selection_changed)
        

        self.view['button_show_ftree'].set_sensitive(False)
        return

    def register_adapters(self):
        self.adapt("sat_bound", "fia_sat_bound")
        return

    def run_fia(self):
        """Makes the model generate fia, and shows the result"""
        fnames = self.model.generate_fia()
        if fnames is None: 
            self.view['label_fia_info'].set_markup(
                "There are <span foreground='red'>no alarms</span> in model. Fault Isolation Analysis is <b>meaningless</b>")
            return
        
        res = self.model.load_fault_trees(fnames)
        if res: msg = "There are <b>%d</b> fault trees. Select one and click <i>Show Fault Tree</i> button" % len(fnames)
        else: msg = "There are <b>%d</b> fault trees available."

        self.view['label_fia_info'].set_markup(msg)
        return
   
    # ---------------------------------------------------------------
    #        Signal handlers
    # ---------------------------------------------------------------    
    def on_tv_fia_selection_changed(self, selection):
        model_iter =  selection.get_selected()
        self.view['button_show_ftree'].set_sensitive(model_iter is not None)
        return

    def on_button_show_ftree_clicked(self, button):
        model_iter = self.view['tv_fia_list'].get_selection().get_selected()
        if model_iter is None: return # should not happen...
        model, iter = model_iter

        process = fsap_process.FTDProcess(model[iter][2], model[iter][3])
        return
    
    # observable properties
    def property_infinite_domain_value_change(self, model, old, new):
        self.view.enable_sat(not new)
        return

    pass # end of class
# ----------------------------------------------------------------------



class FdirFraCtrl  (StartStopActivityCtrl):
    """Controller for Fault Recovery"""

    def __init__(self, model, view):
        StartStopActivityCtrl.__init__(self, model, view, 
                                       self.run_fra, model.stop_nusmv)

        # this sub-ctrl handles the result view
        self.results_ctrl = ResultsCtrl(model, view.fra_results)

        # we observe the number of failures into the slim model to
        # enable the corresponding button. 
        self.model.slim_model.register_observer(self)
        self.prev_trace_ctrl = None # this is used to dispose controllers

        # we also observe the number of selected properties in the
        # properties model, to enable model checking button
        # accordingly
        self.model.filtered_props.register_observer(self)
        self.tvm = TraceViewModel() #the model of the filterbar

        self.fbar_view = FilterBarView(windowed=False)
        self.fbar = self.fbar_view['TBFilter']
        self.fbar_ctrl = FilterBarCtrl(self.tvm, 
                                       self.fbar_view)
    
        return

    def register_view(self, view): 
        StartStopActivityCtrl.register_view(self, view)

        # sets sensitivity of sat stuff depending on the selected tecnology 
        self.view['table_sat'].set_sensitive(not self.model.use_bdd)
        return

    def register_adapters(self):
        self.adapt("use_bdd")
        self.adapt("sat_bound")
        self.adapt("use_sbmc")
        self.adapt("sbmc_complete")   
        return
    
    def run_fra(self):
        """Makes the model run fault recovery analysis, and shows
        the result when ready"""
        self.model.run_fra()
        return

    # ---------------------------------------------------------------
    #        Signal handlers
    # ---------------------------------------------------------------    


    # ---------------------------------------------------------------
    #        Notifications from observable properties
    # ---------------------------------------------------------------    
        
    def property_single_selected_value_change(self, model, old, new):
        self.view.enable_fra(new)
        return

    # when bdd is used, sat stuff is grayed out, and property
    # filtering is set accordingly (no CTL for sat solving)
    def property_use_bdd_value_change(self, model, old, new):
        self.view['table_sat'].set_sensitive(not new)
        if new: self.model.filtered_props.set_filtering((prop_const.Logic.CTL, prop_const.Logic.LTL,))
        else: self.model.filtered_props.set_filtering((prop_const.Logic.LTL,))
        return

    def property_can_use_bdd_value_change(self, model, old, new):
        if not new:
            # checks if bdd is currenlty selected, and switch to
            # SAT if needed (this can happens when the user loads a
            # slim model which has infinite domain)
            if self.view['rb_use_bdd'].get_active():
                self.view['rb_use_bdd'].set_active(False)
                self.view['rb_use_sat'].set_active(True)
                pass
            pass

        self.view['rb_use_bdd'].set_sensitive(new)
        return 

    def property_result_value_change(self, model, old, new):
        """this is used to process results of model checking"""
        if isinstance(new, NusmvControlledException):
            self.view.show_exception(new.get_title(), new.get_description())
            return

        if self.prev_trace_ctrl != None:
            self.prev_trace_ctrl.relieve_model(self.prev_trace_ctrl.model)
            self.prev_trace_ctrl = None
            pass

        if new is None: 
            self.view.fra_results.show_no_results()
            return

        unk_msg = ("true up to bound %d" % new.sat_length) \
            if new.sat_length is not None else "true"

        if new.error is None:
            truth, stock_id = { None : (unk_msg, gtk.STOCK_DIALOG_QUESTION),
                                True : ("true", gtk.STOCK_YES),
                                False : ("false", gtk.STOCK_NO),
                                } [new.truth]
        else: return

        desc = """The <tt>%s</tt> property:
<span background="#feffbd"><tt>%s</tt></span>
has been found <b>%s</b>.""" % tuple (map (escape,
                                           [new.type_name, new.formula_text, truth]))

        # feedback to the user
        if new.truth: msg = "The Fault is <b>recoverable</b>"
        elif new.truth is None: msg = "The Fault <i>may be or may be not</i> recoverable"
        else: msg = "The Fault is <b>not recoverable</b>"
        self.view['label_fra_info'].set_markup(msg)

        trace_view = None
        if new.truth is False:
            desc += """ A counter-example is shown below."""

            trace_view = TraceView(windowed=False, filters=False)
            self.tvm.add_traces( new.trace )
            self.tvm.set_active_trace( new.trace )
            self.prev_trace_ctrl = TraceCtrl(self.tvm, trace_view)
            pass
        
        self.view.show_fra_result(truth, desc, trace_view, stock_id, self.fbar_view)
        return
        
        

    pass #end of class
# ----------------------------------------------------------------------


class FdirFcaCtrl (StartStopActivityCtrl):
    """This is a sub-controller for handling Fault Coverage Analysis activities"""


    ## Global variables representing the name of the headings of the matrix
    HEADING_ENABLED_STR = "Enabled"
    HEADING_FAULT_STR = "Fault"
    HEADING_DELAY_STR = "Detection\nDelay"
    HEADING_TIME_SCALE_STR = "Time Scale"
    HEADING_FINITE_DELAY_STR = "Finite\nDelay"
    HEADING_DETECT_RESULT_STR = "Detectable"
    HEADING_RECOVERABILITY_PROP_STR = "Recoverability\nProperty"
    HEADING_REC_RESULT_STR = "Recoverable"
    HEADING_CLASS_STR = "Classification"
    ## Global variables representing two other values of the liststore beeing
    ## model of the matrix
    FOREGROUND_STR = "Foreground"
    DELAY_ENABLED_STR = "Delay Enabled"


    def __init__(self, model, view):
        StartStopActivityCtrl.__init__(self, model, view, 
                                       self.run_fca, model.stop_fsap)

        ## we must observe both the slim model (for fault injections) and
        ## the properties model (for properties)
        self.observe_model(model.slim_model)
        self.observe_model(model.properties_model)

        self.infinite_domain = False
        self.timed_domain = False
        
        return
   

    def register_view(self, view):
        """GUI Initialisation function.

           Notice that essentially this function aims at defining the starting structure of the matrix for 
           the Fault Coverage Analysis. In particular, the treeview is set up; important to notice is the
           fact that the last values of the liststore are:
           * an HTML code representing the color of the strings of a row (black if the row is enabled, grey otherwise).
           * a boolean value representing the fact whether it is possible to define a delay (yes if the model
             is continuous, no in case the model is discrete)
           Additionally, the drop-down menus and the text entry are enabled or not depending also on the value of the first 
           checkbox (the one saying whether the row is selected or not).
        """
        StartStopActivityCtrl.register_view(self, view)

        ## Initialization of the treeview (only headings added)
        ## We have the following headers:
        ## * enabled -> bool
        ## * fault -> str
        ## * detection delay [fda] -> str
        ## * time scale [fda] -> drop down menu (str)
        ## * detectable [fda result] -> str
        ## * recoverability property [recoverability] -> drop down menu (str)
        ## * recoverable [recoverability result] -> str
        ## * classification -> str
        ## * foreground-color -> str
        ## * delay-editable -> bool

        ## Dictionary used to keep the correspondance between the index of a particular column in the lisstore
        ## In this way, if for instance we want to know the position of the Property column, we can get it
        ## through 'self.headings2int[FdirFcaCtrl.HEADING_RECOVERABILITY_PROP_STR]'
        headings_count = 0
        self.headings2int = dict()

        self.treeliststore = gtk.ListStore(bool,str,str,str,bool,str,str,str,str,str,bool)
        treeview = view['tv_fca_faults_list']
        treeview.set_model(self.treeliststore)

        # Checkbox 'Enabled' Column
        renderer_toggle = gtk.CellRendererToggle()
        renderer_toggle.connect("toggled", self.on_checkbutton_toggled)
        column_toggle = gtk.TreeViewColumn(FdirFcaCtrl.HEADING_ENABLED_STR, renderer_toggle, active=0)
        self.header_check_btn = gtk.CheckButton(FdirFcaCtrl.HEADING_ENABLED_STR)
        self.header_check_btn.set_active(True)
        self.header_check_btn.show()
        column_toggle.set_widget(self.header_check_btn)
        column_toggle.set_clickable(True)
        column_toggle.connect("clicked", self.on_header_checkbutton_clicked)
        treeview.append_column(column_toggle)
        self.headings2int[FdirFcaCtrl.HEADING_ENABLED_STR] = headings_count
        headings_count += 1

        # Fault name Column
        renderer_fault = gtk.CellRendererText()
        column_fault = gtk.TreeViewColumn(FdirFcaCtrl.HEADING_FAULT_STR, renderer_fault, text=1, foreground=9)
        treeview.append_column(column_fault)
        self.headings2int[FdirFcaCtrl.HEADING_FAULT_STR] = headings_count
        headings_count += 1

        # FAULT DETECTION
        # Detection Delay Column
        renderer_delay = gtk.CellRendererText()
        renderer_delay.connect("edited", self.on_delay_edited)
        column_delay =  gtk.TreeViewColumn(FdirFcaCtrl.HEADING_DELAY_STR, renderer_delay, text=2, foreground=9, editable=10)
        treeview.append_column(column_delay)
        self.headings2int[FdirFcaCtrl.HEADING_DELAY_STR] = headings_count
        headings_count += 1

        # Detection Delay time-unit Column
        time_scale_units = gtk.ListStore(str)
        time_scale_units.append(["msec"])
        time_scale_units.append(["sec"])
        time_scale_units.append(["min"])
        time_scale_units.append(["hour"])
        time_scale_units.append(["day"])
        renderer_time_scale = gtk.CellRendererCombo()
        renderer_time_scale.set_property("model", time_scale_units)
        renderer_time_scale.set_property("text-column", 0)
        renderer_time_scale.set_property("has-entry", False)
        renderer_time_scale.connect("edited", self.on_combo_time_scale_changed)
        column_time_scale = gtk.TreeViewColumn(FdirFcaCtrl.HEADING_TIME_SCALE_STR, renderer_time_scale, text=3,
                                               foreground=9, editable=10)
        treeview.append_column(column_time_scale)
        self.headings2int[FdirFcaCtrl.HEADING_TIME_SCALE_STR] = headings_count
        headings_count += 1

        # Fixed Delay Column
        renderer_fixed_delay_toggle = gtk.CellRendererToggle()
        renderer_fixed_delay_toggle.connect("toggled", self.on_fixed_delay_checkbutton_toggled)
        column_fixed_delay = gtk.TreeViewColumn(FdirFcaCtrl.HEADING_FINITE_DELAY_STR, renderer_fixed_delay_toggle, active=4, sensitive=10)
        treeview.append_column(column_fixed_delay)
        self.headings2int[FdirFcaCtrl.HEADING_FINITE_DELAY_STR] = headings_count
        headings_count += 1

        # Detection Result
        renderer_detect_res = gtk.CellRendererText()
        column_detect_res = gtk.TreeViewColumn(FdirFcaCtrl.HEADING_DETECT_RESULT_STR, renderer_detect_res, text=5, foreground=9)
        treeview.append_column(column_detect_res)
        self.headings2int[FdirFcaCtrl.HEADING_DETECT_RESULT_STR] = headings_count
        headings_count += 1

        # RECOVERABILITY
        # Recoverability property Column
        self.recoverability_properties = gtk.ListStore(str)
        self.recoverability_properties.append([""])
        renderer_property = gtk.CellRendererCombo()
        renderer_property.set_property("model", self.recoverability_properties)
        renderer_property.set_property("text-column", 0)
        renderer_property.set_property("has-entry", False)
        renderer_property.connect("edited", self.on_combo_rec_prop_changed)
        column_property = gtk.TreeViewColumn(FdirFcaCtrl.HEADING_RECOVERABILITY_PROP_STR, renderer_property, text=6, 
                                             foreground=9, editable=0)
        treeview.append_column(column_property)
        self.headings2int[FdirFcaCtrl.HEADING_RECOVERABILITY_PROP_STR] = headings_count
        headings_count += 1

        # Recoverability result Column
        renderer_recoverability_res = gtk.CellRendererText()
        column_recoverability_res = gtk.TreeViewColumn(FdirFcaCtrl.HEADING_REC_RESULT_STR, 
                                                       renderer_recoverability_res, text=7, foreground=9)
        treeview.append_column(column_recoverability_res)
        self.headings2int[FdirFcaCtrl.HEADING_REC_RESULT_STR] = headings_count
        headings_count += 1

        # CLASSIFICATION
        renderer_res = gtk.CellRendererText()
        column_res = gtk.TreeViewColumn(FdirFcaCtrl.HEADING_CLASS_STR, renderer_res, text=8, foreground=9)
        treeview.append_column(column_res)
        self.headings2int[FdirFcaCtrl.HEADING_CLASS_STR] = headings_count
        headings_count += 1

        ## Adding the two final values of the listsotre
        self.headings2int[FdirFcaCtrl.FOREGROUND_STR] = headings_count
        headings_count += 1
        self.headings2int[FdirFcaCtrl.DELAY_ENABLED_STR] = headings_count
        headings_count += 1

        ## Setting initial values for the GUI elements
        ## The default value of the combobox is the first one
        view['combobox_fca_overall_time_scale'].set_active(0)
        ## The time unit combobox must be selectable only in case the slim model uses time scaling
        view['combobox_fca_overall_time_scale'].set_sensitive(self.model.slim_model.compile_info.uses_time_scaling)
        ## Overall delay has initial value 0.0
        view['entry_fca_delay'].set_text("0.0")
        ## By default, BDDs are enabled
        view['rb_diag_use_bdd'].set_active(True)
        view['rb_rec_use_bdd'].set_active(True)
        ## By default the delay is finite, hence it is not possible to specify a delay
        ## unless fixed is selected
        self.view['entry_fca_delay'].set_editable(False)
        self.view['entry_fca_delay'].set_sensitive(False)

        return


    def register_adapters(self):
        return


    def run_fca(self):
        """Function which runs the Fault Coverage Analysis for the selected properties, i.e. it runs first Fault Detection
           and then Recoverability

           The workflow is the following:
           * Run Fault Detection
           * Display the result of Fault Detection
           * Run Recoverability
           * Display the result of Recoverability
           * Combine the results of the two analyses
        """

        from src.models.properties import MyTreeStore

        self.clear_active_entries()

        for row in self.treeliststore:
            
            header_fault = self.headings2int[FdirFcaCtrl.HEADING_FAULT_STR]
            header_prop = self.headings2int[FdirFcaCtrl.HEADING_RECOVERABILITY_PROP_STR]
            header_delay = self.headings2int[FdirFcaCtrl.HEADING_DELAY_STR]
            header_finite_delay = self.headings2int[FdirFcaCtrl.HEADING_FINITE_DELAY_STR]
            header_time_scale = self.headings2int[FdirFcaCtrl.HEADING_TIME_SCALE_STR]

            if row[0]:

                ## -- DETECTION

                from property_management.src.properties import Property, PatternInstance, StateFormulaPlaceholderInstance
                from src.models.properties import PatternModel
                from property_management.src.formulae import AtomicProposition
                
                ## TODO: check whether it is correct to build a property in this way
                pd = [p for p in prop_api.get_patterns_document(prop_const.Logic.LTL)]
                assert pd[0].id == "propositional"
                pattern_model = PatternModel(pd[0])
                pattern_instance = PatternInstance.create(pattern_model.logic, pattern_model.id)
                ap = AtomicProposition.create(prop_const.Logic.LTL, row[header_fault])
                placeholder = StateFormulaPlaceholderInstance.create(prop_const.Logic.LTL, "P", ap)
                pattern_instance.add_placeholder_instance(placeholder)
                _property = Property.create(pattern_instance, "propositional 0", str(row[1]))

                ## If the model is discrete, then delay is None
                ## Otherwise take the delay from the GUI
                if not self.infinite_domain:
                    delay = None
                elif row[header_finite_delay]: delay = -1.0
                else: delay = self.model.fca_get_delay(row[header_delay], row[header_delay+1])

                # -- DEBUG CODE
                if False:
                    print "-- Starting Fault Coverage Analysis (Detection part) for fault '%s'" %row[header_fault]
                    print "   with delay %s and SAT bound %s" %(delay, self.view['spinbtn_fca_diag_bound'].get_text())
                # --- END OF DEBUG CODE

                detect_fname = self.model.run_fca_detection(_property, detection_delay=delay, 
                                                            sat_bound=self.view['spinbtn_fca_diag_bound'].get_text())
                if detect_fname is None:
                    detect_msg = "No alarms in model"
                else:
                    res = self.model.load_fca_detection(detect_fname)
                    if not res:
                        detect_msg = "Empty alarms list"
                    else:
                        detect_msg = "Not empty alarms list"
                self.show_detection_result(detect_msg)

                # -- DEBUG CODE
                if False:
                    header_prop = self.headings2int[FdirFcaCtrl.HEADING_RECOVERABILITY_PROP_STR]

                    print "-- Starting Fault Coverage Analysis (Recoverability part) for fault '%s'" %row[header_fault]
                    print "   and property '%s'" %row[header_prop]
                # -- END OF DEBUG
                
                ## -- RECOVERABILITY
                for el in self.model.properties_model.get_properties_list():
                    if str(el[2].name) == str(row[header_prop]):
                        ## TODO: check why it seems not possible to run model checking with a tree store of more
                        ##       than one row
                        rows = MyTreeStore(bool,str,gobject.TYPE_PYOBJECT,bool,bool,
                                           gobject.TYPE_PYOBJECT,gobject.TYPE_PYOBJECT,gobject.TYPE_PYOBJECT,
                                           gobject.TYPE_PYOBJECT)
                        rows.append(None,[el[0],el[1],el[2],el[3],el[4],el[5],el[6],el[7],el[8]])
                        rec_result = self.model.run_fca_rec(rows, sat_bound=self.view['spinbtn_fca_rec_bound'].get_text())
                        self.show_recoverability_result(rec_result)

                ## -- SHOW THE FINAL RESULT
                self.show_classification_result(rec_result, detect_msg)
        return 



    def reset_fca(self):
        return



    def create_fault_inj_rows(self, fault_inj_list):
        """Functions which creates the rows of the matrix for the Fault Coverage Analysis.

           A new row is added for each new fault injection contained in the list passed as parameter.
           By default, every row is initialised as enabled.
        """
        tree_model = self.view['tv_fca_faults_list'].get_model()
        idx = 0

        ## If the list is of faults is empty, we have to delete all existing faults
        if fault_inj_list == []:
            self.treeliststore.clear()
            self.view.update()
            return

        faults = [row[1] for row in self.treeliststore]
       
        for el in fault_inj_list:
            if str(el) not in faults:
                # values: enabled, fault, detection delay, time scale, detectable, recoverability prop, recoverable, classification, foreground, delay_enabled
                self.treeliststore.append([True,str(el),"0.0","msec",True,None,"",None,None,'#000000',
                                           self.infinite_domain and self.view['combobox_fca_overall_time_scale'].get_sensitive()])
                faults.append(str(el))

        ## When the fault injections are modified, we must also check whether some pre-existent fault injections have
        ## not been deleted. To do so, we delete all the faults which belong to FCA but are not present in the fault
        ## injections tab
        idx = 0
        header_fault = self.headings2int[FdirFcaCtrl.HEADING_FAULT_STR]
        for el in self.treeliststore:
            if str(el[header_fault]) not in fault_inj_list:
                _iter = tree_model.get_iter(idx)
                tree_model.remove(_iter)
            idx += 1

        self.view.update()

        return



    def create_required_drop_down_menus(self, rec_props_list):
        """Functions which creates the required drop-down menu, i.e. the one for selecting the property for the 
           recoverability analysis (all properties).

           The information is retrieved from the list passed as parameter; notice that before adding a 
           property we check that this property is not already contained.

           @param rec_props_list: the list of properties selectable for recoverability
           @type  rec_props_list: [str]
        """
        rec_props = [row[0] for row in self.recoverability_properties]
        for el in rec_props_list:
            if str(el) not in rec_props:
                self.recoverability_properties.append([str(el)])

        ## Other-side check; if some properties previously loaded are not more existent, they must be deleted
        ## from the drop-down menu
        columns = self.view['tv_fca_faults_list'].get_columns()
        for col in columns:
            if col.get_title() == FdirFcaCtrl.HEADING_RECOVERABILITY_PROP_STR:
                assert len(col.get_cell_renderers()) == 1
                combo = col.get_cell_renderers()[0]
                combo_model = combo.get_property("model")
                idx = 0

                for el in self.recoverability_properties:
                    if str(el[0]) not in rec_props_list:
                        _iter = combo_model.get_iter(idx)
                        combo_model.remove(_iter)
                        
                        head_rec_prop = self.headings2int[FdirFcaCtrl.HEADING_RECOVERABILITY_PROP_STR]
                        itr = 0
                        while itr < len(self.treeliststore):
                            self.treeliststore[itr][head_rec_prop] = ""
                            itr += 1
                    idx += 1

        self.view.update()

        return



    def show_recoverability_result(self, result):
        """Shows the result of the recoverability analysis. Depending on the value of result.truth
           (True, False, None), a different message is shown.

           @param result: the result of the recoverability analysis
           @type  result: McResultModel
        """

        if result.truth: msg = "Recoverable"
        # elif result.truth is None: msg = "May be recoverable"
        elif result.truth is None: msg = "Recoverable"
        else: msg = "Not recoverable"

        ## We do not have the information about which fault we are referring to
        ## hence, what we can do is check which is the first empty entry in the 
        ## recoverability result whose checkbox is enabled
        ## Notice that we can do that because we execute model checking sequentially

        head = self.headings2int[FdirFcaCtrl.HEADING_REC_RESULT_STR]

        idx = 0
        for row in self.treeliststore:
            if row[0] is True and row[head] is None:
                self.treeliststore[idx][head] = msg
                self.view.update()
                idx +=1
                return
            idx += 1

        return


    def show_detection_result(self, detect_msg):
        """Shows the result of the fault detection analysis. Depending on the value of the detect_msg
           ("No alarms in model", "Empty alarms list", "Not empty alarms list"), a different 
           message is shown.

           @param detect_msg: the result of the Fault Detection Analysis
           @type  detect_msg: string
        """

        assert detect_msg in ["No alarms in model", "Empty alarms list", "Not empty alarms list"]

        if detect_msg == "Not empty alarms list":
            msg = "Detectable"
        else:
            msg = "Not Detectable"
            
        ## We do not have the information about which fault we are referring to
        ## hence, what we can do is check which is the first empty entry in the 
        ## diagnosability result whose checkbox is enabled
        ## Notice that we can do that because we execute model checking sequentially

        head = self.headings2int[FdirFcaCtrl.HEADING_DETECT_RESULT_STR]

        idx = 0
        for row in self.treeliststore:
            if row[0] is True and row[head] is None:
                self.treeliststore[idx][head] = msg
                self.view.update()
                idx +=1
                return
            idx += 1

        return



    def show_classification_result(self, rec_result, detect_msg):
        """Once recoverability and fault detection have been performed for a given fault, their
           results are combined and a result ("Covered" or "Uncovered") is shown in the
           classification column.

           @param rec_result: the result of the recoverability analysis
           @type  rec_result: McResultModel

           @param detect_msg: the result of the fault detection analysis
           @type  detect_msg: str
        """

        assert detect_msg in ["No alarms in model", "Empty alarms list", "Not empty alarms list"]

        if detect_msg == "Not empty alarms list": detect_res = "Detectable"
        else: detect_res = "Not Detectable"

        if rec_result.truth: rec_msg = "Recoverable"
        elif rec_result.truth is None: rec_msg = "Recoverable"
        else: rec_msg = "Not recoverable"
        
        if detect_res == "Not Detectable":
            classification = "Uncovered"
        else:
            if rec_msg == "Recoverable":
                classification = "Covered"
            else:
                classification = "Uncovered"

        head = self.headings2int[FdirFcaCtrl.HEADING_CLASS_STR]
        
        idx = 0
        for row in self.treeliststore:
            if row[0] is True and row[head] is None:
                self.treeliststore[idx][head] = classification
                self.view.update()
                idx += 1
                return
            idx += 1

        return



    def clear_active_entries(self):
        """Removes the results (recoverability, dectection and classification) of the
           active rows of the matrix.

           Notice that this method is called in order to reset the analysis once the user 
           wants to perform it again.
        """

        idx = 0
        head_rec_result = self.headings2int[FdirFcaCtrl.HEADING_REC_RESULT_STR]
        head_detect_result = self.headings2int[FdirFcaCtrl.HEADING_DETECT_RESULT_STR]
        head_class_result = self.headings2int[FdirFcaCtrl.HEADING_CLASS_STR]

        for row in self.treeliststore:
            if row[0] is True:
                self.treeliststore[idx][head_rec_result] = None
                self.treeliststore[idx][head_detect_result] = None
                self.treeliststore[idx][head_class_result] = None
            idx += 1

        self.view.update()
        return



    def set_overall_delay(self, delay):
        """Sets the specified delay for all the entries of the matrix"""

        idx = 0
        head_delay = self.headings2int[FdirFcaCtrl.HEADING_DELAY_STR]

        for row in self.treeliststore:
            if row[0] is True:
                self.treeliststore[idx][head_delay] = delay
            idx+=1

        self.view.update()
        return



    def ready_for_fca(self):
        """Returns true if the tool is ready to perform Fault Coverage Analysis.

           Notice that the tool is ready to perform FCA in case for each enabled entry
           the recoverability property has been selected.
        """

        head_rec_prop = self.headings2int[FdirFcaCtrl.HEADING_RECOVERABILITY_PROP_STR]

        at_least_one_enabled = False

        for row in self.treeliststore:
            if row[0] is True:
                at_least_one_enabled = True
                if row[head_rec_prop] is None or row[head_rec_prop] == "":
                    return False
        return at_least_one_enabled


    # ---------------------------------------------------------------
    #        Signal handlers
    # ---------------------------------------------------------------


    def on_header_checkbutton_clicked(self, widget):
        """Function called when the 'Enabled' heading is clicked"""

        head_color = self.headings2int[FdirFcaCtrl.FOREGROUND_STR]

        if self.header_check_btn.get_active():
            for row in self.treeliststore:
                row[0] = False
                row[head_color] = "#C0C0C0"
            self.header_check_btn.set_active(False)
        else:
            for row in self.treeliststore:
                row[0] = True
                row[head_color] = "#000000"
            self.header_check_btn.set_active(True)

        self.view.get_start_button().set_sensitive(self.ready_for_fca())
        self.view.update()

        return



    def on_checkbutton_toggled(self, widget, path):
        """Functions called when the 'Enabled' checbutton of a specific row is toggled.

           In case a row is enabled (disabled) the color of the various shown strings
           must be changed to black (gray)
        """

        # setting the value of the 'enabled' checkbox
        self.treeliststore[path][0] = not self.treeliststore[path][0]

        if self.treeliststore[path][0]: # row selected
            # changing foreground color to black
            self.treeliststore[path][9] = '#000000'
            pass

        else: # row de-selected
            # changing foreground color to grey
            self.treeliststore[path][9] = '#C0C0C0'
            pass

        self.view.get_start_button().set_sensitive(self.ready_for_fca())
        self.view.update()

        return



    def on_fixed_delay_checkbutton_toggled(self, widget, path):
        """Functions called when the checkbutton for the fixed delay of a single row is toggled
        """
        head = self.headings2int[FdirFcaCtrl.HEADING_FINITE_DELAY_STR]

        self.treeliststore[path][head] = not self.treeliststore[path][head]
        self.view.update()
        return


    def on_combo_rec_prop_changed(self, widget, path, text):
        """Function called when the user changes selection in the drop-down menu for 
           recoverability property
        """
        idx = self.headings2int[FdirFcaCtrl.HEADING_RECOVERABILITY_PROP_STR]
        self.treeliststore[path][idx] = text
        self.view.get_start_button().set_sensitive(self.ready_for_fca())
        return


    def on_combo_time_scale_changed(self, widget, path, text):
        """Function called when the user changes selection in the drop-down menu for 
           the time scale unit for the detection delay
        """
        idx = self.headings2int[FdirFcaCtrl.HEADING_TIME_SCALE_STR]
        self.treeliststore[path][idx] = text
        return


    def on_combobox_fca_overall_time_scale_changed(self, widget):
        """Function called when the user changes the scale unit for the overall delay"""

        idx = self.headings2int[FdirFcaCtrl.HEADING_TIME_SCALE_STR]
        row = 0
        while row < len(self.treeliststore):
            self.treeliststore[row][idx] = widget.get_active_text()
            row += 1
        return


    def on_delay_edited(self, widget, path, value):
        """Function called when the user edits the delay text entry.
        """

        idx = self.headings2int[FdirFcaCtrl.HEADING_DELAY_STR]

        try:
            val = float(value)
            self.treeliststore[path][idx] = value
        except ValueError:
            raise Exception("The delay must be a float'")

        return

    
    def fca_on_overall_delay_clicked(self, widget):
        """Function called when the user sets the overall delay"""

        try:
            val = float(widget.get_text())
            self.set_overall_delay(val)
        except ValueError:
            raise Exception("The delay must be a float")

        return


    def on_rb_rec_use_sat_toggled(self, button):
        """Function called when the user clicks the radio button for enabling sat in recoverability"""
        self.model.rec_use_sat = button.get_active()
        return

    def on_rb_rec_use_bdd_toggled(self, button):
        """Function called when the user clicks the radio button for enabling bdd in recoverability"""
        self.model.rec_use_sat = not button.get_active()
        self.view['spinbtn_fca_rec_bound'].set_editable(not button.get_active())
        return

    def on_rb_diag_use_sat_toggled(self, button):
        """Function called when the user clicks the radio button for enabling sat in diagnosability"""
        self.model.diag_use_sat = button.get_active()
        return

    def on_rb_diag_use_bdd_toggled(self, button):
        """Function called when the user clicks the radio button for enabling bdd in diagnosability"""
        self.model.diag_use_sat = not button.get_active()
        self.view['spinbtn_fca_diag_bound'].set_editable(not button.get_active())
        return


    def on_radiobutton_fca_diag_fixed_delay_toggled(self, button):
        """Function called when the user clicks the radio button for setting finite delay
           (overall case)"""

        self.view['entry_fca_delay'].set_editable(not button.get_active())
        self.view['entry_fca_delay'].set_sensitive(not button.get_active())
        self.view['combobox_fca_overall_time_scale'].set_sensitive(self.model.slim_model.compile_info.uses_time_scaling and not button.get_active())

        idx = 0
        head = self.headings2int[FdirFcaCtrl.HEADING_FINITE_DELAY_STR]

        while idx < len(self.treeliststore):
            self.treeliststore[idx][head] = button.get_active()
            idx += 1

        return

    # ---------------------------------------------------------------
    #        Notifications from observable properties
    # ---------------------------------------------------------------

    def property_failures_value_change(self, model, old, new):
        """Observes the 'failures' observable of the slim model, i.e. it is activated when 
           the number of loaded fault injections is changed"""

        fault_injections_list = self.model.get_fault_injections()
        self.create_fault_inj_rows(fault_injections_list)
        self.view.get_start_button().set_sensitive(self.ready_for_fca())

        return


    def property_infinite_domain_value_change(self, model, old, new):
        """Observes the 'inifinte_domain' observable of the slim model"""

        self.infinite_domain = new
        self.view.enable_sat(not new)

        _iter = 0
        while _iter < len(self.treeliststore):
            # FIXME: get the index of the delay_enabled_str from headings2int
            self.treeliststore[_iter][10] = self.infinite_domain and self.view['combobox_fca_overall_time_scale'].get_sensitive()
            _iter += 1

        self.model.rec_use_sat = not new
        self.model.diag_use_bdd = not new

        return

    
    def property_timed_domain_value_change(self, model, old, new):
        """Observes the 'timed_domain' observable of the slim model"""

        self.timed_domain = new
        return


    def property_loaded_properties_value_change(self, model, old, new):
        """Observes the 'loaded_properties' observable of the properties model, i.e. it is
           activated when the number of loaded properties has changed

           Withing this function, a list of properties is required and passed to 
           the function which must create the drop-down menu for the recoverability 
           property.
        """

        ## -- DEBUG CODE
        if False:
            for el in model.get_properties_list():
                use = el[0] # bool
                description = el[1] # string
                _property = el[2] # gobject.TYPE_PYGOBJECT 
                single_sel = el[3] # bool
                multiple_sel = el[4] # bool
                mc_result = el[5] # gobject.TYPE_PYOBJECT
                ft_generation = el[6] # gobject.TYPE_PYOBJECT
                performability_graph = el[7] # gobject.TYPE_PYOBJECT
                fda_list = el[8] # gobject.TYPE_PYOBJECT
                print "USE: " + str(use)
                print "DESCRIPTION: " + str(description)
                print "PROPERTY: " + str(_property)
                print "PROPERTY PATTERN INSTANCE: " + str(_property.pattern_instance)
                print "PROPERTY NAME: " + str(_property.name)
                print "PROPERTY DESCRIPTION: " + str(_property.description)
                print "SINGLE_SEL: " + str(single_sel)
                print "MULTIPLE SEL: " + str(multiple_sel)
                print "MC RESULT: " + str(mc_result)
                print "FT GENERATION: " + str(ft_generation)
                print "PERFORMABILITY GRAPH: " + str(performability_graph)
                print "FDA LIST: " + str(fda_list)
                pass
            pass
        ## -- END OF DEBUG CODE

        ## We need only the name of the different properties
        rec_properties_list = []

        for el in model.get_properties_list():
            rec_properties_list.append(el[2].name)

        self.create_required_drop_down_menus(rec_properties_list)
        self.view.get_start_button().set_sensitive(self.ready_for_fca())

        return


    pass # end of class
# ----------------------------------------------------------------------
