# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by GMV Aerospace and Defence  
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================
import os.path
import gtk
import gobject
from gtkmvc import Controller
from src.models.mission import PhaseDialogModel, OpModeDialogModel
from src.views.mission_phases_op_modes import PhaseDialogView, OpModeDialogView
from delayedcall import DelayedCall


class PhasesOpModesCtrl (Controller):
    """Represents the Controller of Phases and OpModes."""

    def __init__(self, model, view):
        """Constructor Method"""
        Controller.__init__(self, model, view)
        return


    def register_view(self, view):
        """Sets the windows, dialog, buttons and other parts of View."""
        Controller.register_view(self, view)

        view['button_add_phase'].set_sensitive(True)
        view['button_edit_phase'].set_sensitive(False)
        view['button_remove_phases'].set_sensitive(False)

        view['button_add_op_mode'].set_sensitive(True)
        view['button_edit_op_mode'].set_sensitive(False)
        view['button_remove_op_modes'].set_sensitive(False)

        tv = view['tv_phases']
        tv.set_model(self.model.get_phases_list())
        tv.append_column(self.model.get_phases_col())

        tv = view['tv_op_modes']
        tv.set_model(self.model.get_op_modes_list())
        tv.append_column(self.model.get_op_modes_col())

        # to control selection of phases
        sel = self.view['tv_phases'].get_selection()
        sel.set_mode(gtk.SELECTION_MULTIPLE)
        sel.connect('changed', self.on_phases_selection_changed)

        # to control selection of op-modes
        sel = self.view['tv_op_modes'].get_selection()
        sel.set_mode(gtk.SELECTION_MULTIPLE)
        sel.connect('changed', self.on_op_modes_selection_changed)

        return


    def register_adapters(self):
        return


    def open_confirm_phase_window(self, paths=None):
        """Sets the Confirm Window of a new phase edited.
           @param   paths:  The path of phase.
	   @type    paths:  tuple
        """
        if paths is None:
            old_phase = None
        else:
            old_phase = self.model.get_phases_list()[paths[0][0]][0]

        m = PhaseDialogModel(self.model, old_phase)
        v = PhaseDialogView()
        c = PhaseDialogCtrl(m, v)

        response = v.run()

        if gtk.RESPONSE_APPLY == response:
            if old_phase is None:
                self.model.append_phase(m)
            else:
                self.model.edit_phase(m)

        return


    def open_confirm_op_mode_window(self, paths=None):
        """Sets the Confirm Window of a new op_mode edited.
           @param   paths:  The path of op_mode.
	   @type    paths:  tuple
        """
        if paths is None:
            old_op_mode = None
        else:
            old_op_mode = self.model.get_op_modes_list()[paths[0][0]][0]

        m = OpModeDialogModel(self.model, old_op_mode)
        v = OpModeDialogView()
        c = OpModeDialogCtrl(m, v)

        response = v.run()

        if gtk.RESPONSE_APPLY == response:
            if old_op_mode is None:
                self.model.append_op_mode(m)
            else:
                self.model.edit_op_mode(m)

        return


    # ---------------------------------------------------------------
    #        Signal handlers
    # ---------------------------------------------------------------    
    def _clear_selections(self):
        """Clears the old phases and op_modes selected."""
        sel = self.view['tv_phases'].get_selection()
        sel.unselect_all()
        sel = self.view['tv_op_modes'].get_selection()
        sel.unselect_all()
        return


    def on_button_add_phase_clicked(self, button):
        """Adds a phase to the Phases list.
           @param   button:  The Add Button (self.view['button_add_phase']).
	   @type    button:  gtk.Button
        """
        self.open_confirm_phase_window()
        self._clear_selections()
        return


    def on_button_add_op_mode_clicked(self, button):
        """Adds a op_mode to the OpModes list.
           @param   button:  The Add Button (self.view['button_add_op_mode']).
	   @type    button:  gtk.Button
        """
        self.open_confirm_op_mode_window()
        self._clear_selections()
        return


    def on_button_remove_phases_clicked(self, button):
        """Removes the phase selected in the list.
           @param   button:  The Remove Button (self.view['button_remove_phases']).
	   @type    button:  gtk.Button
        """
        sel = self.view['tv_phases'].get_selection()
        tree_model, paths = sel.get_selected_rows()
        self.model.remove_phases(tree_model, paths)
        self._clear_selections()
        return


    def on_button_remove_op_modes_clicked(self, button):
        """Removes the op_mode selected in the list.
           @param   button:  The Remove Button (self.view['button_remove_op_modes']).
	   @type    button:  gtk.Button
        """
        sel = self.view['tv_op_modes'].get_selection()
        tree_model, paths = sel.get_selected_rows()
        self.model.remove_op_modes(tree_model, paths)
        self._clear_selections()
        return


    def on_phases_selection_changed(self, selection):
        """Changes the phases selected.
           @param   selection:  The selection of phases.
	   @type    selection:  gtk.TreeSelection
        """
        self.view['button_remove_phases'].set_sensitive(
            selection.count_selected_rows() > 0)
        
        """ In addition, the "Edit" button has to become enabled when
        one phase is selected."""
        _, paths = selection.get_selected_rows()
        self.view['button_edit_phase'].set_sensitive(
            selection.count_selected_rows() == 1)
        return


    def on_op_modes_selection_changed(self, selection):
        """Changes the op_modes selected.
           @param   selection:  The selection of op_modes.
	   @type    selection:  gtk.TreeSelection
        """
        self.view['button_remove_op_modes'].set_sensitive(
            selection.count_selected_rows() > 0)
        
        """ In addition, the "Edit" button has to become enabled when
        one op-mode is selected."""
        _, paths = selection.get_selected_rows()
        self.view['button_edit_op_mode'].set_sensitive(
            selection.count_selected_rows() == 1)
        return


    def on_button_edit_phase_clicked(self, button):
        """Edits the phase selected in the list.
           @param   button:  The Edit Button (self.view['button_edit_phase']).
	   @type    button:  gtk.Button
        """
        sel = self.view['tv_phases'].get_selection()
        _, paths = sel.get_selected_rows()
        
        self.open_confirm_phase_window(paths)
        self._clear_selections()
        return


    def on_phase_selected(self, tv, event):
        """Selects the phase from the list.
           @param   tv:  The Tree View of the Phase (self.view['tv_phases']).
	   @type    tv:  gtk.TreeView
           @param   event:  The event associated
	   @type    event:  gtk.gdk.Event
        """
        if event.type == gtk.gdk._2BUTTON_PRESS: 
            sel = self.view['tv_phases'].get_selection()
            _, paths = sel.get_selected_rows()
            if paths: 
                self.open_confirm_phase_window(paths)
                self._clear_selections()
            else: return
        return


    def on_button_edit_op_mode_clicked(self, button):
        """Edits the op_mode selected in the list.
           @param   button:  The Edit Button (self.view['button_edit_op_mode']).
	   @type    button:  gtk.Button
        """
        sel = self.view['tv_op_modes'].get_selection()
        _, paths = sel.get_selected_rows()
        
        self.open_confirm_op_mode_window(paths)
        self._clear_selections()
        return


    def on_op_mode_selected(self, tv, event):
        """Selects the op_mode from the list.
           @param   tv:  The Tree View of the OpMode (self.view['tv_op_modes']).
	   @type    tv:  gtk.Button
           @param   event:  The event associated
	   @type    event:  gtk.gdk.Event
        """
        if event.type == gtk.gdk._2BUTTON_PRESS: 
            sel = self.view['tv_op_modes'].get_selection()
            _, paths = sel.get_selected_rows()
            if paths: 
                self.open_confirm_op_mode_window(paths)
                self._clear_selections()
            else: return
        return 


    def phases_op_modes_busy_value_change(self, model, old, new):
	"""Busy level, in this case, is 0 if AUTOGEF is not running and
           1 otherwise. If AUTOGEF is active, all the informations contained
           in the "Mission" tab cannot be edited. This could otherwise
           lead to misinterpretations after AUTOGEF lead to a result.
           @param   model:  The Mission Model.
	   @type    model:  MissionModel
           @param   old:  The number of old value.
	   @type    old:  int
           @param   new:  The number of new value.
	   @type    new:  int
        """
	if model is self: return # not interested in introspection

	self.view['button_edit_phase'].set_sensitive(new == 0)
	self.view['button_edit_op_mode'].set_sensitive(new == 0)
	self.view['button_remove_phases'].set_sensitive(new == 0)
	self.view['button_remove_op_modes'].set_sensitive(new == 0)
	self.view['button_add_phase'].set_sensitive(new == 0)
	self.view['button_add_op_mode'].set_sensitive(new == 0)
	return

pass # end of PhasesOpModesCtrl class
# ----------------------------------------------------------------------


class PhaseDialogCtrl(Controller):
    """Represents the Dialog Class of Phase"""

    def __init__(self, model, view, loading_phases=False):
        """Constructor Method"""
        Controller.__init__(self, model, view)
        self.loading_phases = loading_phases
	self.effect_valid = False
        return


    def register_view(self, view):
        """Sets the windows, dialog, buttons and other parts of View."""
        self.delay = DelayedCall(100, view.change_color_and_balloon)
        old_phase = self.model.get_old_phase()

        if old_phase:
            self.view['entry_cp_name'].set_text(old_phase.get_name())
            pass

        view['entry_cp_name'].connect(
            "changed", self.on_entry_cp_name_changed)
	view['entry_cp_name'].connect(
	    "activate", self.on_activate_entry_effect) 
        view['button_cp_confirm'].set_sensitive(
            not old_phase is None)

        return


    def register_adapters(self):
        return

    # -------------------------------------------------------------
    #         Signal handlers
    # -------------------------------------------------------------
    def on_button_cp_confirm_clicked(self, button):
        """Confirms the new name of the phase selected in the list.
           @param   button:  The Confirm Button (self.view['button_cp_confirm']).
	   @type    button:  gtk.Button
        """
        phase_name = self.view['entry_cp_name'].get_text().decode('utf_8')
        self.model.instantiate(phase_name)
        self.view['confirm_phase'].response(gtk.RESPONSE_APPLY)
        self.view['confirm_phase'].destroy()
        return


    def on_activate_entry_effect(self, editable):
        """Confirms the new name of the phase selected in the list with the key ENTER.
           @param   editable:  The entry
	   @type    editable:  string
        """
	if self.effect_valid:
            # Simulate clicking of op_mode
            self.on_button_cp_confirm_clicked(None)
        return


    def on_button_cp_discard_clicked(self, button):
        """Discards the new name of the phase selected in the list.
           @param   button:  The Discard Button (self.view['button_cp_discard']).
	   @type    button:  gtk.Button
        """
        self.view['confirm_phase'].response(gtk.RESPONSE_CANCEL)
        self.view['confirm_phase'].destroy()
        return


    def on_entry_cp_name_changed(self, entry):
        """Changes the new name of the phase selected in the list.
           @param   entry:  The new entry name
	   @type    entry:  string
        """
        if len(entry.get_text()) == 0:
            self.view.change_color_and_balloon(entry, ["The name length must be at least 1"])
            self.view['button_cp_confirm'].set_sensitive(False)
        else:
            try:
                entry.get_text().decode('utf_8')
            except UnicodeDecodeError:
                # It was not a ascii-encoded unicode string
                self.delay.change_color_and_balloon(
                    entry, ["Please use UTF-8 characters only"])
                #make sure the "Confirm" button is disabled
                self.view['button_cp_confirm'].set_sensitive(False)

            else:
                if self.model.mission_model.name_exist(
                    entry.get_text(),
                    self.model.get_old_phase().get_name() if self.model.get_old_phase() is not None else None
                    ):
                    
                    if entry.get_text().lower() == "none":
                        self.delay.change_color_and_balloon(entry, ["This name is invalid"])
                    else:
                        self.delay.change_color_and_balloon(entry, ["This name already exist"])
                    self.view['button_cp_confirm'].set_sensitive(False)

                else:
                    self.view.change_color_and_balloon(entry, None)
                    self.view['button_cp_confirm'].set_sensitive(True)
		    self.effect_valid = True

        return

pass # end of PhaseDialogCtrl class
# ----------------------------------------------------------------------


class OpModeDialogCtrl(Controller):
    """Represents the Dialog Class of OpMode"""

    def __init__(self, model, view, loading_op_modes=False):
        """Constructor Method"""
        Controller.__init__(self, model, view)
        self.loading_op_modes = loading_op_modes
        self.effect_valid = False
        return


    def register_view(self, view):
        """Sets the windows, dialog, buttons and other parts of View."""
        self.delay = DelayedCall(100, view.change_color_and_balloon)
        old_op_mode = self.model.get_old_op_mode()

        if old_op_mode:
            self.view['entry_co_name'].set_text(old_op_mode.get_name())
            pass

        view['entry_co_name'].connect(
            "changed", self.on_entry_co_name_changed)
	view['entry_co_name'].connect(
	    "activate", self.on_activate_entry_effect) 
        view['button_co_confirm'].set_sensitive(
            not old_op_mode is None)

        return


    def register_adapters(self):
        return


    # -------------------------------------------------------------
    #         Signal handlers
    # -------------------------------------------------------------
    def on_button_co_confirm_clicked(self, button):
        """Confirms the new name of the op_mode selected in the list.
           @param   button:  The Confirm Button (self.view['button_co_confirm']).
	   @type    button:  gtk.Button
        """
        op_mode_name = self.view['entry_co_name'].get_text().decode('utf_8')
        self.model.instantiate(op_mode_name)
        self.view['confirm_op_mode'].response(gtk.RESPONSE_APPLY)
        self.view['confirm_op_mode'].destroy()
        return


    def on_activate_entry_effect(self, editable):
        """Confirms the new name of the op_mode selected in the list with the key ENTER.
           @param   editable:  The entry
	   @type    editable:  string
        """
	if self.effect_valid:
            # Simulate clicking of op_mode
            self.on_button_co_confirm_clicked(None)
        return


    def on_button_co_discard_clicked(self, button):
        """Discards the new name of the op_mode selected in the list.
           @param   button:  The Discard Button (self.view['button_co_discard']).
	   @type    button:  gtk.Button
        """
        self.view['confirm_op_mode'].response(gtk.RESPONSE_CANCEL)
        self.view['confirm_op_mode'].destroy()
        return


    def on_entry_co_name_changed(self, entry):
        """Changes the new name of the op_mode selected in the list.
           @param   entry:  The new entry name
	   @type    entry:  string
        """
        if len(entry.get_text()) == 0:
            self.view.change_color_and_balloon(entry, ["The name length must be at least 1"])
            self.view['button_co_confirm'].set_sensitive(False)
        else:
            try:
                entry.get_text().decode('utf_8')
            except UnicodeDecodeError:
                # It was not a ascii-encoded unicode string
                self.delay.change_color_and_balloon(
                    entry, ["Please use UTF-8 characters only"])
                #make sure the "Confirm" button is disabled
                self.view['button_co_confirm'].set_sensitive(False)

            else:
                if self.model.mission_model.name_exist(
                    entry.get_text(),
                    self.model.get_old_op_mode().get_name() if self.model.get_old_op_mode() is not None else None
                    ):
                    
                    if entry.get_text().lower() == "none":
                        self.delay.change_color_and_balloon(entry, ["This name already exist"])
                    else:
                        self.delay.change_color_and_balloon(entry, ["This name already exist"])
                    self.view['button_co_confirm'].set_sensitive(False)

                else:
                    self.view.change_color_and_balloon(entry, None)
                    self.view['button_co_confirm'].set_sensitive(True)
		    self.effect_valid = True
        return
        
pass # end of OpModeDialogCtrl class
# ----------------------------------------------------------------------
