# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================


import os.path
import gtk
import gobject

from gtkmvc import Controller

from src.models.properties import PropertyDialogModel
from src.models.properties import PropertiesModel
from src.models.properties import PropertyModel
from src.views.properties import PropertyDialogView
from delayedcall import DelayedCall

import property_management.src.api as api
from property_management.src.properties import *
from property_management.src.constants import PlaceholderType, Logic
import re

# Controller for the property instantiation
class PropertiesCtrl (Controller):
    """Controls property instantiation"""
    
    def __init__(self, m, v):
        Controller.__init__(self, m, v)
        self.iter = None
	# for listening to changes to currently run activities
	m.appl_model.register_observer(self);
        return

    def register_view(self, view):
        """ At this point fill widgets with data """
        
        #view['properties_treeview'].append_column(self.model.use_col)
        view['properties_treeview'].append_column(self.model.properties_col)
        view['properties_treeview'].set_model(self.model.properties_tree)
                
        return

    def register_adapters(self):
        # setup of adapters
        return

    # ---------------------------------------------------------------
    #        Signal handlers
    # ---------------------------------------------------------------

    def on_properties_treeview_cursor_changed(self, treeview):
        self.iter = treeview.get_selection().get_selected()[1]
        
    def on_add_clicked(self, button):
        """This is called when the user wants to add a property"""
        
        m = PropertyDialogModel(self.model.slim_model)
        v = PropertyDialogView(self.model, parent=self.view)
        c = PropertyDialogCtrl(m, v, self.model)
        
        response = v.run()
        
        if gtk.RESPONSE_APPLY == response:
            self.model.prepend_properties(m)
            self.model.property_counter += 1
            self.model.loaded_properties += 1
    
    def on_delete_clicked(self, button):
        if self.iter:
            self.model.remove_property(self.iter)
            self.model.loaded_properties -= 1
        
    def on_edit_clicked(self, button):
        if self.iter:
            iter_old_row = self.iter
            old_prop = self.model.get_properties_tree()[iter_old_row][PropertiesModel.PROPERTY]
            m = PropertyDialogModel(self.model.slim_model, old_prop)
            v = PropertyDialogView(self.model)
            c = PropertyDialogCtrl(m, v, self.model)
            
            response = v.run()
            
            if gtk.RESPONSE_APPLY == response:
                self.model.prepend_properties(m, iter_old_row)
                self.model.remove_property(iter_old_row)
        pass
    
    def on_clone_clicked(self, button):
        """ No cloning will be implemented """
        pass
    
    def save_properties(self):
        """ Save properties to a file with extension *.propxml"""

        dialog = gtk.FileChooserDialog("Save Properties Document",
                                       None,
                                       gtk.FILE_CHOOSER_ACTION_SAVE,
                                       (gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL,
                                       gtk.STOCK_SAVE, gtk.RESPONSE_OK))
        dialog.set_default_response(gtk.RESPONSE_OK)
        filter = gtk.FileFilter()
        filter.set_name("Property Documents")
        filter.add_pattern("*.propxml")
        dialog.add_filter(filter)
        dialog.set_current_name(
            self.model.generate_unique_properties_basename() + ".propxml")
        dialog.set_current_folder(os.path.dirname(
            self.model.generate_unique_properties_basename() + ".propxml"))
        dialog.set_do_overwrite_confirmation(True)    
        dialog.show()
        response = dialog.run()
        filename = dialog.get_filename()
        dialog.destroy()
        if response == gtk.RESPONSE_OK:
            top_level_row_num = len(self.model.get_properties_list())
            # do not take instances of component based properties
            # i.e. take only rows on top level
            props_list = []
            for i in xrange(top_level_row_num):
                props_list.append(self.model.get_properties_list()[i][PropertiesModel.PROPERTY].property)
            fd = open(filename, 'w')
            fd.write(str(PropertiesDocument.create(props_list)))
            fd.close()
            pass
        

    def load_properties(self):
        # Opens the file chooser, and possibly adds properties
        
        chooser = gtk.FileChooserDialog(title="Open Properties Document",
                                        action=gtk.FILE_CHOOSER_ACTION_OPEN,
                                        buttons=(gtk.STOCK_CANCEL,gtk.RESPONSE_CANCEL,
                                                 gtk.STOCK_OPEN,
                                                 gtk.RESPONSE_OK))

        chooser.set_select_multiple(True)
        filter = gtk.FileFilter()
        filter.set_name("Property Documents")
        filter.add_pattern("*.propxml")
        chooser.add_filter(filter)
        filter_all = gtk.FileFilter()
        filter_all.set_name("All Files")
        filter_all.add_pattern("*")
        chooser.add_filter(filter_all)
        chooser.set_current_folder(os.path.dirname(
            self.model.generate_unique_properties_basename() + ".propxml"))

        response = chooser.run()
        filenames = chooser.get_filenames()
        chooser.destroy()
        loaded_properties = []
        if response == gtk.RESPONSE_OK:
            for filename in filenames:
                self.load_properties_file(filename)
            pass

    def load_properties_file(self, filename):
        loaded_properties = []
        doc = parse_properties_document(filename)
        errors = get_syntax_errors_in_document(doc)
        if len(errors) == 0:
            for prop in doc:
                loaded_properties.append(prop)
        else:
            #display dialog saying the file is not valid
            message = "The selected Property Document \
contains syntactical errors. Nothing will be loaded.\nErrors:\n"
            for error in errors:
                message += str(error) + "\n"
            error_msg = gtk.MessageDialog(None, 0, gtk.MESSAGE_ERROR,
                gtk.BUTTONS_OK, message_format=message)
            error_msg.run()
            error_msg.destroy()
            return


        for prop in loaded_properties:
            property = PropertyModel(prop.pattern_instance,
                                                prop.name, prop.description)
            #simulate editing of a property in order to obtain instance based properties
            m = PropertyDialogModel(self.model.slim_model, property)
            #Since this is a fake hide the window from the user
            #Therefore the second arguement is True
            v = PropertyDialogView(self.model, True)
            #Since this is a fake trigger the confirm/discard button
            #Therefore the second arguement is True
            c = PropertyDialogCtrl(m, v, self.model, True)
            
            response = v.run()
            if gtk.RESPONSE_APPLY == response:
                self.model.prepend_properties(m)
                self.model.property_counter += 1
                self.model.loaded_properties += 1
        #validate all properties
        self.model.property_changed_signal_emit(self.model.slim_model,None)

    def property_busy_value_change(self, model, old, new):
        """Busy level is the number of activities currently runiing
        in background. If an activity is active, the properties cannot
        be edited. This could otherwise lead to misinterpretations 
        after the runned activity lead to a result."""
        if model is self: return # not interested in introspection
        
        self.view['edit'].set_sensitive(new == 0)
        self.view['delete'].set_sensitive(new == 0)
        self.view['add'].set_sensitive(new == 0)
        return



    
class PropertyDialogCtrl(Controller):
    
    def __init__(self, m, v, properties_model, loading_properties=False):
        Controller.__init__(self, m, v)
        self.component_implementation_in_property = None
        self.loading_properties = loading_properties
        return
    
    def register_view(self, view):
        """ At this point fill widgets with data """

        # my hack - usually the view does not know that there is a controller
        # This is needed because the view creates entries and connects their
        # signals to methods defined here in the controller
        view.ctrl = self
        
        self.delay = DelayedCall(1000, view.change_color_and_balloon)
        
        view['pattern_treeview'].append_column(self.model.pattern_column)
        view['pattern_treeview'].set_model(self.model.filtered_patterns)
        
        view['categories_treeview'].append_column(self.model.categories_column)
        view['categories_treeview'].set_model(self.model.categories_list)
        
        if self.model.old_prop: # we are editing an existing property
            old_prop = self.model.old_prop
            pattern_id = old_prop.pattern_instance.pattern_id
            logic = old_prop.pattern_instance.get_logic()
            # LTL is subsumed by CTL (see PropertyDialogModel constructor)
            if logic == Logic.LTL: logic = Logic.CTL
            path = None
            # 1.) find this pattern in the list and select it
            for row in self.model.filtered_patterns.get_model():
                pattern_model = row[0]
                if pattern_model.id == pattern_id and (pattern_model.logic == logic or pattern_id == "propositional"):
                    path = row.path
                    break
            assert path is not None, str(pattern_id) +" "+ str(logic)
            view['pattern_treeview'].set_cursor(path)
            
            # 2.) display the story
            # is done automatically since
            # view['pattern_treeview'].set_cursor(path) will trigger
            # on_pattern_treeview_cursor_changed which then displays the story
            
            # 3.) fill the story with the contents of the "old" property
            name = old_prop.name
            self.view['name_entry'].set_text(name)
            txt_placeholders = {}
            for placeholder in old_prop.property.pattern_instance:
                placeholder_key = placeholder.attrib["placeholderId"]
                if isinstance(placeholder, StateFormulaPlaceholderInstance):
                    txt_placeholders[placeholder_key] = placeholder.content.text
                else:
                    txt_placeholders[placeholder_key] = placeholder.content

            buffer_pos = self.view['textview'].get_buffer().get_start_iter()
            while not buffer_pos.is_end():
                if not buffer_pos.get_child_anchor():
                    buffer_pos.forward_char()
                else:
                    placeholderId = buffer_pos.get_child_anchor().get_widgets()[0].get_name().split(".")[0]
                    buffer_pos.get_child_anchor().get_widgets()[0].set_text(str(txt_placeholders[placeholderId]))
                    buffer_pos.forward_char()
                    
        if self.loading_properties:
            if self.view['property_confirm'].get_property('sensitive') or \
               self.component_implementation_in_property is None: 
                self.on_property_confirm_clicked(self.view['property_confirm'])
            else:
                #we do not load component based properties that are invalid
                #because we do not know how to generate the instances
                self.on_property_discard_clicked(self.view['property_discard'])
        return

    def register_adapters(self):
        # setup of adapters
        return

    # ---------------------------------------------------------------
    #        Signal handlers
    # ---------------------------------------------------------------
    
    def on_categories_treeview_cursor_changed(self, treeview):
        model, it = treeview.get_selection().get_selected()
        category = model[it][0]
        self.model.choose_patterns(category)
    
    def on_property_confirm_clicked(self, button):
        #TODO: implement editing
        buf = self.view['textview'].get_buffer()
        start, end = buf.get_bounds()
        slice = buf.get_slice(start, end, True)
        txt_placeholders = []
        #TODO: replacing "0xFFFC" with contents of corresponding entry
        #TODO: is not strictly safe because the user could also enter this character
        for anchor,type in self.view.pattern.anchors:
            entry_content = anchor.get_widgets()[0].get_text()
            #slice = slice.replace(u"\uFFFC", entry_content, 1)
            txt_placeholders.append(entry_content)
        short_name = self.view['name_entry'].get_text().decode('utf_8')
        self.model.instantiate(self.view.pattern, txt_placeholders, short_name, slice)
        self.view['edit_property_dialog'].response(gtk.RESPONSE_APPLY)
        self.view['edit_property_dialog'].destroy()
        return                
        
    def on_property_discard_clicked(self, button):
        self.view['edit_property_dialog'].response(gtk.RESPONSE_CANCEL)
        self.view['edit_property_dialog'].destroy()            
        return
    
    def on_pattern_treeview_cursor_changed(self, treeview):
        iter = treeview.get_selection().get_selected()[1]
        
        #TODO: why is iter None when the property view is initially displayed?
        if iter is None: return

        # display the right story in the textview
        iter = self.model.filtered_patterns.convert_iter_to_child_iter(iter)
        pat = self.model.pattern_list.get_value(iter, 0)
        self.view.property_view_story(pat)
        
    def on_entry_changed(self, entry):
        #possibly extend the entry
        if len(entry.get_text()) > entry.get_width_chars():
            entry.set_width_chars(len(entry.get_text()))
        #check whether the placeholder is syntactically correct
        error_lst, impl_name = self._check_entry_syntax(entry)
        #change the base_color of the entry accordingly
        if not error_lst:
            # disable any delayed messages
            self.delay.change_color_and_balloon(None, None)
            # call the non-delayed version directly to make the entry green
            # immediately.
            self.view.change_color_and_balloon(entry, error_lst)
            self._controll_sensivity(entry, True, impl_name)
        else:
            self.delay.change_color_and_balloon(entry, error_lst)
            #make sure the "Save" button is disabled
            self._controll_sensivity(entry, False, impl_name)
        pass

    def on_name_entry_changed(self, entry):
        """Make sure there are no invalid characters (though it is probably 
        impossible to enter non-unicode characters into an entry)
        """
        #TODO:
        try:
            entry.get_text().decode('utf_8')
        except UnicodeDecodeError:
            # It was not a ascii-encoded unicode string
            self.delay.change_color_and_balloon(entry, ["Please use UTF-8 characters only"])
            #make sure the "Save" button is disabled
            self._controll_sensivity(entry, False, None)
        else:
            # It may have been an ascii-encoded unicode string
            # disable any delayed messages
            self.delay.change_color_and_balloon(None, None)
            # call the non-delayed version directly to make the entry green
            # immediately.
            self.view.change_color_and_balloon(entry, [])
            self._controll_sensivity(entry, True, None)

    def on_entry_activated(self, entry):
        # simulate an enter when it is entered in the placeholders and
        # the placeholders are valid
        if self.view['property_confirm'].get_property('sensitive'): 
            self.on_property_confirm_clicked(self.view['property_confirm'])
        return
        

    def _controll_sensivity(self, entry, syntax_valid, impl_name):
        """Controlls whether the save button is sensitive (enabled) or not.
        Therefore the view has a dictionary with pairs entry:bool. Initially
        all boolean values are set to False indicating that the content of an
        entry does not have the proper syntax. Each time an entry is changed
        this method is called and the truth values are updated.
        
        @param   entry:  The reference to an entry that contains the input.
        @type    entry:  gtk.Entry
        @param   syntax_valid:  True iff the entry's content is syntactically ok
        @type    syntax_valid:  boolean
        """
        if impl_name is not None:
            self.component_implementation_in_property = impl_name
        if entry is not None:
            self.view.entries[entry] = (syntax_valid, impl_name)
        #enable button if all entries' contents have valid syntax
        if syntax_valid and all(entry_syntax_valid
               for entry_syntax_valid, impl_name in self.view.entries.values()):
            if all(impl_name is None or
                   impl_name == self.component_implementation_in_property
                   for entry_syntax_valid, impl_name in self.view.entries.values()):
                #make sure all entries are green
                for entry in self.view.entries.keys():
                    self.view.change_color_and_balloon(entry, None)
                #and the confirm button is sensitive
                self.view['property_confirm'].set_sensitive(True)
            else:
                self.delay.change_color_and_balloon(entry, ["You may refer to at most one component implementation throughout a property"])
                self.view['property_confirm'].set_sensitive(False)
        #disable button as long as there is an entry with incorrect input
        else:
            self.view['property_confirm'].set_sensitive(False)
        pass
        
    def _check_entry_syntax(self, entry):
        """Extracts the entry's type and contents and checks whether there are
        errors in the input. The errors are returned in a list.
        
        @param   entry:  The reference to an entry that contains the input.
        @type    entry:  gtk.Entry
        @return: A list containing the errors. This list will be empty if the 
                 input is correct.
        @rtype:  A list of strings
        """
        contents = entry.get_text()
        type = int(entry.get_name().split(".")[1]) #the name is used to store type information
        
        #given the user's input check whether it is syntactically correct
        if type is PlaceholderType.STATE_FORMULA:
            error_lst, impl_name = self._get_errors_in_atomic_proposition_with_templates(
                        contents)
            return [l.__str__() for l in error_lst], impl_name
        elif type is PlaceholderType.TIME_BOUND:
            r = re.compile("^(-?[0-9]+|infinity)[ ]*(msec|sec|min|hour|day)?$")
            m = r.match(contents.strip())
            if not m:
                return ["The timebound is invalid (use integer optionally followed by time unit)"], None
            timebound = m.group(1)
            unit = m.group(2)
            try:
                timebound = int(m.group(1))
                if timebound >= 0:
                    return [], None
                else:
                    return ["Timebound has to be non-negative"], None
            except Exception as e:
                if timebound == "infinity":
                    return[], None
                else:
                    return ["A time bound is expected but the input is not an integer"], None
        elif type is PlaceholderType.THRESHOLD:
            try:
                threshold = float(contents)
                return [], None
            except:
                return ["A threshold is expected but the input is not a float number"], None
        else:
            assert False, "checking unknown PlaceholderType"
        
    def _get_errors_in_atomic_proposition_with_templates(self, 
                                                         atomic_proposition):
        """Validates the specified atomic proposition (in SLIM syntax) POSSIBLY
        CONTAINING IMPLEMENTATION NAMES with respect to the SLIM model with the
        specified root component implementation and returns a list containing the
        detected errors.
        This is done in three steps:
        1.) make sure only one component implementation name appears in the AP
        2.) make sure there is at least one instance of this component implementation
            that is reachable from the root
        3.) instantiate the name with a path to a component and check this using
            get_errors_in_atomic_proposition (for every possible implementation
            instance)
        
        @param   atomic_proposition:  The atomic proposition in SLIM format.
        @type    atomic_proposition:  string
        @return: A list containing the errors in the atomic proposition. This list
                 will be empty if the atomic proposition is valid (not necessarily
                 true) with respect to the SLIM model.
        @rtype:  A list of errors.InvalidAtomicPropositionError objects. 
        """
    
        # 1st: generate a list of strings of the form "package::component.impl"
        # where "package::" is omitted if package = "__default__"
        impl_names = []
        
        for package in self.model.slim_model.compile_info.packages.values():
            for subpackage in [package.private_subpackage, package.public_subpackage]:
                if subpackage.package_name is "__default__":
                    impl_names.extend(subpackage.implementations.keys())
                else:
                    impl_names.extend([subpackage.package_name + "::" + impl_name
                                       for impl_name
                                       in subpackage.implementations.keys()])
        
        # count how many component implementations the AP refers to
        # and store their names
        
        found_impl_names = []
        for impl_name in impl_names:
            # We have to search for the implementation name followed by a dot
            # in the atomic proposition string. Otherwise substrings might be
            # found.
            if (-1 != atomic_proposition.find(impl_name + ".")):
                found_impl_names.append(impl_name)
            if len(found_impl_names) > 1:
                break
        
        if len(found_impl_names) > 1:
            return ["You may refer to at most one component implementation in \
an atomic proposition"], None
        
        root_implementation = self.model.slim_model.get_extended_root_impl()
        if len(found_impl_names) == 1:
            # 2nd: find all instances of the component implementation that is
            # reachable from root. Then check the AP for these instances
            # NOTE: Because of instance based fault injections we have to check
            # ALL reachable instances
            paths = self._get_instances_of(found_impl_names[0])
            if not paths:
                return ["There is no instance of the specified component \
implementation which is reachable from the root component."], found_impl_names[0]
            
            assert len(paths) > 0, paths
            for path in paths:
                instantiated_atomic_proposition = \
                    atomic_proposition.replace(found_impl_names[0]+".", path)
                errors = api.get_errors_in_atomic_proposition(
                                               instantiated_atomic_proposition,
                                               root_implementation)
                if len(errors) > 0:
                    return errors, found_impl_names[0]
            return [], found_impl_names[0]
    
        else: # There are no templates, directly check syntax
            return api.get_errors_in_atomic_proposition(atomic_proposition,
                                                    root_implementation), None

    def _get_instances_of(self, impl_name):
        """returns a list of qualified subcomponent names which are instances
        of a given subcomponent implementation
        """
        root = self.model.slim_model.root_impl
        paths = []
        if root:
            # The queue is a list of pairs
            # ("path" , ComponentImplementation object)
            queue = [("",root)]
    
            while queue != []:
                current_impl = queue[0][1]
                if current_impl.name == impl_name:
                    paths.append(queue[0][0])
                #add the subcomponents of the current subcomponent to the queue
                for id in current_impl.non_data_subcomponents.keys():
                    path = queue[0][0] + str(id) + "."
                    subcmp = current_impl.non_data_subcomponents[id].get_referenced_component_implementation()
                    queue.append((path, subcmp))
                queue.pop(0)
            
        return paths
