# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================


import thread
import gtk
import tempfile
import os
from gtkmvc import Controller
import gobject

from property_management.src.patterns import *
from property_management.src.constants import Logic
from property_management.src.constants import PlaceholderType
from property_management.src.properties import *
import slim_compiler.src.testexpr as text_expr_module
import utils.fault_tree as fault_tree_module
import property_management.src.api as properties_api
from src.ctrls.activity import ActivityCtrl, StartStopActivityCtrl
from utils.misc import relpath
from src.models.filtered_properties import FilteredPropertiesModel            
import slim_compiler.src.globals as globals
from src.models.performability import ToolSettingsModel
from src.ctrls.performability import ToolSettingsCtrl
from utils import exceptions
from slim_compiler.src.errorManagement.exceptions import InvalidExpressionException

class SafetyCtrl (ActivityCtrl):
    """Controller for the Correctness activity. Two sub-controllers
    control model checking and model simulation panels."""
    def __init__(self, model, view):
        ActivityCtrl.__init__(self, 
                              SafetySubCtrl(self, model, view.subview),
                              model, view)

        # this is used to control the type of selection of properties
        self.mode = None  

        self._ft_col_title = "FT"
        return

    def register_view(self, view):
        ActivityCtrl.register_view(self, view)
        
        self.set_mode("ftg")
        tv = view['tv_properties']
        tv.set_model(self.model.filtered_props.get_properties_list())

        # selection
        def __rend_select(col, rend, model, iter):
            col_num = FilteredPropertiesModel.SELECT_SINGLE \
                if self.mode == "ftg" or self.mode == "ftv" or \
                   self.mode == "fte" \
                else FilteredPropertiesModel.SELECT_MULTI
            rend.set_property("active", model[iter][col_num])
            return
        rend1 = gtk.CellRendererToggle()
        rend1.set_property("activatable", True)
        rend1.connect("toggled", self.on_prop_toggled)
        col = gtk.TreeViewColumn('Name', rend1)
        col.set_cell_data_func(rend1, __rend_select)
        # and name
        rend2 = gtk.CellRendererText()
        col.pack_start(rend2)
        col.set_cell_data_func(rend2, lambda c,r,m,i: r.set_property('text', 
                                            m[i][FilteredPropertiesModel.PROPERTY].name))
        tv.append_column(col)

        # Has FT?
        rend = gtk.CellRendererPixbuf()
        col = gtk.TreeViewColumn(self._ft_col_title, rend)
        col.set_cell_data_func(rend, lambda c,r,m,i: r.set_property('stock-id', 
                                                                    gtk.STOCK_YES if 
                                                                    m[i][FilteredPropertiesModel.FT] is not None 
                                                                    else None))
        tv.append_column(col)

        rend = gtk.CellRendererText()
        col = gtk.TreeViewColumn('Formula', rend)
        col.set_cell_data_func(rend, lambda c,r,m,i: r.set_property('text', 
                                           m[i][FilteredPropertiesModel.PROPERTY].description))
        tv.append_column(col)

        # no selection for this tv:
        tv.get_selection().set_mode(gtk.SELECTION_NONE)

        # to process double click that shows the corresponding ft 
        tv.connect('row-activated', self.on_row_activated, self)
        return
        
    def set_mode(self, mode): 
        """Set current modality: fault tree gen, FMEA or fault tolerance eval"""
        if mode == self.mode: return
        assert mode in ("ftg", "fmea", "ftoe", "ftv", "fte")
        self.mode = mode
        self.view['tv_properties'].queue_draw()
        return

    # ---------------------------------------------------------------
    #        Signal handlers
    # ---------------------------------------------------------------    
    def on_prop_toggled(self, rend, path):
        child_path = self.model.filtered_props.get_properties_list().convert_path_to_child_path(path)
        self.model.props_model.toggle_property(child_path, multi=(self.mode in ("fmea", "ftoe")))
        return

    def on_row_activated(self, tv, path, column, who):
        if who is self and column.get_title() == self._ft_col_title:   
            ft = self.model.filtered_props.get_properties_list()[path][FilteredPropertiesModel.FT]
            if ft: self.model.ftg.load_fault_tree(ft[0], ft[1])
            pass
        return

    pass # end of class
# ----------------------------------------------------------------------    


class SafetySubCtrl (Controller):
    """Sub-controller for the safety controller"""

    def __init__(self, parent, model, view):
        Controller.__init__(self, model, view)
                            
        self.parent = parent

        # sub-controllers
        self.fvg = SafetyFtgCtrl(model.ftg, view.ftg)
        self.fmea = SafetyFmeaCtrl(model.fmea, view.fmea)
        self.ftoe = SafetyFtoeCtrl(model.ftoe, view.ftoe)
        self.ftv = SafetyFtvCtrl(model.ftv, view.ftv)
        self.fte = SafetyFteCtrl(model.fte, view.fte)

        # this controller is interested in observing if properties
        # have been selected
        self.model.filtered_props.register_observer(self)
        return

    # ---------------------------------------------------------------
    #        Signal handlers
    # ---------------------------------------------------------------    

    def on_nb_safety_switch_page(self, nb, page, num): 
        assert num >= 0 and num < 5
        
        self.parent.set_mode({0:"ftg", 1:"fmea", 2:"ftoe", 3:"ftv", 4: "fte"}[num])
        return

    # ---------------------------------------------------------------
    #        Notifications from observable properties
    # ---------------------------------------------------------------    
    def property_single_selected_value_change(self, model, old, new):
        self.view.ftg.enable_ftg(new)
        self.view.ftv.enable_ftv(new)
        self.view.fte.enable_fte(new)
        return

    def property_multi_selected_num_value_change(self, model, old, new):
        self.view.fmea.enable_fmea(new > 0)        
        return

    pass # end of class
# ----------------------------------------------------------------------    


class SafetyFtgCtrl (StartStopActivityCtrl):
    
    def __init__(self, model, view):
        StartStopActivityCtrl.__init__(self, model, view, 
                                       self.run_ftg, model.stop_fsap)

        # this controller is interested in observing if the slim
        # model has infinite domain
        self.observe_model(model.slim_model)
        return
    
    def register_adapters(self):
        self.adapt("dynamic", "cb_dyn_ft")
        self.adapt("sat_bound", "ft_sat_bound")
        self.adapt("use_bdd", "ft_use_bdd")
        self.adapt("sat_sbmc", "ft_use_sbmc")
        self.adapt("sbmc_complete", "ft_sbmc_complete")
        return
    
    def run_ftg(self):
        """Makes the model generated a fault tree and then runs the
        fault tree viewer"""
        ev, gt = self.model.generate_fault_tree()
        self.model.load_fault_tree(relpath(ev), relpath(gt))
        return

    # observable properties
    def property_infinite_domain_value_change(self, model, old, new):
        self.model.can_use_bdd = not new
        if (not self.model.can_use_bdd and self.model.use_bdd):
            self.model.use_bdd = False

        self.view.enable_ft_use_bdd(self.model.can_use_bdd)
        return

    # when bdd is used, sat stuff is grayed out
    def property_use_bdd_value_change(self, model, old, new):
        if (not new): self.view.activate_ft_use_sat(True)
        self.view.enable_ft_dynamic(new)
        self.view.enable_ft_sat_bound(not new)
        self.view.enable_ft_use_sbmc((not new) and self.model.can_use_bdd)
        self.view.enable_ft_sbmc_complete((not new) and \
                  self.model.can_use_bdd and self.model.sat_sbmc)
        return

    # completeness check is available only when sbmc is selected
    def property_sat_sbmc_value_change(self, model, old, new):
        self.view.enable_ft_sbmc_complete(new and self.model.can_use_bdd)
        return

    pass # end of class
# ----------------------------------------------------------------------    


class SafetyFmeaCtrl (StartStopActivityCtrl):
    
    def __init__(self, model, view):
        StartStopActivityCtrl.__init__(self, model, view, 
                                       self.run_fmea_table, model.stop_fsap)

        # this controller is interested in observing if the slim
        # model has infinite domain
        self.observe_model(model.slim_model)
        return

    def register_view(self, view):
        StartStopActivityCtrl.register_view(self, view)

        # disable dynamic initially if compact is enabled
        self.view.enable_dynamic(not self.model.compact)

        tv = view['tv_fmea_table']
        tv.set_model(self.model.fmea_table)

        rend = gtk.CellRendererText()
        col = gtk.TreeViewColumn('Num', rend, text=0)
        tv.append_column(col)

        rend = gtk.CellRendererText()
        col = gtk.TreeViewColumn('ID', rend, text=1)
        tv.append_column(col)

        rend = gtk.CellRendererText()
        col = gtk.TreeViewColumn('Failure Model', rend, text=2)
        tv.append_column(col)

        rend = gtk.CellRendererText()
        col = gtk.TreeViewColumn('Failure Effect', rend, text=3)
        tv.append_column(col)        
        return

    def register_adapters(self):
        self.adapt("cardinality")
        self.adapt("dynamic", "cb_dyn_fmea")
        self.adapt("sat_bound", "fmea_sat_bound")
        self.adapt("compact", "cb_compact_fmea")
        return

        
    def run_fmea_table(self):
        """Makes the model generate the FMEA table and show it"""
        fname = self.model.generate_fmea_table()
        res = self.model.load_fmea_table(fname)
        if not res: self.view.set_info("The resulting <i>FMEA table</i> is <b>empty</b>")
        else: self.view.set_info("The resulting <i>FMEA table</i> is presented <b>underneath</b>")
        return


    # observable properties
    def property_infinite_domain_value_change(self, model, old, new):
        self.view.enable_sat(new)
        self.view.enable_dynamic(not new)
        return

    def property_dynamic_value_change(self, model, old, new):
        self.view.enable_compact(not new)
        return

    def property_compact_value_change(self, model, old, new):
        if not self.model.slim_model.infinite_domain: self.view.enable_dynamic(not new)
        return


    pass # end of class
# ----------------------------------------------------------------------    


class SafetyFtoeCtrl (StartStopActivityCtrl):
    
    def __init__(self, model, view):
        StartStopActivityCtrl.__init__(self, model, view, 
                                       self.run_ftoe, model.stop_tool)

        # this controller is interested in observing if properties
        # have been selected
        self.model.filtered_props.register_observer(self)

        # the controller observes the set of properties, because if
        # changed (for example by FTG which produces a FT) it may
        # happen that FTOE becomes available
        self.model.filtered_props.get_properties_list().connect('row-changed', 
                                                                self.on_properties_row_changed)
        
        return
    
    def register_view(self, view):
        StartStopActivityCtrl.register_view(self, view)                

        # sets up the FTOE table 
        tv = view['tv_ftoe']
        tv.set_model(self.model.ftoe_table)
        
        rend = gtk.CellRendererText()
        col = gtk.TreeViewColumn('Cardinality', rend, text=0)
        tv.append_column(col)

        rend = gtk.CellRendererText()
        col = gtk.TreeViewColumn('Number of Cut Sets', rend, text=1)
        tv.append_column(col)        
        return

    # --------------------------------------------------
    # signals
    # --------------------------------------------------

    def run_ftoe(self):
        self.model.run_fault_tolerance_eval()
        return

    def on_properties_row_changed(self, treemodel, path, iter):        
        self.view.enable_ftoe(len(self.model.get_selected_ft()) > 0)
        return

    # --------------------------------------------------
    # properties
    # --------------------------------------------------

    def property_multi_selected_num_value_change(self, model, old, new):
        # enables ftoe only if there are selected FTs
        if new <= 0: self.view.enable_ftoe(False)
        else: self.view.enable_ftoe(len(self.model.get_selected_ft()) > 0)
        return

    pass # end of class
# ----------------------------------------------------------------------    

class SafetyFtvCtrl (StartStopActivityCtrl):
    
    def __init__(self, model, view):
        StartStopActivityCtrl.__init__(self, model, view,
                                        self.run_ftv,
                                        self.stop_ftv)
        # Observe the list of filtered properties to get notified when the
        # selection is changed. 
        self.model.filtered_props.register_observer(self)

        # Observe when a row is changed in the filtered properties list.
        # TODO: Unnecessary?
        self.model.filtered_props.get_properties_list().connect('row-changed', 
                                                                self.on_properties_row_changed)
        
        # Store a mapping of entry elements to their correctness (i.e. True
        # or False), which will be used to determine the correctness of the
        # entered property.
        self.entry_correctness_map = {}
        
        # Store a mapping of entry elements to tuples of types and IDs, which
        # will be used to determine which kind of checking routine has to be
        # applied for a specific entry.
        self.entry_type_id_map = {}
        
        # Store a mapping of entry elements to formulas, which will be used to
        # construct the formula object for the entered property.
        self.entry_formula_map = {}

        # Store a reference of the pattern object that is currently selected.
        self.pattern = None
        self.tool_settings_ctrl = ToolSettingsCtrl(ToolSettingsModel(), view.ftvsub.tool_settings_sub)
        return
        
        
    def register_view(self, view):
        StartStopActivityCtrl.register_view(self, view)
        self.model.reset()
        self.add_patterns_to_combo_box()
     
        return
    
    def add_patterns_to_combo_box(self):
        """Adds the patterns to the combo box of the view.
        """
        combo_box = self.view.ftvsub['combo_patterns']
        combo_box.set_model(self.model.pattern_list)
        cell = gtk.CellRendererText()
        combo_box.pack_start(cell, True)
        combo_box.add_attribute(cell, 'text', 0)
        # Register for a notification on a changed event of the combo box.
        combo_box.connect('changed', self.on_change_pattern)
        
    def on_change_pattern(self, combo_box):
        """Reacts on a changed event of the combo box displaying the
        available patterns.
        """
        # Retrieve the selected pattern object and display the associated story.
        index = combo_box.get_active()
        pattern_obj = self.model.pattern_list[index][1]
        self.pattern = pattern_obj
        self.show_pattern_story(pattern_obj)
        self.view.ftvsub['vbox_pattern_settings'].show_all()
        
        
    def get_text_buffer_for_pattern(self, pattern):
        # TODO: docstring
        
        # Create a text buffer and retrieve a pointer to its last position.
        text_buffer = gtk.TextBuffer()
        last_pos = text_buffer.get_end_iter()
        
        # Create a list of anchors where entry boxes can be inserted and a
        # mapping of anchors to the IDs they represent.
        anchors = []
        anchor_type_id_map = {}
        
        # Fill the text buffer with the parts of the story.
        for part in pattern.get_story():
            if isinstance(part, PlaceholderInfo):
                # Insert an anchor at the last position if the part of the
                # story is to be entered by the user.
                anchor = text_buffer.create_child_anchor(last_pos)
                anchor_type_id_map[anchor] = (part.type, part.id)
                anchors.append(anchor)
            else:
                # Insert plain text at the last position if the part of the
                # story is descriptive text.
                text_buffer.insert(last_pos, " " + part + " ")
        
        return (text_buffer, anchors, anchor_type_id_map)
        
    def show_pattern_story(self, pattern):
        # TODO: docstring
        pattern_view = self.view.ftvsub
        
        # Retrieve a text buffer, a list of anchors and a mapping of anchors to
        # IDs. The anchors mark the places where entry boxes will be inserted
        # and the mapping is used to build another mapping from entry boxes
        # to the respective IDs that they represent.
        (text_buffer, anchors, anchor_type_id_map) = self.get_text_buffer_for_pattern(pattern)
        pattern_view['txtview_pattern'].set_buffer(text_buffer)
        
        # Clear all mappings.
        self.entry_correctness_map = {}
        self.entry_type_id_map = {}
        self.entry_formula_map = {}
        
        # Insert entry boxes at the anchors.
        for anchor in anchors:
            input = gtk.Entry()
            self.entry_correctness_map[input] = False
            self.entry_type_id_map[input] = anchor_type_id_map[anchor]
            #input.set_has_frame(False)
            input.editable = True
            
            # Register for the changed event of the entry such that a check
            # method can be invoked when something is entered.
            input.connect("changed", self.on_placeholder_entry_changed)
            pattern_view['txtview_pattern'].add_child_at_anchor(input, anchor)
        
    def on_placeholder_entry_changed(self, entry):
        # TODO: docstring
        
        # Retrieve the errors in the entry field, shows them if there exist any
        # and check whether all requirements are met to run the activity.
        errors = self.check_entry(entry)
        if errors != [] or entry.get_text() == "":
            self.entry_correctness_map[entry] = False
        else:
            self.entry_correctness_map[entry] = True
        self.show_errors(entry, errors)
        self.check_requirements_met()

    def show_errors(self, entry, errors):
        # TODO: docstring
        if len(errors) > 0:
            if len(entry.get_text()) > 0:
                color = gtk.gdk.color_parse('#ff0000')
                entry.modify_base(gtk.STATE_NORMAL, color)
            else:
                color = gtk.gdk.color_parse('#ffffff')
                entry.modify_base(gtk.STATE_NORMAL, color)                
        else:
            if len(entry.get_text()) > 0:
                color = gtk.gdk.color_parse('#f6f6b8')
                entry.modify_base(gtk.STATE_NORMAL, color)
            else:
                color = gtk.gdk.color_parse('#ffffff')
                entry.modify_base(gtk.STATE_NORMAL, color)
        

    def get_errors_in_event_expression_string(self, entry):
        # TODO: docstring
        retval = []
        root_node = None
        # If there is a fault_tree selected, check whether the entered
        # expression is valid with respect to the event names of the fault tree.
        if self.model.fault_tree is not None:
            # Set flag to allow only boolean expressions that are legal in
            # expressions over fault tree events and parse the expression
            # afterwards. If there are parsing errors, return those, otherwise
            # check for illegal atomic propositions in the formula resulting
            # from the expression tree.
            globals.PARSER_FT_BOOL_EXPRESSIONS_ONLY = True
            try:
                root_node = text_expr_module.parseExpression(entry.get_text())
            except InvalidExpressionException, exception:
                retval = [error.get_message() for error in exception.error_collector.errors]
            finally:
                globals.PARSER_FT_BOOL_EXPRESSIONS_ONLY = False
                if root_node is not None:
                    formula = \
                            properties_api.create_formula_from_slim_expression(
                                                        Logic.CSL, root_node)
                    aps = properties_api.get_atomic_proposition_strings(formula)
                    valid_event_names = fault_tree_module.get_event_names(
                                                        self.model.fault_tree)
                    invalid_aps = aps - valid_event_names
                    if invalid_aps != set():
                        invalid_aps_str = ", ".join(invalid_aps)
                        retval = ["Invalid event names: " + invalid_aps_str + "."]
                    else:
                        self.entry_formula_map[entry] = formula
                        retval = []

                return retval
        else:
            return ["No fault tree given."]
   
    def check_entry(self, entry):
        # TODO: docstring
        
        # Retrieve the type associated to the entry.
        (type, id) = self.entry_type_id_map[entry]
        text = entry.get_text()
        result = []
        
        # Check the correctness of the text with respect to its type.
        if type is PlaceholderType.STATE_FORMULA:
            if text != "":
                result = self.get_errors_in_event_expression_string(entry)
        elif type is PlaceholderType.TIME_BOUND:
            # Try to convert the time bound to an integer. Return errors if
            # the conversion fails or if time bound is negative.
            try:
                timebound = int(text)
                if timebound < 0:
                    result = ["Timebound must not be non-negative."]
            except:
                result = ["A time bound is expected but the input is not an integer"]
                return result
        elif type is PlaceholderType.THRESHOLD:
            assert False
        else:
            assert False
            
        return result

    def get_property_is_correct(self):
        """Returns whether the entered property is correct, that is, if all
        entry boxes have a legal value.
        """
        return (self.entry_correctness_map.values() != [] and \
                            False not in self.entry_correctness_map.values())

    def get_formula(self):
        property = self.get_property()
        formula = properties_api.create_formula_for_property(property)
        return formula

    def get_property(self):
        if self.get_property_is_correct():
            pattern_instance = PatternInstance.create(Logic.CSL,
                                                      self.pattern.id)
            for entry, (type, id) in self.entry_type_id_map.iteritems():
                if type == PlaceholderType.STATE_FORMULA:
                    placeholder_instance = \
                            StateFormulaPlaceholderInstance.create(
                                    Logic.CSL, id,
                                    self.entry_formula_map[entry].content)
                elif type == PlaceholderType.TIME_BOUND:
                    placeholder_instance = \
                                    TimeBoundPlaceholderInstance.create(
                                            Logic.CSL, id, int(entry.get_text()))
                else:
                    assert False
            
                pattern_instance.add_placeholder_instance(placeholder_instance)
                
            property = Property.create(pattern_instance, "tmp_prop")            
            return property

    def get_time_bounds(self):
        # TODO: docstring
        if self.get_property_is_correct():
            lower = 0
            upper = 0
            # Retrieve the lower and upper time bound from the entry boxes. 
            for entry, (type, id) in self.entry_type_id_map.iteritems():
                if type == PlaceholderType.TIME_BOUND:
                    if id == "Time1":
                        lower = int(entry.get_text())
                    elif id == "Time2":
                        upper = int(entry.get_text())
                    else:
                        assert False
                        
            return (lower, upper)    

    def redraw(self):
        res = self.model.analysis_result
        if res == None:
            return #nothing to do
        if res.has_stepwise_results():
            # Draw the resulting lines on the canvas.
            xpoints, ypoints = (res.probability_steps[0],[res.probability_steps[1]])
            bounds = self.get_time_bounds()
            for ypoints_ in ypoints:
                self.view.draw(bounds, xpoints, ypoints_)
    
            if ypoints != [] and ypoints[0] != []:
                self.view.display_percentage(ypoints[0][len(ypoints[0]) - 1])
        elif res.is_et_analysis():
            value = res.expected_time[0] if res.expected_time[0] == res.expected_time[1] else res.expected_time
            self.view.display_et(value)
        elif res.is_lra_analysis():
            value = res.lra[0] if res.lra[0] == res.lra[1] else res.lra
            self.view.display_lra(value)
        else:
            raise "This should be unreachable"
    
    def check_requirements_met(self):
        if self.model.fault_tree is not None:
            enable = self.get_property_is_correct()
        else:
            enable = False
        self.view.enable_ftv(enable)
        
    def check_all_entries(self):
        for entry in self.entry_correctness_map.keys():
            self.on_placeholder_entry_changed(entry)
        
        
    # --------------------------------------------------
    # signals
    # --------------------------------------------------
    def run_ftv(self):
        try:
            basic_event_rates_file = self.model.get_basic_event_rates_file()
            if basic_event_rates_file is None:
                raise exceptions.UserException("There was at least one event " + \
                                    "for which no rate is specified. "+ \
                                    "Fault Tree Verification cannot " + \
                                    "proceed.")
            self.model.run_dft_verification(self.model.get_selected_ft(),
                                            basic_event_rates_file,
                                            self.get_property(), self.tool_settings_ctrl.model)
            gobject.idle_add(self.redraw)
            #self.redraw()
        except (exceptions.UserException) as e:
            self.stop_ftv()
            gobject.idle_add(self.display_error, str(e))

        
    def display_error(self, message):
        msgbox = gtk.MessageDialog(type=gtk.MESSAGE_ERROR,
                           buttons=gtk.BUTTONS_OK,
                           message_format=message)
        msgbox.run()
        msgbox.destroy()
        
    def stop_ftv(self):
        self.model.stop_dft_verification()
        return
        
    def on_properties_row_changed(self, treemodel, path, iter):
        self.model.reset()
        ft_filename = self.model.get_selected_ft()
        if ft_filename is not None:
            self.model.load_ft(ft_filename)
        self.check_all_entries()
        self.check_requirements_met()
        return
    
    # ---------------------------------------------------------------
    #        Notifications from observable properties
    # ---------------------------------------------------------------            
    def property_single_selected_value_change(self, model, old, new):
        self.model.reset()
        ft_filename = self.model.get_selected_ft()
        if ft_filename is not None:
            self.model.load_ft(ft_filename)
        self.check_all_entries()
        self.check_requirements_met()
        return
    
    pass # end of class
# ----------------------------------------------------------------------    

class SafetyFteCtrl (StartStopActivityCtrl):
    
    def __init__(self, model, view):
        StartStopActivityCtrl.__init__(self, model, view,
                                        self.run_fte,
                                        self.stop_fte)
        # we also observe the number of selected properties in the
        # properties model, to enable the run button accordingly
        self.model.filtered_props.register_observer(self)

        self.model.filtered_props.get_properties_list().connect('row-changed', 
                                                                self.on_properties_row_changed)
        self.entry_correctness_map = {}
        self.tool_settings_ctrl = ToolSettingsCtrl(ToolSettingsModel(), view.tool_settings_sub)
        return

    def register_view(self, view):
        StartStopActivityCtrl.register_view(self, view)
        
#        lower = self.view['entry_lower_time_bound']
#        lower.connect("changed", self.on_time_bound_changed)
#        self.entry_correctness_map[lower] = False
        upper = self.view['entry_upper_time_bound']
        upper.connect("changed", self.on_time_bound_changed)
        self.entry_correctness_map[upper] = False
        criticality = self.view['entry_criticality']
        criticality.connect("changed", self.on_criticality_changed)
        self.entry_correctness_map[criticality] = False
        
        return

    def check_is_int(self, entry):
        result = []
        try:
            value = int(entry.get_text())
        except ValueError:
            result.append("Value must be an integer.")
        finally:
            return result

    def get_entries_correct(self):
        return False not in self.entry_correctness_map.values()

    def show_errors(self, entry, errors):
        self.entry_correctness_map[entry] = False               
        if len(errors) > 0:
            if len(entry.get_text()) > 0:
                color = gtk.gdk.color_parse('#ff0000')
                entry.modify_base(gtk.STATE_NORMAL, color)
            else:
                color = gtk.gdk.color_parse('#ffffff')
                entry.modify_base(gtk.STATE_NORMAL, color) 
        else:
            if len(entry.get_text()) > 0:
                color = gtk.gdk.color_parse('#f6f6b8')
                entry.modify_base(gtk.STATE_NORMAL, color)
                self.entry_correctness_map[entry] = True               
            else:
                color = gtk.gdk.color_parse('#ffffff')
                entry.modify_base(gtk.STATE_NORMAL, color)
        
    def check_requirements_met(self):
        enable = False
        if self.model.fault_tree is not None:
            enable = self.get_entries_correct()
        self.view.enable_fte(enable)

    # --------------------------------------------------
    # signals
    # --------------------------------------------------
    def run_fte(self):
        try:
            basic_event_rates_file = self.model.get_basic_event_rates_file()
            if basic_event_rates_file is None:
                raise exceptions.UserException("There was at least one event " + \
                                    "for which no rate is specified. "+ \
                                    "Fault Tree Evaluation cannot " + \
                                    "proceed.")
                
            # Create two temporary files, which will be used as containers for
            # the gates and events file of the FT+ format.
        
            (gates_file_descriptor, gates_filename) = \
                                    tempfile.mkstemp(suffix=".gates", text=True)
            (events_file_descriptor, events_filename) = \
                                    tempfile.mkstemp(suffix=".events", text=True)
        
            os.close(gates_file_descriptor)
            os.close(events_file_descriptor)
        
            mission_time = int(self.view['entry_upper_time_bound'].get_text())
            criticality = int(self.view['entry_criticality'].get_text())
            
            self.model.run_dft_evaluation(
                self.model.fault_tree, basic_event_rates_file,
                mission_time, criticality, events_filename, gates_filename, self.tool_settings_ctrl.model)

            self.model.load_fault_tree(events_filename, gates_filename)
        
        except (exceptions.UserException) as e:
            self.stop_fte()
            gobject.idle_add(self.display_error, str(e))

        
    def display_error(self, message):
        msgbox = gtk.MessageDialog(type=gtk.MESSAGE_ERROR,
                           buttons=gtk.BUTTONS_OK,
                           message_format=message)
        msgbox.run()
        msgbox.destroy()
        
    def stop_fte(self):
        self.model.stop_dft_evaluation()
        return

    def on_criticality_changed(self, entry):
        text = entry.get_text()
        errors = self.check_is_int(entry)
        self.show_errors(entry, errors)
        self.check_requirements_met()

    def on_time_bound_changed(self, entry):
        text = entry.get_text()
        errors = self.check_is_int(entry)
        self.show_errors(entry, errors)
        self.check_requirements_met()

    def on_properties_row_changed(self, treemodel, path, iter):
        self.model.reset()
        ft_filename = self.model.get_selected_ft()
        if ft_filename is not None:
            self.model.load_ft(ft_filename)
        self.check_requirements_met()       
        return
    
    # ---------------------------------------------------------------
    #        Notifications from observable properties
    # ---------------------------------------------------------------            
    def property_single_selected_value_change(self, model, old, new):
        self.model.reset()
        ft_filename = self.model.get_selected_ft()
        if ft_filename is not None:
            self.model.load_ft(ft_filename)
        self.check_requirements_met()
        return
    
    pass # end of class
# ----------------------------------------------------------------------    
