# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================



from gtkmvc import Controller
from gtkmvc.adapters import Adapter
import gtk

import os

from utils.misc import relpath

from src.models.fault_injection import FaultInjectionModel
from src.ctrls.fault_injection import FaultInjectionCtrl
from src.views.fault_injection import FaultInjectionView
from src.ctrls.console import ConsoleCtrl


# Controller for the slim model panel
class SlimCtrl (Controller):
    """Controls slim models and fault injection panels"""
    
    def __init__(self, m, v):
        Controller.__init__(self, m, v)

        # sub controllers
        self.console = ConsoleCtrl(m.console, v)
        
        # register observer
        self.model.register_observer(self.console)
        return

    def register_view(self, view):
        # sets up all treeviews
        self._setup_treeviews(view)

        # to control selection of slim files
        sel = self.view['tv_slim_models'].get_selection()
        sel.set_mode(gtk.SELECTION_MULTIPLE)
        sel.connect('changed', self.on_slim_models_selection_changed)

        # to control selection of injections
        sel = self.view['tv_fault_injections'].get_selection()
        sel.set_mode(gtk.SELECTION_MULTIPLE)
        sel.connect('changed', self.on_injections_selection_changed)

        return


    def _setup_treeviews(self, view):
        """This is used to setup all treeviews (columns and
        renderers). It exists to make register_view more
        readable"""
        # setting the foreground colours of the SLIM files based
        # on validity
        def __set_slim_file_foreground_color(column, cell, model, iter):
            if model[iter][2]: cell.set_property("foreground", "black")
            else: cell.set_property("foreground", "gray")
            return        

        # used to show filenames as relative paths or absolute,
        # whatever is the shortest
        def __fname(col, rend, model, iter, col_num):
            fname = model.get_value(iter, col_num)
            relfname = relpath(fname)
            if len(fname) < len(relfname): rend.set_property('text', fname)
            else: rend.set_property('text', relfname)
            __set_slim_file_foreground_color(col, rend, model, iter)
            return
            
        # used to show the name of an implementation within a treecolumn
        def __impl_data(col, rend, model, iter, col_num):
            impl = model.get_value(iter, col_num)
            rend.set_property('text', impl.subpackage.package_name + "::" + impl.name)
            return

        # ---- list of slim files ---------------------------------------------
        # sets up the treeview for the slim models
        tv = view['tv_slim_models']
        tv.set_model(self.model.slim_files)        
        
#        col = gtk.TreeViewColumn('Use')
#        rend = gtk.CellRendererToggle()
#        rend.connect("toggled", self.on_toggle_fault_injection)
#        col.pack_start(rend)
##        col.add_attribute(rend, "active", 1)
#        col.add_attribute(rend, "activatable", 2)
#        tv.append_column(col)        
    
        rend = gtk.CellRendererText()
        col = gtk.TreeViewColumn('Filename', rend)
        col.set_cell_data_func(rend, __fname, 0)
        tv.append_column(col)


        # ---- roots ----------------------------------------------------------
        # sets up the treeview of the root implementations
        tv = view['tv_roots']
        tv.set_model(self.model.roots)

        col = gtk.TreeViewColumn('Implementation')
        # root selected
        rend1 = gtk.CellRendererToggle()
        rend1.set_property("activatable", True)
        rend1.connect("toggled", self.on_root_toggled)
        col.pack_start(rend1)
        col.add_attribute(rend1, "active", 2)
        # root name
        rend2 = gtk.CellRendererText()
        col.pack_start(rend2)
        col.set_cell_data_func(rend2, __impl_data, 0)
        tv.append_column(col)

        # this is for the filename containing the root
        rend = gtk.CellRendererText()
        col = gtk.TreeViewColumn('Filename', rend)
        col.set_cell_data_func(rend, __fname, 1)
        tv.append_column(col)
        
        # this is for root selection 
        sel = tv.get_selection()
        sel.set_mode(gtk.SELECTION_NONE)
        
        # ---- fdir components ----------------------------------------------
        # set up the treeview of fdir components
        tv = view['tv_fdir']
        tv.set_model(self.model.fdir_components)
        
        col = gtk.TreeViewColumn('Implementation')
        # fdir component selection
        rend1 = gtk.CellRendererToggle()
        rend1.set_property("activatable", True)
        rend1.connect("toggled", self.on_fdir_toggled)
        col.pack_start(rend1)
        col.add_attribute(rend1, "active", 2)
        # fdir component name
        rend2 = gtk.CellRendererText()
        col.pack_start(rend2)
        col.set_cell_data_func(rend2, __impl_data, 0)
        tv.append_column(col)
        
        
        # this is for the filename containing the root
        rend = gtk.CellRendererText()
        col = gtk.TreeViewColumn('Filename', rend)
        col.set_cell_data_func(rend, __fname, 1)
        tv.append_column(col)
        
        # this is for root selection 
        sel = tv.get_selection()
        sel.set_mode(gtk.SELECTION_NONE)
        
        # ---- fault injections ----------------------------------------------
        tv = view['tv_fault_injections']
        tv.set_model(self.model.fault_injections)
        
        # setting the foreground colours of the fault injections based
        # on validity
        def __set_fault_injection_foreground_color(column, cell, model, iter):
            if model[iter][1]: cell.set_property("foreground", "black")
            else: cell.set_property("foreground", "gray")
            return        
        
        # enable box
        col = gtk.TreeViewColumn('Use')
        rend = gtk.CellRendererToggle()
        rend.connect("toggled", self.on_toggle_fault_injection)
        col.pack_start(rend)
        col.add_attribute(rend, "active", 0)
        col.add_attribute(rend, "activatable", 1)
        tv.append_column(col)
        
        # error implementation
        def __error_implementation(col, rend, model, iter):
            impl = model.get_value(iter, 2)
            rend.set_property('text', impl[0] + "::" + impl[1])
            __set_fault_injection_foreground_color(col, rend, model, iter)
            return
        rend = gtk.CellRendererText()
        col = gtk.TreeViewColumn('Error Implementation', rend)
        col.set_cell_data_func(rend, __error_implementation)
        tv.append_column(col)
        
        # error state
        def __error_state(col, rend, model, iter):
            error_state = model.get_value(iter, 3)
            rend.set_property('text', error_state)
            __set_fault_injection_foreground_color(col, rend, model, iter)            
            return
        rend = gtk.CellRendererText()
        col = gtk.TreeViewColumn('Error State', rend)
        col.set_cell_data_func(rend, __error_state)
        tv.append_column(col)
        
        # effect (instance + data port + effect combined)
        def __effect(col, rend, model, iter):
            component_instance = model.get_value(iter, 4)
            data = model.get_value(iter, 5)
            effect = model.get_value(iter, 6)
            component_instance_str = ".".join(component_instance)            
            rend.set_property('text', component_instance_str 
                                        + "." + data 
                                        + " := " + effect)
            __set_fault_injection_foreground_color(col, rend, model, iter)                                        
            return        
        rend = gtk.CellRendererText()
        col = gtk.TreeViewColumn('Effect', rend)
        col.set_cell_data_func(rend, __effect)
        tv.append_column(col)
        
        
        # set color based on validity
        def foreground_color(column, cell, model, iter, user_data=None):
            if model[iter][PropertiesModel.USE]:
                cell.set_property("foreground", "black")
            else:
                cell.set_property("foreground", "gray")
                
        return

    def load_slim_files(self):
        """This is called when the user wants to open one or more slim files"""
        # Opens the file chooser, and possibly adds the slim files
        # to the model
        
        chooser = gtk.FileChooserDialog(title="Open SLIM Files",
                                        action=gtk.FILE_CHOOSER_ACTION_OPEN,
                                        buttons=(gtk.STOCK_CANCEL,gtk.RESPONSE_CANCEL,
                                                 gtk.STOCK_OPEN,
                                                 gtk.RESPONSE_OK))

        chooser.set_select_multiple(True)
        filter = gtk.FileFilter()
        filter.set_name("SLIM Files")
        filter.add_pattern("*.slim")
        chooser.add_filter(filter)
        filter_all = gtk.FileFilter()
        filter_all.set_name("All Files")
        filter_all.add_pattern("*")
        chooser.add_filter(filter_all)
        chooser.set_current_folder(os.path.dirname(
            self.model.generate_unique_fault_injection_basename() + ".fixml"))
            
        response = chooser.run()
        filenames = chooser.get_filenames()
        chooser.destroy()
        if response == gtk.RESPONSE_OK:
            try:
                if not self.model.add_slim_models(filenames):
                    self.console.expand_output()                
                    pass            
            except IOError, e:    
                err_dia = gtk.MessageDialog(parent=chooser,
                                            buttons=gtk.BUTTONS_CLOSE,
                                            flags=gtk.DIALOG_MODAL,
                                            type=gtk.MESSAGE_ERROR)
                err_dia.set_markup("<big><b>File not found</b></big>")
                err_dia.format_secondary_markup(str(e))
                err_dia.run()
                err_dia.destroy()
                pass
            pass

        
        return


    def save_fault_injections(self):
        """ Save a fault injection document to file with extension .fixml"""

        dialog = gtk.FileChooserDialog("Save Fault Injections",
                                       None,
                                       gtk.FILE_CHOOSER_ACTION_SAVE,
                                       (gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL,
                                       gtk.STOCK_SAVE, gtk.RESPONSE_OK))
        dialog.set_default_response(gtk.RESPONSE_OK)
        filter = gtk.FileFilter()
        filter.set_name("COMPASS Fault Injections")
        filter.add_pattern("*.fixml")
        dialog.add_filter(filter)
        dialog.set_current_name(
            self.model.generate_unique_fault_injection_basename() + ".fixml")
        dialog.set_current_folder(os.path.dirname(
            self.model.generate_unique_fault_injection_basename() + ".fixml"))
        dialog.set_do_overwrite_confirmation(True)
        dialog.show()
        response = dialog.run()
        filename = dialog.get_filename()
        dialog.destroy()
        if response == gtk.RESPONSE_OK:
            name = filename
            fd = open(filename, 'w')
            fd.write(str(self.model.get_fault_injection_document()))
            fd.close()

        return


    def load_fault_injections(self):
        """This is called when the user wants to load a fault injection file"""
        # Opens the file chooser, and possibly adds the slim files
        # to the model
        
        chooser = gtk.FileChooserDialog(title="Open Fault Injections",
                                        action=gtk.FILE_CHOOSER_ACTION_OPEN,
                                        buttons=(gtk.STOCK_CANCEL,gtk.RESPONSE_CANCEL,
                                                 gtk.STOCK_OPEN,
                                                 gtk.RESPONSE_OK))
        chooser.set_select_multiple(False)
        filter = gtk.FileFilter()
        filter.set_name("COMPASS Fault Injections")
        filter.add_pattern("*.fixml")
        chooser.add_filter(filter)
        filter_all = gtk.FileFilter()
        filter_all.set_name("All Files")
        filter_all.add_pattern("*")
        chooser.add_filter(filter_all)
        chooser.set_current_folder(os.path.dirname(
            self.model.generate_unique_fault_injection_basename() + ".fixml"))

        response = chooser.run()
        filename = chooser.get_filename()
        chooser.destroy()
        if response == gtk.RESPONSE_OK:
            res = self.model.load_fault_injections(filename)
            if res is not None:
                self.console.expand_output()
                pass
            pass

        return
        
                

    def open_fault_injection_window(self):
        """This is called when the user wants to do one fault injection"""
        
        m = FaultInjectionModel(self.model.fault_injections, self.model.compile_info, self.model.root_impl)
        v = FaultInjectionView(parent=self.view)
        c = FaultInjectionCtrl(m,v)
        
        response = v.run()
        
        if gtk.RESPONSE_APPLY == response:
            self.model.add_fault_injection(m)
        
        return
    
    
    
        
    
    # ---------------------------------------------------------------
    #        Signal handlers
    # ---------------------------------------------------------------    
    def _clear_selections(self):    
        sel = self.view['tv_slim_models'].get_selection()
        sel.unselect_all()
        sel = self.view['tv_fault_injections'].get_selection()        
        sel.unselect_all()
        sel = self.view['tv_fdir'].get_selection()
        sel.unselect_all()
        return
                
    
    def on_root_toggled(self, rend, path):
        self.model.toggle_root(path)
        self._clear_selections()
        return
    
    def on_fdir_toggled(self, rend, path):
        self.model.toggle_fdir(path)
        self._clear_selections()
        return

    def on_button_add_clicked(self, button):
        """Add slim models has been clicked"""
        self.load_slim_files()
        self._clear_selections()        
        return


    def on_button_remove_clicked(self, button):
        """Remove slim files button"""
        # Retrieve selected slim files
        sel = self.view['tv_slim_models'].get_selection()
        model, paths = sel.get_selected_rows()
        fnames = [model[path][0] for path in paths]
        self.model.remove_slim_models(fnames)
        self._clear_selections()        
        return
    
    
    def on_button_reload_clicked(self, button):
        """remove and add the slim files, i.e., reload"""
        self.model.reload_slim_models()
        self._clear_selections()        
        return
        

    def on_slim_models_selection_changed(self, selection):
        """The selection of slim models has been changed. IFF one
        or more files are selected, button "Remove" must be 
        sensitive"""

        self.view['button_remove'].set_sensitive(
            selection.count_selected_rows() > 0)
        return


    def on_toggle_fault_injection(self, rend, path ):
        self.model.toggle_fault_injection(path)        
        return
        
        
    def on_button_add_injection_clicked(self, button):
        self.open_fault_injection_window()
        self._clear_selections()
        if len(self.model.fault_injections) > 0:
            # set the selection to the most recently added element
            self.view['tv_fault_injections'].set_cursor((len(self.model.fault_injections) - 1,))
        return


    def on_button_remove_injection_clicked(self, button):
        sel = self.view['tv_fault_injections'].get_selection()
        model, paths = sel.get_selected_rows()
        self.model.remove_failures(paths)
        self._clear_selections()        
        return

        
    def on_injections_selection_changed(self, selection):
        """The selection of injections models has been changed. IFF one
        or more files are selected, button "Remove" must be 
        sensitive"""
        self.view['button_remove_injection'].set_sensitive(
            selection.count_selected_rows() > 0)
        
        if self.model.root_impl is not None:
            """ In addition, the clone button has to become enabled when one 
            injection is selected and that injection has clones."""
            _, paths = selection.get_selected_rows()
            self.view['button_clone_injection'].set_sensitive(
                    selection.count_selected_rows() == 1  and
                    len(self.model.get_cloneable_fault_injections(paths[0])) > 0)
        return


    def on_button_clone_injection_enter(self, button):
        """When a mouse pointer is hovered over the clone button,
        we show the list of cloned injections in the fault injection 
        screen."""
        # get the selected fault injection
        sel = self.view['tv_fault_injections'].get_selection()
        model, paths = sel.get_selected_rows()
        if sel.count_selected_rows() != 1:
            return # do nothing if the selection is not a singleton
        path = paths[0]
        
        # get the list of cloneable fault injections associated with it
        self.model.get_cloneable_fault_injections(path)
        return
        
    
    def on_button_clone_injection_leave(self, button):
        """When the mouse pointer leaves the hover, the list of 
        cloned injections are gone. Unless the clone button is clicked,
        then the injections are added"""
        return
        
        
    def on_button_clone_injection_clicked(self, button):
        """When the user clicks on clone, the cloned injections are 
        made permanent by adding them to fault_injections"""
        # get the selected fault injection
        sel = self.view['tv_fault_injections'].get_selection()
        model, paths = sel.get_selected_rows()
        path = paths[0]
        
        # add the list of cloneable fault injections associated with it
        self.model.add_clones_of_fault_injection(path)

        # deactivate the clone button
        self.view['button_clone_injection'].set_sensitive(False)
        self._clear_selections()        
    
        return
    
    # ---------------------------------------------------------------
    #        Notifications from observable properties
    # ---------------------------------------------------------------
    def property_slim_files_num_value_change(self, model, old, new):
        """This notification is called when the number of loaded
        slim files (into the observed slim model) is changed. It is
        used to enable/disable some features based on the presence
        of at least one slim model."""

        self.view['button_add_injection'].set_sensitive(self.model.root_impl is not None and new > 0)
        self.view['button_reload'].set_sensitive(len(self.model.slim_files) > 0)        
        return        

    def property_root_impl_value_change(self, model, old, new):
        """This notification is called when the root implementation 
        is changed. It is used to disable/enable fault injections, 
	because they are dependent on the root implementation."""
        self.view['button_add_injection'].set_sensitive(len(self.model.slim_files) > 0 and new is not None)
        return        

    pass # end of class
