# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================


from gtkmvc import Controller
from src.models.trace_filter import TraceFilter,TraceFilters
from src.models.trace import TraceModel,TraceViewModel            
from src.models.trace import ExtensibleTraceModel
from src.models.correctness import SimResultModel
from src.models.guided import GuidedSimTransitionModel

from slim_compiler.src.dataStructure.nominalbehaviour import DEFAULTINITMODE

import src.views.trace
import src.utils.globals
import gtk,os,xml,re,gobject
from slim_compiler.src.dataStructure.visitors.printVisitor import toSLIMVisitor
import slim_compiler.src.smv.api as smvapi

#the name of the empty filter
QFILTER = "No Filter"

class FilterBarCtrl(Controller):
    """this class handle the trace_filter_bar view"""
    def __init__(self, model, view, filterlist=src.utils.globals.PERMANENT_FILTERS):
        Controller.__init__(self, model, view)
        assert isinstance( model, TraceViewModel ), "The model must be a TraceViewModel"

        #stored filters
        self.filters = model.get_filters()
        self.filterlist = filterlist

        #quick filter
        self.qfilter = TraceFilter()
        self.qfilter.set_filter_name(QFILTER)
        self.filters.add_filters( self.qfilter )

                #filter modify/create view
        self.filterCtrl = FilterCtrl( self.qfilter, 
                                      view.filterView,
                                      self.filters,
                                      self,
                                      [self.qfilter]) #in order to update the values in the treeview
                                     #self.view['CBStoredFilters'])

        return

    def set_model(self, model):
        """
        change the model of the filter bar controller
        @param the new model, (a TraceViewModel instance)
        """
        self.model = model        
        return

    #if model change
    def property__active_trace_value_change(self, model, old, new):
        """
        This method is called when the model changes.
        Do not call it directly.
        """
        #self.qfilter.set_names(['m*'])
        #self.model.set_active_filter(self.qfilter)
        #self.apply_quick_filter(self.view['TTBApply'])
        #return #to remove
        cb = self.view['CBStoredFilters']
        cb_val = cb.get_active()
        tb = self.view['TTBApply']
        
        #apply the filter (priority to the quick filter)
        if tb.get_active():
            self.apply_quick_filter(self.view['TTBApply'])
            pass
        elif cb_val != -1 and cb_val: #0 is the index of no filter
            self.apply_stored_filter(self.view['CBStoredFilters'])
            pass
        
        return

    def add_filter(self, f):
        """
        add a filter in the list of filters.
        @param a TraceFilter instance.
        """
        self.filters.add_filters( f )
        return

    def delete_filter(self, filtername=None):
        """
        as default we delete the quick filter (if filtername is none)
        """
        if not filtername:
            f = self.filters.delete_filter(QFILTER)
        else:
            f = self.filters.delete_filter(filtername)
            pass
        return f
    
    def load_filters(self,fname=None):
        """
        load a filter list from an xml file
        """
        if not fname:
            self.filters.from_xml( self.filterlist )
        else:
            self.filters.from_xml( fname )
            pass
        return
    
    def save_filters(self, where=None, ex=None):
        """
        @param ex, the list of filters that you don't want to save. 
        If none specified, all the filters will be saved, except the "on the fly" filter.
        @param where, the filepath in which you want to save the filters (in xml format)
        """
        
        if not where:                              # if ex is not defined, do not save the quick filter
            self.filters.write_xml(self.filterlist, ex if ex else [self.qfilter])
        else:                             # if ex is not defined, do not save the quick filter
            self.filters.write_xml(where, ex if ex else [self.qfilter])
            pass
        return

    def clear_filters(self):
        """
        delete all the filters handled.
        """
        # clear the present list of filters
        self.filters.clear()
        # readd qfilter
        self.filters.add_filters( self.qfilter )
        return
    
    def setup_filterlist(self, view):
        """
        fill the combobox with the filter list, in the xml passed at init-time.
        """
        
        cb = view['CBStoredFilters']
        if not cb: #if top is evbox
            return
        
        self.clear_filters()
        #read the filter list, and fill the comboBox
        self.filters.from_xml( self.filterlist )
        # self.filters mantains a ListStore, with a list of the names of the filters
        cb.set_model( self.filters.get_ListStore() ) 
        cell = gtk.CellRendererText()
        cb.pack_start(cell,True)
        cb.add_attribute(cell, 'text', 0)
        cb.set_active(0)
        
        return
    
    def register_view(self, view):
        Controller.register_view(self,view)
        self.setup_filterlist(view)
        return

    #signal handler
    # FILTER CALLBACK
    def qfilter_changed(self, editable, reload_view=None):
        """
        Called when the quick filter changes
        """
        toggle = self.view['TTBApply']
        if toggle.get_active():
            toggle.set_active(False) #untoggle!
            pass
        
        self.model.unset_filter()
        return

    def apply_quick_filter(self, widget, reload_view=None ):
        """
        Called when the "enter" button on the keyboard is pressed, or the button "apply" is clicked
        """
        # warning! this method is used if the 'enter' button is pressed down in the entry for the quick filter..
        def apply_filter():
            errors = self.qfilter.set_names(regex)
            if errors:
                self.view.errors_tooltip(errors)
            else:
                self.view.valid_tooltip()
                pass
            
            self.model.set_active_filter(self.qfilter) #redesign!!
            return
        
        toggle_button = self.view['TTBApply']
        text_edit = self.view['TENameFilter']
        cb = self.view['CBStoredFilters']
        cb.set_active(0) #reset combobox
        regex = text_edit.get_text().split(",")

        # if widget is an entry --> if the the user presses the button enter on the keyboard
        if isinstance( widget, gtk.Entry ):
            if not toggle_button.get_active() and any(regex):
                toggle_button.set_active(True)
            elif not regex:
                toggle_button.set_active(False)
                pass
            if regex:
                apply_filter()
            else:
                self.model.unset_filter()
                pass
            return self.qfilter
                
        # if the user clics the button apply
        if toggle_button.get_active(): # toggled
            if any(regex):
                apply_filter()
            else:
                toggle_button.set_active(False) # untoggle!
        else: # untoggled
            self.model.unset_filter()
        
        return self.qfilter

    def apply_stored_filter(self, cb, reload_view=None ):
        
        sfname = cb.get_active_text() #get selected value
        tb = self.view['TTBApply']
        
        if sfname: 
            f = self.filters.get_filter( sfname )
            if tb.get_active():
                tb.set_active(False) #untoggle! if toggled
                pass
            if f != self.qfilter: #quick filter can't be applied here
                self.model.set_active_filter( f ) #redesign!!
                cb.set_active(
                    #the index of the filter f
                    self.model.get_filters().get_filter( f.get_filter_name(), index=True ) 
                    )
                pass
                
            else:
                self.model.unset_filter()
                cb.set_active(0) #reset
                pass
            
        return sfname
    
    #handle modfy/delete buttons. they open a new window
    def create_filter_dialog(self,menu_item):
        
        f = self.model.get_active_filter()
        editFilterName = False #edit the qfilter

        if not f or f == self.qfilter:
            f = TraceFilter()
            f.set_names(self.view['TENameFilter'].get_text()) #copy the qfilter names
            editFilterName = True
            pass

        # the active filter to modify/create
        if self.filterCtrl:
            self.filterCtrl.set_model(f)
            #say if is possible to modify the filter name
            self.filterCtrl.modify_filter_name(editFilterName)
            self.filterCtrl.set_combo_box( self.view['CBStoredFilters'] )
            pass

        self.view.filterView.show() #show the filter dialog
        return

    def remove_filter(self, btn):
        
        cb = self.view['CBStoredFilters']
        i = cb.get_active_iter()
        if not i: return
        fdelete = cb.get_model().get_value(i,0)

        #do not cancel quick filter
        if self.filters.get_filter( fdelete ) == self.qfilter:
            return 
        
        self.filters.delete_filter( fdelete )
        self.model.unset_filter()
        cb.set_active(0) #reset
        self.save_filters() #save the filters set
        
        return fdelete

    def __del__(self):
        self.save_filters()
        return


# ----------------------------------------------------------------------
# normal trace controller
class TraceCtrl (Controller):
    """This is the controller for a single trace."""

    def __init__(self, model, view, name_start_col=True, rename_error_stripes=True):
        """
        @param name_start_col: if true, the 'name' col is the first, else the last
        @param rename_error_stripes: if it is True, the error stripes will be renamed in this way:
        all the strings dot-separated retrieved from the slim_name, 
        that are equal to the property model.ERROR_STRIPE (by default equal to "_errorSubcomponent") 
        will be deleted, 
        and if the last string dot-separated is equal to the value in model.ERROR_OLD (by default equal to "mode")
        it will be replaced by the string in model.ERROR_REPLACE (by default equal to "error")
        """
        
        Controller.__init__(self, model, view)

        #filter bar
        if view.has_filters:
            self.filterBarCtrl = FilterBarCtrl( model, #model
                                                view.filterBarView #view
                                                #list of stored filters set by default
                                                #PREMANENT_FILTERS
                                                )
            pass

        # rename error stripes if specified
        if rename_error_stripes:
            tr = model.get_active_full_trace()
            if tr: tr.rename_error_stripes()
            pass
        
        #sort model in ascending order
        model.sort_trace(reverse=False)
        
        #the treeview that shows the datas
        self.treeview = view['treeview']
        self.rename_error_stripes = rename_error_stripes
        #general purpose utility
        self.last_col_clicked = None
        self.last_step_selected = -1

        self.bg = "#F0F0F0" #base color
        #base error colors
        self.bg_errors = "#FF0400" 
        self.bg_errors_delta = "#ff423e" 
        self.bg_errors_loop = "#FF8080" 
        self.bg_errors_loop_delta = "#c9413f"
        #end base error colors
        #other colors
        self.bg_delta = "#98f3f4"
        self.bg_loop = "#D0D0D0"
        self.bg_loop_delta = "#558d8e"
        self.bg_sel = "#E0E0E0"
        self.bg_loop_sel = "#2020F0"
        #end other colors
        self.fg_text = { 'input' : "#0040A0", 'combinatorial' : "#00A040", 'error' : "#700000" }
        #self.tips = gtk.Tooltip()
        #self.tips.enable()
        self.col_clicked_fun = None
        self.move_columns = {} #remember how to move the columns when the register_view is called
        self.name_first = name_start_col 
        self.name_col = None #the name column

        # This is used to set tooltips, as it is not always available
        self._set_tooltip = gtk.Widget.set_tooltip_text \
            if gtk.pygtk_version >= (2,12,0) else None

        self.trace_tooltips = gtk.Tooltips()
        return

    #setter col clicked
    def set_on_col_click_fun(self, fun):
        """ It must take two params, the first is the gtk.TreeViewColumn, and the second the step in the model """
        self.col_clicked_fun = fun
        return
    #getter col clicked
    def get_on_col_click_fun(self):
        return self.col_clicked_fun

    #getter filterBar
    def get_filter_bar_view(self):
        """
        get the filter bar view, if there is one associated with
        """
        return self.view.filterBarView if self.view.has_filters else None

    def get_filter_bar_ctrl(self):
        """
        get the filter bar controller, if there is one associated with
        """
        return self.filterBarCtrl if self.view.has_filters else None

    def clear_step_selected(self):
        """
        forgot what step was selected
        """
        self.last_col_clicked = None
        self.last_step_selected = -1
        return

    def get_last_col_clicked(self):
        """
        return None if no col is already selected by the user (clicking on it)
        """
        return self.last_col_clicked

    def get_last_step_selected(self):
        """
        return -1 if no col is already selected by the user (clicking on it)
        """
        return self.last_step_selected

    def move_col(self, what, where):
        """
        move the column identified by what, in the place identified by where
        @param what, type int
        @param where, type int
        @return true if succeded, false otherwise.
        """
        if not len(self.treeview.get_columns()):
            self.move_columns[what] = where
            return False

        to_move = self.treeview.get_column(what)
        where_col = self.treeview.get_column(where)
        self.treeview.move_column_after(to_move, where_col)
        
        return True

    #model change
    def property__active_trace_filtered_value_change(self, model, old, new):
        # rename error stripes if specified
        if self.rename_error_stripes:
            tr = self.model.get_active_full_trace()
            if tr: tr.rename_error_stripes()
            pass
        
        self.setup_treeview(self.treeview)
        return

    def cell_name_func(self, col, cell, model, iter):
        """This one is called to show the single trace stripe name"""
        stripe = model.get_value(iter, 0)
        slim_name = model.get_value(iter, 1)
        
        fg = None
        
        assert slim_name, "the name of the stripe cannot be None!"
        
        if stripe.get_error():
            fg = self.fg_text["error"]
            pass

        # show the renamed slim_name 
        cell.set_property("text", slim_name)
        
        if fg is None and stripe.get_info(error=True) in self.fg_text: fg = self.fg_text[stripe.get_info(error=True)]
        
        cell.set_property("foreground", fg)
        return

    def _choose_bg_color(self, model, tv, stripe, iter, step, maxstep, error=True):
        # IS IT POSSIBLE TO JOIN TWO COLORS INTO ANOTHER ONE?
        # if it is, we can do that stuff: bg_errors_delta = join(bg_errors,bg_delta) 
        # colours for cell background
        # errors color
        # check if the value is changed from the previous step
        equal = step <=0 or step >= maxstep or stripe.get_value_at_step(step) == stripe.get_value_at_step(step-1)
        if equal: # change error colours, if the value is changed
            # if the value is higher
            bg_errors = "#f33d38"
            bg_errors_delta = "#ff8f8c"
            bg_errors_loop = "#d27070"
            bg_errors_loop_delta = "#de5e5b"
        else:
            bg_errors = self.bg_errors 
            bg_errors_delta = self.bg_errors_delta 
            bg_errors_loop = self.bg_errors_loop
            bg_errors_loop_delta = self.bg_errors_loop_delta
            pass
        # end errors
        error_cell = error and stripe.get_info(error=True) == 'error'
        
        # iter choosen from self.model, because delta row is filtered..
        dstripe = self.model.get_delta_stripe() #delta stripe
        if dstripe and step < maxstep-1:
            delta = dstripe.get_value_at_step(step)
        else:
            delta = "0" #if delta_iter is not set, delta does not exist, so we use the default color

        if tv.get_selection().iter_is_selected(iter):
            if self.model.is_step_in_loop(step):
                return self.bg_loop_sel
            else:
                return self.bg_sel    
                # selected end    
        elif self.model.is_step_in_loop(step) and error_cell:
            if delta == "0":
                return bg_errors_loop
            else:
                return bg_errors_loop_delta
                
        elif self.model.is_step_in_loop(step):
            return self.bg_loop
        elif error_cell:
            if delta == "0":
                return bg_errors
            else:
                return bg_errors_delta
        else: 
            if delta == "0":
                return self.bg
            else:
                return self.bg_delta
        

    def cell_data_func(self, col, cell, model, iter, args):
        """@param args is (treeview, step)"""
        def fix_stripe_value(value):
            if not value:
                return value
            for prefix in ("mode:", "error:"): 
                res = value[len(prefix):]
                
                # handle exceptions
                if res == DEFAULTINITMODE: res = ""
                if value.startswith(prefix): return res
                pass
            return value

        if cell.__class__ != src.views.trace.CellRendererTrace: #doesn't render col with no renderer trace
            return    
        
        tv, step = args
        stripe = model.get_value(iter, 0)

        maxstep = stripe.get_steps() #the max step for this stripe (not for all the trace)
        is_last_step = (step >= maxstep)
        
        #stripe.get_info() == "state" and 
        if not is_last_step:            
            cell.set_property("is-boolean", stripe.is_boolean())
            cell.set_property("value", fix_stripe_value(stripe.get_value_at_step(step)) )
        else: 
            cell.set_property("value", None)
        
        if step > 0 and step <= maxstep: 
            cell.set_property("prev-value", fix_stripe_value(stripe.get_value_at_step(step-1)))
        else: 
            cell.set_property("prev-value", None)
        
        if step < maxstep-1:     
            cell.set_property("next-value", fix_stripe_value(stripe.get_value_at_step(step+1)))
        else: 
            cell.set_property("next-value", None)
            pass

        # background colour: 
        cell.set_property("cell-background", 
                          self._choose_bg_color( model, tv, stripe, iter, step, maxstep, 
                                                 cell.get_property("value") != None ) 
                          )
                
        return
                   
    def stripe_filter(self, model, iter, data):
        """This function is used when filtering names"""
        slim_name = model.get_value(iter, 1)
        return slim_name is not None and \
            not slim_name.endswith("._errorState") and \
            ("cando_" not in slim_name.split(".")[-1])
    
    def register_view(self, view):
        Controller.register_view(self, view)
        
        def search_func(model, column, key, i, data):
            sn = model.get_value(i,1)#slim_name
            # i don't know why but I have to return false if I found a match
            return not re.match(key+'.*', sn) #key and something after
        
        tv = view['treeview']
        tv.set_search_equal_func(search_func,user_data=None)

        #self.setup_filterlist(view)
        self.setup_treeview(self.treeview)
        return view
    
    def _filter_names(self):
        #helper, filter the names of the model and return a new filtered model
        filtered_model = self.model.filter_new()
        if filtered_model:
            filtered_model.set_visible_func(self.stripe_filter, data=None)
            pass
        return filtered_model

    def hide_col(self, col):
        col.set_visible(False)
        try:
            rend = col.get_cell_renderers()[0]
        except IndexError:
            return
        
        col.set_cell_data_func(rend, None) #unset data func
        return

    def setup_treeview(self, view): 
        tv = view#['treeview']
        tv.set_model(self._filter_names())
        #tv.set_search_equal_func(search_func,user_data=None)
        
        if not self.name_col and self.name_first:
            #create the cols from the begining
            rendtxt = gtk.CellRendererText()
            col = gtk.TreeViewColumn("Name", rendtxt) #the first column 
            #if the cols already exist, these things are already set
            col.set_cell_data_func(rendtxt, self.cell_name_func)
            col.set_resizable(True)
            col.model_col = -1
            tv.append_column(col)
            self.name_col = col
            pass

        cols = tv.get_columns() #existing columns
        lc = len(cols)-1 #remove the "name" cols (the first, or the last one)
        #reuse existing column for new model only adapting it
        li = 0 #the first column that contains the stripe
        
        for i in self.model.valid_cols_iter():
            #hide the cols not used
            #+1 because I cant set invisible the "name" column, or the column just set visible
            for j in xrange(li+1,i+1): 
                if j < lc:
                    self.hide_col(cols[j])
                else: #if does not exist create it and than hide it
                    col = self._append_column(tv) 
                    pass
                pass

            #create the column if needed, and set it visible
            if i >= lc: #the cols does not exist
                col = self._append_column(tv) 
            else:
                # +1 because i have to avoid the first column (the name column)
                # if the name column is at the beginning
                # if the name column is not at the beginning we can take the first one
                column_id = i+1 if self.name_first else i
                col = cols[column_id] 
                pass

            if self.model.get_active_full_trace().is_empty(i): 
                col.set_visible(False)
            else:
                col.set_visible(True)
                rend = self._setup_column(col, i) #automatically label the column as "step i+1"
                self._fill_column(rend, col, tv, i)
                pass
            li = i+1 #last i (i is the column-1, li have to be the number of the column)
            pass

        #now we have to set invisible the extra-cols not used greater than i and lower than lc
        for hc in xrange(li+1, lc+1):
            self.hide_col( cols[hc] )
            pass

        #append the name col at the end if name_first is false
        if not self.name_col and not self.name_first:
            #create the cols from the begining
            rendtxt = gtk.CellRendererText()
            col = gtk.TreeViewColumn("Name", rendtxt) #the first column 
            #if the cols already exist, these things are already set
            col.set_cell_data_func(rendtxt, self.cell_name_func)
            col.set_resizable(True)
            col.model_col = -1
            tv.append_column(col)
            self.name_col = col
            pass

        for what,where in self.move_columns.items():
            self.move_col(what, where)
            #after moving, forget about it
            del self.move_columns[what]
            pass
            
        self.name_col.set_visible(True)
        return
 
    def _setup_column(self, col, i, click=True):
        
        lbl = gtk.Label("Step%d" % (i+1))
        if self.model.is_step_a_loop(i+1):
            # shows an arrow to explicitate that loop begins at that step:
            wid = gtk.HBox()
            wid.show()
            ar = gtk.Arrow(gtk.ARROW_DOWN, gtk.SHADOW_IN) 
            ar.show()
            
            wid.add(ar)
            lbl.show()
            wid.add(lbl)
            if self._set_tooltip: 
                self._set_tooltip(ar, "A loop starts at this step")
                self._set_tooltip(wid,"A loop starts at this step")
                pass
            pass
        else: 
            wid = lbl
            wid.show()
            pass
        #col.clear() #unset previous renderers
        col.set_widget(wid)
        
        rend = col.get_cell_renderers()
        if not rend:
            rend = src.views.trace.CellRendererTrace()

            rend.set_property("cell-background", self.bg)
            rend.set_property("waveform-fg-color", "#105080")
            rend.set_property("waveform-sel-fg-color", "white")
            rend.set_property("bool-waveform-width", 2)
            col.pack_start(rend) 
            #set it clickable (for picking steps for future use values)
            #the number of the column clicked is stored in a variable and can be used
            #as the programmer want to do.
            if click:
                col.set_clickable(True)
                col.connect('clicked', self._column_clicked, (i, self.get_on_col_click_fun()))
                pass
        else:
            rend = rend[0]
            pass

        return rend

    def _fill_column(self, rend, col, tv, i):
        #fill the values in the column, and set the renderer
        col.set_cell_data_func(rend, self.cell_data_func, (tv, i)) #fill column with datas
        col.model_col = i
        return    

    def _append_column(self, tv):
        #create a column and append it to the treeview
        col = gtk.TreeViewColumn()
        tv.append_column( col )
        col.set_visible(False)
        col.set_expand(True)

        return col

    #signals helpers
    #file chooser dialog
    def _fchooser_dialog(self, title, action, buttons):
        # 0 : cancel option
        # 1 : ok option
        #create a file chooser dialog and return it
        o = gtk.FileChooserDialog(title=title, 
                                  action=action, 
                                  buttons=None)
        
        o.add_button(buttons[0], gtk.RESPONSE_CANCEL)
        o.add_button(buttons[1], gtk.RESPONSE_OK)

        return o
    # ----------------------------------------------------------------------
    # Signals
    # ----------------------------------------------------------------------
    def on_window_delete_event(self, win, event=None):
        if self.view.has_filters:
            self.filterBarCtrl.save_filters() 
            pass
        gtk.main_quit()
            
        return True

    #columns clicked
    def _column_clicked(self, col, user_data):
        (step,fun) = user_data
        self.last_col_clicked = col
        self.last_step_selected = step
        if fun:
            fun(col,step)
            pass
        return

    #modify/delete/create filter callbacks
    def create_filter_dialog(self, menu_item):
        return self.filterBarCtrl.create_filter_dialog(menu_item)
    
    def delete_filter(self, menu_item):
        return self.filterBarCtrl.remove_filter(menu_item)
    
    def on_treeview_motion_notify_event(self, tv, event):
        # This is used for the tooltip showing the value and name of signals
        # In particular rationals are shown also as floats

        pos  = tv.get_path_at_pos(int(event.x), int(event.y))            
        if pos:
            path, col, x, y = pos
            model = tv.get_model()
            if col.model_col >= 0:
                stripe = model[path][0]
                name = model[path][1]
                data = str(stripe.get_value_at_step(col.model_col))
                if "/" in data: 
                    a, b = map(float, data.split("/"))
                    data = data + " (%g)" % (a/b)
                    pass
                tip = "%s: %s" % (name, data)
                self.trace_tooltips.set_tip(tv, tip)
                pass                
            else: self.trace_tooltips.set_tip(tv, "")
            pass
        else: self.trace_tooltips.set_tip(tv, "")
        
        return True

    # OPEN/CLOSE/SAVE/LOAD/DELETE CALLBACKS
    #
    def open_filters(self,menu_item):
        """
        @return the filename loaded, or None if no filter list was loaded
        """
        o = self._fchooser_dialog( "open an xml filters list", gtk.FILE_CHOOSER_ACTION_OPEN, ("cancel","open") )
        rid = o.run() 
        fname = None
        if rid == gtk.RESPONSE_OK:
            fname = o.get_filename()
            self.filterBarCtrl.clear_filters()
            #read the filter list, and fill the comboBox
            self.filterBarCtrl.load_filters( fname )

        o.destroy()

        return fname
    #        
    def save_filters(self,menu_item):
        """
        @return the filename on which the filter list was saved, 
        or None if no filter list was saved
        """
        o = self._fchooser_dialog( "save an xml filters list (the file will be overwritten)", 
                              gtk.FILE_CHOOSER_ACTION_SAVE, 
                              ("cancel","save") ) 
        
        rid = o.run()
        fname = None
        if rid == gtk.RESPONSE_OK:
            fname = o.get_filename()
            self.filterBarCtrl.save_filters(where=fname)
            
        o.destroy()
        
        return fname 
            
    def open_trace(self,menu_item, trace_handler=TraceModel, top=None, reload_view=None ):
        #top -> the window in which we had to change title, if none 'window_main' is assumed
        o = self._fchooser_dialog( "open a counter-example trace", gtk.FILE_CHOOSER_ACTION_OPEN, ("cancel","open") )
        m = None
        rid = o.run() 

        if rid == gtk.RESPONSE_OK:
            file_name = o.get_filename()
            o.destroy()
            m = self.model.add_traces_from_file( file_name, th=trace_handler )
            
            if not top:
                top = self.view.get_top_widget()[0] #the main window
                title = top.get_title()
            else:
                title = top.get_title()
                pass

            top.set_title( 
                ' - '.join( [ title.split(' - ')[0], os.path.split(file_name)[1] ] )
                )
            
            self.model.set_active_trace( m ) #redesign!!
            if reload_view:
                method = getattr(self,reload_view)
                method(self.view)
            else:
                self.setup_treeview(self.treeview) 
        else:
            o.destroy()
        
        return m

    # SORT CALLBACKS
    
    def sort_treeview_asc(self, mi ):
        self.model.sort_trace(None, False) #ascending
        self.setup_treeview(self.treeview)
        return

    def sort_treeview_desc(self, mi ):
        self.model.sort_trace(None, True) #descending
        self.setup_treeview(self.treeview)
        return

    def sort_treeview_types(self, mi ):
        """
        sort over types
        """
        def sortfunction( (name1,s1), (name2,s2) ):
            #use ls to compare s1 and s2
            pr_ls = {'state':0,'combinatorial':1,'input':2,'error':3} #Precedence list 

            if pr_ls[s1.get_info(error=True)] < pr_ls[s2.get_info(error=True)]:
                return -1
            elif pr_ls[s1.get_info(error=True)] > pr_ls[s2.get_info(error=True)]:
                return 1
            else:#if they are equal check on names
                return cmp( name1, name2 )
                

            #END UTILITY FUNCTIONS

        self.model.sort_trace( sortfunction )
        self.setup_treeview(self.treeview)
        return

    pass # end of class

#class derived from TraceCtrl that change the behaviour of the treeview in the view
# special trace controller: it shows only the last n rows and permit to add values
class TraceExpansionCtrl(TraceCtrl):
    def __init__(self, m, v, cols_to_show=1):
        """
        @param m, model
        @param v, the view
        @param cols_to_show, how many cols do you want to show (starting from the last one?) default 1.
        """
        
        super(TraceExpansionCtrl,self).__init__(m, v)
        model = m.get_active_full_trace()
    
        if model:
            #the last step
            #how many cols to show
            self.to_show = m.valid_cols()[-cols_to_show:]
            model.append_col()
        else: self.to_show = None

        #last cols, in which were stored the temporary values
        #until it will be forced into the ExtensibleTraceModel
        self.tvvalue_lcol = None

        #setup completion model
        self.completion_model = gtk.ListStore(gobject.TYPE_STRING)
        for x in ['true','false']:
            self.completion_model.append([x])
            pass

        return
    
    def setup_treeview(self, view): 
        #shows the number of cols selected by the user stored in the variable self.to_show
        
        model = self.model.get_active_trace()
        tv = view#['treeview']
        m = self._filter_names()
        tv.set_model(m)        
        #tv.set_search_equal_func(search_func,user_data=None)
        
        if not model:
            return tv

        if not tv.get_columns():
            #create the cols from the begining
            rendtxt = gtk.CellRendererText()
            col = gtk.TreeViewColumn("Name", rendtxt) #the first column 
            #if the cols already exist, these things are already set
            col.set_cell_data_func(rendtxt, self.cell_name_func)
            col.set_resizable(True)
            col.model_col = -1
            tv.append_column(col)
            pass

        cols = tv.get_columns() #existing columns
        
        #self.to_show = cols that the users chooses to show
        for i in xrange(1,len(self.to_show)+1): #avoid the first col (that is the 'name' col
        #create the column if needed, and set it visible
            try:
                col = cols[i] #the data col
            except IndexError:
                col = self._append_column(tv) 
                pass    
            #to_show[i-1] = the step that must be visible in this col
            rend = self._setup_column(col, self.to_show[i-1], click=False) #automatically label the column as "step i+1"
            self._fill_column(rend, col, tv, self.to_show[i-1])
            col.set_visible(True)
            pass
        
        if self.tvvalue_lcol:
            rend = self.tvvalue_lcol.get_cell_renderers()
            assert rend
        
            rend = rend[0]
            rend.connect('edited', self.add_forced_value, m )
            rend.connect('editing-started', self.editing_started, m )
            pass
        
        return tv

    def register_view(self, tv):
        super(TraceExpansionCtrl,self).register_view( tv ) #it calls my setup_treeview
        self.treeview = tv['treeview']
        self.setup_view(self.treeview)
        return tv

    #this is called only once, when register_view is called.
    #it calls setup_treeview too, but in order to refresh the view,
    #we have to call setup_treeview()
    def setup_view(self, tv):
        treeview = tv#['treeview']
        #remove last_col before setup treeview
        if self.tvvalue_lcol and self.tvvalue_lcol in treeview.get_columns():
            #if it is present, remove it in order to avoid problems
            treeview.remove_column( self.tvvalue_lcol )
            pass

        #setup treeview
        self.setup_treeview( treeview ) 
        #create the extra col in 'add' trace viewer
        rend = self.__build_extra_col(treeview) 
        
        #if this is not set, the model is not set
        if not self.tvvalue_lcol:
            return
        
        treeview.append_column( self.tvvalue_lcol ) #add the "new step" column
            
        trace = self.model.get_active_full_trace()
        if not trace:
            self.tvvalue_lcol.set_visible(False)
        else:
            self.tvvalue_lcol.set_visible(True)
            #catch the 'edited' signal in order to add to the model the new stripe value
            m = treeview.get_model()
             #this is only for the first time
            rend.connect('edited', self.add_forced_value, m )
            rend.connect('editing-started', self.editing_started, m )
            pass
        
        return tv

    def __build_extra_col(self, tv):
        """
        @return the renderer of the editable col
        """

        model = self.model.get_active_full_trace()
        if not model: return self.tvvalue_lcol.get_cell_renderers()[0] if self.tvvalue_lcol else None
        
        step = model.get_max_steps()-1 #the added step (empty)
    
        if self.tvvalue_lcol:
            #if exist the extra-col
            rend = self.tvvalue_lcol.get_cell_renderers()[0]
        else:
            rend = gtk.CellRendererText()

            rend.set_property("cell-background", self.bg)
            rend.set_property("editable", True)

            #EDITABLE CELLS
            col = gtk.TreeViewColumn("New Step", rend) #the last column
            col.model_col = -1
            self.tvvalue_lcol = col #set lcol
            #col.clear() #clear previous renderer 
            if step >= 0 : rend.set_property('editable', True)            
            col.set_resizable(True)
            pass

        self.tvvalue_lcol.set_cell_data_func(rend, self.extra_col_data_func, (tv,step))

        return rend

    def extra_col_data_func(self, col, rend, model, iter, args):
        """
        This function is called to refresh the values in the "new step" column, in the guided simulation
        @param args is (treeview, step)
        """
        def fix_stripe_value(value):
            if not value:
                return None
            for prefix in ("mode:", "error:"): 
                if value.startswith(prefix): return value[len(prefix):]
                pass
            return value
                
        tv, step = args
        stripe = model.get_value(iter, 0)

        maxstep = stripe.get_steps() #the max step for this stripe (not for all the trace)
        is_valid_step = (step >= maxstep)

        # background colours:
        rend.set_property("cell-background", 
                          self._choose_bg_color(model, tv, stripe, iter, step, maxstep) )
        
        #stripe.get_info() == "state" and 
        if not is_valid_step:         
            rend.set_property("text", fix_stripe_value(stripe.get_value_at_step(step)))
        else: 
            rend.set_property("text", None)

        return

    #----------#
    # signals  #
    #----------#
    
    #functions that handles the editing of the new step cells
    def editing_started(self, rend, editable, path, user_data ):
        """
        this function is called when the user starts editing the forced value
        (i.e. when the user clicks in the editable field of the left 'small' trace)
        """
        #when the user focus on the editable
        etmodel = user_data #ExtensibleTraceModel (TreeFilter!)
        #set the model
        i = etmodel.get_iter( path )
        estripename = etmodel.get_value( i, 1 ) #the ExtensibleStripe name

        # TODO re-enable completion feature when possible
        # completion
        # if estripe.is_boolean() and not editable.get_completion(): #completion only for booleans
        #     comp = gtk.EntryCompletion()
        #     comp.set_model( self.completion_model )
        #     comp.set_text_column( 0 )
        #     editable.set_completion( comp )

        editable.connect("changed", self.editing, (estripename,etmodel,path,rend) )
        return

    def editing( self, editable, user_data ):
        """
        this function si called while the user is editing the forced value
        """
        estripename,etmodel,path,rend = user_data
        
        model = self.model.get_active_full_trace()
        valid =  gtk.gdk.color_parse("pale green")
        invalid = gtk.gdk.color_parse("indian red")
        valid_input, error = model.validate_input( editable.get_text(), estripename)
        #set background
        editable.modify_base( gtk.STATE_NORMAL, valid if valid_input else invalid )
        if self._set_tooltip:
            self._set_tooltip( editable, error )
            pass

        if valid_input:
            self.add_forced_value(rend, path, editable.get_text(), etmodel)
            pass
        return

    #user data is the treeModel of 'add' (ExtensibleTraceModel)
    def add_forced_value(self, rend, path, new_text, user_data ):
        """
        this function is called when a value is forced into a stripe.
        Warning! call it directly at your own risk.
        """
        etmodel = user_data #ExtensibleTraceModel (TreeFilter!)
        model = self.model.get_active_full_trace()
        #get the right stripe 
        i = etmodel.get_iter( path )
        
        estripename = etmodel.get_value( i, 1 ) 
        estripe = etmodel.get_value( i, 0 )#the ExtensibleStripeModel
        #the step in which we want to add a value
        step = estripe.base_steps()

        #update the value (into another function?)
        valid, error = model.validate_input(new_text, estripename)
        if valid:
            #extend the delta stripe
            ds = self.model.get_delta_stripe()

            # TODO here it seems that this setting is mandatory (or
            # else a bug occurs). Investigate further.
            if ds: ds.force_stripe_value( step, "") 

            # extend the stripe
            estripe.force_stripe_value( step, new_text )
            
            # update the view
            if self.tvvalue_lcol:
                lcol_step = model.get_max_steps()-1 #step to be viewed in the 'new step' col
                self.tvvalue_lcol.set_cell_data_func(rend, 
                                                     self.extra_col_data_func, 
                                                     (self.view['treeview'], lcol_step))
                pass
            pass
            
        return
    #end
    
    #callback signal overloading
    def open_trace(self,menu_item, trace_handler=ExtensibleTraceModel, top=None, reload_view="setup_view" ):
        em = super(TraceExpansionCtrl,self).open_trace(menu_item=menu_item, 
                                                       trace_handler=trace_handler, 
                                                       top=top,
                                                       reload_view=reload_view )
        if em:
            self.model.get_active_full_trace().append_col()
            
        return em

    def apply_quick_filter(self, toggle_button):
        return super(TraceExpansionCtrl,self).apply_quick_filter( toggle_button=toggle_button,
                                                                  reload_view="setup_view" )

    def apply_stored_filter(self, cb):
        return super(TraceExpansionCtrl,self).apply_stored_filter( cb=cb,
                                                                   reload_view="setup_view" )
    pass
#------------------------------------end of the class


class GuidedSimTransitionCtrl (Controller):

    CHECK_DELAY = 1500 # milli-seconds

    def __init__(self, model, view, parent):
        """parent is instance of ModelSimulationCtrl"""
        Controller.__init__(self, model, view)        

        self.parent = parent

        # observes infinite_domain in slim model, to enable/disable
        # timed-transitions pane
        self.observe_model(self.model.slim_model)

        # timers for type checking
        self._timer_check_id = None        

        self._last_constraints = []
        self._last_constraints_truth = True
        return

    def register_view(self, view):
        Controller.register_view(self, view)

        # -------------------------------------------------
        # components
        tv = self.view['tv_components']
        tv.set_model(self.model.slim_tree)

        col = gtk.TreeViewColumn('Name')
        rend_name = gtk.CellRendererText()
        #rend_name.set_property('weight', 1800)
        col.pack_start(rend_name)
        rend_num = gtk.CellRendererText()
        col.pack_end(rend_num, expand=False)
        
        def _inst_name(col, rend, model, iter, col_num):
            row = model[iter]
            name, num = row[col_num], row[6]
            
            rend.set_property('weight-set', num > 0)
            rend.set_property('weight', 800 if num > 0 else 400)
            rend.set_property('text', name)
            return            
        col.set_cell_data_func(rend_name, _inst_name, 2)
        
        def _inst_trans_num(col, rend, model, iter, col_num):
            num = model[iter][col_num]            
            rend.set_property('text', ('(%d)' % num) if num > 0 else "")
            return            
        col.set_cell_data_func(rend_num, _inst_trans_num, 6)

        tv.append_column(col)        

        # ------------------------------
        def _impl(col, rend, model, iter, col_num):            
            impl = model[iter][col_num]
            rend.set_property('text', impl.name)
            return
        rend = gtk.CellRendererText()
        col = gtk.TreeViewColumn('Impl', rend)
        col.set_cell_data_func(rend, _impl, 0)
        tv.append_column(col)        

        sel = tv.get_selection()
        sel.set_mode(gtk.SELECTION_BROWSE)
        sel.connect("changed", self.on_components_sel_changed)

        tv.expand_all()

        # -------------------------------------------------
        # transitions
        tv = self.view['tv_transitions']
        tv.set_model(self.model.trans_list)

        col = gtk.TreeViewColumn('Transitions')

        def _trans_pic(col, rend, model, iter, stocks_map):
            trans = model.get_value(iter, 0)
            val = model.get_value(iter, 1)
            enabled = val != GuidedSimTransitionModel.TRANS_NOT_ALLOWED

            rend.set_property('sensitive', enabled)
            rend.set_property('stock-id', stocks_map[val])
            return
        rend1 = gtk.CellRendererPixbuf()
        stocks = { GuidedSimTransitionModel.TRANS_NOT_ALLOWED : gtk.STOCK_NO, 
                   GuidedSimTransitionModel.TRANS_ALLOWED     : gtk.STOCK_YES, 
                   GuidedSimTransitionModel.TRANS_UNKNOWN     : gtk.STOCK_DIALOG_WARNING,
                   }
        col.pack_start(rend1, expand=False)
        col.set_cell_data_func(rend1, _trans_pic, stocks)

        def _trans_text(col, rend, model, iter):
            trans = model.get_value(iter, 0)
            enabled = model.get_value(iter, 1) != GuidedSimTransitionModel.TRANS_NOT_ALLOWED

            rend.set_property('sensitive', enabled)
            rend.set_property('strikethrough-set', not enabled)
            rend.set_property('text', toSLIMVisitor().visit(trans).strip())
            return
        rend2 = gtk.CellRendererText()
        rend2.set_property('strikethrough', True)
        col.pack_start(rend2)
        col.set_cell_data_func(rend2, _trans_text)

        tv.append_column(col)     
        
        # ----------------------------------------
        sel = tv.get_selection()
        sel.set_mode(gtk.SELECTION_BROWSE)
        sel.connect("changed", self.on_transitions_sel_changed)

        def __trans_sel(path, data):
            model, col = data
            return model[path][col]
        sel.set_select_function(__trans_sel, (self.model.trans_list, 1))

        tv.connect('row-activated', self.on_row_activated)

        # time-transition panel
        cando_time = self.model.cando_time_transition()
        self.view.show_timed_trans_simulation(cando_time) 

        view['sb_time_min'].hide()
        view['sb_time_max'].hide()
        view['cb_time_min_unit'].hide()
        view['cb_time_max_unit'].hide()

        ## The time unit combobox must be selectable only in case the slim model uses time scaling
        view['cb_time_min_unit'].set_sensitive(self.model.slim_model.compile_info.uses_time_scaling)
        view['cb_time_max_unit'].set_sensitive(self.model.slim_model.compile_info.uses_time_scaling)
        return
    
        
    # double-clicked a transition: run the simulation
    def on_row_activated(self, tv, path, column): 
        if self.model.sim_model.busy == 0 and self.model.sim_model.can_simulate: 
            self.parent.start_activity()
        return    

    def on_components_sel_changed(self, sel):        
        model, iter = sel.get_selected() 

        if (self.model.sim_model.result and 
            isinstance(self.model.sim_model.result, SimResultModel) and
            self.model.sim_model.result.trace and 
            isinstance(self.model.sim_model.result.trace, TraceModel)):
            # if there is a trace, use it to filter out transitions
            self.model.build_transitions(iter, self.model.sim_model.result.trace)
        else:
            self.model.build_transitions(iter)
            pass
       
        return

    def on_transitions_sel_changed(self, sel):        
        model, iter = sel.get_selected()         
        self.model.set_trans_constraints(model[iter][2] if iter else [])
        self.model.set_can_simulate(iter is not None)
        self.model.sim_model.can_simulate = iter is not None
        return

    def on_cb_min_changed(self, cb):
        """Function activated when the combox for defining the operator ('>', '>=', '=') of the min value 
           of the duration of a timed transition in guided simulation changes"""
        if cb.get_active() == 0: 
            self.view['sb_time_min'].hide()
            self.view['cb_time_min_unit'].hide()
            self.view.show_max_constraint(True)
        else: 
            self.view['sb_time_min'].show()
            self.view['cb_time_min_unit'].set_active(0)
            self.view['cb_time_min_unit'].show()
            
            if cb.get_active() == 1: # '=' symbol
                self.view.show_max_constraint(False)
            else:
                self.view.show_max_constraint(True)
                pass
            pass
                
        self._delayed_check()
        return

    def on_cb_max_changed(self, cb):
        """Function activated when the combox for defining the operator ('<', '<=') of the max value 
           of the duration of a timed transition in guided simulation changes"""
        if cb.get_active() == 0:
            self.view['sb_time_max'].hide()
            self.view['cb_time_max_unit'].hide()  
        else: 
            self.view['sb_time_max'].show()
            self.view['cb_time_max_unit'].set_active(0)
            self.view['cb_time_max_unit'].show()
            
        self._delayed_check()
        return
    
    
    # LB: added time scale unit
    ## GZ2LB: these two functions are never called/used; I think they should be deleted or they
    ##        should be properly associated to the related combobox
    # def on_cb_time_min_unit_changed(self, cb):
    #     self._delayed_check()
    #     return
        
        
    # def on_cb_time_max_unit_changed(self, cb):                
    #     self._delayed_check()
    #     return
        
        
    def _on_check_values(self):
        min_op, min_val, min_unit = self.view.get_min_constr()
        max_op, max_val, max_unit = self.view.get_max_constr()
                
        constrs = []
        #dv = self.model.converter.data_name(self.model.converter.delta_var())
        dv = self.model.converter.delta_var()

        if min_op: 
            if min_unit:
                min_time_str = str(min_val) + str(min_unit)
            else:
                min_time_str = str(min_val)
            min_val_unit_smv = smvapi.parse_and_convert_numeric_expression(self.model.slim_model.get_extended_root_impl(), 
                                                                           min_time_str, dest_tu=None)
            constrs.append(dv + min_op + str(min_val_unit_smv))

        if max_op: 
            if max_unit:
                max_time_str = str(max_val) + str(max_unit)
            else:
                max_time_str = str(max_val)
            max_val_unit_smv = smvapi.parse_and_convert_numeric_expression(self.model.slim_model.get_extended_root_impl(), 
                                                                           max_time_str, dest_tu=None)
            constrs.append(dv + max_op + str(max_val_unit_smv))

        ok = self.model.cando_time_transition(constrs)
        if not ok and self._last_constraints_truth: self.view.set_error(True)
        elif ok and not self._last_constraints_truth: self.view.set_error(False)
        
        self.model.set_do_time_transition(constrs)
        
        if not self.model.sim_model.can_simulate and \
                self.view.get_guided_simulation_type() == self.view.TIMED:
            # sets only if the view is still showing the timed-trans pane
            self.model.sim_model.can_simulate = ok
            self.model.set_can_simulate(ok)
            pass

        # saves results
        self._last_constraints = constrs
        self._last_constraints_truth = ok

        # resets the timer
        self._timer_check_id = None
        return False        


    def _delayed_check(self):
        self.model.sim_model.can_simulate = False # in the meantime, disable the simulation

        if self._timer_check_id is not None: 
            gobject.source_remove(self._timer_check_id)
            pass
        
        self._timer_check_id = \
            gobject.timeout_add(GuidedSimTransitionCtrl.CHECK_DELAY, 
                                self._on_check_values)
        return

    def on_sb_time_min_value_changed(self, sb):
        """Function activated when changes the value of the spinbutton for defining the min value (i.e. 
           the numeric value) of a timed transition in guided simulation"""
        val = sb.get_value()
        sb_max = self.view['sb_time_max']
        _min, _max = sb_max.get_range()
        sb_max.set_range(val, _max)
        if val > sb_max.get_value(): sb_max.set_value(val)
        
        self._delayed_check()
        return

    def on_sb_time_max_value_changed(self, sb):
        """Function activated when changes the value of the spinbutton for defining the max value (i.e. the 
           numeric value) of a timed transition in guided simulation"""

        self._delayed_check()
        return

    def on_top_sim_guided_transitions_switch_page(self, nb, page, num):
        if num == self.view.DISCRETE:
            sel = self.view['tv_transitions'].get_selection()
            model, iter = sel.get_selected()         
            self.model.set_trans_constraints(model[iter][2] if iter else [])
            self.model.set_can_simulate(iter is not None)
            self.model.sim_model.can_simulate = (iter is not None)
        else:
            self.model.set_do_time_transition(self._last_constraints)
            self.model.sim_model.can_simulate = self._last_constraints_truth
            self.model.set_can_simulate(self._last_constraints_truth)
            pass
        return

    # ----------------------------------------
    # Observable properties
    # ----------------------------------------
    
    def property_infinite_domain_value_change(self, model, old, new):
        if model != self.model.slim_model: return
        cando_time = self.model.cando_time_transition()
        self.view.show_timed_trans_simulation(cando_time)
        return

    def property_changed_signal_emit(self, model, args):
        if model != self.model: return
        
        # checks if time-transitions are enabled
        cando_time = self.model.cando_time_transition()
        self.view.show_timed_trans_simulation(cando_time)            

        sel = self.view['tv_transitions'].get_selection()
        model, iter = sel.get_selected()         
        self.model.set_can_simulate(iter is not None)
        self.model.sim_model.can_simulate = iter is not None
        return

    pass # end of class
# ----------------------------------------------------------------------


#extensible trace controller
class ExtensibleTraceCtrl(Controller):
    """
    This class handles a TraceExpansionCtrl and a TraceCtrl,
    and join them together in order to obtain a guided simulation for a trace,
    instead of the random simulation loaded from xml.

    @param trans_sim_guided_ctrl is a GuidedSimTransitionCtrl
    instance used for controlling the transition-based guided
    simulation.
    """
    def __init__(self, model, view, trans_sim_guided_ctrl):
        Controller.__init__(self, model, view)

        assert model.__class__ == TraceViewModel, "The model is not an TraceViewModel!"
        #base background color
        self.bg = "#F0F0F0"
        
        self.value_ctrl = TraceExpansionCtrl(model, view.value_view, cols_to_show=2)
        self.trans_ctrl = trans_sim_guided_ctrl

        self.tvshow_ctrl = TraceCtrl(model, view.tvshow_view, name_start_col=False)
        
        #filter bar
        if view.has_filters:
            self.filterBarCtrl = FilterBarCtrl( model, #model
                                                view.filterBarView #view
                                                #list of stored filters set by default
                                                )
            pass
        
        self.tvshow = view.tvshow_view['treeview'] #tree view show
        self.tvvalue = view.value_view['treeview'] #tree view add

        self.filters = view.tvshow_view['TBFilter'] #it is None if TBFilter does not exist

        self.first_visible = True
        return

    def get_trans_sim_guided_ctrl(self): 
        """Since self is constructed from scratch every time a
        trace has to be constructed, and as instead the trans-based
        guided simulation has to be reused, this getter gives
        access to the internal GuidedSimTransitionCtrl for being
        passed to the new ExtensibleTraceCtrl instance
        constructor."""
        return self.trans_ctrl

    #def setter for column
    def set_on_col_click_fun(self, fun):
        """
        This method set a function to be called when a column in the TraceView is clicked.
        It must take two params, the first is the gtk.TreeViewColumn clicked, and the second the step in the model 
        """
        self.tvshow_ctrl.set_on_col_click_fun( fun )
        return

    def set_first_col_visible(self, tf):
        #sets the first col of the tvshow visible or not
        self.first_visible = tf
        return

    #getter of the traceviewers
    def get_full_trace_view(self):
        return self.view.tvshow_view
    def get_full_trace_ctrl(self):
        return self.tvshow_ctrl
    def get_small_trace_view(self):
        return self.view.value_view
    def get_small_trace_ctrl(self):
        return self.value_ctrl

    #getter col clicked
    def get_on_col_click_fun(self):
        return self.tvshow_ctrl.get_on_col_click_fun()

    def clear_step_selected(self):
        self.tvshow_ctrl.clear_step_selected()
        return

    #def getter for column
    def get_last_step_selected(self):
        return self.tvshow_ctrl.get_last_step_selected()
    def get_last_col_clicked(self):
        return self.tvshow_ctrl.get_last_col_clicked()

    def setup_btnstep(self):
        if not self.view['BTNStep']:
            return

        if self.model.get_active_full_trace():
            self.view['BTNStep'].set_sensitive(True)
        else:
            self.view['BTNStep'].set_sensitive(False)
            pass
        return

    def setup_tvvalue(self):        
        """ build the treeview of the guided part """
        if not self.model.get_active_trace(): return self.tvvalue
        
        self.value_ctrl.setup_treeview( self.tvvalue )
        return self.tvvalue

    def setup_tvshow(self):
        """ build the treeview of the full trace"""
        if not self.model.get_active_trace():
            return self.tvshow
        
        self.tvshow_ctrl.setup_treeview( self.tvshow )
        self.tvshow.get_column(0).set_visible(self.first_visible)      
        return self.tvshow

    def register_view(self, view):
        Controller.register_view(self, view)    

        # reloads the treeviews (add and show)
        model = self.model.get_active_trace()
        if model:
            self.setup_tvvalue()
            self.setup_tvshow()
            pass
            
        self.setup_btnstep()
        return

    #----SIGNALS---------------------------------------

    def on_window_main_delete_event(self, widget, event):
        gtk.main_quit()
        return

    def load_trace(self, imi):
        #this set the new active trace in the model and update the tvshow treeView
        em = self.tvshow_ctrl.open_trace(imi, trace_handler=ExtensibleTraceModel, top=self.view['window_main'] )
        if em:
            self.setup_tvvalue()
            self.setup_btnstep()
        return

    pass
#---------------------------------------------------------- end of the class

class FilterCtrl(Controller):
    """
    This class implement a controller for the FilterView class.
    It is able to handle the signals from the View and manage it.
    @see FilterView
    """
    RANGES_SEPARATOR = ','
    RANGE_CHAR = '-'

    def __init__(self, model, view, filters, main_ctrl, cb=None):
        """
        @param view, the view FilterView
        @param model, a TraceFilter instance
        @param filters, it is a TraceFilters instance to be expanded with the model
        @param main_ctrl, it is the controller of the TraceFilterBar associated with
        @param cb, it is the combo box in which you would to add the filter created as an entry
        """
        Controller.__init__(self, model, view)
        self.filters = filters
        self.iter = None
        self.create = False
        self.main_ctrl = main_ctrl
        self.cb = cb
        pass

    def set_combo_box(self, cb):
        """
        set the combo box in which the filter will be added as entry.
        """

        self.cb = cb
        return
    
    def get_combo_box(self):
        return self.cb
    
    def set_range_separator(self, char):
        """
        set what is the string that separates the ranges.
        by default this string is ','
        """
        assert len(char) == 1, "The separator have to be length 1"
        assert type(char) is str, "The separator must be a string!"
        
        self.RANGES_SEPARATOR = char
        return

    def set_range_char(self, char):
        """
        set what is the string that specify a range.
        by default this string is '-'
        i.e. 7-10, is the range(7,10)
        """
        assert len(char) == 1, "The character for the range have to be length 1"
        assert type(char) is str, "The character for the range have to be a string!"
        
        self.RANGE_CHAR = char
        
        return

    def get_range_separator(self): return self.RANGES_SEPARATOR
    def get_range_char(self): return self.RANGE_CHAR

    def get_model(self):
        return self.model
    
    def modify_filter_name(self, b):
        self.view['TEFilterName'].set_sensitive(b)
        return
    
    def set_model(self, model):
        self.model = model
        self.observe_model(model)
        self.register_view(self.view)
        return

    def register_view(self, view):  
        Controller.register_view(self, view)
        #set the name
        if self.model.get_filter_name():
            name = self.model.get_filter_name()
        elif self.filters:
            base_name = 'filter_'
            cnt = 1
            # find a valid name
            while 1:
                name = base_name + str(cnt)
                try:
                    self.filters.get_filter(name)
                except KeyError:
                    break # when it is not already in the filters set
                
                cnt += 1
                pass
            pass
        else:
            raise Exception("There is not the set of filters! you have set it at init-time")
        
        view['TEFilterName'].set_text( name )
        view['TEName'].set_text( self.RANGES_SEPARATOR.join(self.model.get_names()) )
        view['TETypes'].set_text( self.RANGES_SEPARATOR.join(self.model.get_types()) )
        view['TESteps'].set_text( self.RANGES_SEPARATOR.join(self.__steps()) )
        return
    
    #HELPERS
    def __steps(self):
        # replace -1 with ''
        ls = []
        for x,y in self.model.get_ranges():
            if x == y:
                # this means that there is only a column in the range
                ls.append(str(x))
                continue

            if x == -1:
                v1 = ''
            else:
                v1 = str(x)
                
            if y == -1:
                v2 = ''
            else:
                v2 = str(y)
                
            ls.append( v1+self.RANGE_CHAR+v2 )

        return ls

    def __parse_steps(self, text ):
        def empty_to_minus1( ls ):
            #transform the list of ranges and set the 'empty' to -1 in order to match the model requirements
            ls1 = []
            for r in ls:
                try:
                    x,y = r
                except ValueError:
                    #this means that there is only a value in the range
                    ls1.append( (int(r[0]),int(r[0])) )
                    continue
                
                if not x:
                    v1 = -1
                else:
                    v1 = int(x)
                    
                if not y:
                    v2 = -1
                else:
                    v2 = int(y)
                ls1.append( (v1,v2) )

            return ls1

        if not text:
            return []
        return empty_to_minus1( [r for r in [se.split(self.RANGE_CHAR) for se in text.split(self.RANGES_SEPARATOR)]] )


    def hide_dialog(self):
        self.view['window_filter'].hide()
        self.relieve_model(self.model)
        return

    #----------#
    #  SIGNALS #
    #----------#
    def on_cancel_clicked(self, btn):
        self.hide_dialog()
        return

    def on_delete_event(self, win, event):
        self.hide_dialog()
        return True # stops propagating the signal, to avoid window destruction

    def create_filter(self, btn):
        
        self.model.set_filter_name( self.view['TEFilterName'].get_text() )
        self.model.set_names( self.view['TEName'].get_text().split(self.RANGES_SEPARATOR) )
        self.model.set_types( self.view['TETypes'].get_text().split(self.RANGES_SEPARATOR) )
        self.model.set_ranges( self.__parse_steps( self.view['TESteps'].get_text() ) )
        
        self.filters.add_filters(self.model) #add the new filter
        
        #an iter that points to the rows inserted in the list store
        i = self.filters.get_ListStore().get_iter( 
            self.filters.get_filter(
                self.model.get_filter_name(), 
                index=True) 
            )

        self.cb.set_active_iter(i) #will raise changed signal
        self.main_ctrl.apply_stored_filter(self.cb)
        self.main_ctrl.save_filters() #save the filter
        
        self.hide_dialog() #hide the dialog
        
        return
    
    #if a text edit change
    def edit_changes(self, te):
        btn = self.view['BTNCreate']
        fname = self.view['TEFilterName'].get_text()
        steps = self.view['TESteps'].get_text()
        #types = self.view['TETypes']
        
        if fname == QFILTER: #no quick filter
            bfname = False #no invalid name
        else:
            bfname = True #ok! valid name

        bfname = bfname and fname # empty is not valid
        
        try:
            lssteps = [r for r in [se.split(self.RANGE_CHAR) for se in steps.split(self.RANGES_SEPARATOR)]
                       #if is valid
                       if (len(r) == 1 and int(r[0])) or ( r[0] == '' or int(r[0]) ) and ( r[1] == '' or int(r[1]) ) ]
            bsteps = True
        except:
            if not steps:
                bsteps = True #empty is a valid range
            else:
                bsteps = False 

        if not bfname or not bsteps: 
            btn.set_sensitive(False)
        else:
            btn.set_sensitive(True)

        return
# ----------------------------------------------------------------------
