# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by GMV Aerospace and Defence  
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================



from gtkmvc import Model

from src.models.slim import SlimModel
from src.models.properties import PropertiesModel
from src.models.mission import MissionModel
from src.models.correctness import CorrectnessModel
from src.models.performability import PerformabilityModel
from src.models.safety import SafetyModel
from src.models.fdir import FdirModel
from src.models.tool_handlers import ToolHandlersModel
from src.models.validation import ValidationModel
from src.models.tfpg import TfpgModel
import src.utils.globals

import utils.process
import options
from utils.log import log_icd
# this is used for the application user-config:
from ConfigParser import SafeConfigParser 

import os.path
import gtk, pango # for the licenses text


class ApplModel (Model):
    
    busy = 0 # busy level 
    __observables__ = ("busy",)

    def __init__(self):
        Model.__init__(self)

        # creates sub-models
        self.tools = ToolHandlersModel()

        self.slim = SlimModel(not bool(options.BASE_FILE_NAME), not bool(options.OUT_DIR_NAME))
        self.properties = PropertiesModel(self, self.slim)
        self.mission = MissionModel(self, self.slim)
        self.correctness = CorrectnessModel(self, self.slim, self.properties)
        self.performability = PerformabilityModel(self, self.slim, self.properties)
        self.safety = SafetyModel(self, self.slim, self.properties)
        self.fdir = FdirModel(self, self.slim, self.properties)
        self.tfpg = TfpgModel(self, self.slim, self.mission)
        self.validation = ValidationModel()

        # observes busy levels
        self.correctness.register_observer(self)
        self.safety.register_observer(self)
        self.fdir.register_observer(self)
        self.performability.register_observer(self)

        # check if running in degraded mode
        try: 
            path = utils.process.get_exec_name(options.NUSMV)
            self.__degraded = False
        except IOError:
            path = utils.process.get_exec_name(options.DEGRADED_NUSMV)
            self.__degraded = True
            pass
        
        # configuration
        self.setup_user_config()
        return

    def setup_user_config(self):
        """This method sets up the user config file and corresponding
        parser"""
        self.config = SafeConfigParser()
        self.config.read(src.utils.globals.USER_CONFIG_FILE)
        
        # creates all missing sections and options
        for section, opts_vals in \
                { 'Dialogs' : (('do_not_show_degraded_mode', 'no'),),
                  'Geometry': ()
                  }.iteritems():
            # creates missing sections
            if not self.config.has_section(section):
                self.config.add_section(section)
                pass

            # creates missing values
            for opt, val in opts_vals:
                if not self.config.has_option(section, opt):
                    self.config.set(section, opt, val)
                    pass
                pass
            pass
                
        return
            

    def reset_user_config(self):
        """Resets the user options to the defaults"""
        for sect in self.config.sections():
            for opt, val in self.config.items(sect):
                self.config.remove_option(sect, opt)
                pass
            pass
        return
            

    @log_icd
    def quit(self):
        """This method is called when it is time for the
        application to shutdown"""
        
        # kills any nusmv process that may be running
        self.correctness.stop_nusmv()
        self.safety.stop_fsap()
        self.fdir.stop_tools()

        # saves the user configuration
        conf_file = None
        try:
            conf_file = open(src.utils.globals.USER_CONFIG_FILE, "w")
            self.config.write(conf_file)
        except: pass 
        finally:
            if conf_file:
                conf_file.close()
            pass
    
        return

    def get_current_root(self):
        """Returns the currently selected root implementation"""
        return self.slim.root_impl

    def can_quit_safely(self):
        """Returns True if there is nothing loaded into the gui, so
        quitting would be not an issue"""
        return (self.busy == 0 and
                len(self.slim.slim_files) == 0 and
                len(self.slim.fdir_components) == 0 and
                len(self.slim.fault_injections) == 0 and
                len(self.properties.get_properties_list()) == 0)

    def is_degraded_mode(self): 
        """Returns True if the application is running in degraded mode"""
        return self.__degraded

    # ---------------------------------------------------------------
    #        Notifications from observable properties
    # ---------------------------------------------------------------    
    def property_busy_value_change(self, model, old, new):
        """Busy level is the number of activities currently runiing
        in background. The controller uses this value to report
        feedback to the user."""
        if model is self: return # not interested in introspection
        self.busy += (new - old)
        return

    pass # end of class ApplModel
# ----------------------------------------------------------------------


class TagTextBuffer (gtk.TextBuffer):
    """A TextBuffer with some tags ready to be used"""
    
    def __init__(self, table=None):        
        gtk.TextBuffer.__init__(self, table)
        
        self.create_tag("important", 
                        weight=pango.WEIGHT_BOLD)
        self.create_tag("comment", 
                        foreground="darkgray")
        return
    pass # end of class 
# ----------------------------------------------------------------------
    
        
class AboutModel (Model):
    def __init__(self, is_degraded):
        Model.__init__(self)

        eval_fname = os.path.join(src.utils.globals.LICENSES_DIR,
                                    "evaluation-license-4.0.txt")
        cpl_fname = os.path.join(src.utils.globals.LICENSES_DIR, 
                             "compass-license-1.0.txt")
        add_fname = os.path.join(src.utils.globals.LICENSES_DIR, 
                             "additional-components-license-1.0.txt")
        esa_fname = os.path.join(src.utils.globals.LICENSES_DIR, 
                                 "ESA-ONLY-additional-components-license-1.0.txt")

        if os.path.exists(eval_fname):
            self._is_evaluation = True
            main_fname = eval_fname
            second_fname = eval_fname
        else:
            self._is_evaluation = False
            main_fname = cpl_fname
            second_fname = add_fname

        # loads the CPL
        self._is_degraded = is_degraded
        self.tb_cpl = TagTextBuffer()

        self.tb_cpl.insert_with_tags_by_name(
            self.tb_cpl.get_end_iter(), 
            "The file containing this License can be found here:\n" +
            main_fname + "\n\n",
            "comment", "important")
        
        f = open(main_fname)
        text = f.read()        
        self.tb_cpl.insert(self.tb_cpl.get_end_iter(), text)
        f.close()

        # loads the additional license
        self.tb_bg = TagTextBuffer()

        if not os.path.exists(second_fname):
            # try with ESA-version
            second_fname = esa_fname
            pass
        
        self.tb_bg.insert_with_tags_by_name(
            self.tb_bg.get_end_iter(), 
            "The file containing this License can be found here:\n" +
            second_fname + "\n\n",
            "comment", "important")

        f = open(second_fname)
        text = f.read()        
        self.tb_bg.insert(self.tb_bg.get_end_iter(), text)
        f.close()

        return

    def is_degraded_mode(self): return self._is_degraded

    def is_evaluation(self): return self._is_evaluation

    pass # end of class ApplModel
# ----------------------------------------------------------------------
    


