# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================


# Model Checking and Simulation model

from gtkmvc import Model, observable

from src.models.activity import ActivityModel
from src.models.filtered_properties import FilteredPropertiesModel
from src.models.tool_handler import NuSMVHandlerModel
from src.models.trace import TraceModel, ExtensibleTraceModel
from src.models.guided import GuidedSimTransitionModel

#convert costraints
from slim_compiler.src.testexpr import parseExpression
from slim_compiler.src.smv.converter import SmvConverter
from slim_compiler.src.parse_property import check_property_tree

from invokers import nusmv
from utils.exceptions import NusmvControlledException

import options
import property_management.src.api as properties_api
import property_management.src.constants as prop_const

import os.path

from slim_compiler.src.testexpr import parseExpression
from slim_compiler.src.dataStructure.visitors.expressionVisitor import (TypeVisitor, )
import slim_compiler.src.smv.api as smvapi
from slim_compiler.src.smv import support
from slim_compiler.src.dataStructure import nominalbehaviour
from slim_compiler.src.parse_property import check_property_tree
from slim_compiler.src import testexpr
from invokers import nusmv
from slim_compiler.src.smv.walkers import ModeCollectorWalker


class CorrectnessModel (ActivityModel):
    """Model of correctness panel. This model contains two
    sub-models for model checking and simulation"""

    def __init__(self, appl_model, slim_model, props_model):

        ActivityModel.__init__(self, appl_model, slim_model, props_model)

        # the properties model
        self.props_model = props_model

        # this is the set of propertis filtered
        self.filtered_props = FilteredPropertiesModel(self.props_model)

        # there are three sub-models
        self.model_check = ModelCheckingModel(slim_model, self.filtered_props)
        self.model_sim = ModelSimulationModel(slim_model, self.filtered_props)
        self.model_dc = DeadlockCheckingModel(slim_model)
        self.model_zeno = ZenoAnalysisModel(slim_model)
        self.model_td = TimeDivergenceModel(slim_model)

        # monitors
        self.appl_model.tools.add_tool("Model Checking",
                                       self.model_check, "Correctness")
        self.appl_model.tools.add_tool("Simulation",
                                       self.model_sim, "Correctness")
        self.appl_model.tools.add_tool("Deadlock Checking",
                                       self.model_dc, "Correctness")
        self.appl_model.tools.add_tool("Zeno Analysis", self.model_zeno, "Correctness")
        self.appl_model.tools.add_tool("Time Divergence", self.model_td, "Correctness")

        # Observes busy levels
        self.model_check.register_observer(self)
        self.model_sim.register_observer(self)
        self.model_dc.register_observer(self)
        self.model_zeno.register_observer(self)
        self.model_td.register_observer(self)
        return

    def stop_nusmv(self):
        """This is called from the upper level when it is time to go sleep"""
        for m in (self.model_check, self.model_sim, self.model_dc, self.model_zeno, self.model_td):
            m.stop_nusmv()
        return

    # ---------------------------------------------------------------
    #        Notifications from observable properties
    # ---------------------------------------------------------------
    def property_busy_value_change(self, model, old, new):
        if model is self:
            return  # not interested in introspection
        self.busy += (new - old)
        return

    pass  # end of class


# ----------------------------------------------------------------------
class ModelCheckingModel (NuSMVHandlerModel):
    """This is the model of model checking"""

    use_bdd = True  # if bdd or sat should be used
    sat_bound = options.NUSMV_BMC_LENGTH
    klive_bound = options.NUSMV_KLIVE_BOUND
    use_sbmc = True
    sbmc_complete = False
    extended_model = False
    can_use_bdd = True  # if BDD can be used

    # FIXME for the moment only one result is supported
    result = None

    __observables__ = ('use_bdd', "result", "can_use_bdd", "extended_model")

    def __init__(self, slim_model, filtered_props_model):
        NuSMVHandlerModel.__init__(self, slim_model)
        # the model of the filtered properties
        self.filtered_props = filtered_props_model

        # monitors some properties of the slim model to change
        # availability of some of the model checking options.
        self.slim_model.register_observer(self)
        return

    def run_model_checking(self):
        self.result = None

        rows = [row for row in self.filtered_props
                if row[FilteredPropertiesModel.SELECT_MULTI]]
        return self._run_model_checking_aux(rows)

    # ---------------------------------------------------------------
    #        Notifications from observable properties
    # ---------------------------------------------------------------

    def property_infinite_domain_value_change(self, model, old, new):
        if model is self.slim_model:
            # infinite domain has to disable bdd-based engines
            self.can_use_bdd = not new

    def property_changed_signal_emit(self, model, arg):
        if model is self.slim_model:
            self.result = None

    # ---------------------------------------------------------------
    #        Internals
    # ---------------------------------------------------------------
    def _run_model_checking_aux(self, rows):
        """Internal method that checks all properties found in
        given rows (ListStorerow). This can be used by derived
        classes like FdirFraModel"""

        self.busy += 1
        try:
            self.start_nusmv(self.extended_model)

            # sets options used for model checking
            options.NUSMV_BMC_LENGTH = self.sat_bound
            options.NUSMV_KLIVE_BOUND = self.klive_bound
            options.NUSMV_USE_BMC = not self.use_bdd
            options.NUSMV_BMC_USE_SBMC = self.use_sbmc
            options.NUSMV_SBMC_IS_COMPLETE = self.sbmc_complete

            # properties are ordered here per logic: those which are
            # not MITL, and those which are MITL.  This minimizes the
            # regeneration of smv files in the two cases.
            ordered_rows = sorted(rows,
                                  key=lambda r: r[
                                      FilteredPropertiesModel.PROPERTY].
                                  property.get_logic() ==
                                  prop_const.Logic.MITL)

            # gets the list of selected properties (currently only one
            # property can be selected)
            properties = [row[FilteredPropertiesModel.PROPERTY].property
                          for row in ordered_rows]

            formulae_doc = properties_api.\
                create_formulae_document_for_properties(properties)
            smv_formulae_doc = properties_api.\
                create_formulae_document_for_properties(properties)

            properties_api.convert_atomic_propositions_to_smv(
                smv_formulae_doc,
                self.slim_model.get_extended_root_impl())

            ress = []
            for formula, smv_formula, row in \
                    zip(formulae_doc, smv_formulae_doc, ordered_rows):
                logic = smv_formula.get_logic()

                result = McResultModel()
                result.formula = smv_formula
                result.formula_text = properties_api.\
                    get_formula_as_smv_string(formula)
                result.type_name = {prop_const.Logic.CTL: "CTL",
                                    prop_const.Logic.LTL: "LTL",
                                    prop_const.Logic.MITL: "LTL"}[logic]
                result.sat_length = (self.sat_bound
                                     if not self.use_bdd
                                     else None)

                # calls the model checker
                if self.get_tool() is not None:

                    # this is done to trigger automatic regeneration
                    # if given formula's logics is not compatible with
                    # the currently generated smv.
                    self.set_timed_trans_fair_alt_mode(
                        logic != prop_const.Logic.MITL)

                    result.truth, trace_fname = \
                        nusmv.model_checking(self.get_tool(), smv_formula)
                    if self.get_tool() is None:
                        # nusmv was stopped
                        result = None
                    else:
                        result.create_trace_from_file(trace_fname)

                else:
                    result.error = "NuSMV died"

                # stores the result
                child_iter = self.filtered_props.get_properties_list().\
                    convert_iter_to_child_iter(row.iter)
                self.filtered_props.props_model.set_mc_result(child_iter,
                                                              result)

                ress.append(result)

            # show results if there is only one:
            if len(ress) == 1:
                self.result = ress[0]

        except NusmvControlledException, e:
            self.result = e
            return False

        finally:
            self.busy -= 1

        return

    pass  # end of class
# ----------------------------------------------------------------------


class ModelSimulationModel (NuSMVHandlerModel):
    """This is the model of model simulation"""
    RANDOM, TRANS_GUIDED, VALUE_GUIDED = range(3)
    types = set((RANDOM, TRANS_GUIDED, VALUE_GUIDED))

    sim_len = options.NUSMV_SIM_LENGTH
    sim_type = RANDOM
    extended_model = False
    extend_trace = False
    wait_for_picking = False #if it is waiting for selecting the simulation starting point
    sim_failed = observable.Signal()
    invalid_trace = observable.Signal()
    deadlock_found = observable.Signal() # signal is emitted along with the trace
    new_trace = observable.Signal() # raised when the user try to simulate a trace after extended_model was changed
    sim_stopped = observable.Signal() # raised when the user stops the simulation
    slim_changed = observable.Signal() # raised when the slim model changes, and the result is already None
    can_simulate = True # simulation can be run

    # footer error message
    footer = """
See the logging tab in the "Model" pane, for more information
on what happened.
"""


    # FIXME for the moment only one result is supported
    result = None
    __observables__ = ("result", "sim_type", "wait_for_picking",
                       "sim_failed", "invalid_trace", "extend_trace",
                       "deadlock_found", "new_trace", "sim_stopped", "slim_changed",
                       "can_simulate", "extended_model")

    def __init__(self, slim_model, properties):
        NuSMVHandlerModel.__init__(self, slim_model, timed_trans_fair_alt_mode=False)
        self.filtered_props = properties
        self.starting_point = 1
        self.prev_trace_fname = None

        # converter for the values added by user
        self.converter = SmvConverter()
        self.converter.set_id_prefix(self.converter.root_var() + ".")
        # this variable is used to stop the simulation thread in the right way
        # otherwise the simulation may think that there is a problem in the model
        self.__stopped_by_user = False

        self.trans_guided = GuidedSimTransitionModel(self.slim_model, self, timed_trans_fair_alt_mode=False)
        self._guided_needs_recalculation = True
        return

    #type of the simulation
    def set_type(self, t):
        """
        set the simulation type in (RANDOM, TRANS_GUIDED, VALUE_GUIDED)
        """
        assert t in self.types

        self.sim_type = t
        return

    def get_type(self):
        """get the simulation type"""
        return self.sim_type

    def get_types(self):
        """get the possible simulation types"""
        return self.types

    def pick_set(self, start):
        """
        select the starting point, and begin to extend the trace
        (with the next call of run_model_simulation())
        """
        assert type(start) is int, "the starting point for the picking must be an integer!"

        self.starting_point = start
        self.pick_selected_state()

        return True


    def pick_unset(self):
        """
        unset the state previously picked
        """
        assert not isinstance( self.result, NusmvControlledException ), "This method cannot be called when the system is in an error state!"
        self.starting_point = self.result.steps() if self.result else 1
        return

    def get_filtered_properties(self):
        """gets the list of selected properties"""
        properties = [ row[FilteredPropertiesModel.PROPERTY].property
                       for row in self.filtered_props if row[FilteredPropertiesModel.SELECT_MULTI] ]
        formulae_doc = properties_api.create_formulae_document_for_properties(properties)
        smv_formulae_doc = properties_api.create_formulae_document_for_properties(properties)
        properties_api.convert_atomic_propositions_to_smv(smv_formulae_doc,
                                                          self.slim_model.get_extended_root_impl())

        smv_formulae = [x for x in smv_formulae_doc]
        smv_prop_formula = [properties_api.get_formula_as_smv_string(f)
                            for f in smv_formulae
                            if properties_api.check_is_propositional_formula(f)]

        return smv_prop_formula

    def get_results_forced_values(self):
        """
        @return the constraints added by user as a list of strings (only in guided simulation)
        """
        nusmv_costraints = []
        if self.result and self.result.trace:
                # I have to add to the costraints, the value inserted in the old trace by the user
            assert self.result.trace.__class__ == ExtensibleTraceModel
            c_dict = self.result.trace.get_forced_values() #costraints dict
                # generate slim_costraints
            slim_costraints = [slim_name+'='+str(value) for slim_name,value in c_dict.iteritems()\
                                   if slim_name and value]
            nusmv_costraints = []
                # convert each costraint in nusmv
            for c in slim_costraints:
                # InvalidExpressionException is avoided by the check at insert-time
                ast = parseExpression(c)
                check_property_tree( self.slim_model.compile_info, ast, self.slim_model.get_extended_root_impl() )
                nusmv_c = self.converter.expression( self.slim_model.get_extended_root_impl(), ast )
                nusmv_costraints.append( nusmv_c )
                pass
            pass
        return nusmv_costraints

    def reload_previous_trace(self):
        """
        reload the previous trace in the system (if someone bad happened)
        and return the filename of the trace loaded.
        """
        trace_fname = None
        nusmv = self.get_tool()

        if nusmv is None:
            raise NusmvControlledException( """There is a communication problem with NuSMV.
Probably the process died.
For this reason it is impossible load the simulation into NuSMV."""+self.footer )

        # try to recover a trace if it is necessary
        if self.prev_trace_fname:
            if self.prev_trace_fname[0] == self.extended_model and os.path.isfile(self.prev_trace_fname[1]):
                # assert that the trace is in the system
                # must return an index if the trace is loaded correctly, None otherwise..
                # we handle it with a signal
                idx = nusmv.read_trace( self.prev_trace_fname[1] )
                if idx:
                    trace_fname = nusmv.goto_state(idx, self.starting_point)
                    if trace_fname is None:
                        raise NusmvControlledException("""An error occurred while generating the trace.
This means that you have to restart the simulation, because the previous trace was lost.
Please report this bug.""")

                else:
                    self.result = None
                    self.invalid_trace.emit()
                    return None

            elif self.prev_trace_fname[0] != self.extended_model:
                # if the old trace was loaded with a different value of extended_model
                # we have to recalculate it
                self.result = None
                #recreated = True
                self.new_trace.emit()
                pass
            pass

        return trace_fname

    # helper function for simulation and picking
    def __pick_selected_state(self, sim_len, trace_fname=None):
        """
        this function (not directly callable) is used to jump to a previous state.
        @param optionally a filename of the trace loaded in the system
        @return the filename of the trace that lasts in the step choosen by user
        (this function is called automatically inside the pick_set() function)
        """
        header = """<span foreground="red" weight="bold">It was not possible to jump.</span>
"""
        no_result = """A trace must be created before jumping to a previous state."""
        nusmv_err = """Probably <b>NuSMV is not working</b>. """
        jump_failed = "Jumping to the state "+str(self.starting_point)+" failed!"

        no_state = "The state "+str(self.starting_point)+" does not exists! This is probably a bug into the tool. Please report this bug."

        if not self.result:
            raise Exception( header+no_result )

        nusmv = self.get_tool()
        if not nusmv:
            raise Exception( header+nusmv_err+self.footer )

        smv_prop_formula = self.get_filtered_properties()

        #create a trace with the end point as user has selected
        #go to state
        if self.starting_point == 1:
            #create a new trace
            trace_fname = nusmv.pick_state_random(smv_prop_formula)
        elif self.starting_point < self.result.steps():
            # I don't use 'not goto' because if goto == None the first branch is choosen
            if trace_fname is None:
                trace_fname = nusmv.goto_state(nusmv.get_traces_num(), self.starting_point)
                pass
        else:
            raise Exception( header+no_state+self.footer )

        if trace_fname is None:
            raise Exception( header+jump_failed+self.footer )

        return trace_fname

    def __pick_initial_state(self, sim_len, trace_fname=None):
        """
        this function (not directly callable) is used to restart the simulation from the step 1.
        It is more efficient than pick_set(1).
        @param optionally the trace filename of the trace loaded in the system.
        @return the filename of the trace that ends in the first step (that has no transitions)
        or None if an error occoured
        """
        err = """<span foreground="red" weight="bold">It was not possible to restart the simulation.</span>
It is <b>impossible to pick an initial state</b>.
Probably the set of initial states is <b>empty</b>
(namely, the constraints are too strong)"""

        nusmv = self.get_tool()
        assert nusmv, "Nusmv handler can't be None!"
        smv_prop_formula = self.get_filtered_properties()
        trace_fname = nusmv.pick_state_random(smv_prop_formula)
        # this means that the set of initial states is EMPTY. No state can be chosen.
        if trace_fname is None:
            raise Exception(err)

        assert type(trace_fname) is str and os.path.isfile(trace_fname), "A filename must be generated! not "+str(trace_fname)
        return trace_fname

    def __update_results(self, trace_fname):
        """
        update the result field, and store the previous one in order to restore it if some errors occurred.
        raises an assertionError if no trace will be loaded.
        @param trace_fname the file name of the newly created trace.
        @return the SimResultModel instance of the last result.
        side effect: stores the result in an internal variable, that can affect the simulation.
        """
        if self.__nusmv_was_stopped():
            return None

        if trace_fname is None:
            self.stop_nusmv()
            self.sim_failed.emit()
            return None

        self.prev_trace_fname = (self.extended_model, trace_fname) #save the name of the last used trace

        result = SimResultModel() #the new result

        result.create_trace_from_file(trace_fname, th=ExtensibleTraceModel)
        assert result.trace, "no trace created!"
        result.trace.set_slim_model(self.slim_model) #set the model for checking
        return result

    def __simulate_random(self, sim_len, trace_fname=None):
        """
        executes a random simulation if it is possible
        return None if an error occoured.
        """
        error_header = """<span foreground="red" weight="bold">It was not possible to simulate.</span>
"""

        smv_prop_formula = self.get_filtered_properties()
        nusmv_costraints = map( lambda x: "next(%s)" % x, self.get_results_forced_values() )
        nusmv = self.get_tool()

        if nusmv is None:
            raise Exception(
                error_header+"""There is a communication problem with NuSMV.
Probably the process died"""+self.footer
            )

        if not self.result or not self.result.trace:
            trace_fname = nusmv.pick_state_random(smv_prop_formula)
            if self.__nusmv_was_stopped(): return None
            if trace_fname is None: return None
            pass

        trace_fname = nusmv.simulate_random(nusmv_costraints+smv_prop_formula, sim_len)
        return trace_fname


    def __simulate_trans_guided(self, sim_len, trace_fname=None):
        """
        executes a guided simulation if it is possible
        return None if an error occoured.
        """
        error_header = """<span foreground="red" weight="bold">It was not possible to simulate.</span>
"""

        smv_prop_formula = self.get_filtered_properties()
        trans_costraints = self.trans_guided.get_trans_constraints()
        nusmv = self.get_tool()

        if nusmv is None:
            raise Exception(
                error_header+"""There is a communication problem with NuSMV.
Probably the process died"""+self.footer
            )

        if not self.result or not self.result.trace:
            trace_fname = nusmv.pick_state_random(smv_prop_formula)
            if self.__nusmv_was_stopped(): return None
            return trace_fname

        trace_fname = nusmv.simulate_random(trans_costraints+smv_prop_formula, 1)
        return trace_fname


    def __nusmv_was_stopped(self):
        #say yes is nusm was stopped, no otherwise
        #if yes, raise the stopped signal
        if self.__stopped_by_user:
                #leave everithing as it is, and launch simulation_stopped signal
            self.sim_stopped.emit()
            self.__stopped_by_user = False
            return True
        else:
            return False

    # end
    ##### Main functions that evaluate a generic computation (like simulation or jump or picking)
    def __run_model_simulation(self, fun, simulation):
        """
        run a generic operation such as simulation, jump or picking.
        raises deadlock_found if a deadlock was found.
        If the function raises an exception, the signal sim_failed will be raised,
        with the exception instance as parameter.
        @param simulation is a boolean, True if this is a simulation, false otherwise
        @param fun is a function that performs a generic simulation.
        It must take a string, that is the trace reloaded (if there is one)
        These function can be __pick_initial_state, __pick_selected_state and __simulate_random.
        @return True if no error was encountered, False otherwise.
        """
        trace_fname = None

        if not isinstance( self.result, ResultModel ) and self.result is not None:
            self.result = None
            pass

        self.busy += 1
        try:
            started = self.start_nusmv(self.extended_model)
            # sets options used for model checking
            sim_len = self.sim_len if self.get_type() == self.RANDOM else 1
            options.NUSMV_SIM_LENGTH = sim_len

            if started:
                trace_fname = self.reload_previous_trace()
                pass

            # simulation
            if callable(fun):
                try:
                    trace_fname = fun(sim_len, trace_fname)
                except Exception, e:
                    self.stop_nusmv()
                    self.sim_failed.emit(e)
                    return False
            else:
                raise ValueError("The parameter must be a function")

            if self.__nusmv_was_stopped(): #do not change anything
                return False

            result = self.__update_results(trace_fname)
            if not result: #if someone goes wrong (like sim_failed)
                return False

            assert isinstance( result, SimResultModel ), "The result must be a SimResultModel! not "+str(result)
            #if the number of steps in the trace is lower or equal than the steps selected
            if simulation and sim_len > result.steps():
                self.starting_point = result.steps()
                self.deadlock_found.emit(result.trace)
                return False

            self.starting_point = result.steps() #set the starting point to the last step
            #update
            self.result = result

        except NusmvControlledException, e:
            self.result = e
            return False

        finally: self.busy -= 1

        return True
    # end

    #function that stops the simulation
    def stop_simulation(self):
        self.stop_nusmv()
        self.__stopped_by_user = True
        return

    # functions that compute simulation or picking
    def run_model_simulation(self):
        """
        create a new trace usign nusmv,
        The newly created trace is stored in the observable properties 'result'
        """
        if self.get_type() == ModelSimulationModel.RANDOM:
            return self.__run_model_simulation(self.__simulate_random, True)

        if self.get_type() == ModelSimulationModel.TRANS_GUIDED:
            return self.__run_model_simulation(self.__simulate_trans_guided, True)

        if self.get_type() == ModelSimulationModel.VALUE_GUIDED:
            return self.__run_model_simulation(self.__simulate_random, True)

        assert False, "no choice"
        return

    def pick_initial_state(self):
        """
        this method is used to restart the simulation.
        """
        return self.__run_model_simulation(self.__pick_initial_state, False)

    def pick_selected_state(self):
        """
        pick the state selected via pick_set().
        This method is automatically called when pick_set() is called.
        Probably it is better to avoid calling it directly.
        """
        return self.__run_model_simulation(self.__pick_selected_state, False)


    def calculate_guided_structure(self, trace_model=None):
        """Calculates and evaluates the structure of guided
        simulation if needed"""
        if self._guided_needs_recalculation:

            if trace_model is None:
                # tries with the trace which is stored currently in
                # the simulation
                if self.result and self.result.trace and \
                        isinstance(self.result.trace, TraceModel):
                    trace_model = self.result.trace
                    pass
                pass

            if trace_model is None: return

            self.trans_guided.evaluate_structure(trace_model)
            self._guided_needs_recalculation = False
            pass

        return


    # ---------------------------------------------------------------
    #        Notifications from observable properties
    # ---------------------------------------------------------------
    def property_changed_signal_emit(self, model, arg):
        super(ModelSimulationModel, self).property_changed_signal_emit(model, arg)
        if model is self.slim_model:
            if not self.result:
                # if the result is already none, the observers have to be notified!
                # it is not sufficient to set result to None
                self.slim_changed.emit()
            else:
                self.result = None
                pass

            self.prev_trace_fname = None

            self.trans_guided.build_structure(self.extended_model)
            self._guided_needs_recalculation = True
            self.can_simulate = True
            pass

        return


    def property_result_value_change(self, model, old, new):
        """A trace has been generated"""
        if model != self: return

        if self.get_type() == self.TRANS_GUIDED:
            trace = new.trace if (new and
                                  new.trace and
                                  isinstance(new.trace, TraceModel)) \
                                  else None

            if self.trans_guided.built_structure_is_extended != self.extended_model:
                self.trans_guided.build_structure(self.extended_model)
                pass

            self._guided_needs_recalculation = True
            self.calculate_guided_structure(trace)
            self.can_simulate = False
            pass
        else: self._guided_needs_recalculation = True
        return

    pass # end of class
# ----------------------------------------------------------------------


# ----------------------------------------------------------------------
class DeadlockCheckingModel (NuSMVHandlerModel):
    """This is the model of deadlock checking"""

    available = True # if the feature can be used
    status = "ready" # can be "ready", "running", "yes", "no", "error"
    extended_model = False

    __observables__ = ('available', "status", "extended_model")

    def __init__(self, slim_model):
        NuSMVHandlerModel.__init__(self, slim_model)

        # monitors some properties of the slim model to change
        # availability.
        self.slim_model.register_observer(self)
        return

    def run_deadlock_checking(self):
        self.busy += 1
        try:
            self.status = "running"
            self.start_nusmv(self.extended_model)
            result = self.get_tool().check_fsm_deadlocks()
            self.status = { True : "yes",
                            False : "no",
                            None : "error" }[result]

        except: self.status = "error"
        finally: self.busy -= 1
        return

    # ---------------------------------------------------------------
    #        Notifications from observable properties
    # ---------------------------------------------------------------

    def property_infinite_domain_value_change(self, model, old, new):
        if model is self.slim_model:
            # infinite domain has to disable bdd-based engines
            self.available = not new
            pass

        return

    # this observes the slim model. If the model changes, resets
    # the status to "ready"
    def property_changed_signal_emit(self, model, arg):
        NuSMVHandlerModel.property_changed_signal_emit(self, model, arg)
        if model is self.slim_model:
            self.available = not self.slim_model.infinite_domain
            if self.available: self.status = "ready"
            pass
        return


    pass # end of class
# ----------------------------------------------------------------------


# ----------------------------------------------------------------------

# LB: the new class for ZENO Analysis Model


class ZenoAnalysisModel (NuSMVHandlerModel):
    """This is the model of zeno analysis"""

    busy = 0
    zeno_result = None
    sat_bound = options.NUSMV_BMC_LENGTH
    k_zeno = options.NUSMV_KLIVE_BOUND
    last_steps = 0
    last_time = '[None]'
    last_mode = '[None]'
    extended_model = False
    zeno_results_list = []
    results_list_changed = observable.Signal()

    # Observable properties
    __observables__ = ("busy", "zeno_result", "sat_bound", "k_zeno", "last_mode", "last_steps", "last_time",
                        "extended_model", "results_list_changed", "zeno_results_list", )


    def __init__(self, slim_model):
        NuSMVHandlerModel.__init__(self, slim_model, timed_trans_fair_alt_mode=False)

        # monitors some properties of the slim model to change
        # availability.
        self.slim_model.register_observer(self)
        return

    def stop_zeno_analysis(self):
        """Stops the Zeno Analysis."""
        self.stop_nusmv()
        self.zeno_result = ZenoResult('stop', None)
        return

    
    def get_zeno_states(self, root_impl):
        """Produces an iterable of tuples (impl_full_name, impl,
        zeno_state_slim, zeno_state_smv)

        where:
          impl_full_name: full name of implementation, from given root
                          Empty string when referring to the root.
          impl: implementation instance

          zeno_state_slim: slim expression (string) representing zeno state

          zeno_state_smv: smv expression (string) to be used as zeno state
                          constraint when calling check_is_state_zeno
        """
        # Gets the modes from the slim_model

        walker = ModeCollectorWalker()

        if root_impl is not None:
            walker.walk(root_impl)
        expr_str_fmt = "{inst_name}mode = mode:{mode_lit}"

        # checks and reports
        for name, impl, modes in walker.get_found_modes():

            for lit in modes:
                # construct the formula for mode
                expr_str = expr_str_fmt.format(
                    inst_name="",
                    mode_lit=lit)

                expr_smv = smvapi.parse_and_convert_predicate(
                    impl, expr_str, id_prefix=name)

                yield (name, impl, expr_str, expr_smv)
                
                
    def analyze_zeno(self, modes):
        """Performs Zeno Analysis and generates the results."""

        # Get the modes selected by the user

        modes_selected = []
        for enum_mode in modes:
            if enum_mode[0]:
                modes_selected.append([enum_mode[1], enum_mode[3]])

        # main zeno validation loop

        self.busy += 1
        try:

            if modes_selected:
            
                for mode in modes_selected:
                   
                    self.start_nusmv(self.extended_model)

                    # Gets the time_var
                    time_var, tu = smvapi.get_smv_timed_var()
                    
                    # Gets the results from NuSMV function
                    ret, val = self.get_tool().is_state_zeno(mode[1], time_var, self.sat_bound, self.k_zeno)
                    
                    if self.get_tool() is None:
                        return  # crash

                    result = self.get_tool().zeno_result_to_string(ret)

                    self.last_mode = mode[0]

                    if result == 'ERROR':
                        self.zeno_result = ZenoResult("error", None)
                        return

                    elif result == 'UNREACHABLE':
                        self.zeno_result = ZenoResult('unreachable', None)

                    elif result == 'NON-ZENO':
                        self.zeno_result = ZenoResult('non_zeno', val)

                    elif result == 'ZENO':
                        # val is the trace filename
                        self.zeno_result = ZenoResult('zeno', val)

                    elif result == 'NOT_FOUND':
                        # val is the number of steps
                        self.last_steps = val
                        self.zeno_result = ZenoResult('unknown_not_found', None)

                    else:
                        #If the result is prove, it returns a SMV rational value -> it is converted in SLIM
                        self.last_time = smvapi.convert_smv_const_to_slim(val)
                        self.zeno_result = ZenoResult('unknown_not_proved', None)

                    # The self.zeno_results_list is the list of all results with the clock
                    # It is used in the function on_results_selection_changed of ctrl.
                    
                    self.zeno_results_list.append([self.zeno_result, self.last_mode])

                    self.stop_nusmv()

            else: self.zeno_result = ZenoResult("selection", None)

        finally: self.busy -= 1

        return


    pass # end of class
# ----------------------------------------------------------------------


class ZenoResult ():
    """Represents the container for the Zeno Analysis result,
        having the success flag (a string) and possibly a ResultModel
        with the trace.
    """

    def __init__(self, _success, trace_fn):
        """Constructor Method"""
        self.success = _success

        # something went wrong
        if _success == "error":
            self.result = None
            return

        # there's a counterexample
        if _success == "zeno" or _success == "non_zeno":
            self.result = ResultModel()
            self.result.create_trace_from_file(trace_fn)

        # there's no counterexample
        else: self.result = None

        return

pass # end of ZenoResult class
# ----------------------------------------------------------------------


# ----------------------------------------------------------------------

# LB: the new class for Time Divergence Model

class TimeDivergenceModel (NuSMVHandlerModel):
    """This is the model of zeno analysis"""

    busy = 0
    td_result = None
    sat_bound = options.NUSMV_BMC_LENGTH
    last_clock = '[None]'
    last_time = '[None]'
    extended_model = False
    td_results_list = []
    results_list_changed = observable.Signal()

    # Observable properties
    __observables__ = ("busy", "td_result", "sat_bound", "last_clock", "extended_model",
                    "results_list_changed", "td_results_list", "last_time", )


    def __init__(self, slim_model):
        NuSMVHandlerModel.__init__(self, slim_model,
                                   timed_trans_fair_alt_mode=False)

        # monitors some properties of the slim model to change
        # availability.
        self.slim_model.register_observer(self)
        return


    def stop_time_divergence(self):
        """Stops the Time Divergence."""
        self.stop_nusmv()
        self.td_result = TimeDivergenceResult('stop', None)
        return


    def analyze_time_divergence(self, clocks):
        """Performs Zeno Analysis and generates the results."""

        # Get the clocks selected by the user

        clocks_bounds_unit_selected = []
        for clock in clocks:
            if clock[0] is True:
                clocks_bounds_unit_selected.append([clock[1], clock[2], clock[3]])

        # main time divergence loop

        self.busy += 1
        try:

            if clocks_bounds_unit_selected:

                for clock_bound_unit in clocks_bounds_unit_selected:

                    self.start_nusmv(self.extended_model)

                    # Gets the formula in SMV format from the root_impl and the slim_clock
                    dsc = support.get_subcomponent_from_path(self.slim_model.get_extended_root_impl(), clock_bound_unit[0])
                    assert isinstance(dsc, nominalbehaviour.DataSubcomponent)
                    assert isinstance(dsc.data_type, nominalbehaviour.Clock)
                    
                    # Converts the bound into SMV format
                    bound_unit = clock_bound_unit[1] + " " + clock_bound_unit[2]
                    
                    if bound_unit not in ("-1", "", None, "None"):
                        bound_smv = smvapi.parse_and_convert_numeric_expression(self.slim_model.get_extended_root_impl(), bound_unit, dest_tu=dsc.data_type.time_unit)
                    else:
                        bound_smv = None
                    
                    clock_smv = smvapi.parse_and_convert_id(self.slim_model.get_extended_root_impl(), clock_bound_unit[0])
                    
                    # Gets the results from NuSMV function
                    ret, val = self.get_tool().check_clock_divergence(clock_smv, bound_smv, self.sat_bound)

                    if self.get_tool() is None: return

                    result = self.get_tool().clock_diverge_result_to_string(ret)

                    self.last_clock = clock_bound_unit[0]


                    if result == 'ERROR':
                        self.td_result = TimeDivergenceResult('error', None)
                        return

                    elif result == 'UNKNOWN':
                        self.td_result = TimeDivergenceResult('unknown', None)

                    elif result == 'BOUNDED':
                        #If the result is bounded, it returns a SMV rational value -> it is converted in SLIM
                        self.last_time = smvapi.convert_smv_const_to_slim(val)
                        self.td_result = TimeDivergenceResult('bounded', None)

                    else:
                        #If the result is unbounded, it returns a trace
                        self.td_result = TimeDivergenceResult('unbounded', val)

                    # The self.td_results_list is the list of all results with the clock
                    # It is used in the function on_results_selection_changed of ctrl.

                    self.td_results_list.append([self.td_result, self.last_clock])

                    self.stop_nusmv()

            else: self.td_result = TimeDivergenceResult("selection", None)

        finally: self.busy -= 1

        return


    pass # end of class
# ----------------------------------------------------------------------


class TimeDivergenceResult ():
    """Represents the container for the TimeDivergence result,
        having the success flag (a string) and possibly a ResultModel
        with the trace.
    """

    def __init__(self, _success, trace_fn):
        """Constructor Method"""
        self.success = _success

        # something went wrong
        if _success == "error":
            self.result = None
            return

        # there's a counterexample
        if _success == "unbounded":
            self.result = ResultModel()
            self.result.create_trace_from_file(trace_fn)

        # there's no counterexample
        else: self.result = None

        return

pass # end of TimeDivergenceResult class
# ----------------------------------------------------------------------



class ResultModel (Model):
    """This is the logics for one result obtained by MC or simulation"""

    trace = None
    reset = False
    __observables__ = ('trace',)

    def steps(self):
        """
        @return the last filled step in the result (can exist empty steps, to be filled manually for instance)
        or 1 if there is no a valid trace.
        """
        bs = self.trace.base_steps() #if bs == 0, none has been added empty steps
        if self.trace:
            result = bs if bs else self.trace.get_max_steps()
        else:
            result = 1
        assert result is not None, "This value cannot be None, but an integer"

        return result

    def create_trace_from_file(self, trace_fname, th=TraceModel):
        """Given a xml file containing a nusmv trace, creates and
        sets a TraceModel instance and sets the internal trace
        field. If None is passed, the trace is set to None"""
        if trace_fname is None: self.trace = None; return

        from src.utils.nusmv import NuSMVTraceReader
        reader = NuSMVTraceReader()
        self.trace = reader.parse(trace_fname, th)
        assert self.trace is not None, "Error while parsing trace file %s" % trace_fname
        return

    pass # end of class
# ----------------------------------------------------------------------


class McResultModel (ResultModel):
    """Result specific for model checking"""
    formula = None
    formula_text = None
    type_name = None
    truth = None
    sat_length = None
    error = None

    def __str__(self):
        return "formula:"+str(self.formula)+"formula_text:"+str(self.formula_text)+"type_name:"+str(self.type_name)+"truth:"+str(self.truth)+"sat_length:"+str(self.sat_length)+"error:"+str(self.error)
    pass # end of class
# ----------------------------------------------------------------------


class SimResultModel (ResultModel):
    """Result specific for model simulation"""
    pass # end of class
# ----------------------------------------------------------------------
