# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================


from gtkmvc import Model

from src.models.activity import ActivityModel
from src.models.filtered_properties import FilteredPropertiesModel
from src.models.safety import FsapHandlerModel
from src.models.correctness import ModelCheckingModel
from src.models.correctness import ResultModel
from src.models.performability import *
from src.models.trace import TraceModel
from slim_compiler.src.smv.api import get_alarms

from utils.main_extension_converter import MainExtensionConverter

from utils import fsap_process
import utils.log
from utils.misc import relpath, get_out_fname
import options

import slim_compiler.src.smv.api as smv_api

import property_management.src.api as properties_api
from property_management.src.properties import Property

from gtk import ListStore
import types
import os
import os.path


class FdirModel (ActivityModel):
    """Model of safety panel. It contains submodels for fault tree
    generation, and failure mode and analysis"""

    busy = 0
    __observables__ = ("busy",)

    def __init__(self, appl_model, slim_model, props_model):        
        ActivityModel.__init__(self, appl_model, slim_model, props_model)
        
        # the properties model
        self.props_model = props_model

        # this is the set of properties filtered
        self.filtered_props = FilteredPropertiesModel(self.props_model)

        # sub-models:
        self.fda = FdirFdaModel(slim_model, self.filtered_props)
        self.fia = FdirFiaModel(slim_model)
        self.fra = FdirFraModel(slim_model, self.filtered_props)
        self.diag = FdirDiagModel(slim_model, self.filtered_props)
        self.fca = FdirFcaModel(slim_model, self.filtered_props, self.props_model)

        # monitors
        self.appl_model.tools.add_tool("Detection", self.fda, "FDIR")
        self.appl_model.tools.add_tool("Identification", self.fia, "FDIR")
        self.appl_model.tools.add_tool("Recovery", self.fra, "FDIR")
        self.appl_model.tools.add_tool("Diagnosability", self.diag, "FDIR")
        self.appl_model.tools.add_tool("Coverage", self.fca, "FDIR")

        # interested in seeing what fra, fda, fia, and diag do
        self.fda.register_observer(self)
        self.fia.register_observer(self)
        self.fra.register_observer(self)
        self.diag.register_observer(self)
        self.fca.register_observer(self)

        # interested in seeing what the slim model does
        slim_model.register_observer(self)

    # ---------------------------------------------------------------
    #        Notifications from observable properties
    # ---------------------------------------------------------------    
    def property_changed_signal_emit(self, model, args):
        """cleanup of submodels is required upon slim model changes."""
        self.fda.fda_list.clear()
        self.fia.ftrees_list.clear()
        return

    def stop_tools(self):
        """This is called from the upper level when it is time to go sleep"""
        for fsap in (self.fda, self.fia, self.diag, self.fca): fsap.stop_fsap()
        self.fra.stop_nusmv()
        return

    # ---------------------------------------------------------------
    #        Notifications from observable properties
    # ---------------------------------------------------------------    
    def property_busy_value_change(self, model, old, new):
        """Busy level is the number of activities currently runiing
        in background. The controller uses this value to report
        feedback to the user."""
        if model is self: return # not interested in introspection
        self.busy += (new - old)
        return


    pass # end of class
# ----------------------------------------------------------------------


class FdirDiagModel (FsapHandlerModel):
    """Model for handling Diagnosability Analysis (DIAG)"""

    sat_bound = 10
    diag_result = None
    diag_result_shown = False
    
    last_property_name = '[None]'
    diag_ctx = None
    diag_delay = -1.0 # It is the Diagnosability Delay (by default -1.0)
    
    __observables__ = ("diag_result", "diag_delay")
    

    def __init__(self, slim_model, properties):
        FsapHandlerModel.__init__(self, slim_model)

        self.filtered_props = properties
        self.restrictions_store = ListStore(gobject.TYPE_PYOBJECT)

        def _display_property_in_cell(column, cell, model, iter):
            # for displaying a property object in a cell
            cell_prop = model.get_value(iter, 0)
            cell_text = cell_prop._get_description()
            if cell_text is None or len(cell_text) < 1:
                cell_text = "[This property has no description.]"
            cell.set_property('text', cell_text)
            return

        rend = gtk.CellRendererText()
        self.restriction_col = gtk.TreeViewColumn('Path Restrictions', rend)
        self.restriction_col.set_cell_data_func(rend, _display_property_in_cell)

        return

    def check_observations(self):
        """Checks if there are observations in the model. Returns
        True if there are observations, or False otherwise."""
        self.start_fsap_obs()
        assert self.failures >= 0 and self.observations >= 0, "Something is broken in the toolchain"
        return self.observations > 0

    def get_restrictions_from_store(self, store):
        """Iterates the list store containing the path restrictions
        and returns a list containing their SMV equivalents as strings."""

        restrictions_buffer = [ row[0] for row in store ]
        smv_formulae_doc = properties_api.create_formulae_document_for_properties(restrictions_buffer)
        properties_api.convert_atomic_propositions_to_smv(smv_formulae_doc, self.slim_model.get_extended_root_impl())

        smv_restrictions = []
        for smv_prop in smv_formulae_doc:
            smv_prop_text = properties_api.get_formula_as_smv_string(smv_prop)
            smv_restrictions.append(smv_prop_text)

        return smv_restrictions

    def set_delay(self, val, unit):
        """Given a value (i.e. an integer) and a time unit, it builds a slim expression and it returns its smv
           translation.
        """
        assert unit in ['msec', 'sec', 'min', 'hour', 'day', None]
        #FIXME: using regular expressions to understand whether it is an integer or a float
        if "." not in val:
            val += ".0"
        if unit is not None:
            time_str = "%s %s" %(str(val), str(unit))
        else:
            time_str = str(val)
        smv_delay = smv_api.parse_and_convert_numeric_expression(self.slim_model.get_extended_root_impl(), time_str)
        if "f'" in str(smv_delay):
            smv_delay = smv_delay.replace("f'", "")
        float_delay = eval(smv_delay)
        self.diag_delay = float_delay

        return

    def go_diag(self):
        """Performs diagnosability analysis and stores the result object
        in the diag_result field of the model."""

        self.busy += 1
        try:
            utils.log.system.info('Starting Diagnosability Analysis.')

            # Set to False to force that we want to see the result (if set to True, no result are shown)
            self.diag_result_shown = False
            self.start_fsap_obs()

            # model checking options
            options.NUSMV_BMC_LENGTH = self.sat_bound

            # diag options
            options.DIAG_DELAY = self.diag_delay

            # verify and prepare observables
            utils.log.system.info('Retrieving names of observable variables.')
            if not self.check_observations():
                raise exceptions.UserException("There are no observables in the model. Broken toolchain.")
            nominal_root_implementation = self.slim_model.get_nominal_root_impl()
            observables = smv_api.get_observations(nominal_root_implementation)

            # prepare the properties
            utils.log.system.info('Retrieving the diagnosability property')
            properties = [ row[FilteredPropertiesModel.PROPERTY].property \
                               for row in self.filtered_props \
                               if row[FilteredPropertiesModel.SELECT_SINGLE] ]
            assert len(properties) == 1 # only one property
            self.last_property_name = properties[0]._get_name()
            formulae_doc = properties_api.create_formulae_document_for_properties(properties)
            smv_formulae_doc = properties_api.create_formulae_document_for_properties(properties)
            properties_api.convert_atomic_propositions_to_smv(smv_formulae_doc, self.slim_model.get_extended_root_impl())

            properties = [(properties_api.get_formula_as_smv_string(slim), 
                           properties_api.get_formula_as_smv_string(smv)) 
                          for slim, smv in zip(formulae_doc, smv_formulae_doc)]
            smv_prop_text = properties[0][1]

            # prepare the diagnosability context, if any
            if self.diag_ctx is not None:
                smv_formulae_doc = properties_api.create_formulae_document_for_properties((self.diag_ctx,))
                properties_api.convert_atomic_propositions_to_smv(smv_formulae_doc, self.slim_model.get_extended_root_impl())
                ctx_smv_string = properties_api.get_formula_as_smv_string(smv_formulae_doc[0])
            else:
                ctx_smv_string = "TRUE"

            # prepare the path restrictions (if any)
            utils.log.system.info('Preparing the path restrictions (if any).')
            path_restrictions = self.get_restrictions_from_store(self.restrictions_store)

            # run diagnosability check
            utils.log.system.info("Diagnosability Check is to be performed for property '%s'" % properties[0][0])
            if self.slim_model.timed_domain:
                delta_t_var_name, tu = smv_api.get_smv_delta_var()
            else:
                delta_t_var_name = None

            # -- DEBUG CODE
            if False:
                print "-------------------------------------------------------------------------"
                print "Running Diagnosability with the following parameters:"
                print "* SMV_PROP_TEXT: %s" %smv_prop_text
                print "* CTX_SMV_STRING: %s" %ctx_smv_string
                print "* OBSERVABLES: %s" %observables
                print "* PATH_RESTRICTIONS: %s" %path_restrictions
                print "* DELTA_T_VAR_NAME: %s" %delta_t_var_name
                print "* DELAY: %s" %options.DIAG_DELAY
            # -- END OF DEBUG CODE

            success, trace_A_fn, trace_B_fn = self.get_tool().check_diagnosability(smv_prop_text, ctx_smv_string, observables, path_restrictions, delta_t_var_name=delta_t_var_name)

            self.diag_result = FdirDiagResult(success, trace_A_fn, trace_B_fn)

            # write result info to log
            utils.log.system.info('Diagnosability Check has been performed for formula "%s".' % (properties[0][0]))
            if success == "Unknown":
                utils.log.system.info('No counterexample to the given diagnosability property found within the given bound.')
            elif success == "Error":
                utils.log.system.info('An error has occurred, NuSMV didn\'t return any result!')
            elif success == "True":
                utils.log.system.info('The property was found to be diagnosable.')
            else:
                utils.log.system.info('The property was found to be not diagnosable.')

            # close nusmv (no incremental diagnosability checks supported yet)
            self.stop_fsap()

            utils.log.system.info('Diagnosability Analysis complete.')

        finally: self.busy -= 1                    
        return

    pass # end of class
# ----------------------------------------------------------------------


class FdirDiagResult ():
    """This is the container for the Diagnosability Check result,
    having the success flag (a string) and possibly two ResultModel with the traces."""

    def __init__(self, _success, trace_A_fn, trace_B_fn):

        self.success = _success

        # something went wrong
        if _success == "Error":
            # raise exceptions.UserException("No result was returned by NuSMV during Diagnosability Analysis.")
            self.result_A = None
            self.result_B = None
            return

        # there's a counterexample
        if _success == "False":
            assert os.path.isfile(trace_A_fn) and os.path.isfile(trace_B_fn),\
                "For unknown reasons DIAG has not generated output files."
            self.result_A = ResultModel()
            self.result_B = ResultModel()
            self.result_A.create_trace_from_file(trace_A_fn)
            self.result_B.create_trace_from_file(trace_B_fn)

        # there's no counterexample
        else:
            self.result_A = None
            self.result_B = None

        return

    pass # end of class
# ----------------------------------------------------------------------


class FdirFdaModel (FsapHandlerModel):
    """Model for handling Fault Detection Analysis (FDA)"""

    fda_list = ListStore(types.StringType)
    sat_bound = options.NUSMV_BMC_LENGTH
    detect_delay = -1.0 # It is the Detection Delay (by default -1.0, i.e. 'FINITE')    

    __observables__ = ("sat_bound", "detect_delay",)

    def __init__(self, slim_model, properties):
        FsapHandlerModel.__init__(self, slim_model)
        self.filtered_props = properties
        return

    def check_alarms(self):
        """Checks if there are alarms in the model. Returns
        True if there are alarms, or False otherwise"""
        self.start_fsap_alm()
        assert self.failures >= 0 and self.alarms >= 0, "Something is broken in the toolchain"
        return self.alarms > 0


    def get_delay(self, val, unit):
        """Given a value (i.e. an integer) and a time unit, it builds a slim expression and it returns its smv
           translation.
        """
        assert unit in ['msec', 'sec', 'min', 'hour', 'day', None]
        #FIXME: using regular expressions to understand whether it is an integer or a float
        if "." not in val:
            val += ".0"
        if unit is not None:
            time_str = "%s %s" %(str(val), str(unit))
        else:
            time_str = str(val)
        smv_delay = smv_api.parse_and_convert_numeric_expression(self.slim_model.get_extended_root_impl(), time_str)
        if "f'" in str(smv_delay):
            smv_delay = smv_delay.replace("f'", "")
        float_delay = eval(smv_delay)
        return float_delay

        
    def generate_fda(self, detection_delay=None):
        """Performs FDA and generates a file containing the
        resulting table. Returns None if there are no
        alarms"""
 
        self.busy += 1

        ## Setting initial value of detect_fname, so that if any exception occurs, a result is returned
        ## and the value is not None
        detect_fname = None

        try:

            ## Every time we run FDA, we need to recreate the smv translated files; since they are regenerated
            ## by the slim model only in case svm_fname is None, we have to explicitly set it to None
            self.slim_model.smv_fname = None

            self.start_fsap_alm()
            if not self.check_alarms(): return None

            # take the property
            properties = [ row[FilteredPropertiesModel.PROPERTY].property \
                               for row in self.filtered_props \
                               if row[FilteredPropertiesModel.SELECT_SINGLE] ]
            assert len(properties) == 1 # only one property
            formulae_doc = properties_api.create_formulae_document_for_properties(properties)
            smv_formulae_doc = properties_api.create_formulae_document_for_properties(properties)
            properties_api.convert_atomic_propositions_to_smv(smv_formulae_doc, self.slim_model.get_extended_root_impl())

            properties = [(properties_api.get_formula_as_smv_string(slim),
                           properties_api.get_formula_as_smv_string(smv))
                          for slim, smv in zip(formulae_doc, smv_formulae_doc)]
            smv_prop_text = properties[0][1]

            if detection_delay is not None:

                if detection_delay == -1.0:
                    detection_delay = None
                    delay_counter_var = None
                    utils.log.system.info("Fault Detection Analysis is to be performed for property '%s'" % properties[0][0])
                else:
                    assert detection_delay >= 0.0
                    delay_counter_bound = None

                    # the bound of the delay counter must be higher than the required detection
                    # delay, otherwise a violation of that bound can not be found
                    if detection_delay >= 1:
                        delay_counter_bound = detection_delay + 1
                    elif 0 == detection_delay:
                        delay_counter_bound = 0.00001
                    else:
                        # upper bound for counter: delay bound plus 10%
                        delay_counter_bound = detection_delay + detection_delay * 0.1

                    self.extend_smv_with_delay_counter(smv_prop_text, delay_counter_bound)
                    delay_counter_var = self.DELAY_COUNTER_VAR
                    utils.log.system.info("Fault Detection Analysis is to be performed for property '%s' and delay %f" % (properties[0][0], detection_delay))
            else:
                utils.log.system.info("Fault Detection Analysis is to be performed for property '%s'" % properties[0][0])
                delay_counter_var = None

            tmp_detect_fname = self.get_tool().compute_fault_detection(smv_prop_text,
                                                                       self.sat_bound,
                                                                       delay_counter_var,
                                                                       detection_delay)

            # rename the files according to the base name
            detect_fname = get_out_fname(options.BASE_FILE_NAME + ".flt_detect")
  
            # If FSAP execution is unsuccessful then an exception is raise
            # and here we assume the run was successful and file has been generated.
            assert os.path.isfile(tmp_detect_fname),\
                "For unknown reasons FSAP has not generated output files"

            os.rename(tmp_detect_fname, detect_fname)
           
            if not options.FSAP_KEEP_SMV_IDS:
                fsap_process.convert_fault_detection(detect_fname, detect_fname, smv_api.convert_smv_const_to_slim)
                pass

            self.stop_fsap()

            utils.log.system.info('Fault Detection has been performed for formula "%s" and the result is stored in file "%s"' %
                                  (properties[0][0], relpath(detect_fname)))

        except Exception:
            if self.get_tool() is not None:
                if self.get_tool().is_alive(): self.stop_fsap()
            detect_fname = None

        finally:
            self.busy -= 1                    
        return detect_fname
        

    def load_fda(self, fda_fname):
        """Load given FDA file into the internal liststore (which
        then can be shown in the view side). Returns False if the
        list is empty, True otherwise"""

        self.busy += 1
        try:
            self.fda_list.clear()
            res = False

            _file = open(fda_fname, "r")
            line = _file.readline()
            while line:
                # the first line contains the headers and it is ignored
                line = _file.readline().strip()
                if line: 
                    fields = line.split()
                    assert len(fields) == 1
                    self.fda_list.append(fields)
                    res = True
                    pass
                
                pass
            _file.close()

        finally: self.busy -= 1
        return res


    def store_fda(self, row):
        """Stores current FDA list contents into appropriate column of
        the gtk property tree viewer model."""
        tmp, itr = ([], iter(self.fda_list))
        try:
            while (True):
                item = itr.next() [ 0 ]
                tmp.append(item)
            
        except StopIteration:
            pass
        
        child_iter = self.filtered_props.get_properties_list().\
            convert_iter_to_child_iter(row.iter)
        self.filtered_props.props_model.set_fda_list(child_iter, tmp)

        return

    def restore_fda(self, fda):
        """ Selects given FDA as current FDA. Given FDA comes
        from a previously stored FDA generation."""

        self.fda_list.clear()
        if fda is None: return
        
        self.busy += 1
        try:
            for item in fda:
                self.fda_list.append([ item ])
            
        finally:
            self.busy -= 1

        return

    pass # end of class
# ----------------------------------------------------------------------


class FdirFiaModel (FsapHandlerModel):
    """Model for handling Fault Isolation Analysis (FIA)"""


    # num, alarm, event_fname, gate_fname
    ftrees_list = ListStore(types.IntType, types.StringType, types.StringType, types.StringType)
    sat_bound = options.NUSMV_BMC_LENGTH

    __observables__ = ("sat_bound",)
    
    def __init__(self, slim_model):
        FsapHandlerModel.__init__(self, slim_model)
        return

    def check_alarms(self):
        """Checks if there are alarms in the model. Returns
        True if there are alarms, or False otherwise"""
        self.start_fsap_alm()
        assert self.failures >= 0 and self.alarms >= 0, "Something is broken in the toolchain"
        return self.alarms > 0
        

    def generate_fia(self):       
        """Performs FIA and generates a set of files containing a
        fault tree each.  Returns None if there are no
        alarms"""

        self.busy += 1
        try:
            self.start_fsap_alm()
            if not self.check_alarms(): return None

            utils.log.system.info("Fault Isolation Analysis is to be performed for %d alarms...", 
                                  self.alarms)      

            tmp_isolat_fname = self.get_tool().compute_fault_isolation(self.alarms, 
                                                                       bmc_length=self.sat_bound)

            # rename the files according to the base name
            out_files = []
            for files, num in zip(tmp_isolat_fname, range(len(tmp_isolat_fname))):
                # If FSAP execution is unsuccessful then an exception is raise
                # and here we assume the run was successful and file has been generated.
                assert os.path.isfile(files[0]) or os.path.isfile(files[1]),\
                    "For unknown reasons FSAP has not generated output files"

                ev_fname = get_out_fname("%s.%d.flt_events" % (options.BASE_FILE_NAME, num))
                gt_fname = get_out_fname("%s.%d.flt_gates" % (options.BASE_FILE_NAME, num))
                out_files.append((relpath(ev_fname), relpath(gt_fname)))

                os.rename(files[0], ev_fname)
                os.rename(files[1], gt_fname)
           
                if not options.FSAP_KEEP_SMV_IDS:
                    fsap_process.convert_fault_tree_events(ev_fname, ev_fname,
                                                           smv_api.convert_smv_const_to_slim)
                    fsap_process.convert_fault_tree_gates(gt_fname, gt_fname,
                                                          smv_api.convert_smv_const_to_slim)
                    pass
                pass
            
            utils.log.system.info('Fault Isolation has been performed and results are stored in files: %s' % 
                                  ", ".join(("<%s,%s>" % p for p in out_files)))

        finally: self.busy -= 1
        return out_files
        

    def load_fault_trees(self, fnames):
        """Load the generated list of fault trees, taking the list
        of generated pair filenames. Returns True there is at least
        one file to be shown"""
        
        self.busy += 1
        try:
            self.ftrees_list.clear()
            alarms = get_alarms(self.slim_model.get_extended_root_impl())
            for pair, alm, i in zip(fnames, alarms, range(len(fnames))): 
                self.ftrees_list.append((i+1, alm, pair[0], pair[1]))

        finally: self.busy -= 1
        return len(fnames) > 0

    pass # end of class
# ----------------------------------------------------------------------


class FdirFraModel (ModelCheckingModel):
    """The model of Fault Recovery Analysis. It is based on the
    model checking model"""
    
    def __init__(self, slim_model, properties):
        ModelCheckingModel.__init__(self, slim_model, properties)

        self.extended_model = True # this is hard-coded for fra
        return


    def run_fra(self):
        rows = [ row for row in self.filtered_props if row[FilteredPropertiesModel.SELECT_SINGLE] ]
        return self._run_model_checking_aux(rows)
        
    pass # end of class
# ----------------------------------------------------------------------


class FdirFcaModel (FsapHandlerModel):
    """Model for handling Fault Coverage Analysis"""

    def __init__(self, slim_model, properties, properties_model):
        """
        """
        FsapHandlerModel.__init__(self, slim_model)

        self.properties_model = properties_model
        self.slim_model = slim_model

        self.rec_use_sat = False   # True if recoverability with SAT, False if recoverability with BDD
        self.diag_use_sat = False  # True if diagnosability with SAT, False if diagnosability with BDD

        return


    def run_fca_rec(self, rows, sat_bound=None):
        """Function which runs the recoverability for the Fault Coverage Analysis. It takes as parameters some
           rows which should be taken from a treeview or something similar, and an optional SAT bound, which is
           None  in case the user wants to use BDDs.

           @param rows: the properties for which recoverability must be run
           @type  rows: list of properties

           @param sat_bound: the SAT bound
           @type  sat_bound: int (default None)

           @return result: the result of the recoverability analysis
           @rtype result McResultModel
        """
        from src.models.correctness import McResultModel
        import property_management.src.constants as prop_const
        from invokers import nusmv

      
        if sat_bound is not None:
            options.NUSMV_BMC_LENGTH = int(sat_bound)

        self.busy += 1
        self.start_fsap()

        properties = [ row[FilteredPropertiesModel.PROPERTY].property for row in rows ]

        formulae_doc = properties_api.create_formulae_document_for_properties(properties)
        smv_formulae_doc = properties_api.create_formulae_document_for_properties(properties)
        properties_api.convert_atomic_propositions_to_smv(smv_formulae_doc, 
                                                          self.slim_model.get_extended_root_impl())

        for formula, smv_formula, row in zip(formulae_doc, smv_formulae_doc, rows):
            result = McResultModel()
            result.formula = smv_formula 
            result.formula_text = properties_api.get_formula_as_smv_string(formula)
            result.type_name = { prop_const.Logic.CTL : "CTL",
                                 prop_const.Logic.LTL : "LTL" }[smv_formula.get_logic()]
            result.sat_length = sat_bound if self.rec_use_sat else None
 
            # calls the model checker
            if self.get_tool() is not None:
                result.truth, trace_fname = nusmv.model_checking(self.get_tool(), smv_formula)
            else:
                raise Exception("An error occurred in running th FSAP tool")
                return

        self.stop_fsap()
        self.busy -= 1

        return result



    def run_fca_detection(self, _property, detection_delay=None, sat_bound=None):
        """Function which runs the fault detection for the Fault Coverage Analysis. A property must 
           be specified, while optionally it is possible to specificy a detection_delay (in case 
           of a continuous model) and a SAT bound.

           @param _property: the property to be analyzed
           @type  _property: PropertyModel

           @param detection_delay: the optional delay in case of continuous models
           @type  detection_delay: float

           @param sat_bound: the bound for the SAT algorithm
           @type  sat_bound: int

           @return detect_fname: the name of the file where the result of the fault detection is stored
           @rtype  detect_fname: str
        """

        self.busy += 1

        try:
            ## Every time we run FDA, we need to recreate the smv translated files; since they are regenerated
            ## by the slim model only in case svm_fname is None, we have to explicitly set it to None
            self.slim_model.smv_fname = None

            self.start_fsap_alm()
            assert self.alarms >= 0
            if self.alarms == 0: return None

            properties = [ _property ]
            assert len(properties) == 1 # only one property

            if sat_bound is not None:
                options.NUSMV_BMC_LENGTH = int(sat_bound)

            if detection_delay is not None:
                options.DIAG_DELAY = float(detection_delay)

            formulae_doc = properties_api.create_formulae_document_for_properties(properties)
            smv_formulae_doc = properties_api.create_formulae_document_for_properties(properties)
            properties_api.convert_atomic_propositions_to_smv(smv_formulae_doc, self.slim_model.get_extended_root_impl())

            properties = [(properties_api.get_formula_as_smv_string(slim),
                           properties_api.get_formula_as_smv_string(smv))
                          for slim, smv in zip(formulae_doc, smv_formulae_doc)]
            smv_prop_text = properties[0][1]

            ## If the dection delay is None or FINITE, the algorithm must be called with None both for it and for the 
            ## deay_counter_var
            if detection_delay is not None:

                if detection_delay == -1.0:
                    detection_delay = None
                    delay_counter_var = None
                    utils.log.system.info("Fault Detection Analysis is to be performed for property '%s'" % properties[0][0])
                else:
                    assert detection_delay >= 0.0

                    # the bound of the delay counter must be higher than the required detection
                    # delay, otherwise a violation of that bound can not be found
                    if detection_delay >= 1:
                        delay_counter_bound = detection_delay + 1
                    elif 0 == detection_delay:
                        delay_counter_bound = 0.00001
                    else:
                        # increase first non-zero digit by one, eg: 0.004823 -> 0.005823
                        factor = 10
                        while detection_delay * factor < 1:
                            factor = factor * 10
                        delay_counter_bound = detection_delay + (1.0 / factor)

                        # alternative: increase by 10%
                        #delay_counter_bound = detection_delay + detection_delay * 0.1

                    self.extend_smv_with_delay_counter(smv_prop_text, delay_counter_bound)
                    delay_counter_var = self.DELAY_COUNTER_VAR
                    utils.log.system.info("Fault Detection Analysis is to be performed for property '%s' and delay %f" % (properties[0][0], detection_delay))
            else:
                utils.log.system.info("Fault Detection Analysis is to be performed for property '%s'" % properties[0][0])
                delay_counter_var = None

            tmp_detect_fname = self.get_tool().compute_fault_detection(smv_prop_text,
                                                                       int(sat_bound),
                                                                       delay_counter_var,
                                                                       detection_delay)

            # rename the files according to the base name
            detect_fname = get_out_fname(options.BASE_FILE_NAME + ".flt_detect")
  
            # If FSAP execution is unsuccessful then an exception is raise
            # and here we assume the run was successful and file has been generated.
            assert os.path.isfile(tmp_detect_fname),\
                "For unknown reasons FSAP has not generated output files"

            os.rename(tmp_detect_fname, detect_fname)
           
            if not options.FSAP_KEEP_SMV_IDS:
                fsap_process.convert_fault_detection(detect_fname, detect_fname, smv_api.convert_smv_const_to_slim)
                pass

            self.stop_fsap()

            utils.log.system.info('Fault Detection has been performed for formula "%s" and the result is stored in file "%s"' %
                                  (properties[0][0], relpath(detect_fname)))

        finally:
            self.busy -= 1                    
        return detect_fname



    def load_fca_detection(self, fda_fname):
        """Function which tries to load the result of the fault detection and returns a boolean to indicate success or failure.

           @param fda_fname: the name of the file in which the result of fault detection was stored
           @type  fda_fname: string

           @return res: the result of loading the fault detection file
           @rtype  res: boolean
        """

        self.busy += 1
        try:
            res = False

            _file = open(fda_fname, "r")
            line = _file.readline()
            while line:
                # the first line contains the headers and it is ignored
                line = _file.readline().strip()
                if line: 
                    fields = line.split()
                    assert len(fields) == 1
                    res = True
                    pass
                
                pass
            _file.close()

        finally: self.busy -= 1
        return res



    def get_fault_injections(self):
        """ Returns the loaded fault injections

            Fault Injections are stored as a liststore containing the following values:
            * bool -> isEnabled
            * bool -> canBeEnabled
            * object -> (package identifier, error model implementation name)
            * types.StringType -> error state
            * object -> component instance tupled path
            * types.StringType -> data port/subcomp
            * types.StringType -> effect

            Notice that what we are interested in is to build a fault injections as follows:
            <component_path>.error = error:<error_state>

            @return fault_inj_list: the list of loaded fault injections
            @rtype  fault_inj_list: [str]
        """

        fault_inj_liststore = self.slim_model.fault_injections
        fault_inj_list = []

        ## -- DEBUG CODE
        if False:
            print "\n-- DEBUG CODE"
            print "Fault Injections for Fault Coverage Analysis:"
            for dbg_el in fault_inj_liststore:
                print "* isEnabled: " + str(dbg_el[0])
                print "* canBeEnabled: " + str(dbg_el[1])
                print "* (pkg_id, error_model_impl_name): " + str(dbg_el[2])
                print "* error state: " + str(dbg_el[3])
                print "* component instance path: " + str(dbg_el[4])
                print "* data port/subcomp: " + str(dbg_el[5])
                print "* effect: " + str(dbg_el[6])
                print "----------------------------------------------------"
            print "-- END OF DEBUG CODE\n"
            pass
        ## -- END OF DEBUG CODE

        for el in fault_inj_liststore:

            from slim_compiler.src.model_extension import fault_injection
            
            assoc = fault_injection.ErrorModelAssociation.create(el[4], el[2][0], el[2][1])
            assoc.effects_specification.add_effect(el[3], el[5] + " := " + el[6])
            if self.slim_model.get_nominal_root_impl() is None:
                ## The slim model has been deleted
                return []
            errors = fault_injection.get_errors_in_error_model_association(assoc, self.slim_model.get_nominal_root_impl())

            ## Adding fault injections only if they are valid
            if el[0] and len(errors) == 0 and el[1]:

                error_state = el[3]
                component_path = ".".join(el[4])
                fault = "%s.error = error:%s" %(component_path, error_state)
                fault_inj_list.append(fault)

        return fault_inj_list



    def fca_get_delay(self, val, unit):
        """Given a value (i.e. an integer) and a time unit, it builds a slim expression and it returns its smv
           translation.
        """
        assert unit in ['msec', 'sec', 'min', 'hour', 'day']
        #FIXME: using regular expressions to understand whether it is an integer or a float
        if "." not in val:
            val += ".0"
        time_str = "%s %s" %(str(val), str(unit))
        smv_delay = smv_api.parse_and_convert_numeric_expression(self.slim_model.get_extended_root_impl(), time_str)
        if "f'" in str(smv_delay):
            smv_delay = smv_delay.replace("f'", "")
        float_delay = eval(smv_delay)
        return float_delay

        

    pass # end of class
# ----------------------------------------------------------------------
