# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================


from gtkmvc import Model, observable

from src.models.tool_handler import NuSMVHandlerModel
from src.models.trace import TraceModel
from slim_compiler.src.smv.walkers import SlimTreeBuilderWalker
from slim_compiler.src.smv import converter, support, api
from slim_compiler.src.dataStructure.visitors.expressionVisitor import ExpressionPostProcessor
from slim_compiler.src.dataStructure.nominalbehaviour \
    import NonDataSubcomponent, ModeTransition
import utils.log

import gtk, gobject
import itertools

# ----------------------------------------------------------------------
class GuidedSimTransitionModel (Model):

    TRANS_NOT_ALLOWED, TRANS_ALLOWED, TRANS_UNKNOWN = range(3)

    changed = observable.Signal() # the evaluation or the structure changed    
    __observables__ = ("changed", )

    def __init__(self, slim_model, sim_model, timed_trans_fair_alt_mode=True):
        Model.__init__(self)
        
        self.slim_model = slim_model
        self.sim_model = sim_model

        # observes the sim_model to see when a trace got generated
        self.observe_model(sim_model)

        # converter (for collecting trans info)
        self.converter = converter.SmvConverter()

        # implementation, transitions, name of the instance, full
        # path name of the instance, parent-implementation, list of
        # tuple coming from evaluate_structure, number of enabled
        # transitions
        self.slim_tree = gtk.TreeStore(object, object, str, str, 
                                       object, object, int)
        self.walker = SlimTreeBuilderWalker(self.slim_tree, self.converter)

        # for showing trans info
        # ModeTransition object, enabled (3 values), list of strings 
        self.trans_list = gtk.ListStore(object, int, object)
        self.trans_list_path = None

        self.trans_constraints = []

        self.built_structure_is_extended = False
        self.build_structure(self.sim_model.extended_model)
        self.can_simulate = True

        # Force alternation of discrete and timed transitions
        self._timed_trans_fair_alt_mode = timed_trans_fair_alt_mode
        return


    def build_structure(self, use_extended_model):
        """Call this method to construct the tree to be used later
        for the evaluation. In particular, everytime the slim model
        gets changed."""

        root = self.slim_model.get_extended_root_impl() \
            if use_extended_model else \
            self.slim_model.get_nominal_root_impl()

        if root is not None:
            import copy
            root = copy.deepcopy(root)
            ExpressionPostProcessor().visit(
                root.subpackage.package.compile_information)

        self.walker.walk(root)

        # stores if extended_model was used or not
        self.built_structure_is_extended = use_extended_model
        return
    

    def evaluate_structure(self, trace_model=None):
        """Traverses all the structure, and associate to each entry
        the set of possible transitions, their triggable truth
        value, and their trigger espressions. This has to be called
        when a trace is available."""

        def __fun_step1(model, path, iter, self_tm_res):            
            self, trace_model, res_list = self_tm_res                         
            info = self.__build_transitions(iter, trace_model)
            model[iter][5] = info
            res_list.append(info)            
            return

        infos = []
        self.slim_tree.foreach(__fun_step1, (self, trace_model, infos))       

        # extracts all pending constraints (which need to be evaluated by nusmv)
        def __fun_step2(_list, res):
            # elements in the list are either a pair or recursively a list
            for el in _list:
                if type(el) is tuple:
                    assert len(el) == 2
                    constr, val = el
                    if val not in (True, False): res.append(val)
                    pass
                else: 
                    assert type(el) is list
                    __fun_step2(el, res)
                pass
            
            return

        unks = []
        for info in infos:
            for trans in info:
                # print "TRANS[1]", trans[1]
                __fun_step2(trans[1], unks)                
                pass
            pass
        
#        print "UNKS: ", unks

        # here unks have to be passed to nusmv, to get a map with truth values
        # WARNING: This is done in the same thread of the gui, making it hang
        print "HERE!!", unks
        res = self.sim_model.get_tool().simulate_check_feasability(unks)
        if res is not None: res_map = dict(zip(unks, res))
        else:
            utils.log.system.error("A problem occurred with an expression sent to nusmv (see logging)")
            res_map = {}
            pass

        # here the truth value have to be evaluated
        #def __fun_step3(model, path, iter, self_tm):
        #    self, trace_model = self_tm                         
        #    model[iter][5], model[iter][6] = self.__build_transitions(iter, trace_model)
        #    if model.get_path(iter) == self.trans_list_path: self.build_transitions(iter, trace_model)
        #    return

        def evaluate_trans_path(trans, res_map):
            # trans can be: [], [path], [cando], [cando, [path]], [cando, cando, ...]
            if len(trans) == 0: return GuidedSimTransitionModel.TRANS_NOT_ALLOWED

            if type(trans[0]) is tuple:
                # this is a path in conjunction
                assert len(trans) <= 2 # premise, and possible path starting from it

                def __is_triggerable(cando, triggerable, res_map):
                    if type(triggerable) in (str, unicode):
                        if res_map.has_key(triggerable):
                            triggerable = GuidedSimTransitionModel.TRANS_ALLOWED \
                                if res_map[triggerable] else GuidedSimTransitionModel.TRANS_NOT_ALLOWED
                        else: triggerable = GuidedSimTransitionModel.TRANS_UNKNOWN
                        pass
                    
                    else: 
                        assert triggerable in (False, True)
                        triggerable = GuidedSimTransitionModel.TRANS_ALLOWED \
                            if triggerable else GuidedSimTransitionModel.TRANS_NOT_ALLOWED
                        pass
                    return triggerable
                    
                
                premise = trans[0]    
                if len(premise) == 2 and type(premise[0]) in (str, unicode):
                    cando, triggerable = premise
                    triggerable = __is_triggerable(cando, triggerable, res_map)
                else:
                    # a disjunction of premises
                    for premise_i in premise:
                        assert len(premise_i) == 2 and type(premise_i[0]) in (str, unicode)
                        cando, triggerable = premise_i
                        triggerable = __is_triggerable(cando, triggerable, res_map)
                        if triggerable is GuidedSimTransitionModel.TRANS_ALLOWED: break
                        pass
                    pass

                # there might be unknown value here if res_map is
                # not fully available
                assert triggerable in (GuidedSimTransitionModel.TRANS_NOT_ALLOWED, 
                                       GuidedSimTransitionModel.TRANS_ALLOWED,
                                       GuidedSimTransitionModel.TRANS_UNKNOWN)
                       
                if triggerable is GuidedSimTransitionModel.TRANS_NOT_ALLOWED: 
                    return triggerable # it's over
                                
                if len(trans) == 2 and trans[1] != []:
                    path_tr = evaluate_trans_path(trans[1], res_map)
                    if path_tr is not GuidedSimTransitionModel.TRANS_ALLOWED: return path_tr
                    pass

                return triggerable

            # this is a disjunction of alternative paths
            assert type(trans[0]) is list, "Not a list:" + str(trans[0])
            disj = GuidedSimTransitionModel.TRANS_NOT_ALLOWED
            for _path in trans:
                assert type(_path) is list
                disj = evaluate_trans_path(_path, res_map) \
                    if disj is GuidedSimTransitionModel.TRANS_NOT_ALLOWED else disj
                if disj is GuidedSimTransitionModel.TRANS_ALLOWED: break
                pass
            return disj
        
        for info in infos:
            for trans in info:   
                trans[2] = evaluate_trans_path(trans[1], res_map)
                pass
            pass

        # sets the number of allowed transitions in the slim_tree
        def __fun_step3(model, path, iter):            
            info = model[iter][5]
            num = 0
            for trans in info:
                if trans[2] is not GuidedSimTransitionModel.TRANS_NOT_ALLOWED: num += 1
                pass
            model[iter][6] = num
            return
        self.slim_tree.foreach(__fun_step3)  

        # redraws trans list:
        if self.trans_list_path:
            self.build_transitions(self.slim_tree.get_iter(self.trans_list_path), trace_model)
            pass

        self.sim_model.can_simulate = False
        self.changed.emit()
        return


    def build_transitions(self, iter, trace_model=None):
        """iter is an iterator of self.slim_tree pointing to the
        row containing the implementation for which we want to
        construct the trans info. This is called when a component
        is selected."""

        self.trans_list_path = self.slim_tree.get_path(iter) if iter else None
        self.trans_list.clear()        
        if iter is None: return

        info = self.slim_tree[iter][5]

        if info is None: 
            # not previously calculated
            info = self.__build_transitions(iter, trace_model)
            self.slim_tree[iter][5] = info
            self.slim_tree[iter][6] = -1
            pass

        for trans, trans_path, triggerable, trigger in info:
            self.trans_list.append((trans, triggerable, trigger))
            pass
        return 


    def set_trans_constraints(self, constraints):
        # sets transition constraints. constr is a list of
        # constraints as strings.
        self.trans_constraints = constraints
        return


    def get_trans_constraints(self):
        # this is called by the simulator, to get the previously
        # set constraints
        return self.trans_constraints

    def cando_time_transition(self, constraints=[]):
        if not self.slim_model.timed_domain:
            return False

# BEGIN(IN)
#         constr = self.converter.root_var() + "." + \
#             self.converter.cando_name(self.converter.delta_var()) + \
#             " & " + "!" + self.converter.done_delta_var() + \
#             "& " + self.converter.templates.do_delta

        if self._timed_trans_fair_alt_mode:
            constr = self.converter.root_var() + "." + \
                self.converter.cando_name(self.converter.timed_event()) + \
                " & " + "!" + self.converter.done_delta_var() + \
                " & " + self.converter.timed_event()
        else:
            constr = self.converter.root_var() + "." + \
                self.converter.cando_name(self.converter.timed_event()) + \
                " & " + self.converter.timed_event()
# END(IN)

        if constraints: constr += " & ".join([""] + list(constraints))
        
        print "HERE2!!", constr
        res = self.sim_model.get_tool().simulate_check_feasability([constr])
        print "RES: " + str(res)
        return all(res) if res is not None else False

    def set_do_time_transition(self, constraints):
        """Sets the constraints to do a timed transition"""

# BEGIN(IN)
#         unks = [self.converter.root_var() + "." + 
#                 self.converter.cando_name(self.converter.delta_var()) +
#                 " & " + "!" + self.converter.done_delta_var() +
#                 "& " + self.converter.templates.do_delta]

        if self._timed_trans_fair_alt_mode:
            constr = self.converter.root_var() + "." + \
                self.converter.cando_name(self.converter.timed_event()) + \
                " & " + "!" + self.converter.done_delta_var() + \
                " & " + self.converter.timed_event()
        else:
            constr = self.converter.root_var() + "." + \
                self.converter.cando_name(self.converter.timed_event()) + \
                " & " + self.converter.timed_event()
        unks = [constr]
# END(IN)

        if constraints: unks.append(" & ".join(constraints))
        self.set_trans_constraints(unks)
        return        
        
    def set_can_simulate(self, flag): self.can_simulate = flag
    def get_can_simulate(self): return self.can_simulate


    # --------------------------------------------------
    #      Observable properties 
    # --------------------------------------------------


    # --------------------------------------------------
    #      Private methods
    # --------------------------------------------------
    def __build_transitions(self, iter, trace_model=None):
        """iter is an iterator of self.slim_tree pointing to the
        row containing the implementation for which we want to
        construct the trans info. The returned list contains
        lists like 
        [trans, triggerable, list_of_feasibile_constraints, trigger_expr]."""
        
        impl = self.slim_tree[iter][0]
        transs = self.slim_tree[iter][1]
        inst_name = self.slim_tree[iter][3] # full name
        par_impl = self.slim_tree[iter][4] # parent implementation

        info = []
        if trace_model:
            # can do names for all transitions

            stripes = [stripe[0] for stripe in trace_model if stripe[1]]
            names = [self.converter.root_var() + "." + self.converter.keep_var_name(stripe[1])
                     for stripe in trace_model if stripe[1]]
            assert len(stripes) == len(names)

#            print "NAMES: ", names

            for trans in transs:
                constrs = transs.get(trans, ())
#                print "CONSTRS: ", constrs
                val = self.__evaluate_constraints(names, stripes, constrs)
#                print "VAL: ", val
                do_it = self.__do_trans(trans, impl, par_impl, inst_name)
#                print "DO_IT: ", do_it
                info.append([trans, val, self.TRANS_UNKNOWN, do_it])
                pass
            pass
                
        else: # no trace is available
            for trans in transs: 
                do_it = self.__do_trans(trans, impl, par_impl, inst_name)
                info.append([trans, [], self.TRANS_ALLOWED, do_it])
                pass
            pass
        
#        print "INFO: ", info
        return info


    def __evaluate_constraints(self, names, stripes, constrs): 
        """Given a complete path for a transition (given with
        constrs), a set of stripes coming from a trace, creates a
        new path where each node is substituted by a pair (node,
        truth), where truth is the trueth value of the node. If not
        evaluable statically, it is a string with the constraint to
        be evaluated later by nusmv."""

        res = []
        for cs in constrs:            
            if type(cs) in (str, unicode):
                # level 0 constraint
                conj_res = True
                conj_unk = [] # unknown constraints

                for c in cs.split("&"):
                    n,v = (x.strip() if x else x for x in (c.split("=") + [None])[0:2])
                    try: 
                        idx = names.index(n)
                        stripe = stripes[idx]
                        if stripe.get_info() == "state": 
                            val = stripe.get_value_at_step(stripe.base_steps() - 1)
                            conj_res = (val.lower() not in ("false", "0")) \
                                       if v is None \
                                       else val == api.convert_smv_const_to_slim(v)
                            pass
                        else:
                            # not a state var: input or combinatorial
                            assert stripe.get_info() in ("input", "combinatorial")                              
                            conj_unk.append(n if v is None else "%s=%s" % (n,v))
                            pass
                        pass                    
            
                    except ValueError: 
                        # here there is no value for, which means that
                        # n contains next, and 'n=v' needs nusmv to get a
                        # truth value
                        conj_unk.append(n if v is None else "%s=%s" % (n,v))
                        pass

                    if not conj_res: break
                    pass

                # GZ CHANGE!!
                if conj_res and conj_unk: res.append((cs, "(%s)" % " & ".join([conj_unk[0]])))
                else: res.append((cs, conj_res))
                
            else: #if cs:
                assert type(cs) is tuple, "cs's unexpected: " + str(cs)
                res.append(self.__evaluate_constraints(names, stripes, cs))
                pass
            
            pass # end of loop
        
        return res

# BEGIN(IN)
#     def __do_trans(self, trans, impl, par_impl, inst_fullname):
#         impl_type = impl.get_referenced_component_type()
#         inst_name = inst_fullname.split(".")[-1]
#         ids_lists = self.converter.create_event_associated_behavior(impl)

#         ev_id_trans = []
#         for ev in impl_type.in_event_ports.keys() + impl_type.out_event_ports.keys() + [self.converter.tau_event()]:
#             behaviors = ids_lists[ev]
#             for elem in behaviors:
#                 ID = self.converter.get_elem_position(behaviors, elem)
#                 if isinstance(elem, basestring) : #there is connection (ev, sc.ev) or (sc.ev, ev)
#                     # event passed down
#                     pass
#                 elif isinstance(elem, NonDataSubcomponent):
#                     # tau is passed down to subcomponent 
#                     pass
#                 elif isinstance(elem, set) : # there are transitions to deal with ev
#                     ev_id_trans.append((ev, ID, elem))
#                     pass
#                 else: raise LookupError, "A list contains unexpected element" + str(type(elem))
#                 pass 
#             pass
        
#         res = []
#         id_prefix = self.converter.root_var() + "." + \
#             ((self.converter.sc_name(inst_fullname) + ".") if inst_fullname else "")
#         self.converter.set_id_prefix(id_prefix)
#         for ev, id, trans_set in ev_id_trans:
#             if trans in trans_set:
#                 # assigns was introduced to fix ticket:482
#                 assigns = " & ".join([
#                     self.converter.data_transition_assigned(
#                         impl_type, impl, dsc_name, value_expr).strip()
#                     for dsc_name, value_expr in 
#                     support.get_trans_assigns(impl, trans, complete=False).iteritems()
#                     ])
#                 if assigns: assigns = "& " + assigns
                
#                 res.append("%s & %s %s" % (self.converter.transition_event_id(ev, id), 
#                                            self.converter.mode_transition(impl, trans),
#                                            assigns))
                
#                 #RC: WARNING! TBD this line was used instead:
#                 #res.append("%s" % self.converter.transition_event_id(ev, id))
#                 pass
#             pass        
#         self.converter.set_id_prefix("")
#         return res

    def __do_trans(self, trans, impl, par_impl, inst_fullname):
        impl_type = impl.get_referenced_component_type()
        inst_name = inst_fullname.split(".")[-1]
        event_trans_map = self.converter.create_event_associated_behavior(impl)

        impl_events = impl_type.in_event_ports.keys()
        impl_events = impl_events + impl_type.out_event_ports.keys()
        impl_events.append(self.converter.tau_event())
        
        ev_id_trans = []

        for ev in impl_events:
            pos = 0
            for tev, k, ts in event_trans_map[ev]:
                if ev == tev and ts:
                    ev_id_trans.append((ev, pos, ts))
                pos = pos + 1

        id_prefix = self.converter.root_var() + "."

        if inst_fullname:
            id_prefix = id_prefix + self.converter.sc_name(inst_fullname) + "."

        self.converter.set_id_prefix(id_prefix)

        res = []

        for ev, pos, ts in ev_id_trans:
            if trans in ts:
                constraint = self.converter.event_mode_transition_constraint(
                    impl, ev, pos, trans)
                res.append(constraint)

        self.converter.set_id_prefix("")
        return res
# END(IN)

    pass # end of class
