# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================


import gtk
import gobject
from gtkmvc import Model
import os
import threading

from src.models.tool_handler import Smv2SigrefHandlerModel, Model2ImcHandlerModel, MRMCHandlerModel, \
IMCAHandlerModel, SlimSimHandlerModel
from src.models.activity import ActivityModel
from src.models.filtered_properties import FilteredPropertiesModel

import options
from invokers import dump, slimsim
import property_management.src.api as properties_api
from property_management.src.constants import Logic
from property_management.src import formulae as formulae_api
import utils
from utils import exceptions, misc, debug_info as debug
from utils.property_checker import PropertyCheckingRecord, PropertyChecker

from lxml import etree
from utils.results import AnalysisResults

class PerformabilityModel (ActivityModel):
    """Model of performability panel."""

    def __init__(self, appl_model, slim_model, props_model):
        ActivityModel.__init__(self, appl_model, slim_model, props_model)

        # the properties model
        self.props_model = props_model        
        # this is the set of propertis filtered
        self.filtered_props = FilteredPropertiesModel(props_model)
        # only list probabilistic models
        self.filtered_props.set_filtering((Logic.CSL,)) 
        # set the model for the probabilistic pane
        self.model_imc = PerformabilityTabIMC(appl_model, slim_model, self.filtered_props)
        self.model_sim = PerformabilityTabSim(appl_model, slim_model, self.filtered_props)
        # monitors
        self.appl_model.tools.add_tool("NuSMV", self.model_imc.smv2sigref, "Performability")
        self.appl_model.tools.add_tool("Model2Imc", self.model_imc.model2imc, "Performability")
        self.appl_model.tools.add_tool("MRMC", self.model_imc.mrmc, "Performability")
        self.appl_model.tools.add_tool("IMCA", self.model_imc.imca, "Performability")
        self.appl_model.tools.add_tool("slimsim", self.model_sim.slimsim, "Performability")
        # observables
        #self.perfModel.smv2sigref.register_observer(self)
        #self.perfModel.sigref.register_observer(self)
        #self.perfModel.mrmc.register_observer(self)
        self.model_imc.register_observer(self)
        self.model_sim.register_observer(self)

    # ---------------------------------------------------------------
    #        Notifications from observable properties
    # ---------------------------------------------------------------    
    def property_busy_value_change(self, model, old, new):
        if model is self: return # not interested in introspection
        self.busy += (new - old)
        return

class ProbabilisticToolHandlers(ActivityModel):
    """ This class is to be used as a superclass of the PerformabilityTab,
    (dynamic) fault tree evaluation and verification tobs. """
    
    # the final probability
    probability = None

    # the property (as an instance of Property) under verification
    csl_property = None    
    # tool instances
    smv2sigref = None
    model2imc = None
    mrmc = None
    imca = None

    
    def __init__(self, appl_model, slim_model, filtered_props):
        ActivityModel.__init__(self, appl_model, slim_model, filtered_props)
        self.slim_model = slim_model
        self.filtered_props = filtered_props
        self.smv2sigref = Smv2SigrefHandlerModel()
        self.model2imc = Model2ImcHandlerModel()
        self.mrmc = MRMCHandlerModel()
        self.imca = IMCAHandlerModel()
        self.slimsim = SlimSimHandlerModel()
        return
    
    
    def reset(self):
        self.probability = None
        self.csl_property = None
        return
    
        
    def get_lower_time_bound(self, prop=None):
        if prop is None:
            #if self.csl_property == None: 
            retval =  0 # default lower bound
            #else: 
            #    pattern_instance = self.csl_property._get_pattern_instance()
            #    placeholder_instance = pattern_instance.get_placeholder_instance("Time1")
            #    retval = placeholder_instance.content
        else:
            pattern_instance = prop._get_pattern_instance()
            placeholder_instance = pattern_instance.get_placeholder_instance("Time1")
            retval = placeholder_instance.content
        return retval
        
    
    def get_upper_time_bound(self, prop=None):
        if prop is None:
            #if self.csl_property == None: 
            retval =  100 # default upper bound
            #else: 
            #    pattern_instance = self.csl_property._get_pattern_instance()
            #    placeholder_instance = pattern_instance.get_placeholder_instance("Time2")
            #    retval = placeholder_instance.content
        else:
            pattern_instance = prop._get_pattern_instance()
            placeholder_instance = pattern_instance.get_placeholder_instance("Time2")
            retval = placeholder_instance.content            
        return retval    

    # end of class ProbabilisticToolHandlers

class PerformabilityTabSim(ProbabilisticToolHandlers):
    # event for notifying that the start-thread has been killed or not
    is_killed_event = threading.Event()

    def __init__(self, appl_model, slim_model, filtered_props):
        ProbabilisticToolHandlers.__init__(self, appl_model, slim_model, filtered_props)

        self.analysis_result = None

        self.error_bound = options.MC_ERRORBOUND
        self.confidence = options.MC_CONFIDENCE
        self.strategy = options.MC_STRATEGY

        self.killed = False

    def reset(self):
        self.analysis_result = None
        ProbabilisticToolHandlers.reset(self)
        return

    def set_property(self):
        properties = [ row for row in 
                        self.filtered_props if row[FilteredPropertiesModel.SELECT_SINGLE] ]
        # only one property is supported
        assert len(properties) == 1

        if properties[0][FilteredPropertiesModel.PROPERTY].property.get_pattern_id() \
            in ["expectedTime", "longRunAverage", "probabilisticResponse"]:
            return False

        self.row = properties[0]
        self.csl_property = properties[0][FilteredPropertiesModel.PROPERTY].property
        return True

    from utils.log import log_icd
    @log_icd
    def run_simulation(self, error_bound, confidence, strategy):
        utils.log.system.debug("Running Simulation")
        self.killed = False

        self.busy += 1

        try:
            self.is_killed_event.clear()

            # run the translation of the extended model
            eroot_impl = self.slim_model.get_extended_root_impl()
            compile_info = eroot_impl.subpackage.package.compile_information
            slimsim.transform_slim(compile_info)

            self.set_property()
            formula = properties_api.create_formula_for_property(self.csl_property)
            slimsim.transform_formula(compile_info, eroot_impl, formula)

            xml_fname = misc.get_out_fname(options.BASE_FILE_NAME + "_sim.xml")
            f_py2xml = open(xml_fname, "w")
            f_py2xml.write(compile_info.to_XML())
            f_py2xml.close()
            self.slim_model.debug_info.slimsim_model = compile_info.to_XML()

            try:
                if self.csl_property.get_pattern_id() == "expectedTime" or \
                   self.csl_property.get_pattern_id() == "longRunAverage" or \
                   self.csl_property.get_pattern_id() == "probabilisticResponse":
                    gobject.idle_add(self.display_error, str("Cannot simulator property of type %s" % self.csl_property.get_pattern_id()))
                    return

                xml_propfname = misc.get_out_fname(options.BASE_FILE_NAME + "_props.xml")
                f_prop2xml = open(xml_propfname, "w")
                sim_formulae = formulae_api.FormulaeDocument.create([formula])
                properties_api.convert_time_bounds_for_performability(sim_formulae)
                f_prop2xml.write( etree.tostring(sim_formulae, pretty_print=True, xml_declaration=True) )
                f_prop2xml.close()

                if self.is_killed_event.isSet(): return
                self.slimsim.start_slimsim(xml_fname, xml_propfname, options.MC_STEPBOUND, self.error_bound, self.confidence, self.strategy)
                if self.slimsim.tool is None:
                    (exit_code, text) = (99, "")
                else:
                    (exit_code, text) = (self.slimsim.tool.returncode, self.slimsim.tool.get_stdout_text())
                self.slim_model.debug_info.slimsim_output = (text, self.slimsim.tool.get_stderr_text())

                if exit_code == 0:
                    probs = slimsim.extract_results_from_output(text)
                    if probs:
                        results = AnalysisResults(self.csl_property, 'slimsim', probability=probs)
                        # stores the result
                        child_iter = self.filtered_props.get_properties_list().convert_iter_to_child_iter(self.row.iter)
                        self.filtered_props.props_model.set_analysis_result(child_iter, results)
                    else:
                        gobject.idle_add(self.display_error, "Simulation error")
                else:
                    gobject.idle_add(self.display_error, "Simulation error: Tool died with code %d" % exit_code)
            except exceptions.ProcessInternalException as e:
                # Most likely caused by interrupt
                utils.log.system.warning("Internal process exception during simulation: " + str(e))
                if not self.killed:
                    raise e
        except exceptions.UserException, e:
            self.stop_performability()
            gobject.idle_add(self.display_error, str(e))
        finally:
            self.busy -= 1
            utils.log.system.debug("Done with Simulation")

    def display_error(self, message):
        msgbox = gtk.MessageDialog(type=gtk.MESSAGE_ERROR,
                           buttons=gtk.BUTTONS_OK,
                           message_format=message)
        msgbox.run()
        msgbox.destroy()

    def stop_simulation(self):
        self.killed = True
        self.is_killed_event.set()
        self.slimsim.stop_slimsim()
        return

class PerformabilityTabIMC(ProbabilisticToolHandlers):

    # event for notifying that the start-thread has been killed or not
    is_killed_event = threading.Event()

    def __init__(self, appl_model, slim_model, filtered_props):
        ProbabilisticToolHandlers.__init__(self, appl_model, slim_model, filtered_props)
        self.infinite_model = False

        # this is used to interrupt the ordering file generation process
        self.__kill_ord_event = None
        self.__kill_ord_event_processed = threading.Event()

        self.analysis_result = None

        self.observe_model(self.slim_model)

        # True if the process was interrupted by the user
        self.killed = False

    def reset(self):
        self.analysis_result = None
        ProbabilisticToolHandlers.reset(self)
        return

    def set_property(self):
        properties = [ row for row in 
                        self.filtered_props if row[FilteredPropertiesModel.SELECT_SINGLE] ]
        # only one property is supported
        assert len(properties) == 1
        self.row = properties[0]
        self.csl_property = properties[0][FilteredPropertiesModel.PROPERTY].property

    from utils.log import log_icd
    @log_icd
    def run_performability(self, tool_settings_model):
        utils.log.system.debug("Running Performability Evaluation")
        self.killed = False

        self.busy += 1

        try:
            self.is_killed_event.clear()

            basename = self.translate_models()
            if basename is None:
                # translation has failed
                return

            #this has to be done AFTER translate_models() where csl_property is set
            formula = properties_api.create_formula_for_property(self.csl_property)
            properties_api.convert_time_bounds_for_performability([formula])
            formula_intermediate_string = properties_api.get_formula_as_intermediate_string(formula)
            label_map = properties_api.create_proposition_to_label_dict([formula])

            if self.is_killed_event.isSet(): return
            self.generate_imc(basename)
            if self.is_killed_event.isSet(): return

            property_checking_record = PropertyCheckingRecord.createEmpty("SMV",[basename+"_smv2sigref.xml"])
            property_checking_record.label_map = label_map
            property_checking_record.addProperty(self.csl_property)
            property_checking_record.writeToFile()

            formula_intermediate_string = properties_api.get_formula_as_intermediate_string(formula)
            formula_number = property_checking_record.property_map[formula_intermediate_string]

            selected_tools = tool_settings_model.get_selected_tools_array()
            self.minimise_imc_to_ctmc([formula_number], os.path.dirname(basename), selected_tools)
            if self.is_killed_event.isSet(): return

            # Update property checking record with results of model2imc
            property_checking_record.mergeWithConversionResults()
            property_checking_record.writeToFile()

            # If a single tool has been selected make sure it is used
            # (i.e. if IMCA is selected use it even though a IMC for MRMC may already exist)    
            force_tool = None if len(selected_tools) > 1 else selected_tools[0]

            property_checker = PropertyChecker(property_checking_record.record_file_path, selected_tools)
            property_checker.set_imca_handler(self.imca)
            property_checker.set_mrmc_handler(self.mrmc)
            property_checker.debug_info = self.slim_model.debug_info

            try:
                if self.csl_property.get_pattern_id() == "expectedTime":
                    (self.analysis_result, error_msg) = property_checker.get_results(self.csl_property, tool_settings_model.get_error_bound(), force_tool)
                elif self.csl_property.get_pattern_id() == "longRunAverage":
                    (self.analysis_result, error_msg) = property_checker.get_results(self.csl_property, tool_settings_model.get_error_bound(), force_tool)
                else:
                    (self.analysis_result, error_msg) = property_checker.get_results(self.csl_property, tool_settings_model.get_error_bound(), force_tool, True, 50)

                # stores the result
                child_iter = self.filtered_props.get_properties_list().convert_iter_to_child_iter(self.row.iter)
                self.filtered_props.props_model.set_analysis_result(child_iter, self.analysis_result)
                if error_msg:
                    gobject.idle_add(self.display_error, error_msg)
            except exceptions.ProcessInternalException as e:
                # Most likely caused by interrupt
                utils.log.system.warning("Internal process exception during performability: " + str(e))
                if not self.killed:
                    raise e
        except exceptions.UserException, e:
            self.stop_performability()
            gobject.idle_add(self.display_error, str(e))
        finally:
            self.busy -= 1
            utils.log.system.debug("Done with Performability Evaluation")

    def display_error(self, message):
        msgbox = gtk.MessageDialog(type=gtk.MESSAGE_ERROR,
                           buttons=gtk.BUTTONS_OK,
                           message_format=message)
        msgbox.run()
        msgbox.destroy()

    def stop_performability(self):
        self.killed = True
        self.is_killed_event.set()
        if self.__kill_ord_event: # waiting for the ordering to be generated
            self.__kill_ord_event_processed.clear()
            self.__kill_ord_event.set()
            self.__kill_ord_event_processed.set()
            pass

        self.smv2sigref.stop_smv_to_sigref()
        self.model2imc.stop_model2imc()
        self.mrmc.stop_mrmc()
        self.imca.stop_imca()
        return

    def translate_models(self):
        """ Translates the SLIM model to SMV. Sets the property considered
        for verification. Dumps the event/probability mapping and the
        atomic propositions
        """

        # run the translation of the extended model
        eroot_impl = self.slim_model.get_extended_root_impl()
        self.__kill_ord_event = threading.Event()

        smv_fname, ord_fname, _, _ = self.slim_model.generate_smv_file(True, self.__kill_ord_event)
        killed = self.__kill_ord_event.isSet()
        self.__kill_ord_event = None
        self.__kill_ord_event_processed.set()
        # return None indicates the translation has failed
        if killed: 
            return None

        # set the basename (used subsequently for dumping of .epm and .map)
        basename = os.path.splitext(smv_fname)[0]
        #options.BASE_FILE_NAME = basename

        # reuse debug info to hold debugging data (in case some step
        # crashes halfway)
        debug_info = self.slim_model.debug_info
        # set the csl formula under verification 
        self.set_property()

        # dump the event-probability mapping
        dump.dump_error_event_probability_map(eroot_impl, debug_info)
        # dump the atomic propositions
        formulae_doc = properties_api.create_formulae_document_for_properties([self.csl_property])
        dump.dump_ap_map_and_replace_aps(eroot_impl, formulae_doc, debug_info)

        return basename    

    def generate_imc(self, basename):
        """ Invokes NuSMV to generate the state space and dump it as
        an Sigref-readable XML """

        self.smv2sigref.busy += 1
        self.smv2sigref.start_smv_to_sigref(basename)
        xml_fname, _ = self.smv2sigref.convert_smv_to_sigref(
                            misc.get_out_fname(basename + ".smv"), 
                            misc.get_out_fname(basename + ".epm"), 
                            misc.get_out_fname(basename + ".ord"), 
                            misc.get_out_fname(basename + ".map"))
        if self.smv2sigref.tool is not None:
            self.slim_model.debug_info.smv2sigref_output = (self.smv2sigref.tool.get_stdout_text(),
                                                            self.smv2sigref.tool.get_stderr_text())
            if self.smv2sigref.tool.returncode != 0 and self.smv2sigref.tool.returncode is not None:
                gobject.idle_add(self.display_error, "IMC generation error: Tool died with code %d" % self.smv2sigref.tool.returncode)
                self.killed = True
                self.is_killed_event.set()
        self.smv2sigref.stop_smv_to_sigref()
        self.smv2sigref.busy -= 1

        return xml_fname

    def minimise_imc_to_ctmc(self, formula_numbers, output_dir, output_formats):
        """ Run Model2Imc to minimise the state space (as an IMC), which 
        hopefully results into a CTMC"""

        # run the minimization to markov chains
        self.model2imc.busy += 1
        self.model2imc.start_model2imc(formula_numbers, output_dir, output_formats)

        if self.model2imc.tool is not None:
            self.slim_model.debug_info.model2imc_output = (self.model2imc.tool.get_stdout_text(),
                                                           self.model2imc.tool.get_stderr_text())
            if self.model2imc.tool.returncode != 0:
                gobject.idle_add(self.display_error, "CTMC/IMC generation error: Tool died with code %d" % self.model2imc.tool.returncode)
                self.killed = True
                self.is_killed_event.set()
        # kill all remainders of the process
        self.model2imc.stop_model2imc()
        self.model2imc.busy -= 1

    # observable properties
    def property_infinite_domain_value_change(self, model, old, new):
        """If the slim domain has infinite domain, performability is not
        available"""
        self.infinite_model = new
    def property_failures_value_change(self, model, old, new):
        """If the slim domain has infinite domain, performability is not
        available"""
        self.failures = new

    # end of class PerformabilityTab    

class ToolSettingsModel(Model):
    valid_tools = ["Auto","MRMC","IMCA"]

    selected_tool = "Auto"
    selected_error_bound = 0.1

    __observables__ = ("selected_tool", "selected_error_bound")

    def get_selected_tool(self):
        return self.selected_tool

    def set_selected_tool(self, tool):
        self.selected_tool = tool

    def get_selected_tools_array(self):
        if self.selected_tool == "Auto":
            return ["mrmc","imca"]
        else:
            return [self.selected_tool.lower()]

    def get_error_bound(self):
        return self.selected_error_bound

    def set_error_bound(self, error_bound):
        self.selected_error_bound = (float)(error_bound)

    def is_error_bound_suitable(self, error_bound):
        try: 
            float(error_bound)
        except ValueError:
            return False
        return float(error_bound) > 0
