# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================


# Model Checking and Simulation model
from __future__ import with_statement
import gtk
import gobject
import re
from gtkmvc import Model, ModelMT
import tempfile

from src.models.activity import ActivityModel
from src.models.filtered_properties import FilteredPropertiesModel
from src.models.tool_handler import ToolHandlerModel, FsapHandlerModel, MRMCHandlerModel
from src.models.performability import ProbabilisticToolHandlers

from utils import fsap_process, model2imc_process, exceptions
from utils.misc import relpath, get_random_name, get_progressive_name, get_out_fname
from utils import cutsets, log
import options

import slim_compiler.src.smv.api as smv_api
import invokers.dump as dump
import utils.fault_tree as fault_tree_module
import property_management.src.api as properties_api
import property_management.src.properties as properties
from property_management.src.constants import Logic

import os
import os.path
from gtk import ListStore
import types
from utils.property_checker import PropertyCheckingRecord, PropertyChecker, DFTPropertyChecker,\
    DFTEvaluater
from invokers.dump import dump_event_name_to_label_map
from tool_handler import Model2ImcHandlerModel, IMCAHandlerModel
import math

class SafetyModel (ActivityModel):
    """Model of safety panel. It contains submodels for fault tree
    generation, and failure mode and analysis"""

    def __init__(self, appl_model, slim_model, props_model):        
        ActivityModel.__init__(self, appl_model, slim_model, props_model)
        
        # this is the set of propertis filtered
        self.filtered_props = FilteredPropertiesModel(props_model)
        self.filtered_props.set_filtering(None) # only propositional

        self.ftg = SafetyFaultTreeGen(slim_model, self.filtered_props)
        self.fmea = SafetyFMEA(slim_model, self.filtered_props)
        self.ftoe = SafetyFaultTolEval(slim_model, self.filtered_props)
        self.ftv = SafetyFaultTreeVer(slim_model, self.filtered_props)
        self.fte = SafetyFaultTreeEval(slim_model, self.filtered_props)

        # monitors
        self.appl_model.tools.add_tool("Fault Tree", self.ftg, "Safety Analysis")
        self.appl_model.tools.add_tool("FMEA", self.fmea, "Safety Analysis")
        self.appl_model.tools.add_tool("Model2Imc (DFT Verification)", self.ftv.model2imc, "Safety Analysis")    
        self.appl_model.tools.add_tool("MRMC (DFT Verification)", self.ftv.mrmc, "Safety Analysis")
        self.appl_model.tools.add_tool("IMCA (DFT Verification)", self.ftv.imca, "Safety Analysis")
        self.appl_model.tools.add_tool("Model2Imc (FT Evaluation)", self.fte.model2imc, "Safety Analysis")
        self.appl_model.tools.add_tool("MRMC (FT Evaluation)", self.fte.mrmc, "Safety Analysis")
        self.appl_model.tools.add_tool("IMCA (FT Evaluation)", self.fte.imca, "Safety Analysis")

        
        # there is no tool behind FTOE:
        #self.appl_model.tools.add_tool("Fault Tolerance", self.ftoe, "Safety Analysis")
        
        # TODO: enable processes view here
        #self.appl_model.tools.add_tool("Fault Tree", self.ftv, "Safety Analysis")

        # Observes busy levels
        self.ftg.register_observer(self)
        self.fmea.register_observer(self)
        self.ftoe.register_observer(self)
        self.ftv.register_observer(self)
        self.fte.register_observer(self)

        return

    def stop_fsap(self):
        """This is called from the upper level when it is time to go sleep"""
        for m in (self.ftg, self.fmea): m.stop_fsap()
        self.ftoe.stop_tool()
        return

    # ---------------------------------------------------------------
    #        Notifications from observable properties
    # ---------------------------------------------------------------    
    def property_busy_value_change(self, model, old, new):
        if model is self: return # not interested in introspection
        self.busy += (new - old)
        return

    pass # end of class
# ----------------------------------------------------------------------



class SafetyFaultTreeGen (FsapHandlerModel):
    """This is the model for safety fault tree generation"""

    use_bdd = True  # if bdd or sat should be used
    sat_sbmc = True # if sbmc or bmc should be used
    sbmc_complete = False # whether to perform completeness check
    can_use_bdd = True # if BDD can be used 

    # dynamic FMEA (applies to BDD only)
    dynamic = False
    sat_bound = options.NUSMV_BMC_LENGTH

    __observables__ = ("dynamic", "sat_bound", "use_bdd", "sat_sbmc", 
                       "sbmc_complete" ,"can_use_bdd")

    def __init__(self, slim_model, properties):
        FsapHandlerModel.__init__(self, slim_model)
        self.filtered_props = properties

        # this model is interested in observing if the slim
        # model has infinite domain
        self.observe_model(slim_model)
        return

    
    def generate_fault_tree(self):
        """using the underlying fsap process, generates two files
        that represent a fault tree. Returns the generated file
        names as a tuple (events_fname, gates_fname, ft_fname). The
        generated files can be passed to method load_fault_tree."""
        self.busy += 1

        try:
            self.start_fsap()

            rows = [ row for row in self.filtered_props if row[FilteredPropertiesModel.SELECT_SINGLE] ]
            assert len(rows) == 1 # there must be one property selected

            properties = [ row[FilteredPropertiesModel.PROPERTY].property 
                           for row in rows ]

            formulae_doc = properties_api.create_formulae_document_for_properties(properties)
            smv_formulae_doc = properties_api.create_formulae_document_for_properties(properties)
            properties_api.convert_atomic_propositions_to_smv(smv_formulae_doc, self.slim_model.get_extended_root_impl())

            properties = [(slim, properties_api.get_formula_as_smv_string(smv)) for slim, smv in zip(formulae_doc, smv_formulae_doc)]
            smv_prop_text = properties[0][1]
            (tmp_ev_fname, tmp_gt_fname, tmp_ft_fname) = \
                self.get_tool().compute_fault_tree(smv_prop_text, 
                                                   dynamic=self.dynamic, 
                                                   bmc_length=self.sat_bound,
                                                   use_bdd =self.use_bdd,
                                                   sat_sbmc=self.sat_sbmc,
                                                   sbmc_completeness=self.sbmc_complete)

            # rename the files according to the base name
            # [MP] rnd_suffix = "_" + get_random_name(6) 
            rnd_suffix = "_" + get_progressive_name(6) 
                
            ev_fname = get_out_fname(options.BASE_FILE_NAME + rnd_suffix + ".flt_events")
            gt_fname = get_out_fname(options.BASE_FILE_NAME + rnd_suffix + ".flt_gates")
            ft_fname = get_out_fname(options.BASE_FILE_NAME + rnd_suffix + "_flt.xml")

            # If FSAP execution is unsuccessful then an exception is raise
            # and here we assume the run was successful and files have been generated.
            assert (os.path.isfile(tmp_ev_fname) or 
                    os.path.isfile(tmp_ft_fname) or
                    os.path.isfile(tmp_gt_fname)), \
                    "Fault Tree files have not been generated by FSAP for unknown reason"

            for source, dest in ((tmp_ev_fname, ev_fname), (tmp_gt_fname, gt_fname), (tmp_ft_fname, ft_fname)):                
                os.rename(source, dest)
                pass

            # Determine the rates of the basic events of the fault tree and
            # store them in a separate file. This is needed in order to
            # perform further analyses on the fault tree (fault tree
            # evaluation or fault tree verification).
            # Note: At this point, the event descriptions in the fault tree
            # XML file must still be in SMV syntax.
            rates_fname = None
            root_impl = self.slim_model.get_extended_root_impl()
            fault_tree = fault_tree_module.parse_xml(ft_fname)
            try:
                event_to_rate_dict = fault_tree_module.\
                        create_basic_event_name_to_rate_dict(fault_tree,
                                                             root_impl)
            except fault_tree_module.FaultTreeError, e:
                # The rates could not be determined. No rates file can be
                # written.
                # TODO: Log a warning here.
                assert rates_fname is None
            else:
                # The rates could be determined. Write them to a file.
                rates_fname = get_out_fname(options.BASE_FILE_NAME + 
                                    rnd_suffix + ".rates_map")
                dump.dump_event_name_to_rate_map(rates_fname,
                                                 event_to_rate_dict)

            if not options.FSAP_KEEP_SMV_IDS:
                fsap_process.convert_fault_tree_events(ev_fname, ev_fname,
                                                       smv_api.convert_smv_const_to_slim)
                fsap_process.convert_fault_tree_gates(gt_fname, gt_fname,
                                                      smv_api.convert_smv_const_to_slim)
                fsap_process.convert_fault_tree_xml(ft_fname, ft_fname,
                                                    smv_api.convert_smv_const_to_slim)
                pass

            # stores the result                
            child_iter = self.filtered_props.get_properties_list().convert_iter_to_child_iter(rows[0].iter)
            self.filtered_props.props_model.set_ft(child_iter, (ev_fname, gt_fname, ft_fname, rates_fname))

        finally: self.busy -= 1

        return (ev_fname, gt_fname)


    def load_fault_tree(self, ev_fname, gt_fname):
        self.busy += 1
        try:
            process = fsap_process.FTDProcess(ev_fname, gt_fname)

        finally: self.busy -= 1
        return
        

    # observable properties
    def property_infinite_domain_value_change(self, model, old, new):
        """If the slim domain has infinite domain, dynamic FMEA and
        FT are not available"""
        if new and self.dynamic: self.dynamic = False
        self.can_use_bdd = new
        return

    pass # end of class
# ----------------------------------------------------------------------


class SafetyFMEA (FsapHandlerModel):
    """This is the model for safety FMEA table generation"""
    
    cardinality = options.FSAP_FAULT_CONFIGURATION_CARDINALITY

    # number, ID, failure mode, failure effect
    fmea_table = ListStore(types.IntType, types.StringType, types.StringType, types.StringType)

    # dynamic FMEA
    dynamic = False
    sat_bound = options.NUSMV_BMC_LENGTH
    compact = options.FSAP_COMPACT_FMEA_TABLE
    __observables__ = ("dynamic", "sat_bound", "compact")


    def __init__(self, slim_model, properties):
        FsapHandlerModel.__init__(self, slim_model)
        self.filtered_props = properties

        # this model is interested in observing if the slim
        # model has infinite domain
        self.observe_model(slim_model)
        return


    def generate_fmea_table(self):  
        """Genrates the fmea table and returns the name of the file containing it"""

        self.busy += 1
        try:
            options.FSAP_FAULT_CONFIGURATION_CARDINALITY = self.cardinality
            self.start_fsap()

            properties = [ row[FilteredPropertiesModel.PROPERTY].property \
                               for row in self.filtered_props \
                               if row[FilteredPropertiesModel.SELECT_MULTI] ]
            assert len(properties) > 0 # at least one property

            formulae_doc = properties_api.create_formulae_document_for_properties(properties)
            smv_formulae_doc = properties_api.create_formulae_document_for_properties(properties)
            properties_api.convert_atomic_propositions_to_smv(smv_formulae_doc, self.slim_model.get_extended_root_impl())

            properties = [(slim, properties_api.get_formula_as_smv_string(smv)) 
                          for slim, smv in zip(formulae_doc, smv_formulae_doc)]
            props = [ x[1] for x in properties ]  # list of smv text formulae

            tmp_fmea_fname = self.get_tool().compute_fmea_table(props,
                                                                options.FSAP_FAULT_CONFIGURATION_CARDINALITY, 
                                                                dynamic=self.dynamic, 
                                                                bmc_length=self.sat_bound, 
                                                                compact=self.compact)

            # rename the files according to the base name
            fmea_fname = get_out_fname(options.BASE_FILE_NAME + ".fmea")
  
            # If FSAP execution is unsuccessful then an exception is raise
            # and here we assume the run was successful and file has been generated.
            assert os.path.isfile(tmp_fmea_fname), \
                "FMEA Table file has not been generated by FSAP for unknown reason (advice: run with '-l debug' to see FSAP messages)"

            os.rename(tmp_fmea_fname, fmea_fname)
           
            if not options.FSAP_KEEP_SMV_IDS:
                fsap_process.convert_fmea_table(fmea_fname, fmea_fname,
                                                smv_api.convert_smv_const_to_slim)
                pass        

        finally: self.busy -= 1
        return fmea_fname


    def load_fmea_table(self, fmea_table_fname):
        """Returns True if the table was generated, or False if the table is empty"""
        self.busy += 1

        try:
            self.fmea_table.clear()
            res = False

            _file = open(fmea_table_fname, "r")
            line = _file.readline()
            while line:
                # the first line contains the headers and it is ignored
                line = _file.readline().strip()
                if line: 
                    fields = line.split("\t")
                    assert len(fields) == 4

                    # fixed fields values
                    fields[0] = int(fields[0])
                    if not fields[1]: fields[1] = "N/A"

                    self.fmea_table.append(fields)
                    res = True
                    pass
                pass
            _file.close()
            
        finally: self.busy -= 1
        return res
    

    # observable properties
    def property_infinite_domain_value_change(self, model, old, new):
        """If the slim domain has infinite domain, dynamic FMEA and
        FT are not available"""
        if self.slim_model is model: 
            if new and self.dynamic: self.dynamic = False
            pass
        return

    pass # end of class
# ----------------------------------------------------------------------



class SafetyFaultTolEval (ToolHandlerModel):
    """This is the model for safety fault tolerance evaluation"""

    # cardinality, number of cutsets
    ftoe_table = ListStore(types.IntType, types.IntType)
  
    def __init__(self, slim_model, properties):
        ToolHandlerModel.__init__(self)

        self.slim_model = slim_model
        self.filtered_props = properties

        # this model is interested in observing if the slim
        # model has infinite domain
        self.observe_model(slim_model)
        
        return

    def get_selected_ft(self):
        """Returns the set of ft (as xml file names) which has been
        selected along with their properties"""
        # 2 is the third element of (event, gates, xml_fname)
        return [row[FilteredPropertiesModel.FT][2] \
                    for row in self.filtered_props \
                    if (row[FilteredPropertiesModel.SELECT_MULTI] and 
                        row[FilteredPropertiesModel.FT] is not None)]

    def run_fault_tolerance_eval(self):
        """Performs FTOE by taking the FTs out of the selected
        preperties (internally stored)"""
        fnames = self.get_selected_ft()        
        self.eval_fault_tolerance(fnames)
        return
        
    def eval_fault_tolerance(self, fts):
        """Performs FTOE. fts is a sequence of xml files each
        containing a fault tree.  Returns a list of pairs. Each
        pair represent a row of the table produced by the Fault
        Tol. Eval (cardinality, cut-sets). Each element of the pair
        is a string."""

        log.system.debug("Running Fault Tolerance Evaluation, taking files: %s" % ", ".join(fts))

        csm = cutsets.get_unique_cut_sets_from_files(fts)
  
        self.ftoe_table.clear()
        for card in csm: self.ftoe_table.append((card, len(csm[card])))

        log.system.debug("Done with Fault Tolerance Evaluation")      
        return [(row[0], row[1]) for row in self.ftoe_table]
    
    pass # end of class
# ----------------------------------------------------------------------

class SafetyFaultTreeVer (ProbabilisticToolHandlers):
    """This is the model for safety fault tree verification."""
    def __init__(self, slim_model, properties):
        ProbabilisticToolHandlers.__init__(self, None, slim_model, properties)        
        # Create a ListStore of CSL patterns for the combo box in the view.
        # The pattern list is a list of tuples, where the first entry is the
        # name of the pattern and the second the pattern object associated
        # to the former name.
        pd = [p for p in properties_api.get_patterns_document(Logic.CSL)]
        self.pattern_list = gtk.ListStore(types.StringType, object)
        for p in pd:
            self.pattern_list.append(["CSL " + p._get_id(), p])

        # Store a reference to the fault tree if one was loaded and not yet
        # unselected.
        self.fault_tree = None

        # Analysis result object
        self.analysis_result = None

        # True if the process was interrupted by the user
        self.killed = False
    
    def get_selected_ft(self):
        # TODO: docstring
        
        # Retrieve a list of all currently selected properties with associated
        # fault trees.
        fts = [row[FilteredPropertiesModel.FT][2] \
                    for row in self.filtered_props \
                    if (row[FilteredPropertiesModel.SELECT_SINGLE] and 
                        row[FilteredPropertiesModel.FT] is not None)]
        
        # If there is exactly one fault tree selected, return this fault tree
        # and None otherwise.
        if len(fts) == 1:
            return fts[0]
        else:
            return None


    def get_basic_event_rates_file(self):
        # TODO: docstring
        
        # Retrieve a list of the names of the rates files for all currently
        # selected properties with associated fault trees.
        rates_filenames = [row[FilteredPropertiesModel.FT][3] \
                           for row in self.filtered_props \
                           if (row[FilteredPropertiesModel.SELECT_SINGLE] and 
                               row[FilteredPropertiesModel.FT] is not None)]
        
        if len(rates_filenames) == 1:
            return rates_filenames[0]
        else:
            return None

    def load_ft(self, ft_filename):
        """Parses the given fault tree file.
        """
        self.fault_tree = fault_tree_module.parse_xml(ft_filename)


    def stop_dft_verification(self):
        self.killed = True
        self.model2imc.stop_model2imc()
        self.mrmc.stop_mrmc()
        self.imca.stop_imca()

    from utils.log import log_icd
    @log_icd
    def run_dft_verification(self, fault_tree_filename, basic_event_rates_filename, property, tool_settings_model):
        # TODO: docstring
        
        self.killed = False
        
        self.busy += 1
        
        log.system.debug("Running Fault Tree Verification")
        
        try:
            properties_document = properties.PropertiesDocument.create([property])
            selected_tools = tool_settings_model.get_selected_tools_array()
            dft_checker = DFTPropertyChecker(fault_tree_filename, basic_event_rates_filename, properties_document, selected_tools)
            dft_checker.set_model2imc_handler(self.model2imc)
            dft_checker.set_mrmc_handler(self.mrmc)
            dft_checker.set_imca_handler(self.imca)
            dft_checker.debug_info = self.slim_model.debug_info
            
            dft_checker.gen_model_files()

            try:
                error_msg = None
                if property.get_pattern_id() == "expectedTime":
                    (self.analysis_result, error_msg) = dft_checker.analyze(property, tool_settings_model.get_error_bound())
                elif property.get_pattern_id() == "longRunAverage":
                    (self.analysis_result, error_msg) = dft_checker.analyze(property, tool_settings_model.get_error_bound())
                else:
                    (self.analysis_result, error_msg) = dft_checker.analyze(property, tool_settings_model.get_error_bound(), True, 50)

                if error_msg:
                    gobject.idle_add(self.display_error, error_msg)
            except exceptions.ProcessInternalException as e:
                # Most like due to process termination
                log.system.warning("Internal process exception: " + str(e))
                if not self.killed:
                    raise e
        finally:
            self.busy -= 1
            log.system.debug("Done with Fault Tree Verification")

    def display_error(self, message):
        msgbox = gtk.MessageDialog(type=gtk.MESSAGE_ERROR,
                           buttons=gtk.BUTTONS_OK,
                           message_format=message)
        msgbox.run()
        msgbox.destroy()

    def reset(self):
        ProbabilisticToolHandlers.reset(self)
        self.fault_tree = None
    
    pass # end of class
# ----------------------------------------------------------------------

class SafetyFaultTreeEval (ToolHandlerModel):
    """This is the model for safety fault tree evaluation"""
  
    def __init__(self, slim_model, properties):
        ToolHandlerModel.__init__(self)

        self.slim_model = slim_model
        self.filtered_props = properties

        self.model2imc = Model2ImcHandlerModel()
        self.mrmc = MRMCHandlerModel()
        self.imca = IMCAHandlerModel()

        self.fault_tree = None

        # True if the process was interrupted by the user
        self.killed = False

    def reset(self):
        self.fault_tree = None

    def get_selected_ft(self):
        # TODO: docstring
        
        # Retrieve a list of all currently selected properties with associated
        # fault trees.
        fts = [row[FilteredPropertiesModel.FT][2] \
                    for row in self.filtered_props \
                    if (row[FilteredPropertiesModel.SELECT_SINGLE] and 
                        row[FilteredPropertiesModel.FT] is not None)]
        
        # If there is exactly one fault tree selected, return this fault tree
        # and None otherwise.
        if fts == [] or len (fts) > 1:
            return None
        else:
            return fts[0]

    def get_basic_event_rates_file(self):
        # TODO: docstring

        # Retrieve a list of the names of the rates files for all currently
        # selected properties with associated fault trees.
        rates_filenames = [row[FilteredPropertiesModel.FT][3] \
                           for row in self.filtered_props \
                           if (row[FilteredPropertiesModel.SELECT_SINGLE] and 
                               row[FilteredPropertiesModel.FT] is not None)]
        
        if len(rates_filenames) == 1:
            return rates_filenames[0]
        else:
            return None

    def load_ft(self, ft_filename):
        """Parses the given fault tree file.
        """
        self.fault_tree = fault_tree_module.parse_xml(ft_filename)


    from utils.log import log_icd
    @log_icd
    def run_dft_evaluation(self, fault_tree,
                           basic_event_rates_filename, mission_time,
                           criticality, events_filename, gates_filename, tool_settings_model):
        # TODO: docstring
        self.killed = False
        
        self.busy += 1
        
        log.system.debug("Running Fault Tree Evaluation")
        
        try:
            # Store the fault tree in a temporary file so that it can be used as input for model2imc.
            file_prefix = "fault_tree"
            (fd, ft_filename) = tempfile.mkstemp(suffix=".xml", prefix=file_prefix,text=True)
            os.write(fd,fault_tree_module.get_as_xml_string(fault_tree))
            os.close(fd)

            selected_tools = tool_settings_model.get_selected_tools_array()
            dft_evaluator = DFTEvaluater(ft_filename, basic_event_rates_filename, events_filename, gates_filename, selected_tools)
            dft_evaluator.set_model2imc_handler(self.model2imc)
            dft_evaluator.set_mrmc_handler(self.mrmc)
            dft_evaluator.set_imca_handler(self.imca)
            dft_evaluator.debug_info = self.slim_model.debug_info

            try:
                dft_evaluator.analyze(mission_time, criticality, tool_settings_model.get_error_bound())
            except exceptions.ProcessInternalException as e:
                # Most like caused by process termination
                log.system.warning("Internal process exception: " + str(e))
                if not self.killed:
                    raise e
        finally:
            self.busy -= 1
            log.system.debug("Done with Fault Tree Evaluation")

    def write_fault_tree_plus_output(self, fault_tree,
                                     event_name_to_probability_dict,
                                     event_name_to_criticality_dict,
                                     events_filename, gates_filename):
        # TODO: docstring
        
        # TODO: Maybe this function should be moved to utils.fault_tree.
        (events_file_content, gates_file_content) = \
                fault_tree_module.get_in_fault_tree_plus_format(
                        fault_tree, event_name_to_probability_dict,
                        event_name_to_criticality_dict)
        with open(events_filename, "w") as events_file:
            events_file.write(events_file_content)
        with open(gates_filename, "w") as gates_file:
            gates_file.write(gates_file_content)


    def load_fault_tree(self, events_filename, gates_filename):
        self.busy += 1
        try:
            process = fsap_process.FTDProcess(events_filename, gates_filename)

        finally: self.busy -= 1
        return


    def stop_dft_evaluation(self):
        self.killed = True

        self.model2imc.stop_model2imc()
        self.mrmc.stop_mrmc()
        self.imca.stop_imca()
        
    pass # end of class
# ----------------------------------------------------------------------
