# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================


import os.path
import types
from gtk import ListStore, SORT_ASCENDING

from gtkmvc import Model
from gtkmvc import observable

# these are from the compass toolchain
import options
from utils import debug_info
from utils.misc import relpath

from invokers import slim, nusmv
import compass
from slim_compiler.src.dataStructure.nominalbehaviour \
    import ComponentImplementation
import slim_compiler.src.smv.support as smvsupport
import utils.log
from slim_compiler.src.model_extension import fault_injection
import slim_compiler.src.model_extension.core as model_extension_core
from slim_compiler.src.model_extension.util import general

from src.models.console import ConsoleModel


class SlimModel (Model):

    # list of open slim models (filenames, isEnabled, isValid)
    # NOTE: isEnabled is currently not used
    slim_files = ListStore(types.StringType, bool, bool)
    slim_files_num = 0  # number of loaded slim models

    # list of ComponentImplementation,
    # corresponding slim filename, selected as root
    roots = ListStore(object, types.StringType, bool)

    # list of FDIR ComponentImplementation
    # fdir component object, component name, selection, replacement object,
    # containing subpackage, key within implementations dictionary
    fdir_components = ListStore(object, types.StringType, bool, object,
                                object, types.StringType)

    # list of FaultInjections
    fault_injections = ListStore(bool,  # isEnabled
                                 bool,  # canBeEnabled
                                 object,  # tuple of package identifier
                                          # and error model implementation name
                                 types.StringType,  # error state
                                 object,  # component instance tupled path
                                 types.StringType,  # data port/subcomp
                                 types.StringType  # effect
                                )
    failures = 0  # number of enabled failures

    # the currently selected root implementation
    root_impl = None

    # the slim model changed:
    changed = observable.Signal()

    # name of the generated smv file, infinite_domain,
    # timed_domain, ordering file name
    smv_fname = None
    infinite_domain = False
    timed_domain = False
    ord_fname = None
    open_world = False

    fdir_comp_counter = 0 #LB: to know if a fdir component is selected
    root_counter = 0 #LB: to know if a root is selected

    # observable properties
    __observables__ = ("root_impl", "changed", "failures", "slim_files_num",
                       "infinite_domain", "timed_domain", "open_world", "fdir_comp_counter", "root_counter", )

    def __init__(self, force_name, force_dir):
        Model.__init__(self)

        # Override base_file_name if loading new root
        self.force_name = force_name
        # Override out_dir if new root
        self.force_dir = force_dir

        # sub models
        self.console = ConsoleModel()

        self.debug_info = debug_info.create_debug_and_compile_information()
        self.compile_info = self.debug_info.compile_info

        # sets up sorting
        for sortable in (self.roots):
            sortable.set_sort_func(3, lambda m, i1, i2:
                                   cmp(m[i1][0].name, m[i2][0].name))
            sortable.set_sort_column_id(3, SORT_ASCENDING)

        # ths used to keep track whether the smv file was generated
        # with extended model or not the last time
        self.last_smv_gen_params = None

        # self is interested in taking some action when changing
        self.register_observer(self)

        # In case of error, debug_info contains useful info
        compass.register_debug_info(self.debug_info)

        return

    def get_name_from_impl_id(self, id):
        """Given an implementation id, returns the name of the
           implementation"""
        return id[0]

    def toggle_root(self, path):
        """Toggles the root implementation found at given path to
        be the root. If not previously selected, the implementation
        becomes the root, and previously selected root (if any)
        becomes unselected. If it was already selected, it is
        toggled to unselected, and no root is selected at all.
        This method has also effects on the nominal implementation
        list, setting visibility of root and non-root
        implementations.

        This method sets also property root_impl accordingly. To know
        what is the currently selected root, refer to property
        root_impl."""
        self._set_root_impl(self.roots[path][0])
        self.root_counter += 1
        return

    def toggle_fdir(self, path):
        """If FDIR component was selected before, deselect it and replace it by
        a dummy (empty) FDIR component of the same type.
        If it was deselected, select it and restore the original
        FDIR component.

        For the storage of the dummy or the original component the field 3 of
        self.fdir_components is used.
        """
 
        row = self.fdir_components[path]
        #invert selection
        row[2] = not row[2]

        if row[2]:  # use the original
            row[4].implementations[row[5]] = row[0]
        else:  # use the dummy
            row[4].implementations[row[5]] = row[3]

        self.fdir_comp_counter += 1

        self.changed.emit()

    def _set_root_impl(self, root_impl):
        """This is used internally to set the current root
        implementation. If None, then the root impl is
        unset. Whatever is needed to set/unset the root
        implementation is done here"""

        # searches for the iterator poiunting to the given root
        # impl, sets all the other to be not the root, toggle the
        # selection value for the root.
        selected = None
        for row in self.roots:
            if row[0] == root_impl:
                selected = row[2]
                row[2] = not selected
            elif row[2]:
                row[2] = False

        # if it was already selected, set to None, otherwise set it
        old_root_impl = self.root_impl
        self.root_impl = None if selected else root_impl

        # if still None and only one root is givem set the root
        if len(self.roots) == 1:
            self.root_impl = self.roots[0][0]
            self.roots[0][2] = True
            pass

        if self.root_impl is not None:
            topslim = os.path.basename(self.root_impl.file)
            if self.force_name:
                options.BASE_FILE_NAME = os.path.splitext(topslim)[0]
            if self.force_dir:
                options.OUT_DIR_NAME = os.path.join(os.path.dirname(
                    self.root_impl.file), options.BASE_FILE_NAME)
            # Force CWD to last loaded SLIM file
            os.chdir(os.path.dirname(self.root_impl.file))

        # report that a change occurred
        if self.root_impl != old_root_impl:
            utils.log.system.info(
                "Root implementation changed: base name is '%s'",
                str(options.BASE_FILE_NAME))
            self.changed.emit()
            self._fix_injections()

        return

    def add_slim_models(self, fnames):
        """Adds a given list of files to the current list of loaded names. It
        invokes a parsing + checking operation to check whether the
        resulting set of files is consistent. If any error occurs, all
        files are marked invalid.
        """

        loaded_fnames = [x[0] for x in self.slim_files]
        new_fnames = [x for x in fnames if x not in loaded_fnames]

        for fname in new_fnames:
            if not os.path.isfile(fname):
                raise IOError("Slim file not found: %s" % fname)

        for fname in new_fnames:
            self.slim_files.append([fname, True, True])

        (cinfo, is_ok) = self._compile_slim_files()

        # clear the roots, they have to be redetermined
        self.roots.clear()
        self._set_root_impl(None)

        # whenever a file is added, all files are reparsed so we clear the
        # fdir component list, too
        self.fdir_components.clear()

        # set the validity of the loaded SLIM files
        for row in self.slim_files:
            row[2] = is_ok

        if is_ok:
            # set the compile info
            self.compile_info = cinfo
            # set the roots
            roots = smvsupport.get_root_components(cinfo)
            for root in roots:
                self.roots.append([root, root.file, False])

            # finally, sets the root implementation if only one root is given
            if len(roots) == 1:
                self._set_root_impl(roots.pop())

            # find FDIR components and append them to the list
            fdirs = self.compile_info.get_fdir_components()
            for fdir in fdirs:
                self.fdir_components.append([fdir, fdir.file, True,
                                             self._create_dummy_for(fdir),
                                             fdir.subpackage, fdir.name])
            pass

        # check fault injections
        self._fix_injections()

        # report that a change occurred
        self.changed.emit()
        self.slim_files_num = len(self.slim_files)

        return is_ok

    def _compile_slim_files(self):
        """ Parses and checks the currently loaded SLIM files. If there are
        any errors, they are shown in the console. """
        from slim_compiler.src.dataStructure.compile import CompileInformation
        self.console.clear_compiler_msg()
        fnames = [x[0] for x in self.slim_files]
        fnames_error = set()

        # parse the files first
        cinfo = CompileInformation()
        for fname in fnames:
            if not slim.parse(cinfo, [fname]):
                fnames_error.add(fname)

        # check the files
        if len(fnames_error) == 0:
            slim.check(cinfo)

        # show errors
        for fname, linenum, id, msg, info in \
                cinfo.error_collector.get_errors():
            fnames_error.add(fname)
            self.console.log_compiler_error(fname, linenum, id, msg, info)

        # show summary per file
        for fname in fnames:
            self.console.log_compiler_msg("Compiling '%s'... " %
                                          relpath(fname))
            if fname in fnames_error:
                self.console.log_compiler_fail("ERROR\n")
            else:
                self.console.log_compiler_ok("OK\n")

        return (cinfo,
                (len(cinfo.error_collector.get_errors()) == 0 and
                 len(fnames_error) == 0))

    def _create_dummy_for(self, fdir):
        dummy = ComponentImplementation(fdir.subpackage, fdir.name,
                                        fdir.category, fdir.fdir, fdir.file,
                                        fdir.line, fdir.pos)
        dummy.add_initial_mode("dummymode", None, None, fdir.line, fdir.pos)
        return dummy

    def remove_slim_models(self, to_remove_fnames):
        """ Remove behaves as remove all loaded files and add
        the remaining SLIM files """
        self.console.clear_compiler_msg()
        if len(to_remove_fnames) == 0:
            return
        loaded_fnames = [x[0] for x in self.slim_files]
        remaining_fnames = [x for x in loaded_fnames
                            if x not in to_remove_fnames]
        self.slim_files.clear()
        self.add_slim_models(remaining_fnames)
        return

    def reload_slim_models(self):
        fnames = [x[0] for x in self.slim_files]
        self.remove_slim_models(fnames)
        self.add_slim_models(fnames)
        self.console.log_compiler_msg("Reloading complete.")
        return

    def load_fault_injections(self, filename):
        """ Loads a set of fault injections from filename """
        self.console.log_compiler_msg("Loading fault injections '%s'... " %
                                      relpath(filename))

        # check whether the filename is valid
        if not os.path.exists(filename):
            self.console.log_compiler_fail("ERROR\n")
            self.console.log_compiler_fail("\t File does not exist.\n")
            return

        # check whether the fault injections document is syntactically correct
        injection_doc = fault_injection.parse(filename)
        syntax_errors = fault_injection.get_syntax_errors_in_injection_document(
               injection_doc)
        if len(syntax_errors) > 0:
            self.console.log_compiler_fail("ERROR\n")
            self.console.log_compiler_fail("\t File contains syntax errors.\n")
            return

        # track the amount of new fault injections added
        num_added = 0
        num_contains = 0
        duplicate_fault_injections = []

        # store the fault injections
        for association in injection_doc:
            effects_spec = association.effects_specification
            for error_state in effects_spec.get_effective_states():
                for effect in effects_spec.get_effects(error_state):
                    # nicely arrange the data in verbose variable names
                    package_name = association.error_model_package_name
                    error_impl_name = association.error_model_implementation_name
                    component_path = association.component_path
                    data_element = effect.split(":=")[0].strip()
                    effect_expr = effect.split(":=")[1].strip()
                    fi_row = [True, True, (package_name, error_impl_name),
                              error_state, component_path, data_element,
                              effect_expr]
                    num_contains += 1

                    # check for uniqueness
                    unique = True
                    for row in self.fault_injections:
                        if row[2] == (package_name, error_impl_name) and \
                                row[3] == error_state and \
                                row[4] == component_path and \
                                row[5] == data_element and \
                                row[6] == effect_expr:
                            unique = False
                            break

                    # add the fault injection if it is unique
                    if unique:
                        self.fault_injections.append(fi_row)
                        num_added += 1
                    else:
                        duplicate_fault_injections.append(fi_row)

        # update the failures var
        self.failures += num_added

        # do a validity check
        self._fix_injections()

        self.console.log_compiler_ok("OK\n")
        for duplicate in duplicate_fault_injections:
            fi_str = (duplicate[2][0] + "::" + duplicate[2][1] + " to " +
                      duplicate[3] + ".".join(duplicate[4]))
            self.console.log_compiler_msg(" > Injection %s already exists.\n" %
                                          fi_str)

        self.console.log_compiler_msg(" > Loaded " + str(num_added) +
                                      " of " + str(num_contains) +
                                      " fault injections.\n")

        return

    def generate_unique_fault_injection_basename(self):
        """ Generates a unique filename for the fault injections upon saving"""
        base = options.BASE_FILE_NAME
        if base == "":
            base = "fault_injections"
        if not os.path.exists(base + ".fixml"):
            return base

        num = 2
        while os.path.exists(base + "-" + str(num) + ".fixml"):
            num += 1
        return base + "-" + str(num)

    def toggle_fault_injection(self, path):
        """ Toggles whether the user wants to enable this fault injection
        or not."""
        self.failures += -1 if self.fault_injections[path][0] else 1
        self.fault_injections[path][0] = not self.fault_injections[path][0]
        self.changed.emit()
        return

    def add_fault_injection(self, fault_injection_model):
        """Adding a fault injections given a fault injection model
        without any checking. Assumption is that the values are correct."""
        self.fault_injections.append(
            [True,
             True,
             (fault_injection_model.error_implementation.subpackage.package_name,
              fault_injection_model.error_implementation.name),
             fault_injection_model.error_state,
             fault_injection_model.nominal_instance.path,
             fault_injection_model.data,
             fault_injection_model.effect, ])
        self.failures += 1
        self.changed.emit()
        return

    def add_clones_of_fault_injection(self, path):
        """Based on a row selected representing the fault injection,
        add the cloned associated with it. """
        clones = self.get_cloneable_fault_injections(path)
        for clone in clones:
            self.fault_injections.append(clone)

        self.failures += len(clones)
        if len(clones) > 0:
            self.changed.emit()
        return

    def _fix_injections(self):
        """This is called after one or more slim files have been
        added or removed, to fix all references to the nominal and
        error implementations stored inside the fault injection
        models"""

        # if the root implementation is gone, then set all injections
        # as invalid
        if self.root_impl is None:
            for injection in self.fault_injections:
                # only decrease the failures if they were enabled
                self.failures += -1 if (injection[0] and injection[1]) else 0
                injection[1] = False
            return

        # otherwise, check injections one by one
        for injection in self.fault_injections:
            # Create an error model association
            assoc = fault_injection.ErrorModelAssociation.create(
                injection[4], injection[2][0], injection[2][1])
            # Add the fault effect of the current injection to the error
            # model association.
            assoc.effects_specification.add_effect(
                injection[3], injection[5] + " := " + injection[6])

            errors = fault_injection.get_errors_in_error_model_association(
                assoc, self.root_impl)

            # update the failures variable based on whether it was enabled
            # and it has errors and whether it *was* valid
            self.failures += (1
                              if (injection[0] and len(errors) == 0 and
                                  not injection[1])
                              else 0)
            self.failures += (-1
                              if (injection[0] and
                                  injection[1] and len(errors) > 0)
                              else 0)
            injection[1] = len(errors) == 0

        # extended model has changed
        self.changed.emit()

        # TODO: show popup to indicate that the injections have changed
        # if any have changed so far.
        return

    def remove_failures(self, paths):
        """Removes the injections stored at the given paths (iterable)"""
        # since the list will change, pre-calculates all the iterators
        for iter in [self.fault_injections.get_iter(path) for path in paths]:
            # we only need to decrement the failures var if
            # the failure was enabled and valid
            was_enabled = self.fault_injections.get_value(iter, 0)
            was_valid = self.fault_injections.get_value(iter, 1)
            self.fault_injections.remove(iter)
            self.failures += -1 if (was_enabled and was_valid) else 0

        self.changed.emit()
        return

    def get_cloneable_fault_injections(self, path):
        """Returns a list of injections that are clones of path (which is
        one fault injection) but differ in the component instance"""

        def __get_component_instance_by_tupled_path(root_impl, tuple):
            # get the set of all component instances
            component_list = general.instance_generator(self.root_impl)
            filtered_list = filter(lambda x: x.path == tupled_ci,
                                   component_list)

            # retrieve the ComponentInstance object related to tupled_ci,
            # the path identifier
            return (filtered_list[0]
                    if filtered_list
                    else None)

        # get the information of the selected fault injection
        iter = self.fault_injections.get_iter(path)
        is_enabled = self.fault_injections.get_value(iter, 0)
        is_valid = self.fault_injections.get_value(iter, 1)
        error_impl = self.fault_injections.get_value(iter, 2)
        error_state = self.fault_injections.get_value(iter, 3)
        tupled_ci = self.fault_injections.get_value(iter, 4)
        component_instance = __get_component_instance_by_tupled_path(
            self.root_impl, tupled_ci)

        if component_instance is None:
            return []
        component_impl = component_instance.implementation
        data_port = self.fault_injections.get_value(iter, 5)
        effect = self.fault_injections.get_value(iter, 6)

        # compute the set of component instances of the same nominal model
        # implementation
        def __is_nominal_implementation(component):
            return (component.implementation.equals(component_impl))

        component_set = set(
            map(lambda x: x.path,
                general.instance_generator(self.root_impl,
                                           __is_nominal_implementation)))

        # compute the set of component instances that have already
        # an error model associated but are different than the
        # error model associated with the selected fault injection
        different_association = set()
        for fault_injection in self.fault_injections:
            other_ci = __get_component_instance_by_tupled_path(
                self.root_impl, fault_injection[4])
            different = (other_ci.implementation.equals(component_impl) and
                         fault_injection[2] != error_impl)
            if different:
                different_association.add(fault_injection[4])

        # substract them from the list of retrieved components
        component_set.difference_update(different_association)

        # compute the set of component instances that have an error model
        # associated with the same error model, on the same error state,
        # and on the same data port as the selected fault injection
        same_injection = set()
        for fault_injection in self.fault_injections:
            other_ci = __get_component_instance_by_tupled_path(
                self.root_impl, fault_injection[4])
            same = (other_ci.implementation.equals(component_impl) and
                    (fault_injection[2] == error_impl) and
                    (fault_injection[3] == error_state) and
                    (fault_injection[5] == data_port))
            if same:
                same_injection.add(fault_injection[4])

        # substract them from the list of retrieved components
        component_set.difference_update(same_injection)

        cloned = [[is_enabled, is_valid, error_impl, error_state,
                   component, data_port, effect]
                  for component in component_set]
        return cloned

    def get_fault_injection_document(self):
        """Creates and returns a fault injection document representing the
        fault injections which were specified by the user.

        @return: A fault injection document based on the user's input.
        @rtype:  fault_injection.FaultInjectionDocument
        """
        # Start with a new empty fault injection document.
        injection_document = fault_injection.FaultInjectionDocument.create()

        # filter away all injections that are disabled and invalid
        injections = filter(lambda x: x[0] and x[1], self.fault_injections)

        # add all remaining injections to the fault injection document
        for injection in injections:
            # The document might already contain an error model
            # association that associates the nominal component and the
            # error model of the current injection (if there are multiple
            # injections for the component). Check if this is the case.
            assoc = injection_document.get_by_component_path(injection[4])

            if assoc is None:
                # The document does not contain an error model association
                # for the current component yet. Add one.
                new_assoc = fault_injection.ErrorModelAssociation.create(
                    injection[4], injection[2][0], injection[2][1])
                injection_document.add_association(new_assoc)
                assoc = injection_document.get_by_component_path(injection[4])
                pass

            assert assoc is not None
            # Add the fault effect of the current injection to the error
            # model association.
            assoc.effects_specification.add_effect(
                injection[3], injection[5] + " := " + injection[6])
            pass
        pass
        return injection_document

    def get_nominal_root_impl(self):
        """Returns the nominal (aka not extended) root
        implementation"""
        return self.root_impl

    def get_extended_root_impl(self):
        """Returns the root implementation of the extended
        model. If no extension was done, returns the currently set
        root implementation (can be None if not selected).

        None is returned if there are more then one implementations
        in the extended model and self.root_impl is None.

        This method triggers the calculation of the extended model"""

        if self.failures == 0:
            return self.root_impl

        roots = smvsupport.get_root_components(self.compile_info)
        # assert len(roots) > 0, "No root implementation found."
        root_implementation = None
        if self.root_impl is not None:
            root_implementation = self.root_impl
        elif len(roots) == 1:
            root_implementation = roots.pop()

        if root_implementation is None:
            return None

        # TODO: Document the invocation of model extension.
        return model_extension_core.create_extended_model(
            root_implementation,
            self.get_fault_injection_document())

    def generate_smv_file(self, use_extended_model=False,
                          kill_event=None,
                          generate_time_var=options.SMV_GENERATE_TIME_VAR,
                          generate_time_fairness_alt=options.SMV_FAIR_TIMED):
        """This method generates the smv file out of the currently
        root implementation, but only if the slim compiled files
        set changed, or if for any reason the smv file needs to be
        re-generated. Returns a tuple (smv_fname, ord_fname,
        need_msat, flag) holding the smv file name and the BDD
        variable ordering file name. Flag is True if the smv file
        was generated, or False if reusing the previously generated
        because nothing has changed.

        if use_extended_model is True, then the smv file of the
        extended model will be generated.

        kill_event is a threading.Event which can be set to
        interrupt the process from other threads.

        Parameters generate_time_var and generate_time_fairness_alt
        control if additional constraints for timed fsm have to be added,
        and will be ignored for discrete (non-timed) FSMs.
        """

        root = (self.get_extended_root_impl()
                if use_extended_model
                else self.root_impl)
        assert root is not None, "A root was expected to be selected"

        # be lazy when generating smv files
        changed = False

        # this is for checking changes wrt previous smv generation, to
        # be used for lazy generation
        gen_pars = (use_extended_model,
                    generate_time_var,
                    generate_time_fairness_alt)

        if (self.smv_fname is None or self.last_smv_gen_params != gen_pars):

            # creates the smv file out of slim compiler information
            self.smv_fname, self.infinite_domain, self.timed_domain = \
                slim.convert_to_smv(
                    root,
                    generate_time_var=generate_time_var,
                    generate_time_fairness_alt=generate_time_fairness_alt)

            self.debug_info.slim2smv_output = open(self.smv_fname, "r").read()
            utils.log.system.info("Generated smv file '%s'", self.smv_fname)

            if self.infinite_domain:
                utils.log.system.info("Generated smv file has infinite domain")
                if self.timed_domain:
                    utils.log.system.info("(and it has also timed domain)")
                    if generate_time_fairness_alt:
                        utils.log.system.info("Constraints for fair time and "
                                              "alternate transitions were "
                                              "generated")
                    else:
                        utils.log.system.info("Constraints for fair time and "
                                              "alternate transitions were "
                                              "NOT generated")

                    if generate_time_var:
                        utils.log.system.info("Generated also time variable")

                else:
                    utils.log.system.info("(but no timed domain)")

            else:
                utils.log.system.info("Generated smv file has finite domain")

            # creates the BDD ordering file
            if not self.infinite_domain:
                self.ord_fname = nusmv.launch_smv_to_bdd_ordering(
                    self.smv_fname, kill_event=kill_event)
                utils.log.system.info("Generated ordering file '%s'",
                                      self.ord_fname)
            else:
                utils.log.system.info("Ordering file not generated as "
                                      "domain is not finite")

            changed = True

        self.last_smv_gen_params = gen_pars
        return (self.smv_fname, self.ord_fname, self.infinite_domain, changed)

    def log_metrics(self):
        if len(self.slim_files) > 0:
            self.compile_info.log_metrics()
            utils.log.metrics.info("Infinite domain for model: %s" %
                                   ("yes" if self.infinite_domain else "no"))
        else:
            utils.log.metrics.info("No SLIM models loaded")

    # ---------------------------------------------------------------
    #        Notifications from observable properties
    # ---------------------------------------------------------------
    def property_changed_signal_emit(self, model, args):
        if model != self:
            return  # interested only in itself

        # this is done to regenerate the smv lazily
        self.smv_fname = None
        self.ord_fname = None

        # searches and sets infinite_domain
        if self.root_impl:
            from slim_compiler.src.smv.walkers \
                import InfiniteDomainFinderWalker
            walker = InfiniteDomainFinderWalker()
            walker.walk(self.root_impl)
            self.infinite_domain = walker.has_found_infinite_data()

            type = self.root_impl.parent.get_component_type(
                self.root_impl.name.split(".")[0])

            port_count = (type.get_total_number_of_outgoing_data_ports() +
                          type.get_total_number_of_incoming_data_ports() +
                          type.get_total_number_of_outgoing_event_ports() +
                          type.get_total_number_of_incoming_event_ports())
            self.open_world = port_count > 0

        return

    pass  # end of class
# ----------------------------------------------------------------------
