# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================



"""
This module provide a way to manage the filters for the traceViewer module.
The filters are saved in xml format.
"""

import xml.dom.minidom as mdom
import xml, gtk, gobject
from gtkmvc import Model
import string
import re

class TraceFilter(Model):
        """
        This class represent a single filter. 
        It is able to load and manage a filter, starting from a filter dom object.
        If a file name will be passed, it will be parsed and the first filter tag 
        will be used as the base filter dom object.
        If there isn't any 'filter' tag in the file, a TypeError will be raised. 
        The data of the filter is avaiable by the functions:
        - get_filter_name()
        - get_filter()
        - get_range_iter()
        - get_ranges()
        - get_type_iter()
        - get_types()
        """
        # {'name':[u'a*','b*'], 'ranges':[(1,5), (20, -1)], 'types':[u'state',u'input']}   
        def __init__(self):
                Model.__init__(self)    
                self.__filter = {'names':[],'names_compiled':[],'ranges':[],'types':[]} #initialize filter container
                self.__filtername = ''
                
                return

        def __filter_text_nodes(self,root):
                       # c == child
                return [c for c in root.childNodes if c.nodeType != xml.dom.Node.TEXT_NODE]

        #utility GET
        def from_dom_xml(self,filter_dom_object):
                """
                load a filter, from a dom filter object, or an xml file
                """
                #if the argument is a string, I assume that is a path of an xml file
                #so I have to parse it and get the dom filter object by myself  
                if type( filter_dom_object ) is str or type( filter_dom_object ) is unicode:
                        filter_obj = mdom.parse( filter_dom_object ).getElementsByTagName('filter')
                        if filter_obj:
                                # I arbitrary choose the first filter tag
                                f = filter_obj[0]
                        else:
                                raise TypeError( "Invalid xml file!" )  
                else:   
                        f = filter_dom_object
        
                self.__filtername = filtername = f.getAttribute('name')
                # get the 3 tags (the filtering is useful in order to ignore the textnodes)
                try:                    
                        name,types,ranges = self.__filter_text_nodes(f) 
                except ValueError:
                        raise ValueError("too many child nodes in filter node. Malformed filter!")
                        
                assert len(name.childNodes) <= 1, "Invalid xml format! field name of filter "+str(filtername)+" has more than 1 child fields"
                if not len(name.childNodes):
                        self.__filter['names'] = []
                else:   
                    #after checks, create new filter
                        self.__filter['names'] = name.childNodes[0].nodeValue.split(',')

                #get types admitted
                for t in self.__filter_text_nodes(types):
                        assert len(t.childNodes) <= 1, "Invalid xml format! some field type of filter "+str(filtername)+" has more than 1 child fields" 
                        self.__filter['types'].append( t.childNodes[0].nodeValue )
                        pass
                
                #get ranges admitted
                for r in self.__filter_text_nodes(ranges):          #add a tuple with start,end range
                                                                            #the special value -1 means:
                                                                            #for start = from the beginning
                                                                            #for end = to the end
                        self.__filter['ranges'].append((int(r.getAttribute('start')),int(r.getAttribute('end'))))
                        pass    
                        
                return 
        

        def to_xml(self):
                """
                return a string like this:
                <filter name="filtro1" >
                <name>simplified_regex1,simplified_regex2</name>
                <types>
                <type>admitted_type1</type>
                <type>admitted_type2</type>
                </types>
                <ranges>
                # example of ranges, -1, for the start field means to the beginning
                # for the end field means until the end.
                <range start="-1" end="3" />
                <range start="10" end="-1" />
                </ranges>
                </filter>
                """
                ff = self
                xml = '\t'#at the last iteration only 0 tab will be written
                xml += '<filter name="'+str(ff.get_filter_name())+'" >\n\t\t'
                xml += '<name>'+str(','.join(ff.get_names()))+'</name>\n\t\t'
                xml += '<types>\n\t\t'

                for t in ff.get_type_iter():
                        xml += '\t'     #at the last iteration only two tabs will be written
                        xml += '<type>'+str(t)+'</type>\n\t\t'
                        pass

                xml += '</types>\n\t\t'
                xml += '<ranges>\n\t\t'

                for r in ff.get_range_iter():
                        xml += '\t'     #at the last iteration only two tabs will be written
                        xml += '<range start="'+str(r[0])+'" end="'+str(r[1])+'" />\n\t\t'
                        pass
                
                xml += '</ranges>\n\t'
                xml += '</filter>\n'
                
                return xml

        def get_filter_name(self):
                return self.__filtername
        
        def get_names(self):
                """ get admitted names """      
                ls = []
                ls.extend( self.__filter['names'] )
                return ls

        def get_range_iter(self):
                for r in self.__filter['ranges']:
                        yield r 
                        
        def get_ranges(self):
                ls = []
                ls.extend(self.__filter['ranges'])      
                return ls

        def get_types(self):
                ls = []
                ls.extend(self.__filter['types'])       
                return ls
        
        def get_type_iter(self):
                for t in self.__filter['types']:
                        yield t 

        #utility SET
        def set_names(self,new):
                """
                set the 'names' field of this filter.
		The names will be filtered with the python function 'strip'.
                @param: the new names (a string of comma separated names)
                @returns the list of invalid names if there are
		"""
		# building list of filters
                if type(new) is list:
			# strip the names
			# and avoid invalid filters
                        ls = [f for f in map( string.strip, new ) if f]     
		elif type(new) is str or type(new) is unicode:
			ls = [f for f in map( string.strip, new.split(',') ) if f]
                # invalid names
		elif type(new) is not str and type(new) is not unicode:
                        raise TypeError( "This "+str(new)+" is not a string" )
		else:
			raise TypeError( "Invalid type passed "+str(new) )

		self.__filter['names'] = []
		self.__filter['names_compiled'] = []

		errors = []
		# strip and avoid invalid filters
		for f in ls: 
			try:
				pattern = re.compile(f)
			except:
				errors.append(f)
				continue
			
			self.__filter['names'].append(f)
			self.__filter['names_compiled'].append(pattern)
			pass
				
                return errors
        
        def set_ranges(self,ranges):
                """
                @param: a list of tuple, each tuple represent a range
                        the values in the tuple must be integers
                        -1 is used to represent "until the end".
                """
                if type(ranges) is not list:
                        raise TypeError( "This "+str(ranges)+" is not a list" )
                
                try:
                        self.__filter['ranges'] = [(int(x),int(y)) for (x,y) in ranges]
                except (TypeError,ValueError):
                        raise TypeError("the parameter passed have to be a list of tuple, with 2 integers inside")      
                        pass    
                
                return

        def set_types(self,types):
                """
                @param: a list of strings, each string represent a type
                admitted type: input, combinatorial, state
                """
		
                if type(types) is list:
			st = map( string.strip , types )
                         #filter the non admitted types
                        self.__filter['types'] = [t for t in st if str(t) == "state" \
							  or str(t) == "input" \
							  or str(t) == "combinatorial"\
							  or str(t) == "error"]
                elif type(types) is str or type(types) is unicode:
                        st = map( string.strip, types.split(',') )
                        self.__filter['types'] = [t for t in st if str(t) == "state" \
							  or str(t) == "input" \
							  or str(t) == "combinatorial"\
							  or str(t) == "error"]
                else:
                        raise TypeError("The parameter have to be a list of strings, or a string of comma-separated values")

                return
        
        def set_filter_name(self,new):
                """
                change the name that identify a filter.
                """
                if type(new) is str or type(new) is unicode:
                        self.__filtername = new
                        return
                
                raise ValueError(str(new)+" is not a string")

        #utility CHECK
        def is_valid_step(self, step):
                """
                take an integer, positive value, and return
                true if the param is in the ranges of the filter,
                false otherwise.
                both start and end values are included.
                @param: must be integer
                """
                #if no ranges specified return true (each range)
                if not self.__filter['ranges']:
                        return True     

                result = False 
                #check if the step statisfy a range
                for r in self.__filter['ranges']:
                        start,end = r
                        #i'll sum 1 because steps start from 0, ranges from 1.
                        bs = step+1 >= start
                        if end != -1:
                                be = step+1 <= end
                        else:
                                be = True
                
                        result = result or (bs and be)
                        
                        pass
                
                return result
        
        def is_valid_type(self,t):
                """
                take a type, and return True if this type is admitted
                false otherwise.
                """
                if not self.__filter['types']:
                        return True
		
		return t in self.__filter['types']
        
        
        def is_valid_name(self,name):
                """
                take a name, and return True if this type is admitted
                false otherwise.
                """                
                def match(regex, name): # return yes if the matched string is the whole string
                        m = regex.match( name )
			if m:
				return m.group() == name
                        else:
                                return False

                #if no names specified return true (each name)
		if not self.__filter['names']:
                        return True     

                result = False
                for regex in self.__filter['names_compiled']:
                        result = result or match(regex, name)        
                        pass
        
                return result


class TraceFilters(Model): #I use this as model for the combobox
        """
        This class save, load and manage all the filters contained in the xml_file, 
        using the TraceFilter class, defined in this module.
        The syntactic definition of the xml is explained in the module documentation.
        @see: TraceFilter
        """
        def __init__(self):
                """
                @param: a file handle with the xml data
                """
                Model.__init__(self)  
                #this class contains a list with the names of all the filters,
                #automatically updated when self.__filters that is a list of TraceFilter,
                #is updated. This ListStore is used as model for the ComboBox in the view
                self.__ls = gtk.ListStore(gobject.TYPE_STRING) 
                #list of filters, the key is the name of the filter
                #the content is the values of the filter
                #like this:
                # >> print self.__filters
                # >> {'filter1':
                #     <TraceFilter instance>,
                #     'filterN':
                #     {...} 
                #    }
                self.__filters = [] #list of filters
                
                return


        def to_xml(self, ex=[]):
                """
                convert the data of the filters in this class in an xml representation
                with this syntax:
                <filters>
                <filter name="filtro2" >
                ...
                </filter>
                </filters>

                where simplified_regex,  mean a string with the special char '*' that means whatever string.
                The ', ' means OR.
		@param ex  is a list of filters that you don't want to include in the xml
                """
                if type(ex) is not list:
			ex = [ex]
			pass
                
		xml = u'<?xml version="1.0" encoding="UTF-8"?>\n'
                xml += '<filters>\n'
		for f in self.__filters:
			if f not in ex:
				xml += f.to_xml()
				pass
			pass

                xml += '</filters>\n'
                return xml

        def write_xml(self, f, ex=[]): 
                """
                write an xml file in the file specified,  
                it can be a file handle.
		@param ex is a list of filters that you don't want to save
                """
                if type(f) is str:
                        file(f, 'w').write( self.to_xml(ex) )
                else:
                        f.write( self.to_xml(ex) )        

                return

        #utility function GET
        def get_filter_names(self):
                return [x.get_filter_name() for x in self.__filters]    

        def get_filters(self):
                return self.__filters

        def get_ListStore(self):
                return self.__ls
                
        def filters_iter(self):
                for x in self.__filters:
                        yield x

        def get_filter(self, name, index=False):
                """
                Get a particular TraceFilter istance
                @param name: the name of the filter
                @param index: If true, an index in the list of filter will be returned
                              (it can be used as a path in function get_iter).
                NB: The filter name is different from the 'names' field of the filter
                """
                for i in xrange(len(self.__filters)):
                        f = self.__filters[i]
                        
                        if f.get_filter_name() == name:
                                if index:
                                        return i
                                else:
                                        return f

                raise KeyError("This filter was not found!")    

        #utility function SET
        def add_filters(self,lfilters):
                """
                add a list of filters.
                If the param is a single filter, not wrapped
                in a list, it works properly.
                @param: a list of filters
                """
                if type(lfilters) is not list:
                        if lfilters not in self.__filters: #does not add a filter twice
                                self.__filters.append(lfilters)
                                self.__ls.append( [lfilters.get_filter_name()] ) #keep ListStore syncronized
                else:
                        #remove existing filters
                        ls = [x for x in lfilters if x not in self.__filters]
                        self.__filters.extend(ls) #append each filter
                        for x in ls: 
                                self.__ls.append( [x.get_filter_name()] ) #keep ListStore syncronized
                               

        def clear(self):
                """
                delete all the filters
                """
                self.__filters = []
                self.__ls.clear()
                
        def from_xml(self, f):
                """
                This function add to the managed data the filters found in 
                the file passed by parameter. 
                The file can be a file handle,  
                or a file name.
                @param: the file name,  or the file handle
                Warning! The old filters with the same name of some new filters
                found in the file will be overwritten by the new filters.
                """
                dom = mdom.parse(f)
                filters = dom.getElementsByTagName("filter")
                #foreach filter in xml
                for f in filters:
                        tf = TraceFilter()
                        tf.from_dom_xml(f)      
                        self.__filters.append( tf )
                        self.__ls.append( [tf.get_filter_name()] )

                return

        def delete_filter(self, name):
                """
                delete an existing filter.
                @param: the filter name.
                
                NB: The filter name is different from the 'names' field of the filter
                """
                pos = 0
                found = False
                for i in xrange(len(self.__filters)):
                        f = self.__filters[i]
                        if f.get_filter_name() == name:
                                self.__filters.remove(f)        
                                found = True
                                pos = i
                                break
                        pass
                
                if not found:
                        raise KeyError("The filter "+str(name)+" does not exist")       
                
                #keep syncronized liststore with __filters
                i = self.__ls.get_iter(pos)
                self.__ls.remove(i)
                
                return f

                
