# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================


# The view for slim models and fault injection

import os.path
import gtk

from gtkmvc import View
import src.utils.globals

from src.views.activity import ActivityView, StartStopActivityView
from balloon import Balloon


class CorrectnessView (ActivityView):    
    
    def __init__(self):
        ActivityView.__init__(self, CorrectnessSubView())
        return

    pass # end of class
# ----------------------------------------------------------------------    


class CorrectnessSubView (View):
    """This is the view of the pure correctness, without
    properties"""

    glade = os.path.join(src.utils.globals.GLADE_DIR, "correctness.glade")
    top = "nb_correctness"

    def __init__(self):
        View.__init__(self)
        
        # sub views for handling model checking, simulation, deadlock checking
        # and zeno analysis
        self.sim = ModelSimulationView()
        self.dc = DeadlockCheckingView()
        self.mc = ModelCheckingView()
        self.zeno = ZenoAnalysisView()
        self.td = TimeDivergenceView()
        
        # connects the subviews
        nb = self['nb_correctness']
        pages = zip((self.sim.get_top_widget(), 
                     self.dc.get_top_widget(),
                     self.mc.get_top_widget(),
                     self.zeno.get_top_widget(),
                     self.td.get_top_widget(),), 
                    (gtk.Label(nb.get_tab_label_text(child)) 
                     for child in nb.get_children())
                    )

        for idx in range(nb.get_n_pages()): nb.remove_page(0)
        for child, lbl in pages: nb.append_page(child, lbl)
        return

    def enable_model_extension(self, flag):
        """Sets all widget to either allow or not the usage of
        model extension"""
        for sv in (self.mc, self.dc, self.sim, self.zeno, self.td): sv.enable_model_extension(flag)        
        return

    pass # end of class
# ----------------------------------------------------------------------    


class ModelCheckingView (StartStopActivityView):    
    glade = os.path.join(src.utils.globals.GLADE_DIR, "correctness.glade")
    top = "vbox_mc"
    
    def __init__(self):
        StartStopActivityView.__init__(self)

        # attaches result view
        self.mc_results = ResultsView()
        self.mc_results['SCVDesc'].show_all()
        self.mc_results['label_res_title'].show()
        
        self['results_mc'].add(self.mc_results.get_top_widget())

        self.enable_model_checking(False)
        return

    def enable_model_checking(self, flag):
        """Sets sensitivity all the widgets needed to perform model
        checking"""
        self['button_run_mc'].set_sensitive(flag)
        self['button_stop_mc'].set_sensitive(flag)
        return

    def enable_model_extension(self, flag):
        """Sets all widget to either allow or not the usage of
        model extension"""
        cb = self['cb_mc_extended_model']
        cb.set_sensitive(flag)
        if not flag and cb.get_active(): cb.set_active(False)
        return

    def enable_sat_options(self, flag):
        """Sets sensitivity of widgets related to SAT options."""
        self['table_sat'].set_sensitive(flag)
        return

    def show_mc_result(self, truth, desc, trace_view, fbarView, stock_id):
        """Show a result coming from model checking. 
        @param truth is astring holding the truth value of a property.
        @param desc is a markup string with description.
        @param trace_view can be None
        @param stock_id is the stock ID for the image of the dialog"""

        title = "<big><b>The property is <tt>%s</tt></b></big>" % truth
        self.mc_results.set_icon_stock(stock_id)
            
        self.mc_results.show_results(title, desc, 
                                     trace_view.get_top_widget() if trace_view else None)

        if trace_view and fbarView: 
            self.mc_results.set_filter_bar( fbarView )
        elif not trace_view:
            self.mc_results.unset_filter_bar()

        return  

    def show_exception(self, title, desc): self.mc_results.show_exception(title, desc)

    # implemention of virtual methods
    def get_start_button(self): return self['button_run_mc']
    def get_stop_button(self): return self['button_stop_mc']

    pass # end of class
# ----------------------------------------------------------------------       



class ModelSimulationView (StartStopActivityView):    
    glade = os.path.join(src.utils.globals.GLADE_DIR, "correctness.glade")
    top = "vbox_sim"
    GUIDED = 1
    RANDOM = 0
    
    def __init__(self):
        StartStopActivityView.__init__(self)
        # attaches result view
        self.sim_results = ResultsView()
        self['results_sim'].add(self.sim_results.get_top_widget())
        
        # what happens when the top page change 
        # callback for the gtk.Notebook of the ResultsView
        def fun(nb, page, page_num, user_data):
            simulation_view = user_data
                                                   # 1 is the index of the 'valid simulation to show' tab
            simulation_view['button_sim_jump'].set_sensitive(True if (page_num == 1) else False)
            
            simulation_view['button_sim_restart'].set_sensitive(True)
            
            if page_num > 2: # if we are switching to an error page
                # if we are in an error state, the only thing that the user
                # can do, is to reset the simulation and than start.
                # so we have to disable all the options.
                self.enable_controls(False)
                simulation_view['button_sim_restart'].set_sensitive(True)
            else:
                self.enable_controls(True)
                pass
            return

        self.sim_results.set_top_page_changed(fun, self) 

        self['cb_sim_type'].set_active(0)
        return

    def select_random(self):
        """Called to select the random simulation"""
        self['hbox_sim_len'].set_sensitive(True)
        return


    def enable_controls(self, flag):
        """Enables/disables all controls for the simulation"""
        self['cb_sim_type'].set_sensitive(flag)
        self['HBStartStop'].set_sensitive(flag)

        for child in self['hb_other'].get_children():
            child.set_sensitive(flag)
            pass

        return

    def show_deadlock(self, trace_view):
        self.sim_results.show_deadlock(trace_view)
        
        # enables only the restart button:
        self.enable_controls(False)        
        self['button_sim_restart'].set_sensitive(True)
        return

    def select_guided(self):
        """Called to select the guided imulation"""
        self['hbox_sim_len'].set_sensitive(False) 
        return


    def enable_model_extension(self, flag):
        """Sets all widget to either allow or not the usage of
        model extension"""
        cb = self['cb_sim_extended_model']
        cb.set_sensitive(flag)
        if not flag and cb.get_active(): cb.set_active(False)
        return

    def show_exception(self, title, desc): self.sim_results.show_exception(title, desc)

    def enable_options(self, flag): 
        """
        set sensitive or insensitive all the options except the
        buttons run and stop, but only if not showing an error
        condition.
        """
        if self.sim_results.get_active_page() <= 2:            
            for child in ("button_sim_restart", "button_sim_jump", 
                          "cb_sim_type", "hbox_sim_len"):
                self[child].set_sensitive(flag)
                pass            
            pass
        return

    def clear_info_label(self):
        self['lbl_sim_info'].set_markup("")
        return

    def set_info_label(self, message):
        """
        @param message: a pango markup string
        """
        self['lbl_sim_info'].set_markup(message)
        return

    def show_sim_result(self, desc, trace_view, trace_view_guided, fbarView):
        """Show a result coming from model simulation
        @param desc is a markup string with description.
        @param trace_view of the simulation (can be None)
        @param trace_view_guided, the trace view of the guided simulation (can be None)
        @param fbarView, the view of the filter bar, if you want to add it. None otherwise.
        """

        title = "<big><b>Simulation</b></big>"
        
        #show the random results
        self.sim_results.show_results(title, desc, 
                                      trace_view.get_top_widget(),
                                      where=0)

        #show the guided results
        self.sim_results.show_results(title, desc, 
                                      trace_view_guided.get_top_widget(), 
                                      where=1) #in the result view number 1
        
        if fbarView and (trace_view and trace_view_guided): 
            self.sim_results.set_filter_bar( fbarView )
        else: self.sim_results.unset_filter_bar()

        self.sim_results.set_icon_stock(gtk.STOCK_NO if trace_view is None else gtk.STOCK_YES)
        return        

    # implemention of virtual methods
    def get_start_button(self): return self['button_run_sim']
    def get_stop_button(self): return self['button_stop_sim']

    pass # end of class
# ----------------------------------------------------------------------       


class DeadlockCheckingView (StartStopActivityView):    
    glade = os.path.join(src.utils.globals.GLADE_DIR, "correctness.glade")
    top = "vbox_dc"
    
    def __init__(self):
        StartStopActivityView.__init__(self)
        self['notebook_dc'].set_show_tabs(False)
        return

    def show_not_avail(self): 
        self['notebook_dc'].set_current_page(3)
        self['button_run_dc'].set_sensitive(False)
        return

    def show_ready(self): 
        self['notebook_dc'].set_current_page(0)
        self['button_run_dc'].set_sensitive(True)
        return

    def is_ready(self): return self['notebook_dc'].get_current_page() == 0

    def show_running(self): 
        self['notebook_dc'].set_current_page(4)
        return

    def is_running(self): return self['notebook_dc'].get_current_page() == 4

    def show_yes(self): 
        self['notebook_dc'].set_current_page(1)
        return

    def show_no(self): 
        self['notebook_dc'].set_current_page(2)
        return

    def show_error(self): 
        self['notebook_dc'].set_current_page(5)
        return
    
    def enable_model_extension(self, flag):
        """Sets all widget to either allow or not the usage of
        model extension"""
        cb = self['cb_dc_extended_model']
        cb.set_sensitive(flag)
        if not flag and cb.get_active(): cb.set_active(False)
        return

    # implemention of virtual methods
    def get_start_button(self): return self['button_run_dc']
    def get_stop_button(self): return self['button_stop_dc']

    pass # end of class
# ----------------------------------------------------------------------       

# LB: the new class for ZENO Analysis View

class ZenoAnalysisView (StartStopActivityView):    
    glade = os.path.join(src.utils.globals.GLADE_DIR, "correctness.glade")
    top = "vbox_zeno"
    
    def __init__(self):
        StartStopActivityView.__init__(self)

        # create and attach result view 
        self.diag_resview = ResultsView()
        self.diag_resview['SCVDesc'].show_all()
        self.diag_resview['label_res_title'].show()
        self.diag_resview['button_detach'].hide()
        self['results_container_trace'].add(self.diag_resview.get_top_widget())
        
        # Traces Window
        w = gtk.Window()
        w.set_border_width(12)
        w.set_default_size(640, 480)
        self['win_zeno_detached'] = w
        
        # At first the detach_traces button is not showed
        self['button_zeno_detach_traces'].hide()
        
        # The list with the messages/results
        self.messages_list = []
        return
        

    # Implemention of virtual methods
    
    def get_start_button(self): return self['button_run_zeno_analysis']
    
    def get_stop_button(self): return self['button_stop_zeno_analysis']
    
    
    def enable_model_extension(self, flag):
        """Sets all widget to either allow or not the usage of model extension"""
        cb = self['cb_zeno_extended_model']
        cb.set_sensitive(flag)
        if not flag and cb.get_active(): cb.set_active(False)
        return
    
        
    def detach(self):
        """Detaches the trace viewer, raising it to its own window"""
        child = self['hbox_results']
        w = self['win_zeno_detached']
        self['button_zeno_detach_traces'].hide()
        child.reparent(w)
        w.show()
        return
        
        
    def attach(self):
        """Re-embeds the previlusly detached trace view into the main GUI"""
        w = self['win_zeno_detached']
        w.hide()
        child = w.get_child()
        self['button_zeno_detach_traces'].show()
        child.reparent(self['vpanel_zeno'])
        return
        
        
    def display_zeno_results(self, mode, steps, time, success, trace_view):
        """Redirects the results of a zeno analysis run to the right GUI widgets.
            @param   mode:  The mode to analyze.
	        @type    mode:  string
	        @param   steps:  The steps to analyze.
	        @type    steps:  string
	        @param   time:  The time to analyze.
	        @type    time:  string
	        @param   success:  The result message of zeno analysis
	        @type    success:  string
	        @param   trace_view:  The TraceView to be shown.
	        @type    trace_view:  TraceView
        """

        self['button_zeno_detach_traces'].hide()
        
        if success == "error":
            self['label_zeno_message'].set_markup('<b>Attention: No result was returned! Error?</b>')
            
        elif success == "selection":
            self['label_zeno_message'].set_markup('<b>Please select at least one mode row from the Treeview to run the Zeno Analysis!</b>')
            
        elif success == "unreachable":
            message = 'The state is UNREACHABLE'
            
        elif success == "unknown_not_found":
            message = 'Cannot find a zeno-path in %s steps.' % (steps)
            
        elif success == "unknown_not_proved":
            message = 'Cannot prove the absence of zeno paths up-to %s.' % (time) 

        elif success == "stop":
            self['label_zeno_message'].set_markup('<b>The analysis has been interrupted! Warning!</b>')

        # When a message/result is selected, the related trace and detach button are shown: (success == "change trace")
        # Else No results is shown and the detach button is hidden: (success == "no trace")
        
        elif success == "change trace":
            self.diag_resview.show_results('<b>Trace</b>', '', 
                                             trace_view.get_top_widget() if trace_view else None)
            
            self.diag_resview.set_icon_stock(gtk.STOCK_YES)            
            self['button_zeno_detach_traces'].show()
            
        elif success == "no trace": 
            self.diag_resview.show_results('<b>No Results to show.</b>', 'There are no results to show at the moment.', None)
            self.diag_resview.set_icon_stock(gtk.STOCK_DIALOG_WARNING)
         
        elif success == "non_zeno":
            message = 'The state is NON-ZENO'
            
        else:
            message = 'The state is ZENO'

        # Only the messages with success = "zeno, non-zeno, unreachable, find and prove" are appended in the TreeView Box.
        # The others are shown as a label, near the Zeno Options.
        if success not in ("error", "stop", "change trace", "no trace", "selection"):
            self.messages_list.append([mode, message])

        return
        
        
    def get_messages_list(self):
        """Gets the messages/results list of Zeno Analysis.
            @return   self.messages_list:  The messages/results list of Zeno Analysis
	        @rtype    self.messages_list:  list of strings
        """ 
        return self.messages_list
        
        
    def clear(self):
        """Clears the Zeno Analysis View."""
        self.messages_list = []
        self['label_zeno_message'].set_markup('')
        self.diag_resview.show_results('<b>No Results to show.</b>', 'There are no results to show at the moment.', None)
        self.diag_resview.set_icon_stock(gtk.STOCK_DIALOG_WARNING)
            
        return
        
        
    def clear_treeview(self):
        """Clears all columns of Treeview every time a change occured in SLIM model."""
        for col in self['tv_zeno_modes_results'].get_columns():
            self['tv_zeno_modes_results'].remove_column(col)
        return

    pass # end of class
# ----------------------------------------------------------------------   


# LB: the new class for Time Divergence View

class TimeDivergenceView (StartStopActivityView):    
    glade = os.path.join(src.utils.globals.GLADE_DIR, "correctness.glade")
    top = "vbox_td"
    
    def __init__(self):
        StartStopActivityView.__init__(self)

        # create and attach result view 
        self.diag_resview = ResultsView()
        self.diag_resview['SCVDesc'].show_all()
        self.diag_resview['label_res_title'].show()
        self.diag_resview['button_detach'].hide()
        self['results_container_trace'].add(self.diag_resview.get_top_widget())
        
        # Traces Window
        w = gtk.Window()
        w.set_border_width(12)
        w.set_default_size(640, 480)
        self['win_td_detached'] = w
        
        # At first the detach_traces button is not showed
        self['button_td_detach_traces'].hide()
        
        # The list with the messages/results
        self.messages_list = []
        return
        

    # Implemention of virtual methods
    
    def get_start_button(self): return self['button_run_td']
    
    
    def get_stop_button(self): return self['button_stop_td']
    
    
    def enable_model_extension(self, flag):
        """Sets all widget to either allow or not the usage of model extension"""
        cb = self['cb_td_extended_model']
        cb.set_sensitive(flag)
        if not flag and cb.get_active(): cb.set_active(False)
        return
    
        
    def detach(self):
        """Detaches the trace viewer, raising it to its own window"""
        child = self['hbox_results']
        w = self['win_td_detached']
        self['button_td_detach_traces'].hide()
        child.reparent(w)
        w.show()
        return
        
        
    def attach(self):
        """Re-embeds the previlusly detached trace view into the main GUI"""
        w = self['win_td_detached']
        w.hide()
        child = w.get_child()
        self['button_td_detach_traces'].show()
        child.reparent(self['vpanel_td'])
        return
        
        
    def display_td_results(self, clock, time, success, trace_view):
        """Redirects the results of a time divergence run to the right GUI widgets.
            @param   clock:  The time to analyze.
	        @type    clock:  string
            @param   time:  The time to analyze.
	        @type    time:  string
	        @param   success:  The result message of zeno analysis
	        @type    success:  string
	        @param   trace_view:  The TraceView to be shown.
	        @type    trace_view:  TraceView
        """
        
        self['button_td_detach_traces'].hide()
        
        if success == "error":
            self['label_td_message'].set_markup('<b>Attention: No result was returned! Error?</b>')
                
        elif success == "selection":
           self['label_td_message'].set_markup('<b>Please select at least one clock row from the Treeview to run the Time Divergence!</b>')
            
        elif success == "bounded":
            message = 'The clock %s is " BOUNDED " by the given bound.' % clock
            
        elif success == "unknown":
            message = 'The check on the clock divergence for %s is " UNKNOWN "' % clock

        elif success == "stop":
            self['label_td_message'].set_markup('<b>The analysis has been interrupted! Warning!</b>')

        # When a message/result is selected, the related trace and detach button are shown: (success == "change trace")
        # Else No results is shown and the detach button is hidden: (success == "no trace")
        
        elif success == "change trace":
            self.diag_resview.show_results('<b>Trace</b>', '', 
                                             trace_view.get_top_widget() if trace_view else None)
            
            self.diag_resview.set_icon_stock(gtk.STOCK_YES)            
            self['button_td_detach_traces'].show()
            
        elif success == "no trace": 
            self.diag_resview.show_results('<b>No Results to show.</b>', 'There are no results to show at the moment.', None)
            self.diag_resview.set_icon_stock(gtk.STOCK_DIALOG_WARNING)
         
        else:
            message = 'The clock %s is " UNBOUNDED "' % clock
            
        # Only the messages with success = "bounded, unbounded, unknown" are appended in the TreeView Box.
        # The others are shown as a label, near the Zeno Options.
        if success not in ("error", "stop", "change trace", "no trace", "selection"):
            self.messages_list.append([clock, message])

        return
        
        
    def get_messages_list(self):
        """Gets the messages/results list of Time Divergence.
            @return   self.messages_list:  The messages/results list of Time Divergence.
	        @rtype    self.messages_list:  list of strings
        """ 
        return self.messages_list
        
        
    def clear(self):
        """Clears the Time Divergence View."""
        self.messages_list = []
        self['label_td_message'].set_markup('')
        self.diag_resview.show_results('<b>No Results to show.</b>', 'There are no results to show at the moment.', None)
        self.diag_resview.set_icon_stock(gtk.STOCK_DIALOG_WARNING)
            
        return
        
        
    def clear_treeview(self):
        """Clears all columns of Treeview every time a change occured in SLIM model."""
        for col in self['tv_td_clocks_results'].get_columns():
            self['tv_td_clocks_results'].remove_column(col)
        return

    pass # end of class
# ----------------------------------------------------------------------       


class ResultsView (View):
    glade = os.path.join(src.utils.globals.GLADE_DIR, "correctness.glade")
    top = ("nb_results_top")
    no_result_to_show = 'There are no results to show at the moment'

    def __init__(self):
        View.__init__(self)
        
        # filters
        w = gtk.Window()
        w.set_border_width(12)
        w.set_default_size(650, 50)
        self['window_filters_detached'] = w
        
        # trace
        w = gtk.Window()
        w.set_border_width(12)
        w.set_default_size(640, 480)
        self['window_detached'] = w

        self['nb_results_top'].set_show_tabs(False)
        self['nb_results'].set_show_tabs(False)
        self['NBResults'].set_show_tabs(False)
        self.show_no_results()
        self.filter_bar = None

        # results boxes (actually three results are supported)
        self.results_map = { # this maps a number into the relative result widget
            0:'VBFirstRes',
            1:'VBSecondRes',
            2:'VbThirdRes'
            }

        return

    def set_filter_bar(self, fbarView ):
        # append the new filter bar
        if fbarView and not self.filter_bar:
            fbar = fbarView.get_top_widget()
            if fbar.get_parent(): #remove from his parent
                fbar.get_parent().remove(fbar)
                pass
            #here
            self['vbox_traces'].pack_end( fbar, expand=False, padding=6)
            fbarView['CBStoredFilters'].set_active(0) #set the first one the active filter
            self.filter_bar = fbar
        elif self.filter_bar and not self.filter_bar.get_parent():
            # it is possible that the filter bar was unset.. so we have to re-set it
            self['vbox_traces'].pack_end( self.filter_bar, expand=False, padding=6)
            pass
        
        if self.filter_bar:
            self.filter_bar.show() #if it is hided..
            pass
        return

    def unset_filter_bar(self):
        if not self.filter_bar:
            return
        if not self.filter_bar.get_parent():
            return
        
        self.filter_bar.get_parent().remove(self.filter_bar)
        self.filter_bar.hide()
        return

    def get_filter_bar(self):
        if self.filter_bar and self.filter_bar.get_parent(): #there is and it is somewhere
            return self.filter_bar
        elif self.filter_bar and not self.filter_bar.get_parent():
            return None #if it is not locate somewhere, it means that someone calls unset_filter after set_filter..
        else:
            return None

    def set_result_page_changed(self, fun, user_data=None):
        """
        connect the function passed to the signal "switch-page"
        of the notebook 'NBResults' of the results view.
        the signature of the function is assumed to be right.
        """
        self['NBResults'].connect( "switch-page", fun, user_data)
        return
    
    def set_top_page_changed(self, fun, user_data=None):
        """
        connect the function passed to the signal "switch-page"
        of the notebook 'nb_results_top' of the results view.
        the signature of the function is assumed to be right.
        """
        self['nb_results_top'].connect( "switch-page", fun, user_data)
        return

    def set_icon_stock(self, stock_id):
        """Sets the icon representing the result."""
        self['image_result'].set_from_stock(stock_id, gtk.ICON_SIZE_BUTTON)
        return

    def get_active_page(self):
        """
        returns the number (between 0 and 5) of the page currently showed
        i.e. the "no results" page, the "results" page (in which there is the results notebook)
        the "deadlock found" page, and so on.
        """
        return self['nb_results_top'].get_current_page()

    def get_active_results_page(self):
        """
        returns the number of the result currently showed (between 0 and 2)
        in the "results" page.
        Note that if the "results" page is not showed, then the active_results_page is not showed too.
        """
        return self['NBResults'].get_current_page()
    
    def show_results_page(self, n):
        """
        shows the selected result.
        The valid results are in range 0-2.
        @see param where in show_results
        """
        nbres = self['NBResults']
        assert n >= 0 and n <= 2, "Invalid range!, only three ranges are supported"
        nbres.set_current_page(n)
        return
    
    def show_page(self, n):
        """
        shows a panel that informa the user about something that happens
        """
        self.attach()
        assert n >= 0 and n <= 5
        self['nb_results_top'].set_current_page(n)
        return

    def show_deadlock(self, trace_view):
        top = self["algn_dead_trace"]
        if top.child: top.remove(top.child)
        top.add(trace_view.get_top_widget())
        self.show_page(5)
        return

    def show_default_error_results(self, desc=None):
        """
        shows a generic error (calling show_no_results()) if no other 
        errors panel is showed.
        """
        if self['nb_results_top'].get_current_page() > 1: # there is an error state currently showed
            return
        
        return self.show_no_results(desc)
    

    def show_no_results(self, desc=None):
        """shows a panel that inform the user that there is no result to show"""

        if desc:
            self['label_no_res_desc'].set_markup(desc)
        else:
            self['label_no_res_desc'].set_markup(self.no_result_to_show)
            pass
        
        self.attach()
        self['nb_results_top'].set_current_page(0)
        return    

    
    def change_results_title(self, title):
        self['label_res_title'].show()
        self['label_res_title'].set_markup(title)
        return
    
    def change_results_description(self, desc):
        self['SCVDesc'].show()
        self['label_desc'].set_markup(desc)
        return

    def show_results(self, title, desc, result, where=0):
        """Plugs in a result. 
        @param title is a markup string
        @param desc is a textual description (with markup)
        @param where an int between 0 and 2 that identify the container in which you would display the result
        (respectively the first, the second or the third one)
        @param result is a a widget to show, or None, if you want to show nothing
        """
        assert self['label_res_title']
        assert self['label_desc']
                
        hide_all = True
        if title:
            hide_all = False
            self['label_res_title'].show()
            self['label_res_title'].set_markup(title)
        else:
            self['label_res_title'].hide()
            pass
        
        if desc:
            hide_all = False
            self['SCVDesc'].show()
            self['label_desc'].set_markup(desc) 
        else:
            self['SCVDesc'].hide()
            pass

        assert self.results_map.has_key(where), \
            "Only three results are supported! the valid position numbers are in range 0-3"

        vbox = self[self.results_map[where]]
        # remove previous results
        for children in vbox.get_children():
            if children != self.filter_bar: #do not remove the filter bar
                vbox.remove(children)
                pass
            pass
        
        nbtop = self['nb_results_top']
        
        if result is not None:
            vbox.add(result)
            self['NBResults'].show()
            hide_all = False
        else:
            # if there is nothing to show
            self['NBResults'].hide()
            pass

        # if everithing is hided
        if hide_all:
            nbtop.hide()
        else:
            nbtop.show()
            pass

        # is it embedded?
        if nbtop.get_current_page() != 2: #if it is 'detached' leave it as is it.
            nbtop.set_current_page(1)
            pass

        return

    def show_exception(self, title, description):
        self['label_err_title'].set_markup("<big>%s</big>\nThe detailed description follows:" % title)
        tv = self['tv_err_desc']
        tv.get_buffer().set_text(description)
        
        # 4 is the index of the page for the nusmvcontrolledexception in the gtk.Notebook in the resultsView
        self['nb_results_top'].set_current_page(4)        
        return

    def add_to(self, where, what):
        if what.get_parent():
            what.get_parent().remove(what)
            pass
        where.pack_end(what, expand=False, padding=6)
        return
    
    def attach(self):
        """re-embeds the previlusly detached trace view into the
        main GUI"""
        nb = self['nb_results_top']
        if nb.get_current_page() != 2: # page is not detached...
            return

        w = self['window_detached']
        w.hide()
        child = w.get_child()
        w.remove(child)

        self['button_detach'].show()
        
        self['results_top_page'].add(child)
        nb.set_current_page(1) # shows the result page
        return

    def detach(self):
        """Detaches the trace viewer, raising it to its own
        window"""
        nb = self['nb_results_top']        
        if nb.get_current_page() == 2: #if it is already detached
            return
        
        nb.set_current_page(2) # shows "page is detached..."
        child = self['vbox_res']
        self['results_top_page'].remove(child)
        
        w = self['window_detached']
        w.add(child)
        self['button_detach'].hide()
        w.show()
        return
        
    pass # end of class
# ----------------------------------------------------------------------        
