# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================


# The view for slim models and fault injection

import os.path
import math
import gtk
from gtkmvc import View

import src.utils.globals
from src.views.activity import ActivityView, StartStopActivityView

from matplotlib.figure import Figure
from matplotlib.backends.backend_gtk import FigureCanvasGTK

class PerformabilityView (ActivityView):
    
    def __init__(self):
        ActivityView.__init__(self, PerformabilitySubView())
        return

    pass # end of class

class PerformabilitySubView (StartStopActivityView):
    """This is the view of the pure performability, without
    properties"""

    glade = os.path.join(src.utils.globals.GLADE_DIR, "performability.glade")
    top = "vbox_performability"

    def __init__(self):
        View.__init__(self)

        # sub views for handling model checking, simulation and deadlock checking
        self.imc = PerformabilityIMCView()
        self.sim = PerformabilitySimulationView()

        # connects the subviews
        nb = self['nb_performability']
        self.pages = zip((self.imc.get_top_widget(),
                     self.sim.get_top_widget(),),
                    (gtk.Label(nb.get_tab_label_text(child)) 
                     for child in nb.get_children())
                    )

        for idx in range(nb.get_n_pages()): nb.remove_page(0)
        for child, lbl in self.pages: nb.append_page(child, lbl)
        self.imc_visible = True
        return

    def display_ports_warning(self, show=False):
        if show:
            self['frame_warning'].show()
        else:
            self['frame_warning'].hide()

    def display_imc(self, show=True):
        if show == self.imc_visible:
            return

        nb = self['nb_performability']
        if show:
            nb.prepend_page(self.pages[0][0], self.pages[0][1])
        else:
            nb.remove_page(0)

        self.imc_visible = show


class PerformabilitySimulationView(StartStopActivityView):
    glade = os.path.join(src.utils.globals.GLADE_DIR, "performability.glade")
    top = "performability_sim"

    def __init__(self):
        StartStopActivityView.__init__(self)

    def get_start_button(self): 
        return self['button_run']
    
    def get_stop_button(self): 
        return self['button_stop']

    def display_prob(self, prob):
        
        self['label_result'].set_markup('[' + str(prob[0]*100.0) + '%, ' + str(prob[1]*100.0) + '%]')

    def reset_prob(self):
        self['label_result'].set_markup('Not yet determined')

class PerformabilityIMCView(StartStopActivityView):
    glade = os.path.join(src.utils.globals.GLADE_DIR, "performability.glade")
    top = "performability_top"
    figure = None
    axes = None
    canvas = None

    def __init__(self):
        StartStopActivityView.__init__(self)
        self.figure = Figure()
        self.axes = self.figure.add_subplot(111)
        self.canvas = FigureCanvasGTK(self.figure)
        self.tool_settings_sub = ToolSettingsSubView(self)
        graph_widget = self['vbox_graph']
        graph_widget.pack_start(self.canvas)
        graph_widget.reorder_child(self.canvas, 0) # reorder the probability entry widget
        hbox_prob = self['hbox_prob']
        graph_widget.reorder_child(hbox_prob, 1)
        tool_settings_sub_widget = self.tool_settings_sub.get_top_widget()
        self['vbox_sub_ftv'].pack_start(tool_settings_sub_widget)
        graph_widget.show_all()
        return
        
    def draw(self, xpoints, ypoints, ltime_bound=0, utime_bound=100):
        self.axes.clear()
        self.axes.set_autoscale_on(False)
        self.axes.set_ylabel("Probability")                                            
        self.axes.set_xlabel("Upper time bound")         
        self.axes.set_ylim(0, 1) # set y-axis range
        self.axes.set_xlim(ltime_bound, utime_bound) # set x-axis range
        self.axes.plot(xpoints, ypoints, 'b-')
        self.canvas.blit()
        self.canvas.draw_idle()
    
    def get_start_button(self): 
        return self['button_run']
    
    def get_stop_button(self): 
        return self['button_stop']
    
    def display_et(self, number):
        self['label1'].set_label("<b>Result</b>")
        self.canvas.set_visible(False)
        self['label3'].set_label("Estimated Time:")
        self['label_prob'].set_markup('<span color="red"><b>' + str(number) + ' </b></span>')
        
    def display_lra(self, number):
        self['label1'].set_label("<b>Result</b>")
        self.canvas.set_visible(False)
        self['label3'].set_label("Long Run Average:")
        self['label_prob'].set_markup('<span color="red"><b>' + str(number) + ' </b></span>')
    
    def reset_labels(self):
        self['label1'].set_label("<b>Cumulative Distribution Function</b>")
        self.canvas.set_visible(True)
        self['label3'].set_label("Probability:")
        self['label_prob'].set_markup('<span color="red"><b> Not determined yet </b></span>')
    
    def display_percentage(self, probability_min_max):
        def __round_percentage(number, digits):
            result = 0
            number = number * math.pow(10, digits)
            if number % 5 <= 5:
                result = math.floor(number / 10) / math.pow(10, (digits - 1))
            else:
                result = math.ceil(number / 10) / math.pow(10, (digits - 1))
            return result
        
        if isinstance(probability_min_max, tuple):
            min_percentage = __round_percentage(probability_min_max[0] * 100, 2)
            max_percentage = __round_percentage(probability_min_max[1] * 100, 2)
            prob_str = "(%s%% , %s%%)"%(min_percentage, max_percentage)
        else:
            prob_str = "%s%%"%__round_percentage(probability_min_max * 100, 2)
        
        self['label1'].set_label("<b>Cumulative Distribution Function</b>")
        self.canvas.set_visible(True)
        self['label3'].set_label("Probability:")
        self['label_prob'].set_markup('<span color="red"><b>' + prob_str + ' </b></span>')

    pass # end of class

class ToolSettingsSubView (StartStopActivityView):
    glade = os.path.join(src.utils.globals.GLADE_DIR, "tool_settings.glade")
    top = "vbox_tool_settings"
    
    def __init__(self, parent):
        StartStopActivityView.__init__(self)
        self.parent = parent
        #base = self[ToolSettingsSubView.top].get_style().bg
        self['label_tool_info'].set_markup("Select <b>one</b> tool to be used for analysis (or <i>auto</i>):")
        self['label_error_bound'].set_label("Error Bound:")

    def tint_error_base(self,colorstring):
        self['txtview_error_bound'].modify_base(gtk.STATE_NORMAL, gtk.gdk.color_parse(colorstring))

    def hide_error_bound(self, hide):
        if hide:
            self['hbox_error_bound'].hide()
        else:
            self['hbox_error_bound'].show()
