# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================


import gtk
import pango
import os.path
from gtkmvc import View
import src.utils.globals
from balloon import Balloon
import subprocess
from tfpg_viewer.src.utils.misc import which
from src.views.activity import ActivityView, StartStopActivityView
from gtk import TextBuffer
from src.views.correctness import ResultsView # result for FRA is similar to MC
import gtk
import gobject


class TfpgView (ActivityView):
    """Represents the View of TFPG."""

    glade = os.path.join(src.utils.globals.GLADE_DIR, "tfpg.glade")
    top = "hpaned_tfpg"
    
    def __init__(self):
        """Constructor Method"""
        ActivityView.__init__(self, TfpgSubView())
        
        # The SubViews of TfpgView
        self.tfpg_viewer_view = TfpgViewerView()
        self.tfpg_slim_assoc_view = TfpgSlimAssocView()
        self.tfpg_synthesis_view = TfpgSynthesisView()
        self.tfpg_behav_val_view = TfpgBehavValView()
        self.tfpg_effect_val_view = TfpgEffectValView()
        self.tfpg_errors_view = TfpgErrorsView()

        # Connects the sub-views into the notebook:
        nb = self['nb_tfpg']
        pages = zip((sv.get_top_widget() 
                     for sv in (self.tfpg_viewer_view, self.tfpg_slim_assoc_view, self.tfpg_synthesis_view, \
                                self.tfpg_behav_val_view, self.tfpg_effect_val_view, self.tfpg_errors_view)),
                    (gtk.Label(nb.get_tab_label_text(child)) 
                     for child in nb.get_children()))
        for idx in range(nb.get_n_pages()): nb.remove_page(0)
        for child, lbl in pages: nb.append_page(child, lbl)

        # Status/Progress Bar
        pb = gtk.ProgressBar()
        pb.set_pulse_step(0.1)
        self['statusbar'].pack_end(pb, expand=False, padding=6)
        self['progress'] = pb

        return


    def enable_progress_bar(self, flag):
        """Enable/disables the progress bar in the statusbar. This
        is used during background activities."""
        pb = self['progress']
        if flag: pb.show() 
        else: pb.hide()
        return

    def set_progress_bar_text(self, text):
        """Sets a text in the progress bar"""
        pb = self['progress']
        pb.set_text(text)
        return

    
    def clear_subviews(self):
        """Clears all the elements of the Subviews."""
        self.tfpg_viewer_view.clear()
        self.tfpg_slim_assoc_view.clear()
        self.tfpg_synthesis_view.clear()
        self.tfpg_behav_val_view.clear()
        self.tfpg_effect_val_view.clear()
        self.tfpg_errors_view.clear()
        return


    def create_subviews_style_tag(self):
        """Creates all subviews tags for styling output"""
        self.tfpg_synthesis_view.create_style_tag()
        self.tfpg_behav_val_view.create_style_tag()
        self.tfpg_effect_val_view.create_style_tag()
        self.tfpg_errors_view.create_style_tag()
        return
        
        
    def enable_buttons(self):
        """Enables all buttons."""
        
        self['button_edit_tfpg'].set_sensitive(True)
        self['button_remove_tfpg'].set_sensitive(True)
        self['button_export_tfpg'].set_sensitive(True)
        
        self.tfpg_viewer_view['button_zoom_in'].set_sensitive(True)
        self.tfpg_viewer_view['button_zoom_out'].set_sensitive(True)
        self.tfpg_viewer_view['button_caption'].set_sensitive(True)
        
        self.tfpg_slim_assoc_view['button_save_associations'].set_sensitive(True)
        self.tfpg_slim_assoc_view['button_load_associations'].set_sensitive(True)

        return
        
        
    def disable_buttons(self):
        """Disables all buttons."""
        
        self['button_edit_tfpg'].set_sensitive(False)
        self['button_remove_tfpg'].set_sensitive(False)
        self['button_export_tfpg'].set_sensitive(False)
        
        self.tfpg_viewer_view['button_zoom_in'].set_sensitive(False)
        self.tfpg_viewer_view['button_zoom_out'].set_sensitive(False)
        self.tfpg_viewer_view['button_caption'].set_sensitive(False)
        
        self.tfpg_slim_assoc_view['button_save_associations'].set_sensitive(False)
        self.tfpg_slim_assoc_view['button_load_associations'].set_sensitive(False)
        return
        
        
    def ui_run_text_editor(self, filename, from_scratch=False):
        """Runs a text editor, if found, and returns its exit code.
            If the parameter from_scratch is True, that means that the user is
            creating a new structural tfpg from scratch, i.e. we must show an
            empty file.
            @param    filename: The name of the file to view in the text editor.
            @type     filename: string
            @param    from_scratch: The value specifying whether it is a new file.
            @type     from_scratch: boolean

	        @return   proc.returncode: Exit status of the subprocess.
	        @rtype    proc.returncode: int
        """

        for program in "gedit emacs".split():
            editor_name = which(program)
            if editor_name: break

        if editor_name is None: raise Exception("Valid textual editor not found in the system")

        if from_scratch:
            with open(filename, "w") as f:
                f.write("\nNAME empty_tfpg\n\n")
                f.write("-- Definition of nodes\n")
                f.write("-- FM fm_node             -- A Failure Mode\n")
                f.write("-- AND and_node MONITORED -- An And Discrepancy\n")
                f.write("-- OR or_node             -- An Or Discrepancy\n\n")
                f.write("-- Definition of the edge modes\n")
                f.write("-- MODES a, b\n\n")
                f.write("-- Definition of edges\n")
                f.write("-- EDGE edge1 src_node dst_node min_time max_time (a,b)")

        proc = subprocess.Popen([editor_name, filename])
        proc.wait()

        return proc.returncode


    def ui_show_message_dialog(self, msg_type, msg_text, secondary_text=None, parent_name=None,
                                          buttons=gtk.BUTTONS_OK_CANCEL):
        """Function which shows a simple message dialog with the edit error.
            @param    msg_type: The type of the message to be shown.
            @type     msg_type: {info, error, warning}
            @param    msg_text: The message text to be shown.
            @type     msg_text: string
            @param    secondary_text: The (optional) secondary message text to be shown.
            @type:    secondary_text: string (default = None)
            @param    parent_name: The name of the container of the dialog.
            @type:    parent_name: string (default = None)
            @param    buttons: The name buttons of the dialog.
            @type:    buttons: gtk.Button

            @return   response:  The response of the gtk.MessageDialog (cancel, quit, ok...).
            @rtype    response:  gtk.RESPONSE_OK, gtk.RESPONSE_CANCEL
        """

        msg_types = { 'info'    : gtk.MESSAGE_INFO,
                      'error'   : gtk.MESSAGE_ERROR, 
                      'warning' : gtk.MESSAGE_WARNING
                      }

        if msg_type not in msg_types.keys(): raise Exception("Invalid message type")
        else:
            if parent_name != None: parent = self[parent_name]
            else: parent = None
            message_dialog = gtk.MessageDialog(parent, 0, msg_type, 
                                                 buttons, msg_text)
            message_dialog.set_markup(str(msg_text))
            message_dialog.show()
            if secondary_text != None:
                message_dialog.format_secondary_text(secondary_text)
                
            response = message_dialog.run()

        message_dialog.destroy()
	
        return response

        
    def get_active_page(self):
        """Gets the number of the current active page.
            self['nb_tfpg'].get_current_page() = 0 -> Tfpg View
            self['nb_tfpg'].get_current_page() = 1 -> Tfpg Model
            self['nb_tfpg'].get_current_page() = 2 -> Tfpg Synthesis
            self['nb_tfpg'].get_current_page() = 3 -> Tfpg Behavioural Validation
            self['nb_tfpg'].get_current_page() = 4 -> Tfpg Effectiveness Validation
            self['nb_tfpg'].get_current_page() = 5 -> Tfpg Errors
            
            @return   self['nb_tfpg'].get_current_page():  The number of the current page.
	        @rtype    self['nb_tfpg'].get_current_page():  int
        """ 
        return self['nb_tfpg'].get_current_page()


    def show_page(self, n):
        """Shows the page:
            n = 0 -> Tfpg View
            n = 1 -> Tfpg Model
            n = 2 -> Tfpg Synthesis
            n = 3 -> Tfpg Behavioural Validation
            n = 4 -> Tfpg Effectiveness Validation
            n = 5 -> Tfpg Errors
            
            @param   n:  The number of the page.
	        @type    n:  int
        """
        assert n >= 0 and n <= 5
        self['nb_tfpg'].set_current_page(n)
        return


    def cell_renderer_set_text(self, cell_renderer, text):
        """Sets the text of the cell_renderers.
            @param   cell_renderer:  The cell_renderer of the columns.
            @type    cell_renderer:  gtk.CellRendererText
            @param   text:  The text to be filled in the cell_renderer of the columns.
            @type    text:  string
	    
        """   
        cell_renderer.set_property("background", "white")     
        if text is None or text == "":
            cell_renderer.set_property('text', "")
        elif text == "None":
            cell_renderer.set_property('text', "none")
        else:
            cell_renderer.set_property('text', text)
            
        return

        
pass # end of TfpgView class
# ----------------------------------------------------------------------    


class TfpgSubView (View):
    """This is the subview of StartStopActivity (synthesis, behavioural and effectiveness validation)"""

    glade = os.path.join(src.utils.globals.GLADE_DIR, "tfpg.glade")
    top = "nb_tfpg"

    def __init__(self):
        """Constructor Method"""
        View.__init__(self)  

        # Sub views StartActivityViews
        self.tfpg_synthesis_view = TfpgSynthesisView()
        self.tfpg_behav_val_view = TfpgBehavValView()
        self.tfpg_effect_val_view = TfpgEffectValView()

        return
        
pass # end of TfpgSubView class
# ----------------------------------------------------------------------


class TfpgViewerView (View):    
    """Represents the TfpgViewer SubView."""
    
    glade = os.path.join(src.utils.globals.GLADE_DIR, "tfpg.glade")
    top = "vbox_viewer"
    
    def __init__(self):
        """Constructor Method"""
        View.__init__(self)      
        return
             
    def clear(self):
        """Clears the Tfpg Graph and Info."""    
        self.clear_tfpg_graph()
        self.clear_tfpg_toolbar()
        self.clear_tfpg_info()
        return
        
    
    def clear_tfpg_graph(self):
        """Clears the Tfpg Graph."""       
        for graph in self['vbox_view'].get_children():
            self['vbox_view'].remove(graph)
        return
        
        
    def clear_tfpg_toolbar(self):
        """Clears the Tfpg Toolbar."""        
        for toolbar in self['vbox_toolbar_viewer'].get_children():
            self['vbox_toolbar_viewer'].remove(toolbar)    
        return
            
            
    def clear_tfpg_info(self):
        """Clears the Tfpg Info."""            
        for info in self['vbox_info'].get_children():
            self['vbox_info'].remove(info)    
        return
        
pass # end of TfpgViewerView class
# ---------------------------------------------------------------------- 


class TfpgSlimAssocView (View):    
    """Represents the TfpgSlimAssoc SubView."""
    
    glade = os.path.join(src.utils.globals.GLADE_DIR, "tfpg.glade")
    top = "vbox_slim_assoc"
    
    def __init__(self):
        """Constructor Method"""
        View.__init__(self) 
        return
        
        
    def clear(self):
        """Clears the Tfpg Slim Associations."""    
          
        for col in self['tv_failure_modes'].get_columns():
            self['tv_failure_modes'].remove_column(col)

        for col in self['tv_monitored_discr'].get_columns():
            self['tv_monitored_discr'].remove_column(col)

        for col in self['tv_not_monitored_discr'].get_columns():
            self['tv_not_monitored_discr'].remove_column(col)
            
        for col in self['tv_tfpg_modes'].get_columns():
            self['tv_tfpg_modes'].remove_column(col)
            
        return
        
    
pass # end of TfpgSlimAssocView class
# ----------------------------------------------------------------------  


class TfpgSynthesisView (StartStopActivityView):    
    """Represents the TfpgSynthesis SubView."""
    
    glade = os.path.join(src.utils.globals.GLADE_DIR, "tfpg.glade")
    top = "vbox_synthesis"
    
    def __init__(self):
        """Constructor Method"""
        StartStopActivityView.__init__(self)

        self.enable_synthesis(False)

        self.synthesis_buf = TextBuffer()
        self.create_style_tag()
        return
        

    def clear(self):
        """Clears the Tfpg Synthesis View."""
        self.synthesis_buf.delete(*self.synthesis_buf.get_bounds())
        return

    def clear_columns(self):
        """Clears the Tfpg Synthesis Slim Associations Columns."""
        
        for col in self['tv_failure_modes_in_synth'].get_columns():
            self['tv_failure_modes_in_synth'].remove_column(col)

        for col in self['tv_monitored_discr_in_synth'].get_columns():
            self['tv_monitored_discr_in_synth'].remove_column(col)

        for col in self['tv_not_monitored_discr_in_synth'].get_columns():
            self['tv_not_monitored_discr_in_synth'].remove_column(col)
            
        for col in self['tv_tfpg_modes_in_synth'].get_columns():
            self['tv_tfpg_modes_in_synth'].remove_column(col)
            
        return

    def create_style_tag(self):
        """Creates synthesis tag for styling output"""
        self.synthesis_buf.create_tag("synthesis", family="Monospace",
                                       left_margin=12, foreground="black")
        return
        
        
    def get_start_button(self): 
        """Gets the run button of the synthesis.
            @return   self['button_run_synthesis']:  The run button of the synthesis.
	        @rtype    self['button_run_synthesis']:  gtk.Button
        """ 
        return self['button_run_synthesis']


    
    def get_stop_button(self): 
        """Gets the stop button of the synthesis.
            @return   self['button_stop_synthesis']:  The stop button of the synthesis.
	        @rtype    self['button_stop_synthesis']:  gtk.Button
        """ 
        return self['button_stop_synthesis']


    def enable_synthesis(self, flag):
        """Enables all widgets for carrying out synthesis accordingly to the given flag.
           @param   flag:  The flag that enables the button self['button_run_synthesis'].
	   @type    flag:  boolean
        """ 
        self['button_run_synthesis'].set_sensitive(flag)
        if flag: msg = "You can run Tfpg Synthesis"
        else: msg = "Create/Load a correct SLIM Associations file to run the Synthesis."
        self['label_synthesis_msg_view'].set_markup(msg)        
        return

        
    
pass # end of TfpgSynthesisView class
# ----------------------------------------------------------------------  

    
class TfpgBehavValView (StartStopActivityView):    
    """Represents the TfpgBehavValView SubView."""
    
    glade = os.path.join(src.utils.globals.GLADE_DIR, "tfpg.glade")
    top = "vbox_behav_val"
    
    def __init__(self):
        """Constructor Method"""
        StartStopActivityView.__init__(self)
        
        self.enable_behav_val(False)
        
        # Create and attach result view of system trace
        self.diag_resview_system = ResultsView()
        self.diag_resview_system['SCVDesc'].show_all()
        self.diag_resview_system['label_res_title'].show()
        self.diag_resview_system['button_detach'].hide()
        self['results_container_system'].add(self.diag_resview_system.get_top_widget())
        
        # Create and attach result view of tfpg trace
        self.diag_resview_tfpg = ResultsView()
        self.diag_resview_tfpg['SCVDesc'].show_all()
        self.diag_resview_tfpg['label_res_title'].show()
        self.diag_resview_tfpg['button_detach'].hide()
        self['results_container_tfpg'].add(self.diag_resview_tfpg.get_top_widget())   
        
        # Traces Window
        w = gtk.Window()
        w.set_border_width(12)
        w.set_default_size(640, 480)
        self['win_beh_val_detached'] = w
        
        # At first the detach_traces button is not showed
        self['button_beh_val_detach_traces'].hide()
                     
        return


    def display_behav_val_results(self, success, cex_description, trace_view_system, trace_view_tfpg):
        """Redirects the results of a tfpg behavioural validation run to the right GUI widgets.
	        @param   success:  The result message of tfpg effectiveness validation (Error, True or False)
	        @type    success:  string
	        @param   trace_view_system:  The 1st TraceView to be shown.
	        @type    trace_view_system:  TraceView
	        @param   trace_view_tfpg:  The 2nd TraceView to be shown.
	        @type    trace_view_tfpg:  TraceView
        """
         
        self['button_beh_val_detach_traces'].hide()
               
        if success == "error":
            self['label_beh_val_message'].set_markup('<b>Attention: No result was returned! Error?</b>')
                
        elif success == "stop":
            self['label_beh_val_message'].set_markup('<b>The analysis has been interrupted! Warning!</b>')

        elif success == "bounded_completeness":
            self['label_beh_val_message'].set_markup('<b>The TFPG is complete with respect to the model. No counterexample was found within the given bound.</b>')
                
        elif success == "complete":
            self['label_beh_val_message'].set_markup('<b>The TFPG is complete with respect to the model.</b>')
            
        elif success == "no associations":
            self['label_beh_val_message'].set_markup('<b>Error: it is not possible to run the TFPG Behavioural Validation if the TFPG SLIM associations, related to the loaded TFPG, are not loaded or completed without errors.</b>')
            
        else:
            self['label_beh_val_message'].set_markup('<b>The TFPG is incomplete with respect to the model. A counterexample is shown (%s).</b>' % cex_description)
            
            self.diag_resview_system.show_results('<b>System Trace</b>', 'Real system behavior.', 
                                             trace_view_system.get_top_widget() if trace_view_system else None)
            
            self.diag_resview_system.set_icon_stock(gtk.STOCK_YES)
            self.diag_resview_tfpg.show_results('<b>TFPG Trace</b>', 'TFPG variables interpreted on system trace.', 
                                             trace_view_tfpg.get_top_widget() if trace_view_tfpg else None)
            self.diag_resview_tfpg.set_icon_stock(gtk.STOCK_NO)
            
            self['button_beh_val_detach_traces'].show()

        return


    def enable_behav_val(self, flag):
        """Enables all widgets for carrying out fra accordingly to the given flag.
            @param   flag:  The flag that enables the button self['button_run_behav_val'].
	        @type    flag:  boolean
        """ 
        self['button_run_behav_val'].set_sensitive(flag)
        if flag: msg = "You can run Tfpg Behavioural Validation"
        else: msg = "Select one tfpg from the list to run Tfpg Behavioural Validation"
        self['label_behav_validated_view'].set_markup(msg)        
        return
        
        
    def clear(self):
        """Clears the Tfpg Behavioural Validation View."""
        self['label_beh_val_message'].set_markup('')
        self.diag_resview_system.show_results('<b>No Results to show.</b>', 'There are no results to show at the moment.', None)
        self.diag_resview_system.set_icon_stock(gtk.STOCK_DIALOG_WARNING)
        self.diag_resview_tfpg.show_results('<b>No Results to show.</b>', 'There are no results to show at the moment.', None)
        self.diag_resview_tfpg.set_icon_stock(gtk.STOCK_DIALOG_WARNING)
        return


    def get_start_button(self): 
        """Gets the run button of the behavioural validation.
            @return   self['button_run_behav_val']:  The run button of the behavioural validation.
	        @rtype    self['button_run_behav_val']:  gtk.Button
        """ 
        return self['button_run_behav_val']

    
    def get_stop_button(self): 
        """Gets the stop button of the behavioural validation.
            @return   self['button_stop_behav_val']:  The stop button of the behavioural validation.
	        @rtype    self['button_stop_behav_val']:  gtk.Button
        """ 
        return self['button_stop_behav_val']
        
        
    def detach(self):
        """Detaches the trace viewer, raising it to its own window"""
        child = self['hbox_results_2']
        self['vbox_b_val'].remove(child)
        
        w = self['win_beh_val_detached']
        w.add(child)
        self['button_beh_val_detach_traces'].hide()
        w.show()
        return
        
        
    def attach(self):
        """Re-embeds the previlusly detached trace view into the main GUI"""
        w = self['win_beh_val_detached']
        w.hide()
        child = w.get_child()
        w.remove(child)

        self['button_beh_val_detach_traces'].show()
        
        self['vbox_b_val'].add(child)
        return


    def show_warning_message(self, msg):

        w = gtk.Window()
        w.set_border_width(12)
        w.set_default_size(240, 180)
        w.set_position(gtk.WIN_POS_CENTER)
        vbox = gtk.VBox()
        label = gtk.Label(msg)
        #label.set_markup(msg)
        button = gtk.Button("OK")
        button.set_size_request(40, 40)
        button.connect('clicked', w.destroy())
        vbox.add(label)
        # vbox.pack_start(button, expand=False, fill=False)
        w.add(vbox)
        w.show_all()

        return
        
    
pass # end of TfpgBehavValView class
# ---------------------------------------------------------------------- 


class TfpgEffectValView (StartStopActivityView):    
    """Represents the TfpgEffectValView SubView."""
    
    glade = os.path.join(src.utils.globals.GLADE_DIR, "tfpg.glade")
    top = "vbox_effect_val"
    
    def __init__(self):
        """Constructor Method"""
        StartStopActivityView.__init__(self)
        
        self.enable_effect_val(False)
        
        # create and attach result view A
        self.diag_resview_A = ResultsView()
        self.diag_resview_A['SCVDesc'].show_all()
        self.diag_resview_A['label_res_title'].show()
        self.diag_resview_A['button_detach'].hide()
        self['results_container_A'].add(self.diag_resview_A.get_top_widget())
        
        # create and attach result view B
        self.diag_resview_B = ResultsView()
        self.diag_resview_B['SCVDesc'].show_all()
        self.diag_resview_B['label_res_title'].show()
        self.diag_resview_B['button_detach'].hide()
        self['results_container_B'].add(self.diag_resview_B.get_top_widget())
        
        # It is the list of results/messages of the TFPG Effectiveness Validation
        self.messages_list = gtk.ListStore(gobject.TYPE_STRING)
        
        # Traces Window
        w = gtk.Window()
        w.set_border_width(12)
        w.set_default_size(640, 480)
        self['win_eff_val_detached'] = w
        
        # At first the detach_traces button is not showed
        self['button_eff_val_detach_traces'].hide()
        
        return


    def display_effect_val_results(self, system_mode, failure_mode, success, trace_view_A, trace_view_B):
        """Redirects the results of a tfpg effectiveness validation run to the right GUI widgets.
            @param   system_mode:  The system mode to diagnose.
	        @type    system_mode:  string
	        @param   failure_mode:  The failure mode to diagnose.
	        @type    failure_mode:  string
	        @param   success:  The result message of tfpg effectiveness validation (Error, True or False)
	        @type    success:  string
	        @param   trace_view_A:  The 1st TraceView to be shown.
	        @type    trace_view_A:  TraceView
	        @param   trace_view_B:  The 2nd TraceView to be shown.
	        @type    trace_view_B:  TraceView
        """
                
        self['button_eff_val_detach_traces'].hide()
        
        if success == "error":
            self['label_eff_val_message'].set_markup('<b>Attention: No result was returned! Error?</b>')
                
        elif success == "diagnosability_unknown":
            message = 'No counterexample for failure mode " %s " of system mode " %s " was found within the given bound'  % (failure_mode, system_mode) 
                
        elif success == "diagnosable":
            message = 'The failure mode " %s " in system mode " %s " was found to be " diagnosable "' % (failure_mode, system_mode)
            
        elif success == "alarms not enabled":
            self['label_eff_val_message'].set_markup('<b>No Failure Modes found! Warning!</b>')
            
        elif success == "stop":
            self['label_eff_val_message'].set_markup('<b>The analysis has been interrupted! Warning!</b>')

        # When a message/result is selected, the related trace and detach button are shown: (success == "change trace")
        # Else No results is shown and the detach button is hidden: (success == "no trace")
        
        elif success == "change trace":
            self.diag_resview_A.show_results('<b>Trace A</b>', 'Compliant with diagnosability property.', 
                                             trace_view_A.get_top_widget() if trace_view_A else None)
            
            self.diag_resview_A.set_icon_stock(gtk.STOCK_YES)
            self.diag_resview_B.show_results('<b>Trace B</b>', 'Not compliant with diagnosability property.', 
                                             trace_view_B.get_top_widget() if trace_view_B else None)
            self.diag_resview_B.set_icon_stock(gtk.STOCK_NO)
            
            self['button_eff_val_detach_traces'].show()
            
        elif success == "no trace": 
            self.diag_resview_A.show_results('<b>No Results to show.</b>', 'There are no results to show at the moment.', None)
            self.diag_resview_A.set_icon_stock(gtk.STOCK_DIALOG_WARNING)
            self.diag_resview_B.show_results('<b>No Results to show.</b>', 'There are no results to show at the moment.', None)
            self.diag_resview_B.set_icon_stock(gtk.STOCK_DIALOG_WARNING)
        
        else:
            message = 'The failure mode " %s " in system mode " %s " was found to be " not diagnosable "' % (failure_mode, system_mode)

        # Only the messages with success = "diagnosable, not diagnosable and diagnosability_unknown" are appended in the TreeView Box.
        # The others are shown as a label, near the Model CHecking Options.
        if success not in ("error", "alarms not enabled", "stop", "change trace", "no trace"):
            self.messages_list.append([message])

        return


    def get_messages_list(self):
        """Gets the messages/results list of TFPG Effectiveness Validation.
            @return   self.messages_list:  The messages/results list of TFPG Effectiveness Validation.
	        @rtype    self.messages_list:  list of strings
        """ 
        return self.messages_list
        
        
        
    def enable_effect_val(self, flag):
        """Enables all widgets for carrying out fra accordingly to the given flag.
            @param   flag:  The flag that enables the button self['button_run_effect_val'].
	        @type    flag:  boolean
        """ 
        self['button_run_effect_val'].set_sensitive(flag)
        if flag: msg = "You can run Tfpg Effectiveness Validation"
        else: msg = "Select one tfpg from the list to run Tfpg Effectiveness Validation"
        self['label_effect_validated_view'].set_markup(msg)        
        return
        
            
    def get_start_button(self): 
        """Gets the run button of the effectiveness validation.
            @return   self['button_run_effect_val']:  The run button of the effectiveness validation.
	        @rtype    self['button_run_effect_val']:  gtk.Button
        """ 
        return self['button_run_effect_val']

    
    def get_stop_button(self): 
        """Gets the stop button of the effectiveness validation.
            @return   self['button_stop_effect_val']:  The stop button of the effectiveness validation.
	        @rtype    self['button_stop_effect_val']:  gtk.Button
        """ 
        return self['button_stop_effect_val']
        
          
    def clear(self):
        """Clears the Tfpg Effectiveness Validation View."""
        self.messages_list.clear()
        self['label_eff_val_message'].set_markup('')
        self.diag_resview_A.show_results('<b>No Results to show.</b>', 'There are no results to show at the moment.', None)
        self.diag_resview_A.set_icon_stock(gtk.STOCK_DIALOG_WARNING)
        self.diag_resview_B.show_results('<b>No Results to show.</b>', 'There are no results to show at the moment.', None)
        self.diag_resview_B.set_icon_stock(gtk.STOCK_DIALOG_WARNING)
        return
        
        
    def detach(self):
        """Detaches the trace viewer, raising it to its own window"""
        child = self['hbox_results']
        self['vbox_e_val'].remove(child)
        
        w = self['win_eff_val_detached']
        w.add(child)
        self['button_eff_val_detach_traces'].hide()
        w.show()
        return
        
        
    def attach(self):
        """Re-embeds the previlusly detached trace view into the main GUI"""
        w = self['win_eff_val_detached']
        w.hide()
        child = w.get_child()
        w.remove(child)

        self['button_eff_val_detach_traces'].show()
        
        self['vbox_e_val'].add(child)
        return
        

    ################################################################################################
    ######## CELL_RENDERERS of PHASES, OP_MODES, CONFIGURATIONS AND COMBINATIONS FUNCTION ########## 
    ################################################################################################
    
    def cell_renderer_set_text(self, cell_renderer, text):
        """Sets the text of the cell_renderers.
            @param   cell_renderer:  The cell_renderer of the columns.
            @type    cell_renderer:  gtk.CellRendererText
            @param   text:  The text to be filled in the cell_renderer of the columns.
            @type    text:  string
        """        
        if text is None or text == "":
            cell_renderer.set_property('text', "")
        elif text == "None":
            cell_renderer.set_property('text', "none")
        else:
            cell_renderer.set_property('text', text)
            
        return

    
pass # end of TfpgEffectValView class
# ---------------------------------------------------------------------- 


class TfpgErrorsView (View):    
    """Represents the TfpgErrors SubView."""
    
    glade = os.path.join(src.utils.globals.GLADE_DIR, "tfpg.glade")
    top = "vbox_errors"
    
    def __init__(self):
        """Constructor Method"""
        View.__init__(self) 
        self.errors_buf = TextBuffer()
        self.create_style_tag()
        return


    def clear(self):
        """Clears the Tfpg Errors View."""
        self.errors_buf.delete(*self.errors_buf.get_bounds())
        return


    def create_style_tag(self):
        """Creates Errors tag for styling output"""
        buf = self.errors_buf
        buf.create_tag("spacer", pixels_above_lines=12)
        buf.create_tag("result_ok", family="Monospace",
                       left_margin=12, foreground="dark green",
                       weight=pango.WEIGHT_BOLD,
                       pixels_below_lines=2)
        buf.create_tag("errors", 
                       weight=pango.WEIGHT_BOLD,
                       background="gray90",
                       foreground="darkred")
        buf.create_tag("normal", family="Sans", foreground="black")
        return
        
    
pass # end of TfpgErrorsView class
# ----------------------------------------------------------------------  
