# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by GMV Aerospace and Defence  
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================

import gtk
import gobject

from hasdel_tools.src.errors import InvalidNameError, \
    NameAlreadyExistsError, FewAssociatedOpModesError, \
    OpModeNotExistError, ExpressionNotCorrectError
from hasdel_tools.src.mission_specification.phases_op_modes import PhaseOpModeListStore

import hasdel_tools.src.functions as functions


def search_configuration(configurations_list, configuration_id):
    """Searches a configuration in the configuration_list with the key configuration_id.
           @param   configurations_list:  The list of Configurations.
	   @type    configurations_list:  list
           @param   configuration_id:  The id of the Configuration.
	   @type    configuration_id:  string
           
           @return   conf:  The found Configuration or None.
	   @rtype    conf:  Configuration or None
    """
    for conf in configurations_list:
        if conf.get_configuration_id() == configuration_id:
            return conf
    return None

pass # end of search_configuration
# -----------------------------------------------------------------------


def load_configurations_list(slim_model, src, dest, existing_names, opm_list):
    """Loads the list of Configurations.
           @param   slim_model:  The slim model.
	   @type    slim_model:  object
           @param   src:  The src list of Configurations.
	   @type    src:  list
           @param   dest:  The dest list of Configurations.
	   @type    dest:  list
           @param   existing_names:  The check of existing names
	   @type    existing_names:  string
           @param   opm_list:  The list of OpModes
	   @type    opm_list:  list
           

           @return   error_list:  The list of errors.
	   @rtype    error_list:  list
           @return   num_added:  The number of configurations added to the dest list.
	   @rtype    num_added:  int
           @return   num_contains:  The number of configurations in the dest list.
	   @rtype    num_contains:  int 
    """

    num_added = 0
    num_contains = 0
    error_list = []

    for configuration in src:
        try:
            configuration_id = configuration.configuration_id
            op_modes_list = configuration.op_modes_list
            spacecraft = configuration.spacecraft

            num_contains += 1
            if (configuration_id.lower() == "none") or (configuration_id == ""):
                raise InvalidNameError(configuration_id)
            elif configuration_id in existing_names:
                raise NameAlreadyExistsError(configuration_id)
            else:
                associated_op_modes_list = PhaseOpModeListStore()

                if len(op_modes_list) == 0:
                    raise FewAssociatedOpModesError(configuration_id)
                
                for op_mode in op_modes_list:
                    op_mode_name = op_mode.name
                    for opm in opm_list:
                        if opm[0].name == op_mode_name:
                            associated_op_modes_list.append(opm)
                            break
                    else:
                        raise OpModeNotExistError(op_mode_name)
                else:
                    error_lst = functions.get_errors_in_atomic_proposition_with_templates(spacecraft, slim_model)[0]

                    if not error_lst:
                        dest.append([Configuration(configuration_id,
                                                   spacecraft,
                                                   associated_op_modes_list)])
                        existing_names.append(configuration_id)
                        num_added += 1
                    else:
                        raise ExpressionNotCorrectError(spacecraft)

                    pass # end of internal for else

                pass # end of name if

            pass # end of try

        except Exception, e:
            error_list.append(e)

        pass # end of external for

    return error_list, num_added, num_contains


pass # end of load_configurations_list
# -----------------------------------------------------------------------


class ConfigurationListStore(gtk.ListStore):
    """Represents a list of Configurations"""

    def __init__(self):
        """Constructor Method"""
        gtk.ListStore.__init__(self, gobject.TYPE_PYOBJECT)
        return


    def sort(self):
        """Sorts the configurations of the list"""
        sup = len(self) - 1
        while sup > 0:
            prev = self.get_value(self[0].iter, 0)
            prev_row = self[0]
            for i, row in enumerate(self):
                curr = self.get_value(row.iter, 0)
                if i > 0:
                    if curr.get_configuration_id() < prev.get_configuration_id():
                        tmp = prev
                        self.set_value(prev_row.iter, 0, curr)
                        self.set_value(row.iter, 0, tmp)
                    else:
                        prev = curr
                prev_row = row
                if i > sup:
                    break
                pass # end of for
            sup += -1
        return
                

pass # end of ConfigurationListStore class
# ------------------------------------------------------------------------------

class Configuration():
    """Represents a Configuration"""

    def __init__(self, configuration_id, spacecraft, associated_op_modes):
        """Constructor Method"""
        self.configuration_id = configuration_id
        self.spacecraft = spacecraft
        self.associated_op_modes = associated_op_modes
        return


    def get_configuration_id(self):
        """Gets the id of the Configuration.
           @return   self.configuration_id:  The id of the Configuration.
	   @rtype    self.configuration_id:  string
        """
        return self.configuration_id


    def set_configuration_id(self, configuration_id):
        """Sets the id of the Configuration.
           @param   configuration_id:  The id of the Configuration.
	   @type    configuration_id:  string
        """
        self.configuration_id = configuration_id
        return


    def get_spacecraft(self):
        """Gets the spacecraft of the Configuration.
           @return   self.spacecraft:  The spacecraft of the Configuration.
	   @rtype    self.spacecraft:  string
        """
        return self.spacecraft


    def set_spacecraft(self, spacecraft):
        """Sets the spacecraft of the Configuration.
           @param   spacecraft:  The spacecraft of the Configuration.
	   @type    spacecraft:  string
        """
        self.spacecraft = spacecraft
        return


    def get_associated_op_modes(self):
        """Gets the list of the op_modes associated to the Configuration.
           @return   self.associated_op_modes:  The list op_modes associated to the Configuration.
	   @rtype    self.associated_op_modes:  list
        """
        return self.associated_op_modes


    def get_associated_op_modes_str(self):
        """Gets the list of the op_modes associated to the Configuration in string format.
           @return   self.associated_op_modes:  The list op_modes associated to the Configuration in string format.
	   @rtype    self.associated_op_modes:  list
        """
        associated_opm = ""
        for op_mode in self.associated_op_modes:
            associated_opm += str(op_mode[0]) + "\n"
        return associated_opm[:-1]

pass # end of Configuration class
# ------------------------------------------------------------------------------
