# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================

import sys
import os
import os.path

sys.path.append('../resources/antlr/antlr-python-runtime-3.1.1/')
sys.path.append('../resources/pychecker-0.8.18/')

import antlr3
import antlr3.tree
import logging
import functools

from slim_compiler.src.generated.SlimLexer import SlimLexer
from slim_compiler.src.generated.SlimWalkerIdentCheck import SlimWalkerIdentCheck
from slim_compiler.src.errorManagement.errors import *
from slim_compiler.src.errorManagement.exceptions import *
from slim_compiler.src.dataStructure.compile import CompileInformation
from slim_compiler.src.dataStructure.nominalbehaviour import *
from slim_compiler.src.dataStructure.derivedparser import DerivedParser
from slim_compiler.src.dataStructure.visitors.expressionVisitor import TypeVisitor
import slim_compiler.src.dataStructure.expressions



def parse_identifier(name):
    qualifications = []
    while name.find(".") != -1:
        qualifications.append(name[0:name.find(".")])
        name = name[name.find(".") + 1:]
        
    return (qualifications, name)


def follow_identifier_path(root_impl, name, error_collector):
    (qualifications, data_field) = parse_identifier(name)
    impl = root_impl
    
    error = False
    for qualification in qualifications:
        if not error:
            if qualification in impl.non_data_subcomponents.keys():
                subcomp = impl.non_data_subcomponents[qualification]
                impl = subcomp.get_referenced_component_implementation()                
            else:
                error_collector.add(WrongQualificationError(name, qualification))
                error = True

    if not error:
        if data_field == "mode" or data_field == "error" or \
                data_field in impl.data_subcomponents.keys():
            error_collector.add(ErrorNotObservableAlarmIdentError(name))
            return None
        else:
            _type = impl.get_referenced_component_type()
            if _type is not None:
                if data_field in _type.in_data_ports.keys():
                    error_collector.add(ErrorNotObservableAlarmIdentError(name))
                    return None
                elif data_field in _type.out_data_ports.keys():
                    port = _type.out_data_ports[data_field]
                    if port.observable or port.alarm:
                        return port
                    else:
                        error_collector.add(ErrorNotObservableAlarmIdentError(name))
                        return None
                else:
                    error_collector.add(WrongQualificationError(name,
                                                                data_field))
                    return None
    else:
        return None


def get_type(data_type):
    if isinstance(data_type, Int):
        return NodeType.INT
    elif isinstance(data_type, Bool):
        return NodeType.BOOL
    elif isinstance(data_type, Real):
        return NodeType.REAL
    elif isinstance(data_type, Clock):
        return NodeType.CLOCK
    elif isinstance(data_type, Continuous):
        return NodeType.CONTINUOUS
    elif isinstance(data_type, Enum):
        return NodeType.ENUM_FIELD
    elif isinstance(data_type, IntRange):
        return NodeType.INT_RANGE
    else:
        assert False



def annotate_definition_via_observable_tree(root_impl, root_node, error_collector):
    if isinstance(root_node, IdentifierTreeNode):
        if root_node.name.find(":") != -1:
            (qualifier, name) = root_node.name.split(":")
            root_node.name = name
            if qualifier == "enum":
                root_node.node_type = NodeType.ENUM_VALUE
            else:
                root_node.node_type = NodeType.ERROR
        else:
            data_elem = follow_identifier_path(root_impl, root_node.name,
                                               error_collector)
            if data_elem is not None:
                if isinstance(data_elem, ComponentImplementation):
                    root_node.node_type = NodeType.ERROR
                elif isinstance(data_elem.data_type, Enum):
                    root_node.enum_values = data_elem.data_type.values
                    root_node.node_type = NodeType.ENUM_FIELD
                else:
                    root_node.node_type = get_type(data_elem.data_type)
            else:
                root_node.node_type = NodeType.ERROR
    elif isinstance(root_node, NumberTreeNode):
        pass
    elif isinstance(root_node, BooleanTreeNode):
        pass
    elif isinstance(root_node, UnaryOperatorTreeNode):
        annotate_definition_via_observable_tree(root_impl, root_node.subexpression,
                                                error_collector)
    elif isinstance(root_node, BinaryOperatorTreeNode):
        annotate_definition_via_observable_tree(root_impl, root_node.left,
                                                error_collector)
        annotate_definition_via_observable_tree(root_impl, root_node.right,
                                                error_collector)
    elif isinstance(root_node, CaseExpressionTreeNode):
        pass
    else:
        assert False
    

def check_definition_via_observable_tree(definition_via_observable_tree, root_impl):
    error_collector = ErrorCollector()
    annotate_definition_via_observable_tree(root_impl, definition_via_observable_tree, error_collector)
    cinfo = CompileInformation()
    cinfo.error_collector = error_collector
    TypeVisitor(cinfo).visit(definition_via_observable_tree)
    if definition_via_observable_tree.node_type != NodeType.BOOL:
        raise InvalidExpressionException(error_collector)
    else:
        # expression was correct
        pass
