# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by GMV Aerospace and Defence  
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================

import os.path
import copy
from lxml import etree

from hasdel_tools.src.errors import ConfigurationAlreadyExistsError
from hasdel_tools.src.xml.opmodes import OpModesDocument, \
    OpMode, _OP_MODES_LIST_TAG, _OP_MODE_TAG

# We import the conversion function used by the property manager 
from property_management.src.api import get_atomic_proposition_in_smv_format as get_smv_format


# MG: We use an additional parameter to indicate format (SLIM, SMV)
#     and the root component that might be useful during conversion.
def create_configurations_document(configurations_list, format="SLIM", slim_root=None): 
    """Creates the Configurations Document.
           @param   configurations_list:  The list of configurations.
	   @type    configurations_list:  list
           @param   format:  The format of file: SLIM
	   @type    format:  string
           @param   slim_root:  The slim root.
	   @type    slim_root:  object
          
           @return   configurations_document:  The document of configurations.
	   @rtype    configurations_document:  object
    """
  
    configurations_document = ConfigurationsDocument.create()

    for configuration in configurations_list:
        element = configurations_document.get_by_configuration_id(configuration[0].get_configuration_id())

        if element is None:
            op_modes_list_element = OpModesDocument.create()

            for op_mode in configuration[0].get_associated_op_modes():
                op_mode_element = op_modes_list_element.get_by_name(str(op_mode[0]))

                if op_mode_element is None:
                    new_op_mode_element = OpMode.create(str(op_mode[0]))
                    op_modes_list_element.add_op_mode(new_op_mode_element)
                    op_mode_element = op_modes_list_element.get_by_name(str(op_mode[0]))
                    pass

                assert op_mode_element is not None

            configuration_expression = configuration[0].get_spacecraft()

            if format=="SMV": # Convert to SMV if needed
                configuration_expression = get_smv_format(configuration_expression, slim_root)

            new_element = Configuration.create(
                configuration[0].get_configuration_id(), 
                op_modes_list_element, 
                 configuration_expression)

            configurations_document.add_configuration(new_element)
            element = configurations_document.get_by_configuration_id(
                configuration[0].get_configuration_id())
            pass
            
        assert element is not None

    return configurations_document

pass # end of create_configurations_document
# ----------------------------------------------------------------------------  
 

class ConfigurationsDocument(etree.ElementBase):
    """Represents the Configurations Document"""

    @staticmethod
    def create():
        """Creates the Configurations Document.     
           @return   list of configurations:  The document of configurations.
	   @rtype    list of configurations:  object
        """
        return _parser.makeelement("{" + _XML_NAMESPACE_URI + "}" +
                                   _CONFIGURATIONS_LIST_TAG)


    def get_by_configuration_id(self, configuration_id):
        """Gets the configuration from the configuration_id.  
           @param   configuration_id:  The configuration_id of the configuration.
	   @type    configuration_id:  string       
          
           @return   configuration:  The configuration found.
	   @rtype    configuration:  Configuration
        """
        for configuration in self:
            if configuration.configuration_id == configuration_id:
                return configuration
        return None


    def add_configuration(self, configuration):
        """Adds the configuration to the configuration document.
           @param   configuration:  The Configuration.
	   @type    configuration:  Configuration
        """
        if self.get_by_configuration_id(configuration.configuration_id):
            raise ConfigurationAlreadyExistsError(configuration.configuration_id)
        else:
            self.append(copy.deepcopy(configuration))
        return


    def __str__(self):
        """Converts the elements in string format"""
        return etree.tostring(self, xml_declaration=True, encoding="UTF-8",
                              pretty_print=True)

pass # end of ConfigurationsDocument class
# ----------------------------------------------------------------------------


class Configuration(etree.ElementBase):
    """Represents the Configuration"""

    @staticmethod
    def create(configuration_id, op_modes_list, spacecraft):
        """Creates the Edge.
           @param   configuration_id:  The configuration_id of the Configuration.
	   @type    configuration_id:  string
           @param   op_modes_list:  The op modes of the Configuration.
	   @type    op_modes_list:  list
           @param   spacecraft:  The spacecraft of the Configuration.
	   @type    spacecraft:  string
         
           @return   configuration:  The document of Configuration.
	   @rtype    configuration:  object
        """

        configuration = _parser.makeelement(
            "{" + _XML_NAMESPACE_URI + "}" + _CONFIGURATION_TAG)

        configuration_id_element = _parser.makeelement(
            "{" + _XML_NAMESPACE_URI + "}" + _CONFIGURATION_ID_TAG)
        configuration_id_element.text = configuration_id
        configuration.append(configuration_id_element)

        configuration.append(op_modes_list)

        spacecraft_element = _parser.makeelement(
            "{" + _XML_NAMESPACE_URI + "}" + _SPACECRAFT_TAG)
        spacecraft_element.text = spacecraft
        configuration.append(spacecraft_element)
        
        return configuration

    configuration_id = property(
        lambda self: self[0].text.strip() if self[0].text is not None else "",
        doc="The id of the Configuration.\n" + 
            "@type: string")

    op_modes_list = property(
        lambda self: self[1],
        doc="The associated Op-Modes of the Configuration.\n" +
            "@type: OpModesDocument")

    spacecraft = property(
        lambda self: self[2].text.strip() if self[2].text is not None else "",
        doc="The allowed spacecraft configuration of the Configuration.\n" +
            "@type: string")

pass # end of Configuration class
# ----------------------------------------------------------------------------


# ----------------------------------------------------------------------------
# Note: The following code will be executed once when this module is imported.
# ----------------------------------------------------------------------------

# Define some constants which are used in this module.
# The XML namespace that is used for configurations documents:
_XML_NAMESPACE_URI = ""
# The filename of the XML Schema document that defines the structure of valid
# configurations documents.
_XSD_FILENAME = os.path.join(os.path.dirname(__file__),
                             "xsd", "configurations.xsd")

# Define constants for the XML tag names that are used in configurations
# documents.
_CONFIGURATIONS_LIST_TAG = "configurationsList"
_CONFIGURATION_TAG = "configuration"
_CONFIGURATION_ID_TAG = "configurationId"
_SPACECRAFT_TAG = "spacecraft"

# Initialize the parser that is used to parse and to create configurations
# specification documents.
# XML comments must be discarded. Otherwise the parser would yield data
# structures cluttered with unwanted elements.
# A lookup scheme is used in order to automatically map certain XML elements
# to instances of the adequate classes which are defined in this module.
_parser = etree.XMLParser(remove_comments=True)
_lookup = etree.ElementNamespaceClassLookup()
_parser.set_element_class_lookup(_lookup)
# Associate the various XML element tags with the corresponding classes. 
# Note that all elements are in the same namespace.
_namespace = _lookup.get_namespace(_XML_NAMESPACE_URI)
_namespace[_CONFIGURATIONS_LIST_TAG] = ConfigurationsDocument
_namespace[_CONFIGURATION_TAG] = Configuration
_namespace[_OP_MODES_LIST_TAG] = OpModesDocument
_namespace[_OP_MODE_TAG] = OpMode

# Parse the XML Schema document that defines the structure of valid configurations
# documents. This is needed for validation.
_configurations_schema_tree = etree.parse(_XSD_FILENAME)
_configurations_schema = etree.XMLSchema(_configurations_schema_tree)

