# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by GMV Aerospace and Defence  
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================

import os.path
import copy
from lxml import etree
from hasdel_tools.src.errors import PhaseAlreadyExistsError


def create_phases_document(phases_list):
    """Creates the Phases Document.
           @param   phases_list:  The list of phases.
	   @type    phases_list:  list
          
           @return   phases_document:  The document of phases.
	   @rtype    phases_document:  object
    """
    phases_document = PhasesDocument.create()

    for phase in phases_list:
        element = phases_document.get_by_name(str(phase[0]))

        if element is None:
            new_element = Phase.create(str(phase[0]))
            phases_document.add_phase(new_element)
            element = phases_document.get_by_name(str(phase[0]))
            pass

        assert element is not None

    return phases_document

pass # end of create_op_modes_document
# ----------------------------------------------------------------------------  
 

class PhasesDocument(etree.ElementBase):
    """Represents the Phases Document"""

    @staticmethod
    def create():
        """Creates the list of Phases.     
           @return   list of Phases:  The document of Phases list.
	   @rtype    list of Phases:  object
        """
        return _parser.makeelement("{" + _XML_NAMESPACE_URI + "}" +
                                   _PHASES_LIST_TAG)


    def get_by_name(self, name):
        """Gets the phase from the name.  
           @param   name:  The name of the op mode.
	   @type    name:  string       
          
           @return   phase:  The phase found.
	   @rtype    phase:  Phase
        """
        for phase in self:
            if phase.name == name:
                return phase
        return None


    def add_phase(self, phase):
        """Adds the phase to the phases list.
           @param   phase:  The Phase.
	   @type    phase:  Phase
        """
        if self.get_by_name(phase.name):
            raise PhaseAlreadyExistsError(phase.name)
        else:
            self.append(copy.deepcopy(phase))
        return

    def __str__(self):
        """Converts the elements in string format"""
        return etree.tostring(self, xml_declaration=True, encoding="UTF-8",
                              pretty_print=True)

pass # end of PhasesDocument class
# ----------------------------------------------------------------------------


class Phase(etree.ElementBase):
    """Represents the Phase"""

    @staticmethod
    def create(name):
        """Creates the Phase.
           @param   name:  The name of the Phase.
	   @type    name:  string         
          
           @return   phase:  The document of Phase.
	   @rtype    phase:  object
        """
        phase = _parser.makeelement(
            "{" + _XML_NAMESPACE_URI + "}" + _PHASE_TAG)
        name_element = _parser.makeelement(
            "{" + _XML_NAMESPACE_URI + "}" + _NAME_TAG)
        name_element.text = name
        phase.append(name_element)
        return phase

    name = property(
        lambda self: self[0].text.strip() if self[0].text is not None else "",
        doc="The name of the Phase.\n" + 
            "@type: string")

pass # end of Phase class
# ----------------------------------------------------------------------------


# ----------------------------------------------------------------------------
# Note: The following code will be executed once when this module is imported.
# ----------------------------------------------------------------------------

# Define some constants which are used in this module.
# The XML namespace that is used for phases documents:
_XML_NAMESPACE_URI = ""
# The filename of the XML Schema document that defines the structure of valid
# phases documents.
_XSD_FILENAME = os.path.join(os.path.dirname(__file__),
                             "xsd", "phases.xsd")

# Define constants for the XML tag names that are used in phases
# documents.
_PHASES_LIST_TAG = "phasesList"
_PHASE_TAG = "phase"
_NAME_TAG = "name"

# Initialize the parser that is used to parse and to create phases
# specification documents.
# XML comments must be discarded. Otherwise the parser would yield data
# structures cluttered with unwanted elements.
# A lookup scheme is used in order to automatically map certain XML elements
# to instances of the adequate classes which are defined in this module.
_parser = etree.XMLParser(remove_comments=True)
_lookup = etree.ElementNamespaceClassLookup()
_parser.set_element_class_lookup(_lookup)
# Associate the various XML element tags with the corresponding classes. 
# Note that all elements are in the same namespace.
_namespace = _lookup.get_namespace(_XML_NAMESPACE_URI)
_namespace[_PHASES_LIST_TAG] = PhasesDocument
_namespace[_PHASE_TAG] = Phase

# Parse the XML Schema document that defines the structure of valid phases
# documents. This is needed for validation.
_phases_schema_tree = etree.parse(_XSD_FILENAME)
_phases_schema = etree.XMLSchema(_phases_schema_tree)

