# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================

import os
import copy
from lxml import etree
from slim_compiler.src.model_extension import errors


def create_associations_document_from_list(associations):
    """Creates the associations document for the given list of associations.
       This function is used when we want to save the associations of the tfpg
       synthesis tab, i.e. we do not have a tfpg associated to them

       @param associations: the list of associations to be converted
       @type  associations: list

       @return AssociationsDocument: the document of associations
       @rtype  AssociationsDocument: object
    """

    failureModes_ls = [(k, (ok, expr, t)) for (k, (ok, expr, t)) in associations.items() if t == "FailureMode"]
    monitoredDiscr_ls = [(k, (ok, expr, t)) for (k, (ok, expr, t)) in associations.items() if t == "MonitoredDiscr"]
    unmonitoredDiscr_ls = [(k, (ok, expr, t)) for (k, (ok, expr, t)) in associations.items() if t == "UnmonitoredDiscr"]
    tfpgModes_ls = [(k, (ok, expr, t)) for (k, (ok, expr, t)) in associations.items() if t == "TfpgMode"]
    
    return AssociationsDocument.create(
        create_failureModes_associations_list(failureModes_ls),
        create_monitoredDiscr_associations_list(monitoredDiscr_ls),
        create_unmonitoredDiscr_associations_list(unmonitoredDiscr_ls),
        create_tfpgModes_associations_list(tfpgModes_ls))

pass


def create_associations_document(model, tfpg_name):
    """Creates the association document for the given tfpg model.
       @param model: the tfpg_model
       @type  model: object
       @param tfpg_name: the name of the tfpg
       @type  tfpg_name: string

       @return AssociationsDocument: the document of associations
       @rtype  AssociationsDocuemnt: object
    """
    
    # LB: the creation of Slim Associations via TFPG Synthesis TAB (it is not necessary create/load a TFPG)

    associations = model.get_slim_associations(tfpg_name)

    failureModes_ls = [(k, (ok, expr, t)) for (k, (ok, expr, t)) in associations.items() if t == "FailureMode"]
    monitoredDiscr_ls = [(k, (ok, expr, t)) for (k, (ok, expr, t)) in associations.items() if t == "MonitoredDiscr"]
    unmonitoredDiscr_ls = [(k, (ok, expr, t)) for (k, (ok, expr, t)) in associations.items() if t == "UnmonitoredDiscr"]
    tfpgModes_ls = [(k, (ok, expr, t)) for (k, (ok, expr, t)) in associations.items() if t == "TfpgMode"]
    
    return AssociationsDocument.create(
        create_failureModes_associations_list(failureModes_ls),
        create_monitoredDiscr_associations_list(monitoredDiscr_ls),
        create_unmonitoredDiscr_associations_list(unmonitoredDiscr_ls),
        create_tfpgModes_associations_list(tfpgModes_ls))

pass #end of create_associations_document
# ----------------------------------------------------------------------------  


def create_failureModes_associations_list(associations):
    """Creates the associations list.
       @param associations: the associations with items
       @type  associations: list

       @return assoc_ls: the list of associations
       @rtype  assoc_ls: AssociationsList
    """
    assoc_ls = FailureModesAssocList.create()

    for (k, (ok, expr, t)) in associations:
        new_assoc = Association.create(k, expr)
        assoc_ls.add_association(new_assoc)

    return assoc_ls

pass #end of create_failureModes_associations_list
# ---------------------------------------------------------------------------- 


def create_monitoredDiscr_associations_list(associations):
    """Creates the associations list.
       @param associations: the associations with items
       @type  associations: list

       @return assoc_ls: the list of associations
       @rtype  assoc_ls: AssociationsList
    """
    assoc_ls = MonitoredDiscrepanciesAssocList.create()

    for (k, (ok, expr, t)) in associations:
        new_assoc = Association.create(k, expr)
        assoc_ls.add_association(new_assoc)

    return assoc_ls

pass #end of create_monitoredDiscr_associations_list
# ---------------------------------------------------------------------------- 


def create_unmonitoredDiscr_associations_list(associations):
    """Creates the associations list.
       @param associations: the associations with items
       @type  associations: list

       @return assoc_ls: the list of associations
       @rtype  assoc_ls: AssociationsList
    """
    assoc_ls = UnmonitoredDiscrepanciesAssocList.create()

    for (k, (ok, expr, t)) in associations:
        new_assoc = Association.create(k, expr)
        assoc_ls.add_association(new_assoc)

    return assoc_ls

pass #end of create_unmonitoredDiscr_associations_list
# ---------------------------------------------------------------------------- 


def create_tfpgModes_associations_list(associations):
    """Creates the associations list.
       @param associations: the associations with items
       @type  associations: list

       @return assoc_ls: the list of associations
       @rtype  assoc_ls: AssociationsList
    """
    assoc_ls = TfpgModesAssocList.create()

    for (k, (ok, expr, t)) in associations:
        new_assoc = Association.create(k, expr)
        assoc_ls.add_association(new_assoc)

    return assoc_ls

pass #end of create_tfpgModes_associations_list
# ---------------------------------------------------------------------------- 


class AssociationsDocument(etree.ElementBase):
    """Represents the document containg tfpg-slim expressions associations"""

    @staticmethod
    def create(failureModes_assoc_ls, monitoredDiscr_assoc_ls,
               unmonitoredDiscr_assoc_ls, tfpgModes_assoc_ls):
        """Creates the AssociationsDocument.
           @param  failureModes_assoc_ls: the list of failure modes associations
           @type   failureModes_assoc_ls: list
           @param  monitoredDiscr_assoc_ls: the list of monitored discrepancies associations
           @type   monitoredDiscr_assoc_ls: list
           @param  unmonitoredDiscr_assoc_ls: the list of unmonitored discrepancies associations
           @type   unmonitoredDiscr_assoc_ls: list
           @param  tfpgModes_assoc_ls: the list of tfpg modes associations
           @type   tfpgModes_assoc_ls: list

           @return associations_doc: the document of associations
           @type   associations_doc: object
        """
        associations_doc = _parser.makeelement(
            "{" + _XML_NAMESPACE_URI + "}" + _ASSOCIATIONS_TAG)

        associations_doc.append(copy.deepcopy(failureModes_assoc_ls))
        associations_doc.append(copy.deepcopy(monitoredDiscr_assoc_ls))
        associations_doc.append(copy.deepcopy(unmonitoredDiscr_assoc_ls))
        associations_doc.append(copy.deepcopy(tfpgModes_assoc_ls))

        return associations_doc


    @staticmethod
    def parse(filename):
        """Parses the AssociationsDocument.
           @param filename: the name of the file to parse
           @type  filename: string

           @return document_tree.getroot(): the document tree parsed
           @type   document_tree.getroot(): etree
        """
        error_list = []

        try:
            document_tree = etree.parse(filename, _parser)
            return document_tree.getroot(), error_list

        except etree.XMLSyntaxError, e:
            error_list.append(errors.NonWellFormedXmlError(str(e)))
            return None, error_list


    @staticmethod
    def get_syntax_errors_in_assoc_document(assoc_document):
        """Gets syntax errors from the Associations Document.
           @param assoc_document: the Associations Document
           @type  assoc_document: object

           @return error_list: the list of syntax errors
           @type   error_list: list
        """
        error_list = []
        try:
            _assoc_schema.assertValid(assoc_document)
        except etree.DocumentInvalid, e:
            error_list.append(errors.DocumentSyntaxError(str(e)))
        return error_list


    failureModes_assoc_ls = property(
        lambda self: self[0],
        doc="The failure modes associations of the Document.\n" +
            "@type: failureModesList")

    monitoredDiscr_assoc_ls = property(
        lambda self: self[1],
        doc="The monitored discrepancies associations of the Document.\n" +
            "@type: monitoredDiscrepanciesList")

    unmonitoredDiscr_assoc_ls = property(
        lambda self: self[2],
        doc="The unmonitored discrepancies associations of the Document.\n" +
            "@type: unmonitoredDiscrepanciesList")

    tfpgModes_assoc_ls = property(
        lambda self: self[3],
        doc="The tfpg modes associations of the Document.\n" +
            "@type: tfpgModesList")


    def __str__(self):
        """Converts the elements in string format"""
        return etree.tostring(self, xml_declaration=True, encoding="UTF-8",
                              pretty_print=True)
    

pass # end of AssociationsDocument class
# ----------------------------------------------------------------------------


class FailureModesAssocList(etree.ElementBase):
    """Represents the list of failure modes associations"""

    @staticmethod
    def create():
        """Creates the list of Failure Modes Associations.     
            @return   list of association:  The document of Failure Modes Associations list.
            @rtype    list of association:  object
        """
        return _parser.makeelement("{" + _XML_NAMESPACE_URI + "}" +
                                   _FAILURE_MODES_ASSOC_TAG)


    def add_association(self, assoc):
        """Adds the association to the list.
            @param   edge:  The Association.
            @type    edge:  Association
        """
        self.append(copy.deepcopy(assoc))
        return


    def __str__(self):
        """Converts the elements in string format"""
        return etree.tostring(self, xml_declaration=True, encoding="UTF-8",
                              pretty_print=True)


pass # end of FailureModesAssocList class
# ----------------------------------------------------------------------------


class MonitoredDiscrepanciesAssocList(etree.ElementBase):
    """Represents the list of monitored discrepancies associations"""

    @staticmethod
    def create():
        """Creates the list of Monitored Discrepancies Associations.     
            @return   list of association:  The document of Monitored Discrepancies Associations list.
            @rtype    list of association:  object
        """
        return _parser.makeelement("{" + _XML_NAMESPACE_URI + "}" +
                                   _MONITORED_DISCR_ASSOC_TAG)


    def add_association(self, assoc):
        """Adds the association to the list.
            @param   edge:  The Association.
            @type    edge:  Association
        """
        self.append(copy.deepcopy(assoc))
        return


    def __str__(self):
        """Converts the elements in string format"""
        return etree.tostring(self, xml_declaration=True, encoding="UTF-8",
                              pretty_print=True)


pass # end of MonitoredDiscrepanciesAssocList class
# ----------------------------------------------------------------------------


class UnmonitoredDiscrepanciesAssocList(etree.ElementBase):
    """Represents the list of unmonitored discrepancies associations"""

    @staticmethod
    def create():
        """Creates the list of Unmonitored Discrepancies Associations.     
            @return   list of association:  The document of Unmonitored Discrepancies Associations list.
            @rtype    list of association:  object
        """
        return _parser.makeelement("{" + _XML_NAMESPACE_URI + "}" +
                                   _UNMONITORED_DISCR_ASSOC_TAG)


    def add_association(self, assoc):
        """Adds the association to the list.
            @param   edge:  The Association.
            @type    edge:  Association
        """
        self.append(copy.deepcopy(assoc))
        return


    def __str__(self):
        """Converts the elements in string format"""
        return etree.tostring(self, xml_declaration=True, encoding="UTF-8",
                              pretty_print=True)


pass # end of UnmonitoredDiscrepanciesAssocList class
# ----------------------------------------------------------------------------


class TfpgModesAssocList(etree.ElementBase):
    """Represents the list of tfpg modes associations"""

    @staticmethod
    def create():
        """Creates the list of Tfpg Modes Associations.     
            @return   list of association:  The document of Tfpg Modes Associations list.
            @rtype    list of association:  object
        """
        return _parser.makeelement("{" + _XML_NAMESPACE_URI + "}" +
                                   _TFPG_MODES_ASSOC_TAG)


    def add_association(self, assoc):
        """Adds the association to the list.
            @param   edge:  The Association.
            @type    edge:  Association
        """
        self.append(copy.deepcopy(assoc))
        return


    def __str__(self):
        """Converts the elements in string format"""
        return etree.tostring(self, xml_declaration=True, encoding="UTF-8",
                              pretty_print=True)


pass # end of TfpgModesAssocList class
# ----------------------------------------------------------------------------


class Association(etree.ElementBase):
    """Represents a single associations"""


    @staticmethod
    def create(_id, expr):
        """Creates and Association.

           @param id: the identifier for the variable to be associated
           @type  id: string

           @param expr: the slim expression to be associated
           @type  expr: string

           @return association: the document of association
           @type   association: object
       """
        association = _parser.makeelement(
            "{" + _XML_NAMESPACE_URI + "}" + _ASSOC_TAG)

        association.set(_ID_ATTRIBUTE, str(_id))
        association.set(_EXPR_ATTRIBUTE, str(expr))

        return association


    _id = property(
        lambda self: self.get(_ID_ATTRIBUTE),
        doc="The variable identifier of the Association\n" +
            "@type: string")

    expr = property(
        lambda self: self.get(_EXPR_ATTRIBUTE),
        doc="The slim expression of the Association\n" +
            "@type: string")

pass # end of Association class
# ----------------------------------------------------------------------------


# ----------------------------------------------------------------------------
# Note: The following code will be executed once when this module is imported.
# ----------------------------------------------------------------------------

# Define some constants which are used in this module.
# The XML namespace that is used for configurations documents:
_XML_NAMESPACE_URI = ""
# The filename of the XML Schema document that defines the structure of valid
# tfpg documents.
_XSD_FILENAME = os.path.join(os.path.dirname(__file__),
                             "xsd", "tfpg_slim_associations.xsd")

# Define constants for the XML tag names that are used in tfpg documents.
_ASSOCIATIONS_TAG = "associations"
_FAILURE_MODES_ASSOC_TAG = "failureModes"
_MONITORED_DISCR_ASSOC_TAG = "monitoredDiscrepancies"
_UNMONITORED_DISCR_ASSOC_TAG = "unmonitoredDiscrepancies"
_TFPG_MODES_ASSOC_TAG = "tfpgModes"
_ASSOC_TAG = "assoc"
_ID_ATTRIBUTE = "id"
_EXPR_ATTRIBUTE = "expr"

# Initialize the parser that is used to parse and to create configurations
# specification documents.
# XML comments must be discarded. Otherwise the parser would yield data
# structures cluttered with unwanted elements.
# A lookup scheme is used in order to automatically map certain XML elements
# to instances of the adequate classes which are defined in this module.
_parser = etree.XMLParser(remove_comments=True)
_lookup = etree.ElementNamespaceClassLookup()
_parser.set_element_class_lookup(_lookup)

# Associate the various XML element tags with the corresponding classes. 
# Note that all elements are in the same namespace.
_namespace = _lookup.get_namespace(_XML_NAMESPACE_URI)
_namespace[_ASSOCIATIONS_TAG] = AssociationsDocument
_namespace[_FAILURE_MODES_ASSOC_TAG] = FailureModesAssocList
_namespace[_MONITORED_DISCR_ASSOC_TAG] = MonitoredDiscrepanciesAssocList
_namespace[_UNMONITORED_DISCR_ASSOC_TAG] = UnmonitoredDiscrepanciesAssocList
_namespace[_TFPG_MODES_ASSOC_TAG] = TfpgModesAssocList
_namespace[_ASSOC_TAG] = Association

# Parse the XML Schema document that defines the structure of valid tfpg assoc
# documents. This is needed for validation.
_assoc_schema_tree = etree.parse(_XSD_FILENAME)
_assoc_schema = etree.XMLSchema(_assoc_schema_tree)
