# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================


# Invokers regarding smv file manipulation throught nusmv (for
# model checking, etc.)

import datetime
import os
import os.path

import options
import utils.process
import slim_compiler.src.smv.api as smvapi
import property_management.src.constants as prop_const
import property_management.src.api as prop_api
from utils.log import log_icd

import string
from lxml import etree


def create_nusmv(smv_fname, ord_fname=None, need_msat=False):
    """Given a smv filename (and an optional ordering filename)
    creates an interactive process instance of class
    NuSMVProcess. If not specified, ord_fname is searched taking
    the smv basename. need_msat is True if the smv file contains
    variables whose domain is infinite (reals, integers)"""
    from utils.nusmv_process import NuSMVProcess
    import subprocess

    return NuSMVProcess(smv_fname, ord_fname, need_msat,
                        stdout=subprocess.PIPE, stderr=subprocess.STDOUT)


@log_icd
def model_checking(nusmv, formula):
    """Given a component implementation and a formula
    (property_management.src.formulae.Formula) predicating
    over it, performs model checking. The result is a possible xml
    trace dumped in file whose name is returned. Returns a pair
    (result, trace) where result is True if property is true, False
    if property does not hold, or None if result is unknown. trace
    is the name of a xml trace file if property is false, or None
    otherwise"""

    if formula.get_logic() == prop_const.Logic.CTL:
        res = model_checking_bdd(nusmv, formula)

    elif formula.get_logic() in [prop_const.Logic.LTL, ]:
        if options.NUSMV_USE_BMC or nusmv.needs_msat():
            res = model_checking_bmc(nusmv, formula)
        else:
            res = model_checking_bdd(nusmv, formula)

    elif formula.get_logic() == prop_const.Logic.MITL:
        assert nusmv.needs_msat()  # timed properties require timed fsm
        res = model_checking_mitl(nusmv, formula)

    else:
        raise ValueError("Only LTL and CTL formulae can be checked by NuSMV")

    return res


def model_checking_bdd(nusmv, formula):
    """Given a NuSMVProcess, builds the BDD FMS and model check
    given formula.  formula is a
    property_management.src.formulae.Formula instance"""
    assert formula.get_logic() in (prop_const.Logic.CTL, prop_const.Logic.LTL)

    if nusmv.needs_msat():
        raise ValueError("BDD cannot be used with infinite domains")
    formula_text = prop_api.get_formula_as_smv_string(formula)
    if formula.get_logic() == prop_const.Logic.CTL:
        res = nusmv.model_check_bdd_ctl(formula_text)
    else:
        res = nusmv.model_check_bdd_ltl(formula_text)
    return res


def model_checking_bmc(nusmv, formula):
    """Given a NuSMVProcess, builds the BMC FMS and model check
    given formula.  formula is a
    property_management.src.formulae.Formula instance"""
    assert formula.get_logic() in (prop_const.Logic.LTL,)

    formula_text = prop_api.get_formula_as_smv_string(formula)

    res = nusmv.model_check_bmc_ltl(formula_text, options.NUSMV_BMC_LENGTH)
    return res


def model_checking_mitl(nusmv, formula):
    """Given a NuSMVProcess, builds the sexp FMS and model check given
    formula.  formula is a property_management.src.formulae.Formula
    instance, whose pattern must be a MITL
    """

    pattern_name, time_bound, specs = \
        prop_api.get_mitl_formula_as_smv_tuple(formula)
    assert 0 < len(specs) <= 2

    cmd_map = {"absence": nusmv.check_pattern_absence,
               "existence": nusmv.check_pattern_existence,
               "universality": nusmv.check_pattern_universality,
               "precedence": nusmv.check_pattern_precedence,
               "responseGlobal": nusmv.check_pattern_response_global}

    assert pattern_name in cmd_map, "Unsupported pattern: " + pattern_name
    checker = cmd_map[pattern_name]

    time_var, tu = smvapi.get_smv_timed_var()

    return checker(*specs,
                   time_bound=time_bound,
                   time_var=time_var,
                   max_search_depth=options.NUSMV_BMC_LENGTH,
                   max_klive_bound=options.NUSMV_KLIVE_BOUND)


def check_is_state_zeno(nusmv, smv_expr):
    """Given a NuSMVProcess, builds the sexp FMS and check a
    state for zenoness.
    smv_expr is a string representing the state to reach.

    Returns one of these pairs:
           (NuSMVProcess.UNREACHABLE, None)
           (NuSMVProcess.NON_ZENO, trace_fname): trace_name is a
                                                 filename (string)
           (NuSMVProcess.ZENO, trace_fname): trace_name is a filename (string)
           (NuSMVProcess.UNKNOWN_NOT_FOUND, steps): steps is an integer value
           (NuSMVProcess.UNKNOWN_NOT_PROVED, time): time is a rational
                 value (string) in default time unit used by SmvConverter
    """

    time_var, tu = smvapi.get_smv_timed_var()

    _ret, _val = nusmv.is_state_zeno(smv_expr,
                                     time_var=time_var,
                                     max_search_depth=options.NUSMV_BMC_LENGTH,
                                     max_klive_bound=options.NUSMV_KLIVE_BOUND)

    if nusmv.UNKNOWN_NOT_PROVED == _ret:
        _val = smvapi.convert_smv_const_to_slim(_val)

    return (_ret, _val)


def check_clock_divergence(nusmv, clock_name, clock_bound=None):
    """Takes clock_name (string), clock_bound (string for rational
        number, or None for no-bound-specification),
        max_search_depth (integer, -1 for no bound).

        Searches for clock diverge and returns one pair of:
          * (NuSMVProcess.BOUNDED, bound)
            Where bound is a string with rational value (slim syntax)

          * (NuSMVProcess.UNBOUNDED, trace)
            Where trace is a file name containing produced trace (iff
            clock_bound was specified), or None if no clock_bound was
            specified.

          * (NuSMVProcess.UNKNOWN, None)
    """

    _ret, _val = nusmv.check_clock_divergence(
        clock_name,
        clock_bound=clock_bound,
        max_search_depth=options.NUSMV_BMC_LENGTH)

    # converts to slim sintax returned bound if any
    if (_ret == nusmv.BOUNDED):
        _val = _val.replace("f'", "")

    return (_ret, _val)


def simulation_random(nusmv, formulas):
    """Given a NuSMVProcess, builds the BDD FSM (of MSAT if needed)
    and simulate the model given formula.  formulas is a list of
    property_management.src.formulae.Formula instance.  Returns a
    trace filename when done"""

    smv_prop_formula = ([prop_api.get_formula_as_smv_string(f)
                         for f in formulas
                         if prop_api.check_is_propositional_formula(f)]
                        if formulas
                        else [])

    nusmv.pick_state_random(smv_prop_formula)
    return nusmv.simulate_random(smv_prop_formula, options.NUSMV_SIM_LENGTH)


def launch_smv_to_sigref(smv_fname, prob_fname, ord_fname=None,
                         label_fname=None,
                         out_fname=None, script_fname=None,
                         tau_name=smvapi.get_top_tau_event()):
    """Invokes binary smv2sigref. If script_fname or out_fname are None,
    the base name will be taken from the input smv_fname."""

    # script for nusmv
    script_templ = """# This file was generated on $gendate
# This file generates sigref output for input smv file '${smvfname}'
set input_file '${smvfname}'
set input_order_file '${ordfname}'
set on_failure_script_quits
dynamic_var_ordering -e sift
go
echo "Generating output in sigref format..."
compass_gen_sigref -i '${probfname}' -t "${tauname}" ${labelarg} -o '${outfname}'
echo "Generated output in sigref format in file '${outfname}'"
write_order -b -o '${new_ordfname}'
quit -x
"""
    basename = os.path.splitext(smv_fname)[0]
    if out_fname is None:
        out_fname = basename + ".xml"
    if script_fname is None:
        script_fname = basename + ".nusmv"

    if ord_fname is None:
        # try to see if there exists a possible ordering file
        ord_fname = basename + ".ord"
        if not os.path.isfile(ord_fname):
            ord_fname = ""

    new_ordfname = ord_fname if ord_fname else basename + ".ord"

    filestocheck = [smv_fname, prob_fname]

    if label_fname:
        labelarg = '-a "%s"' % label_fname
        filestocheck.append(label_fname)
    else:
        labelarg = ""

    if os.path.isfile(ord_fname):
        filestocheck.append(ord_fname)

    # dumps the script file
    script_file = open(script_fname, "w")
    script_file.write(string.Template(script_templ).substitute(
        {'gendate': str(datetime.datetime.utcnow()) + " UTC",
         'smvfname': smv_fname,
         'probfname': prob_fname,
         'labelarg': labelarg,
         'tauname': tau_name,
         'outfname': out_fname,
         'ordfname': ord_fname,
         'new_ordfname': new_ordfname, }))
    script_file.close()

    # checks that file exist
    for fname in filestocheck:
        if not os.path.isfile(fname):
            raise IOError("File not found: " + fname)

    # launch nusmv
    path = utils.process.get_exec_name(options.SMV2SIGREF)
    args = [path, "-load", script_fname]
    output = utils.process.start_process_and_get_results(
        args, None, options.DISPLAY_CALL_OUTPUT)

    normalize_sigref_xml(out_fname)

    return output


def normalize_sigref_xml(filepath):
    """Parses the sigref XML and normalized the node identifiers
    such that model2imc gets the same XML file consistently. This
    allows the results to be cached"""
    p = etree.XMLParser(huge_tree=True)
    tree = etree.parse(filepath, parser=p)
    for dd in tree.getroot().iter('dd'):
        id_map = dict()
        curr_id = 1
        # Store old identifiers and replace by new one
        for dd_node in dd.iter('dd_node'):
            old_id = dd_node.get("id")
            new_id = str(curr_id)
            curr_id += 1

            id_map[old_id] = new_id
            dd_node.set("id", new_id)

        # Replace all references by new identifiers
        for dd_node in dd.iter('dd_then'):
            if "node_ref" in dd_node.keys():
                dd_node.set("node_ref", id_map[dd_node.get("node_ref")])
        for dd_node in dd.iter('dd_else'):
            if "node_ref" in dd_node.keys():
                dd_node.set("node_ref", id_map[dd_node.get("node_ref")])

    # Replace XML file
    with open(filepath, "w") as file:
        file.write(etree.tostring(tree))


def launch_smv_to_bdd_ordering(smv_fname, ord_fname=None, kill_event=None):
    """Calls nusmv to generate a fair bit ordering for BDD
    variables. This is important especially when there are word
    arrays. If ord_fname is not given or None, basename of
    smv_fname is used.

    event 'kill_event' can be set by other threads, to kill the process.

    Returns the name of the ordering file.

    WARNING: Do not call this function if the smv file contains
    infinite domain variables, as nusmv cannot handle this
    condition in a clean manner."""

    if not os.path.isfile(smv_fname):
        raise IOError("File not found: " + smv_fname)

    basename = os.path.splitext(smv_fname)[0]
    if ord_fname is None:
        ord_fname = basename + ".ord"

    # a temp file for first ordering
    import tempfile
    tempf = tempfile.mktemp()

    # launch nusmv (degraded mode is fine for ordering)
    try:
        path = utils.process.get_exec_name(options.NUSMV)
    except IOError:
        path = utils.process.get_exec_name(options.DEGRADED_NUSMV)
        pass

    args = [path, "-int", smv_fname]
    script = """
set on_failure_script_quits
set write_order_dumps_bits
read_model
flatten_hierarchy
encode_variables
write_order -o "%s"
quit
""" % os.path.normpath(tempf)

    _res = utils.process.start_process_and_get_results(
        args, script, options.DISPLAY_CALL_OUTPUT, kill_event)

    if not os.path.isfile(tempf):
        raise IOError("File was expected to be found: " + tempf)

    # now reads back the list from the file
    path = os.path.join(options.BIN_DIR, options.ORDER_BITS)
    args = ["perl", path, "-i", "%s" % tempf, "-o", "%s" % ord_fname]

    utils.process.start_process_and_get_results(args, None, True)
    if os.path.isfile(tempf):
        os.remove(tempf)
    if not os.path.isfile(ord_fname):
        raise IOError("File was expected to be found: " + ord_fname)

    return ord_fname
