# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================


# invokers for slimsim


import os.path

import re
import tempfile
from utils import slimsim_process, log, exceptions
import subprocess
from slim_compiler.src.dataStructure import expressions
from slim_compiler.src import testexpr
from utils.py2xml import Py2XML
from slim_compiler.src.dataStructure.visitors.time import TimeScaleTransformVisitor
from slim_compiler.src.dataStructure.visitors.postProcessing import SimulatorVisitor
from property_management.src.op_elimination import stochastic
from property_management.src import constants
import options

def transform_slim(cinfo):
    TimeScaleTransformVisitor(cinfo).visit(cinfo)
    SimulatorVisitor().visit(cinfo)

def transform_formula(cinfo, root_impl, formula):
    stochastic.eliminate_w(formula, constants.CSL_NAMESPACE)
    stochastic.eliminate_g(formula, constants.CSL_NAMESPACE)
    stochastic.eliminate_f(formula, constants.CSL_NAMESPACE)

    for atomic_proposition in formula.get_atomic_propositions():
        slim_atomic_proposition_tree = testexpr.parseExpression(atomic_proposition.text)
        testexpr.check_property_tree(cinfo,
                         slim_atomic_proposition_tree,
                         root_impl)
        _qualify_property(slim_atomic_proposition_tree)
        xml_tree = Py2XML().parse(slim_atomic_proposition_tree)
        atomic_proposition.text = None
        atomic_proposition.append(xml_tree)

def start_slimsim(xml_filename, prop_filename, step_bound=options.MC_STEPBOUND, error_bound=options.MC_ERRORBOUND,
                  confidence=options.MC_CONFIDENCE, strategy=options.MC_STRATEGY, args=[]):
    # constructs a new instance of slimsim logging file
    fdlog, logfname = tempfile.mkstemp(suffix='.slimsimout', 
                                    prefix=os.path.splitext(os.path.basename(xml_filename))[0],
                                    dir=tempfile.gettempdir(), text=True)
    logfile = os.fdopen(fdlog, "w")
    if not options.MC_IGNOREDEADLOCK:
        args = args + ['--disallowdeadlock']
    if options.MC_IGNORESTEPBOUND:
        args = args + ['--ignorebounderror']

    process = slimsim_process.SlimsimProcess(xml_filename, prop_filename,
                            step_bound=step_bound,
                            error_bound=error_bound, confidence=confidence,
                            strategy=strategy,
                            args=args,
                            # Inherit stdout when running interactively. Only works in terminals etc, not GUI
                            stdout=None if strategy == 'input' else subprocess.PIPE,
                            stderr=subprocess.STDOUT,
                            dumpers=[logfile])

    return process

def run(xml_filename, prop_filename, step_bound=options.MC_STEPBOUND, error_bound=options.MC_ERRORBOUND,
        confidence=options.MC_CONFIDENCE, strategy=options.MC_STRATEGY, args=[]):
    process = start_slimsim(xml_filename, prop_filename, step_bound, error_bound, confidence, strategy, args)

    process.wait()
    if process.returncode != 0:
        raise exceptions.InternalException("Slimsim terminated with error code %d.\n%s" % (process.returncode,process.get_stdout_text()))
    
    return (process.returncode, (process.get_stdout_text(), process.get_stderr_text()))

def extract_results_from_output(slimsim_output):
    concatenated = ''.join(slimsim_output)

    re_property = re.compile( r'(?<=Property holds with probability )\[([0-9.]+),([0-9.]+)\]' )
    
    result = re_property.search(concatenated)
    if not result:
        log.user.error("slimsim did not output result information")
        return None

    (min_str, max_str) = result.group(1,2)

    return (float(min_str), float(max_str))

def _qualify_property(root_node):
    # Assumes only direct comparison between field and value possible, with no subexpressions
    if isinstance(root_node, expressions.BinaryOperatorTreeNode):
        _qualify_property(root_node.left)
        _qualify_property(root_node.right)

        if root_node.left.node_type == expressions.NodeType.MODE_FIELD and \
                root_node.right.node_type == expressions.NodeType.MODE_VALUE:
            root_node.right.name = root_node.left.name + ":" + root_node.right.name
        elif root_node.left.node_type == expressions.NodeType.MODE_VALUE and \
                root_node.right.node_type == expressions.NodeType.MODE_FIELD:
            root_node.left.name = root_node.right.name + ":" + root_node.left.name
        elif root_node.left.node_type == expressions.NodeType.ERROR_FIELD and \
                root_node.right.node_type == expressions.NodeType.ERROR_VALUE:
            root_node.right.name = root_node.left.name + ":" + root_node.right.name
        elif root_node.left.node_type == expressions.NodeType.ERROR_VALUE and \
                root_node.right.node_type == expressions.NodeType.ERROR_FIELD:
            root_node.left.name = root_node.right.name + ":" + root_node.left.name
    elif isinstance(root_node, expressions.UnaryOperatorTreeNode):
        _qualify_property(root_node.subexpression)
    elif isinstance(root_node, expressions.CaseExpressionTreeNode):
        for (cond, res) in root_node.cases:
            _qualify_property(cond)
            _qualify_property(res)
        _qualify_property(root_node.default_expression)

    return
