# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================
from slim_compiler.src.dataStructure.visitors.time import TimeScaleTransformVisitor
from slim_compiler.src.dataStructure.compile import CompileInformation
"""This module provides the public interface for handling properties and
formulae.

To parse an existing formulae document or properties document (in XML format)
and obtain the corresponding data structure, use the parse-methods provided by
the modules "formulae" and "properties".
"""

from lxml import etree

from slim_compiler.src import testexpr
import slim_compiler.src.smv.api as smv_api
import slim_compiler.src.dataStructure.expressions as slim_expressions

from property_management.src import constants
from property_management.src import errors
from property_management.src import exceptions
from property_management.src import util
import property_management.src.formulae as formulae_module
import property_management.src.patterns as patterns_module
from property_management.src.formula_translation import ltl_for_nusmv,\
    csl_for_intermediate
from property_management.src.formula_translation import ctl_for_nusmv
from property_management.src.formula_translation import mitl_for_nusmv
from property_management.src.formula_translation import pctl_for_mrmc
from property_management.src.formula_translation import csl_for_mrmc
from slim_compiler.src.dataStructure.visitors.expressionVisitor import ExpressionPostProcessor,\
    TypeVisitor, ExpressionValueVisitor
from slim_compiler.src.dataStructure.visitors.printVisitor import toSLIMVisitor

# TODO: Maybe this function can be removed.
def get_formulae_document_as_rat_project(formulae, root_impl):
    """ Returns a XML tree that is a valid RAT project"""
    
    # TODO: Printing the formulae should probably be removed.
    for formula in formulae:
        print get_formula_as_smv_string(formula)  
        
    project = etree.Element("project")
    
    # Adding the signals
    signals = etree.Element("signals")
    lang = smv_api.get_smv_language(root_impl)
    for var in lang.keys():
        signal = etree.SubElement(signals, "signal")
        name = etree.SubElement(signal, "name")
        name.text = var        
        type = etree.SubElement(signal, "type")
        type.text = lang[var]
        notes = etree.SubElement(signal, "notes")    
    project.append(signals)

    # Adding the requirements
    reqs = etree.Element("requirements")
    i=0    
    def is_supported(f): return f.get_logic() in constants.SUPPORTED_RAT_LOGICS
    sup_formulae = filter(is_supported, formulae)
    for sup_formula in sup_formulae:
        req = etree.SubElement(reqs, "requirement")
        name = etree.SubElement(req, "name")
        name.text = "R" + str(i)
        i = i + 1
        property = etree.SubElement(req, "property")
        property.text = get_formula_as_smv_string(sup_formula)
        notes = etree.SubElement(req, "notes")
    
    project.append(reqs)

    etree.SubElement(project, "property_assurance")
    etree.SubElement(project, "property_synthesis")
    
    return project


def get_atomic_proposition_strings(formula):
    """Returns the set of all atomic propositions (as strings) in the
    specified formula.
    
    @param   formula: The formula to be considered.
    @type    formula: formulae_module.Formula
    @return: The set of all atomic propositions in the specified formula.
    @rtype:  A set of strings.
    """
    return set([atomic_proposition.text
                for atomic_proposition in formula.get_atomic_propositions()])


def get_errors_in_atomic_proposition(atomic_proposition, root_implementation):
    """Validates the specified atomic proposition (in SLIM syntax) with
    respect to the SLIM model with the specified root component
    implementation and returns a list containing the detected errors.
    
    @param   atomic_proposition:  The atomic proposition in SLIM format.
    @type    atomic_proposition:  string
    @param   root_implementation: The object representing the implementation
                                  of the root component of the SLIM model
                                  which the atomic proposition refers to.
    @type    root_implementation: slim_compiler.src.dataStructure.\
                                  nominalbehaviour.ComponentImplementation
    @return: A list containing the errors in the atomic proposition. This list
             will be empty if the atomic proposition is valid (not necessarily
             true) with respect to the SLIM model.
    @rtype:  A list of errors.InvalidAtomicPropositionError objects. 
    """
    # Initialize the list of errors to be returned.
    error_list = []
    # Get the compile information object representing the SLIM model that
    # contains the specified root component implementation. This will be
    # needed later.
    compile_info = root_implementation.subpackage.package.compile_information
    # Validate the atomic proposition. Add an adequate error object to the
    # list if the atomic proposition is invalid.
    try:
        # Get a tree representation for the specified atomic proposition using
        # the data structures of the SLIM compiler. If the atomic proposition
        # is syntactically incorrect, an exception will be raised.
        atomic_proposition_tree = testexpr.parseExpression(atomic_proposition)
        # Check the tree representation of the atomic proposition. If it is
        # invalid, an exception will be raised.
        # TODO: check_property_tree is actually defined in the parse_property
        # module.
        testexpr.check_property_tree(compile_info,
                                     atomic_proposition_tree,
                                     root_implementation)
    except testexpr.InvalidExpressionException, e:
        error_list.append(errors.InvalidAtomicPropositionError(str(e)))
    
    return error_list

    
def get_errors_in_atomic_propositions_in_formula(formula,
                                                 root_implementation):
    """Validates all atomic propositions in the specified formula with respect
    to the SLIM model with the specified root component implementation and
    returns a list containing the detected errors.
    
    @precondition:
        All atomic propositions in the specified formula must be in SLIM
        syntax.
    
    @param   formula:             The formula to be checked.
    @type    formula:             formulae_module.Formula
    @param   root_implementation: The object representing the implementation
                                  of the root component of the SLIM model
                                  which the formula refers to.
    @type    root_implementation: slim_compiler.src.dataStructure.\
                                  nominalbehaviour.ComponentImplementation
    @return: A list containing the errors in the atomic propositions of the
             formula. This list will be empty if all atomic propositions are
             valid (not necessarily true) with respect to the SLIM model.
    @rtype:  A list of errors.InvalidAtomicPropositionError objects. 
    """
    # Collect the errors for all atomic propositions in the formula.
    # Note that the atomic propositions are represented by objects whose
    # attribute "text" provides the atomic proposition as a string.
    error_list = []
    for proposition_object in formula.get_atomic_propositions():
        error_list += get_errors_in_atomic_proposition(
                            proposition_object.text, root_implementation)
    return error_list
    

def get_atomic_proposition_in_smv_format(slim_atomic_proposition,
                                         root_implementation):
    """Returns the specified atomic proposition in SMV format.
    
    @param   slim_atomic_proposition: The atomic proposition in SLIM format.
    @type    slim_atomic_proposition: string
    @param   root_implementation:     The object representing the
                                      implementation of the root component of
                                      the SLIM model which the atomic
                                      proposition refers to.
    @type    root_implementation:     slim_compiler.src.dataStructure.\
                                      nominalbehaviour.ComponentImplementation
    @return: The atomic proposition in SMV format.
    @rtype:  string
    @raises  testexpr.InvalidExpressionException:
                If the specified atomic proposition is invalid.
    """
    # Get the compile information object representing the SLIM model that
    # contains the specified root component implementation. This will be
    # needed later.
    compile_info = root_implementation.subpackage.package.compile_information
    # Get a tree representation for the specified atomic proposition using the
    # data structures of the SLIM compiler. If the atomic proposition is
    # syntactically incorrect, an exception will be raised.
    slim_atomic_proposition_tree = testexpr.parseExpression(
                                    slim_atomic_proposition)
    # The tree must be checked because it will be annotated while doing this.
    # The annotations are needed in order to translate the atomic proposition
    # tree to a string in SMV syntax. If the atomic proposition is invalid,
    # an exception will be raised in the checking function. 
    # TODO: check_property_tree is actually defined in the parse_property
    # module.
    testexpr.check_property_tree(compile_info,
                                 slim_atomic_proposition_tree,
                                 root_implementation)
    # Translate the atomic proposition from SLIM to SMV syntax.
    return smv_api.convert_expression(root_implementation,
                                      slim_atomic_proposition_tree)

def get_time_bound_in_smv_format(slim_time_bound,
                                         root_implementation):
    """See get_atomic_proposition_in_smv_format, the difference is that the type is numerical"""
    compile_info = root_implementation.subpackage.package.compile_information
    slim_time_bound_tree = testexpr.parseExpression(
                                    slim_time_bound)
    TypeVisitor(compile_info).visit(slim_time_bound_tree)
    # Force time bound to be of real type (it may contain fractions)
    ExpressionPostProcessor().visit(slim_time_bound_tree, slim_expressions.NodeType.REAL)
    # Translate the atomic proposition from SLIM to SMV syntax.
    return smv_api.convert_expression(root_implementation,
                                      slim_time_bound_tree)

def get_time_bound_for_performability(slim_time_bound):
    """See get_atomic_proposition_in_smv_format, the difference is that the type is numerical"""
    #TODO: Performability assumes the base unit hour
    compile_info = CompileInformation()
    compile_info.base_unit = slim_expressions.TimeUnitType.SEC
    compile_info.uses_time_scaling = True
    slim_time_bound_tree = testexpr.parseExpression(
                                    slim_time_bound)
    TypeVisitor(compile_info).visit(slim_time_bound_tree)
    slim_time_bound_tree = TimeScaleTransformVisitor(compile_info).visit(slim_time_bound_tree)
    value = ExpressionValueVisitor(compile_info, False).visit(slim_time_bound_tree)
    return str(float(value))

def create_formula_for_property(property, base_unit = slim_expressions.TimeUnitType.HOUR):
    """Creates and returns the formula that is represented by the specified
    property.
    
    @param   property: The property to be considered.
    @type    property: properties_module.Property
    @return: The logical formula represented by the pattern instance.
    @return: A new object representing the formula that is represented by the
             specified property.
    @rtype:  formulae_module.Formula
    @raises  exceptions.FailedToLoadPatternsDocumentError:
                If the patterns document containing the definition of the
                pattern to be instantiated could not be loaded.
    @raises  exceptions.PatternInstantiationError:
                If the instantiation of the pattern failed. This might happen
                if the specified pattern instance refers to a nonexistent
                pattern, if not all placeholders are instantiated or if a
                placeholder is instantiated twice.
    """
    # Instantiate the pattern that is used to define the specified property.
    # Use the patterns document containing the definition of the patterns
    # for the logic of the property.
    # Note: get_patterns_document may raise an
    # exceptions.FailedToLoadPatternsDocumentError and
    # property.pattern_instance.create_formula may raise an
    # exceptions.PatternInstantiationError.
    patterns_document = get_patterns_document(property.get_logic())
    return property.pattern_instance.create_formula(patterns_document)


def create_formulae_document_for_properties(properties, base_unit = slim_expressions.TimeUnitType.HOUR):
    """Creates and returns a formulae document containing the formulae that
    are represented by the specified properties.    
    
    Example: create_formulae_document([]) will create an empty formulae
    document.
    
    @param   properties: The properties to be considered.
    @type    properties: An iterable over properties.Property objects.
    @return: A new formulae document containing the formulae that are
             represented by the specified properties.
    @rtype:  formulae_module.FormulaeDocument
    @raises  exceptions.FailedToLoadPatternsDocumentError:
                If a pattern document that is needed to create the formula for
                a property could not be loaded.
    @raises  exceptions.PatternInstantiationError:
                If the corresponding formula could not be created for one of
                the specified properties.
    """
    # Create the formulae represented by the specified properties and add them
    # to a new formulae document.
    formulae = result = [create_formula_for_property(p, base_unit) for p in properties]
    return formulae_module.FormulaeDocument.create(formulae)


def convert_atomic_propositions_to_smv(formulae, root_implementation):
    """Replaces all atomic propositions in the specified formulae by their
    correspondents in SMV format.

    Note: The specified formulae will be modified.
    
    @precondition:
        All atomic propositions in the specified formulae must be in SLIM
        syntax.
        
    @param   formulae:            The formulae to be modified.
    @type    formulae:            A sequence of Formula objects. In
                                  particular, this might be a
                                  formulae_module.FormulaeDocument object.
    @param   root_implementation: The object representing the implementation
                                  of the root component of the SLIM model
                                  which the formulae refer to.
    @type    root_implementation: slim_compiler.src.dataStructure.\
                                  nominalbehaviour.ComponentImplementation
    """
    # Convert all atomic propositions in all specified formulae.
    for formula in formulae:
        for atomic_proposition in formula.get_atomic_propositions():
            atomic_proposition.text = get_atomic_proposition_in_smv_format(
                                                    atomic_proposition.text,
                                                    root_implementation)
        for time_bound in formula.get_time_bounds():
            time_bound.text = get_time_bound_in_smv_format(
                                                    time_bound.text,
                                                    root_implementation)

def convert_time_bounds_for_performability(formulae):
    """Replaces all time bounds in the specified formulae by their
    correspondents for performability analysis format.

    Note: The specified formulae will be modified.
    
    @precondition:
        All atomic propositions in the specified formulae must be in SLIM
        syntax.
        
    @param   formulae:            The formulae to be modified.
    @type    formulae:            A sequence of Formula objects. In
                                  particular, this might be a
                                  formulae_module.FormulaeDocument object.
    """
    # Convert all time bounds in all specified formulae.
    for formula in formulae:
        for time_bound in formula.get_time_bounds():
            time_bound.text = get_time_bound_for_performability(
                                                    time_bound.text)

def create_proposition_to_label_dict(formulae):
        atomic_propositions_of_all_formulae = set()
        
        for formula in formulae:
            # The atomic proposition of the formula must be event names of the fault tree.
            atomic_propositions = get_atomic_proposition_strings(formula)
            # XXX: This is a dirty hack. It is needed because the "true" formula is
            # currently treated as if it was a normal atomic proposition.
            atomic_propositions = set(filter(lambda x: x != "tt", atomic_propositions))
            atomic_propositions_of_all_formulae = atomic_propositions_of_all_formulae.union(atomic_propositions)
                    
        # dft2imc will use state labels to indicate that a certain event has
        # occurred. It needs as input a file that maps the relevant event names to
        # the labels to be used. These labels must conform to the syntax for valid
        # MRMC state labels. Therefore we will create a dictionary that maps each
        # event name that is used as an atomic proposition in the formula to an
        # identifier that conforms to the syntax for MRMC state labels.
        proposition_to_label_dict = create_string_to_mrmc_label_dict(list(atomic_propositions_of_all_formulae))

        return proposition_to_label_dict

def create_string_to_mrmc_label_dict(strings):
    """Returns a dictionary that maps each of the specified strings to a
    unique label that complies with the MRMC label syntax.
    
    @param   strings: The strings to be mapped.
    @type    strings: A sequence of strings.
    @return: A dictionary that maps each of the specified strings to a unique
             label that complies with the MRMC label syntax.
    @rtype:  dict: string -> string
    """
    string_to_label_dict = {}
    for (index, current_string) in enumerate(strings):
        if not string_to_label_dict.has_key(current_string):
            string_to_label_dict[current_string] = "label" + str(index + 1)
    return string_to_label_dict

def replace_atomic_propositions(formula, replacement_dict):
    """Replaces the atomic propositions in the specified formula with the
    corresponding values from the specified dictionary.
    
    @param   formula:          The formula to be modified.
    @type    formula:          formulae_module.Formula
    @param   replacement_dict: A dictionary that maps atomic propositions to
                               their replacements. The dictionary may map all
                               atomic propositions that occur in the formula
                               or only a subset of them.
    @type    replacement_dict: dict: string -> string
    @return: A dictionary that maps each of the specified strings to a unique
             label that complies with the MRMC label syntax.
    @rtype:  dict: string -> string
    """
    for atomic_proposition in formula.get_atomic_propositions():
        if atomic_proposition.text in replacement_dict:
            new_text = replacement_dict[atomic_proposition.text]
            atomic_proposition.text = new_text


def get_formula_as_smv_string(formula):
    """Returns a string representation of the specified formula in SMV syntax.
    
    @precondition:
        In order to obtain a formula in valid SMV syntax, the atomic
        propositions of the specified formula must already be in SMV syntax.
        This can be achieved using the function
        convert_atomic_propositions_to_smv.
    
    @param   formula: The formula to be considered. It must be a
                      non-stochastic formula (LTL or CTL).
    @type    formula: formulae_module.Formula (must be LTL or CTL)
    @return: A string representation of the specified formula in SMV syntax.
    @rtype:  string
    @raises  Exception: If the formula is not in LTL or CTL.
    """
    # Determine the logic of the formula and translate accordingly.
    logic = formula.get_logic()
    if logic == constants.Logic.LTL:
        return ltl_for_nusmv.translate_formula(formula)
    elif logic == constants.Logic.CTL:
        return ctl_for_nusmv.translate_formula(formula)
    elif logic == constants.Logic.MITL:
        return mitl_for_nusmv.translate_formula_as_string(formula)
    else:
        # The formula is not in LTL or CTL.
        # TODO: Use a custom exception class.
        raise Exception("The specified formula is not a LTL or CTL formula " +
                        "so it cannot be translated to SMV syntax.")

def get_mitl_formula_as_smv_tuple(formula):
    logic = formula.get_logic()
    if logic == constants.Logic.MITL:
        return mitl_for_nusmv.translate_formula(formula)
    else:
        raise Exception("The specified formula is not a MITL formula " +
                        "so it cannot be translated to SMV syntax.")

def get_formula_as_mrmc_string(formula):
    """Returns a string representation of the specified formula in MRMC
    syntax.
    
    @precondition:
        In order to obtain a formula in valid MRMC syntax, the atomic
        propositions of the specified formula must already be in MRMC syntax.
        This can be achieved using the functions
        create_string_to_mrmc_label_dict and replace_atomic_propositions.
    
    @param   formula: The formula to be considered. It must be a stochastic
                      formula (PCTL or CSL).
    @type    formula: formulae_module.Formula (must be PCTL or CSL)
    @return: A string representation of the specified formula in MRMC syntax.
    @rtype:  string
    @raises  Exception: If the formula is not in PCTL or CSL.
    """
    # Determine the logic of the formula and translate accordingly.
    logic = formula.get_logic()
    if logic == constants.Logic.PCTL:
        return pctl_for_mrmc.translate_formula(formula)
    elif logic == constants.Logic.CSL:
        return csl_for_mrmc.translate_formula(formula)
    else:
        # The formula is not in PCTL of CSL.
        # TODO: Use a custom exception class.
        raise Exception("The specified formula is not a PCTL or CSL " +
                        "formula so it cannot be translated to MRMC syntax.")

def get_csl_bounds(formula):
    if formula.tag == constants.FORMULAE_NAMESPACE + "cslFormula":
        if formula[0][2].tag == constants.CSL_NAMESPACE + "u":
            if formula[0][2][0][0].tag == constants.CSL_NAMESPACE + "value":
                # Any pattern but probabilistic response
                return (formula[0][2][0][0].text, formula[0][2][1][0].text)
            elif formula[0][2][0][1][0][2].tag == constants.CSL_NAMESPACE + "u":
                # Probabilistic response pattern
                return (formula[0][2][0][1][0][2][0][0].text,formula[0][2][0][1][0][2][1][0].text)
            else:
                raise Exception("Malformed CSL formula: has no bounds")
        else:
            raise Exception("Malformed CSL formula: doesn't start with probabilistic until")
    else:
        raise Exception("The specified formula is not a CSL formula")

def get_formula_as_intermediate_string(formula):
    """Returns a string representation of the specified formula in intermediate
    syntax.
    
    @precondition:
        In order to obtain a formula in valid intermediate syntax, the atomic
        propositions of the specified formula must already be in intermediate syntax.
        This can be achieved using the functions
        create_string_to_mrmc_label_dict and replace_atomic_propositions.
    
    @param   formula: The formula to be considered. It must be a CSL formula.
    @type    formula: formulae_module.Formula (must be CSL)
    @return: A string representation of the specified formula in intermediate syntax.
    @rtype:  string
    @raises  Exception: If the formula is not in CSL.
    """
    # Determine the logic of the formula and translate accordingly.
    logic = formula.get_logic()
    if logic == constants.Logic.CSL:
        return csl_for_intermediate.translate_formula(formula)
    else:
        # The formula is not in CSL.
        # TODO: Use a custom exception class.
        raise Exception("The specified formula is not a CSL " +
                        "formula so it cannot be translated to intermediate syntax.")

def getCslFormulaBounds(formula):
    """Returns the lower and upper bound of the probabilistic until at the beginning
    of the formula.
    
    @param   formula: The formula to be considered. It must be a CSL formula.
    @type    formula: formulae_module.Formula (must be CSL)
    @return: A tuple containing the lower and upper bound
    @rtype:  tuple
    """
    if formula.tag == constants.FORMULAE_NAMESPACE + "cslFormula":
        if  formula[0].tag == constants.CSL_NAMESPACE + "p" \
        and formula[0][2].tag == constants.CSL_NAMESPACE + "u":
            return (formula[0][2][0][0].text, formula[0][2][1][0].text)
        else:
            raise Exception("The specified formula is not bounded")
    else:
        raise Exception("The specified formula is not a CSL formula")

def create_formula_from_slim_expression(logic, expression_tree_node):
    """Creates and returns a formula representing the specified boolean SLIM
    expression.
    
    All sub-expressions that are not applications of one of the (boolean)
    operators given below will be used as atomic propositions of the formula.
    More precisely: The string representations of such sub-expressions in SLIM
    syntax will be used as atomic propositions.
    
    Operators that will be translated:
        "and", "or", "imp", "iff", "xor", "xnor"
    
    Example:
        SLIM expression: "(x > 0) and some_boolean"
        Resulting atomic propositions: "x > 0", "some_boolean"
    
    @param   logic:                The logic of the formula.
    @type    logic:                One of the constants in constants.Logic.
    @param   expression_tree_node: A tree representation of the SLIM
                                   expression to be considered. This must be a
                                   boolean expression.
    @type    expression_tree_node: slim_expressions.ExpressionTreeNode
    @return: A formula representing the specified boolean SLIM expression.
    @rtype:  formulae_module.Formula
    """
    # TODO: Use constants instead of string literals for XML tag names in this
    # function.
    
    # Define a recursive helper function:
    def create_xml_element_for_expression(logic, tree_node):
        # Three cases have to be handled:
        # 1) The specified expression is an application of one of the
        #    supported unary boolean operators (this is only the "not"
        #    operator).
        # 2) The specified expression is an application of one of the
        #    supported binary boolean operators.
        # 3) The specified expression is not an application of one of the
        #    supported boolean operators and must thus be interpreted as an
        #    atomic proposition.
        
        # Handle case 1).
        if isinstance(tree_node, slim_expressions.UnaryOperatorTreeNode):
            if tree_node.operator == slim_expressions.OperatorType.NOT:
                xml_element = create_xml_element(logic, "not")
                xml_child_element = create_xml_element_for_expression(
                                        logic, tree_node.subexpression)
                xml_element.append(xml_child_element)
                return xml_element
        
        # Handle case 2).
        if isinstance(tree_node, slim_expressions.BinaryOperatorTreeNode):
            # Most of the binary operators can be mapped directly to a
            # corresponding XML tag name. Use a dictionary for that.
            # Note that the binary XNOR operator is equivalent to the IFF
            # operator here. XNOR may differ from IFF in terms of operator
            # precedence, but since precedence is explicit in the tree
            # representation of the expression, this does not matter.
            # Also note that the XOR operator is not available in the XML
            # format for formulae. It has to be translated separately.
            operator_to_xml_tag_dict = {
                    slim_expressions.OperatorType.AND:  "and",
                    slim_expressions.OperatorType.OR:   "or",
                    slim_expressions.OperatorType.IMP:  "imp",
                    slim_expressions.OperatorType.IFF:  "eq",
                    slim_expressions.OperatorType.XNOR: "eq" }
             
            if tree_node.operator in operator_to_xml_tag_dict:
                tag_name = operator_to_xml_tag_dict[tree_node.operator]
                xml_element = create_xml_element(logic, tag_name)
                xml_child_element1 = create_xml_element_for_expression(
                                                logic, tree_node.left)
                xml_child_element2 = create_xml_element_for_expression(
                                                logic, tree_node.right)
                xml_element.append(xml_child_element1)
                xml_element.append(xml_child_element2)
                return xml_element
            elif tree_node.operator == slim_expressions.OperatorType.XOR:
                # "a XOR b" can be translated to "NOT(a IFF b)".
                eq_xml_element = create_xml_element(logic, "eq")
                xml_child_element1 = create_xml_element_for_expression(
                                                logic, tree_node.left)
                xml_child_element2 = create_xml_element_for_expression(
                                                logic, tree_node.right)
                eq_xml_element.append(xml_child_element1)
                eq_xml_element.append(xml_child_element2)
                xml_element = create_xml_element(logic, "not")
                xml_element.append(eq_xml_element)
                return xml_element

        # Handle case 3).
        expression_string = toSLIMVisitor().visit(tree_node)
        return formulae_module.AtomicProposition.create(logic,
                                                        expression_string)
                
    # Define a helper function for creating XML elements:
    def create_xml_element(logic, tag_name):
        xml_namespace = constants.logic_to_xml_namespace_dict[logic]
        return formulae_module.parser.makeelement(xml_namespace + tag_name)
    
    # Use the recursive helper function to create an XML element representing
    # the SLIM expression. This element is not a formulae_module.Formula
    # object, but can be used as the content of such an object.
    formula_content = create_xml_element_for_expression(logic,
                                                        expression_tree_node)
    formula = formulae_module.Formula.create(logic, formula_content)
    return formula


def check_is_propositional_formula(formula):
    """Returns a boolean value indicating whether the specified formula is a
    propositional formula.
    
    A formula is called propositional if it is composed only of atomic
    propositions and boolean connectives.
    
    @param   formula: The formula to be considered.
    @type    formula: formulae_module.Formula
    @return: True if the specified formula is a propositional one, else False.
    @rtype:  bool
    """
    # Define the tags that are allowed to occur in the XML representation of a
    # propositional formula.
    allowed_tags = ["formula", "true", "atomicProposition",
                    "and", "or", "imp", "eq", "not"]
    
    # Traverse the XML tree representing the formula and check whether only
    # allowed tags are used. Use a recursive function for this.
    def _check_descendants_have_allowed_tags(xml_element):
        # Check the tags of the children of the specified element and continue
        # recursively. Return False as soon as a disallowed tag is found.
        # Note: The *local* tag name has to be checked because the namespace
        # of the elements depends on the logic of the formula.
        for child_element in xml_element:
            if child_element.xpath("local-name()") in allowed_tags:
                if not _check_descendants_have_allowed_tags(child_element):
                    return False
            else:
                return False
        # All descendants have allowed tags.
        return True
    
    # Start the traversal with the specified Formula object itself. This
    # object is the root element of the XML tree representing the formula. 
    return _check_descendants_have_allowed_tags(formula)


def check_is_propositional_property(property):
    """Returns a boolean value indicating whether the specified property
    represents a propositional formula.
    
    See check_is_propositional_formula.
    
    @param   property: The property to be considered.
    @type    property: properties_module.Property
    @return: True if the specified property is a propositional one, else
             False.
    @rtype:  bool
    """
    formula = create_formula_for_property(property)
    return check_is_propositional_formula(formula)


@util.caching_function
def get_patterns_document(logic):
    """Returns an object representing the patterns document for the specified
    logic.
    
    There is exactly one patterns document for each of the following logics:
    LTL, CTL, CSL
    
    Note: The behavior of this function is modified by the function decorator
    util.CachingDecorator:
    When this function is first invoked, the specified patterns document will
    be parsed and the resulting data structure will be cached. When this
    function is called again for the same patterns document, it will return
    exactly the same data structure as before to avoid the parsing overhead.
    
    @param   logic: A constant identifying the logic for which the patterns
                    document shall be returned.
    @type    logic: One of the constants in constants.Logic.
    @return: A tree representation of the patterns document with the specified
             filename.
    @rtype:  patterns_module.PatternsDocument
    @raises  exceptions.FailedToLoadPatternsDocumentError:
                If the patterns document could not be loaded. In particular,
                this will be an exceptions.NoPatternsForSpecifiedLogicError
                object if there are no patterns for the specified logic.
    """
    # Associate the logics with the corresponding patterns document filenames.
    logic_to_filename_dict = {
            constants.Logic.LTL: constants.LTL_PATTERNS_DOCUMENT_PATH,
            constants.Logic.CTL: constants.CTL_PATTERNS_DOCUMENT_PATH,
            constants.Logic.CSL: constants.CSL_PATTERNS_DOCUMENT_PATH,
            constants.Logic.MITL: constants.MITL_PATTERNS_DOCUMENT_PATH}
    # Get the filename for the patterns document for the specified logic.
    try:
        patterns_document_filename = logic_to_filename_dict[logic]
    except KeyError:
        # There is not patterns document for the specified logic.
        raise exceptions.NoPatternsForSpecifiedLogicError(logic)
    # Parse the patterns document and validate the result.
    patterns_document = patterns_module.parse_patterns_document(
                                    patterns_document_filename)
    error_list = patterns_module.get_syntax_errors_in_document(
                                            patterns_document)
    if len(error_list) > 0:
        error_strings = map(str, error_list)
        raise exceptions.FailedToLoadPatternsDocumentError(str(error_strings))
    return patterns_document
