# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================


"""This module provides functionality to translate an MITL formula in XML format
to NuSMV format.
"""

from property_management.src import constants
from property_management.src.formula_translation import exceptions

def get_formula_as_tuple(root):
    """Return a tuple of the MITL formula that is given in XML format as an
    ElementTree rooted at the specified element.
    The tuple consists of 3 elements:
    (formula name, time bound, propositions) where propositions is a list
    of the atomic propositions. The first elements is always proposition 'P'
    the (optional) second is 'Q'
    
    @param   root: The root element of the ElementTree representing the
                   formula.
    @type    root: etree._Element
    @return: A tuple representation of the formula that is given in XML format.
    @rtype:  tuple
    @raises  exceptions.UnexpectedXmlTagError: If the XML representation of
                the formula contains unexpected tags.
    @raises  exceptions.ViolatedRestrictionError: If the formula does not
                satisfy the restrictions that are imposed by the target tool
                (NuSMV).
    """
    # Proceed depending on the tag of the specified root element.
    # TODO: Use constants for the XML tag names.
    if root.tag == constants.FORMULAE_NAMESPACE + "mitlFormula":
        pattern = None
        time = None
        for elem in root:
            if elem.tag == constants.MITL_NAMESPACE + "pattern":
                pattern = elem
            elif elem.tag == constants.MITL_NAMESPACE + "value":
                time = elem.text
        if pattern is None or time is None:
            raise exceptions.ViolatedRestrictionError("Missing pattern or time element")
        propositions = []
        patternName = pattern.get("name")

        if patternName in ["absence", "existence", "universality"]:
            propositions = [pattern[0].text]
        else:
            propositions = [pattern[0].text, pattern[1].text]
        return (patternName, time, propositions)

    else:
        raise exceptions.UnexpectedXmlTagError(root.tag)
    

def translate_formula(formula_root):
    """Eliminates operators from the specified formulae that are not supported
    by NuSMV and returns the resulting formula as a string in NuSMV syntax.
    
    @param   formula_root: The root element of the ElementTree representing
                           the formula.
    @type    formula_root: etree._Element
    @return: The translated formula in NuSMV syntax.
    @rtype:  string
    @raises  exceptions.UnexpectedXmlTagError: If the XML representation of
                the formula contains unexpected tags.
    @raises  exceptions.ViolatedRestrictionError: If the formula does not
                satisfy the restrictions that are imposed by the target tool
                (NuSMV).
    """
    return get_formula_as_tuple(formula_root)

def translate_formula_as_string(formula_root):
    _dict = {
             "absence" : "G_{<=%(T)s} !%(P)s",
             "existence" : "F_{<=%(T)s} %(P)s",
             "universality" : "G_{<=%(T)s} %(P)s",
             "precedence" : "G (%(Q)s -> O_{<=%(T)s} %(P)s)",
             "responseGlobal" : "G (%(P)s -> F_{<=%(T)s} %(Q)s)",
             }
    (name, time, props) = get_formula_as_tuple(formula_root)
    format = _dict[name]
    params = {"T" : str(time), "P" : props[0]}
    if len(props) > 1:
        params["Q"] = props[1]
    return format%params
