# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================


"""This module provides the data structures that are used to represent
formulae documents. It also provides functions to parse formulae documents and
validate them against the adequate XML Schema Definition.
"""

import copy

from lxml import etree

from property_management.src import constants
from property_management.src import xml_tags
from property_management.src import errors
from property_management.src import util


def parse_formulae_document(data_source):
    """Parses the formulae document from the specified data source and returns
    a representation of the document using the data structures defined in this
    module.
    
    The representation of the document will be in tree structure. For a valid
    document, the root element will be a FormulaeDocument object.
    
    Note: The document will not be validated. This must be done explicitly
    using the function get_syntax_errors_in_document.
    
    @param   data_source: The source of the data to be parsed. This may be a
                          filename or a file-like object.
    @type    data_source: A string or a file-like object.
    @return: A tree representation of the document.
    @rtype:  FormulaeDocument
    """
    # Use the parser that is created when this module is imported (see the end
    # of this file). It creates the adequate objects for the XML elements of
    # the document.
    # The result of the parsing process will be an etree._ElementTree object.
    # Return the root element of this tree (a FormulaeDocument object).
    document_tree = etree.parse(data_source, parser)
    return document_tree.getroot()


def get_syntax_errors_in_document(formulae_document):
    """Validates the specified formulae document against the adequate XML
    Schema Definition and returns a list containing the syntactic errors.
    
    Note: Syntactic validation does not include a check of the atomic
    propositions with respect to a certain SLIM model.
    
    @return: A list containing the syntactic errors in the document. This list
             will be empty if the document is valid with respect to the
             adequate XML Schema Definition.
    @rtype:  A list of errors.DocumentSyntaxError objects. 
    """
    # Initialize the list of errors to be returned.
    error_list = []
    # Validate the document. If it is invalid, an exception will be raised
    # which can be used to create an error object with an adequate error
    # description.
    # Note: The XML Schema object that is used for validation is created when
    # this module is imported (see the end of this file).
    try:
        _formulae_schema.assertValid(formulae_document)
    except etree.DocumentInvalid, e:
        error_list.append(errors.DocumentSyntaxError(str(e)))
    return error_list

# ----------------------------------------------------------------------------
# Data structures for formulae documents.
# ----------------------------------------------------------------------------

class FormulaeDocument(util.SerializableXmlElement):
    """A list of logical formulae (in any logic), represented as Formula
    objects.
    
    Important: FomulaeDocument objects must not be created like this:
    x = FomulaeDocument()
    Instead, the static factory method create must be used.
    This is because this class is a subclass of etree.ElementBase, which must
    not be instantiated directly. Refer to the lxml documentation for details.
    """
    
    @staticmethod
    def create(formulae=None):
        """Creates and returns a new FomulaeDocument object.
        
        Note: 
        
        @param   formulae: (Optional) The formulae to be added to the new
                           document. Note: Instead of the formulae objects
                           themselves, deep copies of these objects will be
                           added to the document.
        @type    formulae: An iterable over Formula objects.
        @return: A new FomulaeDocument object.
        @rtype:  FomulaeDocument
        """
        # Note: The parser that is used to create the object is created when
        # this module is imported (see the end of this file).
        formulae_document = parser.makeelement(constants.FORMULAE_NAMESPACE +
                                               xml_tags.FORMULAE_DOCUMENT_TAG)
        if formulae is not None:
            for formula in formulae:
                # Don't add the formula object itself but rather a deep copy.
                # This is necessary for the following reason: The formula
                # object is an XML element, which might already be used as a
                # child element of some other parent element (e.g. a formulae
                # document). Without copying, it would be removed from that
                # parent. See the lxml documentation.
                formulae_document.append(copy.deepcopy(formula))
        return formulae_document

    def __str__(self):
        """Returns a string representation of this document in XML format.
        
        UTF-8 encoding will be used and an XML header will be included.
        
        @return: A string representation of this document in XML format.
        @rtype:  string
        """
        return etree.tostring(self, xml_declaration=True, encoding="UTF-8",
                              pretty_print=True)


class Formula(util.SerializableXmlElement):
    """A logical formula in a certain logic.
    
    Important: Formula objects must not be created like this: x = Formula()
    Instead, the static factory method create must be used.
    This is because this class is a subclass of etree.ElementBase, which must
    not be instantiated directly. Refer to the lxml documentation for details.
    """
    
    @staticmethod
    def create(logic, content):
        """Creates and returns a new logical formula.
        
        @param   logic:   The logic to be used.
        @type    logic:   One of the constants in constants.Logic.
        @param   content: The content of the formula. The content is an XML
                          element representing either an atom (e.g. an atomic
                          proposition) or the top level operator of the
                          formula, including the corresponding subformulae.
                          adequate XML element. This might be an atomic
                          proposition.
                          Note: A deep copy of the content will be created.
        @type    content: etree._Element (e.g. AtomicProposition)
        @return: A new StateFormulaPlaceholderInstance object.
        @rtype:  StateFormulaPlaceholderInstance
        """
        # Get the adequate XML tag name for the specified logic. There is a
        # dictionary for this, but it has to be inverted first.
        logic_to_tag_dict = dict([(_logic, tag_name) for (tag_name, _logic) in
                                  _formula_tag_to_logic_dict.iteritems()])
        tag_name = logic_to_tag_dict[logic]
        # Create a new XML element representing the formula in the specified
        # logic.
        # Note: The parser that is used to create the element is created when
        # this module is imported (see the end of this file).
        formula = parser.makeelement(constants.FORMULAE_NAMESPACE + tag_name)
        # The XML element must have exactly one child element, namely the
        # element that represents the content of the formula.
        # Don't add the specified content object itself but rather a deep
        # copy. This is necessary for the following reason: The content object
        # is an XML element, which might already be used as a child element of
        # some other parent element (e.g. another formula). Without copying,
        # it would be removed from that parent. See the lxml documentation.
        formula.append(copy.deepcopy(content))
        return formula
    
    def get_logic(self):
        """Returns a constant identifying the logic of this formula.
        
        @return: A constant identifying the logic of this formula.
        @rtype:  One of the constants in constants.Logic.
        """
        # The logic can be deduced from the tag name of the underlying XML
        # element.
        tag_name = self.xpath("local-name()")
        return _formula_tag_to_logic_dict[tag_name]

    def get_pattern_id(self):
        return self.xpath("(//@patternId)[1]")

    # This formula is represented by an XML element that has exactly one child
    # element representing the content of the formula.
    content = property(lambda self: self[0],
                       doc="The content of the formula.\n" +
                           "The content is an XML element representing " +
                           "either an atom (e.g. an atomic proposition) or " +
                           "the top level operator of the formula, " +
                           "including the corresponding subformulae.\n" +
                           "@type: etree._Element (e.g. AtomicProposition)")
    
    def get_atomic_propositions(self):
        """Returns a list containing  all atomic propositions that are used in
        this formula.
        
        Note: The atomic propositions are not copied, so modifications of the
        atomic propositions in the returned list directly affect this formula.
        
        @return: A list containing all atomic propositions in this formula.
        @rtype:  A list of AtomicProposition objects.
        """
        # Return all atomicProposition elements in the XML subtree
        # representing this formula, regardless of the namespace (which
        # indicates the logic). Note that due to the lookup scheme used by the
        # parser for formulae documents, the elements will be represented by
        # AtomicProposition objects.
        atomic_propositions = self.xpath(".//*[local-name()='%s']" %
                                         xml_tags.ATOMIC_PROPOSITION_TAG)
        for atomic_proposition in atomic_propositions:
            assert isinstance(atomic_proposition, AtomicProposition)
        return atomic_propositions

    def get_time_bounds(self):
        """Returns a list containing  all time bounds that are used in
        this formula.
        
        Note: The time bounds are not copied, so modifications of the
        time bounds in the returned list directly affect this formula.
        
        @return: A list containing all time bounds in this formula.
        @rtype:  A list of TimeBound objects.
        """
        # Return all atomicProposition elements in the XML subtree
        # representing this formula, regardless of the namespace (which
        # indicates the logic). Note that due to the lookup scheme used by the
        # parser for formulae documents, the elements will be represented by
        # AtomicProposition objects.
        time_bounds = self.xpath(".//*[local-name()='%s']" %
                                         xml_tags.VALUE_TAG)
        for time_bound in time_bounds:
            assert isinstance(time_bound, Value)
        return time_bounds

class AtomicProposition(util.SerializableXmlElement):
    """An atomic proposition of a logical formula (in a certain logic).
    
    Its value can be retrieved and set using the attribute "text".
    
    Important: AtomicProposition objects must not be created like this:
    x = AtomicProposition()
    Instead, the static factory method create must be used.
    This is because this class is a subclass of etree.ElementBase, which must
    not be instantiated directly. Refer to the lxml documentation for details.
    """
    
    @staticmethod
    def create(logic, atomic_proposition_string):
        """Creates and returns a new AtomicProposition object representing the
        atomic proposition that is specified as a string in SLIM syntax.
        
        The atomic proposition will be applicable for the specified logic.
        
        @param   logic:                     A constant identifying the logic
                                            in which the atomic proposition
                                            shall be used.
        @type    logic:                     One of the constants in
                                            constants.Logic.
        @param   atomic_proposition_string: A string representation of the
                                            atomic proposition in SLIM syntax.
        @type    atomic_proposition_string: string
        @return: A new AtomicProposition object.
        @rtype:  AtomicProposition
        """
        # Get the adequate XML namespace for the specified logic.
        xml_namespace = constants.logic_to_xml_namespace_dict[logic]
        # Note: The parser that is used to create the object is created when
        # this module is imported (see the end of this file).
        atomic_proposition = parser.makeelement(
                xml_namespace +
                xml_tags.ATOMIC_PROPOSITION_TAG)
        # The atomic proposition itself is stored as the text content of this
        # XML element.
        atomic_proposition.text = atomic_proposition_string
        return atomic_proposition

class TimeBound(util.SerializableXmlElement):
    """An time bound of a logical formula (in a certain logic).
    
    Its value can be retrieved and set using the attribute "text".
    
    Important: TimeBound objects must not be created like this:
    x = TimeBound()
    Instead, the static factory method create must be used.
    This is because this class is a subclass of etree.ElementBase, which must
    not be instantiated directly. Refer to the lxml documentation for details.
    """
    
    @staticmethod
    def create(logic, time_bound_string):
        """Creates and returns a new TimeBound object representing the
        time bound that is specified as a string in SLIM syntax.
        
        The time bound will be applicable for the specified logic.
        
        @param   logic:                     A constant identifying the logic
                                            in which the atomic proposition
                                            shall be used.
        @type    logic:                     One of the constants in
                                            constants.Logic.
        @param   time_bound_string: A string representation of the
                                    time bound in SLIM syntax.
        @type    time_bound_string: string
        @return: A new TimeBound object.
        @rtype:  TimeBound
        """
        # Get the adequate XML namespace for the specified logic.
        xml_namespace = constants.logic_to_xml_namespace_dict[logic]
        # Note: The parser that is used to create the object is created when
        # this module is imported (see the end of this file).
        time_bound = parser.makeelement(
                xml_namespace +
                xml_tags.TIME_BOUND_TAG)
        # The atomic proposition itself is stored as the text content of this
        # XML element.
        time_bound.text = time_bound_string
        return time_bound

class Value(util.SerializableXmlElement):
    """An time bound of a logical formula (in a certain logic).
    
    Its value can be retrieved and set using the attribute "text".
    
    Important: Value objects must not be created like this:
    x = Value()
    Instead, the static factory method create must be used.
    This is because this class is a subclass of etree.ElementBase, which must
    not be instantiated directly. Refer to the lxml documentation for details.
    """
    
    @staticmethod
    def create(logic, time_bound_string):
        """Creates and returns a new TimeBound object representing the
        time bound that is specified as a string in SLIM syntax.
        
        The time bound will be applicable for the specified logic.
        
        @param   logic:                     A constant identifying the logic
                                            in which the atomic proposition
                                            shall be used.
        @type    logic:                     One of the constants in
                                            constants.Logic.
        @param   time_bound_string: A string representation of the
                                    time bound in SLIM syntax.
        @type    time_bound_string: string
        @return: A new TimeBound object.
        @rtype:  TimeBound
        """
        # Get the adequate XML namespace for the specified logic.
        xml_namespace = constants.logic_to_xml_namespace_dict[logic]
        # Note: The parser that is used to create the object is created when
        # this module is imported (see the end of this file).
        time_bound = parser.makeelement(
                xml_namespace +
                xml_tags.VALUE_TAG)
        # The atomic proposition itself is stored as the text content of this
        # XML element.
        time_bound.text = time_bound_string
        return time_bound

def _get_parser_lookup_scheme():
    """Returns a new parser lookup scheme which is used to map certain XML
    elements in formulae documents to instances of the adequate classes
    defined in this module.
    
    @return: A parser lookup scheme for formulae documents.
    @rtype:  etree.ElementNamespaceClassLookup
    """
    lookup = etree.ElementNamespaceClassLookup()
    
    # Associate the various XML element tags with the corresponding classes.
    # Note: Elements representing atomic propositions are represented by the
    # same class, regardless of their namespace.
    formulae_namespace=lookup.get_namespace(constants.FORMULAE_NAMESPACE_URI)
    formulae_namespace[xml_tags.FORMULAE_DOCUMENT_TAG] = FormulaeDocument
    formulae_namespace[xml_tags.LTL_FORMULA_TAG] = Formula
    formulae_namespace[xml_tags.CTL_FORMULA_TAG] = Formula
    formulae_namespace[xml_tags.MITL_FORMULA_TAG] = Formula
    formulae_namespace[xml_tags.PCTL_FORMULA_TAG] = Formula
    formulae_namespace[xml_tags.CSL_FORMULA_TAG] = Formula

    ltl_namespace = lookup.get_namespace(constants.LTL_NAMESPACE_URI)
    ltl_namespace[xml_tags.ATOMIC_PROPOSITION_TAG] = AtomicProposition
    ltl_namespace[xml_tags.TIME_BOUND_TAG] = TimeBound
    ltl_namespace[xml_tags.VALUE_TAG] = Value

    ctl_namespace = lookup.get_namespace(constants.CTL_NAMESPACE_URI)
    ctl_namespace[xml_tags.ATOMIC_PROPOSITION_TAG] = AtomicProposition
    ctl_namespace[xml_tags.TIME_BOUND_TAG] = TimeBound
    ctl_namespace[xml_tags.VALUE_TAG] = Value

    pctl_namespace = lookup.get_namespace(constants.PCTL_NAMESPACE_URI)
    pctl_namespace[xml_tags.ATOMIC_PROPOSITION_TAG] = AtomicProposition
    pctl_namespace[xml_tags.TIME_BOUND_TAG] = TimeBound
    pctl_namespace[xml_tags.VALUE_TAG] = Value

    csl_namespace = lookup.get_namespace(constants.CSL_NAMESPACE_URI)
    csl_namespace[xml_tags.ATOMIC_PROPOSITION_TAG] = AtomicProposition
    csl_namespace[xml_tags.TIME_BOUND_TAG] = TimeBound
    csl_namespace[xml_tags.VALUE_TAG] = Value

    mitl_namespace = lookup.get_namespace(constants.MITL_NAMESPACE_URI)
    mitl_namespace[xml_tags.ATOMIC_PROPOSITION_TAG] = AtomicProposition
    mitl_namespace[xml_tags.TIME_BOUND_TAG] = TimeBound
    mitl_namespace[xml_tags.VALUE_TAG] = Value

    return lookup


# ----------------------------------------------------------------------------
# Note: The following code will be executed once when this module is imported.
# ----------------------------------------------------------------------------

# Formulae in different logics are represented by XML elements with different
# tag names. Map these tag names to the corresponding logics.
_formula_tag_to_logic_dict = {
        xml_tags.LTL_FORMULA_TAG:  constants.Logic.LTL,
        xml_tags.CTL_FORMULA_TAG:  constants.Logic.CTL,
        xml_tags.MITL_FORMULA_TAG:  constants.Logic.MITL,
        xml_tags.PCTL_FORMULA_TAG: constants.Logic.PCTL,
        xml_tags.CSL_FORMULA_TAG:  constants.Logic.CSL }

# Initialize the parser that is used to parse formulae documents.
# XML comments must be discarded. Otherwise the parser would yield data
# structures cluttered with unwanted elements.
parser = etree.XMLParser(remove_comments=True)
# A lookup scheme is used in order to automatically map certain XML elements
# to instances of the adequate classes which are defined in this module.
parser.set_element_class_lookup(_get_parser_lookup_scheme())

# Parse the XML Schema document that defines the structure of valid formulae
# documents. This is needed for validation.
_formulae_schema_tree = etree.parse(constants.FORMULAE_XSD_PATH)
_formulae_schema = etree.XMLSchema(_formulae_schema_tree)
