# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================


"""This module provides functions to eliminate certain operators from LTL
formulae.

The formulae are given in XML format, represented by ElementTree objects.

Currently, the only operator to be eliminated is W. This can be done using the
following function:
eliminate_w(root, namespace)
"""

import copy
from lxml import etree

# Important note: The parser provided by the formulae-module is used to create
# new XML elements. This is necessary in order to have these XML elements
# represented using the custom classes defined in the formulae-module.
import property_management.src.formulae as formulae_module


def eliminate_w(root, namespace):
    """Eliminate the weak until operator W in the formula that is represented
    by the XML ElementTree rooted a the specified element.
    
    This is done by replacing sub-formulae that contain the operator W by
    semantically equivalent sub-formulae that do not contain W but the strong
    until operator U.
    
    The following equivalence holds:
    x W y = G[x] or (x U y)
    
    @param   root:      The root element of the ElementTree representing the
                        formula.
    @type    root:      etree._Element
    @param   namespace: The XML namespace of the considered logic, surrounded
                        by curly braces. Example:
                        "{http://compass.informatik.rwth-aachen.de/ltl}"
    @type    namespace: string
    @raises  Exception: If the specified formula contains bounded weak until
                        operators. (Bounds are currently not supported.)
                        Note that some unbounded weak until operators might
                        already be eliminated from the formula when the
                        exception is raised.
    """
    # TODO: This function has to be revised. It has to be clarified how the
    # bounded weak until operator can be eliminated. Currently the bounded
    # version is not supported.
    
    # Use a shorter name to refer to the parser for formulae documents, which
    # is used to create new XML elements.
    parser = formulae_module.parser
    
    # Get all "w" elements in the ElementTree representing the formula in
    # depth first order and store them in a list.
    w_elements = list(root.iter(namespace + "w"))
    # Replace these "w" elements by corresponding "u" elements. Do this in
    # reversed order (i.e. bottom-up), to ensure correct results for nested W
    # formulae.
    w_elements.reverse()
    for w_element in w_elements:
        # Abort if the current W formula has bounds. Bounded weak until is
        # currently not supported. Bounds are represented by child elements of
        # the "w" element, in addition to the two child elements representing
        # the sub-formula of the W formula.
        if len(w_element) > 2:
            # TODO: Use a custom exception class.
            raise Exception("Bounded weak until formulae in LTL are not " +
                            "supported.")
        # Create a new element representing the sub-formula "x or y", where x
        # is the first and y the second sub-formula of the current W formula.
        # We are considering a sub-formula of the form "x W y". Get the
        # sub-formulae "x" and "y" which are represented by the last child
        # elements of the "w" element.
        sub_formula_x = w_element[-2]
        sub_formula_y = w_element[-1]
        # Create the sub-formula "x U y". Note that the elements representing
        # "x" and "y" are not copied but moved, so they will be removed from
        # their original parent element.
        u_element = parser.makeelement(namespace + "u")
        # Add the sub-formulae to the "u" element.
        u_element.append(sub_formula_x)
        u_element.append(sub_formula_y)
        
        # Create the sub-formula "G[x]".
        # Note that a deep copy is created of the element representing the
        # sub-formula x. This is necessary because x was already appended to
        # another element and would be removed from there without copying.
        g_element = parser.makeelement(namespace + "g")
        g_element.append(copy.deepcopy(sub_formula_x))

        # Create a new element representing the formula "G[x] or (x U y)".
        # This formula will replace the old W formula.
        # The sub-formulae "G[x]" and "x U y)" have to be wrapped in "formula"
        # elements first.
        g_wrapper = parser.makeelement(namespace + "formula")
        g_wrapper.append(g_element)
        u_wrapper = parser.makeelement(namespace + "formula")
        u_wrapper.append(u_element)
        or_element = parser.makeelement(namespace + "or")
        or_element.append(g_element)
        or_element.append(u_element)

        # Replace the W formula with the newly constructed formula.
        w_element.getparent().replace(w_element, or_element)
