# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================


"""This module provides functions to eliminate certain operators from CSL and
PCTL formulae.

The formulae are given in XML format, represented by ElementTree objects.

Functions:
- eliminate_w(root, namespace)
- eliminate_g(root, namespace)
- eliminate_f(root, namespace)
"""

import copy
from lxml import etree

# Important note: The parser provided by the formulae-module is used to create
# new XML elements. This is necessary in order to have these XML elements
# represented using the custom classes defined in the formulae-module.
import property_management.src.formulae as formulae_module


# Associate the tags of the comparison operator elements with the tags of
# their complements.
# TODO: Use constants here.
_complement_operators = { "lessThan":     "greaterThan",
                          "greaterThan":  "lessThan",
                          "lessEqual":    "greaterEqual",
                          "greaterEqual": "lessEqual" }
    

def eliminate_w(root, namespace):
    """Eliminate the weak until operator W in the formula that is represented
    by the XML ElementTree rooted a the specified element.
    
    This is done by replacing sub-formulae that contain the operator W by
    semantically equivalent sub-formulae that do not contain W but the strong
    until operator U.
    
    The following equivalences hold:
    P{>p}[x W y]  = P{<1-p}[!y U !(x or y)]
    P{>=p}[x W y] = P{<=1-p}[!y U !(x or y)]
    P{<p}[x W y]  = P{>1-p}[!y U !(x or y)]
    P{<=p}[x W y] = P{>=1-p}[!y U !(x or y)]
    
    @param   root:      The root element of the ElementTree representing the
                        formula.
    @type    root:      etree._Element
    @param   namespace: The XML namespace of the considered logic, surrounded
                        by curly braces. Example:
                        "{http://compass.informatik.rwth-aachen.de/csl}"
    @type    namespace: string
    """
    # Use a shorter name to refer to the parser for formulae documents, which
    # is used to create new XML elements.
    parser = formulae_module.parser
    
    # Get all "w" elements in the ElementTree representing the formula in
    # depth first order and store them in a list.
    w_elements = list(root.iter(namespace + "w"))
    # Replace these "w" elements by corresponding "u" elements. Do this in
    # reversed order (i.e. bottom-up), to ensure correct results for nested W
    # formulae.
    w_elements.reverse()
    for w_element in w_elements:
        # Create a new element representing the sub-formula "x or y", where x
        # is the first and y the second sub-formula of the current W formula.
        # These sub-formulae are represented by the last child elements of the
        # "w" element. Note that these children are not copied but moved, so
        # they will be removed from their original parent element.
        sub_formula_x = w_element[-2]
        sub_formula_y = w_element[-1]
        x_or_y = parser.makeelement(namespace + "or")
        x_or_y.append(sub_formula_x)
        x_or_y.append(sub_formula_y)
        
        # Create a new element representing the negation of the previous
        # formula, i.e. "!(x or y)".
        # The formula "x or y" has to be wrapped in a "formula" element first.
        x_or_y_wrapper = parser.makeelement(namespace + "formula")
        x_or_y_wrapper.append(x_or_y)
        not_x_or_y = parser.makeelement(namespace + "not")
        not_x_or_y.append(x_or_y_wrapper)
        
        # Create a new element representing the sub-formula "!y".
        # Note that a deep copy is created of the element representing the
        # sub-formula y. This is necessary because y was already appended to
        # another element and would be removed from there without copying.
        not_y = parser.makeelement(namespace + "not")
        not_y.append(copy.deepcopy(sub_formula_y))
        
        # Create a new element representing the sub-formula "!y U !(x or y)".
        # This will be the path formula of the new P formula.
        # The sub-formulae of the U formula have to be wrapped in "formula"
        # elements first.
        not_y_wrapper = parser.makeelement(namespace + "formula")
        not_y_wrapper.append(not_y)
        not_x_or_y_wrapper = parser.makeelement(namespace + "formula")
        not_x_or_y_wrapper.append(not_x_or_y)
        # Create the element representing the U formula.
        u_element = parser.makeelement(namespace + "u")
        # Reuse the bounds of the W operator if there are any.
        # Note that at this point, the sub-formula children of the "w" element
        # have already been removed.
        if len(w_element) > 0:
            # The W operator has at least one bound. Move it to the "u"
            # element. It will be removed from the "w" element.
            u_element.append(w_element[0])
            if len(w_element) > 0:
                # The W operator has an upper and a lower bound. Move the
                # upper bound element as well. It will be removed from the "w"
                # element.
                u_element.append(w_element[0])
        # Add the sub-formulae to the "u" element.
        u_element.append(not_y_wrapper)
        u_element.append(not_x_or_y_wrapper)
        
        # Get the superordinate "p" element of the "w" element.
        p_element = w_element.getparent()
        # Replace the comparison operator and the threshold of the "p" element
        # with their complements.
        _complement_p_element(p_element, namespace)
        # Replace the old path formula of the "p" element with the new one.
        # The path formula is represented by the third child element of the
        # "p" element.
        p_element[2] = u_element
    

def eliminate_g(root, namespace):
    """Eliminate the operator G ("globally") in the formula that is
    represented by the XML ElementTree rooted a the specified element.
    
    This is done by replacing sub-formulae that contain the operator G by
    semantically equivalent sub-formulae that do not contain G but F
    ("finally").
    
    The following equivalences hold:
    P{>p}[G[y]]  = P{<1-p}[F[!y]]
    P{>=p}[G[y]] = P{<=1-p}[F[!y]]
    P{<p}[G[y]]  = P{>1-p}[F[!y]]
    P{<=p}[G[y]] = P{>=1-p}[F[!y]]
    
    @param   root:      The root element of the ElementTree representing the
                        formula.
    @type    root:      etree._Element
    @param   namespace: The XML namespace of the considered logic, surrounded
                        by curly braces. Example:
                        "{http://compass.informatik.rwth-aachen.de/csl}"
    @type    namespace: string
    """
    # Use a shorter name to refer to the parser for formulae documents, which
    # is used to create new XML elements.
    parser = formulae_module.parser
    
    # Get all "g" elements in the ElementTree representing the formula in
    # depth first order and store them in a list.
    g_elements = list(root.iter(namespace + "g"))
    # Replace these "g" elements by corresponding "f" elements. Do this in
    # reversed order (i.e. bottom-up), to ensure correct results for nested G
    # formulae.
    g_elements.reverse()
    for g_element in g_elements:
        # Create a new element representing the sub-formula "!y", where y is
        # the sub-formula of the G formula.
        # The sub-formula of the G formula is represented by the last child of
        # the "g" element. Note that this child is not copied but moved, so it
        # will be removed from its original parent element.
        not_y = parser.makeelement(namespace + "not")
        not_y.append(copy.deepcopy(g_element[-1]))
        # Wrap the sub-formula "!y" in a "formula" element.
        not_y_wrapper = parser.makeelement(namespace + "formula")
        not_y_wrapper.append(not_y)
        
        # Create a new element representing the sub-formula "F[!y]".
        f_element = parser.makeelement(namespace + "f")
        # Reuse the bounds of the G operator if there are any.
        # Note that at this point, the sub-formula child of the "g" element
        # has already been removed.
        if len(g_element) > 0:
            # The G formula has at least one bound. Move it to the "f"
            # element. It will be removed from the "g" element.
            f_element.append(g_element[0])
            if len(g_element) > 0:
                # The G formula has an upper and a lower bound. Move the upper
                # bound element as well. It will be removed from the "g"
                # element.
                f_element.append(g_element[0])
        f_element.append(not_y_wrapper)

        # Get the superordinate "p" element of the "w" element.
        p_element = g_element.getparent()
        # Replace the comparison operator and the threshold of the "p" element
        # with their complements.
        _complement_p_element(p_element, namespace)
        # Replace the old path formula of the "p" element with the new one.
        # The path formula is represented by the third child element of the
        # "p" element.
        p_element[2] = f_element
    
    
def eliminate_f(root, namespace):
    """Eliminate the operator F ("finally") in the formula that is represented
    by the XML ElementTree rooted a the specified element.
    
    This is done by replacing sub-formulae that contain the operator F by
    semantically equivalent sub-formulae that do not contain F but U (strong
    until).
    
    The following equivalence holds:
    F[y] = true U y
    
    @param   root:      The root element of the ElementTree representing the
                        formula.
    @type    root:      etree._Element
    @param   namespace: The XML namespace of the considered logic, surrounded
                        by curly braces. Example:
                        "{http://compass.informatik.rwth-aachen.de/csl}"
    @type    namespace: string
    """
    # Use a shorter name to refer to the parser for formulae documents, which
    # is used to create new XML elements.
    parser = formulae_module.parser
    
    # Get all "f" elements in the ElementTree representing the formula in
    # depth first order and store them in a list.
    f_elements = list(root.iter(namespace + "f"))
    # Replace these "f" elements by corresponding "u" elements. Do this in
    # reversed order (i.e. bottom-up), to ensure correct results for nested F
    # formulae.
    f_elements.reverse()
    for f_element in f_elements:
        # Create a new element representing the formula "true" and wrap it in
        # a "formula" element.
        true_element = parser.makeelement(namespace + "true")
        true_wrapper = parser.makeelement(namespace + "formula")
        true_wrapper.append(true_element)
        
        # Wrap the element representing the sub-formula of the F formula in
        # a "formula" element. This is needed in order to add the sub-formula
        # to the U formula that will be created. The last child of the "f"
        # element represents its sub-formula. Note that this child element
        # is not copied but moved, so it will be removed from the "f" element.
        sub_formula_wrapper = parser.makeelement(namespace + "formula")
        sub_formula_wrapper.append(f_element[-1])
        
        # Create a new element representing the formula "true U y", where y is
        # the sub-formula of the F formula.
        true_until_y = parser.makeelement(namespace + "u")
        # Reuse the bounds of the G operator if there are any.
        # Note that at this point, the sub-formula child of the "g" element
        # has already been removed.
        if len(f_element) > 0:
            # The F formula has at least one bound. Move it to the "u"
            # element. It will be removed from the "f" element.
            true_until_y.append(f_element[0])
            if len(f_element) > 0:
                # The F formula has an upper and a lower bound. Move the upper
                # bound element as well. It will be removed from the "f"
                # element.
                true_until_y.append(f_element[0])
        # Add the sub-formulae of the U formula.
        true_until_y.append(true_wrapper)
        true_until_y.append(sub_formula_wrapper)
        
        # Replace the F formula with "true U y".
        f_element.getparent().replace(f_element, true_until_y)


def _complement_p_element(p_element, namespace):
    """Replace the comparison operator and the threshold of the specified "p"
    element with their complements.
    
    Example: The formula "P{<p}[y]" will become "P{>1-p}[y]"
    
    @param   p_element: The "p" element to be complemented. This element will
                        be modified.
    @type    p_element: etree._Element
    @param   namespace: The XML namespace of the considered logic, surrounded
                        by curly braces. Example:
                        "{http://compass.informatik.rwth-aachen.de/csl}"
    @type    namespace: string
    """
    # Use a shorter name to refer to the parser for formulae documents, which
    # is used to create new XML elements.
    parser = formulae_module.parser
    
    # Create a new element to represent the complement of the original
    # comparison operator.
    # The original comparison operator element is the child of the first
    # child of the "p" element.
    comparison_op_tag = p_element[0][0].xpath("local-name()")
    new_comparison_op_tag = _complement_operators[comparison_op_tag]
    new_comparison_op = parser.makeelement(namespace + new_comparison_op_tag)
    # Create a new element to represent the complement of the original
    # threshold value.
    # The original threshold value is contained in the child element (with
    # tag "value") of the second child of the "p" element.
    threshold_value = float(p_element[1][0].text)
    new_threshold_element = parser.makeelement(namespace + "value")
    new_threshold_element.text = str(1 - threshold_value)
    # Replace the original comparison operator element and threshold value
    # element with the new ones.
    p_element[0][0] = new_comparison_op
    p_element[1][0] = new_threshold_element
