# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================


"""This module provides the data structures that are used to represent
patterns documents. It also provides functions to parse patterns documents and
validate them against the adequate XML Schema Definition.

Patterns are "templates" for logical formulae. They contain placeholders which
must be replaced in order to obtain a formula.

Some of the classes of this module correspond to types of the XML Schema
documents that define the structure of valid patterns documents.
"""

import copy

from lxml import etree

from property_management.src import constants
from property_management.src import xml_tags
from property_management.src import errors
from property_management.src import util
import property_management.src.formulae as formulae_module


def parse_patterns_document(data_source):
    """Parses the patterns document from the specified data source and returns
    a representation of the document using the data structures defined in this
    module (and in formulae_module).
    
    The representation of the document will be in tree structure. For a valid
    document, the root element will be a PatternsDocument object.
    
    Note: The document will not be validated. This must be done explicitly
    using the function get_syntax_errors_in_document.
    
    @param   data_source: The source of the data to be parsed. This may be a
                          filename or a file-like object.
    @type    data_source: A string or a file-like object.
    @return: A tree representation of the document.
    @rtype:  PatternsDocument
    """
    # Use the parser that is created when this module is imported (see the end
    # of this file). It creates the adequate objects for the XML elements of
    # the document.
    # The result of the parsing process will be an etree._ElementTree object.
    # Return the root element of this tree (a PatternsDocument object).
    document_tree = etree.parse(data_source, parser)
    return document_tree.getroot()


def get_syntax_errors_in_document(patterns_document):
    """Validates the specified patterns document against the adequate XML
    Schema Definition and returns a list containing the syntactic errors.
    
    @return: A list containing the syntactic errors in the document. This list
             will be empty if the document is valid with respect to the
             adequate XML Schema Definition.
    @rtype:  A list of errors.DocumentSyntaxError objects. 
    """
    # Initialize the list of errors to be returned.
    error_list = []
    # Validate the document. If it is invalid, an exception will be raised
    # which can be used to create an error object with an adequate error
    # description.
    # Note: The XML Schema object that is used for validation is created when
    # this module is imported (see the end of this file).
    try:
        _patterns_schema.assertValid(patterns_document)
    except etree.DocumentInvalid, e:
        error_list.append(errors.DocumentSyntaxError(str(e)))
    
    return error_list


# ----------------------------------------------------------------------------
# Data structures for patterns documents.
# ----------------------------------------------------------------------------

class PatternsDocument(util.SerializableXmlElement):
    """A list of patterns for logical formulae.
    
    Patterns can be added to a patterns document using the append-method.
    
    Important: PatternsDocument objects must not be created like this:
    x = PatternsDocument()
    Instead, the static factory method create must be used.
    This is because this class is a subclass of etree.ElementBase, which must
    not be instantiated directly. Refer to the lxml documentation for details.
    """
    
    @staticmethod
    def create(patterns=None):
        """Creates and returns a new PatternsDocument object.
        
        @param   patterns: (Optional) The patterns to be added to the new
                           document. Note: Instead of the pattern objects
                           themselves, deep copies of these objects will be
                           added to the document.
        @type    patterns: An iterable over Pattern objects.
        @return: A new PatternsDocument object.
        @rtype:  PatternsDocument
        """
        # Note: The parser that is used to create the object is created when
        # this module is imported (see the end of this file).
        patterns_document = parser.makeelement(
                                    constants.PATTERNS_NAMESPACE +
                                    xml_tags.PATTERNS_DOCUMENT_TAG)
        if patterns is not None:
            for pattern in patterns:
                # Don't add the pattern object itself but rather a deep copy.
                # This is necessary for the following reason: The pattern
                # object is an XML element, which might already be used as a
                # child element of some other parent element (e.g. a patterns
                # document). Without copying, it would be removed from that
                # parent. See the lxml documentation.
                patterns_document.append(copy.deepcopy(pattern))
        return patterns_document
        
    def get_pattern_by_id(self, pattern_id):
        """Returns the pattern with the specified ID.
        
        @return: The pattern with the specified ID or None if no such pattern
                 exists.
        @rtype:  Pattern or NoneType
        """
        for pattern in self:
            if pattern.id == pattern_id:
                return pattern
        return None

    def __str__(self):
        """Returns a string representation of this document in XML format.
        
        UTF-8 encoding will be used and an XML header will be included.
        
        @return: A string representation of this document in XML format.
        @rtype:  string
        """
        return etree.tostring(self, xml_declaration=True, encoding="UTF-8",
                              pretty_print=True)


class Pattern(util.SerializableXmlElement):
    """A pattern for a formula in a certain logic.
    
    Pattern objects must never be created explicitly. They should only occur
    in the result of parsing a patterns document.
    """
    
    def get_logic(self):
        """Returns a constant identifying the logic that is used by this
        pattern.
        
        @return: A constant identifying the logic that is used by this
                 pattern.
        @rtype:  One of the constants in constants.Logic.
        @raises  Exception: If the logic could not be detected.
        """
        # This pattern is represented by an XML element. The logic is
        # determined by the tag of the child element representing the formula
        # template of the pattern.
        # TODO: Use a dictionary instead of if-elif.
        tag_name = self.formula_template.xpath("local-name()")
        if tag_name == xml_tags.LTL_FORMULA_TEMPLATE_TAG:
            return constants.Logic.LTL
        elif tag_name == xml_tags.CTL_FORMULA_TEMPLATE_TAG:
            return constants.Logic.CTL
        elif tag_name == xml_tags.MITL_FORMULA_TEMPLATE_TAG:
            return constants.Logic.MITL
        elif tag_name == xml_tags.PCTL_FORMULA_TEMPLATE_TAG:
            return constants.Logic.PCTL
        elif tag_name == xml_tags.CSL_FORMULA_TEMPLATE_TAG:
            return constants.Logic.CSL
        else:
            # TODO: Use a custom exception class.
            raise Exception(('The logic could not be detected for the tag ' +
                             'name "%s".') % tag_name)
            
    def _get_formula_template(self):
        # This pattern is represented by an XML element. Return the child
        # element representing its formula template. There must be exactly one
        # such element.
        # Note: There are multiple XML tags that might be used for the formula
        # template element, depending on the logic.
        formula_template_tags = [xml_tags.LTL_FORMULA_TEMPLATE_TAG,
                                 xml_tags.CTL_FORMULA_TEMPLATE_TAG,
                                 xml_tags.MITL_FORMULA_TEMPLATE_TAG,
                                 xml_tags.PCTL_FORMULA_TEMPLATE_TAG,
                                 xml_tags.CSL_FORMULA_TEMPLATE_TAG]
        local_names = ["local-name()='%s'" % tag \
                       for tag in formula_template_tags]
        formula_templates = self.xpath("*[%s]" % " or ".join(local_names))
        assert len(formula_templates) == 1
        return formula_templates[0]
    
    formula_template = property(
            _get_formula_template,
            doc="The formula template of this pattern.\n" +
                "\n" +
                "The formula template is a formula containing " +
                "placeholders. It is represented by an XML element. Note " +
                "that this is not a formulae_module.Formula object because " +
                "formula templates contain placeholders and thus are not " +
                "valid formulae. Replacing the placeholders by adequate " +
                "XML fragments will yield a valid formula.\n" +
                "\n" +
                "@type: etree._Element")
    # TODO: Add a class for formula templates.
            
    def get_story(self):
        """Returns the story for this pattern or None if there is no story.
        
        A story is a description of the formula pattern including information
        about its placeholders. A story is represented by a list of strings
        and PlaceholderInfo objects. Example:
        ['The atomic proposition ', <PlaceholderInfo object>,
         'holds until time ', <PlaceholderInfo object>]
         
        @return: The story for this pattern or None.
        @rtype:  A list of strings and PlaceholderInfo objects or NoneType.
        @raises  Exception: If the story of the pattern contains unknown kinds
                    of placeholders.
        """
        # This pattern is represented by an XML element. Get the child element
        # containing the XML representation of the story. There must be
        # exactly one or no such element.
        story_elements = self.xpath("*[local-name()='%s']"
                                    % xml_tags.PATTERN_STORY_TAG)
        assert len(story_elements) <= 1
        # Return None if this pattern has no story.
        if len(story_elements) == 0:
            return None
        story_element = story_elements[0]
        # The story XML element has mixed content, that means it contains
        # text with interspersed XML elements which represent the
        # placeholders.
        # Define a dictionary to map the tags of the XML elements inside the
        # story to the corresponding constants indicating the type of the
        # placeholder.
        placeholder_info_tag_to_constant_dict = {
                xml_tags.STATE_FORMULA_PLACEHOLDER_INFO_TAG:
                    constants.PlaceholderType.STATE_FORMULA,
                xml_tags.PATH_FORMULA_PLACEHOLDER_INFO_TAG:
                    constants.PlaceholderType.PATH_FORMULA,
                xml_tags.TIME_BOUND_PLACEHOLDER_INFO_TAG:
                    constants.PlaceholderType.TIME_BOUND,
                xml_tags.THRESHOLD_PLACEHOLDER_INFO_TAG:
                    constants.PlaceholderType.THRESHOLD,
                xml_tags.COMPARISON_OPERATOR_PLACEHOLDER_INFO_TAG:
                    constants.PlaceholderType.COMPARISON_OPERATOR}
        # In the list returned by this method we will normalize the text
        # content. This means that all whitespace characters will be replaced
        # by spaces and sequences of whitespaces will be collapsed to just one
        # space.
        # TODO: Revise this decision.
        def _normalize_string(some_string):
            # Exploit the fact that the split-method considers arbitrary
            # sequences of whitespace as separators.
            return " ".join(some_string.split())
        
        # Create a list representing the story.
        # The first part of the text insided the story XML element up to the
        # first placeholder information XML element in the story is accessible
        # through the "text"-attribute of the story element. The elements
        # representing the placeholders are available as the children of the
        # story element. The text between these child elements and the text
        # after the last of these child elements is accessible through the
        # "tail"-attribute of the child element preceding the respective part
        # of the text.
        story_list = []
        if story_element.text:  # Might be None or "".
            story_list.append(_normalize_string(story_element.text))
        for placeholder_info_element in story_element:
            # Add PlaceholderInfo object to represent the current placeholder
            # in the story. It provides information about the type, logic and
            # ID of the placeholder.
            tag = placeholder_info_element.xpath("local-name()")
            try:
                placeholder_type = placeholder_info_tag_to_constant_dict[tag]
            except KeyError:
                # The tag is not a known placeholder info tag.
                # TODO: Use a custom exception class.
                raise Exception('Unexpected kind of placeholder: "%s".' % tag)
            story_list.append(PlaceholderInfo(
                                    self.get_logic(),
                                    placeholder_type,
                                    placeholder_info_element.get("id")))
            # Add the text after the current placeholder info element if there
            # is any text.
            if placeholder_info_element.tail:  # Might be None or "".
                story_list.append(
                        _normalize_string(placeholder_info_element.tail))

        return story_list
    
    # The ID of this pattern is stored as an attribute of this XML element.
    def _get_id(self):
        return self.get("id")
    
    def _set_id(self, new_id):
        self.set("id", new_id)
        
    id = property(_get_id, _set_id, doc="The ID of this pattern.\n" +
                                        "@type: string")
        

class PlaceholderInfo(object):
    """A representation of a placeholder providing information about its
    logic, type and ID.
    """
    
    def __init__(self, logic, placeholder_type, placeholder_id):
        """Initializes the PlaceholderInfo object.
        
        @param   logic:            The logic of the placeholder.
        @type    logic:            One of the constants in constants.Logic.
        @param   placeholder_type: The type of the placeholder.
        @type    placeholder_type: One of the constants in
                                   constants.PlaceholderType.
        @param   placeholder_id:   The ID of the placeholder.
        @type    placeholder_id:   string
        """
        self.logic = logic
        self.type = placeholder_type
        self.id = placeholder_id

# ----------------------------------------------------------------------------
# End of the definition of the data structures.
# ----------------------------------------------------------------------------


def _get_parser_lookup_scheme():
    """Returns a new parser lookup scheme which is used to map certain XML
    elements in patterns documents to instances of the adequate classes
    defined in this module and in the formulae-module.
    
    @return: A parser lookup scheme for patterns documents.
    @rtype:  etree.ElementNamespaceClassLookup
    """
    # Extend the lookup scheme for formulae documents so that the data
    # structures for properties documents and formulae documents are
    # compatible and can be combined.
    lookup = formulae_module._get_parser_lookup_scheme()
    patterns_namespace = lookup.get_namespace(constants.PATTERNS_NAMESPACE_URI)
    patterns_namespace[xml_tags.PATTERNS_DOCUMENT_TAG] = PatternsDocument
    patterns_namespace[xml_tags.PATTERN_TAG] = Pattern
    return lookup


# ----------------------------------------------------------------------------
# Note: The following code will be executed once when this module is imported.
# ----------------------------------------------------------------------------

# Initialize the parser that is used to parse patterns documents.
# XML comments must be discarded. Otherwise the parser would yield data
# structures cluttered with unwanted elements.
parser = etree.XMLParser(remove_comments=True)
# A lookup scheme is used in order to automatically map certain XML elements
# to instances of the adequate classes.
parser.set_element_class_lookup(_get_parser_lookup_scheme())

# Parse the XML Schema document that defines the structure of valid patterns
# documents. This is needed for validation.
_patterns_schema_tree = etree.parse(constants.PATTERNS_XSD_PATH)
_patterns_schema = etree.XMLSchema(_patterns_schema_tree)
