# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================


"""This module provides general utilities which are used by property
management.
"""

from lxml import etree


class SerializableXmlElement(etree.ElementBase):
    """An element of an XML document that can be serialized using the method
    get_as_xml_string.
    
    Important: SerializableXmlElement objects must not be created like this:
    x = SerializableXmlElement()
    Instead, they must be created using an lxml.etree parser.
    This is because this class is a subclass of etree.ElementBase, which must
    not be instantiated directly. Refer to the lxml documentation for details.
    """
    
    def get_as_xml_string(self):
        """Returns the XML fragment rooted at this element as a string.
        
        Pretty printing and UTF-8 encoding will be used.
        
        @return: The XML fragment rooted at this element.
        @rtype:  string
        """
        return etree.tostring(self, xml_declaration=False, encoding="UTF-8",
                              pretty_print=True)


class CachingDecorator(object):
    """A function object that behaves like a given function but caches the
    return values.
    
    Important: This does not work with keyword arguments or unhashable
    arguments!
    """

    def __init__(self, original_function): 
        """Initializes the CachingDecorator object.
        
        @param   original_function: The function to be decorated.
        @type    original_function: A callable object.
        @return: The caching version of the specified function.
        @rtype:  A callable object.
        """
        # Save the specified original function and initialize the cache.
        self.original_function = original_function
        self.args_to_cached_return_value_dict = {} 
        
    def __call__(self, *args):
        """Behaves like the original function but returns cached return values
        immediately.
        
        The original function will only be called if the return value for the
        specified arguments is not yet in the cache.
        
        @param   *args: The (positional) arguments.
        @type    *args: A sequence.
        @return: The corresponding return value of the original function or
                 the cached value if available.
        @rtype:  The type of the return value of the original function.
        """
        if args not in self.args_to_cached_return_value_dict: 
            self.args_to_cached_return_value_dict[args] = \
                    self.original_function(*args)
        return self.args_to_cached_return_value_dict[args]


def caching_function(original_function):
    """A function decorator that returns a caching version of the specified
    function.
    
    See CachingDecorator for details.
    
    @param   original_function: The function to be decorated.
    @param   original_function: A callable object.
    @return: The caching version of the specified function.
    @rtype:  A callable object.
    """
    return CachingDecorator(original_function)
