##  ===========================================================================
##  Author: Roberto Cavada <cavada@fbk.eu>
##
##  Copyright (c) 2005 by FBK 
##
##  This is free software; you can redistribute it and/or
##  modify it under the terms of the GNU Lesser General Public
##  License as published by the Free Software Foundation; either
##  version 2 of the License, or (at your option) any later version.
##
##  This software is distributed in the hope that it will be useful,
##  but WITHOUT ANY WARRANTY; without even the implied warranty of
##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
##  Lesser General Public License for more details.
##
##  You should have received a copy of the GNU Lesser General Public
##  License along with this library; if not, write to the Free Software
##  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA.
##
##  For more information about this software see <http://rat.fbk.eu>
##  or email to the RAT Team <rat@fbk.eu>.
##  Please report bugs to <rat@fbk.eu>.
##
##  ===========================================================================


from gtkmvc.controller import Controller

from controllers.project import ProjectCtrl
from models.application import ApplicationWizardModel

import utils.dialogs
import os.path
import gtk
import gobject
import utils.globals

class ApplicationCtrl (Controller):

    """This is the main application's controller. Mainly this is
    container for the project's controller, and controls menus,
    toolbars, etc.  """

    def __init__(self, model):
        Controller.__init__(self, model)

        self.toggling_view = False
        return
          
    def create_new_project(self, active_view=None):
        """Creates a new project, by creating a new model, view and
        controller for a project"""

        # hides the traces if there are shown traces:
        traces_are_shown = self.model.traces_window_visible 
        self.model.traces_window_visible = False
        
        # model
        self.model.new_project()
        self.model.project.registerObserver(self) # we observe the project model

        # ctrl and view
        self.project = ProjectCtrl(self.model.project, self)
        self.view.new_project(self.project)

        self.view.set_title("RAT - Unnamed Project")

        # set the active view
        if active_view is None:
            self.model.project.active_view = "pa"
            #self.view['rb_prop_assurance'].set_active(True)            
        else: self.model.project.active_view = active_view

        self.model.project.is_new=True

        # show the traces if enabled:
        self.model.traces_window_visible = traces_are_shown
        return


    def quit_application(self):
        if self.model.project.needs_to_be_saved:
            # asks confirmation
            ok = utils.dialogs.open_confirm_action_dialog("Quit",
                                                          "Do you want to quit?",
"""Current project has not been saved, and you
requested to quit the application.

If you choose to quit, any unsaved modification will be lost.""")
            if not ok: return
            pass
        
        self.model.quit()
        gtk.main_quit()
        return


    def ask_to_discharge_project(self):
        return utils.dialogs.open_confirm_action_dialog("Discharge",
                                                        "Discharge current project?",
"""Current project has not been saved, and you
requested to create a new one.

If you choose to discharge the current project,
any unsaved modification will be lost.""")        
    
    def open_new_project_wizard(self):
        if self.model.project.needs_to_be_saved:
            if not self.ask_to_discharge_project(): return
            pass

        m = ApplicationWizardModel();

        c = ApplicationWizardCtrl(m, self.model, self)
        from views.application import ApplicationWizardView
        v = ApplicationWizardView(c)
        return


 
    # ------------------------------------------------------------ #
    #                       Signals Handling                       #
    # ------------------------------------------------------------ #
    def on_rb_prop_assurance_toggled(self, wid):
        if self.toggling_view: return
        if isinstance(wid, gtk.Bin): wid = wid.get_children()[0]
        if wid.get_active():
            self.toggling_view = True
            self.model.project.active_view = "pa"
            self.toggling_view = False
            pass
        return

    def on_rb_prop_simulation_toggled(self, wid):
        if self.toggling_view: return
        if isinstance(wid, gtk.Bin): wid = wid.get_children()[0]
        if wid.get_active():
            self.toggling_view = True
            self.model.project.active_view = "ps"
            self.toggling_view = False
            pass
        return   

    def on_rb_realizability_toggled(self, wid):
        if self.toggling_view: return
        if isinstance(wid, gtk.Bin): wid = wid.get_children()[0]
        if wid.get_active():
            self.toggling_view = True
            self.model.project.active_view = "re"
            self.toggling_view = False
            pass
        return

    def on_quit_activate(self, item):
        self.quit_application()
        return        

    def on_window_main_delete_event(self, window, event):
        self.quit_application()
        return True

    def on_toolbutton_save_clicked(self, button):
        self.model.project.save()
        return

    def on_toolbutton_open_clicked(self, button):
        if self.model.project.needs_to_be_saved:
            if not self.ask_to_discharge_project(): return
            pass

        fn = utils.dialogs.open_project_dialog(self.model.options.last_visited_path)
        if not fn: return

        fn = utils.utilities.normalize_project_filename(fn)
                
        if os.path.isfile(fn):
            # to improve apparent responsiveness:
            while gtk.events_pending(): gtk.main_iteration(False)
            
            self.create_new_project()
            self.model.project.open(fn)
            pass

        return

    def on_project_wizard_activate(self, item): self.open_new_project_wizard()

    def on_pa_project_activate(self, item):
        if self.model.project.needs_to_be_saved:
            if not self.ask_to_discharge_project(): return
            pass
        self.create_new_project("pa")
        return
        
    def on_ps_project_activate(self, item):
        if self.model.project.needs_to_be_saved:
            if not self.ask_to_discharge_project(): return
            pass
        self.create_new_project("ps")
        return        

    def on_toolbutton_new_clicked(self, button): self.open_new_project_wizard()

    def on_save_as_activate(self, item):
        self.model.project.save_as()
        return

    def on_item_show_traces_toggled(self, item):
        if item.get_active() and self.model.traces_window_visible: return
        self.model.traces_window_visible = item.get_active()
        return

    def on_tb_show_traces_toggled(self, button):
        if button.get_active() and self.model.traces_window_visible: return
        self.model.traces_window_visible = button.get_active()
        return

    def on_about_activate(self, item):
        from views.application import ApplicationAboutView
        c = ApplicationAboutCtrl(self.model)
        v = ApplicationAboutView(c)
        v.run()
        self.model.unregisterObserver(c)
        return
    

    # ------------------------------------------------------------ #
    #                       User Callbacks                         #
    # ------------------------------------------------------------ #
    def update_busy_progress(self):
        if self.model.is_busy: self.view.tick_progress_bar()
        return self.model.is_busy


    # ------------------------------------------------------------ #
    #            Notifications from observable properties          #
    # ------------------------------------------------------------ #
    def property_filename_change_notification(self, project, old, new):        
        if new != old:
            fn = os.path.basename(new)
            title, ext = os.path.splitext(fn)
            if ext.lower() != utils.globals.PROJECT_FILE_EXTENSION.lower(): title = fn
            self.view.set_title("RAT - " + title)
        return


    def property_needs_to_be_saved_change_notification(self, project, old, new):        
        if new != old:
            # enables/disable save buttons
            self.view['menuitem_save'].set_sensitive(new)
            self.view['toolbutton_save'].set_sensitive(new)
            pass
        return


    def property_is_new_change_notification(self, model, old, new):
        # enable/disable "save as" and "new" buttons and menu
        self.view['menuitem_save_as'].set_sensitive(not new)
        self.view['menuitem_new'].set_sensitive(not new)
        self.view['toolbutton_new'].set_sensitive(not new)
        return


    def property_is_busy_change_notification(self, model, old, new):
        if new == old: return

        self.view.show_progress_bar(new)
        if new: gobject.timeout_add(150, self.update_busy_progress) #installs a timer
        return

    def property_traces_window_visible_change_notification(self, project, old, new):
        if old == new: return
        items = (self.view['tb_show_traces'], self.view['item_show_traces'])
        for item in items:
            if item.get_active() != new: item.set_active(new)
            pass

        if self.model.project.traces_window_shown != new:
            self.model.project.traces_window_shown = new
            pass
        
        return

    def property_active_view_change_notification(self, project, old, new):
        if self.toggling_view: return
        self.toggling_view = True        
        if new in ("pa", None): self.view['rb_prop_assurance'].set_active(True)
        #elif new == "ps": self.view['rb_prop_simulation'].set_active(True)
        #elif new == "re": self.view['rb_realizability'].set_active(True)

        self.toggling_view = False
        return
            
    pass # end of class



# ----------------------------------------------------------------------
class ApplicationAboutCtrl (Controller):

    def __init__(self, model):
        Controller.__init__(self, model)

        return

    # -------------------------------------------------
    #    GTK signals                                  #
    # -------------------------------------------------    

    def on_expander_credits_activate(self, expander):
        box = expander.get_parent()
        box.set_child_packing(expander, not expander.get_expanded(), True, 0, gtk.PACK_START)
        return

    def on_expander_license_activate(self, expander):
        box = expander.get_parent()
        box.set_child_packing(expander, not expander.get_expanded(), True, 0, gtk.PACK_START)
        return   

    pass # end of class



# ----------------------------------------------------------------------
class ApplicationWizardCtrl (Controller):

    def __init__(self, model, appl_model, parent):
        Controller.__init__(self, model)

        self.appl_model = appl_model
        self.parent = parent
        return

    def registerView(self, view):
        Controller.registerView(self, view)
        view['rb_pa'].toggled()
        return

    # -------------------------------------------------
    #    GTK signals                                  #
    # -------------------------------------------------    
    def on_button_next_clicked(self, button):
        nb = self.view['notebook']
        max_page = nb.get_n_pages() - 1
        page = nb.get_current_page()

        if page < max_page:
            nb.set_current_page(page+1)
            page += 1
            pass

        button.set_sensitive(page < max_page)
        self.view['button_prev'].set_sensitive(page > 0)
        return


    def on_button_prev_clicked(self, button):
        nb = self.view['notebook']
        max_page = nb.get_n_pages() - 1
        page = nb.get_current_page()

        if page > 0:
            nb.set_current_page(page-1)
            page -= 1
            pass

        self.view['button_next'].set_sensitive(page < max_page)
        button.set_sensitive(page > 0)
        return

    def on_button_browse_clicked(self, button):
        fn = utils.dialogs.open_project_dialog(
            self.appl_model.options.last_visited_path)
        self.view['entry_name'].set_text(fn)
        return

    def on_button_cancel_clicked(self, button):
        self.model.unregisterObserver(self)
        self.view.destroy()
        return

    def on_rb_type_toggled(self, rb):
        """Sets the selected project type description by taking it
        from the model"""
        for t in 'pa ps re'.split():
            if rb == self.view['rb_'+t] and rb.get_active():
                self.view.set_description(self.model.get_description(t))
                break
            pass
        return    
    
    def on_button_finish_clicked(self, button):
        fn = self.view['entry_name'].get_text().strip()
        if fn:
            fn = utils.utilities.normalize_project_filename(fn) # fix extention
            if os.path.isfile(fn):
                # requests to confirm overwrite:
                ovrw = utils.dialogs.open_confirm_action_dialog("Overwrite",
                              'Overwrite "%s"?' % os.path.basename(fn),
                              """A file with this name already exists.
If you choose to overwrite this file, the contents will be lost.""")
                
                if not ovrw: return
                pass
            pass
        
        buf = self.view['textview_notes'].get_buffer()
        notes = buf.get_text(buf.get_start_iter(), buf.get_end_iter())
        is_pa = self.view['rb_pa'].get_active()
        is_ps = self.view['rb_ps'].get_active()
        
        self.model.unregisterObserver(self)
    
        self.view.destroy()
        if is_pa: self.parent.create_new_project('pa')
        elif is_ps: self.parent.create_new_project('ps')
        else: self.parent.create_new_project('re')

        self.appl_model.project.notes = notes
        
        if fn:
            self.appl_model.project.filename = fn
            self.appl_model.project.save()
            pass

        return

    pass # end of class
