##  ===========================================================================
##  Author: Roberto Cavada <cavada@fbk.eu>
##
##  Copyright (c) 2005 by FBK 
##
##  This is free software; you can redistribute it and/or
##  modify it under the terms of the GNU Lesser General Public
##  License as published by the Free Software Foundation; either
##  version 2 of the License, or (at your option) any later version.
##
##  This software is distributed in the hope that it will be useful,
##  but WITHOUT ANY WARRANTY; without even the implied warranty of
##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
##  Lesser General Public License for more details.
##
##  You should have received a copy of the GNU Lesser General Public
##  License along with this library; if not, write to the Free Software
##  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA.
##
##  For more information about this software see <http://rat.fbk.eu>
##  or email to the RAT Team <rat@fbk.eu>.
##  Please report bugs to <rat@fbk.eu>.
##
##  ===========================================================================


from gtkmvc.controller import Controller

import gtk

## ----------------------------------------------------------------------
class CheckingPanelCtrl (Controller):
    
    """Controller for the checking panel view embedded within the
    property assurance view"""

    def __init__(self, model, parent):
        Controller.__init__(self, model)
        self.parent = parent
        self.automatic_editing = False

        self.auto_changing_loop = False
        self.switched_from_loop = None

        model.use_tech = "sat"
        return

    def registerView(self, view):
        Controller.registerView(self, view)
        self.setup_view(view)
        return

    def setup_view(self, view):
        # SAT Settings
        cb = view['cb_solver']
        cb.set_model(self.model.solvers)

        view['cb_solver'].set_active(0)
        view['cb_loop'].set_active(0)
        view['sb_length'].set_value(30)

        # BDD Settings
        view['cb_part'].set_active(1)

        view['cb_method'].set_active(0)
        view['cb_method'].set_sensitive(True)
        
        return

    # ------------------------------------------------------------ #
    #                            Signals                           #
    # ------------------------------------------------------------ #
    
    def on_button_check_clicked(self, button):
        # updates the model:
        check_req = self.view['cb_check_reqs_consistency'].get_active()
        self.parent.check(check_req)
        return

    def on_notebook_options_switch_page(self, nb, page, page_num):
        if page_num == 0: self.model.use_tech = "sat"
        else: self.model.use_tech = "bdd"
        return

    def on_cb_inc_toggled(self, cb):
        self.model.use_inc = cb.get_active()
        self.view.enable_inc_sbmc_stuff(self.model.use_inc and self.model.use_sbmc)
        self.view.enable_loop(not (self.model.use_inc and self.model.use_sbmc))
        return

    def on_cb_sbmc_toggled(self, cb):
        self.model.use_sbmc = cb.get_active()
        self.view.enable_inc_sbmc_stuff(self.model.use_inc and self.model.use_sbmc)
        self.view.enable_loop(not (self.model.use_inc and self.model.use_sbmc))        
        return

    def on_cb_comp_toggled(self, cb):
        self.model.use_comp = cb.get_active()
        return

    def on_cb_vu_toggled(self, cb):
        self.model.use_vu = cb.get_active()
        return

    def on_cb_solver_changed(self, cb):
        m = cb.get_model()
        can_be_inc = m.get_value(cb.get_active_iter(), 1)
        self.model.use_inc = can_be_inc and self.view['cb_inc'].get_active()
        self.model.solver = m.get_value(cb.get_active_iter(), 0)

        self.view.enable_inc_stuff(can_be_inc)
        self.view.enable_inc_sbmc_stuff(can_be_inc and self.model.use_sbmc)
        self.view.enable_loop(not (self.model.use_inc and self.model.use_sbmc)) 
        return

    def on_cb_loop_changed(self, cb):
        sbl = self.view['sb_loop']
        allow_value = cb.get_active() == 2
        sbl.set_sensitive(allow_value)

        if allow_value: 
            length = self.view['sb_length'].get_value_as_int()
            cloop = sbl.get_value_as_int()
            sbl.set_range(-length, length-1)
            val = min(cloop, length-1)
            sbl.set_value(val)
            self.model.loop = str(val)
            
        elif cb.get_active() == 0: self.model.loop = "*"
        else: self.model.loop = "X"
            
        if not self.auto_changing_loop: self.switched_from_loop = None
        return

    def on_sb_length_changed(self, sb):

        try: val = int(sb.get_text())
        except ValueError: val = 0
        
        self.model.length = val

        cbl = self.view['cb_loop']
        if val == 0 and self.switched_from_loop is None:
            self.switched_from_loop = cbl.get_active()
            self.auto_changing_loop = True
            cbl.set_active(1)
            cbl.set_sensitive(False)
            self.auto_changing_loop = False
            return
        
        if val > 0 and self.switched_from_loop is not None:
            cbl.set_active(self.switched_from_loop)
            cbl.set_sensitive(True)
            self.switched_from_loop = None
            pass

        if cbl.get_active() == 2:
            sbl = self.view['sb_loop']
            cloop = sbl.get_value_as_int()
            sbl.set_range(-val, val-1)
            sbl.set_value(min(cloop, val-1))
            pass
            
        return

    def on_sb_loop_changed(self, sb):
        self.model.loop = str(sb.get_value_as_int())
        return
        

    def on_cb_dyn_reord_toggled(self, cb):
        val = cb.get_active()
        self.model.use_dyn_reord = val
        self.view['cb_method'].set_sensitive(val)
        return

    def on_cb_use_coi_toggled(self, cb): self.model.use_coi = cb.get_active()

    def on_cb_part_changed(self, cb):
        m = cb.get_model()
        self.model.part_meth = m.get_value(cb.get_active_iter(), 0)
        return

    def on_cb_method_changed(self, cb):
        m = cb.get_model()
        meth = m.get_value(cb.get_active_iter(), 0)
        self.model.reord_meth = meth.lower().replace(" ", "_")
        return      
        

    # ------------------------------------------------------------ #
    #                       Notifications                          #
    # ------------------------------------------------------------ #


    # ---------------------------------------- #
    #   Notifications for observed properties  #
    # ---------------------------------------- #

    def property_busy_level_change_notification(self, model, old, new):
        if old != new: self.view['button_check'].set_sensitive(new == 0)
        return            

    pass # end of class
