##  ===========================================================================
##  Author: Roberto Cavada <cavada@fbk.eu>
##
##  Copyright (c) 2005 by FBK 
##
##  This is free software; you can redistribute it and/or
##  modify it under the terms of the GNU Lesser General Public
##  License as published by the Free Software Foundation; either
##  version 2 of the License, or (at your option) any later version.
##
##  This software is distributed in the hope that it will be useful,
##  but WITHOUT ANY WARRANTY; without even the implied warranty of
##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
##  Lesser General Public License for more details.
##
##  You should have received a copy of the GNU Lesser General Public
##  License along with this library; if not, write to the Free Software
##  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA.
##
##  For more information about this software see <http://rat.fbk.eu>
##  or email to the RAT Team <rat@fbk.eu>.
##  Please report bugs to <rat@fbk.eu>.
##
##  ===========================================================================

from gtkmvc.controller import Controller
from utils.widgets import ParenMatchTextBuffer, get_approx_chars_num
from utils.strings import indent_formula
from views.entry import SignalTypeView
import gtk
import gobject



class EntryCtrl (Controller):
    def __init__(self, model, parent_ctrl=None):  
        Controller.__init__(self, model)
        self.parent_ctrl = parent_ctrl
        return

    def registerView(self, view):
        """Sets up the initial values by taking them from the model"""
        Controller.registerView(self, view)

        buf = ParenMatchTextBuffer()
        self.view.get_textview_other().set_buffer(buf)
        buf.connect('changed', self.on_textview_what_buffer_changed)
        
        self.view.set_name(self.model.name)
        self.view.set_notes(self.model.notes)
        return

    def update_model(self):        
        """Aligns the model content with the data in the view. This
        can be done only when the model can be added successfully to
        the set of entries. The check is performed by the EntriesCtrl
        that is self's parent controller."""
        
        if self.model.name != self.view.get_name(): self.model.name = self.view.get_name()

        if self.model.notes != self.view.get_notes():
            self.model.notes = self.view.get_notes()
            return

    # ------------------------------------------------------------ #
    #                            Signals                           #
    # ------------------------------------------------------------ #

    def on_entry_name_changed(self, entry): self.view.set_window_title()

    def on_window_main_delete_event(self, window, event):
        if self.parent_ctrl: self.parent_ctrl.rollup_entry(self.model)
        else: gtk.main_quit()
        return True

    def on_button_cancel_clicked(self, button):
        """Exits without any further action"""
        if self.parent_ctrl: self.parent_ctrl.rollup_entry(self.model)
        else: gtk.main_quit()
        return

    def on_textview_what_move_cursor(self, tv, step_size, count, extend_selection):
        gobject.idle_add(tv.get_buffer().show_paren_match)
        return
    
    def on_textview_what_button_press_event(self, tv, event):
        gobject.idle_add(tv.get_buffer().show_paren_match)
        return
    
    pass # end of class


## ----------------------------------------------------------------------
class SignalCtrl (EntryCtrl):
    def __init__(self, model, parent_ctrl=None):
        EntryCtrl.__init__(self, model, parent_ctrl)
        self.auto_switching_to_other = False
        self.signal_help = SignalTypeCtrl(model)
        return

    def registerView(self, view):
        EntryCtrl.registerView(self, view)        
        self.view.set_type(self.model.type)
        self.view.set_kind(self.model.kind)
        return

    def update_model(self):
        EntryCtrl.update_model(self)
        if self.model.type != self.view.get_type():
            self.model.type = self.view.get_type()
            pass
        if self.model.kind != self.view.get_kind():
            self.model.kind = self.view.get_kind()
            pass
        return


    # ------------------------------------------------------------ #
    #                            Signals                           #
    # ------------------------------------------------------------ #
    def on_rb_bool_toggled(self, button):
        if button.get_active(): self.view.enable_ok_button(True)
        return

    def on_rb_other_toggled(self, button):
        txt = self.view.get_text_other().strip()
        self.view.enable_ok_button(len(txt) > 0)
        if button.get_active() and not self.auto_switching_to_other:
            self.view.get_textview_other().grab_focus()
            pass
        
        return

    def on_textview_what_buffer_changed(self, buffer):
        txt = self.view.get_text_other().strip()
        
        if self.view['rb_bool'].get_active() and txt:
            self.auto_switching_to_other = True
            self.view['rb_other'].set_active(True)
            self.auto_switching_to_other = False
            return

        if self.view['rb_other'].get_active():
            self.view.enable_ok_button(len(txt) > 0)
            pass
        
        return

    def on_button_ok_clicked(self, button):
        self.signal_help.hide()
        if self.parent_ctrl: self.parent_ctrl.commit_entry(self.model)        
        return

    def on_window_main_delete_event(self, window, event):
        self.signal_help.hide()
        EntryCtrl.on_window_main_delete_event(self, window, event)
        return True

    def on_button_cancel_clicked(self, button):
        self.signal_help.hide()
        EntryCtrl.on_button_cancel_clicked(self, button)
        return


    def on_button_help_types_clicked(self, button):
        self.signal_help.show()
        return

    pass # end of class


## ----------------------------------------------------------------------
class SignalTypeCtrl (Controller):   
    """This is used to show the helper for signal types"""
    def show(self):
        if self.view is None: 
            view = SignalTypeView(self)
            view.show(True)
            pass

        else: self.view.show(True)
        return

    def hide(self):
        if self.view is not None: self.view.show(False)
        return
    
    def on_window_signal_help_delete_event(self, window, event):
        self.hide()
        return True
    
    pass # end of class


## ----------------------------------------------------------------------
class PropertyCtrl (EntryCtrl):
    def __init__(self, model, parent_ctrl=None):
        EntryCtrl.__init__(self, model, parent_ctrl)
        return

    def registerView(self, view):
        EntryCtrl.registerView(self, view)
        self.view.set_property(self.model.property)
        return

    def update_model(self):
        EntryCtrl.update_model(self)
        if self.model.property != self.view.get_property():
            self.model.property = self.view.get_property()
            pass
        return

    # ------------------------------------------------------------ #
    #                            Signals                           #
    # ------------------------------------------------------------ #
    def on_textview_what_buffer_changed(self, buffer):
        self.view.enable_ok_button(self.view.get_property() != "")
        return

    def on_button_ok_clicked(self, button):
        """Saves the model and exists"""
        if self.parent_ctrl: self.parent_ctrl.commit_entry(self.model)
        return

    def on_button_indent_clicked(self, button):
        tv = self.view.get_textview_other()
        width = tv.allocation.width        
        chars = get_approx_chars_num(tv, width)
        chars *= 0.8

        formula = indent_formula(self.view.get_property(), chars)
        self.view.set_property(formula)
        return

    pass # end of class


class RequirementCtrl (PropertyCtrl):
    def __init__(self, model, parent_ctrl=None):
        PropertyCtrl.__init__(self, model, parent_ctrl)
        return

    def registerView(self, view):
        PropertyCtrl.registerView(self, view)
        self.view.set_kind(self.model.kind)
        return

    def update_model(self):
        PropertyCtrl.update_model(self)
        if self.model.kind != self.view.get_kind():
            self.model.kind = self.view.get_kind()
            pass

        return

    pass #end of class
