##  ===========================================================================
##  Authors: Roberto Cavada <cavada@fbk.eu>
##           Ingo Pill <ipill@ist.tugraz.at>
##
##  Copyright (c) 2005 by FBK
##  Copyright (c) 2005 by Graz University of Technology
##
##  This is free software; you can redistribute it and/or
##  modify it under the terms of the GNU Lesser General Public
##  License as published by the Free Software Foundation; either
##  version 2 of the License, or (at your option) any later version.
##
##  This software is distributed in the hope that it will be useful,
##  but WITHOUT ANY WARRANTY; without even the implied warranty of
##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
##  Lesser General Public License for more details.
##
##  You should have received a copy of the GNU Lesser General Public
##  License along with this library; if not, write to the Free Software
##  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA.
##
##  For more information about this software see <http://rat.fbk.eu>
##  or email to the RAT Team <rat@fbk.eu>.
##  Please report bugs to <rat@fbk.eu>.
##
##  ===========================================================================


from gtkmvc.controller import Controller
from controllers.entries import SignalsCtrl, RequirementsCtrl
from controllers.prop_assurance import PropertyAssuranceCtrl
from controllers.prop_simulation import PropertySimulationCtrl
from controllers.realizability import RealizabilityCtrl
from controllers.traces import TracesCtrl

from utils.widgets import ParenMatchTextBuffer
import utils.dialogs

import os.path
import gtk
import gobject

class ProjectCtrl (Controller):

    """This is the project-level controller, that connects the
    project-level model and the project-level view.
    
    This controller contains also a few controllers, that incapsulate
    the controller of respectively the set of Signals, Requirements,
    Property Assurance (PA) and Property Simulation (PS).

    Those have designed to be available as stand alone entities,
    to the top level controller is a glue among them, that can also
    add some further control. 

    The controller of PA contains sub controllers for the set of
    possibilities and assertions, whereas the controller of PS is
    currently left empty (in order to be filled in by TUGRAZ).
    """

    def __init__(self, model, appl_ctrl):
        Controller.__init__(self, model)

        # the contained controllers assume the model to contain
        # the corresponding sub-models
        self.signals = SignalsCtrl(model.signals, self)
        self.requirements = RequirementsCtrl(model.requirements, self)
        self.prop_assurance = PropertyAssuranceCtrl(model.prop_assurance)
        self.prop_simulation = PropertySimulationCtrl(model.prop_simulation)
        self.realizability = RealizabilityCtrl(model.realizability)
        
        self.application = appl_ctrl

        self.traces = None
        return

    def registerView(self, view):
        Controller.registerView(self, view)
        view.create_sub_views(self.signals, self.requirements,
                              self.prop_assurance, self.prop_simulation, self.realizability)

        # sets paren matching for PS formula input:
        #buf = ParenMatchTextBuffer()
        self.view['textview_ps_property'].set_buffer(self.prop_simulation.model.propertyTextBuffer)
        
        return


    def enable_signals_editing(self, enable):
        self.view['button_edit_signal'].set_sensitive(enable)
        self.view['button_del_signal'].set_sensitive(enable)
        return

    def enable_reqs_editing(self, enable):
        self.view['button_edit_req'].set_sensitive(enable)
        self.view['button_del_req'].set_sensitive(enable)
        #HIDE: self.view['button_load_into_ps'].set_sensitive(enable)
        return        


    def show_traces_window(self, show):
        """Opens the project's traces window"""
        self.model.show_traces_window = show
        return

    # ------------------------------------------------------------ #
    #                       Signals Handling                       #
    # ------------------------------------------------------------ #

    def on_button_new_signal_clicked(self, button): self.signals.create_new()
    def on_button_edit_signal_clicked(self, button): self.signals.edit()
    def on_button_del_signal_clicked(self, button): self.signals.delete()
    def on_button_new_req_clicked(self, button): self.requirements.create_new()
    def on_button_edit_req_clicked(self, button): self.requirements.edit()
    def on_button_del_req_clicked(self, button): self.requirements.delete()
    def on_button_load_into_ps_clicked(self, button): self.requirements.load_into_simulation()

    def on_textview_ps_property_move_cursor(self, tv, step_size, count, extend_selection):
        gobject.idle_add(tv.get_buffer().show_paren_match)
        return
        

    # ------------------------------------------------------------ #
    #            Notifications from observable properties          #
    # ------------------------------------------------------------ #

    def property_traces_window_shown_change_notification(self, project, old, new):
        if new == old: return

        if self.traces is None and new: self.traces = TracesCtrl(self.model.traces)
        if self.traces: self.view.show_traces_window(self.traces, new)

        if self.model.traces.has_trace_window != new:
            self.model.traces.has_trace_window = new
            pass
        
        return

    def property_active_view_change_notification(self, project, old, new):
        if new == old: return

        if new == "pa": self.view.set_pa_view()
        elif new == "ps": self.view.set_ps_view()        
        elif new == "re": self.view.set_re_view()
        return
            
    
    pass # end of class
