##  ===========================================================================
##  Author: Roberto Cavada <cavada@fbk.eu>
##
##  Copyright (c) 2005 by FBK 
##
##  This is free software; you can redistribute it and/or
##  modify it under the terms of the GNU Lesser General Public
##  License as published by the Free Software Foundation; either
##  version 2 of the License, or (at your option) any later version.
##
##  This software is distributed in the hope that it will be useful,
##  but WITHOUT ANY WARRANTY; without even the implied warranty of
##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
##  Lesser General Public License for more details.
##
##  You should have received a copy of the GNU Lesser General Public
##  License along with this library; if not, write to the Free Software
##  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA.
##
##  For more information about this software see <http://rat.fbk.eu>
##  or email to the RAT Team <rat@fbk.eu>.
##  Please report bugs to <rat@fbk.eu>.
##
##  ===========================================================================


from gtkmvc.controller import Controller
from controllers.entries import PossibilitiesCtrl, AssertionsCtrl
from controllers.outcomes import OutcomeTracesCtrl
from controllers.check_panel import CheckingPanelCtrl

import utils.globals


## ----------------------------------------------------------------------
class PropertyAssuranceCtrl (Controller):
    
    """The main controller for property assurance. This controller
    contains two sub-controllers for the possibilities and for the
    assertions. Also it handles the creation of the corresponding
    sub-views during the main view registration phase. """

    def __init__(self, model):
        """@param model It is assumed the model contains the
        properties and the assertions sub-models"""

        Controller.__init__(self, model)

        self.possibilities = PossibilitiesCtrl(model.possibilities, self)        
        self.assertions = AssertionsCtrl(model.assertions, self)
        self.outcomes = OutcomeTracesCtrl(model)

        self.check_panel = CheckingPanelCtrl(model.check_panel, self)
        return

    def registerView(self, view):
        Controller.registerView(self, view)
        view.create_sub_views(self.possibilities, self.assertions,
                              self.outcomes, self.check_panel)
        return

    def enable_poss_editing(self, enable):
        self.view['button_edit_poss'].set_sensitive(enable)
        self.view['button_del_poss'].set_sensitive(enable)
        #HIDE self.view['button_load_poss_into_ps'].set_sensitive(enable)
        return

    def enable_asser_editing(self, enable):
        self.view['button_edit_assertion'].set_sensitive(enable)
        self.view['button_del_assertion'].set_sensitive(enable)
        #HIDE self.view['button_load_ass_into_ps'].set_sensitive(enable)
        return        

    def check(self, check_req_consistency):
        self.outcomes.view.show_outcomes(True)
        if self.view.get_current_page() == 0:
            self.model.check_possibilities(self.model.requirements.get_checked_properties(),
                                           self.model.possibilities.get_checked_properties(),
                                           check_req_consistency)
        else: self.model.check_assertions(
            self.model.requirements.get_checked_properties(),
            self.model.assertions.get_checked_properties(),
            check_req_consistency)
        return
        

    # ------------------------------------------------------------ #
    #                            Signals                           #
    # ------------------------------------------------------------ #
    
    def on_notebook_pa_switch_page(self, notebook, page, page_num):
        """Makes a few cosmetics to make more evident the page
        actually selected"""

        assert page_num in (0,1)
        self.view.makes_evidence_page(notebook, page_num)
        return

    def on_button_new_poss_clicked(self, button): self.possibilities.create_new()
    def on_button_edit_poss_clicked(self, button): self.possibilities.edit()
    def on_button_del_poss_clicked(self, button): self.possibilities.delete()
    def on_button_load_poss_into_ps_clicked(self, button): self.possibilities.load_into_simulation()
    
    def on_button_new_assertion_clicked(self, button): self.assertions.create_new()
    def on_button_edit_assertion_clicked(self, button): self.assertions.edit()
    def on_button_del_assertion_clicked(self, button): self.assertions.delete()
    def on_button_load_ass_into_ps_clicked(self, button): self.assertions.load_into_simulation()
 

    # ---------------------------------------- #
    #   Notifications for observed properties  #
    # ---------------------------------------- #
    
    pass # end of class
