##  ===========================================================================
##  Author: Roberto Cavada <cavada@fbk.eu>
##
##  Copyright (c) 2005 by FBK 
##
##  This is free software; you can redistribute it and/or
##  modify it under the terms of the GNU Lesser General Public
##  License as published by the Free Software Foundation; either
##  version 2 of the License, or (at your option) any later version.
##
##  This software is distributed in the hope that it will be useful,
##  but WITHOUT ANY WARRANTY; without even the implied warranty of
##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
##  Lesser General Public License for more details.
##
##  You should have received a copy of the GNU Lesser General Public
##  License along with this library; if not, write to the Free Software
##  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA.
##
##  For more information about this software see <http://rat.fbk.eu>
##  or email to the RAT Team <rat@fbk.eu>.
##  Please report bugs to <rat@fbk.eu>.
##
##  ===========================================================================

from gtkmvc.controller import Controller
import utils.globals

import gtk
import pango
import math

# ----------------------------------------------------------------------
class TraceInfoCtrl (Controller):
    """the controller for the dialog that shows trace information"""

    def __init__(self, trace_model, traces_model, parent_ctrl=None):
        """This needs traces model to access categories information"""
        Controller.__init__(self, trace_model)

        self.traces = traces_model
        self.parent = parent_ctrl

        self.changing_trace = False
        return

    def registerView(self, view):
        Controller.registerView(self, view)
        self.setup_view(view)
        return

    def setup_view(self, view):
        # sets the category combo:
        cb = self.view['cb_category']
        cb.set_model(self.traces.get_categories_model())
        cell = gtk.CellRendererText()
        cb.pack_start(cell, True)
        cb.add_attribute(cell, 'text', 0) # trace's cat name
        
        # info buf tagging preparation:
        buf = self.view['textview_other_info'].get_buffer()
        buf.create_tag("spacer", pixels_above_lines=12)        
        buf.create_tag("title", scale=pango.SCALE_LARGE, foreground="gray50", pixels_below_lines=2)
        buf.create_tag("info", family="Monospace", left_margin=12)
        self.update_view()
        return

    def update_view(self):
        if not self.changing_trace: # it is not self that is changing the trace
            # name:
            self.view['entry_name'].set_text(self.model.name)

            # notes:
            buf = self.view['textview_notes'].get_buffer()
            buf.set_text(self.model.notes)

            # category: searches and sets the current trace's category:
            cat_name = self.traces.get_trace_category(self.model)
            cb = self.view['cb_category']
            m = cb.get_model()
            iter = m.get_iter_first()
            while iter:
                if m.get_value(iter, 0) == cat_name: cb.set_active_iter(iter)
                iter = m.iter_next(iter)
                pass

            # other info:
            buf = self.view['textview_other_info'].get_buffer()
            buf.set_text("")

            # description
            buf.insert_with_tags_by_name(buf.get_end_iter(), "Description\n", "title")
            steps = self.model.get_max_steps()
            if steps == 0: sm = "no steps"
            elif steps == 1: sm = "one single step"
            else: sm = "%d steps" % steps
            buf.insert_with_tags_by_name(buf.get_end_iter(), "This trace contains " + sm, "info")
            loops = len(self.model.loops)
            if loops == 0: sm = "no loop"
            elif loops == 1: sm = "one loop at step %s" % self.model.loops[0]
            else: sm = "multiple loops at steps " + ", ".join(map(str, self.model.loops))
            buf.insert_with_tags_by_name(buf.get_end_iter(), ", and has %s.\n" % sm, "info")

            # dependencies
            buf.insert_with_tags_by_name(buf.get_end_iter(), "Dependencies\n", "title", "spacer")
            buf.insert_with_tags_by_name(buf.get_end_iter(),
                                         ", ".join(self.model.dependencies)+"\n", "info")

            # formula
            buf.insert_with_tags_by_name(buf.get_end_iter(), "Originating formula\n", "title", "spacer")
            txt = self.model.get_formula()
            if not txt: txt = "(Not available)"
            buf.insert_with_tags_by_name(buf.get_end_iter(), txt + "\n", "info")
            
            # other info
            buf.insert_with_tags_by_name(buf.get_end_iter(), "Creation information\n", "title", "spacer")
            txt = self.model.get_creation_info()
            if not txt: txt = "(Not available)"
            buf.insert_with_tags_by_name(buf.get_end_iter(), txt + "\n", "info")


            # history
            buf.insert_with_tags_by_name(buf.get_end_iter(), "History\n", "title", "spacer")
            hlen = len(self.model.history) # length of the history
            # number of digits for alignment:
            hstr = "%0" + "%dd: " % (int(math.log10(30)) + 1)
            for i in range(hlen):
                buf.insert_with_tags_by_name(buf.get_end_iter(),
                                             (hstr % i) + self.model.history[i] + "\n",
                                             "info")
                pass
            
            pass

        return
            

    def update_model(self):
        self.changing_trace = True

        changed = False
        
        # name
        txt = self.view['entry_name'].get_text()
        if self.model.name != txt:
            self.model.name = txt
            self.model.add_event_to_history("The user changed the name to be '%s'" % txt)
            pass

        # notes 
        buf = self.view['textview_notes'].get_buffer()
        txt = buf.get_text(buf.get_start_iter(), buf.get_end_iter())
        if self.model.notes != txt:
            self.model.notes = txt
            changed = True
            pass
        
        # category:
        iter = self.view['cb_category'].get_active_iter()
        new_cat_name = self.view['cb_category'].get_model().get_value(iter, 0)
        cur_cat_name = self.traces.get_trace_category(self.model)
        
        if new_cat_name != cur_cat_name:
            self.model.add_event_to_history("The user changed the category to be '%s'" \
                                            % new_cat_name)
            self.traces.remove_trace(self.model, cur_cat_name)
            self.traces.add_trace(self.model, new_cat_name)
            pass

        if changed: self.model.description_changed = True
        self.changing_trace = False
        return
        

    # ------------------------------------------------------------ #
    #                        GUI's signals                         #
    # ------------------------------------------------------------ #

    def on_entry_name_changed(self, entry):
        self.view.set_title(entry.get_text())
        return

    def on_button_ok_clicked(self, button):
        # first checks that the trace can be committed
        name = self.view['entry_name'].get_text()

        trace = self.traces.get_trace_model(name)
        if trace and trace != self.model:
            # invalid name!
            utils.dialogs.open_invalid_name_dialog()
            self.view['entry_name'].grab_focus()
            self.view['entry_name'].set_position(-1)
        else:
            self.model.unregisterObserver(self)            
            self.update_model()
            if self.parent:
                self.parent.trace_editing_terminated(self.model)
                self.view.destroy()
            else: gtk.main_quit()                
            pass
        
        return

    def on_button_cancel_clicked(self, button):
        self.model.unregisterObserver(self)
        if self.parent:
            self.parent.trace_editing_terminated(self.model)
            self.view.destroy()
        else: gtk.main_quit()
        return

    def on_dialog_top_level_delete_event(self, window, event):
        self.model.unregisterObserver(self)
        if self.parent:
            self.parent.trace_editing_terminated(self.model)
            self.view.destroy()
        else: gtk.main_quit()
        return True


    # -------------------------------------------------------------------- #
    # Observable properties                                                #
    # -------------------------------------------------------------------- #
    def property_name_change_notification(self, model, old, new):
        if new != old and model == self.model: self.update_view()
        return


    def property_description_changed_change_notification(self, model, old, new):
        """For some reason, it is needed a recalculation of trace's description"""
        if model == self.model: self.update_view()
        return
    

    pass # end of class
