##  ===========================================================================
##  Author: Roberto Cavada <cavada@fbk.eu>
##
##  Copyright (c) 2005 by FBK 
##
##  This is free software; you can redistribute it and/or
##  modify it under the terms of the GNU Lesser General Public
##  License as published by the Free Software Foundation; either
##  version 2 of the License, or (at your option) any later version.
##
##  This software is distributed in the hope that it will be useful,
##  but WITHOUT ANY WARRANTY; without even the implied warranty of
##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
##  Lesser General Public License for more details.
##
##  You should have received a copy of the GNU Lesser General Public
##  License along with this library; if not, write to the Free Software
##  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA.
##
##  For more information about this software see <http://rat.fbk.eu>
##  or email to the RAT Team <rat@fbk.eu>.
##  Please report bugs to <rat@fbk.eu>.
##
##  ===========================================================================


import gtk, gobject
import subprocess, os, tempfile
import re

from gtkmvc.model import TreeStoreModel
import models.entry
import utils.globals
from utils.utilities import get_conjuction


## ----------------------------------------------------------------------
class EntriesModel (TreeStoreModel):
    """This is a generic model for those parts of the GUI that require
    an editable TreeView. See for examples frames for Signals,
    Requirements and Possibilities and Assertion within the Property
    Assurance section.

    """

    # ------------------------------------------------------ #
    #                 Observable properties                  #
    # ------------------------------------------------------ #
    __properties__ = {
        'changed' : False  # if the model changed somehow
        }


    def __init__(self, types):
        """@param types is a list of types to be passed to the
           TreeStore constructor"""

        TreeStoreModel.__init__(self, *types)
        return

    def quit(self):
        """Called when it is time to quit the application"""
        return        

    def clear(self):
        """Cleans up the content"""
        gtk.TreeStore.clear(self)
        self.changed = True
        return

    def get_entries_num(self):
        res = 0
        iter = self.get_iter_first()
        while iter:
            res += 1
            iter = self.iter_next(iter)
            pass
        return res        

    def get_new_instance_name(self, prefix):
        name = "%s%d" % (prefix, self.get_counter())
        iter = self.get_iter_first()
        while iter:
            if self.get_value(iter, 0).get_name() == name:
                self.inc_counter()
                return self.get_new_instance_name(prefix)
            iter = self.iter_next(iter)
            pass
        return name


    def get_entry_by_name(self, name):
        """@return the EntryModel that corresponds to the given name,
        or None if the name does not exist"""

        iter = self.get_iter_first()
        while iter:
            elem = self.get_value(iter, 0)
            if elem.get_name() == name: return elem
            iter = self.iter_next(iter)
            pass
        return None

    def update_model(self, model):
        """Returns the iterator of the found model (usable by
        derived classes)"""
        iter = self.get_iter_first()
        while iter:
            elem = self.get_value(iter, 0)
            if elem == model: break
            assert elem.get_name() != model.name
            iter = self.iter_next(iter)
            pass

        assert iter
        self.set_value(iter, 0, model)
        self.changed = True
        return iter
    
        
    def add_new_model(self, model):
        """Adds a new model. The given model must not exist into the
        set of contained entries"""

        # checks that there is no name reuse:
        assert self.get_entry_by_name(model.name) is None

        # appends one new entry
        iter = self.append(None)
        self.set_value(iter, 0, model)
        model.registerObserver(self)
        self.inc_counter()
        self.changed = True
        return


    def to_xml(self, root_doc, parent_node):
        iter = self.get_iter_first()
        while iter:            
            self.get_value(iter, 0).to_xml(root_doc, parent_node)
            iter = self.iter_next(iter)
            pass
        
        return
          
    pass # end of class

    
## ----------------------------------------------------------------------
class SignalsModel (EntriesModel):
    """The model of the Signals frame"""

    __counter = 0 # to count the number of instances

    def __init__(self, proj_model):
        EntriesModel.__init__(self, (gobject.TYPE_PYOBJECT,))
        self.parent = proj_model
        return

    def quit(self):
        """Called when it is time to quit the application"""
        return        

    def clear(self):
        """Cleans up the content"""
        EntriesModel.clear(self)
        SignalsModel.__counter = 0
        return

    def get_counter(self): return SignalsModel.__counter
    def inc_counter(self): SignalsModel.__counter += 1

    def remove(self, iter):
        sgn = self[iter][0]
        EntriesModel.remove(self, iter)
        self.parent.signal_changed_added_or_removed(sgn)
        return        

    def add_new_model(self, model):
        EntriesModel.add_new_model(self, model)
        self.parent.signal_changed_added_or_removed(model)
        return
    
    def update_model(self, model):
        EntriesModel.update_model(self, model)
        self.parent.signal_changed_added_or_removed(model)
        return
 
    def to_xml(self, root_doc, parent_node):
        sgns = root_doc.createElement("signals")
        EntriesModel.to_xml(self, root_doc, sgns)
        parent_node.appendChild(sgns)
        return

    def from_xml(self, root_doc, node):
        signals = node.getElementsByTagName("signal")
        for child in signals:
            model = models.entry.SignalModel("")
            model.from_xml(root_doc, child)
            self.add_new_model(model)
            pass

        return

    def get_boolean_signals_names(self):
        """A utility function, used to improve performances on filters"""

        res = []
        iter = self.get_iter_first()
        while iter:
            sgn = self.get_value(iter,0)
            oftxt = " of "
            pos = sgn.get_type().rfind(oftxt)
            if pos >= 0 and sgn.get_type().strip().find("array ") == 0:
                ty = sgn.get_type()[pos+len(oftxt):].replace(" ", "").strip()
            else: ty = sgn.get_type().replace(" ", "").strip()

            if ty.lower() in ("boolean", "{0,1}", "{1,0}", "0..1", "1..0"):
                res.append(sgn.get_name())
                pass
            iter = self.iter_next(iter)
            pass
        
        return res

    def type_check_signal(self, name, type):
        """Check syntax of given signal. None is returned if the
        check is ok, or the error string is returned in case of
        error"""
        smv, smvfn = tempfile.mkstemp()
        cmd, cmdfn = tempfile.mkstemp()        

        os.write(smv, """MODULE main
VAR %s : %s;
""" % (name, type))
        os.close(smv)
        
        os.write(cmd, "set on_failure_script_quits; read_model; flatten_hierarchy; quit")
        os.close(cmd)


        out, err = subprocess.Popen([utils.globals.NUSMV_PATH, "-source", cmdfn, smvfn],
                                    stdout=subprocess.PIPE, 
                                    stderr=subprocess.PIPE).communicate()
        for fn in (cmdfn, smvfn): os.remove(fn) # cleans up temp files

        errpat = re.compile('^file %s:\sline 2:\s(.+)$' % smvfn, re.MULTILINE)
        m = errpat.search(err) if err else None
        if m: return m.group(1)
        return None

    pass # end of class


## ----------------------------------------------------------------------
class PropertiesModel (EntriesModel):
    """The model of the Requirements frame"""

    def __init__(self, signals, traces=None, prop_simulation=None):

        """@param traces is the model of traces (a TracesModel
        instance) to keep track of those traces that are dependant on
        properties.
        @param prop_simulation is used to load properties into PS"""
        
        EntriesModel.__init__(self,
                              (gobject.TYPE_PYOBJECT,
                               gobject.TYPE_BOOLEAN, # toggled
                               gobject.TYPE_BOOLEAN, # type invalid
                               ))

        self.signals = signals
        self.traces = traces
        self.prop_simu = prop_simulation
        return

    def remove(self, iter):
        prop = self[iter][0]
        EntriesModel.remove(self, iter)
        self.traces.deleted_property(prop)
        return        

    def update_model(self, model):
        iter = EntriesModel.update_model(self, model)
        self.set_value(iter, 2, True)
        return

    def toggle_check(self, path):
        """Called when an event in the view requested to toggle a
        check value. Only correctly type-checked entries can be
        toggled"""
        if self[path][2]: self[path][1] = not self[path][1]
        return

    def get_checked_properties(self):
        """Provides the set of entries that have been checked (selected) by the user.
        @return a list of pair (model, path)"""

        res = []    
        iter = self.get_iter_first()
        while iter:
            if self.get_value(iter, 1): res.append((self.get_value(iter, 0), self.get_path(iter)))
            iter = self.iter_next(iter)
            pass
        
        return res

    def load_props_into_simulation(self, sel_models):
        """Loads the disjunction of selected properties into
        simulation.  @param sel_models is a list of iterators to the
        selected entry models."""

        if len(sel_models) == 0 or self.prop_simu is None: return
        prop_text = get_conjuction(map(lambda x: self.get_value(x,0).get_property(),
                                       sel_models), "\n")

        self.prop_simu.setProperty(prop_text)
        return


    def to_xml(self, root_doc, parent_node):
        iter = self.get_iter_first()
        while iter:            
            self.get_value(iter, 0).to_xml(root_doc, parent_node)

            # adds toggled information
            tog = root_doc.createElement("toggled")
            tog.appendChild(root_doc.createTextNode("%d" % self.get_value(iter, 1)))
            parent_node.lastChild.appendChild(tog)

            iter = self.iter_next(iter)
            pass


    def add_new_model(self, model):
        """Adds a new model. The given model must not exist into the
        set of contained entries"""

        # checks that there is no name reuse:
        assert self.get_entry_by_name(model.name) is None

        # appends one new entry
        iter = self.append(None)
        self.set_value(iter, 0, model)
        self.set_value(iter, 1, int(model.toggled))
        self.set_value(iter, 2, True)
        
        model.registerObserver(self)
        self.inc_counter()
        self.changed = True
        return


    def type_check_all(self):
        """Type checks all the properties. The third column will be
        set accordingly (TRUE=ok, FALSE=type error). Clears also
        the possibly set status of the property (coming from
        previous MC results)"""
        
        for row in self:
            ok = self.type_check_property(row[0].property) is None
            row[2] = ok
            if not ok: 
                row[0].status = None
                self.traces.invalidated_property(row[0])
                row[1] = False # unchecks it
                pass
            pass
        return


    def type_check_property(self, property):
        """Check syntax of given property. None is returned if the
        check is ok, or the error string is returned in case of
        error. 
        @param property is a string with the PSL specification"""
        smv, smvfn = tempfile.mkstemp()
        cmd, cmdfn = tempfile.mkstemp()        

        os.write(smv, "MODULE main\nVAR\n")
        for signal in (row[0] for row in self.signals):
            os.write(smv, "%s : %s;\n" % (signal.name, signal.type))
            pass
        os.write(smv, "PSLSPEC %s;\n" % property)
        os.close(smv)
        
        os.write(cmd, "set on_failure_script_quits; read_model; flatten_hierarchy; quit")
        os.close(cmd)

        out, err = subprocess.Popen([utils.globals.NUSMV_PATH, "-source", cmdfn, smvfn],
                                    stdout=subprocess.PIPE, 
                                    stderr=subprocess.PIPE).communicate()
        for fn in (cmdfn, smvfn): os.remove(fn) # cleans up temp files

        errpat = re.compile('^.+:\s(.+)$', re.MULTILINE)
        m = errpat.search(err) if err else None
        if m: return m.group(1)
        return None


    # -------------------------------------------------------------------- #
    #       Observed properties                                            #
    # -------------------------------------------------------------------- #
    def property_name_change_notification(self, model, old, new):
        if new != old and self.traces:
            self.traces.changed_property_name(model, old)
            pass
        return

    def property_property_change_notification(self, model, old, new):        
        if self._strip_property(new) != self._strip_property(old):
            self.traces.changed_property_specification(model, old)
            pass        
        return

    def property_status_change_notification(self, model, old, new):
        if new != old: self.changed = True
        return


    # -------------------------------------------------------------------- #
    #       Internal services                                              #
    # -------------------------------------------------------------------- #
    def _strip_property(self, prop_str):
        """Internal service used to compare strings of
        specifications. Spaces are removed from given string"""
        return re.sub("\s", "", prop_str)

    pass # end of class


## ----------------------------------------------------------------------
class RequirementsModel (PropertiesModel):
    """The model of the Requirements frame"""

    __counter = 0 # to count the number of instances
    
    def __init__(self, proj_model, 
                 signals, traces=None, prop_simulation=None):

        PropertiesModel.__init__(self, signals, traces, prop_simulation)
        self.parent = proj_model
        return

    def clear(self):
        """Cleans up the content"""
        PropertiesModel.clear(self)
        RequirementsModel.__counter = 0
        return

    def remove(self, iter):
        # this is used to notify the possibilities and the
        # assertions in PA that a requirement has been removed
        prop = self[iter][0]
        PropertiesModel.remove(self, iter)
        self.parent.property_changed_or_removed(prop)
        return


    def get_counter(self): return RequirementsModel.__counter
    def inc_counter(self): RequirementsModel.__counter += 1

    def to_xml(self, root_doc, parent_node):
        props = root_doc.createElement("requirements")
        PropertiesModel.to_xml(self, root_doc, props)
        parent_node.appendChild(props)
        return

    def from_xml(self, root_doc, node):
        props = node.getElementsByTagName("requirement")
        for child in props:
            model = models.entry.RequirementModel("")
            model.from_xml(root_doc, child)
            self.add_new_model(model)
            pass

        return    


    # -------------------------------------------------------------------- #
    #       Observed properties                                            #
    # -------------------------------------------------------------------- #
    def property_property_change_notification(self, model, old, new):
        PropertiesModel.property_property_change_notification(self, model, old, new)

        if self._strip_property(new) != self._strip_property(old):            
            # a requirement changed!
            self.parent.property_changed_or_removed(model)
            pass
        return
    
    pass # end of class


## ----------------------------------------------------------------------
class PossibilitiesModel (PropertiesModel):
    """The model of the Possibilities frame"""

    __counter = 0 # to count the number of instances
    
    def __init__(self, signals, traces=None, prop_simulation=None):
        PropertiesModel.__init__(self, signals, traces, prop_simulation)
        return

    def clear(self):
        """Cleans up the content"""
        PropertiesModel.clear(self)
        PossibilitiesModel.__counter = 0
        return

    def property_changed_or_removed(self, prop):
        """Resets MC results in properties which have a dependency
        wrt given property which has been removed"""
        for row in self:
            if prop.name in row[0].deps: 
                row[0].status = None
                self.row_changed(row.path, row.iter)
                pass
            
            pass
        return

    def get_counter(self): return PossibilitiesModel.__counter
    def inc_counter(self): PossibilitiesModel.__counter += 1

    def to_xml(self, root_doc, parent_node):
        props = root_doc.createElement("possibilities")
        PropertiesModel.to_xml(self, root_doc, props)
        parent_node.appendChild(props)
        return

    def from_xml(self, root_doc, node):
        props = node.getElementsByTagName("possibility")
        for child in props:
            model = models.entry.PossibilityModel("")
            model.from_xml(root_doc, child)
            self.add_new_model(model)
            pass

        return    

    pass # end of class


## ----------------------------------------------------------------------
class AssertionsModel (PropertiesModel):
    """The model of the Assertions frame"""

    __counter = 0 # to count the number of instances
    
    def __init__(self, signals, traces=None, prop_simulation=None):
        PropertiesModel.__init__(self, signals, traces, prop_simulation)
        return

    def clear(self):
        """Cleans up the content"""
        PropertiesModel.clear(self)
        AssertionsModel.__counter = 0
        return

    def property_changed_or_removed(self, prop):
        """Resets MC results in properties which have a dependency
        wrt given property which has been removed"""
        for row in self:
            if prop.name in row[0].deps: 
                row[0].status = None
                self.row_changed(row.path, row.iter)
                pass
            
            pass
        return

    def get_counter(self): return AssertionsModel.__counter
    def inc_counter(self): AssertionsModel.__counter += 1

    def to_xml(self, root_doc, parent_node):
        props = root_doc.createElement("assertions")
        PropertiesModel.to_xml(self, root_doc, props)
        parent_node.appendChild(props)

        return
    
    def from_xml(self, root_doc, node):
        props = node.getElementsByTagName("assertion")
        for child in props:
            model = models.entry.AssertionModel("")
            model.from_xml(root_doc, child)
            self.add_new_model(model)
            pass

        return    
    
    pass # end of class
