##  ===========================================================================
##  Author: Roberto Cavada <cavada@fbk.eu>
##
##  Copyright (c) 2005 by FBK 
##
##  This is free software; you can redistribute it and/or
##  modify it under the terms of the GNU Lesser General Public
##  License as published by the Free Software Foundation; either
##  version 2 of the License, or (at your option) any later version.
##
##  This software is distributed in the hope that it will be useful,
##  but WITHOUT ANY WARRANTY; without even the implied warranty of
##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
##  Lesser General Public License for more details.
##
##  You should have received a copy of the GNU Lesser General Public
##  License along with this library; if not, write to the Free Software
##  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA.
##
##  For more information about this software see <http://rat.fbk.eu>
##  or email to the RAT Team <rat@fbk.eu>.
##  Please report bugs to <rat@fbk.eu>.
##
##  ===========================================================================

from gtkmvc.model import Model
import xml.dom

# ----------------------------------------------------------------------
class EntryModel (Model):

    """This class represents a general entry, like a signal, a requirement, etc.
    Associated information is:
    name: the entry's name
    notes: textual information about the entry"""

    # observable properties:
    __properties__ = {
        "name" : None,
        }

    def __init__(self, name):
        Model.__init__(self)

        self.name = name
        self.notes = ""
        return

    def get_name(self): return self.name
    def get_notes(self): return self.notes

    def to_xml(self, root_doc, parent_node):
        n = root_doc.createElement("name")        
        n.appendChild(root_doc.createTextNode(self.name))
        parent_node.appendChild(n)

        n = root_doc.createElement("notes")        
        n.appendChild(root_doc.createTextNode(self.notes))
        parent_node.appendChild(n)
        return

    def from_xml(self, root_doc, node):
        for child in node.childNodes:
            if child.nodeType == xml.dom.Node.ELEMENT_NODE and hasattr(self, child.tagName):
                data = ""
                for text_child in child.childNodes:
                    if text_child.nodeType == xml.dom.Node.TEXT_NODE:
                        data += text_child.data
                        pass
                    pass

                data = data.strip()
                if data == "None": data = None

                # handles lists
                if type(getattr(self, child.tagName)) == list: data = data.split(",")
                setattr(self, child.tagName, data)
                pass
            pass
        
        return
    
    pass # end of class


# ----------------------------------------------------------------------
class SignalModel (EntryModel):

    """This class represents a single signal. Associated information are:
    name: the signal's name
    type: the type of the variable holding the signal
    notes: textual information about the signal"""

    # observable properties:
    __properties__ = {
        "type" : None,
        "kind" : None, 
        }

    def __init__(self, name, kind="", type=""):
        EntryModel.__init__(self, name)
        self.kind = kind
        self.type = type
        return

    def get_type(self): return self.type
    def get_kind(self): return self.kind

    def to_xml(self, root_doc, parent_node):
        sgn = root_doc.createElement("signal")
        
        EntryModel.to_xml(self, root_doc, sgn)

        n = root_doc.createElement("kind")        
        n.appendChild(root_doc.createTextNode(self.kind))
        sgn.insertBefore(n, sgn.lastChild)

        n = root_doc.createElement("type")        
        n.appendChild(root_doc.createTextNode(self.type))
        sgn.insertBefore(n, sgn.lastChild)

        parent_node.appendChild(sgn)
        return

    pass # end of class


# ----------------------------------------------------------------------
class PropertyModel (EntryModel):

    """This class represents a single property, like a requirment, an
    assertion, etc.  Associated information are:
    * name: the property's name. 
    * property: the property's specification
    * notes: textual information about the signal"""

    # observable properties:
    __properties__ = {
        "property" : None
        }

    conj_symbol = "&"

    def __init__(self, name):
        EntryModel.__init__(self, name)
        self.property = ""
        self.toggled = "0" # this is used only when read from file
        return

    def get_property(self): return self.property

    def to_xml(self, root_doc, parent_node):
        EntryModel.to_xml(self, root_doc, parent_node)
        
        n = root_doc.createElement("property")        
        n.appendChild(root_doc.createTextNode(self.property))
        parent_node.insertBefore(n, parent_node.lastChild)
        return

    pass # end of class


# ----------------------------------------------------------------------
class RequirementModel (PropertyModel):

    # observable properties:
    __properties__ = {
        "kind" : None, 
        }
    
    def __init__(self, name, kind=""):
        PropertyModel.__init__(self, name)

        self.kind = kind
        return

    def get_kind(self): return self.kind

    def to_xml(self, root_doc, parent_node):
        prop = root_doc.createElement("requirement")
        PropertyModel.to_xml(self, root_doc, prop)

        n = root_doc.createElement("kind")
        n.appendChild(root_doc.createTextNode(self.kind))
        prop.insertBefore(n, prop.lastChild)
        
        parent_node.appendChild(prop)
        return
    
    pass # end of class



# ----------------------------------------------------------------------
class PA_PropertyModel (PropertyModel):
    __properties__ = {
        "status" : None,
        "deps" : [], # reqs and other props in the currently shown result
        }
    
    def __init__(self, name):
        PropertyModel.__init__(self, name)
        return

    def to_xml(self, root_doc, parent_node):
        PropertyModel.to_xml(self, root_doc, parent_node)

        # status
        n = root_doc.createElement("status")
        n.appendChild(root_doc.createTextNode(str(self.status)))
        parent_node.insertBefore(n, parent_node.lastChild)

        # deps
        deps_node = root_doc.createElement("deps")
        deps_node.appendChild(root_doc.createTextNode(",".join(self.deps)))
        parent_node.insertBefore(deps_node, parent_node.lastChild)

        return
    

    pass # end of class
        

# ----------------------------------------------------------------------
class PossibilityModel (PA_PropertyModel):
    def __init__(self, name):
        PA_PropertyModel.__init__(self, name)
        return

    def to_xml(self, root_doc, parent_node):
        prop = root_doc.createElement("possibility")
        PA_PropertyModel.to_xml(self, root_doc, prop)
        parent_node.appendChild(prop)
        return
    
    pass # end of class


# ----------------------------------------------------------------------
class AssertionModel (PA_PropertyModel):
    def __init__(self, name):
        PA_PropertyModel.__init__(self, name)
        return

    def to_xml(self, root_doc, parent_node):
        prop = root_doc.createElement("assertion")
        PA_PropertyModel.to_xml(self, root_doc, prop)
        parent_node.appendChild(prop)
        return
    
    pass # end of class

