##  ===========================================================================
##  Author: Roberto Cavada <cavada@fbk.eu>
##
##  Copyright (c) 2005 by FBK 
##
##  This is free software; you can redistribute it and/or
##  modify it under the terms of the GNU Lesser General Public
##  License as published by the Free Software Foundation; either
##  version 2 of the License, or (at your option) any later version.
##
##  This software is distributed in the hope that it will be useful,
##  but WITHOUT ANY WARRANTY; without even the implied warranty of
##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
##  Lesser General Public License for more details.
##
##  You should have received a copy of the GNU Lesser General Public
##  License along with this library; if not, write to the Free Software
##  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA.
##
##  For more information about this software see <http://rat.fbk.eu>
##  or email to the RAT Team <rat@fbk.eu>.
##  Please report bugs to <rat@fbk.eu>.
##
##  ===========================================================================

import gtk
from gtkmvc.model import Model

from models.entries import PossibilitiesModel, AssertionsModel
from models.entry import PropertyModel

from models.check_panel import CheckingPanelModel

from tools.stubs.nusmv import PANuSMVStub

import utils.globals
from utils.utilities import get_conjuction

import os.path

class PropertyAssuranceModel (Model):

    """This is the model of the frame about Property Assurance. The
    model contains two sub-models about possibilities and
    assertions. Also, a checking panel model is contained."""

    # -------------------------------------- #
    #      Set of observable properties      #
    # -------------------------------------- #    
    __properties__ = {
        "changed" : False, # the status changed
        "busy_level" : 0     # some stub is in execution
        }

    def __init__(self, proj_model, 
                 signals, requirements, traces, prop_simulation):
        """
        @param proj_model is the parent model
        @param signals is the SignalsModel instance within the project model
        @param requirements is the RequirementsModel instance within the project model
        @param traces is the TracesModel of the project
        @param prop_simulation is used to load properties into PS        
        """
        
        Model.__init__(self)
        
        self.possibilities = PossibilitiesModel(signals, traces, prop_simulation)
        self.possibilities.registerObserver(self)

        self.assertions = AssertionsModel(signals, traces, prop_simulation)
        self.assertions.registerObserver(self)

        self.check_panel = CheckingPanelModel(self)

        self.parent = proj_model

        # this are used during model checking
        self.signals = signals
        self.requirements = requirements

        # the traces model
        self.traces = traces

        # ------------------------------
        #   Stubs and stub output      -
        # ------------------------------
        self.stub_output_buffer = gtk.TextBuffer()
        
        # this stub is for checking the properties and reqs consistency
        self.stub = PANuSMVStub(utils.globals.NUSMV_PATH, self.stub_output_buffer)

        # self is an observer of itself:
        self.registerObserver(self)

        self.sat_length = None
        return

    def quit(self):
        """Called when it is time to quit the application"""
        self.possibilities.quit()
        self.assertions.quit()
        return

    def property_changed_or_removed(self, prop_model):
        """A requirement or a property got modified or
        removed. This is used to keep track of all dependencies
        among properties, to reset results of model checking"""
        
        self.possibilities.property_changed_or_removed(prop_model)
        self.assertions.property_changed_or_removed(prop_model)
        return
   
    def type_check_all(self):
        """Type checks all the possiblities and assertions"""
        self.possibilities.type_check_all()
        self.assertions.type_check_all()
        return

    def clear(self):
        """Cleans up the model"""
        self.possibilities.clear()
        self.assertions.clear()
        return

    def to_xml(self, root_doc, parent_node):
        pa = root_doc.createElement("property_assurance")
        self.possibilities.to_xml(root_doc, pa)
        self.assertions.to_xml(root_doc, pa)       
        parent_node.appendChild(pa)
        return

    def from_xml(self, root_doc, node):
        poss = node.getElementsByTagName("possibilities")
        for p in poss: self.possibilities.from_xml(root_doc, p)

        ass = node.getElementsByTagName("assertions")
        for a in ass: self.assertions.from_xml(root_doc, a)
        self.changed = False
        return


    def stop_stub(self): self.stub.stop_session()

    def prepare_stub_session(self, stub, signals, property):
        """Prepares the stub for the session: submits the property and
        the signals (provided) and options and the checking command
        (getting them from the check panel's model)"""

        # options
        cmd, opts = self.check_panel.get_command_and_options()
        for opt in opts: stub.add_command(opt)

        # command for technology
        if self.check_panel.get_use_sat():
            stub.add_command("go_bmc")
            self.sat_length = self.check_panel.length
        else:
            stub.add_command("go")
            self.sat_length = None
            pass

        # checking command
        stub.set_checking_command(cmd)

        # signals:
        stub.set_signals(signals)

        # property:
        stub.set_property_to_check(property)
        return

    def check_reqs_consistency(self, reqs):
         # for ICD testing
         print "Call to ICD function rat.src.models.prop_assurance.PropertyAssuranceModel.check_reqs_consistency"

         return self.check_poss_reqs_consistency(reqs, poss=[])

    def check_poss_reqs_consistency(self, reqs, poss):
        """Checks consistency of the requirements for possibilities"""
       
        if len(reqs) == 0: return
        
        # creates !(req1 & req2 & ... & reqN):
        property = "!(%s)" % get_conjuction(map(lambda x: x[0].get_property(), reqs)) 
        
        # sets up the stub for the execution:
        self.stub.set_end_session_callback(self.__reqs_poss_stub_end, (reqs, poss))
        self.prepare_stub_session(self.stub, self.signals, property)

        # launches the stub's session and exits (the result will arrive later
        # to __reqs_poss_stub_end):
        self.busy_level += 1
        self.stub.run_session()
        return


    def check_assert_reqs_consistency(self, reqs, asserts):
        """Checks consistency of the requirements for assertions"""

        if len(reqs) == 0: return
        
        # creates !(req1 & req2 & ... & reqN):
        property = "!(%s)" % get_conjuction(map(lambda x: x[0].get_property(), reqs))

        # sets up the stub for the execution:
        self.stub.set_end_session_callback(self.__reqs_asser_stub_end, (reqs, asserts))
        self.prepare_stub_session(self.stub, self.signals, property)

        # launches the stub's session and exits (the result will arrive later
        # to __reqs_poss_stub_end):
        self.busy_level += 1
        self.stub.run_session()
        return

    def check_possibilities(self, reqs, poss, check_req_consistency):
        # for ICD testing		
        print "Call to ICD function rat.src.models.prop_assurance.PropertyAssuranceModel.check_possibilities"
        # creates req1 & req2 & ... & reqN:
        reqs_str = get_conjuction(map(lambda x: x[0].get_property(), reqs))

        if len(poss) > 0:  # we must check possibilities
            # creates pos1 & pos2 & ... & posN:
            poss_str = get_conjuction(map(lambda x: x[0].get_property(), poss))

            if reqs_str: property = "!(%s)" % get_conjuction([reqs_str, poss_str])
            else: property = "!(%s)" % poss_str

            # sets up the stub for the execution:
            self.stub.set_end_session_callback(self.__possibility_stub_end,
                                               (reqs, poss, check_req_consistency))
            self.prepare_stub_session(self.stub, self.signals, property)

            # launches the stub's session and exits (the result will arrive later
            # to __possibility_stub_end):
            self.busy_level += 1
            self.stub.run_session()

        elif len(reqs) > 0 and check_req_consistency:
            # must check the requirements stand-alone:
            self.check_poss_reqs_consistency(reqs, poss)
            pass

        return
    
    def check_assertions(self, reqs, asserts, check_req_consistency):
        # for ICD testing		
        print "Call to ICD function rat.src.models.prop_assurance.PropertyAssuranceModel.check_possibilities"

        # creates req1 & req2 & ... & reqN:
        reqs_str = get_conjuction(map(lambda x: x[0].get_property(), reqs))

        if len(asserts) > 0:  # we must check assertions
            # creates ass1 & ass2 & ... & assN:
            ass_str = get_conjuction(map(lambda x: x[0].get_property(), asserts))
            
            if reqs_str: property = "((%s) -> (%s))" % (reqs_str, ass_str)
            else: property = ass_str

            # sets up the stub for the execution:
            self.stub.set_end_session_callback(self.__assertion_stub_end,
                          (reqs, asserts, check_req_consistency))
            self.prepare_stub_session(self.stub, self.signals, property)

            # launches the stub's session and exits (the result will arrive later
            # to __assertion_stub_end):
            self.busy_level += 1
            self.stub.run_session()
            pass

        elif len(reqs) > 0 and check_req_consistency:
            # must check the requirements stand-alone:
            self.check_assert_reqs_consistency(reqs, asserts)
            pass
        
        return


    def fix_traces_stripe_type(self, trace):
        """Sets the type of single stripes within the given trace
        model, depending on the signal types. This has influence on
        the way stripes are shown, e.g. if they are boolean or
        enumerative"""

        boolean_sgns = self.signals.get_boolean_signals_names()

        iter = trace.get_iter_first()
        while iter:
            stripe = trace.get_value(iter, 0)
            name = stripe.get_name().split("[")[0].strip() # gets rid of array name
            stripe.set_boolean(name in boolean_sgns)
            
            iter = trace.iter_next(iter)
            pass
        return


    # ---------------------------------------- #
    #   Notifications for observed properties  #
    # ---------------------------------------- #

    def property_changed_change_notification(self, model, old, new):
        """This notification arrives from possibilities and assertions.
        Propagates the notification only if needed."""
        if model is self and not new:
            # Someone has set changed to false. Self reflects on its children:
            self.possibilities.changed = False
            self.assertions.changed = False
            return
        
        if model is not self and new and not self.changed: self.changed = True
        return


    # --------------------------------------- #
    #              User callbacks             #
    # --------------------------------------- #

    def __possibility_stub_end(self, ret, trace, args):
        """This is a callback called when the stub for property checking
        has done its work"""

        reqs, poss_list, check_reqs = args

        if ret is None:
            # set as stopped
            for p in poss_list:
                p[0].status = "stopped"
                self.possibilities.row_changed(p[1], self.possibilities.get_iter(p[1]))
                pass
            self.busy_level -= 1
            return

        if ret != 0:
            # set as error
            for p in poss_list:
                p[0].status = "error"
                self.possibilities.row_changed(p[1], self.possibilities.get_iter(p[1]))
                pass
            self.busy_level -= 1
            return

        deps = [p[0].get_name() for p in reqs + poss_list]
        
        if trace:
            self.fix_traces_stripe_type(trace)
            
            # set as passed, and show the witness
            for p in poss_list:
                p[0].status = "passed"
                p[0].deps = deps
                self.possibilities.row_changed(p[1], self.possibilities.get_iter(p[1]))
                pass

            trace.set_dependencies(deps)    
            self.add_trace(trace)
            
        else: # failed if no reqs check is required
            if check_reqs: self.check_poss_reqs_consistency(reqs, poss_list)
            else:
                for p in poss_list:
                    p[0].status = "failed"
                    p[0].deps = deps
                    self.possibilities.row_changed(p[1], self.possibilities.get_iter(p[1]))
                    pass
            pass

        self.busy_level -= 1
        return


    def __assertion_stub_end(self, ret, trace, args):
        """This is a callback called when the stub for property checking
        has done its work"""

        reqs, asserts, check_reqs = args

        if ret is None:
            # set as stopped
            for a in asserts:
                a[0].status = "stopped"
                self.assertions.row_changed(a[1], self.assertions.get_iter(a[1]))
                pass
            self.busy_level -= 1
            return

        if ret != 0:
            # set as error
            for a in asserts:
                a[0].status = "error"
                self.assertions.row_changed(a[1], self.assertions.get_iter(a[1]))
                pass
            self.busy_level -= 1
            return        

        deps = [p[0].get_name() for p in reqs + asserts]

        if trace:
            self.fix_traces_stripe_type(trace)
           
            # set as passed, and show the witness
            for a in asserts:
                a[0].status = "failed"
                a[0].deps = deps
                self.assertions.row_changed(a[1], self.assertions.get_iter(a[1]))
                pass

            trace.set_dependencies(deps)
            self.add_trace(trace)

        else: # set as passed if no reqs checking is required:
            if check_reqs: self.check_assert_reqs_consistency(reqs, asserts)
            else:
                for a in asserts:
                    a[0].status = "passed"
                    a[0].deps = deps
                    self.assertions.row_changed(a[1], self.assertions.get_iter(a[1]))
                    pass
                pass            
            pass
        
        self.busy_level -= 1
        return

    def add_trace(self, trace_model, cat_name=utils.globals.CAT_NEW_NAME):
        self.traces.add_trace(trace_model, cat_name)
        return

    def remove_trace(self, trace_model, cat_name=None):
        self.traces.remove_trace(trace_model, cat_name)
        return


    def show_inconsistent_reqs_message(self):
        if self.sat_length is None:
            utils.dialogs.open_message_dialog("Requirements are not consistent",
                   """The set of selected requirements are not consistent among themself.
                                             
One or more requirements make the selected requirements not consistent.""")
        else:
            utils.dialogs.open_message_dialog("Requirements might be not consistent",
                 """The set of selected requirements are not consistent
among themself up to the problem depth %d.
                                             
Either one or more requirements make the selected
requirements not consistent, or the given problem length
is not deep enough to produce a witness.""" % self.sat_length)
            pass
        return
        
    
    def __reqs_poss_stub_end(self, ret, trace, reqs_poss_list):
        """This is a callback called when the checking of requirements
        consistency for possibilities ends"""

        reqs, poss_list = reqs_poss_list

        if trace is None:
            # no witness due to inconsistent requirements!
            self.show_inconsistent_reqs_message()
            res = None
            deps = []
            pass
        
        else: 
            deps = [p[0].get_name() for p in reqs + poss_list]
            res = "failed"
            pass
        
        # set the possibilities status accordingly to the result:
        for p in poss_list:
            p[0].status = res
            p[0].deps = deps
            self.possibilities.row_changed(p[1], self.possibilities.get_iter(p[1]))
            pass
        
        self.busy_level -= 1
        return


    def __reqs_asser_stub_end(self, ret, trace, reqs_asserts):
        """This is a callback called when the checking of requirements
        consistency for assertions ends"""

        reqs, asserts = reqs_asserts

        if trace is None:
            # no witness due to inconsistent requirements!
            self.show_inconsistent_reqs_message()            
            res = None
            deps = []
            pass
        
        else: 
            deps = [p[0].get_name() for p in reqs + asserts]
            res = "passed"
            pass
        
        # set the possibilities status accordingly to the result:
        for a in asserts:
            a[0].status = res
            a[0].deps = deps
            self.assertions.row_changed(a[1], self.assertions.get_iter(a[1]))
            pass
        
        self.busy_level -= 1
        return
            
        
    

    pass # end of class

