##  ===========================================================================
##  Author: Roberto Cavada <cavada@fbk.eu>
##
##  Copyright (c) 2005 by FBK 
##
##  This is free software; you can redistribute it and/or
##  modify it under the terms of the GNU Lesser General Public
##  License as published by the Free Software Foundation; either
##  version 2 of the License, or (at your option) any later version.
##
##  This software is distributed in the hope that it will be useful,
##  but WITHOUT ANY WARRANTY; without even the implied warranty of
##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
##  Lesser General Public License for more details.
##
##  You should have received a copy of the GNU Lesser General Public
##  License along with this library; if not, write to the Free Software
##  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA.
##
##  For more information about this software see <http://rat.fbk.eu>
##  or email to the RAT Team <rat@fbk.eu>.
##  Please report bugs to <rat@fbk.eu>.
##
##  ===========================================================================


import os
import os.path

import gobject

import utils.utilities
from utils.launcher import LauncherThread


# ----------------------------------------------------------------------
class ToolStub:

    """A general stub for the invokation/execution of commands by
    abstract tools. An instance of this class can be seen as an
    external abstract tool to be invoked and whose results are
    returned at the end of the execution.

    The term 'tool' is intended in the most general meaning: it can be
    either an external executable program to be executed in batch or
    interactive mode, or it can be a library to be executed in the
    stub's process.

    The execution of a tool stub generally passes through these steps:
     1. Tool initialization        init_tool()
 +-> 2. Session initialization     begin_session()
 |   3. Session execution          run()
 |   4. Result processing          process_session_result()
 +-- 5. Session deinitialization   end_session()
     6. Tool deinitialization      finalize_tool()

    Steps 2-5 can be repeated if the specific implementation of the
    stub allows for repeatition on the same tool execution instance.
    For example a batch execution model would pass through steps 1-6
    without any loop. Instead an interactve session, or a session
    based on a direct library invocation would allow for loops between
    steps 2-5.

    Steps 2-5 are executed in a sequence by calling the method run_session.

    @param buffer is an optional gtk.TextBuffer to possibly send
    the tool's output (both stderr and stdout).


    * Multithreading/sub processing *

    As one session is executed from the main (gtk) thread into a
    parallel thread, the method run_session returns immediately after
    calling and the session is exetuted in background. As soon as the
    session terminates (either because the process terminates, or
    because the user stopped it) after the session has been
    deinitialized a callback is called to inform the main thread and
    to provide it with the session result if there is any.

    End session callback can be set with the method set_end_session_callback
    """

    def __init__(self, tool_name, text_buffer=None):
        self.name = tool_name
        self.text_buffer = text_buffer

        self.end_callback = None
        self.user_arg = None

        self.launcher = None
        self.result = None
        self.ret_value = None
        
        self.init_tool()
        return

    def __del__(self):
        self.finalize_tool()
        return


    def end_callback_wrapper(self):
        """This is due to the need to ensure the user function always
        return False, that is in general not true/secure, so a wrapper
        is provided.

        Threading note: This code is
        executed in the main gtk thread, whereas the session code is
        if selfexecuted in the background thread."""

        if self.end_callback: self.end_callback(self.ret_value,
                                                self.result,
                                                self.user_arg)
        return False  # this is mandatory 


    def session_is_over(self, val, user_arg):
        """This is a callback called by the session execution thread
        when the work is done."""

        # result processing in background, before returning control to main thread
        self.ret_value = val
        self.result = None        
        if val != -1: self.result = self.process_session_result()

        # session cleanup
        self.end_session()

        # Returns to the main (gtk) thread
        gobject.idle_add(self.end_callback_wrapper)                         
        return


    def run_session(self):
        """This is the method to call to run the tool. The result of
        the execution of this top-level method is the result of the
        result processing."""

        self.log_msg("""
# ----------------------------------------------------------------------
# Starting a new session of <%s>
# ----------------------------------------------------------------------
""" % self.name)
    
        self.begin_session()
        self.run()       
        return

    def stop_session(self):
        """Stops the currently run session"""
        if self.is_running: self.launcher.kill()

        self.log_msg("""
# ----------------------------------------------------------------------
# Stopping current session of <%s>
# ----------------------------------------------------------------------
""" % self.name)
        
        return

    def set_end_session_callback(self, function, user_arg=None):
        """Use to tune what happens when the session terminates. 
        @param function is the callback called at the end of the
        session execution.  The callback function must have three arguments:
        - ret_value: the returned value of the process, None if killed by the user
        - session_result: the result (outcome) of the session
        - user_arg: optional argument provided by the user
        
        @param user_arg is the optional argument to be passed when
        function is called.
        """
        
        self.end_callback = function
        self.user_arg = user_arg
        return
    
    def log_msg(self, msg):
        if self.text_buffer: 
            self.text_buffer.insert(self.text_buffer.get_end_iter(),
                                    "%s: %s\n" % (self.name, msg))
            pass
        return


    pass # end of class


# ----------------------------------------------------------------------
class OuterProcStub (ToolStub):

    """Stubbing of the invokation of an external tool intended as a
    stand-alone executable process. When run is executed, the
    executable file is run and the parent process waits its
    termination.
    """
    
    def __init__(self, tool_name, exec_path, text_buffer=None, workdirectory=None):
    
        self.exec_path = exec_path  # name of the executable program
        self.opts = []
        self.arg = None

        self.command = None
        self.workdirectory = workdirectory
        
        ToolStub.__init__(self, tool_name, text_buffer)
        return

    def init_tool(self):
        if not os.path.exists(self.exec_path):
            raise IOError("Stub not found: '%s'" % self.exec_path)        
        return

    def run(self):
        assert self.command is None
        self.launcher = LauncherThread(self.text_buffer, self.session_is_over)
        self.launcher.workdirectory = self.workdirectory
        
        self.command = self.exec_path + " " + " ".join(self.opts)
        if self.arg: self.command += " " + self.arg

        self.launcher.launch_command(self.command)
        return

    def end_session(self):
        self.command = None
        return

    def finalize_tool(self): return # nothing to do

    def add_opt(self, opt): self.opts.append(opt.strip())
    def set_argument(self, arg): self.arg = arg.strip()
    
    def cleanup(self):
        self.opts = []
        self.arg = None
        self.command = None            
        return

    def is_running(self):
        """@return True if the process is currently execution a session, False
        if it is stopped"""
        return self.launcher and self.launcher.isAlive()
        
    
    pass # end of class


# ----------------------------------------------------------------------
class BatchModeStub (OuterProcStub):
    """Stubbing of programs in batch mode"""

    def __init__(self, tool_name, exec_path, text_buffer=None, workdirectory=None):
        OuterProcStub.__init__(self, tool_name, exec_path, text_buffer, workdirectory)
        return # not requested

    pass # end of class


# ----------------------------------------------------------------------
class InteractiveModeStub (OuterProcStub):
    """Stubbing of programs in interactive mode"""

    def __init__(self, tool_name, exec_path, text_buffer=None):
        OuterProcStub.__init__(self, tool_name, exec_path, text_buffer)
        self.commands = []
        return

    def prepend_command(self, command): self.commands.insert(0, command)

    def add_command(self, command): self.commands.append(command.strip())
    
    def cleanup(self):
        OuterProcStub.cleanup(self)
        self.commands = []
        return

    pass # end of class


# ----------------------------------------------------------------------
class ScriptBasedInteractiveModeStub (InteractiveModeStub):
    """Stubbing of programs in interactive mode, whose commands are
    read from a file"""

    def __init__(self, tool_name, exec_path, text_buffer=None, script_name=None):
        """@param script_name is None when a temp file must be created"""
        InteractiveModeStub.__init__(self, tool_name, exec_path, text_buffer)
        self.script_name = script_name

        self.session_script_name = script_name

        if script_name: self.log_msg("Script name is '%s'" % script_name)
        return

    def begin_session(self):       
        """Prepares the commands script"""
        
        if self.script_name is None:
            fd, self.session_script_name = utils.utilities.make_temp_file()
            self.log_msg("Session script name is '%s'" % self.session_script_name)
        else: fd = open(self.script_name, "w")

        
        commands = [('echo Running: "%s"' % c) + os.linesep + c
                    for c in self.commands]
        fd.write(os.linesep.join(commands))
        fd.close()        
        return

    def end_session(self):
        assert self.session_script_name
        try: os.remove(self.session_script_name)
        except: pass

        InteractiveModeStub.end_session(self)
        return
    
    
    pass # end of class

