##  ===========================================================================
##  Author: Roberto Cavada <cavada@fbk.eu>
##
##  Copyright (c) 2005 by FBK 
##
##  This is free software; you can redistribute it and/or
##  modify it under the terms of the GNU Lesser General Public
##  License as published by the Free Software Foundation; either
##  version 2 of the License, or (at your option) any later version.
##
##  This software is distributed in the hope that it will be useful,
##  but WITHOUT ANY WARRANTY; without even the implied warranty of
##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
##  Lesser General Public License for more details.
##
##  You should have received a copy of the GNU Lesser General Public
##  License along with this library; if not, write to the Free Software
##  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA.
##
##  For more information about this software see <http://rat.fbk.eu>
##  or email to the RAT Team <rat@fbk.eu>.
##  Please report bugs to <rat@fbk.eu>.
##
##  ===========================================================================


import os
import os.path
import time

from base import ScriptBasedInteractiveModeStub
from tools.utils.nusmv import NuSMVTraceReader

import utils.utilities
import utils.globals

# ----------------------------------------------------------------------
class NuSMVInteractiveModeStub (ScriptBasedInteractiveModeStub):
    """Stubbing of NuSMV in interactive mode"""

    def __init__(self, nusmv_path, text_buffer=None):
        ScriptBasedInteractiveModeStub.__init__(self, utils.globals.NUSMV_NAME, 
                                                nusmv_path, text_buffer)
        
        self.trace_reader = NuSMVTraceReader()
        return

    def begin_session(self):
        if "quit" not in self.commands: self.add_command("quit")
        
        ScriptBasedInteractiveModeStub.begin_session(self)
        
        assert self.session_script_name
        self.add_opt("-int")
        self.add_opt("-source %s" % self.session_script_name)
        return
    
    pass # end of class



# ----------------------------------------------------------------------
class PANuSMVStub (NuSMVInteractiveModeStub):
    """This is the stub of nusmv used by Property Assurance"""

    def __init__(self, nusmv_path, text_buffer=None):
        NuSMVInteractiveModeStub.__init__(self, nusmv_path, text_buffer)

        self.smv_file = "_input_smv.xml"
        self.trace_file = "_output_trace.xml"

        self.signals = None
        self.property = None
        return

    def set_property_to_check(self, property):
        self.property = utils.utilities.string_normalize_whitespaces(property)
        return
        
    def set_signals(self, signals):
        """@param signals is a SignalsModel instance"""
        self.signals = signals
        return

    def set_checking_command(self, command): self.checking_command = command

    def begin_session(self):
        assert self.property

        if self.text_buffer:
            self.text_buffer.insert(self.text_buffer.get_end_iter(),
                                    """
# Checking command is: '%s -p <PROPERTY>'
""" % self.checking_command)

        self.prepend_command("set verbose_level 1")

        # this should be the first command
        self.prepend_command("set on_failure_script_quits")

        # adds the property and trace production commands
        self.add_command('%s -p "%s"' % (self.checking_command, self.property))
        self.add_command("show_traces -p 4 -o %s" % self.trace_file)

        self.set_argument(self.smv_file)

        NuSMVInteractiveModeStub.begin_session(self)

        # prepares the strings for signals declaration
        signal_strings = []
        iter = self.signals.get_iter_first()
        while iter:
            signal = self.signals.get_value(iter, 0)
            sgn = "%s : %s;" % (utils.utilities.string_normalize_whitespaces(signal.get_name()),
                                utils.utilities.string_normalize_whitespaces(signal.get_type()))

            signal_strings.append(sgn)
            iter = self.signals.iter_next(iter)
            pass

        # generates the source smv file:
        src = """-- Generated by RAT
MODULE main
  VAR
""" + (os.linesep + "    ").join(signal_strings)

        fsmv = open(self.smv_file, "w")
        fsmv.write(src)
        fsmv.close()

        self.log_msg("SMV file is '%s'"%self.smv_file)
        return

    def end_session(self):
        if os.path.exists(self.trace_file): os.remove(self.trace_file)
        if os.path.exists(self.smv_file): os.remove(self.smv_file)
        
        NuSMVInteractiveModeStub.end_session(self)
        self.cleanup()
        return

    def process_session_result(self):
        
        """@return None if no trace was found, or a filled TraceModel
        otherwise"""

        if not os.path.exists(self.trace_file): return None
        trace = self.trace_reader.parse(self.trace_file)

        if trace:
            trace.set_formula(self.property)
            info = """Created by '%s' on %s
Checking command was: '%s'""" % (self.name, time.asctime(), self.checking_command)
            
            trace.set_creation_info(info)
            pass

        return trace

    pass # end of class



class ReNuSMVStub (NuSMVInteractiveModeStub):
    """This is the stub of nusmv used by Realizability"""

    def __init__(self, nusmv_path, text_buffer=None):
        NuSMVInteractiveModeStub.__init__(self, nusmv_path, text_buffer)

        self.smv_file = "_input_re.xml"

        self.signals = None
        self.requirements = None
        return
        
    def set_signals_requirements(self, signals, requirements):
        """@param signals is a SignalsModel instance
        @param requirements is a RequirementsModel instance"""
        self.signals = signals
        self.requirements = requirements
        return

    def set_checking_command(self, command): self.checking_command = command

    def __dump_realizability_problem(self, filename):
        from xml.dom import getDOMImplementation
        impl = getDOMImplementation()
        root = impl.createDocument(None, "project", None)

        self.signals.to_xml(root, root.documentElement)
        self.requirements.to_xml(root, root.documentElement)
        
        f = open(filename, "w")
        root.documentElement.writexml(f, addindent="  ", newl=os.linesep)
        f.close()

        root.unlink()
        return

    def begin_session(self):
        if self.text_buffer:
            self.text_buffer.insert(self.text_buffer.get_end_iter(),
                                    """
# Checking command is: '%s'
""" % self.checking_command)
            pass
        
        self.prepend_command("read_rat_file")
        self.prepend_command("set verbose_level 1")

        # this should be the first command
        self.prepend_command("set on_failure_script_quits")

        # adds the property and trace production commands
        self.add_command('%s' % (self.checking_command,))
        self.set_argument(self.smv_file)

        NuSMVInteractiveModeStub.begin_session(self)

        # generates the source smv file:
        self.__dump_realizability_problem(self.smv_file)
        return

    def end_session(self):
        if os.path.exists(self.smv_file): os.remove(self.smv_file)
        
        NuSMVInteractiveModeStub.end_session(self)
        self.cleanup()
        return

    def process_session_result(self):
        return None

    pass # end of class
    
                         
