##  ===========================================================================
##  Author: Roberto Cavada <cavada@fbk.eu>
##
##  Copyright (c) 2005 by FBK 
##
##  This is free software; you can redistribute it and/or
##  modify it under the terms of the GNU Lesser General Public
##  License as published by the Free Software Foundation; either
##  version 2 of the License, or (at your option) any later version.
##
##  This software is distributed in the hope that it will be useful,
##  but WITHOUT ANY WARRANTY; without even the implied warranty of
##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
##  Lesser General Public License for more details.
##
##  You should have received a copy of the GNU Lesser General Public
##  License along with this library; if not, write to the Free Software
##  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA.
##
##  For more information about this software see <http://rat.fbk.eu>
##  or email to the RAT Team <rat@fbk.eu>.
##  Please report bugs to <rat@fbk.eu>.
##
##  ===========================================================================


import os.path
from utils import globals
from gtkmvc.view import View


## ----------------------------------------------------------------------
class EntryView (View):
    """The base view class for creating and editing signals,
    requirements, etc."""

    GLADE_FILE = os.path.join(globals.GLADE_DIR, "entry.glade")

    def __init__(self, ctrl, top_widget_name, title, name, suffix):        
        self.suffix = suffix

        View.__init__(self, ctrl, self.GLADE_FILE, top_widget_name)

        self.set_name(name)
        self['label_%s_title' % self.suffix].set_markup("<big><b>%s</b></big>" % title)
        return

    def get_name(self): return self['entry_%s_name' % self.suffix].get_text()
    def set_name(self, name):
        entry = self['entry_%s_name' % self.suffix]
        entry.set_text(name)
        entry.set_position(-1)        
        entry.select_region(0,-1)
        return

    def get_notes(self):
        buf = self['textview_%s_notes' % self.suffix].get_buffer()
        start, end = buf.get_bounds()
        return buf.get_text(start, end, False)    

    def set_notes(self, notes):
        buf = self['textview_%s_notes' % self.suffix].get_buffer()
        buf.set_text(notes)
        return

    def get_text_other(self):
        buf = self['textview_%s_what' % self.suffix].get_buffer()
        start, end = buf.get_bounds()
        return buf.get_text(start, end, False)        

    def set_text_other(self, text):
        buf = self['textview_%s_what' % self.suffix].get_buffer()
        buf.set_text(text)
        return    

    def get_textview_other(self): return self['textview_%s_what' % self.suffix]

    def enable_ok_button(self, flag): 
        self['button_%s_ok' % self.suffix].set_sensitive(flag)
        return

    def set_window_title(self): self.get_top_widget().set_title(self.get_name())

    def destroy(self): self.get_top_widget().destroy()

    def set_image(self, fname_png):
        self['%s_image' % self.suffix].set_from_file(os.path.join(globals.IMAGES_DIR, fname_png)) 
        return

    def show_error(self, msg):
        lbl = self['label_%s_err' % self.suffix]
        if msg is None: 
            lbl.set_text("")
            lbl.hide()
        else:
            lbl.set_markup("<span foreground='red'><b>Error:</b></span>\n<tt>%s</tt>" % msg)
            lbl.show()
            pass
        return
    
    pass # end of class
    

## ----------------------------------------------------------------------
class SignalView (EntryView):
    __BOOL_TXT = "boolean"
    __ENV_TXT = "E"
    __SYS_TXT = "S"

    def __init__(self, ctrl, title, name):                
        EntryView.__init__(self, ctrl, 'window_signal', title, name, "sgn")

        self.set_image("dialog/signal.png")
        return

    def get_type(self):
        if self['rb_bool'].get_active(): return self.__BOOL_TXT
        return self.get_text_other()

    def get_kind(self):
        if self['rb_environment'].get_active(): return self.__ENV_TXT
        assert(self['rb_system'].get_active())
        return self.__SYS_TXT

    def set_type(self, text):
        ct = text.strip()
        if not ct or ct == self.__BOOL_TXT:
            self['rb_bool'].set_active(True)
            self.enable_ok_button(True)
            return

        self.set_text_other(text)
        self['rb_other'].set_active(True)
        return

    def set_kind(self, kind):
        ct = kind.strip()
        if not ct or ct == self.__ENV_TXT:
            self['rb_environment'].set_active(True)
            return
        
        assert (ct == self.__SYS_TXT)
        self['rb_system'].set_active(True)
        return
   
    pass # end of class


## ----------------------------------------------------------------------
class SignalTypeView (View):
    
    GLADE_FILE = os.path.join(globals.GLADE_DIR, "entry.glade")

    def __init__(self, ctrl):       
        View.__init__(self, ctrl, self.GLADE_FILE, "window_signal_help")
        
        self['window_signal_help'].set_default_size(450, 350)
        self.pos = None
        return 

    def show(self, flag):
        win = self['window_signal_help']
        if flag:
            if self.pos: win.move(*self.pos)
            win.show()
        else: 
            self.pos = win.get_position()
            win.hide()
            pass
        return
    pass # end of class


## ----------------------------------------------------------------------
class PropertyView (EntryView):

    def __init__(self, ctrl, top_widget_name, title, name, suffix="prop"):        
        EntryView.__init__(self, ctrl, top_widget_name, title, name, 
                           suffix)
        return

    def get_property(self): return self.get_text_other()
    def set_property(self, text): self.set_text_other(text)

    pass # end of class


## ----------------------------------------------------------------------
class RequirementView (PropertyView):
    __ASSUM_TXT = "A"
    __GUAR_TXT = "G"
    
    def __init__(self, ctrl, title, name):        
        PropertyView.__init__(self, ctrl, 'window_req', title, name, "req")
        
        self.set_image("dialog/req.png")
        return

    def get_kind(self):
        if self['rb_assumption'].get_active(): return self.__ASSUM_TXT
        assert(self['rb_guarantee'].get_active())
        return self.__GUAR_TXT

    def set_kind(self, kind):
        ct = kind.strip()
        if not ct or ct == self.__GUAR_TXT:
            self['rb_guarantee'].set_active(True)
            return
        
        assert (ct == self.__ASSUM_TXT)
        self['rb_assumption'].set_active(True)
        return
    
    pass # end of class


## ----------------------------------------------------------------------
class PossibilityView (PropertyView):
    def __init__(self, ctrl, title, name):        
        PropertyView.__init__(self, ctrl, 'window_property', title, name)
        self.set_image("dialog/poss.png")
        return
    pass # end of class


## ----------------------------------------------------------------------
class AssertionView (PropertyView):
    def __init__(self, ctrl, title, name):        
        PropertyView.__init__(self, ctrl, 'window_property', title, name)
        self.set_image("dialog/padlock.png")
        return
    pass # end of class
