##  ===========================================================================
##  Authors: Roberto Cavada <cavada@fbk.eu>
##           Ingo Pill <ipill@ist.tugraz.at>
##
##  Copyright (c) 2005 by FBK
##  Copyright (c) 2005 by Graz University of Technology
##
##  This is free software; you can redistribute it and/or
##  modify it under the terms of the GNU Lesser General Public
##  License as published by the Free Software Foundation; either
##  version 2 of the License, or (at your option) any later version.
##
##  This software is distributed in the hope that it will be useful,
##  but WITHOUT ANY WARRANTY; without even the implied warranty of
##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
##  Lesser General Public License for more details.
##
##  You should have received a copy of the GNU Lesser General Public
##  License along with this library; if not, write to the Free Software
##  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA.
##
##  For more information about this software see <http://rat.fbk.eu>
##  or email to the RAT Team <rat@fbk.eu>.
##  Please report bugs to <rat@fbk.eu>.
##
##  ===========================================================================


import os.path
from utils import globals
from gtkmvc.view import View 

from views.entries import SignalsView, RequirementsView
from views.prop_assurance import PropertyAssuranceView
#from views.prop_simulation import PropertySimulationView
#from views.realizability import RealizabilityView
from views.traces import TracesView


## ----------------------------------------------------------------------
class ProjectView (View):

    """This is the project view. 
    
    This view contains a few other views, that incapsulate
    the views of respectively the set of Signals, Requirements,
    Property Assurance (PA) and Property Simulation (PS).

    Those have designed to be available as stand alone entities,
    to the top level view is a glue among them. 

    The view of PA contains sub views for the set of
    possibilities and assertions.
    """

    GLADE_FILE = os.path.join(globals.GLADE_DIR, "project.glade")

    def __init__(self, ctrl, appl_view):
        View.__init__(self, ctrl, self.GLADE_FILE, "vpaned_top")

        self.application = appl_view

        self.traces = None
        self.curr_view = "pa"
        return
    

    def create_sub_views(self, signals_ctrl, reqs_ctrl,
                         pa_ctrl, ps_ctrl, re_ctrl):
        """This is called by the controller during view's registration.
        @param signals_ctrl is the sub-controller of the Signals set
        @param reqs_ctrl is the sub-controller of the Requirements set
        @param pa_ctrl is the sub-controller of the PA set
        @param ps_ctrl is the sub-controller of the PS set       
        """

        self.signals = SignalsView(signals_ctrl)
        self.requirements = RequirementsView(reqs_ctrl)
        self.prop_assurance = PropertyAssuranceView(pa_ctrl)
        # RC: simulation and realizability are disabled
        #self.prop_simulation = PropertySimulationView(ps_ctrl)
        #self.realizability = RealizabilityView(re_ctrl)

        # attaches the sub-views to the top-level one 'self':
        self['alignment_signals'].add(self.signals.get_top_widget())
        self['alignment_requirements'].add(self.requirements.get_top_widget())

        self['alignment_pa'].add(self.prop_assurance.get_top_widget())
        #self['alignment_ps'].add(self.prop_simulation.get_top_widget())
        #self['alignment_re'].add(self.realizability.get_top_widget())

        # sets a few of aspect flags:
        self['notebook_top'].set_show_tabs(False)
        self['notebook_bottom'].set_show_tabs(False)
        return    

    def _detach_ps(self):
        """Detaches widgets from the PS view if it is the
        case. Returns True if acted, False otherwise"""
        
        if self.curr_view != "ps": return

        # detach ps stuff
        sgn = self['algn_ps_sgns'].get_child()
        self['algn_ps_sgns'].remove(sgn)
        
        rqs = self['algn_ps_reqs'].get_child()
        self['algn_ps_reqs'].remove(rqs)
    
        self['algn_ps_poss'].remove(self['algn_ps_poss'].get_child())
        self['algn_ps_ass'].remove(self['algn_ps_ass'].get_child())

        # attaches pa stuff
        hpaned = self['hpaned_pa_mod']
        self['alignment_signals'].add(sgn)
        self['alignment_requirements'].add(rqs)
        return 


    def set_pa_view(self):
        if self.curr_view == "pa": return
            
        self._detach_ps()
            
        self.prop_assurance.attach_possibilities()
        self.prop_assurance.attach_assertions()
            
        self['notebook_top'].set_current_page(0)
        self['notebook_bottom'].set_current_page(0)            
        self.curr_view = "pa"
        return

    
    def set_ps_view(self):
        if self.curr_view == "ps": return

        # ------------------------------------------------------------
        # 1. Remove PA view
        # ------------------------------------------------------------
            
        # remove signals from pa
        sgn = self.signals.get_top_widget()
        self['alignment_signals'].remove(sgn)

        # remove signals from pa
        rqs = self.requirements.get_top_widget()
        self['alignment_requirements'].remove(rqs)

        # remove possibilities
        poss = self.prop_assurance.detach_possibilities()

        # remove assertions
        ass = self.prop_assurance.detach_assertions()

        # ------------------------------------------------------------
        # 2. Add removed widgets to PS view 
        # ------------------------------------------------------------
        self['algn_ps_sgns'].add(sgn)
        self['algn_ps_reqs'].add(rqs)
        self['algn_ps_poss'].add(poss)
        self['algn_ps_ass'].add(ass)
            
        self['notebook_top'].set_current_page(1)            
        self['notebook_bottom'].set_current_page(1)
        self.curr_view = "ps"
        return

    def set_re_view(self):
        if self.curr_view == "re": return
        self.set_pa_view()
            
        self['notebook_bottom'].set_current_page(2)
        self.curr_view = "re"
        return


    def show_traces_window(self, traces_ctrl, show):
        if self.traces is None: self.traces = TracesView(traces_ctrl)
        return


    pass # end of class
    
    
    

