##  ===========================================================================
##  Author: Roberto Cavada <cavada@fbk.eu>
##
##  Copyright (c) 2005 by FBK 
##
##  This is free software; you can redistribute it and/or
##  modify it under the terms of the GNU Lesser General Public
##  License as published by the Free Software Foundation; either
##  version 2 of the License, or (at your option) any later version.
##
##  This software is distributed in the hope that it will be useful,
##  but WITHOUT ANY WARRANTY; without even the implied warranty of
##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
##  Lesser General Public License for more details.
##
##  You should have received a copy of the GNU Lesser General Public
##  License along with this library; if not, write to the Free Software
##  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA.
##
##  For more information about this software see <http://rat.fbk.eu>
##  or email to the RAT Team <rat@fbk.eu>.
##  Please report bugs to <rat@fbk.eu>.
##
##  ===========================================================================

from gtkmvc.view import View 
from views.entries import PossibilitiesView, AssertionsView
from views.outcomes import OutcomesView
from views.check_panel import CheckingPanelView


import utils.globals

import gtk
import os.path


## ----------------------------------------------------------------------
class PropertyAssuranceView (View):

    """The main view for property assurance. This view contains two
    sub-views for the possibilities and for the assertions.
    Also, it contains a view for checking options and control"""

    GLADE_FILE = os.path.join(utils.globals.GLADE_DIR, "prop_assurance.glade")

    def __init__(self, ctrl):        
        View.__init__(self, ctrl, self.GLADE_FILE, "vbox_top", register=False)
        self.makes_evidence_page(self['notebook_pa'], 0)

        self.poss_are_attached = False
        self.ass_are_attached = False
        self.outs_is_attached = False
        
        ctrl.registerView(self)
        return

    def create_sub_views(self, ctrl_poss, ctrl_ass, ctrl_outcomes, panel_ctrl):
        """This is called by the controller during view's registration.
        @param ctrl_poss is the sub-controller of the Possbilities set
        @param ctrl_ass is the sub-controller of the Assertions set
        @param ctrl_outcomes is the controller for outcomes component
        @param panel_ctrl is the controller for the checking options
        """
        self.possibilities = PossibilitiesView(ctrl_poss)
        self.assertions = AssertionsView(ctrl_ass)
        self.outcomes = OutcomesView(ctrl_outcomes, self)

        self.attach_possibilities()
        self.attach_assertions()
        self.attach_outcomes()

        self.panel = CheckingPanelView(panel_ctrl)
        self['hbox_top'].pack_end(self.panel.get_top_widget(), expand=False, fill=True)
        return

    def detach_possibilities(self):
        """Detaches the possibilities view, and returns the top level widget
        of possibilities. Used by ProjectView when switching between PA to PS view"""
        res = None
        if self.poss_are_attached:
            res = self.possibilities.get_top_widget()
            self['vbox_possibilities'].remove(res)
            self.poss_are_attached = False
            pass
        
        return res

    def detach_assertions(self):
        """Detaches the assertions view, and returns the top level widget
        of assertions. Used by ProjectView when switching between PA to PS view"""
        res = None
        if self.ass_are_attached:
            res = self.assertions.get_top_widget()
            self['vbox_assertions'].remove(res)
            self.ass_are_attached = False
            pass
        
        return res

    def detach_outcomes(self):        
        """Detaches the outcomes view and returns the top-level
        widget. Used to send the outcomes view to its own window"""
        res = None
        if self.outs_is_attached:
            res = self.outcomes.get_top_widget()
            self['vbox_top'].remove(res)
            self.outs_is_attached = False
            pass
        
        return res


    def attach_possibilities(self):
        """Attaches the possibilities view. Used by ProjectView when
        switching between PA to PS view"""
        if not self.poss_are_attached:
            self['vbox_possibilities'].pack_start(self.possibilities.get_top_widget())
            self.poss_are_attached = True
            pass
        
        return

    def attach_assertions(self):
        """Attaches the assertions view. Used by ProjectView when
        switching between PA to PS view"""
        if not self.ass_are_attached:
            self['vbox_assertions'].pack_start(self.assertions.get_top_widget())
            self.ass_are_attached = True
            pass
        
        return

    def attach_outcomes(self):
        """Attaches the outcomes view. Used as initial state and whenever
        the view is embedded from the in-window state"""
        if not self.outs_is_attached:
            wid = self.outcomes.get_top_widget()
            self['vbox_top'].pack_end(wid, expand=False, fill=True)
            self.outs_is_attached = True
            pass
        
        return
    
    def makes_evidence_page(self, nb, page_num):
        texts = ("Possibilities", "Assertions")

        old_page_num = (page_num + 1) % 2
        new_child = nb.get_nth_page(page_num)
        old_child = nb.get_nth_page(old_page_num)

        # makes in evidence the selected page:
        nb.get_tab_label(new_child).set_markup("<b>%s</b>" % texts[page_num])
        nb.set_tab_label_packing(new_child, True, True, gtk.PACK_START)

        # slightly hides the other page:
        nb.get_tab_label(old_child).set_markup(texts[old_page_num])
        nb.set_tab_label_packing(old_child, False, True, gtk.PACK_START)
        return

    def get_current_page(self): return self['notebook_pa'].get_current_page()
    
        
    pass # end of class
    
    
    

