#!/usr/bin/env python

# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================


""" This is basic script. Its main purpose for debugging preliminary
steps, i.e. parsing slim files, performing optional fault injection
and optionally instantiating properties out of a property file.
"""

import booter # import this as first module in scripts

from optparse import OptionParser
import os.path
import sys

from invokers import slim, dump
from utils import debug_info as debug
from utils import exceptions, log
from utils.misc import get_out_fname
import options
import slim_compiler.src.model_extension.core as model_ext_core
import property_management.src.api as properties_api
import property_management.src.constants as properties_const
from property_management.src.properties import Property, PatternInstance, StateFormulaPlaceholderInstance
from property_management.src.formulae import AtomicProposition
from slim_compiler.src.errorManagement.errors import AbstractError
from slim_compiler.src import globals


class Runner (object):
    """Base class to be used by scripts. Actually, derived class
    BaseScript is intended to be used as base class by general tool
    scripts, and ToolScript for SLIM based scripts"""

    __actions = []

    def __init__(self, derived_actions=[]):
        self.__provided_actions = self.__actions + derived_actions

        # creates the parser for options
        self.opt_parser = OptionParser(
            self.get_usage(),
            version="%prog " + ".".join(map(str, options.PROG_VER)))

        # creates a debug info
        self.init_compile_info()
        return

    def init_compile_info(self):
        # creates a debug info
        self.debug_info = debug.create_debug_and_compile_information()

    def get_provided_actions(self):
        """Returns the list of provided actions"""
        return self.__provided_actions

    def run(self):
        """This the method to be invoked to execute the script"""
        # register the options in the whole hierarchy
        self.register_options(self.opt_parser)

        # now it is time to parse the options
        self._parse_options(options)

        # store and check options
        self.check_options(self.opt_parser, options)
        # set all options within module option
        self.store_options(self.opt_parser, options)

        # this comes after logging system has been setup
        try:
            try:
                # do the actual work!
                self.action()
            finally:
                # finally, cleanup and quit
                self.after_action()
        except Exception as e:
            debug.emit_error_and_quit(e, self.debug_info)
        return

    def get_debug_info(self):
        """Returns the internally stored DebugInformation instance"""
        return self.debug_info

    def check_stop_request(self, stop_after_action):
        """This method can be used to check if after given action
        name the execution has to stop. This depends on the user
        request made through the command line with option
        --stop-after. The methods stops the execution if
        needed. Given action must be among those that are supported
        by the script."""
        assert stop_after_action in self.get_provided_actions(), \
            "Unknown action: '%s'" + stop_after_action

        log.system.debug("Successfully executed action '%s'" %
                         stop_after_action)
        if stop_after_action == options.STOP_AFTER:
            log.system.debug("Quitting as requested by the user")
            quit(0)

    # ----------------------------------------------------------------------
    #   To be overridden
    # ----------------------------------------------------------------------
    def get_usage(self):
        return "usage: %prog [options] slim-files"

    def register_options(self, opt_parser):
        """Use this method to register options for the script.
        @param opt_parser is a OptionParser instance that you must
        be used to register the options.  Important: before
        registering the options, call super class method
        register_options"""
        return
    # ----------------------------------------------------------------------
    from utils.log import log_icd

    @log_icd
    def store_options(self, opt_parser, option):
        """Use this method to save all options given throught the
        command line into the given module 'options'. Consider that
        if you register an option with parameter 'dest="OPT_NAME"'
        then the module options will be filled with a variable
        called 'OPT_NAME' automatically. However, under some
        particular condition the values of options might be a
        function of several options given at command line: in this
        case use this function to set their values.

        @param op_parser is an instance of system class optparse.OptionParser

        Important: call super(store_options) before doing anything
        here."""
        # see log_icd decorator
        #print "Call to ICD function runscripts.BaseScript.store_options"
        return
    # ----------------------------------------------------------------------

    def check_options(self, opt_parser, options):
        """This is called after option registration is done to
        check options values. When an error is found, use
        OptionParser.error to break the execution."""
        return
    # ----------------------------------------------------------------------

    def action(self):
        """This code is called when the script can carry out the
        actual work. Important: call super's method action here. If
        an action is through this method, debug.emit_error_and_quit
        will be called and the execution is stopped."""
        return
    # ----------------------------------------------------------------------

    def after_action(self):
        """This method is called after action has been executed, to
        wrap-up, cleanup and quit. super.after_action should be
        called at the end.

        NOTE: this function can be called before self.action is fully
        executed, i.e. be ready for data partially constructed by
        self.action.(This happen if in self.action an exception is
        raised)
        """
        return

    # ----------------------------------------------------------------------
    #   Internals
    # ----------------------------------------------------------------------
    def _parse_options(self, options):
        # sets the given options inside module options
        (opts, args) = self.opt_parser.parse_args()
        for opt, val in opts.__dict__.iteritems():
            if val is not None:
                if hasattr(options, opt):
                    setattr(options, opt, val)
                else:
                    log.system.debug("Not storing automatically option '%s', "
                                     "assuming it will be handled separately" %
                                     opt)

        # the input filenames/positional arguments (alias)
        options.POSITIONAL_ARGS = args
        options.SLIM_FILE_NAMES = args
        return

    pass  # end of class
# ----------------------------------------------------------------------


class BaseScript (Runner):
    """"This is the base class for all tool scripts. It will
    register a default set of options related to input and output."""

    __actions = []

    def __init__(self, derived_actions=[]):
        super(BaseScript, self).__init__(
            derived_actions=self.__actions+derived_actions)

    # ----------------------------------------------------------------------
    def register_options(self, opt_parser):
        super(BaseScript, self).register_options(opt_parser)

        opt_parser.add_option("-f", "--file", dest="BASE_FILE_NAME",
                              help="Write output to files with BASE_FILE_NAME "
                              "base name. Default value is the first "
                              "input file without extension")

        opt_parser.add_option("-d", "--dir", dest="OUT_DIR_NAME",
                              help="Create output/logging files in "
                              "the specified directory")

        opt_parser.add_option("-r", "--root", dest="ROOT_NAME",
                              help="Use given component implementation "
                              "as root")

        opt_parser.add_option("-l", "--level", dest="LOGGING_SYS_LEVEL",
                              help="Set minimum system logging level")

        opt_parser.add_option("-e", "--logfile", dest="LOGGING_FNAME",
                              help="Write logging output to FILE")

        opt_parser.add_option("--quiet", dest="QUIET",
                              action="store_true", default=False,
                              help="Be quiet, do not emit nominal output "
                              "(errors are still emitted though)")

        opt_parser.add_option("--stop-after", dest="STOP_AFTER",
                              help="Stop after given phase: " +
                              ", ".join(self.get_provided_actions()))

        opt_parser.add_option("-a", "--display-call-output",
                              dest="DISPLAY_CALL_OUTPUT",
                              action="store_true", default=False,
                              help="Display call output")

        return

    # ----------------------------------------------------------------------
    def check_options(self, opt_parser, options):
        super(BaseScript, self).check_options(opt_parser, options)

        # set the path to output files and logging files
        if options.OUT_DIR_NAME:
            if not os.path.exists(options.OUT_DIR_NAME):
                opt_parser.error("Fatal Error: directory for output files does"
                                 "not exist : " + options.OUT_DIR_NAME)
        # check stop-after is supported:
        if (options.STOP_AFTER and
                options.STOP_AFTER not in self.get_provided_actions()):
            opt_parser.error("Given action does not exist: '%s'. "
                             "It has to be one of: %s" %
                             (options.STOP_AFTER,
                              ", ".join(self.get_provided_actions())))

        # logging level
        try:
            log.get_level_from_string(options.LOGGING_SYS_LEVEL)
        except ValueError as e:
            opt_parser.error(str(e))

        return

    # ----------------------------------------------------------------------
    def store_options(self, opt_parser, options):
        super(BaseScript, self).store_options(opt_parser, options)

        # set base filename
        if not options.BASE_FILE_NAME:
            topslim = os.path.basename(options.SLIM_FILE_NAMES[0])
            options.BASE_FILE_NAME = os.path.splitext(topslim)[0]

        # user logging level (handling quiet)
        if opt_parser.values.QUIET:
            options.LOGGING_USER_LEVEL = "error"
        elif not options.LOGGING_USER_LEVEL:
            options.LOGGING_USER_LEVEL = "info"

        # set logging file
        if options.LOGGING_FNAME:
            if options.OUT_DIR_NAME:
                options.LOGGING_FNAME = os.path.join(options.OUT_DIR_NAME,
                                                     options.LOGGING_FNAME)

            # absolute path normalization
            options.LOGGING_FNAME = os.path.realpath(options.LOGGING_FNAME)

        # sets up system and user logging
        log.init()

        # prog name and arguments
        options.PROG_NAME = os.path.basename(sys.argv[0])
        import string
        options.PROG_CMDLINE = " ".join(map(string.strip, sys.argv))

        # some logging output
        if options.OUT_DIR_NAME:
            log.system.debug("Set the output directory to '%s'",
                             options.OUT_DIR_NAME)

        log.system.debug("Set the base of output files name to '%s'" %
                         options.BASE_FILE_NAME)
        log.system.debug("Set the logging file name to '%s'" %
                         options.LOGGING_FNAME)
        log.system.debug("Set the logging level to '%s'" %
                         options.LOGGING_SYS_LEVEL)
        log.system.debug("Set the logging user level to '%s'" %
                         options.LOGGING_USER_LEVEL)
        return

    # ----------------------------------------------------------------------
    def action(self):
        super(BaseScript, self).action()


# ----------------------------------------------------------------------
class ToolScript (BaseScript):
    """This is the base class for all SLIM based tool scripts. It will
    register a default set of options related to parsing SLIM files."""

    __actions = ["parsing", "checking", "fault-injection", "smv-dumping"]

    def __init__(self, derived_actions=[]):
        super(ToolScript, self).__init__(
            derived_actions=self.__actions+derived_actions)

        self.compile_info = None
        self.ext_compile_info = None
        self.root_impl = None
        self.formulae_doc = None

        # smv filename
        self.smv_fname = None

        # if mathsat has to be used, as the domain is infinite
        self.infinite = False
        self.need_msat = False

        # if timed variables are used
        self.timed = False

        # this is the set of properties converted into smv
        self.smv_formulae_doc = None

    # ----------------------------------------------------------------------
    def register_options(self, opt_parser):
        super(ToolScript, self).register_options(opt_parser)

        opt_parser.add_option("-s", "--enable-smv-debug",
                              dest="SMV_ENABLE_DEBUG",
                              action="store_true", default=False,
                              help="Enable the generation of SMV "
                              "debugging artifacts")

        opt_parser.add_option("-i", "--injection",
                              dest="FAULT_INJECTION_FILE_NAME",
                              help="Fault injection file")

        opt_parser.add_option("--output-fault-extended",
                              dest="OUTPUT_FAULT_EXTENDED",
                              action="store_true", default=False,
                              help="Outputs the fault extended model into "
                              ".err_slim file. This is purely a "
                              "debugging option.")

        opt_parser.add_option("-p", "--property", dest="PROPERTIES_FILE_NAME",
                              help="Properties file")

        def __check_wordsize(option, opt_str, value, parser):
            if value <= 0:
                parser.error("Word size must be greater than 0.")
            if value > 64:
                parser.error("Word size must be less than or equal to 64.")
            setattr(parser.values, option.dest, value)
            return

        opt_parser.add_option("-w", "--wordsize",
                              action="callback", callback=__check_wordsize,
                              dest="SMV_WORD_SIZE", type="int",
                              help="Size of words used to represent int type")

        opt_parser.add_option("--accept-extended",
                              dest="PARSER_ACCEPT_EXTENDED_MODEL",
                              action="store_true", default=False,
                              help="Allow parser to read extended models. "
                              "This is purely a debugging option.")

        return

    # ----------------------------------------------------------------------
    def check_options(self, opt_parser, options):

        super(ToolScript, self).check_options(opt_parser, options)

        if not options.SLIM_FILE_NAMES:
            opt_parser.error("Slim files were not specified.")
        # RC: Now fail injection is optional: it is up to the scripts to
        # check when mandatory
        #if (not options.FAULT_INJECTION_FILE_NAME and
        #        options.STOP_AFTER not in ("checking", "parsing")):
        #    opt_parser.error("Failure injection file was not specified.")

        for _file in options.SLIM_FILE_NAMES:
            if not os.path.exists(_file):
                opt_parser.error("Fatal Error: Slim file %s not found." %
                                 _file)

        # check that output of fault extended model is only allowed if there
        # is fault injection
        if (options.OUTPUT_FAULT_EXTENDED and
                not options.FAULT_INJECTION_FILE_NAME):
            opt_parser.error("Output of fault extended model is only allowed "
                             "if fault injection file is specified "
                             "(check option -i).")

        return

    # ----------------------------------------------------------------------
    def store_options(self, opt_parser, options):
        super(ToolScript, self).store_options(opt_parser, options)

        return

    def get_compile_info(self):
        return (self.ext_compile_info
                if self.ext_compile_info
                else self.compile_info)

    # ----------------------------------------------------------------------
    def action(self):
        super(ToolScript, self).action()

        # parse and check slim
        self._parse_slim_files()

        # extend with failure
        self._do_fault_injection()

        # find the root implementation
        self._extract_root_implementation()

        # get the formulae if specified
        self._extract_formulae()

        # output debug fault extended SLIM model
        if options.OUTPUT_FAULT_EXTENDED:
            assert self.ext_compile_info, \
                ("Internal error. this option --output-fault-extended is only "
                 "possible if -i was given => extended mode must have "
                 "been created")
            dump.dump_fault_extended_slim_model(
                self.ext_compile_info,
                get_out_fname(options.BASE_FILE_NAME + ".err_slim"))

        # generate smv
        self.smv_fname, self.infinite, self.timed = slim.convert_to_smv(
            self.root_impl,
            generate_time_var=options.SMV_GENERATE_TIME_VAR,
            generate_time_fairness_alt=options.SMV_FAIR_TIMED)

        self.need_msat = self.infinite
        self.debug_info.slim2smv_output = open(self.smv_fname, "r").read()

        # properties conversion
        if options.PROPERTIES_FILE_NAME is not None and options.PROPERTIES_FILE_NAME != '':
            try:
                import property_management.src.properties as prop_api

                loaded_properties = []
                doc = prop_api.parse_properties_document(options.PROPERTIES_FILE_NAME)
                errors = prop_api.get_syntax_errors_in_document(doc)
                if len(errors) == 0:
                    for prop in doc:
                        loaded_properties.append(prop)
                else:
                    raise exceptions.UserException('ERROR!! The properties file contains some syntax errors')

                properties_list = []
                wrong_properties = 0
                for prop_idx, prop in enumerate(loaded_properties):
                    formula = properties_api.create_formula_for_property(prop)
                    errors = properties_api.get_errors_in_atomic_propositions_in_formula(formula, self.root_impl)
                
                    if len(errors) > 0:
                        # Not valid formula
                        print 'Warning!! Property %d contains some errors; indeed it will not be analyzed' %prop_idx
                        wrong_properties = wrong_properties + 1

                        pattern_instance = PatternInstance.create(properties_const.Logic.LTL, "propositional")
                        atomic_prop = AtomicProposition.create(properties_const.Logic.LTL, "false")
                        placeholder = StateFormulaPlaceholderInstance.create(properties_const.Logic.LTL, "P", atomic_prop)
                        pattern_instance.add_placeholder_instance(placeholder)
                        _property = Property.create(pattern_instance, "wrong_property_%d" %wrong_properties)
                        properties_list.append(_property)

                        if options.FSAP_PROPERTY_NUMBER == prop_idx:
                            raise exceptions.UserException("The specified property is not valid.")

                    else:
                        properties_list.append(prop)

                if wrong_properties == len(properties_list):
                    raise exceptions.UserException("ERROR!! All the specified properties are wrong; hence the analysis will not be performed")

                self.smv_formulae_doc = properties_api.create_formulae_document_for_properties(properties_list)
                properties_api.convert_atomic_propositions_to_smv(self.smv_formulae_doc, self.root_impl)
                
            except Exception as e:
                raise exceptions.UserException("ERROR! It was not possible to properly convert the properties file.\n%s" %e)

        self.check_stop_request("smv-dumping")

        return

    def _parse_slim_files(self):
        # check for debugging flags
        globals.PARSER_ACCEPT_EXTENDED_MODEL = \
            options.PARSER_ACCEPT_EXTENDED_MODEL

        # parse the slim files
        self.compile_info = self.debug_info.compile_info
        parsing_result = slim.parse(self.compile_info, options.SLIM_FILE_NAMES)

        # Restore debug flags
        globals.PARSER_ACCEPT_EXTENDED_MODEL = False

        # skip spurious-errors when options.PARSER_ACCEPT_EXTENDED_MODEL is set
        skippableErrors = filter(lambda x: (isinstance(x, AbstractError) and
                                            x.identifier == 'F-6'),
                                 self.compile_info.error_collector.errors)
        realErrors = filter(lambda x: x not in skippableErrors,
                            self.compile_info.error_collector.errors)

        # check parsing result
        if not parsing_result and len(realErrors) != 0:
            slim.emit_errors(self.compile_info)
            quit(-1)

        self.check_stop_request("parsing")

        checking_result = slim.check(self.compile_info)

        if not checking_result:
            slim.emit_errors(self.compile_info)
            if not options.PARSER_ACCEPT_EXTENDED_MODEL:
                quit(-1)
            else:
                log.system.warn('Continuing execution as '
                                'options.PARSER_ACCEPT_EXTENDED_MODEL is set '
                                'and errors might be spurious')

        self.check_stop_request("checking")
        return

    # ----------------------------------------------------------------------
    def _do_fault_injection(self):
        """Creates the extended model.
        @raises model_ext_core.InvalidFaultInjectionDocumentError:
                    If the fault injection document is invalid.
        """
        # Do fault injection. If exceptions are raised due to invalid fault
        # injection documents, convert them to UserException objects and quit.
        # TODO: The type of these exceptions might be changed in the future.
        # See slim.do_fault_injection.
        try:
            if os.path.isfile(options.FAULT_INJECTION_FILE_NAME):
                self.debug_info.fault_injection_file_name = \
                    options.FAULT_INJECTION_FILE_NAME
            self.ext_compile_info = slim.do_fault_injection(self.compile_info)

        except model_ext_core.InvalidFaultInjectionDocumentError, e:
            debug.emit_error_and_quit(exceptions.UserException(str(e)),
                                      self.debug_info)

        # store the error-extended SLIM file
        self.debug_info.extended_compile_info = self.ext_compile_info

        self.check_stop_request("fault-injection")
        return

    def _extract_root_implementation(self):
        cinfo = (self.ext_compile_info
                 if self.ext_compile_info
                 else self.compile_info)
        root_name = None if self.ext_compile_info else options.ROOT_NAME

        self.root_impl = slim.get_root_impl(cinfo, root_name)

        self.debug_info.root_impl = self.root_impl.name
        return

    def _extract_formulae(self):
        self.formulae_doc = slim.get_formulae_doc(self.root_impl,
                                                  options.PROPERTIES_FILE_NAME)
        self.debug_info.formulae_document = self.formulae_doc
        return

    pass  # end of class ToolScript
# ----------------------------------------------------------------------


if __name__ == "__main__":

    main = ToolScript()
    main.run()
