#!/usr/bin/env python

# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================



"""Converter for fault injection documents.

This script converts a given fault injection document from the deprecated
implementation-based format to the component-based format.

In the implementation-based format, error models are assigned to component
implementations.
In the component-based format, error models are assigned to component
instances.

Note: XML comments will be discarded in the resulting document.

This script can be executed directly from the command line.
The fault injection document to be converted must be specified via standard
input. The converted document will be written to standard output.
The document can only be converted with respect to a specific SLIM model.
The name(s) of the file(s) containing that SLIM model must be specified as
arguments. The SLIM model must contain the error models that are referenced in
the fault injection document.

Example usage:
  convert_fault_injection_document.py \
        slim_model.slim \
        < old_fault_injection_doc.xml \
        > new_fault_injection_doc.xml
"""

import booter # import this as first module in scripts

import sys
import os
from utils import log

from lxml import etree

from runscripts import basic
from slim_compiler.src.model_extension import convert
from slim_compiler.src.model_extension import fault_injection


class FaultInjectionDocumentConverter(basic.BaseScript):
    """A script that converts fault injection documents from the
    implementation-based format to the component-based format.
    """

    def check_options(self, opt_parser, options):
        super(FaultInjectionDocumentConverter, self).check_options(opt_parser,
                                                                   options)
        if not options.SLIM_FILE_NAMES:
            opt_parser.error("Slim files were not specified.")
        for _file in options.SLIM_FILE_NAMES:
            if not os.path.exists(_file):
                opt_parser.error(
                        "Fatal Error: Slim file %s not found." % _file)
    
    
    def store_options(self, opt_parser, options):
        super(FaultInjectionDocumentConverter, self).store_options(opt_parser,
                                                                   options)
        if not options.BASE_FILE_NAME:
            topslim = os.path.basename(options.SLIM_FILE_NAMES[0])
            options.BASE_FILE_NAME = os.path.splitext(topslim)[0]
        # Set up system and user logging. 
        log.init()
    

    def action(self):
        """Writes the converted fault injection document to standard output.
        """
        super(FaultInjectionDocumentConverter, self).action()

        if root_impl is None:
            return

        # Parse the fault injection document, which is provided via standard
        # input.
        old_document = convert.parse_implementation_based_injection_document(
                                                                sys.stdin)
        # Convert the fault injection document to the component-based format.
        # Note that the root element of the document must be passed, not the
        # tree itself.
        new_document = convert.create_component_based_injection_document(
                                                old_document, root_impl)
        # Validate the converted fault injection document.
        # Note: The validity will be checked on a syntactic level but also
        # with respect to the SLIM model.
        errors = fault_injection.get_errors_in_injection_document(
                                        new_document, root_impl)
        if len(errors) > 0:
            error_message = ("The converted document is invalid:\n" +
                             "\n".join(map(str, errors)))
            raise Exception(error_message)        
        
        # By default, lxml uses "n0", "n1", ... as namespace prefixes. To get
        # nicer prefixes, we use a dirty trick: Create a new element with a
        # custom namespace prefix map and add the root element of the
        # document as a child element. This causes the custom namespace
        # prefixes to be used when the document is serialized.
        # XXX: This is a dirty hack.
        namespace_map = {"fault": fault_injection._XML_NAMESPACE_URI}
        dirty_root_element = etree.Element("dirty", nsmap=namespace_map)
        dirty_root_element.append(new_document)
        output = etree.tostring(new_document, xml_declaration=True,
                                encoding="UTF-8", pretty_print=True)
        # Print the document to standard output.
        sys.stdout.write(output)


if __name__ == "__main__":
    main = FaultInjectionDocumentConverter()
    main.run()
