#!/usr/bin/env python
# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================



"""Fault Tree Evaluation

TODO: Complete this docstring.
"""
import booter # import this as first module in scripts

import os
import options
import re
from runscripts.basic import BaseScript
from utils import misc, log
from utils.property_checker import DFTEvaluater, PropertyCheckingRecord

class FaultTreeEvaluation (BaseScript):
    """This script executes performance evaluation.
    It replaces performability_evaluation__srs_022.py"""

    _actions = ["to-sigref", "to-imc", "evaluation"]

    def __init__(self, derived_actions=[]):
        super(FaultTreeEvaluation, self).__init__(derived_actions=derived_actions+self._actions)
        
        self.fault_tree_input_filename = None
        self.basic_event_rates_filename = None
        self.mission_time = None
        self.events_output_filename = None
        self.gates_output_filename = None
        
        return
    
    def get_usage(self):
        return "usage: %prog [options] fault_tree_filename basic_event_rates_filename mission_time events_output_filename gates_output_filename"

    def register_options(self, opt_parser):
        super(FaultTreeEvaluation, self).register_options(opt_parser)

        def __check_severity(option, opt_str, value, parser):
            if value < 0: parser.error("Severity must be positive.")
            return

        opt_parser.add_option("-s", "--severity", dest="SEVERITY",
                              action="callback",callback=__check_severity, type="float",
                              help="Set severity of top-level event")

        def __check_tool(option, opt_str, value, parser):
            tools = ["imca", "mrmc"]
            if value == "auto":
                options.MODEL_CHECKERS = tools
            elif value in tools:
                options.MODEL_CHECKERS = [value]
            else:
                parser.error("Invalid tool selected: %s." % value)

            return
        
        opt_parser.add_option("--tool", callback=__check_tool, 
                              action="callback", type="string",
                              default="auto", help="Tool to use for performability analysis (imca, mrmc, auto)")

        def __check_error_bound(option, opt_str, value, parser):
            if value <= 0: parser.error("Error bound must be positive.")
            #TODO: no logging yet
            if value >= 1.0: print "Error bound possibly too high"
            return

        opt_parser.add_option("--error-bound", callback=__check_error_bound, 
                              action="callback", dest="ERROR_BOUND", type="float",
                              default=1e-3, help="Error bound for IMCA process")

        return

    def check_options(self, opt_parser, options):
        super(FaultTreeEvaluation, self).check_options(opt_parser, options)
        
        # Five positional arguments are required:
        # - The name of the fault tree input file.
        # - The name of the file containing the mapping from event names to rates.
        # - The mission time.
        # - The name of the output file for the events in FaultTree+ format.
        # - The name of the output file for the gates in FaultTree+ format.
        if len(options.POSITIONAL_ARGS) != 5: opt_parser.error("Incorrect number of arguments.")

        self.fault_tree_input_filename = options.POSITIONAL_ARGS[0]
        self.basic_event_rates_filename = options.POSITIONAL_ARGS[1]
        
        r = re.compile("^(-?[0-9]+|infinity)[ ]*(msec|sec|min|hour|day)?$")
        m = r.match(options.POSITIONAL_ARGS[2].strip())
        if not m:
            opt_parser.error("The timebound is invalid (use integer optionally followed by time unit)")
        timebound = m.group(1)
        try:
            timebound = int(m.group(1))
            if timebound < 0:
                opt_parser.error("Timebound has to be non-negative")
        except Exception as e:
            if timebound != "infinity":
                opt_parser.error("A time bound is expected but the input is not an integer")

        self.mission_time = options.POSITIONAL_ARGS[2].strip()
        self.events_output_filename = options.POSITIONAL_ARGS[3]
        self.gates_output_filename = options.POSITIONAL_ARGS[4]
    
        return
    

    def action(self):
        super(FaultTreeEvaluation, self).action()

        dft_evaluator = DFTEvaluater(self.fault_tree_input_filename, self.basic_event_rates_filename,
                                     self.events_output_filename, self.gates_output_filename, options.MODEL_CHECKERS)
        dft_evaluator.debug_info = self.debug_info

        (relpaths, error_msg) = dft_evaluator.analyze(self.mission_time, options.SEVERITY, options.ERROR_BOUND)
        if error_msg:
                log.system.error("Possible error during performability analysis")
        output_template = ("The fault tree has been evaluated. The annotated " + 
                   "fault tree can be found in the files %s and %s.")
        log.user.info(output_template % relpaths)

    pass # end of class

# ----------------------------------------------------------------------
        
if __name__ == "__main__":
    
    main = FaultTreeEvaluation()
    main.run()
