#!/usr/bin/env python

# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================



"""Failure Detection Analysis

Check whether a fault is detectable in a model.

Inputs      M : an operational model.
            FM : a fault model.
            FDM : an FDIR model (optional).
            Alm : a set of alarms.
            P : a property (a fault).

Outputs     A set of alarms O (the set of candidate detection means)
            if the fault is detectable. Empty set if the fault is not
            detectable.

Processing  Using model checking techniques, the COMPASS toolset checks
            if a fault is detectable in an operational model, extended
            with a fault model and optionally extended an FDIR
            model. A fault is considered detectable if there exists an
            alarm in the input set of alarms, such that
            every occurrence of the fault eventually causes the
            alarm to be true.
"""

import booter # import this as first module in scripts

import property_management.src.constants as prop_const
import property_management.src.api as prop_api
from utils import fsap_process
from utils import exceptions, log
from utils.misc import relpath, get_out_fname
import options
import slim_compiler.src.smv.api as smv_api
from utils.main_extension_converter import MainExtensionConverter
import subprocess
import os

from runscripts.basic import ToolScript
from runscripts.fault_tree_generation__srs_017 import FsapGeneralScript


class FsapFaultDetection(FsapGeneralScript):
    """This is the script that can be used to run FSAP to perform
    fault detection analysis"""

    def register_options(self, opt_parser):
        # Specific options 
        super(FsapFaultDetection, self).register_options(opt_parser)

        opt_parser.add_option("--property-num", dest="FSAP_PROPERTY_NUMBER",
                              action="store", type = "int", default=0,
                              help="Specifies the number of a property to deal with. Count begins with 0 which is the default value.")

        opt_parser.add_option("--detection-delay", dest="DETECTION_DELAY",
                              action="store", type="string", default=None,
                              help="Specifies the detection delay (do not specify in case of discrete model)")

        opt_parser.add_option("--bmc-length", dest="NUSMV_BMC_LENGTH", 
                              action="store", type="int", default=10,
                              help="Maximum path length for BMC.")
        return

    def check_options(self, opt_parser, options):
        super(FsapFaultDetection, self).check_options(opt_parser, options)
        
        # properties are mandatory
        if not options.PROPERTIES_FILE_NAME:
            opt_parser.error("Properties file was not specified.")
            pass

        # the property number has to be >= 0.
        if options.FSAP_PROPERTY_NUMBER < 0:
            opt_parser.error("The number of a property has to be 0 or greater.")
            pass
        
        return


    def store_options(self, opt_parser, options):
        super(FsapFaultDetection, self).store_options(opt_parser, options)

        if opt_parser.values.DETECTION_DELAY:
            self.DETECTION_DELAY = opt_parser.values.DETECTION_DELAY
        else:
            self.DETECTION_DELAY = -1.0

        assert opt_parser.values.NUSMV_BMC_LENGTH
        options.NUSMV_BMC_LENGTH = opt_parser.values.NUSMV_BMC_LENGTH

        return


    def action(self):
        """Builds FSAP and performing Fault Detection Analysis for one
        propositional property"""
        super(FsapFaultDetection, self).action()

        assert self._alarms_num >= 0, "Tool chain is broken. The number of alarms must have been calculated during failures/alarms dumping"

        if self._alarms_num == 0:
            raise exceptions.UserException("There are no alarms in model. Fault Detection Analysis is meaningless in such situation")
        
        if self.formulae_doc is None:
            raise exceptions.UserException("Properties file is empty")

        if options.FSAP_PROPERTY_NUMBER < 0 or options.FSAP_PROPERTY_NUMBER >= len(self.formulae_doc):
            raise exceptions.UserException("Error: the specified number of a property is outside of the range: %d is not in [0, %d)" %\
                             (options.FSAP_PROPERTY_NUMBER, len(self.formulae_doc)))
        else:
            log.user.info("Fault Detection Analysis is to be performed for property number %d..." %
                          (options.FSAP_PROPERTY_NUMBER))
            pass
  
        orig_prop = self.formulae_doc[options.FSAP_PROPERTY_NUMBER]
        orig_prop_text = prop_api.get_formula_as_smv_string(orig_prop)
        
        smv_prop = self.smv_formulae_doc[options.FSAP_PROPERTY_NUMBER]
        smv_prop_text = prop_api.get_formula_as_smv_string(smv_prop)

        if not prop_api.check_is_propositional_formula(smv_prop):
            raise exceptions.UserException("Fault Detection cannot be computed for non-propositional formulas : %s"\
                             % orig_prop_text)
    
        if self.DETECTION_DELAY is None or self.DETECTION_DELAY == -1.0:
            delay = self.DETECTION_DELAY
        else:
            delay = self.get_delay_value(self.DETECTION_DELAY)

        if delay is not None:

            if delay == -1.0:
                delay = None
                delay_counter_var = None
            else:
                assert delay >= 0.0
                delay_counter_bound = None

                if delay >= 1:
                    delay_counter_bound = delay + 1
                elif delay == 0:
                    delay_counter_bound  = 0.00001
                else:
                    # increase first non-zero digit by one, eg: 0.004823 -> 0.005823
                    factor = 10
                    while delay * factor < 1:
                        factor = factor * 10
                    delay_counter_bound = delay + (1.0 / factor)

                self.extend_smv_with_delay_counter(smv_prop_text, delay_counter_bound, "fsap_delay_counter",
                                                   self.smv_fname)
                delay_counter_var = "fsap_delay_counter"
                log.user.info("Fault Detection is to be performed with delay: %s..." %str(delay))

        else: # delay is None
            delay_counter_var = None


        tmp_detect_fname = self.fsap.compute_fault_detection(smv_prop_text, options.NUSMV_BMC_LENGTH, delay_counter_var, delay)

        # rename the files according to the base name
        detect_fname = get_out_fname(options.BASE_FILE_NAME + ".flt_detect")
  
        # If FSAP execution is unsuccessful then an exception is raise
        # and here we assume the run was successful and file has been generated.
        assert os.path.isfile(tmp_detect_fname),\
               "For unknown reasons FSAP has not generated output files (advice: run with '-l debug' to see FSAP messages)"

        os.rename(tmp_detect_fname, detect_fname)
           
        if not options.FSAP_KEEP_SMV_IDS:
            fsap_process.convert_fault_detection(detect_fname, detect_fname,\
                                                 smv_api.convert_smv_const_to_slim)
            pass

        log.user.info('Fault Detection has been performed for formula "%s" and the result is stored in file "%s"' %
                      (orig_prop_text, relpath(detect_fname)))
        return
    pass # end of class


    def get_delay_value(self, delay_str):

        values = delay_str.split(" ")

        if len(values) == 1:
            # It must be an integer/float
            try:
                delay = float(delay_str)
                return delay
            except ValueError:
                raise exceptions.UserException("Error: the specified delay is not valid. %s is not a float or a "\
                                                   "valid slim expression" %delay_str)
        else:
            # It must be an expression of the type <numeric value> <time unit>
            if len(values) != 2:
                raise exceptions.UserException("Error: the specified delay is not valid. %s is not a float or a "\
                                                   "valid slim expression" %delay_str)
            time_val = values[0]
            #FIXME: using regular expressions to understand whether it is an integer or a float
            if "." not in time_val:
                time_val += ".0"
            time_unit = values[1]
            if time_unit not in ["msec", "sec", "min", "hour", "day"]:
                raise exceptions.UserException("Error: the time unit is not valid. %s must be one of the following "\
                                                   "values: ['msec', 'sec', 'min', 'hour', 'day']" %time_unit)
            delay_str = time_val + time_unit
            smv_delay = smv_api.parse_and_convert_numeric_expression(self.root_impl, delay_str)
            if "f'" in str(smv_delay):
                smv_delay = smv_delay.replace("f'", "")
            delay = eval(smv_delay)
            return delay


    def extend_smv_with_delay_counter(self, smv_prop, delay_counter_bound, delay_counter_var, smv_fname):
        assert delay_counter_bound > 0
        assert smv_fname is not None

        conv = MainExtensionConverter()

        delta_var = None

        from slim_compiler.src.smv.walkers import InfiniteDomainFinderWalker
        walker = InfiniteDomainFinderWalker()
        walker.walk(self.root_impl)

        infinite_domain = walker.has_found_infinite_data()
        timed_domain = walker.has_found_clock_data()

        if timed_domain:
            # count time
            delta_var = conv.delta_var()
        else:
            # count steps
            delta_var = "1"

        delay_counter_bound_str = None
        delay_var_type = None

        if infinite_domain:
            # delay counter bound is float
            delay_counter_bound_str = "%f" % delay_counter_bound
            delay_var_type = "real"
        else:
            # delay bound is int
            assert delay_counter_bound % 1 == 0
            delay_counter_bound_str = "%d" % delay_counter_bound
            delay_var_type = "-1..%d" % delay_counter_bound

        rnd_choice_var = "%s_rnd" % delay_counter_var

        main_extension = \
            "  VAR %s : %s;\n" % (delay_counter_var, delay_var_type) + \
            "  VAR %s : boolean;\n" % rnd_choice_var + \
            "\n" + \
            "  ASSIGN init(%s) :=\n" % (delay_counter_var) + \
            "    case\n" + \
            "      %s & %s : 0;\n" % (smv_prop, rnd_choice_var) + \
            "      TRUE : -1;\n" + \
            "    esac;\n" + \
            "\n"

        main_extension += \
            "  ASSIGN next(%s) :=\n" % (delay_counter_var) + \
            "    case\n" + \
            "      %s = -1 & !next(%s) : -1;\n" % (delay_counter_var, smv_prop) + \
            "      %s = -1 &  next(%s) & !%s : -1;\n" % (delay_counter_var, smv_prop, rnd_choice_var) + \
            "      %s = -1 &  next(%s) &  %s :  0;\n" % (delay_counter_var, smv_prop, rnd_choice_var) + \
            "      %s + %s < %s : %s + %s;\n" % (delay_counter_var, delta_var, delay_counter_bound_str, delay_counter_var, delta_var) + \
            "      TRUE : %s;\n" % (delay_counter_bound_str) + \
            "    esac;\n" + \
            "\n"

        conv.set_main_extension(main_extension)

        out = open(smv_fname, 'w')
        conv.dump(self.root_impl, out)
        out.close()

        return

# ----------------------------------------------------------------------
        
if __name__ == "__main__":
    main = FsapFaultDetection()
    main.run()

    
