#!/usr/bin/env python

# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================



"""Fault Tree Generation

Generates a fault tree given an operational model extended with a
fault model, optionally extended with an FDIR model and a property
representing the top level event.

Inputs      M : an operational model.
            FM : a fault model.
            FDM : an FDIR model (optional).
            P : a set of properties.

Outputs     A fault tree in two files, one with the events and one with gates.

Processing  Using model checking techniques, the COMPASS toolset
            generates a fault tree staring from an operational model
            extended with a fault model, optionally extended with an
            FDIR model, and a property representing the top level
            event. The fault tree is returned as two files, one
            with events and another with gates, in the formal suitable
            for the fault tree viewer.
"""

import booter # import this as first module in scripts

import property_management.src.constants as prop_const
import property_management.src.api as prop_api
from utils import fsap_process
from utils import exceptions, log
from utils.misc import relpath, get_out_fname
import utils.fault_tree as fault_tree_module
from invokers import dump
import options
import slim_compiler.src.smv.api as smv_api

import subprocess
import os

from runscripts.basic import ToolScript
from runscripts.model_checking__srs_015 import NuSMVBasedScript

class FsapGeneralScript (NuSMVBasedScript):
    """This script can be used to create a FSAP process.
    Whatever will be done with the FSAP process depends on the derived classes."""

    # the number of observables found during the failures-observables-alarms dump
    _observables_num = -1 # default incorrect value is -1
    _alarms_num = -1 # default incorrect value is -1

    def __init__(self):
        super(FsapGeneralScript, self).__init__()

        self.fsap = None  # the FSAP process to control
        self.fail_obs_alm_fname = None # name of a file with failures, observables and alarms
        return


    def register_options(self, opt_parser):
        # Specific options for FSAP
        super(FsapGeneralScript, self).register_options(opt_parser)

        opt_parser.add_option("--keep-smv-ids", dest="FSAP_KEEP_SMV_IDS",
                              action="store_true", default=False,
                              help="Disables restoring identifiers in the generated files to their original SLIM form (default behavior). "\
                              "With this option the identifiers will be in the SMV (internally used) form.")
        opt_parser.add_option("--output-xml-and-rates",
                              dest="FSAP_OUTPUT_FAULT_TREE_XML_AND_RATES",
                              action="store_true", default=False,
                              help="Output the fault tree in XML format " +
                                   "and output the rates of the basic " +
                                   "events. This output can be used for " +
                                   "Fault Tree Evaluation and Fault Tree " +
                                   "Verification.")
        return


    def check_options(self, opt_parser, options):

        super(FsapGeneralScript, self).check_options(opt_parser, options)

        if not options.FAULT_INJECTION_FILE_NAME:
            opt_parser.error("Failure injection file was not specified.")
            pass
        pass


    def action(self):
        super(FsapGeneralScript, self).action()
        
        # generate the failures/observables/alarms file
        self.fail_obs_alm_fname = get_out_fname(options.BASE_FILE_NAME + ".fail_obs_alm")
        
        fo_file = open(self.fail_obs_alm_fname, "w")
        fails, num_obs, num_alm = smv_api.dump_failures_observations_alarms(self.root_impl, fo_file)
        self._observables_num = num_obs # the found number of observables
        self._alarms_num = num_alm # the found number of alarms
        fo_file.close()
        
        fo_file = open(self.fail_obs_alm_fname, "r")
        self.debug_info.failures_observations_alarms = fo_file.read()
        fo_file.close()

        logfname = os.path.splitext(self.smv_fname)[0] + ".fsapout"
        log.system.debug("Invoking FSAP, with redirected to " + logfname)

        # creates the process
        self.fsap = fsap_process.FsapProcess(self.smv_fname, self.fail_obs_alm_fname,
                                             need_msat=self.need_msat, 
                                             stdout=subprocess.PIPE, stderr=subprocess.STDOUT, 
                                             dumpers=[open(logfname, "w")])

        # set a prefix for temporary files.
        # this prefix is unusual enough not to interfere with normal files
        self.fsap.set_tmp_filename_prefix(get_out_fname(options.BASE_FILE_NAME + "__temp_fsap__"))

        return


    def after_action(self):
        """Quits the process"""
        if self.fsap: self.fsap.kill()
        super(FsapGeneralScript, self).after_action()
        return

    pass # end of class
# ----------------------------------------------------------------------


class FsapFaultDetection (FsapGeneralScript):
    """This is the script that can be used to generate a fault tree
    for a propositional property by FSAP"""

    def register_options(self, opt_parser):
        # Specific options for Fault Detection
        super(FsapFaultDetection, self).register_options(opt_parser)

        opt_parser.add_option("--property-num", dest="FSAP_PROPERTY_NUMBER",
                              action="store", type = "int", default=0,
                              help="Specifies the number of a property to deal with. Count begins with 0 which is the default value.")

        opt_parser.add_option("--dynamic", dest="FSAP_USE_DYNAMIC",
                              action="store", default=False,
                              help="Enable the generation of Dynamic Fault Trees")
        return


    def check_options(self, opt_parser, options):
        super(FsapFaultDetection, self).check_options(opt_parser, options)
        
        # properties are mandatory
        if not options.PROPERTIES_FILE_NAME:
            opt_parser.error("Properties file was not specified.")
            pass

        # the property number has to be >= 0.
        if options.FSAP_PROPERTY_NUMBER < 0:
            opt_parser.error("The number of a property has to be 0 or greater.")
            pass
            
        return


    def action(self):
        """Builds FSAP and performing Fault Tree generation for one
        propositional properties"""
        super(FsapFaultDetection, self).action()
        
        if self.formulae_doc is None:
            raise exceptions.UserException("Properties file is empty")

        if options.FSAP_PROPERTY_NUMBER < 0 or options.FSAP_PROPERTY_NUMBER >= len(self.formulae_doc):
            raise exceptions.UserException("Error: the specified number of a property is outside of the range: %d is not in [0, %d)" %\
                             (options.FSAP_PROPERTY_NUMBER, len(self.formulae_doc)))
        else:
            log.user.info("Fault Tree is to be generated for property number %d..." %
                          (options.FSAP_PROPERTY_NUMBER))
            pass

        orig_prop = self.formulae_doc[options.FSAP_PROPERTY_NUMBER]
        orig_prop_text = prop_api.get_formula_as_smv_string(orig_prop)

        smv_prop = self.smv_formulae_doc[options.FSAP_PROPERTY_NUMBER]
        smv_prop_text = prop_api.get_formula_as_smv_string(smv_prop)

        if not prop_api.check_is_propositional_formula(smv_prop):
            raise exceptions.UserException("Fault Tree cannot be computed for non-propositional formula : %s"\
                             % orig_prop_text)

        # Generates the FT
        if options.FSAP_USE_DYNAMIC:
            (tmp_ev_fname, tmp_gt_fname, tmp_xml_fname) = self.fsap.compute_dyn_fault_tree(smv_prop_text)
        else: (tmp_ev_fname, tmp_gt_fname, tmp_xml_fname) = self.fsap.compute_fault_tree(smv_prop_text) 

        # rename the files according to the base name
        ev_fname = get_out_fname(options.BASE_FILE_NAME + ".flt_events")
        gt_fname = get_out_fname(options.BASE_FILE_NAME + ".flt_gates")

        # If FSAP execution is unsuccessful then an exception is raise
        # and here we assume the run was successful and files have been generated.
        assert os.path.isfile(tmp_ev_fname) or os.path.isfile(tmp_gt_fname),\
               "Fault Tree files have not been generated by FSAP for unknown reason (advice: run with '-l debug' to see FSAP messages)"

        os.rename(tmp_ev_fname, ev_fname)
        os.rename(tmp_gt_fname, gt_fname)
        
        if not options.FSAP_KEEP_SMV_IDS:
            fsap_process.convert_fault_tree_events(
                    ev_fname, ev_fname, smv_api.convert_smv_const_to_slim)
            fsap_process.convert_fault_tree_gates(
                    gt_fname, gt_fname, smv_api.convert_smv_const_to_slim)
            pass

        log_template = ('%sFault Tree has been generated for formula "%s". ' +
                        'It has been stored in FaultTree+ format in files '+
                        '"%s" and "%s".')
        log.user.info(log_template %
                      ("Dynamic " if options.FSAP_USE_DYNAMIC else "", 
                       orig_prop_text, relpath(ev_fname), relpath(gt_fname)))
        
        # If the corresponding option is set, store the fault tree in XML
        # format and also the rates of the basic events.
        if options.FSAP_OUTPUT_FAULT_TREE_XML_AND_RATES:
            # Rename the fault tree XML file according to the base name.
            xml_fname = get_out_fname(options.BASE_FILE_NAME + "_flt.xml")
            os.rename(tmp_xml_fname, xml_fname)
            
            # Determine the rates of the basic events of the fault tree and
            # store them in a separate file. This is needed in order to
            # perform further analyses on the fault tree (fault tree
            # evaluation or fault tree verification).
            # Note: At this point, the event descriptions in the fault tree
            # XML file must still be in SMV syntax.
            fault_tree = fault_tree_module.parse_xml(xml_fname)
            try:
                event_to_rate_dict = fault_tree_module.\
                        create_basic_event_name_to_rate_dict(fault_tree,
                                                             self.root_impl)
            except fault_tree_module.FaultTreeError, e:
                log.user.info('Could not determine the rates of the basic ' +
                              'events, because an error occurred:\n' +
                              str(e))
            else:
                # The rates could be determined. Write them to a file.
                rates_fname = get_out_fname(options.BASE_FILE_NAME +
                                            "_flt.rates_map")
                dump.dump_event_name_to_rate_map(rates_fname,
                                                 event_to_rate_dict)
                rates_log_template = ('The rates of the basic events have ' +
                                      'been stored in file "%s".')
                log.user.info(rates_log_template % relpath(rates_fname))
            
            if not options.FSAP_KEEP_SMV_IDS:
                fsap_process.convert_fault_tree_xml(
                        xml_fname, xml_fname,
                        smv_api.convert_smv_const_to_slim)

            xml_log_template = ('%sFault Tree has been stored in XML ' +
                                'format in file "%s".')
            log.user.info(xml_log_template %
                          ("Dynamic " if options.FSAP_USE_DYNAMIC else "", 
                           relpath(xml_fname)))

        return
    pass # end of class
# ----------------------------------------------------------------------
        
if __name__ == "__main__":
    main = FsapFaultDetection()
    main.run()

    
