#!/usr/bin/env python
# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================
import booter

from runscripts.basic import ToolScript
from utils import misc, log
from invokers import dump
import options

from lxml import etree
from property_management.src import api as property_api, formulae as formulae_api
from invokers import slimsim
from property_management.src.constants import Logic
import copy

class MonteCarloSimulation (ToolScript):
    _actions = ["convert-slim", "convert-property", "simulation"]

    def __init__(self, derived_actions=[]):
        super(MonteCarloSimulation, self).__init__(derived_actions=derived_actions+self._actions)

    def register_options(self, opt_parser):
        # Specific options for Monte Carlo simulation
        super(MonteCarloSimulation, self).register_options(opt_parser)

        opt_parser.add_option("-c", "--confidence", dest="MC_CONFIDENCE",
                              action="store", type = "float", default=options.MC_CONFIDENCE,
                              help="Specifies the confidence of the simulation output. Default is " + str(options.MC_CONFIDENCE))

        opt_parser.add_option("--error-bound", dest="MC_ERRORBOUND",
                              action="store", type = "float", default=options.MC_ERRORBOUND,
                              help="Specifies the error bound for the simulation. Default is " + str(options.MC_ERRORBOUND))

        opt_parser.add_option("--interactive", dest="MC_INTERACTIVE",
                              action="store_true",
                              help="Enables interactive simulation")

        opt_parser.add_option("--strategy", dest="MC_STRATEGY",
                              action="store", type = "string", default=options.MC_STRATEGY,
                              help="Specifies the strategy used during the simulation. Default is " + str(options.MC_STRATEGY))

    def store_options(self, opt_parser, options):
        super(MonteCarloSimulation, self).store_options(opt_parser, options)

        (opts, _) = opt_parser.parse_args()
        options.MC_INTERACTIVE = bool(opts.MC_INTERACTIVE)
        if options.MC_INTERACTIVE:
            options.MC_STRATEGY = 'input'

    def check_options(self, opt_parser, options):
        super(MonteCarloSimulation, self).check_options(opt_parser, options)

        # properties are mandatory
        if not options.PROPERTIES_FILE_NAME:
            opt_parser.error("Properties file was not specified.")

        if options.MC_CONFIDENCE <= 0 or options.MC_CONFIDENCE >= 1.0:
            opt_parser.error("Confidence must be between 0 and 1")

        if options.MC_ERRORBOUND <= 0 or options.MC_CONFIDENCE >= 1.0:
            opt_parser.error("Error bound must be between 0 and 1")

    def action(self):
        #super(MonteCarloSimulation, self).action()

        # parse and check slim
        self._parse_slim_files()

        # extend with failure
        self._do_fault_injection()

        # find the root implementation
        self._extract_root_implementation()

        # get the formulae if specified
        self._extract_formulae()

        # output debug fault extended SLIM model
        if options.OUTPUT_FAULT_EXTENDED:
            assert self.ext_compile_info, "Internal error. this option --output-fault-extended is only possible if -i was given => extended mode must have been created"
            dump.dump_fault_extended_slim_model(self.ext_compile_info, misc.get_out_fname(options.BASE_FILE_NAME + ".err_slim"))

        cinfo = self.ext_compile_info if self.ext_compile_info else self.compile_info


        slimsim.transform_slim(cinfo)
        #print SimulatorPrintVisitor().visit(cinfo)

        xml_fname = misc.get_out_fname(options.BASE_FILE_NAME + "_sim.xml")
        f_py2xml = open(xml_fname, "w")
        f_py2xml.write(cinfo.to_XML())
        f_py2xml.close()
        self.debug_info.slimsim_model = cinfo.to_XML()

        self.check_stop_request("convert-slim")

        assert self.formulae_doc is not None, "Needs formulae doc"

        property_api.convert_time_bounds_for_performability(self.formulae_doc)
        formulae = [formula for formula in self.formulae_doc if formula.get_logic() == Logic.CSL and etree.QName(formula[0].tag).localname not in ['et', 'lra']]

        for formula in formulae:
            orig_formula = copy.deepcopy(formula)
            slimsim.transform_formula(cinfo, self.root_impl, formula)

            xml_propfname = misc.get_out_fname(options.BASE_FILE_NAME + "_props.xml")
            f_prop2xml = open(xml_propfname, "w")
            sim_formulae = formulae_api.FormulaeDocument.create([formula])
            f_prop2xml.write( etree.tostring(sim_formulae, pretty_print=True, xml_declaration=True) )
            f_prop2xml.close()

            self.check_stop_request("convert-property")

            (code, output) = slimsim.run(xml_fname, xml_propfname, options.MC_STEPBOUND, options.MC_ERRORBOUND,
                                       options.MC_CONFIDENCE, options.MC_STRATEGY)
            self.debug_info.slimsim_output = output

            if code == 0:
                probs = slimsim.extract_results_from_output(output[0])
                if probs:
                    self.print_results(orig_formula, probs)
                else:
                    log.user.error("Simulation error")
            else:
                log.user.error("Simulation error")
                break;

        self.check_stop_request("simulation")

        return



    def print_results(self, formula, probs):
        desc = property_api.get_formula_as_mrmc_string(formula)
        log.user.info("Monte Carlo Simulation of '%s' returns\n%s" % (desc, probs))
        return

if __name__ == "__main__":
    main = MonteCarloSimulation()
    main.run()
