#!/usr/bin/env python

# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================


"""Model Checking

Verify whether a model (optionally extended with a fault and/or FDIR
model) satisfies a property.

Inputs      M : an operational model.
            FM : a fault model (optional).
            FDM : an FDIR model (optional).
            P : a set of properties.

Outputs     Ok shall be returned if the property holds, a counterexample
            trace Tr otherwise.

Processing  Using model checking techniques, the COMPASS toolset
            shall check if the property holds in the given
            operational model (optionally extended with a fault
            model and an FDIR model). A counterexample in XML
            format shall be generated if the property does not
            hold.
"""

import booter  # import this as first module in scripts

import property_management.src.api as properties_api
import property_management.src.constants as prop_const

from invokers import nusmv
from utils.nusmv_process import NuSMVProcess
from utils import log
from utils.misc import relpath
import options

import subprocess
import os.path

from runscripts.basic import ToolScript


class NuSMVBasedScript (ToolScript):
    """This script can be used to generate smv and ordering files
    out of a set of slim files"""

    __actions = ["ord-dumping"]

    def __init__(self, derived_actions=[]):
        super(NuSMVBasedScript, self).__init__(
            derived_actions=self.__actions+derived_actions)

        self.ord_fname = None

    def action(self):
        super(NuSMVBasedScript, self).action()

        # BDD ordering
        if not self.need_msat:
            self.ord_fname = nusmv.launch_smv_to_bdd_ordering(self.smv_fname)
        self.check_stop_request("ord-dumping")
        return

    pass  # end of class
# ----------------------------------------------------------------------


class NuSMVGeneral (NuSMVBasedScript):
    """This script can be used to create a nusmv process out of a
    smv (and optional ordering) file. Whatever will be done with
    the nusmv process depends on the derived classes"""

    def __init__(self, derived_actions=[]):
        super(NuSMVGeneral, self).__init__(derived_actions=derived_actions)

        self.nusmv = None  # the nusmv process to control
        return

    def action(self):
        super(NuSMVGeneral, self).action()

        logfname = os.path.splitext(self.smv_fname)[0] + ".nusmvout"
        log.system.debug("Invoking NuSMV, with redirected to " + logfname)

        # creates the process
        self.nusmv = NuSMVProcess(self.smv_fname,
                                  need_msat=self.need_msat,
                                  stdout=subprocess.PIPE,
                                  stderr=subprocess.STDOUT,
                                  dumpers=[open(logfname, "w")])

        log.system.debug("NuSMV has finished")
        return

    def after_action(self):
        """Quits the process"""
        if self.nusmv:
            self.nusmv.kill()
            self.nusmv = None

        super(NuSMVGeneral, self).after_action()
        return

    pass  # end of class
# ----------------------------------------------------------------------


class NuSMVModelChecking (NuSMVGeneral):
    """This is the script that can be used to perform model
    checking with nusmv"""

    def register_options(self, opt_parser):
        # Specific options for nusmv
        super(NuSMVGeneral, self).register_options(opt_parser)

        opt_parser.add_option("--bmc", dest="NUSMV_USE_BMC",
                              action="store_true", default=False,
                              help="use Bounded Model Checking when "
                              "performing LTL model checking")

        def __check_positive(option, opt_str, value, parser):
            if value < 0:
                parser.error("BMC problem length must be positive.")
            setattr(parser.values, option.dest, value)
            return

        opt_parser.add_option(
            "--bmc-length",
            action="callback", callback=__check_positive,
            dest="NUSMV_BMC_LENGTH", type="int",
            help="Problem max length used when performing BMC")

        opt_parser.add_option(
            "--klive-bound",
            action="callback", callback=__check_positive,
            dest="NUSMV_KLIVE_BOUND", type="int",
            help="Problem max bound for kliveness used when "
            "performing MITL checking")

        return

    def check_options(self, opt_parser, options):
        super(NuSMVModelChecking, self).check_options(opt_parser, options)

        # properties are mandatory
        if not options.PROPERTIES_FILE_NAME:
            opt_parser.error("Properties file was not specified.")

        return

    def action(self):
        """Builds nusmv and performs model checking on all
        properties that is possible to model check"""
        super(NuSMVModelChecking, self).action()

        last_logic_mitl = False

        if self.formulae_doc is not None:
            res = []

            # properties are grouped to be either mitl or not, to
            # minimize slim reconstruction and process reinit
            _list = sorted(
                zip(self.formulae_doc, self.smv_formulae_doc),
                key=lambda f: f[0].get_logic() == prop_const.Logic.MITL)

            for formula, smv_formula in _list:
                is_mitl = formula.get_logic() == prop_const.Logic.MITL
                if last_logic_mitl != is_mitl:
                    options.SMV_FAIR_TIMED = not is_mitl
                    options.SMV_GENERATE_TIME_VAR = is_mitl
                    last_logic_mitl = is_mitl

                    # reset and regenerate slim
                    self.after_action()
                    self.init_compile_info()
                    super(NuSMVModelChecking, self).action()

                hold, trace = nusmv.model_checking(self.nusmv, smv_formula)

                res.append((properties_api.get_formula_as_smv_string(formula),
                            hold, trace))
                if self.nusmv and not self.nusmv.is_alive():
                    log.system.error("Process died unexpectedly. "
                                     "Aborting remaining checks")
                    break

            # show results:
            for form, hold, trace in res:
                if hold:
                    log.user.info("Model Checking: Formula '%s' is true", form)
                elif hold is None:
                    log.user.info("Model Checking: Formula '%s' is unknown",
                                  form)
                else:
                    log.user.info("Model Checking: Formula '%s' is false",
                                  form)
                    log.user.info("Counterexample in file '%s'",
                                  relpath(trace))

        return

    pass  # end of class
# ----------------------------------------------------------------------

if __name__ == "__main__":

    main = NuSMVModelChecking()
    main.run()
