#!/usr/bin/env python

# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================
import booter # import this as first module in scripts

from utils import log, misc
from utils.property_checker import SLIMPropertyChecker, PropertyCheckingRecord
from invokers import slim
from property_management.src import properties as properties_module
from property_management.src import api as properties_api
from runscripts.model_checking__srs_015 import NuSMVBasedScript
import options
import os

#ERROR REPORT MODES
REPORT_VIA_LOG = 0
REPORT_VIA_RETURN = 1
ERROR_REPORT_MODES = [REPORT_VIA_LOG, REPORT_VIA_RETURN]

class PerformabilityEvaluation (NuSMVBasedScript):
    """This script executes performance evaluation.
    It replaces performability_evaluation__srs_022.py"""

    _actions = ["to-sigref", "to-imc", "evaluation"]

    def __init__(self, derived_actions=[]):
        super(PerformabilityEvaluation, self).__init__(derived_actions=derived_actions+self._actions)
        return

    def register_options(self, opt_parser):
        super(PerformabilityEvaluation, self).register_options(opt_parser)

        opt_parser.add_option("--stepwise", dest="STEPWISE",
                              action="store_true", default=False,
                              help="Calculate probabilities stepwise")

        def __check_positive(option, opt_str, value, parser):
            if value < 1: parser.error("Granularity must be positive.")
            return

        opt_parser.add_option("--granularity", callback=__check_positive, 
                              action="callback", dest="STEPWISE_GRANULARITY", type="int",
                              default=10, help="Number of steps in stepwise calculation")
        
        def __check_tool(option, opt_str, value, parser):
            tools = ["imca", "mrmc"]
            if value == "auto":
                options.MODEL_CHECKERS = tools
            elif value in tools:
                options.MODEL_CHECKERS = [value]
            else:
                parser.error("Invalid tool selected: %s." % value)

            return
        
        opt_parser.add_option("--tool", callback=__check_tool, 
                              action="callback", type="string",
                              default="auto", help="Tool to use for performability analysis (imca, mrmc, auto)")
        
        def __check_error_bound(option, opt_str, value, parser):
            if value <= 0: parser.error("Error bound must be positive.")
            #TODO: no logging yet
            if value >= 1.0:
                print "Error bound possibly too high"
            options.ERROR_BOUND = value
            return

        opt_parser.add_option("--error-bound", callback=__check_error_bound, 
                              action="callback", dest="ERROR_BOUND", type="float",
                              help="Error bound for IMCA process")

        return

    def check_options(self, opt_parser, options):
        super(PerformabilityEvaluation, self).check_options(opt_parser, options)
        
        # properties are mandatory
        if not options.PROPERTIES_FILE_NAME:
            opt_parser.error("Properties file was not specified.")
            pass
        return
    

    def action(self):
        super(PerformabilityEvaluation, self).action()
        
        # Check if the root has no event or data ports
        self.check_root_closed(self.root_impl)
        # Check if the root is nonhybrid (no infinite domain)
        self.check_nonhybrid(self.root_impl)
        
        # JK: This is obviously not exactly a nice way of doing things,
        # but for consistency reasons I chose to stick with the options module
        slim_filepaths = options.SLIM_FILE_NAMES
        fault_injection_filepath = options.FAULT_INJECTION_FILE_NAME
        properties_filepath = options.PROPERTIES_FILE_NAME
        tools = options.MODEL_CHECKERS
        stepwise = options.STEPWISE
        error_bound = options.ERROR_BOUND
        granularity = 1
        if stepwise:
            granularity = options.STEPWISE_GRANULARITY

        perf_evaluater = SLIMPropertyChecker(slim_filepaths, fault_injection_filepath, properties_filepath, tools)
        
        # TODO: modifying internal data
        perf_evaluater.debug_info = self.debug_info
        perf_evaluater.cinfo = self.compile_info
        perf_evaluater.ext_cinfo = self.ext_compile_info
        
        perf_evaluater.smv_fname = self.smv_fname
        perf_evaluater.ord_fname = self.ord_fname
        
        # TODO: calling internal functionality
        formulae_doc = slim.get_formulae_doc(self.root_impl, properties_filepath)
        formulae_doc = perf_evaluater._fix_formulae_doc_wrt_atomic_proposition_templates(formulae_doc)

        #perf_evaluater.gen_smv not required and can be skipped
        perf_evaluater.gen_sigref(self.root_impl, formulae_doc)
        
        self.check_stop_request("to-sigref")
        
        perf_evaluater.gen_property_record()
        perf_evaluater.gen_model_files(formulae_doc)
        
        self.check_stop_request("to-imc")

        for property in properties_module.parse_properties_document(properties_filepath):
            (res, error_msg) = perf_evaluater.analyze(property, error_bound, stepwise, granularity)
            self.print_results(property, res)
            if error_msg:
                log.user.warning("Possible error during performability analysis")
        
        self.check_stop_request("evaluation")

        return

    def check_nonhybrid(self, root_impl):
        from slim_compiler.src.smv.walkers import InfiniteDomainFinderWalker 
        walker = InfiniteDomainFinderWalker()
        walker.walk(root_impl)
        infinite_domain = walker.has_found_infinite_data()
        if infinite_domain:
            # raise exceptions.UserException("SLIM model has an infinite domain, performability analysis is not possible.")
            log.system.error("SLIM model has an infinite domain, performability analysis is not possible.")
            #log.user.error("SLIM model has an infinite domain, performability analysis is not possible.")
            import sys
            sys.exit(1)

    def check_root_closed(self, root_impl):
        type = root_impl.parent.get_component_type(root_impl.name.split(".")[0])
        port_count = type.get_total_number_of_outgoing_data_ports() + type.get_total_number_of_incoming_data_ports() + \
                 type.get_total_number_of_outgoing_event_ports() + type.get_total_number_of_incoming_event_ports()
        if port_count > 0:
            log.system.warning("The root component contains ports, performability assumes a closed world.")
            log.user.warning("The root component contains ports, performability assumes a closed world.")
        pass

    def print_results(self, formula, probs):
        desc = formula.description
        if desc == '':
            desc = formula.name
        log.user.info("Performability Evaluation of '%s' returns\n%s" % (desc, probs))
        return

    pass # end of class

# ----------------------------------------------------------------------
        
if __name__ == "__main__":
    
    main = PerformabilityEvaluation()
    main.run()
