#!/usr/bin/env python

# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================


""" TFPG Completeness Analysis """

import booter  # import this as first module in scripts

import options, tempfile, os, subprocess
from utils import log

from runscripts.model_checking__srs_015 import NuSMVGeneral
from invokers import nusmv
from utils.misc import get_out_fname

from utils.fame_tfpg_process import FameTfpgProcess


class TfpgCompleteness (NuSMVGeneral):

    def register_options(self, opt_parser):
        # Specific options for nusmv
        super(TfpgCompleteness, self).register_options(opt_parser)

        opt_parser.add_option("--tfpg-file", dest="TFPG_FILE_NAME",
                              help="TFPG xml file")

        opt_parser.add_option("--associations-file", dest="TFPG_ASSOC_FILE_NAME",
                              help="TFPG SLIM associations file")

        opt_parser.add_option("--bmc-length", dest="NUSMV_BMC_LENGTH",
            action="store", type="int", default=10, help="Maximum path length for BMC.")

        return

    def check_options(self, opt_parser, options):
        super(TfpgCompleteness, self).check_options(opt_parser, options)
        
        if not options.FAULT_INJECTION_FILE_NAME:
            opt_parser.error("Failure injection file was not specified.")

        if not options.TFPG_FILE_NAME:
            opt_parser.error("The TFPG file was not specified.")
        
        if not options.TFPG_ASSOC_FILE_NAME:
            opt_parser.error("The TFPG SLIM associations file was not specified.")
 
        return

    def store_options(self, opt_parser, options):
        super(TfpgCompleteness, self).store_options(opt_parser, options)

        assert opt_parser.values.TFPG_FILE_NAME
        options.TFPG_FILE_NAME = opt_parser.values.TFPG_FILE_NAME

        assert opt_parser.values.TFPG_ASSOC_FILE_NAME
        options.TFPG_ASSOC_FILE_NAME = opt_parser.values.TFPG_ASSOC_FILE_NAME

        assert opt_parser.values.NUSMV_BMC_LENGTH
        options.NUSMV_BMC_LENGTH = opt_parser.values.NUSMV_BMC_LENGTH

        return

    def action(self):
        """Builds nusmv and performs model checking on all
        properties that is possible to model check"""

        super(TfpgCompleteness, self).action()

        from tfpg2smv.tfpg2smv import dump_extended_sys_as_smv
        from tfpg2smv.tfpg2smv import parse_to_internal_struct
        from tfpg2smv.tfpg2smv import get_slim_associations_from_file

        log_out, logfname = tempfile.mkstemp(suffix='.fametfpgout',
            prefix=os.path.splitext(self.smv_fname)[0], dir=".", text=True)

        print "log_out:",log_out
        print "logfname:",logfname

        log.system.debug("Invoking FAME-TFPG, with redirected to " + logfname)

        # parse TFPG xml
        print "--> parsing TFPG file"
        tfpg = parse_to_internal_struct(options.TFPG_FILE_NAME)

        # parse associations file
        print "--> parsing associations file"
        associations = get_slim_associations_from_file(options.TFPG_ASSOC_FILE_NAME)

        # dump extended file
        print "--> dumping extended smv file"
        dump_extended_sys_as_smv(self.root_impl, tfpg, associations, self.smv_fname)

        # instantiate TFPG FAME process
        print "--> instantiating tfpg fame process"
        fame_tfpg_proc = FameTfpgProcess(self.smv_fname, ord_fname=None, \
            need_msat=True, args=[], stdout=subprocess.PIPE,             \
            stderr=subprocess.STDOUT, dumpers=[os.fdopen(log_out, "w")])
        fame_tfpg_proc.set_tmp_filename_prefix(get_out_fname(options.BASE_FILE_NAME + "__temp_fametfpg__"))

        # call completeness validation function
        print "--> calling completeness validation function"
        res = fame_tfpg_proc.check_tfpg_completeness(tfpg, options.NUSMV_BMC_LENGTH)

        # present results
        print "Result:", res['result']
        print "Trace:", res['counterexample']

        # close process
        fame_tfpg_proc.kill()

        return

    pass  # end of class
# ----------------------------------------------------------------------

if __name__ == "__main__":

    main = TfpgCompleteness()
    main.run()

