#!/usr/bin/env python

# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================


"""State Zeno Check

Verify whether a model (optionally extended with a fault and/or FDIR
model) contains zenoness conditions.

TBD
"""

import booter  # import this as first module in scripts

from slim_compiler.src.smv.walkers import ModeCollectorWalker
import slim_compiler.src.smv.api as smvapi

from utils import log

from invokers import nusmv
import options

from runscripts.model_checking__srs_015 import NuSMVGeneral


class NuXMVZenoChecking (NuSMVGeneral):
    """This is the script that can be used to perform zeno
    and other time-related checking with nuXmv"""

    def register_options(self, opt_parser):
        # Specific options for nusmv
        super(NuSMVGeneral, self).register_options(opt_parser)

        def __check_positive(option, opt_str, value, parser):
            if value < 0:
                parser.error("Value must be positive.")
            setattr(parser.values, option.dest, value)
            return

        opt_parser.add_option(
            "--bmc-length",
            action="callback", callback=__check_positive,
            dest="NUSMV_BMC_LENGTH", type="int",
            help="Problem max length used when performing BMC")

        opt_parser.add_option(
            "--klive-bound",
            action="callback", callback=__check_positive,
            dest="NUSMV_KLIVE_BOUND", type="int",
            help="Problem max bound for kliveness used when "
            "performing zenoness checking")

        return

    def check_options(self, opt_parser, options):
        super(NuXMVZenoChecking, self).check_options(opt_parser, options)

        return

    def action(self):
        """Builds nusmv and performs model checking on all
        properties that is possible to model check"""

        options.SMV_FAIR_TIMED = False
        options.SMV_GENERATE_TIME_VAR = True

        super(NuXMVZenoChecking, self).action()

        if not self.timed:
            log.user.info(
                '\nZenoness Checking: Model is not timed, skipping checking')
            return

        for name, impl, zeno_state_slim, zeno_state_smv in \
                self.get_zeno_states(self.root_impl):
            log.user.info(
                '\nZenoness Checking: In "{impl}" checking of:'.format(
                    impl=name if name else "(root)"))

            # invokes xsap
            _ret, _val = nusmv.check_is_state_zeno(self.nusmv, zeno_state_smv)
            result = self.nusmv.zeno_result_to_string(_ret)
            if _ret in (self.nusmv.ZENO, self.nusmv.NON_ZENO):
                result += " (trace in '%s')" % _val
            elif self.nusmv.UNKNOWN_NOT_FOUND == _ret:
                result += " (steps: '%d')" % _val
            elif self.nusmv.UNKNOWN_NOT_PROVED == _ret:
                result += " (time: '%s')" % _val

            msg = '    "{expr}" is {result}'.format(
                expr=zeno_state_smv,
                result=result)
            log.user.info(msg)

    def get_zeno_states(self, root_impl):
        """Produces an iterable of tuples (impl_full_name, impl,
        zeno_state_slim, zeno_state_smv)

        where:
          impl_full_name: full name of implementation, from given root
                          Empty string when referring to the root.
          impl: implementation instance

          zeno_state_slim: slim expression (string) representing zeno state

          zeno_state_smv: smv expression (string) to be used as zeno state
                          constraint when calling check_is_state_zeno
        """
        walker = ModeCollectorWalker()
        walker.walk(root_impl)
        expr_str_fmt = "{inst_name}mode = mode:{mode_lit}"

        # checks and reports
        for name, impl, modes in walker.get_found_modes():

            for lit in modes:
                # construct the formula for mode
                expr_str = expr_str_fmt.format(
                    inst_name="",
                    mode_lit=lit)

                expr_smv = smvapi.parse_and_convert_predicate(
                    impl, expr_str, id_prefix=name)

                yield (name, impl, expr_str, expr_smv)

    pass  # end of class
# ----------------------------------------------------------------------

if __name__ == "__main__":

    main = NuXMVZenoChecking()
    main.run()
