# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================

from slim_compiler.src.dataStructure.located_info import LocatedInformation

class NodeType(object):
    BOOL = 0
    INT = 1
    REAL = 2
    ENUM_FIELD = 3
    ENUM_VALUE = 4
    CLOCK = 5
    CONTINUOUS = 6
    TRAJECTORY_EXPRESSION = 7
    TRAJECTORY_EQUATION = 8
    ERROR = 9
    # Deprecated, check operatortype
    #ASSIGN = 10
    MODE_FIELD = 12
    MODE_VALUE = 13
    ERROR_FIELD = 14
    ERROR_VALUE = 15
    INT_RANGE = 16
    INT_CONST = 19
    REAL_CONST = 20

    def type_to_str(cls, type):
        if type == cls.BOOL:
            return "bool"
        elif type == cls.INT:
            return "int"
        elif type == cls.REAL:
            return "real"
        elif type == cls.ENUM_FIELD:
            return "enum field"
        elif type == cls.ENUM_VALUE:
            return "enum value"
        elif type == cls.CLOCK:
            return "clock"
        elif type == cls.CONTINUOUS:
            return "continuous"
        elif type == cls.TRAJECTORY_EXPRESSION:
            return "trajectory expression"
        elif type == cls.TRAJECTORY_EQUATION:
            return "trajectory equation"
        elif type == cls.ERROR:
            return "error"
        #elif type == cls.ASSIGN:
        #    return "assign"
        elif type == cls.MODE_FIELD:
            return "mode field"
        elif type == cls.MODE_VALUE:
            return "mode value"
        elif type == cls.ERROR_FIELD:
            return "error field"
        elif type == cls.ERROR_VALUE:
            return "error value"
        elif type == cls.INT_RANGE:
            return "int range"
        elif type == cls.INT_CONST:
            return "int const"
        elif type == cls.REAL_CONST:
            return "real const"
    
    type_to_str = classmethod(type_to_str)

class ExpressionTreeNode(LocatedInformation):
    def __init__(self, line, pos):
        LocatedInformation.__init__(self, None, "expression", None, line, pos)
        self.node_type = None

    def __repr__(self):
        from slim_compiler.src.dataStructure.visitors.printVisitor import toSLIMVisitor
        return toSLIMVisitor().visit(self)

    def clone(self):
        new_node = ExpressionTreeNode(self.line, self.pos)
        new_node.file = self.file
        new_node.node_type = self.node_type
        return new_node

class OperatorTreeNode(ExpressionTreeNode):
    def __init__(self, operator, line, pos):
        ExpressionTreeNode.__init__(self, line, pos)
        self.operator = operator

    # Ignoring: line, pos, number
    def almostEqual(self, other):
        if not isinstance(other, OperatorTreeNode): return False
        return self.operator == other.operator

    def clone(self):
        new_node = OperatorTreeNode(self.operator, self.line, self.pos)
        new_node.file = self.file
        new_node.node_type = self.node_type
        return new_node

class BinaryOperatorTreeNode(OperatorTreeNode):
    def __init__(self, left, right, operator, line, pos):
        OperatorTreeNode.__init__(self, operator, line, pos)
        self.left = left
        self.right = right

        left.parent = self
        right.parent = self

    def clone(self):
        new_node = BinaryOperatorTreeNode(self.left.clone(), self.right.clone(), self.operator, self.line, self.pos)
        new_node.file = self.file
        new_node.node_type = self.node_type
        return new_node

    @classmethod
    def createNode(cls, left_expr, right_expr, operator, use_timescale):
        """Returns a tree representation of the operator applied to the two specified
        SLIM expressions.
        
        Note: One or both specified expression parameters may be None.
        If one parameter is None, the other parameter will be returned.
        If both parameters are None, then None will be returned.
        
        @param   left_expr:     The tree representation of the first
                                expression.
        @type    left_expr:     expressions.ExpressionTreeNode
        @param   right_expr:    The tree representation of the second
                                expression.
        @type    right_expr:    expressions.ExpressionTreeNode
        @param   use_timescale: Boolean indicating if time scales are used
        @type    use_timescale: bool
        @return: A tree representation of the operator applied to the specified boolean
                 expressions.
        @rtype:  expressions.BinaryOperatorTreeNode
        """
        if left_expr is None:
            return right_expr
        if right_expr is None:
            return left_expr

        node = BinaryOperatorTreeNode(left_expr, right_expr, operator, right_expr.line, right_expr.pos)
        node.node_type = get_result_table_for_operator(operator, use_timescale)[(left_expr.node_type, right_expr.node_type)]
        node.file = right_expr.file
        return node

    # TODO: This is a copy & paste from model_extension.util.expressions
    # => Keep in only one place.
    @classmethod
    def create_and(cls, boolean_expression_tree1, boolean_expression_tree2):
        return BinaryOperatorTreeNode.createNode(boolean_expression_tree1, boolean_expression_tree2, OperatorType.AND, False)

    @classmethod
    def create_or(cls, boolean_expression_tree1, boolean_expression_tree2):
        return BinaryOperatorTreeNode.createNode(boolean_expression_tree1, boolean_expression_tree2, OperatorType.OR, False)


    @classmethod
    def createDiv(cls, nominator_expr, denominator_str, use_timescale):
        denominator_node = NumberTreeNode(denominator_str, None, nominator_expr.line, nominator_expr.pos)
        denominator_node.file = nominator_expr.file
        return BinaryOperatorTreeNode.createNode(nominator_expr, denominator_node, OperatorType.DIVIDE, use_timescale)

    @classmethod
    def createMul(cls, left_expr, right_str, use_timescale):
        right_node = NumberTreeNode(right_str, None, left_expr.line, left_expr.pos)
        right_node.file = left_expr.file
        return BinaryOperatorTreeNode.createNode(left_expr, right_node, OperatorType.TIMES, use_timescale)

    def get_value(self):
        return self.number
    
    # Ignoring: line, pos, number
    def almostEqual(self, other):
        if not isinstance(other, BinaryOperatorTreeNode): return False
        return  self.operator == other.operator \
            and self.node_type == other.node_type \
            and self.left.almostEqual(other.left) \
            and self.right.almostEqual(other.right) \

class UnaryOperatorTreeNode(OperatorTreeNode):
    def __init__(self, subexpression, operator, line, pos):
        OperatorTreeNode.__init__(self, operator, line, pos)
        self.subexpression = subexpression
        self.node_type = None

        subexpression.parent = self

    def clone(self):
        new_node = UnaryOperatorTreeNode(self.subexpression.clone(), self.operator, self.line, self.pos)
        new_node.file = self.file
        new_node.node_type = self.node_type
        return new_node

    @classmethod
    def createNode(cls, sub_expr, operator, use_timescale):
        if sub_expr is None:
            return None

        node = UnaryOperatorTreeNode(sub_expr, operator, sub_expr.line, sub_expr.pos)
        node.node_type = get_result_table_for_operator(operator, use_timescale)[sub_expr.node_type]
        node.file = sub_expr.file
        return node

    # Ignoring: line, pos, number
    def almostEqual(self, other):
        if not isinstance(other, UnaryOperatorTreeNode): return False
        return  self.operator == other.operator \
            and self.node_type == other.node_type \
            and self.subexpression.almostEqual(other.subexpression) \
    
class IdentifierTreeNode(ExpressionTreeNode):
    def __init__(self, name, line, pos):
        ExpressionTreeNode.__init__(self, line, pos)
        self.name = name
        
        # Meta data for checking enums/ranges
        self.enum_type = []
        self.enum_values = []
        self.bounds = None
        # Reference to object that declared this identifier
        self.ref = None

    def clone(self):
        new_node = IdentifierTreeNode(self.name, self.line, self.pos)
        new_node.file = self.file
        new_node.node_type = self.node_type
        # Note: metadata is not cloned, but reference copy
        new_node.enum_type = self.enum_type
        new_node.enum_values = self.enum_values
        new_node.bounds = self.bounds
        new_node.ref = self.ref
        return new_node

    def get_value(self):
        return self.name
    
    # Ignoring: line, pos, number, bounds
    def almostEqual(self, other):
        if not isinstance(other, IdentifierTreeNode): return False
        return  self.name == other.name \
            and self.node_type == other.node_type \
            and cmp(self.enum_values, other.enum_values) == 0 \
            and cmp(self.enum_type, other.enum_type) == 0 \
            and cmp(self.bounds, other.bounds) == 0

class BooleanTreeNode(ExpressionTreeNode):
    def __init__(self, value, line, pos):
        ExpressionTreeNode.__init__(self, line, pos)
        self.value = value
        self.node_type = NodeType.BOOL

    def get_value(self):
        return self.value

    def clone(self):
        new_node = BooleanTreeNode(self.value, self.line, self.pos)
        new_node.file = self.file
        new_node.node_type = self.node_type
        return new_node

    # Ignoring: line, pos
    def almostEqual(self, other):
        if not isinstance(other, BooleanTreeNode): return False
        return  self.value == other.value \
            and self.node_type == other.node_type

class NumberTreeNode(ExpressionTreeNode):
    @classmethod
    def createZero(cls, file=None, line=None, pos=None):
        num_node = NumberTreeNode("0", None, line, pos)
        num_node.file = file
        return num_node
    
    def __init__(self, number, time_unit, line, pos):
        ExpressionTreeNode.__init__(self, line, pos)
        self.number = number
        self.time_unit = time_unit
        self.node_type = self.get_type()

    def get_type(self):
        if self.number.find(".") != -1:
            return NodeType.REAL_CONST
        else:
            return NodeType.INT_CONST

    def get_value(self):
        ret = 0
        try:
            ret = int(self.number)
        except ValueError:
            ret = float(self.number)
        return ret

    def clone(self):
        new_node = NumberTreeNode(self.number, self.time_unit, self.line, self.pos)
        new_node.file = self.file
        new_node.node_type = self.node_type
        return new_node

    # Ignoring: line, pos, number
    def almostEqual(self, other):
        if not isinstance(other, NumberTreeNode): return False
        return  self.number == other.number \
            and self.time_unit == other.time_unit \
            and self.node_type == other.node_type

class TrajectoryEquationTreeNode(ExpressionTreeNode):
    def __init__(self, variable, trajectory_expression, line, pos, time_unit=None):
        ExpressionTreeNode.__init__(self, line, pos)
        self.variable = variable
        self.trajectory_expression = trajectory_expression
        self.node_type = NodeType.TRAJECTORY_EQUATION
        self.time_unit=time_unit

    def clone(self):
        new_node = TrajectoryEquationTreeNode(self.variable.clone(), self.trajectory_expression.clone(), self.line, self.pos, self.time_unit)
        new_node.file = self.file
        new_node.node_type = self.node_type
        return new_node

    # Ignoring: line, pos
    def almostEqual(self, other):
        if not isinstance(other, TrajectoryEquationTreeNode): return False
        return  self.variable.almostEqual(other.variable) \
            and self.trajectory_expression.almostEqual(other.trajectory_expression) \
            and self.node_type == other.node_type

# TODO: Remove this one. Seems to be a deprecated remnant (never used)
class TrajectoryExpressionTreeNode(ExpressionTreeNode):
    def __init__(self, variable, times, offset, line, pos):
        ExpressionTreeNode.__init__(self, line, pos)
        self.variable = variable
        self.times = times
        self.offset = offset
        self.node_type = NodeType.TRAJECTORY_EXPRESSION

    # Ignoring: line, pos
    def almostEqual(self, other):
        if not isinstance(other, TrajectoryExpressionTreeNode): return False
        return  self.variable == other.variable \
            and self.times == other.times \
            and self.offset == other.offset \
            and self.node_type == other.node_type

class CaseExpressionTreeNode(OperatorTreeNode):
    def __init__(self, cases, default_expression, line, pos):
        OperatorTreeNode.__init__(self, "case", line, pos)
        self.cases = cases
        self.default_expression = default_expression
        self.node_type = None
        self.name = ""
        for (cond, expr) in self.cases:
            cond.parent = self
            expr.parent = self
        default_expression.parent = self

    def clone(self):
        new_cases = []
        for (cond, expr) in self.cases:
            new_cases.append((cond.clone(), expr.clone()))
        new_node = CaseExpressionTreeNode(new_cases, self.default_expression.clone(), self.line, self.pos)
        new_node.file = self.file
        new_node.node_type = self.node_type
        return new_node

    # Ignoring: line, pos
    def almostEqual(self, other):
        if not isinstance(other, CaseExpressionTreeNode): return False
        if len(self.cases) != len(other.cases):
            return False
        for ( (cond_self, expr_self), (cond_other, expr_other)) in map(None, self.cases, other.cases):
            if not cond_self.almostEqual(cond_other) or not expr_self.almostEqual(expr_other):
                return False
        return  self.default_expression.almostEqual(other.default_expression) \
            and self.node_type == other.node_type

class OperatorType(object):
    LESS = 0
    LESSOREQUAL = 1
    GREATER = 2
    GREATEROREQUAL = 3
    EQUAL = 4
    NOTEQUAL = 5
    PLUS = 6
    MINUS = 7
    TIMES = 8
    DIVIDE = 9
    AND = 10
    ASSIGN = 11
    OR = 12
    XOR = 13
    XNOR = 14
    NOT = 15
    UNARY_MINUS = 16
    IFF = 17
    IMP = 18
    MOD = 19
    MSEC = 20
    SEC = 21
    MIN = 22
    HOUR = 23
    DAY = 24
    # See ticket 543 to x_TOREAL requirement
    REAL_TOMSEC = 20
    REAL_TOSEC = 21
    REAL_TOMIN = 22
    REAL_TOHOUR = 23
    REAL_TODAY = 24

    MSEC_TOREAL = 30
    SEC_TOREAL = 31
    MIN_TOREAL = 32
    HOUR_TOREAL = 33
    DAY_TOREAL = 34

    def operator_to_str(cls, operator):
        if operator == cls.LESS:
            return "<"
        elif operator == cls.LESSOREQUAL:
            return "<="
        elif operator == cls.GREATER:
            return ">"
        elif operator == cls.GREATEROREQUAL:
            return ">="
        elif operator == cls.EQUAL:
            return "="
        elif operator == cls.NOTEQUAL:
            return "!="
        elif operator == cls.PLUS:
            return "+"
        elif operator == cls.MINUS:
            return "-"
        elif operator == cls.TIMES:
            return "*"
        elif operator == cls.DIVIDE:
            return "/"
        elif operator == cls.AND:
            return "and"
        elif operator == cls.ASSIGN:
            return ":="
        elif operator == cls.OR:
            return "or"
        elif operator == cls.XOR:
            return "xor"
        elif operator == cls.XNOR:
            return "xnor"
        elif operator == cls.NOT:
            return "not"
        elif operator == cls.UNARY_MINUS:
            return "-"
        elif operator == cls.IFF:
            return "iff"
        elif operator == cls.IMP:
            return "imp"
        elif operator == cls.MOD:
            return "mod"
        elif operator == cls.MSEC or operator == cls.MSEC_TOREAL or operator == cls.MSEC_TOREAL:
            return "msec"
        elif operator == cls.SEC or operator == cls.SEC_TOREAL or operator == cls.SEC_TOREAL:
            return "sec"
        elif operator == cls.MIN or operator == cls.MIN_TOREAL or operator == cls.MIN_TOREAL:
            return "min"
        elif operator == cls.HOUR or operator == cls.HOUR_TOREAL or operator == cls.HOUR_TOREAL:
            return "hour"
        elif operator == cls.DAY or operator == cls.DAY_TOREAL or operator == cls.DAY_TOREAL:
            return "day"

        assert False

    operator_to_str = classmethod(operator_to_str)


class TimeUnitType(object):
    MSEC = 0
    SEC = 1
    MIN = 2
    HOUR = 3
    DAY = 4
    NONE = 5

    def str_to_time_unit_type(cls, text):
        if (text == "msec"):
            return cls.MSEC
        elif (text == "sec"):
            return cls.SEC
        elif (text == "min"):
            return cls.MIN
        elif (text == "hour"):
            return cls.HOUR
        elif (text == "day"):
            return cls.DAY
        else:
            assert False, "Unknown time unit: %s" % text

    def time_unit_type_to_str(cls, time_unit):
        if (time_unit == cls.MSEC):
            return "msec"
        elif (time_unit == cls.SEC):
            return "sec"
        elif (time_unit == cls.MIN):
            return "min"
        elif (time_unit == cls.HOUR):
            return "hour"
        elif (time_unit == cls.DAY):
            return "day"
        elif (time_unit == cls.NONE):
            return ""
        else:
            assert False, "Unknown time unit: %d" % time_unit

    str_to_time_unit_type = classmethod(str_to_time_unit_type)
    time_unit_type_to_str = classmethod(time_unit_type_to_str)

    def to_operator(cls, time_unit):
        if (time_unit == cls.MSEC):
            return OperatorType.MSEC
        elif (time_unit == cls.SEC):
            return OperatorType.SEC
        elif (time_unit == cls.MIN):
            return OperatorType.MIN
        elif (time_unit == cls.HOUR):
            return OperatorType.HOUR
        elif (time_unit == cls.DAY):
            return OperatorType.DAY
        elif (time_unit == cls.NONE):
            return OperatorType.NONE
    to_operator = classmethod(to_operator)

    @classmethod
    def from_operator(cls, op):
        if op in (OperatorType.MSEC,
                  OperatorType.REAL_TOMSEC,
                  OperatorType.MSEC_TOREAL):
            return cls.MSEC
        elif op in (OperatorType.SEC,
                    OperatorType.REAL_TOSEC,
                    OperatorType.SEC_TOREAL):
            return cls.SEC
        elif op in (OperatorType.MIN,
                    OperatorType.REAL_TOMIN,
                    OperatorType.MIN_TOREAL):
            return cls.MIN
        elif op in (OperatorType.HOUR,
                    OperatorType.REAL_TOHOUR,
                    OperatorType.HOUR_TOREAL):
            return cls.HOUR
        elif op in (OperatorType.DAY,
                    OperatorType.REAL_TODAY,
                    OperatorType.DAY_TOREAL):
            return cls.DAY
        elif op == OperatorType.NONE:
            return cls.NONE
        else:
            assert False, "Invalid TU operator: %d" % op


numeric_types = [NodeType.INT_CONST, NodeType.REAL_CONST, NodeType.INT_RANGE,
                 NodeType.INT, NodeType.REAL, NodeType.CLOCK,
                 NodeType.CONTINUOUS]

numeric_types_set = set(numeric_types)

constant_types = [NodeType.INT_CONST, NodeType.REAL_CONST]

constant_types_set = set(constant_types)

enum_types = [NodeType.ENUM_FIELD, NodeType.ENUM_VALUE,
              NodeType.MODE_FIELD, NodeType.MODE_VALUE,
              NodeType.ERROR_FIELD, NodeType.ERROR_VALUE]

enum_types_set = set(enum_types)

binary_plus_result = \
        {
            (NodeType.INT_CONST, NodeType.INT_CONST) : NodeType.INT_CONST,
            (NodeType.INT_CONST, NodeType.REAL_CONST) : NodeType.REAL_CONST,
            (NodeType.REAL_CONST, NodeType.INT_CONST) : NodeType.REAL_CONST,
            (NodeType.INT_CONST, NodeType.INT_RANGE) : NodeType.INT_RANGE,
            (NodeType.INT_RANGE, NodeType.INT_CONST) : NodeType.INT_RANGE,
            (NodeType.INT_CONST, NodeType.INT) : NodeType.INT,
            (NodeType.INT, NodeType.INT_CONST) : NodeType.INT,
            (NodeType.INT_CONST, NodeType.REAL) : NodeType.REAL,
            (NodeType.REAL, NodeType.INT_CONST) : NodeType.REAL,
            (NodeType.INT_CONST, NodeType.CLOCK) : NodeType.CLOCK,
            (NodeType.CLOCK, NodeType.INT_CONST) : NodeType.CLOCK,
            (NodeType.INT_CONST, NodeType.CONTINUOUS) : NodeType.CONTINUOUS,
            (NodeType.CONTINUOUS, NodeType.INT_CONST) : NodeType.CONTINUOUS,

            (NodeType.REAL_CONST, NodeType.REAL_CONST) : NodeType.REAL_CONST,
            (NodeType.REAL_CONST, NodeType.INT) : NodeType.REAL,
            (NodeType.INT, NodeType.REAL_CONST) : NodeType.REAL,
            (NodeType.REAL_CONST, NodeType.REAL) : NodeType.REAL,
            (NodeType.REAL, NodeType.REAL_CONST) : NodeType.REAL,
            (NodeType.REAL_CONST, NodeType.CLOCK) : NodeType.CLOCK,
            (NodeType.CLOCK, NodeType.REAL_CONST) : NodeType.CLOCK,
            (NodeType.REAL_CONST, NodeType.CONTINUOUS) : NodeType.CONTINUOUS,
            (NodeType.CONTINUOUS, NodeType.REAL_CONST) : NodeType.CONTINUOUS,

            (NodeType.INT_RANGE, NodeType.INT_RANGE) : NodeType.INT_RANGE,

            (NodeType.INT, NodeType.INT) : NodeType.INT,
            (NodeType.INT, NodeType.REAL) : NodeType.REAL,
            (NodeType.REAL, NodeType.INT) : NodeType.REAL,
            (NodeType.INT, NodeType.CLOCK) : NodeType.CLOCK,
            (NodeType.CLOCK, NodeType.INT) : NodeType.CLOCK,
            (NodeType.INT, NodeType.CONTINUOUS) : NodeType.CONTINUOUS,
            (NodeType.CONTINUOUS, NodeType.INT) : NodeType.CONTINUOUS,

            (NodeType.REAL, NodeType.REAL) : NodeType.REAL,
            (NodeType.REAL, NodeType.CLOCK) : NodeType.CLOCK,
            (NodeType.CLOCK, NodeType.REAL) : NodeType.CLOCK,
            (NodeType.REAL, NodeType.CONTINUOUS) : NodeType.CONTINUOUS,
            (NodeType.CONTINUOUS, NodeType.REAL) : NodeType.CONTINUOUS,

            (NodeType.CLOCK, NodeType.CLOCK) : NodeType.CLOCK,
            (NodeType.CLOCK, NodeType.CONTINUOUS) : NodeType.CONTINUOUS,
            (NodeType.CONTINUOUS, NodeType.CLOCK) : NodeType.CONTINUOUS,

            (NodeType.CONTINUOUS, NodeType.CONTINUOUS) : NodeType.CONTINUOUS
        }

binary_minus_result = \
        {
            (NodeType.INT_CONST, NodeType.INT_CONST) : NodeType.INT_CONST,
            (NodeType.INT_CONST, NodeType.REAL_CONST) : NodeType.REAL_CONST,
            (NodeType.REAL_CONST, NodeType.INT_CONST) : NodeType.REAL_CONST,
            (NodeType.INT_CONST, NodeType.INT_RANGE) : NodeType.INT_RANGE,
            (NodeType.INT_RANGE, NodeType.INT_CONST) : NodeType.INT_RANGE,
            (NodeType.INT_CONST, NodeType.INT) : NodeType.INT,
            (NodeType.INT, NodeType.INT_CONST) : NodeType.INT,
            (NodeType.INT_CONST, NodeType.REAL) : NodeType.REAL,
            (NodeType.REAL, NodeType.INT_CONST) : NodeType.REAL,
            (NodeType.INT_CONST, NodeType.CLOCK) : NodeType.CLOCK,
            (NodeType.CLOCK, NodeType.INT_CONST) : NodeType.CLOCK,
            (NodeType.INT_CONST, NodeType.CONTINUOUS) : NodeType.CONTINUOUS,
            (NodeType.CONTINUOUS, NodeType.INT_CONST) : NodeType.CONTINUOUS,

            (NodeType.REAL_CONST, NodeType.REAL_CONST) : NodeType.REAL_CONST,
            (NodeType.REAL_CONST, NodeType.INT) : NodeType.REAL,
            (NodeType.INT, NodeType.REAL_CONST) : NodeType.REAL,
            (NodeType.REAL_CONST, NodeType.REAL) : NodeType.REAL,
            (NodeType.REAL, NodeType.REAL_CONST) : NodeType.REAL,
            (NodeType.REAL_CONST, NodeType.CLOCK) : NodeType.CLOCK,
            (NodeType.CLOCK, NodeType.REAL_CONST) : NodeType.CLOCK,
            (NodeType.REAL_CONST, NodeType.CONTINUOUS) : NodeType.CONTINUOUS,
            (NodeType.CONTINUOUS, NodeType.REAL_CONST) : NodeType.CONTINUOUS,

            (NodeType.INT_RANGE, NodeType.INT_RANGE) : NodeType.INT_RANGE,

            (NodeType.INT, NodeType.INT) : NodeType.INT,
            (NodeType.INT, NodeType.REAL) : NodeType.REAL,
            (NodeType.REAL, NodeType.INT) : NodeType.REAL,
            (NodeType.INT, NodeType.CLOCK) : NodeType.CLOCK,
            (NodeType.CLOCK, NodeType.INT) : NodeType.CLOCK,
            (NodeType.INT, NodeType.CONTINUOUS) : NodeType.CONTINUOUS,
            (NodeType.CONTINUOUS, NodeType.INT) : NodeType.CONTINUOUS,

            (NodeType.REAL, NodeType.REAL) : NodeType.REAL,
            (NodeType.REAL, NodeType.CLOCK) : NodeType.CLOCK,
            (NodeType.CLOCK, NodeType.REAL) : NodeType.CLOCK,
            (NodeType.REAL, NodeType.CONTINUOUS) : NodeType.CONTINUOUS,
            (NodeType.CONTINUOUS, NodeType.REAL) : NodeType.CONTINUOUS,

            (NodeType.CLOCK, NodeType.CLOCK) : NodeType.CLOCK,
            (NodeType.CLOCK, NodeType.CONTINUOUS) : NodeType.CONTINUOUS,
            (NodeType.CONTINUOUS, NodeType.CLOCK) : NodeType.CONTINUOUS,

            (NodeType.CONTINUOUS, NodeType.CONTINUOUS) : NodeType.CONTINUOUS
        }

unary_minus_result = \
        {
            NodeType.INT_CONST : NodeType.INT_CONST,
            NodeType.REAL_CONST : NodeType.REAL_CONST,
            NodeType.INT_RANGE : NodeType.INT_RANGE,
            NodeType.INT : NodeType.INT,
            NodeType.REAL : NodeType.REAL,
            NodeType.CLOCK: NodeType.CLOCK,
            NodeType.CONTINUOUS: NodeType.CONTINUOUS
        }

binary_times_result = \
        {
            (NodeType.INT_CONST, NodeType.INT_CONST) : NodeType.INT_CONST,
            (NodeType.INT_CONST, NodeType.REAL_CONST) : NodeType.REAL_CONST,
            (NodeType.REAL_CONST, NodeType.INT_CONST) : NodeType.REAL_CONST,
            (NodeType.INT_CONST, NodeType.INT_RANGE) : NodeType.INT_RANGE,
            (NodeType.INT_RANGE, NodeType.INT_CONST) : NodeType.INT_RANGE,
            (NodeType.INT_CONST, NodeType.INT) : NodeType.INT,
            (NodeType.INT, NodeType.INT_CONST) : NodeType.INT,
            (NodeType.INT_CONST, NodeType.REAL) : NodeType.REAL,
            (NodeType.REAL, NodeType.INT_CONST) : NodeType.REAL,
            (NodeType.INT_CONST, NodeType.CLOCK) : NodeType.CLOCK,
            (NodeType.CLOCK, NodeType.INT_CONST) : NodeType.CLOCK,
            (NodeType.INT_CONST, NodeType.CONTINUOUS) : NodeType.CONTINUOUS,
            (NodeType.CONTINUOUS, NodeType.INT_CONST) : NodeType.CONTINUOUS,

            (NodeType.REAL_CONST, NodeType.REAL_CONST) : NodeType.REAL_CONST,
            (NodeType.REAL_CONST, NodeType.INT) : NodeType.REAL,
            (NodeType.INT, NodeType.REAL_CONST) : NodeType.REAL,
            (NodeType.REAL_CONST, NodeType.REAL) : NodeType.REAL,
            (NodeType.REAL, NodeType.REAL_CONST) : NodeType.REAL,
            (NodeType.REAL_CONST, NodeType.CLOCK) : NodeType.CLOCK,
            (NodeType.CLOCK, NodeType.REAL_CONST) : NodeType.CLOCK,
            (NodeType.REAL_CONST, NodeType.CONTINUOUS) : NodeType.CONTINUOUS,
            (NodeType.CONTINUOUS, NodeType.REAL_CONST) : NodeType.CONTINUOUS,

            (NodeType.INT_RANGE, NodeType.INT_RANGE) : NodeType.INT_RANGE,

            (NodeType.INT, NodeType.INT) : NodeType.INT,
            (NodeType.INT, NodeType.REAL) : NodeType.REAL,
            (NodeType.REAL, NodeType.INT) : NodeType.REAL,
            (NodeType.INT, NodeType.CLOCK) : NodeType.CLOCK,
            (NodeType.CLOCK, NodeType.INT) : NodeType.CLOCK,
            (NodeType.INT, NodeType.CONTINUOUS) : NodeType.CONTINUOUS,
            (NodeType.CONTINUOUS, NodeType.INT) : NodeType.CONTINUOUS,

            (NodeType.REAL, NodeType.REAL) : NodeType.REAL,
            (NodeType.REAL, NodeType.CLOCK) : NodeType.CLOCK,
            (NodeType.CLOCK, NodeType.REAL) : NodeType.CLOCK,
            (NodeType.REAL, NodeType.CONTINUOUS) : NodeType.CONTINUOUS,
            (NodeType.CONTINUOUS, NodeType.REAL) : NodeType.CONTINUOUS,

            (NodeType.CLOCK, NodeType.CLOCK) : NodeType.CLOCK,
            (NodeType.CLOCK, NodeType.CONTINUOUS) : NodeType.CONTINUOUS,
            (NodeType.CONTINUOUS, NodeType.CLOCK) : NodeType.CONTINUOUS,

            (NodeType.CONTINUOUS, NodeType.CONTINUOUS) : NodeType.CONTINUOUS
        }

binary_division_result = \
        {
            (NodeType.INT_CONST, NodeType.INT_CONST) : NodeType.INT_CONST,
            (NodeType.INT_CONST, NodeType.REAL_CONST) : NodeType.REAL_CONST,
            (NodeType.REAL_CONST, NodeType.INT_CONST) : NodeType.REAL_CONST,
            (NodeType.INT_CONST, NodeType.INT_RANGE) : NodeType.INT_RANGE,
            (NodeType.INT_RANGE, NodeType.INT_CONST) : NodeType.INT_RANGE,
            (NodeType.INT_CONST, NodeType.INT) : NodeType.INT,
            (NodeType.INT, NodeType.INT_CONST) : NodeType.INT,
            (NodeType.INT_CONST, NodeType.REAL) : NodeType.REAL,
            (NodeType.REAL, NodeType.INT_CONST) : NodeType.REAL,
            (NodeType.INT_CONST, NodeType.CLOCK) : NodeType.CLOCK,
            (NodeType.CLOCK, NodeType.INT_CONST) : NodeType.CLOCK,
            (NodeType.INT_CONST, NodeType.CONTINUOUS) : NodeType.CONTINUOUS,
            (NodeType.CONTINUOUS, NodeType.INT_CONST) : NodeType.CONTINUOUS,

            (NodeType.REAL_CONST, NodeType.REAL_CONST) : NodeType.REAL_CONST,
            (NodeType.REAL_CONST, NodeType.INT) : NodeType.REAL,
            (NodeType.INT, NodeType.REAL_CONST) : NodeType.REAL,
            (NodeType.REAL_CONST, NodeType.REAL) : NodeType.REAL,
            (NodeType.REAL, NodeType.REAL_CONST) : NodeType.REAL,
            (NodeType.REAL_CONST, NodeType.CLOCK) : NodeType.CLOCK,
            (NodeType.CLOCK, NodeType.REAL_CONST) : NodeType.CLOCK,
            (NodeType.REAL_CONST, NodeType.CONTINUOUS) : NodeType.CONTINUOUS,
            (NodeType.CONTINUOUS, NodeType.REAL_CONST) : NodeType.CONTINUOUS,

            (NodeType.INT_RANGE, NodeType.INT_RANGE) : NodeType.INT_RANGE,

            (NodeType.INT, NodeType.INT) : NodeType.INT,
            (NodeType.INT, NodeType.REAL) : NodeType.REAL,
            (NodeType.REAL, NodeType.INT) : NodeType.REAL,
            (NodeType.INT, NodeType.CLOCK) : NodeType.CLOCK,
            (NodeType.CLOCK, NodeType.INT) : NodeType.CLOCK,
            (NodeType.INT, NodeType.CONTINUOUS) : NodeType.CONTINUOUS,
            (NodeType.CONTINUOUS, NodeType.INT) : NodeType.CONTINUOUS,

            (NodeType.REAL, NodeType.REAL) : NodeType.REAL,
            (NodeType.REAL, NodeType.CLOCK) : NodeType.CLOCK,
            (NodeType.CLOCK, NodeType.REAL) : NodeType.CLOCK,
            (NodeType.REAL, NodeType.CONTINUOUS) : NodeType.CONTINUOUS,
            (NodeType.CONTINUOUS, NodeType.REAL) : NodeType.CONTINUOUS,

            (NodeType.CLOCK, NodeType.CLOCK) : NodeType.CLOCK,
            (NodeType.CLOCK, NodeType.CONTINUOUS) : NodeType.CONTINUOUS,
            (NodeType.CONTINUOUS, NodeType.CLOCK) : NodeType.CONTINUOUS,

            (NodeType.CONTINUOUS, NodeType.CONTINUOUS) : NodeType.CONTINUOUS
        }

binary_mod_result = \
        {
            (NodeType.INT_CONST, NodeType.INT_CONST) : NodeType.INT_CONST,
            (NodeType.INT, NodeType.INT_CONST) : NodeType.INT,
            (NodeType.INT_RANGE, NodeType.INT_CONST) : NodeType.INT_RANGE,
            (NodeType.INT_CONST, NodeType.INT) : NodeType.INT,
            (NodeType.INT, NodeType.INT) : NodeType.INT,
            (NodeType.INT_RANGE, NodeType.INT) : NodeType.INT_RANGE
        }

binary_equality_result = \
        {
            (NodeType.BOOL, NodeType.BOOL) : NodeType.BOOL,

            (NodeType.INT_CONST, NodeType.INT_CONST) : NodeType.BOOL,

            (NodeType.INT_CONST, NodeType.REAL_CONST) : NodeType.BOOL,
            (NodeType.REAL_CONST, NodeType.INT_CONST) : NodeType.BOOL,
            (NodeType.INT_CONST, NodeType.INT_RANGE) : NodeType.BOOL,
            (NodeType.INT_RANGE, NodeType.INT_CONST) : NodeType.BOOL,
            (NodeType.INT_CONST, NodeType.INT) : NodeType.BOOL,
            (NodeType.INT, NodeType.INT_CONST) : NodeType.BOOL,
            (NodeType.INT_CONST, NodeType.REAL) : NodeType.BOOL,
            (NodeType.REAL, NodeType.INT_CONST) : NodeType.BOOL,
            (NodeType.INT_CONST, NodeType.CLOCK) : NodeType.BOOL,
            (NodeType.CLOCK, NodeType.INT_CONST) : NodeType.BOOL,
            (NodeType.INT_CONST, NodeType.CONTINUOUS) : NodeType.BOOL,
            (NodeType.CONTINUOUS, NodeType.INT_CONST) : NodeType.BOOL,

            (NodeType.REAL_CONST, NodeType.REAL_CONST) : NodeType.BOOL,
            (NodeType.REAL_CONST, NodeType.REAL) : NodeType.BOOL,
            (NodeType.REAL, NodeType.REAL_CONST) : NodeType.BOOL,
            (NodeType.REAL_CONST, NodeType.CLOCK) : NodeType.BOOL,
            (NodeType.CLOCK, NodeType.REAL_CONST) : NodeType.BOOL,
            (NodeType.REAL_CONST, NodeType.CONTINUOUS) : NodeType.BOOL,
            (NodeType.CONTINUOUS, NodeType.REAL_CONST) : NodeType.BOOL,

            (NodeType.INT_RANGE, NodeType.INT_RANGE) : NodeType.BOOL,
            (NodeType.INT_RANGE, NodeType.INT) : NodeType.BOOL,
            (NodeType.INT, NodeType.INT_RANGE) : NodeType.BOOL,
            (NodeType.INT_RANGE, NodeType.REAL) : NodeType.BOOL,
            (NodeType.REAL, NodeType.INT_RANGE) : NodeType.BOOL,
            (NodeType.INT_RANGE, NodeType.CLOCK) : NodeType.BOOL,
            (NodeType.CLOCK, NodeType.INT_RANGE) : NodeType.BOOL,
            (NodeType.INT_RANGE, NodeType.CONTINUOUS) : NodeType.BOOL,
            (NodeType.CONTINUOUS, NodeType.INT_RANGE) : NodeType.BOOL,

            (NodeType.INT, NodeType.INT) : NodeType.BOOL,
            (NodeType.INT, NodeType.REAL) : NodeType.BOOL,
            (NodeType.REAL, NodeType.INT) : NodeType.BOOL,
            (NodeType.INT, NodeType.CLOCK) : NodeType.BOOL,
            (NodeType.CLOCK, NodeType.INT) : NodeType.BOOL,
            (NodeType.INT, NodeType.CONTINUOUS) : NodeType.BOOL,
            (NodeType.CONTINUOUS, NodeType.INT) : NodeType.BOOL,

            (NodeType.REAL, NodeType.REAL) : NodeType.BOOL,
            (NodeType.REAL, NodeType.CLOCK) : NodeType.BOOL,
            (NodeType.CLOCK, NodeType.REAL) : NodeType.BOOL,
            (NodeType.REAL, NodeType.CONTINUOUS) : NodeType.BOOL,
            (NodeType.CONTINUOUS, NodeType.REAL) : NodeType.BOOL,

            (NodeType.CLOCK, NodeType.CLOCK) : NodeType.BOOL,
            (NodeType.CLOCK, NodeType.CONTINUOUS) : NodeType.BOOL,
            (NodeType.CONTINUOUS, NodeType.CLOCK) : NodeType.BOOL,
            (NodeType.CONTINUOUS, NodeType.CONTINUOUS) : NodeType.BOOL,

            (NodeType.ENUM_FIELD, NodeType.ENUM_FIELD) : NodeType.BOOL,
            (NodeType.ENUM_FIELD, NodeType.ENUM_VALUE) : NodeType.BOOL,
            (NodeType.ENUM_VALUE, NodeType.ENUM_FIELD) : NodeType.BOOL,
            (NodeType.MODE_FIELD, NodeType.MODE_VALUE) : NodeType.BOOL,
            (NodeType.MODE_VALUE, NodeType.MODE_FIELD) : NodeType.BOOL,
            (NodeType.MODE_FIELD, NodeType.MODE_FIELD) : NodeType.BOOL,
            (NodeType.ERROR_FIELD, NodeType.ERROR_VALUE) : NodeType.BOOL,
            (NodeType.ERROR_VALUE, NodeType.ERROR_FIELD) : NodeType.BOOL,
            (NodeType.ERROR_FIELD, NodeType.ERROR_FIELD) : NodeType.BOOL,
        }

binary_inequality_result = \
        {
            (NodeType.BOOL, NodeType.BOOL) : NodeType.BOOL,

            (NodeType.INT_CONST, NodeType.INT_CONST) : NodeType.BOOL,

            (NodeType.INT_CONST, NodeType.REAL_CONST) : NodeType.BOOL,
            (NodeType.REAL_CONST, NodeType.INT_CONST) : NodeType.BOOL,
            (NodeType.INT_CONST, NodeType.INT_RANGE) : NodeType.BOOL,
            (NodeType.INT_RANGE, NodeType.INT_CONST) : NodeType.BOOL,
            (NodeType.INT_CONST, NodeType.INT) : NodeType.BOOL,
            (NodeType.INT, NodeType.INT_CONST) : NodeType.BOOL,
            (NodeType.INT_CONST, NodeType.REAL) : NodeType.BOOL,
            (NodeType.REAL, NodeType.INT_CONST) : NodeType.BOOL,
            (NodeType.INT_CONST, NodeType.CLOCK) : NodeType.BOOL,
            (NodeType.CLOCK, NodeType.INT_CONST) : NodeType.BOOL,
            (NodeType.INT_CONST, NodeType.CONTINUOUS) : NodeType.BOOL,
            (NodeType.CONTINUOUS, NodeType.INT_CONST) : NodeType.BOOL,

            (NodeType.REAL_CONST, NodeType.REAL_CONST) : NodeType.BOOL,
            (NodeType.REAL_CONST, NodeType.REAL) : NodeType.BOOL,
            (NodeType.REAL, NodeType.REAL_CONST) : NodeType.BOOL,
            (NodeType.REAL_CONST, NodeType.CLOCK) : NodeType.BOOL,
            (NodeType.CLOCK, NodeType.REAL_CONST) : NodeType.BOOL,
            (NodeType.REAL_CONST, NodeType.CONTINUOUS) : NodeType.BOOL,
            (NodeType.CONTINUOUS, NodeType.REAL_CONST) : NodeType.BOOL,

            (NodeType.INT_RANGE, NodeType.INT_RANGE) : NodeType.BOOL,
            (NodeType.INT_RANGE, NodeType.INT) : NodeType.BOOL,
            (NodeType.INT, NodeType.INT_RANGE) : NodeType.BOOL,
            (NodeType.INT_RANGE, NodeType.REAL) : NodeType.BOOL,
            (NodeType.REAL, NodeType.INT_RANGE) : NodeType.BOOL,
            (NodeType.INT_RANGE, NodeType.CLOCK) : NodeType.BOOL,
            (NodeType.CLOCK, NodeType.INT_RANGE) : NodeType.BOOL,
            (NodeType.INT_RANGE, NodeType.CONTINUOUS) : NodeType.BOOL,
            (NodeType.CONTINUOUS, NodeType.INT_RANGE) : NodeType.BOOL,

            (NodeType.INT, NodeType.INT) : NodeType.BOOL,
            (NodeType.INT, NodeType.REAL) : NodeType.BOOL,
            (NodeType.REAL, NodeType.INT) : NodeType.BOOL,
            (NodeType.INT, NodeType.CLOCK) : NodeType.BOOL,
            (NodeType.CLOCK, NodeType.INT) : NodeType.BOOL,
            (NodeType.INT, NodeType.CONTINUOUS) : NodeType.BOOL,
            (NodeType.CONTINUOUS, NodeType.INT) : NodeType.BOOL,

            (NodeType.REAL, NodeType.REAL) : NodeType.BOOL,
            (NodeType.REAL, NodeType.CLOCK) : NodeType.BOOL,
            (NodeType.CLOCK, NodeType.REAL) : NodeType.BOOL,
            (NodeType.REAL, NodeType.CONTINUOUS) : NodeType.BOOL,
            (NodeType.CONTINUOUS, NodeType.REAL) : NodeType.BOOL,

            (NodeType.CLOCK, NodeType.CLOCK) : NodeType.BOOL,
            (NodeType.CLOCK, NodeType.CONTINUOUS) : NodeType.BOOL,
            (NodeType.CONTINUOUS, NodeType.CLOCK) : NodeType.BOOL,
            (NodeType.CONTINUOUS, NodeType.CONTINUOUS) : NodeType.BOOL,

            (NodeType.ENUM_FIELD, NodeType.ENUM_FIELD) : NodeType.BOOL,
            (NodeType.ENUM_FIELD, NodeType.ENUM_VALUE) : NodeType.BOOL,
            (NodeType.ENUM_VALUE, NodeType.ENUM_FIELD) : NodeType.BOOL,
            (NodeType.MODE_FIELD, NodeType.MODE_VALUE) : NodeType.BOOL,
            (NodeType.MODE_VALUE, NodeType.MODE_FIELD) : NodeType.BOOL,
            (NodeType.MODE_FIELD, NodeType.MODE_FIELD) : NodeType.BOOL,
            (NodeType.ERROR_FIELD, NodeType.ERROR_VALUE) : NodeType.BOOL,
            (NodeType.ERROR_VALUE, NodeType.ERROR_FIELD) : NodeType.BOOL,
            (NodeType.ERROR_FIELD, NodeType.ERROR_FIELD) : NodeType.BOOL,
        }

binary_less_result = \
        {
            (NodeType.INT_CONST, NodeType.INT_CONST) : NodeType.BOOL,

            (NodeType.INT_CONST, NodeType.REAL_CONST) : NodeType.BOOL,
            (NodeType.REAL_CONST, NodeType.INT_CONST) : NodeType.BOOL,
            (NodeType.INT_CONST, NodeType.INT_RANGE) : NodeType.BOOL,
            (NodeType.INT_RANGE, NodeType.INT_CONST) : NodeType.BOOL,
            (NodeType.INT_CONST, NodeType.INT) : NodeType.BOOL,
            (NodeType.INT, NodeType.INT_CONST) : NodeType.BOOL,
            (NodeType.INT_CONST, NodeType.REAL) : NodeType.BOOL,
            (NodeType.REAL, NodeType.INT_CONST) : NodeType.BOOL,
            (NodeType.INT_CONST, NodeType.CLOCK) : NodeType.BOOL,
            (NodeType.CLOCK, NodeType.INT_CONST) : NodeType.BOOL,
            (NodeType.INT_CONST, NodeType.CONTINUOUS) : NodeType.BOOL,
            (NodeType.CONTINUOUS, NodeType.INT_CONST) : NodeType.BOOL,

            (NodeType.REAL_CONST, NodeType.REAL_CONST) : NodeType.BOOL,
            (NodeType.REAL_CONST, NodeType.REAL) : NodeType.BOOL,
            (NodeType.REAL, NodeType.REAL_CONST) : NodeType.BOOL,
            (NodeType.REAL_CONST, NodeType.CLOCK) : NodeType.BOOL,
            (NodeType.CLOCK, NodeType.REAL_CONST) : NodeType.BOOL,
            (NodeType.REAL_CONST, NodeType.CONTINUOUS) : NodeType.BOOL,
            (NodeType.CONTINUOUS, NodeType.REAL_CONST) : NodeType.BOOL,

            (NodeType.INT_RANGE, NodeType.INT_RANGE) : NodeType.BOOL,
            (NodeType.INT_RANGE, NodeType.INT) : NodeType.BOOL,
            (NodeType.INT, NodeType.INT_RANGE) : NodeType.BOOL,
            (NodeType.INT_RANGE, NodeType.REAL) : NodeType.BOOL,
            (NodeType.REAL, NodeType.INT_RANGE) : NodeType.BOOL,
            (NodeType.INT_RANGE, NodeType.CLOCK) : NodeType.BOOL,
            (NodeType.CLOCK, NodeType.INT_RANGE) : NodeType.BOOL,
            (NodeType.INT_RANGE, NodeType.CONTINUOUS) : NodeType.BOOL,
            (NodeType.CONTINUOUS, NodeType.INT_RANGE) : NodeType.BOOL,

            (NodeType.INT, NodeType.INT) : NodeType.BOOL,
            (NodeType.INT, NodeType.REAL) : NodeType.BOOL,
            (NodeType.REAL, NodeType.INT) : NodeType.BOOL,
            (NodeType.INT, NodeType.CLOCK) : NodeType.BOOL,
            (NodeType.CLOCK, NodeType.INT) : NodeType.BOOL,
            (NodeType.INT, NodeType.CONTINUOUS) : NodeType.BOOL,
            (NodeType.CONTINUOUS, NodeType.INT) : NodeType.BOOL,

            (NodeType.REAL, NodeType.REAL) : NodeType.BOOL,
            (NodeType.REAL, NodeType.CLOCK) : NodeType.BOOL,
            (NodeType.CLOCK, NodeType.REAL) : NodeType.BOOL,
            (NodeType.REAL, NodeType.CONTINUOUS) : NodeType.BOOL,
            (NodeType.CONTINUOUS, NodeType.REAL) : NodeType.BOOL,

            (NodeType.CLOCK, NodeType.CLOCK) : NodeType.BOOL,
            (NodeType.CLOCK, NodeType.CONTINUOUS) : NodeType.BOOL,
            (NodeType.CONTINUOUS, NodeType.CLOCK) : NodeType.BOOL,
            (NodeType.CONTINUOUS, NodeType.CONTINUOUS) : NodeType.BOOL,
        }

binary_greater_result = \
        {
            (NodeType.INT_CONST, NodeType.INT_CONST) : NodeType.BOOL,

            (NodeType.INT_CONST, NodeType.REAL_CONST) : NodeType.BOOL,
            (NodeType.REAL_CONST, NodeType.INT_CONST) : NodeType.BOOL,
            (NodeType.INT_CONST, NodeType.INT_RANGE) : NodeType.BOOL,
            (NodeType.INT_RANGE, NodeType.INT_CONST) : NodeType.BOOL,
            (NodeType.INT_CONST, NodeType.INT) : NodeType.BOOL,
            (NodeType.INT, NodeType.INT_CONST) : NodeType.BOOL,
            (NodeType.INT_CONST, NodeType.REAL) : NodeType.BOOL,
            (NodeType.REAL, NodeType.INT_CONST) : NodeType.BOOL,
            (NodeType.INT_CONST, NodeType.CLOCK) : NodeType.BOOL,
            (NodeType.CLOCK, NodeType.INT_CONST) : NodeType.BOOL,
            (NodeType.INT_CONST, NodeType.CONTINUOUS) : NodeType.BOOL,
            (NodeType.CONTINUOUS, NodeType.INT_CONST) : NodeType.BOOL,

            (NodeType.REAL_CONST, NodeType.REAL_CONST) : NodeType.BOOL,
            (NodeType.REAL_CONST, NodeType.REAL) : NodeType.BOOL,
            (NodeType.REAL, NodeType.REAL_CONST) : NodeType.BOOL,
            (NodeType.REAL_CONST, NodeType.CLOCK) : NodeType.BOOL,
            (NodeType.CLOCK, NodeType.REAL_CONST) : NodeType.BOOL,
            (NodeType.REAL_CONST, NodeType.CONTINUOUS) : NodeType.BOOL,
            (NodeType.CONTINUOUS, NodeType.REAL_CONST) : NodeType.BOOL,

            (NodeType.INT_RANGE, NodeType.INT_RANGE) : NodeType.BOOL,
            (NodeType.INT_RANGE, NodeType.INT) : NodeType.BOOL,
            (NodeType.INT, NodeType.INT_RANGE) : NodeType.BOOL,
            (NodeType.INT_RANGE, NodeType.REAL) : NodeType.BOOL,
            (NodeType.REAL, NodeType.INT_RANGE) : NodeType.BOOL,
            (NodeType.INT_RANGE, NodeType.CLOCK) : NodeType.BOOL,
            (NodeType.CLOCK, NodeType.INT_RANGE) : NodeType.BOOL,
            (NodeType.INT_RANGE, NodeType.CONTINUOUS) : NodeType.BOOL,
            (NodeType.CONTINUOUS, NodeType.INT_RANGE) : NodeType.BOOL,

            (NodeType.INT, NodeType.INT) : NodeType.BOOL,
            (NodeType.INT, NodeType.REAL) : NodeType.BOOL,
            (NodeType.REAL, NodeType.INT) : NodeType.BOOL,
            (NodeType.INT, NodeType.CLOCK) : NodeType.BOOL,
            (NodeType.CLOCK, NodeType.INT) : NodeType.BOOL,
            (NodeType.INT, NodeType.CONTINUOUS) : NodeType.BOOL,
            (NodeType.CONTINUOUS, NodeType.INT) : NodeType.BOOL,

            (NodeType.REAL, NodeType.REAL) : NodeType.BOOL,
            (NodeType.REAL, NodeType.CLOCK) : NodeType.BOOL,
            (NodeType.CLOCK, NodeType.REAL) : NodeType.BOOL,
            (NodeType.REAL, NodeType.CONTINUOUS) : NodeType.BOOL,
            (NodeType.CONTINUOUS, NodeType.REAL) : NodeType.BOOL,

            (NodeType.CLOCK, NodeType.CLOCK) : NodeType.BOOL,
            (NodeType.CLOCK, NodeType.CONTINUOUS) : NodeType.BOOL,
            (NodeType.CONTINUOUS, NodeType.CLOCK) : NodeType.BOOL,
            (NodeType.CONTINUOUS, NodeType.CONTINUOUS) : NodeType.BOOL,
        }

binary_lessorequal_result = \
        {
            (NodeType.INT_CONST, NodeType.INT_CONST) : NodeType.BOOL,

            (NodeType.INT_CONST, NodeType.REAL_CONST) : NodeType.BOOL,
            (NodeType.REAL_CONST, NodeType.INT_CONST) : NodeType.BOOL,
            (NodeType.INT_CONST, NodeType.INT_RANGE) : NodeType.BOOL,
            (NodeType.INT_RANGE, NodeType.INT_CONST) : NodeType.BOOL,
            (NodeType.INT_CONST, NodeType.INT) : NodeType.BOOL,
            (NodeType.INT, NodeType.INT_CONST) : NodeType.BOOL,
            (NodeType.INT_CONST, NodeType.REAL) : NodeType.BOOL,
            (NodeType.REAL, NodeType.INT_CONST) : NodeType.BOOL,
            (NodeType.INT_CONST, NodeType.CLOCK) : NodeType.BOOL,
            (NodeType.CLOCK, NodeType.INT_CONST) : NodeType.BOOL,
            (NodeType.INT_CONST, NodeType.CONTINUOUS) : NodeType.BOOL,
            (NodeType.CONTINUOUS, NodeType.INT_CONST) : NodeType.BOOL,

            (NodeType.REAL_CONST, NodeType.REAL_CONST) : NodeType.BOOL,
            (NodeType.REAL_CONST, NodeType.REAL) : NodeType.BOOL,
            (NodeType.REAL, NodeType.REAL_CONST) : NodeType.BOOL,
            (NodeType.REAL_CONST, NodeType.CLOCK) : NodeType.BOOL,
            (NodeType.CLOCK, NodeType.REAL_CONST) : NodeType.BOOL,
            (NodeType.REAL_CONST, NodeType.CONTINUOUS) : NodeType.BOOL,
            (NodeType.CONTINUOUS, NodeType.REAL_CONST) : NodeType.BOOL,

            (NodeType.INT_RANGE, NodeType.INT_RANGE) : NodeType.BOOL,
            (NodeType.INT_RANGE, NodeType.INT) : NodeType.BOOL,
            (NodeType.INT, NodeType.INT_RANGE) : NodeType.BOOL,
            (NodeType.INT_RANGE, NodeType.REAL) : NodeType.BOOL,
            (NodeType.REAL, NodeType.INT_RANGE) : NodeType.BOOL,
            (NodeType.INT_RANGE, NodeType.CLOCK) : NodeType.BOOL,
            (NodeType.CLOCK, NodeType.INT_RANGE) : NodeType.BOOL,
            (NodeType.INT_RANGE, NodeType.CONTINUOUS) : NodeType.BOOL,
            (NodeType.CONTINUOUS, NodeType.INT_RANGE) : NodeType.BOOL,

            (NodeType.INT, NodeType.INT) : NodeType.BOOL,
            (NodeType.INT, NodeType.REAL) : NodeType.BOOL,
            (NodeType.REAL, NodeType.INT) : NodeType.BOOL,
            (NodeType.INT, NodeType.CLOCK) : NodeType.BOOL,
            (NodeType.CLOCK, NodeType.INT) : NodeType.BOOL,
            (NodeType.INT, NodeType.CONTINUOUS) : NodeType.BOOL,
            (NodeType.CONTINUOUS, NodeType.INT) : NodeType.BOOL,

            (NodeType.REAL, NodeType.REAL) : NodeType.BOOL,
            (NodeType.REAL, NodeType.CLOCK) : NodeType.BOOL,
            (NodeType.CLOCK, NodeType.REAL) : NodeType.BOOL,
            (NodeType.REAL, NodeType.CONTINUOUS) : NodeType.BOOL,
            (NodeType.CONTINUOUS, NodeType.REAL) : NodeType.BOOL,

            (NodeType.CLOCK, NodeType.CLOCK) : NodeType.BOOL,
            (NodeType.CLOCK, NodeType.CONTINUOUS) : NodeType.BOOL,
            (NodeType.CONTINUOUS, NodeType.CLOCK) : NodeType.BOOL,
            (NodeType.CONTINUOUS, NodeType.CONTINUOUS) : NodeType.BOOL,
        }

binary_greaterorequal_result = \
        {
            (NodeType.INT_CONST, NodeType.INT_CONST) : NodeType.BOOL,

            (NodeType.INT_CONST, NodeType.REAL_CONST) : NodeType.BOOL,
            (NodeType.REAL_CONST, NodeType.INT_CONST) : NodeType.BOOL,
            (NodeType.INT_CONST, NodeType.INT_RANGE) : NodeType.BOOL,
            (NodeType.INT_RANGE, NodeType.INT_CONST) : NodeType.BOOL,
            (NodeType.INT_CONST, NodeType.INT) : NodeType.BOOL,
            (NodeType.INT, NodeType.INT_CONST) : NodeType.BOOL,
            (NodeType.INT_CONST, NodeType.REAL) : NodeType.BOOL,
            (NodeType.REAL, NodeType.INT_CONST) : NodeType.BOOL,
            (NodeType.INT_CONST, NodeType.CLOCK) : NodeType.BOOL,
            (NodeType.CLOCK, NodeType.INT_CONST) : NodeType.BOOL,
            (NodeType.INT_CONST, NodeType.CONTINUOUS) : NodeType.BOOL,
            (NodeType.CONTINUOUS, NodeType.INT_CONST) : NodeType.BOOL,

            (NodeType.REAL_CONST, NodeType.REAL_CONST) : NodeType.BOOL,
            (NodeType.REAL_CONST, NodeType.REAL) : NodeType.BOOL,
            (NodeType.REAL, NodeType.REAL_CONST) : NodeType.BOOL,
            (NodeType.REAL_CONST, NodeType.CLOCK) : NodeType.BOOL,
            (NodeType.CLOCK, NodeType.REAL_CONST) : NodeType.BOOL,
            (NodeType.REAL_CONST, NodeType.CONTINUOUS) : NodeType.BOOL,
            (NodeType.CONTINUOUS, NodeType.REAL_CONST) : NodeType.BOOL,

            (NodeType.INT_RANGE, NodeType.INT_RANGE) : NodeType.BOOL,
            (NodeType.INT_RANGE, NodeType.INT) : NodeType.BOOL,
            (NodeType.INT, NodeType.INT_RANGE) : NodeType.BOOL,
            (NodeType.INT_RANGE, NodeType.REAL) : NodeType.BOOL,
            (NodeType.REAL, NodeType.INT_RANGE) : NodeType.BOOL,
            (NodeType.INT_RANGE, NodeType.CLOCK) : NodeType.BOOL,
            (NodeType.CLOCK, NodeType.INT_RANGE) : NodeType.BOOL,
            (NodeType.INT_RANGE, NodeType.CONTINUOUS) : NodeType.BOOL,
            (NodeType.CONTINUOUS, NodeType.INT_RANGE) : NodeType.BOOL,

            (NodeType.INT, NodeType.INT) : NodeType.BOOL,
            (NodeType.INT, NodeType.REAL) : NodeType.BOOL,
            (NodeType.REAL, NodeType.INT) : NodeType.BOOL,
            (NodeType.INT, NodeType.CLOCK) : NodeType.BOOL,
            (NodeType.CLOCK, NodeType.INT) : NodeType.BOOL,
            (NodeType.INT, NodeType.CONTINUOUS) : NodeType.BOOL,
            (NodeType.CONTINUOUS, NodeType.INT) : NodeType.BOOL,

            (NodeType.REAL, NodeType.REAL) : NodeType.BOOL,
            (NodeType.REAL, NodeType.CLOCK) : NodeType.BOOL,
            (NodeType.CLOCK, NodeType.REAL) : NodeType.BOOL,
            (NodeType.REAL, NodeType.CONTINUOUS) : NodeType.BOOL,
            (NodeType.CONTINUOUS, NodeType.REAL) : NodeType.BOOL,

            (NodeType.CLOCK, NodeType.CLOCK) : NodeType.BOOL,
            (NodeType.CLOCK, NodeType.CONTINUOUS) : NodeType.BOOL,
            (NodeType.CONTINUOUS, NodeType.CLOCK) : NodeType.BOOL,
            (NodeType.CONTINUOUS, NodeType.CONTINUOUS) : NodeType.BOOL,
        }

unary_not_result = \
        {
            NodeType.BOOL : NodeType.BOOL
        }

binary_and_result = \
        {
            (NodeType.BOOL, NodeType.BOOL) : NodeType.BOOL,
            (NodeType.TRAJECTORY_EQUATION, NodeType.BOOL) : NodeType.BOOL,
            (NodeType.BOOL, NodeType.TRAJECTORY_EQUATION) : NodeType.BOOL,
            (NodeType.TRAJECTORY_EQUATION, NodeType.TRAJECTORY_EQUATION) : NodeType.BOOL,
        }

binary_iff_result = \
        {
            (NodeType.BOOL, NodeType.BOOL) : NodeType.BOOL
        }

binary_imp_result = \
        {
            (NodeType.BOOL, NodeType.BOOL) : NodeType.BOOL
        }

binary_or_result = \
        {
            (NodeType.BOOL, NodeType.BOOL) : NodeType.BOOL
        }

binary_xnor_result = \
        {
            (NodeType.BOOL, NodeType.BOOL) : NodeType.BOOL
        }

binary_xor_result = \
        {
            (NodeType.BOOL, NodeType.BOOL) : NodeType.BOOL
        }

assign_result = \
        {
            (NodeType.BOOL, NodeType.BOOL) : NodeType.BOOL,

            (NodeType.INT_RANGE, NodeType.INT_CONST) : NodeType.INT_RANGE,
            (NodeType.INT_RANGE, NodeType.INT_RANGE) : NodeType.INT_RANGE,

            (NodeType.INT, NodeType.INT) : NodeType.INT,
            (NodeType.INT, NodeType.INT_CONST) : NodeType.INT,
            (NodeType.INT, NodeType.INT_RANGE) : NodeType.INT,

            (NodeType.REAL, NodeType.INT_CONST) : NodeType.REAL,
            (NodeType.REAL, NodeType.REAL_CONST) : NodeType.REAL,
            (NodeType.REAL, NodeType.INT_RANGE) : NodeType.REAL,
            (NodeType.REAL, NodeType.INT) : NodeType.REAL,
            (NodeType.REAL, NodeType.REAL) : NodeType.REAL,
            (NodeType.REAL, NodeType.CLOCK) : NodeType.REAL,
            (NodeType.REAL, NodeType.CONTINUOUS) : NodeType.REAL,

            (NodeType.CLOCK, NodeType.INT_CONST) : NodeType.CLOCK,
            (NodeType.CLOCK, NodeType.REAL_CONST) : NodeType.CLOCK,

            (NodeType.CONTINUOUS, NodeType.INT_CONST) : NodeType.CONTINUOUS,
            (NodeType.CONTINUOUS, NodeType.REAL_CONST) : NodeType.CONTINUOUS,

            (NodeType.ENUM_FIELD, NodeType.ENUM_FIELD) : NodeType.ENUM_FIELD,
            (NodeType.ENUM_FIELD, NodeType.ENUM_VALUE) : NodeType.ENUM_FIELD
        }

unary_time_scale_result = \
        {
            NodeType.CLOCK : NodeType.REAL,
            NodeType.INT : NodeType.CLOCK,
            NodeType.INT_CONST : NodeType.CLOCK,
            NodeType.REAL_CONST : NodeType.CLOCK,
            NodeType.REAL : NodeType.CLOCK,
            NodeType.CONTINUOUS : NodeType.CLOCK
        }

def __init__(self):
        pass

def get_result_table_for_operator(operator, cinfo_uses_time_scales):
        result_table = None
        if operator == OperatorType.PLUS:
            result_table = binary_plus_result
        elif operator == OperatorType.MINUS:
            result_table = binary_minus_result
        elif operator == OperatorType.UNARY_MINUS:
            result_table = unary_minus_result
        elif operator == OperatorType.TIMES:
            result_table = binary_times_result
        elif operator == OperatorType.DIVIDE:
            result_table = binary_division_result
        elif operator == OperatorType.MOD:
            result_table = binary_mod_result
        elif operator == OperatorType.EQUAL:
            result_table = binary_equality_result
        elif operator == OperatorType.NOTEQUAL:
            result_table = binary_inequality_result
        elif operator == OperatorType.LESS:
            result_table = binary_less_result
        elif operator == OperatorType.GREATER:
            result_table = binary_greater_result
        elif operator == OperatorType.LESSOREQUAL:
            result_table = binary_lessorequal_result
        elif operator == OperatorType.GREATEROREQUAL:
            result_table = binary_greaterorequal_result
        elif operator == OperatorType.AND:
            result_table = binary_and_result
        elif operator == OperatorType.IFF:
            result_table = binary_iff_result
        elif operator == OperatorType.IMP:
            result_table = binary_imp_result
        elif operator == OperatorType.NOT:
            result_table = unary_not_result
        elif operator == OperatorType.OR:
            result_table = binary_or_result
        elif operator == OperatorType.XNOR:
            result_table = binary_xnor_result
        elif operator == OperatorType.XOR:
            result_table = binary_xor_result
        elif operator == OperatorType.ASSIGN:
            result_table = assign_result
        elif OperatorType.MSEC <= operator <= OperatorType.DAY or\
             OperatorType.REAL_TOMSEC <= operator <= OperatorType.REAL_TODAY or\
             OperatorType.MSEC_TOREAL <= operator <= OperatorType.DAY_TOREAL:
            result_table = unary_time_scale_result

        assert (result_table is not None)

        # If time scales are used NodeType.CLOCK may only be related to another
        # NodeType.CLOCK --> remove invalid entries
        # Exceptions are: TIMES, DIVIDE (scaling), ASSIGN (reset)
        if cinfo_uses_time_scales:
            if operator == OperatorType.ASSIGN:
                result = dict()
                for (k1,k2) in result_table.iterkeys():
                    if (k1 != NodeType.CLOCK and k2 != NodeType.CLOCK) \
                       or (k1 == NodeType.CLOCK and (k2 == NodeType.CLOCK or k2 in constant_types)):
                        result[(k1,k2)] = result_table[(k1,k2)]
                return result
            elif operator not in {OperatorType.TIMES, OperatorType.DIVIDE} \
                 and isinstance(result_table.keys()[0], tuple):
                result = dict()
                for (k1,k2) in result_table.iterkeys():
                    if (k1 != NodeType.CLOCK and k2 != NodeType.CLOCK) \
                    or (k1 == NodeType.CLOCK and k2 == NodeType.CLOCK):
                        result[(k1,k2)] = result_table[(k1,k2)]
                return result

        return result_table

def get_all_identifiers(root_node):
    """Retrieves all identifiers in the expression rooted by the given node.

    @param    root_node:
                The root node of the expression tree.
    @type     root_node:
                An object of a subclass of expressions.ExpressionTreeNode.
    @result:  A set of all identifiers occuring in the expression.
    @rtype:   A set of strings.
    """
    if isinstance(root_node, BinaryOperatorTreeNode):
        return (get_all_identifiers(root_node.left) |
                get_all_identifiers(root_node.right))
    elif isinstance(root_node, UnaryOperatorTreeNode):
        return (get_all_identifiers(root_node.subexpression))
    elif isinstance(root_node, CaseExpressionTreeNode):
        result = set()
        for (cond, expr) in root_node.cases:
            result |= (get_all_identifiers(cond) | get_all_identifiers(expr))
        result |= get_all_identifiers(root_node.default_expression)
        return result
    elif isinstance(root_node, TrajectoryEquationTreeNode):
        return set([root_node.variable])
    elif isinstance(root_node, TrajectoryExpressionTreeNode):
        return set()
    elif isinstance(root_node, NumberTreeNode):
        return set()
    elif isinstance(root_node, BooleanTreeNode):
        return set()
    elif isinstance(root_node, IdentifierTreeNode):
        return set([root_node.name])
    else:
        assert False

def get_AST(root_node):
    result = None

    if root_node is not None:
        if isinstance(root_node, IdentifierTreeNode):
            result = root_node.node_type
        elif isinstance(root_node, BooleanTreeNode):
            result = root_node.node_type
        elif isinstance(root_node, NumberTreeNode):
            result = root_node.node_type
        elif isinstance(root_node, TrajectoryExpressionTreeNode):
            result = root_node.node_type

        if result:
            return str(result)

        #complex case
        if isinstance(root_node, BinaryOperatorTreeNode):
            result = "(" + get_AST(root_node.left) + ") " + \
                     OperatorType.operator_to_str(root_node.operator) + \
                     " (" + get_AST(root_node.right) + ")"
        elif isinstance(root_node, UnaryOperatorTreeNode):
            result = OperatorType.operator_to_str(root_node.operator) + \
                     "(" + get_AST(root_node.subexpression) + ")"
        elif isinstance(root_node, TrajectoryEquationTreeNode):
            result = "(" + str(root_node.variable.node_type) + "' = " + str(root_node.trajectory_expression.node_type) + ")"
        elif isinstance(root_node, CaseExpressionTreeNode):
            result = "case"
            for (cond, expr) in root_node.cases:
                result += "(" + get_AST(cond) + ": " + get_AST(expr) + ")"
            result += " otherwise " + get_AST(root_node.default_expression) + " end"

        assert result is not None
        return str(result)
