# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================
"""This module provides classes shared by all visitors.
Take care to apply modification to one visitor to the others as well
"""

from functools import partial
from utils.exceptions import InternalException

class BaseVisitor(object):
    """This class is the basic visitor class all visitors need to derive from.
    It supports features polymorphism and expects the visit-methods to be named according
    to the pattern: visit_<type>. The parameter count of the visit-methods is variable.
    """

    def _get_method(self, node, name='visit', optional=False):
        """Retrieved the visit method matching to the closest related type
        of the passed node. If optional, None may be returned if no such
        function, otherwise a generic function will be returned raising an
        exception"""
        assert node is not None

        meth = None
        for cls in node.__class__.__mro__:
            meth_name = name + '_' + cls.__name__
            meth = getattr(self, meth_name, None)
            if callable(meth):
                break

        if meth is None:
            if not optional:
                meth = partial(self._generic_visit, name=name)
            else:
                meth = None

        return meth

    def _generic_visit(self, node, name, *args):
        """Raises error when a visit-method was expected to exist but couldn't be found
        """
        raise InternalException('No {} method for type {}'.format(name, type(node).__name__))

    def visit(self, node, *args):
        meth = self._get_method(node, name='visit')
        return meth(node, *args)

class ActionVisitor(BaseVisitor):
    """Class like BaseVisitor, but visit methods return values may
    be post processed by an action method."""
    def __init__(self, pass_result=True):
        super(ActionVisitor, self).__init__()

        self.pass_result = pass_result
        if self.pass_result:
            self._generic_visit = self._generic_visit_data

    def visit(self, node, *args):
        """Does visit, action. Return value of visit passed to action,
        returns return value of action, otherwise visit"""
        visit_meth = self._get_method(node, name='visit')
        value = visit_meth(node, *args)

        post_meth = self._get_method(node, name='action', optional=True)
        if post_meth:
            if self.pass_result:
                value = post_meth(node, value, *args)
            else:
                value = post_meth(node, *args)

        return value

    def _generic_visit_data(self, node, data, name, *args):
        """Raises error when a visit-method was expected to exist but couldn't be found
        """
        raise InternalException('No {} method for type {}'.format(name, type(node).__name__))

class CompileVisitor(ActionVisitor):
    """Simple compileinfo visitor. Simply walks the tree. Actions
    should not modify the tree (but may modify attributes).
    The visit return value is NOT passed to action.
    default visit mathods accept any number of arguments,
    but do not act upon them. For usage see expression postprocessor"""
    def __init__(self):
        super(CompileVisitor, self).__init__(pass_result=False)
        pass

    def _visit_dict(self, dict_arg, *args):
        for v in dict_arg.values():
            self.visit(v, *args)

    def _visit_list(self, list_arg, *args):
        for x in list_arg:
            self.visit(x, *args)

    def visit_CompileInformation(self, node, *args, **kwargs):
        self._visit_dict(node.packages)

    def visit_Package(self, node, *args, **kwargs):
        self.visit(node.public_subpackage)
        self.visit(node.private_subpackage)

    def visit_Subpackage(self, node, *args, **kwargs):
        self._visit_dict(node.types)
        self._visit_dict(node.implementations)
        self._visit_dict(node.error_types)
        self._visit_dict(node.error_implementations)

    def visit_ComponentType(self, node, *args, **kwargs):
        self._visit_dict(node.in_data_ports)
        self._visit_dict(node.out_data_ports)
        self._visit_dict(node.in_event_ports)
        self._visit_dict(node.out_event_ports)

    def visit_ComponentImplementation(self, node, *args, **kwargs):
        self._visit_dict(node.non_data_subcomponents)
        self._visit_dict(node.data_subcomponents)
        self._visit_list(node.data_port_connections, "data")
        self._visit_list(node.event_port_connections, "event")
        self._visit_list(node.flows)
        self._visit_dict(node.modes)
        self._visit_list(node.normal_mode_transitions)
        self._visit_list(node.probability_mode_transitions)
        self._visit_list(node.reactivation_mode_transitions)

    def visit_InEventPort(self, node, *args, **kwargs):
        pass

    def visit_OutEventPort(self, node, *args, **kwargs):
        pass

    def visit_DataPort(self, node, *args, **kwargs):
        if node.default_value is not None:
            self.visit(node.default_value)

    def visit_OutDataPort(self, node, *args, **kwargs):
        if node.default_value is not None:
            self.visit(node.default_value)

    def visit_ErrorEventPort(self, node, *args, **kwargs):
        pass

    def visit_PortConnection(self, node, *args, **kwargs):
        pass

    def visit_Flow(self, node, *args, **kwargs):
        self.visit(node.expression)

    def visit_NonDataSubcomponent(self, node, *args, **kwargs):
        pass

    def visit_DataSubcomponent(self, node, *args, **kwargs):
        self.visit(node.data_type)
        if node.default_value is not None:
            self.visit(node.default_value)

    def visit_Mode(self, node, *args, **kwargs):
        if node.urgent is not None:
            self.visit(node.urgent)
        if node.invariant is not None:
            self.visit(node.invariant)

    def visit_ModeTransition(self, node, *args, **kwargs):
        self.visit(node.label)

    def visit_ModeTransitionLabel(self, node, *args, **kwargs):
        if node.time_delay is not None:
            self.visit(node.time_delay[0])
            self.visit(node.time_delay[1])
        if node.guard is not None:
            self.visit(node.guard)
        if node.effects is not None:
            self._visit_list(node.effects)

    def visit_ProbabilityModeTransitionLabel(self, node, *args, **kwargs):
        pass

    def visit_ReactivationModeTransitionLabel(self, node, *args, **kwargs):
        if node.effects is not None:
            self._visit_list(node.effects)

    def visit_ErrorModelType(self, node, *args, **kwargs):
        self._visit_dict(node.states)
        self._visit_dict(node.error_propagations)

    def visit_ErrorModelImplementation(self, node, *args, **kwargs):
        self._visit_dict(node.error_events)
        self._visit_dict(node.clocks)
        self._visit_dict(node.states)
        self._visit_list(node.error_transitions)
        self._visit_list(node.reactivation_transitions)

    def visit_ErrorState(self, node, *args, **kwargs):
        if node.urgent is not None:
            self.visit(node.urgent)
        if node.invariant is not None:
            self.visit(node.invariant)

    def visit_ErrorTransition(self, node, *args, **kwargs):
        if node.time_delay is not None:
            self.visit(node.time_delay[0])
            self.visit(node.time_delay[1])
        if node.guard is not None:
            self.visit(node.guard)
        if node.effects is not None:
            self._visit_list(node.effects)

    def visit_ErrorEvent(self, node, *args, **kwargs):
        pass 

    def visit_ErrorPropagation(self, node, *args, **kwargs):
        pass

    def visit_AbstractDataType(self, node, *args, **kwargs):
        pass

    #######################
    #     Expressions     #
    #######################

    def visit_IdentifierTreeNode(self, root_node, *args, **kwargs):
        pass

    def visit_BooleanTreeNode(self, root_node, *args, **kwargs):
        pass

    def visit_NumberTreeNode(self, root_node, *args, **kwargs):
        pass

    def visit_TrajectoryExpressionTreeNode(self, root_node, *args, **kwargs):
        #deprecated
        pass

    def visit_BinaryOperatorTreeNode(self, root_node, *args, **kwargs):
        self.visit(root_node.left)
        self.visit(root_node.right)

    def visit_UnaryOperatorTreeNode(self, root_node, *args, **kwargs):
        self.visit(root_node.subexpression)

    def visit_TrajectoryEquationTreeNode(self, root_node, *args, **kwargs):
        self.visit(root_node.variable)
        self.visit(root_node.trajectory_expression)

    def visit_CaseExpressionTreeNode(self, root_node, *args, **kwargs):
        for (cond, res) in root_node.cases:
            self.visit(cond)
            self.visit(res)
        self.visit(root_node.default_expression)

class TransformVisitor(ActionVisitor):
    """Transformation visitor. Visit method returns the transformed
    object. By default, children get replaced by this object in visit
    methods. Note that 'action' may be used to transform the actual
    tree"""
    def __init__(self, pass_result=True):
        super(TransformVisitor, self).__init__(pass_result)
        pass

    def _transform_dict(self, dict_arg, *args):
        for k, v in dict_arg.iteritems():
            dict_arg[k] = self.visit(v, *args)

    def _transform_list(self, list_arg, *args):
        list_arg[:] = [self.visit(x, *args) for x in list_arg]

    def visit_CompileInformation(self, node, *args, **kwargs):
        self._transform_dict(node.packages)

        return node

    def visit_Package(self, node, *args, **kwargs):
        node.public_subpackage = self.visit(node.public_subpackage)
        node.private_subpackage = self.visit(node.private_subpackage)

        return node

    def visit_Subpackage(self, node, *args, **kwargs):
        self._transform_dict(node.types)
        self._transform_dict(node.implementations)
        self._transform_dict(node.error_types)
        self._transform_dict(node.error_implementations)

        return node

    def visit_ComponentType(self, node, *args, **kwargs):
        self._transform_dict(node.in_data_ports)
        self._transform_dict(node.out_data_ports)
        self._transform_dict(node.in_event_ports)
        self._transform_dict(node.out_event_ports)

        return node

    def visit_ComponentImplementation(self, node, *args, **kwargs):
        self._transform_dict(node.non_data_subcomponents)
        self._transform_dict(node.data_subcomponents)
        self._transform_list(node.data_port_connections, "data")
        self._transform_list(node.event_port_connections, "event")
        self._transform_list(node.flows)
        self._transform_dict(node.modes)
        self._transform_list(node.normal_mode_transitions)
        self._transform_list(node.probability_mode_transitions)
        self._transform_list(node.reactivation_mode_transitions)

        return node

    def visit_InEventPort(self, node, *args, **kwargs):
        return node

    def visit_OutEventPort(self, node, *args, **kwargs):
        return node

    def visit_DataPort(self, node, *args, **kwargs):
        if node.default_value is not None:
            node.default_value = self.visit(node.default_value)

        return node

    def visit_OutDataPort(self, node, *args, **kwargs):
        if node.default_value is not None:
            node.default_value = self.visit(node.default_value)

        return node

    def visit_ErrorEventPort(self, node, *args, **kwargs):
        return node

    def visit_PortConnection(self, node, *args, **kwargs):
        return node

    def visit_Flow(self, node, *args, **kwargs):
        node.expression = self.visit(node.expression)

        return node

    def visit_NonDataSubcomponent(self, node, *args, **kwargs):
        return node

    def visit_DataSubcomponent(self, node, *args, **kwargs):
        node.data_type = self.visit(node.data_type)
        if node.default_value is not None:
            node.default_value = self.visit(node.default_value)

        return node

    def visit_Mode(self, node, *args, **kwargs):
        if node.urgent is not None:
            node.urgent = self.visit(node.urgent)
        if node.invariant is not None:
            node.invariant = self.visit(node.invariant)

        return node

    def visit_ModeTransition(self, node, *args, **kwargs):
        node.label = self.visit(node.label)

        return node

    def visit_ModeTransitionLabel(self, node, *args, **kwargs):
        if node.time_delay is not None:
            node.time_delay = (self.visit(node.time_delay[0]), self.visit(node.time_delay[1]))
        if node.guard is not None:
            node.guard = self.visit(node.guard)
        if node.effects is not None:
            self._transform_list(node.effects)

        return node

    def visit_ProbabilityModeTransitionLabel(self, node, *args, **kwargs):
        return node

    def visit_ReactivationModeTransitionLabel(self, node, *args, **kwargs):
        if node.effects:
            self._transform_list(node.effects)

        return node

    def visit_ErrorModelType(self, node, *args, **kwargs):
        self._transform_dict(node.states)
        self._transform_dict(node.error_propagations)

        return node

    def visit_ErrorModelImplementation(self, node, *args, **kwargs):
        self._transform_dict(node.error_events)
        self._transform_dict(node.clocks)
        self._transform_dict(node.states)
        self._transform_list(node.error_transitions)
        self._transform_list(node.reactivation_transitions)

        return node

    def visit_ErrorState(self, node, *args, **kwargs):
        if node.urgent is not None:
            node.urgent = self.visit(node.urgent)
        if node.invariant is not None:
            node.invariant = self.visit(node.invariant)

        return node

    def visit_ErrorTransition(self, node, *args, **kwargs):
        if node.time_delay is not None:
            node.time_delay = (self.visit(node.time_delay[0]), self.visit(node.time_delay[1]))
        if node.guard is not None:
            node.guard = self.visit(node.guard)
        if node.effects:
            self._transform_list(node.effects)

        return node

    def visit_ErrorEvent(self, node, *args, **kwargs):
        return node 

    def visit_ErrorPropagation(self, node, *args, **kwargs):
        return node

    def visit_AbstractDataType(self, node, *args, **kwargs):
        return node

    #######################
    #     Expressions     #
    #######################

    def visit_IdentifierTreeNode(self, root_node, *args, **kwargs):
        return root_node

    def visit_BooleanTreeNode(self, root_node, *args, **kwargs):
        return root_node

    def visit_NumberTreeNode(self, root_node, *args, **kwargs):
        return root_node

    def visit_TrajectoryExpressionTreeNode(self, root_node, *args, **kwargs):
        #deprecated
        return root_node

    def visit_BinaryOperatorTreeNode(self, root_node, *args, **kwargs):
        root_node.left = self.visit(root_node.left)
        root_node.left.parent = root_node
        root_node.right = self.visit(root_node.right)
        root_node.right.parent = root_node

        return root_node

    def visit_UnaryOperatorTreeNode(self, root_node, *args, **kwargs):
        root_node.subexpression = self.visit(root_node.subexpression)
        root_node.subexpression.parent = root_node

        return root_node

    def visit_TrajectoryEquationTreeNode(self, root_node, *args, **kwargs):
        root_node.variable = self.visit(root_node.variable)
        root_node.trajectory_expression = self.visit(root_node.trajectory_expression)
        return root_node

    def visit_CaseExpressionTreeNode(self, root_node, *args, **kwargs):
        root_node.cases = [(self.visit(cond), self.visit(res)) for (cond, res) in root_node.cases]
        root_node.default_expression = self.visit(root_node.default_expression)

        return root_node

class AttrDict(dict):
    """Class to generate simple dictionary object, used by EvalVisitor.
    Note: requires python 2.7.3, otherwise mem-leak"""
    def __init__(self, *args, **kwargs):
        super(AttrDict, self).__init__(*args, **kwargs)
        self.__dict__ = self

class EvalVisitor(ActionVisitor):
    """Evaluation visitor. Visit method sets attributes of data object with.
    This object is passed as a parameter to action methods. Action methods
    can aggregate this result into a new object for the current node (by
    default is None).
    Data object attributes correspond to node attributes, but with aggregated
    data.
    Can be used to generate string representation or to evaluate data"""
    def __init__(self):
        super(EvalVisitor, self).__init__(pass_result=True)
        pass

    def visit(self, node, *args):
        """Does visit, action. Return value of visit passed to action,
        returns return value of action, otherwise visit"""

        data = AttrDict()

        visit_meth = self._get_method(node, name='visit')
        visit_meth(node, data, *args)

        post_meth = self._get_method(node, name='action', optional=True)
        if post_meth:
            value = post_meth(node, data, *args)
        else:
            value = None

        return value

    def _eval_dict(self, dict_arg, data_arg, *args):
        for k, v in dict_arg.iteritems():
            data_arg[k] = self.visit(v, *args)

    def _eval_list(self, list_arg, data_arg, *args):
        data_arg[:] = [self.visit(x, *args) for x in list_arg]

    def visit_CompileInformation(self, node, data, *args, **kwargs):
        data.packages = {}
        self._eval_dict(node.packages, data.packages)

    def visit_Package(self, node, data, *args, **kwargs):
        data.public_subpackage = self.visit(node.public_subpackage)
        data.private_subpackage = self.visit(node.private_subpackage)

    def visit_Subpackage(self, node, data, *args, **kwargs):
        data.types = {}
        self._eval_dict(node.types, data.types)
        data.implementations = {}
        self._eval_dict(node.implementations, data.implementations)
        data.error_types = {}
        self._eval_dict(node.error_types, data.error_types)
        data.error_implementations = {}
        self._eval_dict(node.error_implementations, data.error_implementations)

    def visit_ComponentType(self, node, data, *args, **kwargs):
        data.in_data_ports = {}
        self._eval_dict(node.in_data_ports, data.in_data_ports)
        data.out_data_ports = {}
        self._eval_dict(node.out_data_ports, data.out_data_ports)
        data.in_event_ports = {}
        self._eval_dict(node.in_event_ports, data.in_event_ports)
        data.out_event_ports = {}
        self._eval_dict(node.out_event_ports, data.out_event_ports)

    def visit_ComponentImplementation(self, node, data, *args, **kwargs):
        data.non_data_subcomponents = {}
        self._eval_dict(node.non_data_subcomponents, data.non_data_subcomponents)
        data.data_subcomponents = {}
        self._eval_dict(node.data_subcomponents, data.data_subcomponents)
        data.data_port_connections = []
        self._eval_list(node.data_port_connections, data.data_port_connections, "data")
        data.event_port_connections = []
        self._eval_list(node.event_port_connections, data.event_port_connections, "event")
        data.flows = []
        self._eval_list(node.flows, data.flows)
        data.modes = {}
        self._eval_dict(node.modes, data.modes)
        data.normal_mode_transitions = []
        self._eval_list(node.normal_mode_transitions, data.normal_mode_transitions)
        data.probability_mode_transitions = []
        self._eval_list(node.probability_mode_transitions, data.probability_mode_transitions)
        data.reactivation_mode_transitions = []
        self._eval_list(node.reactivation_mode_transitions, data.reactivation_mode_transitions)

    def visit_InEventPort(self, node, data, *args, **kwargs):
        pass

    def visit_OutEventPort(self, node, data, *args, **kwargs):
        pass

    def visit_DataPort(self, node, data, *args, **kwargs):
        data.data_type = self.visit(node.data_type)
        if node.default_value is not None:
            data.default_value = self.visit(node.default_value)

    def visit_OutDataPort(self, node, data, *args, **kwargs):
        data.data_type = self.visit(node.data_type)
        if node.default_value is not None:
            data.default_value = self.visit(node.default_value)

    def visit_ErrorEventPort(self, node, data, *args, **kwargs):
        pass

    def visit_PortConnection(self, node, data, *args, **kwargs):
        pass

    def visit_Flow(self, node, data, *args, **kwargs):
        data.expression = self.visit(node.expression)

    def visit_NonDataSubcomponent(self, node, data, *args, **kwargs):
        pass

    def visit_DataSubcomponent(self, node, data, *args, **kwargs):
        data.data_type = self.visit(node.data_type)
        if node.default_value is not None:
            data.default_value = self.visit(node.default_value)

    def visit_Mode(self, node, data, *args, **kwargs):
        if node.urgent is not None:
            data.urgent = self.visit(node.urgent)
        if node.invariant is not None:
            data.invariant = self.visit(node.invariant)

    def visit_ModeTransition(self, node, data, *args, **kwargs):
        data.label = self.visit(node.label)

    def visit_ModeTransitionLabel(self, node, data, *args, **kwargs):
        if node.time_delay is not None:
            data.time_delay = (self.visit(node.time_delay[0]), self.visit(node.time_delay[1]))
        if node.guard is not None:
            data.guard = self.visit(node.guard)
        if node.effects:
            data.effects = []
            self._eval_list(node.effects, data.effects)

    def visit_ProbabilityModeTransitionLabel(self, node, data, *args, **kwargs):
        pass

    def visit_ReactivationModeTransitionLabel(self, node, data, *args, **kwargs):
        if node.effects:
            data.effects = []
            self._eval_list(node.effects, data.effects)

    def visit_ErrorModelType(self, node, data, *args, **kwargs):
        data.states = {}
        self._eval_dict(node.states, data.states)
        data.error_propagations = {}
        self._eval_dict(node.error_propagations, data.error_propagations)

    def visit_ErrorModelImplementation(self, node, data, *args, **kwargs):
        data.error_events = {}
        self._eval_dict(node.error_events, data.error_events)
        data.clocks = {}
        self._eval_dict(node.clocks, data.clocks)
        data.states = {}
        self._eval_dict(node.states, data.states)
        data.error_transitions = []
        self._eval_list(node.error_transitions, data.error_transitions)
        data.reactivation_transitions = []
        self._eval_list(node.reactivation_transitions, data.reactivation_transitions)

    def visit_ErrorState(self, node, data, *args, **kwargs):
        if node.urgent is not None:
            data.urgent = self.visit(node.urgent)
        if node.invariant is not None:
            data.invariant = self.visit(node.invariant)

    def visit_ErrorTransition(self, node, data, *args, **kwargs):
        if node.time_delay is not None:
            data.time_delay = (self.visit(node.time_delay[0]), self.visit(node.time_delay[1]))
        if node.guard is not None:
            data.guard = self.visit(node.guard)
        if node.effects is not None:
            data.effects = []
            self._eval_list(node.effects, data.effects)

    def visit_ErrorEvent(self, node, data, *args, **kwargs):
        pass 

    def visit_ErrorPropagation(self, node, data, *args, **kwargs):
        pass

    def visit_AbstractDataType(self, node, data, *args, **kwargs):
        pass

    #######################
    #     Expressions     #
    #######################

    def visit_IdentifierTreeNode(self, root_node, data, *args, **kwargs):
        pass

    def visit_BooleanTreeNode(self, root_node, data, *args, **kwargs):
        pass

    def visit_NumberTreeNode(self, root_node, data, *args, **kwargs):
        pass

    def visit_TrajectoryExpressionTreeNode(self, root_node, data, *args, **kwargs):
        #deprecated
        pass

    def visit_BinaryOperatorTreeNode(self, root_node, data, *args, **kwargs):
        data.left = self.visit(root_node.left)
        data.right = self.visit(root_node.right)


    def visit_UnaryOperatorTreeNode(self, root_node, data, *args, **kwargs):
        data.subexpression = self.visit(root_node.subexpression)


    def visit_TrajectoryEquationTreeNode(self, root_node, data, *args, **kwargs):
        data.variable = self.visit(root_node.variable)
        data.trajectory_expression = self.visit(root_node.trajectory_expression)

    def visit_CaseExpressionTreeNode(self, root_node, data, *args, **kwargs):
        data.cases = [(self.visit(cond), self.visit(res)) for (cond, res) in root_node.cases]
        data.default_expression = self.visit(root_node.default_expression)

