# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================
"""This module provides visitors on the CompileInformation-structure
"""

from slim_compiler.src.dataStructure.nominalbehaviour import Category, Enum, ModeTransitionLabel, ProbabilityModeTransitionLabel, ModeType, ErrorEventPort,\
    DataPort, Continuous, Clock, TIME_DELAY_CLK_NAME,\
    ReactivationModeTransitionLabel, IntRange, ModeTransition,\
    ComponentImplementation, Bool
from slim_compiler.src.dataStructure.errorbehaviour import ErrorStateType, DistributionType, ErrorPropagationType,\
    ErrorTransition
from slim_compiler.src.dataStructure.visitors import expressionVisitor
from slim_compiler.src.dataStructure.visitors.baseVisitor import BaseVisitor,\
    TransformVisitor, CompileVisitor
from slim_compiler.src.dataStructure.expressions import TimeUnitType, NodeType,\
    IdentifierTreeNode, BooleanTreeNode, BinaryOperatorTreeNode, OperatorType,\
    UnaryOperatorTreeNode, TrajectoryEquationTreeNode, NumberTreeNode
import slim_compiler.src.dataStructure.expressions as expressions
from slim_compiler.src.errorManagement.errors import ProcessHasNoThreadAndThreadGroupSubcomponentError, EnumElementsNotDistinctError, IntRangeBoundsInvalidError,\
    ObservablePortInFDIRComponentError, AlarmPortInNonFDIRComponentError,\
    ObservableAlarmPortComponentError,\
    EnumValuesNotDistinctAndNotIdenticalInTypeError,\
    EnumValuesDataPortsNotDistinctError,\
    FDIRComponentTypeHasMultipleImplementationsError,\
    FDIRComponentTypeHasNonFDIRImplementationError,\
    ComponentImplementationRefersToUnknownTypeError,\
    ComponentImplementationCrossReferenceToTypeError,\
    ImplementationCategoryMustMatchTypeCategoryError,\
    InDataPortNotTargetInModeMissingDefaultValueError,\
    OutDataPortNotTargetInStartingModeMissingDefaultValueError,\
    OutDataPortNotTargetInModeMissingDefaultValueError,\
    DataPortOfRootComponentHasDefaultValueError,\
    DataPortDependencyRelationCyclicError,\
    DataSubcomponentsDataPortsNotDistinctError,\
    EnumValuesDataSubcomponentsNotDistinctError,\
    EnumValuesNotDistinctAndNotIdenticalInImplError,\
    RecursiveComponentHierarchyError, PortConnectionHasSameSourceAsTargetError,\
    PassiveEventPortConnectionIsNotOutToInError,\
    EventPortConnectionReferencesDataPortError,\
    PortConnectionReferencesUnknownPortError,\
    PortConnectionReferencesUnknownSubcomponentError,\
    PortConnectionReferencesInactiveSubcomponentError,\
    PortConnectionViolatatesPossibleTopologyError,\
    InModesListRefersToUnkonwnModeError, InModesListElementsNotDistinctError,\
    DataPortConnectionReferencesDifferentDataTypesError,\
    DataPortConnectionReferencesEventPortError,\
    DataPortIsTargetOfFlowsConnsAndEffectsError, DataPortHasFanInError,\
    EventPortHasFanOutToSameTargetComponentError,\
    EventPortIsTargetOfPassiveAndActivePortConnections,\
    ComponentImplementationSpecifiesMultipleStartingModeError,\
    ComponentImplementationSpecifiesNoStartingModeError,\
    InvariantTrajectoryEquationRefersToNonContinuousSubcomponentError,\
    InvariantSpecifiesMultipleTrajectoryEquationsForOneVariableError,\
    InitValuesViolateStartingModeInvariant, InvariantHasIllegalTypeError, \
    ModeTransitionRefersToUnknownModeError, TriggerIsInvalidError,\
    TriggerRefersToUnknownSubomponentError, TriggerRefersToDataPortError,\
    TriggerRefersToUnknownPortError, TriggerRefersToInactiveSubomponentError,\
    GuardHasIllegalTypeError, DataComponentSetInMultipleEffectsError,\
    ModeIsNotReachableError, IntRangeValueIllegalError, EnumValueIllegalError,\
    ClocksMustNotBeInitializedError, InitializationValueTypeMismatchError,\
    SubcomponentReferencesUnknownPackageError,\
    SubcomponentReferencesNotUniqueComponentTypeError,\
    SubcomponentReferencesComponentTypeWithoutImplementationsError,\
    SubcomponentReferencesUnknownComponentTypeError,\
    SubcomponentReferencesUnknownImplementationError,\
    SubcomponentReferencesFDIRImplementationError,\
    SubcomponentDeclarationMustMatchImplementationCategoryError,\
    SubcomponentBindingReferencesUnknownSubcomponentError,\
    SubcomponentBindingReferencesWrongCategoryError,\
    SubcomponentBusAccessesItselfError,\
    SubcomponentIsActiveInModeWhereBindingIsNotError,\
    DataSubcomponentMissingDefaultValueError,\
    ErrorTypeSpecifiesMultipleStartingStatesError,\
    ErrorTypeSpecifiesNoStartingStateError,\
    PropagationNameAlreadyUsedAsStateNameError,\
    ErrorImplementationCrossReferenceToTypeError,\
    ErrorImplementationRefersToUnknownTypeError,\
    ErrorStateInImplAlreadyDeclaredInTypeError,\
    ErrorStateTransitionAlreadyDeclaredError, ErrorTriggerIsInvalidError,\
    ErrorStateTransitionRefersToUnknownStateError,\
    ErrorStateNotReachableError, ErrorStatesInBothTypeAndImplementationError,\
    ErrorImplSpecifiesNoStartingStateError,\
    ErrorImplSpecifiesMultipleStartingStatesError, InvariantIsNotConvexError,\
    ClockNotScaledInTimeScaledModel,\
    ErrorStateInvariantForProbabilisticStateError,\
    ErrorEventNotScaledInTimeScaledModel,\
    ErrorTransitionIsProbabilisticButHasGuardError,\
    NonFDIRComponentTypeHasFDIRImplementationError, ModeUrgencyIsNegative,\
    InvalidWithinBoundsError, RateIsNotPositiveError,\
    ErrorStateUrgencyForProbabilisticStateError,\
    ErrorTransitionIsProbabilisticButHasDelayError,\
    ErrorEventAlreadyDeclaredInTypeError, StateUrgencyIsNegative,\
    AlarmPortNotBoolError,\
    ModeTransitionWithinNotScaledInTimeScaledModel,\
    ModeUrgentNotScaledInTimeScaledModel,\
    ErrorStateUrgentNotScaledInTimeScaledModel,\
    ErrorTransitionWithinNotScaledInTimeScaledModel,\
    ErrorTransitionsLeavingStateAreNondeterministicError,\
    ModeNotUrgentWithDelayedTransitionError,\
    StateNotUrgentWithDelayedTransitionError

import itertools
from slim_compiler.src.dataStructure.visitors.expressionVisitor import InvariantToNNFVisitor,\
    InvariantIsConvexVisitor, ExpressionValueVisitor, SimplifyCaseVisitor

class compareVisitor(BaseVisitor):
    """This class provides visit-methods for the CompileInformation-structure.
    The visit-methods return whether the visited nodes are equal (ignoring
    information as line number, position and file).
    """ 
    def visit_OutEventPort(self, origOutEventPort, parsedOutEventPort):
        if not origOutEventPort.name == parsedOutEventPort.name: return False
        if not origOutEventPort.type == parsedOutEventPort.type: return False
        if not origOutEventPort.nonblocking == parsedOutEventPort.nonblocking: return False
        if isinstance(origOutEventPort, ErrorEventPort):
            if not origOutEventPort.distribution == parsedOutEventPort.distribution: return False
            if not origOutEventPort.parameter == parsedOutEventPort.parameter: return False
            if not origOutEventPort.time_unit == parsedOutEventPort.time_unit: return False
        return True
    
    def visit_InEventPort(self, origInEventPort, parsedInEventPort):
        if not (origInEventPort.name == parsedInEventPort.name): return False
        if not (origInEventPort.type == parsedInEventPort.type) : return False
        return True
    
    def visit_InDataPort(self, origInDataPort, parsedInDataPort):
        if not (origInDataPort.name == parsedInDataPort.name) : return False
        if not (origInDataPort.type == parsedInDataPort.type) : return False
        if not (origInDataPort.data_type.type == parsedInDataPort.data_type.type) : return False
        if not (origInDataPort.default_value.type == parsedInDataPort.default_value.type) : return False
        if not (self,origInDataPort.default_value.almostEqual(parsedInDataPort.default_value)) : return False
        return True
    
    def visit_OutDataPort(self, origOutDataPort, parsedOutDataPort):
        if not (origOutDataPort.name == parsedOutDataPort.name) : return False
        if not (origOutDataPort.type == parsedOutDataPort.type) : return False
        if not (origOutDataPort.data_type.type == parsedOutDataPort.data_type.type) : return False
        if not (origOutDataPort.default_value.type == parsedOutDataPort.default_value.type) : return False
        if not (self,origOutDataPort.default_value.almostEqual(parsedOutDataPort.default_value)) : return False
        if not (origOutDataPort.alarm == parsedOutDataPort.alarm) : return False
        if not (origOutDataPort.observable == parsedOutDataPort.observable) : return False
        return True
    
    def visit_ComponentType(self, origType, parsedType):
        if not (origType.category == parsedType.category) : return False
        if not (origType.fdir == parsedType.fdir) : return False
        if not (origType.name == parsedType.name) : return False
        if not (origType.type == parsedType.type) : return False
        return True
        
        # Compare in_event_ports
        if not (len(origType.in_event_ports) == len(parsedType.in_event_ports)) : return False
        keys = origType.in_event_ports.keys()
        for key in keys:
            origInEventPort = origType.in_event_ports[key]
            parsedInEventPort = parsedType.in_event_ports[key]
            if not self.visit(origInEventPort, parsedInEventPort) : return False
            
        # Compare out_event_ports
        if not (len(origType.out_event_ports) == len(parsedType.out_event_ports)) : return False
        keys = origType.out_event_ports.keys()
        for key in keys:
            origOutEventPort = origType.out_event_ports[key]
            parsedOutEventPort = parsedType.out_event_ports[key]
            if not self.visit(origOutEventPort, parsedOutEventPort) : return False
        
        # Compare in_data_ports
        if not (len(origType.in_data_ports) == len(parsedType.in_data_ports)) : return False
        keys = origType.in_data_ports.keys()
        for key in keys:
            origInDataPort = origType.in_data_ports[key]
            parsedInDataPort = parsedType.in_data_ports[key]
            if not self.visit(origInDataPort, parsedInDataPort) : return False
    
        # Compare out_data_ports
        if not (len(origType.out_data_ports) == len(parsedType.out_data_ports)) : return False
        keys = origType.out_data_ports.keys()
        for key in keys:
            origOutDataPort = origType.out_data_ports[key]
            parsedOutDataPort = parsedType.out_data_ports[key]
            if not self.visit(origOutDataPort, parsedOutDataPort) : return False
        return True
    
    def visit_PortConnection(self, origPortConnection, parsedPortConnection):
        #if not (origPortConnection.number == parsedPortConnection.number) : return False
        if not (origPortConnection.passive == parsedPortConnection.passive) : return False
        if not (origPortConnection.source_port == parsedPortConnection.source_port) : return False
        if not (origPortConnection.target_port == parsedPortConnection.target_port) : return False
        if not (origPortConnection.type == parsedPortConnection.type) : return False
        
        # Compare in_modes
        if not (len(origPortConnection.in_modes) == len(parsedPortConnection.in_modes)) : return False
        if not (cmp(origPortConnection.in_modes, parsedPortConnection.in_modes) == 0) : return False
        return True
    
    def visit_Mode(self, origMode, parsedMode):
        if not (origMode.name == parsedMode.name) : return False
        if not (origMode.type == parsedMode.type) : return False
        if origMode.invariant != None:
            if not (origMode.invariant.almostEqual(parsedMode.invariant)) : return False
        return True
    
    def visit_NonDataSubcomponent(self, origNonDataSubcomp, parsedNonDataSubcomp):
        # Note: No need to compare compare implementations here. They they are compared at some other point to avoid recursion.
        if not (origNonDataSubcomp.name == parsedNonDataSubcomp.name) : return False
        if not (origNonDataSubcomp.type == parsedNonDataSubcomp.type) : return False
        if not (origNonDataSubcomp.category == parsedNonDataSubcomp.category) : return False
        if not (origNonDataSubcomp.implementation_ref == parsedNonDataSubcomp.implementation_ref) : return False
    
        # Compare accesses
        if not (cmp(origNonDataSubcomp.accesses, parsedNonDataSubcomp.accesses) == 0) : return False
        
        # Compare in_modes
        if not (cmp(origNonDataSubcomp.in_modes, parsedNonDataSubcomp.in_modes) == 0) : return False
        
        # Compare running_on
        if not (cmp(origNonDataSubcomp.running_on, parsedNonDataSubcomp.running_on) == 0) : return False
        
        # Compare stored_in
        if not (cmp(origNonDataSubcomp.stored_in, parsedNonDataSubcomp.stored_in) == 0) : return False
        return True
    
    def visit_ModeTransitionLabel(self, origLabel, parsedLabel):
        if not (origLabel.trigger == parsedLabel.trigger) : return False
        
        # Compare effects
        if not (len(origLabel.effects) == len(parsedLabel.effects)) : return False
        for i in range(len(origLabel.effects)):
            if not (origLabel.effects[i].almostEqual(parsedLabel.effects[i])) : return False
        
        # Compare guard
        if(origLabel.guard != None):
            if not (origLabel.guard.almostEqual(parsedLabel.guard)) : return False
        return True
    
    def visit_NormalModeTransition(self, origTransition, parsedTransition):
        if not (origTransition.internal_number == parsedTransition.internal_number) : return False
        #if not (origTransition.number == parsedTransition.number) : return False
        if not (origTransition.source_mode == parsedTransition.source_mode) : return False
        if not (origTransition.target_mode == parsedTransition.target_mode) : return False
        if not (origTransition.type == parsedTransition.type) : return False
        
        # Compare label
        if not self.visit(origTransition.label, parsedTransition.label) : return False
        return True
    
    def visit_ProbabilityModeTransition(self, origTransition, parsedTransition):
        # TODO: Implement
        raise Exception("Not implemented")
    
    def visit_AbstractDataType(self, origDataType, parsedDataType):
        if not (origDataType.type == parsedDataType.type) : return False
        if not (type(origDataType) == type(parsedDataType)) : return False
        return True
    
    def visit_DataSubcomponent(self, origDataSubcomp, parsedDataSubcomp):
        # Note: No need to compare compare implementations here. They they are compared at some other point to avoid recursion.
        if not (origDataSubcomp.name == parsedDataSubcomp.name) : return False
        if not (origDataSubcomp.type == parsedDataSubcomp.type) : return False
    
        # Compare data_type
        if not self.visit(origDataSubcomp.data_type, parsedDataSubcomp.data_type) : return False
    
        # Compare default_value
        if not (origDataSubcomp.default_value.almostEqual(parsedDataSubcomp.default_value)) : return False
    
        # Compare in_modes
        if not (cmp(origDataSubcomp.in_modes, parsedDataSubcomp.in_modes) == 0) : return False
        return True
    
    def visit_Flow(self, origFlow, parsedFlow):
        # TODO: Implement
        raise Exception("Not implemented")
    
    def visit_ComponentImplementation(self, origImplementation, parsedImplementation):
        if not (origImplementation.name == parsedImplementation.name) : return False
        if not (origImplementation.type == parsedImplementation.type) : return False
        if not (origImplementation.category == parsedImplementation.category) : return False
        if not (origImplementation.fdir == parsedImplementation.fdir) : return False
        if not (origImplementation.is_error_component_implementation == parsedImplementation.is_error_component_implementation) : return False
        if not (origImplementation.nextFlowID == parsedImplementation.nextFlowID) : return False
        if not (origImplementation.nextModeTransitionID == parsedImplementation.nextModeTransitionID) : return False
        if not (origImplementation.nextPortConnectionID == parsedImplementation.nextPortConnectionID) : return False
        
        # Compare data_port_connections
        if not (len(origImplementation.data_port_connections) == len(parsedImplementation.data_port_connections)) : return False
        for i in range(len(origImplementation.data_port_connections)):
            if not self.visit(origImplementation.data_port_connections[i], parsedImplementation.data_port_connections[i]) : return False
        
        # Compare event_port_connections
        if not (len(origImplementation.event_port_connections) == len(parsedImplementation.event_port_connections)) : return False
        for i in range(len(origImplementation.event_port_connections)):
            if not self.visit(origImplementation.event_port_connections[i], parsedImplementation.event_port_connections[i]) : return False
        
        # Compare data_subcomponents
        if not (len(origImplementation.data_subcomponents) == len(parsedImplementation.data_subcomponents)) : return False
        keys = origImplementation.data_subcomponents.keys()
        for key in keys:
            origDataSubcomp = origImplementation.data_subcomponents[key]
            parsedDataSubcomp = parsedImplementation.data_subcomponents[key]
            if not self.visit(origDataSubcomp, parsedDataSubcomp) : return False
        
        # Compare flows
        if not (len(origImplementation.flows) == len(parsedImplementation.flows)) : return False
        for i in range(len(origImplementation.flows)):
            if not self.visit(origImplementation.flows[i], parsedImplementation.flows[i]) : return False
        
        # Compare modes
        if not (len(origImplementation.modes) == len(parsedImplementation.modes)) : return False
        keys = origImplementation.modes.keys()
        for key in keys:
            origMode = origImplementation.modes[key]
            parsedMode = parsedImplementation.modes[key]
            if not self.visit(origMode, parsedMode) : return False
        
        # Compare non_data_subcomponents
        if not (len(origImplementation.non_data_subcomponents) == len(parsedImplementation.non_data_subcomponents)) : return False
        keys = origImplementation.non_data_subcomponents.keys()
        for key in keys:
            origNonDataSubcomponent = origImplementation.non_data_subcomponents[key]
            parsedNonDataSubcomponent = parsedImplementation.non_data_subcomponents[key]
            if not self.visit(origNonDataSubcomponent, parsedNonDataSubcomponent) : return False
        
        # Compare normal_mode_transitions
        if not (len(origImplementation.normal_mode_transitions) == len(parsedImplementation.normal_mode_transitions)) : return False
        for i in range(len(origImplementation.normal_mode_transitions)):
            if not self.visit_NormalModeTransition(origImplementation.normal_mode_transitions[i], parsedImplementation.normal_mode_transitions[i]) : return False    
        
        # Compare probability_mode_transitions
        if not (len(origImplementation.probability_mode_transitions) == len(parsedImplementation.probability_mode_transitions)) : return False
        for i in range(len(origImplementation.probability_mode_transitions)):
            if not self.visit_ProbabilityModeTransition(origImplementation.probability_mode_transitions[i], parsedImplementation.probability_mode_transitions[i]) : return False
        return True   
    
    def visit_Subpackage(self, origSubpackage, parsedSubpackage):
        # Note: Compare error_implementations, error_types are never contained in extended model and thus never compared
        if not (origSubpackage.package_name == parsedSubpackage.package_name) : return False
        if not (origSubpackage.type == parsedSubpackage.type)  : return False
        
        # Compare types
        if not (len(origSubpackage.types) == len(parsedSubpackage.types)) : return False
        keys = origSubpackage.types.keys()
        for key in keys:
            origType = origSubpackage.types[key]
            parsedType = parsedSubpackage.types[key]
            if not self.visit(origType, parsedType) : return False
            
        # Compare implementations
        if not (len(origSubpackage.implementations) == len(parsedSubpackage.implementations)) : return False
        keys = origSubpackage.implementations.keys()
        for key in keys:
            origImplementation = origSubpackage.implementations[key]
            parsedImplementation = parsedSubpackage.implementations[key]
            if not self.visit(origImplementation, parsedImplementation) : return False
        return True
    
    def visit_Package(self, origPackage, parsedPackage):
        if not (origPackage.name == parsedPackage.name) : return False
        if not (origPackage.type == parsedPackage.type) : return False
        if not self.visit(origPackage.private_subpackage, parsedPackage.private_subpackage) : return False
        if not self.visit(origPackage.public_subpackage, parsedPackage.public_subpackage) : return False
        return True
    
    def visit_CompileInformation(self,ext_cinfo, cinfoFromExtended):
        if not len(ext_cinfo.packages) == len(cinfoFromExtended.packages) : return False
        keys = ext_cinfo.packages.keys()
        for key in keys:
            origPackage = ext_cinfo.packages[key]
            parsedPackage = cinfoFromExtended.packages[key]
            if not self.visit(origPackage, parsedPackage) : return False
        return True
    
class CheckVisitor(BaseVisitor):
    """This class provides visit-methods for the CompileInformation-structure.
    The visit-methods check the visited nodes for consistency with respect to
    the SLIM syntax document.
    """
    def __init__(self, compile_info):
        self.compile_info = compile_info
        self.evaluator = ExpressionValueVisitor(self.compile_info)

    def visit_CompileInformation(self, node):
        for package in node.packages.values():
            self.visit(package)
            
    def visit_Package(self, node):
        self.visit(node.public_subpackage)
        self.visit(node.private_subpackage)
        
    def visit_Subpackage(self, node):
        for type in node.types.values():
            self.visit(type)
        for impl in node.implementations.values():
            self.visit(impl)
        for err_type in node.error_types.values():
            self.visit(err_type)
        for err_impl in node.error_implementations.values():
            self.visit(err_impl)
            
    def visit_ComponentType(self, node):
        for port in node.in_data_ports.values():
            self.visit(port)
        for port in node.out_data_ports.values():
            self.visit(port)
        for port in node.in_event_ports.values():
            self.visit(port)
        for port in node.out_event_ports.values():
            self.visit(port)
        
        if node.fdir:
            self.check_ComponentType_at_most_one_implementation(node)
        self.check_ComponentType_enum_values_and_data_ports_distinct(node)
        self.check_ComponentType_enums_disjoint_or_identical(node)

    def visit_ComponentImplementation(self, node):
        self.check_ComponentImplementation_referenced_component_type_exists(node)
        self.check_ComponentImplementation_referenced_component_type_matches_category(node)
        self.check_ComponentImplementation_referenced_component_type_matches_fdir(node)
                
        self.check_ComponentImplementation_default_values_missing(node)
        self.check_ComponentImplementation_default_values_root_ports(node)
        self.check_ComponentImplementation_port_dependency_relation_is_acyclic(node)
        self.check_ComponentImplementation_data_subcomponents(node)
        
        # check all subcomponent declarations
        for non_data_subcomp in node.non_data_subcomponents.values():
            self.visit(non_data_subcomp)
        for data_subcomp in node.data_subcomponents.values():
            self.visit(data_subcomp)
        
        if node.category == Category.PROCESS:
            # check process has at least one thread or thread_group
            has_thread_or_thread_group = False
            for subcomp in node.non_data_subcomponents.values():
                if subcomp.category == Category.THREAD or subcomp.category == Category.THREAD_GROUP:
                    has_thread_or_thread_group = True
                    
            if not has_thread_or_thread_group:
                node.add_error(ProcessHasNoThreadAndThreadGroupSubcomponentError(node.file, node.line, node.pos, "F-6", node))
            
        self.check_ComponentImplementation_data_identifiers_distinct(node)
        self.check_ComponentImplementation_non_recursiveness(node)

        self.check_ComponentImplementation_connections_and_flows(node)
        self.check_ComponentImplementation_modes(node)
        self.check_ComponentImplementation_mode_transitions(node)
        self.check_ComponentImplementation_each_mode_reachable(node)
 
    def visit_Port(self, node):
        pass
    
    def visit_DataPort(self, node):
        self.visit(node.data_type, node.parent.add_error, node)
        self.check_DataPort_default_value(node)
 
    def visit_OutDataPort(self, node):
        self.visit_DataPort(node)
        if node.observable and node.parent.fdir:
            node.parent.add_error(ObservablePortInFDIRComponentError(node.file, node.line, node.pos, "D-7", node))
        if node.observable and not isinstance(node.data_type, Bool):
            #TODO: SMV does not support this, but currently no official restriction
            #node.parent.add_error(ObservablePortNotBoolError(node.file, node.line, node.pos, "D-9", node))
            pass
        if node.alarm and not node.parent.fdir:
            node.parent.add_error(AlarmPortInNonFDIRComponentError(node.file, node.line, node.pos, "D-8", node))
        if node.alarm and not isinstance(node.data_type, Bool):
            node.parent.add_error(AlarmPortNotBoolError(node.file, node.line, node.pos, "D-8", node))
        if node.alarm and node.observable:
            # Mutex due to the fact that one is for fdir, the other is not
            # Disallowed by compiler
            # No restriction explicitly specified
            #node.parent.add_error(ObservableAlarmPortComponentError(node.file, node.line, node.pos, "D-9", node))
            pass

    def visit_NonDataSubcomponent(self, node):
        self.check_NonDataSubcomponent_implementation_reference(node)
        self.check_NonDataSubcomponent_binding_references(node)
        self.check_NonDataSubcomponent_in_modes_list(node)

    def visit_DataSubcomponent(self, node):
        self.check_DataSubcomponent_default_value(node)
        self.check_DataSubcomponent_in_modes_list(node)  
        self.visit(node.data_type, node.add_error, node) 

    def visit_AbstractDataType(self, node, *args):
        pass

    def visit_Enum(self, node, error_add_method, obj):
        duplicates = node.get_duplicates()
        for elem in duplicates:
            error_add_method(EnumElementsNotDistinctError(node.file, node.line, node.pos, elem, "C-1", obj))
            
    def visit_IntRange(self, node, error_add_method, obj):
        if not node.bounds_are_valid():
            error_add_method(IntRangeBoundsInvalidError(node.file, node.line, node.pos, "C-2", obj))
            
    def visit_ErrorModelType(self, node):
        self.check_ErrorModelType_exactly_one_starting_state(node)
        self.check_ErrorModelType_error_states_and_propagations_distinct(node)
 
    def visit_ErrorModelImplementation(self, node):
        self.check_ErrorModelImplementation_clocks(node);
        self.check_ErrorModelImplementation_states(node);
        self.check_ErrorModelImplementation_events(node)
        self.check_ErrorModelImplementation_referenced_error_type_exists(node)
        self.check_ErrorModelImplementation_states_either_here_or_in_type(node)
        self.check_ErrorModelImplementation_exactly_one_starting_state(node)
        self.check_ErrorModelImplementation_error_transitions(node)
        self.check_ErrorModelImplementation_only_single_type_of_transitions_per_state(node)
        self.check_ErrorModelImplementation_each_state_reachable(node)
        self.check_ErrorModelImplementation_probabilistic_guards(node)
       
    def check_ComponentType_enums_disjoint_or_identical(self, node):
        """Checks whether the list of value identifiers of all enum types are
        either disjoint or identical.
        """
        value_lists_to_check = []
        violating_pairs = []

        for port in node.in_data_ports.values():
            if isinstance(port.data_type, Enum):
                value_lists_to_check.append((port, port.data_type.values))
                
        for port in node.out_data_ports.values():
            if isinstance(port.data_type, Enum):
                value_lists_to_check.append((port, port.data_type.values))
                                
        for i in range(0, len(value_lists_to_check) - 1):
            (port1, list1) = value_lists_to_check[i]
            for j in range(i + 1, len(value_lists_to_check)):
                (port2, list2) = value_lists_to_check[j]
                if not ((set(list1) == set(list2)) or \
                       (set(list1) & set(list2) == set())):
                    if (port1, port2) not in violating_pairs:
                        violating_pairs.append((port1, port2))
        
        error_references = []
        for (entry1, entry2) in violating_pairs:
            if not entry1 in error_references:
                error_references.append(entry1)
            if not entry2 in error_references:
                error_references.append(entry2)
                                    
        if violating_pairs:
            node.add_error(EnumValuesNotDistinctAndNotIdenticalInTypeError(node.file, node.line, node.pos, "D-6", node, violating_pairs, error_references))
    
    def check_ComponentType_enum_values_and_data_ports_distinct(self, node):
        """Checks whether the value identifiers of all enum types are distinct
        from the port identifiers.
        """
        data_port_identifier = set(node.in_data_ports.keys() + node.out_data_ports.keys())
        enum_value_list = set()
        
        for port in node.in_data_ports.values():
            if isinstance(port.data_type, Enum):
                enum_value_list.update(port.data_type.values)
                        
        for port in node.out_data_ports.values():
            if isinstance(port.data_type, Enum):
                enum_value_list.update(port.data_type.values)

        duplicates = data_port_identifier & enum_value_list

        if not duplicates == set():
            error_references = [port for (port, _, _) in [node.get_port(port_name) for port_name in duplicates]]
            for port in node.in_data_ports.values() + node.out_data_ports.values():
                if isinstance(port.data_type, Enum):
                    local_duplicates = duplicates & set(port.data_type.values)
                    if not local_duplicates == set():
                        if port not in error_references:
                            error_references.append(port)
                            
            node.add_error(EnumValuesDataPortsNotDistinctError(node.file, node.line, node.pos, "D-5", node, duplicates, error_references))

    def check_ComponentType_at_most_one_implementation(self, node):
        """Checks whether there is at most one component implementation
        referring to this component type (e.g. required for FDIR component
        types).
        """
        impls = node.parent.parent.get_implementations_for_type(node.name, False)
        if len(impls) > 1:
            node.add_error(FDIRComponentTypeHasMultipleImplementationsError(node.file, node.line, node.pos, "B-4", node, impls))
 
    def check_ComponentImplementation_referenced_component_type_exists(self, node):
        """Checks whether the component type that the component implementation
        refers to exists.
        """
        type = node.get_referenced_component_type()
        if type is None:
            # found no type, check for it in the other subpackage to make 
            # the error message more precise
            if node.parent == node.parent.parent.public_subpackage:
                other_subpackage = node.parent.parent.private_subpackage
            else:
                other_subpackage = node.parent.parent.public_subpackage
            
            type = other_subpackage.get_component_type(node.name.split(".")[0])
            if type is None:
                node.add_error(ComponentImplementationRefersToUnknownTypeError(node.file, node.line, node.pos, "E-2", node))
            else:
                node.add_error(ComponentImplementationCrossReferenceToTypeError(node.file, node.line, node.pos, "A-3", node))
    
    def check_ComponentImplementation_referenced_component_type_matches_category(self, node):
        """Checks whether the component type that this component implementation
        refers to has the same category.
        """
        type = node.get_referenced_component_type()
        if type is not None:
            if type.category != node.category:
                node.add_error(ImplementationCategoryMustMatchTypeCategoryError(node.file, node.line, node.pos, "E-3", node, Category.category_to_str(type.category)))
    
    def check_ComponentImplementation_referenced_component_type_matches_fdir(self, node):
        """Checks whether the implementation belonging to this component type
        is marked as FDIR.
        """
        type = node.get_referenced_component_type()
        if type is not None:
            if type.fdir != node.fdir:
                if type.fdir:
                    node.add_error(FDIRComponentTypeHasNonFDIRImplementationError(node.file, node.line, node.pos, "E-5", type, node))
                else:
                    node.add_error(NonFDIRComponentTypeHasFDIRImplementationError(node.file, node.line, node.pos, "E-5", type, node))
    
    def check_ComponentImplementation_default_values_missing(self, node):
        """Checks whether there are default values missing for some data port or
        data subcomponent.
        """
        self.check_ComponentImplementation_default_values_missing_for_out_data_ports(node)
        self.check_ComponentImplementation_default_values_missing_for_in_data_ports_of_subcomponents(node)
    
    def check_ComponentImplementation_default_values_missing_for_out_data_ports(self, node):
        """Checks whether there are default values missing for some data port.
        """
        # Ticket 527: Do not enforce default values for out data ports of root components
        if node in node.parent.parent.parent.get_root_components():
            return
        type = node.get_referenced_component_type()
        if type is not None:
            for port in type.out_data_ports.values():
                if port.default_value is None:
                    starting_mode_str = "_None"
                    if node.get_starting_mode() is not None:
                        starting_mode_str = node.get_starting_mode().name

                    if not node.port_is_target_of_connection_or_flow(port.name, starting_mode_str):
                        node.add_error(OutDataPortNotTargetInStartingModeMissingDefaultValueError(node.file, port.line, port.pos, "G-6", port, starting_mode_str))
                    
                    seen_modes = set()
                    
                    for trans in node.normal_mode_transitions:
                        if trans.target_mode not in seen_modes:
                            seen_modes.add(trans.target_mode)
                            if node.port_is_target_of_connection_or_flow(port.name, trans.source_mode):
                                if not node.port_is_target_of_connection_or_flow(port.name, trans.target_mode):
                                    is_reset_along_transition = False
                                    for effect in trans.label.effects:
                                        if effect.left.name == port.name:
                                            is_reset_along_transition = True
                                        
                                    if not is_reset_along_transition:
                                        node.add_error(OutDataPortNotTargetInModeMissingDefaultValueError(node.file, port.line, port.pos, "G-6", port, trans.target_mode))
    
    def check_ComponentImplementation_default_values_missing_for_in_data_ports_of_subcomponents(self, node):
        """Checks whether there are default values missing for some incoming
        data ports of a subcomponent.
        """
        for subcomp in node.non_data_subcomponents.values():
            type = subcomp.get_referenced_component_type()
            if type is not None:
                for port in type.in_data_ports.values():
                    if port.default_value is None:
                        if subcomp.in_modes:
                            modes = set(subcomp.in_modes)
                        else:
                            modes = set(node.modes.keys())
                        
                        for conn in node.data_port_connections:
                            if conn.target_port == subcomp.name + "." + port.name:
                                if not conn.in_modes:
                                    modes = set()
                                else:
                                    modes -= set(conn.in_modes)
                                    
                        if modes != set():
                            node.add_error(InDataPortNotTargetInModeMissingDefaultValueError(node.file, subcomp.line, subcomp.pos, "G-5", subcomp, port.name, modes))
    
    def check_ComponentImplementation_default_values_root_ports(self, node):
        """Checks whether there are default values specified for root component
        in data ports
        """
        # See also ticket 527
        if not node in node.parent.parent.parent.get_root_components():
            return
        type = node.get_referenced_component_type()
        if type is not None:
            for port in type.in_data_ports.values():
                if not port.default_value is None:
                    node.add_error(DataPortOfRootComponentHasDefaultValueError(node.file, port.line, port.pos, "G-14", port))
    
    def check_ComponentImplementation_port_dependency_relation_is_acyclic(self, node):
        """Checks whether an overapproximation of the data port dependency
        relation is acyclic.
        """
        starting_ports = []
        for conn in node.data_port_connections:
            if conn.target_port.find(".") != -1:
                [subcomp, port] = conn.target_port.split(".")
                if subcomp in node.non_data_subcomponents:
                    impl = node.non_data_subcomponents[subcomp].get_referenced_component_implementation()
                    if impl is not None:
                        type = impl.get_referenced_component_type()
                        if type is not None:
                            if port in type.in_data_ports:
                                new_entry = (impl, subcomp, port, True) 
                                if new_entry not in starting_ports:
                                    starting_ports.append(new_entry)
                                    
        for flow in node.flows:
            if flow.expression.left.name.find(".") != -1:
                [subcomp, port] = flow.expression.left.name.split(".")
                if subcomp in node.non_data_subcomponents:
                    impl = node.non_data_subcomponents[subcomp].get_referenced_component_implementation()
                    if impl is not None:
                        type = impl.get_referenced_component_type()
                        if type is not None:
                            if port in type.in_data_ports:
                                new_entry = (impl, subcomp, port, True) 
                                if new_entry not in starting_ports:
                                    starting_ports.append(new_entry)

        for starting_port in starting_ports:
            res = self.check_ComponentImplementation_port_dependency_relation_is_acyclic_recursive(node, [starting_port])
            if res != True:
                trace = []
                prefix = []
                last_in_port = False
                
                for elem in res:
                    (impl, subcomp, port, in_port) = elem
                    if in_port:
                        prefix.append(subcomp)
                        
                    new_entry = ""
                    for pre in prefix:
                        new_entry += pre + "."
                    new_entry += port + " (" + impl.name + ")"
                    
                    trace.append(new_entry)
                    
                    if not in_port:
                        prefix.pop()
                
                (_, subcomp_name, port_name, _) = starting_port
                subcomp = node.non_data_subcomponents[subcomp_name]   
                node.add_error(DataPortDependencyRelationCyclicError(node.file, node.line, node.pos, node, trace, "H-5", subcomp))
    
    def check_ComponentImplementation_port_dependency_relation_is_acyclic_recursive(self, node, stack):
        """Checks whether an overapproximation of the data port dependency
        relation is acyclic in a recursive manner.
        """
        (impl, subcomp, port, in_port) = stack[-1]
                
        if in_port:
            for conn in impl.data_port_connections:
                if conn.source_port == port:
                    if conn.target_port.find(".") != -1:
                        [new_subcomp, new_port] = conn.target_port.split(".")
                    
                        if new_subcomp in impl.non_data_subcomponents:
                            new_impl = impl.non_data_subcomponents[new_subcomp].get_referenced_component_implementation()
                            if new_impl is not None:
                                new_entry = (new_impl, new_subcomp, new_port, True)
                                if new_entry not in stack:
                                    res = self.check_ComponentImplementation_port_dependency_relation_is_acyclic_recursive(node, stack + [new_entry])
                                    if res != True: return res
                                else:
                                    level = 0
                                    for i in range(0, len(stack) - 1):
                                        (_, _, _, in_port) = stack[i]
                                        if in_port:
                                            level += 1
                                        else:
                                            level -= 1
                                            
                                    if new_entry == stack[0] and level == 0:
                                        return stack
                            
            for flow in impl.flows:
                if port in expressions.get_all_identifiers(flow.expression.right):
                    if flow.expression.left.name.find(".") == -1:
                        new_entry = (impl, subcomp, flow.expression.left.name, False)
                    else:
                        [new_subcomp, new_port] = flow.expression.left.name.split(".")
                        new_entry = (impl, new_subcomp, new_port, True)
                    if new_entry not in stack:
                        res = self.check_ComponentImplementation_port_dependency_relation_is_acyclic_recursive(node, stack + [new_entry])
                        if res != True: return res
                    else:
                        level = 0
                        for i in range(0, len(stack) - 1):
                            (_, _, _, in_port) = stack[i]
                            if in_port:
                                level += 1
                            else:
                                level -= 1
                                            
                        if new_entry == stack[0] and level == 0:
                            return stack
            
        else:
            impl, in_port = None, None
            backward_in_index = 1
            pos = len(stack) - 2
            
            while (pos >= 0):
                (impl, _, _, in_port) = stack[pos]
                if not in_port:
                    backward_in_index += 1
                else:
                    if backward_in_index == 0:
                        break
                    else:
                        backward_in_index -= 1
                        
                pos -= 1
                                    
            if impl is not None:
                if pos == -1:
                    data_port_connections = node.data_port_connections
                    flows = node.flows
                else:
                    data_port_connections = impl.data_port_connections
                    flows = impl.flows
                    
                for conn in data_port_connections:
                    if conn.source_port == subcomp + "." + port:
                        if conn.target_port.find(".") != -1:
                            [new_subcomp, new_port] = conn.target_port.split(".")
                            new_entry = None
                            if pos == -1:
                                if new_subcomp in node.non_data_subcomponents:
                                    impl = node.non_data_subcomponents[new_subcomp].get_referenced_component_implementation()
                                    new_entry = (impl, new_subcomp, new_port, True)
                            else:
                                new_entry = (impl, new_subcomp, new_port, True)
                                
                            if new_entry is not None:
                                if new_entry not in stack:
                                    res = self.check_ComponentImplementation_port_dependency_relation_is_acyclic_recursive(node, stack + [new_entry])
                                    if res != False: return res
                                else:
                                    level = 0
                                    for i in range(0, len(stack)):
                                        (_, _, _, in_port) = stack[i]
                                        if in_port:
                                            level += 1
                                        else:
                                            level -= 1
                                                                                        
                                if new_entry == stack[0] and level == 0:
                                    return stack
                                
                        else:
                            new_entry = (impl, subcomp, conn.target_port, False)
                            if new_entry not in stack:
                                res = self.check_ComponentImplementation_port_dependency_relation_is_acyclic_recursive(node, stack + [new_entry])
                                if res != False: return res
                            else:
                                level = 0
                                for i in range(0, len(stack) - 1):
                                    (_, _, _, in_port) = stack[i]
                                    if in_port:
                                        level += 1
                                    else:
                                        level -= 1
                                            
                                if new_entry == stack[0] and level == 0:
                                    return stack

                for flow in flows:
                    if subcomp + "." + port in \
                        expressions.get_all_identifiers(flow.expression.right):
                        if flow.expression.left.name.find(".") != -1:
                            [new_subcomp, new_port] = flow.expression.left.name.split(".")
                            new_entry = None
                            if pos == -1:
                                if new_subcomp in node.non_data_subcomponents:
                                    impl = node.non_data_subcomponents[new_subcomp].get_referenced_component_implementation()
                                    new_entry = (impl, new_subcomp, new_port, True)
                            else:
                                new_entry = (impl, new_subcomp, new_port, True)

                            if new_entry is not None:
                                if new_entry not in stack:
                                    res = self.check_ComponentImplementation_port_dependency_relation_is_acyclic_recursive(node, stack + [new_entry])
                                    if res != False: return res
                                else:
                                    level = 0
                                    for i in range(0, len(stack)):
                                        (_, _, _, in_port) = stack[i]
                                        if in_port:
                                            level += 1
                                        else:
                                            level -= 1
                                                                                        
                                if new_entry == stack[0] and level == 0:
                                    return stack
                        else:
                            new_entry = (impl, subcomp, flow.expression.left.name, False)
                            if new_entry not in stack:
                                res = self.check_ComponentImplementation_port_dependency_relation_is_acyclic_recursive(node, stack + [new_entry])
                                if res != False: return res
                            else:
                                level = 0
                                for i in range(0, len(stack) - 1):
                                    (_, _, _, in_port) = stack[i]
                                    if in_port:
                                        level += 1
                                    else:
                                        level -= 1
                                            
                                if new_entry == stack[0] and level == 0:
                                    return stack
        return True
    
    def check_ComponentImplementation_data_identifiers_distinct(self, node):
        """Checks whether the data identifiers of certain components are
        distinct.
        """
        self.check_ComponentImplementation_subcomp_and_ports_distinct(node)
        self.check_ComponentImplementation_enum_values_and_data_subcomponents_distinct(node)
        self.check_ComponentImplementation_enums_disjoint_or_identical(node)
    
    def check_ComponentImplementation_subcomp_and_ports_distinct(self, node):
        """Checks whether subcomponent identifiers and port identifiers are
        distinct.
        """
        subcomp_idents = set(node.data_subcomponents.keys())
        
        type = node.get_referenced_component_type()
        if type is not None:
            data_port_idents = set(type.in_data_ports.keys()) | set(type.out_data_ports.keys())
            duplicates = subcomp_idents & data_port_idents
            objects = []
            for duplicate in duplicates:
                objects.append(node.data_subcomponents[duplicate])
                if duplicate in type.in_data_ports:
                    objects.append(type.in_data_ports[duplicate])
                else:
                    objects.append(type.out_data_ports[duplicate])
            
            if not duplicates == set():
                node.add_error(DataSubcomponentsDataPortsNotDistinctError(node.file, node.line, node.pos, "F-13", node, duplicates, objects))
    
    def check_ComponentImplementation_enum_values_and_data_subcomponents_distinct(self, node):
        """Checks whether enum value identifiers and subcomponent identifiers
        are distinct.
        """
        subcomp_idents = set(node.data_subcomponents.keys())
        enum_value_idents = set()
        
        for subcomp in node.data_subcomponents.values():
            if isinstance(subcomp.data_type, Enum):
                enum_value_idents |= set(subcomp.data_type.values)
                        
        type = node.get_referenced_component_type()
        port_idents = set()
        if type is not None:
            port_idents = set(type.in_data_ports.keys() + type.out_data_ports.keys())
                        
        duplicates = enum_value_idents & (subcomp_idents | port_idents)
        
        if not duplicates == set():
            node.add_error(EnumValuesDataSubcomponentsNotDistinctError(node.file, node.line, node.pos, "F-14", node, duplicates))
    
    def check_ComponentImplementation_enums_disjoint_or_identical(self, node):
        """Checks whether enum value identifiers are disjoint or identical
        for any two data elements of type enum in the implementation.
        """
        type_value_lists = []
        impl_value_lists = []
        type_violating_pairs = []
        impl_violating_pairs = []
        
        type = node.get_referenced_component_type()
        if type is not None:
            for port in type.in_data_ports.values():
                if isinstance(port.data_type, Enum):
                    type_value_lists.append((port, port.data_type.values))
            for port in type.out_data_ports.values():
                if isinstance(port.data_type, Enum):
                    type_value_lists.append((port, port.data_type.values))

        for subcomp in node.data_subcomponents.values():
            if isinstance(subcomp.data_type, Enum):
                impl_value_lists.append((subcomp, subcomp.data_type.values))
                                
        for (port1, list1) in type_value_lists:
            for (subcomp2, list2) in impl_value_lists:
                set1, set2 = set(list1), set(list2)
                intersection = set1 & set2
                if not (intersection == set() or (set1 == set2)):
                    if (port1, subcomp2) not in type_violating_pairs:
                        type_violating_pairs.append((port1, subcomp2))
                        
        for i in range(0, len(impl_value_lists) - 1):
            (subcomp1, list1) = impl_value_lists[i]
            for j in range(i + 1, len(impl_value_lists)):
                (subcomp2, list2) = impl_value_lists[j]
                set1, set2 = set(list1), set(list2)
                intersection = set1 & set2
                if not (intersection == set() or (set1 == set2)):
                    if (subcomp1, subcomp2) not in impl_violating_pairs:
                        impl_violating_pairs.append((subcomp1, subcomp2))
        
        type_error_references = []
        for (entry1, entry2) in type_violating_pairs:
            if entry1 not in type_error_references:
                type_error_references.append(entry1)
            if entry2 not in type_error_references:
                type_error_references.append(entry2)
        
        impl_error_references = []
        for (entry1, entry2) in impl_violating_pairs:
            if entry1 not in impl_error_references:
                impl_error_references.append(entry1)
            if entry2 not in impl_error_references:
                impl_error_references.append(entry2)

        if type_violating_pairs:
            node.add_error(EnumValuesNotDistinctAndNotIdenticalInImplError(node.file, node.line, node.pos, "F-18", node, type_violating_pairs, type_error_references))
        if impl_violating_pairs:
            node.add_error(EnumValuesNotDistinctAndNotIdenticalInImplError(node.file, node.line, node.pos, "F-17", node, impl_violating_pairs, impl_error_references))
    
    def check_ComponentImplementation_non_recursiveness(self, node):
        """Checks the component hierarchy for recursiveness.
        """
        impl_stack = []
        visited_list = []
        
        # queue all subcomponent implementations
        for subcomp_name in node.non_data_subcomponents.keys():
            subcomp = node.non_data_subcomponents[subcomp_name]
            new_impl = subcomp.get_referenced_component_implementation()
            if new_impl is not None:
                impl_stack.append((new_impl, [node.get_fully_qualified_name(), 
                                        new_impl.get_fully_qualified_name()]))
                visited_list.append(new_impl.get_fully_qualified_name())
        
        while len(impl_stack) > 0:
            # dequeue topmost element
            (impl, trace) = impl_stack[0]
            del impl_stack[0]

            full_name = impl.get_fully_qualified_name()
            if full_name == node.get_fully_qualified_name():
                node.add_error(RecursiveComponentHierarchyError(node.file, node.line, node.pos, "F-12", node, trace))                    
            else:
                for subcomp_name in impl.non_data_subcomponents.keys():
                    subcomp = impl.non_data_subcomponents[subcomp_name]
                    new_impl = subcomp.get_referenced_component_implementation()
                    if new_impl is not None:
                        full_name = new_impl.get_fully_qualified_name()
                        if full_name not in visited_list: 
                            new_trace = trace + [full_name]
                            impl_stack.append((new_impl, new_trace))
                            visited_list.append(full_name)
    
    def check_ComponentImplementation_connections_and_flows(self, node):
        """Checks whether connections and flows are legal.
        """
        self.check_ComponentImplementation_port_connections(node)
        self.check_ComponentImplementation_port_flows(node)
        self.check_ComponentImplementation_targets_of_conn_flows_and_effects_mutually_exclusive(node)
        self.check_ComponentImplementation_ports_have_no_fan_in(node)
        self.check_ComponentImplementation_event_port_fan_out(node)
        self.check_ComponentImplementation_event_port_passivity(node)
    
    def check_ComponentImplementation_port_connections(self, node):
        """Checks port connections for their validity.
        """
        self.check_ComponentImplementation_event_port_connections(node)
        self.check_ComponentImplementation_data_port_connections(node)
    
    def check_ComponentImplementation_event_port_connections(self, node):
        """Checks the event port connections for validity.
        """
        for conn in node.event_port_connections:
            s_subcomp_name = None
            s_port_name = None
            t_subcomp_name = None
            t_port_name = None
            source_subcomp = None
            target_subcomp = None
            
            error = False
            if conn.source_port.find(".") != -1: # subcomponent port
                [s_subcomp_name, s_port_name] = conn.source_port.split(".")
                source_subcomp = node.get_non_data_subcomponent(s_subcomp_name)
                if source_subcomp is not None:
                    source_type = source_subcomp.get_referenced_component_type()
                else:
                    node.add_error(PortConnectionReferencesUnknownSubcomponentError(conn.file, conn.line, conn.pos, "G-10", conn, s_subcomp_name))
                    error = True
            else:                                # component port
                s_port_name = conn.source_port
                source_type = node.get_referenced_component_type()
                
            if conn.target_port.find(".") != -1:
                [t_subcomp_name, t_port_name] = conn.target_port.split(".")
                target_subcomp = node.get_non_data_subcomponent(t_subcomp_name)
                if target_subcomp is not None:
                    target_type = target_subcomp.get_referenced_component_type()
                else:
                    node.add_error(PortConnectionReferencesUnknownSubcomponentError(conn.file, conn.line, conn.pos, "G-10", conn, t_subcomp_name))
                    error = True
            else:
                t_port_name = conn.target_port
                target_type = node.get_referenced_component_type()

            if s_subcomp_name:
                self.check_ComponentImplementation_subcomp_active_in_modes(node, s_subcomp_name, conn.in_modes, conn, conn.line, conn.pos)
            if t_subcomp_name:
                self.check_ComponentImplementation_subcomp_active_in_modes(node, t_subcomp_name, conn.in_modes, conn, conn.line, conn.pos)
                
            if not error and source_type and target_type:
                (s_port, s_in_port, s_event_port) = source_type.get_port(s_port_name)
                (t_port, t_in_port, t_event_port) = target_type.get_port(t_port_name)
                if not s_port:
                    if s_subcomp_name:
                        node.add_error(PortConnectionReferencesUnknownPortError(conn.file, conn.line, conn.pos, "G-1", conn, s_subcomp_name + "." + s_port_name))
                    else:
                        node.add_error(PortConnectionReferencesUnknownPortError(conn.file, conn.line, conn.pos, "G-1", conn, s_port_name))
                    error = True 
                else:
                    if not s_event_port:
                        node.add_error(EventPortConnectionReferencesDataPortError(conn.file, conn.line, conn.pos, "G-1", conn, s_port_name))
                        error = True
                if not t_port:
                    if t_subcomp_name:
                        node.add_error(PortConnectionReferencesUnknownPortError(conn.file, conn.line, conn.pos, "G-1", conn, t_subcomp_name + "." + t_port_name))
                    else:
                        node.add_error(PortConnectionReferencesUnknownPortError(conn.file, conn.line, conn.pos, "G-1", conn, t_port_name))                        
                    error = True
                else: 
                    if not t_event_port:
                        node.add_error(EventPortConnectionReferencesDataPortError(conn.file, conn.line, conn.pos, "G-1", conn, t_port_name))
                        error = True
                
                # check that the connection is out-to-in if it is marked as passive
                if conn.passive:
                    if s_in_port or not t_in_port:
                        node.add_error(PassiveEventPortConnectionIsNotOutToInError(conn.file, conn.line, conn.pos, "G-12", conn))
                        error = True        
                
                # check source and target component are different
                if s_subcomp_name == t_subcomp_name:
                    node.add_error(PortConnectionHasSameSourceAsTargetError(conn.file, conn.line, conn.pos, "G-1", conn))

                if not error:
                    self.check_ComponentImplementation_connection_topology_valid(node, source_type, s_in_port,
                                                                                target_type, t_in_port,
                                                                                not s_subcomp_name,
                                                                                not t_subcomp_name,
                                                                                source_subcomp,
                                                                                target_subcomp,
                                                                                node.category, conn,
                                                                                conn.line, conn.pos)
                    
            self.check_ComponentImplementation_in_modes_list(node, conn.in_modes, "G-3", conn, conn.file, conn.line, conn.pos)
    
    def check_ComponentImplementation_subcomp_active_in_modes(self, node, subcomp_name, in_modes, conn, line, pos):
        """Checks whether the given subcomponent is active in the given modes.
                
        Note: This is intended as a subroutine for the checking of port
        connections.
                
        @param    subcomp_name:
                    The name of the subcomponent to check.
        @type     subcomp_name:
                    string
        @param    in_modes:
                    A list of modes to check.
        @type     in_modes:
                    A list of strings.
        @param    conn:
                    The port connection errors should be associated with.
        @type     conn:
                    A PortConnection object.
        @param    line:
                    The line in which the port connection is defined.
        @type     line:
                    int
        @param    pos:
                    The column position at which the port connection is defined.
        @type     pos:
                    int
        """
        if subcomp_name:
            if not in_modes:
                conn_modes = set(node.modes.keys())
            else:
                conn_modes = set(in_modes)
                
            subcomp = node.get_non_data_subcomponent(subcomp_name)
            
            if subcomp:
                if subcomp.in_modes:
                    subcomp_modes = set(subcomp.in_modes)
                else:
                    subcomp_modes = set(node.modes)
                inactive_modes = (conn_modes - subcomp_modes) & set(node.modes.keys())
            
                for mode in inactive_modes:
                    node.add_error(PortConnectionReferencesInactiveSubcomponentError(node.file, line, pos, "G-4", subcomp_name, conn, mode))    
    
    def check_ComponentImplementation_connection_topology_valid(self, node, source_type, s_in_port,
                                        target_type, t_in_port,
                                        source_contained, target_contained,
                                        s_subcomp, t_subcomp,
                                        container_category, conn, line, pos):
        """Checks whether the given port connection topology is valid.
        
        @param    source_type:
                    The component type of the source of the connection.
        @type     source_type:
                    nominalbehaviour.ComponentType
        @param    s_in_port:
                    Indicates whether the source is an incoming port.
        @type     s_in_port:
                    bool
        @param    target_type:
                    The component type of the target of the connection.
        @type     target_type:
                    nominalbehaviour.ComponentType
        @param    t_in_port:
                    Indicates whether the target is an incoming port.
        @type     t_in_port:
                    bool
        @param    source_contained:
                    Indicates whether the source port is contained in the
                    current implementation.
        @type     source_contained:
                    bool
        @param    target_contained:
                    Indicates whether the target port is contained in the
                    current implementation.
        @type     target_contained:
                    bool
        @param    s_subcomp:
                    The subcomponent name of the source port.
        @type     s_subcomp:
                    string
        @param    t_subcomp:
                    The subcomponent name of the target port.
        @type     t_subcomp:
                    string
        @param    container_category:
                    The category of the container of the connection.
        @type     container_category:
                    A constant defined in nominalbehaviour.Category.
        @param    conn:
                    The port connection object to which errors should be
                    associated
        @type     conn:
                    A PortConnection object.
        @param    line:
                    The line in which the port connection was defined.
        @type     line:
                    int
        @param    pos:
                    The column position at which the port connection was
                    defined.
        @type     pos:
                    int
        """
        error = False
        code = ""
        if s_in_port and t_in_port:
            if target_contained:
                error = True
                code = "In to in connection must target subcomponent"
            elif not source_contained:
                error = True
                code = "In to in connection cannot originate from subcomponent"
            else:
                if ((source_type.category == Category.PROCESS or \
                     source_type.category == Category.THREAD_GROUP) and \
                    ((target_type.category == Category.THREAD or \
                     target_type.category == Category.THREAD_GROUP))) or \
                   (source_type.category == Category.SYSTEM):
                    pass
                else:
                    error = True
                    code = "In to in connection not supported between these component categories"
        elif not s_in_port and not t_in_port:
            if not target_contained:
                error = True
                code = "Out to out connection cannot target subcomponent"
            elif source_contained:
                error = True
                code = "Out to out connection must originate from subcomponent"
            else:
                if ((source_type.category == Category.THREAD or \
                     source_type.category == Category.THREAD_GROUP) and \
                    ((target_type.category == Category.PROCESS or \
                     target_type.category == Category.THREAD_GROUP))) or \
                   (target_type.category == Category.SYSTEM):
                    pass
                else:
                    error = True
                    code = "Out to out connection not supported between these component categories"
        elif not s_in_port and t_in_port:
            if source_contained or target_contained:
                error = True
                code = "Out to in connection only supported between subcomponents"
            else:
                if s_subcomp.name == t_subcomp.name:
                    error = True
                    code = "Source and target must differ"
                elif (source_type.category == Category.PROCESSOR or \
                    source_type.category == Category.DEVICE or \
                    source_type.category == Category.SYSTEM) and \
                   (target_type.category == Category.PROCESSOR or \
                    target_type.category == Category.DEVICE or \
                    target_type.category == Category.SYSTEM):
                    #if common bus: okay
                    #otherwise: error
                    if set(s_subcomp.accesses) & set(t_subcomp.accesses) == set():
                        error = True
                        code = "Source and target require access to the same bus"
                elif (source_type.category == Category.PROCESS and \
                      target_type.category == Category.PROCESSOR):
                    if not t_subcomp.name in s_subcomp.running_on:
                        error = True
                        code = "Process not bound to processor"
                elif (source_type.category == Category.PROCESSOR and \
                      target_type.category == Category.PROCESS):
                    if not s_subcomp.name in t_subcomp.running_on:
                        error = True
                        code = "Process not bound to processor"
                elif (container_category == Category.PROCESS or \
                      container_category == Category.THREAD_GROUP):
                    if (source_type.category != Category.THREAD_GROUP and\
                        source_type.category != Category.THREAD) or \
                       (target_type.category != Category.THREAD_GROUP and
                        target_type.category != Category.THREAD):
                        error = True
                        code = "Source and target must be of thread or thread group category"
                else:
                    error = True
                    code = "In to in connection not supported between these component categories"
        else:
            error = True
            code = "In to out connections are not allowed"

        if error:
            node.add_error(PortConnectionViolatatesPossibleTopologyError(node.file, line, pos, "G-9", code, conn))

    def check_ComponentImplementation_in_modes_list(self, node, in_modes, error_identifier, obj, file, line, pos):
        """Checks the given In-Modes list for validity.
        
        @param    in_modes:
                    The In-Modes list to check.
        @type     in_modes:
                    A list of strings.
        @param    error_identifier:
                    The error identifier to add when errors are detected.
        @type     error_identifier:
                    string
        @param    obj:
                    The object to which the error should be assicated.
        @type     obj:
                    An arbitrary object.
        @param    file:
                    The file in which the In-Modes list is defined.
        @type     file:
                    string
        @param    line:
                    The line in which the In-Modes list is defined.
        @type     line:
                    int
        @param    pos:
                    The column position at which the In-Modes list is defined.
        @type     pos:
                    int
        """
        if in_modes:
            duplicates = set([elem for elem in in_modes if in_modes.count(elem) > 1])
            for mode in duplicates:
                node.add_error(InModesListElementsNotDistinctError(file, line, pos, error_identifier, mode, obj))
            
            for mode in in_modes:
                if mode not in node.modes.keys():
                    node.add_error(InModesListRefersToUnkonwnModeError(file, line, pos, error_identifier, mode, obj))
    
    def check_ComponentImplementation_data_port_connections(self, node):
        """Checks the data port connections for validity.
        """
        modes_active = {}
        for conn in node.data_port_connections:
            s_subcomp_name = None
            s_port_name = None
            t_subcomp_name = None
            t_port_name = None
            source_subcomp = None
            target_subcomp = None
            error = False
            if conn.source_port.find(".") != -1: # subcomponent port
                [s_subcomp_name, s_port_name] = conn.source_port.split(".")
                source_subcomp = node.get_non_data_subcomponent(s_subcomp_name)
                if source_subcomp is not None:
                    source_type = source_subcomp.get_referenced_component_type()
                else:
                    node.add_error(PortConnectionReferencesUnknownSubcomponentError(conn.file, conn.line, conn.pos, "G-10", conn, s_subcomp_name))
                    error = True
            else:                                # component port
                s_port_name = conn.source_port
                source_type = node.get_referenced_component_type()
                                
            if conn.target_port.find(".") != -1:
                [t_subcomp_name, t_port_name] = conn.target_port.split(".")
                target_subcomp = node.get_non_data_subcomponent(t_subcomp_name)
                if target_subcomp is not None:
                    target_type = target_subcomp.get_referenced_component_type()
                else:
                    node.add_error(PortConnectionReferencesUnknownSubcomponentError(conn.file, conn.line, conn.pos, "G-10", conn, t_subcomp_name))
                    error = True
            else:
                t_port_name = conn.target_port
                target_type = node.get_referenced_component_type()
                
            if conn.target_port not in modes_active:
                modes_active[conn.target_port] = []

            modes = modes_active[conn.target_port]

            if conn.in_modes:
                modes_to_add = conn.in_modes
            else:
                modes_to_add = node.modes.keys()
            if not modes_to_add:
                modes_to_add = ["__default__"]
            for mode in modes_to_add:
                modes.append(mode)
                
            if s_subcomp_name:
                self.check_ComponentImplementation_subcomp_active_in_modes(node, s_subcomp_name, conn.in_modes, conn, conn.line, conn.pos)
            if t_subcomp_name:
                self.check_ComponentImplementation_subcomp_active_in_modes(node, t_subcomp_name, conn.in_modes, conn, conn.line, conn.pos)
                
            if not error and source_type and target_type:
                (s_port, s_in_port, s_event_port) = source_type.get_port(s_port_name)
                (t_port, t_in_port, t_event_port) = target_type.get_port(t_port_name)
                if not s_port:
                    if s_subcomp_name:
                        node.add_error(PortConnectionReferencesUnknownPortError(conn.file, conn.line, conn.pos, "G-2", conn, s_subcomp_name + "." +  s_port_name))
                    else:
                        node.add_error(PortConnectionReferencesUnknownPortError(conn.file, conn.line, conn.pos, "G-2", conn, s_port_name))                        
                    error = True 
                if not t_port:
                    if t_subcomp_name:
                        node.add_error(PortConnectionReferencesUnknownPortError(conn.file, conn.line, conn.pos, "G-2", conn, t_subcomp_name + "." + t_port_name))
                    else:
                        node.add_error(PortConnectionReferencesUnknownPortError(conn.file, conn.line, conn.pos, "G-2", conn, t_port_name))
                    error = True
                if s_event_port:
                    if s_subcomp_name:
                        node.add_error(DataPortConnectionReferencesEventPortError(conn.file, conn.line, conn.pos, "G-2", conn, s_subcomp_name + "." + s_port_name))
                    else:
                        node.add_error(DataPortConnectionReferencesEventPortError(conn.file, conn.line, conn.pos, "G-2", conn, s_port_name))                        
                    error = True
                if t_event_port:
                    if t_subcomp_name:
                        node.add_error(DataPortConnectionReferencesEventPortError(conn.file, conn.line, conn.pos, "G-2", conn, t_subcomp_name + "." + t_port_name))
                    else:
                        node.add_error(DataPortConnectionReferencesEventPortError(conn.file, conn.line, conn.pos, "G-2", conn, t_port_name))                        
                    error = True
                if s_port and t_port and not s_event_port and not t_event_port:
                    if s_port.data_type.get_name() != t_port.data_type.get_name():
                        node.add_error(DataPortConnectionReferencesDifferentDataTypesError(conn.file, conn.line, conn.pos, "G-2", conn))
                    elif s_port.data_type.get_name() == "enum":
                        s_enum_values = s_port.data_type.values
                        t_enum_values = t_port.data_type.values

                        if not set(s_enum_values) == set(t_enum_values):
                            node.add_error(DataPortConnectionReferencesDifferentDataTypesError(conn.file, conn.line, conn.pos, "G-2", conn))
                
                if not error:            
                    self.check_ComponentImplementation_connection_topology_valid(node, source_type, s_in_port,
                                                                                 target_type, t_in_port,
                                                                                 not s_subcomp_name,
                                                                                 not t_subcomp_name,
                                                                                 source_subcomp,
                                                                                 target_subcomp,
                                                                                 node.category, conn,
                                                                                 conn.line, conn.pos)

            self.check_ComponentImplementation_in_modes_list(node, conn.in_modes, "G-3", conn, conn.file, conn.line, conn.pos) 
    
    def check_ComponentImplementation_port_flows(self, node):
        """Checks the flows defined for ports.
        """
        if node.flows:
            for flow in node.flows:
                self.check_ComponentImplementation_in_modes_list(node, flow.in_modes, "H-6", flow, node.file, flow.line, flow.pos)
    
    def check_ComponentImplementation_targets_of_conn_flows_and_effects_mutually_exclusive(self, node):
        # first, check own ports
        type = node.get_referenced_component_type()
        if type is not None:
            data_ports = set(type.in_data_ports.values() + type.out_data_ports.values())                
            for port in data_ports:
                if node.port_is_target_of_connection_or_flow(port.name) and node.port_is_target_of_effect(port.name):
                    node.add_error(DataPortIsTargetOfFlowsConnsAndEffectsError(port.file, port.line, port.pos, "G-7", port,
                                                                               list(node.port_is_target_of_connection_or_flow(port.name)),
                                                                               list(node.port_is_target_of_effect(port.name))))
    
    def check_ComponentImplementation_ports_have_no_fan_in(self, node):
        """Checks whether there are data ports that have a fan-in.
        """
        type = node.get_referenced_component_type()
        if type is not None:
            ports = type.out_data_ports.keys()
            port_targets = {}
            for mode in node.modes.keys():
                for conn in node.data_port_connections:
                    if conn.target_port in ports:
                        if not conn.in_modes or mode in conn.in_modes:
                            if (mode, conn.target_port) in port_targets:
                                node.add_error(DataPortHasFanInError(node.file, conn.line, conn.pos, "G-8", conn, conn.target_port, mode))
                            else:
                                port_targets[(mode, conn.target_port)] = True
            
                for flow in node.flows:
                    if flow.expression.left.name in ports:
                        if not flow.in_modes or mode in flow.in_modes:
                            if (mode, flow.expression.left.name) in port_targets:
                                node.add_error(DataPortHasFanInError(node.file, flow.line, flow.pos, "G-8", flow, flow.expression.left.name, mode))
                            else:
                                port_targets[(mode, flow.expression.left.name)] = True
    
    def check_ComponentImplementation_event_port_fan_out(self, node):
        """Checks whether there are event port connections declaring a fan-out
        to the same target component.
        """
        source_subcomp_to_ports_map = {}
        source_subcomp_to_conns_map = {}
        for conn in node.event_port_connections:
            (subcomp, _, port) = conn.target_port.partition(".")
            if port is None or port == "":
                port = subcomp
                subcomp = None
            
            if not (conn.source_port, subcomp) in source_subcomp_to_ports_map:
                source_subcomp_to_ports_map[(conn.source_port, subcomp)] = [port]
                source_subcomp_to_conns_map[(conn.source_port, subcomp)] = [conn]
            else:
                other_ports = source_subcomp_to_ports_map[(conn.source_port, subcomp)] 
                if port not in other_ports:
                    source_subcomp_to_conns_map[(conn.source_port, subcomp)].append(conn)
                    source_subcomp_to_ports_map[(conn.source_port, subcomp)].append(port)
                    
        for (key, conns) in source_subcomp_to_conns_map.iteritems():
            if len(conns) > 1:
                # Check if connections are mutex wrt modes
                target_port_names = source_subcomp_to_ports_map[key]
                (source_port_name, subcomp_name) = key
                modes_sets = [conn.in_modes for conn in conns]
                # Generate combinations of connection mode lists
                combinations = itertools.combinations(modes_sets, 2)
                for (set1, set2) in combinations:
                    # If either list is empty (all modes) or intersection != 0, error G-11
                    if len(set1) == 0 or len(set2) == 0 or len(set.intersection(set(set1), set(set2))) > 0:
                        node.add_error(
                                EventPortHasFanOutToSameTargetComponentError(
                                conns[1].file, conns[1].line, conns[1].pos, "G-11",
                                conns, source_port_name, subcomp_name,
                                target_port_names))
                        break #Only report once for all targets
    
    def check_ComponentImplementation_event_port_passivity(self, node):
        """Checks that there does not exist passive as well as non-passive
        port connections targeting a single port."""
        active_connection_target_ports_to_connection_map = {}
        passive_connection_target_ports_to_connection_map = {}
        for conn in node.event_port_connections:
            if conn.passive:
                if conn.target_port in active_connection_target_ports_to_connection_map:
                    node.add_error(
                            EventPortIsTargetOfPassiveAndActivePortConnections(
                            conn.file, conn.line, conn.pos, "G-13",
                            active_connection_target_ports_to_connection_map[conn.target_port], conn))
                else:
                    passive_connection_target_ports_to_connection_map[conn.target_port] = conn
            else:
                if conn.target_port in passive_connection_target_ports_to_connection_map:
                    node.add_error(
                            EventPortIsTargetOfPassiveAndActivePortConnections(
                            conn.file, conn.line, conn.pos, "G-13",
                            passive_connection_target_ports_to_connection_map[conn.target_port], conn))
                else:
                    active_connection_target_ports_to_connection_map[conn.target_port] = conn
    
    def check_ComponentImplementation_modes(self, node):
        if node.modes:
            self.check_ComponentImplementation_category_supports_modes(node)
            self.check_ComponentImplementation_exactly_one_starting_mode(node)
            self.check_ComponentImplementation_mode_invariants(node)
            self.check_ComponentImplementation_mode_urgency(node)
    
    def check_ComponentImplementation_category_supports_modes(self, node):
        """Checks whether the category of the current implementation supports
        the definition of modes.
        """
        # all categories allowed
        pass
    
    def check_ComponentImplementation_exactly_one_starting_mode(self, node):
        """Checks whether there is exactly one starting mode.
        """
        starting_modes_found = 0
        for mode in node.modes.values():
            if mode.type == ModeType.INITIAL or \
               mode.type == ModeType.ACTIVATION:
                starting_modes_found += 1
        if starting_modes_found == 0:
            node.add_error(
                    ComponentImplementationSpecifiesNoStartingModeError(
                    node.file, node.line, node.pos, "I-3", node))
        elif starting_modes_found > 1:
            node.add_error(
                    ComponentImplementationSpecifiesMultipleStartingModeError(
                    node.file, node.line, node.pos, "I-3", node))

    def check_ComponentImplementation_mode_invariants(self, node):
        """Checks whether the mode invariants are legal. 
        """
        # Note: part of invariant checks also handled by expression checker
        for mode in node.modes.values():
            if mode.invariant is not None:
                res = mode.invariant.node_type
                if res == NodeType.ERROR:
                    continue

                if res != NodeType.BOOL and res != NodeType.TRAJECTORY_EQUATION:
                    node.add_error(
                                InvariantHasIllegalTypeError(
                                        node.file, mode.line, mode.pos, "I-7",
                                        NodeType.type_to_str(res)))
                else:
                    # Make sure the invariant is convex
                    simple_invariant = SimplifyCaseVisitor().replace_case(mode.invariant.clone())
                    nnf_invariant = InvariantToNNFVisitor().visit(simple_invariant)
                    is_convex = InvariantIsConvexVisitor().visit(nnf_invariant)
                    if not is_convex:
                        node.add_error(InvariantIsNotConvexError(node.file, mode.line, mode.pos, "I-19", mode.invariant, mode))
                    if mode.type == ModeType.INITIAL or mode.type == ModeType.ACTIVATION:
                        sat = self.evaluator.visit(mode.invariant)
                        if sat is None or not sat:
                            node.add_error(
                                InitValuesViolateStartingModeInvariant(
                                        node.file, mode.line, mode.pos, "I-9",
                                        mode))

                traj_eq_vars = node.get_all_trajectory_equation_variables(mode.invariant)
                duplicates = set([var for var in traj_eq_vars if traj_eq_vars.count(var) > 1])
                for elem in duplicates:
                    node.add_error(
                            InvariantSpecifiesMultipleTrajectoryEquationsForOneVariableError(
                                    node.file, mode.line, mode.pos, "I-8", elem, mode))

    def check_ComponentImplementation_mode_urgency(self, node):
        for mode in node.modes.values():
            if mode.urgent is not None and mode.urgent.node_type != NodeType.ERROR:
                if self.compile_info.uses_time_scaling and mode.urgent.node_type != NodeType.CLOCK:
                    node.add_error(ModeUrgentNotScaledInTimeScaledModel(mode.urgent.file, mode.urgent.line, mode.urgent.pos, "I-25", mode))
                value = self.evaluator.visit(mode.urgent)
                if value is None or value < 0.0:
                    node.add_error(ModeUrgencyIsNegative(mode.urgent.file, mode.urgent.line, mode.urgent.pos, "I-20", mode.urgent, mode))
            else:
                for trans in node.normal_mode_transitions:
                    if trans.source_mode == mode.name and trans.label.time_delay is not None:
                        node.add_error(ModeNotUrgentWithDelayedTransitionError(mode.file, mode.line, mode.pos, "I-27", mode))
                        break

    def check_ComponentImplementation_mode_transitions(self, node):
        """Checks whether the mode transitions are legal.
        """
        self.check_ComponentImplementation_normal_mode_transitions(node)
        self.check_ComponentImplementation_probability_mode_transitions(node)
    
    def check_ComponentImplementation_normal_mode_transitions(self, node):
        """Checks whether the non-probability mode transitions are legal.
        """
        for trans in node.normal_mode_transitions + node.reactivation_mode_transitions:
            comps_seen = [] # components seen in left sides of effects
            s_mode = node.get_mode(trans.source_mode)
            t_mode = node.get_mode(trans.target_mode)
            if not s_mode:
                node.add_error(
                        ModeTransitionRefersToUnknownModeError(trans.file,
                                                               trans.line,
                                                               trans.pos, "I-10",
                                                               trans,
                                                               trans.source_mode))
            if not t_mode:
                node.add_error(
                        ModeTransitionRefersToUnknownModeError(trans.file,
                                                               trans.line,
                                                               trans.pos, "I-10",
                                                               trans,
                                                               trans.target_mode))
            if isinstance(trans.label, ModeTransitionLabel):
                self.check_ComponentImplementation_valid_trigger(node, trans)
                self.check_ComponentImplementation_valid_guard(node, trans)
                self.check_ComponentImplementation_valid_within(node, trans)
            self.check_ComponentImplementation_valid_effects(node, trans)
    
    def check_ComponentImplementation_valid_trigger(self, node, trans):
        """Checks whether the given trigger is valid.
        
        @param    trans:
                    The transition whose trigger to check.
        @type     trans:
                    ModeTransition object
        """
        label = trans.label
        trigger = label.trigger
        
        trigger_valid = False
        if trigger is None:
            trigger_valid = True
        elif trigger == "reset":
            trigger_valid = True
        else:                    
            # check if trigger refers to an event port of a subcomponent
            if trigger.find(".") != -1:
                [subcomp_name, port_name] = trigger.split(".")
                subcomp = node.get_non_data_subcomponent(subcomp_name)
                if subcomp is not None:
                    if subcomp.in_modes and not trans.source_mode in subcomp.in_modes:
                        node.add_error(TriggerRefersToInactiveSubomponentError(trans.file, trans.line, trans.pos, "I-11", subcomp_name, trans))
                    impl = subcomp.get_referenced_component_implementation()
                    if impl is not None:
                        type = impl.get_referenced_component_type()
                        if type is not None:
                            (port, in_port, event_port) = type.get_port(port_name)
                            if port is not None and event_port:
                                trigger_valid = True
                            elif port is None:
                                node.add_error(TriggerRefersToUnknownPortError(trans.file, trans.line, trans.pos, "I-11", subcomp_name, port_name, trans))
                            elif not event_port:
                                node.add_error(TriggerRefersToDataPortError(trans.file, trans.line, trans.pos, "I-11", subcomp_name, port_name, trans))
                else:
                    node.add_error(TriggerRefersToUnknownSubomponentError(trans.file, trans.line, trans.pos, "I-11", subcomp_name, trans))
            else:
                # check if trigger is an incoming or outgoing event port of the
                # current component
                type = node.get_referenced_component_type()
                if type is not None:
                    (port, in_port, event_port) = type.get_port(trigger)
                    if port is not None and event_port:
                        trigger_valid = True
                        
        if not trigger_valid:
            node.add_error(TriggerIsInvalidError(trans.file, trans.line, trans.pos,"I-11", trigger, trans))
    
    def check_ComponentImplementation_valid_guard(self, node, trans):
        """Checks whether the given guard is legal in a transition leaving
        the given mode.
        
        @param    trans:
                    The transition whose guard to check.
        @type     guard:
                    ModeTransition object
        """
        label = trans.label
        guard = label.guard

        if guard is not None:
            guard_type = guard.node_type

            if guard_type != NodeType.BOOL and guard_type != NodeType.ERROR:
                node.add_error(
                        GuardHasIllegalTypeError(trans.file, trans.line,
                                                 trans.pos, "I-12", NodeType.type_to_str(guard_type), trans))

    def check_ComponentImplementation_valid_within(self, node, trans):
        if trans.label.time_delay is not None:
            (lower, upper) = trans.label.time_delay
            if not (lower.node_type == NodeType.ERROR or upper.node_type == NodeType.ERROR):
                (lval, uval) = self.evaluator.visit(lower), self.evaluator.visit(upper)
                if self.compile_info.uses_time_scaling and (lower.node_type != NodeType.CLOCK or upper.node_type != NodeType.CLOCK):
                    node.add_error(ModeTransitionWithinNotScaledInTimeScaledModel(trans.file, trans.line, trans.pos, "I-26", trans))
                elif lval is None or uval is None or lval > uval or lval < 0.0:
                    node.add_error(InvalidWithinBoundsError(trans.file, trans.line, trans.pos, "I-21", lval, uval, trans))

    def check_ComponentImplementation_probability_mode_transitions(self, node):
        """Checks whether the probability mode transitions are legal.
        """
        pass
    
    def check_ComponentImplementation_valid_effects(self, node, trans):
        """Checks whether the given effects are legal in a transition between
        the two given states.
        
        @param    trans:
                    The transition whose effects to check.
        @type     trans:
                    ModeTransition object
        """
        label = trans.label
        effects = label.effects
        comps_set = []
        if effects:
            for effect in effects:
                comps_set.append(effect.left.name)
            # check whether elements were set multiple times
            duplicates = set([element for element in comps_set if comps_set.count(element) > 1])
            for element in duplicates:
                node.add_error(DataComponentSetInMultipleEffectsError(trans.file, trans.line, trans.pos, "I-13", element, trans))
    
    def check_ComponentImplementation_each_mode_reachable(self, node):
        """Checks whether each mode that was declared is reachable.
        """
        if node.modes.keys():
            reachable_modes = []
            start = node.get_starting_mode()
            if start is not None:
                reachable_modes.append(start.name)
                changed = True
                while changed:
                    changed = False
                    for mode in reachable_modes:
                        for trans in node.normal_mode_transitions + node.reactivation_mode_transitions:
                            if trans.source_mode == mode:
                                if trans.target_mode not in reachable_modes:
                                    reachable_modes.append(trans.target_mode)
                                    changed = True
                                
                for mode_name in node.modes.keys():
                    if mode_name not in reachable_modes:
                        mode = node.modes[mode_name]
                        if mode.type != ModeType.INITIAL and mode.type != ModeType.ACTIVATION:
                            node.add_error(ModeIsNotReachableError(mode.file, mode.line, mode.pos, "I-4", mode))

    def check_ComponentImplementation_data_subcomponents(self, node):
        for data_subcomp in node.data_subcomponents.values():
            if isinstance(data_subcomp.data_type, Clock):
                clock = data_subcomp
                if self.compile_info.uses_time_scaling and clock.data_type.time_unit == None:
                    self.compile_info.add_error(ClockNotScaledInTimeScaledModel(clock.file, clock.line, clock.pos, "F-19", clock))

    def check_DataPort_default_value(self, node):
        """Checks whether a supplied default value matches the data type of
        the port.
        """
        
        # Note: Enum values already checked by annotationvisitor
        if node.default_value and node.default_value.node_type != NodeType.ERROR:
            # use the assignment operator table to check if compatible
            table = expressions.get_result_table_for_operator(OperatorType.ASSIGN, self.compile_info.uses_time_scaling)
            if not (node.data_type.get_node_type(), node.default_value.node_type) in table:
                self.compile_info.add_error(InitializationValueTypeMismatchError(node.file,
                                                                                 node.line,
                                                                                 node.pos,
                                                                                 "D-4", node.data_type.get_name(), node))
                return

            if isinstance(node.data_type, IntRange):
                if not node.data_type.value_is_valid(
                                                     node.default_value.get_value()):
                    self.compile_info.add_error(IntRangeValueIllegalError(node.file,
                                                                          node.line,
                                                                          node.pos,
                                                                          "D-4", node.default_value.get_value(), node))
            elif isinstance(node.data_type, Clock):
                # Note: This ought not to happen, as clock data ports are not supported
                self.compile_info.add_error(ClocksMustNotBeInitializedError(node.file,
                                                                            node.line,
                                                                            node.pos,
                                                                            "F-5", node))

    def check_NonDataSubcomponent_implementation_reference(self, node):
        """Checks the implementation reference for validity.
        """
        error = False
        package_name = None
        type_name = None
        implementation_name = None
        impl = None
        
        (package_name, type_name, implementation_name) = node.parse_implementation_reference()
                
        # find the right package
        if package_name:
            ci = node.parent.parent.parent.parent
            package = ci.get_package(package_name)
            if package is None:
                node.add_error(SubcomponentReferencesUnknownPackageError(node.file, node.line, node.pos, "F-2", node, package_name))
                error = True
        else:
            package = node.parent.parent.parent

        if not error:
            # only reference to type given
            if not implementation_name:
                if package_name:
                    type = package.get_component_type(type_name, True)
                else:
                    type = package.get_component_type(type_name, False)
                    
                if type is not None:
                    # find all implementations for that type
                    impls = package.get_implementations_for_type(type_name, False)
                
                    if len(impls) > 1:
                        node.add_error(SubcomponentReferencesNotUniqueComponentTypeError(node.file, node.line, node.pos, "F-2", node, type_name))
                    elif len(impls) == 0:
                        node.add_error(SubcomponentReferencesComponentTypeWithoutImplementationsError(node.file, node.line, node.pos, "F-2", node, type_name))
                    elif len(impls) == 1:
                        impl = impls[0]
                else:
                    node.add_error(SubcomponentReferencesUnknownComponentTypeError(node.file, node.line, node.pos, "F-2", node, type_name))
                    error = True
                    
            # implementation directly referenced
            else:
                if package_name:
                    impl = package.get_component_implementation(
                            type_name + "." + implementation_name, True)
                else:
                    impl = package.get_component_implementation(
                            type_name + "." + implementation_name, False)
                    
                if not impl:
                    node.add_error(SubcomponentReferencesUnknownImplementationError(node.file, node.line, node.pos, "F-2", node, type_name + "." + implementation_name))
                    error = True
                elif impl.fdir:
                    node.add_error(SubcomponentReferencesFDIRImplementationError(node.file, node.line, node.pos, "F-2", node))
        
        if not error and impl:
            if impl.category != node.category:
                node.add_error(SubcomponentDeclarationMustMatchImplementationCategoryError(node.file, node.line, node.pos, "F-3", node, impl, Category.category_to_str(impl.category)))
                error = True

    def check_NonDataSubcomponent_binding_references(self, node):
        """Checks whether the binding references specified are valid.
        """
        if node.accesses:
            for bus in node.accesses:
                subcomp = node.parent.get_non_data_subcomponent(bus)
                if subcomp is None:
                    node.add_error(
                            SubcomponentBindingReferencesUnknownSubcomponentError(
                            node.file, node.line, node.pos, "F-9", node, bus))
                else:
                    if subcomp.category != Category.BUS:
                        node.add_error(
                                SubcomponentBindingReferencesWrongCategoryError(
                                node.file, node.line, node.pos, "F-9", node,
                                bus, Category.category_to_str(subcomp.category),
                                "bus", "accesses"))
                    elif subcomp.name == node.name:
                        node.add_error(SubcomponentBusAccessesItselfError(
                                       node.file, node.line, node.pos, "F-9",
                                       node))
                    self.check_NonDataSubcomponent_binding_active_in_all_modes(node, subcomp)
                        
        if node.stored_in:
            for memory in node.stored_in:
                subcomp = node.parent.get_non_data_subcomponent(memory)
                if subcomp is None:
                    node.add_error(
                            SubcomponentBindingReferencesUnknownSubcomponentError(
                            node.file, node.line, node.pos, "F-7", node,memory))
                else:
                    if subcomp.category != Category.MEMORY:
                        node.add_error(
                                SubcomponentBindingReferencesWrongCategoryError(
                                node.file, node.line, node.pos, "F-7",
                                node, memory,
                                Category.category_to_str(subcomp.category),
                                "memory", "stored in"))
                    self.check_NonDataSubcomponent_binding_active_in_all_modes(node, subcomp)
                
        if node.running_on:
            for processor in node.running_on:
                subcomp = node.parent.get_non_data_subcomponent(processor)
                if subcomp is None:
                    node.add_error(
                            SubcomponentBindingReferencesUnknownSubcomponentError(
                            node.file, node.line, node.pos, "F-8", node,
                            processor))
                else:
                    if subcomp.category != Category.PROCESSOR:
                        node.add_error(
                                SubcomponentBindingReferencesWrongCategoryError(
                                node.file, node.line, node.pos, "F-8", 
                                node, processor,
                                Category.category_to_str(subcomp.category),
                                "processor", "running on"))
                    self.check_NonDataSubcomponent_binding_active_in_all_modes(node, subcomp)

    def check_NonDataSubcomponent_binding_active_in_all_modes(self, node, binding_partner):
        """Checks whether the given binding partner is active in all the modes
        in which the subcomponent is active.
        """
        if not node.in_modes:
            my_modes = set(node.parent.modes.keys())
        else:
            my_modes = set(node.in_modes)
            
        my_modes &= set(node.parent.modes.keys())
                        
        if not binding_partner.in_modes:
            other_modes = set(binding_partner.parent.modes.keys())
        else:
            other_modes = set(binding_partner.in_modes)
                        
        for mode in my_modes:
            if not mode in other_modes:
                node.add_error(SubcomponentIsActiveInModeWhereBindingIsNotError(
                        node.file, node.line, node.pos, "F-10", node,
                        binding_partner.name, mode))

    def check_NonDataSubcomponent_in_modes_list(self, node):
        """Checks the In-Modes list.
        """
        if node.in_modes is not None:
            duplicates = set([elem for elem in node.in_modes if node.in_modes.count(elem) > 1])
            
            for elem in duplicates:
                node.add_error(InModesListElementsNotDistinctError(node.file, node.line, node.pos, "F-11", elem, node))
            
            for mode in node.in_modes:
                if not mode in node.parent.modes.keys():
                    node.add_error(InModesListRefersToUnkonwnModeError(node.file, node.line, node.pos, "F-11", mode, node))
   
    def check_DataSubcomponent_default_value(self, node):
        """Checks the default value for validity.
        """
        # Note: Enum values already checked by annotationvisitor
        if node.default_value and node.default_value.node_type != NodeType.ERROR:
            # use the assignment operator table to check if compatible
            table = expressions.get_result_table_for_operator(OperatorType.ASSIGN, self.compile_info.uses_time_scaling)
            if not (node.data_type.get_node_type(), node.default_value.node_type) in table:
                self.compile_info.add_error(InitializationValueTypeMismatchError(node.file,
                                                                                 node.line,
                                                                                 node.pos,
                                                                                 "F-4", node.data_type.get_name(), node))
                return

            if isinstance(node.data_type, IntRange):
                if not node.data_type.value_is_valid(
                                                     node.default_value.get_value()):
                    self.compile_info.add_error(IntRangeValueIllegalError(node.file,
                                                                          node.line,
                                                                          node.pos,
                                                                          "F-4", node.default_value.get_value(), node))
            elif isinstance(node.data_type, Clock):
                self.compile_info.add_error(ClocksMustNotBeInitializedError(node.file,
                                                                            node.line,
                                                                            node.pos,
                                                                            "F-5", node))
        elif not node.default_value:
            if not isinstance(node.data_type, Clock):
                self.compile_info.add_error(DataSubcomponentMissingDefaultValueError(node.file, node.line, node.pos, "I-16", node))

    def check_DataSubcomponent_in_modes_list(self, node):
        """Checks the In-Modes list for validity.
        """
        if node.in_modes:
            duplicates = set([elem for elem in node.in_modes if node.in_modes.count(elem) > 1])            
            for elem in duplicates:
                node.add_error(InModesListElementsNotDistinctError(node.file, node.line, node.pos, "F-11", elem, node))
            for mode in node.in_modes:
                if not mode in node.parent.modes.keys():
                    node.add_error(InModesListRefersToUnkonwnModeError(node.file, node.line, node.pos, "F-11", mode, node))
    
    def check_ErrorModelType_exactly_one_starting_state(self, node):
        """Checks whether there exists exactly one starting state in the error
        model type. The set of states of an error type may be empty though if
        error implementations contain states.
        """
        startingStates = node.get_all_starting_states()
        impls = node.get_referenced_error_implementations()
        
        some_impl_has_states = False
        for impl in impls:
            if len(impl.states)>0:
                some_impl_has_states = True
                break
        
        if len(node.states) == 0 and some_impl_has_states:
            # It's fine. However the implementations will be checked
            # at some later point - they must contain starting states.
            pass
        else:
            if len(startingStates) < 1:
                node.add_error(ErrorTypeSpecifiesNoStartingStateError(node.file, node.line, node.pos, "J-3", node))
            elif len(startingStates) > 1:
                node.add_error(ErrorTypeSpecifiesMultipleStartingStatesError(node.file, node.line, node.pos, "J-3", node))
    
    def check_ErrorModelType_error_states_and_propagations_distinct(self, node):
        """Checks whether all error state names and the names of all error
        propagations are distinct.
        """
        states = set(node.states.keys())
        props = set(node.error_propagations.keys())
        duplicates = states & props
        for identifier in duplicates:
            prop = node.error_propagations[identifier]
            node.add_error(PropagationNameAlreadyUsedAsStateNameError(prop.file, prop.line, prop.pos, "J-2", prop))

    def check_ErrorModelImplementation_exactly_one_starting_state(self, node):
        """Checks whether there exists exactly one starting state in the error
        implementation type. The set of states of an error implementation may
        be empty though if the error type contains states.
        """
        startingStates = node.get_all_starting_states()
        type = node.get_referenced_error_type()
        if type is not None:
            type_has_states = len(type.states)>0 
            if len(node.states) == 0 and type_has_states:
                # It's fine. However the state must contain starting states.
                # The error type is checked separately.
                pass
            else:
                if len(startingStates) < 1:
                    node.add_error(ErrorImplSpecifiesNoStartingStateError(node.file, node.line, node.pos, "K-16", node))
                elif len(startingStates) > 1:
                    node.add_error(ErrorImplSpecifiesMultipleStartingStatesError(node.file, node.line, node.pos, "K-16", node))

    def check_ErrorModelImplementation_clocks(self, node):
        for clock in node.clocks.itervalues():
            if self.compile_info.uses_time_scaling and clock.data_type.time_unit == None:
                self.compile_info.add_error(ClockNotScaledInTimeScaledModel(node.file, node.line, node.pos, "K-34", clock)) 

    def check_ErrorModelImplementation_states(self, node):
        """Checks whether the error states have valid invariants"""
        for state in node.states.itervalues():
            if state.invariant is not None and state.invariant.node_type != NodeType.ERROR:
                if state.invariant.node_type != NodeType.BOOL:
                    node.add_error(InvariantHasIllegalTypeError(node.file, state.line, state.pos, "K-20",
                                                                NodeType.type_to_str(state.invariant.node_type)
                                                                ))
                else:
                    # Make sure the invariant is convex
                    simple_invariant = SimplifyCaseVisitor().replace_case(state.invariant.clone())
                    nnf_invariant = InvariantToNNFVisitor().visit(simple_invariant)
                    is_convex = InvariantIsConvexVisitor().visit(nnf_invariant)
                    if not is_convex:
                        node.add_error(InvariantIsNotConvexError(node.file, state.line, state.pos, "K-19", state.invariant, state))

            if state.urgent is not None and state.urgent.node_type != NodeType.ERROR:
                if self.compile_info.uses_time_scaling and state.urgent.node_type != NodeType.CLOCK:
                    node.add_error(ErrorStateUrgentNotScaledInTimeScaledModel(state.urgent.file, state.urgent.line, state.urgent.pos, "K-35", state))

                value = self.evaluator.visit(state.urgent)
                if value is None or value < 0.0:
                    node.add_error(StateUrgencyIsNegative(state.urgent.file, state.urgent.line, state.urgent.pos, "K-21", state.urgent, state))
            else:
                for trans in node.error_transitions:
                    if trans.source_state == state.name and trans.time_delay is not None:
                        node.add_error(StateNotUrgentWithDelayedTransitionError(state.file, state.line, state.pos, "K-37", state))
                        break

            for trans in node.error_transitions:
                if trans.source_state == state.name and (trans.trigger in node.error_events) \
                    and node.error_events[trans.trigger].distribution != None:
                        if state.invariant is not None:
                            self.compile_info.add_error(ErrorStateInvariantForProbabilisticStateError(state.file, state.line, state.pos,
                                                                                                      "K-27", state))
                        if state.urgent is not None:
                            self.compile_info.add_error(ErrorStateUrgencyForProbabilisticStateError(state.file, state.line, state.pos,
                                                                                                    "K-28", state))

    def check_ErrorModelImplementation_events(self, node):
        for event in node.error_events.values():
            if event.distribution == DistributionType.POISSON and event.parameter is not None:
                rate = float(event.parameter)
                if rate <= 0.0:
                    node.add_error(RateIsNotPositiveError(event.file, event.line, event.pos, "K-13", rate, event))
                pass
            if self.compile_info.uses_time_scaling and event.distribution is not None and event.time_unit is None:
                self.compile_info.add_error(ErrorEventNotScaledInTimeScaledModel(event.file, event.line, event.pos, "K-33", event))

        t = node.get_referenced_error_type()
        if t is not None:
            duplicates = set(node.error_events.keys()) & set(t.error_propagations.keys())

            for id in duplicates:
                event = node.error_events[id]
                prop = t.error_propagations[id]
                self.compile_info.add_error(ErrorEventAlreadyDeclaredInTypeError(event.file, event.line, event.pos, "K-12", t, event, prop))

    def check_ErrorModelImplementation_referenced_error_type_exists(self, node):
        """Checks whether the referenced error type exists.
        """
        type = node.get_referenced_error_type() 
        if type is None:
            # found no type, check for it in the other subpackage to make 
            # the error message more precise
            if node.parent == node.parent.parent.public_subpackage:
                other_subpackage = node.parent.parent.private_subpackage
            else:
                other_subpackage = node.parent.parent.public_subpackage
            
            type = other_subpackage.get_error_type(node.name.split(".")[0])
            if type is None:
                node.add_error(ErrorImplementationRefersToUnknownTypeError(node.file, node.line, node.pos, "K-2", node))
            else:
                node.add_error(ErrorImplementationCrossReferenceToTypeError(node.file, node.line, node.pos, "A-3", node))

    def check_ErrorModelImplementation_states_either_here_or_in_type(self, node):
        """Checks whether the states are defined either in the error model
        implementation or the corresponding error model type. They may not be
        defined in both
        """
        type = node.get_referenced_error_type()
        if type is not None:
            if (len(type.states.keys()) != 0
            and len(node.states.keys()) != 0):
                state = node.states.values()[0]
                node.add_error(ErrorStatesInBothTypeAndImplementationError(state.file, state.line, state.pos, "K-15", type, node))

    def check_ErrorModelImplementation_error_transitions(self, node):
        """Checks all declared transitions in the error model implementation.
        """
        states = node.states.keys()
        type = node.get_referenced_error_type()
        transitions = {}
        if type is not None:
            states += type.states.keys()
        
        for trans in node.error_transitions + node.reactivation_transitions:
            # check if source state declared
            if trans.source_state not in states:
                node.add_error(ErrorStateTransitionRefersToUnknownStateError(trans.file, trans.line, trans.pos, "K-5", trans, trans.source_state))
            # check if target state declared
            if trans.target_state not in states:
                node.add_error(ErrorStateTransitionRefersToUnknownStateError(trans.file, trans.line, trans.pos, "K-5", trans, trans.target_state))
            
            # check if trigger valid (ignored for reactivation_transitions)
            if not trans.is_reactivation:
                trigger_valid = False
                if trans.trigger == "reset" or trans.trigger in node.error_events:
                    trigger_valid = True
                else:
                    if type is not None:
                        if trans.trigger in type.error_propagations:
                            trigger_valid = True
                            
                if not trigger_valid:
                    node.add_error(ErrorTriggerIsInvalidError(trans.file, trans.line, trans.pos, "K-6", trans.trigger, trans))
                
            # check if transition was already declared
            if (trans.source_state, trans.trigger, trans.target_state) in \
               transitions:
                old_trans = transitions[(trans.source_state, trans.trigger, trans.target_state)]
                node.add_error(ErrorStateTransitionAlreadyDeclaredError(trans.file, trans.line, trans.pos, "K-10", node, old_trans))
            else:
                transitions[(trans.source_state, trans.trigger, trans.target_state)] = trans
            
            # check guard
            self.check_ErrorModelImplementation_valid_guard(node, trans)
            self.check_ErrorModelImplementation_valid_within(node, trans)
            
            # check effects
            self.check_ErrorModelImplementation_valid_effects(node, trans)

        # Check for non-deterministic transition triggers
        # (Not supported by model extension, as FI are
        # ambiguous then)
        state_trigger_map = {}
        for state_name in states:
            transitions = node.get_transitions_leaving_state(state_name)
            state_triggers = []

            for trans in transitions:
                if trans.trigger in state_triggers:
                    if state_name in node.states:
                        state = node.states[state_name]
                    else:
                        state = type.states[state_name]
                    node.add_error(ErrorTransitionsLeavingStateAreNondeterministicError(
                                trans.file, trans.line, trans.pos, "K-8",
                                state, trans.trigger))
                    break
                else:
                    state_triggers.append(trans.trigger)

    def check_ErrorModelImplementation_valid_guard(self, node, trans):
        """Checks whether the given guard is legal in a transition leaving
        the given state.
        
        @param    trans:
                    The transition whose guard to check.
        @type     guard:
                    ErrorTransition object
        """
        guard = trans.guard

        if guard is not None:
            guard_type = guard.node_type

            if guard_type != NodeType.BOOL and guard_type != NodeType.ERROR:
                node.add_error(
                        GuardHasIllegalTypeError(trans.file, trans.line,
                                                 trans.pos, "K-23", NodeType.type_to_str(guard_type), trans))

    def check_ErrorModelImplementation_valid_within(self, node, trans):
        if trans.time_delay is not None:
            (lower, upper) = trans.time_delay
            if not (lower.node_type == NodeType.ERROR or upper.node_type == NodeType.ERROR):
                (lval, uval) = self.evaluator.visit(lower), self.evaluator.visit(upper)
                if self.compile_info.uses_time_scaling and (lower.node_type != NodeType.CLOCK or upper.node_type != NodeType.CLOCK):
                    node.add_error(ErrorTransitionWithinNotScaledInTimeScaledModel(trans.file, trans.line, trans.pos, "K-36", trans))
                elif lval is None or uval is None or lval > uval or lval < 0.0:
                    node.add_error(InvalidWithinBoundsError(trans.file, trans.line, trans.pos, "K-24", lval, uval, trans))

    def check_ErrorModelImplementation_valid_effects(self, node, trans):
        """Checks whether the given effects are legal in a transition between
        the two given states.
        
        @param    trans:
                    The transition whose effects to check.
        @type     trans:
                    ErrorTransition object
        """
        effects = trans.effects
        comps_set = []
        if effects:
            for effect in effects:
                comps_set.append(effect.left.name)
            # check whether elements were set multiple times
            duplicates = set([element for element in comps_set if comps_set.count(element) > 1])    
            for element in duplicates:
                node.add_error(DataComponentSetInMultipleEffectsError(trans.file, trans.line, trans.pos, "K-25", element, trans))
    
    def check_ErrorModelImplementation_only_single_type_of_transitions_per_state(self, node):
        """Checks whether the transitions leaving any state are either all
        probabilistic or all non-probabilistic.
        """
        state_names = node.states.keys()
        type = node.get_referenced_error_type()
        if type is not None:
            state_names += type.states.keys()
            
        for state_name in state_names:
            transitions = node.get_transitions_leaving_state(state_name)
            if state_name in node.states:
                state = node.states[state_name]
            else:
                state = type.states[state_name]
            prob_transitions = []
            non_prob_transitions = []
            trigger_to_trans_map = {}
            
            for trans in transitions:
                if trans.trigger in trigger_to_trans_map:
                    trigger_to_trans_map[trans.trigger].append(trans)
                else:
                    trigger_to_trans_map[trans.trigger] = [trans]
                  
                if trans.trigger in node.error_events:
                    event = node.error_events[trans.trigger]
                    if event.distribution == DistributionType.POISSON:
                        prob_transitions.append(trans)
                    else:
                        non_prob_transitions.append(trans)
                elif type is not None and trans.trigger in type.error_propagations:
                    non_prob_transitions.append(trans)
                elif trans.trigger == "reset":
                    non_prob_transitions.append(trans)

            if len(prob_transitions) > 0 and len(non_prob_transitions) > 0:
                pass # Ticket 564
                #node.add_error(
                #        ErrorStateHasDifferentTypesOfLeavingTransitionsError(
                #                state.file, state.line, state.pos, "K-7",
                #                state, prob_transitions, non_prob_transitions))
    
    def check_ErrorModelImplementation_each_state_reachable(self, node):
        """Checks whether each state is reachable by a sequence of transitions.
        """
        type = node.get_referenced_error_type()
        states = set(node.states.keys())
        visited_states = set()
        new_states = set()
        if type is not None:
            states |= set(type.states.keys())

            starting_states = type.get_all_starting_states() | node.get_all_starting_states()
            for starting_state in starting_states:
                new_states.add(starting_state.name)
                visited_states.add(starting_state.name)
            
                while len(new_states) > 0:
                    cur_state = new_states.pop()
                    transitions = node.get_transitions_leaving_state(cur_state)
                
                    for trans in transitions:
                        if not trans.target_state in visited_states:
                            new_states.add(trans.target_state)
                            visited_states.add(trans.target_state)
                        
            if visited_states != states:
                missing_states = states - visited_states
                for state in missing_states:
                    if state in node.states:
                        state_obj = node.states[state]
                    else:
                        state_obj = type.states[state]
                    node.add_error(ErrorStateNotReachableError(state_obj.file, state_obj.line, state_obj.pos, "K-9", state_obj))

    def check_ErrorModelImplementation_probabilistic_guards(self, node):
        # Probabilistic transitions (those triggered by probabilistic events) may not have guards
        for trans in node.error_transitions:
            if (trans.trigger in node.error_events) and node.error_events[trans.trigger].distribution != None:
                if trans.guard is not None:
                    self.compile_info.add_error(ErrorTransitionIsProbabilisticButHasGuardError(trans.file, trans.line, trans.pos,
                                                                                               "K-29", trans))
                if trans.time_delay is not None:
                    self.compile_info.add_error(ErrorTransitionIsProbabilisticButHasDelayError(trans.file, trans.line, trans.pos,
                                                                                               "K-30", trans))
