# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================
"""This module provides visitors on the CompileInformation-structure
that output text formats
"""
from slim_compiler.src.dataStructure.visitors.baseVisitor import BaseVisitor,\
    EvalVisitor
from slim_compiler.src.dataStructure.nominalbehaviour import Category, ModeType, \
    DEFAULTINITMODE
from slim_compiler.src.dataStructure.errorbehaviour import DistributionType, \
    ErrorStateType, ErrorPropagationType
from slim_compiler.src.dataStructure.expressions import TimeUnitType,\
    OperatorType, BinaryOperatorTreeNode, UnaryOperatorTreeNode
import string

class toSLIMVisitor(EvalVisitor):
    """This class provides visit-methods for the CompileInformation-structure.
    The visit-methods return the SLIM-representation of the visited nodes.

    Note: for larger segments list concatenation is used as it seems to be
    a bit faster.
    """
    def __init__(self):
        super(toSLIMVisitor, self).__init__()

        self.exprToSLIMVisitor = ExprToSLIMVisitor()
        self.indent_ch = '    '
        self.indent_sec = ''

    def indent(self, text):
        return self.indent_ch + ('\n'+self.indent_ch).join(text.split('\n'))

    def action_CompileInformation(self, node, data):
        result = '\n\n'.join(data.packages.values())
        return result

    def action_Package(self, node, data):
        result = []
        if node.name != "__default__":
            result += ["package " + node.name + "\n"]
            result += [self.indent_sec + "public\n"]
            result += [self.indent(data.public_subpackage) + '\n']
        else:
            result += [data.public_subpackage]
        if not node.private_subpackage.is_empty():
            result += [self.indent_sec + "private\n"]
            result += [self.indent(data.private_subpackage) + '\n']
        if node.name != "__default__":
            result += ["end " + node.name + ";"]
        return ''.join(result)

    def action_Subpackage(self, node, data):
        result = ""
        items = data.types.values() + data.implementations.values() \
                + data.error_types.values() + data.error_implementations.values()
        result += '\n\n'.join(items)
        return result

    def action_ComponentType(self, node, data):
        result = []
        fdir = "fdir " if node.fdir else ""
        result += [fdir + Category.category_to_str(node.category) + " " + node.name + "\n"]
        if data.in_data_ports or data.out_data_ports or \
           data.in_event_ports or data.out_event_ports:
            result += [self.indent_sec + "features\n"]
            for port in data.in_data_ports.values():
                result += [self.indent(port)] + ["\n"]
            for port in data.out_data_ports.values():
                result += [self.indent(port)] + ["\n"]
            for port in data.in_event_ports.values():
                result += [self.indent(port)] + ["\n"]
            for port in data.out_event_ports.values():
                result += [self.indent(port)] + ["\n"]
        result += ["end " + node.name + ";"]
        return ''.join(result)

    def action_ComponentImplementation(self, node, data):
        result = []
        fdir = "fdir " if node.fdir else ""
        result += [fdir + Category.category_to_str(node.category) + \
                " implementation " + node.name + "\n"]
        if node.data_subcomponents or node.non_data_subcomponents:
            result += [self.indent_sec + "subcomponents\n"]
            for sub in data.data_subcomponents.values():
                result += [self.indent(sub)] + ["\n"]
            for sub in data.non_data_subcomponents.values():
                result += [self.indent(sub)] + ["\n"]
        if node.data_port_connections or node.event_port_connections:
            result += [self.indent_sec + "connections\n"]
            for conn in data.data_port_connections:
                result += [self.indent(conn)] + ['\n']
            for conn in data.event_port_connections:
                result += [self.indent(conn)] + ['\n']
        if node.flows:
            result += [self.indent_sec + "flows\n"]
            for flow in data.flows:
                result += [self.indent(flow)] + ["\n"]
        if node.modes:
            result += [self.indent_sec + "modes\n"]
            for mode in data.modes.values():
                result += [self.indent(mode)] + ["\n"]
        if node.normal_mode_transitions or node.probability_mode_transitions or node.reactivation_mode_transitions:
            result += [self.indent_sec + "transitions\n"]
            for trans in data.normal_mode_transitions:
                result += [self.indent(trans)] + ["\n"]
            for trans in data.reactivation_mode_transitions:
                result += [self.indent(trans)] + ["\n"]
            for trans in data.probability_mode_transitions:
                result += [self.indent(trans)] + ["\n"]
        result += ["end " + node.name + ";"]

        return ''.join(result)

    def action_InEventPort(self, node, data):
        result = node.name + ": in event port;"
        return result

    def action_OutEventPort(self, node, data):
        nonblock = " nonblocking" if node.nonblocking else ""
        result = node.name + ": out event port" + nonblock + ";"
        return result

    def action_DataPort(self, node, data):
        defval = " default " + data.default_value if node.default_value is not None else ""
        result = node.name + ": in data port " + ''.join(data.data_type) + defval + ";"
        return result

    def action_OutDataPort(self, node, data):
        alarm = " alarm" if node.alarm else ""
        observable = " observable" if node.observable else ""
        defval = " default " + data.default_value if node.default_value is not None else ""
        result = node.name + ": out data port " + ''.join(data.data_type) + defval + alarm + observable + ";"
        return result

    def action_ErrorEventPort(self, node, data):
        occurrence = ""
        if node.distribution is not None:
            occurrence += " " + DistributionType.dist_type_to_str(node.distribution) + \
                          " " + node.parameter
            if node.time_unit is not None:
                occurrence += " per " + TimeUnitType.time_unit_type_to_str(node.time_unit)
        result = node.name + ": error event port" + occurrence + ";"
        return result

    def _inmodes(self, in_modes):
        result = ""
        if in_modes:
            result += " in modes (" + ', '.join(in_modes) + ")"
        return result

    def action_PortConnection(self, node, data, type):
        passive = " passive" if node.passive else ""
        modes = self._inmodes(node.in_modes)
        result = type + " port " + node.source_port + \
                " -> " + node.target_port + modes + passive + ";"
        return result

    def action_Flow(self, node, data):
        modes = self._inmodes(node.in_modes)
        result = data.expression + modes + ";"
        return result

    def action_NonDataSubcomponent(self, node, data):
        result = node.name + ": " + \
                Category.category_to_str(node.category) + " " + \
                node.implementation_ref
        if node.accesses:
            for binding in node.accesses:
                result += " accesses " + binding
        if node.stored_in:
            for binding in node.stored_in:
                result += " stored in " + binding
        if node.running_on:
            for binding in node.running_on:
                result += " running on " + binding
        if node.in_modes:
            result += self._inmodes(node.in_modes)
        result += ";"
        return result

    def action_DataSubcomponent(self, node, data):
        result = node.name + ": data " + data.data_type
        if node.default_value is not None:
            result += " default " + data.default_value
        if node.in_modes:
            result += self._inmodes(node.in_modes)
        result += ";"
        return result

    def action_Mode(self, node, data):
        result = node.name + " :"
        if node.type == ModeType.INITIAL:
            result += " initial"
        elif node.type == ModeType.ACTIVATION:
            result += " activation"
        elif node.type == ModeType.ERROR:
            result += " error"
        result += " mode"
        if node.urgent:
            result += " urgent in " + data.urgent
        if node.invariant:
            result += " while " + data.invariant
        result += ";"
        return result

    def action_ModeTransition(self, node, data):
        result = node.source_mode + " -[" + data.label + "]-> " + node.target_mode + ";"
        return result

    def action_ModeTransitionLabel(self, node, data):
        result = ""
        if node.trigger is not None:
            if not isinstance(node.trigger, basestring):
                if not node.trigger[0] is None:
                    result += ' or '.join(node.trigger)
            else:
                result += node.trigger
        if node.time_delay:
            result += " within " + data.time_delay[0] + " to " + data.time_delay[1]
        if node.guard:
            result += " when " + data.guard
        if len(node.effects) > 0:
            result += " then "
            result += '; '.join(data.effects)
        return result

    def action_ReactivationModeTransitionLabel(self, node, data):
        result = "@activation"
        if len(node.effects) > 0:
            result += " then "
            result += '; '.join(data.effects)
        return result

    def action_ProbabilityModeTransitionLabel(self, node, data):
        return "prob " + str(node.probability)

    def action_ErrorModelType(self, node, data):
        result = ["error model " + node.name + "\n"]
        if node.states or node.error_propagations:
            result += [self.indent_sec + "features\n"]
            for state in data.states.values():
                result += [self.indent(state)] + ["\n"]
            for propagation in data.error_propagations.values():
                result += [self.indent(propagation)] + ["\n"]
        result += ["end " + node.name + ";"]

        return ''.join(result)

    def action_ErrorModelImplementation(self, node, data):
        result = ["error model implementation " + node.name + "\n"]
        if node.error_events:
            result += [self.indent_sec + "events\n"]
            for e in data.error_events.values():
                result += [self.indent(e)] + ["\n"]
        if node.clocks:
            result += [self.indent_sec + "clocks\n"]
            for c in data.clocks.values():
                result += [self.indent(c)] + ["\n"]
        if node.states:
            result += [self.indent_sec + "states\n"]
            for s in data.states.values():
                result += [self.indent(s)] + ["\n"]
        if node.error_transitions or node.reactivation_transitions:
            result += [self.indent_sec + "transitions\n"]
            for trans in data.error_transitions:
                result += [self.indent(trans)] + ["\n"]
            for trans in data.reactivation_transitions:
                result += [self.indent(trans)] + ["\n"]
        result += ["end " + node.name + ";"]

        return ''.join(result)

    def action_ErrorState(self, node, data):
        result = node.name + ": " + \
                ErrorStateType.err_state_type_to_str(node.type) + " state"
        if node.urgent:
            result += " urgent in " + data.urgent
        if node.invariant:
            result += " while " + data.invariant
        result += ";"
        return result

    def action_ErrorTransition(self, node, data):
        result = node.source_state
        result += "-[" + ('@activation' if node.is_reactivation else node.trigger)
        if node.time_delay:
            result += " within " + data.time_delay[0] + " to " + data.time_delay[1]
        if node.guard:
            result += " when " + data.guard
        if node.effects:
            result += " then " + '; '.join(data.effects)
        result += "]->" + node.target_state + ";"
        return result

    def action_ErrorEvent(self, node, data):
        occurrence = ""
        if node.distribution is not None:
            occurrence += " " + DistributionType.dist_type_to_str(node.distribution) + \
                          " " + node.parameter
            if node.time_unit is not None:
                occurrence += " per " + TimeUnitType.time_unit_type_to_str(node.time_unit)
        result = node.name + ": error event " + occurrence + ";"
        return result

    def action_ErrorPropagation(self, node, data):
        result = node.name + ": " + \
                ErrorPropagationType.prop_type_to_str(node.type) + \
                " error propagation;"
        return result

    def action_AbstractDataType(self, node, data):
        result = node.get_name()
        return result;

    def action_Clock(self, node, data):
        result = node.get_name()
        if node.time_unit:
            result += " " + TimeUnitType.time_unit_type_to_str(node.time_unit)
        return result;

    def action_IntRange(self, node, data):
        result = "[" + str(node.lower) + ".." + str(node.upper) + "]"
        return result;

    def action_Enum(self, node, data):
        result = node.get_name() + "(" + ", ".join(node.values) + ")"
        return result;

    def action_ExpressionTreeNode(self, node, data):
        # TODO: Defer to ExprToSLIMVisitor class. Makes paren parameter handling easier,
        # but should be refactored into this class
        return self.exprToSLIMVisitor.visit(node)

class ExprToSLIMVisitor(BaseVisitor):
    """This class provides visit-methods for the expression-structure.
    The visit-methods return the SLIM-representation of the visited nodes.
    Do not use the class directly, rather use ToSLIMVisitor
    """

    operator_precedence = {
        # 1 is for () and case

        OperatorType.NOT : 2,
        OperatorType.UNARY_MINUS : 2,

        OperatorType.MSEC : 3,
        OperatorType.SEC : 3,
        OperatorType.MIN : 3,
        OperatorType.HOUR : 3,
        OperatorType.DAY : 3,
        OperatorType.MSEC_TOREAL : 3,
        OperatorType.SEC_TOREAL : 3,
        OperatorType.MIN_TOREAL : 3,
        OperatorType.HOUR_TOREAL : 3,
        OperatorType.DAY_TOREAL : 3,
        OperatorType.REAL_TOMSEC : 3,
        OperatorType.REAL_TOSEC : 3,
        OperatorType.REAL_TOMIN : 3,
        OperatorType.REAL_TOHOUR : 3,
        OperatorType.REAL_TODAY : 3,

        OperatorType.TIMES : 4,
        OperatorType.DIVIDE : 4,
        OperatorType.MOD : 4,

        OperatorType.PLUS : 5,
        OperatorType.MINUS : 5,

        OperatorType.LESS : 6,
        OperatorType.LESSOREQUAL : 6,
        OperatorType.GREATER : 6,
        OperatorType.GREATEROREQUAL : 6,
        OperatorType.EQUAL : 6,
        OperatorType.NOTEQUAL : 6,

        OperatorType.AND : 7,

        OperatorType.OR : 8,
        OperatorType.XOR : 8,
        OperatorType.XNOR : 8,

        OperatorType.IFF : 9,

        OperatorType.IMP : 10,

        OperatorType.ASSIGN : 11,
    }

    non_associative_operators = [OperatorType.DIVIDE, OperatorType.MOD,
                                 OperatorType.MINUS, OperatorType.IMP,
                                 OperatorType.ASSIGN
                                 ]

    def visit_IdentifierTreeNode(self, root_node, paren = False):
        result = root_node.name
        if paren:
            return "(" + result + ")"
        return result

    def visit_BooleanTreeNode(self, root_node, paren = False):
        result = string.lower(str(root_node.value))
        if paren:
            return "(" + result + ")"
        return result

    def visit_NumberTreeNode(self, root_node, paren = False):
        result = str(root_node.number)
        if paren:
            return "(" + result + ")"
        return result

    def visit_TrajectoryExpressionTreeNode(self, root_node, paren = False):
        # deprecated
        return str(root_node.offset)

    def visit_BinaryOperatorTreeNode(self, root_node, paren = False):
        next_level_paren_left = False
        next_level_paren_right = False

        root_precedence = self.operator_precedence[root_node.operator]

        if isinstance(root_node.left, BinaryOperatorTreeNode) or isinstance(root_node.left, UnaryOperatorTreeNode):
            left_precedence = self.operator_precedence[root_node.left.operator]
            if left_precedence > root_precedence:
                next_level_paren_left = True
            elif left_precedence == root_precedence:
                if root_node.left.operator != root_node.operator or root_node.operator in ExprToSLIMVisitor.non_associative_operators:
                    # Only assignment right associative
                    next_level_paren_left = (root_node.left.operator == OperatorType.ASSIGN)
            pass

        if isinstance(root_node.right, BinaryOperatorTreeNode) or isinstance(root_node.right, UnaryOperatorTreeNode):
            right_precedence = self.operator_precedence[root_node.right.operator]
            if right_precedence > root_precedence:
                next_level_paren_right = True
            elif right_precedence == root_precedence:
                if root_node.right.operator != root_node.operator or root_node.operator in ExprToSLIMVisitor.non_associative_operators:
                    # All binary operators left associative, except assignment
                    next_level_paren_right = (root_node.right.operator != OperatorType.ASSIGN)
            pass

        result = self.visit(root_node.left, next_level_paren_left) + " " + \
                 OperatorType.operator_to_str(root_node.operator) + " " + \
                 self.visit(root_node.right, next_level_paren_right)
        if paren: return "(" + result + ")"
        return result

    def visit_UnaryOperatorTreeNode(self, root_node, paren = False):
        # Unary operator has highest precedence
        # TODO: precedence difference between time unit and others, update logic
        next_level_paren = False

        root_precedence = self.operator_precedence[root_node.operator]

        if isinstance(root_node.subexpression, BinaryOperatorTreeNode) or isinstance(root_node.subexpression, UnaryOperatorTreeNode):
            sub_precedence = self.operator_precedence[root_node.subexpression.operator]
            if sub_precedence > root_precedence:
                next_level_paren = True
            elif root_node.subexpression.operator == root_node.operator == OperatorType.UNARY_MINUS:
                # Special case, otherwise it'll be a comment
                next_level_paren = True

        if OperatorType.MSEC <= root_node.operator <= OperatorType.DAY or\
           OperatorType.MSEC_TOREAL <= root_node.operator <= OperatorType.DAY_TOREAL or\
           OperatorType.REAL_TOMSEC <= root_node.operator <= OperatorType.REAL_TODAY:
            result = self.visit(root_node.subexpression, next_level_paren) + \
                     " " + OperatorType.operator_to_str(root_node.operator)
        else:
            space = " " if root_node.operator == OperatorType.NOT else ""
            result = OperatorType.operator_to_str(root_node.operator) + \
                     space + self.visit(root_node.subexpression, next_level_paren)
        if paren: return "(" + result + ")"
        return result

    def visit_TrajectoryEquationTreeNode(self, root_node, paren = False):
        # next level paren False, as this node has pretty much lowest precedence
        next_level_paren = False
        result = self.visit(root_node.variable, next_level_paren) + "' = " + \
               self.visit(root_node.trajectory_expression, next_level_paren)
        if root_node.time_unit:
            result += " per " + TimeUnitType.time_unit_type_to_str(root_node.time_unit)
        if paren: return "(" + result + ")"
        return result

    def visit_CaseExpressionTreeNode(self, root_node, paren = False):
        next_level_paren = False
        result = "case "
        pairs = [self.visit(cond, next_level_paren) + " : " + self.visit(res, next_level_paren) for (cond, res) in root_node.cases]
        result += "; ".join(pairs)
        result += " otherwise " + self.visit(root_node.default_expression, next_level_paren)
        result += " end"
        if paren: return "(" + result + ")"
        return result

class toSXMLVisitor(BaseVisitor):
    """This class provides visit-methods for the CompileInformation-structure.
    The visit-methods return the SLIM-representation of the visited nodes.
    """
    def __init__(self):
        self.toSLIMVisitor = toSLIMVisitor()

    def visit_CompileInformation(self, node, name, assertNoErrors):
        if assertNoErrors:
            assert node.error_collector.has_no_errors(), "Model contains errors."
        assert len(node.packages) == 1, "Model contains more than one package."

        root_components = node.get_root_components()
        error_root_components = node.get_error_root_components()

        result = "<CGM Version=\"1.1.0\">\n"
        result += "\t<CGMProjectObject id=\"" + name + "\" label=\"" + name + "\">\n"
        for root_impl in root_components:
            result += "\t\t<CHILD reference=\"" + name + "." + root_impl.name.split(".")[0] + "\"/>\n"
        for root_impl in error_root_components:
            result += "\t\t<CHILD reference=\"" + name + "." + root_impl.name + "\"/>\n"
        result += "\t</CGMProjectObject>\n"

        impl_instance_map = {}
        type_instance_map = {}
        for root_impl in root_components:
            result += self.visit(root_impl, name, root_impl.name.split(".")[0], type_instance_map, impl_instance_map)
            pass

        impl_instance_map = {}
        type_instance_map = {}
        package = node.packages.values()[0]
        for subpackage in [package.public_subpackage, package.private_subpackage]:
            for error_impl in subpackage.error_implementations.values():
                result += self.visit(error_impl, name, error_impl.name, type_instance_map, impl_instance_map)

        result += "</CGM>"
        return result

    def visit_ComponentImplementation(self, node, prefix, relativeName, type_instance_map, impl_instance_map, in_modes = [], accesses = [], runningOn = [], storedIn = []):
        """Returns the SXML representation corresponding to the
        component implementation object.
        
        @return:    The SXML representation corresponding to the
                    component implementation object.
        @rtype:     string
        """
        category_string = ""
        if node.category == Category.SYSTEM:
            category_string += "CGMSystemObject"
        elif node.category == Category.PROCESS:
            category_string += "CGMProcessObject"
        elif node.category == Category.THREAD_GROUP:
            category_string += "CGMThreadGroupObject"
        elif node.category == Category.THREAD:
            category_string += "CGMThreadObject"
        elif node.category == Category.DATA:
            category_string += "CGMDataObject"
        elif node.category == Category.PROCESSOR:
            category_string += "CGMProcessorObject"
        elif node.category == Category.MEMORY:
            category_string += "CGMMemoryObject"
        elif node.category == Category.DEVICE:
            category_string += "CGMDeviceObject"
        elif node.category == Category.BUS:
            category_string += "CGMBusObject"

        # Keep two maps: one of component type to the id of the first component instance used, and one of component implementation to the id of the first component instance used. Subsequent use of the same type/implementation have to refer to their id. Also see ticket #100 on Ellidiss Trac.
        classifier = ""
        (c_type, c_impl) = node.name.split(".")
        if c_type in type_instance_map:
            classifier += type_instance_map[c_type] + " "
        else:
            classifier += c_type + ". "
            type_instance_map[c_type] = prefix + "." + relativeName
        if node.name in impl_instance_map:
            classifier += impl_instance_map[node.name]
        else:
            impl_instance_map[node.name] = prefix + "." + relativeName
            classifier += c_impl


        result = "\t<" + category_string + " id=\"" + prefix + "." + relativeName + "\" classifier=\"" + classifier + "\" label=\"" + relativeName + "\""
        if runningOn is not None and len(runningOn) > 0:
            result += " runningOn=\""
            for i in range(0, len(runningOn)):
                runningOnElement = prefix + "." + runningOn[i]
                result += runningOnElement
                if i < len(runningOn) - 1:
                    result += " "
            result += "\""
        if accesses is not None and len(accesses) > 0:
            result += " accesses=\""
            for i in range(0, len(accesses)):
                accessesElement = prefix + "." + accesses[i]
                result += accessesElement
                if i < len(accesses) - 1:
                    result += " "
            result += "\""
        if storedIn is not None and len(storedIn) > 0:
            result += " storedIn=\""
            for i in range(0, len(storedIn)):
                storedInElement = prefix + "." + storedIn[i]
                result += storedInElement
                if i < len(storedIn) - 1:
                    result += " "
            result += "\""
        result += " fdir=\"" + str(node.fdir).lower() + "\">\n"

        type = node.get_referenced_component_type()


        if type is not None:
            for port in type.in_data_ports.values() + type.out_data_ports.values():
                result += "\t\t<FEATURE reference=\"" + prefix + "." + relativeName + ".dataPort-" + port.name + "\"/>\n"
            for port in type.in_event_ports.values() + type.out_event_ports.values():
                result += "\t\t<FEATURE reference=\"" + prefix + "." + relativeName + ".eventPort-" + port.name + "\"/>\n"

        for subcomp in node.non_data_subcomponents.values() + node.data_subcomponents.values():
            result += "\t\t<CHILD reference=\"" + prefix + "." + relativeName + "." + subcomp.name + "\"/>\n"

        for mode in node.modes.values():
            if mode.name is not DEFAULTINITMODE:
                result += "\t\t<CHILD reference=\"" + prefix + "." + relativeName + "." + mode.name + "\"/>\n"

        # handling of anymode transitions
        node.normal_mode_transitions = filter(lambda x: x.internal_number <= 1, node.normal_mode_transitions)
        # if exists any generated anymode transitions
        if any(map(lambda x: x.internal_number > 0, node.normal_mode_transitions)):
            result += "\t\t<CHILD reference=\"" + prefix + "." + relativeName + "._AnyMode\"/>\n"
        # update transitions with anymode information
        for transition in filter(lambda x: x.internal_number > 0, node.normal_mode_transitions):
            transition.source_mode = "_AnyMode"

        for flow in node.flows:
            result += "\t\t<CHILD reference=\"" + prefix + "." + relativeName + ".F" + str(flow.number) + "\"/>\n"

        for mode in in_modes:
            result += "\t\t<INMODE reference=\"" + prefix + "." + mode + "\"/>\n"

        result += "\t</" + category_string + ">\n";

        if type is not None:
            for port in type.in_data_ports.values() + type.out_data_ports.values():
                result += self.visit(port, prefix + "." + relativeName, port.name)
            for port in type.in_event_ports.values() + type.out_event_ports.values():
                result += self.visit(port, prefix + "." + relativeName, port.name)

        for subcomp in node.data_subcomponents.values():
            result += self.visit(subcomp, prefix + "." + relativeName, subcomp.name)

        for mode in node.modes.values():
            if mode.name is not DEFAULTINITMODE: result += self.visit(mode, prefix + "." + relativeName, mode.name)

        # handling of anymode transitions
        if any(map(lambda x: x.internal_number > 0, node.normal_mode_transitions)):
            result += self.visit(mode, prefix + "." + relativeName, "_AnyMode")

        for transition in node.normal_mode_transitions:
            result += self.visit(transition, prefix + "." + relativeName, str(transition.number))

        for connection in node.data_port_connections:
            result += "\t<CGMPortConnectionObject id=\"" + prefix + "." + relativeName + ".C" + str(connection.number) + "\" label=\"\">\n"
            if connection.source_port.find(".") == -1:
                result += "\t\t<OBJECT1 reference=\"" + prefix + "." + relativeName + ".dataPort-" + connection.source_port + "\"/>\n"
            else:
                (comp_name, port_name) = connection.source_port.split(".")
                result += "\t\t<OBJECT1 reference=\"" + prefix + "." + relativeName + "." + comp_name + ".dataPort-" + port_name + "\"/>\n"
            if connection.target_port.find(".") == -1:
                result += "\t\t<OBJECT2 reference=\"" + prefix + "." + relativeName + ".dataPort-" + connection.target_port + "\"/>\n"
            else:
                (comp_name, port_name) = connection.target_port.split(".")
                result += "\t\t<OBJECT2 reference=\"" + prefix + "." + relativeName + "." + comp_name + ".dataPort-" + port_name + "\"/>\n"
            for mode in connection.in_modes:
                result += "\t\t<INMODE reference=\"" + prefix + "." + relativeName + "." + mode + "\"/>\n"
            result += "\t</CGMPortConnectionObject>\n"

        for connection in node.event_port_connections:
            # support for passive port connections
            if connection.passive:
                passive_str = "true"
            else:
                passive_str = "false"

            result += "\t<CGMPortConnectionObject id=\"" + prefix + "." + relativeName + ".C" + str(connection.number) + "\" label=\"\"" + " passive=\"" + passive_str + "\">\n"

            if connection.source_port.find(".") == -1:
                result += "\t\t<OBJECT1 reference=\"" + prefix + "." + relativeName + ".eventPort-" + connection.source_port + "\"/>\n"
            else:
                (comp_name, port_name) = connection.source_port.split(".")
                result += "\t\t<OBJECT1 reference=\"" + prefix + "." + relativeName + "." + comp_name + ".eventPort-" + port_name + "\"/>\n"
            if connection.target_port.find(".") == -1:
                result += "\t\t<OBJECT2 reference=\"" + prefix + "." + relativeName + ".eventPort-" + connection.target_port + "\"/>\n"
            else:
                (comp_name, port_name) = connection.target_port.split(".")
                result += "\t\t<OBJECT2 reference=\"" + prefix + "." + relativeName + "." + comp_name + ".eventPort-" + port_name + "\"/>\n"
            for mode in connection.in_modes:
                result += "\t\t<INMODE reference=\"" + prefix + "." + relativeName + "." + mode + "\"/>\n"
            result += "\t</CGMPortConnectionObject>\n"

        for flow in node.flows:
            result += "\t<CGMFlowBoxObject id=\"" + prefix + "." + relativeName + ".F" + str(flow.number) + "\" label=\"" + self.toSLIMVisitor.visit(flow.expression.right).replace(">", "&gt;").replace("<", "&lt;").replace("=", "&eq;") + "\">\n"
            for mode in flow.in_modes:
                result += "\t\t<INMODE reference=\"" + prefix + "." + relativeName + "." + mode + "\"/>\n"
            result += "\t\t<FEATURE reference=\"" + prefix + "." + relativeName + ".GC" + str(flow.number) + "\" />\n"
            result += "\t</CGMFlowBoxObject>\n"

            result += "\t<GenericConnectorObject id=\"" + prefix + "." + relativeName + ".GC" + str(flow.number) + "\" label=\"\" direction=\"out\" >\n"
            result += "\t\t<COMPONENT reference=\"" + prefix + "." + relativeName + ".F" + str(flow.number) + "\" />\n"
            result += "\t</GenericConnectorObject>\n"

            result += "\t<CGMFlowArrowObject id=\"" + prefix + "." + relativeName + ".FA" + str(flow.number) + "\" label=\"\">\n"
            result += "\t\t<OBJECT1 reference=\"" + prefix + "." + relativeName + ".dataPort-" + self.toSLIMVisitor.visit(flow.expression.left) + "\" />\n"
            result += "\t\t<OBJECT2 reference=\"" + prefix + "." + relativeName + ".GC" + str(flow.number) + "\" />\n"
            result += "\t</CGMFlowArrowObject>\n"

        for subcomp in node.non_data_subcomponents.values():
            impl = subcomp.get_referenced_component_implementation()
            if impl is not None:
                result += self.visit(impl, prefix + "." + relativeName, subcomp.name, type_instance_map, impl_instance_map, subcomp.in_modes, subcomp.accesses, subcomp.running_on, subcomp.stored_in)

        return result


    def visit_InEventPort(self, node, prefix, relativeName):
        """Returns the SXML representation corresponding to the incoming
        event port object.
        
        @return:    The SXML representation corresponding to the incoming
                    event port object.
        @rtype:     string
        """
        result = "\t<CGMEventPortObject id=\"" + prefix + ".eventPort-" + relativeName + "\" "
        result += "label=\"" + relativeName + "\" direction=\"in\">\n"
        result += "\t\t<COMPONENT reference=\"" + prefix + "\"/>\n"
        result += "\t</CGMEventPortObject>\n"

        return result


    def visit_OutEventPort(self, node, prefix, relativeName):
        """Returns the SXML representation corresponding to the outgoing
        event port object.
        
        @return:    The SXML representation corresponding to the outgoing
                    event port object.
        @rtype:     string
        """
        result = "\t<CGMEventPortObject id=\"" + prefix + ".eventPort-" + relativeName + "\" "
        result += "label=\"" + node.name + "\" direction=\"out\">\n"
        result += "\t\t<COMPONENT reference=\"" + prefix + "\"/>\n"
        result += "\t</CGMEventPortObject>\n"

        return result

    def visit_DataPort(self, node, prefix, relativeName):
        """Returns the SXML representation corresponding to the incoming
        data port object.
        
        @return:    The SXML representation corresponding to the incoming
                    event port object.
        @rtype:     string
        """
        result = "\t<CGMDataPortObject id=\"" + prefix + ".dataPort-" + relativeName + "\" "
        result += "label=\"" + relativeName + "\" direction=\"in\">\n"
        result += "\t\t<COMPONENT reference=\"" + prefix + "\"/>\n"
        result += "\t\t<TYPE reference=\"" + prefix + ".dataPort-" + relativeName + ".type\" />\n"
        result += "\t</CGMDataPortObject>\n"
        if node.default_value is None:
            result += self.visit(node.data_type, prefix + ".dataPort-" + relativeName, "")
        else:
            result += self.visit(node.data_type, prefix + ".dataPort-" + relativeName, node.default_value)

        return result

    def visit_OutDataPort(self, node, prefix, relativeName):
        """Returns the SXML representation corresponding to the outgoing
        data port object.
        
        @return:    The SXML representation corresponding to the outgoing
                    event port object.
        @rtype:     string
        """
        result = "\t<CGMDataPortObject id=\"" + prefix + ".dataPort-" + relativeName + "\" "

        result += "label=\"" + relativeName + "\""
        if node.observable or node.alarm:
            result += " modifier=\"true\" "
        else:
            result += " modifier=\"false\" "
        result += "direction=\"out\" "
        result += ">\n"

        result += "\t\t<COMPONENT reference=\"" + prefix + "\"/>\n"
        result += "\t\t<TYPE reference=\"" + prefix + ".dataPort-" + relativeName + ".type\" />\n"

        result += "\t</CGMDataPortObject>\n"

        if node.default_value is None:
            result += self.visit(node.data_type, prefix + ".dataPort-" + relativeName, "")
        else:
            result += self.visit(node.data_type, prefix + ".dataPort-" + relativeName, node.default_value)

        return result

    def visit_DataSubcomponent(self, node, prefix, relativeName):
        """Returns the SXML representation corresponding to the outgoing
        data port object.
        
        @return:    The SXML representation corresponding to the outgoing
                    event port object.
        @rtype:     string
        """
        result = "\t<CGMDataObject id=\"" + prefix + "." + relativeName + "\" classifier=\"\" "

        result += "label=\"" + relativeName + "\">\n"

        result += "\t\t<TYPE reference=\"" + prefix + "." + relativeName + ".type\" />\n"

        for mode in node.in_modes:
            result += "\t\t<INMODE reference=\"" + prefix + "." + mode + "\"/>\n"

        result += "\t</CGMDataObject>\n"

        if node.default_value is None:
            result += self.visit(node.data_type, prefix + "." + relativeName, "")
        else:
            result += self.visit(node.data_type, prefix + "." + relativeName, node.default_value)

        return result


    def visit_Bool(self, node, parent, default_value):
        """Returns the SXML representation corresponding to the data type
        object.
        
        @return:    The SXML representation corresponding to the data type
                    object.
        @rtype:     string
        """
        if default_value:
            result = "\t<CGMDataTypeObject id=\"" + parent + ".type" + "\" typeName=\"bool\" typeDefaultValue=\"" + self.toSLIMVisitor.visit(default_value).lower() + "\" >\n"
        else:
            result = "\t<CGMDataTypeObject id=\"" + parent + ".type" + "\" typeName=\"bool\" >\n"

        result += "\t\t<COMPONENT reference=\"" + parent + "\" />\n"
        result += "\t</CGMDataTypeObject>\n"

        return result

    def visit_Clock(self, node, parent, default_value):
        """Returns the SXML representation corresponding to the data type
        object.
        
        @return:    The SXML representation corresponding to the data type
                    object.
        @rtype:     string
        """
        if default_value:
            result = "\t<CGMDataTypeObject id=\"" + parent + ".type" + "\" typeName=\"clock\" typeDefaultValue=\"" + self.toSLIMVisitor.visit(default_value).lower() + "\" >\n"
        else:
            result = "\t<CGMDataTypeObject id=\"" + parent + ".type" + "\" typeName=\"clock\" >\n"

        result += "\t\t<COMPONENT reference=\"" + parent + "\" />\n"
        result += "\t</CGMDataTypeObject>\n"

        return result


    def visit_Continuous(self, node, parent, default_value):
        """Returns the SXML representation corresponding to the data type
        object.
        
        @return:    The SXML representation corresponding to the data type
                    object.
        @rtype:     string
        """
        if default_value:
            result = "\t<CGMDataTypeObject id=\"" + parent + ".type" + "\" typeName=\"continuous\" typeDefaultValue=\"" + self.toSLIMVisitor.visit(default_value).lower() + "\" >\n"
        else:
            result = "\t<CGMDataTypeObject id=\"" + parent + ".type" + "\" typeName=\"continuous\" >\n"

        result += "\t\t<COMPONENT reference=\"" + parent + "\" />\n"
        result += "\t</CGMDataTypeObject>\n"

        return result

    def visit_Enum(self, node, parent, default_value):
        """Returns the SXML representation corresponding to the data type
        object.
        
        @return:    The SXML representation corresponding to the data type
                    object.
        @rtype:     string
        """
        if default_value:
            result = "\t<CGMDataTypeObject id=\"" + parent + ".type" + "\" typeName=\"enum\" typeDefaultValue=\"" + self.toSLIMVisitor.visit(default_value) + "\" "
        else:
            result = "\t<CGMDataTypeObject id=\"" + parent + ".type" + "\" typeName=\"enum\" "

        result += "typeDefinition=\"("
        for i in range(0, len(node.values)):
            result += node.values[i]
            if i < len(node.values) - 1:
                result += ", "
        result += ")\" >\n"

        result += "\t\t<COMPONENT reference=\"" + parent + "\" />\n"
        result += "\t</CGMDataTypeObject>\n"

        return result


    def visit_Int(self, node, parent, default_value):
        """Returns the SXML representation corresponding to the data type
        object.
        
        @return:    The SXML representation corresponding to the data type
                    object.
        @rtype:     string
        """
        if default_value:
            result = "\t<CGMDataTypeObject id=\"" + parent + ".type" + "\" typeName=\"int\" typeDefaultValue=\"" + self.toSLIMVisitor.visit(default_value).lower() + "\" >\n"
        else:
            result = "\t<CGMDataTypeObject id=\"" + parent + ".type" + "\" typeName=\"int\" >\n"

        result += "\t\t<COMPONENT reference=\"" + parent + "\" />\n"
        result += "\t</CGMDataTypeObject>\n"

        return result


    def visit_IntRange(self, node, parent, default_value):
        """Returns the SXML representation corresponding to the data type
        object.
        
        @return:    The SXML representation corresponding to the data type
                    object.
        @rtype:     string
        """
        if default_value:
            result = "\t<CGMDataTypeObject id=\"" + parent + ".type" + "\" typeName=\"range\" typeDefaultValue=\"" + self.toSLIMVisitor.visit(default_value).lower() + "\" typeDefinition=\"[" + str(node.lower) + ".." + str(node.upper) + "]\" >\n"
        else:
            result = "\t<CGMDataTypeObject id=\"" + parent + ".type" + "\" typeName=\"range\" typeDefinition=\"[" + str(node.lower) + ".." + str(node.upper) + "]\" >\n"

        result += "\t\t<COMPONENT reference=\"" + parent + "\" />\n"
        result += "\t</CGMDataTypeObject>\n"

        return result


    def visit_Real(self, node, parent, default_value):
        """Returns the SXML representation corresponding to the data type
        object.
        
        @return:    The SXML representation corresponding to the data type
                    object.
        @rtype:     string
        """
        if default_value:
            result = "\t<CGMDataTypeObject id=\"" + parent + ".type" + "\" typeName=\"real\" typeDefaultValue=\"" + self.toSLIMVisitor.visit(default_value).lower() + "\" >\n"
        else:
            result = "\t<CGMDataTypeObject id=\"" + parent + ".type" + "\" typeName=\"real\" >\n"

        result += "\t\t<COMPONENT reference=\"" + parent + "\" />\n"
        result += "\t</CGMDataTypeObject>\n"

        return result


    def visit_Mode(self, node, prefix, relativeName):
        """Returns the SXML representation corresponding to the mode object.
        
        @return:    The SXML representation corresponding to the mode object.
        @rtype:     string
        """
        label = relativeName
        if relativeName.startswith('_AnyMode'):
            # node.name = "_AnyMode"
            result = "\t<CGMAnyModeObject id=\"" + prefix + "." + relativeName + "\" "
            label = "*"
        elif node.type == ModeType.NOMINAL:
            result = "\t<CGMModeObject id=\"" + prefix + "." + relativeName + "\" "
        elif node.type == ModeType.INITIAL:
            result = "\t<CGMInitialModeObject id=\"" + prefix + "." + relativeName + "\" "
        elif node.type == ModeType.ACTIVATION:
            result = "\t<CGMActivationModeObject id=\"" + prefix + "." + relativeName + "\" "

        result += "label=\"" + label + "\""
        if node.invariant is not None:
            result += " invariant=\"" + self.toSLIMVisitor.visit(node.invariant).replace(">", "&gt;").replace("<", "&lt;").replace("=", "&eq;") + "\""
        result += ">\n"

        for transition in node.parent.normal_mode_transitions:
            if transition.source_mode == relativeName:
                result += "\t\t<FEATURE reference=\"" + prefix + ".TC" + str(transition.number) + "-" + str(transition.internal_number) + "-1\" />\n"
            if transition.target_mode == relativeName:
                result += "\t\t<FEATURE reference=\"" + prefix + ".TC" + str(transition.number) + "-" + str(transition.internal_number) + "-2\" />\n"

        if relativeName.startswith('_AnyMode'):
            result += "\t</CGMAnyModeObject>\n"
        elif node.type == ModeType.NOMINAL:
            result += "\t</CGMModeObject>\n"
        elif node.type == ModeType.INITIAL:
            result += "\t</CGMInitialModeObject>\n"
        elif node.type == ModeType.ACTIVATION:
            result += "\t</CGMActivationModeObject>\n"

        return result

    def visit_ModeTransition(self, node, prefix, relativeName):
        """Returns the SXML representation corresponding to the mode transition
        object.
        
        @return:    The SXML representation corresponding to the mode transition
                    object.
        @rtype:     string
        """
        result = "\t<CGMModeTransitionObject id=\"" + prefix + ".T" + relativeName + "\" "
        result += "label=\""
        if node.label.trigger is not None:
            result += node.label.trigger
            if node.label.guard is not None or (node.label.effects is not None and len(node.label.effects) > 0):
                result += " "
        if node.label.guard is not None:
            result += "when " + self.toSLIMVisitor.visit(node.label.guard).replace(">", "&gt;").replace("<", "&lt;").replace("=", "&eq;")
            if node.label.effects is not None and len(node.label.effects) > 0:
                result += " "
        if node.label.effects and len(node.label.effects) > 0 is not None:
            result += "then "
            for i in range (0, len(node.label.effects)):
                effect = node.label.effects[i]
                result += self.toSLIMVisitor.visit(effect).replace(">", "&gt;").replace("<", "&lt;").replace("=", "&eq;")
                if i < len(node.label.effects) - 1:
                    result += "; "

        result += "\">\n"

        result += "\t\t<OBJECT1 reference=\"" + prefix + ".TC" + relativeName + "-" + str(node.internal_number) + "-1\"/>\n"
        result += "\t\t<OBJECT2 reference=\"" + prefix + ".TC" + relativeName + "-" + str(node.internal_number) + "-2\"/>\n"

        result += "\t</CGMModeTransitionObject>\n"

        result += "\t<GenericConnectorObject id=\"" + prefix + ".TC" + relativeName + "-" + str(node.internal_number) + "-1\" label=\"\" direction=\"out\" >\n"
        result += "\t\t<COMPONENT reference=\"" + prefix + "." + node.source_mode + "\" />\n"
        result += "\t</GenericConnectorObject>\n"
        result += "\t<GenericConnectorObject id=\"" + prefix + ".TC" + relativeName + "-" + str(node.internal_number) + "-2\" label=\"\" direction=\"in\" >\n"
        result += "\t\t<COMPONENT reference=\"" + prefix + "." + node.target_mode + "\" />\n"
        result += "\t</GenericConnectorObject>\n"

        return result

    #########################################################


    def visit_ErrorModelImplementation(self, node, prefix, relativeName, type_instance_map, impl_instance_map):
        """Returns the SXML representation corresponding to the error model
        implementation object.
        
        @return:    The SXML representation corresponding to the error model
                    implementation object.
        @rtype:     string
        """

        # Keep two maps: one of component type to the id of the first component instance used, and one of component implementation to the id of the first component instance used. Subsequent use of the same type/implementation have to refer to their id. Also see ticket #100 on Ellidiss Trac.
        classifier = ""
        (c_type, c_impl) = node.name.split(".")
        if c_type in type_instance_map:
            classifier += type_instance_map[c_type] + " "
        else:
            classifier += c_type + ". "
            type_instance_map[c_type] = prefix + "." + relativeName
        if node.name in impl_instance_map:
            classifier += impl_instance_map[node.name]
        else:
            impl_instance_map[node.name] = prefix + "." + relativeName
            classifier += c_impl


        result = "\t<CGMErrorModelObject id=\"" + prefix + "." + relativeName + "\" label=\"" + classifier + "\">\n"

        # handling of any error state, i.e. * -[]-> s1
        generated_transitions = {}
        exists_any_state = False
        for transition in node.error_transitions:
            if transition.number in generated_transitions.keys():
                transition.source_state = "_AnyState"
                exists_any_state = True
            generated_transitions[transition.number] = transition
        node.error_transitions = generated_transitions.values()

        error_type = node.get_referenced_error_type()
        if error_type is not None:
            for state in error_type.states.values():
                result += "\t\t<CHILD reference=\"" + prefix + "." + relativeName + "." + state.name + "\"/>\n"
            # handling of any error transition
            if exists_any_state:
                result += "\t\t<CHILD reference=\"" + prefix + "." + relativeName + "._AnyState\"/>\n"

            for propagation in error_type.error_propagations.values():
                result += "\t\t<FEATURE reference=\"" + prefix + "." + relativeName + "." + propagation.name + "\"/>\n"

        for event in node.error_events.values():
            result += "\t\t<FEATURE reference=\"" + prefix + "." + relativeName + "." + event.name + "\"/>\n"

        result += "\t</CGMErrorModelObject>\n"

        if error_type is not None:
            for state in error_type.states.values():
                result += self.visit(state, prefix + "." + relativeName, state.name, node.error_transitions)
            # handling of any states
            if exists_any_state:
                result += self.visit(state, prefix + "." + relativeName, "*", node.error_transitions)

        for propagation in error_type.error_propagations.values():
            result += self.visit(propagation, prefix + "." + relativeName, propagation.name)

        for event in node.error_events.values():
            result += self.visit(event, prefix + "." + relativeName, event.name)

        for transition in node.error_transitions:
            result += self.visit(transition, prefix + "." + relativeName, str(transition.number))

        return result


    def visit_ErrorState(self, node, prefix, relativeName, error_transitions):
        """Returns the SXML representation corresponding to the error model
        implementation object.
        
        @return:    The SXML representation corresponding to the error model
                    implementation object.
        @rtype:     string
        """
        label = relativeName
        if relativeName == "*":
            result = "\t<CGMAnyStateObject "
            label = "*"
            relativeName = "_AnyState"
            # node.name = "_AnyState"
        elif node.type == ErrorStateType.ERROR:
            result = "\t<CGMErrorStateObject "
        elif node.type == ErrorStateType.ACTIVATION:
            result = "\t<CGMActivationStateObject "
        elif node.type == ErrorStateType.INITIAL:
            result = "\t<CGMInitialStateObject "

        result += "id =\"" + prefix + "." + relativeName + "\" label=\"" + label + "\">\n"

        for transition in error_transitions:
            if transition.source_state == relativeName:
                result += "\t\t<FEATURE reference=\"" + prefix + ".ETC" + str(transition.number) + "-1\" />\n"
            if transition.target_state == relativeName:
                result += "\t\t<FEATURE reference=\"" + prefix + ".ETC" + str(transition.number) + "-2\" />\n"

        if relativeName == "_AnyState":
            result += "\t</CGMAnyStateObject>\n"
        elif node.type == ErrorStateType.ERROR:
            result += "\t</CGMErrorStateObject>\n"
        elif node.type == ErrorStateType.ACTIVATION:
            result += "\t</CGMActivationStateObject>\n"
        elif node.type == ErrorStateType.INITIAL:
            result += "\t</CGMInitialStateObject>\n"

        return result

    def visit_ErrorTransition(self, node, prefix, relativeName):
        """Returns the SXML representation corresponding to the error transition
        object.
        
        @return:    The SXML representation corresponding to the error model
                    implementation object.
        @rtype:     string
        """
        result = "\t<CGMErrorTransitionObject id=\"" + prefix + ".ET" + relativeName + "\" label=\"" + node.trigger + "\">\n"
        result += "\t\t<OBJECT1 reference=\"" + prefix + ".ETC" + relativeName + "-1\"/>\n"
        result += "\t\t<OBJECT2 reference=\"" + prefix + ".ETC" + relativeName + "-2\"/>\n"
        result += "\t</CGMErrorTransitionObject>\n"

        result += "\t<GenericConnectorObject id=\"" + prefix + ".ETC" + relativeName + "-1\" label=\"\" direction=\"out\" >\n"
        result += "\t\t<COMPONENT reference=\"" + prefix + "." + node.source_state + "\" />\n"
        result += "\t</GenericConnectorObject>\n"
        result += "\t<GenericConnectorObject id=\"" + prefix + ".ETC" + relativeName + "-2\" label=\"\" direction=\"in\" >\n"
        result += "\t\t<COMPONENT reference=\"" + prefix + "." + node.target_state + "\" />\n"
        result += "\t</GenericConnectorObject>\n"

        return result

    def visit_ErrorEvent(self, node, prefix, relativeName):
        """Returns the SXML representation corresponding to the error model
        implementation object.
        
        @return:    The SXML representation corresponding to the error model
                    implementation object.
        @rtype:     string
        """
        result = "\t<CGMErrorEventObject id=\"" + prefix + "." + relativeName + "\" errorEventName=\"" + relativeName + "\""
        if node.distribution == DistributionType.POISSON and node.parameter is not None:
            result += " errorEventType=\"poisson\" errorEventValue=\"" + node.parameter + "\""
        result += ">\n"
        result += "\t\t<COMPONENT reference=\"" + prefix + "\"/>\n"
        result += "\t</CGMErrorEventObject>\n"

        return result

    def visit_ErrorPropagation(self, node, prefix, relativeName):
        """Returns the SXML representation corresponding to the error propagation.
        
        @return:    The SXML representation corresponding to the error propagation object.
        @rtype:     string
        """
        result = "\t<CGMErrorPropagationObject id=\"" + prefix + "." + relativeName + "\" label=\"" + relativeName + "\" direction=\""

        if node.type == ErrorPropagationType.IN:
            result += "in"
        if node.type == ErrorPropagationType.OUT:
            result += "out"

        result += "\">\n"
        result += "\t\t<COMPONENT reference=\"" + prefix + "\"/>\n"
        result += "\t</CGMErrorPropagationObject>\n"

        return result
