# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================

from slim_compiler.src.deprecated import deprecated

"""Provides all error classes."""

class ErrorCollector:
    """Provides the tracking of error objects."""
            
    """Initializes the error collector instance."""
    def __init__(self):
        self.errors = []
        self.file_dict = {}
            
    def add(self, error):
        """Adds an error to the list of errors.
        
        @param error: An error object derived from errors.AbstractError."""
                
        self.errors.append(error)
        return

    def get_errors(self):
        """Returns a sequence of strings, each containing the
        errors that were recorded. Each entry is a tuple of five
        elements: (filename, linenum, identifier, message, info). If filename
        is not available, filename and linenum are respectively the
        empty string and -1, info is the empty string."""
        
        self.read_all_files()
        
        res = []

        for error in self.errors:
            if error.line > 0 and error.file in self.file_dict.keys():
                # info is available
                error_line = (self.file_dict[error.file])[error.line - 1]
                error_line = error_line.replace("\n", "")
                error_line = error_line.replace("\r", "")
                offset = self._calculate_line_offset(error_line, error.pos)
                marker = offset + "^"
                info = error_line + "\n" + marker
            else: info = ""
            if not error.file: res.append(("", -1, error.identifier, error.get_message(), info))
            else: res.append((error.file, error.line, error.identifier, error.get_message(), info))
            pass

        res.sort()

        return res
        
    def get_error_id_set(self):
        return set([error.identifier for error in self.errors])
        
    def contains_only(self, error_ids):
        real_error_ids = [error.identifier for error in self.errors]
                
        if set(error_ids) == set(real_error_ids):
            return True
        else:
            print error_ids, real_error_ids
            #print self
            return False
        
    def has_no_errors(self):
        if len(self.errors) > 0:
            #print self
            return False
        else:
            return True

    def to_SXML(self):
        result = '<CGM Version="1.1.0">\n'
        result += '\t<CGMProjectObject id="" classifier="" label="" loadedfiles="" />\n'
        result += '\t<CGMErrors>\n'
        for error in self.errors:
            result += '\t\t<error refersTo=\"'
            for i in range(0, len(error.get_references())):
                reference = error.get_references()[i]
                result += reference.get_ref_identifier()
                if i < len(error.get_references()) - 1:
                    result += ' '
            result += '\" message=\"%s\" />\n' % error.get_sxml_message().replace(">","&gt;").replace("<", "&lt;").replace("=", "&eq;").replace('"', "&quot;")
        result += '\t</CGMErrors>\n'
        result += '</CGM>'
        return result

    def __str__(self):
        """Prints all errors that were recorded."""
        errorString = ""
        for fname, linenum, identifier, msg, marker in self.get_errors():        

            if fname: errorString +=  "%s:%d:(%s): " % (fname, linenum, identifier)
            errorString += msg + "\n\n"
            if marker: errorString += marker + "\n"
            pass
                
        return errorString;
    
    @deprecated    
    def output(self):
        """Prints all errors that were recorded."""
        self.read_all_files()
        #self.print_files()
        for error in self.errors:
            print "%s:%d:%d: %s\n" % (
                        error.file, error.line, error.pos, error.get_message())
            # can error line be printed?
            if error.line > 0:
                error_line = (self.file_dict[error.file])[error.line - 1]
                error_line = error_line.replace("\n", "")
                error_line = error_line.replace("\r", "")
                print "\t" + error_line
                (tabs, spaces) = self._calculate_line_offset(error_line, error.pos)
                marker = ""
                for i in range(tabs + 1):
                    marker += "\t"
                for i in range(spaces):
                    marker += " "
                marker += "^\n"
                print marker
            
    def _calculate_line_offset(self, line, pos):
        spaces = 0
        tabs = 0
        offset = []
        for i in range(pos - 1):
            if line[i] == "\t":
                offset.append("\t")
            else:
                offset.append(" ")
                
        return ''.join(offset)
    
    @deprecated        
    def print_files(self):
        self.read_all_files()
        for file in self.file_dict.keys():
            print file, self.file_dict[file]
            
    def read_all_files(self):
        for error in self.errors:
            import os
            if error.file and error.file not in self.file_dict and \
               os.path.isfile(error.file):
                file = open(error.file, "r")
                self.file_dict[error.file] = file.readlines()
                file.close()

class AbstractError:
    """Encapsulates an abstract error."""
    
    def __init__(self, file, line, pos, identifier = "A-0", references=[]):
        """Initializes an abstract error object.
        
        @param file: The file the error is located in.
        @param line: The line the error is located in.
        @param pos: The position in the line the error is located in.
        @param identifier: The error identifier in the SLIM syntax document.
        """
        self.file = file
        self.line = line
        self.pos = pos
        self.identifier = identifier
        self.references = references
        
    def get_message(self):
        pass
        
    def get_sxml_message(self):
        return self.get_message().replace('"', '&quot;')
    
    def get_references(self):
        return self.references

class PackageAlreadyDeclaredError(AbstractError):
    def __init__(self, file, line, pos, identifier, package):
        AbstractError.__init__(self, file, line, pos, identifier, [package])
        self.package = package
        self.error_string = "Package \"%s\" already declared in %s at line %d:%d."
        self.sxml_error_string = "Package '%s' already declared."
        
    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string \
            % (self.package.name, self.package.file, self.package.line,
               self.package.pos)
        
    def get_sxml_message(self):
        return self.sxml_error_string % self.package.name
                
class ComponentTypeAlreadyDeclaredError(AbstractError):
    def __init__(self, file, line, pos, identifier, type):
        AbstractError.__init__(self, file, line, pos, identifier, [type])
        self.type = type
        self.error_string = \
            "Component Type \"%s\" already declared in package \"%s\" in %s at " \
            + "line %d:%d."
        self.sxml_error_string = "Component Type '%s' already declared in " \
            + "package '%s'."
        
    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string \
            % (self.type.name, self.type.parent.package_name, self.type.file,
               self.type.line, self.type.pos) 

    def get_sxml_message(self):
        return self.sxml_error_string % (self.type.name, self.type.parent.parent.name)
                
class ComponentImplementationAlreadyDeclaredError(AbstractError):
    def __init__(self, file, line, pos, identifier, impl):
        AbstractError.__init__(self, file, line, pos, identifier, [impl])
        self.impl = impl
        self.error_string = \
            "Component Implementation \"%s\" already declared in package \"%s\" " \
            + "in %s at line %d:%d."
        self.sxml_error_string = \
            "Component Implementation '%s' already declared in package '%s'."
        
    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string \
            % (self.impl.name, self.impl.parent.package_name, self.impl.file,
               self.impl.line, self.impl.pos)                
                
    def get_sxml_message(self):
        return self.sxml_error_string % (self.impl.name, self.impl.parent.parent.name)

class ErrorTypeAlreadyDeclaredError(AbstractError):
    def __init__(self, file, line, pos, identifier, type):
        AbstractError.__init__(self, file, line, pos, identifier, [type])
        self.type = type
        self.error_string = \
            "Error model \"%s\" already declared in package \"%s\" in %s at " \
            + "line %d:%d."
        self.sxml_error_string = \
            "Error model type '%s' already declared in package '%s'."
        
    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string \
            % (self.type.name, self.type.parent.name, self.type.file,
               self.type.line, self.type.pos)
        
    def get_sxml_message(self):
        return self.sxml_error_string % (self.type.name, self.type.parent.parent.name)

class ErrorImplementationAlreadyDeclaredError(AbstractError):
    def __init__(self, file, line, pos, identifier, impl):
        AbstractError.__init__(self, file, line, pos, identifier, [impl])
        self.impl = impl
        self.error_string = \
            "Error model implementation \"%s\" already declared in package \"%s\" " \
            + "in %s at line %d:%d."
        self.sxml_error_string = \
            "Error model implementation '%s' already declared in package '%s'."
        
    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % \
            (self.impl.name, self.impl.parent.name, self.impl.file,
             self.impl.line, self.impl.pos)

    def get_sxml_message(self):
        return self.sxml_error_string % (self.impl.name, self.impl.parent.parent.name)
        
class EnumElementsNotDistinctError(AbstractError):
    def __init__(self, file, line, pos, element_name, identifier, obj):
        AbstractError.__init__(self, file, line, pos, identifier, [obj])
        self.element_name = element_name
        self.error_string = "Enumeration contains element \"%s\" multiple times."
        
    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % self.element_name
        
class IntRangeBoundsInvalidError(AbstractError):
    def __init__(self, file, line, pos, identifier, obj):
        AbstractError.__init__(self, file, line, pos, identifier, [obj])
        self.error_string = "Illegal bounds for integer range."
        
    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string
        
class PortAlreadyDeclaredError(AbstractError):
    def __init__(self, file, line, pos, identifier, port):
        AbstractError.__init__(self, file, line, pos, identifier, [port])
        self.port = port
        self.error_string = \
            "Port \"%s\" already declared in component type \"%s\" in %s at " \
            + "line %d:%d."
        self.sxml_error_string = \
            "Port '%s' already declared in component type '%s'."
        
    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % \
            (self.port.name, self.port.parent.name, self.port.file,
             self.port.line, self.port.pos)
            
    def get_sxml_message(self):
        return self.sxml_error_string % (self.port.name, self.port.parent.name)

class DataPortDependencyRelationCyclicError(AbstractError):
    def __init__(self, file, line, pos, impl, dependencies, identifier, subcomp):
        AbstractError.__init__(self, file, line, pos, identifier, [subcomp])
        self.dependencies = dependencies
        self.impl = impl
        self.error_string = "The dependency relation of data ports " + \
            "in the component implementation \"%s\" contains cycles: "
        
        for dependency in self.dependencies:
            self.error_string += dependency + " -> "
            
        self.error_string += self.dependencies[0]
        
    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % (self.impl.name)

class DataPortIsTargetOfFlowsConnsAndEffectsError(AbstractError):
    def __init__(self, file, line, pos, identifier, port, flows_and_conns,
                 transitions):
        AbstractError.__init__(self, file, line, pos, identifier, [port] +
                               flows_and_conns + transitions)
        self.port = port
        self.error_string = "The data port \"%s\" is used as a target for " + \
        "in port connections or flows and " + \
        "transition effects in the component implementation \"%s\" but " + \
        "ports can only be the target in either one of those possibilities."
        self.sxml_error_string = "The data port \"%s\" is used as a target for " + \
        "in port connections or flows and " + \
        "transition effects. but " + \
        "ports can only be the target in either one of those possibilities."
        
    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % (self.port.name, self.port.parent.name)

    def get_sxml_message(self):
        return self.sxml_error_string % (self.port.name)

class EndIdentifierDoesNotMatchError(AbstractError):
    def __init__(self, file, line, pos, identifier, obj, type, name,
                 end_identifier):
        AbstractError.__init__(self, file, line, pos, identifier, [obj])
        self.name = name
        self.end_identifier = end_identifier
        self.type = type
        self.error_string = \
        "End identifier \"%s\" must match the %s name, namely \"%s\"."
    
    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string \
        % (self.end_identifier, self.type, self.name)

class EndIdentifierOfPackageDoesNotMatchError(EndIdentifierDoesNotMatchError):
    def __init__(self, file, line, pos, identifier, package, end_identifier):
        EndIdentifierDoesNotMatchError.__init__(self, file, line, pos,
                                                identifier, package, "package",
                                                package.name, end_identifier)
        
class EndIdentifierOfComponentTypeDoesNotMatchError(EndIdentifierDoesNotMatchError):
    def __init__(self, file, line, pos, identifier, type, end_identifier):
        EndIdentifierDoesNotMatchError.__init__(self, file, line, pos,
                                                identifier, type,
                                                "component type", type.name,
                                                end_identifier)
                
class EndIdentifierOfComponentImplementationDoesNotMatchError(EndIdentifierDoesNotMatchError):
    def __init__(self, file, line, pos, identifier, impl, end_identifier):
        EndIdentifierDoesNotMatchError.__init__(self, file, line, pos,
                                                identifier, impl,
                                                "component implementation",
                                                impl.name, end_identifier)
        
class EndIdentifierOfErrorTypeDoesNotMatchError(EndIdentifierDoesNotMatchError):
    def __init__(self, file, line, pos, identifier, type, end_identifier):
        EndIdentifierDoesNotMatchError.__init__(self, file, line, pos,
                                                identifier, type, 
                                                "error model type", type.name,
                                                end_identifier)
                
class EndIdentifierOfErrorImplementationDoesNotMatchError(EndIdentifierDoesNotMatchError):
    def __init__(self, file, line, pos, identifier, impl, end_identifier):
        EndIdentifierDoesNotMatchError.__init__(self, file, line, pos,
                                                identifier, impl,
                                                "error model implementation",
                                                impl.name, end_identifier)
        
class ComponentImplementationRefersToUnknownTypeError(AbstractError):
    def __init__(self, file, line, pos, identifier, impl):
        AbstractError.__init__(self, file, line, pos, identifier, [impl])
        self.impl = impl
        self.error_string = \
        "Component implementation \"%s\" refers to the unknown component type " \
        + "\"%s\"."
               
    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % (self.impl.name, self.impl.name.split(".")[0])
        
class ComponentImplementationCrossReferenceToTypeError(AbstractError):
    def __init__(self, file, line, pos, identifier, impl):
        AbstractError.__init__(self, file, line, pos, identifier, [impl])
        self.impl = impl
        self.error_string = \
        "Component implementation \"%s\" has a cross-reference to the " + \
        "component type \"%s\"."
               
    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % (self.impl.name, self.impl.name.split(".")[0])
        
class ErrorImplementationRefersToUnknownTypeError(AbstractError):
    def __init__(self, file, line, pos, identifier, impl):
        AbstractError.__init__(self, file, line, pos, identifier, [impl])
        self.impl = impl
        self.error_string = \
        "Error model implementation \"%s\" refers to the unknown error " + \
        "model type \"%s\"."
        
    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % (self.impl.name, self.impl.name.split(".")[0])
        
class ErrorImplementationCrossReferenceToTypeError(AbstractError):
    def __init__(self, file, line, pos, identifier, impl):
        AbstractError.__init__(self, file, line, pos, identifier, [impl])
        self.impl = impl
        self.error_string = \
        "Error model implementation \"%s\" has a cross-reference to the " + \
        "error model type \"%s\"."
               
    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % (self.impl.name, self.impl.name.split(".")[0])        

class ErrorStatesInBothTypeAndImplementationError(AbstractError):
    def __init__(self, file, line, pos, identifier, type, impl):
        AbstractError.__init__(self, file, line, pos, identifier, [type, impl])
        self.type = type
        self.impl = impl
        self.error_string = \
        "Both the error type \"%s\" and the error implementation \"%s\" define error states. " + \
        "Declaration must occur in either the type or the implementation." 
        self.sxml_error_string = \
        "Both the error type \"%s\" and the error implementation \"%s\" define error states."
               
    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % (self.type.name, self.impl.name)
        
    def get_sxml_message(self):
        return self.sxml_error_string % (self.type.name,
                                         self.impl.name)
        
class ErrorStateInImplAlreadyDeclaredInTypeError(AbstractError):
    def __init__(self, file, line, pos, identifier, impl, new_state, old_state):
        AbstractError.__init__(self, file, line, pos, identifier, [new_state])
        self.impl = impl
        self.old_state = old_state
        self.error_string = \
        "Error state \"%s\" in \"%s\" already declared in error model type \"%s\" in " + \
        "%s at line %d:%d."
        self.sxml_error_string = \
        "Error state \"%s\" already declared in error model type '%s'."
               
    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % (self.old_state.name, self.impl.name,
                                    self.old_state.parent.name,
                                    self.old_state.file, self.old_state.line,
                                    self.old_state.pos)
        
    def get_sxml_message(self):
        return self.sxml_error_string % (self.old_state.name,
                                         self.old_state.parent.name)
        
class ErrorStateTransitionAlreadyDeclaredError(AbstractError):
    def __init__(self, file, line, pos, identifier, impl, old_trans):
        AbstractError.__init__(self, file, line, pos, identifier, [old_trans])
        self.impl = impl
        self.old_trans = old_trans
        self.error_string = \
        "The error state transition %s -[%s]-> %s in the error model " + \
        "implementation \"%s\" was already declared in %s at line %d:%d."
        self.sxml_error_string =\
        "The error state transition %s -[%s]-> %s was already declared."
        
    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % (self.old_trans.source_state,
                                    self.old_trans.trigger,
                                    self.old_trans.target_state,
                                    self.impl.name, self.old_trans.file,
                                    self.old_trans.line, self.old_trans.pos)
        
    def get_sxml_message(self):
        return self.sxml_error_string % (self.old_trans.source_state,
                                         self.old_trans.trigger,
                                         self.old_trans.target_state)
        
class ErrorStateHasDifferentTypesOfLeavingTransitionsError(AbstractError):
    def __init__(self, file, line, pos, identifier, state, prob_trans,
                 non_prob_trans):
        AbstractError.__init__(self, file, line, pos, identifier, prob_trans + 
                               non_prob_trans)
        self.state = state
        self.error_string = \
        "There are probabilistic and non-probabilistic transitions leaving " + \
        "the error state \"%s\"."
        
    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % (self.state.name)
        
class ErrorTransitionsLeavingStateAreNondeterministicError(AbstractError):
    def __init__(self, file, line, pos, identifier, state, trigger):
        AbstractError.__init__(self, file, line, pos, identifier, [state])
        self.state = state
        self.trigger = trigger
        self.error_string = \
        "The choice of transition leaving the error state \"%s\" is " + \
        "nondeterministic, because the trigger \"%s\" is used multiple " + \
        "times."
        
    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % (self.state.name, self.trigger)
    
class ErrorStateNotReachableError(AbstractError):
    def __init__(self, file, line, pos, identifier, state):
        AbstractError.__init__(self, file, line, pos, identifier, [state])
        self.state = state
        self.error_string = \
        "The error state \"%s\" is not reachable."
        
    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % (self.state.name)
        
class PropagationNameAlreadyUsedAsStateNameError(AbstractError):
    def __init__(self, file, line, pos, identifier, propagation):
        AbstractError.__init__(self, file, line, pos, identifier, [propagation])
        self.propagation = propagation
        self.error_string = \
        "Propagation name \"%s\" already used as an error state name."
        
    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % (self.propagation.name)
        
class EventPortHasFanOutToSameTargetComponentError(AbstractError):
    def __init__(self, file, line, pos, identifier, conns, source_port_name,
                 subcomp_name, target_port_names):
        AbstractError.__init__(self, file, line, pos, identifier, conns)
        self.source_port_name = source_port_name
        self.subcomp_name = subcomp_name
        self.target_port_names = target_port_names
        if self.subcomp_name is not None:
            self.error_string = \
            "Port connection declares a fan out for the source port \"%s\" " + \
            "to the subcomponent \"%s\" as the same source port is already " + \
            "connected to the following ports of that subcomponent: "
        else:
            self.error_string = \
            "Port connection declares a fan out for the source port \"%s\" " + \
            "to the current component as the same source port is already " + \
            "connected to the ports: "
            
        for i in range(0, len(target_port_names) - 1):
            self.error_string += target_port_names[i]
            if i < len(target_port_names) - 1:
                self.error_string += ", "
            else:
                self.error_string += "."
            
        
    def get_message(self):
        """Retrieves the error message associated with the error"""
        if self.subcomp_name is not None:
            return self.error_string % (self.source_port_name, self.subcomp_name)
        else:
            return self.error_string % (self.source_port_name)
        
class EventPortIsTargetOfPassiveAndActivePortConnections(AbstractError):
    def __init__(self, file, line, pos, identifier, conn1, conn2):
        AbstractError.__init__(self, file, line, pos, identifier, [conn1, conn2])
        self.conn1 = conn1
        self.conn2 = conn2
        self.error_string = \
        "Event Port \"%s\" is target of passive and active port connections, " \
        "because of the previous port connection\n%s."
        
    def get_message(self):
        """Retrieves the error message associated with the error"""
        from slim_compiler.src.dataStructure.visitors.printVisitor import toSLIMVisitor
        return self.error_string % (self.conn1.target_port, toSLIMVisitor().visit(self.conn1,"event"))
                
class ComponentCategoryCannotDeclarePortsError(AbstractError):
    def __init__(self, file, line, pos, identifier, type, category):
        AbstractError.__init__(self, file, line, pos, identifier, [type])
        self.type = type
        self.category = category
        self.error_string = \
        "Component type \"%s\" of category \"%s\" declares ports, but " \
        + "components of that category cannot declare ports."
        
    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % (self.type.name, self.category)
        
class AnalogDataPortsNotSupportedError(AbstractError):
    def __init__(self, file, line, pos, identifier, port, data_type):
        AbstractError.__init__(self, file, line, pos, identifier, [port])
        self.port = port
        self.data_type = data_type
        self.error_string = \
        "Port \"%s\" is of type \"%s\", but data ports cannot be of that type."
        
    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % (self.port.name, self.data_type)

class ObservablePortInFDIRComponentError(AbstractError):
    def __init__(self, file, line, pos, identifier, port):
        AbstractError.__init__(self, file, line, pos, identifier, [port])
        self.port = port
        self.error_string = \
        "FDIR type tags port \"%s\" as observable."

    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % (self.port.name)

class ObservablePortNotBoolError(AbstractError):
    def __init__(self, file, line, pos, identifier, port):
        AbstractError.__init__(self, file, line, pos, identifier, [port])
        self.port = port
        self.error_string = \
        "The port \"%s\" is tagged as observable, but is not of boolean type."

    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % (self.port.name)

class AlarmPortInNonFDIRComponentError(AbstractError):
    def __init__(self, file, line, pos, identifier, port):
        AbstractError.__init__(self, file, line, pos, identifier, [port])
        self.port = port
        self.error_string = \
        "Non-FDIR type tags port \"%s\" as alarm."

    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % (self.port.name)

class AlarmPortNotBoolError(AbstractError):
    def __init__(self, file, line, pos, identifier, port):
        AbstractError.__init__(self, file, line, pos, identifier, [port])
        self.port = port
        self.error_string = \
        "The port \"%s\" is tagged as alarm, but is not of boolean type."

    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % (self.port.name)

class ObservableAlarmPortComponentError(AbstractError):
    def __init__(self, file, line, pos, identifier, port):
        AbstractError.__init__(self, file, line, pos, identifier, [port])
        self.port = port
        self.error_string = \
        "Port \"%s\" can't be of both type observable and alarm."

    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % (self.port.name)
        
class ImplementationCategoryMustMatchTypeCategoryError(AbstractError):
    def __init__(self, file, line, pos, identifier, impl, category):
        AbstractError.__init__(self, file, line, pos, identifier, [impl])
        self.impl = impl
        self.category = category
        self.error_string = \
        "Component implementation \"%s\" must be of the same category as the " \
        + "corresponding component type \"%s\", namely \"%s\"."
        
    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % \
        (self.impl.name, self.impl.name.split(".")[0], self.category)
        
class SubcomponentAlreadyDeclaredError(AbstractError):
    def __init__(self, file, line, pos, identifier, subcomponent):
        AbstractError.__init__(self, file, line, pos, identifier, [subcomponent])
        self.subcomponent = subcomponent
        self.error_string = \
        "Subcomponent \"%s\" already declared in %s at line %d:%d."
        self.sxml_error_string = \
        "Subcomponent \"%s\" declared multiple times."
        
    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string \
        % (self.subcomponent.name, self.subcomponent.file,
           self.subcomponent.line, self.subcomponent.pos)
        
    def get_sxml_message(self):
        return self.sxml_error_string % (self.subcomponent.name)
        
class SubcomponentReferencesUnknownPackageError(AbstractError):
    def __init__(self, file, line, pos, identifier, subcomponent, package_name):
        AbstractError.__init__(self, file, line, pos, identifier, [subcomponent])
        self.subcomponent = subcomponent
        self.package_name = package_name
        self.error_string = \
        "Subcomponent declaration of \"%s\" refers to the unknown package \"%s\"."
        
    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string \
        % (self.subcomponent.name, self.package_name)

class SubcomponentReferencesUnknownTypeError(AbstractError):
    def __init__(self, file, line, pos, identifier, subcomponent, type_name):
        AbstractError.__init__(self, file, line, pos, identifier, [subcomponent])
        self.subcomponent = subcomponent
        self.type_name = type_name
        self.error_string = \
        "Subcomponent declaration of \"%s\" refers to the unknown component " \
        + "type \"%s\"."
        
    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string \
        % (self.subcomponent.name, self.type_name)

class SubcomponentReferencesUnknownImplementationError(AbstractError):
    def __init__(self, file, line, pos, identifier, subcomponent, impl_name):
        AbstractError.__init__(self, file, line, pos, identifier, [subcomponent])
        self.subcomponent = subcomponent
        self.impl_name = impl_name
        self.error_string = \
        "Subcomponent declaration of \"%s\" refers to the unknown component " \
        + "implementation \"%s\"."
        
    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string \
        % (self.subcomponent.name, self.impl_name)

class SubcomponentReferencesNotUniqueComponentTypeError(AbstractError):
    def __init__(self, file, line, pos, identifier, subcomponent, type_name):
        AbstractError.__init__(self, file, line, pos, identifier, [subcomponent])
        self.subcomponent = subcomponent
        self.type_name = type_name
        self.error_string = \
        "Subcomponent declaration of \"%s\" refers to the component type " \
        + "\"%s\", but there is more than one component implementation for " \
        + "this type, so one specific one must be referenced."
        
    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % (self.subcomponent.name, self.type_name)

class SubcomponentReferencesComponentTypeWithoutImplementationsError(AbstractError):
    def __init__(self, file, line, pos, identifier, subcomponent, type_name):
        AbstractError.__init__(self, file, line, pos, identifier, [subcomponent])
        self.subcomponent = subcomponent
        self.type_name = type_name
        self.error_string = \
        "Subcomponent declaration of \"%s\" refers to the component type " \
        + "\"%s\", but there is no component implementation for this type."
        
    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % (self.subcomponent.name, self.type_name)

class SubcomponentReferencesUnknownComponentTypeError(AbstractError):
    def __init__(self, file, line, pos, identifier, subcomponent, type_name):
        AbstractError.__init__(self, file, line, pos, identifier, [subcomponent])
        self.subcomponent = subcomponent
        self.type_name = type_name
        self.error_string = \
        "Subcomponent declaration of \"%s\" refers to the unknown component " \
        + "type \"%s\"."
        
    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % (self.subcomponent.name, self.type_name)
        
class SubcomponentDeclarationMustMatchTypeCategoryError(AbstractError):
    def __init__(self, file, line, pos, identifier, subcomponent, type_name,
                 category):
        AbstractError.__init__(self, file, line, pos, identifier, [subcomponent])
        self.subcomponent = subcomponent
        self.type_name = type_name
        self.category = category
        self.error_string = \
        "Subcomponent declaration of \"%s\" refers to the component type " \
        + "\"%s\", and therefore must match its category, namely \"%s\"."
        
    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string \
        % (self.subcomponent.name, self.type_name, self.category)
        
class SubcomponentDeclarationMustMatchImplementationCategoryError(AbstractError):
    def __init__(self, file, line, pos, identifier, subcomponent, impl, category):
        AbstractError.__init__(self, file, line, pos, identifier, [subcomponent])
        self.subcomponent = subcomponent
        self.impl = impl
        self.category = category
        self.error_string = \
        "Subcomponent declaration of \"%s\" refers to the component " \
        + "implementation \"%s\", and therefore must match its category, namely" \
        + " \"%s\"."
        
    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string \
        % (self.subcomponent.name, self.impl.name, self.category)
        
class ImplementationCategoryDoesNotSupportSubcomponentsError(AbstractError):
    def __init__(self, file, line, pos, identifier, impl, impl_category,
                 unsupported_category):
        AbstractError.__init__(self, file, line, pos, identifier, [impl])
        self.impl = impl
        self.impl_category = impl_category
        self.unsupported_category = unsupported_category
        self.error_string = \
        "Component implementation \"%s\" of category \"%s\" does not support " \
        + "subcomponents of category \"%s\"."
        
    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string \
            % (self.impl.name, self.impl_category, self.unsupported_category)
        
class SubcomponentDeclarationUsesUnsupportedBindingError(AbstractError):
    def __init__(self, file, line, pos, identifier, subcomponent, category, binding):
        AbstractError.__init__(self, file, line, pos, identifier, [subcomponent])
        self.subcomponent = subcomponent
        self.category = category
        self.binding = binding
        self.error_string = \
        "Subcomponent declaration of \"%s\" of category \"%s\" cannot use " \
        + "\"%s\" binding."
        
    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string \
            % (self.subcomponent.name, self.category, self.binding)
            
class SubcomponentBindingReferencesUnknownSubcomponentError(AbstractError):
    def __init__(self, file, line, pos, identifier, subcomponent,
                 other_subcomponent_name):
        AbstractError.__init__(self, file, line, pos, identifier, [subcomponent])
        self.subcomponent = subcomponent
        self.other_subcomponent_name = other_subcomponent_name
        self.error_string = "Subcomponent declaration of \"%s\" refers to " \
        + " unknown subcomponent \"%s\"."

    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % (self.subcomponent.name,
                                    self.other_subcomponent_name)    
    
class SubcomponentBindingReferencesWrongCategoryError(AbstractError):
    def __init__(self, file, line, pos, identifier, subcomponent,
                 subcomponent_reference, wrong_category, right_category,
                 binding):
        AbstractError.__init__(self, file, line, pos, identifier, [subcomponent])
        self.subcomponent = subcomponent
        self.subcomponent_reference = subcomponent_reference
        self.wrong_category = wrong_category
        self.right_category = right_category
        self.binding = binding
        self.error_string = "Subcomponent declaration of \"%s\" uses binding \"%s" \
        + " with subcomponent \"%s\", which is of category \"%s\", but \"%s\"" \
        + " must refer to subcomponent of category \"%s\"."
        
    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % (self.subcomponent.name, self.binding,
                                   self.subcomponent_reference,
                                   self.wrong_category,
                                   self.binding, self.right_category)
        
class SubcomponentBusAccessesItselfError(AbstractError):
    def __init__(self, file, line, pos, identifier, subcomponent):
        AbstractError.__init__(self, file, line, pos, identifier, [subcomponent])
        self.subcomponent = subcomponent
        self.error_string = "Subcomponent declaration of \"%s\" of category " \
        + "\"bus\" cannot use \"accesses\" binding to itself."
        
    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % (self.subcomponent.name)
    
class SubcomponentIsActiveInModeWhereBindingIsNotError(AbstractError):
    def __init__(self, file, line, pos, identifier, subcomponent,
                 binding_partner_name, mode_name):
        AbstractError.__init__(self, file, line, pos, identifier, [subcomponent])
        self.subcomponent = subcomponent
        self.binding_partner_name = binding_partner_name
        self.mode_name = mode_name
        self.error_string = "Subcomponent \"%s\" is bound to subcomponent " \
        + "\"%s\", but is active in mode \"%s\", where \"%s\" is not active."
        
    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % (self.subcomponent.name,
                                    self.binding_partner_name,
                                    self.mode_name, self.binding_partner_name)    

class InModesListElementsNotDistinctError(AbstractError):
    def __init__(self, file, line, pos, identifier, mode, obj):
        AbstractError.__init__(self, file, line, pos, identifier, [obj])
        self.mode = mode
        self.error_string = "In Modes list contains element \"%s\" multiple times."
        
    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % self.mode

class InModesListRefersToUnkonwnModeError(AbstractError):
    def __init__(self, file, line, pos, identifier, unknown_mode, obj):
        AbstractError.__init__(self, file, line, pos, identifier, [obj])
        self.unknown_mode = unknown_mode
        self.error_string = "In Modes list refers to unknown mode \"%s\"."
        
    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % (self.unknown_mode)
    
class DataSubcomponentsDataPortsNotDistinctError(AbstractError):
    def __init__(self, file, line, pos, identifier, impl, duplicates, objects):
        AbstractError.__init__(self, file, line, pos, identifier, objects)
        self.impl = impl
        self.duplicates = duplicates
        self.error_string = "In the component implementation \"%s\" the " + \
        "identifiers of all subcomponents and the identifiers of all data " + \
        "ports declared in the corresponding type specification have to be " + \
        "distinct. The following identifiers appear" + \
        " multiple times:\n"

        for (i,name) in enumerate(self.duplicates):
            self.error_string += "\t" + str(i + 1) + ". " + name + "\n"

    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % (self.impl.name)
    
class EnumValuesDataSubcomponentsNotDistinctError(AbstractError):
    def __init__(self, file, line, pos, identifier, impl, duplicates):
        AbstractError.__init__(self, file, line, pos, identifier, [impl])
        self.impl = impl
        self.duplicates = duplicates
        self.error_string = "In the component implementation \"%s\" the " + \
        "identifiers of all enum values and the identifiers of all data " + \
        "subcomponents and the identifiers of all data ports in the " + \
        "corresponding component type have to be distinct. The following " + \
        "identifiers appear multiple times:\n"

        for (i,name) in enumerate(duplicates):
            self.error_string += "\t" + str(i + 1) + ". " + name + "\n"

    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % (self.impl.name)
    
class EnumValuesDataPortsNotDistinctError(AbstractError):
    def __init__(self, file, line, pos, identifier, type, duplicates,
                 error_references):
        AbstractError.__init__(self, file, line, pos, identifier,
                               error_references)
        self.type = type
        self.duplicates = duplicates
        self.error_string = "In the component type \"%s\" the " + \
        "identifiers of all enum values and the identifiers of all data " + \
        "ports have to be distinct. The following identifiers " + \
        "appear multiple times:\n"

        for (i,name) in enumerate(self.duplicates):
            self.error_string += "\t" + str(i + 1) + ". " + name + "\n"

    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % (self.type.name)
    
class EnumValuesNotDistinctAndNotIdenticalInTypeError(AbstractError):
    def __init__(self, file, line, pos, identifier, type, violating_pairs,
                 error_references):
        AbstractError.__init__(self, file, line, pos, identifier,
                               error_references)
        self.type = type
        self.violating_pairs = violating_pairs
        self.error_string = "The list of value identifiers of enum " + \
        "types in a component type have to be distinct or identical. The " + \
        "value identifiers of the enums associated to the following " + \
        "identifiers violate this in the component type \"%s\":\n"

        for (i,(port1, port2)) in enumerate(violating_pairs):
            self.error_string += "\t" + str(i + 1) + ". " + port1.name + " and " + \
            port2.name + "\n"

    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % (self.type.name)    
    
class EnumValuesNotDistinctAndNotIdenticalInImplError(AbstractError):
    def __init__(self, file, line, pos, identifier, impl, violating_pairs,
                 error_references):
        AbstractError.__init__(self, file, line, pos, identifier,
                               error_references)
        self.impl = impl
        self.violating_pairs = violating_pairs
        self.error_string = "The list of value identifiers of enum " + \
        "types in a component implementation " + \
        "have to be distinct or identical. The value identifiers " + \
        "of the enums associated to the following identifiers violate this " + \
        "in the component implementation \"%s\":\n"

        for (i,(entry1, entry2)) in enumerate(violating_pairs):
            self.error_string += "\t" + str(i + 1) + ". " + entry1.name + " and " + \
            entry2.name + "\n"

    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % (self.impl.name)

class EnumValuesNotDistinctAndNotIdenticalInImplAndTypeError(AbstractError):
    def __init__(self, file, line, pos, identifier, impl, violating_pairs,
                 error_references):
        AbstractError.__init__(self, file, line, pos, identifier,
                               error_references)
        self.impl = impl
        self.violating_pairs = violating_pairs
        self.error_string = "The list of value identifiers of enum " + \
        "types in a component implementation and the corresponding component type " + \
        "have to be distinct or identical. The value identifiers " + \
        "of the enums associated to the following identifiers violate this " + \
        "in the component implementation \"%s\":\n"

        for (i,(entry1, entry2)) in enumerate(violating_pairs):
            self.error_string += "\t" + str(i + 1) + ". " + entry1.name + " and " + \
            entry2.name + "\n"

    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % (self.impl.name)

class RecursiveComponentHierarchyError(AbstractError):
    def __init__(self, file, line, pos, identifier, impl, trace):
        AbstractError.__init__(self, file, line, pos, identifier, [impl])
        self.impl = impl
        self.trace = trace
        
    def get_message(self):
        """Retrieves the error message associated with the error"""
        result = "The component hierarchy is recursive, namely component " + \
                 "implementation \"" + self.impl.name + "\" is an " + \
                 "(indirect) subcomponent of itself. See the following trace " + \
                 "for details:\n"
                                  
        for i in range(len(self.trace)):
            result += "\t"
            if i > 0:
                result += "-> "
            result += self.trace[i]
            if i < len(self.trace) - 1:
                if i > 0:
                    result += " which"
                result += " has a subcomponent referencing " + \
                          "the component implementation\n"
            else:
                result += "."
        
        return result

class PortConnectionReferencesUnknownSubcomponentError(AbstractError):
    def __init__(self, file, line, pos, identifier, conn, subcomponent_name):
        AbstractError.__init__(self, file, line, pos, identifier, [conn])
        self.subcomponent_name = subcomponent_name
        self.error_string = "Port Connection references unknown subcomponent " + \
        "\"%s\"."

    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % (self.subcomponent_name) 

class EventPortConnectionReferencesDataPortError(AbstractError):
    def __init__(self, file, line, pos, identifier, conn, port_name):
        AbstractError.__init__(self, file, line, pos, identifier, [conn])
        self.port_name = port_name
        self.error_string = "Event Port Connection references data port " + \
        "\"%s\"."

    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % (self.port_name) 

class PassiveEventPortConnectionIsNotOutToInError(AbstractError):
    def __init__(self, file, line, pos, identifier, conn):
        AbstractError.__init__(self, file, line, pos, identifier, [conn])
        self.error_string = "Passive Event Port Connection is not connecting " + \
        "an out event port with an in event port."

    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string

class PortConnectionReferencesUnknownPortError(AbstractError):
    def __init__(self, file, line, pos, identifier, conn, port_name):
        AbstractError.__init__(self, file, line, pos, identifier, [conn])
        self.port_name = port_name
        self.error_string = "Port Connection references unknown port " + \
        "\"%s\"."

    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % (self.port_name) 

class FlowReferencesUnknownIncomingDataPortError(AbstractError):
    def __init__(self, file, line, pos, identifier, port_name, flow):
        AbstractError.__init__(self, file, line, pos, identifier, [flow])
        self.port_name = port_name
        self.error_string = "Flow refers to unknown incoming data port \"%s\"." 

    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % (self.port_name) 

class FlowReferencesUnknownIncomingDataPortOfSubcompError(AbstractError):
    def __init__(self, file, line, pos, identifier, subcomp_name, port_name,
                 flow):
        AbstractError.__init__(self, file, line, pos, identifier, [flow])
        self.port_name = port_name
        self.subcomp_name = subcomp_name
        self.error_string = "Flow refers to unknown incoming data port " + \
        "\"%s\" of subcomponent \"%s\"." 

    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % (self.port_name, self.subcomp_name) 
    
class FlowReferencesUnknownOutgoingDataPortError(AbstractError):
    def __init__(self, file, line, pos, identifier, port_name, flow):
        AbstractError.__init__(self, file, line, pos, identifier, [flow])
        self.port_name = port_name
        self.error_string = "Flow refers to unknown outgoing data port \"%s\"." 

    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % (self.port_name) 

class FlowReferencesUnknownOutgoingDataPortOfSubcompError(AbstractError):
    def __init__(self, file, line, pos, identifier, subcomp_name, port_name,
                 flow):
        AbstractError.__init__(self, file, line, pos, identifier, [flow])
        self.port_name = port_name
        self.subcomp_name = subcomp_name
        self.error_string = "Flow refers to unknown outgoing data port " + \
        "\"%s\" of subcomponent \"%s\"." 

    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % (self.port_name, self.subcomp_name) 

class FlowReferencesUnknownSubcomponentError(AbstractError):
    def __init__(self, file, line, pos, identifier, subcomp_name, flow):
        AbstractError.__init__(self, file, line, pos, identifier, [flow])
        self.subcomp_name = subcomp_name
        self.error_string = "Flow refers to unknown subcomponent \"%s\"." 

    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % (self.subcomp_name) 

class PortConnectionHasSameSourceAsTargetError(AbstractError):
    def __init__(self, file, line, pos, identifier, conn):
        AbstractError.__init__(self, file, line, pos, identifier, [conn])
        self.error_string = "Port Connection has the same source and " + \
        "target component."

    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string 

class DataPortConnectionReferencesEventPortError(AbstractError):
    def __init__(self, file, line, pos, identifier, conn, port_name):
        AbstractError.__init__(self, file, line, pos, identifier, [conn])
        self.port_name = port_name
        self.error_string = "Data Port Connection references event port " + \
        "\"%s\"."

    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % (self.port_name) 

class DataPortConnectionReferencesDifferentDataTypesError(AbstractError):
    def __init__(self, file, line, pos, identifier, conn):
        AbstractError.__init__(self, file, line, pos, identifier, [conn])
        self.error_string = "Data Port Connection references data ports " + \
        "with different data types."

    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string

class PortConnectionReferencesInactiveSubcomponentError(AbstractError):
    def __init__(self, file, line, pos, identifier, subcomponent_name, conn,
                 mode):
        AbstractError.__init__(self, file, line, pos, identifier, [conn])
        self.subcomponent_name = subcomponent_name
        self.mode = mode
        self.error_string = "Port Connection is active in mode \"%s\"" + \
        " and references subcomponent \"%s\" which is inactive in that mode."

    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % (self.mode, self.subcomponent_name)

class PortFlowReferencesInactiveSubcomponentError(AbstractError):
    def __init__(self, file, line, pos, identifier, subcomponent_name, mode,
                 flow):
        AbstractError.__init__(self, file, line, pos, identifier, [flow])
        self.subcomponent_name = subcomponent_name
        self.mode = mode
        self.error_string = "Flow is active in mode \"%s\"" + \
        " and references subcomponent \"%s\" which is inactive in that mode."

    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % (self.mode, self.subcomponent_name)

class PortConnectionViolatatesPossibleTopologyError(AbstractError):
    def __init__(self, file, line, pos, identifier, message, conn):
        AbstractError.__init__(self, file, line, pos, identifier, [conn])
        self.message = message
        self.error_string = "Port Connection is invalid, because it " + \
        "violates the rules for possible connection topologies: %s."

    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string%self.message

class DataPortHasFanInError(AbstractError):
    def __init__(self, file, line, pos, identifier, obj, port_name, mode_name):
        AbstractError.__init__(self, file, line, pos, identifier, [obj])
        self.port_name = port_name
        self.mode_name = mode_name
        self.error_string = "Data Port \"%s\" has a fan-in as it is target " + \
        "of multiple connections and/or flows in mode \"%s\"."

    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % (self.port_name, self.mode_name)

class OutDataPortNotTargetInModeMissingDefaultValueError(AbstractError):
    def __init__(self, file, line, pos, identifier, port, mode_name):
        AbstractError.__init__(self, file, line, pos, identifier, [port])
        self.port = port
        self.mode_name = mode_name
            
        self.error_string = "Out Data Port \"%s\" misses a default value, " + \
        "because it is not the target of a flow or port connection in " + \
        "mode \"%s\"."

    def get_message(self):
        """Retrieves the error message associated with the er.or"""
        return self.error_string % (self.port.name, self.mode_name)
    
class OutDataPortNotTargetInStartingModeMissingDefaultValueError(AbstractError):
    def __init__(self, file, line, pos, identifier, port, mode_name):
        AbstractError.__init__(self, file, line, pos, identifier, [port])
        self.port = port
        self.mode_name = mode_name
            
        self.error_string = "Data Port \"%s\" misses a default value, " + \
        "because it is not target of a flow or port connection in the " + \
        "starting mode \"%s\"."

    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % (self.port.name, self.mode_name)

# TODO: This error has been obsoleted by the inverse InDataPortOfRootComponentHasDefaultValueError error
# See ticket 527
class InDataPortOfRootComponentMissingDefaultValueError(AbstractError):
    def __init__(self, file, line, pos, identifier, port):
        AbstractError.__init__(self, file, line, pos, identifier)
        self.port = port
            
        self.error_string = "In Data Port \"%s\" misses a default value, " + \
        "because it is an incoming port of a possible root component."

    def get_message(self):
        """Retrieves the error message associated with the error"""
        result = self.error_string % (self.port.name)
        return result

class DataPortOfRootComponentHasDefaultValueError(AbstractError):
    def __init__(self, file, line, pos, identifier, port):
        AbstractError.__init__(self, file, line, pos, identifier)
        self.port = port
            
        self.error_string = "Data Port \"%s\" specifies a default value, " + \
        "but this is not allowed for a root component."

    def get_message(self):
        """Retrieves the error message associated with the error"""
        result = self.error_string % (self.port.name)
        return result
    
class InDataPortNotTargetInModeMissingDefaultValueError(AbstractError):
    def __init__(self, file, line, pos, identifier, subcomponent, port_name,
                 mode_names):
        AbstractError.__init__(self, file, line, pos, identifier,
                               [subcomponent])
        self.port_name = port_name
        self.mode_names = mode_names
            
        self.error_string = "In Data Port \"%s\" misses a default value, " + \
        "because it not target of a port connection in the " + \
        "following mode(s): "

    def get_message(self):
        """Retrieves the error message associated with the error"""
        result = self.error_string % (self.port_name)
        for mode_name in self.mode_names:
            result += "\"" + mode_name + "\", "
        result = result[0 : len(result) - 2] + "."
        return result

class DataSubcomponentMissingDefaultValueError(AbstractError):
    def __init__(self, file, line, pos, identifier, subcomponent):
        AbstractError.__init__(self, file, line, pos, identifier, [subcomponent])
        self.subcomponent = subcomponent
        self.error_string = "Data Subcomponent \"%s\" misses a default " + \
        "value."

    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % (self.subcomponent.name)

class ModeAlreadyDeclaredError(AbstractError):
    def __init__(self, file, line, pos, identifier, old_mode):
        AbstractError.__init__(self, file, line, pos, identifier, [old_mode])
        self.old_mode = old_mode
        self.error_string = \
        "Mode \"%s\" already declared in component implementation \"%s\" in " + \
        "%s at line %d:%d."
        self.sxml_error_string = \
        "Mode '%s' declared multiple times in component implementation '%s'."
        
    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % \
        (self.old_mode.name, self.old_mode.parent.name, self.old_mode.file,
         self.old_mode.line, self.old_mode.pos)

    def get_sxml_message(self):
        return self.sxml_error_string % (self.old_mode.name,
                                         self.old_mode.parent.name)

class ComponentImplementationSpecifiesNoStartingModeError(AbstractError):
    def __init__(self, file, line, pos, identifier, impl):
        AbstractError.__init__(self, file, line, pos, identifier, [impl])
        self.impl = impl
        self.error_string = "Component Implementation \"%s\" specifies modes" + \
        ", but no starting mode."
        
    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % (self.impl.name)

class ComponentImplementationSpecifiesMultipleStartingModeError(AbstractError):
    def __init__(self, file, line, pos, identifier, impl):
        AbstractError.__init__(self, file, line, pos, identifier, [impl])
        self.impl = impl
        self.error_string = "Component Implementation \"%s\" specifies " + \
        "multiple starting modes."
        
    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % (self.impl.name)

class ErrorTypeSpecifiesMultipleStartingStatesError(AbstractError):
    def __init__(self, file, line, pos, identifier, type):
        AbstractError.__init__(self, file, line, pos, identifier, [type])
        self.type = type
        self.error_string = "Error Model Type \"%s\" specifies " + \
        "multiple starting states."
        
    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % (self.type.name)
    
class ErrorImplSpecifiesMultipleStartingStatesError(AbstractError):
    def __init__(self, file, line, pos, identifier, impl):
        AbstractError.__init__(self, file, line, pos, identifier, [impl])
        self.impl = impl
        self.error_string = "Error model implementation \"%s\" specifies " + \
        "multiple starting states."
        
    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % (self.impl.name)

class ErrorTypeSpecifiesNoStartingStateError(AbstractError):
    def __init__(self, file, line, pos, identifier, type):
        AbstractError.__init__(self, file, line, pos, identifier, [type])
        self.type = type
        self.error_string = "Error Model Type \"%s\" specifies " + \
        "no starting state."
        
    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % (self.type.name)

class ErrorImplSpecifiesNoStartingStateError(AbstractError):
    def __init__(self, file, line, pos, identifier, impl):
        AbstractError.__init__(self, file, line, pos, identifier, [impl])
        self.impl = impl
        self.error_string = "Error model implementation \"%s\" specifies " + \
        "no starting state."
        
    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % (self.impl.name)

class ModeTransitionRefersToUnknownModeError(AbstractError):
    def __init__(self, file, line, pos, identifier, trans, mode_name):
        AbstractError.__init__(self, file, line, pos, identifier, [trans])
        self.mode_name = mode_name
        self.error_string = "Mode transition refers to the unknown mode \"%s\"."
        
    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % (self.mode_name)

class ErrorStateTransitionRefersToUnknownStateError(AbstractError):
    def __init__(self, file, line, pos, identifier, trans, state_name):
        AbstractError.__init__(self, file, line, pos, identifier, [trans])
        self.state_name = state_name
        self.error_string = "Error transition refers to the unknown mode \"%s\"."
        
    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % (self.state_name)

class TriggerRefersToInactiveSubomponentError(AbstractError):
    def __init__(self, file, line, pos, identifier, subcomp_name, trans):
        AbstractError.__init__(self, file, line, pos, identifier, [trans])
        self.subcomp_name = subcomp_name
        self.trans = trans
        self.error_string = "Trigger refers to the subcomponent \"%s\", but " + \
        "that subcomponent is inactive in the source mode of the transition, " + \
        "namely \"%s\"."
        
    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % (self.subcomp_name, self.trans.source_mode)
    
class TriggerRefersToUnknownSubomponentError(AbstractError):
    def __init__(self, file, line, pos, identifier, subcomp_name, trans):
        AbstractError.__init__(self, file, line, pos, identifier, [trans])
        self.subcomp_name = subcomp_name
        self.error_string = "Trigger refers to the unknown subcomponent \"%s\"."
        
    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % (self.subcomp_name)

class TriggerRefersToUnknownPortError(AbstractError):
    def __init__(self, file, line, pos, identifier, subcomp_name, port_name,
                 trans):
        AbstractError.__init__(self, file, line, pos, identifier, [trans])
        self.subcomp_name = subcomp_name
        self.port_name = port_name
        self.error_string = "Trigger refers to the unknown port \"%s\" of " + \
        " subcomponent \"%s\"."
        
    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % (self.port_name, self.subcomp_name)

class TriggerRefersToDataPortError(AbstractError):
    def __init__(self, file, line, pos, identifier, subcomp_name, port_name,
                 trans):
        AbstractError.__init__(self, file, line, pos, identifier, [trans])
        self.subcomp_name = subcomp_name
        self.port_name = port_name
        self.error_string = "Trigger refers to a data port \"%s\" of " + \
        " subcomponent \"%s\", instead of an event port."
        
    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % (self.port_name, self.subcomp_name)

class TriggerIsInvalidError(AbstractError):
    def __init__(self, file, line, pos, identifier, trigger, trans):
        AbstractError.__init__(self, file, line, pos, identifier, [trans])
        self.trigger = trigger
        self.error_string = "Trigger \"%s\" is invalid. It is neither empty, " + \
        " nor refers to an event port of the current component, nor to an " + \
        "event port of a subcomponent that is active in the starting mode " + \
        "of the transition, nor is it the keyword \"reset\"."
        
    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % (self.trigger)
    
class ErrorTriggerIsInvalidError(AbstractError):
    def __init__(self, file, line, pos, identifier, trigger, trans):
        AbstractError.__init__(self, file, line, pos, identifier, [trans])
        self.trigger = trigger
        self.error_string = "Trigger \"%s\" is invalid. It is neither " + \
        "\"reset\",  nor refers to an error propagation of the current " + \
        "error model, nor to an error event."
        
    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % (self.trigger)

class ModeIsNotReachableError(AbstractError):
    def __init__(self, file, line, pos, identifier, mode):
        AbstractError.__init__(self, file, line, pos, identifier, [mode])
        self.mode = mode
        self.error_string = "Mode \"%s\" is not reachable."
        
    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % (self.mode.name)

class ExpressionTypeMismatchError(AbstractError):
    def __init__(self, file, line, pos, left_type, right_type, operator,
                 error_reference_object):
        AbstractError.__init__(self, file, line, pos,
                               references=[error_reference_object])
        self.identifier = "C-3"
        self.left_type = left_type
        self.right_type = right_type
        self.operator = operator
        self.error_string = "Expression type mismatch, cannot combine " + \
        "\"%s\" and \"%s\" with \"%s\"."
        
    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % (self.left_type, self.right_type,
                                    self.operator)

class ExpressionIsNonLinearError(AbstractError):
    def __init__(self, file, line, pos, left_type, right_type, operator,
                 error_reference_object):
        AbstractError.__init__(self, file, line, pos,
                               references=[error_reference_object])
        self.identifier = "C-5"
        self.left_type = left_type
        self.right_type = right_type
        self.operator = operator
        self.error_string = "Nonlinear operations on continuous data are not" + \
        " supported: Cannot combine \"%s\" and \"%s\" with \"%s\"."
        
    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % (self.left_type, self.right_type,
                                    self.operator)
        
class InitializationValueTypeMismatchError(AbstractError):
    def __init__(self, file, line, pos, identifier, expected_type,
                 error_reference_object):
        AbstractError.__init__(self, file, line, pos, identifier,
                               references=[error_reference_object])
        self.expected_type = expected_type
        self.error_string = "Default value does not match data type, " + \
        "expecting \"%s\"."
        
    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % (self.expected_type)
    
class ClocksMustNotBeInitializedError(AbstractError):
    def __init__(self, file, line, pos, identifier, error_reference_object):
        AbstractError.__init__(self, file, line, pos, identifier,
                               references=[error_reference_object])
        self.error_string = "Clocks must not be initialized."
        
    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string
    
class EnumValueIllegalError(AbstractError):
    def __init__(self, file, line, pos, identifier, illegal_values,
                 error_reference_object):
        AbstractError.__init__(self, file, line, pos, identifier,
                               references=[error_reference_object])
        self.illegal_values = illegal_values
        self.error_string = "The values \"%s\" are illegal for this enum."
        
    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % (", ".join(self.illegal_values))
    
class IntRangeValueIllegalError(AbstractError):
    def __init__(self, file, line, pos, identifier, illegal_value,
                 error_reference_object):
        AbstractError.__init__(self, file, line, pos, identifier,
                               references=[error_reference_object])
        self.illegal_value = illegal_value
        self.error_string = "The value %d is illegal for this int range"
        
    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % (self.illegal_value)
    
class ClockSetToNonZeroError(AbstractError):
    def __init__(self, file, line, pos, error_reference_object):
        AbstractError.__init__(self, file, line, pos,
                               references=[error_reference_object])
        self.identifier = 67
        self.error_string = "Clocks can only be set to 0."
        
    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string

class ContinuousNotSetToRealError(AbstractError):
    def __init__(self, file, line, pos, error_reference_object):
        AbstractError.__init__(self, file, line, pos,
                               references=[error_reference_object])
        self.identifier = 68
        self.error_string = "Continuous data can only be set to a real value."
        
    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string
    
class DataComponentSetInMultipleEffectsError(AbstractError):
    def __init__(self, file, line, pos, identifier, comp_name, trans):
        AbstractError.__init__(self, file, line, pos, identifier, [trans])
        self.comp_name = comp_name
        self.error_string = "Data subcomponent \"%s\" is set in multiple effects."
        
    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % (self.comp_name)

class DataSubcomponentSetAndNotActiveError(AbstractError):
    def __init__(self, file, line, pos, identifier, comp_name, mode, trans):
        AbstractError.__init__(self, file, line, pos, identifier, [trans])
        self.comp_name = comp_name
        self.mode = mode
        self.error_string = "Data subcomponent \"%s\" is set, but is inactive" + \
        " in mode \"%s\"."
        
    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % (self.comp_name, self.mode)

class EffectRefersToUnknownIdentifierLhsError(AbstractError):
    def __init__(self, file, line, pos, identifier, unknown_identifier, trans):
        AbstractError.__init__(self, file, line, pos, identifier, [trans])
        self.unknown_identifier = unknown_identifier
        self.error_string = "Left hand side of effect refers to unknown " + \
        "subcomponent or outgoing data port \"%s\"."
        
    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % (self.unknown_identifier)
    
class EffectRefersToUnknownIdentifierRhsError(AbstractError):
    def __init__(self, file, line, pos, identifier, unknown_identifier, trans):
        AbstractError.__init__(self, file, line, pos, identifier, [trans])
        self.unknown_identifier = unknown_identifier
        self.error_string = "Right hand side of effect refers to unknown " + \
        "subcomponent or port or enum value \"%s\"."
        
    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % (self.unknown_identifier)
    
class EffectRefersToInactiveSubcomponentRhsError(AbstractError):
    def __init__(self, file, line, pos, identifier, subcomp_name, mode, trans):
        AbstractError.__init__(self, file, line, pos, identifier, [trans])
        self.subcomp_name = subcomp_name
        self.mode = mode
        self.error_string = "Right hand side of effect refers to " + \
        "subcomponent \"%s\", but that subcomponent is inactive in mode \"%s\"."
        
    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % (self.subcomp_name, self.mode)
    
class TrajectoryEquationReferencesIllegalType(AbstractError):
    def __init__(self, file, line, pos, identifier, illegal_type,
                 error_reference_object):
        AbstractError.__init__(self, file, line, pos, identifier,
                               [error_reference_object])
        self.illegal_type = illegal_type
        self.error_string = "Trajectory expression refers to illegal type \"%s\"."
        
    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % (self.illegal_type)

class SubcomponentReferencesFDIRImplementationError(AbstractError):
    def __init__(self, file, line, pos, identifier, subcomp):
        AbstractError.__init__(self, file, line, pos, identifier, [subcomp])
        self.subcomp = subcomp
        self.error_string = "Subcomponent \"%s\" references an FDIR " + \
        "implementation, but for FDIR subcomponents only the component " + \
        "type must be given."
        
    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % (self.subcomp.name)
    
class FDIRComponentTypeHasMultipleImplementationsError(AbstractError):
    def __init__(self, file, line, pos, identifier, type, impls):
        AbstractError.__init__(self, file, line, pos, identifier, [type] + impls)
        self.type = type
        self.error_string = "Component type \"%s\" is annotated as FDIR, but" + \
        " has multiple implementations."
        
    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % (self.type.name)

class FDIRComponentTypeHasNonFDIRImplementationError(AbstractError):
    def __init__(self, file, line, pos, identifier, type, impl):
        AbstractError.__init__(self, file, line, pos, identifier, [type, impl])
        self.type = type
        self.impl = impl
        self.error_string = "Component type \"%s\" is annotated as FDIR, but" + \
        " has the non-FDIR implementation \"%s\"."
        
    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % (self.type.name, self.impl.name)

class NonFDIRComponentTypeHasFDIRImplementationError(AbstractError):
    def __init__(self, file, line, pos, identifier, type, impl):
        AbstractError.__init__(self, file, line, pos, identifier, [type, impl])
        self.type = type
        self.impl = impl
        self.error_string = "Component type \"%s\" is not annotated as FDIR, but" + \
        " has the FDIR implementation \"%s\"."
        
    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % (self.type.name, self.impl.name)

class ProcessHasNoThreadAndThreadGroupSubcomponentError(AbstractError):
    def __init__(self, file, line, pos, identifier, impl):
        AbstractError.__init__(self, file, line, pos, identifier, [impl])
        self.impl = impl
        self.error_string = "Component implementation \"%s\" of category " + \
        "\"process\" must contain at least one thread or thread group " + \
        "subcomponent."
        
    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % (self.impl.name)

class ModeInvariantRefersToIllegalDataTypeError(AbstractError):
    def __init__(self, file, line, pos, identifier, subcomponent_name,
                 wrong_type, mode):
        AbstractError.__init__(self, file, line, pos, identifier, [mode])
        self.subcomponent_name = subcomponent_name
        self.wrong_type = wrong_type
        self.error_string = "Mode invariant refers to subcomponent \"%s\" " + \
        "of type \"%s\", but can only refer to subcomponents of type " + \
        "\"continuous\" or \"clock\"."

    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % (self.subcomponent_name, self.wrong_type)

class ModeInvariantRefersToInactiveSubcomponentError(AbstractError):
    def __init__(self, file, line, pos, identifier, subcomponent_name, mode):
        AbstractError.__init__(self, file, line, pos, identifier, [mode])
        self.subcomponent_name = subcomponent_name
        self.mode = mode
        self.error_string = "Mode invariant refers to subcomponent \"%s\", " + \
        "but that subcomponent is inactive in mode \"%s\"."

    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % (self.subcomponent_name, self.mode.name)

class ModeInvariantRefersToUnknownSubcomponentError(AbstractError):
    def __init__(self, file, line, pos, identifier, subcomponent_name, mode):
        AbstractError.__init__(self, file, line, pos, identifier, [mode])
        self.subcomponent_name = subcomponent_name
        self.error_string = "Mode invariant refers to unknown subcomponent " + \
        "\"%s\"."

    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % (self.subcomponent_name)
    
class StateInvariantRefersToUnknownClockError(AbstractError):
    def __init__(self, file, line, pos, identifier, clock_name, state):
        AbstractError.__init__(self, file, line, pos, identifier, [state])
        self.clock_name = clock_name
        self.error_string = "Mode invariant refers to unknown clock " + \
        "\"%s\"."

    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % (self.clock_name)
    
    
class InvariantIsNotConvexError(AbstractError):
    def __init__(self, file, line, pos, identifier, invariant, state):
        AbstractError.__init__(self, file, line, pos, identifier, [state])
        self.invariant = invariant
        self.error_string = "Invariant \"%s\" is not convex."

    def get_message(self):
        """Retrieves the error message associated with the error"""
        from slim_compiler.src.dataStructure.visitors.printVisitor import toSLIMVisitor
        invariant_str = toSLIMVisitor().visit(self.invariant)
        return self.error_string % (invariant_str)

class ModeUrgencyIsNegative(AbstractError):
    def __init__(self, file, line, pos, identifier, urgent, mode):
        AbstractError.__init__(self, file, line, pos, identifier, [mode])
        self.urgent = urgent
        self.mode = mode
        self.error_string = "Urgency of mode %s is not allowed to be negative."

    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % (self.mode.name)

class StateUrgencyIsNegative(AbstractError):
    def __init__(self, file, line, pos, identifier, urgent, state):
        AbstractError.__init__(self, file, line, pos, identifier, [state])
        self.urgent = urgent
        self.state = state
        self.error_string = "Urgency of state %s is not allowed to be negative."

    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % (self.state.name)

class StateNotUrgentWithDelayedTransitionError(AbstractError):
    def __init__(self, file, line, pos, identifier, state):
        AbstractError.__init__(self, file, line, pos, identifier, [state])
        self.state_name = state.name
        self.error_string = \
        "State \"%s\" must specify 'urgent'-notation as at least one transition uses 'within'-syntax."
        self.sxml_error_string = \
        "State \"%s\" must specify 'urgent'-notation as at least one transition uses 'within'-syntax."

    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % self.state_name

class UnaryExpressionTypeMismatchError(AbstractError):
    def __init__(self, file, line, pos, type, operator, error_reference_object):
        AbstractError.__init__(self, file, line, pos,
                               references=[error_reference_object])
        self.identifier = "C-3"
        self.type = type
        self.operator = operator
        self.error_string = "Expression type mismatch, cannot use \"%s\" " + \
        "with \"%s\"."
        
    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % (self.type, self.operator)

class EnumsIncomparableError(AbstractError):
    def __init__(self, file, line, pos, enum1, enum2, error_reference_object):
        AbstractError.__init__(self, file, line, pos,
                               references=[error_reference_object])
        self.identifier = 87
        self.enum1 = enum1
        self.enum2 = enum2
        self.error_string = "Enums \"%s\" and \"%s\" are incomparable."
    
    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % (self.enum1, self.enum2)    
    
class GuardRefersToInactiveSubcomponentError(AbstractError):
    def __init__(self, file, line, pos, identifier, subcomp_name, trans):
        AbstractError.__init__(self, file, line, pos, identifier, [trans])
        self.subcomp_name = subcomp_name
        self.trans = trans
        self.error_string = "Guard refers to subcomponent \"%s\", which is " + \
        "inactive in the starting mode \"%s\" of the transition."
        
    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % (self.subcomp_name, self.trans.source_mode)    

class GuardRefersToUnknownClockError(AbstractError):
    def __init__(self, file, line, pos, identifier, clock_name, error_impl, trans):
        AbstractError.__init__(self, file, line, pos, identifier, [trans, error_impl])
        self.clock_name = clock_name
        self.error_impl = error_impl
        self.trans = trans
        self.error_string = "Guard \"%s\" refers to clock \"%s\", which is " + \
        "not defined in the error implementation \"%s\""
        
    def get_message(self):
        """Retrieves the error message associated with the error"""
        from slim_compiler.src.dataStructure.visitors.printVisitor import toSLIMVisitor
        guard_str = toSLIMVisitor().visit(self.trans.guard)
        return self.error_string % (guard_str, self.clock_name, self.error_impl.name)

class EffectRefersToUnknownClockError(AbstractError):
    def __init__(self, file, line, pos, identifier, clock_name, error_impl, trans):
        AbstractError.__init__(self, file, line, pos, identifier, [trans, error_impl])
        self.clock_name = clock_name
        self.error_impl = error_impl
        self.trans = trans
        self.error_string = "Effect \"%s\" refers to clock \"%s\", which is " + \
        "not defined in the error implementation \"%s\""
        
    def get_message(self):
        """Retrieves the error message associated with the error"""
        from slim_compiler.src.dataStructure.visitors.printVisitor import toSLIMVisitor
        effects_str = ';'.join((toSLIMVisitor().visit(effect) for effect in self.trans.effects))
        return self.error_string % (effects_str, self.clock_name, self.error_impl.name)
        
class GuardRefersToUnknownSubcomponentOrDataPortError(AbstractError):
    def __init__(self, file, line, pos, identifier, unknown_identifier, trans):
        AbstractError.__init__(self, file, line, pos, identifier, [trans])
        self.unknown_identifier = unknown_identifier
        self.error_string = "Guard refers to unknown subcomponent or " + \
        "data port \"%s\"."
        
    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % (self.unknown_identifier)    
        
class GuardHasIllegalTypeError(AbstractError):
    def __init__(self, file, line, pos, identifier, type, trans):
        AbstractError.__init__(self, file, line, pos, identifier, [trans])
        self.type = type
        self.error_string = "Guard is of type \"%s\", but should be of type" + \
        "\"bool\"."
        
    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % (self.type)

class InvalidWithinBoundsError(AbstractError):
    def __init__(self, file, line, pos, identifier, lval, uval, trans):
        AbstractError.__init__(self, file, line, pos, identifier, [trans])
        self.trans = trans
        self.lval = lval
        self.uval = uval
        self.error_string = "Time bounds [%d..%d] of transition are invalid, must be non-negative and convex."

    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % (self.lval, self.uval)

class InvariantSpecifiesMultipleTrajectoryEquationsForOneVariableError(AbstractError):
    def __init__(self, file, line, pos, identifier, variable, mode):
        AbstractError.__init__(self, file, line, pos, identifier, [mode])
        self.variable = variable
        self.error_string = "Mode invariant may specify at most one trajectory " + \
        "equation for variable \"%s\"."
        
    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % (self.variable)   
    
class InvariantTrajectoryEquationRefersToNonContinuousSubcomponentError(AbstractError):             
    def __init__(self, file, line, pos, identifier, subcomp, mode):
        AbstractError.__init__(self, file, line, pos, identifier, [mode])
        self.subcomp = subcomp
        self.error_string = "Mode invariant refers to non-continuous " + \
        "subcomponent \"%s\" in trajectory equation."
        
    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % (self.subcomp)
    
    def get_references(self):
        return self.references   
        
class InitValuesViolateStartingModeInvariant(AbstractError):
    def __init__(self, file, line, pos, identifier, mode):
        AbstractError.__init__(self, file, line, pos, identifier, [mode])
        self.mode = mode
        self.error_string = "Mode invariant of starting mode \"%s\" is not " + \
        "satisfied by default values."
        
    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % (self.mode.name)      

class InvariantHasIllegalTypeError(AbstractError):
    def __init__(self, file, line, pos, identifier, type):
        AbstractError.__init__(self, file, line, pos, identifier)
        self.type = type
        self.error_string = "Invariant is of type \"%s\", but should be of type " + \
        "\"bool\"."

    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % (self.type)

class CaseConditionNotOfTypeBoolError(AbstractError):
    def __init__(self, file, line, pos, wrong_type, error_reference_object):
        AbstractError.__init__(self, file, line, pos,
                               references=[error_reference_object])
        self.identifier = "C-3"
        self.wrong_type = wrong_type
        self.error_string = "Case condition is of type \"%s\", expecting " + \
        "expression of type \"bool\"."
        
    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % (self.wrong_type)                
        
class CaseExpressionsHaveMultipleTypesError(AbstractError):
    def __init__(self, file, line, pos, types, error_reference_object):
        AbstractError.__init__(self, file, line, pos,
                               references=[error_reference_object])
        self.identifier = "C-3"
        self.types = types
        self.error_string = "Case expression returns multiple different " + \
        "types, which can not be implicitly converted:\n"
        for i in range(0, len(self.types)):
            self.error_string += "\t" + str(i + 1) + ". " + self.types[i]
            if i < len(self.types) - 1:
                self.error_string += "\n"
        
    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string
        
class CaseExpressionValuesNotValidForEnumsError(AbstractError):
    def __init__(self, file, line, pos, error_reference_object):
        AbstractError.__init__(self, file, line, pos,
                               references=[error_reference_object])
        self.identifier = "C-3"
        self.error_string = "Case expression returns enum values that are " + \
        "not valid for all returned enum identifiers."
        
    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string

class CaseExpressionsHaveDifferentValidEnumsError(AbstractError):
    def __init__(self, file, line, pos, error_reference_object):
        AbstractError.__init__(self, file, line, pos,
                               references=[error_reference_object])
        self.identifier = "C-3"
        self.error_string = "Case expression returns enum values that are " + \
        "not valid for all returned enum identifiers."
        
    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string
        
class CaseExpressionsHaveDifferentValidEnumValuesError(AbstractError):
    def __init__(self, file, line, pos, error_reference_object):
        AbstractError.__init__(self, file, line, pos,
                               references=[error_reference_object])
        self.identifier = "C-3"
        self.error_string = "Case expression returns enum identifiers " + \
        " with different value sets."
        
    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string
    
class CaseExpressionSpecifiesFieldsAndValuesValuesMismatchError(AbstractError):
    def __init__(self, file, line, pos, error_reference_object):
        AbstractError.__init__(self, file, line, pos,
                               references=[error_reference_object])
        self.identifier = "C-3"
        self.error_string = "Case expression returns enum identifiers " + \
        "and enum values, but not all enum values are valid for all of the " + \
        "enum identifiers."
        
    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string    

class ErrorClockAlreadyDeclaredError(AbstractError):
    def __init__(self, file, line, pos, identifier, other_clock):
        AbstractError.__init__(self, file, line, pos, identifier, [other_clock])
        self.other_clock = other_clock
        self.error_string = \
        "Clock \"%s\" already declared in %s at line %d:%d."
        self.sxml_error_string = \
        "Clock \"%s\" declared multiple times."
        
    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string \
        % (self.other_clock.name, self.other_clock.file,
           self.other_clock.line, self.other_clock.pos)
        
    def get_sxml_message(self):
        return self.sxml_error_string % (self.subcomponent.name)

class ErrorStateInvariantInTypeError(AbstractError):
    def __init__(self, file, line, pos, identifier, error_state_name):
        AbstractError.__init__(self, file, line, pos, identifier, [])
        self.error_state_name = error_state_name;
        self.error_string = \
        "Error state \"%s\" has an invariant. Invariants are not allowed in error types" + \
        " - only in error implementations"
        self.sxml_error_string = \
        "Error state \"%s\" has an invariant. These are only allowed in error implementations."
        
    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string%self.error_state_name
        
    def get_sxml_message(self):
        return self.sxml_error_string % (self.error_state_name)
    
class ErrorStateUrgentInTypeError(AbstractError):
    def __init__(self, file, line, pos, identifier, error_state_name):
        AbstractError.__init__(self, file, line, pos, identifier, [])
        self.error_state_name = error_state_name;
        self.error_string = \
        "Error state \"%s\" uses 'urgent' notation (implicit invariant). Invariants are not allowed in error types" + \
        " - only in error implementations"
        self.sxml_error_string = \
        "Error state \"%s\" uses 'urgent' notation (implicit invariant). These are only allowed in error implementations."
        
    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string%self.error_state_name
        
    def get_sxml_message(self):
        return self.sxml_error_string % (self.error_state_name)
    
class ErrorTransitionIsProbabilisticButHasGuardError(AbstractError):
    def __init__(self, file, line, pos, identifier, error_transition):
        AbstractError.__init__(self, file, line, pos, identifier, [error_transition])
        self.error_transition = error_transition
        self.error_string = \
        "Error transition \"%s\" has a guard. However, probabilistic transitions are not allowed to have guards."
        self.sxml_error_string = \
        "Probabilistic error transition \"%s\" has a guard."
        
    def get_message(self):
        """Retrieves the error message associated with the error"""
        from slim_compiler.src.dataStructure.visitors.printVisitor import toSLIMVisitor
        return self.error_string % toSLIMVisitor().visit(self.error_transition)
        
    def get_sxml_message(self):
        from slim_compiler.src.dataStructure.visitors.printVisitor import toSLIMVisitor
        return self.sxml_error_string % toSLIMVisitor().visit(self.error_transition)

class ErrorTransitionIsProbabilisticButHasDelayError(AbstractError):
    def __init__(self, file, line, pos, identifier, error_transition):
        AbstractError.__init__(self, file, line, pos, identifier, [error_transition])
        self.error_transition = error_transition
        self.error_string = \
        "Error transition \"%s\" has a time delay constraint. However, probabilistic transitions are not allowed to have time constraints."
        self.sxml_error_string = \
        "Probabilistic error transition \"%s\" has a time constraint."

    def get_message(self):
        """Retrieves the error message associated with the error"""
        from slim_compiler.src.dataStructure.visitors.printVisitor import toSLIMVisitor
        return self.error_string % toSLIMVisitor().visit(self.error_transition)

    def get_sxml_message(self):
        from slim_compiler.src.dataStructure.visitors.printVisitor import toSLIMVisitor
        return self.sxml_error_string % toSLIMVisitor().visit(self.error_transition)

class ErrorStateInvariantForProbabilisticStateError(AbstractError):
    def __init__(self, file, line, pos, identifier, state):
        AbstractError.__init__(self, file, line, pos, identifier, [state])
        self.state = state;
        self.error_string = \
        "Error state \"%s\" has an invariant. Invariants are not allowed in probabilistic states"
        self.sxml_error_string = \
        "Probabilistic error state \"%s\" has an invariant."
        
    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string%self.state.name
        
    def get_sxml_message(self):
        return self.sxml_error_string % (self.error_state_name)

class ErrorStateUrgencyForProbabilisticStateError(AbstractError):
    def __init__(self, file, line, pos, identifier, state):
        AbstractError.__init__(self, file, line, pos, identifier, [state])
        self.state = state;
        self.error_string = \
        "Error state \"%s\" is urgent. Urgency is not allowed for probabilistic states"
        self.sxml_error_string = \
        "Probabilistic error state \"%s\" has urgency."
        
    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string%self.state.name
        
    def get_sxml_message(self):
        return self.sxml_error_string % (self.error_state_name)

class ErrorStateAlreadyDeclaredInTypeError(AbstractError):
    def __init__(self, file, line, pos, identifier, state, type):
        AbstractError.__init__(self, file, line, pos, identifier, [state])
        self.type = type
        self.state = state
        self.error_string = \
        "Error state \"%s\" already declared in error model type \"%s\" in " + \
        "%s at line %d:%d."
        self.sxml_error_string = \
        "Error state '%s' declared multiple times."
        
    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % \
        (self.state.name, self.type.name, self.state.file, self.state.line,
         self.state.pos)
        
    def get_sxml_message(self):
        return self.sxml_error_string % (self.state.name)
        
class ErrorStateAlreadyDeclaredInImplementationError(AbstractError):
    def __init__(self, file, line, pos, identifier, impl, state):
        AbstractError.__init__(self, file, line, pos, identifier, [state])
        self.impl = impl
        self.state = state
        self.error_string = \
        "Error state \"%s\" already declared in error model implementation" + \
        "\"%s\" in %s at line %d:%d."
        self.sxml_error_string = \
        "Error state '%s' declared multiple times."
        
    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % \
        (self.state.name, self.impl.name, self.state.file, self.state.line,
         self.state.pos)
        
    def get_sxml_message(self):
        return self.sxml_error_string % (self.state.name)

class ErrorPropagationAlreadyDeclaredInTypeError(AbstractError):        
    def __init__(self, file, line, pos, identifier, type, prop):
        AbstractError.__init__(self, file, line, pos, identifier, [prop])
        self.type = type
        self.prop = prop
        self.error_string = \
        "Error propagation \"%s\" already declared in error model type \"%s\"" + \
        "in \"%s\" at line %d:%d."
        self.sxml_error_string = \
        "Error propagation '%s' declared multiple times."
        
    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % \
        (self.prop.name, self.type.name, self.prop.file, self.prop.line,
         self.prop.pos)
        
    def get_sxml_message(self):
        return self.sxml_error_string % (self.prop.name)

class ErrorEventAlreadyDeclaredInTypeError(AbstractError):
    def __init__(self, file, line, pos, identifier, type, event, prop):
        AbstractError.__init__(self, file, line, pos, identifier, [event])
        self.type = type
        self.event = event
        self.prop = prop
        self.error_string = \
        "Error event \"%s\" already declared as error propagation in error model type \"%s\"" + \
        " at line %d:%d."
        self.sxml_error_string = \
        "Error event '%s' already declared as error propagation."
        
    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % \
        (self.event.name, self.type.name, self.prop.line, self.prop.pos)
        
    def get_sxml_message(self):
        return self.sxml_error_string % (self.new_id.name)

class RateIsNotPositiveError(AbstractError):
    def __init__(self, file, line, pos, identifier, rate, event):
        AbstractError.__init__(self, file, line, pos, identifier, [event])
        self.rate = rate
        self.event = event
        self.error_string = \
        "Error event \"%s\" specifies a non-positive rate %d."

    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % \
        (self.event.name, self.rate)

class ErrorPropagationAlreadyDeclaredInImplementationError(AbstractError):
    def __init__(self, file, line, pos, name, implname, event):
        AbstractError.__init__(self, file, line, pos, "No identifier yet", [event])
        self.name = name
        self.implname = implname
        self.event = event
        self.error_string = \
        "Error propagation \"%s\" already declared as error event in error model implementation type \"%s\"" + \
        " at line %d:%d."
        self.sxml_error_string = \
        "Error propagation '%s' already declared as error event."
        
    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % \
        (self.name, self.implname, self.event.line, self.event.pos)
        
    def get_sxml_message(self):
        return self.sxml_error_string % (self.name)

class ErrorEventAlreadyDeclaredError(AbstractError):        
    def __init__(self, file, line, pos, identifier, impl, event):
        AbstractError.__init__(self, file, line, pos, identifier, [event])
        self.impl = impl
        self.event = event
        self.error_string = \
        "Error event \"%s\" already declared in error model implementation " + \
        "\"%s\" in \"%s\" at line %d:%d."
        self.sxml_error_string = \
        "Error event '%s' declared multiple times."
        
    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % \
        (self.event.name, self.impl.name, self.event.file, self.event.line,
         self.event.pos)
        
    def get_sxml_message(self):
        return self.sxml_error_string % (self.event.name)
    
class ReactivationTransitionHasGuardError(AbstractError):
    def __init__(self, file, line, pos, identifier, transition):
        AbstractError.__init__(self, file, line, pos, identifier, [transition])
        self.transition = transition;
        self.error_string = \
        "Reactivation transition \"%s\" has a guard. Guards are not allowed in reactivation transitions."
        self.sxml_error_string = \
        "Reactivation transition \"%s\" has a guard."
        
    def get_message(self):
        """Retrieves the error message associated with the error"""
        from slim_compiler.src.dataStructure.visitors.printVisitor import toSLIMVisitor
        return self.error_string % toSLIMVisitor().visit(self.transition)
        
    def get_sxml_message(self):
        from slim_compiler.src.dataStructure.visitors.printVisitor import toSLIMVisitor
        return self.sxml_error_string % toSLIMVisitor().visit(self.transition)
    
class ReactivationTransitionHasTimeDelayError(AbstractError):
    def __init__(self, file, line, pos, identifier, transition):
        AbstractError.__init__(self, file, line, pos, identifier, [transition])
        self.transition = transition;
        self.error_string = \
        "Reactivation transition \"%s\" uses 'within' notation (implicit guard). Guards are not allowed in reactivation transitions."
        self.sxml_error_string = \
        "Reactivation transition \"%s\" uses 'within' notation (implicit guard)."
        
    def get_message(self):
        """Retrieves the error message associated with the error"""
        from slim_compiler.src.dataStructure.visitors.printVisitor import toSLIMVisitor
        return self.error_string % toSLIMVisitor().visit(self.transition)
        
    def get_sxml_message(self):
        from slim_compiler.src.dataStructure.visitors.printVisitor import toSLIMVisitor
        return self.sxml_error_string % toSLIMVisitor().visit(self.transition)

class ClockNotScaledInTimeScaledModel(AbstractError):
    def __init__(self, file, line, pos, identifier, clock_subcomponent):
        AbstractError.__init__(self, file, line, pos, identifier, [clock_subcomponent])
        self.clock_name = clock_subcomponent.name
        self.error_string = \
        "Model uses time scaling but clock \"%s\" has no time scaling operator."
        self.sxml_error_string = \
        "Clock \"%s\" has no time scaling operator."
        
    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % self.clock_name
        
    def get_sxml_message(self):
        return self.sxml_error_string % self.clock_name

class ErrorEventNotScaledInTimeScaledModel(AbstractError):
    def __init__(self, file, line, pos, identifier, error_event):
        AbstractError.__init__(self, file, line, pos, identifier, [error_event])
        self.event_name = error_event.name
        self.error_string = \
        "Model uses time scaling but error event \"%s\" has no time scaling operator."
        self.sxml_error_string = \
        "Error event \"%s\" has no time scaling operator."
        
    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % self.event_name
        
    def get_sxml_message(self):
        return self.sxml_error_string % self.event_name


class ModeUrgentNotScaledInTimeScaledModel(AbstractError):
    def __init__(self, file, line, pos, identifier, mode):
        AbstractError.__init__(self, file, line, pos, identifier, [mode])
        self.mode_name = mode.name
        self.error_string = \
        "Mode \"%s\" uses 'urgent'-notation but doesn't use time scaling operator although model is time scaled."
        self.sxml_error_string = \
        "Mode \"%s\" uses 'urgent'-notation but misses time scaling operator."
        
    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % self.mode_name
        
    def get_sxml_message(self):
        return self.sxml_error_string % self.mode_name

class ModeNotUrgentWithDelayedTransitionError(AbstractError):
    def __init__(self, file, line, pos, identifier, mode):
        AbstractError.__init__(self, file, line, pos, identifier, [mode])
        self.mode_name = mode.name
        self.error_string = \
        "Mode \"%s\" must specify 'urgent'-notation as at least one transition uses 'within'-syntax."
        self.sxml_error_string = \
        "Mode \"%s\" must specify 'urgent'-notation."

    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % self.mode_name

    def get_sxml_message(self):
        return self.sxml_error_string % self.mode_name

class TrajectoryExpressionNotScaledInTimeScaledModel(AbstractError):
    def __init__(self, file, line, pos, identifier, traj_equation, obj):
        AbstractError.__init__(self, file, line, pos, identifier, [obj])
        self.traj_equation = traj_equation
        self.error_string = \
        "Trajectory equation \"%s\" doesn't use time scaling although model is time scaled."
        self.sxml_error_string = \
        "Trajectory equation \"%s\" misses time scaling operator."
        
    def get_message(self):
        """Retrieves the error message associated with the error"""
        from slim_compiler.src.dataStructure.visitors.printVisitor import toSLIMVisitor
        return self.error_string % (toSLIMVisitor().visit(self.traj_equation))
        
    def get_sxml_message(self):
        from slim_compiler.src.dataStructure.visitors.printVisitor import toSLIMVisitor
        return self.sxml_error_string % (toSLIMVisitor().visit(self.traj_equation), self.mode_name)    

class ModeTransitionWithinNotScaledInTimeScaledModel(AbstractError):
    def __init__(self, file, line, pos, identifier, transition):
        AbstractError.__init__(self, file, line, pos, identifier, [transition])
        self.transition = transition
        self.error_string = \
        "Mode transition uses 'within'-notation but doesn't use time scaling operator although model is time scaled."
        self.sxml_error_string = \
        "Mode transition uses 'within'-notation but misses time scaling operator."
        
    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string
        
    def get_sxml_message(self):
        return self.sxml_error_string


class ErrorStateUrgentNotScaledInTimeScaledModel(AbstractError):
    def __init__(self, file, line, pos, identifier, error_state):
        AbstractError.__init__(self, file, line, pos, identifier, [error_state])
        self.state_name = error_state.name
        self.error_string = \
        "Error state \"%s\" uses 'urgent'-notation but doesn't use time scaling operator although model is time scaled."
        self.sxml_error_string = \
        "Error state \"%s\" uses 'urgent'-notation but misses time scaling operator."
        
    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % self.state_name
        
    def get_sxml_message(self):
        return self.sxml_error_string % self.state_name

class ErrorTransitionWithinNotScaledInTimeScaledModel(AbstractError):
    def __init__(self, file, line, pos, identifier, transition):
        AbstractError.__init__(self, file, line, pos, identifier, [transition])
        self.transition = transition
        self.error_string = \
        "Error transition \"%s\" uses 'within'-notation but doesn't use time scaling operator although model is time scaled."
        self.sxml_error_string = \
        "Error transition \"%s\" uses 'within'-notation but misses time scaling operator."
        
    def get_message(self):
        """Retrieves the error message associated with the error"""
        from slim_compiler.src.dataStructure.visitors.printVisitor import toSLIMVisitor
        return self.error_string % toSLIMVisitor().visit(self.transition)
        
    def get_sxml_message(self):
        from slim_compiler.src.dataStructure.visitors.printVisitor import toSLIMVisitor
        return self.sxml_error_string % toSLIMVisitor().visit(self.transition)

# start parsing errors ---------------------------------------------------------

class ParserError(AbstractError):
    def __init__(self, file, line, pos, rule, found, expecting):
        AbstractError.__init__(self, file, line, pos)
        self.identifier = 114
        self.rule = rule
        self.found = found
        if self.found != "'<EOF>'":
            if self.found:
                self.found = self.found[2:-1] # correcting token string
        else:
            if self.found:
                self.found = self.found[1:-1]
        if expecting == "IDENT":
            self.expecting = "identifier"
        elif expecting == "identifier":
            self.expecting = "identifier"
        else:
            self.expecting = expecting
        
    def get_message(self):
        pass

class MismatchedTokenError(ParserError):
    def __init__(self, file, line, pos, rule, found, expecting):
        ParserError.__init__(self, file, line, pos, rule, found, expecting)
        self.identifier = 115
        self.error_message = "Mismatched input in %s: found \"%s\"," + \
        " expecting %s."
        
    def get_message(self):
        return self.error_message % (self.rule, self.found, self.expecting)
    
class MissingTokenError(ParserError):
    def __init__(self, file, line, pos, rule, found, expecting):
        ParserError.__init__(self, file, line, pos, rule, found, expecting)        
        self.identifier = 116
        self.error_message = "Missing %s in %s at \"%s\"."
        
    def get_message(self):
        return self.error_message % (self.expecting, self.rule, self.found)
    
class UnwantedTokenError(ParserError):
    def __init__(self, file, line, pos, rule, found, expecting):
        ParserError.__init__(self, file, line, pos, rule, found, expecting)        
        self.identifier = 117
        self.error_message = "Extraneous input \"%s\" in %s, expecting %s."
        
    def get_message(self):
        return self.error_message % (self.found, self.rule, self.expecting)
    
class NoViableAltError(ParserError):
    def __init__(self, file, line, pos, rule, found):
        ParserError.__init__(self, file, line, pos, rule, found, None)
        self.identifier = 118
        self.error_message = "No viable alternative in %s at input \"%s\"."
        
    def get_message(self):
        return self.error_message % (self.rule, self.found)
    
class EarlyExitError(ParserError):
    def __init__(self, file, line, pos, rule, found):
        ParserError.__init__(self, file, line, pos, rule, found, None)
        self.identifier = 119
        self.error_message = "Required (...)+ loop in %s did not match " + \
        "anything at input \"%s\"."
        
    def get_message(self):
        return self.error_message % (self.rule, self.found)
    
class MismatchedSetError(ParserError):
    def __init__(self, file, line, pos, rule, found, expecting):
        ParserError.__init__(self, file, line, pos, rule, found, expecting)
        self.identifier = 120
        self.error_message = "Mismatched input \"%s\" in %s, expecting %s."
        
    def get_message(self):
        return self.error_message % (self.found, self.rule, self.expecting)
    
class MismatchedNotSetError(ParserError):
    def __init__(self, file, line, pos, rule, found, expecting):
        ParserError.__init__(self, file, line, pos, rule, found, expecting)
        self.identifier = 121
        self.error_message = "Mismatched input \"%s\" in %s, expecting %s."
        
    def get_message(self):
        return self.error_message % (self.found, self.rule, self.expecting)
    
class MismatchedRangeError(ParserError):
    def __init__(self, file, line, pos, rule, found, expecting_from,
                 expecting_to):
        ParserError.__init__(self, file, line, pos, rule, found, expecting_from)
        self.identifier = 122
        self.expecting_to = expecting_to
        self.error_message = "Mismatched input \"%s\" in \"%s\", expecting " + \
        "set \"%s\" .. \"%s\"."
        
    def get_message(self):
        return self.error_message % (self.found, self.rule, self.expecting,
                                     self.expecting_to)

class FailedPredicateError(ParserError):
    def __init__(self, file, line, pos, rule, predicate):
        ParserError.__init__(self, file, line, pos, rule, None, None)
        self.identifier = 123
        self.predicate = predicate
        self.error_message = "Failed predicate {%s} in rule %s."
        
    def get_message(self):
        return self.error_message % (self.predicate, self.rule)

class UnsupportedFeatureError(AbstractError):
    def __init__(self, file, line, pos, feature):
        AbstractError.__init__(self, file, line, pos)
        self.identifier = 124
        self.feature = feature
        self.error_message = "Unsupported feature %s."
        
    def get_message(self):
        return self.error_message % (self.feature)
    
class SLIMParserError(ParserError):
    def __init__(self, file, line, pos, rule, message):
        ParserError.__init__(self, file, line, pos, rule, None, None)
        self.identifier = 131
        self.message = message
        self.error_message = "Parser error in %s: %s"
    
    def get_message(self):
        return self.error_message % (self.rule, self.message)

# end parsing errors -----------------------------------------------------------
# begin properties and effects errors ------------------------------------------

class WrongQualificationError(AbstractError):
    def __init__(self, full_qualification, missing_subcomp):
        AbstractError.__init__(self, "", 0, 0)
        self.identifier = 125
        self.full_qualification = full_qualification
        self.missing_subcomp = missing_subcomp
        self.error_string = "Could not follow qualification path \"%s\". " + \
        "Unknown subcomponent \"%s\"."
        
    def get_message(self):
        return self.error_string % (self.full_qualification,
                                    self.missing_subcomp)
        
class InjectionRefersToUnknownIdentifierLhsError(AbstractError):
    def __init__(self, identifier):
        AbstractError.__init__(self, "", 0, 0)
        self.identifier = 126
        self.identifier = identifier
        self.error_string = "Left-hand side of failure injection refers to " + \
        "the unknown identifier \"%s\"."
        
    def get_message(self):
        return self.error_string % (self.identifier)

class InjectionRefersToUnknownIdentifierRhsError(AbstractError):
    def __init__(self, identifier):
        AbstractError.__init__(self, "", 0, 0)
        self.identifier = 127
        self.identifier = identifier
        self.error_string = "Right-hand side of failure injection refers to " + \
        "the unknown identifier \"%s\"."
        
    def get_message(self):
        return self.error_string % (self.identifier)
    
class ModeValueIllegalError(AbstractError):
    def __init__(self, illegal_mode, error_reference_object):
        AbstractError.__init__(self, "", 0, 0,
                               references=[error_reference_object])
        self.identifier = 128
        self.illegal_mode = illegal_mode
        self.error_string = "The value \"%s\" is not a valid mode."
        
    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % (self.illegal_mode)

class ErrorStateValueIllegalError(AbstractError):
    def __init__(self, illegal_state, error_reference_object):
        AbstractError.__init__(self, "", 0, 0,
                               references=[error_reference_object])
        self.identifier = 129
        self.illegal_state = illegal_state
        self.error_string = "The value \"%s\" is not a valid error state."
        
    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % (self.illegal_state)

class ModeFieldsIncomparableError(AbstractError):
    def __init__(self, field1, field2, error_reference_object):
        AbstractError.__init__(self, "", 0, 0,
                               references=[error_reference_object])
        self.identifier = 130
        self.field1 = field1
        self.field2 = field2
        self.error_string = "Mode fields \"%s\" and \"%s\" are incomparable."
    
    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % (self.field1, self.field2)    

class ErrorFieldsIncomparableError(AbstractError):
    def __init__(self, field1, field2, error_reference_object):
        AbstractError.__init__(self, "", 0, 0,
                               references=[error_reference_object])
        self.identifier = 131
        self.field1 = field1
        self.field2 = field2
        self.error_string = "Error fields \"%s\" and \"%s\" are incomparable."
    
    def get_message(self):
        """Retrieves the error message associated with the error"""
        return self.error_string % (self.field1, self.field2) 
    
class ErrorTypeMissingError(AbstractError):
    def __init__(self, file, line, pos, impl_name):
        AbstractError.__init__(self, file, line, pos)
        self.impl_name = impl_name
        self.error_message = "You have declared an error model implementation " + \
        "'%s' but you did not declare an apt error model type."
        
    def get_message(self):
        return self.error_message % (self.impl_name)

# end properties and effects errors --------------------------------------------
