# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================


"""This module provides utility functions for handling SLIM expressions."""
# TODO: In this file, a maximum line length of 80 is used, instead of 79 as
# specified in PEP8.

from slim_compiler.src.dataStructure import expressions
from slim_compiler.src import testexpr


def create_slim_assignment(left_hand_side_identifier,
                           right_hand_side_identifier):
    """Returns a tree representation of the SLIM assignment that assigns the
    second specified identifier to the first one.
    
    Note: No checking (for example type checking) will be performed, because
    the context of the assignment is not known.
    
    @param   left_hand_side_identifier:  The identifier to be used on the
                                         left-hand side of the assignment.
    @type    left_hand_side_identifier:  string
    @param   right_hand_side_identifier: The identifier to be used on the
                                         right-hand side of the assignment.
    @type    right_hand_side_identifier: string
    @return: A tree representation of the assignment.
    @rtype:  expressions.BinaryOperatorTreeNode (representing an assignment)
    """
    left_hand_side = expressions.IdentifierTreeNode(
                        left_hand_side_identifier, 0, 0) 
    right_hand_side = expressions.IdentifierTreeNode(
                        right_hand_side_identifier, 0, 0)
    return expressions.BinaryOperatorTreeNode(
            left_hand_side, right_hand_side, expressions.OperatorType.ASSIGN,
            0, 0)


def create_slim_enum_value_condition(enum_element_name, allowed_values):
    """Returns a tree representation of the condition that is true iff the SLIM
    enum element with the specified name has one of the specified values.
    
    Note: It will not be checked whether the specified values are valid for the
    enum element because the context of the element is not known.
    
    @param   enum_element_name: The name of the enum data element to be
                                considered.
    @type    enum_element_name: string
    @param   allowed_values:    An iterable object providing the allowed values
                                for the enum element.
    @type    allowed_values:    An iterable of strings.
    @return: A tree representation of the condition.
    @rtype:  expressions.BinaryOperatorTreeNode (representing a boolean
             expression)
    """
    
    # If no values are allowed for the enum data element, then return the
    # boolean node representing "false".
    if len(allowed_values) == 0:
        return expressions.BooleanTreeNode(False, 0, 0)
    
    # There is at least one allowed value. Create the condition that is true
    # iff the enum element has one of these values.
    # The variable to represent this condition is initialized with None.
    condition_tree = None
    for value in allowed_values:
        # Create a tree node representing the enum element name.
        name_node = expressions.IdentifierTreeNode(enum_element_name, 0, 0)
        # Create a tree node representing the current value.
        value_node = expressions.IdentifierTreeNode(value, 0, 0)
        
        # Check whether the current value is the first allowed value.
        if condition_tree == None:
            # The current value is the first allowed value, so initialize the
            # condition such that it is true iff the enum element has the
            # current value.
            condition_tree = expressions.BinaryOperatorTreeNode(
                            name_node, value_node,
                            expressions.OperatorType.EQUAL, 0, 0)
        else:
            # The current value is not the first allowed value, so create the
            # disjunction of the current condition and the condition that is
            # true iff the enum element has the current value.
            additional_condition_tree = expressions.BinaryOperatorTreeNode(
                    name_node, value_node,
                    expressions.OperatorType.EQUAL, 0, 0)
            condition_tree = expressions.BinaryOperatorTreeNode(
                    condition_tree, additional_condition_tree,
                    expressions.OperatorType.OR, 0, 0)
            
    # Return the constructed tree representation of the desired condition.
    return condition_tree


def create_conjunction_of_slim_expressions(boolean_expression_tree1,
                                           boolean_expression_tree2):
    """Returns a tree representation of the conjunction of the two specified
    boolean SLIM expressions.
    
    Note: One or both specified parameters may be None.
    If one parameter is None, the other parameter will be returned.
    If both parameters are None, then None will be returned.
    
    @param   boolean_expression_tree1: The tree representation of the first
                                       boolean expression.
    @type    boolean_expression_tree1: expressions.ExpressionTreeNode
    @param   boolean_expression_tree2: The tree representation of the second
                                       boolean expression.
    @type    boolean_expression_tree2: expressions.ExpressionTreeNode
    @return: A tree representation of the conjunction of the specified boolean
             expressions.
    @rtype:  expressions.ExpressionTreeNode
    """
    # If both parameters are None, return None.
    if (boolean_expression_tree1 is None and
            boolean_expression_tree2 is None):
        return None
    
    # At least one parameter is not None. If one parameter is None, return the
    # other one.
    if boolean_expression_tree1 is None:
        return boolean_expression_tree2
    elif boolean_expression_tree2 is None:
        return boolean_expression_tree1
    
    # Both parameters are not None. Return the tree representing the
    # conjunction of the two boolean expressions.
    return expressions.BinaryOperatorTreeNode(boolean_expression_tree1,
                                              boolean_expression_tree2,
                                              expressions.OperatorType.AND,
                                              0, 0)


def get_slim_assignments_union(assignments1, assignments2):
    """Returns a list representing the union of the specified containers of
    SLIM assignments, where assignments of the first container override those
    of the second container.
    
    All assignments must be in tree representation.
    
    Assignments can be used as effects of mode transitions and as effects of
    faults (when performing fault injection).
    An assignment has the form "x := some_expression".
    If both specified containers contain an assignment for the same variable,
    say x, then the union will only contain the assignment from the first
    container.
    
    The order of the assignments will be preserved and the assignments of the
    first container will precede those of the second container.
    
    @precondition:
        There must not be multiple assignments for the same variable within
        one of the specified container.
    
    @param   assignments1: The first container of assignments. It is
                           prioritized over the second one.
    @type    assignments1: A container of expressions.BinaryOperatorTreeNode
                           objects.
    @param   assignments2: The second container of assignments.
    @type    assignments2: A container of expressions.BinaryOperatorTreeNode
                           objects.
    @return: The union of the assignments.
    @rtype:  A list of expressions.BinaryOperatorTreeNode objects.
    """
    # Store all variable names, for which there is an assignment in the first
    # specified container, in a set.
    variables_set = set([get_slim_assignment_variable_name(assignment)
                        for assignment in assignments1])
    # Initialize the list to be returned with the elements in the first
    # specified assignment list.
    union = assignments1
    # Compute the list of all assignments in the second specified container
    # that do not conflict with the assignments of the first container (i.e. do
    # not assign to the same variables).
    # Add these assignments to the list to be returned.
    filtered_assignments2 = [assignment for assignment in assignments2
                            if get_slim_assignment_variable_name(assignment)
                                not in variables_set]
    union.extend(filtered_assignments2)
    return union


def get_slim_assignment_variable_name(assignment_tree):
    """Returns the name of the variable on the left-hand side of the specified
    assignment.
    
    The assignment must be specified in tree representation.
    
    Example: For an assignment of the form "x := some_expression" (but
    represented as a tree) "x" will be returned.

    @param   assignment_tree:    The assignment to be considered.
    @type    assignment_tree:    expressions.BinaryOperatorTreeNode
    @return: The name of the variable on the left-hand side of the specified
             assignment.
    @rtype:  string
    """
    # Assert that the specified tree represents an assignment and that the left
    # hand side represents some identifier.
    assert isinstance(assignment_tree, expressions.BinaryOperatorTreeNode)
    assert assignment_tree.operator == expressions.OperatorType.ASSIGN
    assert isinstance(assignment_tree.left, expressions.IdentifierTreeNode)
    # Return the name of the variable on the left-hand side of the assignment.
    return assignment_tree.left.name


def parse_slim_assignment(assignment_string):
    """Parses the specified SLIM assignment string.
    
    @param   assignment_string: The SLIM assignment to be parsed.
    @type    assignment_string: string
    @return: A tree representation of the specified SLIM assignment.
    @rtype:  expressions.BinaryOperatorTreeNode
    @raises  testexpr.InvalidExpressionException: If the specified assignment
                is syntactically incorrect.
    """
    # Note: If the effect is syntactically incorrect, a
    # testexpr.InvalidExpressionException will be raised.
    return testexpr.parseAssignment(assignment_string)
