# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================


"""This module assembles various functions needed for renaming elements.

The following aspects are handled:
    - Underscore Elimination (no "_" on access paths)
    - Safe Names for Properties & Placeholders (valid as variable and file names)
    - Fill up empty in modes lists (make implicit understanding that they contain every mode explicit)
"""

from shared import *





# ==================================================================================================
# Underscore Elimination
# ==================================================================================================


def recursivelyEliminateUnderscoresInComponentsDataElementsAndSubcomponents(componentImplementation):
    """Eliminates underscores in names of all data elements (in data port, out data port, data subcomponents) and of non-data subcomponents of the given component implementation and its type and recursively for all contained non-data subcomponents.

    This has to be done since we use underscores in variable names to separate the steps
    in an access path. Assume a component has a data element "some_name" and a non-data
    subcomponent "some" with a data element "name" of the same type. Both would lead to
    the variable name "dscStore_<accessPathToComponent>_some_name" which causes errors.
    Therefore the elimination has to be done.

    Of course we cannot just delete underscores but have to make sure that no name
    clashes are introduced. Therefore the names in the component implementation, in its
    type and in all other implementations of that component type have to be considered
    (e.g. renaming a data port in the component type requires a name disjoint from all
    data element names in all implementations), cf. the syntactic restrictions:
        D-2 All (event or data) port identifiers declared within a component type have to be distinct.
        D-5 The value identifiers of all enum types used in a component type have to be distinct from the data port identifiers of that component.
        F-1 All declared subcomponent identifiers have to be distinct.
        F-13 For each component implementation, the identifiers of all data subcomponents and the identifiers of all data ports declared in the corresponding type specification have to be distinct.
        F-14 The value identifiers of all enum types used in a component implementation have to be distinct from the data port and the data subcomponent identifiers of that component.

    It does not suffice to change the declarations of names only, but every usage of
    a name (in port connections/flows, expressions for guards/effects/properties) has
    to be renamed accordingly. As this is not limited to the current component but also
    has to be done in supercomponents and properties, an additional renaming mapping is
    stored in each type and in each implementation for later reuse. The renaming mapping
    is a dictionary of the form:
        { (oldName, typ): newName }
            , where typ is a constant in [DATAELEMENT, NONDATASUBCOMPONENT]
              - The typ distinction is important as some data elements might have the same original name as a non data subcomponent but are renamed differently.
              - The renaming mapping of component types only can contain DATAELEMENT
    
    The names of enums, modes, and event ports do not have to be considered as they are
    not used to generate variable names.

    @type componentImplementation:  instance of slim_compiler.src.dataStructure.nominalbehaviour.ComponentImplementation
    @param componentImplementation: Root of the component tree for which renaming is
                                    to be done. The changes are done in place in the
                                    componentImplementation so that it stays in the
                                    compile information for later reuse. However,
                                    if you do not want to loose the original compile
                                    info, use copy.deepcopy before.
    """


    # *** Before eliminating underscores in current component, recursively do so for its non-data subcomponents ***
    for scObject in componentImplementation.non_data_subcomponents.values():
        recursivelyEliminateUnderscoresInComponentsDataElementsAndSubcomponents(scObject.get_referenced_component_implementation())

    
    # *** First, collect all data element/port/subcomponent names to simplify checking for name clashes ***

    # Ports are defined as features in the component type and not its implementation(s) => get component type object
    componentType = componentImplementation.get_referenced_component_type()

    # Considering only one implementation and its type does not suffice (see docstring) => get list of other implementations of same type (which must be in the same package)
    otherImplementations = componentType.subpackage.package.get_implementations_for_type(componentType.name, False)
    otherImplementations.remove(componentImplementation) # current implementation is not an "other" one

    # Initialize lists for names - different ones as some elements are allowed to have the same name (e.g., event ports and data subcomponents)
    ownDataElementNames = []    # enum names might occur more than once in the list, but as they do not have to be renamed = deleted from list, this does not matter
    othersDataElementNames = [] # names might occure more than once in this list, but as they are not renamed = deleted from list in this function call, this does not matter
    portNames = []
    ownSubcomponentNames = []

    # In Data Ports
    # componentType.*_*_ports = dict(portName: string, portObject: instance(slim_compiler.src.dataStructure.nominalbehaviour.*Port))
    for (portName, portObject) in componentType.in_data_ports.items():
        ownDataElementNames.append(portName)
        portNames.append(portName)
        if isinstance(portObject.data_type, Enum):
            ownDataElementNames.extend(portObject.data_type.values)

    # Out Data Ports
    # componentType.*_*_ports = dict(portName: string, portObject: instance(slim_compiler.src.dataStructure.nominalbehaviour.*Port))
    for (portName, portObject) in componentType.out_data_ports.items():
        ownDataElementNames.append(portName)
        portNames.append(portName)
        if isinstance(portObject.data_type, Enum):
            ownDataElementNames.extend(portObject.data_type.values)

    # In & Out Event Ports
    portNames.extend(componentType.in_event_ports.keys())
    portNames.extend(componentType.out_event_ports.keys())

    # Own Data Subcomponents
    # componentImplementation.data_subcomponents = dict(scName: string, scObject: instance(slim_compiler.src.dataStructure.nominalbehaviour.DataSubcomponent))
    for (scName, scObject) in componentImplementation.data_subcomponents.items():
        ownDataElementNames.append(scName)
        ownSubcomponentNames.append(scName)
        if isinstance(scObject.data_type, Enum):
            ownDataElementNames.extend(scObject.data_type.values)

    # Own Non-Data Subcomponents
    # componentImplementation.non_data_subcomponents = dict(scName: string, scObject: instance(slim_compiler.src.dataStructure.nominalbehaviour.NonDataSubcomponent))
    ownSubcomponentNames.extend(componentImplementation.non_data_subcomponents.keys())

    # Others Data Subcomponents & Enum Names
    for otherImplementation in otherImplementations:
        for (scName, scObject) in otherImplementation.data_subcomponents.items():
            othersDataElementNames.append(scName)
            if isinstance(scObject.data_type, Enum):
                othersDataElementNames.extend(scObject.data_type.values)


    # *** Now, eliminate underscores in declarations of data ports (in Type) and subcomponents (in Implementation) ***

    # An inner function which helps to find a new name without "_" and which does not collide with a list of names
    def newName(oldName, aNameList):
        if oldName.find("_") >= 0:
            oldName = oldName.replace("_", "X") + "X"
            cnt = 0
            while (oldName + str(cnt)) in aNameList:
                cnt = cnt + 1
            oldName = oldName + str(cnt)   
        return oldName

    # Add slim2promela_renamingMapping attribute to componentType & componentImplementation
    if not hasattr(componentType, "slim2promela_renamingMapping"):
        componentType.slim2promela_renamingMapping = {}
    if not hasattr(componentImplementation, "slim2promela_renamingMapping"):
        componentImplementation.slim2promela_renamingMapping = {}
    
    # In Data Ports
    # componentType.*_*_ports = dict(portName: string, portObject: instance(slim_compiler.src.dataStructure.nominalbehaviour.*Port))
    for (portName, portObject) in filter(lambda (portName, portObject): portName.find("_") >= 0, componentType.in_data_ports.items()):
        # delete data port with old name from lists
        ownDataElementNames.remove(portName)
        portNames.remove(portName)
        componentType.in_data_ports.pop(portName)
        # adapt name (only in portObject.name - portName stays the old name!)
        portObject.name = newName(portName, ownDataElementNames + portNames + othersDataElementNames)
        # add data port with new name to lists
        ownDataElementNames.append(portObject.name)
        portNames.append(portObject.name)
        componentType.in_data_ports.update({portObject.name: portObject})
        # store renaming in mapping
        componentType.slim2promela_renamingMapping.update({ (portName, DATAELEMENT) : portObject.name })

    # Out Data Ports
    # componentType.*_*_ports = dict(portName: string, portObject: instance(slim_compiler.src.dataStructure.nominalbehaviour.*Port))
    for (portName, portObject) in filter(lambda (portName, portObject): portName.find("_") >= 0, componentType.out_data_ports.items()):
        # delete data port with old name from lists
        ownDataElementNames.remove(portName)
        portNames.remove(portName)
        componentType.out_data_ports.pop(portName)
        # adapt name (only in portObject.name - portName stays the old name!)
        portObject.name = newName(portName, ownDataElementNames + portNames + othersDataElementNames)
        # add data port with new name to lists
        ownDataElementNames.append(portObject.name)
        portNames.append(portObject.name)
        componentType.out_data_ports.update({portObject.name: portObject})
        # store renaming in mapping
        componentType.slim2promela_renamingMapping.update({ (portName, DATAELEMENT) : portObject.name })

    # Data Subcomponents
    # componentImplementation.data_subcomponents = dict(scName: string, scObject: instance(slim_compiler.src.dataStructure.nominalbehaviour.DataSubcomponent))
    for (scName, scObject) in filter(lambda (scName, scObject): scName.find("_") >= 0, componentImplementation.data_subcomponents.items()):
        # delete subcomponent with old name from lists
        ownDataElementNames.remove(scName)
        ownSubcomponentNames.remove(scName)
        componentImplementation.data_subcomponents.pop(scName)
        # adapt name (only in scObject.name - scName stays the old name!)
        scObject.name = newName(scName, ownDataElementNames + ownSubcomponentNames)
        # add subcomponent with new name to lists
        ownDataElementNames.append(scObject.name)
        ownSubcomponentNames.append(scObject.name)
        componentImplementation.data_subcomponents.update({scObject.name: scObject})
        # store renaming in mapping
        componentImplementation.slim2promela_renamingMapping.update({ (scName, DATAELEMENT) : scObject.name })

    # Non-Data Subcomponents
    # componentImplementation.non_data_subcomponents = dict(scName: string, scObject: instance(slim_compiler.src.dataStructure.nominalbehaviour.NonDataSubcomponent))
    for (scName, scObject) in filter(lambda (scName, scObject): scName.find("_") >= 0, componentImplementation.non_data_subcomponents.items()):
        # delete subcomponent with old name from lists
        ownSubcomponentNames.remove(scName)
        componentImplementation.non_data_subcomponents.pop(scName)
        # adapt name (only in scObject.name - scName stays the old name!)
        scObject.name = newName(scName, ownSubcomponentNames)
        # add subcomponent with new name to lists
        ownSubcomponentNames.append(scObject.name)
        componentImplementation.non_data_subcomponents.update({scObject.name: scObject})
        # store renaming in mapping
        componentImplementation.slim2promela_renamingMapping.update({ (scName, NONDATASUBCOMPONENT) : scObject.name })


    # *** Finally, apply renaming to all usages of data ports and subcomponents ***

    # An inner function which walks through an expression and applies renaming in it
    def adaptExpression(anExpressionTreeNode, componentImplementation = componentImplementation, componentType = componentType):
        # Distinguish the different Node Types and adapt them appropriately
        if anExpressionTreeNode is None:
            pass
        elif isinstance(anExpressionTreeNode, BinaryOperatorTreeNode):
            adaptExpression(anExpressionTreeNode.left)
            adaptExpression(anExpressionTreeNode.right)
        elif isinstance(anExpressionTreeNode, UnaryOperatorTreeNode):
            adaptExpression(anExpressionTreeNode.subexpression)
        elif isinstance(anExpressionTreeNode, CaseExpressionTreeNode):
            # Cases: anExpressionTreeNode.cases = list(condition: expression, expression: expression)
            for aCase in anExpressionTreeNode.cases:
                adaptExpression(aCase[0])
                adaptExpression(aCase[1])
            # Default Case
            adaptExpression(anExpressionTreeNode.default_expression)
        # - IDENTIFIER -                    
        elif isinstance(anExpressionTreeNode, IdentifierTreeNode):
            # a) non-data subcomponent's data port
            if anExpressionTreeNode.name.find(".") >= 0:
                (scName, scPort) = anExpressionTreeNode.name.split(".")
                if (scName, NONDATASUBCOMPONENT) in componentImplementation.slim2promela_renamingMapping:
                    scName = componentImplementation.slim2promela_renamingMapping[(scName, NONDATASUBCOMPONENT)]
                if (scPort, DATAELEMENT) in componentImplementation.non_data_subcomponents[scName].get_referenced_component_type().slim2promela_renamingMapping:
                    scPort = componentImplementation.non_data_subcomponents[scName].get_referenced_component_type().slim2promela_renamingMapping[(scPort, DATAELEMENT)]
                anExpressionTreeNode.name = scName + "." + scPort
            # b) own data port or data subcomponent
            else:
                if (anExpressionTreeNode.name, DATAELEMENT) in componentType.slim2promela_renamingMapping:
                    anExpressionTreeNode.name = componentType.slim2promela_renamingMapping[(anExpressionTreeNode.name, DATAELEMENT)]
                elif (anExpressionTreeNode.name, DATAELEMENT) in componentImplementation.slim2promela_renamingMapping:
                    anExpressionTreeNode.name = componentImplementation.slim2promela_renamingMapping[(anExpressionTreeNode.name, DATAELEMENT)]
        # - /IDENTIFIER -
        elif isinstance(anExpressionTreeNode, BooleanTreeNode) or isinstance(anExpressionTreeNode, NumberTreeNode):
            pass
        elif isinstance(anExpressionTreeNode, FlowEquationTreeNode) or isinstance(anExpressionTreeNode, FlowExpressionTreeNode):
            pass # TODO: "ERR: Flow Equations for continuous variables are not supported"
        else:
            pass # TODO: "ERR: Unknown ExpressionTreeNode"

    # Data Port Connections
    # componentImplementation.data_port_connections = list(connection: instance(slim_compiler.src.dataStructure.nominalbehaviour.PortConnection))
    for connection in componentImplementation.data_port_connections:
        
        # connection.source_port
        if connection.source_port.find(".") >= 0: # => non-data subcomponent's data port
            (scName, scPort) = connection.source_port.split(".")
            if (scName, NONDATASUBCOMPONENT) in componentImplementation.slim2promela_renamingMapping:
                scName = componentImplementation.slim2promela_renamingMapping[(scName, NONDATASUBCOMPONENT)]
            if (scPort, DATAELEMENT) in componentImplementation.non_data_subcomponents[scName].get_referenced_component_type().slim2promela_renamingMapping:
                scPort = componentImplementation.non_data_subcomponents[scName].get_referenced_component_type().slim2promela_renamingMapping[(scPort, DATAELEMENT)]
            connection.source_port = scName + "." + scPort
        else: # => own data port 
            if (connection.source_port, DATAELEMENT) in componentType.slim2promela_renamingMapping:
                connection.source_port = componentType.slim2promela_renamingMapping[(connection.source_port, DATAELEMENT)]

        # connection.target_port
        if connection.target_port.find(".") >= 0: # => non-data subcomponent's data port
            (scName, scPort) = connection.target_port.split(".")
            if (scName, NONDATASUBCOMPONENT) in componentImplementation.slim2promela_renamingMapping:
                scName = componentImplementation.slim2promela_renamingMapping[(scName, NONDATASUBCOMPONENT)]
            if (scPort, DATAELEMENT) in componentImplementation.non_data_subcomponents[scName].get_referenced_component_type().slim2promela_renamingMapping:
                scPort = componentImplementation.non_data_subcomponents[scName].get_referenced_component_type().slim2promela_renamingMapping[(scPort, DATAELEMENT)]
            connection.target_port = scName + "." + scPort
        else: # => own data port 
            if (connection.target_port, DATAELEMENT) in componentType.slim2promela_renamingMapping:
                connection.target_port = componentType.slim2promela_renamingMapping[(connection.target_port, DATAELEMENT)]

    # Event Port Connections
    for connection in componentImplementation.event_port_connections:

        # connection.source_port
        if connection.source_port.find(".") >= 0: # => non-data subcomponent's event port
            (scName, scPort) = connection.source_port.split(".")
            if (scName, NONDATASUBCOMPONENT) in componentImplementation.slim2promela_renamingMapping:
                scName = componentImplementation.slim2promela_renamingMapping[(scName, NONDATASUBCOMPONENT)]
            connection.source_port = scName + "." + scPort

        # connection.target_port
        if connection.target_port.find(".") >= 0: # => non-data subcomponent's event port
            (scName, scPort) = connection.target_port.split(".")
            if (scName, NONDATASUBCOMPONENT) in componentImplementation.slim2promela_renamingMapping:
                scName = componentImplementation.slim2promela_renamingMapping[(scName, NONDATASUBCOMPONENT)]
            connection.target_port = scName + "." + scPort

    # Flows
    for flow in componentImplementation.flows:
        adaptExpression(flow.expression)

    # Normal Mode Transitions: Trigger, Guard, Effects
    for transition in componentImplementation.normal_mode_transitions:
        
        # transition.label.trigger
        if transition.label.trigger is not None and transition.label.trigger.find(".") >= 0: # => non-data subcomponent's event port
            (scName, scPort) = transition.label.trigger.split(".")
            if (scName, NONDATASUBCOMPONENT) in componentImplementation.slim2promela_renamingMapping:
                scName = componentImplementation.slim2promela_renamingMapping[(scName, NONDATASUBCOMPONENT)]
            transition.label.trigger = scName + "." + scPort

        # transition.label.guard
        adaptExpression(transition.label.guard)

        # transition.label.effects
        for effect in transition.label.effects:
            adaptExpression(effect)
            
# </def _recursivelyEliminateUnderscoresInComponentsDataElementsAndSubcomponents: ...>



# --------------------------------------------------------------------------------------------------



def performUnderscoreEliminationInAccessPath(accessPath, componentImplementation):
    """TODO: Describe.
    Returns the new string.
    @return: Tuple (variableNameFragment: string, typ: constant defined in XXX)
    """
    
    # A) Descend to next non-data subcomponent
    if accessPath.find(".") >= 0:
        # Split identifier of next non-data subcomponent in accessPath off
        (scName, accessPathRest) = accessPath.split(".", 1)
        # Perform renaming, if necessary
        if (scName, NONDATASUBCOMPONENT) in componentImplementation.slim2promela_renamingMapping:
            scName = componentImplementation.slim2promela_renamingMapping[(scName, NONDATASUBCOMPONENT)]
        # Retriev subcomponent's implementation
        if scName in componentImplementation.non_data_subcomponents:
            scImplementation = componentImplementation.non_data_subcomponents[scName].get_referenced_component_implementation()
        else:
            return "(ERR: Subcomponent '" + scName + "' does not exist!).\"" + accessPathRest + "\""   # TODO: Throw error
        # Recursively proceed translation
        accessPathTail = performUnderscoreEliminationInAccessPath(accessPathRest, scImplementation)
        # Return combined result
        return scName + "." + accessPathTail

    
    # B) Reached last entry of accessPath
    
    # B.1) nominal/error mode
    elif accessPath in ["mode", "error"]:
        return accessPath
    
    # B.3) data element (data port or data subcomponent)
    else:
        # We also need to consider the component type
        componentType = componentImplementation.get_referenced_component_type()
        # Perform renaming, if necessary
        if (accessPath, DATAELEMENT) in componentImplementation.slim2promela_renamingMapping: # must be a DSC, as stored in Implementation
            accessPath = componentImplementation.slim2promela_renamingMapping[(accessPath, DATAELEMENT)]
        elif (accessPath, DATAELEMENT) in componentType.slim2promela_renamingMapping: # must be a DP, as stored in Type
            accessPath = componentType.slim2promela_renamingMapping[(accessPath, DATAELEMENT)]
        # Return result
        return accessPath


def performUnderscoreEliminationInAtomicProposition(anExpressionTreeNode, rootComponentImplementation):
    """Makes the changes in place - use copy.deepcopy before, if you want to!"""

    if anExpressionTreeNode is None:
        pass
    
    elif isinstance(anExpressionTreeNode, BinaryOperatorTreeNode):
        performUnderscoreEliminationInAtomicProposition(anExpressionTreeNode.left, rootComponentImplementation)
        performUnderscoreEliminationInAtomicProposition(anExpressionTreeNode.right, rootComponentImplementation)

    elif isinstance(anExpressionTreeNode, UnaryOperatorTreeNode):
        performUnderscoreEliminationInAtomicProposition(anExpressionTreeNode.subexpression, rootComponentImplementation)
        
    elif isinstance(anExpressionTreeNode, CaseExpressionTreeNode):
        # Cases: anExpressionTreeNode.cases = list(condition: expression, expression: expression)
        for (b, e) in anExpressionTreeNode.cases:
            performUnderscoreEliminationInAtomicProposition(b)
            performUnderscoreEliminationInAtomicProposition(e)
        # Default Case
        performUnderscoreEliminationInAtomicProposition(anExpressionTreeNode.default_expression)
        
    # - IDENTIFIER -                    
    elif isinstance(anExpressionTreeNode, IdentifierTreeNode):
        anExpressionTreeNode.name = performUnderscoreEliminationInAccessPath(anExpressionTreeNode.name, rootComponentImplementation)
    # - /IDENTIFIER -
    
    else: # esp. for BooleanTreeNode, NumberTreeNode, FlowEquationTreeNode, FlowExpressionTreeNode
        pass


def iterativelyPerformUnderscoreEliminationInAtomicPropositions(atomicPropositions, rootComponentImplementation):
    for (placeholderName, atomicProposition) in atomicPropositions.items():
        performUnderscoreEliminationInAtomicProposition(atomicProposition, rootComponentImplementation)


def iterativelyPerformUnderscoreEliminationInProperties(propertiesDict, rootComponentImplementation):
    for (propertyName, (safeName, patternId, atomicPropositions)) in propertiesDict.items():
        iterativelyPerformUnderscoreEliminationInAtomicPropositions(atomicPropositions, rootComponentImplementation)


# ==================================================================================================
# /Underscore Elimination
# ==================================================================================================










# ==================================================================================================
# Safe Names for Properties & Placeholders
# ==================================================================================================

def _getNewSafeName(oldName, listOfNames):
    """TODO: Describe.
    @return: ...
    """

    # 1.) Remove all invalid characters
    newName = string.translate(oldName, string.maketrans("", ""), '\x00\x01\x02\x03\x04\x05\x06\x07\x08\t\n\x0b\x0c\r\x0e\x0f\x10\x11\x12\x13\x14\x15\x16\x17\x18\x19\x1a\x1b\x1c\x1d\x1e\x1f !"#$%&\'()*+,-./:;<=>?@[\\]^_`{|}~\x7f\x80\x81\x82\x83\x84\x85\x86\x87\x88\x89\x8a\x8b\x8c\x8d\x8e\x8f\x90\x91\x92\x93\x94\x95\x96\x97\x98\x99\x9a\x9b\x9c\x9d\x9e\x9f\xa0\xa1\xa2\xa3\xa4\xa5\xa6\xa7\xa8\xa9\xaa\xab\xac\xad\xae\xaf\xb0\xb1\xb2\xb3\xb4\xb5\xb6\xb7\xb8\xb9\xba\xbb\xbc\xbd\xbe\xbf\xc0\xc1\xc2\xc3\xc4\xc5\xc6\xc7\xc8\xc9\xca\xcb\xcc\xcd\xce\xcf\xd0\xd1\xd2\xd3\xd4\xd5\xd6\xd7\xd8\xd9\xda\xdb\xdc\xdd\xde\xdf\xe0\xe1\xe2\xe3\xe4\xe5\xe6\xe7\xe8\xe9\xea\xeb\xec\xed\xee\xef\xf0\xf1\xf2\xf3\xf4\xf5\xf6\xf7\xf8\xf9\xfa\xfb\xfc\xfd\xfe')

    # 2.) Avoid nameclashes by adding a number
    if newName in listOfNames:
        cnt = 2
        while newName + str(cnt) in listOfNames:
            cnt += 1
        newName += str(cnt)

    # 3.) Return result
    return newName


def setSafeNamesForPlaceholders(atomicPropositions):
    alreadyUsedSafeNames = atomicPropositions.keys()
    
    for (oldPlaceholderName, atomicProposition) in atomicPropositions.items():
        atomicPropositions.pop(oldPlaceholderName)
        alreadyUsedSafeNames.remove(oldPlaceholderName)
        newPlaceholderName = _getNewSafeName(oldPlaceholderName, alreadyUsedSafeNames)
        alreadyUsedSafeNames.append(newPlaceholderName)
        atomicPropositions[newPlaceholderName] = atomicProposition


def setSafeNamesForProperties(propertiesDict):

    alreadyUsedSafeNames = [safeName for (safeName, patternId, atomicPropostions) in propertiesDict.values()]

    for (propertyName, (safeName, patternId, atomicPropositions)) in propertiesDict.items():
        alreadyUsedSafeNames.remove(safeName)
        newSafeName = _getNewSafeName(safeName, alreadyUsedSafeNames)
        alreadyUsedSafeNames.append(newSafeName)

        setSafeNamesForPlaceholders(atomicPropositions)
        
        propertiesDict[propertyName] = (newSafeName, patternId, atomicPropositions)
        
# ==================================================================================================
# /Safe Names for Properties & Placeholders
# ==================================================================================================










# ==================================================================================================
# Fill up empty in modes lists
# ==================================================================================================

def fillUpEmptyInModesListsInComponent(componentImplementation):
    # List of all mode names
    completeModeList = componentImplementation.modes.keys()

    # Look at each object that has an in modes list ...
    for element in componentImplementation.non_data_subcomponents.values() + \
                   componentImplementation.data_subcomponents.values() + \
                   componentImplementation.data_port_connections + \
                   componentImplementation.event_port_connections + \
                   componentImplementation.flows:
        # ... set in modes to all modes, if empty
        if element.in_modes in [[], None]:
            element.in_modes = completeModeList # HINT: Lists are elementary data types (no objects) so they are indeed copied.
            

def recursivelyFillUpEmptyInModesLists(componentImplementation):
    # Check valid parameter
    if componentImplementation == None or not isinstance(componentImplementation, ComponentImplementation):
        return

    # Fill up empty in modes lists in this component
    fillUpEmptyInModesListsInComponent(componentImplementation)

    # Recursively proceed with non-data subcomponents
    for scObject in componentImplementation.non_data_subcomponents.values():
        recursivelyFillUpEmptyInModesLists(scObject.get_referenced_component_implementation())
    
# ==================================================================================================
# /Fill up empty in modes lists
# ==================================================================================================
