# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================


from shared import *
from translation import *
from slim_compiler.src.slicing.slicing import *

import getopt
import subprocess
import slim_compiler.src.smv.api
import utils.debug_info

from slim_compiler.src.dataStructure.visitors.printVisitor import toSLIMVisitor


# ==================================================================================================
# Preconfigured Examples
# ==================================================================================================

# Format: (SLIMFILENAMES, FAULTINJECTIONFILENAME, PROPERTIESDOCUMENTFILENAME, ROOTNAME)

adder = (["examples\\Adder.slim"], "", "", "")

halfAdder = (["examples\\HalfAdder.slim"], "", "examples\\HalfAdder.xml", "")

adderWithFaultInjection = (["examples\\Adder.slim", "examples\\Adder_err.slim"], "examples\\Adder-correctFaultInjection.xml", "", "")

properties2promela_Test = (["tests\\properties2promela-concrete.slim"], "", "tests\\properties2promela-concrete.prop.xml", "")

transitionChangesDataPort = (["examples\\TransitionChangesDataPort.slim"], "", "", "")

acyclicConnectionButCyclicOnComponentLevel = (["examples\\acyclicConnectionButCyclicOnComponentLevel.slim"], "", "", "")

checkCheckerAndAnnotations = (["tests\\checkCheckerAndAnnotations.slim"], "", "", "")

checkExpressionNodeTypes = (["tests\\checkExpressionNodeTypes.slim"], "", "", "")



chosenExample = properties2promela_Test # halfAdder # adder # checkCheckerAndAnnotations # checkExpressionNodeTypes # acyclicConnectionButCyclicOnComponentLevel #transitionChangesDataPort #



if False: # set this to true to use preconfigured examples instead of input parameters
    (SLIMFILENAMES, FAULTINJECTIONFILENAME, PROPERTIESDOCUMENTFILENAME, ROOTNAME) = chosenExample
    sys.argv = ["x"]
    sys.argv.extend(SLIMFILENAMES)
    if FAULTINJECTIONFILENAME != "":
        sys.argv.extend(["-f", FAULTINJECTIONFILENAME])
    if PROPERTIESDOCUMENTFILENAME != "":
        sys.argv.extend(["-p", PROPERTIESDOCUMENTFILENAME])
    if ROOTNAME != "":
        sys.argv.extend(["-r", ROOTNAME])
    sys.argv.append("--noslicing")
    sys.argv.append("--nospin")
    sys.argv.append("--generatesmv")

# ==================================================================================================
# /Preconfigured Examples
# ==================================================================================================





# ==================================================================================================
# Generate SMV Details
# ==================================================================================================

class DummyAtomicProposition:
    def __init__(self, text):
        self.text = text
    
class DummyFormula:
    def __init__(self, listOfAtomicPropositionTrees):
        self.atomicPropositionsSLIM = []
        for atomicPropositionTree in listOfAtomicPropositionTrees:
            atomicPropositionSLIM = toSLIMVisitor().visit(atomicPropositionTree)
            self.atomicPropositionsSLIM.append(DummyAtomicProposition(atomicPropositionSLIM))
        
    def get_atomic_propositions(self):
        return self.atomicPropositionsSLIM



def _generateSMV(rootComponent, listOfAtomicPropositionDicts, baseFileName):
    # Prepare
    options.BASE_FILE_NAME = baseFileName
    myDebugInfo = utils.debug_info.create_debug_and_compile_information()

    # Translate to SMV
    sys.stdout.write("Translate to SMV ... ")
    sys.stdout.flush()
    smvFile = open(baseFileName + ".smv", "w")
    slim_compiler.src.smv.api.convert_to_smv(rootComponent, smvFile)
    smvFile.close()
    sys.stdout.write("OK.\n")
    if VERBOSE:
        printOutputBlock("Translation Result:   The SMV code", open(baseFileName + ".smv").read())

    # Dump State Labelling
    sys.stdout.write("Dump state labelling ... ")
    sys.stdout.flush()
    if listOfAtomicPropositionDicts != None:
        # Normally, we would create a formulae document on the basis of the properties document:
        #     formulaeDocument = api.create_formulae_document_for_properties(propertiesDocument)
        # However, we want to do so for selected properties only. Because of that we
        # create "DummyFormulae" from the given atomic propositions which mimick
        # the behaviour of "real" Formulas. The list of formulas is then used
        # instead of the formulae document.
        listOfDummyFormulae = [DummyFormula(atomicPropositions.values()) for atomicPropositions in listOfAtomicPropositionDicts]
        invokers.dump.dump_ap_map_and_replace_aps(rootComponent, listOfDummyFormulae, myDebugInfo)
        sys.stdout.write("OK\n")
        if VERBOSE:
            printOutputBlock("AP Dump", str(myDebugInfo.label_ap_map))
    else:
        sys.stdout.write("SKIP (No atomic propositions available.)\n")

    # Dump Probability Map
    sys.stdout.write("Dump probability map ... ")
    sys.stdout.flush()
    invokers.dump.dump_error_event_probability_map(rootComponent, myDebugInfo)
    sys.stdout.write("OK\n")
    if VERBOSE:
        printOutputBlock("Probability Dump", str(myDebugInfo.error_event_probability_map))
    
# ==================================================================================================
# /Generate SMV Details
# ==================================================================================================





# ==================================================================================================
# Invoke SPIN Details
# ==================================================================================================

def executionCommands(patternId, atomicPropositions, pmlFileName):
    
    propertyToSpinFormulaDict = {
        "propositional":      ("!( gflagGoNextTransition U (!gflagGoNextTransition && {Placeholder:P}) )", ["-DSAFETY"], []),
        "absenceGlobal":      ("![]!(!gflagGoNextTransition && {Placeholder:P})", ["-DSAFETY"], []),
        "existenceGlobal":    ("!<>(!gflagGoNextTransition && {Placeholder:P})", [], ["-a"]), # acceptance cycles still needed!!!
        "universalityGlobal": ("![](!gflagGoNextTransition -> {Placeholder:P})", ["-DSAFETY"], []),
        "precedenceGlobal":   ("!(  [](!gflagGoNextTransition -> !{Placeholder:S})  ||  ( (!gflagGoNextTransition -> !{Placeholder:S}) U (!gflagGoNextTransition && {Placeholder:P}) )  )", ["-DSAFETY"], []), # does not need acceptance cycles any more: [], ["-a"]),
        "responseGlobal":     ("![]( (!gflagGoNextTransition && {Placeholder:P}) -> <>(!gflagGoNextTransition && {Placeholder:S}) )", [], ["-a"]) # acceptance cycles still needed!!!
        }

    if patternId not in propertyToSpinFormulaDict:
        printErrorMessageAndStop("Unknown LTL-formula pattern \"" + str(patternId) + "\" - STOP.", 1)

    (formula, gccParams, panParams) = propertyToSpinFormulaDict[patternId]

    # Additional parameters for memory management
    if COLLAPSESTATEVECTOR:
        gccParams.append("-DCOLLAPSE")
    if STORESTATESINMA:
        gccParams.append("-DMA=" + str(STORESTATESINMA))
    if MEMORYLIMIT:
        gccParams.append("-DMEMLIM=" + str(MEMORYLIMIT))
    if VECTORSIZE:
        gccParams.append("-DVECTORSZ=" + str(VECTORSIZE))
    if USEBFSFORSAFETY and ("-DSAFETY" in gccParams):
        gccParams.append("-DBFS")

    # Generate complete formula
    #for (placeholderName, atomicProposition) in atomicPropositions.items():
    #    formula = formula.replace("{Placeholder:" + placeholderName + "}", "atomicProposition_" + placeholderName)

    #spinCmd = ["spin", "-a", "-f",  formula, pmlFileName]
    spinCmd = ["spin", "-a", "-N", INVOCATIONDIRECTORY + "/neverClaims/neverClaim_" + patternId, pmlFileName]
    gccCmd  = ["gcc"] + gccParams + ["-o", "pan", "pan.c"]
    panCmd  = ["./pan", "-n"] + panParams

    return (spinCmd, gccCmd, panCmd)



def runProgram(name, cmd):
    sys.stdout.write("Run " + name + " (" + " ".join(cmd) + ") ... ")
    sys.stdout.flush()
    
    try:
        if os.name == "nt":
            sys.stdout.write("SIMULATE OUTPUT ON WINDOWS\n")
            return (True, 0, ["pan: acceptance cycle (at depth <Nr>)\n", "  OR\n", "pan: claim violated! (at depth <Nr>)\n", "pan: wrote Example.pml.trail\n", "\n", "(Spin Version 5.1.7 -- 23 December 2008)\n",
                              "Warning: Search not completed\n", "    + Partial Order Reduction\n", "State-vector 20 byte, depth reached 81, errors: 1\n"],
                    ["Dummy\n", "error\n"])
            
        programProcess = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        returnCode = programProcess.wait()
        
        if returnCode == 0:
            sys.stdout.write("OK\n")
            return (True, programProcess.returncode, programProcess.stdout.readlines(), programProcess.stderr.readlines())
        else:
            sys.stdout.write("ERROR\n")
            print "* " + name + " returned with error code: " + str(returnCode)
            print "* Output:"
            stdoutList = programProcess.stdout.readlines()
            for line in stdoutList:
                sys.stdout.write("\t" + line);
            print "* Errors:"
            stderrList = programProcess.stderr.readlines()
            for line in stderrList:
                sys.stdout.write("\t" + line);
            return (False, programProcess.returncode, stdoutList, stderrList)
        
    except BaseException, e:
        sys.stdout.write("ERROR\n")
        print "* " + name + "could not be executed:"
        print "\t" + str(e)
        return (False, 0, [], [])

# ==================================================================================================
# /Invoke SPIN Details
# ==================================================================================================










# ==================================================================================================
# MAIN SCRIPT
# ==================================================================================================

# *** Print Start-of-Program Message ***
print "START OF <promela\script.py>\n\n\n"





# *** Read Input Parameters ***
try:
    (optlist, args) = getopt.gnu_getopt(sys.argv[1:], "f:p:o:r:v",
                                        ["help",
                                         "storefaultinjectionresult", "noslicing", "cycleslicing=", "storeslicingresult", "verboseslicing", "generatesmv", "simulate=", "nospin",
                                         "sdstart=", "sdfactor=", "sdadd=", "sdmax=",
                                         "collapse", "ma=", "memlim=", "vectorsz=", "usebfsforsafety"])
except getopt.GetoptError, err:
    printErrorMessageAndStop("Invalid parameters (" + str(err) + ") - STOP!", 1)
optdict = dict(optlist)

# Show Usage Info, if necessary
if "--help" in optdict:
    print "python script.py -f -p -o -r -v --help --storefaultinjectionresult --noslicing --cycleslicing --storeslicingresult --verboseslicing --generatesmv --simulate --nospin --sdstart --sdfactor --sdadd --sdmax --collapse --ma= --memlim= --vectorsz= --usebfsforsafety"
    sys.exit(1)

# Output ...
printOutputBlock_Header("Input Parameters:")

# Slim Input Files
if len(args) < 1:
    printErrorMessageAndStop("No SLIM input file given - STOP!", 1)
SLIMFILENAMES = [os.path.abspath(relFileName) for relFileName in args]
print "SLIM Input Files:\t\t" + "\n\t\t\t\t".join(SLIMFILENAMES)

# Fault Injection File
if "-f" in optdict:
    FAULTINJECTIONFILENAME = os.path.abspath(optdict["-f"])
else:
    FAULTINJECTIONFILENAME = ""
print "Fault Injection File:\t\t" + FAULTINJECTIONFILENAME

# Properties Document File
if "-p" in optdict:
    PROPERTIESDOCUMENTFILENAME = os.path.abspath(optdict["-p"])
else:
    PROPERTIESDOCUMENTFILENAME = ""
print "Properties Document File:\t" + PROPERTIESDOCUMENTFILENAME

# Output Base File Name
if "-o" in optdict:
    OUTPUTBASEFILENAME = os.path.abspath(optdict["-o"])
else:
    OUTPUTBASEFILENAME = os.path.splitext(SLIMFILENAMES[0])[0]

# Save invocation directory
INVOCATIONDIRECTORY = os.getcwd()
print "Invocation Directory:\t\t" + INVOCATIONDIRECTORY

# Go to working directory (as SPIN puts outputs there from where it is called)
(relWorkingPath, OUTPUTBASEFILENAME) = os.path.split(OUTPUTBASEFILENAME)
if os.path.exists(os.path.abspath(relWorkingPath)):
    os.chdir(os.path.abspath(relWorkingPath))
else:
    printErrorMessageAndStop("Working Path \"" + os.path.abspath(relWorkingPath) + "\" does not exist - STOP!", 1)
print "Output & Working Directory:\t" + os.getcwd()

# Ensure that OutputBaseFileName is not empty
if OUTPUTBASEFILENAME == "":
    OUTPUTBASEFILENAME = "slim2promela"
print "Output Base File Name:\t\t" + OUTPUTBASEFILENAME

# Name of root component
if "-r" in optdict:
    ROOTNAME = optdict["-r"]
else:
    ROOTNAME = ""
print "Root Name:\t\t\t" + ROOTNAME
options.ROOT_NAME = ROOTNAME

# Simulation
if "--simulate" in optdict:
    try:
        SIMULATIONDEPTH = int(optdict["--simulate"])
    except BaseException, e:
        printErrorMessageAndStop("Value \"" + optdict["--simulate"] + "\" is invalid for --simulate.", 1)
    SIMULATION = True
    print "Simulation:\t\t\tTrue, depth = " + str(SIMULATIONDEPTH)
    print "\t\t\t\tWARNING: Model Checking with enabled simulation info might be slower!"
else:
    SIMULATION = False
    print "Simulation:\t\t\tFalse"

# Search Depth (Start, Factor, Add, Max)
if "--sdstart" in optdict:
    try:
        SEARCHDEPTH_START = int(optdict["--sdstart"])
    except BaseException, e:
        printErrorMessageAndStop("Value \"" + optdict["--sdstart"] + "\" is invalid for --sdstart.", 1)
else:
    SEARCHDEPTH_START = 1000

if "--sdfactor" in optdict:
    try:
        SEARCHDEPTH_FACTOR = int(optdict["--sdfactor"])
    except BaseException, e:
        printErrorMessageAndStop("Value \"" + optdict["--sdfactor"] + "\" is invalid for --sdfactor.", 1)
else:
    SEARCHDEPTH_FACTOR = 10

if "--sdadd" in optdict:
    try:
        SEARCHDEPTH_ADD = int(optdict["--sdadd"])
    except BaseException, e:
        printErrorMessageAndStop("Value \"" + optdict["--sdadd"] + "\" is invalid for --sdadd.", 1)
else:
    SEARCHDEPTH_ADD = 0

if "--sdmax" in optdict:
    try:
        SEARCHDEPTH_MAX = int(optdict["--sdmax"])
    except BaseException, e:
        printErrorMessageAndStop("Value \"" + optdict["--sdmax"] + "\" is invalid for --sdmax.", 1)
else:
    SEARCHDEPTH_MAX = 1000000000

print "Search Depth:\t\t\tfrom %d to %d, increasing by *%d +%d" % (SEARCHDEPTH_START, SEARCHDEPTH_MAX, SEARCHDEPTH_FACTOR, SEARCHDEPTH_ADD)

# Memory Management
COLLAPSESTATEVECTOR = "--collapse" in optdict
if "--ma" in optdict:
    try:
        STORESTATESINMA = int(optdict["--ma"])
    except BaseException, e:
        printErrorMessageAndStop("Value \"" + optdict["--ma"] + "\" is invalid for --ma.", 1)
else:
    STORESTATESINMA = False
if "--memlim" in optdict:
    try:
        MEMORYLIMIT = int(optdict["--memlim"])
    except BaseException, e:
        printErrorMessageAndStop("Value \"" + optdict["--memlim"] + "\" is invalid for --memlim.", 1)
else:
    MEMORYLIMIT = False
if "--vectorsz" in optdict:
    try:
        VECTORSIZE = int(optdict["--vectorsz"])
    except BaseException, e:
        printErrorMessageAndStop("Value \"" + optdict["--vectorsz"] + "\" is invalid for --vectorsz.", 1)
else:
    VECTORSIZE = False
USEBFSFORSAFETY = "--usebfsforsafety" in optdict

print "Memory Management:\t\tCollapse: " + str(COLLAPSESTATEVECTOR) + ", MinAutomaton: " + str(STORESTATESINMA) + ", Memory Limit: " + str(MEMORYLIMIT) + ", VectorSize: " + str(VECTORSIZE) + ", BFS (Safety only): " + str(USEBFSFORSAFETY)

# Control handling of cycles in Slicing
CYCLESLICING = CYCLESLICING_AUTO
if "--cycleslicing" in optdict:
    if optdict["--cycleslicing"] == "on":
        CYCLESLICING = CYCLESLICING_ON
    elif optdict["--cycleslicing"] == "off":
        CYCLESLICING = CYCLESLICING_OFF
print "CycleSlicing:\t\t\t" + str(CYCLESLICING)

# On/Off-Switches
VERBOSE = "-v" in optdict
print "Verbose:\t\t\t" + str(VERBOSE)
STOREFAULTINJECTIONRESULT = "--storefaultinjectionresult" in optdict
print "StoreFaultInjectionResult:\t" + str(STOREFAULTINJECTIONRESULT)
NOSLICING = "--noslicing" in optdict
print "NoSlicing:\t\t\t" + str(NOSLICING)
STORESLICINGRESULT = "--storeslicingresult" in optdict
print "StoreSlicingResult:\t\t" + str(STORESLICINGRESULT)
VERBOSESLICING = "--verboseslicing" in optdict
print "VerboseSlicing:\t\t\t" + str(VERBOSESLICING)
GENERATESMV = "--generatesmv" in optdict
print "GenerateSMV:\t\t\t" + str(GENERATESMV)
NOSPIN = "--nospin" in optdict
print "NoSPIN:\t\t\t\t" + str(NOSPIN)

# ... Output
printOutputBlock_Footer()

#printErrorMessageAndStop("STOP AFTER READING INPUT PARAMETERS.", 2)





# *** Parse SLIM Files ***
sys.stdout.write("Parse SLIM Files ... ")
sys.stdout.flush()

for fileName in SLIMFILENAMES:
    if not os.path.isfile(fileName):
        sys.stdout.write("ERROR\n")
        printErrorMessageAndStop("SLIM input file '%s' does not exist - STOP!" % fileName, 1)

originalCompileInfo = slim_compiler.src.dataStructure.compile.CompileInformation()
parsingResult = invokers.slim.parse(originalCompileInfo, SLIMFILENAMES)
if not parsingResult:
    sys.stdout.write("ERROR\n")
    printErrorCollector(originalCompileInfo.error_collector)
    printErrorMessageAndStop("SLIM input files could not be parsed correctly - STOP!", 1)
else:
    sys.stdout.write("OK\n")

if VERBOSE:
    printOutputBlock("SLIM Parsing Result:   the original Compile Info", toSLIMVisitor().visit(originalCompileInfo))
    #originalRootComponent = invokers.slim.get_root_impl(originalCompileInfo)
    #printOutputBlock("SLIM Parsing Result:   the original expressions", printAllExpressions(originalRootComponent))

#printErrorMessageAndStop("STOP AFTER PARSING SLIM FILES.", 2)





# *** Check parsed SLIM files ***
sys.stdout.write("Check parsed SLIM Files ... ")
sys.stdout.flush()

checkingResult = invokers.slim.check(originalCompileInfo)
if not checkingResult:
    sys.stdout.write("ERROR\n")
    printErrorCollector(originalCompileInfo.error_collector)
    printErrorMessageAndStop("Parsed SLIM input files did not meet all syntactical restrictions - STOP!", 1)
else:
    sys.stdout.write("OK\n")

if VERBOSE:
    printOutputBlock("SLIM Checking Result:   the checked (& annotated) Compile Info", toSLIMVisitor().visit(originalCompileInfo))
    #originalRootComponent = invokers.slim.get_root_impl(originalCompileInfo)
    #printOutputBlock("SLIM Checking Result:   the checked (& annotated) expressions", printAllExpressions(originalRootComponent))

#printErrorMessageAndStop("STOP AFTER CHECKING PARSED SLIM FILES.", 2)





# *** Fault Injection ***
extendedCompileInfo = None


if FAULTINJECTIONFILENAME != "":
    sys.stdout.write("Do Fault Injection ... ")
    sys.stdout.flush()
    
    if not os.path.isfile(FAULTINJECTIONFILENAME):
        sys.stdout.write("ERROR\n")
        printErrorMessageAndStop("Fault Injection file '%s' does not exist - STOP!" % FAULTINJECTIONFILENAME, 1)
        
    options.FAULT_INJECTION_FILE_NAME = FAULTINJECTIONFILENAME
    try:
        extendedCompileInfo = invokers.slim.do_fault_injection(originalCompileInfo)
    except BaseException, e:
        sys.stdout.write("ERROR\n")
        printErrorMessageAndStop("Fault Injection could not be done correctly ('%s') - STOP!" % str(e), 1)

    options.ROOT_NAME = "" # HINT: After fault injection, the root is unique but its name has changed => simply reset ROOT_NAME

    sys.stdout.write("OK\n")

    if STOREFAULTINJECTIONRESULT:
        storeInFile(toSLIMVisitor().visit(extendedCompileInfo), OUTPUTBASEFILENAME + ".slim---FaultInjectionResult.txt")

    if VERBOSE:
        printOutputBlock("Fault Injection / Model Extension Result:   the extended Compile Info", toSLIMVisitor().visit(extendedCompileInfo))


if extendedCompileInfo != None:
    currentCompileInfo = extendedCompileInfo
else:
    currentCompileInfo = originalCompileInfo
currentRootComponent = invokers.slim.get_root_impl(currentCompileInfo)

#printErrorMessageAndStop("STOP AFTER FAULT INJECTION.", 2)





# *** Parse Properties Document ***
propertiesDocument = None
propertiesDict = {}

if PROPERTIESDOCUMENTFILENAME != "":
    sys.stdout.write("Parse Properties Document ... ")
    sys.stdout.flush()

    # Check file existance
    if not os.path.isfile(PROPERTIESDOCUMENTFILENAME):
        sys.stdout.write("ERROR\n")
        printErrorMessageAndStop("Properties Document file '%s' does not exist - STOP!" % PROPERTIESDOCUMENTFILENAME, 1)

    # Parse
    propertiesDocument = properties.parse_properties_document(PROPERTIESDOCUMENTFILENAME)

    # Check for syntactical validity
    errorList = properties.get_syntax_errors_in_document(propertiesDocument)
    if len(errorList) > 0:
        sys.stdout.write("ERROR\n")
        printErrors(errorList)
        printErrorMessageAndStop("Properties Document is invalid - STOP!", 1)
    
    # Collect all Atomic Propositions for all Properties
    for aProperty in filter(lambda (aProperty): aProperty.get_logic() == constants.Logic.LTL, propertiesDocument):
        atomicPropositions = {}
        for aPlaceholder in aProperty.pattern_instance:
            atomicPropositionAsTree = testexpr.parseExpression(aPlaceholder.content.text)
            try:
                parse_property.check_property_tree(currentCompileInfo, atomicPropositionAsTree, currentRootComponent)
            except BaseException, e:
                printErrorMessageAndStop("Properties Document contains invalid atomic proposition \"" + aPlaceholder.content.text + \
                                         "\" in property \"" + aProperty.name + "\" (" + str(e) + ") - STOP!", 1)
            atomicPropositions[aPlaceholder.placeholder_id] = atomicPropositionAsTree
        propertiesDict[aProperty.name] = (aProperty.name, aProperty.pattern_instance.pattern_id, atomicPropositions)

    sys.stdout.write("OK\n")

    if VERBOSE:
        printOutputBlock_Header("Properties Parsing Result:   the LTL-Properties")
        for (propertyName, (safeName, patternId, atomicPropositions)) in propertiesDict.items():
            print propertyName + " (" + safeName + " : " + str(patternId) + "):"
            for (placeholderName, atomicProposition) in atomicPropositions.items():
                print "- " + placeholderName + ": " + expressionToString(atomicProposition)
        printOutputBlock_Footer()

#printErrorMessageAndStop("STOP AFTER PARSE PROPERTIES DOCUMENT.", 2)





# *** Preprocess compileInfo & properties ***
sys.stdout.write("Preprocess specification and properties ... ")
sys.stdout.flush()


# 0.) Deepcopy compileInfo
preprocessedCompileInfo = copy.deepcopy(currentCompileInfo)
preprocessedRootComponent = invokers.slim.get_root_impl(preprocessedCompileInfo)

# 1.) Eliminate underscores on access paths
recursivelyEliminateUnderscoresInComponentsDataElementsAndSubcomponents(preprocessedRootComponent)
iterativelyPerformUnderscoreEliminationInProperties(propertiesDict, preprocessedRootComponent)

# 2.) Fill up empty in modes lists
recursivelyFillUpEmptyInModesLists(preprocessedRootComponent)

# 3.) Give properties a safe name
setSafeNamesForProperties(propertiesDict)


sys.stdout.write("OK\n")

if VERBOSE:
    printOutputBlock("Preprocessing Result:   the Compile Info w/o underscores on access paths and non-empty in modes lists", toSLIMVisitor().visit(preprocessedCompileInfo))
    printOutputBlock_Header("Preprocessing Result:   the Properties w/o underscores and with safe names")
    for (propertyName, (safeName, patternId, atomicPropositions)) in propertiesDict.items():
        print propertyName + " (" + safeName + " : " + str(patternId) + "):"
        for (placeholderName, atomicProposition) in atomicPropositions.items():
             print "- " + placeholderName + ": " + expressionToString(atomicProposition)
    printOutputBlock_Footer()

currentCompileInfo = preprocessedCompileInfo
currentRootComponent = invokers.slim.get_root_impl(currentCompileInfo)

#printErrorMessageAndStop("STOP AFTER PREPROCESSING SPECIFICATION AND PROPERTIES.", 2)





# *** For each property:   Slice,  Translate,  Invoke SPIN   ***
if len(propertiesDict) > 0:
    summaryModelCheckingResults = {}
    
    for (propertyName, (safeName, patternId, atomicPropositions)) in propertiesDict.items():

        # Mark Beginning of Section
        print "\n\n\n\n\n================================================================================"
        print "CHECK PROPERTY: " + propertyName + " (" + safeName + ") : " + str(patternId)
        print "================================================================================\n"



        # Add property as unchecked to Model Checking Results
        if propertyName in summaryModelCheckingResults:
            printErrorMessageAndStop("Property name '" + propertyName + "' is not unique - STOP!", 1)
        summaryModelCheckingResults[propertyName] = PROPERTYSTATUS_UNCHECKED



        # Slicing
        sys.stdout.write("Do Slicing ... ")
        sys.stdout.flush()
        
        if not NOSLICING:

            if CYCLESLICING == CYCLESLICING_ON:
                enforceKeepingCycles = True
            elif CYCLESLICING == CYCLESLICING_OFF:
                enforceKeepingCycles = False
            else:
                enforceKeepingCycles = (patternId in ["existenceGlobal", "responseGlobal"])
            sys.stdout.write("(enforceKeepingCycles=" + str(enforceKeepingCycles) + ") ...")

            if VERBOSESLICING: sys.stdout.write("\n")
            
            (slicedCompileInfo, slicedRootComponent) = doSlicing(currentRootComponent,
                                                                 atomicPropositions.values(),
                                                                 enforceKeepingCycles,
                                                                 verbose = VERBOSESLICING)   # Slicing itself is verbose only if additional parameter --verboseslicing is set.

            translationRootComponent = slicedRootComponent

            sys.stdout.write("OK\n")

            if STORESLICINGRESULT:
                storeInFile(toSLIMVisitor().visit(slicedCompileInfo), OUTPUTBASEFILENAME + "-" + safeName + ".slim---SlicingResult.txt")

            if VERBOSE and not VERBOSESLICING:
                printOutputBlock("Slicing Result:   the sliced Compile Info", toSLIMVisitor().visit(slicedCompileInfo))

        else:
            translationRootComponent = currentRootComponent

            sys.stdout.write("DISABLED\n")



        # Generate SMV
        if GENERATESMV:
            sys.stdout.write("Generate SMV ... ")
            sys.stdout.flush()
            _generateSMV(translationRootComponent,
                         [atomicPropositions],
                         OUTPUTBASEFILENAME + "-" + safeName)
            sys.stdout.write("OK\n")
            


        # Translate to Promela
        sys.stdout.write("Translate to Promela ... ")
        sys.stdout.flush()

        (translationResult, errors) = doTranslation(translationRootComponent, atomicPropositions, OUTPUTBASEFILENAME + "-" + safeName + ".pml", verbose = VERBOSE, addPrintfs = SIMULATION)

        if len(errors.errors) > 0:
            sys.stdout.write("ERRORS\n")
            printErrorCollector(errors, "The   T R A N S L A T I O N   E R R O R S   are:")
            if VERBOSE:
                printOutputBlock("Translation Result:   the resulting Promela code", translationResult)
            print "Due to translation errors this property is omitted. Continue with next one."
            print "================================================================================\n"
            continue
        else:
            sys.stdout.write("OK\n")
            if VERBOSE:
                printOutputBlock("Translation Result:   the resulting Promela code", translationResult)



        # Simulate using SPIN
        if SIMULATION and SIMULATIONDEPTH > 0:
            print ""
            simulationResult = runProgram("SPIN-Simulation", ["spin", "-u" + str(SIMULATIONDEPTH), "-B", OUTPUTBASEFILENAME + "-" + safeName + ".pml"])
            if simulationResult[0]:
                print "\nSimulation Errors:"
                for line in simulationResult[3]:
                    sys.stdout.write("\t" + line)
                print "\nSimulation Output:"
                for line in simulationResult[2]:
                    sys.stdout.write("\t" + line)
            else:
                print "SPIN-Simulation could not be run successfully - ignored."
            


        # Invoke SPIN, GCC, PAN
        if not NOSPIN:
            print ""

            # - Retrieve command lines
            (spinCmd, gccCmd, panCmd) = executionCommands(patternId, atomicPropositions, OUTPUTBASEFILENAME + "-" + safeName + ".pml")

            # - SPIN
            if not runProgram("SPIN", spinCmd)[0]:
                summaryModelCheckingResults[propertyName] = PROPERTYSTATUS_ERROR
                print "Since SPIN could not be run successfully this property is omitted. Continue with next one."
                print "================================================================================\n"
                continue

            # - GCC
            if not runProgram("GCC", gccCmd)[0]:
                summaryModelCheckingResults[propertyName] = PROPERTYSTATUS_ERROR
                print "Since GCC could not be run successfully this property is omitted. Continue with next one."
                print "================================================================================\n"
                continue

            # - PAN (incl.: iterate over max search depth)
            currentSearchDepth = SEARCHDEPTH_START
            continueSearchDepthIteration = True
            successfulSearchDepthIteration = False

            def checkPanResultForLineBeginning(panStdout, lineBeginning):
                    for line in panStdout:
                        if line.find(lineBeginning) == 0:
                            return True
                    return False

            while True:

                # Invoke PAN
                panResult = runProgram("PAN", panCmd + ["-m" + str(currentSearchDepth)])

                # On invocation errors: omit
                if not panResult[0]:
                    print "Since PAN could not be run successfully this property is omitted. Continue with next one."
                    break

                # On "out of memory": omit
                if checkPanResultForLineBeginning(panResult[2], "pan: out of memory"):
                    print "Not enough memory to run PAN successfully. Omit this property and continue with next one."
                    break
                    
                # On "max search depth too small": increase, if possible
                if checkPanResultForLineBeginning(panResult[2], "error: max search depth too small"):
                    newSearchDepth = min(SEARCHDEPTH_FACTOR * currentSearchDepth + SEARCHDEPTH_ADD, SEARCHDEPTH_MAX)
                    if newSearchDepth > currentSearchDepth:
                        print "Search depth " + str(currentSearchDepth) + " was too small. Try again with search depth " + str(newSearchDepth) + "."
                        currentSearchDepth = newSearchDepth
                    else:
                        print "Search depth " + str(currentSearchDepth) + " was too small but is already the chosen maximum."
                        print "Since max search depth is too small to run PAN successfully this property is omitted. Continue with next one."
                        break

                # If current search depth sufficed: stop successful iteration
                else:
                    print "Search depth " + str(currentSearchDepth) + " sufficed."
                    successfulSearchDepthIteration = True
                    break
                
            if not successfulSearchDepthIteration:
                summaryModelCheckingResults[propertyName] = PROPERTYSTATUS_ERROR
                print "================================================================================\n"
                continue

            # - Analyse PAN's output
            print "\nModel Checking Errors:"
            for line in panResult[3]:
                sys.stdout.write("\t" + line)
            print "\nModel Checking Output:"
            for line in panResult[2]:
                sys.stdout.write("\t" + line)
                
            print "\nModel Checking Result:"
            reportedViolation, reportedAcceptCycle, reportedTrail, reportedErrors, reportedPanErrors = "", "", "", None, None
            for line in panResult[2]:
                if line.find("pan: claim violated! (at depth ") == 0:
                    sys.stdout.write("\t- Violation:\t" + line)
                    reportedViolation = line
                if line.find("pan: acceptance cycle (at depth ") == 0:
                    sys.stdout.write("\t- Accept Cycle:\t" + line)
                    reportedAcceptCycle = line
                if line.find("pan: wrote ") == 0:
                    sys.stdout.write("\t- Error Trail:\t" + line)
                    reportedTrail = line
                if line.find("pan: error") == 0:
                    sys.stdout.write("\t- Pan Error:\t" + line)
                    reportedPanErrors = True
                reSearchResult = re.search("^State-vector (\d+) byte, depth reached (\d+), errors: (\d+)\n$", line)
                if reSearchResult:
                    reportedErrors = reSearchResult.groups()[2]
                    try:
                        reportedErrors = int(reportedErrors)
                        sys.stdout.write("\t- Errors:\t" + str(reportedErrors) + "\n")
                    except BaseException, e:
                        reportedErrors = None

            if reportedErrors > 0 and not reportedPanErrors and reportedTrail and (reportedViolation or reportedAcceptCycle):
                summaryModelCheckingResults[propertyName] = PROPERTYSTATUS_FALSE
                print "==> PROPERTY DOES NOT HOLD <=="
            elif reportedErrors == 0 and not reportedPanErrors and not reportedTrail and not reportedViolation and not reportedAcceptCycle:
                summaryModelCheckingResults[propertyName] = PROPERTYSTATUS_TRUE
                print "==> PROPERTY HOLDS <=="
            elif reportedErrors > 0 and reportedPanErrors:
                summaryModelCheckingResults[propertyName] = PROPERTYSTATUS_ERROR
                print "--> PAN reported errors <--"
            else:
                summaryModelCheckingResults[propertyName] = PROPERTYSTATUS_UNCLEAR
                print "--> PAN's output is confusing <--"

            # - CLEANUP
            print "\nHINT: Now you can delete the files \"pan*\", \"" + OUTPUTBASEFILENAME + "-" + safeName + ".pml" + "\", \"" + OUTPUTBASEFILENAME + "-" + safeName + ".pml" + ".trail\"."



        # Mark End of Section
        print "\n================================================================================"
        print "/ CHECK PROPERTY: " + propertyName
        print "================================================================================\n"



    # Show Summary of Model Checking Results
    print "\n\n\n\n\n================================================================================"
    print "SUMMARY:   MODEL CHECKING RESULTS"
    print "================================================================================\n"

    for (propertyName, checkResult) in summaryModelCheckingResults.items():
        if checkResult == PROPERTYSTATUS_UNCHECKED:
            txt = "- unchecked -"
        elif checkResult == PROPERTYSTATUS_TRUE:
            txt = "TRUE"
        elif checkResult == PROPERTYSTATUS_FALSE:
            txt = "FALSE"
        elif checkResult == PROPERTYSTATUS_UNCLEAR:
            txt = "- unclear -"
        else:
            txt = "- error -"
        print propertyName + ":\t" + txt

    print "\n================================================================================"
    print "/ SUMMARY:   MODEL CHECKING RESULTS"
    print "================================================================================\n"

        
     
else:

    # Print hint that just translated w/o properties
    print "\nHINT: No properties given, so just do a simple translation."
    
    # Generate SMV
    if GENERATESMV:
        sys.stdout.write("Generate SMV ... ")
        sys.stdout.flush()
        _generateSMV(currentRootComponent,
                     [],
                     OUTPUTBASEFILENAME)
        sys.stdout.write("OK\n")

            
    # Translate to Promela
    sys.stdout.write("Translate to Promela ... ")
    sys.stdout.flush()

    (translationResult, errors) = doTranslation(currentRootComponent, {}, OUTPUTBASEFILENAME + ".pml", verbose = VERBOSE, addPrintfs = SIMULATION)

    if len(errors.errors) > 0:
        sys.stdout.write("ERRORS\n")
        printErrorCollector(errors, "The   T R A N S L A T I O N   E R R O R S   are:")
        if VERBOSE:
            printOutputBlock("Translation Result:   the resulting Promela code", translationResult)
    else:
        sys.stdout.write("OK\n")
        if VERBOSE:
            printOutputBlock("Translation Result:   the resulting Promela code", translationResult)


    # Simulate using SPIN
    if SIMULATION and SIMULATIONDEPTH > 0 and len(errors.errors) == 0:
        print ""
        simulationResult = runProgram("SPIN-Simulation", ["spin", "-u" + str(SIMULATIONDEPTH), "-B", OUTPUTBASEFILENAME + ".pml"])
        if simulationResult[0]:
            print "\nSimulation Errors:"
            for line in simulationResult[3]:
                sys.stdout.write("\t" + line)
            print "\nSimulation Output:"
            for line in simulationResult[2]:
                sys.stdout.write("\t" + line)
        else:
            print "SPIN-Simulation could not be run successfully - ignored."





# *** Print End-of-Program Message ***
print "\n\n\nEND OF <promela\script.py>"

# ==================================================================================================
# /MAIN SCRIPT
# ==================================================================================================
