# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================


# Usage:
#     from shared import *
# => Same names useable as with own imports, no shared.xxx necessary!





# ==================================================================================================
# Imports
# ==================================================================================================

import sys
import os

if os.name == "nt":
    sys.path.append(os.path.abspath(".\\..\\..\\..")) # 3 levels up to tools directory (WIN)
    sys.path.append('.\\..\\..\\..\\..\\resources/antlr/antlr-python-runtime-3.1.1/') # add path to antlr runtime
else: 
    sys.path.append(os.path.abspath("./../../../"))   # 3 levels up to tools directory (UNIX)
    sys.path.append('./../../../../resources/antlr/antlr-python-runtime-3.1.1/') # add path to antlr runtime

# --------------------------------------------------------------------------------------------------

import options
import invokers.slim
import invokers.dump
import slim_compiler.src.dataStructure.compile
from slim_compiler.src.dataStructure.nominalbehaviour import *
from slim_compiler.src.dataStructure.expressions import *
from slim_compiler.src.errorManagement.errors import *

import slim_compiler.src.testexpr as testexpr
import slim_compiler.src.parse_property as parse_property
import property_management.src.properties as properties
import property_management.src.constants as constants
import property_management.src.api as api

# --------------------------------------------------------------------------------------------------

import string
import re
import copy

# --------------------------------------------------------------------------------------------------

#Importing already imported is somehow harmful. An if-test with globals() does not help.
#from renaming import *    
#from translation import *
#from slim_compiler.src.slicing.slicing import *

# ==================================================================================================
# /Imports
# ==================================================================================================





# ==================================================================================================
# Global Constants
# ==================================================================================================

# Used in _accessPathToPromelaVariable
UNKNOWN = 0
DATASUBCOMPONENT = 1
INDATAPORT = 2
OUTDATAPORT = 3
NOMINALMODE = 4
ERRORSTATE = 5

# Used in slim2promela_renamingMapping
DATAELEMENT = 0
NONDATASUBCOMPONENT = 1

# Used in script to store status of properties
PROPERTYSTATUS_UNCHECKED = 0
PROPERTYSTATUS_TRUE = 1
PROPERTYSTATUS_FALSE = 2
PROPERTYSTATUS_UNCLEAR = 3
PROPERTYSTATUS_ERROR = 4

# Used in script to control enforceKeepingCycles for Slicing
CYCLESLICING_AUTO = 0
CYCLESLICING_ON = 1
CYCLESLICING_OFF = 2

# ==================================================================================================
# /Global Constants
# ==================================================================================================





# ==================================================================================================
# Convenience Functions
# ==================================================================================================

def printErrorMessageAndStop(errorMessage, exitCode):
    sys.stderr.write(errorMessage + "\n")
    sys.exit(exitCode)



def printOutputBlock_Header(title, outputPipe = sys.stdout):
    outputPipe.write("--------------------------------------------------------------------------------\n")
    outputPipe.write(title + "\n")
    outputPipe.write("--------------------------------------------------------------------------------\n\n")

def printOutputBlock_Content(text, outputPipe = sys.stdout):
    outputPipe.write(text + "\n")

def printOutputBlock_Footer(outputPipe = sys.stdout):
    outputPipe.write("\n--------------------------------------------------------------------------------\n\n\n\n")
    
def printOutputBlock(title, text, outputPipe = sys.stdout):
    printOutputBlock_Header(title, outputPipe)
    printOutputBlock_Content(text, outputPipe)
    printOutputBlock_Footer(outputPipe)



def printErrorCollector(errorCollector, title = "", outputPipe = sys.stderr):
    """Prints all parsing errors or "No parsing errors." to STDERR.
    """

    if title == "":
        title = "The   E R R O R S   are:"

    if len(errorCollector.errors) > 0:
        printOutputBlock(title, str(errorCollector), outputPipe)
    else:
        printOutputBlock(title, "No errors.", outputPipe)

def printErrors(errors, title = "", outputPipe = sys.stderr):
    if title == "":
        title = "The   E R R O R S   are:"

    if len(errors) > 0:
        printOutputBlock(title, "\n".join(map(str, errors)), outputPipe)
    else:
        printOutputBlock(title, "No errors.", outputPipe)
    


def storeInFile(content, fileName):
    if fileName:
        outputFileObject = open(fileName, "w")
        outputFileObject.write(content)
        outputFileObject.close()

# ==================================================================================================
# /Convenience Functions
# ==================================================================================================





# ==================================================================================================
# Functions for Expression Handling
# ==================================================================================================

def expressionReadsFrom(anExpressionTreeNode):
    """Recursively extracts a list of data element names which are read in the expression.

    @param    anExpressionTreeNode:
                The expression in question.
    @type     anExpressionTreeNode:
                expressions.ExpressionTreeNode
    @return:  List of data element names read in the expression.
    @rtype:   list(string)
    """
    
    # Initialize
    readsFrom = []

    # Distinguish by ExpressionTreeNode types
    if anExpressionTreeNode == None:
        pass

    elif isinstance(anExpressionTreeNode, BinaryOperatorTreeNode):
        if anExpressionTreeNode.operator != OperatorType.ASSIGN:
            readsFrom.extend(expressionReadsFrom(anExpressionTreeNode.left))
        readsFrom.extend(expressionReadsFrom(anExpressionTreeNode.right))

    elif isinstance(anExpressionTreeNode, UnaryOperatorTreeNode):
        readsFrom.extend(expressionReadsFrom(anExpressionTreeNode.subexpression))

    elif isinstance(anExpressionTreeNode, CaseExpressionTreeNode):
        for (b, e) in anExpressionTreeNode.cases:
            readsFrom.extend(expressionReadsFrom(b))
            readsFrom.extend(expressionReadsFrom(e))
        readsFrom.extend(expressionReadsFrom(anExpressionTreeNode.default_expression))

    elif isinstance(anExpressionTreeNode, IdentifierTreeNode):
        # We only consider data elements (no ENUM constants, no mode/error information)
        if anExpressionTreeNode.node_type not in [None, NodeType.ENUM_VALUE, NodeType.MODE_FIELD, NodeType.MODE_VALUE, NodeType.ERROR_FIELD, NodeType.ERROR_VALUE]:
            readsFrom.append(anExpressionTreeNode.name)

    else: # Irrelevant: BooleanTreeNode, NumberTreeNode, TrajectoryEquationTreeNode, TrajectoryExpressionTreeNode
        pass

    # Return result
    return readsFrom


def expressionWritesTo(anExpressionTreeNode):
    """Extracts the name of the data element to which the expression assignes a value.

    @param    anExpressionTreeNode:
                The expression in question.
    @type     anExpressionTreeNode:
                expressions.ExpressionTreeNode
    @return:  Data element name on left hand side of the assignment
              or None if not a valid assignment.
    @rtype:   string or None
    """
    
    if isinstance(anExpressionTreeNode, BinaryOperatorTreeNode):
        if anExpressionTreeNode.operator == OperatorType.ASSIGN:
            if isinstance(anExpressionTreeNode.left, IdentifierTreeNode):
                if anExpressionTreeNode.left.node_type in [NodeType.BOOL, NodeType.INT, NodeType.INT_RANGE, NodeType.REAL, NodeType.ENUM_FIELD, NodeType.CLOCK, NodeType.CONTINUOUS, NodeType.TIME_VALUE]:
                    # Valid assignment => return name on LHS
                    return anExpressionTreeNode.left.name
    return None


def trajectoryEquationsWriteTo(anExpressionTreeNode):
    """Extracts the name of all data elements for which a trajectory equation exists in the expression.

    @param    anExpressionTreeNode:
                The expression in question.
    @type     anExpressionTreeNode:
                expressions.ExpressionTreeNode
    @return:  List of all names of data elements for which a trajectory
              equation exists in the expression.
    @rtype:   list(string)
    """
    
    # Initialize
    writesTo = []

    # Distinguish by ExpressionTreeNode types
    if anExpressionTreeNode == None:
        pass

    elif isinstance(anExpressionTreeNode, BinaryOperatorTreeNode):
        if anExpressionTreeNode.operator == OperatorType.AND:
            writesTo.extend(trajectoryEquationsWriteTo(anExpressionTreeNode.left))
            writesTo.extend(trajectoryEquationsWriteTo(anExpressionTreeNode.right))

    elif isinstance(anExpressionTreeNode, TrajectoryEquationTreeNode):
        if isinstance(anExpressionTreeNode.variable, IdentifierTreeNode):
            writesTo.append(anExpressionTreeNode.variable.name)

    else: # Irrelevant: UnaryOperatorTreeNode, IdentifierTreeNode, BooleanTreeNode, NumberTreeNode, TrajectoryExpressionTreeNode
        pass

    # Return result
    return writesTo



def checkSyntacticalIdentityOfExpressions(anExpressionTreeNodeA, anExpressionTreeNodeB):
    """Checks whether two expressions are syntactically identical.

    @param    anExpressionTreeNodeA:
                Root of first (A) expression.
    @type     anExpressionTreeNodeA:
                expression.ExpressionTreeNode
    @param    anExpressionTreeNodeB:
                Root of second (B) expression.
    @type     anExpressionTreeNodeB:
                expression.ExpressionTreeNode
    @return:  True, iff both expressions are syntactically identical.
              Otherwise, false.
    @rtype:   boolean
    """

    # *** Obvious cases ***
    if anExpressionTreeNodeA == anExpressionTreeNodeB:
        return True
    elif anExpressionTreeNodeA == None or anExpressionTreeNodeB == None:
        return False   # both None => already A = B

    # *** Distinguish by node type ***
    elif isinstance(anExpressionTreeNodeA, BinaryOperatorTreeNode) and \
         isinstance(anExpressionTreeNodeB, BinaryOperatorTreeNode):
        return anExpressionTreeNodeA.operator == anExpressionTreeNodeB.operator and \
               checkSyntacticalIdentityOfExpressions(anExpressionTreeNodeA.left, anExpressionTreeNodeB.left) and \
               checkSyntacticalIdentityOfExpressions(anExpressionTreeNodeA.right, anExpressionTreeNodeB.right)
        
    elif isinstance(anExpressionTreeNodeA, UnaryOperatorTreeNode) and \
         isinstance(anExpressionTreeNodeB, UnaryOperatorTreeNode):
        return anExpressionTreeNodeA.operator == anExpressionTreeNodeB.operator and \
               checkSyntacticalIdentityOfExpressions(anExpressionTreeNodeA.subexpression, anExpressionTreeNodeB.subexpression)
        
    elif isinstance(anExpressionTreeNodeA, CaseExpressionTreeNode) and \
         isinstance(anExpressionTreeNodeB, CaseExpressionTreeNode):
        return identicalLists(anExpressionTreeNodeA.cases, anExpressionTreeNodeB.cases,
                              lambda (bA, eA), (bB, eB): checkSyntacticalIdentityOfExpressions(bA, bB) and
                                                         checkSyntacticalIdentityOfExpressions(eA, eB)       ) \
               and \
               checkSyntacticalIdentityOfExpressions(anExpressionTreeNodeA.default_expression, anExpressionTreeNodeB.default_expression)
        
    elif isinstance(anExpressionTreeNodeA, IdentifierTreeNode) and \
         isinstance(anExpressionTreeNodeB, IdentifierTreeNode):
        return anExpressionTreeNodeA.name == anExpressionTreeNodeB.name
        
    elif isinstance(anExpressionTreeNodeA, BooleanTreeNode) and \
         isinstance(anExpressionTreeNodeB, BooleanTreeNode):
        return anExpressionTreeNodeA.value == anExpressionTreeNodeB.value
        
    elif isinstance(anExpressionTreeNodeA, NumberTreeNode) and \
         isinstance(anExpressionTreeNodeB, NumberTreeNode):
        return anExpressionTreeNodeA.number == anExpressionTreeNodeB.number and \
               anExpressionTreeNodeA.time_unit == anExpressionTreeNodeB.time_unit
        
    elif isinstance(anExpressionTreeNodeA, TrajectoryEquationTreeNode) and \
         isinstance(anExpressionTreeNodeB, TrajectoryEquationTreeNode):
        return anExpressionTreeNodeA.variable == anExpressionTreeNodeB.variable and \
               checkSyntacticalIdentityOfExpressions(anExpressionTreeNodeA.trajectory_expression, anExpressionTreeNodeB.trajectory_expression)
        
    elif isinstance(anExpressionTreeNodeA, TrajectoryExpressionTreeNode) and \
         isinstance(anExpressionTreeNodeB, TrajectoryExpressionTreeNode):
        return anExpressionTreeNodeA.times == anExpressionTreeNodeB.times and \
               anExpressionTreeNodeA.variable == anExpressionTreeNodeB.variable and \
               anExpressionTreeNodeA.offset == anExpressionTreeNodeB.offset

    # *** Different node types => not identical ***
    else:
        return False
        
    

def expressionsToString(anExpressionList):
    if len(anExpressionList) > 0:
        txt = ""
        for expression in anExpressionList:
            txt += expressionToString(expression) + "; "
        return txt
    else:
        return "EmptyExpressionList"



def expressionToString(anExpressionTreeNode):
    # Alternatively, use: "else: txt = toSLIMVisitor().visit(anExpressionTreeNode)", but that sometimes misses s.th.
    if anExpressionTreeNode == None:
        txt = "EmptyExpression"   
    elif isinstance(anExpressionTreeNode, BinaryOperatorTreeNode):
        txt = "(" + expressionToString(anExpressionTreeNode.left) + ") " + str(OperatorType.operator_to_str(anExpressionTreeNode.operator)) + " (" + expressionToString(anExpressionTreeNode.right) + ")"
    elif isinstance(anExpressionTreeNode, UnaryOperatorTreeNode):
        txt = OperatorType.operator_to_str(anExpressionTreeNode.operator) +  "(" + expressionToString(anExpressionTreeNode.subexpression) + ")"
    elif isinstance(anExpressionTreeNode, CaseExpressionTreeNode):
        txt = "case "
        for (b, e) in anExpressionTreeNode.cases:
            txt += expressionToString(b) + " : " + expressionToString(e) + "; "
        txt += "otherwise " + expressionToString(anExpressionTreeNode.default_expression) + " end"
    elif isinstance(anExpressionTreeNode, IdentifierTreeNode):
        txt = str(anExpressionTreeNode.name)
    elif isinstance(anExpressionTreeNode, BooleanTreeNode):
        txt = str(anExpressionTreeNode.value).lower()
    elif isinstance(anExpressionTreeNode, NumberTreeNode):
        txt = str(anExpressionTreeNode.number)
    elif isinstance(anExpressionTreeNode, FlowEquationTreeNode):
        txt = anExpressionTreeNode.variable + "' = " + expressionToString(anExpressionTreeNode.flow_expression)
    elif isinstance(anExpressionTreeNode, FlowExpressionTreeNode):
        txt = ""
        if anExpressionTreeNode.times != None:
            txt += str(anExpressionTreeNode.times) + " * "
        txt += expressionToString(anExpressionTreeNode.variable)
        if anExpressionTreeNode.offset != None:
            txt += " + " + str(anExpressionTreeNode.offset)
    else:
        txt = "ERR: Unknown Expression Construct"
    return txt



def expressionToDebugString(anExpressionTreeNode, indent=0):
    if anExpressionTreeNode == None:
        txt = "<EmptyExpression>"
        
    elif isinstance(anExpressionTreeNode, BinaryOperatorTreeNode):
        txt = indent * "\t" + "BinaryOperatorTreeNode:" + \
              " operator=" + str(OperatorType.operator_to_str(anExpressionTreeNode.operator)) + \
              " node_type=\"" + str(NodeType.type_to_str(anExpressionTreeNode.node_type)) + "\"" + \
              " bounds=" + str(anExpressionTreeNode.bounds) + \
              " number=" + str(anExpressionTreeNode.number) + "\n"
        txt += indent * "\t" +  "-LEFT:\n" + expressionToDebugString(anExpressionTreeNode.left, indent+1)
        txt += indent * "\t" +  "-RIGHT:\n" + expressionToDebugString(anExpressionTreeNode.right, indent+1)
        
    elif isinstance(anExpressionTreeNode, UnaryOperatorTreeNode):
        txt = indent * "\t" + "UnaryOperatorTreeNode:" + \
              " operator=" + str(OperatorType.operator_to_str(anExpressionTreeNode.operator)) + \
              " node_type=\"" + str(NodeType.type_to_str(anExpressionTreeNode.node_type)) + "\"" + \
              " bounds=" + str(anExpressionTreeNode.bounds) + \
              " number=" + str(anExpressionTreeNode.number) + "\n"
        txt += indent * "\t" +  "-SUBEXPRESSION:\n" + expressionToDebugString(anExpressionTreeNode.subexpression, indent+1)
        
    elif isinstance(anExpressionTreeNode, CaseExpressionTreeNode):
        txt = indent * "\t" + "CaseExpressionTreeNode:" + \
              " operator=" + str(OperatorType.operator_to_str(anExpressionTreeNode.operator)) + \
              " node_type=\"" + str(NodeType.type_to_str(anExpressionTreeNode.node_type)) + "\"" + \
              " number=" + str(anExpressionTreeNode.number) + \
              " name=" + str(anExpressionTreeNode.name) + \
              " valid_enums=" + str(anExpressionTreeNode.valid_enums) + \
              " enum_values=" + str(anExpressionTreeNode.enum_values) + "\n"
        for (b, e) in anExpressionTreeNode.cases:
            txt += indent * "\t" +  "-CASE.CONDITION:\n" + expressionToDebugString(b, indent+1)
            txt += indent * "\t" +  "-CASE.EXPRESSION:\n" + expressionToDebugString(e, indent+1)
        txt += indent * "\t" +  "-DEFAULTEXPRESSION:\n" + expressionToDebugString(anExpressionTreeNode.default_expression, indent+1)
        
    elif isinstance(anExpressionTreeNode, IdentifierTreeNode):
        txt = indent * "\t" + "IdentifierTreeNode:" + \
              " name=" + str(anExpressionTreeNode.name) + \
              " node_type=\"" + str(NodeType.type_to_str(anExpressionTreeNode.node_type)) + "\"" + \
              " bounds=" + str(anExpressionTreeNode.bounds) + \
              " number=" + str(anExpressionTreeNode.number) + \
              " valid_enums=" + str(anExpressionTreeNode.valid_enums) + \
              " enum_values=" + str(anExpressionTreeNode.enum_values) + "\n"
        
    elif isinstance(anExpressionTreeNode, BooleanTreeNode):
        txt = indent * "\t" + "BooleanTreeNode:" + \
              " value=" + str(anExpressionTreeNode.value).lower() + \
              " node_type=\"" + str(NodeType.type_to_str(anExpressionTreeNode.node_type)) + "\"\n"
        
    elif isinstance(anExpressionTreeNode, NumberTreeNode):
        txt = indent * "\t" + "NumberTreeNode:" + \
              " number=" + str(anExpressionTreeNode.number) + \
              " node_type=\"" + str(NodeType.type_to_str(anExpressionTreeNode.node_type)) + "\"" + \
              " time_unit=" + str(anExpressionTreeNode.time_unit) + "\n"
        
##    elif isinstance(anExpressionTreeNode, FlowEquationTreeNode):
##        txt = anExpressionTreeNode.variable + "' = " + expressionToString(anExpressionTreeNode.flow_expression)
##        
##    elif isinstance(anExpressionTreeNode, FlowExpressionTreeNode):
##        txt = ""
##        if anExpressionTreeNode.times != None:
##            txt += str(anExpressionTreeNode.times) + " * "
##        txt += expressionToString(anExpressionTreeNode.variable)
##        if anExpressionTreeNode.offset != None:
##            txt += " + " + str(anExpressionTreeNode.offset)
            
    else:
        txt = "<UnknownExpressionConstruct>"
        
    return txt


def expressionsToDebugString(anExpressionList):
    if len(anExpressionList) > 0:
        txt = ""
        for expression in anExpressionList:
            txt += expressionToDebugString(expression)
        return txt
    else:
        return "<EmptyExpressionList>"


def printAllExpressions(componentImplementation):
    print "-------------------------------------------"
    print "COMPONENT: " + componentImplementation.name

    componentType = componentImplementation.get_referenced_component_type()

    for scObject in componentImplementation.data_subcomponents.values():
        print "* DATA SUBCOMPONENT: " + scObject.name + "\n" + expressionToDebugString(scObject.default_value)

    for portObject in componentType.in_data_ports.values():
        print "* IN DATA PORT: " + portObject.name + "\n" + expressionToDebugString(portObject.default_value)
        
    for portObject in componentType.out_data_ports.values():
        print "* OUT DATA PORT: " + portObject.name + "\n" + expressionToDebugString(portObject.default_value)

    for flow in componentImplementation.flows:
        print "* FLOW:\n" + expressionToDebugString(flow.expression)

    for transition in componentImplementation.normal_mode_transitions:
        print "* TRANSITION:\n" + expressionsToDebugString(transition.label.effects)

    #probability_mode_transitions

    for scObject in componentImplementation.non_data_subcomponents.values():
        printAllExpressions(scObject.get_referenced_component_implementation())
    
    print "..........................................."

# ==================================================================================================
# /Functions for Expression Handling
# ==================================================================================================





# ==================================================================================================
# Supporting functions
# ==================================================================================================

def includeInList(existingList, includeList):
    """Adds every element from includeList to existingList that is not already
    contained in existingList and reports whether anythin was added at all.

    @param    existingList:
                The list in which the elements are to be added.
                Note: As lists are like objects, the changes are done directly
                      in the original list.
    @type     existingList:
                list
    @param    includeList:
                The list of elements which are to be added, if not already contained.
    @type     includeList:
                list
    @return:  True, iff some element was added to existing list.
              False, otherwise.
    @rtype:   boolean
    """

    # Initialize changed-flag
    changed = False

    # For every element in includeList ...
    for element in includeList:
        # ... if not already in existingList, add it!
        if element not in existingList:
            existingList.append(element)
            changed = True

    # Return changed-flag
    return changed


def isIntersectionNotEmpty(a, b):
    """Checks whether the two collections have some element(s) in common.

    As multiple occurences of elements as well as their order does not matter
    the collections are transformed to sets before the intersection is taken.

    @param     a:
                 First collection of elements.
    @type      a:
                 iterable, e.g. list
    @param     b:
                 Second collection of elements.
    @type      b:
                 iterable, e.g. list
    @return:   True, iff a, b have some element(s) in common.
               False, otherwise.
    @rtype:    boolean
    """
    
    return len(set(a) & set(b)) > 0


def identicalLists(a, b, compareFunction):
    """Checks whether the two lists contain the same elements
    in the same order.
    The elements are compared using the function compareFunction.

    Example:
    >>> identicalLists([1, 2, 3], [1, 2, 3], lambda x, y: x == y)
    True
    >>> identicalLists([1, 2], [2, 1], lambda x, y: x == y)
    False
    >>> identicalLists([myCar1, yourCar1], [myCar2, yourCar2], lambda x, y: x.owner == y.owner)
    True

    Note: In the last example, myCar1 and myCar2 do not have to be equal objects,
    it suffices that myCar1.owner == myCar2.owner (same with yourCar[1|2]).

    @param     a:
                 First list.
    @type      a:
                 list
    @param     b:
                 Second list.
    @type      b:
                 list
    @param     compareFunction:
                 Boolean function to compare two elements.
    @type      compareFunction:
                 Function: (aElement, bElement) -> boolean
    @return:   True, iff a, b have the same elements in the same order.
               False, otherwise.
    @rtype:    boolean
    """
    
    if len(a) != len(b):
        return False
    
    for pos in range(len(a)):
        if not compareFunction(a[pos], b[pos]):
            return False

    return True
                                              

def identicalUnorderedContent(a, b, compareFunction):
    """Checks whether the two collections contain the same elements.
    The order of elements and multiple occurences are ignored.
    The elements are compared using the function compareFunction.

    Example:
    >>> identicalUnorderedContent([1, 2, 3], [3, 2, 1, 2], lambda x, y: x == y)
    True
    >>> identicalUnorderedContent([1, 2], [0, 1, 2], lambda x, y: x == y)
    False
    >>> identicalUnorderedContent([myCar1, yourCar1], [myCar2, yourCar2], lambda x, y: x.owner == y.owner)
    True

    Note: In the last example, myCar1 and myCar2 do not have to be equal objects,
    it suffices that myCar1.owner == myCar2.owner (same with yourCar[1|2]).

    @param     a:
                 First collection of elements.
    @type      a:
                 iterable, e.g. list
    @param     b:
                 Second collection of elements.
    @type      b:
                 iterable, e.g. list
    @param     compareFunction:
                 Boolean function to compare two elements.
    @type      compareFunction:
                 Function: (aElement, bElement) -> boolean
    @return:   True, iff a, b have the same elements.
               False, otherwise.
    @rtype:    boolean
    """
    
    return checkUniversality(lambda aElement:   checkExistence(lambda bElement: compareFunction(aElement, bElement), b),   a) and \
           checkUniversality(lambda bElement:   checkExistence(lambda aElement: compareFunction(aElement, bElement), a),   b)


def strictOr(listOfTruthValues, initialValue = False):
    """Takes a list of truth values and calculates the strict or over them.

    This function is useful in situations where several functions are called
    in sequence and in the end we need to know whether at least one call
    returned True. Concatenating the calls by "or" does not help since
    sequential evaluation causes that as soon as one function returned true
    the following functions are not called any more. Example:

    >>> def f():
            print "f"
            return True
    >>> def g():
            print "g"
            return True
    >>> f() or g()
    f
    True

    Instead, use strictOr like this:
    >>> strictOr([f(), g()])
    f
    g
    True


    @param    listOfTruthValues:
                The list of truth values for which the strict or is to be calculated.
    @type     listOfTruthValues:
                list(boolean)
    @param    initialValue:
                Optional initial value. If the list is empty, this value is returned,
                otherwise it is just added to the calculation of the strict or.
    @type     initialValue:
                boolean
    @return:  Strict or over the given values.
    @rtype:   boolean
    """
    
    return reduce(lambda x, y: x or y, listOfTruthValues, initialValue)


def checkExistence(booleanMapping, listOfElements):
    """Checks whether in the given listOfElements some element(s) exist which
    fulfill the property described by booleanMapping.

    @param    booleanMapping:
                Method that checks whether a single element of the list fulfills
                the desired property.
    @type     booleanMapping:
                Function: element -> boolean
    @param    listOfElements:
                The list of elements to be considered.
    @type     listOfElements:
                list
    @return:  True, iff at least one element exists for that booleanMapping
              evaluates to true. Otherwise, false.
    @rtype:   boolean
    """
    
    return strictOr( map(booleanMapping, listOfElements) )


def checkUniversality(booleanMapping, listOfElements):
    """Checks whether in the given listOfElements all elements
    fulfill the property described by booleanMapping.

    @param    booleanMapping:
                Method that checks whether a single element of the list fulfills
                the desired property.
    @type     booleanMapping:
                Function: element -> boolean
    @param    listOfElements:
                The list of elements to be considered.
    @type     listOfElements:
                list
    @return:  True, iff for all elements booleanMapping
              evaluates to true. Otherwise, false.
    @rtype:   boolean
    """
     
    return reduce(lambda x, y: x and y, 
                  map(booleanMapping, listOfElements),
                  True)

# ==================================================================================================
# /Supporting functions
# ==================================================================================================
