# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================


"""This module contains everything for the translation:   SLIM -> Promela

Contents:  - Imports: adapt path, import compiler, import python modules
           - class Slim2PromelaTranslationError
           - class ProctypeInfo
           - class Slim2Promela_Translator
           - Main Program (executes if this module is loaded as __main__)

           To read and understand this source code start with the Slim2Promela_Translator class.
           

Author:   Maximilian R. Odenbrett, maximilian.odenbrett@rwth-aachen.de
Position: Diploma Thesis at MOVES, RWTH Aachen University
Date:     June 2009 - approx. December 2009
"""

from shared import *
from renaming import *

from slim_compiler.src.dataStructure.visitors.printVisitor import toSLIMVisitor



# ==================================================================================================
# class Slim2PromelaTranslationError
# ==================================================================================================

class Slim2PromelaTranslationError(slim_compiler.src.errorManagement.errors.AbstractError):
    """This class represents an error which occured during translation of SLIM to Promela."""
    
    def __init__(self, msg):
        """Initializes the Slim2PromelaTranslationError error.

        Usage in Slim2Promela_Translator:  self.error_collector.add(Slim2PromelaTranslationError("..."))
        Usage in ProctypeInfo:  self.translator.error_collector.add(Slim2PromelaTranslationError("..."))

        Add location infos: self.translator.error_collector.add(Slim2PromelaTranslationError("... '" + ... + "' ... (in " + self.accessPath + " : " + self.packageName + "::" + self.componentTypeName + "." + self.componentImplementationName + ")"))

        @type msg:  string
        @param msg: The message describing the error.
        """
        slim_compiler.src.errorManagement.errors.AbstractError.__init__(self, "", -1, -1)
        self.message = msg
        
    def get_message(self):
        """Returns the message describing the error.

        @rtype:  string
        @return: The message describing the error.
        """
        return self.message

# ==================================================================================================
# </class Slim2PromelaTranslationError: ...>
# ==================================================================================================










# ==================================================================================================
# class ProctypeInfo
# ==================================================================================================

class ProctypeInfo:
    """This class is responsible for translating one single SLIM component instance to a Promela proctype and maintaining all therefore necessary information.

    Usage: 1.) Create a ProctypeInfo instance
               >>> pi = ProctypeInfo(<Slim2Promela_Translator>, <accessPath>)
               
           2.) Extract the required information from the compile information / an component implementation
               >>> pi._collectProctypeInfo(<componentImplementation>)
               
           3.) Perform the translation and retrieve a tuple (global variable declarations, proctype definition)
               >>> (gvds, ptds) = pi.toPromela()

    Errors occuring in ProctypeInfo.toPromela() are written to the resulting Promela code and
    are added as Slim2PromelaTranslationError to <Slim2Promela_Translator>.error_collector.
    """



    # ----------------------------------------------------------------------------------------------



    def __init__(self, translator, componentImplementation, accessPath):
        """Initializes a ProctypeInfo instance.

        @type translator:  instance(Slim2Promela_Translator)
        @param translator: reference to the translator object used, needed since it manages the global list of enum names, mode names and errors
        @type accessPath:  string
        @param accessPath: identifies the component in the component tree and thus becomes part of the proctype's name
                           (Here the Promela-notation with underscore as seperator is to be used!)
        """
        
        self.translator = translator    # reference to Slim2Promela_Translator - instance (for .enumNames, .modeNames, .error_collector)
        self.verbose = translator.verbose        # Copy flags of translator to own flags.
        self.addPrintfs = translator.addPrintfs  # (for easyer access and more independence)
        self.accessPath = accessPath    # a string: <root>_<sc>_..._<thisComponent>
        self._clearProctypeInfo()       # initialize internal representation of component
        self._collectProctypeInfo(componentImplementation)


    def _clearProctypeInfo(self):
        """Initialize/Reset internal representation of component."""
        
        self.packageName = ""                 # a string
        self.componentTypeName = ""           # a string
        self.componentImplementationName = "" # a string
        
        self.in_data_ports = []               # contains tuples:   (Name: string, Type: Bool/Int/Real/..., Default: ***TreeNode)
        self.out_data_ports = []              # contains tuples:   (Name: string, Type: Bool/Int/Real/..., Default: ***TreeNode)
        self.in_event_ports = []              # contains strings:   names of in event ports
        self.out_event_ports = []             # contains strings:   names of out event ports

        self.subcomponents = []               # contains tuples:   (Name: string, InDataPorts: list, OutDataPorts: list, InEventPorts: list, OutEventPorts: list, InModes: list, ModeHistory: boolean)
        self.datasubcomponents = []           # contains tuples:   (Name: string, Type: Bool/Int/Real/..., Default: ***TreeNode, InModes: list)

        self.data_port_connections = []       # contains tuples:   (SourcePort: string, DestinationPort: string, InModes: list)
        self.event_port_connections = []      # contains tuples:   (SourcePort: string, DestinationPort: string, InModes: list)
        self.flows = []                       # contains tuples:   (Expression: BinaryOperatorTreeNode, InModes: list)
        
        self.modes = []                       # contains tuples:   (ModeName: string, ModeType: int (nominalbehaviour.ModeType.INITIAL/ACTIVATION/...))
        self.transitions = []                 # contains tuples:   (StartMode: string, Trigger: string, Guard: ***TreeNode, Effects: list(BinaryOperatorTreeNodes), DestinationMode: string)
        # DEPRECATED: self.probabilistictransitions = []    # contains tuples:   (StartMode: string, Rate: string, DestinationMode: string)

        self.startMode = None                 # Stores the name: string of the start (initial/activation) mode
        self.modeHistory = None               # Stores whether (boolean) this component supports mode history, i.e. has an initial mode.
        self.toBeBuffered = set()             # Stores the names (set(string)) of all data elements (interesting: data subcomponents) which need to be buffered somewhen.


    def _collectProctypeInfo(self, componentImplementation):
        """Collects all the necessary information from the given componentImplementation and stores it in ProctypeInfo's internal data structures (see _clearProctypeInfo).

        @type componentImplementation:  instance(slim_compiler.src.dataStructure.nominalbehaviour.ComponentImplementation)
        @param componentImplementation: compiler's data structure of the component
        """

        # DEBUG: enable/disable
        showDebugInfo = False

        
        # *** Clear all older Info ***
        self._clearProctypeInfo()

        
        # *** Get corresponding Component Type & Names ***
        componentType = componentImplementation.get_referenced_component_type()
        self.packageName = componentImplementation.subpackage.package_name
        self.componentTypeName = componentType.name
        self.componentImplementationName = componentImplementation.name.split(".")[1]
        if showDebugInfo: print "[Collect Proctype Info for:   componentImplementation.name = " + componentImplementation.name + ", componentType.name = " + componentType.name + ", package: " + componentImplementation.subpackage.package_name + "]"


        # *** Collect info about Non-Data Subcomponents ***
        # componentImplementation.non_data_subcomponents = dict(scName: string, scObject: instance(slim_compiler.src.dataStructure.nominalbehaviour.NonDataSubcomponent))

        if showDebugInfo: print "\tNon-Data Subcomponents:" #+ " " + str(componentImplementation.non_data_subcomponents)

        # for each subcomponent: ...
        for (scName, scObject) in componentImplementation.non_data_subcomponents.items():

            # Get Type & Implementation corresponding to that subcomponent
            scType = scObject.get_referenced_component_type()
            scImpl = scObject.get_referenced_component_implementation()

            if showDebugInfo: print "\t\t%s/%s : %s/%s (used in %s in modes %s)" % (scName, scObject.name, scObject.implementation_ref, scType.name, scObject.implementation.name, scObject.in_modes)

            # Subcomponent's In Data Ports
            if showDebugInfo: print "\t\t\t|- In Data Ports:" #+ " " + str(scType.in_data_ports)
            scIDPs = []
            for (portName, portObject) in scType.in_data_ports.items():
                if showDebugInfo: print "\t\t\t\t%s/%s : %s = %s" % (portName, portObject.name, portObject.data_type.get_name(), portObject.default_value)
                scIDPs.append((portName, portObject.data_type, portObject.default_value))
                if isinstance(portObject.data_type, Enum):
                    if showDebugInfo: print "\t\t\t\t\t=> EnumNames: %s" % portObject.data_type.values
                    self.translator._addEnumNames(portObject.data_type.values)

            # Subcomponent's Out Data Ports (including ErrorEventPorts)
            if showDebugInfo: print "\t\t\t|- Out Data Ports:" #+ " " + str(scType.out_data_ports)
            scODPs = []
            for (portName, portObject) in scType.out_data_ports.items():
                if showDebugInfo: print "\t\t\t\t%s/%s : %s = %s" % (portName, portObject.name, portObject.data_type.get_name(), portObject.default_value)
                scODPs.append((portName, portObject.data_type, portObject.default_value))
                if isinstance(portObject.data_type, Enum):
                    if showDebugInfo: print "\t\t\t\t\t=> EnumNames: %s" % portobject.data_type.values
                    self.translator._addEnumNames(portObject.data_type.values)

            # Subcomponent's In Event Ports
            if showDebugInfo: print "\t\t\t|- In Event Ports:" #+ " " + str(scType.in_event_ports)
            scIEPs = []
            for portName in scType.in_event_ports.keys():
                if showDebugInfo: print "\t\t\t\t%s" % (portName)
                scIEPs.append(portName)

            # Subcomponent's Out Event Ports
            if showDebugInfo: print "\t\t\t|- Out Event Ports:" #+ " " + str(scType.out_event_ports)
            scOEPs = []
            for portName in scType.out_event_ports.keys():
                if showDebugInfo: print "\t\t\t\t%s" % (portName)
                scOEPs.append(portName)

            # Subcomponent's Mode History
            scModeHistory = checkExistence(lambda modeObject: modeObject.type == slim_compiler.src.dataStructure.nominalbehaviour.ModeType.INITIAL, scImpl.modes.values())
            if showDebugInfo: print "\t\t\t|- Mode History:" + " " + str(scModeHistory)
            
            # Add Subcomponent's info to list of subcomponents
            self.subcomponents.append((scName, scIDPs, scODPs, scIEPs, scOEPs, scObject.in_modes, scModeHistory))
        # </for each subcomponent: ...>


        # *** Collect info about Data Subcomponents ***
        # componentImplementation.data_subcomponents = dict(scName: string, scObject: instance(slim_compiler.src.dataStructure.nominalbehaviour.DataSubcomponent))
            
        if showDebugInfo: print "\tData Subcomponents:" #+ " " + str(componentImplementation.data_subcomponents)
        for (scName, scObject) in componentImplementation.data_subcomponents.items():
            if showDebugInfo: print "\t\t%s/%s : %s = %s (in modes %s)" % (scName, scObject.name, scObject.data_type.get_name(), scObject.default_value, scObject.in_modes)
            self.datasubcomponents.append((scName, scObject.data_type, scObject.default_value, scObject.in_modes))
            if isinstance(scObject.data_type, Enum):
                if showDebugInfo: print "\t\t\t=> EnumNames: %s" % scObject.data_type.values
                self.translator._addEnumNames(scObject.data_type.values)


        # *** Collect info about Ports ***
        # componentType.*_*_ports = dict(portName: string, portObject: instance(slim_compiler.src.dataStructure.nominalbehaviour.*Port))

        # In Data Ports
        if showDebugInfo: print "\tIn Data Ports:" #+ " " + str(componentType.in_data_ports)
        for (portName, portObject) in componentType.in_data_ports.items():
            if showDebugInfo: print "\t\t%s/%s : %s = %s" % (portName, portObject.name, portObject.data_type.get_name(), portObject.default_value)
            self.in_data_ports.append((portName, portObject.data_type, portObject.default_value))
            if isinstance(portObject.data_type, Enum):
                if showDebugInfo: print "\t\t\t=> EnumNames: %s" % portObject.data_type.values
                self.translator._addEnumNames(portObject.data_type.values)

        # Out Data Ports
        if showDebugInfo: print "\tOut Data Ports:" #+ " " + str(componentType.out_data_ports)
        for (portName, portObject) in componentType.out_data_ports.items():
            if showDebugInfo: print "\t\t%s/%s : %s = %s" % (portName, portObject.name, portObject.data_type.get_name(), portObject.default_value)
            self.out_data_ports.append((portName, portObject.data_type, portObject.default_value))
            if isinstance(portObject.data_type, Enum):
                if showDebugInfo: print "\t\t\t=> EnumNames: %s" % portObject.data_type.values
                self.translator._addEnumNames(portObject.data_type.values)

        # In Event Ports
        if showDebugInfo: print "\tIn Event Ports:" #+ " " + str(componentType.in_event_ports)
        for portName in componentType.in_event_ports.keys():
            if showDebugInfo: print "\t\t%s" % (portName)
            self.in_event_ports.append(portName)

        # Out Event Ports (including ErrorEventPorts)
        if showDebugInfo: print "\tOut Event Ports:" #+ " " + str(componentType.out_event_ports)
        for portName in componentType.out_event_ports.keys():
            if showDebugInfo: print "\t\t%s" % (portName)
            self.out_event_ports.append(portName)


        # *** Collect info about Port Connections & Flows ***
        # componentImplementation.*_port_connections = list(connection: instance(slim_compiler.src.dataStructure.nominalbehaviour.PortConnection))
        # componentImplementation.flows = list(flow: instance(slim_compiler.src.dataStructure.nominalbehaviour.Flow))

        # Data Port Connections
        if showDebugInfo: print "\tData Port Connections:" #+ " " + str(componentImplementation.data_port_connections)
        for connection in componentImplementation.data_port_connections:
            if showDebugInfo: print "\t\t%s -> %s, in modes %s" % (connection.source_port, connection.target_port, connection.in_modes)
            self.data_port_connections.append((connection.source_port, connection.target_port, connection.in_modes))

        # Event Port Connections
        if showDebugInfo: print "\tEvent Port Connections:" #+ " " + str(componentImplementation.event_port_connections)
        for connection in componentImplementation.event_port_connections:
            if showDebugInfo: print "\t\t%s -> %s, in modes %s" % (connection.source_port, connection.target_port, connection.in_modes)
            self.event_port_connections.append((connection.source_port, connection.target_port, connection.in_modes))

        # Flows
        if showDebugInfo: print "\tFlows:" #+ " " + str(componentImplementation.flows)
        for flow in componentImplementation.flows:
            if showDebugInfo: print "\t\t%s, in modes %s)" % (expressionToString(flow.expression), flow.in_modes)
            self.flows.append((flow.expression, flow.in_modes))


        # *** Collect info about Modes & Transitions ***
        # componentImplementation.modes = dict(modeName: string, modeObject: instance(slim_compiler.src.dataStructure.nominalbehaviour.Mode))
        # componentImplementation.*_mode_transition = list(transition: instance(...))

        # Modes
        if showDebugInfo: print "\tModes:" + " " + str(componentImplementation.modes)
        for (modeName, modeObject) in componentImplementation.modes.items():
            if showDebugInfo: print "\t\t%s/%s (%s)" % (modeName, modeObject.name, modeObject.type)
            self.modes.append((modeName, modeObject.type))
        self.translator._addModeNames(componentImplementation.modes.keys())

        # Normal Mode Transitions
        if showDebugInfo: print "\tNormal Mode Transitions:" + " " + str(componentImplementation.normal_mode_transitions)
        for transition in componentImplementation.normal_mode_transitions:
            if showDebugInfo: print "\t\t%s -[%s when %s then %s]-> %s" % (transition.source_mode, transition.label.trigger, transition.label.guard, expressionsToString(transition.label.effects), transition.target_mode)
            self.transitions.append((transition.source_mode, transition.label.trigger, transition.label.guard, transition.label.effects, transition.target_mode))

        # DEPRECATED: Probability Mode Transitions
        if hasattr(componentImplementation, "probability_mode_transitions") and \
           componentImplementation.probability_mode_transitions != None and \
           componentImplementation.probability_mode_transitions != []:
            self.translator.error_collector.add(Slim2PromelaTranslationError(self.accessPath + " : " + self.packageName + "::" + self.componentTypeName + "." + self.componentImplementationName + " contains deprecated probability mode transitions!"))
        #if showDebugInfo: print "\tProbability Mode Transitions:" + " " + str(componentImplementation.probability_mode_transitions)
        #for transition in componentImplementation.probability_mode_transitions:
        #    if showDebugInfo: print "\t\t%s -[prob %s]-> %s" % (transition.source_mode, transition.label.probability, transition.target_mode)
        #    self.probabilistictransitions.append((transition.source_mode, transition.label.probability, transition.target_mode))      

    # </def _collectProctypeInfo(): ...>



    # ----------------------------------------------------------------------------------------------



    def _getOwnInDataPortByName(self, aName):
        matchings = filter(lambda (name, typ, default): name == aName, self.in_data_ports)
        if len(matchings) > 0:
            return matchings[0]
        else:
            return None
    

    def _isOwnInDataPort(self, aName):
        return self._getOwnInDataPortByName(aName) != None
    

    def _getOwnOutDataPortByName(self, aName):
        matchings = filter(lambda (name, typ, default): name == aName, self.out_data_ports)
        if len(matchings) > 0:
            return matchings[0]
        else:
            return None
        

    def _isOwnOutDataPort(self, aName):
        return self._getOwnOutDataPortByName(aName) != None


    def _getOwnDataSubcomponentByName(self, aName):
        matchings = filter(lambda (name, typ, default, ims): name == aName, self.datasubcomponents)
        if len(matchings) > 0:
            return matchings[0]
        else:
            return None


    def _isOwnDataSubcomponent(self, aName):
        return self._getOwnDataSubcomponentByName(aName) != None


    def _getSubcomponentsInDataPortByName(self, aName):
        parts = aName.split(".")
        if (len(parts) == 2):
            # Try to find appropriate subcomponent ...
            for (scname, scidp, scodp, sciep, scoep, ims, mh) in filter(lambda (scname, scidp, scodp, sciep, scoep, ims, mh): scname == parts[0], self.subcomponents):
                # ... containing that data port
                matchings = filter(lambda (name, typ, default): name == parts[1], scidp)
                if len(matchings) > 0:
                    (name, typ, default) = matchings[0]
                    return (scname, name, typ, default)

        # If no subcomponent with that data port found:
        return None


    def _isSubcomponentsInDataPort(self, aName):
        return self._getSubcomponentsInDataPortByName(aName) != None


    def _getSubcomponentsOutDataPortByName(self, aName):
        parts = aName.split(".")
        if (len(parts) == 2):
            # Try to find appropriate subcomponent ...
            for (scname, scidp, scodp, sciep, scoep, ims, mh) in filter(lambda (scname, scidp, scodp, sciep, scoep, ims, mh): scname == parts[0], self.subcomponents):
                # ... containing that data port
                matchings = filter(lambda (name, typ, default): name == parts[1], scodp)
                if len(matchings) > 0:
                    (name, typ, default) = matchings[0]
                    return (scname, name, typ, default)

        # If no subcomponent with that data port found:
        return None


    def _isSubcomponentsOutDataPort(self, aName):
        return self._getSubcomponentsOutDataPortByName(aName) != None


    def _isSubcomponentsInEventPort(self, aTrigger):
        parts = aTrigger.split(".")
        if len(parts) == 2:
            for (scname, scidp, scodp, sciep, scoep, ims, mh) in filter(lambda (scname, scidp, scodp, sciep, scoep, ims, mh): scname == parts[0], self.subcomponents):
                if parts[1] in sciep:
                    return True
        return False
        
        
    def _isSubcomponentsOutEventPort(self, aName):
        parts = aName.split(".")
        if len(parts) == 2:
            for (scname, scidp, scodp, sciep, scoep, ims, mh) in filter(lambda (scname, scidp, scodp, sciep, scoep, ims, mh): scname == parts[0], self.subcomponents):
                if parts[1] in scoep:
                    return True
        return False


    def _getSubcomponentByName(self, aName):
        aName = aName.split(".")[0]    # If aName is <scname>.<subname>, e.g. for SC's data port, only take scname
        for sc in self.subcomponents:
            if sc[0] == aName:
                return sc
        return None


    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


    def _getTypeOfVariable(self, aName):
        # Own In Data Port
        if self._isOwnInDataPort(aName):
            return self._getOwnInDataPortByName(aName)[1]
        # Own Out Data Port
        elif self._isOwnOutDataPort(aName):
            return self._getOwnOutDataPortByName(aName)[1]
        # Own Data Subcomponent
        elif self._isOwnDataSubcomponent(aName):
            return self._getOwnDataSubcomponentByName(aName)[1]
        # Subcomponent's In Data Port
        elif self._isSubcomponentsInDataPort(aName):
            return self._getSubcomponentsInDataPortByName(aName)[3]
        # Subcomponent's Out Data Port
        elif self._isSubcomponentsOutDataPort(aName):
            return self._getSubcomponentsOutDataPortByName(aName)[3]
        # None of all (enum constant or invalid)
        else:
            return None


    def _isVariableOfTypeReal(self, aName):
        typ = self._getTypeOfVariable(aName)
        if typ != None:
            return isinstance(typ, Real)


    @deprecated
    def _isVariableOfTypeIntRange(self, aName):
        typ = self._getTypeOfVariable(aName)
        if typ != None:
            return isinstance(typ, IntRange)

        
    def _doesExpressionRequireEmbeddedCCode(self, anExpressionTreeNode):
        if anExpressionTreeNode == None:
            result = None
            
        elif isinstance(anExpressionTreeNode, BinaryOperatorTreeNode):
            # For boolean opearators (and, iff, imp, or, xnor, xor): always false
            if anExpressionTreeNode.operator in [OperatorType.AND, OperatorType.IFF, OperatorType.IMP, OperatorType.OR, OperatorType.XNOR, OperatorType.XOR]:
                result = False
            # For all others (+, -, *, ...; =, !=, <, ...): check subexpressions
            else:
                result = self._doesExpressionRequireEmbeddedCCode(anExpressionTreeNode.left) or self._doesExpressionRequireEmbeddedCCode(anExpressionTreeNode.right)
                
        elif isinstance(anExpressionTreeNode, UnaryOperatorTreeNode):
            result = self._doesExpressionRequireEmbeddedCCode(anExpressionTreeNode.subexpression)
            
        elif isinstance(anExpressionTreeNode, IdentifierTreeNode):
            result = self._isVariableOfTypeReal(anExpressionTreeNode.name)
            
        elif isinstance(anExpressionTreeNode, NumberTreeNode):
            result = anExpressionTreeNode.node_type == NodeType.REAL

        elif isinstance(anExpressionTreeNode, CaseExpressionTreeNode):
            # HINT: Occurence of reals inside an b_i do not matter because this is handeled in translation of them. Only reals inside an e_i matter for the surrounding expression!
            result = reduce(lambda b1, b2: b1 or b2, map(lambda (b, e): self._doesExpressionRequireEmbeddedCCode(e), anExpressionTreeNode.cases), False)
            result = result or self._doesExpressionRequireEmbeddedCCode(anExpressionTreeNode.default_expression)
            
        elif isinstance(anExpressionTreeNode, BooleanTreeNode) or isinstance(anExpressionTreeNode, CaseExpressionTreeNode) or \
             isinstance(anExpressionTreeNode, FlowEquationTreeNode) or isinstance(anExpressionTreeNode, FlowExpressionTreeNode):
            result = False
            
        else:
            result = None
            
        return result


    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


    @staticmethod
    def _datatypeToPromela(aDatatype):
        # SLIM          Object      Promela/C
        # ----          ------      ---------
        # bool          Bool        bool
        # clock         Clock       ---
        # continuous    Continues   ---
        # enum          Enum        translate names to ints (globally: translator.enumNames)
        # int           Int         int
        # real          Real        int + c_code{float}
        # [l..u]        IntRange    int + modulo
        if isinstance(aDatatype, Bool):
            txt = "bool"
        elif isinstance(aDatatype, Enum):
            txt = "enumIntType"
        elif isinstance(aDatatype, Int) or isinstance(aDatatype, Real) or isinstance(aDatatype, IntRange):
            txt = "int"
        else:
            txt = "ERR: unsupported type '" + aDatatype.get_name() + "'"
        return txt


    def _identifierToPromela(self, anIdentifier, buffered = False):
        """Translates a SLIM identifier to the corresponding Promela variable.

        For naming convention of Promela variables, see ProctypeInfo.toPromela()

        @type anIdentifier:  string
        @param anIdentifier: identifier of a data port or data subcomponent in the SLIM component specification
        @type buffered:  boolean, default: False
        @param buffered: Controls whether buffered (True: "New") or unbuffered (False: "Store") variable name is returned
        @rtype:  string
        @return: name of corresponding Promela variable
        """

        # First, fix appropriate storage label
        if buffered:
            storageLabel = "New"
        else:
            storageLabel = "Store"

        # a) Assign to own out data port
        if self._isOwnOutDataPort(anIdentifier):
            return "odp" + storageLabel + "_" + self.accessPath + "_" + anIdentifier
        # b) Assign to own data subcomponent
        elif self._isOwnDataSubcomponent(anIdentifier):
            return "dsc" + storageLabel + "_" + self.accessPath + "_" + anIdentifier
        # c) Assign to subcomponent's in data port -- only allowed in initialization and reset
        elif self._isSubcomponentsInDataPort(anIdentifier):
            return "idp" + storageLabel + "_" + self.accessPath + "_" + anIdentifier.split(".")[0] + "_" + anIdentifier.split(".")[1]
        # d) Assign to own in data port -- only allowed in initialization
        elif self._isOwnInDataPort(anIdentifier):
            return "idp" + storageLabel + "_" + self.accessPath + "_" + anIdentifier
        # e) Assign to subcomponent's out data port -- only allowed in initialization
        elif self._isSubcomponentsOutDataPort(anIdentifier):
            return "odp" + storageLabel + "_" + self.accessPath + "_" + anIdentifier.split(".")[0] + "_" + anIdentifier.split(".")[1]
        # f) Invalid assignment
        else:
            self.translator.error_collector.add(Slim2PromelaTranslationError("'" + anIdentifier + "' is an invalid SLIM identifier and cannot be translated to a Promela variable. (in " + self.accessPath + " : " + self.packageName + "::" + self.componentTypeName + "." + self.componentImplementationName + ")"))
            return "ERROR: '" + anIdentifier + "' is an invalid SLIM identifier and cannot be translated to a Promela variable"


    def _guardToPromela(self, aGuard):
        if aGuard == None:
            return "true"
        else:
            return self._expressionToPromela(aGuard)


    def _splitEffectToAssignment(self, anExpressionTreeNode):
        if isinstance(anExpressionTreeNode, BinaryOperatorTreeNode):
            if anExpressionTreeNode.operator == OperatorType.ASSIGN:
                if isinstance(anExpressionTreeNode.left, IdentifierTreeNode):
                    if self._isOwnOutDataPort(anExpressionTreeNode.left.name) or self._isOwnDataSubcomponent(anExpressionTreeNode.left.name) \
                       or self._isSubcomponentsInDataPort(anExpressionTreeNode.left.name):
                        return (anExpressionTreeNode.left.name, anExpressionTreeNode.right, True, None)
                    else:
                        return (None, None, False, "Effect's left hand side '" + anExpressionTreeNode.left.name + "' does not identify an out data port, data subcomponent, or subcomponents in data port")
                else:
                    return (None, None, False, "Effect's left hand side is no identifier")
            else:
                return (None, None, False, "Effect is no assignment")
        else:
            return (None, None, False, "Effect does not have a binary operator as root")


    @staticmethod
    def _getBufferRequirementsFromEffects(effects):
        writtenTo = set()
        needBuffer = set()
        
        for effect in effects:
            needBuffer.update( writtenTo & set(expressionReadsFrom(effect)) )
            writtenTo.update( [expressionWritesTo(effect)] )

        return needBuffer
            

    def _effectsToPromela(self, anExpressionList, aPrefix = "\t\t"):
        assignToNew = ""
        copyNewToStore = ""

        needBuffer = self._getBufferRequirementsFromEffects(anExpressionList)
                
        for anExpressionTreeNode in anExpressionList:
            (target, rhs, ok, err) = self._splitEffectToAssignment(anExpressionTreeNode)
            
            if ok:
                # Perform valid assignment.
                # - Buffer new value ...
                #       * out data ports: always, since needed for fixpoint iteration
                #       * data subcomponent: only when the old value is read in a later effect of same transition!
                # - Of course the old values are read in _assignmentToPromela / _expressionToPromela from the unbuffered version!
                if self._isOwnOutDataPort(target):
                    assignToNew += aPrefix + self._assignmentToPromela(target, rhs, True) + ";\n"
                    # New value for out data port is not copied to store since this will be done by parent in fixpoint iteration.
                elif self._isOwnDataSubcomponent(target):
                    assignToNew += aPrefix + self._assignmentToPromela(target, rhs, target in needBuffer) + ";\n"
                    if target in needBuffer:
                        copyNewToStore += aPrefix + self._identifierToPromela(target, False) + " = " + self._identifierToPromela(target, True) + "; "
                        copyNewToStore += self._identifierToPromela(target, True) + " = 0;  // Copy & reset buffer\n"
                # For flows to subcomponents in data ports
                elif self._isSubcomponentsInDataPort(target):
                    assignToNew += aPrefix + self._assignmentToPromela(target, rhs, False) + ";\n"
                    # No buffering for subcomponents in data ports necessary.
            else:
                self.translator.error_collector.add(Slim2PromelaTranslationError(err + ". (in " + self.accessPath + " : " + self.packageName + "::" + self.componentTypeName + "." + self.componentImplementationName + ")"))
                assignToNew += aPrefix + "ERROR: " + err + ";\n"
                
        return assignToNew + copyNewToStore


    def _assignmentToPromela(self, target, expression, buffered = False):
        """Translates target = expression to Promela. If buffered, its not stored to '...Store_*' but to '...New_*'

        # Possible Targets of Assignments and when they are allowed:
        #
        #                          Init   Reset   Effect   Update (DataPortCon., Flow)
        #                          ----   -----   ------   ------
        #
        # d) own in data port      (+) ~> Parent always sets new value before sending !update.
        # a) own out data port     +      +       +        +
        #
        # b) own data subcomponent +      +       +        -
        #
        # c) sc's in data port     +  ??????????????????????????????    +
        # e) sc's out data port    (+) ~> Value will always be received from subcomponent via ?update.
        """
        
        # Target is a real variable => start embedded c-code
        if self._isVariableOfTypeReal(target):
            txt = "c_code { *((float *)&(now." + self._identifierToPromela(target, buffered) + ")) = "  #Abuse int as float:   *((float *)&(now.myGlobalIntVar)) = ...
            txt += self._expressionToPromela(expression, True) + "; }"

        # Target is a "normal" variable => just assign
        # HINT: Modulo calculation for IntRange variables is already added by the checker.
        else:
            txt = self._identifierToPromela(target, buffered) + " = " + self._expressionToPromela(expression)

        return txt

    
    def _expressionToPromela(self, anExpressionTreeNode, asEmbeddedCCode = False):
        """Translates SLIM-expression stored in anExpressionTreeNode to Promela. Assumes, it is not None (otherwise this Expression should not be used!). Does not handle assignments (use _effectsToPromela instead).
        The value for identifiers is always read from the unbuffered version! (for _identifierToPromela, buffered is by default false)"""

        # First check, whether embedded C-Code has to be opened for subexpression (which must be a comparison)
        if (not asEmbeddedCCode) and self._doesExpressionRequireEmbeddedCCode(anExpressionTreeNode):
            return "c_expr { " + self._expressionToPromela(anExpressionTreeNode, True) + " }"
            
        # Distinguish the different Node Types and generate appropriate Promela expressions
        if isinstance(anExpressionTreeNode, BinaryOperatorTreeNode):
            left  = "(" + self._expressionToPromela(anExpressionTreeNode.left, asEmbeddedCCode)  + ")"
            right = "(" + self._expressionToPromela(anExpressionTreeNode.right, asEmbeddedCCode) + ")"

            # SLIM      OperatorType       Promela (int)   C (real)
            # ----      ------------       ----         -
            # +         .PLUS              +            +
            # -b        .MINUS             -            -
            # *         .TIMES             *            *
            # /         .DIVIDE            /            /
            # mod       .MOD             ! %          ! %
            #
            # =         .EQUAL           ! ==           ==
            # !=        .NOTEQUAL          !=           !=
            # <         .LESS              <            <
            # >         .GREATER           >            >
            # <=        .LESSOREQUAL       <=           <=
            # >=        .GREATEROREQUAL    >=           >=
            #
            # and       .AND             ! &&         ! &&
            # iff       .IFF            !! ...       !! ...
            # imp       .IMP            !! ...       !! ...
            # or        .OR              ! ||         ! ||
            # xnor      .XNOR           !! ...       !! ...
            # xor       .XOR            !! ...       !! ...
            if anExpressionTreeNode.operator == OperatorType.MOD:
                txt = left + " % " + right
            elif anExpressionTreeNode.operator == OperatorType.EQUAL:
                txt = left + " == " + right
            elif anExpressionTreeNode.operator == OperatorType.AND:
                txt = left + " && " + right
            elif anExpressionTreeNode.operator == OperatorType.OR:
                txt = left + " || " + right
            elif anExpressionTreeNode.operator in [OperatorType.IFF, OperatorType.XNOR]:
                # a <=> b   <---->   (a and b) or (!a and !b)  /  (a => b) and (b => a)  <->  (b or !a) and (a or !b)
                # a XNOR b   <---->   a <=> b   <---->   ...
                txt = "(" + left + " && " + right + ") or (!" + left + " && !" + right + ")"
            elif anExpressionTreeNode.operator == OperatorType.IMP:
                # a => b   <---->   b or !a
                txt = right + " || !" + left
            elif anExpressionTreeNode.operator == OperatorType.XOR:
                # a XOR b   <---->   (a and !b) or (!a and b)
                txt = "(" + left + " && !" + right + ") || (!" + left + " && " + right + ")"
            elif anExpressionTreeNode.operator in [OperatorType.PLUS, OperatorType.MINUS, OperatorType.TIMES, OperatorType.DIVIDE, OperatorType.NOTEQUAL,
                                                   OperatorType.LESS, OperatorType.GREATER, OperatorType.LESSOREQUAL, OperatorType.GREATEROREQUAL]:
                txt = left + " " + OperatorType.operator_to_str(anExpressionTreeNode.operator) + " " + right
            else:
                self.translator.error_collector.add(Slim2PromelaTranslationError("'" + OperatorType.operator_to_str(anExpressionTreeNode.operator) + "' unsupported as binary operator. (in " + self.accessPath + " : " + self.packageName + "::" + self.componentTypeName + "." + self.componentImplementationName + ")"))
                txt = left + "ERROR: '" + OperatorType.operator_to_str(anExpressionTreeNode.operator) + "' unsupported as binary operator" + right

        elif isinstance(anExpressionTreeNode, UnaryOperatorTreeNode):
            right = "(" + self._expressionToPromela(anExpressionTreeNode.subexpression, asEmbeddedCCode) + ")"
            
            # SLIM      OperatorType       Promela (int)
            # ----      ------------       ----
            # -u        .UNARY_MINUS       -
            # not       .NOT             ! !
            if anExpressionTreeNode.operator == OperatorType.UNARY_MINUS:
                txt = "-" + right
            elif anExpressionTreeNode.operator == OperatorType.NOT:
                txt = "!" + right
            else:
                self.translator.error_collector.add(Slim2PromelaTranslationError("'" + OperatorType.operator_to_str(anExpressionTreeNode.operator) + "' unsupported as unary operator. (in " + self.accessPath + " : " + self.packageName + "::" + self.componentTypeName + "." + self.componentImplementationName + ")"))
                txt = "ERROR: '" + OperatorType.operator_to_str(anExpressionTreeNode.operator) + "' unsupported as unary operator" + right
            
        elif isinstance(anExpressionTreeNode, IdentifierTreeNode):
            # TODO: This is not safe, when enum identifiers are identical to a data element name. Consider node type!!!
            # Read own in/out data port, data subcomponent, subcomponent's in/out data port
            if self._isOwnInDataPort(anExpressionTreeNode.name) or \
               self._isOwnOutDataPort(anExpressionTreeNode.name) or \
               self._isOwnDataSubcomponent(anExpressionTreeNode.name) or \
               self._isSubcomponentsInDataPort(anExpressionTreeNode.name) or \
               self._isSubcomponentsOutDataPort(anExpressionTreeNode.name):
                if asEmbeddedCCode:
                    if self._isVariableOfTypeReal(anExpressionTreeNode.name):
                        # Typecast-Trick for abusing Promela-ints as floats:   *((float *)&(now.myGlobalIntVar))
                        txt = "*((float *)&(now." + self._identifierToPromela(anExpressionTreeNode.name) + "))"
                    else:
                        txt = "now." + self._identifierToPromela(anExpressionTreeNode.name)
                else:
                    txt = self._identifierToPromela(anExpressionTreeNode.name)
            # Identifier is constant of an ENUM data type   (should hold: .node_type == NodeType.ENUM_VALUE)
            elif (self.translator != None) and (anExpressionTreeNode.name in self.translator.enumNames):
                txt = "enum_" + anExpressionTreeNode.name
            # Invalid Read
            else:
                self.translator.error_collector.add(Slim2PromelaTranslationError("Invalid identifier '" + anExpressionTreeNode.name + "' on effect's right hand side. (in " + self.accessPath + " : " + self.packageName + "::" + self.componentTypeName + "." + self.componentImplementationName + ")"))
                txt = "ERROR: invalid identifier '" + anExpressionTreeNode.name + "' on effect's right hand side"

        elif isinstance(anExpressionTreeNode, BooleanTreeNode):
            txt = str(anExpressionTreeNode.value).lower()

        elif isinstance(anExpressionTreeNode, NumberTreeNode):
            if anExpressionTreeNode.node_type == NodeType.REAL:
                if asEmbeddedCCode:
                    txt = str(anExpressionTreeNode.number)
                    if anExpressionTreeNode.get_type() == NodeType.REAL_CONST:
                        txt += "f"
                    elif anExpressionTreeNode.get_type() == NodeType.INT_CONST:
                        txt += ".0f"
                    else:
                        self.translator.error_collector.add(Slim2PromelaTranslationError("Value '" + str(anExpressionTreeNode.number) + "' cannot be used in expression of type real. (in " + self.accessPath + " : " + self.packageName + "::" + self.componentTypeName + "." + self.componentImplementationName + ")"))
                        txt = "ERROR: Value '" + str(anExpressionTreeNode.number) + "' cannot be used in expression of type real"
                else:
                    self.translator.error_collector.add(Slim2PromelaTranslationError("Value '" + str(anExpressionTreeNode.number) + "' . (in " + self.accessPath + " : " + self.packageName + "::" + self.componentTypeName + "." + self.componentImplementationName + ")"))
                    txt = "ERROR: Value '" + str(anExpressionTreeNode.number) + "' of type real can be used in embedded C-code only"
            else:
                txt = str(anExpressionTreeNode.number)

        elif isinstance(anExpressionTreeNode, CaseExpressionTreeNode):
            if asEmbeddedCCode:
                sign = " ? "
            else:
                sign = " -> "
            
            # Translate first case
            txt = "(" + self._expressionToPromela(anExpressionTreeNode.cases[0][0], asEmbeddedCCode)
            txt += sign + self._expressionToPromela(anExpressionTreeNode.cases[0][1], asEmbeddedCCode) + " : "
            # Only one case => translate "otherwise"
            if len(anExpressionTreeNode.cases) == 1:
                txt += self._expressionToPromela(anExpressionTreeNode.default_expression, asEmbeddedCCode)
            # Not yet "otherwise" => translate the remaining cases
            else:
                smallerCaseExpression = copy.copy(anExpressionTreeNode)
                smallerCaseExpression.cases = smallerCaseExpression.cases[1:] # Slice to ignore previously translated first case
                txt += self._expressionToPromela(smallerCaseExpression, asEmbeddedCCode)
            txt += ")"

        elif isinstance(anExpressionTreeNode, FlowEquationTreeNode) or isinstance(anExpressionTreeNode, FlowExpressionTreeNode):
            self.translator.error_collector.add(Slim2PromelaTranslationError("Flow equations for contunuous variables are not supported. (in " + self.accessPath + " : " + self.packageName + "::" + self.componentTypeName + "." + self.componentImplementationName + ")"))
            txt = "ERROR: Flow equations for continuous variables are not supported"

        else:
            self.translator.error_collector.add(Slim2PromelaTranslationError("Unknown type of ExpressionTreeNode occured. (in " + self.accessPath + " : " + self.packageName + "::" + self.componentTypeName + "." + self.componentImplementationName + ")"))
            txt = "ERROR: Unknown type of ExpressionTreeNode occured."

        return txt


    def _transitionCoreToPromela(self, SourceMode, Effects, DestinationMode, aPrefix = "\t\t", adjustSubcomponentsActivation = True):
        txt = ""
        
        # Reset Data Subcomponents which are reactivated
        # - Do so before Effects are translated so that they overwrite the defaults.
        # - Can directly write to dscStore_* (unbuffered => don't need to copy later), since old value may not be read in effects.
        # - Not reset on deactivation as not necessary when never reactivated (value shouldn't be read when deactivated).
        for (name, typ, default, ims) in filter(lambda (name, typ, default, ims): (SourceMode not in ims) and (DestinationMode in ims), self.datasubcomponents):
            if default != None:
                txt += aPrefix + self._assignmentToPromela(name, default) + ";\n"
            else:
                txt += aPrefix + self._identifierToPromela(name) + " = 0;\n"


        # Reset Data Ports which are disconnected from Connections/Flows
        # - Do so before Effects are translated so that they overwrite the defaults (mixing of effects and connections/flows in different modes is allowed).
        # - New value for out data ports are always written to odpNew_* (buffered), but the copy to odpStore_* (unbuffered) is done by supercomponent in Fixpoint Iteration!
        # - Opposite to Data Subcomponents, reset for Out Data Ports is necessary already on disconnection (ports can't be deactivated and must always have a defined value).
        # - New connected Connections/Flows are considered in the update methods and overwrite the defaults.

        # 1.) Collect disconnected data ports
        # HINT: Every data port occures at most once, because fan-in is not possible.
        # 1.a) Disconnected from Data Port Connection
        disconnectedDataPorts = [dp for (sp, dp, ims) in filter(lambda (sp, dp, ims): (SourceMode in ims) and (DestinationMode not in ims), self.data_port_connections)]
        # 1.b) Disconnected from Flow
        for (expr, ims) in filter(lambda (expr, ims): (SourceMode in ims) and (DestinationMode not in ims), self.flows):
            (target, rhs, ok, err) = self._splitEffectToAssignment(expr)
            if ok:
                disconnectedDataPorts.append(target)
            else:
                self.translator.error_collector.add(Slim2PromelaTranslationError("Invalid flow ('" + err + "') is disconnected. (in " + self.accessPath + " : " + self.packageName + "::" + self.componentTypeName + "." + self.componentImplementationName + ")"))
                txt += aPrefix + "ERROR: Invalid flow ('" + err + "') is disconnected\n"

        # 2.) Generate reset code
        for dp in disconnectedDataPorts:
            # Own out data port (in - out / out - out)
            if self._isOwnOutDataPort(dp):
                (name, typ, default) = self._getOwnOutDataPortByName(dp)
                if default != None:
                    txt += aPrefix + self._assignmentToPromela(dp, default, True) + ";\n"
                else:
                    txt += aPrefix + self._identifierToPromela(dp, True) + " = 0;\n"
            # Subcomponents' in data port (in - in / out - in)
            # - Subcomponents' in data ports do not have to be buffered since they cannot be read in Effects!
            elif self._isSubcomponentsInDataPort(dp):
                (scname, name, typ, default) = self._getSubcomponentsInDataPortByName(dp)
                if default != None:
                    txt += aPrefix + self._assignmentToPromela(dp, default) + ";\n"
                else:
                    txt += aPrefix + self._identifierToPromela(dp) + " = 0;\n"
            # Everything else in invalid!
            else:
                self.translator.error_collector.add(Slim2PromelaTranslationError("Invalid data port identifier '" + dp + "' disconnected from data port connection or flow. (in " + self.accessPath + " : " + self.packageName + "::" + self.componentTypeName + "." + self.componentImplementationName + ")"))
                txt += aPrefix + "ERROR: Invalid data port identifier '" + dp + "' disconnected from data port connection or flow.\n"


        # Translate the Effects
        txt += self._effectsToPromela(Effects, aPrefix)


        # Change the stored Nominal Mode
        # - HINT: "}" closes d_step, which must be opened before _transitionCoreToPromela!
        txt += aPrefix + "nominalMode_" + self.accessPath + " = modeID_" + DestinationMode + "; }\n"


        # Adjust Subcomponents Activation
        # - HINT: For Reactive Transitions to a subcomponent's out event (= Trymultiway incomming from a subcomponent) the adjustment of subcomponents' activation has to be delayed!
        if adjustSubcomponentsActivation:
            txt += self._adjustSubcomponentsActivation(SourceMode, DestinationMode, aPrefix)

        return txt


    def _adjustSubcomponentsActivation(self, SourceMode, DestinationMode, aPrefix = "\t\t"):
        txt = ""
        
        # Deactivate those, active in SourceMode but inactive in DestinationMode
        for (scname, scidp, scodp, sciep, scoep, ims, mh) in filter(lambda (scname, scidp, scodp, sciep, scoep, ims, mh): (SourceMode in ims) and (DestinationMode not in ims), self.subcomponents):
            txt += aPrefix + "chan_" + self.accessPath + "_" + scname + "!deactivate, 0;\n"
            txt += aPrefix + "chan_" + self.accessPath + "_" + scname + "?deactivate, _;\n"

        # Activate those, inactive in SourceMode but active in DestinationMode
        for (scname, scidp, scodp, sciep, scoep, ims, mh) in filter(lambda (scname, scidp, scodp, sciep, scoep, ims, mh): (SourceMode not in ims) and (DestinationMode in ims), self.subcomponents):
            txt += aPrefix + "chan_" + self.accessPath + "_" + scname + "!activate, 0;\n"
            txt += aPrefix + "chan_" + self.accessPath + "_" + scname + "?activate, _;\n"

        return txt

    
    @staticmethod
    def _EventToNumber(anEventPort, inEvents, outEvents):
        if anEventPort in inEvents:
            return 1 + inEvents.index(anEventPort)
        elif anEventPort in outEvents:
            return len(inEvents) + 1 + outEvents.index(anEventPort)
        else:
            return 0


    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


    def _prepare(self):
        """Perform some preparation on the collected data to simplify the translation to Promela."""

        # Search start mode (initial or activation) & check whether mode history supported
        startModes = filter(lambda (modeName, modeType): modeType in [slim_compiler.src.dataStructure.nominalbehaviour.ModeType.INITIAL, slim_compiler.src.dataStructure.nominalbehaviour.ModeType.ACTIVATION], self.modes)
        if len(startModes) == 1:
            self.startMode = startModes[0][0]
            self.modeHistory = startModes[0][1] == slim_compiler.src.dataStructure.nominalbehaviour.ModeType.INITIAL
        else:
            self.translator.error_collector.add(Slim2PromelaTranslationError("No or no unique start mode available. (in " + self.accessPath + " : " + self.packageName + "::" + self.componentTypeName + "." + self.componentImplementationName + ")"))
            self.startMode = "ERROR: No or no unique start mode available."
            self.modeHistory = False

        # Determine all data subcomponents that need a buffer variable
        for (sm, tr, gu, effects, dm) in self.transitions:
            self.toBeBuffered.update(self._getBufferRequirementsFromEffects(effects))
            
        # Sort port lists to make sure that event ports always get the same number
        # (Number depends on order of elements, but the lists of ports are created from dictionaries, which may change their ordered!)
        # Hint: Sorting changes the list in place - do not assign the sorted list to the original list!
        self.in_event_ports.sort()
        self.out_event_ports.sort()
        for (scname, scidp, scodp, sciep, scoep, ims, mh) in self.subcomponents:
            sciep.sort()
            scoep.sort()      



    # ----------------------------------------------------------------------------------------------    



    def toPromela(self):
        """Translates the SLIM component (from which all relevant information was collected by calling _collectProctypeInfo(<componentImplementation>) beforehand) to Promela.
        A declaration block for global variables and a proctype definition together with inline functions are generated and returned as a tuple.

        # Naming convention:   <Kind:idp,odp,dsc><Buffer:Store,New>_<accessPath>_<elementName>
        # => (Own) In Data Port:   idp Store     _(<componentName>)+_<elementName>
        # => (Own) Out Data Port:  odp(Store|New)_(<componentName>)+_<elementName>
        # => (Own) Data Subcomp.:  dsc(Store|New)_(<componentName>)+_<elementName>
        # ~> (SC's In Data Port:   idp(Store|New)_(<componentName>)+_<subcomponentName>_<elementName> ) -- generated by SC
        # ~> (SC's Out Data P.:    odp(Store|New)_(<componentName>)+_<subcomponentName>_<elementName> ) -- generated by SC
        #
        # * Channel to components: chan_<accessPath>
        # * Proctype of component: process_<accessPath>
        # * Inlines of component:  inlineUpdateMode<modeName>_<accessPath>
        # * Enum Names:            enum_<enumName>
        # * Mode IDs:              modeID_<modeName>
        # * Variable storing mode: nominalMode_<accessPath>
        # * Mode Labels:           mode_<modeName>

        @return: tuple (global variable declarations: string, proctype definition: string)
        """

        # *** Preparations ***
        globalVariableDeclarations = ""
        inlineDefinitions = ""
        proctypeDefinition = ""
        self._prepare()



        # *** Global Variables for storing Inputs, Outputs, Subcomponents ***

        globalVariableDeclarations += "// *** " + self.accessPath + " : " + self.packageName + "::" + self.componentTypeName + "." + self.componentImplementationName + " ***\n"
        globalVariableDeclarations += "chan chan_" + self.accessPath + " = [0] of { mtype, int };\n"
        globalVariableDeclarations += "modeIntType nominalMode_" + self.accessPath + ";\n"
        self.translator._addGlobalStateVariable(self.accessPath + ".mode", "nominalMode_" + self.accessPath)

        # Own Data Ports
        for (name, typ, default) in self.in_data_ports:
            globalVariableDeclarations += self._datatypeToPromela(typ) + " idpStore_" + self.accessPath + "_" + name + ";\n"
            self.translator._addGlobalStateVariable(self.accessPath + "." + name, "idpStore_" + self.accessPath + "_" + name)
        for (name, typ, default) in self.out_data_ports:
            globalVariableDeclarations += self._datatypeToPromela(typ) + " odpStore_" + self.accessPath + "_" + name + "; "
            # HINT: Buffer is always needed, for fixpoint iteration.
            globalVariableDeclarations += self._datatypeToPromela(typ) + " odpNew_" + self.accessPath + "_" + name + ";\n"
            self.translator._addGlobalStateVariable(self.accessPath + "." + name, "odpStore_" + self.accessPath + "_" + name)

        # Own Data Subcomponents
        for (name, typ, default, ims) in self.datasubcomponents:
            globalVariableDeclarations += self._datatypeToPromela(typ) + " dscStore_" + self.accessPath + "_" + name + ";"
            # HINT: Buffer is only needed if a transition with effects exist where assignment takes place before last reading access.
            if name in self.toBeBuffered:
                globalVariableDeclarations += " " + self._datatypeToPromela(typ) + " dscNew_" + self.accessPath + "_" + name + ";"
            globalVariableDeclarations += "\n"
            self.translator._addGlobalStateVariable(self.accessPath + "." + name, "dscStore_" + self.accessPath + "_" + name)


        
        # *** Head of proctype ***
        # HINT: Length of proctype name must be at most 55! Therefore, slice it "[:55]". To avoid name clashes, each proctype gets a unique number (1..254).
        inlineDefinitions += "// ----- " + self.accessPath + " : " + self.packageName + "::" + self.componentTypeName + "." + self.componentImplementationName + " --------------------------------------------------\n\n"
        proctypeDefinition = "active proctype " + ("process" + str(self.translator._getNextProctypeNumber()) + "_" + self.accessPath)[:55] + "()\n"
        proctypeDefinition += "{\n"
        if self.addPrintfs: proctypeDefinition += "\tprintf(\"### " + self.accessPath + "(%d) ### started\\n\", _pid);\n"



        # *** Wait for first activation ***
        proctypeDefinition += "\n\n\n\t// *** Wait for first activation ***\n"
        proctypeDefinition += "FirstActivationOrReactivationWithoutModeHistory:\n"
        proctypeDefinition += "\tchan_" + self.accessPath + "?activate, _;\n"
        proctypeDefinition += "\td_step {\n"

        # Initialize Own Data Ports
        proctypeDefinition += "\n\t\t// Initialize Own Data Ports\n"
        for (name, typ, default) in self.in_data_ports:
            if default != None: # otherwise: SPIN's default = 0
                proctypeDefinition += "\t\t" + self._assignmentToPromela(name, default) + ";\n"
        for (name, typ, default) in self.out_data_ports:
            if default != None: # otherwise: SPIN's default = 0
                proctypeDefinition += "\t\t" + self._assignmentToPromela(name, default, True) + "; "
                proctypeDefinition += self._assignmentToPromela(name, default) + ";\n"

        # Initialize Data Subcomponents (only those active in start mode)
        # TODO: - THINK: Should also the inactive data subcomponents be initialized? => ~No problem, when initialized in declaration.
        proctypeDefinition += "\n\t\t// Initialize Data Subcomponents (only those active in start mode)\n"
        for (name, typ, default, ims) in filter(lambda (name, typ, default, ims): self.startMode in ims, self.datasubcomponents):
            if default != None: # otherwise: SPIN's default = 0
                proctypeDefinition += "\t\t" + self._assignmentToPromela(name, default) + ";\n"

        # Initialize Mode Info
        proctypeDefinition += "\n\t\t// Initialize Mode Info\n"
        proctypeDefinition += "\t\tnominalMode_" + self.accessPath + " = modeID_" + self.startMode + ";\n"

        proctypeDefinition += "\t}\n"

        # Activate Non-Data Subcomponents (only those active in start mode)
        proctypeDefinition += "\n\t// Activate Non-Data Subcomponents (only those active in start mode)\n"
        for (scname, scidp, scodp, sciep, scoep, ims, mh) in filter(lambda (scname, scidp, scodp, sciep, scoep, ims, mh): self.startMode in ims, self.subcomponents):
            proctypeDefinition += "\tchan_" + self.accessPath + "_" + scname + "!activate, 0;\n"
            proctypeDefinition += "\tchan_" + self.accessPath + "_" + scname + "?activate, _;\n"

        # Activation completed
        proctypeDefinition += "\n\t// Activation completed\n"
        if self.addPrintfs: proctypeDefinition += "\tprintf(\"### " + self.accessPath + "(%d) ### activated\\n\", _pid);\n"
        proctypeDefinition += "\tchan_" + self.accessPath + "!activate, 0;\n"
        proctypeDefinition += "\tgoto mode_" + self.startMode + ";\n"



        # *** Run as automaton & React to messages ***
        proctypeDefinition += "\n\n\n\t// *** Run as automaton & React to messages ***\n"

        # Iterate over all modes
        for (modeName, modeType) in self.modes:


            # Label
            proctypeDefinition += "\n\n\t// --- Mode: " + modeName + " ---\nmode_" + modeName + ":\n"
            if self.addPrintfs: proctypeDefinition += "\tprintf(\"### " + self.accessPath + "(%d) ### enters state '" + modeName + "'\\n\", _pid);\n"


            # Start mode's do-loop
            proctypeDefinition += "\tdo\n"


            # Transitions
            for (sm, tr, gu, ef, dm) in filter(lambda (sm, tr, gu, ef, dm): sm == modeName, self.transitions):

                # Proactive Transition (w/o event)
                if tr == None:
                    proctypeDefinition += "\n\t// Proactive Transition (w/o event): %s -[when %s then %s]-> %s\n" % (sm, expressionToString(gu), expressionsToString(ef), dm)
                    proctypeDefinition += "\t:: nT(" + self._guardToPromela(gu) + ") -> d_step {\n"
                    proctypeDefinition += self._transitionCoreToPromela(sm, ef, dm)
                    if self.addPrintfs:
                        proctypeDefinition += "\t\tprintf(\"### " + self.accessPath + "(%d) ### takes proactive transition: " + sm + \
                                              " -[when " + expressionToString(gu) + " then " + expressionsToString(ef) + "]-> " + dm + "\\n\", _pid);\n"
                    proctypeDefinition += "\t\tchan_" + self.accessPath + "!invalidate, 0;  // Propagate changes of data ports & fT by root! (Could be avoided when no changes to data ports.)\n"
                    proctypeDefinition += "\t\tgoto mode_" + dm + ";\n"

                # Master Transition (own out event)
                # HINT: As ErrorEventPorts are stored in <ComponentType>.out_event_ports, this section handles "master error transitions" as well.
                #       However, the rate distribution is not considered here. TODO: Give warning on that!!!
                elif (tr.find(".") == -1) and (tr in self.out_event_ports):
                    proctypeDefinition += "\n\t// Master Transition (own out event): %s -[%s when %s then %s]-> %s\n" % (sm, tr, expressionToString(gu), expressionsToString(ef), dm)
                    proctypeDefinition += "\t:: nT(" + self._guardToPromela(gu) + ") ->\n"
                    proctypeDefinition += "\t\t// We know: gflagTrymultiway == false, since reset after every trymultiway!\n"
                    proctypeDefinition += "\t\tchan_" + self.accessPath + "!trymultiway, " + str(self._EventToNumber(tr, self.in_event_ports, self.out_event_ports)) + ";\n"
                    proctypeDefinition += "\t\tchan_" + self.accessPath + "?trymultiway, _;\n"
                    proctypeDefinition += "\t\tif\n"
                    proctypeDefinition += "\t\t:: d_step { gflagTrymultiway ->\n"
                    proctypeDefinition += "\t\t\tgflagTrymultiway = false;   // Reset after usage => false for next trymultiway and outside trymultiway\n"
                    proctypeDefinition += self._transitionCoreToPromela(sm, ef, dm, "\t\t\t")
                    if self.addPrintfs:
                        proctypeDefinition += "\t\t\tprintf(\"### " + self.accessPath + "(%d) ### takes master transition: " + sm + \
                                              " -[" + tr + " when " + expressionToString(gu) + " then " + expressionsToString(ef) + "]-> " + dm + "\\n\", _pid);\n"
                    proctypeDefinition += "\t\t\tchan_" + self.accessPath + "!invalidate, 0; // Supercomponent expects to retrieve this message! fT by root!\n"
                    proctypeDefinition += "\t\t\tgoto mode_" + dm + ";\n"
                    proctypeDefinition += "\t\t:: !gflagTrymultiway -> fT; // no communication partner => continue with loop (no jump/reentrance of state)\n"
                    proctypeDefinition += "\t\tfi;\n"

                # Master Transition (subcomponent's in event)
                elif (tr.find(".") != -1) and (self._isSubcomponentsInEventPort(tr)):
                    (scname, scidp, scodp, sciep, scoep, ims, mh) = self._getSubcomponentByName(tr)
                    # Send trymultiway only to active control subcomponents (otherwise, the few intermediate steps leading into deadlock can be skipped)
                    if modeName in ims:
                        proctypeDefinition += "\n\t// Master Transition (subcomponent's in event): %s -[%s when %s then %s]-> %s\n" % (sm, tr, expressionToString(gu), expressionsToString(ef), dm)
                        proctypeDefinition += "\t:: nT(" + self._guardToPromela(gu) + ") ->\n"
                        proctypeDefinition += "\t\t// We know: gflagTrymultiway == false, since reset after every trymultiway!\n"
                        proctypeDefinition += "\t\tchan_" + self.accessPath + "_" + scname + "!trymultiway, " + str(self._EventToNumber(tr.split(".")[1], sciep, scoep)) + ";\n"
                        proctypeDefinition += "\t\tchan_" + self.accessPath + "_" + scname + "?trymultiway, _;\n"
                        proctypeDefinition += "\t\tif\n"
                        proctypeDefinition += "\t\t:: d_step { gflagTrymultiway ->\n"
                        proctypeDefinition += "\t\t\tgflagTrymultiway = false;   // Reset after usage => false for next trymultiway and outside trymultiway\n"
                        proctypeDefinition += self._transitionCoreToPromela(sm, ef, dm, "\t\t\t")
                        if self.addPrintfs:
                            proctypeDefinition += "\t\t\tprintf(\"### " + self.accessPath + "(%d) ### takes master transition: " + sm + \
                                                  " -[" + tr + " when " + expressionToString(gu) + " then " + expressionsToString(ef) + "]-> " + dm + "\\n\", _pid);\n"
                        proctypeDefinition += "\t\t\tchan_" + self.accessPath + "!invalidate, 0; // fT by root!\n"
                        proctypeDefinition += "\t\t\tgoto mode_" + dm + ";\n"
                        proctypeDefinition += "\t\t:: !gflagTrymultiway -> fT; // no communication partner => continue with loop (no jump/reentrance of state)\n"
                        proctypeDefinition += "\t\tfi;\n"
                    else:
                        proctypeDefinition += "\n\t// Omitted Master Transition (in event port of deactive subcomponent): %s -[%s when %s then %s]-> %s\n" % (sm, tr, expressionToString(gu), expressionsToString(ef), dm)

                # Reactive Transition => to be treated in "Incomming Message: trymultiway"-section!
                else:
                    proctypeDefinition += "\n\t// No proactive/master transition: %s -[%s when %s then %s]-> %s\n" % (sm, tr, expressionToString(gu), expressionsToString(ef), dm)


            # DEPRECATED: Probabilistic Transitions
            #for (sm, rate, dm) in filter(lambda (sm, rate, dm): sm == modeName, self.probabilistictransitions):
            #    proctypeDefinition += "\n\t// Probabilistic Transition: %s -[prob %s]-> %s\n" % (sm, rate, dm)
            #    self.translator.error_collector.add(Slim2PromelaTranslationError("(WARNING) Probabilictic Transitions are not fully supported - they are transformed into tau-transitions! (in " + self.accessPath + " : " + self.packageName + "::" + self.componentTypeName + "." + self.componentImplementationName + ")"))
            #    proctypeDefinition += "\t// WARNING: Probabilistic Transitions are not fully supported - they are transformed into tau-transitions!\n"
            #    proctypeDefinition += "\t:: nT(true) -> d_step { \n"
            #    proctypeDefinition += self._transitionCoreToPromela(sm, [], dm)
            #    proctypeDefinition += "\t\tchan_" + self.accessPath + "!invalidate, 0;  // propagate changes; fT by root!\n"
            #    proctypeDefinition += "\t\tgoto mode_" + dm + ";\n"
                

            # Incomming Message: trymultiway (from parent)
            for inEventPort in self.in_event_ports:
                proctypeDefinition += "\n\t// Incomming Message: trymultiway (from parent: " + inEventPort + " (" + str(self._EventToNumber(inEventPort, self.in_event_ports, self.out_event_ports)) + ") )\n"

                # Receive incomming trymultiway
                proctypeDefinition += "\t:: chan_" + self.accessPath + "?trymultiway, " + str(self._EventToNumber(inEventPort, self.in_event_ports, self.out_event_ports)) + " ->\n"
                if self.addPrintfs:
                    proctypeDefinition += "\t\tprintf(\"### " + self.accessPath + "(%d) ### receives trymultiway request for " + inEventPort + \
                                          " (No. " + str(self._EventToNumber(inEventPort, self.in_event_ports, self.out_event_ports)) + ")\\n\", _pid);\n"

                # Ask subcomponents via Event Port Connections
                proctypeDefinition += "\n\t\t// Ask subcomponents via Event Port Connections\n"
                forwardingConnections = filter(lambda (sp, dp, ims): sp == inEventPort and self._isSubcomponentsInEventPort(dp) and modeName in ims, self.event_port_connections) 
                for (sp, dp, ims) in forwardingConnections:
                    (scname, scidp, scodp, sciep, scoep, ims, mh) = self._getSubcomponentByName(dp)
                    proctypeDefinition += "\n\t\t\t// Event Port Connection: %s -> %s (and current mode %s is in modes %s)\n" % (sp, dp, modeName, ims)
                    proctypeDefinition += "\t\t\tchan_" + self.accessPath + "_" + scname + "!trymultiway, " + str(self._EventToNumber(dp.split(".")[1], sciep, scoep)) + ";\n"
                    proctypeDefinition += "\t\t\tchan_" + self.accessPath + "_" + scname + "?trymultiway, _;\n"

                # Reactive Transitions (own in event)
                reactiveTransitions = filter(lambda (sm, tr, gu, ef, dm): sm == modeName and tr == inEventPort, self.transitions)
                if len(reactiveTransitions) > 0:
                    proctypeDefinition += "\n\t\t// Reactive Transitions\n"
                    proctypeDefinition += "\t\tif\n";
                    for (sm, tr, gu, ef, dm) in reactiveTransitions:
                        proctypeDefinition += "\n\t\t\t// Reactive Transition (own in event): %s -[%s when %s then %s]-> %s\n" % (sm, tr, expressionToString(gu), expressionsToString(ef), dm)
                        proctypeDefinition += "\t\t\t:: d_step { " + self._guardToPromela(gu) + " ->\n"
                        proctypeDefinition += "\t\t\t\tgflagTrymultiway = true;\n"
                        proctypeDefinition += self._transitionCoreToPromela(sm, ef, dm, "\t\t\t\t")
                        if self.addPrintfs:
                            proctypeDefinition += "\t\t\t\tprintf(\"### " + self.accessPath + "(%d) ### takes reactive transition: " + sm + \
                                                  " -[" + tr + " when " + expressionToString(gu) + " then " + expressionsToString(ef) + "]-> " + dm + "\\n\", _pid);\n"
                        proctypeDefinition += "\t\t\t\t// Answer to trymultiway, no invalidate necessary since sended by Master Transition\n"
                        if self.addPrintfs:
                            proctypeDefinition += "\t\t\t\tprintf(\"### " + self.accessPath + "(%d) ### answers to trymultiway: %d\\n\", _pid, gflagTrymultiway);\n"
                        proctypeDefinition += "\t\t\t\tchan_" + self.accessPath + "!trymultiway, 0;\n"
                        proctypeDefinition += "\t\t\t\tgoto mode_" + dm + ";\n"
                    proctypeDefinition += "\n\t\t\t// No Reactive Transition enabled => just answer to trymultiway\n"
                    proctypeDefinition += "\t\t\t:: else ->\n"
                    if self.addPrintfs:
                        proctypeDefinition += "\t\t\tprintf(\"### " + self.accessPath + "(%d) ### answers to trymultiway: %d\\n\", _pid, gflagTrymultiway);\n"
                    proctypeDefinition += "\t\t\t\tchan_" + self.accessPath + "!trymultiway, 0;\n"
                    proctypeDefinition += "\n\t\tfi;\n"
                
                # Answer incomming trymultiway (if no reactive transition)
                else:
                    proctypeDefinition += "\n\t\t// No Reactive Transitions => just answer to trymultiway\n"
                    if self.addPrintfs:
                        proctypeDefinition += "\t\tprintf(\"### " + self.accessPath + "(%d) ### answers to trymultiway: %d\\n\", _pid, gflagTrymultiway);\n"
                    proctypeDefinition += "\t\tchan_" + self.accessPath + "!trymultiway, 0;\n"


            # Incomming Message: trymultiway (from active subcomponent)
            # HINT: As ErrorEventPorts are stored in <ComponentType>.out_event_ports, this section handles "reactive error transitions" as well.
            #       However, the rate distribution is not considered here.
            for (scname, scidp, scodp, sciep, scoep, ims, mh) in filter(lambda (scname, scidp, scodp, sciep, scoep, ims, mh): modeName in ims, self.subcomponents):
                for outEventPort in scoep:
                    proctypeDefinition += "\n\t// Incomming Message: trymultiway (from active subcomponent " + scname + ": " + outEventPort + " (" + str(self._EventToNumber(outEventPort, sciep, scoep)) + ") )\n"

                    # Receive incomming trymultiway
                    proctypeDefinition += "\t:: chan_" + self.accessPath + "_" + scname + "?trymultiway, " + str(self._EventToNumber(outEventPort, sciep, scoep)) + " ->\n"
                    if self.addPrintfs:
                        proctypeDefinition += "\t\tprintf(\"### " + self.accessPath + "(%d) ### receives trymultiway request for " + scname + "." + outEventPort + \
                                              " (No. " + str(self._EventToNumber(outEventPort, sciep, scoep)) + ")\\n\", _pid);\n"

                    # Ask parent via Event Port Connections
                    forwardingConnections = filter(lambda (sp, dp, ims): sp == scname + "." + outEventPort and (dp in self.out_event_ports) and (modeName in ims), self.event_port_connections) 
                    proctypeDefinition += "\n\t\t// Ask parent via Event Port Connections\n"
                    for (sp, dp, ims) in forwardingConnections:
                        proctypeDefinition += "\n\t\t\t// Event Port Connection: %s -> %s (and current mode %s is in modes %s)\n" % (sp, dp, modeName, ims)
                        proctypeDefinition += "\t\t\tchan_" + self.accessPath + "!trymultiway, " + str(self._EventToNumber(dp, self.in_event_ports, self.out_event_ports)) + ";\n"
                        proctypeDefinition += "\t\t\tchan_" + self.accessPath + "?trymultiway, _;\n"
                        
                    # Ask other subcomponents via Event Port Connections
                    forwardingConnections = filter(lambda (sp, dp, ims): sp == scname + "." + outEventPort and self._isSubcomponentsInEventPort(dp) and (modeName in ims), self.event_port_connections) 
                    proctypeDefinition += "\n\t\t// Ask subcomponents via Event Port Connections\n"
                    for (sp, dp, ims) in forwardingConnections:
                        (scname2, scidp2, scodp2, sciep2, scoep2, ims2, mh2) = self._getSubcomponentByName(dp)
                        proctypeDefinition += "\n\t\t\t// Event Port Connection: %s -> %s (and current mode %s is in modes %s)\n" % (sp, dp, modeName, ims)
                        proctypeDefinition += "\t\t\tchan_" + self.accessPath + "_" + scname2 + "!trymultiway, " + str(self._EventToNumber(dp.split(".")[1], sciep2, scoep2)) + ";\n"
                        proctypeDefinition += "\t\t\tchan_" + self.accessPath + "_" + scname2 + "?trymultiway, _;\n"

                    # Reactive Transitions (subcomponent's out event)
                    reactiveTransitions = filter(lambda (sm, tr, gu, ef, dm): sm == modeName and tr == scname + "." + outEventPort, self.transitions)
                    if len(reactiveTransitions) > 0:
                        proctypeDefinition += "\n\t\t// Reactive Transitions\n"
                        proctypeDefinition += "\t\tif\n";
                        for (sm, tr, gu, ef, dm) in reactiveTransitions:
                            proctypeDefinition += "\n\t\t\t// Reactive Transition (subcomponent's out event): %s -[%s when %s then %s]-> %s\n" % (sm, tr, expressionToString(gu), expressionsToString(ef), dm)
                            proctypeDefinition += "\t\t\t:: d_step { " + self._guardToPromela(gu) + " ->\n"
                            proctypeDefinition += "\t\t\t\tgflagTrymultiway = true;\n"
                            proctypeDefinition += self._transitionCoreToPromela(sm, ef, dm, "\t\t\t\t", adjustSubcomponentsActivation = False)
                            if self.addPrintfs:
                                proctypeDefinition += "\t\t\tprintf(\"### " + self.accessPath + "(%d) ### takes reactive transition: " + sm + \
                                                      " -[" + tr + " when " + expressionToString(gu) + " then " + expressionsToString(ef) + "]-> " + dm + "\\n\", _pid);\n"
                            proctypeDefinition += "\t\t\t\t// Answer to trymultiway\n"
                            if self.addPrintfs:
                                proctypeDefinition += "\t\t\t\tprintf(\"### " + self.accessPath + "(%d) ### answers to trymultiway: %d\\n\", _pid, gflagTrymultiway);\n"
                            proctypeDefinition += "\t\t\t\tchan_" + self.accessPath + "_" + scname + "!trymultiway, 0;\n"
                            # IMPORTANT: Catch invalidate of to-be-deactivated subcomponent before it is deactivated!
                            proctypeDefinition += "\t\t\t\t// Receive invalidate from subcomponent containing the master transition (before possible deactivating it!)\n"
                            proctypeDefinition += "\t\t\t\tchan_" + self.accessPath + "_" + scname + "?invalidate, _;\n"
                            proctypeDefinition += "\t\t\t\t// Adjust Subcomponents' Activation (now it is even safe to deactivate the subcomponent containing the master)\n"
                            proctypeDefinition += self._adjustSubcomponentsActivation(sm, dm, "\t\t\t\t")
                            proctypeDefinition += "\t\t\t\t// Forward invalidate\n"
                            proctypeDefinition += "\t\t\t\tchan_" + self.accessPath + "!invalidate, 0;\n"
                            proctypeDefinition += "\t\t\t\tgoto mode_" + dm + ";\n"
                        proctypeDefinition += "\n\t\t\t// No Reactive Transition enabled => just answer to trymultiway\n"
                        proctypeDefinition += "\t\t\t:: else ->\n"
                        if self.addPrintfs:
                            proctypeDefinition += "\t\t\t\tprintf(\"### " + self.accessPath + "(%d) ### answers to trymultiway: %d\\n\", _pid, gflagTrymultiway);\n"
                        proctypeDefinition += "\t\t\t\tchan_" + self.accessPath + "_" + scname + "!trymultiway, 0;\n"
                        proctypeDefinition += "\n\t\tfi;\n"
                    
                    # Answer incomming trymultiway (if no reactive transition)
                    else:
                        proctypeDefinition += "\n\t\t// No Reactive Transitions => just answer to trymultiway\n"
                        if self.addPrintfs:
                            proctypeDefinition += "\t\tprintf(\"### " + self.accessPath + "(%d) ### answers to trymultiway: %d\\n\", _pid, gflagTrymultiway);\n"
                        proctypeDefinition += "\t\tchan_" + self.accessPath + "_" + scname + "!trymultiway, 0;\n"


            # Incomming Message: update (from parent)
            proctypeDefinition += "\n\t// Incomming Message: update (from parent)\n"
            proctypeDefinition += "\t:: chan_" + self.accessPath + "?update, _ ->\n"

            # - Display text message: input parameters
            if self.addPrintfs: 
                proctypeDefinition += "\n\t\t// Display text message: input parameters\n"
                proctypeDefinition += "\t\tprintf(\"### " + self.accessPath + "(%d) ### update with new input values: "
                proctypeDefinition += ", ".join([name+"=%d" for (name, typ, default) in self.in_data_ports])
                proctypeDefinition += "\\n\", _pid"
                proctypeDefinition += "".join([", idpStore_" + self.accessPath + "_" + name for (name, typ, default) in self.in_data_ports])
                proctypeDefinition += ");\n"

            # - Update Subcomponents, if there are any active
            activeSubcomponents = filter(lambda (scname, scidp, scodp, sciep, scoep, ims, mh): modeName in ims, self.subcomponents)
            if len(activeSubcomponents) > 0:
                
                # -- First Try
                proctypeDefinition += "\n\t\t// Update active subcomponents: First Iteration\n"
                proctypeDefinition += "\t\tinlineUpdateMode" + modeName + "_" + self.accessPath + "();\n"

                # -- Prepare: Is Fixpoint Iteration required? (exists data port connection or flow from subcomponent to (same or another) subcomponent?)
                dataPortConnectionsBetweenSubcomponents = filter(lambda (sp, dp, ims): sp.find(".") >= 0 and dp.find(".") >= 0 and (modeName in ims), self.data_port_connections)
                dataFlowsBetweenSubcomponents = filter(lambda (expr, ims): expressionWritesTo(expr).find(".") >= 0 and \
                                                                           checkExistence(lambda sp: sp.find(".") >= 0, expressionReadsFrom(expr)) and \
                                                                           (modeName in ims)
                                                       , self.flows)
                fixpointRelevantSubcomponentsOutDataPorts = [sp for (sp, dp, ims) in dataPortConnectionsBetweenSubcomponents]
                for (expr, ims) in dataFlowsBetweenSubcomponents:
                    fixpointRelevantSubcomponentsOutDataPorts.extend(filter(lambda sp: sp.find(".") >= 0, expressionReadsFrom(expr)))
                fixpointIterationIsRequired = len(fixpointRelevantSubcomponentsOutDataPorts) > 0
                
                # -- Prepare: Fixpoint Copy (needed anyways, not empty when active subcomponent with out data port exists, thus at least when fixpoint iteration is required)
                doNonFixpointCopy = ""
                for (scname, scidp, scodp, sciep, scoep, ims, mh) in activeSubcomponents:
                    for (name, typ, default) in scodp:
                        # HINT: Direct equality-check and copy of PROMELA ints which contain reals works. Thus, c_code (as done by _expressionToPromela) not necessary!
                        doNonFixpointCopy  += "odpStore_" + self.accessPath + "_" + scname + "_" + name +  " = odpNew_" + self.accessPath + "_" + scname + "_" + name + "; "

                # -- Prepare: Fixpoint Check (not empty when fixpoint iteration required)
                doNonFixpointCheck = ""
                for sp in fixpointRelevantSubcomponentsOutDataPorts:
                    # HINT: Direct equality-check and copy of PROMELA ints which contain reals works. Thus, c_code (as done by _expressionToPromela) not necessary!
                    doNonFixpointCheck += "odpStore_" + self.accessPath + "_" + sp.replace(".", "_") + " == odpNew_" + self.accessPath + "_" + sp.replace(".", "_") + " && "
                doNonFixpointCheck = doNonFixpointCheck[:-4]    # Slicing with -4 removes last " && "
                
                # -- Fixpoint Iteration   OR   just copy once
                if fixpointIterationIsRequired:
                    proctypeDefinition += "\n\t\t// Update active subcomponents: Fixpoint Iteration over subcomponents' out data ports forwarded to subcomponents' in data ports\n"
                    proctypeDefinition += "\t\tdo\n"
                    proctypeDefinition += "\t\t:: d_step { !(" + doNonFixpointCheck + ") ->\n"
                    proctypeDefinition += "\t\t\t" + doNonFixpointCopy + "}\n"
                    proctypeDefinition += "\t\t\tinlineUpdateMode" + modeName + "_" + self.accessPath + "()\n"
                    proctypeDefinition += "\t\t:: d_step { (" + doNonFixpointCheck + ") ->\n"
                    proctypeDefinition += "\t\t\t" + doNonFixpointCopy + "}\n"
                    proctypeDefinition += "\t\t\tbreak\n"
                    proctypeDefinition += "\t\tod;\n"
                else:
                    proctypeDefinition += "\n\t\t// Update active subcomponents: No Fixpoint Iteration needed (since no data port connection/flow between subcomponents), just copy buffered values to store\n"
                    if doNonFixpointCopy: proctypeDefinition += "\t\td_step { " + doNonFixpointCopy + " }\n"

                # -- Mode dependend internal update (inline function)
                inlineDefinitions += "inline inlineUpdateMode" + modeName + "_" + self.accessPath + "()\n"
                inlineDefinitions += "{\n\n"

                for (scname, scidp, scodp, sciep, scoep, ims, mh) in activeSubcomponents:
                    inlineDefinitions += "\t// Update active subcomponent " + scname + "\n"

                    # Forward new inputs to subcomponents in data ports
                    inlineDefinitions += "\t// 1.) Forward new inputs to subcomponents in data ports\n"
                    # a) (Collect) Via Data Port Connections to subcomponents in data port
                    forwardingConnections = filter(lambda (sp, dp, ims): dp.find(".") >= 0 and dp.split(".")[0] == scname and (modeName in ims), self.data_port_connections)
                    # b) (Collect) Via Flows to subcomponents in data port
                    forwardingFlows = filter(lambda (expr, ims): self._splitEffectToAssignment(expr)[2] and \
                                                                 self._isSubcomponentsInDataPort(self._splitEffectToAssignment(expr)[0]) and \
                                                                 (self._splitEffectToAssignment(expr)[0]).split(".")[0] == scname and \
                                                                 (modeName in ims)
                                             , self.flows)
                    # -> Generate output for a) & b), if not empty
                    if len(forwardingConnections) + len(forwardingFlows) > 0:
                        inlineDefinitions += "\td_step {\n"
                        # a) (Output) Via Data Port Connections to subcomponents in data port
                        inlineDefinitions += "\t\t// 1.a) Via Data Port Connections to subcomponents in data port\n"
                        for (sp, dp, ims) in forwardingConnections:
                            # HINT: Direct copy of PROMELA ints which contain reals works. Thus, c_code (as done by _assignmentToPromela) not necessary!
                            inlineDefinitions += "\t\t" + self._identifierToPromela(dp) + " = " + self._identifierToPromela(sp) + ";\n"
                        # b) (Output) Via Flows to subcomponents in data port
                        inlineDefinitions += "\t\t// 1.b) Via Flows to subcomponents in data port\n"
                        for (expr, ims) in forwardingFlows:
                            inlineDefinitions += self._effectsToPromela([expr])
                        inlineDefinitions += "\t}\n"

                    # Send update message to subcomponent & receive acknowledgement
                    # (Necessary even if no new inputs are forwarded, since inputs could have changed due to (de)activation of connections, flows, components.)
                    inlineDefinitions += "\t// 2.) Send update message to subcomponent & receive acknowledgement\n"
                    inlineDefinitions += "\tchan_" + self.accessPath + "_" + scname + "!update, 0;\n"
                    inlineDefinitions += "\tchan_" + self.accessPath + "_" + scname + "?update, _;\n\n"
                
                inlineDefinitions += "}\n\n"

            else: # No active subcomponents
                proctypeDefinition += "\n\t\t// No active subcomponents to update\n"

            # - Update own out data ports
            proctypeDefinition += "\n\t\t// Update own out data ports\n"

            forwardingConnections = filter(lambda (sp, dp, ims): self._isOwnOutDataPort(dp) and (modeName in ims), self.data_port_connections) # a)
            forwardingFlows = filter(lambda (expr, ims): self._splitEffectToAssignment(expr)[2] and \
                                                         self._isOwnOutDataPort(self._splitEffectToAssignment(expr)[0]) and \
                                                        (modeName in ims)
                                     , self.flows) # b)

            if len(forwardingConnections) + len(forwardingFlows) > 0:
                proctypeDefinition += "\t\tskip; // skip necessary as break cannot jump to d_step\n"
                proctypeDefinition += "\t\td_step {\n"
            
                # -- a) Via Data Port Connections to own out data port (from subcomponents' out data ports)
                proctypeDefinition += "\t\t// a) Via Data Port Connections to own out data port (from subcomponents' out data ports)\n"
                for (sp, dp, ims) in forwardingConnections:
                    # HINT: Direct copy of PROMELA ints which contain reals works. Thus, c_code (as done by _assignmentToPromela) not necessary!
                    proctypeDefinition += "\t\t" + self._identifierToPromela(dp, True) + " = " + self._identifierToPromela(sp) + ";\n"

                # -- b) Via Flows to own out data port (from own in data ports and subcomponents' out data ports)
                proctypeDefinition += "\t\t// b) Via Flows to own out data port (from own in data ports and subcomponents' out data ports)\n"
                for (expr, ims) in forwardingFlows:
                    proctypeDefinition += self._effectsToPromela([expr])

                proctypeDefinition += "\t\t}\n"

            # - Display text message: output parameters
            if self.addPrintfs: 
                proctypeDefinition += "\n\t\t// Display text message: output parameters\n"
                proctypeDefinition += "\t\tprintf(\"### " + self.accessPath + "(%d) ### update finally lead to new output values: "
                proctypeDefinition += ", ".join([name+"=%d" for (name, typ, default) in self.out_data_ports])
                proctypeDefinition += "\\n\", _pid"
                proctypeDefinition += "".join([", odpNew_" + self.accessPath + "_" + name for (name, typ, default) in self.out_data_ports])
                proctypeDefinition += ");\n"

            # - Send update acknowledge to parent
            proctypeDefinition += "\n\t\t// Send update acknowledge to parent\n"
            proctypeDefinition += "\t\tchan_" + self.accessPath + "!update, 0;\n"
            
            # --- /update ---

                    
            # Incomming Message: invalidate (from active subcomponents only)
            proctypeDefinition += "\n\t// Incomming Message: invalidate (from subcomponents only)\n"
            for (scname, scidp, scodp, sciep, scoep, ims, mh) in filter(lambda (scname, scidp, scodp, sciep, scoep, ims, mh): modeName in ims, self.subcomponents):
                proctypeDefinition += "\t:: chan_" + self.accessPath + "_" + scname + "?invalidate, _ ->\n"
                proctypeDefinition += "\t\tchan_" + self.accessPath + "!invalidate, 0;\n"

                    
            # Incomming Message: deactivate (from parent)
            if self.modeHistory:
                proctypeDefinition += "\n\t// Incomming Message: deactivate (from parent) - keep mode history\n"
                
                proctypeDefinition += "\t:: chan_" + self.accessPath + "?deactivate, _ ->\n"
                proctypeDefinition += "\t\t// deactivate active subcomponents\n"
                for (scname, scidp, scodp, sciep, scoep, ims, mh) in filter(lambda (scname, scidp, scodp, sciep, scoep, ims, mh): modeName in ims, self.subcomponents):
                    proctypeDefinition += "\t\tchan_" + self.accessPath + "_" + scname + "!deactivate, 0;\n"
                    proctypeDefinition += "\t\tchan_" + self.accessPath + "_" + scname + "?deactivate, _;\n"
                if self.addPrintfs: proctypeDefinition += "\t\tprintf(\"### " + self.accessPath + "(%d) ### was deactivated (with mode history)\\n\", _pid);\n"
                proctypeDefinition += "\t\tchan_" + self.accessPath + "!deactivate, 0;\n"
                
                proctypeDefinition += "\t\t// reactivate\n"
                proctypeDefinition += "\t\tchan_" + self.accessPath + "?activate, _;\n"
                for (scname, scidp, scodp, sciep, scoep, ims, mh) in filter(lambda (scname, scidp, scodp, sciep, scoep, ims, mh): modeName in ims, self.subcomponents):
                    proctypeDefinition += "\t\tchan_" + self.accessPath + "_" + scname + "!activate, 0;\n"
                    proctypeDefinition += "\t\tchan_" + self.accessPath + "_" + scname + "?activate, _;\n"
                if self.addPrintfs: proctypeDefinition += "\t\tprintf(\"### " + self.accessPath + "(%d) ### was reactivated (with mode history)\\n\", _pid);\n"
                proctypeDefinition += "\t\tchan_" + self.accessPath + "!activate, 0;\n"
                
            else:
                proctypeDefinition += "\n\t// Incomming Message: deactivate (from parent) - reset mode history\n"       
                proctypeDefinition += "\t:: chan_" + self.accessPath + "?deactivate, _ ->\n"

                proctypeDefinition += "\t\t// reset active subcomponents\n"
                for (scname, scidp, scodp, sciep, scoep, ims, mh) in filter(lambda (scname, scidp, scodp, sciep, scoep, ims, mh): modeName in ims, self.subcomponents):
                    proctypeDefinition += "\t\tchan_" + self.accessPath + "_" + scname + "!reset, 0;\n"
                    proctypeDefinition += "\t\tchan_" + self.accessPath + "_" + scname + "?reset, _;\n"

                proctypeDefinition += "\t\t// activate already deactivated subcomponents to reset them (only those with mode history)\n"
                for (scname, scidp, scodp, sciep, scoep, ims, mh) in filter(lambda (scname, scidp, scodp, sciep, scoep, ims, mh): (modeName not in ims) and mh, self.subcomponents):
                    proctypeDefinition += "\t\tchan_" + self.accessPath + "_" + scname + "!activate, 0;\n"
                    proctypeDefinition += "\t\tchan_" + self.accessPath + "_" + scname + "?activate, _;\n"
                    proctypeDefinition += "\t\tchan_" + self.accessPath + "_" + scname + "!reset, 0;\n"
                    proctypeDefinition += "\t\tchan_" + self.accessPath + "_" + scname + "?reset, _;\n"

                proctypeDefinition += "\t\t// confirm deactivation\n"
                if self.addPrintfs: proctypeDefinition += "\t\tprintf(\"### " + self.accessPath + "(%d) ### was deactivated (without mode history)\\n\", _pid);\n"
                proctypeDefinition += "\t\tchan_" + self.accessPath + "!deactivate, 0;\n"
                proctypeDefinition += "\t\tgoto FirstActivationOrReactivationWithoutModeHistory;\n"
                
            
            # Incomming Message: reset (from parent)   ---   nearly identical to deactivate, but mode history always gets lost
            proctypeDefinition += "\n\t// Incomming Message: reset (from parent)\n"       
            proctypeDefinition += "\t:: chan_" + self.accessPath + "?reset, _ ->\n"
            
            proctypeDefinition += "\t\t// reset active subcomponents\n"
            for (scname, scidp, scodp, sciep, scoep, ims, mh) in filter(lambda (scname, scidp, scodp, sciep, scoep, ims, mh): modeName in ims, self.subcomponents):
                proctypeDefinition += "\t\tchan_" + self.accessPath + "_" + scname + "!reset, 0;\n"
                proctypeDefinition += "\t\tchan_" + self.accessPath + "_" + scname + "?reset, _;\n"

            proctypeDefinition += "\t\t// activate already deactivated subcomponents to reset them (only those with mode history)\n"
            for (scname, scidp, scodp, sciep, scoep, ims, mh) in filter(lambda (scname, scidp, scodp, sciep, scoep, ims, mh): (modeName not in ims) and mh, self.subcomponents):
                proctypeDefinition += "\t\tchan_" + self.accessPath + "_" + scname + "!activate, 0;\n"
                proctypeDefinition += "\t\tchan_" + self.accessPath + "_" + scname + "?activate, _;\n"
                proctypeDefinition += "\t\tchan_" + self.accessPath + "_" + scname + "!reset, 0;\n"
                proctypeDefinition += "\t\tchan_" + self.accessPath + "_" + scname + "?reset, _;\n"

            proctypeDefinition += "\t\t// confirm reset\n"
            if self.addPrintfs: proctypeDefinition += "\t\tprintf(\"### " + self.accessPath + "(%d) ### is deactivated and forced to be reset\\n\", _pid);\n"
            proctypeDefinition += "\t\tchan_" + self.accessPath + "!reset, 0;\n"
            proctypeDefinition += "\t\tgoto FirstActivationOrReactivationWithoutModeHistory;\n"


            # End of mode's do-loop
            proctypeDefinition += "\n\tod;\n\t// --- /Mode: " + modeName + " ---\n"

        # </for (modeName, modeType) in self.modes: ... "do-loop" ...>



        # *** Return completed proctype definition ***
        return (globalVariableDeclarations, inlineDefinitions + "\n" + proctypeDefinition + "\n}\n")
    # </def toPromela()>:

    
# ==================================================================================================
# </class ProctypeInfo: ...>
# ==================================================================================================










# ==================================================================================================
# class Slim2Promela_Translator
# ==================================================================================================

class Slim2Promela_Translator:
    """This class contains the framework for translating a SLIM specification to Promela.

    Usage: TODO
    """



    # ----------------------------------------------------------------------------------------------



    def __init__(self, rootComponent, atomicPropositions, verbose = False, addPrintfs = False):
        """Initializes a new instance of Slim2Promela_Translator."""

        self.theRootComponent = rootComponent            # Stores the root component (Implementation, from a CompileInfo data structure)
        self.theAtomicPropositions = atomicPropositions  # Stores a dictionary of placeholderNames and atomicPropositions
        self.verbose = verbose
        self.addPrintfs = addPrintfs
        
        self.proctypeCount = 0           # Counter for proctypes (max. 255 possible).
        self.enumNames = {}              # Stores all enum names found in the SLIM input together with an unique integer ID.
        self.modeNames = {}              # Stores all mode names found in the SLIM input together with an unique integer ID.
        self.globalStateVariables = {}   # dict(displayName: string, promelaVariable: string)
        
        self.error_collector = slim_compiler.src.errorManagement.errors.ErrorCollector()
                                         # This ErrorCollector collects all errors which occure during translation itself.
                                         
        self.translationResult = ""      # Stores the resulting Promela code.



    # ----------------------------------------------------------------------------------------------



    def _getNextProctypeNumber(self):
        self.proctypeCount += 1
        return self.proctypeCount
    

    def _addEnumNames(self, aList):
        """Adds enum names to the list of enum names stored in the translator (self.enumNames) and assigns a unique integer identifier to them.
        Enum names which are already in the list are not readded, but ignored.
        The "global" list of enum names is necessary to define integer constants for them in the Promela header which then are used to represend enum values.

        @type aList:  list(string)
        @param aList: each item is one enum name to be added
        """
        
        for enumName in aList:
            if enumName not in self.enumNames:
                self.enumNames.update({enumName: len(self.enumNames)})


    def _addModeNames(self, aList):
        """Adds mode names to the list of mode names stored in the translator (self.modeNames) and assigns a unique integer identifier to them.
        Mode names which are already in the list are not readded, but ignored.
        The "global" list of mode names is necessary to define integer constants for them in the Promela header which then are used to store modes in variables.

        @type aList:  list(string)
        @param aList: each item is one mode name to be added
        """

        for modeName in aList:
            if modeName not in self.modeNames:
                self.modeNames.update({modeName: len(self.modeNames)})


    def _addGlobalStateVariable(self, displayName, promelaVariable):
        self.globalStateVariables.update({ displayName.replace("_", ".") : promelaVariable })



    # ----------------------------------------------------------------------------------------------



    def _accessPathFragmentToPromelaVariableFragment(self, accessPath, componentImplementation):
        """TODO: Describe.
        @return: Tuple (variableNameFragment: string, typ: constant defined in XXX)
        """
        
        # A) Descend to next non-data subcomponent
        if accessPath.find(".") >= 0:
            # Split identifier of next non-data subcomponent in accessPath off
            (scName, accessPathRest) = accessPath.split(".", 1)
            # Retriev subcomponent's implementation
            if scName in componentImplementation.non_data_subcomponents:
                scImplementation = componentImplementation.non_data_subcomponents[scName].get_referenced_component_implementation()
            else:
                return ("// ERR: Subcomponent '" + scName + "' does not exist!", UNKNOWN, False)
            # Recursively proceed translation
            (variableNameFragment, typ, isReal) = self._accessPathFragmentToPromelaVariableFragment(accessPathRest, scImplementation)
            # Return combined result
            return ("_" + scName + variableNameFragment, typ, isReal)
        
        # B) Reached last entry of accessPath
        # B.1) nominal mode
        elif accessPath in ["mode", "error"]:
            # HINT: *.error has been translated to *._errorSubcomponent.error, so we do not have to distinguish!
            # HINT: We don't have to ask whether identifierExpressionTreeNode.node_type in [NodeType.MODE_FIELD, NodeType.ERROR_FIELD]. The textual suffix suffices!
            return ("", NOMINALMODE, False)
        # B.3) data element (data port or data subcomponent)
        else:
            # We also need to consider the component type
            componentType = componentImplementation.get_referenced_component_type()
            # Find out which type
            if accessPath in componentImplementation.data_subcomponents:
                typ = DATASUBCOMPONENT
                isReal = isinstance(componentImplementation.data_subcomponents[accessPath].data_type, Real)
            elif accessPath in componentType.in_data_ports:
                typ = INDATAPORT
                isReal = isinstance(componentType.in_data_ports[accessPath].data_type, Real)
            elif accessPath in componentType.out_data_ports:
                typ = OUTDATAPORT
                isReal = isinstance(componentType.out_data_ports[accessPath].data_type, Real)
            else:
                typ = UNKNOWN
                isReal = False
            # Return result
            return ("_" + accessPath, typ, isReal)


    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


    def _accessPathToPromelaVariable(self, accessPath):
        """TODO: Describe.
        @return: Tuple (variableNameFragment: string, typ: constant defined in XXX)
        """

        (variableNameFragment, typ, isReal) = self._accessPathFragmentToPromelaVariableFragment(accessPath, self.theRootComponent)
        if typ == DATASUBCOMPONENT:
            return "dscStore_root" + variableNameFragment
        elif typ == INDATAPORT:
            return "idpStore_root" + variableNameFragment
        elif typ == OUTDATAPORT:
            return "odpNew_root" + variableNameFragment
        elif typ == NOMINALMODE:
            # HINT: This applies for error states as well.
            return "nominalMode_root" + variableNameFragment
        else:
            return "// ERR: Erronous element \"" + variableNameFragment + "\" "


    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


    def _isAccessPathOfTypeReal(self, accessPath):
        (variableNameFragment, typ, isReal) = self._accessPathFragmentToPromelaVariableFragment(accessPath, self.theRootComponent)
        return isReal


    def _doesAtomicPropositionRequireEmbeddedCCode(self, anExpressionTreeNode):
        if anExpressionTreeNode == None:
            return None
            
        elif isinstance(anExpressionTreeNode, BinaryOperatorTreeNode):
            # For boolean opearators (and, iff, imp, or, xnor, xor): always false
            if anExpressionTreeNode.operator in [OperatorType.AND, OperatorType.IFF, OperatorType.IMP, OperatorType.OR, OperatorType.XNOR, OperatorType.XOR]:
                return False
            # For all others (+, -, *, ...; =, !=, <, ...): check subexpressions
            else:
                return self._doesAtomicPropositionRequireEmbeddedCCode(anExpressionTreeNode.left) or \
                       self._doesAtomicPropositionRequireEmbeddedCCode(anExpressionTreeNode.right)
                
        elif isinstance(anExpressionTreeNode, UnaryOperatorTreeNode):
            return self._doesAtomicPropositionRequireEmbeddedCCode(anExpressionTreeNode.subexpression)
            
        elif isinstance(anExpressionTreeNode, IdentifierTreeNode):
            if not anExpressionTreeNode.node_type in [NodeType.ENUM_VALUE, NodeType.MODE_VALUE, NodeType.ERROR_VALUE]:
                return self._isAccessPathOfTypeReal(anExpressionTreeNode.name)
            
        elif isinstance(anExpressionTreeNode, NumberTreeNode):
            return anExpressionTreeNode.node_type == NodeType.REAL

        elif isinstance(anExpressionTreeNode, CaseExpressionTreeNode):
            return reduce(lambda x, y: x or y,
                          map(lambda (b, e): self._doesAtomicPropositionRequireEmbeddedCCode(e), anExpressionTreeNode.cases),
                          self._doesAtomicPropositionRequireEmbeddedCCode(anExpressionTreeNode.default_expression))
            
        elif isinstance(anExpressionTreeNode, BooleanTreeNode) or isinstance(anExpressionTreeNode, CaseExpressionTreeNode) or \
             isinstance(anExpressionTreeNode, FlowEquationTreeNode) or isinstance(anExpressionTreeNode, FlowExpressionTreeNode):
            return False
            
        return None


    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


    def _atomicPropositionToPromela(self, anExpressionTreeNode, asEmbeddedCCode = False):
        """TODO: Describe.
        @return: ...
        """
        
        # Similar to _expressionToPromela, BUT:
        # - The case handling IdentifierTreeNode is different!
        #     - Idea: Try to join them: _expressionToSPIN(..., identifierHandler) ~> call identifierHandler(anExpressionTreeNode.name)
        # . "myTranslator" used instead of "self"
        # . Name in recursive calls changed
        # - doesExpressionRequireEmbeddedCCode nutzt _isAccessPathOfTypeReal

        
        # First check, whether embedded C-Code has to be opened for subexpression (which must be a comparison)
        if (not asEmbeddedCCode) and self._doesAtomicPropositionRequireEmbeddedCCode(anExpressionTreeNode):
            return "c_expr { " + self._atomicPropositionToPromela(anExpressionTreeNode, True) + " }"

            
        # Distinguish the different Node Types and generate appropriate Promela expressions
        if isinstance(anExpressionTreeNode, BinaryOperatorTreeNode):
            left  = "(" + self._atomicPropositionToPromela(anExpressionTreeNode.left, asEmbeddedCCode)  + ")"
            right = "(" + self._atomicPropositionToPromela(anExpressionTreeNode.right, asEmbeddedCCode) + ")"

            # SLIM      OperatorType       Promela (int)C (real)
            # ----      ------------       ----         -
            # +         .PLUS              +            +
            # -b        .MINUS             -            -
            # *         .TIMES             *            *
            # /         .DIVIDE            /            /
            # mod       .MOD             ! %          ! %
            #
            # =         .EQUAL           ! ==           ==
            # !=        .NOTEQUAL          !=           !=
            # <         .LESS              <            <
            # >         .GREATER           >            >
            # <=        .LESSOREQUAL       <=           <=
            # >=        .GREATEROREQUAL    >=           >=
            #
            # and       .AND             ! &&         ! &&
            # iff       .IFF            !! ...       !! ...
            # imp       .IMP            !! ...       !! ...
            # or        .OR              ! ||         ! ||
            # xnor      .XNOR           !! ...       !! ...
            # xor       .XOR            !! ...       !! ...
            if anExpressionTreeNode.operator == OperatorType.MOD:
                txt = left + " % " + right
            elif anExpressionTreeNode.operator == OperatorType.EQUAL:
                txt = left + " == " + right
            elif anExpressionTreeNode.operator == OperatorType.AND:
                txt = left + " && " + right
            elif anExpressionTreeNode.operator == OperatorType.OR:
                txt = left + " || " + right
            elif anExpressionTreeNode.operator in [OperatorType.IFF, OperatorType.XNOR]:
                # a <=> b   <---->   (a and b) or (!a and !b)  /  (a => b) and (b => a)  <->  (b or !a) and (a or !b)
                # a XNOR b   <---->   a <=> b   <---->   ...
                txt = "(" + left + " && " + right + ") or (!" + left + " && !" + right + ")"
            elif anExpressionTreeNode.operator == OperatorType.IMP:
                # a => b   <---->   b or !a
                txt = right + " || !" + left
            elif anExpressionTreeNode.operator == OperatorType.XOR:
                # a XOR b   <---->   (a and !b) or (!a and b)
                txt = "(" + left + " && !" + right + ") || (!" + left + " && " + right + ")"
            elif anExpressionTreeNode.operator in [OperatorType.PLUS, OperatorType.MINUS, OperatorType.TIMES, OperatorType.DIVIDE, OperatorType.NOTEQUAL,
                                                   OperatorType.LESS, OperatorType.GREATER, OperatorType.LESSOREQUAL, OperatorType.GREATEROREQUAL]:
                txt = left + " " + OperatorType.operator_to_str(anExpressionTreeNode.operator) + " " + right
            else:
                txt = left + "ERR: '" + OperatorType.operator_to_str(anExpressionTreeNode.operator) + "' unsupported as binary operator" + right # TODO: Add error to error_collector

        elif isinstance(anExpressionTreeNode, UnaryOperatorTreeNode):
            right = "(" + self._atomicPropositionToPromela(anExpressionTreeNode.subexpression, asEmbeddedCCode) + ")"
            
            # SLIM      OperatorType       Promela (int)
            # ----      ------------       ----
            # -u        .UNARY_MINUS       -
            # not       .NOT             ! !
            if anExpressionTreeNode.operator == OperatorType.UNARY_MINUS:
                txt = "-" + right
            elif anExpressionTreeNode.operator == OperatorType.NOT:
                txt = "!" + right
            else:
                txt = "ERR: '" + OperatorType.operator_to_str(anExpressionTreeNode.operator) + "' unsupported as unary operator" + right # TODO: Add error to error_collector

        elif isinstance(anExpressionTreeNode, IdentifierTreeNode):
   
            # A) Enum Constant (was: "enum:xxx")
            if anExpressionTreeNode.node_type == NodeType.ENUM_VALUE:
                if anExpressionTreeNode.name in self.enumNames:
                    return str(self.enumNames[anExpressionTreeNode.name]) # HINT: defined names "enum_xxx" cannot be used in #define <AP> as CPP does not replace them in there.
                else:
                    return "ERR: undefined enum identifier '" + anExpressionTreeNode.name + "'" # TODO: Add error

            # B) Mode/Error Constant (was: "mode:xxx" or "error:xxx")
            elif anExpressionTreeNode.node_type in [NodeType.MODE_VALUE, NodeType.ERROR_VALUE]:
                if anExpressionTreeNode.name in self.modeNames:
                    return str(self.modeNames[anExpressionTreeNode.name]) # HINT: defined names "modeId_xxx" cannot be used in #define <AP> as CPP does not replace them in there.
                else:
                    return "ERR: undefined mode/error state identifier '" + anExpressionTreeNode.name + "'" # TODO: Add error

            # C) Access path 
            else:
                if asEmbeddedCCode:
                    if self._isAccessPathOfTypeReal(anExpressionTreeNode.name):
                        # Typecast-Trick for abusing Promela-ints as floats:   *((float *)&(now.myGlobalIntVar))
                        txt = "*((float *)&(now." + self._accessPathToPromelaVariable(anExpressionTreeNode.name) + "))"
                    else:
                        txt = "now." + self._accessPathToPromelaVariable(anExpressionTreeNode.name)
                else:
                    txt = self._accessPathToPromelaVariable(anExpressionTreeNode.name)

        elif isinstance(anExpressionTreeNode, BooleanTreeNode):
            txt = str(anExpressionTreeNode.value).lower()

        elif isinstance(anExpressionTreeNode, NumberTreeNode):
            if anExpressionTreeNode.node_type == NodeType.REAL:
                if asEmbeddedCCode:
                    txt = str(anExpressionTreeNode.number)
                    if anExpressionTreeNode.get_type() == NodeType.REAL_CONST:
                        txt += "f"
                    elif anExpressionTreeNode.get_type() == NodeType.INT_CONST:
                        txt += ".0f"
                    else:
                        self.translator.error_collector.add(Slim2PromelaTranslationError("Value '" + str(anExpressionTreeNode.number) + "' cannot be used in expression of type real. (in " + self.accessPath + " : " + self.packageName + "::" + self.componentTypeName + "." + self.componentImplementationName + ")"))
                        txt = "ERROR: Value '" + str(anExpressionTreeNode.number) + "' cannot be used in expression of type real"
                else:
                    self.translator.error_collector.add(Slim2PromelaTranslationError("Value '" + str(anExpressionTreeNode.number) + "' of type real can be used in embedded C-code only. (in " + self.accessPath + " : " + self.packageName + "::" + self.componentTypeName + "." + self.componentImplementationName + ")"))
                    txt = "ERROR: Value '" + str(anExpressionTreeNode.number) + "' of type real can be used in embedded C-code only"
            else:
                txt = str(anExpressionTreeNode.number)

        elif isinstance(anExpressionTreeNode, CaseExpressionTreeNode):
            if asEmbeddedCCode:
                sign = " ? "
            else:
                sign = " -> "
            
            # Translate first case
            txt = "(" + self._atomicPropositionToPromela(anExpressionTreeNode.cases[0][0], asEmbeddedCCode)
            txt += sign + self._atomicPropositionToPromela(anExpressionTreeNode.cases[0][1], asEmbeddedCCode) + " : "
            # Only one case => translate "otherwise"
            if len(anExpressionTreeNode.cases) == 1:
                txt += self._atomicPropositionToPromela(anExpressionTreeNode.default_expression, asEmbeddedCCode)
            # Not yet "otherwise" => translate the remaining cases
            else:
                smallerCaseExpression = copy.copy(anExpressionTreeNode)
                smallerCaseExpression.cases = smallerCaseExpression.cases[1:] # Slice to ignore previously translated first case
                txt = self._atomicPropositionToPromela(smallerCaseExpression, asEmbeddedCCode)
            txt += ")"

        elif isinstance(anExpressionTreeNode, FlowEquationTreeNode) or isinstance(anExpressionTreeNode, FlowExpressionTreeNode):
            txt = "ERR: Flow Equations for continuous variables are not supported" # TODO: Add error

        else:
            txt = "ERR: Unknown ExpressionTreeNode" # TODO: Add error

        return txt



    # ----------------------------------------------------------------------------------------------



    def _translateComponentTree(self, componentImplementation, accessPath):
        """Translates a component implementation instance and recursively all its subcomponents to Promela.
        To start the translation call this method without parameters, since then the root component will be searched and then used.

        @precondition: ...
        @type accessPath:  string
        @param accessPath: access path to the currently considered component implementation
        @type componentImplementation:  instance of slim_compiler.src.dataStructure.nominalbehaviour.ComponentImplementation
        @param componentImplementation: compiler's data structure of the component implementation which is to be considered
        @rtype:  A tuple of two strings (gvds: string, ptds: string)
        @return: A pair in which gvds contains Promela declarations of global variables while ptds contains Promela proctype definitions
        """

        # *** Generate Proctype Info & Translate to Promela ***
        pi = ProctypeInfo(self, componentImplementation, accessPath)
        (gvds, ptds) = pi.toPromela()

        # *** Descend Component Tree:   for each Non-Data Subcomponent: invoke translateComponentTree() ***
        # componentImplementation.non_data_subcomponents = dict(scName: string, scObject: instance(slim_compiler.src.dataStructure.nominalbehaviour.NonDataSubcomponent))
        for (scName, scObject) in componentImplementation.non_data_subcomponents.items():
            scImplementation = scObject.get_referenced_component_implementation()         
            (gvdsAppend, ptdsAppend) = self._translateComponentTree(scImplementation, accessPath + "_" + scName)
            (gvds, ptds) = (gvds + "\n" + gvdsAppend, ptds + "\n\n\n" + ptdsAppend)

        # *** Return (global vars, proctypes) ***
        return (gvds, ptds)



    def _translateHeader(self):
        """Generates the header of Promela code, including definition of message types, macros, and integer IDs for enum and mode names.

        @precondition: This method is to be called after (!) _translateComponentTree(), since required information in self.enumNames/modeNames is collected by it.
        @rtype:  string
        @return: contains the header
        """

        # *** Generator Hint ***
        result = "// This file has been generated by Slim2Promela.\n"

        
        # *** Message Types ***
        result += "\n\n\n// ----- Message Types ----------------------------------------------------------------------------\n\n"
        result += "mtype = { update, invalidate, trymultiway, activate, deactivate, reset };\n"
        

        # *** Macros for ensuring atomicity of transitions ***
        result += "\n\n\n// ----- MACROs for ensuring atomicity of transitions ---------------------------------------------\n\n"
        
        result += "bool gflagGoNextTransition = true;  // initially true for Environment proctype\n\n"
        
        result += "#define nT(guard) d_step { \\\n"
        result += "\tguard && !gflagGoNextTransition; \\\n"
        result += "\tgflagGoNextTransition = true"
        if self.addPrintfs:
            result += "; \\\n\tprintf(\"### (%d) ### Process %d is allowed to take its next transition!\\n\", _pid, _pid)"
        result += " }\n"

        result += "#define fT d_step { \\\n"
        result += "\tgflagGoNextTransition = false"
        if self.addPrintfs:
            result += "; \\\n\tprintf(\"### (%d) ### Process %d returned right to take next transition!\\n\", _pid, _pid); \\\n"
            listOfGlobalStateVariables = self.globalStateVariables.items()
            showValues = ", ".join([k+"=%d" for (k,v) in listOfGlobalStateVariables])
            accessVariables = ", ".join([v for (k,v) in listOfGlobalStateVariables])
            result += "\tprintf(\"NEW GLOBAL STATE: " + showValues + "\\n\", " + accessVariables + ")"
        result += " }\n"


        # *** Flag for Multiway Event Communication ***
        result += "\n\n\n// ----- Flag for Multiway Event Communication ----------------------------------------------------\n\n"
        result += "bool gflagTrymultiway = false;\n"

        
        # *** Integer Type & Constants for ENUMs ***
        result += "\n\n\n// ----- DEFINE integer type & constants for enum data types --------------------------------------\n\n"

        # Integer Type depends on #enums:
        if len(self.enumNames) <= 2:
            result += "#define enumIntType bit\n"
        elif len(self.enumNames) <= 256:
            result += "#define enumIntType byte\n"
        elif len(self.enumNames) <= 32768:
            result += "#define enumIntType short\n"
        elif len(self.enumNames) <= 2147483648:
            result += "#define enumIntType int\n"
        else: # TODO: What happens inside Python when len(list) > MAXINT?
            self.error_collector.add(Slim2PromelaTranslationError("Too many enum names - do not fit into Promela int."))
            result += "ERROR: Too many enum names - do not fit into Promela int.\n"

        # For each enumName, define its enumID
        for (enumName, enumID) in self.enumNames.items():
            result += "#define enum_%s %s\n" % (enumName, enumID)


        # *** Integer Type & Constants for nominal modes ***
        result += "\n\n\n// ----- DEFINE integer type & constants for nominal modes ----------------------------------------\n\n"

        # Integer Type depends on #modes:
        if len(self.modeNames) <= 2:
            result += "#define modeIntType bit\n"
        elif len(self.modeNames) <= 256:
            result += "#define modeIntType byte\n"
        elif len(self.modeNames) <= 32768:
            result += "#define modeIntType short\n"
        elif len(self.modeNames) <= 2147483648:
            result += "#define modeIntType int\n"
        else: # TODO: What happens inside Python when len(list) > MAXINT?
            self.error_collector.add(Slim2PromelaTranslationError("Too many mode names - do not fit into Promela int."))
            result += "ERROR: Too many mode names - do not fit into Promela int.\n"

        # For each modeName, define its modeID
        for (modeName, modeID) in self.modeNames.items():
            result += "#define modeID_%s %s\n" % (modeName, modeID)


        # *** Placeholders for atomic propositions ***
        result += "\n\n\n// ----- DEFINE placeholders with atomic propositions ---------------------------------------------\n"
        result += "// Note: Defined names for enums/modes cannot be used in properties as CPP does not replace text in later #defines.\n\n"
        for (placeholderName, atomicProposition) in self.theAtomicPropositions.items():
            result += "#define atomicProposition_" + placeholderName + " " + self._atomicPropositionToPromela(atomicProposition) + "\n"


        # *** Header completed ***
        return result

    # </def _translateHeader: ...>

        

    def _translateEnvironment(self):
        """Generates the Environment proctype which is responsible for activating and communicating with the root component.

        @rtype:  string
        @return: contains the Environment proctype
        """
        
        # *** Head ***
        result = "// ----- Environment ------------------------------------------------------------------------------\n\n"
        result += "active proctype Environment()\n"
        result += "{\n"
        if self.addPrintfs: result += "\tprintf(\"### ENV(%d) ### Environment started.\\n\", _pid);\n\n\n"


        # *** Sequence to copy buffered (New) values of roots out data ports to Store. ***
        # (Similar to Fixpoint Iteration on update in non-root components.)
        # componentType.out_data_ports = dict(portName: string, portObject: instance(slim_compiler.src.dataStructure.nominalbehaviour.*Port))
        componentType = self.theRootComponent.get_referenced_component_type()

        odpNewToStore = ""
        for (portName, portObject) in componentType.out_data_ports.items():
            odpNewToStore  += "odpStore_root_" + portName +  " = odpNew_root_" + portName + "; "

        if len(componentType.out_data_ports) > 1:
            odpNewToStore = "d_step { " + odpNewToStore + "}"
        

        # *** Activate & Update root component ***
        result += "\t// *** Activate & Update root component ***\n"

        result += "\tchan_root!activate, 0;\n"
        result += "\tchan_root?activate, _;\n\n"

        result += "\t// For tests: manipulate root's input data port variables here\n\n"
        
        result += "\tchan_root!update, 0;\n"
        result += "\tchan_root?update, _;\n"
        if odpNewToStore: result += "\t" + odpNewToStore + "\n"

        result += "\n\tfT;  // Allow components to take transitions (initially nT reserved for Environment)\n"

        
        # *** React to Messages from root component ***
        result += "\n\n\n\t// *** React to Messages from root component ***\n"
        result += "\tdo\n"

        result += "\n\t// Invalidate: react with update\n"
        result += "\t:: chan_root?invalidate, _ ->\n"
        result += "\t\tchan_root!update, 0;\n"
        result += "\t\tchan_root?update, _;\n"
        if odpNewToStore: result += "\t\t" + odpNewToStore + "\n"
        result += "\t\tfT;  // Environment responsible for fT after invalidate-update-cycle!\n"

        result += "\n\t// Upward Trymultiway:\n"
        result += "\t:: chan_root?trymultiway, _ ->\n"
        result += "\t\tgflagTrymultiway = true;\n"
        result += "\t\tchan_root!trymultiway, 0;\n"          

        result += "\n\t// Downward Trymultiway:\n"
        rootInEvents = componentType.in_event_ports.keys()
        rootInEvents.sort()
        for idp in rootInEvents:
            result += "\t:: nT(true) ->\n"
            result += "\t\tchan_root!trymultiway, " + str(ProctypeInfo._EventToNumber(idp, rootInEvents, [])) + ";\n"
            result += "\t\tchan_root?trymultiway, _;\n"
            result += "\t\tif\n"
            result += "\t\t:: d_step { gflagTrymultiway ->\n"
            result += "\t\t\tgflagTrymultiway = false; }\n"
            result += "\t\t\tchan_root!update, 0;\n"
            result += "\t\t\tchan_root?update, _;\n"
            if odpNewToStore: result += "\t\t\t" + odpNewToStore + "\n"
            result += "\t\t\tfT;  // Environment responsible for fT after invalidate-update-cycle!\n"
            result += "\t\t:: else -> fT\n"
            result += "\t\tfi;\n"

        result += "\n\tod;\n"


        # *** Environment proctype completed ***
        result += "}\n"
        return result
    
    # </def _translateEnvironment: ...>



    # ----------------------------------------------------------------------------------------------



    def toPromela(self):
        """Translates the SLIM specification (which has been parsed from an input file) to Promela.
        The resulting Promela code as a string is returned and stored internally in translationResult.

        @precondition: ...
        @rtype:  string
        @return: the generated Promela code
        """

        # First, adapt atomic propositions: ensure valid Promela names for atomic proposition placeholders
        setSafeNamesForPlaceholders(self.theAtomicPropositions)
        
        # Second, eliminate underscores in whole component tree under root and in atomic propositions
        recursivelyEliminateUnderscoresInComponentsDataElementsAndSubcomponents(self.theRootComponent)
        iterativelyPerformUnderscoreEliminationInAtomicPropositions(self.theAtomicPropositions, self.theRootComponent)

        # Third, fill up empty in modes lists
        recursivelyFillUpEmptyInModesLists(self.theRootComponent)
        
        # Now, walk trough component tree and translate each component implementation instance in it
        (gvds, ptds) = self._translateComponentTree(self.theRootComponent, "root")

        # - - - - -

        # Output: Header
        self.translationResult += self._translateHeader()
        self.translationResult += "\n\n\n\n\n\n// ================================================================================================\n\n\n\n\n\n\n"

        # Output: Global Variable Declarations
        self.translationResult += "// ----- Global Variable Declarations for storing Components' Data --------------------------------\n\n"
        self.translationResult += gvds
        self.translationResult += "\n\n\n\n\n\n// ================================================================================================\n\n\n\n\n\n\n"

        # Output: Environment Proctype
        self.translationResult += self._translateEnvironment()
        self.translationResult += "\n\n\n\n\n\n// ================================================================================================\n\n\n\n\n\n\n"
        
        # Implementations
        self.translationResult += ptds

        # Promela Code completed
        return self.translationResult

    # </def toPromela: ....>

# ==================================================================================================
# </class Slim2Promela_Translator: ...>
# ==================================================================================================










# ==================================================================================================
# Convenience Method for invoking the SLIM2Promela - Translation
# ==================================================================================================

def doTranslation(rootComponentImplementation, atomicPropositions = {}, outputFileName = "", verbose = False, addPrintfs = False):
    # Attention: We perform the following renamings in place.
    #     - iterativelyRewriteErrorModesInAtomicPropositions(self.theAtomicPropositions)
    #     - setSafeNamesForPlaceholders(self.theAtomicPropositions)
    #     - recursivelyEliminateUnderscoresInComponentsDataElementsAndSubcomponents(self.theRootComponent)
    #     - iterativelyPerformUnderscoreEliminationInAtomicPropositions(self.theAtomicPropositions, self.theRootComponent)
    # Use copy.deepcopy before, if you want to.
    
    # Initiate translation
    myTranslator = Slim2Promela_Translator(rootComponentImplementation, atomicPropositions, verbose, addPrintfs)
    translationResult = myTranslator.toPromela()
        
    # Save translation result to file
    if outputFileName != "":
        outputFileObject = open(outputFileName, "w")
        outputFileObject.write(translationResult)
        outputFileObject.close()

    # Return Translation result & error collector
    return (translationResult, myTranslator.error_collector)

# ==================================================================================================
# /Convenience Method for invoking the SLIM2Promela - Translation
# ==================================================================================================










# ==================================================================================================
# Main Program
# ==================================================================================================

if __name__ == "__main__":
    
    # Print Start-of-Program Message
    print "START OF SLIM2PROMELA - Translation (v2)\n\n\n"
    propertiesInputFileName = ""


    # Input File Names
    slimInputFileName = "tests\\slim2promela\\slim2promela.slim"
    #slimInputFileName = "tests\\manyUnderscores.slim"
    #slimInputFileName = "tests\\severalRoots.slim" # add: rootComponentName = "MainA.Impl2"
    #slimInputFileName = "tests\\properties2promela-concrete.slim"
    #propertiesInputFileName = "tests\\properties2promela-concrete.xml"
    slimInputFileName = "examples\\HalfAdder\\HalfAdder.slim"
    #slimInputFileName = "examples\\FullAdder.slim"
    #slimInputFileName = "examples\\RedundantBatterySystem2.slim"
    #slimInputFileName = "examples\\TransitionChangesDataPort.slim"
    #slimInputFileName = "examples\\Adder.slim"
    #slimInputFileName = "examples\\Adder_X.slim"
    #slimInputFileName = "examples\\Negator.slim"


    # Parse & Check SLIM Input File
    originalCompileInfo = slim_compiler.src.dataStructure.compile.CompileInformation()
    
    parsingResult = invokers.slim.parse(originalCompileInfo, [slimInputFileName])
    if not parsingResult:
        printErrorCollector(originalCompileInfo.error_collector)
        printErrorMessageAndStop("SLIM input files could not be parsed correctly - STOP!", 1)

    checkingResult = invokers.slim.check(originalCompileInfo)
    if not checkingResult:
        printErrorCollector(originalCompileInfo.error_collector)
        printErrorMessageAndStop("Parsed SLIM input files did not meet all syntactical restrictions - STOP!", 1)

    printOutputBlock("SLIM Parsing Result:   the parsed & checked Compile Info", toSLIMVisitor().visit(originalCompileInfo))
    

    # Find root component
    rootComponentImplementation = invokers.slim.get_root_impl(originalCompileInfo)


    # Properties ...
    propertiesDict = {}
    if propertiesInputFileName != "":
        propertiesDocument = properties.parse_properties_document(propertiesInputFileName)
        
        # Collect all Atomic Propositions for all Properties
        for aProperty in filter(lambda (aProperty): aProperty.get_logic() == constants.Logic.LTL, propertiesDocument):
            atomicPropositions = {}
            for aPlaceholder in aProperty.pattern_instance:
                atomicPropositions[aPlaceholder.placeholder_id] = testexpr.parseExpression(aPlaceholder.content.text)
            propertiesDict[aProperty.name] = (aProperty.pattern_instance.pattern_id, atomicPropositions)

    atomicPropositionsDict = {}
    print "--------------------------------------------------------------------------------"
    print "Properties Parsing Result:   the LTL-Properties"
    print "--------------------------------------------------------------------------------\n"
    for (propertyName, (patternId, atomicPropositions)) in propertiesDict.items():
        print propertyName + " (" + str(patternId) + "):"
        for (placeholderName, atomicProposition) in atomicPropositions.items():
            print "- " + placeholderName + ": " + expressionToString(atomicProposition)
            print "=> " + propertyName + "XXX" + placeholderName + ": " + expressionToString(atomicProposition)
            atomicPropositionsDict[propertyName + "XXX" + placeholderName] = atomicProposition
    print "\n--------------------------------------------------------------------------------\n\n\n"
    

    # Do the translation
    (translationResult, errors) = doTranslation(rootComponentImplementation, atomicPropositionsDict, "", True, True)
    printErrorCollector(errors)
    printOutputBlock("SLIM2Promela - Translation Result:", translationResult)


    # Print End-of-Program Message
    print "\n\n\nEND OF SLIM2PROMELA - Translation (v2)"

# ==================================================================================================
# </Main Program>
# ==================================================================================================
