# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================


# Here are all the functions to converter a SLIM parse tree
# into some format.
#
# To make the convertion functions as general as possible they
# generate text according to templates.
# This means that in order to output SLIM in some other format
# it may be enough to change the templates (defined in template.py)
# but not the functions themselves.
# See templates.py for more information.

# Every template "tmpl" has a corresponding function dump_tmlp.

import options
import os
import string
import sys
import traceback
import collections
import operator
import itertools

from utils import log

from slim_compiler.src.dataStructure import expressions
from slim_compiler.src.dataStructure import nominalbehaviour
from slim_compiler.src.smv import support
from slim_compiler.src.dataStructure.visitors.printVisitor import toSLIMVisitor
from slim_compiler.src.dataStructure.compile import CompileInformation

# this enables/disables generation of debug properties and other
# constructs in the generated SMV
DEBUG = 0


class Converter (object):
    """Class for supporting SLIM model translations.

    This class contains functionalities that support the encoding
    process, but are not specific to the target encoding.

    Attributes:
        _templates: Templates used to output the encoding.
        _id_prefix: A prefix of identifier; no prefix by default.
        _re_mulnl: A regex for removing repeted empty lines.
        _re_int: A regex for recognizing integer constants.
        _re_float: A regex for recognizing real constants.
        _re_word: A regex for recognizing word constants.
    """

    def __init__(self, templates):
        """Inits the converter object.

        Args:
            template: A template used to output the encoding.
        """

        self._templates = templates
        self._id_prefix = ""

        import re

        # Regex for removing repeated empty lines.
        self._re_mulnl = re.compile(("^\s*" + os.linesep) * 2, re.MULTILINE)

        # Regex for recognizing integer constants.
        self._re_int = re.compile("(-?)0[us]?d\d+_(\d+)")

        # Regex for recognizing real constants.
        self._re_float = re.compile("(-?\d+)(\.?\d*$)")

        # Regex for recognizing word constants.
        self._re_word = re.compile("(-?)0[us]?d\d+_(\d+$)")

    def dump(self, root_impl, ofile):
        """Dumps root implementation."""
        pass

    def set_id_prefix(self, id_prefix):
        """Sets the prefix for identifier.

        This method will be used during the encoding of expressions,
        data names, subcomponent names, port names, etc.

        Args:
            prefix: A new prefix.
        """
        self._id_prefix = id_prefix

    def get_id_prefix(self):
        """Gets the prefix for identifier.

        Returns:
            The prefix used for encoding identifiers.
        """
        return self._id_prefix

    ## The methods below provide accesses to template elements.

    def mode_var(self):
        """Gets the name of the mode variable.

        Returns:
            The name of mode variable.
        """
        return self._templates.MODE_VAR

    def root_var(self):
        """Gets the name of the root variable.

        Returns:
            The name of root variable.
        """
        return self._templates.ROOT_VAR

    def active_name(self):
        """Gets the name of active definition.

        Returns:
            The name of active definition
        """
        return self._templates.ACTIVE_DEF

    def tau_event(self):
        """Gets the name of tau event.

        Returns:
            The name of tau event.
        """
        return self._templates.TAU_EVENT

    def timed_event(self):
        """Gets the name of timed event.

        Returns:
            The name of timed event.
        """
        return self._templates.TIMED_EVENT

    def stutter_event(self):
        """Gets the name of stutter event.

        Returns:
            The name of stutter event.
        """
        return self._templates.STUTTER_EVENT

    def delta_var(self):
        """Gets the name of delta variable.

        Returns:
            The name of delta variable.
        """
        return self._templates.DELTA_VAR

    def time_var(self):
        """Gets the name of time variable.

        Important: this variable is introduced only if requested so at
        construction time.

        Returns:
            The name of time variable.
        """
        return self._templates.TIME_VAR

    def time_var_tu(self):
        """Gets the time unit
        `slim_compiler.src.dataStructure.expressions.TimeUnitType` enumeration
        used to encode time"""
        return self.def_base_unit

    def done_delta_var(self):
        """Returns the name of done-delta variable.

        Returns:
            The name of the done-delta variable
        """
        return self._templates.DONE_DELTA_VAR

    ## The methods below provide naming convention.

    def impl_name(self, name):
        """Normalizes implementation name.

        This method removes white spaces in the given name, prepends the name
        with a prefix signifying implementation, and encloses it with
        double quotes if it contains dots.

        Args:
            name: A name for an implementation.

        Returns:
            The normalized implementation name.
        """
        name = name.replace(' ', '')
        name = self._templates.IMPL_PREFIX + name

        if "." in name:
            name = '"' + name + '"'

        return name

    def event_name(self, name):
        """Normalizes event name.

        Args:
            name: An event name.

        Returns:
            The normalized event name.
        """
        return self._rename(name,
                            self._templates.SUBCOMP_PREFIX,
                            self._templates.EVENT_PREFIX)

    def event_par_name(self, name):
        """Gets formal parameter name for an event.

        Args:
            name: An event name.

        Returns:
            The formal parameter name for the specified event.
        """
        return self._templates.EVENT_PAR_PREFIX + self.event_name(name)

    def data_name(self, name):
        """Normalizes data name.

        Args:
            name: A data name.

        Returns:
            The normalized data name.
        """
        return self._rename(name,
                            self._templates.SUBCOMP_PREFIX,
                            self._templates.DATA_PREFIX)

    def data_par_name(self, name):
        """Gets formal parameter name for data.

        Args:
            name: A data name.

        Returns:
            The formal parameter name for data.
        """
        return self._templates.DATA_PAR_PREFIX + self.data_name(name)

    def sc_name(self, name):
        """Normalizes subcomponent name.

        Args:
            name: A subcomponent name.

        Returns:
            The normalized subcomponent name.
        """
        return self._rename(name,
                            self._templates.SUBCOMP_PREFIX,
                            self._templates.SUBCOMP_PREFIX)

    def const_name(self, name):
        """Normalizes enumeration constant.

        Args:
            name: An enumeration name.

        Returns:
            The normalized enumeration name.
        """
        assert "." not in name, name  # No complex name.
        return self._templates.CONST_PREFIX + name

    def mode_var_name(self, name):
        """Normalizes mode variable name.

        Args:
            name: A mode variable name.

        Returns:
            The normalized mode variable name.
        """
        return self._rename(name, self._templates.SUBCOMP_PREFIX, "")

    # TODO(IN): Added due to legacy.
    def keep_var_name(self, name):
        """I have no idea why this method existed."""
        return self._rename(name, self._templates.SUBCOMP_PREFIX, "")

    def error_var_name(self, name):
        """Normalizes error state variable name.

        Error state variable is encoded as mode variable. For example,
        a.b.c.error is first renamed to a.b.c.mode before being renamed
        further to follow the normalized mode variable name.

        Args:
            name: An error state variable name.

        Returns:
            The normalized error state variable name.
        """
        temp = name.split(".")[:-1] + [self.mode_var()]
        return self.mode_var_name(".".join(temp))

    def id_var_name(self):
        """Normalizes variable name for transition id.

        Returns:
            The normalized id variable name
        """
        return self._rename(self._templates.ID_VAR,
                            self._templates.SUBCOMP_PREFIX,
                            "")

    def mode_lit_name(self, name):
        """Normalizes mode literal name.

        Args:
            name: A mode literal name.

        Returns:
            The normalized mode literal name.
        """
        assert "." not in name, name  # No complex name.
        return self._templates.MODE_LIT_PREFIX + name

    def error_lit_name(self, name):
        """Normalizes error literal name.

        Args:
            name: An error literal name.

        Returns:
            The normalized error literal name.
        """
        assert "." not in name, name  # No complex name.
        return self._templates.ERROR_LIT_PREFIX + name

    def mode_name(self, impl, name):
        """Normalizes mode name according to the kind of implementation.

        The mode is an actual mode if the implementation describes nominal
        behavior, or is an error state if the implementation describes
        error behavior. This mode calls mode_lit_name in the former
        case, and error_lit_name in the latter.

        Args:
            impl: An implementation.
            name: A mode name.

        Returns:
            The normalized mode name according to the implementation.
        """
        if impl.is_error_impl():
            return self.error_lit_name(name)
        return self.mode_lit_name(name)

    def cando_name(self, name):
        """Produces an identifier denoting transition's precondition.

        This method takes as an input a name of a kind transition and
        returns an identifier denoting the precondition of performing
        that kind of transition. The kind of transition is determined
        by the cause of the transition:
        - mode transition,
        - input and output event port, and
        - and tau.

        Args:
            name: A name of transition.

        Returns:
            The identifier denoting transition's precondition.
        """
        return self._rename(name,
                            self._templates.SUBCOMP_PREFIX,
                            self._templates.CANDO_PREFIX)

    def resolved_name(self, name):
        """Produces an identifier denoting resolved condition of output event.

        This method takes as an input a name of an output event
        and returns an identifier denoting resolved condition of the event.

        Args:
            name: A name of output event.

        Returns:
            The identifier denoting resolved condition of output event.
        """
        return self._rename(name,
                            self._templates.SUBCOMP_PREFIX,
                            self._templates.RESOLVED_PREFIX)

    def sc_do_event_name(self, name, event_name):
        """Produces an identifier denoting event transition's obligation.

        This method takes as inputs a name of a subcomponent and
        a name of an event, including tau, and returns an identifier
        denoting the obligation of a subcomponent to perform a transition
        associated with the event.

        Args:
            name: A name of subcomponent.
            event_name: A name of event, including tau.

        Returns:
            The identifier denoting event transition's obligation.
        """
        sc_name = self.sc_name(name)
        return self._substitute(
            self._templates.sc_do_event,
            {"sc_name": sc_name,
             "event_name": event_name})

    def sc_activated_name(self, name):
        """Produces an identifier indicating if subcomponent is activated.

        This method takes as an input a name of a subcomponent and returns
        an identifier that indicates if the subcomponent is activated.

        Args:
            name: A name of subcomponent.

        Returns:
            The identifier denoting if a subcomponent is activated.
        """
        sc_name = self.sc_name(name)
        return self._substitute(self._templates.sc_activated,
                                {"sc_name": sc_name})

    def sc_reactivated_name(self, name):
        """Produces an identifier indicating if subcomponent is reactivated.

        This method takes as an input a name of a subcomponent and returns
        an identifier that indicates if the subcomponent is reactivated.

        Args:
            name: A name of subcomponent.

        Returns:
            The identifier denoting if a subcomponent is reactivated.
        """
        sc_name = self.sc_name(name)
        return self._substitute(self._templates.sc_reactivated,
                                {"sc_name": sc_name})

    def sc_delta_name(self, name):
        """Produces an identifier indicating the delta value for subcomponent.

        This method takes as an input a name of a subcomponent and returns
        an identifier that indicates the delta value for subcomponent.

        Args:
            name: A name of subcomponent.

        Returns:
            The identifier denoting the delta value.
        """
        sc_name = self.sc_name(name)
        return self._substitute(self._templates.sc_delta,
                                {"sc_name": sc_name})

    def sc_val_in_data_name(self, name, data_name):
        """Produces an identifier denoting the value of input data port.

        This method takes as inputs a name of a subcomponent and
        a name of input data port of the subcomponent, and return
        an identifier indicating the value of the input data port.

        Args:
            name: A name of subcomponent.
            data_name: A name of input data port.

        Returns:
            The identifier denoting value of input data port.
        """
        sc_name = self.sc_name(name)
        return self._substitute(
            self._templates.sc_in_data,
            {"sc_name": sc_name,
             "data_name": data_name})

    ## Private methods.

    def _write(self, s, ofile=None):
        """Writes string to an output file or to the standard output.

        Args:
            s : The string to be written.
            ofile : An output file.
        """

        s = self._re_mulnl.sub(os.linesep, s)  # removes double new lines

        if ofile:
            ofile.write(s)
        else:
            print s

    def _rename(self, name, prefix1, prefix2):
        """Renames dot-separated identifiers by inserting prefixes.

        This method takes a name in the form of dot-separated
        identifiers and two prefixes, and prepends all but the last
        identifiers with the first prefix and the last one with
        the second prefix. For example, given a.b.c as the input
        name and p1, p2 as, respectively, the input first and the second
        prefixes, this method returns p1a.p1b.p2c.

        Args:
            name: A string of dot-separated identifiers.
            prefix1: The first prefix.
            prefix2: The second prefix.

        Returns:
            The renamed dot-separated identifiers as described.
        """

        split_name = name.split(".")
        heads = map(lambda x: prefix1 + x, split_name[:-1])
        last = prefix2 + split_name[-1] if split_name[-1] else ""
        return self._id_prefix + ".".join(heads + [last])

    def _comment(self, msg):
        """Makes a comment out of a given message

        Args:
            msg: A string message.

        Returns:
            The comment containing the given message.
        """
        msgs = map(lambda x: "-- " + x + os.linesep, msg.splitlines())
        return "".join(msgs)

    def _banner(self):
        """Creates banner.

        Returns:
           The banner.
        """

        import datetime

        version = ".".join(map(str, options.PROG_VER))
        date = str(datetime.datetime.utcnow()) + " UTC"
        ban_str = self._substitute(self._templates.banner,
                                   {"progname": options.PROG_NAME,
                                    "progver": version,
                                    "date": date,
                                    "cmdline": options.PROG_CMDLINE, })
        return self._comment(ban_str)

    def _substitute(self, template, substitution):
        """Substitutes elements in template based on a given substitution.

        This method can be thought of as a shorthand for
        t = string.Template(template)
        return t.substitute(substitution)

        Args:
            template: A template.
            substitution: A substitution.

        Returns:
            The substituted template.
        """
        return string.Template(template).safe_substitute(substitution)

    def create_event_associated_behavior(self, impl):
        """Creates event associated behavior.

        For details, see implementation documentation.

        Args:
            impl: A component implementation.

        Returns:
            The mapping from events, including tau, to their
            corresponding behavior.
        """

        impl_type = impl.get_referenced_component_type()

        event_trans_map = {}

        # For each OUT event.
        for out_event in impl_type.out_event_ports:
            event_trans_map[out_event] = self._create_out_event_behavior(
                impl,
                out_event)

        # For each IN event.
        for in_event in impl_type.in_event_ports:
            event_trans_map[in_event] = self._create_in_event_behavior(
                impl,
                in_event)

        event_trans_map[self.tau_event()] = self._create_tau_event_behavior(
            impl)

        if DEBUG:
            print "=========== DEBUG Component %s ==========" % impl.name
            for ev in sorted(event_trans_map):
                print ev
                print "---------"
                ev_list = event_trans_map[ev]
                for n, k, ts in ev_list:
                    print "-- (EVENT: %s, TYPE: %d)" % (n, k)
                    if ts:
                        for t in ts:
                            print toSLIMVisitor().visit(t)
                            print "\n"

        return event_trans_map

    def _create_out_event_behavior(self, impl, out_event):
        """Creates event behavior associated to an output event.

        For details, see implementation documentation.

        Args:
            impl: A component implementation.
            out_event: An output event.

        Returns:
            The behavior associated to the output event.
        """

        out_event_trans = []

        # Type 0 transition.
        out_trans = support.get_trans_with_event(impl, out_event)
        if out_trans:
            out_event_trans.append((out_event, 0, out_trans))

        # Get all sc.OUT such that there exists connection (sc.OUT,OUT).
        out_to_out_conns = support.get_out_to_out_connections_with_target(
            impl, out_event)
        sc_out_events = set([c.source_port for c in out_to_out_conns])

        for sc_out_event in sorted(sc_out_events):

            # Type 1 transition
            sc_out_trans = support.get_trans_with_event(impl, sc_out_event)
            if sc_out_trans:
                out_event_trans.append((sc_out_event, 1, sc_out_trans))

            # Get multiway connection (sc.OUT,sc'.IN).
            mway_conns = support.get_out_to_in_connections_with_source(
                impl, sc_out_event)

            # Filter for passive.
            mway_conns = [c for c in mway_conns if not c.passive]

            if mway_conns:
                # Type 2 transition.
                out_event_trans.append((sc_out_event, 2, []))

            # Type 3 transition
            out_event_trans.append((sc_out_event, 3, []))

        return out_event_trans

    def _create_in_event_behavior(self, impl, in_event):
        """Creates event behavior associated to an input event.

        For details, see implementation documentation.

        Args:
            impl: A component implementation.
            in_event: An input event.

        Returns:
            The behavior associated to the input event.
        """

        in_event_trans = []

        # Type 0 transition.
        in_trans = support.get_trans_with_event(impl, in_event)
        if in_trans:
            in_event_trans.append((in_event, 0, in_trans))

        # Get all sc.IN such that there exists connection (IN,sc.IN).
        in_to_in_conns = support.get_in_to_in_connections_with_source(
            impl, in_event)
        sc_in_events = set([c.target_port for c in in_to_in_conns])

        # sc_in_events = support.get_sc_in_events_with_event(impl, in_event)

        if sc_in_events:
            # Type 1 transition.
            in_event_trans.append((in_event, 1, []))

        return in_event_trans

    def _create_tau_event_behavior(self, impl):
        """Creates event behavior associated to the tau event.

        For details, see implementation documentation.

        Args:
            impl: A component implementation.

        Returns:
            The behavior associated to the tau event.
        """

        tau_event_trans = []

        # Type 0 transition.
        tau_trans = support.get_trans_with_tau(impl)
        if tau_trans:
            tau_event_trans.append((self.tau_event(), 0, tau_trans))

        for sc_name in sorted(impl.non_data_subcomponents):
            # Type 1 transition.
            tau_event_trans.append((sc_name, 1, []))

        for sc_name in sorted(impl.non_data_subcomponents):

            sc = impl.non_data_subcomponents[sc_name]
            sc_type = sc.get_referenced_component_type()

            for e in sorted(sc_type.in_event_ports):
                sc_in_event = sc.name + "." + e

                # Type 2 transition.
                sc_in_trans = support.get_trans_with_event(impl,
                                                           sc_in_event)
                if sc_in_trans:
                    tau_event_trans.append((sc_in_event, 2, sc_in_trans))

            for e in sorted(sc_type.out_event_ports):
                sc_out_event = sc.name + "." + e

                # Get connections (sc.OUT,OUT)
                conns = support.get_out_to_out_connections_with_source(
                    impl, sc_out_event)

                # Get modes where there is no (sc.OUT,OUT)
                conns_modes = support.get_connections_modes(impl, conns)
                all_modes = support.get_component_impl_modes(impl)
                no_conns_modes = all_modes - conns_modes

                if no_conns_modes:
                    sc_out_trans = support.get_trans_with_event(impl,
                                                                sc_out_event)

                    # Type 3 transition.
                    if sc_out_trans:
                        tau_event_trans.append((sc_out_event, 3, sc_out_trans))

                    # Type 4 transition.
                    tau_event_trans.append((sc_out_event, 4, []))

        return tau_event_trans


class SmvConverter (Converter):
    """Class for converting SLIM models into SMV models."""

    def __init__(self,
                 generate_timed_fsm=False,
                 generate_time_var=options.SMV_GENERATE_TIME_VAR,
                 generate_time_fairness_alt=options.SMV_FAIR_TIMED):
        """Inits the SMV converter objects.

        This initialization method is optionally parameterized by
        a Boolean flag indicating if delta should be considered.
        If this flag is set to true, then the encoding that supports
        timed (or delta) transition is generated.

        Args:
            generate_timed_fsm: A Boolean flag indicating if delta should
                be considered.

            generate_time_var: A boolean flag indicating if time
                variable should be generated.  IMPORTANT: when True,
                generate_timed_fsm must be True as well.

            generate_time_fairness_alt=True): When generating timed
            FSM, if this parameter is True additional constraint are
            added to force timed transitions in fair paths, and force
            alternating time and discrete transitions.
        """
        from slim_compiler.src.smv import templates
        super(SmvConverter, self).__init__(templates)

        assert not generate_time_var or generate_timed_fsm, \
            "generate_time_var assumes generate_timed_fsm"

        self.generate_timed_fsm = generate_timed_fsm
        self.generate_time_var = generate_time_var
        self.generate_time_fairness_alt = generate_time_fairness_alt

        # sets of used and declared constants are used when converting
        # constants (see bug #80)
        self._used_constants = set()
        self._decl_constants = set()

        self._assigned_odps_cache = {}
        self._trans_id_cache = {}
        self._port_conns_cache = {}

        # time-unit convertion ratio [from][to]
        self._tu_map_ratio = {
            expressions.TimeUnitType.MSEC:
            {expressions.TimeUnitType.MSEC: "1",
             expressions.TimeUnitType.SEC: "f'1/1000",
             expressions.TimeUnitType.MIN: "f'1/60000",
             expressions.TimeUnitType.HOUR: "f'1/3600000",
             expressions.TimeUnitType.DAY: "f'1/86400000"},

            expressions.TimeUnitType.SEC:
            {expressions.TimeUnitType.MSEC: "1000",
             expressions.TimeUnitType.SEC: "1",
             expressions.TimeUnitType.MIN: "f'1/60",
             expressions.TimeUnitType.HOUR: "f'1/3600",
             expressions.TimeUnitType.DAY: "f'1/86400"},

            expressions.TimeUnitType.MIN:
            {expressions.TimeUnitType.MSEC: "3600000",
             expressions.TimeUnitType.SEC: "60",
             expressions.TimeUnitType.MIN: "1",
             expressions.TimeUnitType.HOUR: "f'1/60",
             expressions.TimeUnitType.DAY: "f'1/1440"},

            expressions.TimeUnitType.HOUR:
            {expressions.TimeUnitType.MSEC: "3600000",
             expressions.TimeUnitType.SEC: "3600",
             expressions.TimeUnitType.MIN: "60",
             expressions.TimeUnitType.HOUR: "1",
             expressions.TimeUnitType.DAY: "f'1/24"},

            expressions.TimeUnitType.DAY:
            {expressions.TimeUnitType.MSEC: "86400000",
             expressions.TimeUnitType.SEC: "86400",
             expressions.TimeUnitType.MIN: "1440",
             expressions.TimeUnitType.HOUR: "24",
             expressions.TimeUnitType.DAY: "1"},
        }

        # this should be a class property
        self.def_base_unit = CompileInformation().base_unit

    def dump(self, root_impl, ofile):
        """Dumps SMV model of the SLIM model into an output file.

        Args:
            root_impl: The root implementation of SLIM model.
            ofile: An output file.
        """
        self._used_constants = set()
        self._decl_constants = set()

        log.system.info("Converting SLIM model to SMV...")
        try:
            result = self._afile(root_impl)
        except Exception, e:
            exc_type, exc_value, exc_traceback = sys.exc_info()
            traceback.print_tb(exc_traceback)
            log.system.error("An error has occurred during "
                             "SLIM to SMV conversion:")
            log.system.error(str(e))
            raise e
        log.system.info("Done SLIM to SMV conversion")
        self._write(result, ofile)

    def _afile(self, root_impl):
        """Translates each component implementation.

        This method is an auxiliary method that helps the dump
        method. See dump.

        Args:
            root_impl: The root implementation of SLIM model.

        Returns:
            The results of translation.
        """
        impls = support.get_all_children_implementations(root_impl)
        impls.add(root_impl)

        impl_map = {}
        for impl in impls:
            impl_map[impl.name] = impl

        results = []
        for impl_name in sorted(impl_map):
            results.append(self._implementation(impl_map[impl_name]))

        main_module = self._main_module(root_impl)

        return self._substitute(self._templates.afile,
                                {"banner": self._banner(),
                                 "implementation_list": "".join(results),
                                 "main_module": main_module})

    def _implementation(self, impl):

        event_trans_map = self.create_event_associated_behavior(impl)

        substitution = {}

        impl_name = impl.name
        if impl.subpackage.package.name != "__default__":
            impl_name = impl.subpackage.package.name + "." + impl_name

        substitution["implementation_name"] = self.impl_name(impl_name)

        ev_def_list = self._event_definition_list(impl)
        substitution["event_definition_list"] = ev_def_list

        in_data_def_list = self._input_data_definition_list(impl)
        substitution["input_data_definition_list"] = in_data_def_list

        substitution["module_parameters"] = self._module_parameters(impl)
        substitution["active_definition"] = self._templates.active_definition

        substitution["mode_declaration"] = self._mode_declaration(impl)

        (max_id, id_decl) = self._id_declaration(event_trans_map)
        substitution["id_declaration"] = id_decl

        substitution["mode_invariant_list"] = self._mode_invariant_list(impl)

        event_id_invar_list = self._event_id_invariant_list(event_trans_map,
                                                            max_id)
        substitution["event_id_invariant_list"] = event_id_invar_list

        odps = self._output_data_port_list(impl)
        substitution["output_data_port_list"] = odps

        dsc_decl_list = self._data_subcomponent_declaration_list(impl)
        substitution["data_subcomponent_declaration_list"] = dsc_decl_list

        csc_decl_list = self._control_subcomponent_declaration_list(impl)
        substitution["control_subcomponent_declaration_list"] = csc_decl_list

        init_list = self._initialization_list(impl)
        substitution["initialization_list"] = init_list

        cando_mt_list = self._cando_mode_trans_definition_list(impl)
        substitution["cando_mode_trans_definition_list"] = cando_mt_list

        cando_event_list = self._cando_event_definition_list(impl,
                                                             event_trans_map)
        substitution["cando_event_definition_list"] = cando_event_list

        trans_list = self._transition_list(impl, event_trans_map)
        substitution["transition_list"] = trans_list

        resolved_list = self._resolved_definition_list(impl, event_trans_map)
        substitution["resolved_definition_list"] = resolved_list

        act_list = self._control_subcomponent_active_definition_list(impl)
        substitution["control_subcomponent_active_definition_list"] = act_list

        idp_sc_list = self._input_data_subcomponent_definition_list(impl)
        substitution["input_data_subcomponent_definition_list"] = idp_sc_list

        do_list = self._do_control_subcomponent_event_definition_list(
            impl,
            event_trans_map)
        substitution["do_control_subcomponent_event_definition_list"] = do_list

        return self._substitute(self._templates.implementation,
                                substitution)

    def _main_module(self, root_impl):
        return self._root_declaration(root_impl)

    def _module_parameters(self, impl):
        """Generate module parameters."""

        substitution = {}
        impl_type = impl.get_referenced_component_type()

        substitution["tau_event_par"] = self.event_par_name(self.tau_event())

        in_events = map(self.event_par_name, sorted(impl_type.in_event_ports))
        in_event_par = ", ".join(in_events)
        if in_events:
            in_event_par = ", " + in_event_par

        substitution["in_event_par_list"] = in_event_par

        out_events = map(self.event_par_name,
                         sorted(impl_type.out_event_ports))
        out_event_par = ", ".join(out_events)
        if out_events:
            out_event_par = ", " + out_event_par

        substitution["out_event_par_list"] = out_event_par

        in_data = map(self.data_par_name, sorted(impl_type.in_data_ports))
        in_data_par = ", ".join(in_data)
        if in_data:
            in_data_par = ", " + in_data_par

        substitution["in_data_par_list"] = in_data_par

        if self.generate_timed_fsm:
            substitution["delta"] = ", " + self.delta_var()
        else:
            substitution["delta"] = ""

        return self._substitute(self._templates.module_parameters,
                                substitution)

    def _event_definition_list(self, impl):
        """Generates a list of event definitions."""

        impl_type = impl.get_referenced_component_type()

        events = []
        result = []
        for ev in sorted(impl_type.in_event_ports):
            ev_name = self.event_name(ev)
            ev_par_name = self.event_par_name(ev)
            ev_def = self._substitute(self._templates.event_definition,
                                      {"event_name": ev_name,
                                       "event_par_name": ev_par_name})
            events.append(ev_name)
            result.append(ev_def)

        for ev in sorted(impl_type.out_event_ports):
            ev_name = self.event_name(ev)
            ev_par_name = self.event_par_name(ev)
            ev_def = self._substitute(self._templates.event_definition,
                                      {"event_name": ev_name,
                                       "event_par_name": ev_par_name})
            events.append(ev_name)
            result.append(ev_def)

        tau_ev_name = self.event_name(self.tau_event())
        tau_ev_par_name = self.event_par_name(self.tau_event())
        tau_def = self._substitute(self._templates.event_definition,
                                   {"event_name": tau_ev_name,
                                    "event_par_name": tau_ev_par_name})
        events.append(tau_ev_name)
        result.append(tau_def)

        if self.generate_timed_fsm:
            events.append(self.timed_event())
            result.append(self._templates.timed_event_definition)

        stutter_def = self._substitute(self._templates.stutter_definition,
                                       {"event_list": " | ".join(events)})
        result.append(stutter_def)

        return "".join(result)

    def _input_data_definition_list(self, impl):
        """Generates a list of input data definitions."""

        impl_type = impl.get_referenced_component_type()

        result = []
        for in_data in sorted(impl_type.in_data_ports):
            in_data_def = self._substitute(
                self._templates.input_data_definition,
                {"in_data_name": self.data_name(in_data),
                 "in_data_par_name": self.data_par_name(in_data)})
            result.append(in_data_def)

        return "".join(result)

    def _mode_declaration(self, impl):
        """Generates mode declaration."""

        assert impl.modes, "No modes"  # There must be a mode.

        modes = map(lambda m: self.mode_name(impl, m), sorted(impl.modes))
        return self._substitute(self._templates.mode_declaration,
                                {"mode_list": ", ".join(modes)})

    def _mode_invariant_list(self, impl):
        """Generates mode invariants."""

        assert impl.modes, "No modes"  # There must be a mode.

        invars = []
        for m in sorted(impl.modes):
            m_invar = impl.modes[m].invariant

            if m_invar is not None:
                m_invar_const = support.get_constraints_list_from_invariants(
                    m_invar)

                if m_invar_const:
                    const_list = map(lambda c: self.expression(impl, c),
                                     m_invar_const)
                    invar = self._substitute(self._templates.mode_invariant,
                                             {"mode": self.mode_name(impl, m),
                                              "invar": " & ".join(const_list)})
                else:
                    invar = self._substitute(self._templates.mode_invariant,
                                             {"mode": self.mode_name(impl, m),
                                              "invar": "TRUE"})
                invars.append(invar)

        return "".join(invars)

    def _id_declaration(self, event_trans_map):
        """Generates id declaration."""

        max_id = 0
        for ts in event_trans_map.itervalues():
            if max_id < len(ts):
                max_id = len(ts)

        if max_id == 0:
            max_id = 1

        decl = self._substitute(self._templates.id_declaration,
                                {"max_id": max_id})

        return (max_id, decl)

    def _event_id_invariant_list(self, event_trans_map, max_id):
        """Generates invariants of event-associated id."""

        invars = []
        for ev in sorted(event_trans_map):
            ts = event_trans_map[ev]
            if ts and len(ts) != max_id:
                invar_ev = self._substitute(self._templates.event_id_invar,
                                            {"event": self.event_name(ev),
                                             "max_event_id": str(len(ts))})
                invars.append(invar_ev)

        return "".join(invars)

    def _output_data_port_list(self, impl):
        """Generates declarations and definitions for output data ports."""

        assigned = self._assigned_output_data_port_declaration_list(impl)
        unassigned = self._unassigned_output_data_port_definition_list(impl)
        return assigned + unassigned

    def _assigned_output_data_port_declaration_list(self, impl):
        """Generates declarations for assigned output data port."""

        impl_type = impl.get_referenced_component_type()
        # assigned_odps = support.get_assigned_odp(impl)
        assigned_odps = self._assigned_odps(impl)

        decls = []

        for odp in sorted(assigned_odps):
            odp_type = impl_type.out_data_ports[odp].data_type
            odp_decl = self._substitute(
                self._templates.assigned_odp_declaration,
                {"name": self.data_name(odp),
                 "type": self.type_name(odp_type)})
            decls.append(odp_decl)

        return "".join(decls)

    def _unassigned_output_data_port_definition_list(self, impl):
        """Generates definitions for unassigned output data port."""

        impl_type = impl.get_referenced_component_type()
        impl_modes = support.get_component_impl_modes(impl)

        # assigned_odps = support.get_assigned_odp(impl)
        assigned_odps = self._assigned_odps(impl)
        unassigned_odps = set(impl_type.out_data_ports) - assigned_odps

        odp_defs = []

        for odp in sorted(unassigned_odps):
            conn_values = []
            flow_values = []
            odp_modes = set([])

            conns = support.get_connections_with_target_data(impl, odp)
            for c in conns:
                odp_modes |= support.get_connection_modes(impl, c)
                val = self._odp_connection_value(impl, c)
                conn_values.append(val)

            flows = support.get_flows_with_target_data(impl, odp)
            for f in flows:
                odp_modes |= support.get_flow_modes(impl, f)
                val = self._odp_flow_value(impl, f)
                flow_values.append(val)

            (odp_default_value, is_undef) = self.default_value(
                impl,
                impl_type.out_data_ports[odp])

            if is_undef and (not impl_modes <= odp_modes):
                odp_type = impl_type.out_data_ports[odp].data_type
                nondet_var_decl = self._substitute(
                    self._templates.odp_nondet_declaration,
                    {"name": self.data_name(odp),
                     "type": self.type_name(odp_type)})
                odp_default_value = self._substitute(
                    self._templates.odp_nondet, {"name": self.data_name(odp)})
                odp_defs.append(nondet_var_decl)

            odp_def = ""

            if not flows and not conns:
                odp_def = self._substitute(
                    self._templates.unassigned_odp_definition_opt,
                    {"name": self.data_name(odp),
                     "default_value": odp_default_value})
            else:
                if impl_modes <= odp_modes:
                    odp_def = self._substitute(
                        self._templates.unassigned_odp_definition_wo_default,
                        {"name": self.data_name(odp),
                         "odp_connection_values": "".join(conn_values),
                         "odp_flow_values": "".join(flow_values)})
                else:
                    odp_def = self._substitute(
                        self._templates.unassigned_odp_definition,
                        {"name": self.data_name(odp),
                         "odp_connection_values": "".join(conn_values),
                         "odp_flow_values": "".join(flow_values),
                         "default_value": odp_default_value})

            odp_defs.append(odp_def)

        return "".join(odp_defs)

    def _odp_connection_value(self, impl, connection):
        """Generates connection value for output data port."""

        conn_modes = support.get_connection_modes(impl, connection)
        mode_cond = self._mode_cond(impl, conn_modes)
        source = self.data_name(connection.source_port)

        return self._substitute(self._templates.odp_connection_value,
                                {"mode_cond": mode_cond,
                                 "source": source})

    def _odp_flow_value(self, impl, flow):
        """Generates flow value for output data port."""

        flow_modes = support.get_flow_modes(impl, flow)
        mode_cond = self._mode_cond(impl, flow_modes)

        # TODO(IN): Tidy up this dirty code.
        assert (isinstance(flow.expression,
                           expressions.BinaryOperatorTreeNode) and
                isinstance(flow.expression.left,
                           expressions.IdentifierTreeNode)), \
            "Flow has unexpected structure"

        flow_expr = self.expression(impl, flow.expression.right)

        return self._substitute(self._templates.odp_flow_value,
                                {"mode_cond": mode_cond,
                                 "flow_expr": flow_expr})

    def _data_subcomponent_declaration_list(self, impl):
        """Generates declarations for data subcomponents."""

        decls = []
        for dsc_name in sorted(impl.data_subcomponents):
            dsc = impl.data_subcomponents[dsc_name]
            decl = self._substitute(
                self._templates.data_subcomponent_declaration,
                {"name": self.sc_name(dsc.name),
                 "type": self.type_name(dsc.data_type)})
            decls.append(decl)

        return "".join(decls)

    def _control_subcomponent_declaration_list(self, impl):
        """Generates declarations for control (or non-data) subcomponents."""

        decls = []

        for ndsc_name in sorted(impl.non_data_subcomponents):
            csc = impl.non_data_subcomponents[ndsc_name]
            csc_name = self.sc_name(csc.name)

            csc_impl = csc.get_referenced_component_implementation()
            csc_impl_name = csc_impl.name
            if csc_impl.subpackage.package.name != "__default__":
                pack_name = csc_impl.subpackage.package.name
                csc_impl_name = pack_name + "." + csc_impl_name

            csc_impl_name = self.impl_name(csc_impl_name)

#             csc_impl_name = self.impl_name(
#                 csc.get_referenced_component_implementation().name)
            csc_activated = self.sc_activated_name(csc.name)
            csc_reactivated = self.sc_reactivated_name(csc.name)

            do_tau_event = self.sc_do_event_name(csc.name, self.tau_event())

            csc_type = csc.get_referenced_component_type()

            do_ievs = []
            for iev in sorted(csc_type.in_event_ports):
                do_iev = self.sc_do_event_name(csc.name, iev)
                do_ievs.append(do_iev)

            do_in_events = ", ".join(do_ievs)
            if do_ievs:
                do_in_events = ", " + do_in_events

            do_oevs = []
            for oev in sorted(csc_type.out_event_ports):
                do_oev = self.sc_do_event_name(csc.name, oev)
                do_oevs.append(do_oev)

            do_out_events = ", ".join(do_oevs)
            if do_oevs:
                do_out_events = ", " + do_out_events

            idp_vals = []
            for idp in sorted(csc_type.in_data_ports):
                idp_val = self.sc_val_in_data_name(csc.name, idp)
                idp_vals.append(idp_val)

            in_data_vals = ", ".join(idp_vals)
            if idp_vals:
                in_data_vals = ", " + in_data_vals

            delta = ""
            if self.generate_timed_fsm:
                # delta = ", " + self.delta_var()
                delta = ", " + self.sc_delta_name(csc.name)

            csc_decl = self._substitute(
                self._templates.control_subcomponent_declaration,
                {"name": csc_name,
                 "impl_name": csc_impl_name,
                 "activated": csc_activated,
                 "reactivated": csc_reactivated,
                 "do_tau_event": do_tau_event,
                 "do_in_event_list": do_in_events,
                 "do_out_event_list": do_out_events,
                 "in_data_val_list": in_data_vals,
                 "delta": delta})

            decls.append(csc_decl)

        return "".join(decls)

    ### INIT #################################################################

    def _initialization_list(self, impl):
        """Generates initializations."""

        mode_init = self._mode_init(impl)
        dsc_inits = self._data_subcomponent_inits(impl)
        odp_inits = self._output_data_port_inits(impl)

        return mode_init + dsc_inits + odp_inits

    def _mode_init(self, impl):
        """Generates initialization for mode."""

        modes = support.get_component_impl_modes(impl)

        assert modes  # Nonempty set of modes.

        if len(modes) == 1:
            return ""
        else:
            init_mode = support.get_initial_or_activation_mode(impl)
            return self._substitute(self._templates.mode_init,
                                    {"init_mode": self.mode_name(impl,
                                                                 init_mode)})

    def _data_subcomponent_inits(self, impl):
        """Generates initializations for data subcomponents."""

        init_mode = support.get_initial_or_activation_mode(impl)

        dsc_inits = []

        for dsc_name in impl.data_subcomponents:
            dsc = impl.data_subcomponents[dsc_name]
            dsc_modes = support.get_subcomponent_modes(impl, dsc)
            dsc_init = ""
            if init_mode in dsc_modes:
                (dsc_init_val, is_undef) = self.default_value(impl, dsc)
                if not is_undef:
                    dsc_init = self._substitute(
                        self._templates.data_subcomponent_init,
                        {"name": self.sc_name(dsc.name),
                         "initial_value": dsc_init_val})
            else:
                if isinstance(dsc.data_type, nominalbehaviour.Clock):
                    dsc_init = self._substitute(
                        self._templates.data_clock_init_inactive,
                        {"name": self.sc_name(dsc.name)})

            dsc_inits.append(dsc_init)

        return "".join(dsc_inits)

    def _output_data_port_inits(self, impl):
        """Generates initializations for output data port."""

        impl_type = impl.get_referenced_component_type()
        # assigned_odps = support.get_assigned_odp(impl)
        assigned_odps = self._assigned_odps(impl)

        odp_inits = []

        for odp in sorted(assigned_odps):
            odp_obj = impl_type.out_data_ports[odp]
            (odp_init_val, is_undef) = self.default_value(impl,
                                                          odp_obj)

            if not is_undef:
                odp_init = self._substitute(
                    self._templates.output_data_port_init,
                    {"name": self.data_name(odp),
                     "initial_value": odp_init_val})
                odp_inits.append(odp_init)

        return "".join(odp_inits)

    ### CANDO ################################################################

    def _cando_mode_trans_definition_list(self, impl):
        """Generates cando definitions for mode transitions."""

        defs = []
        for t in impl.normal_mode_transitions:
            cando_name = self.cando_name(self._trans_id(impl, t))
            mode_cond = self._mode_cond(impl, set([t.source_mode]))

            guard = ""
            if t.label.guard is not None:
                guard = " & " + self.expression(impl, t.label.guard)

            t_def = self._substitute(
                self._templates.cando_mode_trans_definition,
                {"cando_name": cando_name,
                 "mode_cond": mode_cond,
                 "guard": guard})

            defs.append(t_def)

        return "".join(defs)

    def _cando_event_definition_list(self, impl, event_trans_map):
        """Generates cando definition for event transitions."""

        impl_type = impl.get_referenced_component_type()

        cando_events = []

        for out_event in sorted(impl_type.out_event_ports):
            cando_out = self._cando_out_event_definition(
                impl, event_trans_map, out_event)
            cando_events.append(cando_out)

        for in_event in sorted(impl_type.in_event_ports):
            cando_in = self._cando_in_event_definition(
                impl, event_trans_map, in_event)
            cando_events.append(cando_in)

        cando_tau = self._cando_tau_event_definition(impl, event_trans_map)
        cando_events.append(cando_tau)

        if self.generate_timed_fsm:
            cando_events.append(
                self._templates.cando_timed_transition_definition)

        return "".join(cando_events)

    def _cando_out_event_definition(self, impl, event_trans_map, out_event):
        """Generates cando definition for output event transition."""

        cando_conds = []

        for ev, kind, trans_list in event_trans_map[out_event]:

            if kind == 0:
                assert trans_list  # Nonempty list of mode transitions.

                for t in trans_list:
                    tid = self._trans_id(impl, t)
                    cando_trans_name = self.cando_name(tid)
                    cando_conds.append(cando_trans_name)

            elif kind == 3:
                assert not trans_list  # Empty list of mode transitions.

                cando_sc_out = self.cando_name(ev)

                conns = support.get_connections_with_events(impl,
                                                            ev,
                                                            out_event)
                modes = support.get_connections_modes(impl, conns)
                or_mode_conds = self._mode_cond(impl, modes)

                if or_mode_conds != self._templates.mode_cond_all:
                    cando_sc_out = "(%s & %s)" % (cando_sc_out,
                                                  or_mode_conds)

                cando_conds.append(cando_sc_out)

            else:
                pass  # Do nothing.

        out_cond = ""

        if cando_conds:
            out_cond = " | ".join(cando_conds)
            if len(cando_conds) > 1:
                out_cond = "(" + out_cond + ")"
        else:
            out_cond = "FALSE"

        return self._substitute(self._templates.cando_event_trans_definition,
                                {"cando_name": self.cando_name(out_event),
                                 "cando_cond": out_cond})

    def _cando_in_event_definition(self, impl, event_trans_map, in_event):
        """Generates cando definition for input event transition."""

        cando_conds = []

        for ev, kind, trans_list in event_trans_map[in_event]:

            if kind == 0:
                assert trans_list  # Nonempty list of mode transitions.

                for t in trans_list:
                    tid = self._trans_id(impl, t)
                    cando_trans_name = self.cando_name(tid)
                    cando_conds.append(cando_trans_name)

            elif kind == 1:
                assert not trans_list  # Empty list of mode transitions.

                in_to_in_conns = support.get_in_to_in_connections_with_source(
                    impl, in_event)
                sc_in_events = set([c.target_port for c in in_to_in_conns])

                for sc_ev in sorted(sc_in_events):

                    cando_sc_in = self.cando_name(sc_ev)

                    conns = [c
                             for c in in_to_in_conns
                             if c.target_port == sc_ev]

                    modes = support.get_connections_modes(impl, conns)
                    or_mode_conds = self._mode_cond(impl, modes)

                    if or_mode_conds != self._templates.mode_cond_all:
                        cando_sc_in = "(%s & %s)" % (cando_sc_in,
                                                     or_mode_conds)

                    cando_conds.append(cando_sc_in)

            else:
                assert False  # Unreachable code.

        in_cond = ""

        if cando_conds:
            in_cond = " | ".join(cando_conds)
            if len(cando_conds) > 1:
                in_cond = "(" + in_cond + ")"
        else:
            in_cond = "FALSE"

        return self._substitute(self._templates.cando_event_trans_definition,
                                {"cando_name": self.cando_name(in_event),
                                 "cando_cond": in_cond})

    def _cando_tau_event_definition(self, impl, event_trans_map):
        """Generates cando definition for tau event transition."""

        cando_conds = []
        for ev, kind, trans_list in event_trans_map[self.tau_event()]:

            if kind == 0:
                assert trans_list  # Nonempty list of mode transitions.

                for t in trans_list:
                    tid = self._trans_id(impl, t)
                    cando_trans_name = self.cando_name(tid)
                    cando_conds.append(cando_trans_name)

            elif kind == 1:
                assert not trans_list  # Empty list of mode transitions.

                # Get subcomponent tau event.
                ev = ev + "." + self.tau_event()
                cando_sub_tau = self.cando_name(ev)
                cando_conds.append(cando_sub_tau)

            elif kind == 2:
                assert trans_list  # Nonempty list of mode transitions.

                cando_sub_in = self.cando_name(ev)

                cando_sub_in_trans = []
                for t in trans_list:
                    tid = self._trans_id(impl, t)
                    cando_trans_name = self.cando_name(tid)
                    cando_sub_in_trans.append(cando_trans_name)

                cando_sub_in = "%s & (%s)" % (cando_sub_in,
                                              " | ".join(cando_sub_in_trans))
                cando_conds.append(cando_sub_in)

            elif kind == 3 or kind == 4:

                cando_sub_out = self.cando_name(ev)

                # Get connections (sc.OUT,OUT)
                conns = support.get_out_to_out_connections_with_source(impl,
                                                                       ev)
                # Get modes where there is (sc.OUT,OUT)
                modes = support.get_connections_modes(impl, conns)

                if modes:
                    or_mode_conds = self._mode_cond(impl, modes)
                    cando_sub_out = "(%s & !(%s))" % (cando_sub_out,
                                                      or_mode_conds)

                if kind == 3:
                    assert trans_list  # Nonempty list of mode transitions.
                    tau3 = self._cando_tau_event_3_aux(impl, trans_list)
                    cando_sub_out = cando_sub_out + " & " + tau3
                elif kind == 4:
                    tau4 = self._cando_tau_event_4_aux(impl,
                                                       ev,
                                                       event_trans_map)
                    cando_sub_out = cando_sub_out + " & " + tau4

                cando_conds.append(cando_sub_out)

            else:
                assert False  # Unreachable code.

        tau_cond = ""

        if cando_conds:
            tau_cond = " | ".join(cando_conds)
            if len(cando_conds) > 1:
                tau_cond = "(" + tau_cond + ")"
        else:
            tau_cond = "FALSE"

        tau_event = self.tau_event()

        return self._substitute(self._templates.cando_event_trans_definition,
                                {"cando_name": self.cando_name(tau_event),
                                 "cando_cond": tau_cond})

    def _cando_tau_event_3_aux(self, impl, trans_list):
        """Generates cando constraint for tau event transition type 3."""

        assert trans_list  # Nonempty list of mode transitions.

        cando_tau = []
        for t in trans_list:
            tid = self._trans_id(impl, t)
            cando_trans_name = self.cando_name(tid)
            cando_tau.append(cando_trans_name)

        return " | ".join(cando_tau)

    def _cando_tau_event_4_aux(self, impl, sc_out_event, event_trans_map):
        """Generates cando constraint for tau event transition type 4."""

        (bhv3, pos) = self._find_event_behavior(event_trans_map,
                                                self.tau_event(),
                                                sc_out_event,
                                                3)

        cannot_do_3 = ""

        if bhv3 is not None:
            (ev, k, trans_list) = bhv3

            assert ev == sc_out_event and k == 3 and trans_list

            cando_tau_event_3 = self._cando_tau_event_3_aux(impl, trans_list)

            if len(trans_list) > 1:
                cannot_do_3 = "!(" + cando_tau_event_3 + ")"
            else:
                cannot_do_3 = "!" + cando_tau_event_3

        # get the output port given its name
        output_port = support.get_output_event(impl, sc_out_event)

        # if the output port is nonblocking, always allow the
        # tau event transition, but it must consider cannot_do_3 below
        if output_port.nonblocking:
            result = "TRUE"
        else:
            resolved_sc_out = self.resolved_name(sc_out_event)
            result = resolved_sc_out

            conns = support.get_out_to_in_connections_with_source(impl,
                                                                  sc_out_event)
            # Filter for passive.
            conns = [c for c in conns if not c.passive]

            if conns:
                cando_sc_ins = []
                for c in conns:
                    cando_sc_in = self.cando_name(c.target_port)

                    modes = support.get_connection_modes(impl, c)
                    or_mode_conds = self._mode_cond(impl, modes)

                    if or_mode_conds != self._templates.mode_cond_all:
                        cando_sc_in = "(%s & %s)" % (cando_sc_in,
                                                     or_mode_conds)

                    cando_sc_ins.append(cando_sc_in)

                if len(conns) > 1:
                    result = "(%s | (!%s & (%s)))" % (result,
                                                      resolved_sc_out,
                                                      " | ".join(cando_sc_ins))
                else:
                    result = "(%s | (!%s & %s))" % (result,
                                                    resolved_sc_out,
                                                    cando_sc_ins[0])

        if cannot_do_3:
            result = cannot_do_3 + " & " + result

        return result

    ### TRANSITIONS ##########################################################

    def _transition_list(self, impl, event_trans_map):
        """Generates transitions."""

        impl_type = impl.get_referenced_component_type()

        all_trans = []

        for out_event in sorted(impl_type.out_event_ports):
            out_event_trans = self._out_event_transition(
                impl, event_trans_map, out_event)
            all_trans.append(out_event_trans)

        for in_event in sorted(impl_type.in_event_ports):
            in_event_trans = self._in_event_transition(
                impl, event_trans_map, in_event)
            all_trans.append(in_event_trans)

        tau_event_trans = self._tau_event_transition(impl, event_trans_map)
        all_trans.append(tau_event_trans)

        stutter_trans = self._stutter_transition(impl)
        all_trans.append(stutter_trans)

        reactivation_trans = self._reactivation_transition(impl)
        all_trans.append(reactivation_trans)

        if self.generate_timed_fsm:
            all_trans.append(self._timed_transition(impl))

        return "".join(all_trans)

    def _out_event_transition(self, impl, event_trans_map, out_event):
        """Generates transitions associated with output event."""

        out_trans = []
        event_name = self.event_name(out_event)

        tid = 1

        for ev, kind, trans_list in event_trans_map[out_event]:

            constraint = ""

            if kind == 0:
                constraint = self._out_event_transition_0(impl, trans_list)
            elif kind == 1:
                constraint = self._out_event_transition_1(impl,
                                                          out_event,
                                                          ev,
                                                          trans_list)
            elif kind == 2:
                constraint = self._out_event_transition_2(impl,
                                                          out_event,
                                                          ev,
                                                          event_trans_map)

            elif kind == 3:
                constraint = self._out_event_transition_3(impl,
                                                          out_event,
                                                          ev,
                                                          event_trans_map)
            else:
                assert False  # Unreachable code.

            trans = self._substitute(self._templates.event_transition,
                                     {"event": event_name,
                                      "id": tid,
                                      "constraint": constraint})

            out_trans.append(trans)

            tid = tid + 1

        return "".join(out_trans)

    def _in_event_transition(self, impl, event_trans_map, in_event):
        """Generates transitions associated with input event."""

        in_trans = []
        event_name = self.event_name(in_event)

        tid = 1

        for ev, kind, trans_list in event_trans_map[in_event]:

            constraint = ""

            if kind == 0:
                constraint = self._in_event_transition_0(impl, trans_list)
            elif kind == 1:
                constraint = self._in_event_transition_1(impl,
                                                         in_event,
                                                         ev,
                                                         event_trans_map)
            else:
                assert False  # Unreachable code.

            trans = self._substitute(self._templates.event_transition,
                                     {"event": event_name,
                                      "id": tid,
                                      "constraint": constraint})

            in_trans.append(trans)

            tid = tid + 1

        return "".join(in_trans)

    def _tau_event_transition(self, impl, event_trans_map):
        """Generate transitions associated with tau event."""

        tau_trans = []
        event_name = self.event_name(self.tau_event())

        tid = 1

        for ev, kind, trans_list in event_trans_map[self.tau_event()]:

            constraint = ""

            if kind == 0:
                constraint = self._tau_event_transition_0(impl, trans_list)
            elif kind == 1:
                constraint = self._tau_event_transition_1(impl, ev)
            elif kind == 2:
                constraint = self._tau_event_transition_2(impl, ev, trans_list)
            elif kind == 3:
                constraint = self._tau_event_transition_3(impl, ev, trans_list)
            elif kind == 4:
                constraint = self._tau_event_transition_4(impl,
                                                          ev,
                                                          event_trans_map)
            else:
                assert False  # Unreachable code.

            trans = self._substitute(self._templates.event_transition,
                                     {"event": event_name,
                                      "id": tid,
                                      "constraint": constraint})

            tau_trans.append(trans)

            tid = tid + 1

        return "".join(tau_trans)

    def _stutter_transition(self, impl):
        """Generates constraint for stutter transition."""

        effects = self._non_mode_transition_constraint(impl)

        return self._substitute(self._templates.stutter_transition,
                                {"effects": effects})

    def _reactivation_transition(self, impl):
        """Generates constraint for reactivation transition."""

        # here activation and reactivation transition are handled
        # (see issue 536 for details)
        reactivation_trans_cases = []
        if impl.reactivation_mode_transitions:
            # collects by mode:
            mode_src_getter = operator.attrgetter("source_mode")
            transs = sorted(impl.reactivation_mode_transitions,
                            key=mode_src_getter)

            # groups by common source mode:
            reactivation_trans_cases = []
            for k, g in itertools.groupby(transs, mode_src_getter):
                cm_trans = sorted(g)

                mode_reactivation_trans_ors = []
                for trans in cm_trans:
                    __effects = self._reactivation_trans_effects(
                        impl, trans)
                    effects = (" & " + __effects
                               if __effects
                               else "")

                    mode_reactivation_trans = self._substitute(
                        self._templates.mode_reactivation_trans,
                        {"mode": self.mode_name(impl, trans.target_mode),
                         "effects": effects})
                    mode_reactivation_trans_ors.append(
                        "(" + mode_reactivation_trans + ")")

                if mode_reactivation_trans_ors:
                    reactivation_trans_case = self._substitute(
                        self._templates.reactivation_trans_case,
                        {"mode": self.mode_name(impl, k),
                         "mode_reactivation_trans_or":
                         " |\n         ".join(
                             mode_reactivation_trans_ors)})

                    reactivation_trans_cases.append(
                        reactivation_trans_case)

        # here default case is set
        if support.has_history(impl):
            effects = self._non_mode_transition_constraint(impl)
            reactivation_def_trans = (effects
                                      if effects
                                      else "TRUE")
        else:
            # default case for reactivation
            mode = self.mode_name(
                impl, support.get_initial_or_activation_mode(impl))

            effects = ""
            assigned_odps = self._assigned_output_data_port_dfl_effects(impl)
            if assigned_odps:
                effects = effects + " & " + assigned_odps
            data_sc = self._data_subcomponent_dfl_effects(impl)
            if data_sc:
                effects = effects + " & " + data_sc
            reactivation_def_trans = self._substitute(
                self._templates.reactivation_def_trans,
                {"mode": mode, "effects": effects})

        # result is composed
        r_trans = self._substitute(
            self._templates.reactivation_transition_act,
            {"reactivation_trans_cases":
             "\n      ".join(reactivation_trans_cases),
             "reactivation_def_trans": reactivation_def_trans})

        return r_trans

    def _out_event_transition_0(self, impl, trans_list):
        """Generates constraint for transition type 0 of output event."""

        constraint = []

        for t in trans_list:
            t_constraint = self._mode_transition_constraint(impl, t)
            constraint.append(t_constraint)

        return " | ".join(constraint)

    def _out_event_transition_1(self, impl,
                                out_event, trigger_event, trans_list):
        """Generates constraint for transition type 1 of output event."""

        out_trans = self._out_event_transition_trigger_event(impl,
                                                             out_event,
                                                             trigger_event)
        mode_trans_constraint = []
        for t in trans_list:
            t_constraint = self._mode_transition_constraint(impl, t)
            mode_trans_constraint.append(t_constraint)

        if mode_trans_constraint:
            if len(mode_trans_constraint) > 1:
                out_trans = "%s & (%s)" % (out_trans,
                                           " | ".join(mode_trans_constraint))
            else:
                out_trans = "%s & %s" % (out_trans,
                                         mode_trans_constraint[0])
        return out_trans

    def _out_event_transition_2(self, impl,
                                out_event, trigger_event, event_trans_map):
        """Generates constraint for transition type 2 of output event."""

        out_trans = self._out_event_transition_trigger_event(impl,
                                                             out_event,
                                                             trigger_event)
        (bhv1, pos) = self._find_event_behavior(event_trans_map,
                                                out_event,
                                                trigger_event,
                                                1)

        if bhv1 is not None:

            (e, k, mts) = bhv1
            assert mts  # Nonempty list of mode transitions.

            cando_mts = []

            for t in mts:
                tid = self._trans_id(impl, t)
                cando_t = self.cando_name(tid)
                cando_mts.append(cando_t)

            if len(cando_mts) > 1:
                out_trans = "%s & !(%s)" % (out_trans, " | ".join(cando_mts))
            else:
                out_trans = "%s & !%s" % (out_trans, cando_mts[0])

        mway_conns = support.get_out_to_in_connections_with_source(
            impl, trigger_event)

        # Filter for passive.
        mway_conns = [c for c in mway_conns if not c.passive]

        assert mway_conns  # Nonempty multiway connections.

        cando_mway = self._cando_or_target_event_port_connections(impl,
                                                                  mway_conns)

        effects = self._non_mode_transition_constraint(impl)

        out_trans = "%s & (%s) & %s" % (out_trans, cando_mway, effects)

        return out_trans

    def _out_event_transition_3(self, impl,
                                out_event, trigger_event, event_trans_map):
        """Generates constraint for transition type 3 of output event."""

        out_trans = self._out_event_transition_trigger_event(impl,
                                                             out_event,
                                                             trigger_event)
        (bhv1, pos1) = self._find_event_behavior(event_trans_map,
                                                 out_event,
                                                 trigger_event,
                                                 1)

        if bhv1 is not None:

            (e, k, mts) = bhv1
            assert mts  # Nonempty list of mode transitions.

            cando_mts = []

            for t in mts:
                tid = self._trans_id(impl, t)
                cando_t = self.cando_name(tid)
                cando_mts.append(cando_t)

            if len(cando_mts) > 1:
                out_trans = "%s & !(%s)" % (out_trans, " | ".join(cando_mts))
            else:
                out_trans = "%s & !%s" % (out_trans, cando_mts[0])

        (bhv2, pos2) = self._find_event_behavior(event_trans_map,
                                                 out_event,
                                                 trigger_event,
                                                 2)

        if bhv2 is not None:

            mway_conns = support.get_out_to_in_connections_with_source(
                impl, trigger_event)

            # Filter for passive.
            mway_conns = [c for c in mway_conns if not c.passive]

            assert mway_conns  # Nonempty multiway connections.

            cando_mway = self._cando_or_target_event_port_connections(
                impl,
                mway_conns)

            out_trans = "%s & !(%s)" % (out_trans, cando_mway)

        effects = self._non_mode_transition_constraint(impl)

        out_trans = "%s & %s" % (out_trans, effects)

        return out_trans

    def _out_event_transition_trigger_event(self,
                                            impl,
                                            out_event,
                                            trigger_event):
        """Generates constraint for trigger event of output event."""

        cando_name = self.cando_name(trigger_event)

        conns = support.get_connections_with_events(impl,
                                                    trigger_event,
                                                    out_event)
        modes = support.get_connections_modes(impl, conns)
        or_mode_conds = self._mode_cond(impl, modes)

        result = cando_name

        if or_mode_conds != self._templates.mode_cond_all:
            result = "%s & (%s)" % (result, or_mode_conds)

        return result

    def _in_event_transition_0(self, impl, trans_list):
        """Generates constraint for transition type 0 of input event."""

        constraint = []

        for t in trans_list:
            t_constraint = self._mode_transition_constraint(impl, t)
            constraint.append(t_constraint)

        return " | ".join(constraint)

    def _in_event_transition_1(self, impl,
                               in_event, trigger_event, event_trans_map):
        """Generates constraint for transition type 1 of input event."""

        in_trans = ""

        (bhv0, pos) = self._find_event_behavior(event_trans_map,
                                                in_event,
                                                trigger_event,
                                                0)

        if bhv0 is not None:

            (e, k, mts) = bhv0
            assert mts  # Nonempty list of mode transitions.

            cando_mts = []

            for t in mts:
                tid = self._trans_id(impl, t)
                cando_t = self.cando_name(tid)
                cando_mts.append(cando_t)

            in_trans = " | ".join(cando_mts)

            if len(cando_mts) > 1:
                in_trans = "!(" + in_trans + ")"
            else:
                in_trans = "!" + in_trans

        in_to_in_conns = support.get_in_to_in_connections_with_source(
            impl, in_event)
        sc_in_events = set([c.target_port for c in in_to_in_conns])

        assert sc_in_events

        candos = []

        for sc_ev in sorted(sc_in_events):

            cando_sc_in = self.cando_name(sc_ev)
            conns = [c
                     for c in in_to_in_conns
                     if c.target_port == sc_ev]

            modes = support.get_connections_modes(impl, conns)
            or_mode_conds = self._mode_cond(impl, modes)

            if or_mode_conds != self._templates.mode_cond_all:
                cando_sc_in = "(%s & %s)" % (cando_sc_in,
                                             or_mode_conds)
            candos.append(cando_sc_in)

        or_candos = ""

        if len(candos) > 1:
            or_candos = " | ".join(candos)
        else:
            or_candos = candos[0]

        if not in_trans:
            in_trans = or_candos
        else:
            in_trans = "%s & %s" % (in_trans, or_candos)

        effects = self._non_mode_transition_constraint(impl)

        in_trans = "%s & %s" % (in_trans, effects)

        return in_trans

    def _tau_event_transition_0(self, impl, trans_list):
        """Generates constraint for transition type 0 of tau event."""

        constraint = []

        for t in trans_list:
            t_constraint = self._mode_transition_constraint(impl, t)
            constraint.append(t_constraint)

        return " | ".join(constraint)

    def _tau_event_transition_1(self, impl, sc):
        """Generates constraint for transition type 1 of tau event."""

        sc_tau = sc + "." + self.tau_event()
        tau_trans = self.cando_name(sc_tau)

        effects = self._non_mode_transition_constraint(impl)

        tau_trans = "%s & %s" % (tau_trans, effects)

        return tau_trans

    def _tau_event_transition_2(self, impl, trigger_event, trans_list):
        """Generates constraint for transition type 2 of tau event."""

        tau_trans = self.cando_name(trigger_event)

        constraint = []

        for t in trans_list:
            t_constraint = self._mode_transition_constraint(impl, t)
            constraint.append(t_constraint)

        if len(constraint) > 1:
            tau_trans = "%s & (%s)" % (tau_trans, " | ".join(constraint))
        else:
            tau_trans = "%s & %s" % (tau_trans, constraint[0])

        return tau_trans

    def _tau_event_transition_3(self, impl, trigger_event, trans_list):
        """Generates constraint for transition type 3 of tau event."""

        cando_sub_out = self.cando_name(trigger_event)

        # Get connections (sc.OUT,OUT)
        conns = support.get_out_to_out_connections_with_source(
            impl, trigger_event)

        # Get modes where there is (sc.OUT,OUT)
        modes = support.get_connections_modes(impl, conns)

        if modes:
            or_mode_conds = self._mode_cond(impl, modes)
            cando_sub_out = "(%s & !(%s))" % (cando_sub_out,
                                              or_mode_conds)

        tau_trans = cando_sub_out

        constraint = []

        for t in trans_list:
            t_constraint = self._mode_transition_constraint(impl, t)
            constraint.append(t_constraint)

        if len(constraint) > 1:
            tau_trans = "%s & (%s)" % (tau_trans, " | ".join(constraint))
        else:
            tau_trans = "%s & %s" % (tau_trans, constraint[0])

        return tau_trans

    def _tau_event_transition_4(self, impl, trigger_event, event_trans_map):
        """Generates constraint for transition type 4 of tau event."""

        cando_sub_out = self.cando_name(trigger_event)

        cando = self._cando_tau_event_4_aux(impl,
                                            trigger_event,
                                            event_trans_map)

        if cando:
            cando = cando_sub_out + " & " + cando
        else:
            cando = cando_sub_out

        effects = self._non_mode_transition_constraint(impl)

        tau_trans = cando + " & " + effects

        return tau_trans

    def event_mode_transition_constraint(self, impl,
                                         event, pos, trans):
        """Generates event mode transition constraint.

        This method is used by guided simulation.
        """

        event_name = self.event_name(event)
        id_var = self.id_var_name()
        id_str = str(pos + 1)
        mode_trans = self._mode_transition_constraint(impl, trans)

        return self._substitute(self._templates.event_mode_transition,
                                {"event": event_name,
                                 "id_var": id_var,
                                 "id": id_str,
                                 "mode_trans": mode_trans})

    def _mode_transition_constraint(self, impl, trans):
        """Generates constraints for mode transition."""

        tid = self._trans_id(impl, trans)
        cando_name = self.cando_name(tid)
        next_mode = self.mode_name(impl, trans.target_mode)

        assigns = support.get_trans_assigns(impl, trans, complete=False)

        effects = ""

        odp_effects = self._assigned_output_data_port_upd_effects(impl,
                                                                  assigns)

        if odp_effects:
            effects = effects + " & " + odp_effects

        dsc_effects = self._data_subcomponent_upd_effects(impl,
                                                          trans,
                                                          assigns)

        if dsc_effects:
            effects = effects + " & " + dsc_effects

        mode_var = self.mode_var_name(self.mode_var())

        return self._substitute(self._templates.mode_transition,
                                {"cando_name": cando_name,
                                 "mode_var": mode_var,
                                 "next_mode": next_mode,
                                 "effects": effects})

    def _non_mode_transition_constraint(self, impl):
        """Generates constraints for non-mode transition"""

        effects = ""

        odp_effects = self._assigned_output_data_port_upd_effects(impl, {})

        if odp_effects:
            effects = effects + " & " + odp_effects

        dsc_effects = self._data_subcomponent_upd_effects(impl, None, {})

        if dsc_effects:
            effects = effects + " & " + dsc_effects

        mode_var = self.mode_var_name(self.mode_var())

        return self._substitute(self._templates.non_mode_transition,
                                {"mode_var": mode_var,
                                 "effects": effects})

    def _assigned_output_data_port_upd_effects(self, impl, assigns):
        """Generates update effects for assigned output data ports."""

        # assigned_odps = support.get_assigned_odp(impl)
        assigned_odps = self._assigned_odps(impl)

        odp_upds = []

        for odp in sorted(assigned_odps):

            odp_upd = ""

            if odp in assigns:
                upd_expr = self.expression(impl, assigns[odp])
                odp_upd = self._substitute(
                    self._templates.var_ass_transition,
                    {"var_name": self.data_name(odp),
                     "value": upd_expr})
            else:
                odp_upd = self._substitute(
                    self._templates.stutter_var_transition,
                    {"var_name": self.data_name(odp)})

            odp_upds.append(odp_upd)

        return " & ".join(odp_upds)

    def _data_subcomponent_upd_effects(self, impl, trans, assigns):
        """Generates update effects for data subcomponents."""

        dsc_upds = []

        if trans is None:
            assert not assigns  # No mode transition, no assignments.
            for dsc_name in sorted(impl.data_subcomponents):
                dsc = impl.data_subcomponents[dsc_name]
                dsc_upd = self._substitute(
                    self._templates.stutter_var_transition,
                    {"var_name": self.sc_name(dsc.name)})

                dsc_upds.append(dsc_upd)

        else:

            for dsc_name in sorted(impl.data_subcomponents):
                dsc = impl.data_subcomponents[dsc_name]
                dsc_upd = ""
                if dsc.name in assigns:
                    expr = self.expression(impl, assigns[dsc.name])
                    dsc_upd = self._substitute(
                        self._templates.var_ass_transition,
                        {"var_name": self.sc_name(dsc.name),
                         "value": expr})
                else:
                    dsc_modes = support.get_subcomponent_modes(impl, dsc)

                    if ((not trans.source_mode in dsc_modes) and
                            trans.target_mode in dsc_modes):

                        (dsc_def_val, is_undef) = self.default_value(impl,
                                                                     dsc)
                        if not is_undef:
                            dsc_upd = self._substitute(
                                self._templates.var_ass_transition,
                                {"var_name": self.sc_name(dsc.name),
                                 "value": dsc_def_val})

                    else:
                        dsc_upd = self._substitute(
                            self._templates.stutter_var_transition,
                            {"var_name": self.sc_name(dsc.name)})

                dsc_upds.append(dsc_upd)

        return " & ".join(dsc_upds)

    def _assigned_output_data_port_dfl_effects(self, impl):
        """Generates default effects for assigned output data ports."""

        impl_type = impl.get_referenced_component_type()
        # assigned_odps = support.get_assigned_odp(impl)
        assigned_odps = self._assigned_odps(impl)

        odp_upds = []

        for odp in sorted(assigned_odps):

            (odp_default_value, is_undef) = self.default_value(
                impl,
                impl_type.out_data_ports[odp])

            if not is_undef:
                odp_upd = self._substitute(
                    self._templates.var_ass_transition,
                    {"var_name": self.data_name(odp),
                     "value": odp_default_value})
                odp_upds.append(odp_upd)

        return " & ".join(odp_upds)

    def _data_subcomponent_dfl_effects(self, impl):
        """Generates default effects for data subcomponents."""

        dsc_upds = []

        for dsc_name in sorted(impl.data_subcomponents):
            dsc = impl.data_subcomponents[dsc_name]
            (dsc_def_val, is_undef) = self.default_value(impl,
                                                         dsc)
            if not is_undef:
                dsc_upd = self._substitute(
                    self._templates.var_ass_transition,
                    {"var_name": self.sc_name(dsc.name),
                     "value": dsc_def_val})
                dsc_upds.append(dsc_upd)

        return " & ".join(dsc_upds)

    def _reactivation_trans_effects(self, impl, trans):
        """Generates effects for data subcomponents and output data ports, for
        given activation transition. Transistion effects are
        considered, if not affected, value remains unchanged,
        effectively supporting history for reactivation transitions (see issue
        https://compassinternal.informatik.rwth-aachen.de/redmine/issues/536
        """

        ass = support.get_trans_assigns(impl, trans, complete=False)
        def_ass = support.get_assignable_data(impl)
        unass = [d for d in def_ass if d not in ass]

        upds = []
        for d in sorted(ass):
            _name = (self.sc_name(d)
                     if d in impl.data_subcomponents
                     else self.data_name(d))
            _value = self.expression(impl, ass[d])

            upd = self._substitute(
                self._templates.var_ass_transition,
                {"var_name": _name, "value": _value})

            upds.append(upd)

        for d in sorted(unass):
            _name = (self.sc_name(d)
                     if d in impl.data_subcomponents
                     else self.data_name(d))

            upd = None
            if d in impl.data_subcomponents:
                # check if was active or not
                dsc = impl.data_subcomponents[d]
                d_modes = support.get_subcomponent_modes(impl, dsc)
                if trans.source_mode not in d_modes:
                    # now activates dsc, reset to init value
                    def_val, undef = self.default_value(impl, dsc)
                    upd = self._substitute(
                        self._templates.var_ass_transition,
                        {"var_name": _name,
                         "value": def_val})

            if upd is None:
                # unchanged if were active, or for ODP
                upd = self._substitute(
                    self._templates.stutter_var_transition,
                    {"var_name": _name})

            upds.append(upd)

        return " & ".join(upds)

    def _cando_or_target_event_port_connections(self, impl, conns):
        """Generates cando of target event port of connections."""

        candos = []
        for c in conns:
            candos.append(self._cando_target_event_port_connection(impl, c))

        return " | ".join(candos)

    def _cando_target_event_port_connection(self, impl, conn):
        """Generates cando of target event port of connection."""

        cando_name = self.cando_name(conn.target_port)
        modes = support.get_connection_modes(impl, conn)
        or_mode_conds = self._mode_cond(impl, modes)
        cando = ""

        if or_mode_conds:
            cando = "%s & %s" % (cando_name, or_mode_conds)

        return cando

    def _resolved_definition_list(self, impl, event_trans_map):
        """Generates list of resolved definitions."""

        impl_type = impl.get_referenced_component_type()

        resolved_defs = []

        for out_event in sorted(impl_type.out_event_ports):

            resolved_name = self.resolved_name(out_event)
            event_name = self.event_name(out_event)

            bhvs1 = self._find_event_behaviors_with_kind(event_trans_map,
                                                         out_event,
                                                         1)
            bhvs2 = self._find_event_behaviors_with_kind(event_trans_map,
                                                         out_event,
                                                         2)
            ids = [pos + 1 for bhv, pos in bhvs1 + bhvs2]

            id_cond = ""

            if ids:
                id_eqs = []
                for i in ids:
                    id_eq = self._substitute(self._templates.id_equal,
                                             {"id": str(i)})
                    id_eqs.append(id_eq)

                id_cond = " | ".join(id_eqs)

                if len(ids) > 1:
                    id_cond = "(" + id_cond + ")"
            else:
                id_cond = "FALSE"

            resolved_def = self._substitute(
                self._templates.resolved_definition,
                {"resolved_name": resolved_name,
                 "event": event_name,
                 "id_cond": id_cond})

            resolved_defs.append(resolved_def)

        return "".join(resolved_defs)

    ### PARAMETERS ###########################################################

    def _control_subcomponent_active_definition_list(self, impl):
        """Generates (re)activation definitions for control subcomponents."""

        active_defs = []

        for ndsc_name in sorted(impl.non_data_subcomponents):
            csc = impl.non_data_subcomponents[ndsc_name]
            csc_name = self.sc_name(csc.name)
            csc_activated = self.sc_activated_name(csc.name)
            csc_reactivated = self.sc_reactivated_name(csc.name)
            modes = support.get_subcomponent_modes(impl, csc)
            mode_cond = self._mode_cond(impl, modes)

            active_def = self._substitute(
                self._templates.control_sc_activation_definition,
                {"sc_active_name": csc_activated,
                 "mode_cond": mode_cond})
            active_defs.append(active_def)

            reactive_def = self._substitute(
                self._templates.control_sc_reactivation_definition,
                {"sc_reactive_name": csc_reactivated,
                 "sc_name": csc_name,
                 "sc_active_name": csc_activated})
            active_defs.append(reactive_def)

        return "".join(active_defs)

    def _input_data_subcomponent_definition_list(self, impl):
        """Generates input data definitions."""

        impl_modes = support.get_component_impl_modes(impl)
        idp_defs = []

        for ndsc_name in sorted(impl.non_data_subcomponents):
            csc = impl.non_data_subcomponents[ndsc_name]
            csc_type = csc.get_referenced_component_type()
            csc_modes = support.get_subcomponent_modes(impl, csc)
            csc_impl = csc.get_referenced_component_implementation()

            for idp in sorted(csc_type.in_data_ports):

                idp_val = self.sc_val_in_data_name(csc.name, idp)
                full_idp_name = csc.name + "." + idp

                conn_values = []
                flow_values = []
                idp_modes = set([])

                conns = support.get_connections_with_target_data(
                    impl, full_idp_name)

                for c in conns:
                    idp_modes |= support.get_connection_modes(impl, c)
                    val = self._idp_connection_value(impl, c)
                    conn_values.append(val)

                flows = support.get_flows_with_target_data(impl,
                                                           full_idp_name)

                for f in flows:
                    idp_modes |= support.get_flow_modes(impl, f)
                    val = self._idp_flow_value(impl, f)
                    flow_values.append(val)

                (idp_default_value, is_undef) = self.default_value(
                    csc_impl,
                    csc_type.in_data_ports[idp])

                arbitrary_value = idp_default_value

                if is_undef and (not csc_modes <= idp_modes):
                    idp_type = csc_type.in_data_ports[idp].data_type
                    nondet_var_decl = self._substitute(
                        self._templates.idp_nondet_declaration,
                        {"name": idp_val,
                         "type": self.type_name(idp_type)})
                    idp_default_value = self._substitute(
                        self._templates.idp_nondet, {"name": idp_val})
                    idp_defs.append(nondet_var_decl)

#                 if (not is_undef) and (not (impl_modes <= idp_modes)):
#                     idp_type = csc_type.in_data_ports[idp].data_type
#                     type_name = self.type_name(idp_type)
#                     if type_name.startswith("{") and type_name.endswith("}"):
#                         def_enum_decl = self._substitute(
#                             self._templates.idp_enum_default_declaration,
#                             {"name": idp_val,
#                              "type": type_name})
#                         def_enum_invar = self._substitute(
#                             self._templates.idp_enum_default_invar,
#                             {"name": idp_val,
#                              "value": idp_default_value})
#                         idp_defs.append(def_enum_decl)
#                         idp_defs.append(def_enum_invar)
#                         idp_default_value = self._substitute(
#                             self._templates.idp_enum_default,
#                             {"name": idp_val})

                idp_def = ""

                if not flows and not conns:
                    idp_def = self._substitute(
                        self._templates.idp_value_definition_opt,
                        {"name": idp_val,
                         "default_value": idp_default_value})
                else:
                    if csc_modes <= idp_modes:
                        if impl_modes <= idp_modes:
                            idp_def = self._substitute(
                                self._templates.idp_value_definition_wo_default,
                                {"val_name": idp_val,
                                 "idp_connection_values": "".join(conn_values),
                                 "idp_flow_values": "".join(flow_values)})
                        else:
                            idp_def = self._substitute(
                                self._templates.idp_value_definition,
                                {"val_name": idp_val,
                                 "idp_connection_values": "".join(conn_values),
                                 "idp_flow_values": "".join(flow_values),
                                 "default_value": arbitrary_value})
                    else:
                        idp_def = self._substitute(
                            self._templates.idp_value_definition,
                            {"val_name": idp_val,
                             "idp_connection_values": "".join(conn_values),
                             "idp_flow_values": "".join(flow_values),
                             "default_value": idp_default_value})

                idp_defs.append(idp_def)

        return "".join(idp_defs)

    # TODO(IN): _idp_connection_value can be merged with the _odp one.
    def _idp_connection_value(self, impl, connection):
        """Generates connection value for input data port."""

        conn_modes = support.get_connection_modes(impl, connection)
        mode_cond = self._mode_cond(impl, conn_modes)
        source = self.data_name(connection.source_port)

        return self._substitute(self._templates.idp_connection_value,
                                {"mode_cond": mode_cond,
                                 "source": source})

    # TODO(IN): _idp_flow_value can be merged with the _idp one.
    def _idp_flow_value(self, impl, flow):
        """Generates flow value for input data port."""

        flow_modes = support.get_flow_modes(impl, flow)
        mode_cond = self._mode_cond(impl, flow_modes)

        # TODO(IN): Tidy up this dirty code.
        assert (isinstance(flow.expression,
                           expressions.BinaryOperatorTreeNode) and
                isinstance(flow.expression.left,
                           expressions.IdentifierTreeNode)), \
            "Flow has unexpected structure"

        flow_expr = self.expression(impl, flow.expression.right)

        return self._substitute(self._templates.idp_flow_value,
                                {"mode_cond": mode_cond,
                                 "flow_expr": flow_expr})

    ### DO ###################################################################

    def _do_control_subcomponent_event_definition_list(self,
                                                       impl,
                                                       event_trans_map):
        """Generates do definition list."""

        do_defs = []

        for ndsc_name in sorted(impl.non_data_subcomponents):
            csc = impl.non_data_subcomponents[ndsc_name]
            csc_type = csc.get_referenced_component_type()

            for oev in sorted(csc_type.out_event_ports):
                oev_do_def = self._sc_out_do_definition(event_trans_map,
                                                        csc,
                                                        oev)
                do_defs.append(oev_do_def)

            for iev in sorted(csc_type.in_event_ports):
                iev_do_def = self._sc_in_do_definition(impl,
                                                       event_trans_map,
                                                       csc,
                                                       iev)
                do_defs.append(iev_do_def)

            tau_do_def = self._sc_tau_do_definition(event_trans_map,
                                                    csc)
            do_defs.append(tau_do_def)

            if self.generate_timed_fsm:
                timed_do_def = self._sc_timed_do_definition(csc)
                do_defs.append(timed_do_def)

        return "".join(do_defs)

    def _sc_out_do_definition(self, event_trans_map, csc, oev):
        """Generates do definition for sc.OUT"""

        do_oev = self.sc_do_event_name(csc.name, oev)
        oev_fullname = csc.name + "." + oev
#         cando_oev = self.cando_name(oev_fullname)

        oev_bhvs = self._find_event_behaviors_with_trigger(event_trans_map,
                                                           oev_fullname)

        oev_conds = []
        for ev in sorted(oev_bhvs):

            ev_name = self.event_name(ev)
            assert oev_bhvs[ev]

            id_eqs = []
            for bhv, pos in oev_bhvs[ev]:
                id_eq = self._substitute(self._templates.id_equal,
                                         {"id": str(pos + 1)})
                id_eqs.append(id_eq)

            oev_ev_cond = ""

            if len(id_eqs) > 1:
                oev_ev_cond = "(%s & (%s))" % (ev_name, " | ".join(id_eqs))
            else:
                oev_ev_cond = "(%s & %s)" % (ev_name, id_eqs[0])

            oev_conds.append(oev_ev_cond)

        conds = ""

        if oev_conds:
            if len(oev_conds) > 1:
                conds = "(" + " | ".join(oev_conds) + ")"
            else:
                conds = oev_conds[0]
        else:
            conds = "FALSE"

        return self._substitute(self._templates.do_wo_cando_definition,
                                {"do_sc_event": do_oev,
                                 "conds": conds})

#         return self._substitute(self._templates.do_definition,
#                                 {"do_sc_event": do_oev,
#                                  "cando_sc_event": cando_oev,
#                                  "conds": conds})

    def _sc_in_do_definition(self, impl, event_trans_map, csc, iev):
        """Generates do definition for sc.IN"""

        do_iev = self.sc_do_event_name(csc.name, iev)
        iev_fullname = csc.name + "." + iev
        cando_iev = self.cando_name(iev_fullname)

        need_cando = False

        iev_bhvs = self._get_sc_in_event_involvement(impl,
                                                     event_trans_map,
                                                     iev_fullname)

        iev_conds = []
        for ev in sorted(iev_bhvs):

            ev_name = self.event_name(ev)
            assert iev_bhvs[ev]

            id_eqs = []
            for bhv, modes, pos in iev_bhvs[ev]:
                id_eq = self._substitute(self._templates.id_equal,
                                         {"id": str(pos + 1)})
                mode_cond = self._mode_cond(impl, modes)

                if mode_cond != self._templates.mode_cond_all:
                    id_eq = "%s & (%s)" % (id_eq, mode_cond)

                id_eqs.append(id_eq)

                (tev, k, ts) = bhv

                if (not need_cando) and (tev != self.tau_event() or k != 2):
                    need_cando = True

            iev_ev_cond = ""

            if len(id_eqs) > 1:
                iev_ev_cond = "(%s & (%s))" % (ev_name, " | ".join(id_eqs))
            else:
                iev_ev_cond = "(%s & %s)" % (ev_name, id_eqs[0])

            iev_conds.append(iev_ev_cond)

        conds = ""

        if iev_conds:
            if len(iev_conds) > 1:
                conds = "(" + " | ".join(iev_conds) + ")"
            else:
                conds = iev_conds[0]
        else:
            conds = "FALSE"

        if need_cando:
            return self._substitute(self._templates.do_definition,
                                    {"do_sc_event": do_iev,
                                     "cando_sc_event": cando_iev,
                                     "conds": conds})
        else:
            return self._substitute(self._templates.do_wo_cando_definition,
                                    {"do_sc_event": do_iev,
                                     "conds": conds})

    def _sc_tau_do_definition(self, event_trans_map, csc):
        """Generates do definition for sc.TAU"""

        do_tau = self.sc_do_event_name(csc.name, self.tau_event())
        # tau_fullname = csc.name + "." + self.tau_event()
        # cando_tau = self.cando_name(tau_fullname)

        sc_tau_id = -1

        if event_trans_map[self.tau_event()]:
            p = 0
            for tev, k, trans_list in event_trans_map[self.tau_event()]:
                if tev == csc.name:
                    sc_tau_id = p
                    break
                p = p + 1

        conds = ""

        if sc_tau_id == -1:
            conds = "FALSE"
        else:
            id_eq = self._substitute(self._templates.id_equal,
                                     {"id": str(sc_tau_id + 1)})
            conds = "(%s & %s)" % (self.event_name(self.tau_event()),
                                   id_eq)

        return self._substitute(self._templates.do_tau_definition,
                                {"do_sc_event": do_tau,
                                 "conds": conds})

    def _sc_timed_do_definition(self, csc):
        """Generates do definition for timed transition."""

        sc_delta = self.sc_delta_name(csc.name)
        sc_cando = self.cando_name(csc.name + "." + self.timed_event())

        return self._substitute(self._templates.do_timed_definition,
                                {"sc_delta": sc_delta,
                                 "sc_cando": sc_cando})

    def _get_sc_in_event_involvement(self, impl, event_trans_map, iev):
        """Gets all sc.IN involvement."""

        iev_involve = {}

        for ev in sorted(event_trans_map):
            p = 0
            for tev, k, trans_list in event_trans_map[ev]:

                if tev == iev:

                    assert ev == self.tau_event()  # Event must be tau.
                    assert k == 2  # Transition type must be 2.
                    assert trans_list  # Nonempty transistion list .

                    modes = set()
                    for t in trans_list:
                        modes.add(t.source_mode)

                    if ev in iev_involve:
                        iev_involve[ev].append(((tev, k, trans_list),
                                                modes,
                                                p))
                    else:
                        iev_involve[ev] = [((tev, k, trans_list), modes, p)]

                else:
                    conns = self._connections_with_events(impl, tev, iev)
                    if conns:
                        modes = support.get_connections_modes(impl, conns)
                        if ev in iev_involve:
                            iev_involve[ev].append(((tev, k, trans_list),
                                                    modes,
                                                    p))
                        else:
                            iev_involve[ev] = [((tev, k, trans_list),
                                                modes, p)]

                p = p + 1

        return iev_involve

    ### MAIN MODULE ##########################################################

    def _root_declaration(self, impl):
        """Generates declarations for root."""

        impl_type = impl.get_referenced_component_type()

        event_decls = []
        event_consts = []

        tau_event = self.event_name(self.tau_event())
        tau_event_decl = self._substitute(
            self._templates.root_event_declaration,
            {"event": tau_event})
        event_decls.append(tau_event_decl)

        tau_event_cando = self.cando_name(self.tau_event())
        tau_event_const = self._substitute(
            self._templates.root_event_constraint,
            {"event": tau_event,
             "cando_event": tau_event_cando})
        event_consts.append(tau_event_const)

        do_ievs = []

        for iev in sorted(impl_type.in_event_ports):
            do_iev = self.event_name(iev)
            iev_decl = self._substitute(self._templates.root_event_declaration,
                                        {"event": do_iev})
            event_decls.append(iev_decl)
            iev_cando = self.cando_name(iev)
            iev_const = self._substitute(self._templates.root_event_constraint,
                                         {"event": do_iev,
                                          "cando_event": iev_cando})
            event_consts.append(iev_const)
            do_ievs.append(do_iev)

        do_ievs_par = ""
        if do_ievs:
            do_ievs_par = ", " + ", ".join(do_ievs)

        do_oevs = []

        for oev in sorted(impl_type.out_event_ports):
            do_oev = self.event_name(oev)
            oev_decl = self._substitute(self._templates.root_event_declaration,
                                        {"event": do_oev})
            event_decls.append(oev_decl)
            oev_cando = self.cando_name(oev)
            oev_const = self._substitute(self._templates.root_event_constraint,
                                         {"event": do_oev,
                                          "cando_event": oev_cando})
            event_consts.append(oev_const)
            do_oevs.append(do_oev)

        do_oevs_par = ""
        if do_oevs:
            do_oevs_par = ", " + ", ".join(do_oevs)

        all_events = [tau_event] + do_ievs + do_oevs

        root_timed_transition = ""
        delta_par = ""
        if self.generate_timed_fsm:
            root_time_var = (self._templates.root_time_var
                             if self.generate_time_var
                             else "")
            root_time_fairness_alt = (self._templates.root_time_fairness_alt
                                      if self.generate_time_fairness_alt
                                      else "")

            root_timed_transition = self._substitute(
                self._templates.root_timed_transition,
                {"root_time_var": root_time_var,
                 "root_time_fairness_alt": root_time_fairness_alt})

            delta_par = ", " + self.delta_var()
            all_events.append(self.timed_event())

        at_most_one_const = self._substitute(
            self._templates.root_atmost_one_event_constraint,
            {"event_list": ', '.join(all_events)})

        idp_decls = []
        idp_consts = []
        idp_vars = []

        for idp in sorted(impl_type.in_data_ports):
            idp_var = self.data_name(idp)
            idp_type = self.type_name(impl_type.in_data_ports[idp].data_type)
            (idp_val, is_undef) = self.default_value(
                impl,
                impl_type.in_data_ports[idp])
            # MG: This assertion might fail since we relaxed the
            # condition on root
            #    components not having default values.
            # assert not is_undef

            idp_decl = self._substitute(
                self._templates.root_input_data_declaration,
                {"ivar": idp_var,
                 "type": idp_type})
            idp_decls.append(idp_decl)

            idp_const = self._substitute(
                self._templates.root_input_data_constraint,
                {"ivar": idp_var,
                 "val": idp_val})
            idp_consts.append(idp_const)

            idp_vars.append(idp_var)

        idp_vars_par = ""
        if idp_vars:
            idp_vars_par = ", " + ", ".join(idp_vars)

        impl_name = impl.name
        if impl.subpackage.package.name != "__default__":
            pack_name = impl.subpackage.package.name
            impl_name = pack_name + "." + impl_name

        root_comp = self._substitute(
            self._templates.root_component_declaration,
            {"name": self.impl_name(impl_name),
             "do_tau_event": tau_event,
             "do_in_event_list": do_ievs_par,
             "do_out_event_list": do_oevs_par,
             "in_data_val_list": idp_vars_par,
             "delta": delta_par})

        substitution = {}
        substitution["root_event_declaration_list"] = "".join(event_decls)
        substitution["root_input_data_declaration_list"] = "".join(idp_decls)
        substitution["root_event_constraint_list"] = "".join(event_consts)
#        substitution["root_input_data_constraint_list"] = "".join(idp_consts)
        substitution["root_input_data_constraint_list"] = ""
        substitution["root_timed_transition"] = root_timed_transition
        substitution["root_at_most_one_event_constraint"] = at_most_one_const
        substitution["root_component_declaration"] = root_comp
        substitution["pending_constants"] = self._pending_constants()

        return self._substitute(self._templates.main_module, substitution)

    def _pending_constants(self):
        """Generates pending constants."""

        result = ""
        pending = self._used_constants - self._decl_constants

        if pending:
            result = self._substitute(
                self._templates.pending_constants,
                {"constant_list": ", ".join(sorted(pending))})

        return result

    ### TIMED TRANSITION #####################################################

    def _timed_transition(self, impl):
        """Generates timed transition."""

        stutters = []

        for dsc in sorted(support.get_discrete_data_subcomponents(impl)):
            dsc_stutter = self._substitute(
                self._templates.stutter_var_transition,
                {"var_name": self.sc_name(dsc)})

            stutters.append(dsc_stutter)

        odp_stutter = self._assigned_output_data_port_upd_effects(impl, {})

        if odp_stutter:
            stutters.append(odp_stutter)

        effects = " & ".join([self._trajectories(impl)] + stutters)

        if effects:
            effects = " & " + effects

        return self._substitute(self._templates.timed_transition,
                                {"effects": effects})

    def _trajectories(self, impl):
        """Generates encoding for continous trajectories."""

        mode_trajectories = []
        for mode_name in sorted(impl.modes):
            mt = self._continuous_mode_trajectory(impl, mode_name)
            mode_trajectories.append(mt)

        stutters = []
        for dsc_name in sorted(support.get_continuous_data_subcomponent(impl)):
            dsc_stutter = self._substitute(
                self._templates.stutter_var_transition,
                {"var_name": self.sc_name(dsc_name)})
            stutters.append(dsc_stutter)

        mt_list = "".join(mode_trajectories)
        s_list = ""

        if stutters:
            s_list = " & ".join(stutters)
        else:
            s_list = "TRUE"

        return self._substitute(self._templates.trajectories,
                                {"mode_trajectory_list": mt_list,
                                 "stuttering_list": s_list})

    def _continuous_mode_trajectory(self, impl, mode_name):
        """Generates encoding for continuous trajectory in a mode."""

        trajectories = support.get_continuos_trajectories(impl, mode_name)
        updates = []
        for dsc_name in sorted(trajectories):
            expr = trajectories[dsc_name]
            if isinstance(expr, expressions.IdentifierTreeNode):
                # Stuttering.
                stutter = self._substitute(
                    self._templates.stutter_var_transition,
                    {"var_name": self.sc_name(dsc_name)})
                updates.append(stutter)

            else:  # This is on a trajectory
                dsc = impl.data_subcomponents[dsc_name]
                if expr is None:
                    # Clock
                    assert isinstance(dsc.data_type, nominalbehaviour.Clock)
                    tu_ratio = self._get_time_units_ratio(
                        expressions.TimeUnitType.NONE,
                        dsc.data_type.time_unit)
                    delta = ("%s * %s" % (self.delta_var(), tu_ratio)
                             if tu_ratio != "1"
                             else self.delta_var())
                    clock = self._substitute(
                        self._templates.traj_time_transition,
                        {"var_name": self.sc_name(dsc_name),
                         "delta": delta})

                    updates.append(clock)

                else:
                    # Continuous
                    assert isinstance(expr,
                                      expressions.TrajectoryEquationTreeNode)
                    tu_ratio = self._get_time_units_ratio(
                        expressions.TimeUnitType.NONE,
                        expr.time_unit)
                    delta = "%s * %s" % (
                        self.expression(impl, expr.trajectory_expression),
                        self.delta_var())
                    if tu_ratio != "1":
                        delta = delta + " * %s" % tu_ratio

                    cont = self._substitute(
                        self._templates.traj_time_transition,
                        {"var_name": self.sc_name(dsc_name),
                         "delta": delta})
                    updates.append(cont)

        if updates:
            effects = " & ".join(updates)
        else:
            effects = "TRUE"

        return self._substitute(self._templates.mode_trajectory,
                                {"mode": self.mode_name(impl, mode_name),
                                 "effects": effects})

    ### GENERICS #############################################################

    def _mode_cond(self, impl, modes):
        """Generates mode condition."""

        assert modes  # Nonempty set of modes.
        if support.are_modes_complete(impl, modes):
            return self._templates.mode_cond_all
        elif len(modes) == 1:
            mode_var = self.mode_var_name(self.mode_var())
            mode_name = self.mode_name(impl, modes.pop())
            return self._substitute(self._templates.mode_cond_single,
                                    {"mode_var": mode_var,
                                     "mode": mode_name})
        else:
            mode_var = self.mode_var_name(self.mode_var())
            mode_names = map(lambda m: self.mode_name(impl, m), sorted(modes))
            return self._substitute(self._templates.mode_cond,
                                    {"mode_var": mode_var,
                                     "modes": ", ".join(mode_names)})

    # TODO(IN): Added to support legacy.
    def get_in_modes(self, impl, modes):
        """Generates mode condition."""

        assert modes  # Nonempty set of nodes.
        in_modes = self._mode_cond(impl, modes)

        if in_modes == self._templates.mode_cond_all:
            return ""

        return in_modes

    # TODO(IN): Added due to legacy.
    def get_mode_definition(self, impl):
        """Gets mode definition (declaration)"""

        modes = support.get_component_impl_modes(impl)

        return (self.mode_var(),
                map(lambda m: self.mode_name(impl, m), sorted(modes)))

    # TODO(IN): Added due to legacy.
    def mode_definition(self, impl):
        """Gets declaration of mode."""

        modes_info = self.get_mode_definition(impl)
        return self._substitute(self._templates.mode_declaration,
                                {"mode_list": ", ".join(modes_info[1])})

    def _find_event_behavior(self, event_trans_map,
                             event, trigger_event, kind):
        """Finds behavior associated with an event."""

        bhv = None

        if event in event_trans_map:
            p = 0
            for ev, k, trans_list in event_trans_map[event]:
                if trigger_event == ev and kind == k:
                    return ((ev, k, trans_list), p)
                p = p + 1

        return (bhv, -1)

    def _find_event_behaviors_with_kind(self, event_trans_map,
                                        event, kind):
        """Finds behavior associated with an event."""

        bhvs = []
        if event in event_trans_map:
            p = 0
            for ev, k, trans_list in event_trans_map[event]:
                if kind == k:
                    bhvs.append(((ev, k, trans_list), p))
                p = p + 1

        return bhvs

    def _find_event_behaviors_with_trigger(self, event_trans_map,
                                           triggered_event):
        """Finds behavior associated with a triggered event."""

        bhvs = {}
        for ev in sorted(event_trans_map):
            p = 0
            for tev, k, trans_list in event_trans_map[ev]:
                if tev == triggered_event:
                    if ev in bhvs:
                        bhvs[ev].append(((tev, k, trans_list), p))
                    else:
                        bhvs[ev] = [((tev, k, trans_list), p)]
                p = p + 1

        return bhvs

    def type_name(self, a_type):
        """Encodes SLIM data types into SMV types."""

        if isinstance(a_type, nominalbehaviour.Bool):
            return self._templates.BOOL_TYPE
        elif isinstance(a_type, nominalbehaviour.Enum):
            consts = map(self.const_name, a_type.values)
            self._decl_constants |= set(consts)
            return "{" + ", ".join(consts) + "}"
        elif isinstance(a_type, nominalbehaviour.IntRange):
            return "%d .. %d" % tuple(a_type.get_bounds())
        elif isinstance(a_type, nominalbehaviour.Int):
            word_type = ""
            if options.SMV_WORD_SIGNED:
                word_type = self._templates.SIGNED_WORD_TYPE
            else:
                word_type = self._templates.UNSIGNED_WORD_TYPE
            return "%s[%d]" % (word_type, options.SMV_WORD_SIZE)
        elif isinstance(a_type, nominalbehaviour.Real):
            return self._templates.REAL_TYPE
        elif isinstance(a_type, nominalbehaviour.Continuous):
            return self._templates.REAL_TYPE
        elif isinstance(a_type, nominalbehaviour.Clock):
            return self._templates.REAL_TYPE
        else:
            # Not supported type.
            assert False, str(type(a_type)) + str(a_type)
        return

    def expression(self, impl, expr, time_unit_dest=None):
        """Encodes SLIM expression into SMV expressions."""

        # 1. Base cases
        if expr is None:
            return ""

        # this is to handle possible first call
        if time_unit_dest is None:
            time_unit_dest = self.get_preferred_time_unit_in_expr(impl, expr)

        # 1.1 identifier
        if isinstance(expr, expressions.IdentifierTreeNode):

            # for mode and error special fields and values:
            _map = {nominalbehaviour.NodeType.MODE_FIELD: self.mode_var_name,
                    nominalbehaviour.NodeType.MODE_VALUE: (
                        self.mode_lit_name
                        if not impl.is_error_impl()
                        else self.error_lit_name),
                    nominalbehaviour.NodeType.ERROR_FIELD: self.error_var_name,
                    nominalbehaviour.NodeType.ERROR_VALUE: self.error_lit_name,
                    }
            if expr.node_type in _map:
                return _map[expr.node_type](expr.name)

            # check if port or dsc
            names = expr.name.split(".")
            # sc1.sc2.sc3 (can be empty string):
            scs_name = ".".join(names[:-1])
            data_name = names[-1]  # name of data referring

            impl = support.get_subcomponent_impl_from_name(impl, scs_name)
            ctype = impl.get_referenced_component_type()
            for cm, foo in ((ctype.in_data_ports, self.data_name),
                            (ctype.out_data_ports, self.data_name),
                            (ctype.in_event_ports, self.event_name),
                            (ctype.out_event_ports, self.event_name),
                            (impl.data_subcomponents, self.sc_name),):
                if data_name in cm:
                    _res = foo(expr.name)

                    # check if it is a clock, and normalize accordingly
                    member = cm[data_name]
                    if (isinstance(member,
                                   nominalbehaviour.DataSubcomponent) and
                            isinstance(member.data_type,
                                       nominalbehaviour.Clock)):
                        # applies rate
                        tu_from = member.data_type.time_unit
                        tu_ratio = self._get_time_units_ratio(tu_from,
                                                              time_unit_dest)
                        if tu_ratio != "1":
                            _res = "%s * %s" % (_res, tu_ratio)

                    return _res
            
            # WARNING: here we do not check enum constant literals,
            # we assume they are correct
            assert expr.node_type == nominalbehaviour.NodeType.ENUM_VALUE, \
                "Expected enum literal (ticket:35), found %s type:%s" \
                % (expr.name, str(expr.node_type))
            assert not scs_name

            cons = self.const_name(expr.name)
            # add the constant to the set of pending constants
            self._used_constants.add(cons)
            return cons

        # 1.2 number
        if isinstance(expr, expressions.NumberTreeNode):
            if nominalbehaviour.NodeType.INT == expr.node_type:
                val = int(expr.number)
                signed = ("0sd"
                          if options.SMV_WORD_SIGNED
                          else "0d")
                word_prefix = "%s%d_" % (signed, options.SMV_WORD_SIZE)
                if val < 0:
                    return "-%s%d" % (word_prefix, abs(val))
                return "%s%d" % (word_prefix, val)

            if expr.node_type in (nominalbehaviour.NodeType.REAL,
                                  nominalbehaviour.NodeType.CONTINUOUS,
                                  nominalbehaviour.NodeType.CLOCK):
                return expr.number

            assert nominalbehaviour.NodeType.INT_RANGE == expr.node_type, \
                "Found unexpected number type: %s" % str(expr.node_type)
            return expr.number

        # 1.3 boolean constant
        if isinstance(expr, expressions.BooleanTreeNode):
            if expr.value:
                return "TRUE"
            return "FALSE"

        # This is for non-regression of ticket:189
        if isinstance(expr, nominalbehaviour.DataSubcomponent):
            raise NotImplementedError("FOUND DataSubcomponent in expression: "
                                      "(ticket:189)" + str(expr) +
                                      " (%s)" % expr.name)

        # 1.4 Trajectories
        if isinstance(expr, expressions.TrajectoryEquationTreeNode):
            assert isinstance(expr.variable, expressions.IdentifierTreeNode), \
                "Unexpected type for var in trajectory (ticket:190): " \
                + str(expr.variable)
            raise NotImplementedError("Trajectories are not supported in this "
                                      "point, as they are expected not to "
                                      "occur here")

        # 1.99 unsupported leaves
        if isinstance(expr, (expressions.TrajectoryEquationTreeNode,
                             expressions.TrajectoryExpressionTreeNode,)):
            raise NotImplementedError(
                "Unsupported node type in expression: %s" % expr)

        # 2. Induction step
        if isinstance(expr, expressions.UnaryOperatorTreeNode):
            # check if it is a time-unit operator
            if support.is_expression_timed_unit_op(expr):
                # Time unit operators are handled here
                tu_from = expressions.TimeUnitType.from_operator(expr.operator)
                tu_ratio = self._get_time_units_ratio(tu_from, time_unit_dest)
                sub = self.expression(impl, expr.subexpression, tu_from)

                return ("%s * %s" % (sub, tu_ratio)
                        if tu_ratio != "1" and expr.operator in
                        (expressions.OperatorType.REAL_TOMSEC,
                         expressions.OperatorType.REAL_TOSEC,
                         expressions.OperatorType.REAL_TOMIN,
                         expressions.OperatorType.REAL_TOHOUR,
                         expressions.OperatorType.REAL_TODAY)
                        else sub)

            # fallback to numeric unary operators
            ops = {
                expressions.OperatorType.NOT: "!%s",
                expressions.OperatorType.UNARY_MINUS: "-%s",
            }
            return ("(%s)" % (ops[expr.operator] %
                              self.expression(impl, expr.subexpression,
                                              time_unit_dest)))

        if isinstance(expr, expressions.BinaryOperatorTreeNode):
            ops = {
                expressions.OperatorType.LESS: "%s < %s",
                expressions.OperatorType.LESSOREQUAL: "%s <= %s",
                expressions.OperatorType.GREATER: "%s > %s",
                expressions.OperatorType.GREATEROREQUAL: "%s >= %s",
                expressions.OperatorType.EQUAL: "%s = %s",
                expressions.OperatorType.NOTEQUAL: "%s != %s",
                expressions.OperatorType.PLUS: "%s + %s",
                expressions.OperatorType.MINUS: "%s - %s",
                expressions.OperatorType.TIMES: "%s * %s",
                expressions.OperatorType.DIVIDE: "%s / %s",
                expressions.OperatorType.AND: "%s & %s",
                expressions.OperatorType.ASSIGN: "next(%s) = %s",
                expressions.OperatorType.OR: "%s | %s",
                expressions.OperatorType.XOR: "%s xor %s",
                expressions.OperatorType.XNOR: "%s xnor %s",
                expressions.OperatorType.IFF: "%s <-> %s",
                expressions.OperatorType.IMP: "%s -> %s",
                expressions.OperatorType.MOD: "%s mod %s",
                }

            return "(%s)" % (ops[expr.operator] %
                             (self.expression(impl, expr.left,
                                              time_unit_dest),
                              self.expression(impl, expr.right,
                                              time_unit_dest)))

        if isinstance(expr, expressions.CaseExpressionTreeNode):
            if not expr.cases:
                return self.expression(impl, expr.default_expression,
                                       time_unit_dest)
            return """(case
 %s
 TRUE : %s;
 esac)""" % (os.linesep.join(["%s : %s; " %
                              (self.expression(impl, cond, time_unit_dest),
                               self.expression(impl, then, time_unit_dest))
                              for cond, then in expr.cases]),
             self.expression(impl, expr.default_expression, time_unit_dest))

        raise Exception("we should never reach this point. Expr is:" +
                        str(expr))

    def get_preferred_time_unit_in_expr(self, impl, expr):
        """Returns the preferred time_unit to be used when
        converting the expression. This is internally used to compute
        a good time_unit to normalize TUs
        """
        res = expressions.TimeUnitType.NONE
        _dict = collections.defaultdict(int)
        self._get_clock_tu_in_expr_aux(impl, expr, _dict)
        if _dict:
            best = max(_dict.items(), key=operator.itemgetter(1))
            res = best[0]

        return res

    def default_value(self, impl, data):
        """Generates default value for data port or data subcomponent.

        Given a data port of a data subcomponent, this method returns
        a pair consisting of a value and a boolean flag indicating
        if the default value is defined. If the default value is
        undefined, then the first element of the pair is the default
        value itself, otherwise it is an arbitrary constant of the
        type of the data port of the data subcomponent.

        If datatype is clock, (0, False) is returned.
        """

        assert isinstance(data, nominalbehaviour.DataPort) or \
            isinstance(data, nominalbehaviour.DataSubcomponent)

        if data.default_value is not None:
            val = self.expression(impl, data.default_value)
            return (val, False)
        else:
            arb_val = self.expression(impl,
                                      self._constant_of_type(data.data_type))

            # clocks has no default by construction, so (0, False) is
            # returned to indicate a default value exist
            return (arb_val, not isinstance(data.data_type,
                                            nominalbehaviour.Clock))

    # TODO(IN): Tidy up this dirty method.
    def _constant_of_type(self, a_type):
        """Generate an arbitrary constant of a given type."""

        if isinstance(a_type, nominalbehaviour.Bool):
            return expressions.BooleanTreeNode(False, -1, -1)

        elif isinstance(a_type, nominalbehaviour.Enum):
            idf = expressions.IdentifierTreeNode(a_type.values[0], -1, -1)
            idf.node_type = expressions.NodeType.ENUM_VALUE
            return idf

        elif isinstance(a_type, nominalbehaviour.IntRange):
            num = expressions.NumberTreeNode("%d" % a_type.get_bounds()[0],
                                             expressions.TimeUnitType.NONE,
                                             -1, -1)
            num.node_type = expressions.NodeType.INT_RANGE
            return num
        try:
            _cnst, _type = {nominalbehaviour.Int:
                            ("0", expressions.NodeType.INT),
                            nominalbehaviour.Real:
                            ("0", expressions.NodeType.REAL),
                            nominalbehaviour.Continuous:
                            ("0", expressions.NodeType.CONTINUOUS),
                            nominalbehaviour.Clock:
                            ("0", expressions.NodeType.CLOCK),
                            }[a_type.__class__]
            num = expressions.NumberTreeNode(_cnst,
                                             expressions.TimeUnitType.NONE,
                                             -1, -1)
            num.node_type = _type
            return num

        except KeyError:
            # not implemented type
            raise NotImplementedError("Unsupported type:" + str(a_type))
        return

    # TODO(IN): Added due to legacy.
    def smv_const_to_slim_const(self, const):
        """Converts SMV constants to SLIM constants."""

        try:
            # Try evaluate the input constant in an empty environment.
            # Integer, float, and fixed constants can be evaluated
            # in such a way.
            eval(const)
            return const
        except (NameError, SyntaxError):
            pass

        # Rational constant in the form f'q/d?
        if const.startswith("f'") and "/" in const[2:]:
            return const[2:]

        # Word constant.
        m = self._re_word.match(const)
        if m:
            return "".join(m.group(1, 2))  # to join possible '-'.

        const_map = {
            "TRUE": "True",
            "FALSE": "False",
            }

        if const in const_map:
            return const_map[const]

        # Identifier (enumerator).
        return self.smv_id_to_slim_id(const)

    # TODO(IN): Added due to legacy.
    def smv_id_to_slim_id(self, name, can_be_enum_literal=True):
        """Converts SMV identifier to SLIM identifer."""

        split_name = name.split(".")

        # Check if prefix is the "root" name and remove it.
        if split_name[0] == self.root_var():
            split_name = split_name[1:]
        elif not can_be_enum_literal:
            return None  # id not beginning with root is an artifact in main.

        def _remove_prefix(simple_name,
                           debug=options.SMV_ENABLE_INTERNAL_SLIM_NAMES):
            """Removes a prefix, if any, from a simple name."""

            # Exact names to be kept as they are.
            if simple_name in (self._templates.MODE_VAR,):
                return simple_name

            # Prefix to keep for debugging.
            if debug:
                for pre in (self._templates.CANDO_PREFIX, ):
                    if simple_name.startswith(pre):
                        return simple_name

            # The following prefixes are possibly removed.
            for pre in (self._templates.EVENT_PREFIX,
                        self._templates.DATA_PREFIX,
                        self._templates.SUBCOMP_PREFIX,
                        self._templates.CONST_PREFIX):
                if simple_name.startswith(pre):
                    return simple_name[len(pre):]

            # For mode and error literals.
            for pre in (self._templates.MODE_LIT_PREFIX,
                        self._templates.ERROR_LIT_PREFIX):
                if simple_name.startswith(pre):
                    return "%s:%s" % (pre[:-1], simple_name[len(pre):])

            # Otherwise.
            return None

        for i in range(len(split_name)):
            n = _remove_prefix(split_name[i])
            if n is None:
                return None
            split_name[i] = n

        # Now check if it is a tau which is artificial but looks like
        # a normal event.
        if split_name[-1] == self.tau_event():
            return None

        return ".".join(split_name)

    def _assigned_odps(self, impl):
        impl_name = impl.name
        if impl.subpackage.package.name != "__default__":
            impl_name = impl.subpackage.package.name + "." + impl_name

        if impl_name in self._assigned_odps_cache:
            return self._assigned_odps_cache[impl_name]
        else:
            odps = support.get_assigned_odp(impl)
            self._assigned_odps_cache[impl_name] = odps
            return odps

    def _trans_id(self, impl, trans):
        if trans in self._trans_id_cache:
            return self._trans_id_cache[trans]
        else:
            tid = support.get_trans_id(impl, trans)
            self._trans_id_cache[trans] = tid
            return tid

    def _connections_with_events(self, impl, src_event, tgt_event):
        if impl in self._port_conns_cache:
            pass
        else:
            impl_conns = {}
            for c in impl.event_port_connections:
                if c.source_port in impl_conns:
                    src_conns = impl_conns[c.source_port]
                    if c.target_port in src_conns:
                        src_conns[c.target_port].append(c)
                    else:
                        src_conns[c.target_port] = [c]
                else:
                    src_conns = {c.target_port: [c]}
                    impl_conns[c.source_port] = src_conns

            self._port_conns_cache[impl] = impl_conns

        impl_conns = self._port_conns_cache[impl]
        if src_event in impl_conns:
            src_conns = impl_conns[src_event]
            if tgt_event in src_conns:
                return src_conns[tgt_event]
            else:
                return []
        else:
            return []

    def _get_time_units_ratio(self, tu_from, tu_to):
        """Get a (possibly rational) number (as string) that represent the
        constant ratio to be applied to time-unit tu_from to get
        corresponding time-unit tu_to. E.g. (MIN, SEC) returns "60",
        while (HOUR, DAY) returns "f'1/24".

        If TimeUnitType.NONE is passed as tu_from or tu_to, the
        default Time Unit is used.
        """
        if tu_from == expressions.TimeUnitType.NONE or tu_from is None:
            tu_from = self.def_base_unit

        if tu_to == expressions.TimeUnitType.NONE or tu_to is None:
            tu_to = self.def_base_unit

        return self._tu_map_ratio[tu_from][tu_to]

    def _get_clock_tu_in_expr_aux(self, impl, expr, _dict):
        """Fills a map Encodes SLIM expression into SMV expressions.
        map entries are: {time_unit: count}
        """

        # 1. Base cases
        if expr is None:
            return

        # 1.1 identifier
        if isinstance(expr, expressions.IdentifierTreeNode):
            names = expr.name.split(".")

            # sc1.sc2.sc3 (can be empty string):
            scs_name = ".".join(names[:-1])

            data_name = names[-1]  # name of data referring
            impl = support.get_subcomponent_impl_from_name(impl, scs_name)
            ctype = impl.get_referenced_component_type()
            for cm in (ctype.in_data_ports,
                       ctype.out_data_ports,
                       ctype.in_event_ports,
                       ctype.out_event_ports,
                       impl.data_subcomponents, ):
                if data_name in cm:
                    # check if it is a clock, and normalize accordingly
                    member = cm[data_name]
                    if (isinstance(member,
                                   nominalbehaviour.DataSubcomponent) and
                            isinstance(member.data_type,
                                       nominalbehaviour.Clock)):
                        _dict[member.data_type.time_unit] += 1
                    break

            return

        # 1.2 number, 1.3 boolean constant, 1.4 Trajectoies
        if isinstance(expr, (expressions.NumberTreeNode,
                             expressions.BooleanTreeNode,
                             nominalbehaviour.DataSubcomponent,
                             expressions.TrajectoryEquationTreeNode)):
            return

        # 1.99 unsupported leaves
        if isinstance(expr, (expressions.TrajectoryEquationTreeNode,
                             expressions.TrajectoryExpressionTreeNode,)):
            raise NotImplementedError(
                "Unsupported node type in expression: %s" % expr)

        # 2. Induction step
        if isinstance(expr, expressions.UnaryOperatorTreeNode):
            self._get_clock_tu_in_expr_aux(impl, expr.subexpression,
                                           _dict)

        elif isinstance(expr, expressions.BinaryOperatorTreeNode):
            self._get_clock_tu_in_expr_aux(impl, expr.left, _dict)
            self._get_clock_tu_in_expr_aux(impl, expr.right, _dict)

        elif isinstance(expr, expressions.CaseExpressionTreeNode):
            if expr.cases:
                for cond, then in expr.cases:
                    self._get_clock_tu_in_expr_aux(impl, cond, _dict)
                    self._get_clock_tu_in_expr_aux(impl, then, _dict)

            self._get_clock_tu_in_expr_aux(impl, expr.default_expression,
                                           _dict)

        return

    pass  # end of class
