# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================


from slim_compiler.src.smv import support
from slim_compiler.src.dataStructure import nominalbehaviour
from slim_compiler.src.dataStructure import expressions
from slim_compiler.src.model_extension import constants
import itertools


# These constants can be used to choose the traversal ordering
WALKER_PRE_ORDER, WALKER_POST_ORDER = range(2)


class Walker (object):
    """This is the base class for walkers that can traverse the
    compile info. In particular traversal is carried out through
    non-data subcomponents, in post-order (using pre-order is
    possible). Derived class are expected to derive from Walker
    class and implement method _action"""
    def __init__(self, order=WALKER_POST_ORDER):
        self.__curr_path = []
        self.__order = order
        return

    def walk(self, root_impl):
        """Post order traversal is implemented"""
        self.__curr_path = [(root_impl, "")]
        try:
            self.__walk_aux(root_impl, None)
        except StopIteration:
            pass

    def __walk_aux(self, impl, parent):
        """Recursive service for walk"""

        _type = impl.get_referenced_component_type()

        if self.__order == WALKER_PRE_ORDER:
            self._action(_type, impl, parent, tuple(self.__curr_path))

        for sc_name in sorted(impl.non_data_subcomponents):
            sc = impl.non_data_subcomponents[sc_name]
            sc_impl = sc.get_referenced_component_implementation()

            self.__curr_path.append((sc_impl, sc_name))
            self.__walk_aux(sc_impl, impl)
            del self.__curr_path[-1]

        if self.__order == WALKER_POST_ORDER:
            self._action(_type, impl, parent, tuple(self.__curr_path))

        return

    def _action(self, comp_type, comp_impl, parent_impl, inst_full_path):
        """This is the method to be implemented from derived
           classes.  It gets the currently visited component
           implementation and type, and a tuple of pairs (impl,
           inst-name) containing the complete path up to the actual
           non-data subcomponent instance.

           If the walker wants
           to stop the iteration, it can raise the StopIteration
           exception."""

        return

    pass  # end of class
# ----------------------------------------------------------------------


class ErrorEventProbWalker (Walker):
    """This walker is used to collect info about the mapping between
    error events and probabilities"""
    def __init__(self, converter):
        super(ErrorEventProbWalker, self).__init__()

        self.converter = converter
        self.result = []
        return

    def _action(self, comp_type, comp_impl, parent_impl, inst_full_path):
        # searches for the error event data ports
        ls = []
        for oep_name in sorted(comp_type.out_event_ports):
            oep = comp_type.out_event_ports[oep_name]
            if isinstance(oep, nominalbehaviour.ErrorEventPort):
                ls.append((oep.name, oep.get_parameter()))

        # re-adapt by adding prefix from inst_full_path
        self.result += [(self.converter.event_name(
            ".".join([n for i, n in inst_full_path[1:]] + [ev_name])),
            par)
            for ev_name, par in ls if par is not None]
        return

    pass  # end of class
# ----------------------------------------------------------------------


class SmvDataTypeWalker (Walker):
    """This is used to collect data types of data ports, data
    sub-components and compoenent modes. The result is a map of SMV
    data/map names and corresponding SMV type that can be used to
    generate the language section of the RAT project"""

    def __init__(self, converter):
        super(SmvDataTypeWalker, self).__init__()

        self.converter = converter
        self.result = {}
        return

    def _action(self, comp_type, comp_impl, parent_impl, inst_full_path):
        # collects data ports, data components and modes

        # data ports and data subcomponents
        for _map in (comp_type.in_data_ports, comp_type.out_data_ports,
                     comp_impl.data_subcomponents):
            for pn in sorted(_map):
                p = _map[pn]
                data_path = (".".join((n for i, n in inst_full_path[1:])) +
                             "." + pn)
                self.result[self.converter.data_name(data_path)] = \
                    self.converter.type_name(p.data_type)

        # modes are converted only if not optimized away
        minfo = self.converter.get_mode_definition(comp_impl)
        if minfo:
            mode_var_path = (".".join(map(self.converter.sc_name,
                                          (n for i, n in inst_full_path[1:])) +
                                      [minfo[0]]))
            self.result[mode_var_path] = "{ %s }" % ", ".join(minfo[1])
        return

    pass  # end of class
# ----------------------------------------------------------------------


class InfiniteDomainFinderWalker (Walker):
    """This is used to search for infinite domain data into data
    sub-components. The results are:

    1. a boolean value which is set to True if there exists at
       least one data sub-component which requires time-related
       structure to represent the hybrid aspects.

    2. a boolean value which is set to True if there exists at
    least one data sub-component or data port whose domain is not
    finite (like reals, clock, and continuous)."""

    def __init__(self):
        super(InfiniteDomainFinderWalker, self).__init__()

        self.has_timed = False
        self.has_infinite = False
        return

    def _action(self, comp_type, comp_impl, parent_impl, inst_full_path):

        # searches in data ports
        for dp in itertools.chain(comp_type.in_data_ports.itervalues(),
                                  comp_type.out_data_ports.itervalues()):
            assert isinstance(dp, nominalbehaviour.DataPort), \
                "Unexpected type for data port"
            if isinstance(dp.data_type, (nominalbehaviour.Real,)):
                self.has_infinite = True
                # stops the iteration, continuing the search for timed
                # in data-subcomponents
                break

        # searches in data subcomponents
        for name, dsc in comp_impl.data_subcomponents.iteritems():
            if isinstance(dsc.data_type,
                          (nominalbehaviour.Clock,
                           nominalbehaviour.Continuous)):
                self.has_timed = True
                self.has_infinite = True
                # stops the iteration, as all needed information have
                # been found
                raise StopIteration

            if isinstance(dsc.data_type, (nominalbehaviour.Real,)):
                self.has_infinite = True
                # continues the search, to see if there are timed
                # components

        # the search continues (in different implementations)
        return

    def has_found_clock_data(self):
        """After the walk, this method will return True if any
        time-related data-sub-component has been found"""
        return self.has_timed

    def has_found_infinite_data(self):
        """After the walk, this method will return True if any
        infinite-domain data-sub-component has been found"""
        return self.has_infinite

    pass  # end of class
# ----------------------------------------------------------------------


class ClockCollectorWalker (Walker):
    """This is used to collect the list of all clock variables.

    In particular this can be used when checking clock divergence.
    """

    def __init__(self):
        super(ClockCollectorWalker, self).__init__()
        self.clocks = []

    def _action(self, comp_type, comp_impl, parent_impl, inst_full_path):

        # searches in data subcomponents
        for name, dsc in comp_impl.data_subcomponents.iteritems():
            if isinstance(dsc.data_type, nominalbehaviour.Clock):
                full_name = ".".join((n for i, n in inst_full_path[1:]))
                clock_full_name = (full_name + "." + name
                                   if full_name
                                   else name)

                self.clocks.append((clock_full_name, dsc,
                                    full_name, comp_impl))

        # the search continues (in different implementations)
        return

    def get_found_clocks(self):
        """After the walk, this method will return the list of
        found clocks as tuple
        (clock_full_name, DataSubcomponent instance, impl_fullname, impl)"""
        return self.clocks

    pass  # end of class
# ----------------------------------------------------------------------


class ModeCollectorWalker (Walker):
    """This is used to collect the list of all modes for all implementations.

    In particular this can be used when checking zenoness.
    """

    def __init__(self):
        super(ModeCollectorWalker, self).__init__()
        self.modes = []

    def _action(self, comp_type, comp_impl, parent_impl, inst_full_path):
        full_name = ".".join((n for i, n in inst_full_path[1:]))

        impl_modes = support.get_component_impl_modes(comp_impl)
        assert inst_full_path not in self.modes
        self.modes.append((full_name, comp_impl, sorted(impl_modes)))

        # the search continues (in different implementations)
        return

    def get_found_modes(self):
        """After the walk, this method will return the map of
        found clocks as pairs (fullname, impl, modes)"""
        return self.modes

    pass  # end of class
# ----------------------------------------------------------------------


class FailureObservationAlarmControllableWalker (Walker):
    """This walker is used to collect info about the failures, their
    nominal modes, observables, alarms and controllables"""
    def __init__(self, converter):
        super(FailureObservationAlarmControllableWalker, self).__init__()

        self.converter = converter
        # list of (smv name of failure event, its nominal value (always FALSE))
        self.failure_event = []
        # list of (smv name of failure state, its nominal mode)
        self.failure_mode = []
        self.observables = []  # list of smv names of observables
        self.alarms = []  # list of smv names of alarms
        self.in_events = []  # list of smv names of input events
        self.out_events = []  # list of smv names of output events
        return

    def _action(self, comp_type, comp_impl, parent_impl, inst_full_path):
        """ searches "_errorState" out data ports in "_errorSubcomponent"
        (they are failures),
        out data ports/sub-components with "observable" flag set up
        and out data ports with "alarm" flag set up"""
        full_name = ".".join((n for i, n in inst_full_path[1:]))
        if full_name:
            full_name += "."

        # failure is an out data port with name "_errorState" in
        # "_errorSubcomponent" control sub-component.
        # Those names are taken from the module constants in
        # model_extension package
        full_error_name = (full_name +
                           constants.ERROR_SUBCOMPONENT_IDENTIFIER + ".")
        for sc in sorted(comp_impl.non_data_subcomponents):
            if sc == constants.ERROR_SUBCOMPONENT_IDENTIFIER:
                sc_impl = support.get_subcomponent_impl_from_name(
                    comp_impl, sc)
                sc_type = sc_impl.get_referenced_component_type()
                for odp_name in sorted(sc_type.out_data_ports):
                    odp = sc_type.out_data_ports[odp_name]
                    if odp.name == constants.ERROR_STATES_ENUM_PORT_NAME:
                        f = self.converter.data_name(
                            full_error_name + odp.name)
                        m = self.converter.default_value(sc_impl, odp)[0]
                        self.failure_mode.append((f, m))

        # Currently, only out data ports are expected to be failures.
        assert (constants.ERROR_STATES_ENUM_PORT_NAME not in
                comp_impl.data_subcomponents.keys())
        assert (constants.ERROR_STATES_ENUM_PORT_NAME not in
                comp_type.in_data_ports.keys())

        # extracts failure events
        for port_name in sorted(comp_type.out_event_ports):
            port = comp_type.out_event_ports[port_name]
            if isinstance(port, nominalbehaviour.ErrorEventPort):
                f = self.converter.event_name(full_name + port.name)
                v = self.converter.expression(
                    comp_impl, expressions.BooleanTreeNode(False, -1, -1))
                self.failure_event.append((f, v))

        # extracts info for observables
        for port_name in sorted(comp_type.out_data_ports):
            port = comp_type.out_data_ports[port_name]
            if port.observable:
                name = self.converter.data_name(full_name + port.name)
                self.observables.append(name)

        # extracts info for alarms
        for port_name in sorted(comp_type.out_data_ports):
            port = comp_type.out_data_ports[port_name]
            if port.alarm:
                name = self.converter.data_name(full_name + port.name)
                self.alarms.append(name)

        # extracts input events other than failure events
        for port_name in sorted(comp_type.in_event_ports):
            port = comp_type.in_event_ports[port_name]
            if not isinstance(port, nominalbehaviour.ErrorEventPort):
                f = self.converter.event_name(full_name + port.name)
                self.in_events.append(f)

        # extracts output events other than failure events
        for port_name in comp_type.out_event_ports:
            port = comp_type.out_event_ports[port_name]
            if not isinstance(port, nominalbehaviour.ErrorEventPort):
                f = self.converter.event_name(full_name + port.name)
                self.out_events.append(f)

    def dump_in_file_obs(self, file):
        self.dump_in_file(file, "Observable")

    def dump_in_file_alm(self, file):
        self.dump_in_file(file, "Alarm")

    def dump_in_file(self, file, obs_alm=None):
        """This function dumps the collected information into a file
        already opened"""
        # the head text
        file.write("<?xml version=\"1.0\"?>\n<compass>\n<fmlist>\n")

        # output the failures
        for (event, value) in self.failure_event:
            file.write("<fm name=\"" + event + "\" nominal_value=\"" +
                       value + "\"/>\n")

        # intermediate text
        file.write("</fmlist>\n<obslist>\n")

        if obs_alm is None:
            # output the observables
            for observable in self.observables:
                file.write("<obs name=\"" + observable + "\"/>\n")

            # output the alarms
            for alarm in self.alarms:
                file.write("<obs name=\"" + alarm + "\"/>\n")

        elif obs_alm == "Observable":
            # output the observables
            for observable in self.observables:
                file.write("<obs name=\"" + observable + "\"/>\n")

        elif obs_alm == "Alarm":
            # output the alarms
            for alarm in self.alarms:
                file.write("<obs name=\"" + alarm + "\"/>\n")

        # ending text
        file.write("</obslist>\n</compass>\n")
        return

    def dump_in_file_obs_alm_contr(self, file):
        """This function dumps the collected information into a file
        already opened"""
        # the head text
        file.write("<?xml version=\"1.0\"?>\n<autogef>\n<fmlist>\n")

        # output the failures
        for (event, value) in self.failure_event:
            file.write("<fm name=\"" + event + "\" nominal_value=\"" +
                       value + "\"/>\n")

        # intermediate text
        file.write("</fmlist>\n<obslist>\n")

        # output the observables
        for observable in self.observables:
            file.write("<obs name=\"" + observable + "\"/>\n")

        # intermediate text
        file.write("</obslist>\n<alarmlist>\n")

        # output the alarms
        for alarm in self.alarms:
            file.write("<alarms name=\"" + alarm + "\"/>\n")

        # intermediate text
        file.write("</alarmlist>\n<eventlist>\n")

        # output the events
        for event in self.in_events:
            file.write("<event name=\"" + event + "\" type=\"IN\"/>\n")

        for event in self.out_events:
            file.write("<event name=\"" + event + "\" type=\"OUT\"/>\n")

        # ending text
        file.write("</eventlist>\n</autogef>\n")
        return

    def get_observables_num(self):
        """Returns the number of observables collected so far"""
        return len(self.observables)

    def get_observables(self):
        """Returns the names of observables collected so far"""
        import copy
        return copy.deepcopy(self.observables)

    def get_alarms_num(self):
        """Returns the number of alarms collected so far"""
        return len(self.alarms)

    def get_alarms(self):
        """Returns the names of alarms collected so far"""
        import copy
        return copy.deepcopy(self.alarms)

    def get_failures_num(self):
        """Returns the number of failures collected so far"""
        return len(self.failure_mode)

    def get_in_events(self):
        """Returns the names of input events collected so far"""
        import copy
        return copy.deepcopy(self.in_events)

    def get_out_events(self):
        """Returns the names of output events collected so far"""
        import copy
        return copy.deepcopy(self.out_events)

    def get_in_events_num(self):
        """Returns the number of input events collected so far"""
        return len(self.in_events)

    def get_out_events_num(self):
        """Returns the number of output events collected so far"""
        return len(self.out_events)

    pass  # end of class
# ----------------------------------------------------------------------


class SlimTreeBuilderWalker (Walker):

    def __init__(self, treestore, smv_converter):
        super(SlimTreeBuilderWalker, self).__init__(order=WALKER_PRE_ORDER)

        self.tree = treestore
        self.converter = smv_converter
        self.__parent2iter = {}
        return

    def _action(self, comp_type, comp_impl, parent_impl, inst_full_path):
        piter = (None
                 if parent_impl is None
                 else self.__parent2iter[parent_impl])

        transitions = {}

        comp_inst_name = inst_full_path[-1][1]
        if comp_inst_name:
            comp_inst_name += "."

        comp_inst_fullname = ".".join(n for i, n in inst_full_path[1:])
        if comp_inst_fullname:
            comp_inst_fullname += "."

        # transitions triggered by events
        in_out_events = sorted(comp_type.in_event_ports.keys())
        in_out_events = (in_out_events +
                         sorted(comp_type.out_event_ports.keys()))

        for ev, trans in [(ev, support.get_trans_with_event(comp_impl, ev))
                          for ev in in_out_events]:

            __get_evs = self.__get_evs(inst_full_path[:-1],
                                       comp_inst_name + ev,
                                       ev in comp_type.in_event_ports,
                                       ev in comp_type.out_event_ports.keys() and comp_type.out_event_ports[ev].nonblocking,
                                       False) # a 'sender' port cannot be passive

            for t in trans:
                cando_trans = (
                    self.converter.root_var() + "." +
                    self.converter.sc_name(comp_inst_fullname) +
                    self.converter.active_name() +
                    " & " +
                    self.converter.root_var() + "." +
                    self.converter.cando_name(
                        comp_inst_fullname +
                        support.get_trans_id(comp_impl, t)))

                transitions[t] = (cando_trans, __get_evs)

        # local trans triggered by tau
        for t in support.get_trans_with_tau(comp_impl):
            cando_trans = (
                self.converter.root_var() + "." +
                self.converter.sc_name(comp_inst_fullname) +
                self.converter.active_name() +
                " & " +
                self.converter.root_var() + "." +
                self.converter.cando_name(comp_inst_fullname +
                                          support.get_trans_id(comp_impl, t)))
            transitions[t] = (cando_trans,)

        # tau transitions made with internal sub-components:
        for ndsc_name in sorted(comp_impl.non_data_subcomponents):
            sc = comp_impl.non_data_subcomponents[ndsc_name]
            sct = sc.get_referenced_component_type()
            sct_evs = sorted(sct.in_event_ports.keys())
            sct_evs = sct_evs + sorted(sct.out_event_ports.keys())
            for ev in sct_evs:
                ev_fullname = sc.name + "." + ev
                for t in support.get_trans_with_event(comp_impl, ev_fullname):
                    cando_trans = (
                        self.converter.root_var() + "." +
                        self.converter.sc_name(
                            comp_inst_fullname+sc.name + ".") +
                        self.converter.active_name() +
                        " & " +
                        self.converter.root_var() + "." +
                        self.converter.cando_name(
                            comp_inst_fullname +
                            support.get_trans_id(comp_impl, t)))

                    cando_scev = (
                        self.converter.root_var() + "." +
                        self.converter.cando_name(comp_inst_fullname +
                                                  ev_fullname))

                    transitions[t] = (cando_trans, (cando_scev, ))

        # significant leaf?
        if len(comp_impl.non_data_subcomponents) > 0 or len(transitions) > 0:
            iter = self.tree.append(
                piter,
                (comp_impl,
                 transitions,
                 inst_full_path[-1][1] if inst_full_path else "",
                 ".".join((n for i, n in inst_full_path[1:])),
                 parent_impl, None, 0))
            self.__parent2iter[comp_impl] = iter

        return

    def __get_evs(self, inst_name_path, sc_ev, is_input, is_nonblocking, is_passive):

        if len(inst_name_path) < 1 or is_nonblocking:
            return ()  # reached the top

        res = []
        inst_full_name = ".".join([n for i, n in inst_name_path[1:]]).strip()
        if inst_full_name:
            inst_full_name += "."

        pimpl, pname = inst_name_path[-1]
        if pname:
            pname += "."
        ptype = pimpl.get_referenced_component_type()
        ptype_ports = (ptype.in_event_ports.keys() +
                       ptype.out_event_ports.keys())

        ptrans = support.get_trans_with_event(pimpl, sc_ev)
        # cando names for all transitions
        tids = tuple((self.converter.root_var() + "." +
                      self.converter.cando_name(inst_full_name +
                                                support.get_trans_id(pimpl, t))
                      for t in ptrans))

        if len(tids) > 1:
            tids = tuple(((tid,) for tid in tids))

        for pev, conn in (
                (c.source_port if is_input else c.target_port, c)
                for c in (
                    support.get_connections_with_target_event(pimpl, sc_ev)
                    if is_input
                    else support.get_connections_with_source_event(pimpl,
                                                                   sc_ev))):
            if pev in ptype_ports:  # passed up/down
                pev_res = tuple(self.__get_evs(inst_name_path[:-1],
                                               pname + pev,
                                               is_input,
                                               is_nonblocking,
                                               conn.passive))
            else:
                # passed to neighbour
                pev_res = (self.converter.root_var() + "." +
                           self.converter.cando_name(inst_full_name + pev),)

            prefix = (self.converter.root_var() + "." +
                      self.converter.sc_name(inst_full_name))

            old_prefix = self.converter.get_id_prefix()
            self.converter.set_id_prefix(prefix)
            tmp = self.converter.cando_name(sc_ev)
            in_modes = self.converter.get_in_modes(
                pimpl,
                support.get_connection_modes(pimpl, conn))
            if in_modes:
                tmp = tmp + " & " + in_modes
            self.converter.set_id_prefix(old_prefix)

            if not conn.passive and not is_passive:
                res.append((tmp, pev_res))

        return (tids, tuple(res))

    def walk(self, root_impl):
        self.tree.clear()
        self.__parent2iter = {}
        if root_impl is not None:
            super(SlimTreeBuilderWalker, self).walk(root_impl)

    pass  # end of class
