# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================
import sys
sys.path.append('../../../resources/antlr/antlr-python-runtime-3.1.1/')
sys.path.append('../../')

import unittest

import antlr3
from slim_compiler.src import globals
from slim_compiler.src.dataStructure.visitors.expressionVisitor import ExpressionValueVisitor,\
    AnnotateVisitor, TypeVisitor
from slim_compiler.src.dataStructure.derivedlexer import DerivedLexer
from slim_compiler.src.dataStructure.derivedparser import DerivedParser
from slim_compiler.src.generated.SlimWalkerIdentCheck import SlimWalkerIdentCheck
from slim_compiler.src.dataStructure.nominalbehaviour import Int, IntRange, Bool,\
    Real, Continuous, Clock, Enum, DataSubcomponent
from slim_compiler.src.dataStructure.expressions import NodeType, TimeUnitType,\
    NumberTreeNode, IdentifierTreeNode, BooleanTreeNode
from slim_compiler.src.dataStructure.compile import CompileInformation
from slim_compiler.src.errorManagement.errors import GuardRefersToUnknownSubcomponentOrDataPortError
from slim_compiler.src.dataStructure.visitors.time import TimeScaleTransformVisitor
from fractions import Fraction

def parse_expression(cinfo, exprString, invariant = True):
    """ Parses a string as an expression and returns the expression tree
    associated with it. If it is an invalid expression, returns None.
    """
    globals.PARSER_EXPRESSIONS_ONLY = True
    char_stream = antlr3.ANTLRStringStream(exprString)
    lexer = DerivedLexer(cinfo, char_stream)
    tokens = antlr3.CommonTokenStream(lexer)
    parser = DerivedParser(cinfo, tokens)
    if not invariant:
        r = parser.expressionsOnly()
    else:
        r = parser.invariant()

    if parser.has_parsing_errors() or lexer.has_lexing_errors():
        print cinfo.error_collector
        return None

    nodes = antlr3.tree.CommonTreeNodeStream(r.tree)
    nodes.setTokenStream(tokens)
    expressionTree = SlimWalkerIdentCheck(nodes).expr(True)

    return expressionTree.value

def annotate_expression(cinfo, scope, root_node):
    annotator = AnnotateVisitor(cinfo)
    annotator.annotation_meth = scope.get_elem_type
    annotator.visit(root_node)

    if not cinfo.error_collector.has_no_errors():
        print cinfo.error_collector

    return cinfo.error_collector.has_no_errors()

def check_expression(cinfo, root_node):
    checker = TypeVisitor(cinfo)
    checker.visit(root_node)

    if not cinfo.error_collector.has_no_errors():
        print cinfo.error_collector

    return cinfo.error_collector.has_no_errors() and root_node.node_type != NodeType.ERROR

def evaluate_expression(cinfo, root_node):
    evaluator = ExpressionValueVisitor(cinfo)
    return evaluator.visit(root_node)

class TestScope(object):
    def __init__(self, compile_info):
        self.compile_info = compile_info
        self.types = {
                      "ivar" :     (Int(None, 0, 0), NumberTreeNode("1", None, 0, 0)),
                      "rvar" :     (Real(None, 0, 0), NumberTreeNode("1.0", None, 0, 0)),
                      "rangevar" : (IntRange(1,3, None, 0, 0), NumberTreeNode("1", None, 0, 0)),
                      "bvar" :     (Bool(None, 0, 0), BooleanTreeNode(False, 0, 0)),
                      "clkvar" :   ((Clock(TimeUnitType.SEC, None, 0, 0)), None),
                      "contvar" :  (Continuous(None, 0, 0), NumberTreeNode("1.0", None, 0, 0)),
                      "evar" :     (Enum(["a", "b", "c"], None, 0, 0), IdentifierTreeNode("a", 0, 0)),
                      "evar2" :    (Enum(["d", "e", "f"], None, 0, 0), IdentifierTreeNode("d", 0, 0))
                      }

    def get_elem_type(self, name, line, pos):
        ad = DataSubcomponent(None, "name", None, None, None, None, 0, 0)
        if name in self.types:
            ad.data_type = self.types[name][0]
            ad.default_value = self.types[name][1]
            enum_type = self.types[name][0].values if name in ["evar", "evar2"] else None
            return (self.types[name][0].get_node_type(), ad, enum_type)

        if name in self.types["evar"][0].values:
            ad.data_type = self.types["evar"][0]
            ad.default_value = self.types["evar"][1]
            return (NodeType.ENUM_VALUE, ad, [name])
        if name in self.types["evar2"][0].values:
            ad.data_type = self.types["evar2"][0]
            ad.default_value = self.types["evar2"][1]
            return (NodeType.ENUM_VALUE, ad, [name])

        self.compile_info.add_error(GuardRefersToUnknownSubcomponentOrDataPortError(None, 0, 0, "?-?", name, None))
        return (NodeType.ERROR, None, None)

class ExpressionBaseTest(unittest.TestCase):
    def setUp(self):
        self.cinfo = CompileInformation()
        self.cinfo.uses_time_scaling = True
        self.cinfo.base_unit = TimeUnitType.HOUR

        self.scope = TestScope(self.cinfo)

    def _testSyntax(self, expr_str):
        tree = parse_expression(self.cinfo, expr_str)
        return tree != None

    def _testType(self, expr_str):
        tree = parse_expression(self.cinfo, expr_str)
        if tree == None:
            return False
        if not annotate_expression(self.cinfo, self.scope, tree):
            return False
        if not check_expression(self.cinfo, tree):
            return False
        return True

    def _testValue(self, expr_str, value):
        tree = parse_expression(self.cinfo, expr_str)
        if tree == None:
            return False
        if not annotate_expression(self.cinfo, self.scope, tree):
            return False
        if not check_expression(self.cinfo, tree):
            return False
        if not evaluate_expression(self.cinfo, tree) == value:
            print str(evaluate_expression(self.cinfo, tree)) + " != " + str(value)
            return False
        return True

    def _testEqual(self, expr_str1, expr_str2):
        tree1 = parse_expression(self.cinfo, expr_str1)
        if tree1 == None:
            return False
        if not annotate_expression(self.cinfo, self.scope, tree1):
            return False
        if not check_expression(self.cinfo, tree1):
            return False

        tree2 = parse_expression(self.cinfo, expr_str2)
        if tree2 == None:
            return False
        if not annotate_expression(self.cinfo, self.scope, tree2):
            return False
        if not check_expression(self.cinfo, tree2):
            return False

        if not evaluate_expression(self.cinfo, tree1) == evaluate_expression(self.cinfo, tree2):
            print str(evaluate_expression(self.cinfo, tree1)) + " != " + str(evaluate_expression(self.cinfo, tree2))
            return False
        return True

class ExpressionTests(ExpressionBaseTest):
    def test_syntax_int(self):
        self.assertTrue(self._testType("1 + 2 - 3 * 4 / 5 mod 6"))
        pass
    def test_syntax_real(self):
        self.assertTrue(self._testType("1.0 + 2.0 - 3.0 * 4.0 / 5.0"))
    def test_syntax_bool(self):
        self.assertTrue(self._testType("true and false or true imp false iff true xor false xnor true"))
    def test_syntax_relational(self):
        self.assertTrue(self._testType("1 > 2 or 1 < 2 or 1 >= 2 or 1 <= 2 or 1 = 2 or 1 != 2"))
    def test_syntax_case_int(self):
        self.assertTrue(self._testType("case true: 1; false: 2 otherwise 3 end + case true: 1; false: 2 otherwise 3 end"))
    def test_syntax_case_real(self):
        self.assertTrue(self._testType("case true: 1.0; false: 2.0 otherwise 3.0 end + case true: 1.0; false: 2.0 otherwise 3.0 end"))
    def test_syntax_case_bool(self):
        self.assertTrue(self._testType("case true: true; false: false otherwise false end or case true: true; false: false otherwise false end"))
    def test_syntax_case_enum(self):
        self.assertTrue(self._testType("case true: a; false: b otherwise c end != evar or case true: d; false: e otherwise f end != evar2"))

    def test_syntax_unary(self):
        self.assertTrue(self._testType("not not false"))
        self.assertTrue(self._testType("-1.0 - -(1.0) - -1 - -(1)"))

class TimedExpressionTests(ExpressionBaseTest):
    # Note: time tests check if typing works OK and evaluator works OK. Does NOT check the transform visitor
    def test_time(self):
        self.assertTrue(self._testEqual("3600 sec", "1 hour"))

    def test_time2(self):
        self.assertTrue(self._testValue("3600000 msec", 1.0))
        self.assertTrue(self._testValue("3600 sec", 1.0))
        self.assertTrue(self._testValue("60 min", 1.0))
        self.assertTrue(self._testValue("1 hour", 1.0))
        self.assertTrue(self._testValue("(1/24) day", 1.0))

    def test_time3(self):
        self.assertTrue(self._testValue("clkvar", 0.0))
        self.assertTrue(self._testValue("clkvar + 1.0 hour", 1.0))
        self.assertTrue(self._testValue("(clkvar + 1.0 hour) msec", 3600000))
        self.assertTrue(self._testValue("(clkvar + 1.0 hour) sec", 3600))
        self.assertTrue(self._testValue("(clkvar + 1.0 hour) min", 60))
        self.assertTrue(self._testValue("(clkvar + 1.0 hour) hour", 1))
        self.assertTrue(self._testValue("(clkvar + 1.0 hour) day", 1.0/24))

    def test_time4(self):
        self.assertTrue(self._testValue("1 sec sec sec sec", 1))
        self.assertTrue(self._testValue("1 sec min min sec", 1))

class TimeTransformTests(ExpressionBaseTest):
    def _testValue(self, expr_str, value):
        tree = parse_expression(self.cinfo, expr_str)
        if tree == None:
            return False
        if not annotate_expression(self.cinfo, self.scope, tree):
            return False
        if not check_expression(self.cinfo, tree):
            return False
        tree = TimeScaleTransformVisitor(self.cinfo).visit(tree)
        if not evaluate_expression(self.cinfo, tree) == value:
            print str(evaluate_expression(self.cinfo, tree)) + " != " + str(value) + " : " + str(tree)
            return False
        return True

    def test_time(self):
        self.assertTrue(self._testEqual("3600 sec", "1 hour"))

    def _test_time2(self):
        self.assertTrue(self._testValue("3600000 msec", 1.0))
        self.assertTrue(self._testValue("3600 sec", 1.0))
        self.assertTrue(self._testValue("60 min", 1.0))
        self.assertTrue(self._testValue("1 hour", 1.0))
        self.assertTrue(self._testValue("(1/24) day", 1.0))

    def test_time3(self):
        self.assertTrue(self._testValue("clkvar", 0.0))
        self.assertTrue(self._testValue("clkvar + 1.0 hour", 1.0))
        self.assertTrue(self._testValue("(clkvar + 1.0 hour) msec", 3600000))
        self.assertTrue(self._testValue("(clkvar + 60.0 min) sec", 3600))
        self.assertTrue(self._testValue("(clkvar + 3600.0 sec) min", 60))
        self.assertTrue(self._testValue("(clkvar + 1.0 hour) hour", 1))
        self.assertTrue(self._testValue("(clkvar + 1.0 hour) day", Fraction(1,24)))

    def test_time4(self):
        self.assertTrue(self._testValue("1 sec sec sec sec", 1))
        self.assertTrue(self._testValue("1 sec min min sec", 1))

    def test_time_traj(self):
        expr = parse_expression(self.cinfo, "contvar' = 1 per min", True)
        TimeScaleTransformVisitor(self.cinfo).visit(expr)
        self.assertTrue(self._testValue(repr(expr.trajectory_expression), 60))

if __name__ == "__main__":
    unittest.main()
