#!/usr/bin/python 
# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================



"""
Organisation of this test suite. 
================================

This test suite is organized around the sections occurring in the SLIM
specification document. Every (sub)section is a separate class. Every 
class that is a test suite must derive from BasicTestCase.

This is the test suite for negative test cases, i.e. the cases which are
expected to fail with certain errors. 
"""


import sys
sys.path.append('../../../resources/antlr/antlr-python-runtime-3.1.1/')
sys.path.append('../../')

from invokers import slim
from slim_compiler.src.dataStructure.compile import CompileInformation
from slim_compiler.src.dataStructure.nominalbehaviour import ComponentImplementation
from slim_compiler.src.dataStructure.errorbehaviour import ErrorModelImplementation
import unittest
import tempfile

import functools

def INVALID_CGM_TESTCASE(func):
    """This is a decorator which can be used to mark testcases
    invalid for generation of the CGM integration test case.
    It will result in a warning being emitted
    when the testcase is traversed. """

    @functools.wraps(func)
    def new_func(*args, **kwargs):
        global sxml_output
        
        if sxml_output:
            print "\nPassing generation of invalid CGM test case %s." % func.__name__
            return None
        else:
            return func(*args, **kwargs)
    return new_func


global name_to_number
name_to_number = {}

class BasicTestCaseNeg(unittest.TestCase):

    def _parse_only(self, slim_snippet):
        global sxml_output, name_to_number
        # generate temp file
        f = tempfile.NamedTemporaryFile()
        try:
            f.write(slim_snippet)
            f.flush()
            # set up new compile info object
            cinfo = CompileInformation()
            slim.parse(cinfo, [f.name])
            slim.check(cinfo)
            ComponentImplementation.reset()
            ErrorModelImplementation.reset()
            
            if sxml_output: 
                error_identifier_str = ""
                for error in cinfo.error_collector.get_error_id_set():
                    error_identifier_str += "_" + str(error)
                testcase_name = 'testcase' + error_identifier_str
                if testcase_name in name_to_number:
                    next_number = name_to_number[testcase_name]
                    name_to_number[testcase_name] += 1
                else:
                    name_to_number[testcase_name] = 1
                    next_number = 0
                fsxml = open(testcase_name + ("-" + str(next_number) if next_number > 0 else "")  + ".xml", 
'w')
                fsxml.write(cinfo.to_SXML(testcase_name, False))
                fsxml.close()
            
        finally:
            f.close()
        return cinfo

    def _expect_errors(self, cinfo, *error_ids):
        return cinfo.error_collector.contains_only(error_ids)

    def print_SXML(self, cinfo):
        global instanceCounter
        print "errors"
        print cinfo.error_collector.to_SXML()
        print "model"
        print cinfo.to_SXML('testcase' + str(instanceCounter), False)

pass

class Test_02SystemTestCaseNeg(BasicTestCaseNeg):
    
    @INVALID_CGM_TESTCASE
    def test_package_identifiers_not_distinct(self):
        slim = """
package aPackage
    public
        system aSys
        end aSys;
end aPackage;

package aPackage
    public
        system aSys2
        end aSys2;
end aPackage;
"""
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "A-2");
        return
    
    @INVALID_CGM_TESTCASE
    def test_cross_reference_type_and_impl_1(self):
        slim = """
package aPackage
    public
        system aSys
        end aSys;
    private
        system implementation aSys.Impl
        end aSys.Impl;
end aPackage;
"""
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "A-3");
        return

    @INVALID_CGM_TESTCASE
    def test_cross_reference_type_and_impl_2(self):
        slim = """
package aPackage
    public
        system implementation aSys.Impl
        end aSys.Impl;
    private
        system aSys
        end aSys;
end aPackage;
"""
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "A-3");
        return
    
    @INVALID_CGM_TESTCASE
    def test_cross_reference_type_and_impl_3(self):
        slim = """
package aPackage
    public
        error model aErr
        features
          e1 : initial state;
        end aErr;
    private
        error model implementation aErr.Impl
        end aErr.Impl;
end aPackage;
"""
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "A-3");
        return
    
    @INVALID_CGM_TESTCASE
    def test_cross_reference_type_and_impl_4(self):
        slim = """
package aPackage
    public
        error model implementation aErr.Impl
        end aErr.Impl;
    private
        error model aErr
        features
          e1 : initial state;
        end aErr;
end aPackage;
"""
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "A-3");
        return
        
    @INVALID_CGM_TESTCASE        
    def test_package_end_identifier_mismatch(self):
        slim = """
package aPackage
    public
        system aSys
        end aSys;
end anotherPackage;
"""
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "A-4");
        return
        
pass
        
class Test_03TypeTestCaseNeg(BasicTestCaseNeg):

    @INVALID_CGM_TESTCASE        
    def test_type_identifiers_not_distinct_1(self):
        slim = """
package aPackage
    public
        system aSys
        end aSys;
        
        device aSys
        end aSys;
end aPackage;
"""
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "B-1");
        return

    @INVALID_CGM_TESTCASE    
    def test_type_identifiers_not_distinct_2(self):
        slim = """
package aPackage
    public
        system aSys
        end aSys;
    private
        device aSys
        end aSys;
end aPackage;
"""
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "B-1");
        return

    @INVALID_CGM_TESTCASE
    def test_type_identifiers_not_distinct_3(self):
        slim = """
package aPackage
    public
        system anotherSys
        end anotherSys;
    private
        system aSys
        end aSys;
        
        device aSys
        end aSys;
end aPackage;
"""
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "B-1");
        return

    # B-2 is restricted by grammar

    @INVALID_CGM_TESTCASE        
    def test_type_end_identifier_not_matching(self):
        slim = """
package aPackage
    public
        system aSys
        end anotherSys;
end aPackage;
"""
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "B-3");
        return
         
    @INVALID_CGM_TESTCASE        
    def test_fdir_type_has_multiple_implementations_1(self):
        slim = """
package aPackage
    public
        fdir system aSys
        end aSys;
    
        fdir system implementation aSys.Impl1
        end aSys.Impl1;
        
        fdir system implementation aSys.Impl2
        end aSys.Impl2;
end aPackage;
"""
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "B-4");
        return

    @INVALID_CGM_TESTCASE
    def test_fdir_type_has_multiple_implementations_2(self):
        slim = """
package aPackage
    public
        system anotherSys
        end anotherSys;
    private
        fdir system aSys
        end aSys;
    
        fdir system implementation aSys.Impl1
        end aSys.Impl1;
        
        fdir system implementation aSys.Impl2
        end aSys.Impl2;
end aPackage;
"""
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "B-4");
        return

class Test_04DataTypesTestCaseNeg(BasicTestCaseNeg):
    @INVALID_CGM_TESTCASE    
    def test_enum_values_not_distinct(self):
        slim = """
package aPackage
    public
        system aSys
        end aSys;
        
        system implementation aSys.Impl
            subcomponents
                myData: data enum(val1, val2, val1) default val1;
        end aSys.Impl;
end aPackage;
"""        
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "C-1");
        return
    
    @INVALID_CGM_TESTCASE    
    def test_illegal_integer_range(self):
        slim = """
package aPackage
    public
        system aSys
        end aSys;
        
        system implementation aSys.Impl
            subcomponents
                myData: data [5..4] default 4;
        end aSys.Impl;
end aPackage;
"""        
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "C-2", "F-4");
        return

    @INVALID_CGM_TESTCASE    
    def test_illegal_integer_range_2(self):
        slim = """
package aPackage
    public
        system aSys
        end aSys;
        
        system implementation aSys.Impl
            subcomponents
                myData: data [4..4] default 4;
        end aSys.Impl;
end aPackage;
"""        
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "C-2");
        return

    """this test is for the CGM only"""
    def test_wrong_typing_of_expressions(self):
        slim = """
package aPackage
    public     
        system aSys
        end aSys;
        
        system implementation aSys.Impl
            subcomponents
                varBool: data bool default true;
                varInt: data int default 5;
                varRange: data [1..5] default 5;
                varEnum: data enum(NUM1, NUM2) default NUM1;
                varReal: data real default 1.5;
            modes
                myMode1: initial mode;
                myMode2: mode;
            transitions
                myMode1 -[then varBool := varInt]-> myMode2;
                myMode1 -[then varBool := varRange]-> myMode2;
                myMode1 -[then varBool := varEnum]-> myMode2;
                myMode1 -[then varBool := varReal]-> myMode2;
                myMode1 -[when varBool > 5]-> myMode2;
                myMode1 -[when varBool < 5.1]-> myMode2;
                myMode1 -[when varBool <= NUM1]-> myMode2;
                myMode1 -[when varBool >= varRange]-> myMode2;
        end aSys.Impl;
end aPackage;
"""        
        ci = self._parse_only(slim)
        #assert self._expect_errors(ci, "C-3");
        return

    @INVALID_CGM_TESTCASE
    def test_enum_invalid_value_case(self):
        slim = """
system Sys
  features
    inData: in data port enum (a, b, c, d);
    outData: out data port enum (e, f, g, h) default e;
end Sys;

system implementation Sys.impl
  flows
    outData := case inData=a:a; inData=b:b; inData=c:c; inData=d:d otherwise a end;
end Sys.impl;
"""
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "A-0");
        return

class Test_05PortTestCaseNeg(BasicTestCaseNeg):

    @INVALID_CGM_TESTCASE    
    def test_category_does_not_support_ports_1(self):
        slim = """
package aPackage
    public
        memory aMemory
            features
                input: in event port;
        end aMemory;
end aPackage;
"""        
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "D-1");
        return

    @INVALID_CGM_TESTCASE
    def test_category_does_not_support_ports_2(self):
        slim = """
package aPackage
    public
        memory aMemory
            features
                output: out event port;
        end aMemory;
end aPackage;
"""        
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "D-1");
        return

    @INVALID_CGM_TESTCASE
    def test_category_does_not_support_ports_3(self):
        slim = """
package aPackage
    public
        memory aMemory
            features
                input: in data port bool;
        end aMemory;

        memory implementation aMemory.Impl
        end aMemory.Impl;
end aPackage;
"""        
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "D-1");
        return

    @INVALID_CGM_TESTCASE
    def test_category_does_not_support_ports_4(self):
        slim = """
package aPackage
    public
        memory aMemory
            features
                output: out data port bool;
        end aMemory;
end aPackage;
"""        
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "D-1");
        return

    @INVALID_CGM_TESTCASE
    def test_category_does_not_support_ports_5(self):
        slim = """
package aPackage
    public
        bus aBus
            features
                input: in event port;
        end aBus;
end aPackage;
"""        
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "D-1");
        return

    @INVALID_CGM_TESTCASE
    def test_category_does_not_support_ports_6(self):
        slim = """
package aPackage
    public
        bus aBus
            features
                output: out event port;
        end aBus;
end aPackage;
"""        
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "D-1");
        return

    @INVALID_CGM_TESTCASE
    def test_category_does_not_support_ports_7(self):
        slim = """
package aPackage
    public
        bus aBus
            features
                input: in data port bool;
        end aBus;
end aPackage;
"""        
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "D-1");
        return

    @INVALID_CGM_TESTCASE
    def test_category_does_not_support_ports_8(self):
        slim = """
package aPackage
    public
        bus aBus
            features
                output: out data port bool;
        end aBus;
end aPackage;
"""        
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "D-1");
        return
    
    @INVALID_CGM_TESTCASE
    def test_port_identifiers_not_distinct_1(self):
        slim = """
package aPackage
    public
        system aSys
            features
                input: in event port;
                input: out event port;
        end aSys;

	system implementation aSys.impl
	end aSys.impl;
end aPackage;
"""        
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "D-2");
        return
        
    @INVALID_CGM_TESTCASE
    def test_port_identifiers_not_distinct_2(self):
        slim = """
package aPackage
    public
        system aSys
            features
                input: in event port;
                input: out data port bool;
        end aSys;

	system implementation aSys.impl
	end aSys.impl;
end aPackage;
"""        
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "D-2");
        return

    @INVALID_CGM_TESTCASE
    def test_port_type_not_supported_1(self):
        slim = """
package aPackage
    public
        system aSys
            features
                input: in data port continuous;
        end aSys;
end aPackage;
"""        
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "D-3");
        return

    @INVALID_CGM_TESTCASE
    def test_port_type_not_supported_2(self):
        slim = """
package aPackage
    public
        system aSys
            features
                input: in data port clock;
        end aSys;
end aPackage;
"""        
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "D-3");
        return
    
    def test_port_default_value_type_mismatch_1(self):
        slim = """
package aPackage
    public
        system Root
        end Root;
        
        system implementation Root.Impl
        subcomponents
            asys : system aSys.Impl;
        end Root.Impl;
        
        system aSys
            features
                input: in data port bool default 0;
        end aSys;
        
        system implementation aSys.Impl
        end aSys.Impl;
end aPackage;
"""        
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "D-4");
        return

    def test_port_default_value_type_mismatch_2(self):
        slim = """
package aPackage
    public
        system Root
        end Root;
        
        system implementation Root.Impl
        subcomponents
            asys : system aSys.Impl;
        end Root.Impl;
        
        system aSys
            features
                input: in data port bool default test;
        end aSys;

        system implementation aSys.Impl
        end aSys.Impl;
end aPackage;
"""        
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "D-4");
        return

    def test_port_default_value_type_mismatch_3(self):
        slim = """
package aPackage
    public
        system Root
        end Root;
        
        system implementation Root.Impl
        subcomponents
            asys : system aSys.Impl;
        end Root.Impl;
        
        system aSys
            features
                input: in data port int default 0.4;
        end aSys;

        system implementation aSys.Impl
        end aSys.Impl;
end aPackage;
"""        
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "D-4");
        return

    def test_port_default_value_type_mismatch_4(self):
        slim = """
package aPackage
    public
        system Root
        end Root;
        
        system implementation Root.Impl
        subcomponents
            asys : system aSys.Impl;
        end Root.Impl;
        
        system aSys
            features
                input: in data port [0..5] default -0.5;
        end aSys;

        system implementation aSys.Impl
        end aSys.Impl;
end aPackage;
"""        
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "D-4");
        return

    def test_port_default_value_type_mismatch_5(self):
        slim = """
package aPackage
    public
        system Root
        end Root;
        
        system implementation Root.Impl
        subcomponents
            asys : system aSys.Impl;
        end Root.Impl;
        
        system aSys
            features
                input: in data port int default true;
        end aSys;

        system implementation aSys.Impl
        end aSys.Impl;
end aPackage;
"""        
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "D-4");
        return

    def test_port_default_value_type_mismatch_6(self):
        slim = """
package aPackage
    public
        system Root
        end Root;
        
        system implementation Root.Impl
        subcomponents
            asys : system aSys.Impl;
        end Root.Impl;
        
        system aSys
            features
                input: in data port enum(val1, val2, val3) default true;
        end aSys;

        system implementation aSys.Impl
        end aSys.Impl;
end aPackage;
"""        
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "D-4");
        return

    def test_port_default_value_type_mismatch_7(self):
        slim = """
package aPackage
    public
        system Root
        end Root;
        
        system implementation Root.Impl
        subcomponents
            asys : system aSys.Impl;
        end Root.Impl;
        
        system aSys
            features
                input: in data port enum(val1, val2, val3) default 5;
        end aSys;

        system implementation aSys.Impl
        end aSys.Impl;
end aPackage;
"""        
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "D-4");
        return

    def test_root_data_port_default_1(self):
        slim = """
package aPackage
    public    
        system aSys
            features
                input: in data port bool default false;
        end aSys;
        
        system implementation aSys.Impl
        end aSys.Impl;
end aPackage;
"""
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "G-14");
        return

    def test_port_enum_values_and_port_identifiers_not_distinct(self):
        slim = """
package aPackage
    public
        system aSys
            features
                input: in data port enum(val1, val2, val3);
                val2: out data port bool default false;
        end aSys;

        system implementation aSys.Impl
        end aSys.Impl;
end aPackage;
"""        
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "D-5");
        return

    def test_port_enum_values_not_disjoint_or_identical(self):
        slim = """
package aPackage
    public
        system aSys
            features
                input: in data port enum(val1, val2, val3);
                output: out data port enum(val1, val2, val3_1) default val1;
        end aSys;

        system implementation aSys.Impl
        end aSys.Impl;
end aPackage;
"""        
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "D-6");
        return

    # Port tags for IN data port restricted by compiler, no test

    def test_port_observable_FDIR_1(self):
        slim = """
package aPackage
    public
        fdir system aSys
            features
                output: out data port bool observable;
        end aSys;

        fdir system implementation aSys.Impl
        end aSys.Impl;
end aPackage;
"""
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "D-7");
        return

    def test_port_alarm_FDIR_1(self):
        slim = """
package aPackage
    public
        fdir system aSys
            features
                output: out data port int alarm;
        end aSys;

        fdir system implementation aSys.Impl
        end aSys.Impl;
end aPackage;
"""
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "D-8");
        return

    def test_port_alarm_FDIR_2(self):
        slim = """
package aPackage
    public
        system aSys
            features
                output: out data port bool alarm;
        end aSys;

        system implementation aSys.Impl
        end aSys.Impl;
end aPackage;
"""
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "D-8");
        return

pass

class Test_06ImplementationTestCaseNeg(BasicTestCaseNeg):

    @INVALID_CGM_TESTCASE    
    def test_implementation_identifiers_not_distinct_1(self):
        slim = """
package aPackage
    public
        system aSys
        end aSys;
        
        system implementation aSys.Impl
        end aSys.Impl;
        
        system implementation aSys.Impl
        end aSys.Impl;
end aPackage;
"""        
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "E-1");
        return

    @INVALID_CGM_TESTCASE    
    def test_implementation_identifiers_not_distinct_2(self):
        slim = """
package aPackage
    public
        system anotherSys
        end anotherSys;
    private
        system aSys
        end aSys;
        
        system implementation aSys.Impl
        end aSys.Impl;
        
        system implementation aSys.Impl
        end aSys.Impl;
end aPackage;
"""        
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "E-1");
        return

    @INVALID_CGM_TESTCASE    
    def test_implementation_refers_to_unkown_type(self):
        slim = """
package aPackage
    public
        system implementation aSys.Impl
        end aSys.Impl;
end aPackage;
"""        
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "E-2");
        return

    @INVALID_CGM_TESTCASE
    def test_implementation_and_type_category_mismatch(self):
        slim = """
package aPackage
    public
        system aSys
        end aSys;
        
        device implementation aSys.Impl
        end aSys.Impl;
end aPackage;
"""        
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "E-3");
        return

    @INVALID_CGM_TESTCASE
    def test_implementation_FDIR(self):
        slim = """
package aPackage
    public
        system aSys
        end aSys;
        
        fdir system implementation aSys.Impl
        end aSys.Impl;
end aPackage;
"""
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "E-5");
        return

    @INVALID_CGM_TESTCASE
    def test_implementation_end_identifier_mismatch(self):
        slim = """
package aPackage
    public
        system aSys
        end aSys;
        
        system implementation aSys.Impl
        end anotherSys.Impl;
end aPackage;
"""        
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "E-6");
        return

pass

class Test_07SubcomponentTestCaseNeg(BasicTestCaseNeg):

    @INVALID_CGM_TESTCASE    
    def test_subcomponent_identifiers_not_distinct(self):
        slim = """
package aPackage
    public
        system aSys
        end aSys;
        
        system implementation aSys.Impl
            subcomponents
                subcomp: data bool default true;
                subcomp: data int default 1;
        end aSys.Impl;
end aPackage;
"""        
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "F-1");
        return

    @INVALID_CGM_TESTCASE    
    def test_subcomponent_references_unknown_implementation_1(self):
        slim = """
package aPackage
    public
        system aSys
        end aSys;
        
        system implementation aSys.Impl
            subcomponents
                subcomp: system anotherSys.Impl;
        end aSys.Impl;
end aPackage;
"""        
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "F-2");
        return

    @INVALID_CGM_TESTCASE    
    def test_subcomponent_references_unknown_package(self):
        slim = """
package aPackage
    public
        system anotherSys
        end anotherSys;
        
        system implementation anotherSys.Impl
        end anotherSys.Impl;
    
        system aSys
        end aSys;
        
        system implementation aSys.Impl
            subcomponents
                subcomp: system anotherPackage::anotherSys.Impl;
        end aSys.Impl;
end aPackage;
"""        
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "F-2");
        return

    @INVALID_CGM_TESTCASE
    def test_subcomponent_references_unknown_implementation_2(self):
        slim = """
package aPackage
    public
        system anotherSys
        end anotherSys;
        
        system implementation anotherSys.Impl
        end anotherSys.Impl;
    
        system aSys
        end aSys;
        
        system implementation aSys.Impl
            subcomponents
                subcomp: system anotherPackage::anotherSys.Impl;
        end aSys.Impl;
end aPackage;

package anotherPackage
    public
        system thirdSys
        end thirdSys;
end anotherPackage;
"""        
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "F-2");
        return

    @INVALID_CGM_TESTCASE    
    def test_subcomponent_references_unknown_type(self):
        slim = """
package aPackage
    public    
        system aSys
        end aSys;
        
        system implementation aSys.Impl
            subcomponents
                subcomp: system anotherSys;
        end aSys.Impl;
end aPackage;
"""        
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "F-2");
        return

    @INVALID_CGM_TESTCASE    
    def test_subcomponent_references_type_with_multiple_implementations(self):
        slim = """
package aPackage
    public    
        system anotherSys
        end anotherSys;

        system implementation anotherSys.Impl1
        end anotherSys.Impl1;

        system implementation anotherSys.Impl2
        end anotherSys.Impl2;

        system aSys
        end aSys;
        
        system implementation aSys.Impl
            subcomponents
                subcomp: system anotherSys ;
        end aSys.Impl;
end aPackage;
"""        
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "F-2");
        return

    @INVALID_CGM_TESTCASE    
    def test_subcomponent_references_type_without_implementations(self):
        slim = """
package aPackage
    public    
        system anotherSys
        end anotherSys;

        system aSys
        end aSys;
        
        system implementation aSys.Impl
            subcomponents
                subcomp: system anotherSys;
        end aSys.Impl;
end aPackage;
"""        
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "F-2");
        return

    @INVALID_CGM_TESTCASE    
    def test_subcomponent_references_impl_with_FDIR_type(self):
        slim = """
package aPackage
    public    
        fdir system anotherSys
        end anotherSys;

        fdir system implementation anotherSys.Impl
        end anotherSys.Impl;

        system aSys
        end aSys;
        
        system implementation aSys.Impl
            subcomponents
                subcomp: system anotherSys.Impl;
        end aSys.Impl;
end aPackage;
"""        
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "F-2");
        return

    @INVALID_CGM_TESTCASE    
    def test_subcomponent_reference_category_mismatch_1(self):
        slim = """
package aPackage
    public    
        device anotherSys
        end anotherSys;

        device implementation anotherSys.Impl
        end anotherSys.Impl;

        system aSys
        end aSys;
        
        system implementation aSys.Impl
            subcomponents
                subcomp: system anotherSys.Impl;
        end aSys.Impl;
end aPackage;
"""        
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "F-3");
        return

    @INVALID_CGM_TESTCASE    
    def test_subcomponent_reference_category_mismatch_2(self):
        slim = """
package aPackage
    public    
        device anotherSys
        end anotherSys;

        device implementation anotherSys.Impl
        end anotherSys.Impl;

        system aSys
        end aSys;
        
        system implementation aSys.Impl
            subcomponents
                subcomp: system anotherSys;
        end aSys.Impl;
end aPackage;
"""        
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "F-3");
        return
    
    def test_subcomponent_default_value_type_mismatch_1(self):
        slim = """
package aPackage
    public    
        system aSys
        end aSys;
        
        system implementation aSys.Impl
            subcomponents
                subcomp: data int default 0.5;
        end aSys.Impl;
end aPackage;
"""        
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "F-4");
        return

    def test_subcomponent_default_value_type_mismatch_2(self):
        slim = """
package aPackage
    public    
        system aSys
        end aSys;
        
        system implementation aSys.Impl
            subcomponents
                subcomp: data [0..7] default 0.5;
        end aSys.Impl;
end aPackage;
"""        
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "F-4");
        return
    
    def test_subcomponent_default_value_type_mismatch_3(self):
        slim = """
package aPackage
    public    
        system aSys
        end aSys;
        
        system implementation aSys.Impl
            subcomponents
                subcomp: data bool default 3;
        end aSys.Impl;
end aPackage;
"""        
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "F-4");
        return

    def test_subcomponent_default_value_type_mismatch_4(self):
        slim = """
package aPackage
    public    
        system aSys
        end aSys;
        
        system implementation aSys.Impl
            subcomponents
                subcomp: data enum(val1, val2, val3) default true;
        end aSys.Impl;
end aPackage;
"""        
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "F-4");
        return

    def test_subcomponent_default_value_type_mismatch_5(self):
        slim = """
package aPackage
    public    
        system aSys
        end aSys;
        
        system implementation aSys.Impl
            subcomponents
                subcomp: data real default false;
        end aSys.Impl;
end aPackage;
"""        
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "F-4");
        return

    @INVALID_CGM_TESTCASE    
    def test_subcomponent_clock_has_default_value(self):
        slim = """
package aPackage
    public    
        system aSys
        end aSys;
        
        system implementation aSys.Impl
            subcomponents
                subcomp: data clock default 2;
        end aSys.Impl;
end aPackage;
"""        
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "F-5");
        return

    def test_subcomponent_process_has_no_thread_group(self):
        slim = """
package aPackage
    public    
        process aProc
        end aProc;
        
        process implementation aProc.Impl
            subcomponents
                subcomp: data int default 0;
        end aProc.Impl;
end aPackage;
"""        
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "F-6");
        return

    @INVALID_CGM_TESTCASE
    def test_subcomponent_category_not_supported_1(self):
        slim = """
package aPackage
    public    
        process aProc
        end aProc;
        
        process implementation aProc.Impl
            subcomponents
                subcomp: bus aBus.Impl;
        end aProc.Impl;
        
        bus aBus
        end aBus;
        
        bus implementation aBus.Impl
        end aBus.Impl;
end aPackage;
"""        
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "F-6");
        return

    @INVALID_CGM_TESTCASE
    def test_subcomponent_category_not_supported_2(self):
        slim = """
package aPackage
    public    
        thread aThread
        end aThread;
        
        thread implementation aThread.Impl
            subcomponents
                subcomp: thread group aThreadGroup.Impl;
        end aThread.Impl;
        
        thread group aThreadGroup
        end aThreadGroup;
        
        thread group implementation aThreadGroup.Impl
        end aThreadGroup.Impl;
end aPackage;
"""        
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "F-6");
        return

    @INVALID_CGM_TESTCASE    
    def test_subcomponent_category_not_supported_3(self):
        slim = """
package aPackage
    public    
        memory aMem
        end aMem;
        
        memory implementation aMem.Impl
            subcomponents
                subcomp: data bool default true;
        end aMem.Impl;
end aPackage;
"""        
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "F-6");
        return

    @INVALID_CGM_TESTCASE
    def test_subcomponent_illegal_stored_in_1(self):
        slim = """
package aPackage
    public    
        system aSys
        end aSys;
        
        system implementation aSys.Impl
            subcomponents
                myProc: process aProc.Impl stored in myMem;
        end aSys.Impl;
        
        process aProc
        end aProc;
        
        process implementation aProc.Impl
            subcomponents
                myThread: thread aThread.Impl;
        end aProc.Impl;
        
        thread aThread
        end aThread;
        
        thread implementation aThread.Impl
        end aThread.Impl; 
end aPackage;
"""        
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "F-7");
        return

    @INVALID_CGM_TESTCASE
    def test_subcomponent_illegal_stored_in_2(self):
        slim = """
package aPackage
    public    
        system aSys
        end aSys;
        
        system implementation aSys.Impl
            subcomponents
                mySys: system anotherSys.Impl stored in myMem;
                myMem: memory aMem.Impl;
        end aSys.Impl;
        
        memory aMem
        end aMem;
        
        memory implementation aMem.Impl
        end aMem.Impl;
        
        system anotherSys
        end anotherSys;
        
        system implementation anotherSys.Impl
        end anotherSys.Impl;
end aPackage;
"""        
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "F-7");
        return
    
    @INVALID_CGM_TESTCASE
    def test_subcomponent_illegal_stored_in_3(self):
        slim = """
package aPackage
    public    
        system aSys
        end aSys;
        
        system implementation aSys.Impl
            subcomponents
                myProc: process aProc.Impl stored in myBus;
                myBus: bus aBus.Impl;
        end aSys.Impl;
        
        bus aBus
        end aBus;
        
        bus implementation aBus.Impl
        end aBus.Impl;
        
        process aProc
        end aProc;
        
        process implementation aProc.Impl
            subcomponents
                myThread: thread aThread.Impl;
        end aProc.Impl;
        
        thread aThread
        end aThread;
        
        thread implementation aThread.Impl
        end aThread.Impl; 
end aPackage;
"""        
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "F-7");
        return

    @INVALID_CGM_TESTCASE    
    def test_subcomponent_illegal_running_on_1(self):
        slim = """
package aPackage
    public    
        system aSys
        end aSys;
        
        system implementation aSys.Impl
            subcomponents
                myProc: process aProc.Impl running on myProcessor;
        end aSys.Impl;
        
        process aProc
        end aProc;
        
        process implementation aProc.Impl
            subcomponents
                myThread: thread aThread.Impl;
        end aProc.Impl;
        
        thread aThread
        end aThread;
        
        thread implementation aThread.Impl
        end aThread.Impl; 
end aPackage;
"""        
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "F-8");
        return

    @INVALID_CGM_TESTCASE
    def test_subcomponent_illegal_running_on_2(self):
        slim = """
package aPackage
    public    
        system aSys
        end aSys;
        
        system implementation aSys.Impl
            subcomponents
                mySys: system anotherSys.Impl running on myProcessor;
                myProcessor: processor aProcessor.Impl;
        end aSys.Impl;
        
        processor aProcessor
        end aProcessor;
        
        processor implementation aProcessor.Impl
        end aProcessor.Impl;
        
        system anotherSys
        end anotherSys;
        
        system implementation anotherSys.Impl
        end anotherSys.Impl;
end aPackage;
"""        
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "F-8");
        return
    
    @INVALID_CGM_TESTCASE
    def test_subcomponent_illegal_running_on_3(self):
        slim = """
package aPackage
    public    
        system aSys
        end aSys;
        
        system implementation aSys.Impl
            subcomponents
                myProc: process aProc.Impl running on myBus;
                myBus: bus aBus.Impl;
        end aSys.Impl;
        
        bus aBus
        end aBus;
        
        bus implementation aBus.Impl
        end aBus.Impl;
        
        process aProc
        end aProc;
        
        process implementation aProc.Impl
            subcomponents
                myThread: thread aThread.Impl;
        end aProc.Impl;
        
        thread aThread
        end aThread;
        
        thread implementation aThread.Impl
        end aThread.Impl; 
end aPackage;
"""        
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "F-8");
        return

    @INVALID_CGM_TESTCASE
    def test_subcomponent_illegal_accesses_1(self):
        slim = """
package aPackage
    public    
        system aSys
        end aSys;
        
        system implementation aSys.Impl
            subcomponents
                mySys: system anotherSys.Impl accesses myBus;
        end aSys.Impl;
                
        system anotherSys
        end anotherSys;
        
        system implementation anotherSys.Impl
        end anotherSys.Impl; 
end aPackage;
"""        
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "F-9");
        return

    @INVALID_CGM_TESTCASE
    def test_subcomponent_illegal_accesses_2(self):
        slim = """
package aPackage
    public    
        system aSys
        end aSys;
        
        system implementation aSys.Impl
            subcomponents
                mySys: process aProc.Impl accesses myBus;
                myBus: bus aBus.Impl;
        end aSys.Impl;
                
        process aProc
        end aProc;
        
        process implementation aProc.Impl
            subcomponents
                myThread: thread aThread.Impl;
        end aProc.Impl;
                
        thread aThread
        end aThread;
        
        thread implementation aThread.Impl
        end aThread.Impl;
        
        bus aBus
        end aBus;
        
        bus implementation aBus.Impl
        end aBus.Impl;
end aPackage;
"""        
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "F-9");
        return
    
    @INVALID_CGM_TESTCASE
    def test_subcomponent_illegal_accesses_3(self):
        slim = """
package aPackage
    public    
        system aSys
        end aSys;
        
        system implementation aSys.Impl
            subcomponents
                myProc: system anotherSys.Impl accesses myMem;
                myMem: memory aMem.Impl;
        end aSys.Impl;
        
        memory aMem
        end aMem;
        
        memory implementation aMem.Impl
        end aMem.Impl;
        
        system anotherSys
        end anotherSys;
        
        system implementation anotherSys.Impl
        end anotherSys.Impl; 
end aPackage;
"""        
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "F-9");
        return  

    @INVALID_CGM_TESTCASE    
    def test_subcomponent_illegal_accesses_4(self):
        slim = """
package aPackage
    public    
        system aSys
        end aSys;
        
        system implementation aSys.Impl
            subcomponents
                myBus: bus aBus.Impl accesses myBus;
        end aSys.Impl;
        
        bus aBus
        end aBus;
        
        bus implementation aBus.Impl
        end aBus.Impl;
end aPackage;
"""        
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "F-9");
        return  

    def test_subcomponent_binding_partner_is_inactive(self):
        slim = """
package aPackage
    public    
        system aSys
        end aSys;
        
        system implementation aSys.Impl
            subcomponents
                mySys: system anotherSys.Impl accesses myBus;
                myBus: bus aBus.Impl in modes (myMode1);
            modes
                myMode1: initial mode;
                myMode2: mode;
            transitions
                myMode1 -[]-> myMode2;
        end aSys.Impl;
        
        bus aBus
        end aBus;
        
        bus implementation aBus.Impl
        end aBus.Impl;
        
        system anotherSys
        end anotherSys;
        
        system implementation anotherSys.Impl
        end anotherSys.Impl;
end aPackage;
"""        
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "F-10");
        return  

    @INVALID_CGM_TESTCASE
    def test_subcomponent_in_modes_list_references_unknown_mode(self):
        slim = """
package aPackage
    public    
        system aSys
        end aSys;
        
        system implementation aSys.Impl
            subcomponents
                mySys: system anotherSys.Impl accesses myBus in modes (myMode);
                myBus: bus aBus.Impl;
            modes
                myMode1: initial mode;
                myMode2: mode;
            transitions
                myMode1 -[]-> myMode2;
        end aSys.Impl;
        
        bus aBus
        end aBus;
        
        bus implementation aBus.Impl
        end aBus.Impl;
        
        system anotherSys
        end anotherSys;
        
        system implementation anotherSys.Impl
        end anotherSys.Impl;
end aPackage;
"""        
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "F-11");
        return

    @INVALID_CGM_TESTCASE
    def test_subcomponent_in_modes_list_duplicate_mode(self):
        slim = """
package aPackage
    public    
        system aSys
        end aSys;
        
        system implementation aSys.Impl
            subcomponents
                mySys: system anotherSys.Impl accesses myBus in modes (myMode1,myMode1);
                myBus: bus aBus.Impl;
            modes
                myMode1: initial mode;
                myMode2: mode;
            transitions
                myMode1 -[]-> myMode2;
        end aSys.Impl;
        
        bus aBus
        end aBus;
        
        bus implementation aBus.Impl
        end aBus.Impl;
        
        system anotherSys
        end anotherSys;
        
        system implementation anotherSys.Impl
        end anotherSys.Impl;
end aPackage;
"""        
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "F-11");
        return

    @INVALID_CGM_TESTCASE    
    def test_subcomponent_component_hierarchy_is_recursive_1(self):
        slim = """
package aPackage
    public    
        system aSys
        end aSys;
        
        system implementation aSys.Impl
            subcomponents
                mySys: system aSys.Impl;
        end aSys.Impl;
end aPackage;
"""        
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "F-12");
        return

    @INVALID_CGM_TESTCASE    
    def test_subcomponent_component_hierarchy_is_recursive_2(self):
        slim = """
package aPackage
    public    
        system aSys
        end aSys;
        
        system implementation aSys.Impl
            subcomponents
                mySys: system anotherSys.Impl;
        end aSys.Impl;

        system anotherSys
        end anotherSys;
        
        system implementation anotherSys.Impl
            subcomponents
                mySys: system aSys.Impl;
        end anotherSys.Impl;
end aPackage;
"""        
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "F-12");
        return

    @INVALID_CGM_TESTCASE    
    def test_subcomponent_and_data_port_identifiers_not_distinct(self):
        slim = """
package aPackage
    public    
        system aSys
            features
                input: in data port bool;
        end aSys;
        
        system implementation aSys.Impl
            subcomponents
                input: data int default 1;
        end aSys.Impl;
end aPackage;
"""
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "F-13");
        return
        
    def test_subcomponent_enum_values_and_data_ports_and_data_subcomponents_not_distinct_1(self):
        slim = """
package aPackage
    public    
        system aSys
        end aSys;
        
        system implementation aSys.Impl
            subcomponents
                mySubcomp1: data int default 1;
                mySubcomp2: data enum(val1, mySubcomp1) default val1;
        end aSys.Impl;
end aPackage;
"""
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "F-14");
        return

    def test_subcomponent_enum_values_and_data_ports_and_data_subcomponents_not_distinct_2(self):
        slim = """
package aPackage
    public    
        system aSys
            features
                val1: in data port bool;
        end aSys;
        
        system implementation aSys.Impl
            subcomponents
                mySubcomp1: data enum(val1, mySubcomp1) default val1;
        end aSys.Impl;
end aPackage;
"""
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "F-14");
        return

    def test_subcomponent_enum_values_not_identical_or_distinct_1(self):
        slim = """
package aPackage
    public    
        system aSys
        end aSys;
        
        system implementation aSys.Impl
            subcomponents
                mySubcomp1: data enum(val1, val2, val3) default val1; 
                mySubcomp2: data enum(val1, val2, val3_1) default val1;
        end aSys.Impl;
end aPackage;
"""
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "F-17");
        return

    def test_subcomponent_enum_values_not_identical_or_distinct_2(self):

        slim = """
package aPackage
    public    
        system aSys
            features
                myPort: in data port enum(val1, val2, val3);
        end aSys;
        
        system implementation aSys.Impl
            subcomponents
                mySubcomp: data enum(val1, val2, val3_1) default val1;
        end aSys.Impl;
end aPackage;
"""
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "F-18");
        return
pass

class Test_08PortConnectionTestCaseNeg(BasicTestCaseNeg):


    @INVALID_CGM_TESTCASE    
    def test_port_conn_refers_to_unknown_event_port(self):
        slim = """
package aPackage
    public    
        system aSys
        end aSys;
        
        system implementation aSys.Impl
            subcomponents
                mySys: system anotherSys.Impl;
            connections
                event port input -> mySys.input;
        end aSys.Impl;
        
        system anotherSys
            features
                input: in event port;
        end anotherSys;
        
        system implementation anotherSys.Impl
        end anotherSys.Impl;
end aPackage;
"""
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "G-1");
        return

    @INVALID_CGM_TESTCASE        
    def test_event_port_conn_references_same_component(self):
        slim = """
package aPackage
    public    
        system aSys
        end aSys;
        
        system implementation aSys.Impl
            subcomponents
                mySys: system anotherSys.Impl accesses myBus;
                myBus: bus aBus.Impl;
            connections
                event port mySys.output -> mySys.input;
        end aSys.Impl;
        
        bus aBus
        end aBus;
        
        bus implementation aBus.Impl
        end aBus.Impl;
        
        system anotherSys
            features
                input: in event port;
                output: out event port;
        end anotherSys;
        
        system implementation anotherSys.Impl
        end anotherSys.Impl;
end aPackage;
"""        
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "G-1", "G-9");
        return
    
    @INVALID_CGM_TESTCASE
    def test_data_port_conn_refers_to_unknown_port(self):
        slim = """
package aPackage
    public    
        system aSys
        end aSys;
        
        system implementation aSys.Impl
            subcomponents
                mySys: system anotherSys.Impl;
            connections
                data port input -> mySys.input;
        end aSys.Impl;
                
        system anotherSys
            features
                input: in data port bool;
        end anotherSys;
        
        system implementation anotherSys.Impl
        end anotherSys.Impl;
end aPackage;
"""        
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "G-2");
        return
    
    def test_data_port_conn_different_data_types(self):
        slim = """
package aPackage
    public    
        system aSys
            features
                input: in data port bool;
        end aSys;
        
        system implementation aSys.Impl
            subcomponents
                mySys: system anotherSys.Impl;
            connections
                data port input -> mySys.input;
        end aSys.Impl;
                
        system anotherSys
            features
                input: in data port int;
        end anotherSys;
        
        system implementation anotherSys.Impl
        end anotherSys.Impl;
end aPackage;
"""
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "G-2");
        return

    def test_data_port_conn_in_modes_list_not_distinct(self):
        slim = """
package aPackage
    public    
        system aSys
            features
                input: in data port bool;
        end aSys;
        
        system implementation aSys.Impl
            subcomponents
                mySys: system anotherSys.Impl;
            connections
                data port input -> mySys.input in modes(aMode1, aMode2, aMode1);
            modes
                aMode1: initial mode;
                aMode2: mode;
            transitions
                aMode1 -[]-> aMode2;
        end aSys.Impl;
                
        system anotherSys
            features
                input: in data port bool;
        end anotherSys;
        
        system implementation anotherSys.Impl
        end anotherSys.Impl;
end aPackage;
"""
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "G-3");
        return
        
    @INVALID_CGM_TESTCASE
    def test_data_port_conn_in_modes_list_refers_to_unkown_mode(self):
        slim = """
package aPackage
    public    
        system aSys
            features
                input: in data port bool;
        end aSys;
        
        system implementation aSys.Impl
            subcomponents
                mySys: system anotherSys.Impl;
            connections
                data port input -> mySys.input in modes(aMode1, aMode2, aMode3);
            modes
                aMode1: initial mode;
                aMode2: mode;
            transitions
                aMode1 -[]-> aMode2;
        end aSys.Impl;
                
        system anotherSys
            features
                input: in data port bool;
        end anotherSys;
        
        system implementation anotherSys.Impl
        end anotherSys.Impl;
end aPackage;
"""
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "G-3");
        return
    
    def test_data_port_conn_in_modes_list_references_inactive_subcomp(self):
        slim = """
package aPackage
    public    
        system aSys
            features
                input: in data port bool;
        end aSys;
        
        system implementation aSys.Impl
            subcomponents
                mySys: system anotherSys.Impl in modes(aMode1);
            connections
                data port input -> mySys.input;
            modes
                aMode1: initial mode;
                aMode2: mode;
            transitions
                aMode1 -[]-> aMode2;
        end aSys.Impl;
                
        system anotherSys
            features
                input: in data port bool;
        end anotherSys;
        
        system implementation anotherSys.Impl
        end anotherSys.Impl;
end aPackage;
"""
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "G-4");
        return
    
    def test_data_port_conn_missing_default_value_1(self):
        slim = """
package aPackage
    public    
        system aSys
            features
                input: in data port bool;
        end aSys;
        
        system implementation aSys.Impl
            subcomponents
                mySys: system anotherSys.Impl;
            connections
                data port input -> mySys.input in modes(aMode1);
            modes
                aMode1: initial mode;
                aMode2: mode;
            transitions
                aMode1 -[]-> aMode2;
        end aSys.Impl;
                
        system anotherSys
            features
                input: in data port bool;
        end anotherSys;
        
        system implementation anotherSys.Impl
        end anotherSys.Impl;
end aPackage;
"""
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "G-5");
        return
    
    def test_data_port_conn_missing_default_value_2(self):
        slim = """
package aPackage
    public    
        system Root
        end Root;
        
        system implementation Root.Impl
        subcomponents
            asys : system aSys.Impl;
        end Root.Impl;
        
        system aSys
            features
                output: out data port bool;
        end aSys;
        
        system implementation aSys.Impl
            modes
                aMode1: initial mode;
                aMode2: mode;
            transitions
                aMode1 -[]-> aMode2;
        end aSys.Impl;
end aPackage;
"""
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "G-6");
        return
    
    def test_data_port_conn_missing_default_value_3(self):
        slim = """
package aPackage
    public    
        system aSys
            features
                output: out data port bool;
        end aSys;
        
        system implementation aSys.Impl
            subcomponents
                mySys: system anotherSys.Impl;
            connections
                data port mySys.output -> output in modes(aMode1);
            modes
                aMode1: initial mode;
                aMode2: mode;
            transitions
                aMode1 -[]-> aMode2;
        end aSys.Impl;
       
        system anotherSys
            features
                output: out data port bool;
        end anotherSys;
        
        system implementation anotherSys.Impl
        end anotherSys.Impl;
end aPackage;
"""
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "G-6");
        return

    def test_data_port_target_of_flow_and_assignment(self):
        slim = """
package aPackage
    public    
        system aSys
            features
                output: out data port bool;
        end aSys;
        
        system implementation aSys.Impl
            subcomponents
                mySys: system anotherSys.Impl;
            connections
                data port mySys.output -> output in modes(aMode1);
            flows
                output := true in modes(aMode2);
            modes
                aMode1: initial mode;
                aMode2: mode;
            transitions
                aMode1 -[]-> aMode2;
                aMode2 -[then output := true]-> aMode1;
        end aSys.Impl;
       
        system anotherSys
            features
                output: out data port bool default true;
        end anotherSys;
        
        system implementation anotherSys.Impl
        end anotherSys.Impl;
end aPackage;
"""
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "G-7");
        return

    def test_data_port_target_of_flow_and_connection(self):
        slim = """
package aPackage
    public    
        system aSys
            features
                output: out data port bool;
        end aSys;
        
        system implementation aSys.Impl
            subcomponents
                mySys: system anotherSys.Impl;
            connections
                data port mySys.output -> output;
            flows
                output := true;
        end aSys.Impl;
       
        system anotherSys
            features
                output: out data port bool default true;
        end anotherSys;
        
        system implementation anotherSys.Impl
        end anotherSys.Impl;
end aPackage;
"""
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "G-8");
        return

    def test_data_port_fan_in_1(self):
        slim = """
package aPackage
    public    
        system aSys
            features
                output: out data port bool;
                input1: in data port bool;
                input2: in data port bool;
        end aSys;
        
        system implementation aSys.Impl
            subcomponents
                mySys: system anotherSys.Impl;
            connections
                data port mySys.output1 -> output;
                data port mySys.output2 -> output;
            modes
                aMode1: initial mode;
                aMode2: mode;
            transitions
                aMode1 -[]-> aMode2;
        end aSys.Impl;
       
        system anotherSys
            features
                output1: out data port bool default true;
                output2: out data port bool default false;
        end anotherSys;
        
        system implementation anotherSys.Impl
        end anotherSys.Impl;
end aPackage;
"""
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "G-8");
        return

    def test_data_port_fan_in_2(self):
        slim = """
package aPackage
    public    
        system aSys
            features
                output: out data port bool;
                input1: in data port bool;
                input2: in data port bool;
        end aSys;
        
        system implementation aSys.Impl
            subcomponents
                mySys: system anotherSys.Impl;
            connections
                data port mySys.output -> output;
            flows
                output := input1 or input2;
        end aSys.Impl;
       
        system anotherSys
            features
                output: out data port bool default true;
        end anotherSys;
        
        system implementation anotherSys.Impl
        end anotherSys.Impl;
end aPackage;
"""
        # This test case is disabled because due to the 
        # SLIM restrictions as of 26 Apr 2010, flows
        # and port connections have to be mutually
        # exclusive.

        #ci = self._parse_only(slim)
        #assert self._expect_errors(ci, "G-8");
        return

    def test_data_port_fan_in_3(self):
        slim = """
package aPackage
    public    
        system aSys
            features
                output: out data port bool;
                input1: in data port bool;
                input2: in data port bool;
        end aSys;
        
        system implementation aSys.Impl
            flows
                output := input1 or input2;
                output := input1 xor input2;
        end aSys.Impl;
end aPackage;
"""
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "G-8");
        return

    def test_data_port_conn_violates_possible_topologies_1(self):
        slim = """
package aPackage
    public    
        system aSys
            features
                output: out data port bool;
                input: in data port bool;
        end aSys;
        
        system implementation aSys.Impl
            connections
                data port input -> output;
        end aSys.Impl;
end aPackage;
"""
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "G-9");
        return

    def test_event_port_subcomponent_exists(self):
        slim = """
package aPackage
    public    
        system aSys
            features
                out1: out event port;
        end aSys;
        
        system implementation aSys.Impl
            subcomponents
                subSys: system anotherSys.Impl;
            connections
                event port subSysa.outport -> out1;
        end aSys.Impl;
        
        system anotherSys
            features
                outport: out event port;
        end anotherSys;
        
        system implementation anotherSys.Impl
        end anotherSys.Impl;
end aPackage;
"""
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "G-10");
        return

    def test_event_port_fan_out_1(self):
        slim = """
package aPackage
    public    
        system aSys
            features
                inPort1: in event port;
        end aSys;
        
        system implementation aSys.Impl
            subcomponents
                subSys: system anotherSys.Impl;
            connections
                event port inPort1 -> subSys.inPort1;
                event port inPort1 -> subSys.inPort2;
        end aSys.Impl;
        
        system anotherSys
            features
                inPort1: in event port;
                inPort2: in event port;
        end anotherSys;
        
        system implementation anotherSys.Impl
        end anotherSys.Impl;
end aPackage;
"""
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "G-11");
        return

    def test_event_port_fan_out_2(self):
        slim = """
package aPackage
    public    
        system aSys
        end aSys;

        system implementation aSys.Impl
            subcomponents
                subSys1: system anotherSys.Impl accesses bbus;
                subSys2: system anotherSys.Impl accesses bbus;
                bbus: bus Bus.Impl;
            connections
                event port subSys1.outport -> subSys2.inPort1;
                event port subSys1.outport -> subSys2.inPort2;
        end aSys.Impl;
        
        system anotherSys
            features
                outport: out event port;
                inPort1: in event port;
                inPort2: in event port;
        end anotherSys;
        
        system implementation anotherSys.Impl
        end anotherSys.Impl;
        
        bus Bus
        end Bus;
        bus implementation Bus.Impl
        end Bus.Impl;
end aPackage;
"""
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "G-11");
        return

    def test_event_port_fan_out_3(self):
        slim = """
package aPackage
    public    
        system aSys
            features
                out1: out event port;
                out2: out event port;
        end aSys;
        
        system implementation aSys.Impl
            subcomponents
                subSys: system anotherSys.Impl;
            connections
                event port subSys.outport -> out1;
                event port subSys.outport -> out2;
        end aSys.Impl;
        
        system anotherSys
            features
                outport: out event port;
        end anotherSys;
        
        system implementation anotherSys.Impl
        end anotherSys.Impl;
end aPackage;
"""
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "G-11");
        return

    def test_event_port_passive_in_in(self):
        slim = """
package aPackage
    public    
        system aSys
            features
                in1: in event port;
        end aSys;
        
        system implementation aSys.Impl
            subcomponents
                subSys: system anotherSys.Impl;
            connections
                event port in1 -> subSys.in1 passive;
        end aSys.Impl;
        
        system anotherSys
            features
                in1: in event port;
        end anotherSys;
        
        system implementation anotherSys.Impl
        end anotherSys.Impl;
end aPackage;
"""
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "G-12");
        return

    def test_event_port_passive_out_out(self):
        slim = """
package aPackage
    public    
        system aSys
            features
                out1: out event port;
        end aSys;
        
        system implementation aSys.Impl
            subcomponents
                subSys: system anotherSys.Impl;
            connections
                event port subSys.out1 -> out1 passive;
        end aSys.Impl;
        
        system anotherSys
            features
                out1: out event port;
        end anotherSys;
        
        system implementation anotherSys.Impl
        end anotherSys.Impl;
end aPackage;
"""
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "G-12");
        return

    def test_event_port_passive_mixed(self):
        slim = """
package aPackage
    public
        bus bBus
        end bBus;
        bus implementation bBus.Impl
        end bBus.Impl;

        system aSys
        end aSys;
        
        system implementation aSys.Impl
            subcomponents
                subSys1: system anotherSys1.Impl accesses bus1;
                subSys2: system anotherSys1.Impl accesses bus1;
                subSys3: system anotherSys2.Impl accesses bus1;
                bus1 : bus bBus.Impl;
            connections
                event port subSys1.outport -> subSys3.inport passive;
                event port subSys2.outport -> subSys3.inport;
        end aSys.Impl;
        
        system anotherSys1
            features
                outport: out event port;
        end anotherSys1;
        
        system implementation anotherSys1.Impl
        end anotherSys1.Impl;
        
        system anotherSys2
            features
                inport: in event port;
        end anotherSys2;
        
        system implementation anotherSys2.Impl
        end anotherSys2.Impl;
end aPackage;
"""
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "G-13");
        return

    def test_event_port_passive_mixed_modes(self):
        slim = """
package aPackage
    public
        bus bBus
        end bBus;
        bus implementation bBus.Impl
        end bBus.Impl;

        system aSys
        end aSys;
        
        system implementation aSys.Impl
            subcomponents
                subSys1: system anotherSys1.Impl accesses bus1;
                subSys2: system anotherSys1.Impl accesses bus1;
                subSys3: system anotherSys2.Impl accesses bus1;
                bus1 : bus bBus.Impl;
            connections
                event port subSys1.outport -> subSys3.inport in modes (m0) passive;
                event port subSys2.outport -> subSys3.inport in modes (m1);
            modes
                m0 : initial mode;
                m1 : mode;
            transitions
                m0 -[]-> m1;
                m1 -[]-> m0;
        end aSys.Impl;
        
        system anotherSys1
            features
                outport: out event port;
        end anotherSys1;
        
        system implementation anotherSys1.Impl
        end anotherSys1.Impl;
        
        system anotherSys2
            features
                inport: in event port;
        end anotherSys2;
        
        system implementation anotherSys2.Impl
        end anotherSys2.Impl;
end aPackage;
"""
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "G-13");
        return

        # G-14: See PortTestCaseNeg

    def test_eventport_references_dataport(self):
        slim = """
package aPackage
    public    
        system aSys
            features
                outPort: out data port bool;
        end aSys;
        
        system implementation aSys.Impl
            subcomponents
                subSys: system anotherSys.Impl;
            connections
                event port subSys.outPort -> outPort;
        end aSys.Impl;
        
        system anotherSys
            features
                outPort: out event port;
        end anotherSys;
        
        system implementation anotherSys.Impl
        end anotherSys.Impl;
end aPackage;
"""
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "G-1");
        return

    def test_dataport_references_eventport(self):
        slim = """
package aPackage
    public    
        system aSys
            features
                outPort: out event port;
        end aSys;
        
        system implementation aSys.Impl
            subcomponents
                subSys: system anotherSys.Impl;
            connections
                data port subSys.outPort -> outPort;
        end aSys.Impl;
        
        system anotherSys
            features
                outPort: out data port bool default false;
        end anotherSys;
        
        system implementation anotherSys.Impl
        end anotherSys.Impl;
end aPackage;
"""
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "G-2");
        return

pass

class Test_09FlowTestCaseNeg(BasicTestCaseNeg):

    @INVALID_CGM_TESTCASE        
    def test_flow_references_unknown_out_port(self):
        slim = """
package aPackage
    public    
        system aSys
            features
                input: in data port bool;
        end aSys;
        
        system implementation aSys.Impl
            flows
                output := not input;
        end aSys.Impl;
end aPackage;
"""
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "H-1");
        return

    @INVALID_CGM_TESTCASE        
    def test_flow_has_fan_in(self):
        slim = """
package aPackage
    public    
        system aSys
            features
                input1: in data port bool;
                input2: in data port bool;
        end aSys;
        
        system implementation aSys.Impl
            flows
                output := not input1;
                output := not input2;
        end aSys.Impl;
end aPackage;
"""
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "H-1");
        return
    
    def test_flow_references_unknown_in_port(self):
        slim = """
package aPackage
    public    
        system aSys
            features
                output: out data port bool;
        end aSys;
        
        system implementation aSys.Impl
            flows
                output := not input;
        end aSys.Impl;
end aPackage;
"""
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "H-2");
        return

    def test_flow_bad_type(self):
        slim = """
package aPackage
    public    
        system aSys
            features
                output: out data port bool;
                input: in data port int;
        end aSys;
        
        system implementation aSys.Impl
            flows
                output := input - 1;
        end aSys.Impl;
end aPackage;
"""
        ci = self._parse_only(slim)
        # Note: expression is parsed as a whole, so H-2 is subsumed by
        # expression type checks (Assignment operator)
        #assert self._expect_errors(ci, "H-2");
        assert self._expect_errors(ci, "C-3");
        return


    def test_data_port_target_of_flow_and_assignment(self):
        slim = """
package aPackage
    public    
        system aSys
            features
                output: out data port bool;
        end aSys;
        
        system implementation aSys.Impl
            subcomponents
                mySys: system anotherSys.Impl;
            flows
                output := true in modes(aMode2);
            modes
                aMode1: initial mode;
                aMode2: mode;
            transitions
                aMode1 -[]-> aMode2;
                aMode2 -[then output := true]-> aMode1;
        end aSys.Impl;
       
        system anotherSys
            features
                output: out data port bool default true;
        end anotherSys;
        
        system implementation anotherSys.Impl
        end anotherSys.Impl;
end aPackage;
"""
        ci = self._parse_only(slim)
        # H-3 covered by G-7
        assert self._expect_errors(ci, "G-7");
        return

    # H-4 subsumed by G-6

    def test_port_dependency_relation_cyclic(self):
        slim = """
package aPackage
    public    
        system aSys
        end aSys;
        
        system implementation aSys.Impl
            subcomponents
                mySys1: system anotherSys.Impl1 accesses myBus;
                mySys2: system anotherSys.Impl2 accesses myBus;
                myBus: bus aBus.Impl;
            connections
                data port mySys1.output -> mySys2.input;
                data port mySys2.output -> mySys1.input;
        end aSys.Impl;
        
        bus aBus
        end aBus;
        
        bus implementation aBus.Impl
        end aBus.Impl;
        
        system anotherSys
            features
                input: in data port bool default false;
                output: out data port bool default true;
        end anotherSys;
        
        system implementation anotherSys.Impl1
            flows
                output := not input;
        end anotherSys.Impl1;
        
        system implementation anotherSys.Impl2
            flows
                output := input and input;
        end anotherSys.Impl2;
end aPackage;
"""
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "H-5");
        return
        
    @INVALID_CGM_TESTCASE    
    def test_flow_references_unknown_mode(self):
        slim = """
package aPackage
    public    
        system aSys
            features
                output: out data port bool default false;
                input: in data port bool;
        end aSys;
        
        system implementation aSys.Impl
            flows
                output := not input in modes (nonExistentMode);
        end aSys.Impl;
end aPackage;
"""
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "H-6");
        return
    
pass

class Test_10ModeTestCaseNeg(BasicTestCaseNeg):
    # I-1: modes are supported by all categories

    @INVALID_CGM_TESTCASE    
    def test_mode_names_not_distinct(self):
        slim = """
package aPackage
    public
        system aSys
        end aSys;
    
        system implementation aSys.Impl
            modes
                aMode: initial mode;
                aMode: mode;
        end aSys.Impl;
end aPackage;
        """
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "I-2");
        return
    
    def test_no_starting_mode(self):
        slim = """
package aPackage
    public
        system aSys
        end aSys;
    
        system implementation aSys.Impl
            modes
                aMode: mode;
        end aSys.Impl;
end aPackage;
        """
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "I-3");
        return

    def test_multiple_starting_mode(self):
        slim = """
package aPackage
    public
        system aSys
        end aSys;
    
        system implementation aSys.Impl
            modes
                aMode1: initial mode;
                aMode2: activation mode;
        end aSys.Impl;
end aPackage;
        """
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "I-3");
        return

    def test_mode_not_reachable(self):
        slim = """
package aPackage
    public
        system aSys
        end aSys;
    
        system implementation aSys.Impl
            modes
                aMode1: initial mode;
                aMode2: mode;
        end aSys.Impl;

end aPackage;
        """
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "I-4");
        return

    def test_mode_trajectory_refers_to_int_subcomponent(self):
        slim = """
package aPackage
    public
        system aSys
        end aSys;
    
        system implementation aSys.Impl
            subcomponents
                myData: data int default 0;
            modes
                aMode1: initial mode while myData' = 5;
        end aSys.Impl;

end aPackage;
        """
        ci = self._parse_only(slim)
        # Related to I-5, but subsumed by I-7 (int component is out of scope)
        assert self._expect_errors(ci, "I-7");
        return

    def test_mode_trajectory_refers_to_clock(self):
        slim = """
package aPackage
    public
        system aSys
        end aSys;
    
        system implementation aSys.Impl
            subcomponents
                myData: data clock;
            modes
                aMode: initial mode while myData' = 4;
        end aSys.Impl;

end aPackage;
        """
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "I-5");
        return

    @INVALID_CGM_TESTCASE
    def test_mode_invariant_expression_bad_type(self):
        slim = """
package aPackage
    public
        system aSys
        end aSys;
    
        system implementation aSys.Impl
            subcomponents
                myData: data continuous default 0;
            modes
                aMode1: initial mode while myData + 2;
        end aSys.Impl;

end aPackage;
        """
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "I-7");
        return

    def test_mode_invariant_refers_to_nonexistant_subcomponent(self):
        slim = """
package aPackage
    public
        system aSys
        end aSys;
    
        system implementation aSys.Impl
            subcomponents
                myData: data int default 0;
            modes
                aMode1: initial mode while myData2 < 5;
        end aSys.Impl;

end aPackage;
        """
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "I-7");
        return

    def test_mode_invariant_refers_to_int_subcomponent(self):
        slim = """
package aPackage
    public
        system aSys
        end aSys;
    
        system implementation aSys.Impl
            subcomponents
                myData: data int default 0;
            modes
                aMode1: initial mode while myData < 5;
        end aSys.Impl;

end aPackage;
        """
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "I-7");
        return

    def test_mode_invariant_refers_to_inactive_subcomponent(self):
        slim = """
package aPackage
    public
        system aSys
        end aSys;
    
        system implementation aSys.Impl
            subcomponents
                myData: data continuous default 0 in modes(aMode2);
            modes
                aMode1: initial mode while myData < 5;
                aMode2: mode;
            transitions
                aMode1 -[]-> aMode2;
        end aSys.Impl;

end aPackage;
        """
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "I-7");
        return
    
    def test_mode_invariant_flow_equation_refers_to_inactive_subcomp(self):
        slim = """
package aPackage
    public
        system aSys
        end aSys;
    
        system implementation aSys.Impl
            subcomponents
                myData: data continuous default 0 in modes(aMode2);
            modes
                aMode1: initial mode while myData' = 4;
                aMode2: mode;
            transitions
                aMode1 -[]-> aMode2;
        end aSys.Impl;

end aPackage;
        """
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "I-7");
        return

    def test_mode_invariant_non_linear_1(self):
        slim = """
package aPackage
    public
        system aSys
        end aSys;
    
        system implementation aSys.Impl
            subcomponents
                myData: data continuous default 0;
            modes
                aMode1: initial mode while myData * myData >= 0;
                aMode2: mode;
            transitions
                aMode1 -[]-> aMode2;
        end aSys.Impl;

end aPackage;
        """
        ci = self._parse_only(slim)
        # Special case of the fact that all continuous expression have to be linear
        assert self._expect_errors(ci, "C-5");
        return

    def test_mode_invariant_non_linear_2(self):
        slim = """
package aPackage
    public
        system aSys
        end aSys;
    
        system implementation aSys.Impl
            subcomponents
                myData: data continuous default 0;
            modes
                aMode1: initial mode while 1 / myData >= 0;
                aMode2: mode;
            transitions
                aMode1 -[]-> aMode2;
        end aSys.Impl;

end aPackage;
        """
        ci = self._parse_only(slim)
        # Special case of the fact that all continuous expression have to be linear
        assert self._expect_errors(ci, "C-5");
        return

    @INVALID_CGM_TESTCASE    
    def test_mode_invariant_multiple_flow_equations_for_one_subcomp(self):
        slim = """
package aPackage
    public
        system aSys
        end aSys;
    
        system implementation aSys.Impl
            subcomponents
                myData: data continuous default 0;
            modes
                aMode1: initial mode while myData' = 4 and myData' = 5;
        end aSys.Impl;

end aPackage;
        """
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "I-8");
        return

    @INVALID_CGM_TESTCASE
    def test_mode_invariant_invalid_initial_value_1(self):
        slim = """
package aPackage
    public
        system aSys
        end aSys;
    
        system implementation aSys.Impl
            subcomponents
                myData: data continuous default 0;
            modes
                aMode1: initial mode while myData >= (3 mod 2);
        end aSys.Impl;

end aPackage;
        """
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "I-9");
        return

    @INVALID_CGM_TESTCASE
    def test_mode_invariant_invalid_initial_value_2(self):
        slim = """
package aPackage
    public
        system aSys
        end aSys;
    
        system implementation aSys.Impl
            subcomponents
                myData: data continuous default 0;
            modes
                aMode1: initial mode while case myData >= (3 mod 2): true otherwise false end;
        end aSys.Impl;

end aPackage;
        """
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "I-9");
        return

    @INVALID_CGM_TESTCASE
    def test_mode_transition_refers_to_unknown_mode(self):
        slim = """
package aPackage
    public
        system aSys
        end aSys;
    
        system implementation aSys.Impl
            modes
                aMode1: initial mode;
            transitions
                aMode2 -[]-> aMode2;
        end aSys.Impl;

end aPackage;
        """
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "I-10");
        return

    def test_mode_transition_invalid_trigger_1(self):
        slim = """
package aPackage
    public
        system aSys
        end aSys;
    
        system implementation aSys.Impl
            subcomponents
                mySys: system anotherSys.Impl; 
            modes
                aMode1: initial mode;
            transitions
                aMode1 -[mySys.anotherPort]-> aMode1;
        end aSys.Impl;

        system anotherSys
            features
                aPort: out event port;
        end anotherSys;
        
        system implementation anotherSys.Impl
        end anotherSys.Impl;
end aPackage;
        """
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "I-11");
        return

    def test_mode_transition_invalid_trigger_2(self):
        slim = """
package aPackage
    public
        system aSys
        end aSys;
    
        system implementation aSys.Impl
            subcomponents
                mySys: system anotherSys.Impl in modes(aMode2); 
            modes
                aMode1: initial mode;
                aMode2: mode;
            transitions
                aMode1 -[mySys.aPort]-> aMode2;
        end aSys.Impl;

        system anotherSys
            features
                aPort: out event port;
        end anotherSys;
        
        system implementation anotherSys.Impl
        end anotherSys.Impl;
end aPackage;
        """
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "I-11");
        return

    def test_mode_transition_invalid_trigger_3(self):
        slim = """
package aPackage
    public
        system aSys
            features
                input: in event port;
        end aSys;
    
        system implementation aSys.Impl
            subcomponents
                mySys: system anotherSys.Impl; 
            modes
                aMode1: initial mode;
            transitions
                aMode1 -[mySys.aPort or anotherInput]-> aMode1;
        end aSys.Impl;

        system anotherSys
            features
                aPort: out event port;
        end anotherSys;
        
        system implementation anotherSys.Impl
        end anotherSys.Impl;
end aPackage;
        """
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "I-11");
        return

    def test_mode_transition_guard_refers_to_unknown_identifier(self):
        slim = """
package aPackage
    public
        system aSys
        end aSys;
    
        system implementation aSys.Impl
            modes
                aMode1: initial mode;
            transitions
                aMode1 -[when input]-> aMode1;
        end aSys.Impl;
end aPackage;
        """
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "I-12");
        return

    def test_mode_transition_guard_refers_to_inactive_subcomp(self):
        slim = """
package aPackage
    public
        system aSys
        end aSys;
    
        system implementation aSys.Impl
            subcomponents
                myData: data int default 3 in modes(aMode2); 
            modes
                aMode1: initial mode;
                aMode2: mode;
            transitions
                aMode1 -[when myData = 7]-> aMode2;
        end aSys.Impl;
end aPackage;
        """
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "I-12");
        return

    def test_mode_transition_guard_wrong_type(self):
        slim = """
package aPackage
    public
        system aSys
        end aSys;
    
        system implementation aSys.Impl
            subcomponents
                myData: data int default 3; 
            modes
                aMode1: initial mode;
                aMode2: mode;
            transitions
                aMode1 -[when myData]-> aMode2;
        end aSys.Impl;
end aPackage;
        """
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "I-12");
        return

    def test_mode_transition_effect_assigns_multiple_values_to_one_subcomp(self):
        slim = """
package aPackage
    public
        system aSys
        end aSys;
    
        system implementation aSys.Impl
            subcomponents
                myData: data int default 1; 
            modes
                aMode1: initial mode;
                aMode2: mode;
            transitions
                aMode1 -[then myData := 7; myData := 8]-> aMode2;
        end aSys.Impl;
end aPackage;
        """
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "I-13");
        return

    def test_mode_transition_effect_refers_to_inactive_subcomp_1(self):
        slim = """
package aPackage
    public
        system aSys
        end aSys;
    
        system implementation aSys.Impl
            subcomponents
                myData: data int default 1 in modes(aMode1); 
            modes
                aMode1: initial mode;
                aMode2: mode;
            transitions
                aMode1 -[then myData := 7]-> aMode2;
        end aSys.Impl;
end aPackage;
        """
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "I-13");
        return
    
    def test_mode_transition_effect_refers_to_inactive_subcomp_2(self):
        slim = """
package aPackage
    public
        system aSys
        end aSys;
    
        system implementation aSys.Impl
            subcomponents
                myData1: data int default 1;
                myData2: data int default 3 in modes(aMode2); 
            modes
                aMode1: initial mode;
                aMode2: mode;
            transitions
                aMode1 -[then myData1 := myData2 + 2]-> aMode2;
        end aSys.Impl;
end aPackage;
        """
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "I-14");
        return    
    
    
    def test_mode_transition_effect_refers_to_unknown_identifier(self):
        slim = """
package aPackage
    public
        system aSys
        end aSys;
    
        system implementation aSys.Impl
            subcomponents
                myData1: data int default 1;
            modes
                aMode1: initial mode;
                aMode2: mode;
            transitions
                aMode1 -[then myData1 := myData2 + 2]-> aMode2;
        end aSys.Impl;
end aPackage;
        """
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "I-14");
        return 

    # I-15 subsumed by expression type checks
    
    def test_subcomp__default_value_missing_1(self):
        slim = """
package aPackage
    public
        system aSys
        end aSys;
    
        system implementation aSys.Impl
            subcomponents
                myData: data int;
            modes
                aMode1: initial mode;
                aMode2: mode;
                aMode3: mode;
            transitions
                aMode1 -[]-> aMode2;
                aMode2 -[when myData = 3]-> aMode3;
        end aSys.Impl;
end aPackage;
        """
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "I-16");
        return    

    def test_subcomp__default_value_missing_2(self):
        slim = """
package aPackage
    public
        system aSys
        end aSys;
    
        system implementation aSys.Impl
            subcomponents
                myData: data continuous;
            modes
                aMode1: initial mode;
                aMode2: mode;
                aMode3: mode while myData < 3;
            transitions
                aMode1 -[]-> aMode2;
                aMode2 -[]-> aMode3;
        end aSys.Impl;
end aPackage;
        """
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "I-16");
        return

    # I-17 deprecated
    # I-18 subsumed by G-7

    def test_mode_invariant_non_convex(self):
        slim = """
package aPackage
    public
        system aSys
        end aSys;
    
        system implementation aSys.Impl
            subcomponents
                myData: data continuous default 0;
            modes
                aMode1: initial mode while myData <= 0 or myData >= 2;
                aMode2: mode;
            transitions
                aMode1 -[]-> aMode2;
        end aSys.Impl;

end aPackage;
        """
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "I-19");
        return

    def test_mode_urgent_bounds(self):
        slim = """
package aPackage
    public
        system aSys
        end aSys;
    
        system implementation aSys.Impl
            modes
                aMode1: initial mode urgent in -1.0;
        end aSys.Impl;

end aPackage;
        """
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "I-20");
        return

    def test_trans_within_bounds(self):
        slim = """
package aPackage
    public
        system aSys
        end aSys;
    
        system implementation aSys.Impl
            subcomponents
                myData: data continuous default 0;
            modes
                aMode1: initial mode urgent in 1.0;
                aMode2: mode;
            transitions
                aMode1 -[within 2.0 to 1.0]-> aMode2;
        end aSys.Impl;

end aPackage;
        """
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "I-21");
        return

    def test_trans_reactivation_has_guard(self):
        slim = """
package aPackage
    public
        system aSys
        end aSys;
    
        system implementation aSys.Impl
            subcomponents
                myData: data continuous default 0;
            modes
                aMode1: initial mode;
                aMode2: mode;
            transitions
                aMode1 -[@activation when true]-> aMode2;
        end aSys.Impl;

end aPackage;
        """
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "I-22");
        return

    def test_trans_reactivation_has_within(self):
        slim = """
package aPackage
    public
        system aSys
        end aSys;
    
        system implementation aSys.Impl
            subcomponents
                myData: data continuous default 0;
            modes
                aMode1: initial mode;
                aMode2: mode;
            transitions
                aMode1 -[@activation within 1.0 to 2.0]-> aMode2;
        end aSys.Impl;

end aPackage;
        """
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "I-23");
        return

class Test_11ErrorModelTypesTestCaseNeg(BasicTestCaseNeg):
    @INVALID_CGM_TESTCASE
    def test_same_errormodel_type_default(self):
        slim = """
        error model defaultErr
            features
                state1: initial state;
        end defaultErr;

        error model defaultErr
            features
                state2: error state;
        end defaultErr;

	error model implementation defaultErr.impl       
		events 
			fault: error event;
			fault2: error event;
		transitions
			state1 -[fault]-> state1;
			state1 -[fault2]-> state2;
	end defaultErr.impl;
        """
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "J-1");



    @INVALID_CGM_TESTCASE
    def test_same_errormodel_type_public_private(self):
        slim = """
package BigPackage
    public
        error model defaultErr
            features
                state1: initial state;
        end defaultErr;
    private
        error model defaultErr
            features
                state2: error state;
        end defaultErr;
end BigPackage;"""
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "J-1");    
        

    @INVALID_CGM_TESTCASE    
    def test_same_error_state_identifiers_within_model3(self):
        slim = """
error model defaultErr
    features
        deState2: initial state;
        deState2: error state;
end defaultErr;

error model implementation defaultErr.impl       
	events
		fault: error event;
	transitions
		deState2 -[fault]-> deState2;
end defaultErr.impl;
        """
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "J-2");



    @INVALID_CGM_TESTCASE    
    def test_same_error_propagation_identifiers(self):
        slim = """
error model defaultErr
    features
        deOutProp1: out error propagation;
        deOutProp1: in error propagation;
        state1: initial state;
end defaultErr;

error model implementation defaultErr.impl       
end defaultErr.impl;
	"""
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "J-2");

    @INVALID_CGM_TESTCASE    
    def test_same_error_state_and_propagation_identifiers(self):
        slim = """
error model defaultErr
    features
        state1: out error propagation;
        state1: initial state;
end defaultErr;

error model implementation defaultErr.impl       
end defaultErr.impl;
    """
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "J-2");

    def test_multiple_starting_error_states(self):
        slim = """
package BigPackage
    public
        error model defaultErr
            features
                deState1: activation state;
                deState2: error state;
                deState3: initial state;
        end defaultErr;

	error model implementation defaultErr.impl       
		events 
			fault: error event;
		transitions
			deState3 -[fault]-> deState1;
			deState2 -[fault]-> deState1;
			deState1 -[fault]-> deState2;
	end defaultErr.impl;
end BigPackage;"""
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "J-3");


    def test_missing_starting_error_state_in_type(self):
        slim = """
package BigPackage
    public
        error model defaultErr
            features
                State1: error state;
        end defaultErr;

	error model implementation defaultErr.impl       
		events 
			fault: error event;
		transitions
			State1 -[fault]-> State1;
	end defaultErr.impl;
end BigPackage;"""
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "J-3", "K-9");

    @INVALID_CGM_TESTCASE    
    def test_same_error_type_identifiers(self):
        slim = """
error model defaultErr
    features
        state1: initial state;
end defaultErr1;
    """
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "J-4");

pass    


class Test_12ErrorModelImplementationTestCaseNeg(BasicTestCaseNeg):

    @INVALID_CGM_TESTCASE
    def test_same_errormodel_implementation_default(self):
        slim = """
error model defaultErr
    features
        state1: initial state;
end defaultErr;

error model implementation defaultErr.impl       
end defaultErr.impl;

error model implementation defaultErr.impl
end defaultErr.impl;"""
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "K-1");



    @INVALID_CGM_TESTCASE
    def test_same_errormodel_implementation_public_private(self):
        slim = """
package BigPackage
    public        
        error model defaultErr
            features
                state1 : initial state;
        end defaultErr;
        
        error model implementation defaultErr.impl
        end defaultErr.impl;
    
    private
    
        error model implementation defaultErr.impl
        end defaultErr.impl;
end BigPackage;"""
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "K-1");

    @INVALID_CGM_TESTCASE
    def test_missing_errormodel_type_for_implementation_1(self):
        slim = """
package BigPackage
    public        
        error model implementation anotherErr.impl
        end anotherErr.impl;

        error model defaultErr
            features
                state1: initial state;
        end defaultErr;
end BigPackage;"""
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "K-2");

    @INVALID_CGM_TESTCASE    
    def test_same_error_event_identifiers(self):
        slim = """
error model defaultErr
    features
        state1: initial state;
end defaultErr;

error model implementation defaultErr.impl
        events
                eve2: error event occurrence poisson 25; 
                eve2: error event;
end defaultErr.impl;"""
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "K-4");

    @INVALID_CGM_TESTCASE
    def test_error_transition_source_not_exists(self):
        slim = """
error model defaultErr
    features
                hidden1: initial state;
end defaultErr;

error model implementation defaultErr.impl
        transitions
                hidden2345 -[reset]-> hidden1;
end defaultErr.impl;"""
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "K-5");



    @INVALID_CGM_TESTCASE
    def test_error_transition_target_not_exists(self):
        slim = """
error model defaultErr
    features
        hidden1: initial state;
end defaultErr;

error model implementation defaultErr.impl
        transitions
                hidden1 -[reset]-> hidden2345;
end defaultErr.impl;"""
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "K-5");


    def test_error_transition_trigger_not_exists(self):
        slim = """
error model defaultErr
    features
        hidden1: initial state;
        hidden2: error state;
end defaultErr;

error model implementation defaultErr.impl
        transitions
                hidden1 -[teser]-> hidden2;
end defaultErr.impl;"""
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "K-6");
            
            
    def test_inconsistent_error_transitions(self):
        slim = """
error model defaultErr
    features
        st1: activation state;
        st2: error state;
end defaultErr;

error model implementation defaultErr.impl
        events
                eve1: error event occurrence poisson 300; 
        transitions
                st1 -[eve1]-> st2;
                st1 -[reset]-> st1;
                
end defaultErr.impl;"""
        ci = self._parse_only(slim)
        # Ticket 564
        #assert self._expect_errors(ci, "K-7");            

    def test_nondeterminism_in_errormodel(self):
        slim = """
 error model defaultErr
     features
         st1: activation state;
         st2: error state;
 end defaultErr;
 
 error model implementation defaultErr.impl
         events
                 eve1: error event occurrence poisson 300; 
         transitions
                 st1 -[eve1]-> st2;
                 st1 -[eve1]-> st1;
 end defaultErr.impl;"""
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "K-8");

    def test_nondeterminism_in_errormodel2(self):
        slim = """
 error model defaultErr
     features
         st1: activation state;
         st2: error state;
 end defaultErr;
 
 error model implementation defaultErr.impl
         events
                 eve1: error event occurrence poisson 300; 
         transitions
                 st1 -[@activation]-> st2;
                 st1 -[@activation]-> st1;
 end defaultErr.impl;"""
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "K-8");

    def test_error_state_unreachable(self):
        slim = """
error model defaultErr
    features
        start: activation state;
        stop : error state;
end defaultErr;

error model implementation defaultErr.impl
        events
                eve1: error event occurrence poisson 300;
        transitions
                start -[eve1]-> start;
                stop -[reset]-> stop;
end defaultErr.impl;"""
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "K-9");


    def test_same_error_transition_1(self):
        slim = """
error model defaultErr
    features
        hidden1: initial state;
        hidden3: error state;
end defaultErr;

error model implementation defaultErr.impl
        transitions
                hidden1 -[reset]-> hidden3;
                hidden1 -[reset]-> hidden3;
end defaultErr.impl;"""
        ci = self._parse_only(slim)
        # previously "K-8", too
        assert self._expect_errors(ci, "K-8", "K-10"); 


    @INVALID_CGM_TESTCASE
    def test_same_error_transition_2(self):
        slim = """
error model defaultErr
    features
        deInProp1: in error propagation;
        hidden1: initial state;
        hidden3: error state;
end defaultErr;

error model implementation defaultErr.impl
        transitions
                hidden1 -[deInProp1]-> hidden3;
                hidden1 -[deInProp1]-> hidden3;
end defaultErr.impl;"""
        ci = self._parse_only(slim)
        # previously "K-8", too
        assert self._expect_errors(ci, "K-8", "K-10");


    @INVALID_CGM_TESTCASE
    def test_mismatch_end_error_impl_identifier(self):
        slim = """
error model defaultErr
    features
        deInProp1: in error propagation;
        hidden1: initial state;
        hidden3: error state;
end defaultErr;

error model implementation defaultErr.impl
        transitions
                hidden1 -[deInProp1]-> hidden3;
end faultyIdentifier.impl;"""
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "K-11");

    @INVALID_CGM_TESTCASE
    def test_propagation_event_same_identifier(self):
        slim = """
error model defaultErr
    features
        deInProp1: in error propagation;
end defaultErr;

error model implementation defaultErr.impl
    events
        deInProp1: error event;
    states
        state1: initial state;
end defaultErr.impl;"""
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "K-12");

    @INVALID_CGM_TESTCASE
    def test_poisson_rate_1(self):
        slim = """
error model defaultErr
    features
        state1: initial state;
end defaultErr;

error model implementation defaultErr.impl
    events
        evt : error event occurrence poisson 0.0;
end defaultErr.impl;"""
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "K-13");

    @INVALID_CGM_TESTCASE
    def test_poisson_rate_2(self):
        slim = """
error model defaultErr
    features
        state1: initial state;
end defaultErr;

error model implementation defaultErr.impl
    events
        evt : error event occurrence poisson -0.1;
end defaultErr.impl;"""
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "K-13");

    @INVALID_CGM_TESTCASE
    def test_distinct_state(self):
        slim = """
error model defaultErr
end defaultErr;

error model implementation defaultErr.impl
    events
        evt: error event;
    states
        state1: initial state;
        state1: error state;
    transitions
        state1 -[evt]-> state1;
end defaultErr.impl;"""
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "K-14");

    @INVALID_CGM_TESTCASE
    def test_state_impl_and_type(self):
        slim = """
error model defaultErr
    features
        state1: initial state;
end defaultErr;

error model implementation defaultErr.impl
    events
        evt: error event;
    states
        state1: initial state;
end defaultErr.impl;"""
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "K-15");

    @INVALID_CGM_TESTCASE
    def test_starting_state_1(self):
        slim = """
error model defaultErr
end defaultErr;

error model implementation defaultErr.impl
    states
        state1: error state;
end defaultErr.impl;"""
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "K-16", "K-9");

    @INVALID_CGM_TESTCASE
    def test_starting_state_2(self):
        slim = """
error model defaultErr
end defaultErr;

error model implementation defaultErr.impl
    events
        evt: error event;
    states
        state1: initial state;
        state2: activation state;
    transitions
        state1 -[evt]-> state2;
end defaultErr.impl;"""
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "K-16");

    @INVALID_CGM_TESTCASE
    def test_distinct_clock(self):
        slim = """
error model defaultErr
end defaultErr;

error model implementation defaultErr.impl
    clocks
        clk1: data clock;
        clk1: data clock;
    states
        state1: initial state;
end defaultErr.impl;"""
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "K-17");

    @INVALID_CGM_TESTCASE
    def test_invariant_refers_to_clock(self):
        slim = """
error model defaultErr
end defaultErr;

error model implementation defaultErr.impl
    clocks
        clk1: data clock;
    states
        state1: initial state while clk2 >= 0;
end defaultErr.impl;"""
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "K-18");

    @INVALID_CGM_TESTCASE
    def test_invariant_is_convex(self):
        slim = """
error model defaultErr
end defaultErr;

error model implementation defaultErr.impl
    clocks
        clk1: data clock;
    states
        state1: initial state while clk1 >= 3 or clk1 <= 1;
end defaultErr.impl;"""
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "K-19");

    @INVALID_CGM_TESTCASE
    def test_invariant_type(self):
        slim = """
error model defaultErr
end defaultErr;

error model implementation defaultErr.impl
    clocks
        clk1: data clock;
    states
        state1: initial state while clk1 * 2;
end defaultErr.impl;"""
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "K-20");

    @INVALID_CGM_TESTCASE
    def test_urgent_bounds(self):
        slim = """
error model defaultErr
end defaultErr;

error model implementation defaultErr.impl
    states
        state1: initial state urgent in -1.0;
end defaultErr.impl;"""
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "K-21");

    @INVALID_CGM_TESTCASE
    def test_guard_refers_clocks(self):
        slim = """
error model defaultErr
end defaultErr;

error model implementation defaultErr.impl
    events
        evt: error event;
    clocks
        clk1: data clock;
    states
        state1: initial state;
    transitions
        state1 -[evt when clk3 = 0]-> state1;
end defaultErr.impl;"""
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "K-22");

    @INVALID_CGM_TESTCASE
    def test_guard_type(self):
        slim = """
error model defaultErr
end defaultErr;

error model implementation defaultErr.impl
    events
        evt: error event;
    clocks
        clk1: data clock;
    states
        state1: initial state;
    transitions
        state1 -[evt when clk1 + 0]-> state1;
end defaultErr.impl;"""
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "K-23");

    @INVALID_CGM_TESTCASE
    def test_within_bounds(self):
        slim = """
error model defaultErr
end defaultErr;

error model implementation defaultErr.impl
    events
        evt: error event;
    states
        state1: initial state urgent in 0.0;
    transitions
        state1 -[evt within 1.0 to 0.0]-> state1;
end defaultErr.impl;"""
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "K-24");

    @INVALID_CGM_TESTCASE
    def test_distinct_effect(self):
        slim = """
error model defaultErr
end defaultErr;

error model implementation defaultErr.impl
    events
        evt: error event;
    clocks
        clk1: data clock;
    states
        state1: initial state;
    transitions
        state1 -[evt then clk1 := 0; clk1 := 0]-> state1;
end defaultErr.impl;"""
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "K-25");

    @INVALID_CGM_TESTCASE
    def test_effect_clock_exists(self):
        slim = """
error model defaultErr
end defaultErr;

error model implementation defaultErr.impl
    events
        evt: error event;
    clocks
        clk1: data clock;
    states
        state1: initial state;
    transitions
        state1 -[evt then clk1 := 0; clk3 := 0]-> state1;
end defaultErr.impl;"""
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "K-26");

    @INVALID_CGM_TESTCASE
    def test_prob_state_has_invariant(self):
        slim = """
error model defaultErr
end defaultErr;

error model implementation defaultErr.impl
    events
        evt: error event occurrence poisson 0.5;
    clocks
        clk1: data clock;
    states
        state1: initial state while clk1 >= 0;
    transitions
        state1 -[evt]-> state1;
end defaultErr.impl;"""
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "K-27");

    @INVALID_CGM_TESTCASE
    def test_prob_state_is_urgent(self):
        slim = """
error model defaultErr
end defaultErr;

error model implementation defaultErr.impl
    events
        evt: error event occurrence poisson 0.5;
    states
        state1: initial state urgent in 1.0;
    transitions
        state1 -[evt]-> state1;
end defaultErr.impl;"""
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "K-28");

    @INVALID_CGM_TESTCASE
    def test_prob_state_has_guard(self):
        slim = """
error model defaultErr
end defaultErr;

error model implementation defaultErr.impl
    events
        evt: error event occurrence poisson 0.5;
    clocks
        clk1: data clock;
    states
        state1: initial state;
    transitions
        state1 -[evt when clk1 >= 0]-> state1;
end defaultErr.impl;"""
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "K-29");

    @INVALID_CGM_TESTCASE
    def test_prob_state_has_within(self):
        slim = """
error model defaultErr
end defaultErr;

error model implementation defaultErr.impl
    events
        evt: error event occurrence poisson 0.5;
    states
        state1: initial state urgent in 1.0;
    transitions
        state1 -[evt within 0.0 to 1.0]-> state1;
end defaultErr.impl;"""
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "K-28", "K-30");

    def test_trans_reactivation_has_guard(self):
        slim = """
error model defaultErr
end defaultErr;

error model implementation defaultErr.impl
    clocks
        clk: data clock;
    states
        state1: initial state;
    transitions
        state1 -[@activation when clk >= 0.0]-> state1;
end defaultErr.impl;"""
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "K-31");
        return

    def test_trans_reactivation_has_within(self):
        slim = """
error model defaultErr
end defaultErr;

error model implementation defaultErr.impl
    states
        state1: initial state;
    transitions
        state1 -[@activation within 0.0 to 1.0]-> state1;
end defaultErr.impl;"""
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "K-32");
        return

pass

class Test_13TimeScalesTestCaseNeg(BasicTestCaseNeg):
    @INVALID_CGM_TESTCASE
    def test_unscaled_clock(self):
        slim = """
system aSysScaled
end aSysScaled;
system implementation aSysScaled.Impl
subcomponents
    clk : data clock min;
    sub : system aSys;
end aSysScaled.Impl;

system aSys
end aSys;
system implementation aSys.Impl
subcomponents
    clk : data clock;
end aSys.Impl;"""
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "F-19");

    @INVALID_CGM_TESTCASE
    def test_unscaled_trajectory(self):
        slim = """
system aSysScaled
end aSysScaled;
system implementation aSysScaled.Impl
subcomponents
    clk : data clock min;
    sub : system aSys;
end aSysScaled.Impl;

system aSys
end aSys;
system implementation aSys.Impl
subcomponents
    cnt : data continuous default 0.0;
modes
    m0 : initial mode while cnt' = 0;
end aSys.Impl;"""
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "I-24");

    @INVALID_CGM_TESTCASE
    def test_unscaled_invariant(self):
        slim = """
system aSysScaled
end aSysScaled;
system implementation aSysScaled.Impl
subcomponents
    clk : data clock min;
    sub : system aSys;
end aSysScaled.Impl;

system aSys
end aSys;
system implementation aSys.Impl
subcomponents
    clk : data clock min;
modes
    m0 : initial mode while clk < 12.0;
end aSys.Impl;"""
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "C-3");

    @INVALID_CGM_TESTCASE
    def test_unscaled_urgent(self):
        slim = """
system aSysScaled
end aSysScaled;
system implementation aSysScaled.Impl
subcomponents
    clk : data clock min;
    sub : system aSys;
end aSysScaled.Impl;

system aSys
end aSys;
system implementation aSys.Impl
modes
    m0 : initial mode urgent in 4.0;
end aSys.Impl;"""
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "I-25");

    @INVALID_CGM_TESTCASE
    def test_unscaled_guard(self):
        slim = """
system aSysScaled
end aSysScaled;
system implementation aSysScaled.Impl
subcomponents
    clk : data clock min;
    sub : system aSys;
end aSysScaled.Impl;

system aSys
end aSys;
system implementation aSys.Impl
subcomponents
    clk : data clock min;
modes
    m0 : initial mode;
transitions
    m0 -[when clk > 2.0]-> m0;
end aSys.Impl;"""
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "C-3");

    @INVALID_CGM_TESTCASE
    def test_unscaled_within_1(self):
        slim = """
system aSysScaled
end aSysScaled;
system implementation aSysScaled.Impl
subcomponents
    clk : data clock min;
    sub : system aSys;
end aSysScaled.Impl;

system aSys
end aSys;
system implementation aSys.Impl
subcomponents
    clk : data clock min;
modes
    m0 : initial mode urgent in 2.0 sec;
transitions
    m0 -[within 1.0 to 2.0]-> m0;
end aSys.Impl;"""
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "I-26");

    @INVALID_CGM_TESTCASE
    def test_unscaled_within_2(self):
        slim = """
system aSysScaled
end aSysScaled;
system implementation aSysScaled.Impl
subcomponents
    clk : data clock min;
    sub : system aSys;
end aSysScaled.Impl;

system aSys
end aSys;
system implementation aSys.Impl
subcomponents
    clk : data clock min;
modes
    m0 : initial mode urgent in 2.0 sec;
transitions
    m0 -[within 1.0 to 2.0 sec]-> m0;
end aSys.Impl;"""
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "I-26");

    @INVALID_CGM_TESTCASE
    def test_unscaled_within_3(self):
        slim = """
system aSysScaled
end aSysScaled;
system implementation aSysScaled.Impl
subcomponents
    clk : data clock min;
    sub : system aSys;
end aSysScaled.Impl;

system aSys
end aSys;
system implementation aSys.Impl
subcomponents
    clk : data clock min;
modes
    m0 : initial mode urgent in 2.0 sec;
transitions
    m0 -[within 1.0 sec to 2.0]-> m0;
end aSys.Impl;"""
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "I-26");

    @INVALID_CGM_TESTCASE
    def test_unscaled_effect(self):
        slim = """
system aSysScaled
end aSysScaled;
system implementation aSysScaled.Impl
subcomponents
    clk : data clock min;
    sub : system aSys;
end aSysScaled.Impl;

system aSys
end aSys;
system implementation aSys.Impl
subcomponents
    clk : data clock min;
    r : data real default 0.0;
modes
    m0 : initial mode;
transitions
    m0 -[then r := clk]-> m0;
end aSys.Impl;"""
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "C-3");

    pass

class Test_14TimeScalesErrTestCaseNeg(BasicTestCaseNeg):
    @INVALID_CGM_TESTCASE
    def test_unscaled_clock(self):
        slim = """
system aSysScaled
end aSysScaled;
system implementation aSysScaled.Impl
subcomponents
    clk : data clock min;
end aSysScaled.Impl;

error model aErr
end aErr;
error model implementation aErr.Impl
clocks
    clk : data clock;
states
    e0 : initial state;
end aErr.Impl;"""
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "K-34");

    @INVALID_CGM_TESTCASE
    def test_unscaled_invariant(self):
        slim = """
system aSysScaled
end aSysScaled;
system implementation aSysScaled.Impl
subcomponents
    clk : data clock min;
end aSysScaled.Impl;

error model aErr
end aErr;
error model implementation aErr.Impl
clocks
    clk : data clock min;
states
    e0 : initial state while clk < 12.0;
end aErr.Impl;"""
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "C-3");

    @INVALID_CGM_TESTCASE
    def test_unscaled_urgent(self):
        slim = """
system aSysScaled
end aSysScaled;
system implementation aSysScaled.Impl
subcomponents
    clk : data clock min;
end aSysScaled.Impl;

error model aErr
end aErr;
error model implementation aErr.Impl
states
    e0 : initial state urgent in 4.0;
end aErr.Impl;"""
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "K-35");

    @INVALID_CGM_TESTCASE
    def test_unscaled_event(self):
        slim = """
system aSysScaled
end aSysScaled;
system implementation aSysScaled.Impl
subcomponents
    clk : data clock min;
end aSysScaled.Impl;

error model aErr
end aErr;
error model implementation aErr.Impl
events
    evt : error event occurrence poisson 0.05;
states
    e0 : initial state;
end aErr.Impl;"""
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "K-33");

    @INVALID_CGM_TESTCASE
    def test_unscaled_guard(self):
        slim = """
system aSysScaled
end aSysScaled;
system implementation aSysScaled.Impl
subcomponents
    clk : data clock min;
end aSysScaled.Impl;

error model aErr
end aErr;
error model implementation aErr.Impl
events
    evt : error event;
clocks
    clk : data clock min;
states
    e0 : initial state;
transitions
    e0 -[evt when clk > 2.0]-> e0;
end aErr.Impl;"""
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "C-3");

    @INVALID_CGM_TESTCASE
    def test_unscaled_within_1(self):
        slim = """
system aSysScaled
end aSysScaled;
system implementation aSysScaled.Impl
subcomponents
    clk : data clock min;
end aSysScaled.Impl;

error model aErr
end aErr;
error model implementation aErr.Impl
events
    evt : error event;
states
    e0 : initial state urgent in 2.0 sec;
transitions
    e0 -[evt within 1.0 to 2.0]-> e0;
end aErr.Impl;"""
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "K-36");

    @INVALID_CGM_TESTCASE
    def test_unscaled_within_2(self):
        slim = """
system aSysScaled
end aSysScaled;
system implementation aSysScaled.Impl
subcomponents
    clk : data clock min;
end aSysScaled.Impl;

error model aErr
end aErr;
error model implementation aErr.Impl
events
    evt : error event;
states
    e0 : initial state urgent in 2.0 sec;
transitions
    e0 -[evt within 1.0 to 2.0 sec]-> e0;
end aErr.Impl;"""
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "K-36");

    @INVALID_CGM_TESTCASE
    def test_unscaled_within_3(self):
        slim = """
system aSysScaled
end aSysScaled;
system implementation aSysScaled.Impl
subcomponents
    clk : data clock min;
end aSysScaled.Impl;

error model aErr
end aErr;
error model implementation aErr.Impl
events
    evt : error event;
states
    e0 : initial state urgent in 2.0 sec;
transitions
    e0 -[evt within 1.0 sec to 2.0]-> e0;
end aErr.Impl;"""
        ci = self._parse_only(slim)
        assert self._expect_errors(ci, "K-36");

    pass


if __name__ == "__main__":
    global sxml_output
    sxml_output = False
    if len(sys.argv) > 1:
        if sys.argv[1] == "-sxml":
            sxml_output = True
            sys.argv = [sys.argv[0]]
    unittest.main()
