#!/usr/bin/python

# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================



"""
Organisation of this test suite. 
================================

This test suite is organised around the sections occurring in the SLIM
specification document. Every (sub)section is a separate class. Every 
class that is a test suite must derive from BasicTestCase. 

This is the test suite for positive test cases, i.e. the cases which should
not rise any errors.
"""


import sys
sys.path.append('../resources/antlr/antlr-python-runtime-3.1.1/')
sys.path.append('../../')

from invokers import dump, mrmc, nusmv, slim
from utils import debug_info as debug
from utils import log
import options
from slim_compiler.src.dataStructure.compile import CompileInformation

import unittest
import tempfile



class BasicTestCasePos(unittest.TestCase):

    def _parse_only(self, slim_snippet):
        # generate temp file
        f = tempfile.NamedTemporaryFile()
        try:
            f.write(slim_snippet)
            f.flush()
            # set up new compile info object
            cinfo = CompileInformation()
            slim.parse(cinfo, [f.name])
            slim.check(cinfo)
        finally:
            f.close()
        return cinfo
    
    def _has_no_errors(self, cinfo):
        no_error = cinfo.error_collector.has_no_errors()
        if not no_error:
            print cinfo.error_collector.get_error_id_set()
        return no_error

pass

class SystemTestCasePos(BasicTestCasePos):
    def test_nonblocking_ports(self):
        slim = """
system sys
end sys;

system implementation sys.impl
    subcomponents
        subsys_out : system subsystem_out.impl;
        subsys_in1 : system subsystem_in.impl in modes(m_enabled);
        subsys_in2 : system subsystem_in.impl;
--    connections
--        event port subsys_out.evt -> subsys_in1.evt;
--        subsys_out.evt -> subsys_in2.evt passive;
    modes
        m_enabled : initial mode;
        m_disabled : mode;
    transitions
        m_enabled -[ ]-> m_disabled;
end sys.impl;

system subsystem_out
    features
        evt : out event port nonblocking;
end subsystem_out;

system implementation subsystem_out.impl
    modes
        m0 : initial mode;
    transitions
        m0 -[ evt ]-> m0;
end subsystem_out.impl;

system subsystem_in
    features
        evt : in event port;
end subsystem_in;

system implementation subsystem_in.impl
    modes
        m0 : initial mode;
    transitions
        m0 -[ evt ]-> m0;
end subsystem_in.impl;
"""
        ci = self._parse_only(slim)
        print ci.error_collector
        self.assertTrue(self._has_no_errors(ci))
    
    def test_package_identifiers_not_distinct(self):
        slim = """
package aPackage
    public
        system aSys
        end aSys;
end aPackage;

package anotherPackage
    public
        system aSys
        end aSys;
end anotherPackage;
"""
        ci = self._parse_only(slim)
        self.assertTrue( self._has_no_errors(ci) )

    def test_multiple_event_trigger(self):
        slim = """
device D
    features
        e1: in event port;
        e2: in event port;
        e3: in event port;
end D;

device implementation D.Imp
    modes
        m: initial mode;
    transitions
        m -[e1 or e2]-> m;
        m -[e1 or e2 or e3]-> m;
end D.Imp;
"""
        ci = self._parse_only(slim)
        self.assertTrue( self._has_no_errors(ci) )

    def test_event_port_fan_out_mutex(self):
        slim = """
package aPackage
    public    
        system aSys
            features
                out1: out event port;
                out2: out event port;
        end aSys;
        
        system implementation aSys.Impl
            subcomponents
                subSys: system anotherSys.Impl;
            connections
                event port subSys.outport -> out1 in modes (m0);
                event port subSys.outport -> out2 in modes (m1);
            modes
                m0 : initial mode;
                m1 : mode;
            transitions
                m0 -[]-> m1;
                m1 -[]-> m0;
        end aSys.Impl;
        
        system anotherSys
            features
                outport: out event port;
        end anotherSys;
        
        system implementation anotherSys.Impl
        end anotherSys.Impl;
end aPackage;
"""
        ci = self._parse_only(slim)
        self.assertTrue( self._has_no_errors(ci) )

    def test_root_data_no_default(self):
        slim = """
package aPackage
    public
        system aSys
        features
            outData : out data port bool default false;
            inData : in data port bool;
        end aSys;
        
        system implementation aSys.impl
        subcomponents
            bsys : system bSys.impl;
        end aSys.impl;
        
        system bSys
        features
            outData : out data port bool default true;
            inData : in data port bool default true;
        end bSys;
        
        system implementation bSys.impl
        end bSys.impl;
end aPackage;
"""
        ci = self._parse_only(slim)
        self.assertTrue( self._has_no_errors(ci) )

    def test_transition_multiple_event(self):
        slim = """
system Sys
  features
    ev1: in event port;
    ev2: in event port;
    ev3: in event port;
end Sys;

system implementation Sys.Imp
  subcomponents
    delay : data clock;
  modes
    m0 : initial mode;
  transitions
    m0 -[ ev1 or ev2 or ev3 then delay := 0 ]-> m0;
end Sys.Imp;
"""
        ci = self._parse_only(slim)
        self.assertTrue( self._has_no_errors(ci) )
        
    def test_transition_multiple_source(self):
        slim = """
system Sys
  features
    ev1: in event port;
end Sys;

system implementation Sys.Imp
  subcomponents
    delay : data clock;
  modes
    m0 : initial mode;
    m1 : mode;
    m2 : mode;
  transitions
    * -[ ev1 then delay := 0 ]-> m0;
    m0 -[]-> m1;
    m0 -[]-> m2;
end Sys.Imp;
"""
        ci = self._parse_only(slim)
        self.assertTrue( self._has_no_errors(ci) )

    def test_case_enums(self):
        slim = """
system Sys
  features
    inData: in data port enum (a, b, c, d);
    outData: out data port enum (e, f, g, h) default e;
end Sys;

system implementation Sys.impl
  flows
    outData := case inData=a:e; inData=b:f; inData=c:g; inData=d:h otherwise e end;
end Sys.impl;
"""
        ci = self._parse_only(slim)
        self.assertTrue( self._has_no_errors(ci) )
        
    def test_case_enums2(self):
        slim = """
system Sys
  features
    inData: in data port enum (a, b, c, d);
    outData: out data port enum (e, f, g, h) default e;
end Sys;

system implementation Sys.impl
  flows
    outData := case inData=a:e; inData=b:f; inData=c:g; inData=d:h otherwise e end;
end Sys.impl;
"""
        ci = self._parse_only(slim)
        self.assertTrue( self._has_no_errors(ci) )
        
    def test_mixed_error_transitions(self):
        slim = """
error model defaultErr
    features
        st1: activation state;
        st2: error state;
end defaultErr;

error model implementation defaultErr.impl
        events
                eve1: error event occurrence poisson 300; 
        transitions
                st1 -[eve1]-> st2;
                st1 -[reset]-> st1;
                
end defaultErr.impl;"""
        ci = self._parse_only(slim)
        self.assertTrue( self._has_no_errors(ci) )

    def test_mixed_error_transitions_2(self):
        slim = """
error model defaultErr
    features
        st1: activation state;
        st2: error state;
        prop: in error propagation;
end defaultErr;

error model implementation defaultErr.impl
        events
                eve1: error event occurrence poisson 300; 
        transitions
                st1 -[eve1]-> st2;
                st1 -[prop]-> st1;
                
end defaultErr.impl;"""
        ci = self._parse_only(slim)
        self.assertTrue( self._has_no_errors(ci) )
    def test_mixed_error_transitions_3(self):
        slim = """
error model defaultErr
    features
        st1: activation state;
        st2: error state;
        prop: out error propagation;
end defaultErr;

error model implementation defaultErr.impl
        events
                eve1: error event occurrence poisson 300; 
        transitions
                st1 -[eve1]-> st2;
                st1 -[prop]-> st1;
                
end defaultErr.impl;"""
        ci = self._parse_only(slim)
        self.assertTrue( self._has_no_errors(ci) )
pass

if __name__ == "__main__":
    unittest.main()
