# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================

import antlr3

from src.models.tfpg_structural import Tfpg as TfpgStructural
from grammar.generated.TfpgLexer import TfpgLexer
from grammar.generated.TfpgParser import TfpgParser


class ConcreteSyntaxParser():
    """Class that converts the TFPG described by a human-readable file
    into an internal tfpg_graph"""

    def __init__(self):
        """Class constructor"""
        self.tfpg_graph = TfpgStructural()
        return


    def build_graph(self, input_file):
        """Parses the human-readable file and builds the graph

        :param input_file: the name of the input human-readable file
        "type input_file: string

        """

        # build the tree to parse properly the file
        char_stream = antlr3.ANTLRFileStream(input_file)
        lexer = TfpgLexer(char_stream)
        tokens = antlr3.CommonTokenStream(lexer)
        parser = TfpgParser(tokens)

        r = parser.tfpg()
        t = r.tree

        for child in t.getChildren():

            if str(child) == "NAME":
                self.tfpg_graph.set_name(str(child.getChildren()[0]))

            elif str(child) == "FM":
                if len(child.getChildren()) == 1:
                    self.tfpg_graph.add_fm_node(str(child.getChildren()[0]))
                else:
                    errors.append(FMNotCorrectError)

            elif str(child) == "AND":
                if len(child.getChildren()) == 2:
                    self.tfpg_graph.add_and_node(str(child.getChildren()[0]), True)
                else:
                    self.tfpg_graph.add_and_node(str(child.getChildren()[0]), False)

            elif str(child) == "OR":
                if len(child.getChildren()) == 2:
                    self.tfpg_graph.add_or_node(str(child.getChildren()[0]), True)
                else:
                    self.tfpg_graph.add_or_node(str(child.getChildren()[0]), False)

            elif str(child) == "MODES":
                modes = []
                for ch in child.getChildren():
                    modes.append(str(ch))
                self.tfpg_graph.set_modes_list(modes)

            elif str(child) == "EDGE":
                edg_name = str(child.getChildren()[0])
                edg_src = str(child.getChildren()[1])
                edg_dst = str(child.getChildren()[2])
                edg_tMin = float(str(child.getChildren()[3]))
                if str(child.getChildren()[4]) == "INFINITY":
                    edg_tMax = -1
                else:
                    edg_tMax = float(str(child.getChildren()[4]))
                 
                 
                edg_modes = [] 
                if str(child.getChildren()[5]) == "ALL": 
                    for m in modes:
                        edg_modes.append(str(m))
                else:      
                    for ch in (child.getChildren()[5]).getChildren():
                        edg_modes.append(str(ch))
                        
                self.tfpg_graph.add_edge(self.tfpg_graph.get_node_by_name(edg_src), 
                                         edg_tMin, edg_tMax, edg_modes,
                                         self.tfpg_graph.get_node_by_name(edg_dst), 
                                         edg_name)

        errors = self.tfpg_graph.get_errors_list()
        if errors != []:
            for err in errors:
                raise err
       
        return self.tfpg_graph
