# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================

import os.path
import copy

from lxml import etree

# from slim_compiler.src.model_extension import errors
from src.models.tfpg_structural import Tfpg as Tfpg_Struct
from src.models.tfpg_structural import Edge as Tfpg_Edge
from src.models.tfpg_structural import Node as Tfpg_Node
from src.models.tfpg_structural import NodeType as Tfpg_NodeType
from src.models.tfpg_structural_errors import NotExistingNode, FMNotCorrectError


def tfpg_doc_to_tfpg_internal_struct(tfpg_doc):
    """Converts the TfpgDocument to the internal structure Tfpg.
       @param   tfpg_doc:  The tfpg document.
       @type    tfpg_doc:  TfpgDocument
          
       @return   tfpg_internal_struct:  The internal structure of Tfpg.
       @rtype    tfpg_internal_struct:  Tfpg
    """
    tfpg_internal_struct = Tfpg_Struct()

    nodes = tfpg_doc.nodes_list
    edges = tfpg_doc.edges_list
    modes = tfpg_doc.modes_list
    name = tfpg_doc.name

    tfpg_internal_struct_name = tfpg_internal_struct.set_name(name)

    for node in nodes:
        node_type = node.type
        node_name = node.name
        if node.isMonitored.lower() == "true":
            node_isMonitored = True
        else:
            node_isMonitored = False

        if node_type == "FM":
            if node_isMonitored is True:
                tfpg_internal_struct.get_errors_list().append(FMNotCorrectError(node_name))
                return tfpg_internal_struct
            else:
                tfpg_internal_struct.add_fm_node(node_name)
            
        elif node_type == "AND":
            tfpg_internal_struct.add_and_node(node_name, node_isMonitored)

        elif node_type == "OR":
            tfpg_internal_struct.add_or_node(node_name, node_isMonitored)

    lst = []
    for md in modes:
        lst.append(str(md.text))
    tfpg_internal_struct.set_modes_list(lst)

    for edge in edges:

        edge_srcNode_name = edge.srcNode
        edge_tMin = edge.tMin
        edge_tMax = edge.tMax
        edge_modes = edge.modes
        edge_destNode_name = edge.destNode
        edge_name = edge.name

        edge_modes_list = []
        for mode in edge_modes:
            edge_modes_list.append(mode.text)

        ## GZ: in a TfpgDocument structure, the source node is not represented as a Node,
        ##     but as a name
        if tfpg_internal_struct.has_node_by_name(edge_srcNode_name) is True:
            edge_srcNode = tfpg_internal_struct.get_node_by_name(edge_srcNode_name)
        else:
            tfpg_internal_struct.get_errors_list().append(NotExistingNode(edge_srcNode_name, 'src'))
            return tfpg_internal_struct

        ## GZ: in a TfpgDocument structure, the destination node is not represented as a Node,
        ##     but as a name
        if tfpg_internal_struct.has_node_by_name(edge_destNode_name) is True:
            edge_destNode = tfpg_internal_struct.get_node_by_name(edge_destNode_name)
        else:
            tfpg_internal_struct.get_errors_list().append(NotExistingNode(edge_destNode_name, 'dst'))
            return tfpg_internal_struct

        tfpg_internal_struct.add_edge(edge_srcNode, edge_tMin, edge_tMax, edge_modes_list,
                                      edge_destNode, edge_name)


    return tfpg_internal_struct

pass # end of tfpg_doc_to_tfpg_internal_struct
# ----------------------------------------------------------------------------    


def tfpg_internal_struct_to_tfpg_doc(tfpg_internal_struct):
    """Converts the internal structure Tfpg to TfpgDocument.
       @param   tfpg_internal_struct:  The tfpg internal structure
       @type    tfpg_internal_struct:  Tfpg
          
       @return   tfpg_doc:  The tfpg document.
       @rtype    tfpg_doc:  TfpgDocument
    """
    nodes_list = nodesList()
    edges_list = edgesList()
    modes_list = modesList()
    
    nodes = tfpg_internal_struct.get_nodes_list()
    edges = tfpg_internal_struct.get_edges_list()
    modes = tfpg_internal_struct.get_modes_list()
    name = tfpg_internal_struct.get_name()


    for node in nodes:
        node_type = node.get_type()
        node_name = node.get_name()
        node_isMonitored = str(node.is_monitored()).lower()

        if node_type == Tfpg_NodeType.FM:
            node_type_str = "FM"
        elif node_type == Tfpg_NodeType.AND:
            node_type_str = "AND"
        elif node_type == Tfpg_NodeType.OR:
            node_type_str = "OR"
        n = Node().create(node_type_str, node_name, node_isMonitored)
        nodes_list.append(n)

    for md in modes:
        modes_list.add_mode(md)

    for edge in edges:
        edge_srcNode = edge.get_srcNode()
        edge_tMin = edge.get_tMin()
        edge_tMax = edge.get_tMax()
        edge_modes = edge.get_modes()
        edge_destNode = edge.get_destNode()
        edge_name = edge.get_name()

        edge_modes_list = modesList()
        for mode in edge_modes:
            edge_modes_list.add_mode(mode)

        e = Edge().create(edge_srcNode, str(edge_tMin), str(edge_tMax), 
                          edge_modes_list, edge_destNode, edge_name)
        edges_list.append(e)

    tfpg_doc = TfpgDocument().create(nodes_list, modes_list, edges_list, name)

    return tfpg_doc

pass # end of tfpg_internal_struct_to_tfpg_doc
# ----------------------------------------------------------------------------   


## GZ: !!!warning!!! This function is not used, and consequently, neither the 
##     functions get_nodes_list(), get_edges_list(), get_modes_list() and get_tfpg_name()
##     of the TfpgModel class
def create_tfpg_document(model):
    """Creates the Tfpg Document.
       @param   model:  The tfpg model.
       @type    model:  TfpgModel
          
       @return   TfpgDocument:  The document of tfpg.
       @rtype    TfpgDocument:  TfpgDocument
    """
    return TfpgDocument.create(
        create_nodes_document(model.get_nodes_list(str(model.get_tfpg_name()))),
        create_edges_document(model.get_edges_list(str(model.get_tfpg_name()))),
        create_modes_document(model.get_modes_list(str(model.get_tfpg_name()))),
        str(model.get_tfpg_name()))

pass # end of create_tfpg_document
# ----------------------------------------------------------------------------  
 

def create_modes_document(modes_list, format="SLIM", slim_root=None):
    """Creates the Mode Document.
       @param   modes_list:  The list of modes.
       @type    modes_list:  list
       @param   format:  The format of file: SLIM
       @type    format:  string
       @param   slim_root:  The slim root.
       @type    slim_root:  None
          
       @return   modes_document:  The document of modes.
       @rtype    modes_document:  Mode
    """
    modes_document = modesList.create()

    for md in modes_list:
        md_name = md.mode
        element = modes_document.get_by_name(md_name)

    return modes_document

pass # end of create_modes_document
# ----------------------------------------------------------------------------  
 

def create_edges_document(edges_list, format="SLIM", slim_root=None):
    """Creates the Edge Document.
       @param   edges_list:  The list of edges.
       @type    edges_list:  list
       @param   format:  The format of file: SLIM
       @type    format:  string
       @param   slim_root:  The slim root.
       @type    slim_root:  None
          
       @return   edges_document:  The document of edges.
       @rtype    edges_document:  Edge
    """
    edges_document = edgesList.create()

    for edge in edges_list:
        srcNode = edge.srcNode
        tMin = edge.tMin
        tMax = edge.tMax
        modes = edge.modes
        destNode = edge.destNode
        name = edge.name

        element = edges_document.get_by_name(name)

        if element is None:
            new_element = Edge.create(srcNode, tMin, tMax, modes, destNode, name)
            edges_document.add_edge(new_element)
            element = edges_document.get_by_name(name)
 
    return edges_document

pass # end of create_edges_document
# ----------------------------------------------------------------------------  
 

def create_nodes_document(nodes_list, format="SLIM", slim_root=None):
    """Creates the Node Document.
       @param   nodes_list:  The list of nodes.
       @type    nodes_list:  list
       @param   format:  The format of file: SLIM
       @type    format:  string
       @param   slim_root:  The slim root.
       @type    slim_root:  None
          
       @return   nodes_document:  The document of nodes.
       @rtype    nodes_document:  Node
    """
    nodes_document = nodesList.create()

    for node in nodes_list:
        type = node.type
        name = node.name
        isMonitored = str(node.isMonitored)

        if (isMonitored is None):
            isMonitored = False
            

        element = nodes_document.get_by_type_name(type, name)

        if element is None:
            new_element = Node.create(type, name, isMonitored)
            nodes_document.add_node(new_element)
            element = nodes_document.get_by_type_name(type, name)

    return nodes_document

pass # end of create_nodes_document
# ----------------------------------------------------------------------------  
 

class TfpgDocument(etree.ElementBase):
    """Represents the TFPG Document"""

    @staticmethod
    def create(nodes_list, modes_list, edges_list, name):
        """Creates the Tfpg Document.
           @param   nodes_list:  The list of nodes.
	   @type    nodes_list:  list
           @param   modes_list:  The list of modes.
	   @type    modes_list:  list
           @param   edges_list:  The list of edges.
	   @type    edges_list:  list
           @param   name:  The name of the tfpg.
	   @type    name:  string          
          
           @return   edges_document:  The document of tfpg.
	   @rtype    edges_document:  object
        """
        tfpg = _parser.makeelement(
            "{" + _XML_NAMESPACE_URI + "}" + _TFPG_TAG)

        tfpg.append(copy.deepcopy(nodes_list)) 
        tfpg.append(copy.deepcopy(modes_list))
        tfpg.append(copy.deepcopy(edges_list))
        tfpg.set(_TFPG_NAME_ATTRIBUTE, str(name))
        
        return tfpg


    @staticmethod
    def parse(filename):
        """Parse the Tfpg Document.
           @param   filename:  The name of the file to parse.
	   @type    filename:  string
          
           @return   document_tree.getroot():  The document tree parsed.
	   @rtype    document_tree.getroot():  etree
        """
        error_list = []

        try:
            document_tree = etree.parse(filename, _parser)
            return document_tree.getroot(), error_list
            
        except etree.XMLSyntaxError, e:
            error_list.append(errors.NonWellFormedXmlError(str(e)))
            return None, error_list


    @staticmethod
    def get_syntax_errors_in_tfpg_document(tfpg_document):
        """Gets syntax errors from the Tfpg Document.
           @param   tfpg_document:  The Tfpg Document.
	   @type    tfpg_document:  object
          
           @return   error_list:  The list of syntax errors.
	   @rtype    error_list:  list
        """
        error_list = []
        try:
            _tfpg_schema.assertValid(tfpg_document)
        except etree.DocumentInvalid, e:
            error_list.append(errors.DocumentSyntaxError(str(e)))
        return error_list


    nodes_list = property(
        lambda self: self[0],
        doc="The nodes of the Tfpg.\n" +
            "@type: nodesList")

    modes_list = property(
        lambda self: self[1],
        doc="The modes of the Tfpg.\n" +
            "@type: modesList")

    edges_list = property(
        lambda self: self[2],
        doc="The edges of the Tfpg.\n" +
            "@type: edgesList")

    name = property(
        lambda self: self.get(_TFPG_NAME_ATTRIBUTE),
        doc="The Name of the Tfpg.\n" +
            "@type: string")


    def __str__(self):
        """Converts the elements in string format"""
        return etree.tostring(self, xml_declaration=True, encoding="UTF-8",
                              pretty_print=True)

pass # end of TfpgDocument class
# ----------------------------------------------------------------------------


class edgesList(etree.ElementBase):
    """Represents the list of Edges"""

    @staticmethod
    def create():
        """Creates the list of Edges.     
           @return   list of edges:  The document of edges list.
	   @rtype    list of edges:  list
        """
        return _parser.makeelement("{" + _XML_NAMESPACE_URI + "}" +
                                   _EDGES_LIST_TAG)

    def get_by_name(self, name):
        """Gets the edge from the name.  
           @param   name:  The name of the edge.
	   @type    name:  string       
          
           @return   edge:  The edge found.
	   @rtype    edge:  Edge
        """
        for edge in self:
            if edge.name == name:
                return edge
        return None


    def add_edge(self, edge):
        """Adds the edge to the edges list.
           @param   edge:  The added Edge.
	   @type    edge:  Edge
        """
        if not self.get_by_name(edge.name):
            self.append(copy.deepcopy(edge))
        return


    def __str__(self):
        """Converts the elements in string format"""
        return etree.tostring(self, xml_declaration=True, encoding="UTF-8",
                              pretty_print=True)

pass # end of edgesList class
# ----------------------------------------------------------------------------


class Edge(etree.ElementBase):
    """Represents the Edge"""

    @staticmethod
    def create(srcNode, tMin, tMax, modes, destNode, name):
        """Creates the Edge.
           @param   srcNode:  The source node of the Edge.
	   @type    srcNode:  string
           @param   tMin:  The initial time of the Edge.
	   @type    tMin:  float
           @param   tMax:  The final time of the Edge.
	   @type    tMax:  float
           @param   modes:  The modes of the Edge.
	   @type    modes:  list
           @param   destNode:  The destination node of the Edge.
	   @type    destNode:  string
           @param   name:  The name of the Edge.
	   @type    name:  string         
          
           @return   edge:  The document of edge.
	   @rtype    edge:  Edge
        """

        edge = _parser.makeelement(
            "{" + _XML_NAMESPACE_URI + "}" + _EDGE_TAG)

        srcNode_element = _parser.makeelement(
            "{" + _XML_NAMESPACE_URI + "}" + _SRC_NODE_TAG)
        srcNode_element.text = srcNode.get_name()
        edge.append(srcNode_element)

        tMin_element = _parser.makeelement(
            "{" + _XML_NAMESPACE_URI + "}" + _T_MIN_TAG)
        tMin_element.text = tMin
        edge.append(tMin_element)

        tMax_element = _parser.makeelement(
            "{" + _XML_NAMESPACE_URI + "}" + _T_MAX_TAG)
        tMax_element.text = tMax
        edge.append(tMax_element)

	edge.append(copy.deepcopy(modes))

        destNode_element = _parser.makeelement(
            "{" + _XML_NAMESPACE_URI + "}" + _DEST_NODE_TAG)
        destNode_element.text = destNode.get_name()
        edge.append(destNode_element)

        edge.set(_EDGE_NAME_ATTRIBUTE, str(name))

        return edge


    srcNode = property(
        lambda self: self[0].text.strip() if self[0].text is not None else "",
        doc="The name of the srcNode.\n" + 
            "@type: string")

    tMin = property(
        lambda self: self[1].text.strip() if self[1].text is not None else "",
        doc="The initial time of the Propagation Interval.\n" + 
            "@type: float")

    tMax = property(
        lambda self: self[2].text.strip() if self[2].text is not None else "",
        doc="The final time of the Propagation Interval.\n" + 
            "@type: float")

    modes = property(
        lambda self: self[3],
        doc="The modes of the Edge.\n" +
            "@type: modesList")

    destNode = property(
        lambda self: self[4].text.strip() if self[4].text is not None else "",
        doc="The name of the destNode.\n" + 
            "@type: string")

    name = property(
        lambda self: self.get(_EDGE_NAME_ATTRIBUTE),
        doc="The Name of the Edge.\n" +
            "@type: string")

pass # end of Edge class
# ----------------------------------------------------------------------------


class nodesList(etree.ElementBase):
    """Represents the list of Nodes"""

    @staticmethod
    def create():
        """Creates the list of Nodes.     
           @return   list of nodes:  The document of nodes list.
	   @rtype    list of nodes:  list
        """
        return _parser.makeelement("{" + _XML_NAMESPACE_URI + "}" +
                                   _NODES_LIST_TAG)


    def get_by_type_name(self, type, name):
        """Gets the node from the type and the name.  
           @param   type:  The type of the node.
	   @type    type:  string    
           @param   name:  The name of the node.
	   @type    name:  string       
          
           @return   node:  The node found.
	   @rtype    node:  Node
        """
        for node in self:
            if (node.type == type) and (node.name == name):
                return node
        return None


    def add_node(self, node):
        """Adds the node to the nodes list.
           @param   node:  The added Node.
	   @type    node:  Node  
        """
        if not self.get_by_type_name(node.type, node.name):
            self.append(copy.deepcopy(node))
        return


    def __str__(self):
        """Converts the elements in string format"""
        return etree.tostring(self, xml_declaration=True, encoding="UTF-8",
                              pretty_print=True)

pass # end of nodesList class
# ----------------------------------------------------------------------------


class Node(etree.ElementBase):
    """Represents the Node"""

    @staticmethod
    def create(type, name, isMonitored):
        """Creates the Node.
           @param   type:  The type of the Node.
	   @type    type:  string
           @param   name:  The name of the Node.
	   @type    name:  string   
           @param   isMonitored:  If the Node is monitored.
	   @type    isMonitored:  boolean

           @return   node:  The document of node.
	   @rtype    node:  Node
        """

        node = _parser.makeelement(
            "{" + _XML_NAMESPACE_URI + "}" + _NODE_TAG)

        type_element = _parser.makeelement(
            "{" + _XML_NAMESPACE_URI + "}" + _TYPE_TAG)
        type_element.text = type
        node.append(type_element)

        node.set(_NODE_NAME_ATTRIBUTE, str(name))
        node.set(_IS_MONITORED_ATTRIBUTE, str(isMonitored))
        
        return node


    type = property(
        lambda self: self[0].text.strip() if self[0].text is not None else "",
        doc="The type of the Node (FM, AND or OR).\n" + 
            "@type: string")

    name = property(
        lambda self: self.get(_NODE_NAME_ATTRIBUTE),
        doc="The Name of the Node.\n" +
            "@type: string")

    isMonitored = property(
        lambda self: self.get(_IS_MONITORED_ATTRIBUTE),
        doc="If the Node (AND or OR) is monitored.\n" +
            "@type: boolean")

pass # end of Node class
# ----------------------------------------------------------------------------


class modesList(etree.ElementBase):
    """Represents the list of Modes"""

    @staticmethod
    def create():
        """Creates the list of Modes. 
           @return   modes_list:  The document of modes list.
	   @rtype    modes_list:  list
        """
        return _parser.makeelement("{" + _XML_NAMESPACE_URI + "}" + _MODES_LIST_TAG)


    def add_mode(self, mode_name):
        """Adds the mode to the modes list.
           @param    mode: The name of the Mode
           @type     mode: string
        """
        if not self.get_by_name(mode_name):
    
            mode_element = _parser.makeelement(
                "{" + _XML_NAMESPACE_URI + "}" + _MODE_TAG)
            mode_element.text = str(mode_name)
            self.append(mode_element)
        return


    def get_by_name(self, mode_name):
        """Gets the mode from the name.
           @param    name: The name of the mode
           @type     name: string

           @return   mode: The mode found (if any)
           @rtype    mode: Mode
        """
        for m in self:
            if m.text == mode_name:
                return m
        return None


    def get_by_mode(self, mode):
        """Gets the mode from the name of the mode.
           @param   mode:  The mode of the modes_list.
	   @type    mode:  Mode
          
           @return   m:  The mode found.
	   @rtype    m:  Mode
        """
        for m in self:
            if m.mode == mode:
                return m
        return None

    mode = property(
        lambda self: self[0].text.strip() if self[0].text is not None else "",
        doc="The name of the Mode.\n" + 
            "@type: string")

pass # end of modesList class
# ----------------------------------------------------------------------------


# ----------------------------------------------------------------------------
# Note: The following code will be executed once when this module is imported.
# ----------------------------------------------------------------------------

# Define some constants which are used in this module.
# The XML namespace that is used for configurations documents:
_XML_NAMESPACE_URI = ""
# The filename of the XML Schema document that defines the structure of valid
# tfpg documents.
_XSD_FILENAME = os.path.join(os.path.dirname(__file__),
                             "xsd", "tfpg.xsd")

# Define constants for the XML tag names that are used in tfpg
# documents.
_TFPG_TAG = "tfpg"
_TFPG_NAME_ATTRIBUTE = "name"
_EDGES_LIST_TAG = "edgesList"
_NODES_LIST_TAG = "nodesList"
_EDGE_TAG = "edge"
_SRC_NODE_TAG = "srcNode"
_T_MIN_TAG = "tMin"
_T_MAX_TAG = "tMax"
_DEST_NODE_TAG = "destNode"
_EDGE_NAME_ATTRIBUTE = "name"
_NODE_TAG = "node"
_TYPE_TAG = "type"
_NODE_NAME_ATTRIBUTE = "name"
_IS_MONITORED_ATTRIBUTE = "isMonitored"
_MODES_LIST_TAG = "modesList"
_MODE_TAG = "mode"

# Initialize the parser that is used to parse and to create configurations
# specification documents.
# XML comments must be discarded. Otherwise the parser would yield data
# structures cluttered with unwanted elements.
# A lookup scheme is used in order to automatically map certain XML elements
# to instances of the adequate classes which are defined in this module.
_parser = etree.XMLParser(remove_comments=True)
_lookup = etree.ElementNamespaceClassLookup()
_parser.set_element_class_lookup(_lookup)
# Associate the various XML element tags with the corresponding classes. 
# Note that all elements are in the same namespace.
_namespace = _lookup.get_namespace(_XML_NAMESPACE_URI)
_namespace[_TFPG_TAG] = TfpgDocument
_namespace[_EDGES_LIST_TAG] = edgesList
_namespace[_NODES_LIST_TAG] = nodesList
_namespace[_EDGE_TAG] = Edge
_namespace[_NODE_TAG] = Node
_namespace[_MODES_LIST_TAG] = modesList

# Parse the XML Schema document that defines the structure of valid tfpg
# documents. This is needed for validation.
_tfpg_schema_tree = etree.parse(_XSD_FILENAME)
_tfpg_schema = etree.XMLSchema(_tfpg_schema_tree)


