# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================

from tfpg_errors import NodeBigActivationTime, EdgeBigActivationTime


class TFPGTrace:
    """Represents a tfpg trace"""

    tfpg_states = None


    def __init__(self):
        self.tfpg_states = []
        return


    def add_tfpg_state(self, tfpg_state):
        """Adds a tfpg state to the tfpg trace structure

        Notice that no checks are performed, 
        since the validation is done in the TFPGState class

        :param tfpg_state: the state to be added
        :type tfpg_state: TFPGState
        """
        self.tfpg_states.append(tfpg_state)
        return tfpg_state


    def get_tfpg_states(self):
        """Return the list of tfpg states contained in the trace"""
        return self.tfpg_states


    def __len__(self):
        return len(self.tfpg_states)


    def __getitem__(self, idx):
        return self.tfpg_states[idx]


    def __str__(self):
        out = []
        for (i, state) in enumerate(self.tfpg_states):
            out.append("#### Step {%d} ####" %(i))
            out.append(str(state))

        return "\n".join(out)




class TFPGState:
    """Represents a state of a tfpg trace"""

    nodes_state = None
    nodes_act_time = None
    edges_state = None
    edges_act_time = None
    global_time = None
    mode = None


    def __init__(self):
        self.nodes_state = dict()
        self.nodes_act_time = dict()
        self.edges_state = dict()
        self.edges_act_time = dict()
        global_time = 0.0
        mode = ""


    def set_node_state(self, node, value):
        """Updates the state of a given node

        The state of a given node can be True or
        False, meaning that this node is active or not.
        In case the node is active (i.e., the 
        activation time is greater than 0.0), the state
        of the node is set to True, and its activation
        time is properly updated.
        Notice that in this function validation-checks are
        performed, i.e., the activation time must be a float
        lower than the global time.

        :param node: the node to be updated
        :type node: TFPGNode

        :param value: the activation time of the node
        :type value: float

        """
        try: 
            val = float(value)

            if val < self.global_time:
                if val == 0.0:
                    self.nodes_state[node] = False
                elif val > 0.0:
                    self.nodes_state[node] = True
                else:
                    raise NodeNegativeActivationTime(node)
                self.nodes_act_time[node] = val
                return

            else:
                raise NodeBigActivationTime(node)

        except ValueError:
            print "Not a float"


    def set_edge_state(self, edge, value):
        """Updates the state of a given edge

        The state of a given edge can be True or
        False, meaning that this edge is active or not.
        In case the edge is active (i.e., the 
        activation time is greater than 0.0), the state
        of the edge is set to True, and its activation
        time is properly updated.
        Notice that in this function validation-checks are
        performed, i.e., the activation time must be a float
        lower than the global time.

        :param node: the edge to be updated
        :type node: TFPGEdge

        :param value: the activation time of the edge
        :type value: float

        """
        try:
            val = float(value)

            if val < self.global_time:
                
                if val == 0.0:
                    self.edges_state[edge] = False
                elif val > 0.0:
                    self.edges_state[edge] = True
                else:
                    raise EdgeNegativeActivationTime(node)
                self.edges_act_time[edge] = val
                return

            else:
                raise EdgeBigActivationTime(edge)

        except ValueError:
            print "Not a float"


    def set_global_time(self, value):
        """Sets the global time

        :param value: the global time value
        :type value: float
        """
        try:
            self.global_time = float(value)
            return

        except ValueError:
            print "Not a float"


    def set_mode(self, value):
        """Sets the state relative mode

        :param value: the mode of the current state
        :type value: string
        """
        self.mode = value
        return


    def get_nodes_state(self):
        """Return the dict of nodes states (True or False)"""
        return self.nodes_state


    def get_edges_state(self):
        """Returns the dict of edges states (True or False)"""
        return self.edges_state


    def get_nodes_act_time(self):
        """Return the node activation times"""
        return self.nodes_act_time


    def get_edges_act_time(self):
        """Returns the edges activation times"""
        return self.edges_act_time


    def get_global_time(self):
        """Return the global time value"""
        return self.global_time


    def get_mode(self):
        """Return the state's mode"""
        return self.mode


    def __str__(self):
        out = []
        out.append("ctime = {%s}" %(self.global_time))
        for (key, value) in self.nodes_state.items():
            out.append("{%s} = {%s}" %(key, value))
        for (key, value) in self.nodes_act_time.items():
            out.append("{%s} = {%s}" %(key, value))
        for (key, value) in self.edges_state.items():
            out.append("{%s} = {%s}" %(key, value))
        for (key, value) in self.edges_act_time.items():
            out.append("{%s} = {%s}" %(key, value))

        return "\n".join(out)

