# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================


# This file contains some utility for the analysis of fault tree
# represented as XML files


from lxml import etree


# ----------------------------------------------------------------------
# Public

def get_unique_cut_sets_from_files(fnames):
    """Given a set of xml files, returns a map {cardinality:
    list_of_cut_set_xml_elements} where duplicates in the xml
    elements have been removed"""
    res = {}
    for fname in fnames:
        et = etree.parse(fname)
        root = et.getroot()
        csm = get_cut_sets(root)
        
        for card, cs in csm.iteritems(): res[card] = res.get(card, []) + cs
        pass
        
    return get_unique_cut_sets(res)


# ----------------------------------------------------------------------
# Internals

def get_unique_cut_sets(cs_map):
    """Given a map as returned by get_cut_set returns a new map, in
    which duplicates have been removed"""
    res = {}
    for card, cs_list in cs_map.iteritems():
        new_cs_list = []
        unique_cs = set()
        for cs in cs_list:
            cs_str = cut_set_to_string(cs).strip()
            if cs_str not in unique_cs: 
                unique_cs.add(cs_str)
                new_cs_list.append(cs)
                pass            
            pass
        
        res[card] = new_cs_list
        pass

    return res        


def cut_set_to_string(root):
    """Traverses the tree, and dumps only the representative
    (i.e. functional to comparison) parts into a string"""
    if root.tag == "gate": 
        tp = root.get("type")
        assert tp in ("AND", "PAND", "NULL")
        return " ".join([tp] + list(map(cut_set_to_string, root)))
    
    assert root.tag == "event"
    return root.get("description").strip()


def get_cut_set_leaves(root):
    """Given a cat set, returns the set of all leaves. The empty set is returned for
    true (empty cs), None for false (indicating no-cut-set)"""

    if root.tag == "gate": 
        tp = root.get("type")
        assert tp in ("AND", "PAND", "NULL")
        css = filter(lambda c: c is not None, (get_cut_set_leaves(sel) for sel in root))
        return reduce(set.union, css) if len(css) > 0 else None
    
    assert root.tag == "event"
    name = root.get("name")
    if name.lower() == "false": return None # no CS
    if name.lower() == "true": return set() # empty set
    return set((name,))


def get_cut_sets(root):
    """Returns a map {cardinality: list_of_xml_elements} of the cut
    sets found in root fault tree"""

    assert root.tag == "fault-tree"
    res_list = []
    
    def get_cut_sets_recur(root, res):
        assert root.tag == "gate", "Unexpected tag:" + str(root)
        root_tp = root.get("type")
        
        if root_tp == "OR": 
            for el in root: get_cut_sets_recur(el, res)
            pass
        else:
            # this is a cut set
            cs_set = get_cut_set_leaves(root)
            if cs_set is not None: res.append((root, len(cs_set)))
            pass
        return

    for _or in root: get_cut_sets_recur(_or, res_list)

    # now organizes the list as a map
    res = {}
    for cs,card in res_list:
        cs_card = res.get(card, [])
        cs_card.append(cs)
        res[card] = cs_card
        pass

    return res


# testing code
if __name__ == "__main__":
    import sys
    res = get_unique_cut_sets_from_files(sys.argv[1:])
    print res
