# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================


# Module containing utilisties for supportgin debugging and
# production of debugging information

import os.path
import sys

import options
from utils import exceptions, log
from utils.misc import get_out_fname
from slim_compiler.src.dataStructure.visitors.printVisitor import toSLIMVisitor

def create_debug_and_compile_information():
    """Creates an instance of class DebugInformation and returns
    it. This function creates also an instance of a
    CompileInformation, and set it into the DebugInformation's
    instance field called 'compile_info'.
    """
    from slim_compiler.src.dataStructure.compile import CompileInformation
    dinfo = DebugInformation()

    # calculate full command used to invoke the main script
    dinfo.command_line = " ".join(sys.argv)
    dinfo.compile_info = CompileInformation()

    return dinfo   


# ----------------------------------------------------------------------
def emit_error_and_quit(err, dinfo):
    """Logs the output message and quits the execution"""
    dinfo.error = err

    if isinstance(err, exceptions.UserException): 
        log.user.error("ERROR IN INPUT")
        log.user.error(err)

    elif isinstance(err, exceptions.InternalException): 
        log.system.error("INTERNAL ERROR")
        log.system.error(err)

    else: 
        log.system.error("UNIDENTIFIED ERROR") # general unidentified error (a bug in the implementation)
        log.system.error(err)
        pass    
    
    # dump all input files into the debug object
    if dinfo.compile_info is not None:
        for filename in dinfo.compile_info.files:
            if dinfo.slim_files is None: dinfo.slim_files = {}
            dinfo.slim_files[filename] = open(filename).read()
            pass
        pass
    
    if dinfo.fault_injection_file_name is not None:
        dinfo.fault_injection_file = open(dinfo.fault_injection_file_name).read()
        pass
    
    if options.LOGGING_FNAME: logfile_name = options.LOGGING_FNAME
    else: logfile_name = get_out_fname(options.BASE_FILE_NAME + ".log")

    dbg_fname = get_out_fname(options.BASE_FILE_NAME + ".dbg")
    
    # shut down logging system and save log file
    import logging
    logging.shutdown()
        
    if os.path.isfile(logfile_name): dinfo.log = open(logfile_name).read()
    
    dinfo.write_to_file(dbg_fname)
    
    print "Errors occurred during the run. The file " + \
        dbg_fname + " has been generated. "
        
    if not isinstance(err, exceptions.UserException):
        print "Please attach this file in the bug report!"
        pass

    quit(-1)
    return


# ----------------------------------------------------------------------
# ----------------------------------------------------------------------

class DebugInformation():
    """This class can be used to collect and dump debugging
    information"""
    def __init__(self):
        self.command_line = None
        self.error = None
        self.slim_files = None
        self.compile_info = None
        self.extended_compile_info = None
        self.root_impl = None
        self.formulae_document = None
        self.formulae_strings = None
        self.label_ap_map = None
        self.error_event_probability_map = None
        self.smv_language = None
        self.failures_observations_alarms = None
        self.slim2smv_output = None
        self.smv2sigref_output = None
        self.model2imc_output = None
        self.mrmc_output = None
        self.imca_output = None
        self.slimsim_output = None
        self.slimsim_model = None
        self.log = None
        self.fault_injection_file_name = None
        self.fault_injection_file = None
        self.tfpg_filename = None
        return 

    def write_to_file(self, filename):
        _file = open(filename, "w")
        if self.command_line is not None:
            _file.write("Invocation command:-----------------------------------")
            _file.write("\n")
            _file.write(self.command_line)
            _file.write("\n")
        if self.error is not None:
            _file.write("Error and traceback:----------------------------------")
            _file.write("\n")
            _file.write(str(self.error))
            _file.write("\n")
            import traceback
            traceback.print_tb(sys.exc_traceback, None, _file)
        if self.log is not None:
            _file.write("Logging output:---------------------------------------")
            _file.write("\n")
            _file.write(self.log)
            _file.write("\n")
        if self.slim_files is not None:
            _file.write("\n\nProcessed SLIM _files:---------------------------------")
            _file.write("\n")
            for slim_file_name, slim_file in self.slim_files.iteritems():
                _file.write(slim_file_name + ": ###############################")
                _file.write("\n")
                _file.write(slim_file)
                _file.write("\n")
        if self.compile_info is not None:
            _file.write("SLIM representation of compile information:-----------")
            _file.write("\n")
            _file.write(toSLIMVisitor().visit(self.compile_info))
            _file.write("\n")
        if self.extended_compile_info is not None:
            _file.write("SLIM representation of extended compile information:--")
            _file.write("\n")
            _file.write(toSLIMVisitor().visit(self.extended_compile_info))
            _file.write("\n")
        if self.slim2smv_output is not None:
            _file.write("SLIM2SMV output:--------------------------------------")
            _file.write("\n")
            _file.write(self.slim2smv_output)
            _file.write("\n")
        if self.root_impl is not None:
            _file.write("Root implementation:----------------------------------")
            _file.write("\n")
            _file.write(self.root_impl)
            _file.write("\n")
        if self.formulae_document is not None:
            _file.write("Formula document:-------------------------------------")
            _file.write("\n")
            _file.write(str(self.formulae_document))
            _file.write("\n")
        if self.formulae_strings is not None:
            _file.write("Formulae strings:-------------------------------------")
            _file.write("\n")
            for formula in self.formulae_strings:
                _file.write(formula)
                _file.write("\n")
        if self.fault_injection_file is not None:
            _file.write("Fault injection file:---------------------------------")
            _file.write("\n")
            _file.write(self.fault_injection_file_name + ": ###################")
            _file.write("\n")
            _file.write(self.fault_injection_file)
            _file.write("\n")
        if self.label_ap_map is not None:
            _file.write("Label to atomic propositions map:---------------------")
            _file.write("\n")
            _file.write(self.label_ap_map)
            _file.write("\n")
        if self.error_event_probability_map is not None:
            _file.write("Error event probability map:--------------------------")
            _file.write("\n")
            _file.write(self.error_event_probability_map)
            _file.write("\n")
        if self.smv_language is not None:
            _file.write("SMV language:-----------------------------------------")
            _file.write("\n")
            _file.write(self.smv_language)
            _file.write("\n")
        if self.failures_observations_alarms is not None:
            _file.write("Failure observations alarms:--------------------------")
            _file.write("\n")
            _file.write(self.failures_observations_alarms)
            _file.write("\n")
        if self.smv2sigref_output is not None:
            (stdout, stderr) = self.smv2sigref_output
            _file.write("SMV2Sigref output:------------------------------------")
            _file.write("\n")
            _file.write("stdout:\n")
            if stdout is not None:
                _file.write(stdout)
            _file.write("\nstderr:\n")
            if stderr is not None:
                _file.write(stderr)
            _file.write("\n")
        if self.model2imc_output is not None:
            (stdout, stderr) = self.model2imc_output
            _file.write("Model2IMC output:-----------------------------------")
            _file.write("\n")
            _file.write("stdout:\n")
            if stdout is not None:
                _file.write(stdout)
            _file.write("\nstderr:\n")
            if stderr is not None:
                _file.write(stderr)
            _file.write("\n")
        if self.mrmc_output is not None:
            (stdout, stderr) = self.mrmc_output
            _file.write("MRMC output:------------------------------------------")
            _file.write("\n")
            _file.write("stdout:\n")
            if stdout is not None:
                _file.write(stdout)
            _file.write("\nstderr:\n")
            if stderr is not None:
                _file.write(stderr)
            _file.write("\n")
        if self.imca_output is not None:
            (stdout, stderr) = self.imca_output
            _file.write("IMCA output:------------------------------------------")
            _file.write("\n")
            _file.write("stdout:\n")
            if stdout is not None:
                _file.write(stdout)
            _file.write("\nstderr:\n")
            if stderr is not None:
                _file.write(stderr)
            _file.write("\n")
        if self.slimsim_output is not None:
            (stdout, stderr) = self.slimsim_output
            _file.write("slimsim output:------------------------------------------")
            _file.write("\n")
            _file.write("stdout:\n")
            if stdout is not None:
                _file.write(stdout)
            _file.write("\nstderr:\n")
            if stderr is not None:
                _file.write(stderr)
            _file.write("\n")
        if self.slimsim_model is not None:
            _file.write("slimsim model output:------------------------------------------")
            _file.write(self.slimsim_model)
        if self.tfpg_filename is not None:
            _file.write("TFPG output:------------------------------------------")
            _file.write("\n")
            _file.write(self.tfpg_filename)
            try:
                _file.write(open(self.tfpg_filename).read())
            except:
                _file.write("Unable to read file\n")
        _file.close()
        return
    pass # end of class DebugInformation


