# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================
from __future__ import with_statement

import os
import re
import subprocess
import imca_process
import exceptions
from utils.results import AnalysisResults
import tempfile
from utils.imca_process import ImcaProcess

def convert_mrmc_to_imca(tra_filename, lab_filename, labels_filename, 
                        init_filename, ma_filename):
    goalstates = extract_goalstates(lab_filename, labels_filename)
    write_mafile(init_filename, goalstates, tra_filename, ma_filename)
    return

def extract_goalstates(lab_filename, labels_filename):
    label = ''
    goalstates = []
    with open(labels_filename, 'r') as labels_file:
        label = labels_file.readline().split()[1]
    
    with open(lab_filename, 'r') as labelling_file:
        #skip header (there must be a better way.)
        for line in labelling_file:
            if(line == '#END\n'):
                break
        # TODO check for EOF...        
        #
        for line in labelling_file:
            if(line.find(label) != -1):
                goalstates.append(line.split()[0])
    
    #print goalstates
    return goalstates;

def write_mafile(init_filename, goalstates, tra_filename, ma_filename):
    with open(ma_filename, 'w') as ma_file:
        ma_file.write('#INITIALS\n')
        with open(init_filename, 'r') as init_file:
            init_file.readline();
            # The second row contains all initial states.            
            initstates = init_file.readline().strip().split();
            for s in initstates:            
                ma_file.write(s + '\n')
        ma_file.write('#GOALS\n')
        for s in goalstates:
            ma_file.write(s + '\n')
        ma_file.write('#TRANSITIONS\n')
        with open(tra_filename, 'r') as tra_file:
            #skip first two lines
            tra_file.next()
            tra_file.next()
            last_outgoing_state = -1
            for line in tra_file:
                linecontent = line.split()
                if(int(linecontent[0]) != last_outgoing_state):
                    ma_file.write(linecontent[0] + ' !\n* ' + linecontent[1] +
                                   ' ' + linecontent[2] + '\n')
                    last_outgoing_state = int(linecontent[0])
                else:
                    ma_file.write('* ' + linecontent[1] + 
                                  ' ' + linecontent[2] + '\n')
    return

def start_imca(ma_filename, computation_type, lower_time_bound=None, upper_time_bound=None, error_bound=0.1, stepwise=False, granularity=100, args=[]):
    # constructs a new instance of imca logging file
    fdlog, logfname = tempfile.mkstemp(suffix='.imcaout', 
                                    prefix=os.path.splitext(ma_filename)[0], 
                                    dir=tempfile.gettempdir(), text=True)

    process = ImcaProcess(ma_filename, computation_type,
                            lower_time_bound=lower_time_bound, upper_time_bound=upper_time_bound,
                            error_bound=error_bound, stepwise=stepwise, granularity=granularity,
                            args=[],
                            stdout=subprocess.PIPE,
                            stderr=subprocess.STDOUT,
                            dumpers=[os.fdopen(fdlog, "w")])
        
    return process
    
def run(ma_filename, computation_type, lower_time_bound=None, upper_time_bound=None, error_bound=0.1, stepwise=False, granularity=10):
    process = start_imca(ma_filename, computation_type, lower_time_bound, upper_time_bound, error_bound, stepwise, granularity)

    process.wait()
    if process.returncode != 0:
        raise exceptions.InternalException("IMCA terminated with error code %d." % process.returncode)
    
    return (process.returncode, (process.get_stdout_text(), process.get_stderr_text()))

def extract_results_from_output(property, imca_output):
    concated = ''.join(imca_output)
    patternId = property.get_pattern_id()

    if patternId == "expectedTime":
        return _extract_et_results(property, imca_output)
    elif patternId == "longRunAverage":
        return _extract_lra_results(property, imca_output)
    else:
        return _extract_probability_results(property, imca_output)

def _extract_min_max_results(concated, re_min, re_max):
    minvals = re_min.findall(concated)
    maxvals = re_max.findall(concated)

    assert(len(minvals) == 1)
    assert(len(maxvals) == 1)

    return float(minvals[0]), float(maxvals[0])

def _extract_et_results(property, concated):
    # Extract min and max expected time
    re_min = re.compile( r'(?<=Minimal expected time\: )[\d\w\.\-\+]+' )
    re_max = re.compile( r'(?<=Maximal expected time\: )[\d\w\.\-\+]+' )
    
    minET, maxET = _extract_min_max_results(concated, re_min, re_max)

    return AnalysisResults(property, 'imca',
                          expected_time = (minET, maxET))

def _extract_lra_results(property, concated):
    # Extract min and max expected time
    re_min = re.compile( r'(?<=Minimal LRA\: )[\d\w\.\-\+]+' )
    re_max = re.compile( r'(?<=Maximal LRA\: )[\d\w\.\-\+]+' )
    
    minLRA, maxLRA = _extract_min_max_results(concated, re_min, re_max)

    return AnalysisResults(property, 'imca', lra = (minLRA, maxLRA))

def get_iteration_count(imca_output):
    re_iterations = re.compile( r'(?<=iterations: )[\d]+')
    iterations = re_iterations.findall(imca_output)
    # Iteration count is expected to be the same for both min and max
    if len(iterations)>0:
        for i in iterations:
            assert(iterations[0]==i)
        return int(iterations[0])
    else:
        return None    

class ImcaIterationCountZeroException(exceptions.UserException):
    def __init__(self):
        error_msg = "Analysis took IMCA 0 iterations. You might need to use a smaller error bound."
        super(ImcaIterationCountZeroException, self).__init__(error_msg)

def _extract_probability_results(property, concated):
    # Extract min and max probabilities
    rmin = re.compile( r'(?<=Minimal time-bounded reachability probability\: )[\d\w\.\+\-]+' )
    rmax = re.compile( r'(?<=Maximal time-bounded reachability probability\: )[\d\w\.\+\-]+' )
    
    minProbsStr = rmin.findall(concated)
    maxProbsStr = rmax.findall(concated)

    minProbs = map(lambda x: float(x), minProbsStr)
    maxProbs = map(lambda x: float(x), maxProbsStr)

    assert(len(minProbs) >= 1)
    assert(len(maxProbs) >= 1)
    assert(len(maxProbs) == len(minProbs))
    assert(float(minProbs[0]) <= (maxProbs[0]))

    stepwise = len(minProbs) > 1
    if stepwise:
        # Must also find the x values if multiple steps
        rxvals = re.compile(r'^tb=(\d+(\.\d+)?)', re.MULTILINE)
        # This will find each time bound twice (and convert the two parts of the number to a float),
        # so the second half must be thrown away
        xvals = map(lambda x: float(x[0]),rxvals.findall(concated))
        xvals = xvals[:len(minProbs)]
    
    # Only return one prob if min and max prob are the same
    probs = [p1 if p1 == p2 else (p1,p2) for (p1,p2) in zip(minProbs,maxProbs)]

    # Note that we only return the tail of the lists for consistency with mrmc output
    # (The head is 0 in both lists and not in the list returned for mrmc)
    return (AnalysisResults(property, 'imca',
                           probability_steps=(xvals[1:], probs[1:]))
           if stepwise
           else AnalysisResults(property, 'imca', probability=probs[0]))
