# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================
import os
import options
import utils.process
import exceptions



class ImcaProcess (utils.process.InteractiveProcess):
    """Process handler for running dft2mrmc.
    """

    def __init__(self, ma_filename, computation_type, 
                  minmax = 'both', lower_time_bound=None,
                  upper_time_bound=None,error_bound=None,
                  stepwise=False,granularity=100,
                  args=[],
                  stdout=None,stdin=None, stderr=None, dumpers=[]):

        # The specified input files must exist.
        if not os.path.isfile(ma_filename):
            raise IOError("File not found: " + ma_filename)
       
        path = utils.process.get_exec_name(options.IMCA)
     
        args = [ma_filename] + args
        
        if (minmax == 'min' or minmax == 'max'):
            args.append('-' + minmax)
        elif minmax == 'both':
            args.append('-min')
            args.append('-max')
        else:
            raise exceptions.InternalException("Imca needs a min/max/both argument")

        if stepwise and lower_time_bound and float(lower_time_bound) != 0:
            utils.log.system.info("WARNING: Illegal lower time bound for IMCA" +
                                  " (must be 0, is " + lower_time_bound + ")")
            utils.log.system.info("WARNING: Will overwrite lower time bound with 0")
            lower_time_bound = 0

        if(lower_time_bound != None or upper_time_bound != None):
            if(lower_time_bound != None):
                args.append("--from")
                args.append(str(lower_time_bound))
            if(upper_time_bound != None):
                args.append("--to")
                args.append(str(upper_time_bound))     

        args.append(computation_type)

        if stepwise:
            args.append("-i")

            # IMCA can only compute multiple steps if lower bound = 0
            # The following code will raise an exception in this case
            # The currently active code above will instead overwrite the lb with 0
            # if lower_time_bound and int(lower_time_bound) != 0:
            #     raise exceptions.InternalException("Illegal lower time bound for IMCA" +
            #                                       " (must be 0, is " + lower_time_bound + ")")
            
            if not upper_time_bound:
                raise exceptions.InternalException("Step-wise computation only"
                                                   + " possible if time bound is given")
            
            # Compute interval step from number of steps (granularity)
            step = float(upper_time_bound) / granularity
            args.append(str(step))

        if(error_bound != None):
            args.append("-e")
            args.append(str(error_bound))

        #print ' '.join(args)
        super(ImcaProcess, self).__init__(path, args, stdout=stdout, 
                                              stdin=stdin, stderr=stderr, 
                                              dumpers=dumpers)

    
pass # end of class
