# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================


# This module contains class ProcMonitor, which is used by
# interactive processes to monitor and control memory and cputime
# usage.


from __future__ import with_statement # needed in 2.5
import subprocess
import threading
import platform
from utils import log
import gtk
import gobject


class ProcMonitor (threading.Thread):
    # March 29, 2010 by Viet Yen Nguyen in response to ticket #368
    # The following two attributes are modified to have faster polling 
    # times
    busy_sleep_time = 1.0
    idle_sleep_time = 4.0

    def __init__(self, process, memout=-1, timeout=-1):
        """process is an InteractiveProcess instance. Memory out is
        expressed in KBytes, timeout in seconds, -1 means no
        limit"""

        threading.Thread.__init__(self)
        self.proc = process
        self.memout = int(memout) if int(memout) >= 0 else -1
        self.timeout = int(timeout) if int(timeout) >= 0 else -1
        self.acc_time = 0
        self.tot_time = 0 
        self.tot_mem = 0
        self.max_mem = 0
        self.status_lock = threading.Lock()                        
        self._wait = threading.Event()        
        self.__set_status("STOP")
        return
    

    def set_timeout(self, timeout): 
        """Sets the timeout for the monitor. -1 means not limit"""
        self.timeout = int(timeout) if int(timeout) >= 0 else -1  
        log.system.debug("Monitor: set timeout %d for process '%s'" % (self.timeout, self.proc.path))
        return


    def set_memout(self, memout): 
        """Sets the memout in KB for the monitor. -1 means not limit"""
        self.memout = int(memout) if int(memout) >= 0 else -1
        log.system.debug("Monitor: set memout %d for process '%s'" % (self.memout, self.proc.path))
        return


    def stop(self):
        """Stops the monitor"""
        if self.isAlive():
            self.__set_status("STOP")  
            self._wait.set() # exits any waiting run
            pass
        return


    def get_info(self):
        """Returns process (status, total cputime, partial cputime, memory usage, memory peak).
        Times in seconds, memory in KBytes."""
        return (self.__get_status(), 
                self.tot_time, self.tot_time-self.acc_time, 
                self.tot_mem, 
                self.max_mem)


    def __set_status(self, status):
        with self.status_lock: self.status = status
        return


    def __get_status(self):
        with self.status_lock: return self.status

    
    def run(self):
        log.system.debug("Monitor: started monitoring '%s'" % self.proc.path)
        log.system.debug("Monitor: TIMEOUT is %d secs., MEMOUT is %d KB" % (self.timeout, self.memout))
        idle_times = 0
        last_time = -1        
        self.__set_status("BUSY")
        while (self.__get_status() in ("BUSY", "IDLE")):

            r = get_pid_time_mem(self.proc.pid)
            if r is None: 
                # March 29, 2010 by Viet Yen Nguyen in response to ticket #368
                # The following two statements are commented, preventing
                # the statistics to be reset when a proces dies. 

                # self.tot_mem = 0
                # self.max_mem = 0
                self.__set_status("DEAD")
                log.system.debug("Monitor: '%s' is dead, exiting" % self.proc.path)
                break

            self.tot_time, self.tot_mem = r
            self.max_mem = max(self.max_mem, self.tot_mem)
            # check activity: is it idle?
            if self.tot_time == last_time:
                sleep_time = ProcMonitor.idle_sleep_time
                if self.__get_status() != "IDLE": idle_times += 1 
                pass
            else: 
                sleep_time = ProcMonitor.busy_sleep_time
                self.__set_status("BUSY")
                idle_times = 0
                pass                
            
            if idle_times > 2: 
                self.__set_status("IDLE")
                self.acc_time = self.tot_time
            else:
                if self.timeout >= 0 and self.tot_time - self.acc_time >= self.timeout:
                    self.__set_status("TIMEOUT")
                    break
                
                elif self.memout >= 0 and self.tot_mem >= self.memout:
                    self.__set_status("MEMOUT")
                    break
                pass

            last_time = self.tot_time
            self._wait.wait(sleep_time)
            pass
        
        # Here the monitor is exiting the loop
        status, tt, pt, mem, peak = self.get_info()
        if status in ("TIMEOUT", "MEMOUT"):
            # kills the process
            msg = "Monitor: killing '%s' due to '%s'" % (self.proc.path, status)
            log.system.debug(msg)
            self.proc.kill()
            # Check if the kill did something, if so display message
            if self.proc.returncode is None or self.proc.returncode != 0:
                # Display message to the user
                gobject.idle_add(display_timeout, msg)
            pass

        # Monitor's goodbye message:
        log.system.debug("Monitor: Exiting the monitoring loop of '%s'" % self.proc.path)
        log.system.debug("Monitor: exit status is: %s" % status)
        log.system.debug("Monitor: total time was: %d secs" % tt)
        log.system.debug("Monitor: partial time was: %d secs" % pt)
        log.system.debug("Monitor: total memory usage was: %d KB" % mem)    
        log.system.debug("Monitor: peak memory usage was: %d KB" % peak)
        return

    pass # end of class

def display_timeout(message):
    msgbox = gtk.MessageDialog(type=gtk.MESSAGE_ERROR,
                       buttons=gtk.BUTTONS_OK,
                       message_format=message)
    msgbox.run()
    msgbox.destroy()

def get_pid_time_mem(pid):
    """Returns (cputime, memory) or None. Time in seconds, memory in Kbytes"""
    if platform.system() == "Windows": return (0,0) # for windows it is not available
    cmd = "ps -o vsize=,cputime= %s" % pid
    r = subprocess.Popen(cmd.split(), stdout=subprocess.PIPE).communicate()[0].strip()
    res = None
    if r:
        mem, cputime = r.split()
        if ( -1 != cputime.find('-')): (ds, ts) = cputime.split('-')
        else: (ds, ts) = ("0", cputime)

        (dd, hh, mm, ss) = map( lambda x: int(x), [ds,] + ts.split(":") )
        res = (dd*86400 + hh*3600 + mm*60+ ss, int(mem))
        pass
    return res
    

