# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================
from utils import exceptions, imca, debug_info, model2imc_process, log
import property_management.src.properties as properties_module
import property_management.src.api as properties_api
import utils.fault_tree as fault_tree_module
from utils.results import AnalysisResults
import tempfile
import os
from invokers import slim, dump, nusmv, mrmc
from slim_compiler.src.dataStructure.compile import CompileInformation
from utils.misc import get_out_fname
import options
import invokers
import hashlib
from property_management.src.formulae import Formula
from utils import misc
from property_management.src.properties import PropertiesDocument
from slim_compiler.src.dataStructure.visitors.time import TimeScaleTransformVisitor

tool2suffix_map = {"mrmc":["tra", "lab", "init"],
                   "imca":["ma"]}

def hashModelFile(model_type, filepath):
    if not os.path.exists(filepath):
        raise exceptions.InternalException("Model file '%s' does not exist"%os.path.abspath(filepath))
    
    with open(filepath) as mfile:
        if model_type == "SMV":
            lines = filter(lambda x:not x.startswith("--"), mfile.readlines())
        else:
            lines = mfile.readlines()
    
    return hashlib.md5("".join(lines)).hexdigest()

class PropertyCheckingRecord(object):
    default_filename = "property_checking_record.txt"
    default_conversion_filename="property_checking_conversion_result.txt"
    supported_types = {"DFT","SMV"}
    
    # Do not use this! Use the create-methods instead
    def __init__(self, model_type=None, model_entries=[], record_file_path=default_filename, basefolder=None, intermediate_strings=None, property_map=None, label_map=None, records=None):
        if intermediate_strings == None: intermediate_strings = set()
        if property_map == None: property_map = dict()
        if label_map == None: label_map = dict()
        if records == None: records = dict()

        self.model_type = model_type             # type of the model: DFT or SMV
        self.model_entries = model_entries       # list of tuples (model_path, md5 hash of content)
        self.record_file_path = record_file_path # file to write record to
        self.basefolder = basefolder             # files referenced in the records are created in this folder
        self.intermediate_strings = intermediate_strings # data about intermediate representations; only used in C part
        self.property_map = property_map         # property string --> index/key (type int) in records table
        self.label_map = label_map               # normal label --> mrmc label
        self.records = records
    
    @classmethod
    def create(cls, model_type, model_filepaths, record_file_path=None, basefolder=None):
        """ Creates PropertyCheckingRecord from an existing file if a valid file exists in the basefolder.
            Otherwise an empty one will be created.
        """
        if not record_file_path:
            record_file_path = misc.get_out_fname(PropertyCheckingRecord.default_filename)
        if not basefolder:
            basefolder = os.path.dirname(record_file_path);

        # TODO: Record file somehow retains state information, recreate it
        if True: #not os.path.exists(record_file_path):
            return PropertyCheckingRecord.createEmpty(model_type, model_filepaths, record_file_path, basefolder)
        else:
            try:
                old_property_checking_record = PropertyCheckingRecord.createFromRecordsFile(record_file_path)
                if old_property_checking_record.isValid(model_type):
                    return old_property_checking_record
                else:
                    log.system.info("Records file '%s' is not valid. Type or hash mismatch. A new one will be created instead."%record_file_path)
                    return PropertyCheckingRecord.createEmpty(model_type, model_filepaths, record_file_path, basefolder)
            except (exceptions.InternalException, exceptions.UserException) as err:
                # If input is bad simply create a new record
                log.system.warning(err)
                return PropertyCheckingRecord.createEmpty(model_type, model_filepaths, record_file_path, basefolder)
    
    @classmethod
    def createEmpty(cls, model_type, model_filepaths, record_file_path=None, basefolder=None):
        if not record_file_path:
            record_file_path = misc.get_out_fname(PropertyCheckingRecord.default_filename)
        if not basefolder:
            basefolder = os.path.dirname(record_file_path);

        if not model_type in PropertyCheckingRecord.supported_types:
            raise exceptions.InternalException("Model type '%s' is unknown. Expected one of the following: %s"%(model_type,PropertyCheckingRecord.supported_types))

        # Create (path, md5hash) tuples if models exist
        model_entries = []
        for model_filepath in model_filepaths:
            model_entries.append((model_filepath, hashModelFile(model_type, model_filepath)))

        res = PropertyCheckingRecord(model_type, model_entries=model_entries, record_file_path=record_file_path, basefolder=basefolder)

        return res

    @classmethod
    def createFromRecordsFile(cls, record_file_path):
        if not os.path.exists(record_file_path):
            raise exceptions.InternalException("Records file '%s' does not exist"%os.path.abspath(record_file_path))

        with open(record_file_path) as recordsFile:
            lines = [line.replace("\n","").rstrip() for line in recordsFile.readlines()]
            lines = [line for line in lines if line] # Non-blank lines
            try:
                type_index = lines.index("#TYPE")
                model_index = lines.index("#MODEL")
                basefolder_index = lines.index("#BASEFOLDER")
                intermediate_index = lines.index("#INTERMEDIATE")
                label_index = lines.index("#LABEL_MAP")
                property_index = lines.index("#PROPERTY_MAP")
                record_index = lines.index("#RECORDS")
            except ValueError:
                raise exceptions.InternalException("Malformed records file. One of the hashtag-lines is not present")
            
            # Assert the correct order of sections 
            index_list = [type_index,model_index,basefolder_index,intermediate_index,label_index,property_index,record_index]
            if not index_list == sorted(index_list):
                raise exceptions.InternalException("Sections in records file '%s' are in wrong order"%os.path.abspath(record_file_path))  
            
            # Parse model type
            type = lines[type_index+1].strip()
            
            # Parse model entries
            model_entries = []
            for line in lines[model_index+1:basefolder_index]:
                separator = line.find(":")
                filepath = line[:separator]
                md5hash = line[separator+1:]
                model_entries.append((filepath, md5hash))
            
            # Parse basefolder
            basefolder = lines[basefolder_index+1]
            
            # Parse intermediate data
            intermediate_strings = set()
            for line in lines[intermediate_index+1:label_index]:
                if line not in intermediate_strings:
                    intermediate_strings.add(line)
            
            # Parse label map
            label_map = dict()
            for line in lines[label_index+1:property_index]:
                separator = line.rfind(":")
                label = line[:separator]
                new_label = line[(separator+1):]
                label_map[label] = new_label
            
            # Parse property map
            property_map = dict()
            for line in lines[property_index+1:record_index]:
                separator = line.rfind(":")
                property = line[:separator]
                index = int(line[separator+1:])
                property_map[property] = index
                
            # Parse records
            records = dict()
            for line in lines[record_index+1:]:
                (index, status, tool, basename) = _parseRecordsLine(line)
                if records.has_key(index):
                    records[index].append((status, tool, basename))
                else:
                    records[index] = [(status, tool, basename)]

        # Create instance
        res = PropertyCheckingRecord(type, model_entries, record_file_path, basefolder, intermediate_strings, property_map, label_map, records) 
        return res

    def mergeWithConversionResults(self,conversion_results_filename=default_conversion_filename):
        conversion_results_filepath = misc.get_out_fname(conversion_results_filename)
        if not os.path.exists(conversion_results_filepath):
            raise exceptions.InternalException("Conversion results file '%s'not found."%conversion_results_filepath)
        
        # Read conversion results
        with open(conversion_results_filepath) as conversion_results_file:
            lines = [line.replace("\n","").rstrip() for line in conversion_results_file.readlines()]
            try:
                intermediate_index = lines.index("#INTERMEDIATE")
                record_index = lines.index("#RECORDS")
            except ValueError:
                raise exceptions.InternalException("Malformed records file. One of the hashtag-lines is not present")

            # Add new intermediate data
            for line in lines[intermediate_index+1:record_index]:
                if line not in self.intermediate_strings:
                    self.intermediate_strings.add(line)
            
            # Add new records
            for line in lines[record_index+1:]:
                (index, status, tool, basename) = _parseRecordsLine(line)
                if self.records.has_key(index):
                    if (status, tool, basename) not in self.records[index]: 
                        self.records[index].append((status, tool, basename))
                else:
                    self.records[index] = [(status, tool, basename)]

    def isValid(self, expectedType):
        if not self.model_type == expectedType:
            return False
        
        for (filepath, md5hash) in self.model_entries:
            current_hash = hashModelFile(self.model_type, filepath)
            if md5hash != current_hash:
                return False
        return True

    def getLabelMap(self):
        return self.label_map

    def addProperty(self, property):
        formula = properties_api.create_formula_for_property(property)
        properties_api.convert_time_bounds_for_performability([formula])
        formula_intermediate_string = properties_api.get_formula_as_intermediate_string(formula)
        
        # Add formula to mapping
        if self.property_map.has_key(formula_intermediate_string):
            # Skip formulas have already been processed
            return

        # Map formula to to highest unused index
        highest_unused_index = 0 if len(self.property_map)==0 else int(max(self.property_map.values())) + 1
        self.property_map[formula_intermediate_string] = highest_unused_index
        
        # Extend label map if new atomic propositions are encountered
        atomic_proposition_strings = properties_api.get_atomic_proposition_strings(formula)
        atomic_proposition_strings = set(filter(lambda x: x != "tt", atomic_proposition_strings))
        for ap_string in atomic_proposition_strings:
            # If map is empty the first mapping must be ap_string --> label1
            if len(self.label_map) == 0:
                self.label_map[ap_string] = "label1"
            # Add new mapping if not existent yet
            elif not self.label_map.has_key(ap_string):
                # Find 'highest' unused label
                try:
                    highest_used_label_num = max(sorted([int(label.strip("label")) for label in self.label_map.values()]))
                except ValueError:
                    raise exceptions.UserException("Malformed labels. Should be of the form label<integer>.")
                new_label = "label"+str(highest_used_label_num+1)
                self.label_map[ap_string] = new_label
    
    def writeToFile(self):
        #whereToCreatePath = misc.get_out_fname(self.record_file_path)
        with open(self.record_file_path,"w") as recordsFile:
            recordsFile.write(str(self))
        log.system.info("Records file '%s' was written "%self.record_file_path)
            
    def __str__(self):
        # Model type
        type_str = "\n".join(["#TYPE", self.model_type])
        # Model files and timestamps
        models_str = "\n".join(["#MODEL"] + [str(filepath) + ":" + str(timestamp) for (filepath, timestamp) in self.model_entries])
        # Basefolder
        output_str="#BASEFOLDER\n"+os.path.abspath(self.basefolder)
        # Intermediate data as string
        intermediate_str = "\n".join(["#INTERMEDIATE"] +  list(self.intermediate_strings))
        # Label map
        label_str = "\n".join(["#LABEL_MAP"] + [str(key) + ":" + str(value) for (key, value) in self.label_map.iteritems()])
        # Property map
        property_str = "\n".join(["#PROPERTY_MAP"] + [str(key) + ":" + str(value) for (key, value) in self.property_map.iteritems()])
        # The actual records
        records = []
        for (key, value) in self.records.iteritems():
            for (status, tool, basename) in value:
                records.append(str(key)+","+str(status)+","+str(tool)+","+basename)
        records_str = "\n".join(["#RECORDS"] + records)
        return "\n".join([type_str, models_str, output_str, intermediate_str, label_str, property_str, records_str])

def _parseRecordsLine(line):
    parts = line.split(",")
    if not len(parts) == 4:
        raise exceptions.InternalException("Record '%s' doesn't consist of 4 values" % line)
    index = int(parts[0])
    status = int(parts[1])
    tool = parts[2].strip()
    basename = parts[3].strip()
    return (index, status, tool, basename)

class Checker:
    def __init__(self, tools=["mrmc", "imca"]):
        # Create debug_info instance for easier debugging
        self.debug_info = debug_info.create_debug_and_compile_information()

        # Set tools for checking
        self.tools = tools

        # Set default handlers for tools (for link with GUI)
        self.model2imc_handler = None
        self.mrmc_handler = None
        self.imca_handler = None

    def analyze(self):
        raise NotImplementedError

    def set_model2imc_handler(self, handler):
        self.model2imc_handler = handler

    def set_mrmc_handler(self, handler):
        self.mrmc_handler = handler

    def set_imca_handler(self, handler):
        self.imca_handler = handler

class PropertyChecker(Checker):
    def __init__(self, property_checking_record_path, tools=["mrmc", "imca"]):
        """property_checking_record_path is a required parameter and must contain an existing records file."""
        Checker.__init__(self, tools)
        self.property_checking_record = PropertyCheckingRecord.createFromRecordsFile(property_checking_record_path)

    def invoke_imca(self, formula, patternId, basename, error_bound, stepwise=False, granularity=100):

        # Determine IMCA's computation mode based on the patternId
        def computation_type(patternId):
            if patternId == "expectedTime":
                return "-et"
            elif patternId == "longRunAverage":
                return "-lra"
            else:
                return "-tb"

        suffixes = tool2suffix_map["imca"]
        ma_filename = os.path.join(self.property_checking_record.basefolder, basename + "." + suffixes[0])

        # Set time bounds if applicable
        if not (patternId in ["expectedTime","longRunAverage"]):
            lower_time_bound, upper_time_bound = properties_api.get_csl_bounds(formula)
        else:
            lower_time_bound, upper_time_bound = None, None

        # If run from GUI
        if self.imca_handler:
            self.imca_handler.busy += 1
            self.imca_handler.start_imca(ma_filename, computation_type(patternId), lower_time_bound, upper_time_bound, error_bound, stepwise, granularity)
            self.imca_handler.busy -= 1

            if not self.imca_handler.tool:
                # Process was most likely killed
                return (99, "")

            # Wait for IMCA to finish.
            exit_code = self.imca_handler.tool.returncode
            imca_out = (self.imca_handler.tool.get_stdout_text(), self.imca_handler.tool.get_stderr_text())
            self.debug_info.imca_output = imca_out
        else:
            exit_code, imca_out = imca.run(ma_filename, computation_type(patternId), lower_time_bound, upper_time_bound, error_bound, stepwise, granularity)
            self.debug_info.imca_output = imca_out
        return (exit_code, imca_out[0])

    def invoke_mrmc(self, formula, basename, stepwise=False, granularity=100):
        suffixes = tool2suffix_map["mrmc"]
        tra_filename  = os.path.join(self.property_checking_record.basefolder, basename + "." + suffixes[0])
        lab_filename  = os.path.join(self.property_checking_record.basefolder, basename + "." + suffixes[1])

        properties_api.replace_atomic_propositions(formula, self.property_checking_record.label_map)
        log.system.debug("(Replaced) MRMC formula string: %s"%properties_api.get_formula_as_mrmc_string(formula))
        
        # Run MRMC to compute the probability of the specified formula.
        formula_string = properties_api.get_formula_as_mrmc_string(formula)
        
        # If run from GUI
        if self.mrmc_handler:
            self.imca_handler.busy += 1
            self.mrmc_handler.start_mrmc(tra_filename, lab_filename, formula_string, stepwise, granularity)
            self.imca_handler.busy -= 1

            if not self.mrmc_handler.tool:
                # Process was most likely killed
                return (99, "")

            # Wait for MRMC to finish.
            exit_code = self.mrmc_handler.tool.returncode
            output = (self.mrmc_handler.tool.get_stdout_text(), self.mrmc_handler.tool.get_stderr_text())
            self.debug_info.mrmc_output = output
        else:
            (exit_code, output) = mrmc.run(tra_filename, lab_filename, formula_string, stepwise, granularity)
            self.debug_info.mrmc_output = output
        return (exit_code, output[0])


    def analyze(self, property, error_bound=0.1, force_tool=None, stepwise=False, granularity=100):
        """
        This will read the property_checking_record.txt-file and invoke the respective tools.
        If a specific tool is passed (via force_tool) this one will explicitly be used for checking
        and others (even if a suitable IMC has already been generated) will be ignored.
        
        Note that the error bound parameter is only used by IMCA.
        """
        formula = properties_api.create_formula_for_property(property)
        properties_api.convert_time_bounds_for_performability([formula])
        formula_intermediate_string = properties_api.get_formula_as_intermediate_string(formula)
        
        # If the formula has been processed before, the old IMC can be reused
        if not self.property_checking_record.property_map.has_key(formula_intermediate_string):
            # TODO: RE-GENERATE MODELS
            raise exceptions.InternalException("No files to analyze generated yet")
        property_index = self.property_checking_record.property_map[formula_intermediate_string]
        
        # Get corresponding formula
        if not self.debug_info.formulae_strings:
            self.debug_info.formulae_strings = []
        self.debug_info.formulae_strings.append(formula_intermediate_string)
        
        imc_conv_result = self.property_checking_record.records[property_index]
        tool_output = None
        for (status, tool, basename) in imc_conv_result:
            # If a specific tool was provided make sure only this tool is used
            if force_tool and tool != force_tool:
                log.system.info("Skip checking %s on %s with %s since a specific tool (%s) has been selected" % (formula_intermediate_string, basename, tool, force_tool))
                continue
            
            # Complain if only one tool has been selected and it happened to be the unsuitable one
            if len(self.tools) == 1 and int(status) != 0:
                raise ToolNotSuitableException(property.description, self.tools[0]) 
            log.system.info("Checking %s on %s with %s.." % (formula_intermediate_string, basename, tool))
            
            if int(status) != 0:
                log.system.info("Skipping probability computation of %s with %s. Conversion error code: %s" % (formula_intermediate_string, tool, status))
                continue
            
            if (tool == 'imca'):
                (exit_code, imca_out) = self.invoke_imca(formula, property.get_pattern_id(), basename, error_bound, stepwise, granularity)

                # Check whether IMCA has terminated successfully.
                if exit_code != 0:
                    if exit_code == None: exit_code = 99
                    raise exceptions.ProcessInternalException("IMCA", "Terminated with error code %d." % exit_code)
                
                # Check whether result is reasonable
                iter_count = imca.get_iteration_count(imca_out)
                if iter_count is not None and iter_count==0:
                    error_msg = "Analysis took IMCA 0 iterations. You might need to use a smaller error bound, or the model is not probabilistic"
                    log.user.warning(error_msg)
                    tool_output = ('imca', imca_out, basename, error_msg)
                    #raise imca.ImcaIterationCountZeroException() 
                else:
                    tool_output = ('imca', imca_out, basename, None)
                
                # It suffices to get a result from one tool -> break on success
                break
            elif (tool == 'mrmc'):
                (exit_code, mrmc_output) = self.invoke_mrmc(formula, basename, stepwise, granularity)

                # Check whether MRMC has terminated successfully.
                if exit_code != 0:
                    if exit_code == None: exit_code = 99
                    raise exceptions.ProcessInternalException("MRMC", "Terminated with error code %d." % exit_code)

                # Check for overflows
                if "ERROR: Fox-Glynn:" in ''.join(mrmc_output):
                    error_msg = "Errors in the Fox-Glynn algorithm, results can be unreliable"
                    log.system.error(error_msg)
                    tool_output = ('mrmc', mrmc_output, basename, error_msg)
                else:
                    tool_output = ('mrmc', mrmc_output, basename, None)
                # It suffices to get a result from one tool -> break on success
                break
            else:
                raise exceptions.InternalException("Unsupported tool %d." % self.chosen_tool)
        return tool_output
    
    # Extracts results from tool output
    # Result is a list of probabilities (ordered by order of given tools) 
    # TODO: Make sure all references set force_tool appropriately 
    def get_results(self, property, error_bound=0.1, force_tool=None, stepwise=False, granularity=100):
        """
        Invokes the analysis of the property with either a specific tool (passed via force_tool) or
        all available tools (if force_tool is None).
        
        Returns an AnalysisResult instance
        """
        (tool, output, basename, error_msg) = self.analyze(property, error_bound, force_tool, stepwise, granularity)

        result = None
        if tool == 'imca':
            result = imca.extract_results_from_output(property, output)
        elif tool == 'mrmc':
            init_filename = os.path.join(self.property_checking_record.basefolder, basename + "." + tool2suffix_map['mrmc'][2])
            if stepwise:
                probs = invokers.mrmc._parse_mrmc_cdf_data(output, init_filename)
                result = AnalysisResults(property, 'mrmc', probability_steps=probs)
            else:
                prob = invokers.mrmc.extract_probabilities(output, init_filename)
                result = AnalysisResults(property, 'mrmc', probability=prob)
        else:
            raise exceptions.InternalException("Unsupported tool")

        # Invert result probabilities if necessary
        invertedProps = ["probabilisticInvariance", "probabilisticPrecedence"]
        if result.property.get_pattern_id() in invertedProps:
            log.system.debug("Property pattern id: %s => must invert"% property.get_pattern_id())
            log.system.debug("Before inversion: %s"%result)
            result.invertProbabilities()
        return (result, error_msg)
    
    def _is_empty_tra(self, trafilename): 
        trans = int(open(trafilename).readlines()[1].split(' ')[1])
        
        return trans == 0

class ToolNotSuitableException(exceptions.UserException):
    def __init__(self,property_string, tool):
        error_msg = "Probability computation of \"%s\" is not possible with \"%s\""%(property_string.strip(), tool)
        super(ToolNotSuitableException, self).__init__(error_msg)        

# TODO: Move into DFTPropertyChecker, DFTEvaluater & SLIMPropertyChecker into an own module
# TODO: functionality separated into blocks to be called from runscripts/gui, in order for 
# runscript to be able to reuse basic functionality. Need to split such functionality into
# other utilities at a later point (out of runscripts/GUI). Ex generating smv files, xml files
class SLIMPropertyChecker(Checker):
    def __init__(self, slim_files, fault_injection_filename, properties_filename, tools=["mrmc", "imca"]):
        Checker.__init__(self, tools)
        
        self.properties_filename = properties_filename
        self.slim_files = slim_files
        self.fault_injection_filename = fault_injection_filename

        self.cinfo = None
        self.ext_cinfo = None
        self.root_impl = None
        
        self.property_checking_record = None
        
        pass
        
    def load(self):
        # SLIM to SMV
        self.gen_smv()
        # Parse formulae
        self.gen_formulae()
        # SMV to BDD
        self.gen_sigref()
        # BDD in record
        self.gen_property_record()

    def gen_smv(self):
        # Compile slim files
        self.cinfo = CompileInformation()
        if not slim.parse(self.cinfo, self.slim_files):
            log.system.error("Input model couldn't be parsed - invalid syntax.\n%s"%self.cinfo.error_collector)
            exit()
        self.debug_info.compile_info = self.cinfo
        slim.check(self.cinfo) # TODO: check for errors
        
        # Convert timescales
        TimeScaleTransformVisitor(self.cinfo).visit(self.cinfo)
        
        # Do fault-injection
        self.ext_cinfo = slim.do_fault_injection(self.cinfo, self.fault_injection_filename)
        self.debug_info.extended_compile_info = self.ext_cinfo
        
        # Extract root implementation
        self.root_impl = self._extract_root_implementation()
        self.debug_info.root_impl = self.root_impl

        # Transform SLIM to SMV
        self.smv_fname, need_msat, _ = slim.convert_to_smv(self.root_impl, get_out_fname(options.BASE_FILE_NAME + ".smv"))
        self.debug_info.slim2smv_output = open(self.smv_fname, "r").read()

        # Transform properties to SMV
        #smv_formulae_doc = slim.get_formulae_doc(self.root_impl, self.properties_filename)
        #properties_api.convert_atomic_propositions_to_smv(smv_formulae_doc, self.root_impl)
        
        # BDD ordering
        if not need_msat:
            self.ord_fname = get_out_fname(options.BASE_FILE_NAME + ".ord")
            nusmv.launch_smv_to_bdd_ordering(smv_fname=self.smv_fname,
                                             ord_fname=self.ord_fname)
        
        # Check if the root has no event or data ports
        self.check_root_closed()
        
        # Check if the root is nonhybrid (no infinite domain)
        self.check_nonhybrid()
        
    def gen_formulae(self):
        # Extract formulae document
        self.formulae_doc = slim.get_formulae_doc(self.root_impl, self.properties_filename)
        self.formulae_doc = self._fix_formulae_doc_wrt_atomic_proposition_templates(self.formulae_doc)
        self.debug_info.formulae_document = self.formulae_doc
        pass
        
    def gen_sigref(self, root_impl=None, formulae_doc=None):
        if root_impl is None:
            root_impl = self.root_impl
        if formulae_doc is None:
            formulae_doc = self.formulae_doc

        # dump the event probability mapping
        dump.dump_error_event_probability_map(root_impl, self.debug_info)
        
        # perform mapping to mrmc compatible labels
        #self.proposition_to_label_dict = properties_api.create_proposition_to_label_dict(formulae_doc)
        dump.dump_ap_map(root_impl, formulae_doc, self.debug_info)

        # run smv2sigref
        self.xml_filepath = get_out_fname(options.BASE_FILE_NAME + ".xml")
        self.debug_info.smv2sigref_output = nusmv.launch_smv_to_sigref(
                self.smv_fname,
                get_out_fname(options.BASE_FILE_NAME + ".epm"),
                ord_fname=self.ord_fname,
                label_fname=get_out_fname(options.BASE_FILE_NAME + ".map"),
                out_fname=self.xml_filepath)

    def gen_property_record(self):
        self.property_checking_record = PropertyCheckingRecord.create("SMV",[self.xml_filepath])
        pass

    def gen_model_files(self, formulae_doc=None):
        if formulae_doc is None:
            formulae_doc = self.formulae_doc

        properties_document = properties_module.parse_properties_document(self.properties_filename)
        
        formulae = []
        for property in properties_document:
            formula = properties_api.create_formula_for_property(property)
            formulae.append(formula)
        self.property_checking_record.label_map = properties_api.create_proposition_to_label_dict(formulae)
        
        for property in properties_document:
            self.property_checking_record.addProperty(property)

        self.property_checking_record.writeToFile()
        # Run model2imc to convert the XML BDD to a CTMC in MRMC and IMCA format.
        formula_numbers = self.property_checking_record.property_map.values()
        if self.model2imc_handler:
            self.model2imc_handler.busy += 1
            self.model2imc_handler.start_model2imc(formula_numbers, self.property_checking_record.basefolder, self.tools)
            self.model2imc_handler.busy -= 1

            if not self.model2imc_handler.tool:
                # Process was most likely killed
                return (99, "")

            # Wait for model2imc to finish.
            exit_code = self.model2imc_handler.tool.returncode
            output = (self.model2imc_handler.tool.get_stdout_text(), self.model2imc_handler.tool.get_stderr_text())
            self.debug_info.model2imc_output = output
        else:
            (exit_code, model2imc_out) = model2imc_process.run(formula_numbers, self.property_checking_record.basefolder, self.tools)
            self.debug_info.model2imc_output = model2imc_out
        
        # Check whether model2imc has terminated successfully.
        if exit_code != 0:
            if exit_code == None: exit_code = 99
            raise exceptions.ProcessInternalException("model2imc", "Terminated with error code %d." % exit_code)
            
        # Update property checking record with results of dft2imc
        self.property_checking_record.mergeWithConversionResults()
        self.property_checking_record.writeToFile()
    
    def analyze(self, property, error_bound=0.1, stepwise=False, granularity=100):
        property_checker = PropertyChecker(self.property_checking_record.record_file_path, self.tools)
        property_checker.set_model2imc_handler(self.model2imc_handler)
        property_checker.set_imca_handler(self.imca_handler)
        property_checker.set_mrmc_handler(self.mrmc_handler)
        property_checker.debug_info = self.debug_info
        force_tool = None if len(self.tools) > 1 else self.tools[0]
        return property_checker.get_results(property, error_bound, force_tool, stepwise, granularity)
    
    def check_root_closed(self):
        type_ = self.root_impl.parent.get_component_type(self.root_impl.name.split(".")[0])
        port_count = type_.get_total_number_of_outgoing_data_ports() + type_.get_total_number_of_incoming_data_ports() + \
                 type_.get_total_number_of_outgoing_event_ports() + type_.get_total_number_of_incoming_event_ports()
        if port_count > 0:
            log.system.info("The root component contains ports, performability assumes a closed world.")
        pass
    
    def check_nonhybrid(self):
        from slim_compiler.src.smv.walkers import InfiniteDomainFinderWalker 
        walker = InfiniteDomainFinderWalker()
        walker.walk(self.root_impl)
        infinite_domain = walker.has_found_infinite_data()
        if infinite_domain:
            raise exceptions.UserException("SLIM model has an infinite domain, performability analysis is not possible.")
    
    def _extract_root_implementation(self):
        cinfo = self.ext_cinfo if self.ext_cinfo else self.cinfo
        root_impl = slim.get_root_impl(cinfo, None)
        self.debug_info.root_impl = root_impl.name
        return root_impl


    def _fix_formulae_doc_wrt_atomic_proposition_templates(self, formulae_doc):
        old_ap_2_derived_aps_map = dict()
        for formula in formulae_doc:
            for atomic_proposition in properties_api.get_atomic_proposition_strings(formula):
                old_ap_2_derived_aps_map[atomic_proposition] = self._get_atomic_proposition_with_templates(atomic_proposition)
        
        derived_formulae = self._derive_formulae(formulae_doc, old_ap_2_derived_aps_map)
        return properties_module.formulae_module.FormulaeDocument.create(derived_formulae)
    
    def _derive_formulae(self, formulae_doc, old_ap_2_derived_aps_map):
        res = formulae_doc
        for old_ap in old_ap_2_derived_aps_map.keys():
            new_aps = old_ap_2_derived_aps_map[old_ap]
            res = [self._derive_formula_helper(formula, old_ap, new_aps) for formula in res]
            #flatten
            res = [item for sublist in res for item in sublist]
        return res
    
    def _derive_formula_helper(self, formula, old_ap, new_aps):
        res = []
        for new_ap in new_aps:
            new_formula = Formula.create(formula.get_logic(), formula.content)
            properties_api.replace_atomic_propositions(new_formula, {old_ap:new_ap})
            res.append(new_formula)
        return res
    
    def _get_atomic_proposition_with_templates(self, atomic_proposition):
        """Validates the specified atomic proposition (in SLIM syntax) POSSIBLY
        CONTAINING IMPLEMENTATION NAMES with respect to the SLIM model with the
        specified root component implementation and returns a list of derived
        atomic propositions
        This is done in three steps:
        1.) make sure only one component implementation name appears in the AP
        2.) make sure there is at least one instance of this component implementation
            that is reachable from the root
        3.) instantiate the name with a path to a component and check this using
            get_errors_in_atomic_proposition (for every possible implementation
            instance)
        
        @param   atomic_proposition:  The atomic proposition in SLIM format.
        @type    atomic_proposition:  string
        """
    
        # 1st: generate a list of strings of the form "package::component.impl"
        # where "package::" is omitted if package = "__default__"
        impl_names = []
        
        for package in self.cinfo.packages.values():
            for subpackage in [package.private_subpackage, package.public_subpackage]:
                if subpackage.package_name is "__default__":
                    impl_names.extend(subpackage.implementations.keys())
                else:
                    impl_names.extend([subpackage.package_name + "::" + impl_name
                                       for impl_name
                                       in subpackage.implementations.keys()])
        
        # count how many component implementations the AP refers to
        # and store their names
        found_impl_names = []
        for impl_name in impl_names:
            # We have to search for the implementation name followed by a dot
            # in the atomic proposition string. Otherwise substrings might be
            # found.
            if (-1 != atomic_proposition.find(impl_name + ".")):
                found_impl_names.append(impl_name)
            if len(found_impl_names) > 1:
                break
        
        if len(found_impl_names) > 1:
            raise exceptions.UserException("You may refer to at most one component implementation in an atomic proposition")
        root_implementation = slim.get_root_impl(self.ext_cinfo, None)
        res = []
        if len(found_impl_names) == 1:
            # 2nd: find all instances of the component implementation that is
            # reachable from root. Then check the AP for these instances
            # NOTE: Because of instance based fault injections we have to check
            # ALL reachable instances
            paths = self._get_instances_of(found_impl_names[0])
            if not paths:
                raise exceptions.UserException("There is no instance of the specified component implementation which is reachable from the root component.")
            
            assert len(paths) > 0, paths
            for path in paths:
                instantiated_atomic_proposition = atomic_proposition.replace(found_impl_names[0]+".", path)
                res.append(instantiated_atomic_proposition)
                error_list = properties_api.get_errors_in_atomic_proposition(instantiated_atomic_proposition, root_implementation)
        else: # There are no templates, directly check syntax
            res = [atomic_proposition]
            error_list = properties_api.get_errors_in_atomic_proposition(atomic_proposition, root_implementation)
        
        if len(error_list) > 0:
            raise exceptions.UserException("\n".join(error_list))
        return res
        
    def _get_instances_of(self, impl_name):
        """returns a list of qualified subcomponent names which are instances
        of a given subcomponent implementation
        """
        root = slim.get_root_impl(self.cinfo, None)
        paths = []
        if root:
            # The queue is a list of pairs
            # ("path" , ComponentImplementation object)
            queue = [("",root)]
    
            while queue != []:
                current_impl = queue[0][1]
                if current_impl.name == impl_name:
                    paths.append(queue[0][0])
                #add the subcomponents of the current subcomponent to the queue
                for id in current_impl.non_data_subcomponents.keys():
                    path = queue[0][0] + str(id) + "."
                    subcmp = current_impl.non_data_subcomponents[id].get_referenced_component_implementation()
                    queue.append((path, subcmp))
                queue.pop(0)
        return paths

# TODO: Move into DFTPropertyChecker, DFTEvaluater & SLIMPropertyChecker into an own module
class DFTPropertyChecker(Checker):
    def __init__(self, fault_tree_filename, basic_event_rates_filename, properties_filename, tools=["mrmc","imca"]):
        Checker.__init__(self, tools)
        
        self.properties_filename = properties_filename
        self.fault_tree_filename = fault_tree_filename
        self.basic_event_rates_filename = basic_event_rates_filename
        self.property_checking_record = PropertyCheckingRecord.create("DFT",[self.fault_tree_filename, self.basic_event_rates_filename])

        # Parse the specified properties document and check it for errors
        if not isinstance(self.properties_filename, PropertiesDocument):
            errors = properties_module.get_syntax_errors_in_document(properties_module.parse_properties_document(self.properties_filename))
            if len(errors) > 0:
                # Create an error message by concatenating the descriptions of the
                # individual errors, separated by newline characters
                error_strings = map(str, errors)
                error_message = ("Errors in properties document:\n" + "\n".join(error_strings))
                raise exceptions.UserException(error_message)
    
    def gen_model_files(self):
        if not isinstance(self.properties_filename, PropertiesDocument):
            properties_document = properties_module.parse_properties_document(self.properties_filename)
        else:
            properties_document = self.properties_filename

        for property in properties_document:
            self.property_checking_record.addProperty(property)

        self.property_checking_record.writeToFile()
        # Run model2imc to convert the fault tree to a CTMC in MRMC and IMCA format.
        formula_numbers = self.property_checking_record.property_map.values()
        if self.model2imc_handler:
            self.model2imc_handler.busy += 1
            self.model2imc_handler.start_model2imc(formula_numbers, self.property_checking_record.basefolder, self.tools)
            self.model2imc_handler.busy -= 1

            if not self.model2imc_handler.tool:
                # Process was most likely killed
                (exit_code, _) = (99, "")
            else:
                # Wait for model2imc to finish.
                exit_code = self.model2imc_handler.tool.returncode
                output = (self.model2imc_handler.tool.get_stdout_text(), self.model2imc_handler.tool.get_stderr_text())
                self.debug_info.model2imc_output = output
        else:
            (exit_code, output) = model2imc_process.run(formula_numbers, self.property_checking_record.basefolder, self.tools)
            self.debug_info.model2imc_output = output

        # Check whether model2imc has terminated successfully.
        if exit_code != 0:
            if exit_code == None: exit_code = 99
            raise exceptions.ProcessInternalException("model2imc", "Terminated with error code %d." % exit_code)
        
        # Update property checking record with results of dft2imc
        self.property_checking_record.mergeWithConversionResults()
        self.property_checking_record.writeToFile()

    def analyze(self, property, error_bound, stepwise=False, granularity=100):
        property_checker = PropertyChecker(self.property_checking_record.record_file_path, self.tools)
        force_tool = None if len(self.tools) > 1 else self.tools[0]
        property_checker.set_model2imc_handler(self.model2imc_handler)
        property_checker.set_imca_handler(self.imca_handler)
        property_checker.set_mrmc_handler(self.mrmc_handler)
        property_checker.debug_info = self.debug_info
        return property_checker.get_results(property, error_bound, force_tool, stepwise, granularity)

# TODO: Move into DFTPropertyChecker, DFTEvaluater & SLIMPropertyChecker into an own module 
class DFTEvaluater(Checker):
    def __init__(self, fault_tree_filename, basic_event_rates_filename, events_output_filename="events.txt", gates_output_filename="gates.txt", tools=["mrmc", "imca"]):
        Checker.__init__(self, tools)
        self.fault_tree_filename = fault_tree_filename
        self.basic_event_rates_filename = basic_event_rates_filename
        self.gates_output_filename = gates_output_filename
        self.events_output_filename = events_output_filename

    def analyze(self, mission_time, severity, error_bound):
        (fault_tree, event_name_to_probability_dict, error_msg) = self.evaluate_fault_tree(mission_time, error_bound)

        # If a severity number is specified, create a dictionary that maps the
        # name of the top-level event of the specified fault tree to the
        # criticality of that event. The criticality is the product of the
        # probability of the event and its severity.
        event_names_to_criticality_dict = None
        if severity is not None:
            severity = float(severity)
            top_name = fault_tree_module.get_top_level_event_name(fault_tree)
            top_probability = event_name_to_probability_dict[top_name]
            # TODO: This is currently broken, because top_probability is in fact an instance of Result, not a probability
            event_names_to_criticality_dict = {top_name: top_probability * severity}

        events_fname = misc.get_out_fname(self.events_output_filename)
        gates_fname = misc.get_out_fname(self.gates_output_filename)

        fault_tree_module.write_fault_tree_plus_output(fault_tree,
                                     event_name_to_probability_dict,
                                     event_names_to_criticality_dict,
                                     events_fname,
                                     gates_fname)

        relpaths = tuple(map(misc.relpath, [events_fname, gates_fname]))
        return (relpaths, error_msg)

    def evaluate_fault_tree(self, mission_time, error_bound):
        # Parse the fault tree and compute for each event of its events the
        # probability of occurrence in the specified mission time.
        # This is done by evaluating for each event the subtree below that event.
        # The probabilities will be stored in a dictionary, where the event names
        # are the keys.
        event_name_to_probability_dict = {}
        error_msg = None
        fault_tree = fault_tree_module.parse_xml(self.fault_tree_filename)
        event_names = []
        properties = []
        for sub_fault_tree in fault_tree_module.sub_fault_tree_generator(fault_tree):
            # Get the name of the top-level event of the subtree. This is the
            # event whose probability is to be computed.
            event_name = fault_tree_module.get_top_level_event_name(sub_fault_tree)
            event_names.append(event_name)
            properties.append(fault_tree_module.create_reachability_property(event_name, mission_time, "prop"))
            
        # Compute the probability for the events
        properties_document = properties_module.PropertiesDocument.create(properties)
        dft_checker = DFTPropertyChecker(self.fault_tree_filename, self.basic_event_rates_filename, properties_document, self.tools)
        dft_checker.set_model2imc_handler(self.model2imc_handler)
        dft_checker.set_imca_handler(self.imca_handler)
        dft_checker.set_mrmc_handler(self.mrmc_handler)
        dft_checker.debug_info = self.debug_info

        dft_checker.gen_model_files()
        
        for index in range(len(properties)):
            (result, last_err) = dft_checker.analyze(properties[index], error_bound) # always uses result of first tool
            event_name_to_probability_dict[event_names[index]] = result.probability
            if last_err:
                error_msg = last_err

        return (fault_tree, event_name_to_probability_dict, last_err)
