# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================
import types
from lxml import etree
from slim_compiler.src.dataStructure.expressions import ExpressionTreeNode
from slim_compiler.src.dataStructure.located_info import LocatedInformation

class Py2XML():
    def __init__(self):
        self.known_object_ids = set()
        self.used_obj_ids = set()

    def get_ids(self, obj):
        """Collects the object IDs that occur
        throughout the data structures
        """
        if obj is None:
            return
        if self.isIntrinsic(obj):
            return

        obj_id = id(obj)

        if obj_id in self.known_object_ids:
            self.used_obj_ids.add(obj_id)
        else:
            self.known_object_ids.add(obj_id)
            #Recurse

            # Custom class
            if hasattr(obj, "__dict__"):
                for var in obj.__dict__:
                    if not self.filter(obj, var):
                        continue
                    attr = getattr(obj, var)
                    self.get_ids(attr)
            # List or Tuple
            elif isinstance(obj, types.ListType) or isinstance(obj, types.TupleType):
                for l in obj:
                    self.get_ids(l)
            # Dict
            elif isinstance(obj, types.DictType):
                for (key,value) in obj.items():
                    self.get_ids(value)
            else:
                print obj
                raise NotImplementedError

    def parse_obj(self, obj, tag=None, key=None, expr=False):
        """Build XML structure based on given object. If tag is None,
        XML tag wil lbe the name of the type, otherwise the given value.
        If key is not None, the key attribute will be set.
        Expr is true if currently traversing an expression tree or at least
        one parent is an expression.
        """
        if obj is None:
            return None

        if tag is None:
            # If tag is empty, create element based on type
            root = etree.Element(type(obj).__name__)
        else:
            root = etree.Element(tag)

        root.set("type", type(obj).__name__)

        # Set key attribute if object is key in dictionary
        if key:
            root.set("key", key)

        # Dump string representation of intrinsic types
        if self.isIntrinsic(obj):
            root.text = str(obj)
            return root

        # True if expression node
        exprNode = isinstance(obj, ExpressionTreeNode)

        obj_id = id(obj)
        if obj_id in self.known_object_ids and (obj_id in self.used_obj_ids
                                                or (expr and not exprNode)):
            # Add a reference if object is of reference-type and already in XML
            # Also is the object occurs as a child in an expression
            # (makes it more readable)
            root.set("objref", str(obj_id))
            return root
        elif obj_id in self.known_object_ids:
            # Add ID if object is known to be reused
            root.set("objid", str(obj_id))
            self.used_obj_ids.add(obj_id)

        # Custom class
        if hasattr(obj, "__dict__"):
            for var in obj.__dict__:
                if not self.filter(obj, var):
                    continue
                attr = getattr(obj, var)
                child = self.parse_obj(attr, var, None, expr or exprNode)
                if child is not None:
                    root.append(child)
        # List or Tuple
        elif isinstance(obj, types.ListType) or isinstance(obj, types.TupleType):
            for l in obj:
                child = self.parse_obj(l, None, None, expr or exprNode)
                if child is not None:
                    root.append(child)
        # Dict
        elif isinstance(obj, types.DictType):
            for (key,value) in obj.items():
                child = self.parse_obj(value, None, key, expr or exprNode)
                if child is not None:
                    root.append(child)
        else:
            raise NotImplementedError

        return root

    def parse(self, cinfo):
        # re-initialize before parsing a new CompileInfo structure
        self.__init__()

        # Get list of IDs occuring in the cinfo structure and the one that are reused
        self.get_ids(cinfo)
        # Initialize list of reused object ids so the parser can use it
        self.known_object_ids = self.used_obj_ids
        self.used_obj_ids = set()
        # Parse cinfo into lxml etree structure
        root = self.parse_obj(cinfo)
        assert(root is not None)

        return root

    def convert(self, cinfo):
        """Dump the CompileInformation as XML, returns a string"""
        assert type(cinfo).__name__ == "CompileInformation", "XML converter requires CompileInformation as argument"

        root = self.parse(cinfo)

        # Dump to string
        return etree.tostring(root, pretty_print=True, xml_declaration=True)

    def isIntrinsic(self, obj):
        """Returns true if the obj is an intrinsic type for the purpose of the XML dumper"""
        return type(obj) in [types.NoneType,
                             types.TypeType,
                             types.BooleanType,
                             types.IntType,
                             types.LongType,
                             types.FloatType,
                             types.ComplexType,
                             types.StringType,
                             types.UnicodeType]

    def filter(self, obj, attr):
        _any_filter = ['subpackage']
        _loc_filter = ['file', 'line', 'pos', 'parent']
        _expr_filter = ['enum_type', 'enum_values', 'type', 'checked']

        if attr in _any_filter:
            return False
        
        if isinstance(obj, LocatedInformation):
            if attr in _loc_filter:
                return False

        if isinstance(obj, ExpressionTreeNode):
            if attr in _expr_filter:
                return False
        return True
