# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================

import sys
import re
import random
import itertools

############################# CLASS TfpgSynthesizer ############################


class TfpgSynthesizer (object):

    ###################################################### CONSTRUCTOR & HELPERS

    def __init__ (self, system_modes, failure_modes, discrepancies, minimal_cutsets):

        # input parameters
        assert len(set(failure_modes).intersection(set(discrepancies))) is 0
        assert len(failure_modes) > 0
        assert len(discrepancies) > 0
        self.system_modes = system_modes
        self.failure_modes = failure_modes
        self.discrepancies = discrepancies
        self.minimal_cutsets = minimal_cutsets
        self.generated_node_contexts = {}

        self.discrepancy_names = set([d[0] for d in self.discrepancies])
        self.monitored_discrepancies = set([d[0] for d in self.discrepancies if d[1] == "monitored"])

        # synthesis helpers
        self.gen_id = 0
        self.nodes_being_processed = set()

        # synthesis results
        self.causality_graph = {}

    def is_generated_node_name (self, name):
        if isinstance(name,int) or re.search(r'^\d*$', name):
            return True
        else:
            return False

    ########################################################### INPUT VALIDATION

    def validate_node_names (self):
        for fm in self.failure_modes:
            if self.is_generated_node_name(fm):
                return "illegal failure mode name (%s)" % fm
        for d in self.discrepancy_names:
            if self.is_generated_node_name(d):
                return "illegal discrepancy name (%s)" % d
        return "ok"

    def validate_cutsets (self):
        assert self.minimal_cutsets is not None

        res = self.exists_unreachable_discrepancy(self.minimal_cutsets)
        if res[0]:
            return "unreachable discrepancy within analysis bound (%s)" % res[1]

        res = self.exists_spontaneous_discrepancy_activation(self.failure_modes, self.minimal_cutsets)
        if res[0]:
            return "spontaneous discrepancy activations (%s)" % res[1]

        res = self.exist_correlated_discrepancy_activations(self.minimal_cutsets)
        if res[0]:
            return "correlated discrepancy activations (%s, %s)" % (res[1], res[2])

        return "ok"

    def exists_unreachable_discrepancy(self, minimal_cutsets):
        for d in minimal_cutsets:
            if minimal_cutsets[d] is None:
                return (True, d)
        return (False, None)

    def exist_correlated_discrepancy_activations (self, minimal_cutsets):

        # check whether some discrepancy in some failure configurations it always
        # fires simultaneously with some other discrepancy
        #
        # for this we need to see whether there exist two discrepancies d and d'
        # which have a shared minimal cutset, disregarding d and d' themselves as
        # cutset elements.

        for d in minimal_cutsets:
            # skip failure modes
            if len(minimal_cutsets[d]) is 0: continue

            for d_prime in minimal_cutsets:
                # skip d
                if d_prime is d: continue
                # skip failure modes
                if len(minimal_cutsets[d_prime]) is 0: continue

                for mcs in minimal_cutsets[d]:
                    # skip non-matching cutsets
                    if d_prime not in mcs: continue

                    # first cutset without the second discrepancy
                    tmp_mcs = [x for x in mcs if x != d_prime]

                    for mcs_prime in minimal_cutsets[d_prime]:
                        # skip non-matching cutsets
                        if d not in mcs_prime: continue

                        # second cutset without the first discrepancy
                        tmp_mcs_prime = [x for x in mcs_prime if x != d]

                        if set(tmp_mcs) == set(tmp_mcs_prime): return (True, d, d_prime)

        return (False, None, None)

    def exists_spontaneous_discrepancy_activation (self, failuremodes, minimal_cutsets):

        for d in minimal_cutsets:
            if d not in failuremodes and len(minimal_cutsets[d]) is 0:
                # d is not a failure mode, but can happen spontaneously,
                # i.e. it is caused by the empty cutset
                return (True, d)

        # only failure modes are caused by empty cutsets
        return (False, None)

    ################################################## CAUSALITY GRAPH SYNTHESIS

    def synthesize_causality_graph (self, discrepancies):
        # assert preconditions
        assert self.gen_id == 0
        assert len(self.nodes_being_processed) == 0
        assert self.minimal_cutsets is not None

        # call synthesis routine
        self.synthesize_causality_graph_rec(discrepancies)

        # assert postconditions
        assert self.gen_id >= 0
        assert len(self.nodes_being_processed) == 0
        assert self.discrepancy_names <= set(self.causality_graph.keys())
        assert self.minimal_cutsets is not None

    def synthesize_causality_graph_rec (self, discrepancies):

        while True:

            # ignore nodes that already have been processed
            unresolved_nodes = set(discrepancies) - set(self.causality_graph.keys())

            # we furthermore ignore nodes being processed at the moment,
            # to avoid infinite recursion
            unresolved_nodes -= self.nodes_being_processed

            # stop if all nodes have been processed
            if len(unresolved_nodes) is 0: break

            # pick randomly a new node that still needs to be processed
            n = random.sample(unresolved_nodes, 1)[0]

            # mark it as being processed, in order to prevent infinite recursions
            assert n not in self.nodes_being_processed
            self.nodes_being_processed.add(n)

            # retrieve all minimal cutsets of the selected node (list of lists)
            assert n in self.minimal_cutsets

            ls_mcs = self.minimal_cutsets[n]

            # n is a basic fault event (no minimal cutsets, i.e. only the empty one)
            if len(ls_mcs) is 0:
                self.causality_graph[n] = {'type':'FM', 'children':None}

            # n is a consequence fault event (there are some non-empty minimal cutsets)
            else:

                children_of_n = []

                for mcs in ls_mcs:

                    # mcs is a basic fault event
                    if len(mcs) == 1:
                        if mcs[0] not in self.causality_graph.keys():
                            self.causality_graph[mcs[0]] = {'type':'FM', 'children':None}

                    # mcs consists of more than one event
                    else:
                        assert len(mcs) > 1
                        self.synthesize_causality_graph_rec(mcs)

                    # generate id for new node
                    tmp_id = self.gen_id
                    self.gen_id += 1

                    # generate and node representing the minimal cutset
                    self.causality_graph[tmp_id] = {'type':'AND', 'children':mcs}
                    self.generated_node_contexts[tmp_id] = mcs

                    # register new node as a child of n
                    children_of_n.append(tmp_id)

                # create fault tree for current node
                self.causality_graph[n] = {'type':'OR', 'children':children_of_n}

            self.nodes_being_processed.remove(n)

    ##################################################### CUTSETS SIMPLIFICATION

    def appears_in_all_lists (self, elem, lists):
        if len(lists) is 0: return False
        for l in lists:
            if elem not in l: return False
        return True

    def compute_minimal_hitting_sets (self, list_of_lists_of_lists):

        # input:
        #   [[[scs1], [scs2], [ncs1], [ncs2]],
        #    [[scs1], [scs2], [ncs3], [ncs4]],
        #    [[scs1], [scs2], [ncs5]]]

        result = []
        stripped_of_common_cutsets = []

        for outer_list in list_of_lists_of_lists:
            stripped_of_common_cutsets.append([])
            for mcs in outer_list:
                if mcs in result:
                    assert self.appears_in_all_lists(mcs, list_of_lists_of_lists)
                    continue
                elif self.appears_in_all_lists(mcs, list_of_lists_of_lists):
                    result.append(mcs)
                else:
                    stripped_of_common_cutsets[-1].append(mcs)
            if len(stripped_of_common_cutsets[-1]) is 0:
                # this means that one of the shared nodes has as minimal cutsets
                # just the ones that are common to all shared nodes. this means
                # that there exists another node shared by all tle cutsets that
                # can be caused by a cutset which is minimal for that node but which
                # is not part of any minimal cutset for the tle. therefore only
                # the shared minimal cutsets need to be considered.
                return result

        # result := [scs1, scs2]

        stripped_of_common_cutsets = [x for x in stripped_of_common_cutsets if len(x) > 0]

        print "result:",result
        print "stripped_of_common_cutsets:",stripped_of_common_cutsets

        # stripped_of_common_cutsets := [[[ncs1], [ncs2]],
        #                                [[ncs3], [ncs4]],
        #                                [[ncs5]]]

        products = itertools.product(*stripped_of_common_cutsets)

        # products := [[[ncs1], [ncs3], [ncs5]],
        #              [[ncs2], [ncs3], [ncs5]],
        #              [[ncs1], [ncs4], [ncs5]],
        #              [[ncs2], [ncs4], [ncs5]]]

        print "products:"
        for res_cs in products:
            print "res_cs:",res_cs
            tmp = set()
            for x in res_cs:
                print " x:",x
                tmp.add(*x)
            result.append(list(tmp))

        print "result:",result

        # result := [scs1, scs2, ncs1ncs3ncs5, ncs1ncs4ncs5, ncs2ncs3ncs5, ncs2ncs4ncs5]

        return result

    def simplify_cutsetbased (self):

        for node in self.minimal_cutsets:
            if self.minimal_cutsets[node] is not None and len(self.minimal_cutsets[node]) > 1:
                set_intersect = set.intersection(*map(set, self.minimal_cutsets[node]))
                if len(set_intersect) > 0:

                    print "testing simplification based on shared implicants of '%s'" % node
                   
                    print "  > compute minimal hitting sets of list of minimal cutset lists of all shared nodes"

                    list_of_lists_of_minimal_cutsets = [self.minimal_cutsets[n] for n in set_intersect]
                    minimal_hitting_sets = self.compute_minimal_hitting_sets(list_of_lists_of_minimal_cutsets)

                    for l in minimal_hitting_sets: l.sort()
                    minimal_hitting_sets.sort()

                    print "  > subtract shared nodes from minimal cutsets of '%s'" % node

                    minimal_cutsets_minus_shared_nodes = [list(set(cs)-set_intersect) for cs in self.minimal_cutsets[node]]

                    for l in minimal_cutsets_minus_shared_nodes: l.sort()
                    minimal_cutsets_minus_shared_nodes.sort()

                    #print "minimal_cutsets_minus_shared_nodes:",minimal_cutsets_minus_shared_nodes
                    #print "list_of_lists_of_minimal_cutsets:",list_of_lists_of_minimal_cutsets
                    #print "minimal_hitting_sets:",minimal_hitting_sets

                    print "  > compare the two results"

                    can_do_simplified_graph_insertion = minimal_hitting_sets == minimal_cutsets_minus_shared_nodes

                    print "  > simplification possible:", ("yes" if can_do_simplified_graph_insertion else "no")

                    if can_do_simplified_graph_insertion:
                        # nb: In case an entry for 'node' already exists, it is OK to be overwritten/redeclared.
                        print "  > instantiating '%s' as AND node, with incoming edges from:"
                        print "    %s" % ", ".join(set_intersect)
                        self.causality_graph[node] = {'type':'AND', 'children':list(set_intersect)}

    ############################################# CAUSALITY GRAPH SIMPLIFICATION

    def simplify_causality_graph (self, causality_graph):

        num_nodes = len(causality_graph)
        num_children = sum([len(causality_graph[k]['children']) for k in causality_graph if causality_graph[k]['children'] is not None])

        while True:
            #print "  transitive links ..."
            self.simplify_causality_graph_transitive_links(causality_graph)
            #print "  single generated nodes ..."
            self.simplify_causality_graph_single_generated_nodes(causality_graph)
            #print "  or to and ..."
            self.simplify_causality_graph_or_to_and(causality_graph)

            num_nodes_new = len(causality_graph)
            num_children_new = sum([len(causality_graph[k]['children']) for k in causality_graph if causality_graph[k]['children'] is not None])

            if num_nodes_new == num_nodes and num_children_new == num_children:
                print "simplification methods converged"
                break

            num_nodes = num_nodes_new
            num_children = num_children_new

        while True:
            #print "  transitive links bis"
            self.simplify_causality_graph_transitive_links_bis(causality_graph)

            num_nodes_new = len(causality_graph)
            num_children_new = sum([len(causality_graph[k]['children']) for k in causality_graph if causality_graph[k]['children'] is not None])

            if num_nodes_new == num_nodes and num_children_new == num_children:
                print "simplification bis methods converged"
                break

            num_nodes = num_nodes_new
            num_children = num_children_new

        self.simplify_causality_graph_transitive_links_w_context(causality_graph)
        self.simplify_causality_graph_single_generated_nodes(causality_graph)
        self.simplify_causality_graph_duplicate_edges(causality_graph)
        self.simplify_causality_graph_and_to_or(causality_graph)

    def simplify_causality_graph_transitive_links (self, causality_graph):
        # remove transitive links where possible (works through side-effect
        # on causality_graph)

        for d in causality_graph:

            if causality_graph[d]['type'] != 'AND': continue
            assert len(causality_graph[d]['children']) > 0

            for child_one in causality_graph[d]['children']:
                for child_two in causality_graph[d]['children']:
                    if child_one is child_two: continue

                    if causality_graph[child_two]['children'] is None: continue
                    if (causality_graph[child_two]['type'] == 'AND' or \
                        len(causality_graph[child_two]['children']) == 1) and \
                        child_one in causality_graph[child_two]['children']:

                        # child_one is guaranteed to occur by child_two; it's
                        # safe to remove edge from child_one to d

                        print "SIMP TRANS1 > remove edge '%s -> %s'" % (child_one, d)

                        causality_graph[d]['children'] = [x for x in causality_graph[d]['children'] if x != child_one]

                        return

    def simplify_causality_graph_transitive_links_bis (self, causality_graph):
        # remove transitive links where possible (works through side-effect
        # on causality_graph)

        for d in causality_graph:

            if causality_graph[d]['type'] != 'AND': continue
            assert len(causality_graph[d]['children']) > 0

            for child_one in causality_graph[d]['children']:
                for child_two in causality_graph[d]['children']:
                    if child_one is child_two: continue

                    #print "> '%s' is in all mcs of '%s'?" % (child_one, child_two)
                    #print "> mcs of '%s':" % child_two, self.minimal_cutsets[child_two]

                    if self.appears_in_all_lists(child_one, self.minimal_cutsets[child_two]):

                        print "SIMP TRANS2 > remove edge '%s -> %s'" % (child_one, d)

                        causality_graph[d]['children'] = [x for x in causality_graph[d]['children'] if x != child_one]
                        return

    def get_active_cutsets (self, discrepancy, context):
        # discrepancy: the node whose cutsets we want to consider
        # context: the list of active nodes
        
        active_cutsets = []

        for mcs in self.minimal_cutsets[discrepancy]:

            mcs_is_active = True
            for d in mcs:
                if d not in context:
                    mcs_is_active = False
                    break

            if mcs_is_active:
                active_cutsets.append(mcs)

        return active_cutsets

    def simplify_causality_graph_transitive_links_w_context (self, causality_graph):
        # remove transitive edges implied in a local context (cutset)

        # for each generated cutset node c
        for c in causality_graph:
            if not self.is_generated_node_name(c): continue

            # for each incoming edge (d,c)
            for d in causality_graph[c]['children']:

                # for any other incoming edge (d',c), with d' != d
                for d_prime in causality_graph[c]['children']:
                    if d is d_prime: continue

                    # if d is in all active cutsets of d' (i.e. the occurrence of d'
                    # implies the occurrence of d in the context of c)

                    context = self.generated_node_contexts[c]
                    active_cutsets = self.get_active_cutsets(d_prime, context)
                    is_present_in_all_active_cutsets = len(active_cutsets) > 0

                    for cs in active_cutsets:
                        if d not in cs:
                            is_present_in_all_active_cutsets = False
                            break

                    # then drop edge (d,c) from graph
                    if is_present_in_all_active_cutsets:

                        print "SIMP TRANS3 > remove edge '%s -> %s'" % (d, c)

                        #print "active cutsets of '%s':" % d_prime, active_cutsets
                        #print "context:",context

                        causality_graph[c]['children'] = [n for n in causality_graph[c]['children'] if n != d]

    def simplify_causality_graph_single_generated_nodes (self, causality_graph):
        # remove generated cutset AND nodes with a single input (works through
        # side-effect on causality_graph)

        keys = causality_graph.keys()
        while len(keys) > 0:
            d = keys[0]

            if self.is_generated_node_name(d):
                assert self.causality_graph[d]['type'] == 'AND'
                assert len(causality_graph[d]['children']) > 0
                if len(causality_graph[d]['children']) == 1:

                    # 'd' can be removed from the graph
                    print "SIMP SINGLE > remove node '%s'" % (d)

                    # identify single child node
                    single_child = causality_graph[d]['children'][0]

                    # for all other nodes that have the generated node as a child
                    for d_prime in causality_graph:
                        if d_prime is d: continue
                        if causality_graph[d_prime]['children'] is None: continue

                        # replace the generated node with the single child node
                        # in their cutsets
                        if d in causality_graph[d_prime]['children']:
                            causality_graph[d_prime]['children'].remove(d)
                            causality_graph[d_prime]['children'].append(single_child)

                    del causality_graph[d]

            keys = keys[1:]
        pass

    def simplify_causality_graph_or_to_and (self, causality_graph):
        # simplify an OR node with a single cutset by changing it to an AND node,
        # linking the incoming edges of the generated cutset AND node to the new
        # AND node, and deleting the generated cutset AND node. (works through
        # side-effect on causality_graph)

        # for each discrepancy:
        keys = causality_graph.keys()
        while len(keys) > 0:
            d = keys[0]

            # if not generated and has only one child which is a generated node:
            if d in self.discrepancy_names and \
               causality_graph[d]['children'] is not None and \
               len(causality_graph[d]['children']) is 1 and \
               self.is_generated_node_name(causality_graph[d]['children'][0]):

                print "SIMP OR2AND > redeclare '%s' as AND node" % (d)

                # get generated node id
                child = causality_graph[d]['children'][0]

                # duplicate 'child' with name of 'd'
                causality_graph[d] = causality_graph[child]
                assert causality_graph[d]['type'] == 'AND'

                # remove generated node
                del causality_graph[child]

            keys = keys[1:]
        pass

    def simplify_causality_graph_and_to_or (self, causality_graph):
        # an AND discrepancy that has a single input is redeclared as an OR
        for d in causality_graph:
            if causality_graph[d]['type'] == 'AND' and len(causality_graph[d]['children']) == 1:
                print "SIMP AND2OR > redeclare '%s' as OR node" % (d)
                causality_graph[d]['type'] = 'OR'

    def simplify_causality_graph_duplicate_edges (self, causality_graph):
        for d in causality_graph:
            if causality_graph[d]['children'] is not None and len(causality_graph[d]['children']) > 0:
                causality_graph[d]['children'] = list(set(causality_graph[d]['children']))

    ##################################################################### OUTPUT


    def print_human_readable_causality_graph (self, f, tfpg_name): 
    
        # tfpg_name: it is the arbitrary name of the tfpg: str
        
        assert self.causality_graph is not None

        # TFPG name

        if tfpg_name is None or tfpg_name == "":
            f.write("NAME DependencyGraph\n")
        else: f.write("NAME " + tfpg_name + "\n")

        # nodes

        for d in self.causality_graph:
            d_type = self.causality_graph[d]['type']
            f.write("%s %s\n" % (d_type, d))

        # modes

        all_modes = ','.join(self.system_modes)
        f.write("MODES %s\n" % all_modes)

        # edges

        edge_counter = 1

        for target_node in self.causality_graph:
            if self.causality_graph[target_node]['children'] is None: continue

            for source_node in self.causality_graph[target_node]['children']:
                e_name = "EDGE%d" % edge_counter
                edge_counter += 1
                f.write("EDGE %s %s %s 0.0 -1.0 (%s)\n" % (e_name, source_node, target_node, all_modes))



    def print_xml_causality_graph (self, f, tfpg_name):
    
        # tfpg_name: it is the arbitrary name of the tfpg: str
        
        assert self.causality_graph is not None

        
        f.write("<?xml version='1.0' encoding='UTF-8'?>\n")
        
        if tfpg_name is None or tfpg_name == "":
            f.write('<tfpg name="DependencyGraph">\n')
        else: f.write('<tfpg name="' + tfpg_name + '">\n')
        
        print '<tfpg name="' + tfpg_name + '">\n'
        
        f.write("  <nodesList>\n")

        for d in self.causality_graph:
            f.write('    <node name="%s" isMonitored="%s">\n' % (d, "true" if d in self.monitored_discrepancies else "false"))
            f.write("      <type>%s</type>\n" % self.causality_graph[d]['type'])
            f.write("    </node>\n")

        f.write("  </nodesList>\n")
        f.write("  <modesList>\n")
        for mode in self.system_modes:
            f.write("    <mode>%s</mode>\n" % mode)
        f.write("  </modesList>\n")

        f.write("  <edgesList>\n")

        edge_counter = 0
        for target_node in self.causality_graph:
            if self.causality_graph[target_node]['children'] is None: continue

            for source_node in self.causality_graph[target_node]['children']:
                edge_counter += 1

                f.write('    <edge name="EDGE%d">\n' % edge_counter)
                f.write("      <srcNode>%s</srcNode>\n" % source_node)
                f.write("      <tMin>0.0</tMin>\n")
                f.write("      <tMax>-1.0</tMax>\n")
                f.write("      <modesList>\n")
                for mode in self.system_modes:
                    f.write("    <mode>%s</mode>\n" % mode)
                f.write("      </modesList>\n")
                f.write("      <destNode>%s</destNode>\n" % target_node)
                f.write("    </edge>\n")

        f.write("  </edgesList>\n")
        f.write("</tfpg>\n")


################################# CLASS CutSet #################################


class CutSet (object):

    def __init__ (self, top_level_event, cutset_events):
        assert isinstance(cutset_events, list)
        self.top_level_event = top_level_event
        self.cutset_events = [x for x in cutset_events if x != top_level_event]


    def get_tle (self):
        return self.top_level_event

    def get_events (self):
        return self.cutset_events

    @staticmethod
    def parse_cutset_list (filename):
        """ Static method for parsing list of cutsets from an SA log file. """

        f = open(filename,'r')

        cutset_list = []
        current_cutset = None

        top_level_event = None

        for line in f:

            match = re.search(r'-- minimal cut sets for specification (\S+)', line)
            if match:
                assert top_level_event is None
                top_level_event = match.group(1)
                continue

            # skip everything until the minimal cut sets
            if top_level_event is None:
                continue

            match = re.search(r'\[no cut set found\]', line)
            if match:
                assert top_level_event is not None
                cutset_list = None
                continue

            match = re.search(r'\d+\) -------------------------', line)
            if match:
                assert top_level_event is not None
                #create obj for prev cutset
                if current_cutset is not None:
                    cutset_list.append(CutSet(top_level_event, current_cutset))
                # set-up new cutset
                current_cutset = []
                continue

            match = re.search(r'^\s*(_hist_var_\d+_)?(\S+)_TRUE = TRUE', line)
            if match:
                assert top_level_event is not None
                # add to current cutset
                current_cutset.append(match.group(2))
                continue

            match = re.search(r'\[cut set with no variables\]', line)
            if match:
                assert top_level_event is not None
                for cs in cutset_list:
                    assert cs.top_level_event != top_level_event
                cutset_list.append(CutSet(top_level_event, []))
                current_cutset = []
                continue

            pass

        f.close()

        if cutset_list is None: return None

        #create obj for last cutset
        if current_cutset is not None and len(current_cutset) > 0:
            cutset_list.append(CutSet(top_level_event, current_cutset))

        return cutset_list










########################################################################### MAIN

"""
create main function that takes as input a file with all the minimal cutsets,
the discrepancies, and the failuremodes, and creates an output file with the
resulting TFPG.
"""

# ok
tfpg_01 = ('01',
           {'F1':[],
            'F2':[],
            'A':[['F1']],
            'B':[['F1','A','F2']]})

# ok
tfpg_02 = ('02',
           {'F1':[],
            'F2':[],
            'A':[['F1'],['F2']]})

# ok
tfpg_03 = ('03',
           {'F1':[],
            'F2':[],
            'A':[['F1','F2']]})

# ok
tfpg_04 = ('04',
           {'F1':[],
            'A':[['F1']],
            'B':[['A','F1']]})

# ??
tfpg_05 = ('05',
           {'F1':[],
            'F2':[],
            'A':[['F1'],['F2','B','C']],
            'B':[['F2'],['F1','A']],
            'C':[['F2','B'],['F1','A','B']]})

# ??
tfpg_06 = ('06',
           {'F1':[],
            'F2':[],
            'A':[['F1'],['F2','B']],
            'B':[['F2'],['F1','A']]})

# circular dependency (output warning)
tfpg_07 = ('07',
           {'F1':[],
            'alpha':[['F1']],
            'beta':[['F1']],
            'delta':[['F1']],
            'A':[['alpha','B'],['delta','B']],
            'B':[['beta'],['alpha','A'],['delta','A']]})


ls_all_tfpg = [tfpg_01, tfpg_02, tfpg_03, tfpg_04, tfpg_05, tfpg_06]


if __name__ == "__main__":

    my_tfpg = tfpg_05

    system_modes = ['std']
    failure_modes = [x for x in my_tfpg[1] if x[0] == 'F']
    discrepancies = [x for x in my_tfpg[1] if x[0] != 'F']
    minimal_cutsets = my_tfpg[1]

    tfpg_synthesizer = TfpgSynthesizer(system_modes, failure_modes, discrepancies, minimal_cutsets)

    print "validating node names and cutsets ...",
    res_nodenames = tfpg_synthesizer.validate_node_names()
    res_cutsets = tfpg_synthesizer.validate_cutsets()

    if res_nodenames == "ok" and res_cutsets == "ok":
        print "calling synthesis ...",
        tfpg_synthesizer.synthesize_causality_graph(discrepancies)
        print "done"

        print "calling simplification ...",
        tfpg_synthesizer.simplify_causality_graph(tfpg_synthesizer.causality_graph)
        print "done"

        print
        tfpg_synthesizer.print_human_readable_causality_graph(sys.stdout, "")

    else:
        print
        print "res_nodenames:", res_nodenames
        print "res_cutsets:", res_cutsets






############################################################### TFPG SYNTHESIZER




###################################################################### OLD STUFF




"""
global tfpg
tfpg = tfpg_05[1]

compute_causality_graph(tfpg.keys())

for ft in causality_graph.keys():
    print ft,":",causality_graph[ft]['type'],",",list(causality_graph[ft]['children'])

# validation
tfpg_keys = {}

for my_tfpg in ls_all_tfpg:
    for _ in range(10):
        #global n_id
        n_id = 0

        #global causality_graph
        causality_graph = {}

        #global tfpg
        tfpg = my_tfpg[1]

        name = my_tfpg[0]
        compute_causality_graph(tfpg.keys())

        cur_keys = set(causality_graph.keys())

        if name not in tfpg_keys:
            tfpg_keys[name] = cur_keys
        else:
            if tfpg_keys[name] != cur_keys:
                print name
                print "tfpg_keys[name] =",tfpg_keys[name]
                print "causality_graph.keys() =",causality_graph.keys()
            assert tfpg_keys[name] == cur_keys
"""

